package com.sas;

/**
 * DeepClonable provides an interface for objects
 * which can be cloned shallowly by default but which can also be cloned "deeply".
 * <p>
 * For example, most collections in the <code>com.sas.collection</code>
 * package perform shallow cloning: when you clone a collection, it simply
 * returns a new collection which contains the same objects.
 * A deep clone of an object will also clone such contained objects.
 * <p>
 * DeepClonable also supports cloning which preserves the initial object structure
 * of the source object. This means that a {@link com.sas.DeepCloneSupport} can be used
 * to keep track of clones as the deep clone recurses. This prevents one object
 * from being cloned more that one time if the same object is referenced more than
 * once during the recursion.
 *
 * <h2>Usage</h2>
 *
 * The proper way to clone one or more objects is to request
 * a DeepCloneSupport object to clone them:
 * <pre>
 *  DeepCloneSupport support = new DeepCloneSupport();
 *  cloneA = (TypeA) support.clone(a, true);
 *  cloneB = (TypeB) support.clone(b, true);
 *  cloneC = (TypeC) support.clone(c, true);
 * </pre>
 * This ensures all objects are registered when cloned and no
 * object gets clone more than once.
 *
 * <h2>Example</h2>
 *
 * For example, consider a class which implements a linked list interface:
 * <pre>
 * public class ListElement
 *        implements com.sas.PublicClonable,
 *                   com.sas.DeepClonable
 * {
 *    ListElement next;
 *    public Object clone() throws CloneNotSupportedException
 *    {
 *      ListElement clone = (ListElement) super.clone();
 *      next = null;
 *      return clone;
 *    }
 * }
 * </pre>
 * The linked list may be circular: the last element in the list points back to
 * the first element in the list via it's <var>next</var> field.
 * A deep clone of a ListElement would clone the node referenced by the
 * <var>next</var> field:
 * <pre>
 * // This method is called from the DeepCloneSupport's
 * {@link com.sas.DeepCloneSupport#clone(Object, boolean) clone(Object, boolean)} method.
 * public Object clone(DeepCloneSupport support, boolean deeply) throws CloneNotSupportedException
 *   {
 *      Object clone = clone();
 *      if (support != null)
 *         support.register(this, clone); // do this before recursion
 *      if (deeply && next != null)
 *         // caller wants to create a deep clone of the list
 *         if (support != null)
 *            // Get the clone of the next item. It may have been
 *            // cloned already.
 *            clone.next = (ListElement) support.clone(next, deeply);
 *         else
 *            // no DeepCloneSupport object passed in. Clone the next ListElement
 *            // but this may yield a StackOverflowError if the list is circular
 *            clone.next = next.clone(support, deeply);
 *      return clone;
 *   }
 * </pre>
 * Consider a linked list with nodes <code> a, b </code>, and <code>c</code>, with
 * <code>a.next == b, b.next = c</code> and
 * <code>c.next = a</code>.
 * <pre>
 *
 *     .---.      .---.     .---.
 *     | a |----->| b |---->| c |--.
 *     `___'      `___'     `___'  |
 *       ^                         |
 *       |                         |
 *       `-------------------------'
 *
 * </pre>
 * Calling <code>new DeepCloneSupport().clone(a, true)</code>
 * will result in a new linked list with three new nodes <code>a'</code>,
 * <code>b'</code>, and <code>c'</code> with the same structure as the
 * original list: <code>a'.next == b', b'.next = c'</code> and
 * <code>c'.next = a'</code>:
 * <pre>
 *
 *     .---.      .---.     .---.
 *     | a'|----->| b'|---->| c'|--.
 *     `___'      `___'     `___'  |
 *       ^                         |
 *       |                         |
 *       `-------------------------'
 *
 * </pre>
 * That is, <code>new DeepCloneSupport().clone(a, true)</code>
 * will call a.clone(DeepCloneSupport, boolean) which
 * will clone <code>a</code> and
 * <code>a</code>'s <var>next</var> field (<code>b</code>)
 * with:
 * <code>support.clone(next, deeply)</code> which will create
 * <code>b'</code>,
 * then it will in turn call
 * <code>support.clone(next, deeply)</code> which will create
 * <code>c'</code>.
 * When <code>c</code> is cloned,
 * it calls <code>support.clone(next, deeply)</code> which
 * returns <code>a'</code> instead of cloning <code>a</code>
 * again.
 *
 * @version 1.0
 * @since 1.2
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @see com.sas.PublicClonable
 * @see com.sas.DeepCloneSupport
 */
public interface DeepClonable
{
  /**
   * Clone the object deeply.
   * The object will clone not just itself but
   * also other objects that it contains or references.
   * The {@link com.sas.DeepCloneSupport support} object is used to support the deep clone by
   * keeping track of clones, so that a structural clone can be achieved
   * while avoiding infinite recursion. For example, some object structures
   * may contain cycles in the object references, and a default or context-free
   * deep clone would continue to clone objects over and over. The <var>support</var>
   * object provides a way to prevent this.
   * <p>
   * You normally do <strong>not</strong> call this method directly.
   * Rather, it is called from the DeepCloneSupport object after
   * a call to the support object's
   * {@link com.sas.DeepCloneSupport#clone(Object, boolean) clone(Object, boolean)}
   * method.
   * @param support an object to support a structural or deep clone by preventing
   * objects from being cloned more than once. If <var>support</var> is null,
   * try to perform a deep clone, but objects may be cloned more than once
   * and/or infinite loops may occur when cloning cyclic structures.
   * @parma deeply if true, perform a deep clone. If false, perform a
   * default, shallow clone.
   */
  public Object clone(DeepCloneSupport support, boolean deeply) throws CloneNotSupportedException;
}
