package com.sas;
import com.sas.collection.LongKeyDictionary;
import com.sas.security.BaseSecuritySupport;
import java.text.MessageFormat;

/**
 * A <code>DeepCloneSupport</code> object is used to support a deep clone by
 * keeping track of clones, so that a structural clone can be achieved
 * while avoiding infinite recursion. For example, some object structures
 * such as collections
 * may contain cycles in the object references. (A collection may contain itself
 * as an element.)  A default or context-free
 * deep clone would continue to clone objects over and over. The <code>DeepCloneSupport</code>
 * object provides a way to prevent this. As the <code>DeepCloneSupport</code> object clones objects,
 * they register themselves and their clones with the <code>DeepCloneSupport</code> object. By calling
 * the {@link #clone(Object object, boolean deeply)} method, if the <var>object</var>
 * has already been cloned and registered with the <code>DeepCloneSupport</code>, the clone
 * is returned immediately. Otherwise, the object's
 * {@link com.sas.DeepClonable#clone(com.sas.DeepCloneSupport support, boolean deeply)} method
 * will be called to clone the object. The <code>DeepCloneSupport</code> object registers (remmebers) the value
 * returned from this clone call.
 *
 * <h2>Usage</h2>
 * The proper way to clone one or more objects is to request
 * a DeepCloneSupport object to clone them:
 * <pre>
 *  DeepCloneSupport support = new DeepCloneSupport();
 *  cloneA = (TypeA) support.clone(a, true);
 *  cloneB = (TypeB) support.clone(b, true);
 *  cloneC = (TypeC) support.clone(c, true);
 * </pre>
 * This ensures all objects are registered when cloned and no
 * object gets cloned more than once.
 * @version 1.0
 * @since 1.2
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @see com.sas.DeepClonable
 */

public class DeepCloneSupport
{

  /**
   * The table which tracks clones. The keys are the
   * <a href="../../java/lang/System#identityHashCode(java.lang.Object)">java.lang.System.identityHashCode(java.lang.Object)</a>
   * for each source object and the values are the clones
   * of the source objects.
   */
  protected LongKeyDictionary table;

  /**
   * Default constructor.
   */
  public DeepCloneSupport ()
    {
       table = new LongKeyDictionary();
    }

  /**
   * Register an object and it's clone with this <code>DeepCloneSupport</code> object.
   * This is normally called from the object's {@link com.sas.DeepClonable#clone(com.sas.DeepCloneSupport support, boolean deeply)}
   * method, after the initial (shallow) clone, but before cloning other objects.
   * After registering, future calls to clone the <var>object</var>
   * will return the registered <var>clone</var> directly.
   * @param object A source object that has been cloned.
   * @param clone The clone of the source <var>object</var>
   */
  public void register(Object object, Object clone)
    {
      long key = System.identityHashCode(object);
      table.set(key, clone);
    }

  /**
   * clone is a utility method which clones the passed in Object.
   * If the Object implements the PublicClonable interface, that interface
   * will be used to call clone (since it is faster than using reflection).
   * Otherwise, reflection is used to determine if a public clone method
   * exists and calls it if it does.
   * <P>
   * This method will return null of the object passed in does not implement the
   * java.lang.Cloneable interface.  All other exceptions will be reported with
   * a java.lang.CloneNotSupportedException.
   * <P>
   * @param The object to clone
   * @return The clone of the object.
   * @exception CloneNotSupportedException is thrown if the object does not implement PubliClonable
   * and does not have a public clone() method callable via reflection.
   */
   public static Object clone( Object objectToClone ) throws CloneNotSupportedException
     {
       if ( objectToClone instanceof Cloneable )
       {
         if ( objectToClone instanceof PublicClonable )
         {
           return ((PublicClonable)objectToClone).clone();
         }
         else
         {
           // Use reflection to call clone() method.
           Object clonedObject = null;

           try
           {
             java.lang.reflect.Method cloneMethod;
             cloneMethod = BaseSecuritySupport.securitySupport.getMethod( objectToClone.getClass(), "clone", null);

             if ( cloneMethod == null )
             {
               String message = com.sas.RB.getStringResource( "DeepCloneSupport.noCloneMethod.ex.txt" );
               throw new CloneNotSupportedException( MessageFormat.format( message, new Object[] {objectToClone.getClass().getName()} ) );
             }

             clonedObject = BaseSecuritySupport.securitySupport.invokeMethod( cloneMethod, objectToClone, null );

             return clonedObject;
           } // try
           catch ( Exception exception )
           {
             String message = com.sas.RB.getStringResource( "DeepCloneSupport.cloneException.ex.txt" );
             throw new CloneNotSupportedException( MessageFormat.format( message, new Object[] {objectToClone.getClass().getName()} ) );
           } // catch
         } // else
       } // if

       return null;
     }

  /**
   * Return a clone of an object, optionally performing a deep clone.
   * The object and its clone are registered with this <code>DeepCloneSupport</code> instance.
   * @param object the object to clone. If the object has already registered a clone,
   * return that clone. Else, if the object is an instance of DeepClonable
   * then return
   * {@link DeepClonable#clone(DeepCloneSupport support, boolean deeply) object.clone(this, <var>deeply</var>)}.
   * Otherwise, clone the object.
   * @param deeply If true, try to clone the object deeply, if the object supports it.
   * @return the clone of the <var>object</var>.  If the <var>object</var> has already
   * been cloned and registered with this <code>DeepCloneSupport</code> instance, then return
   * the registered clone. Otherwise, clone it.
   */
  public Object clone(Object object, boolean deeply) throws CloneNotSupportedException
    {
      if (object instanceof DeepClonable)
        return clone((DeepClonable) object, deeply);
      Object clone = registered(object);
      if (clone != null)
        return clone;
      clone = clone(object);
      register(object, clone);
      return clone;
    }

  /**
   * Return a clone of an object, optionally performing a deep clone.
   * The object and its clone are registered with this <code>DeepCloneSupport</code> instance.
   * @param object the object to clone. If the object has already registered a clone,
   * return that clone. Otherwise, call its
   * {@link com.sas.DeepClonable#clone(com.sas.DeepCloneSupport, boolean) clone(this, boolean)}
   * method to clone it and register the clone with this <code>DeepCloneSupport</code> instance.
   * @param deeply If true, try to clone the object deeply. Otherwise, perform
   * a shallow clone.
   * @return the clone of the <var>object</var>.  If the <var>object</var> has already
   * been cloned and registered with this <code>DeepCloneSupport</code> instance, then return
   * the registered clone.
   */
  public Object clone(DeepClonable object, boolean deeply) throws CloneNotSupportedException
    {
      Object clone = registered(object);
      if (clone == null)
        {
          clone = object.clone(this, deeply);
          register(object, clone);
        }
      return clone;
    }

  /**
   * Test if an object has a clone registered with this <code>DeepCloneSupport</code>.
   * @param object an object
   * @return null if there is no clone registered for this object.
   * Return the clone if one is registered.
   */
  public Object registered(Object object)
    {
      long key = System.identityHashCode(object);
      Object clone = table.get (key, table.NO_SUCH_ELEMENT);
      return (   (clone == table.NO_SUCH_ELEMENT)
               ? null
               : clone );
    }

}
