package com.sas;

import com.sas.util.Enum;  // base class and static Enum methods
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import com.sas.util.Countable;

/**
 * An Enumeration class. Use the static final variables
 * as enumeration values: discrete named values from a set of
 * 29 values.
 * Note there are no public
 * constructors, so the final static variables listed below
 * will be the only instances of the class.
 *
 * @see com.sas.util.Enum
 */

final public class MissingValues extends Enum
	implements Countable
{
	/**
	* Currently only one missing value of float type.
	**/
	private static final int FLOAT_MISSING_VALUE = 0xFFD10000;

	private static MissingValues enums[] = new MissingValues [29];
	/**
	 * first enum in the set, or null if none.
	 * The constructors build a linked list of enums via
	 * this first enum and each enum's next.
	 */
	private static MissingValues first = null;

	/**
	* long value used to store raw value of MissingValues object.
	**/
	private long longValue;

	/**
	 * Number of enumerations.
	 */
	public static final int count = 29;

	/**
	 * Return the number of enums
	 * @return the number of enums
	 */
	public int count()
	{
		return count;
	}

	/**
	 * Return the next MissingValues in the sequence
	 * @return the next MissingValues in the sequence
	 */
	public MissingValues getNext()
	{
		return (MissingValues) next;
	}

	/**
	 * Return the next MissingValues in the sequence
	 * @return the next MissingValues in the sequence
	 */
	public Enum getNextEnum()
	{
		return next;
	}

	/**
	 * Return the ordinal position of an enum. The ordinal
	 * position is it's position relative to the others;
	 * this is a value from 0 to 28
	 * @return int
	 */
	public int getIndex()
	{
		return getIndex(this, first);
	}

	/**
	 * Return the enum corresponding to an index, or ordinal position.
	 * @param index The relative position [0,28]
	 * @return MissingValues
	 */
	public static MissingValues get(int index)
	{
		return (MissingValues) get(index, first);
	}

	/**
	 * Return the enum containing a specified value
	 * @param value A value of a MissingValues to find.
	 * @return MissingValues
	 */
	public static MissingValues getEnum(int value)
	{
		return (MissingValues) getEnum(value, first);
	}

	/**
	 * Return the enum corresponding to an enum name
	 * @param name The enumeration name to find
	 * @return MissingValues, null if no such enumeration
	 */
	public static MissingValues get(String name)
	{
		MissingValues enum = (MissingValues) get(name, first);

		/*
		 * Special case lookup for MissingValue (see MissingValues.get)
		 * If not found by regular name, try alternative names.
		 * that match use in SAS.
		 */
		if (enum==null && name != null)
		{
			if (".".equals(name))
				enum =Dot;
			else if ("._".equals(name))
				enum = _;
			else if (name.startsWith(".") && name.length() == 2)
				enum = get(name.substring(1, 2).toUpperCase());/*I18nOK:MTH*/
			else
			{
				String upperName = name.toUpperCase();/*I18nOK:MTH*/
				if ( !upperName.equals( name ) )
				{
					enum = get( upperName );
				}
			}
		}

		return enum;
	}

	/**
	* Return the MissingValues object for the specified float value.
	* There is currently only one float missing value.  That value is
	* returned from all the MissingValues in the getFloatValue(), but
	* this method will only return the MissingValues.Dot object.
	* @param floatValue the float value to convert to a MissingValues object.
	* @return MissingValues.Dot if floatValue rawIntBits equals the one 
	* float missing value, otherwise returns null.
	**/
	public static MissingValues toMissingValue( float floatValue )
	{
		int rawValue = Float.floatToRawIntBits( floatValue );
		if ( rawValue == FLOAT_MISSING_VALUE )
			return Dot;

		return null;
	}

	/**
	* Return the MissingValues object for the specified double value.
	* @param doubleValue the double value to convert to a MissingValues object.
	* @return the MissingValues object with the long value equal to the value of 
	* Double.doubleToRawLongBits of the doubleValue specified, otherwise returns null.
	**/
	public static MissingValues toMissingValue( double doubleValue )
	{
		if ( !Double.isNaN( doubleValue ) )
			return null;

		long rawValue = Double.doubleToRawLongBits( doubleValue );
		for ( MissingValues e = getFirst(); e != null; e = e.getNext() )
		{
			if ( e.longValue == rawValue )
				return e;
		}

		return null;
	}


	/**
	* Get the float value for this MissingValues object.
	* Currently there is only one float missing value for all the MissingValues.
	* The float returned will be a NaN value.
	* return the Float.intBitsToFloat(int) of the one float missing value.
	**/
	public float getFloatValue()
	{
		return Float.intBitsToFloat( FLOAT_MISSING_VALUE );
	}

	/**
	* Get the double value for this MissingValues object.
	* The double returned will be a NaN value.
	* return the Double.longBitsToDouble(long) value of this MissingValues object.
	**/
	public double getDoubleValue()
	{
		return Double.longBitsToDouble( longValue );
	}

	/**
	 * Support for Serialization. Write a MissingValues to
	 * an ObjectOutputStream.  Use this method to serialize
	 * a MissingValues held in an instance variable if you want
	 * to make the holding object Serializable. Enum objects
	 * are not Serializable because they are immutable singletons.
	 * (The benefit of enums is there is only one of each value.)
	 * @param os an Object Output Stream.
	 */
	public void write(ObjectOutputStream os) throws IOException
	{
		os.writeInt(getValue());
	}

	/**
	 * Deserialization support. Read a MissingValues from an
	 * ObjectInputStream. Note that MissingValues objects are not serializable;
	 * this prevents creation of multiple objects with the same
	 * value and/or name. Call this from your class' private readObject
	 * method if your class is Serializable and keeps MissingValues objects;
	 * the MissingValues instance variable in your class should be transient.
	 * @param is the ObjectInputStream
	 */
	public static MissingValues read(ObjectInputStream is) throws IOException
	{
		int value = is.readInt();
		MissingValues enum = getEnum(value);
		if (enum == null)
			throw new IOException();

		return enum;
	}

	/**
	 * Get the first enum in the sequence.
	 * @return the next Enum, or null if this is the last in the sequence.
	 */
	public static MissingValues getFirst()
	{
		return (MissingValues) first;
	}

	/**
	 * Return an array of all the MissingValues enum items.
	 * @return an array of all the MissingValues enum items.
	 */
	public static MissingValues[] getEnums()
	{
		MissingValues enums[] = new MissingValues[count];
		MissingValues e = getFirst();
		for (int i = 0; e != null; e = e.getNext())
			enums[i++] = e;

		return enums;
	}

	/**
	 * Construct a new MissingValues enum from a name
	 * and its integer value.
	 * @param name The name (identifier) for this enumeration
	 * @param value An int value to associate with the enumeration
	 */
	private MissingValues(String name, int value, long rawValue )
	{
		super(name, value);
		// insert at beginning of static list -- constructors are
		// called in reverse order!

		longValue = rawValue;
		this.next = first;
		first     = this;
	}

	// Generate the enums in reverse order so the linked list is built correctly.

	/**
	  * .Z missing value
	  */
	public static final MissingValues Z = new MissingValues ("Z", 28, 0xFFFFE40000000000L );


	/**
	  * .Y missing value
	  */
	public static final MissingValues Y = new MissingValues ("Y", 27, 0xFFFFE50000000000L );


	/**
	  * .X missing value
	  */
	public static final MissingValues X = new MissingValues ("X", 26, 0xFFFFE60000000000L );


	/**
	  * .W missing value
	  */
	public static final MissingValues W = new MissingValues ("W", 25, 0xFFFFE70000000000L );


	/**
	  * .V missing value
	  */
	public static final MissingValues V = new MissingValues ("V", 24, 0xFFFFE80000000000L );


	/**
	  * .U missing value
	  */
	public static final MissingValues U = new MissingValues ("U", 23, 0xFFFFE90000000000L );


	/**
	  * .T missing value
	  */
	public static final MissingValues T = new MissingValues ("T", 22, 0xFFFFEA0000000000L );


	/**
	  * .S missing value
	  */
	public static final MissingValues S = new MissingValues ("S", 21, 0xFFFFEB0000000000L );


	/**
	  * .R missing value
	  */
	public static final MissingValues R = new MissingValues ("R", 20, 0xFFFFEC0000000000L );


	/**
	  * .Q missing value
	  */
	public static final MissingValues Q = new MissingValues ("Q", 19, 0xFFFFED0000000000L );


	/**
	  * .P missing value
	  */
	public static final MissingValues P = new MissingValues ("P", 18, 0xFFFFEE0000000000L );


	/**
	  * .O missing value
	  */
	public static final MissingValues O = new MissingValues ("O", 17, 0xFFFFEF0000000000L );


	/**
	  * .N missing value
	  */
	public static final MissingValues N = new MissingValues ("N", 16, 0xFFFFF00000000000L );


	/**
	  * .M missing value
	  */
	public static final MissingValues M = new MissingValues ("M", 15, 0xFFFFF10000000000L );


	/**
	  * .L missing value
	  */
	public static final MissingValues L = new MissingValues ("L", 14, 0xFFFFF20000000000L );


	/**
	  * .K missing value
	  */
	public static final MissingValues K = new MissingValues ("K", 13, 0xFFFFF30000000000L );


	/**
	  * .J missing value
	  */
	public static final MissingValues J = new MissingValues ("J", 12, 0xFFFFF40000000000L );


	/**
	  * .I missing value
	  */
	public static final MissingValues I = new MissingValues ("I", 11, 0xFFFFF50000000000L );


	/**
	  * .H missing value
	  */
	public static final MissingValues H = new MissingValues ("H", 10, 0xFFFFF60000000000L );


	/**
	  * .G missing value
	  */
	public static final MissingValues G = new MissingValues ("G", 9, 0xFFFFF70000000000L );


	/**
	  * .F missing value
	  */
	public static final MissingValues F = new MissingValues ("F", 8, 0xFFFFF80000000000L );


	/**
	  * .E missing value
	  */
	public static final MissingValues E = new MissingValues ("E", 7, 0xFFFFF90000000000L );


	/**
	  * .D missing value
	  */
	public static final MissingValues D = new MissingValues ("D", 6, 0xFFFFFA0000000000L );


	/**
	  * .C missing value
	  */
	public static final MissingValues C = new MissingValues ("C", 5, 0xFFFFFB0000000000L );


	/**
	  * .B missing value
	  */
	public static final MissingValues B = new MissingValues ("B", 4, 0xFFFFFC0000000000L );


	/**
	  * .A missing value
	  */
	public static final MissingValues A = new MissingValues ("A", 3, 0xFFFFFD0000000000L );


	/**
	  * (.) missing value
	  */
	public static final MissingValues Dot = new MissingValues ("Dot"/*I18nOK:EMS*/, 2, 0xFFFFFE0000000000L );


	/**
	  * ._ missing value
	  */
	public static final MissingValues _ = new MissingValues ("_", 1, 0xFFFFFF0000000000L );


	/**
	  * SCL uninitialized value
	  */
	public static final MissingValues _BLANK_ = new MissingValues ("_BLANK_", 0, 0xFFFFFFFFFFFFFFFFL );

}

