/* Copyright (c) 1998-2000 by SAS Institute Inc., Cary, NC 27513 */

package com.sas;

import java.text.NumberFormat;
import java.util.Locale;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.StringTokenizer;
import java.net.URL;

/**
 * Version contains the build version of the webAF class library.
 */

public class Version
{
	static Manifest manifest = null;

  /**
   * Print a copyright notice and the webAF class library version information to System.out.
   */
  public static void main(String args[])
  {
    System.out.println( "Copyright (c) 1998-2003 by SAS Institute Inc., Cary, NC 27513"/*I18nOK:EMS*/);
    System.out.println( "webAF Class Library Version Info"/*I18nOK:EMS*/ );
    System.out.println( "   Major Version: " + getMajorVersion() ); /*I18nOK:LINE*/
    System.out.println( "   Minor Version: " + getMinorVersion() ); /*I18nOK:LINE*/
    System.out.println( "   Micro Version: " + getMicroVersion() ); /*I18nOK:LINE*/
    System.out.println( "   Build Date: " + getBuildDate() ); /*I18nOK:LINE*/
    System.out.println( "   Build Number: " + getBuildNumber() ); /*I18nOK:LINE*/
    System.out.println( "   Extra Information: " + getExtraInfo() ); /*I18nOK:LINE*/
  }

  /**
   * Return the current version string.
   * The content of the version string is subject to change, so it should be treated as a
   * string with no defined structure.
   */
  public static String getVersion()
  {
	NumberFormat numberFormatter = NumberFormat.getNumberInstance( Locale.US );
	numberFormatter.setMinimumIntegerDigits( 3 );

	return getMajorVersion() + "." +
		  getMinorVersion() + "." +
		  getBuildDate() + "." +
		  numberFormatter.format( getBuildNumber() );
  }

  /**
   * Return the current version string.
   * The version string consists of four elements
   * in the format "V.v.m.yyyyMMdd.bbb[.x]"
   * where V is the major version number, v is the minor version number,
   * m is the micro version number,
   * yyyyMMdd the build date, bbb is the build number, and x is any extra information
   * about the build version that is available.  if there is no extra information,
   * the separator will not be included in the output.
   * A '.' separates each element.
   */
  public static String getFullVersion()
  {
	NumberFormat numberFormatter = NumberFormat.getNumberInstance( Locale.US );
	numberFormatter.setMinimumIntegerDigits( 3 );

	StringBuffer buffer = new StringBuffer();
	buffer.append( getMajorVersion() + "." );
	buffer.append( getMinorVersion() + "." );
	buffer.append( getMicroVersion() + "." );
	buffer.append( getBuildDate() + "." );
	buffer.append( numberFormatter.format( getBuildNumber() ) );

	String extraInfo = getExtraInfo();
	if ( extraInfo != null  &&  extraInfo.length() > 0 )
	{
		buffer.append( "." );
		buffer.append( extraInfo );
	}

	return buffer.toString(); /*I18nOK:MTH*/
  }

	/**
	 * Returns the major release version for the webAF class library.
	 * The current major release version is 3.
	 */
	public static int getMajorVersion()
	{
		String version = getManifestValue("Specification-Version"/*I18nOK:EMS*/);
		StringTokenizer tokenizer = new StringTokenizer(version, ".");
		Integer value = new Integer(tokenizer.nextToken());
		return value.intValue();
	}

	/**
	 * Returns the minor release version for the webAF class library.
	 * The current minor release version is 0.
	 */
	public static int getMinorVersion()
	{
		String version = getManifestValue("Specification-Version"/*I18nOK:EMS*/);
		StringTokenizer tokenizer = new StringTokenizer(version, ".");
		tokenizer.nextToken();
		Integer value = new Integer(tokenizer.nextToken());
		return value.intValue();
	}

	/**
	 * Returns the micro version of the webAF class library.  This generally is used
	 * to identify versions of released JARS with the same major and minor versions.
	 * The current micro version is .
	 */
	public static int getMicroVersion()
	{
		String version = getManifestValue("Specification-Version"/*I18nOK:EMS*/);
		StringTokenizer tokenizer = new StringTokenizer(version, ".");
		tokenizer.nextToken();
		tokenizer.nextToken();
		Integer value = new Integer(tokenizer.nextToken());
		return value.intValue();
	}

	/**
	 * Returns a String containing the daily build number.
	 * The current build number is .
	 */
	public static int getBuildNumber()
	{
		String version = getManifestValue("Specification-Version"/*I18nOK:EMS*/);
		StringTokenizer tokenizer = new StringTokenizer(version, ".");
		tokenizer.nextToken();
		tokenizer.nextToken();
		tokenizer.nextToken();
		Integer value = new Integer(tokenizer.nextToken());
		return value.intValue();
	}

	/**
	 * Returns a String containing extra information about the build version.
	 * The format of the build date is "yyyyMMdd" where "yyyy" is the year,
	 * "MM" is the month (1-12), and "dd" is the day of the month (1-31).
	 * The current build date is "20020325".
	 * @return the String containing the extra information about the build version
	 */
	public static String getBuildDate()
	{
		return getManifestValue("Implementation-Build-Date"/*I18nOK:EMS*/);
	}

	/**
	 * Returns a String containing extra information about the build version.
	 * The format and content of this information is subject to change.
	 * The current extra information is "".
	 */
	public static String getExtraInfo()
	{
		return "";  // EXTRA_INFO_KEY
	}

	protected static String getManifestValue(String key)
	{
		getManifest();

		// Get the specified attribute from the manifest.
		Attributes mainAttributes = manifest.getMainAttributes();
		return mainAttributes.getValue(key);
	}

	protected static void getManifest()
	{
		if (manifest != null)
			return;

		try
		{
			final String fileProtocol = "file:/";

			// First get the URL of the jar containing this class.
			// Must be a privileged call since we might be running in an applet sandbox environment.
			URL url = (URL)java.security.AccessController.doPrivileged( 
				new java.security.PrivilegedAction(){
					public final Object run(){
						return Version.class.getClassLoader().getResource("com/sas/Version.class");
					}
				}
			);

			String fileURL = url.getFile();

			// And extract just the file system path of the jar.
			int start = fileURL.indexOf(fileProtocol);
			if (start != -1)
			{
				start += fileProtocol.length();
				int end = fileURL.indexOf('!');
				if (end != -1)
				{
					fileURL = fileURL.substring(start, end);
					final String URL = fileURL;

					// Now get a JarFile for this path.
					// Must be a privileged call since we might be running in an applet sandbox environment.
					JarFile jarFile = (JarFile)java.security.AccessController.doPrivileged( 
						new java.security.PrivilegedAction(){
							public final Object run(){
								try {
									return new JarFile(URL);
								}
								catch (Exception e)
								{
									// Intentionally do nothing here
								}
								return null;
							}
						}
					);

					// Get the manifest from the jar file.
					manifest = jarFile.getManifest();
				}
			}
		}
		catch (Exception e)
		{
			// Intentionally do nothing here.
//			e.printStackTrace();
		}
	}
}
