// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.beans;

import com.sas.GlobalInfoTable;

/**
 * BeanDescriptor is an extension of <code>java.beans.BeanDescriptor</code>
 * which overrides getCustomizerClass() to lookup the Customizer
 * for ROCF interfaces.  A <code>com.sas.beans.BeanDescriptor</code> may be returned 
 * by the <code>com.sas.beans.Introspector</code> when introspecting
 * a class.
 */
public class BeanDescriptor extends java.beans.BeanDescriptor
{
	protected Class customizerClass;

	/**
	 * Constructor.  Will cause the lookup of the Customizer by name.
	 */
	public BeanDescriptor( Class beanClass )
	{
		super( beanClass );
	}
	
	/**
	 * Constructor with specified Customizer class.
	 */
	public BeanDescriptor( Class beanClass, Class customizerClass )
	{
		super( beanClass, customizerClass );
		
		this.customizerClass = customizerClass;
	}

    /**
	 * Returns the Class object for the JavaBean's Customizer.
     * @return The Class object for the JavaBean's Customizer.  This may
     * be null if the bean doesn't have a Customizer.
     */
    public Class getCustomizerClass()
	{
		if ( customizerClass == null )
		{
			customizerClass = lookupCustomizerClass(getBeanClass());
		}
		
		return customizerClass;
    }

    /**
	 * Sets the Class object for the JavaBean's Customizer.
     * @param customizerClass The Class object for the JavaBean's Customizer.  This may
     * be null if the Customizer should be looked up by name.
     */
	public void setCustomizerClass( Class customizerClass )
	{
		this.customizerClass = customizerClass;
	}

	/**
	 * This is a helper method that looks up the JavaBean's Customizer.
	 * The Customizer is first looked up by appending "Customizer" to
	 * the JavaBean's class name.  If that class exists, it is returned.
	 * If it does not exist and the class implements the IntrospectInterfaceForBeanInfo
	 * interface, the base name of the class is determined and a Customizer
	 * is looked up for that class.
	 * @param beanClass The class to determine the Customizer for.
	 * @return The class of the Customizer or null if no Customizer is found.
	 */
	protected static Class lookupCustomizerClass( Class beanClass )
	{
//System.out.println("Looking up customizer class for : " + beanClass );

		// Try to look up the Customizer based on the Bean's class name

		{
			String customizerName = beanClass.getName() + "Customizer"/*I18nOK:LINE*/;
			Class customizerClass = loadCustomizer( beanClass, customizerName );
			if ( customizerClass != null )
				return customizerClass;
		}
		{
//DA 07/07/00:		   Class jcomponentClass = GlobalInfoTable._loadClass( "javax.swing.JComponent" );
//DA 07/07/00:		   if (jcomponentClass.isAssignableFrom (beanClass)){
			  String customizerName = "com.sas.beans.customizers." + beanClass.getName() + "Customizer"/*I18nOK:LINE*/;
			  Class customizerClass = loadCustomizer( beanClass, customizerName );
			  if ( customizerClass != null )
               	return customizerClass;
		}
			 
//DA 07/07/00:		  } //if jcomponentClass

		// If the class is a ROCF class, see if there's a Customizer for the 
		// interface instead of the instance.

		Class introspectInterfaceForBeanInfoClass = GlobalInfoTable._loadClass( "com.sas.beans.IntrospectInterfaceForBeanInfo" );
		if ( introspectInterfaceForBeanInfoClass.isAssignableFrom( beanClass ) )
		{
			String customizerName = Introspector.getRemoteClassBaseName( beanClass.getName() );
			customizerName = customizerName + "Customizer"/*I18nOK:LINE*/;

//System.out.println("Looking for REMOTE customizer : " + customizerName );
			Class customizerClass = loadCustomizer( beanClass, customizerName );
			if ( customizerClass != null )
			{
//System.out.println("Found REMOTE customizer class for : " + beanClass );
				return customizerClass;
			}
		} // if

		return null;
	} // lookupCustomizerClass

	/**
	 * This is a helper method to load the Class of the Customizer.
	 * @param beanClass The class to load the Customizer for.
	 * @param customizerName The name of the Customizer class.
	 * @return The class of the Customizer or null if no Customizer is found.
	 */
	protected static Class loadCustomizer( Class beanClass, String customizerName )
	{
		try
		{
			if ( beanClass.getClassLoader() == null )
			{
				
				Class beansCustomizerClass = java.beans.Customizer.class;
				Class customizerClass = GlobalInfoTable.loadClass( customizerName );
				if ( beansCustomizerClass.isAssignableFrom( customizerClass ) )
				  return customizerClass;
				
			} // if
			else
			{
				Class customizerClass = beanClass.getClassLoader().loadClass(customizerName);
				return customizerClass;
			} // else
		} // try
		catch ( Exception e )
		{
			// Class not found.  Fall through and return null...
		} // catch

		return null;
	}
}
