// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.beans;

/**
 * ExtendedBeanInfo is a helper class that lets components specify BeanInfo for their class
 * without writing a BeanInfo class.
 * <P>
 * Note that the information provided by ExtendedBeanInfo is only available when performing
 * Introspection using com.sas.beans.Introspector.
 * <P>
 * Default input and output properties may be specified so that default linking is easier.
 * The default input property should be the property which will most often be used as the 
 * target for a link.
 * The default ouput property should be the property which will most often be used as the 
 * source for a link.
 * <P>
 * The shortDescription can be used to set the description of the Bean class.  This will be
 * returned when the getShortDescription method is called on the BeanDescriptor returned from
 * BeanInfo.getBeanDescriptor.
 * <P>
 * The ExtendedBeanInfo class provides a bunch of static final String definitions to help
 * in defining the property metadata.  See below for an example of the metadata format.
 */
public class ExtendedBeanInfo
{
   // Non-translated Strings

   public static final String NAME                  = "Name"/*I18nOK:EMS*/;
   public static final String DEFAULT_VALUE         = "Default value"/*I18nOK:EMS*/;
   public static final String SHORT_DESCRIPTION     = "Short description"/*I18nOK:EMS*/;
   public static final String DESCRIPTION           = "Description"/*I18nOK:EMS*/;
   public static final String READ_METHOD           = "Read method"/*I18nOK:EMS*/;
   public static final String WRITE_METHOD          = "Write method"/*I18nOK:EMS*/;
   public static final String BOUND                 = "Bound"/*I18nOK:EMS*/;
   public static final String LINKABLE              = "Linkable"/*I18nOK:EMS*/;
   public static final String COMPARE_VALUE         = "Compare value"/*I18nOK:EMS*/;
   public static final String PERSIST               = "Persist"/*I18nOK:EMS*/;
   public static final String FIELD_NAME            = "Field name"/*I18nOK:EMS*/;
   public static final String CONSTRAINED           = "Constrained"/*I18nOK:EMS*/;
   public static final String PROPERTY_EDITOR_CLASS = "Property editor class"/*I18nOK:EMS*/;
   public static final String EXPERT                = "Expert"/*I18nOK:EMS*/;
   public static final String HIDDEN                = "Hidden"/*I18nOK:EMS*/;
   public static final String MASKED                = "Masked"/*I18nOK:EMS*/;
   public static final String SEND_EVENT            = "Send event"/*I18nOK:EMS*/;
   public static final String STATIC                = "Static"/*I18nOK:EMS*/;
   public static final String CATEGORY              = "Category"/*I18nOK:EMS*/;

	/**
	 * Constructor.
	 */
	public ExtendedBeanInfo()
	{
		defaultInputPropertyName = "";
		defaultOutputPropertyName = "";
		shortDescription = "";
	}

	/**
	 * The name of the property which will be used as the default when the Bean is used as 
	 * the target for a link.
	 */
	public String defaultInputPropertyName;

	/**
	 * The name of the property which will be used as the default when the Bean is used as 
	 * the source for a link.
	 */
	public String defaultOutputPropertyName;

	/**
	 * An three dimensional array of information about the properties in the class.
	 * The format of the array is like the following.  Each entry in the second
	 * dimension describes a single property.  The first entry in each of the the third
	 * dimension's arrays will be the property name (specified with a "Name" tag).
	 * 
	 * <PRE>
	 * {
	 *    {
	 *       { ExtendedBeanInfo.NAME, "somePropertyName" },
	 *       { ExtendedBeanInfo.DESCRIPTION, "A hidden property" },
	 *       { ExtendedBeanInfo.HIDDEN, "true" }
	 *    },
	 *    {
	 *       { ExtendedBeanInfo.NAME, "someOtherPropertyName" },
	 *       { ExtendedBeanInfo.DESCRIPTION, "An expert property" },
	 *       { ExtendedBeanInfo.EXPERT, "true" }
	 *    }
	 * }
	 *  
	 * </PRE>
	 */
	public String[][][] propertyMetadata;
	
	/**
	 * Corresponds to the BeanDescriptor's shortDescription
	 */
	public String shortDescription;


  /**
    * Composite the passed in ExtendedBeanInfo with the information that is in
    * this ExtendedBeanInfo. The passed in information is overrides the information
    * that is in this ExtendedBeanInfo.
    *
    * @param ebi ExtendedBeanInfo object to merge with
    */
  public void composite(ExtendedBeanInfo ebi)
	{
          // Override the default input and output properties if they are set in
          // "ebi"
		
          if (ebi.defaultInputPropertyName.length() > 0) {
            defaultInputPropertyName = ebi.defaultInputPropertyName;
          }
          
          if (ebi.defaultOutputPropertyName.length() > 0) {
            defaultOutputPropertyName = ebi.defaultOutputPropertyName;
          }
          
          // Override the short description
          if (ebi.shortDescription.length() > 0) {
            shortDescription = ebi.shortDescription;
          }

          // Merge the property metadata
          if (propertyMetadata == null) {
            
            // Nothing to merge; just set it directly
            propertyMetadata = ebi.propertyMetadata;
          }
          if (ebi.propertyMetadata != null) {

            // Each object has metadata; merge them
            java.util.Vector v = new java.util.Vector();

            // First dump the contents of the given bean info into the vector.
            // This will serve as our base. Any duplicate information will be
            // merged together with any base information overriding any
            // information in this bean info
            for (int i = 0; i < ebi.propertyMetadata.length; i++) {
              v.addElement(ebi.propertyMetadata[i]);
            }

            // We now have a new vector with all of the metadata. Go through
            // our metadata and merge
            for (int i = 0; i < propertyMetadata.length; i++) {

              // Get the 'Name' of the property
              String name = findName(propertyMetadata[i]);

              if (name != null) {

                // We have a named property. See if we already have one in the
                // vector
                int index = findProperty(name, v);

                if (index >= 0) {

                  // The metadata already exists. Merge them together.
                  mergeMetadata(v, index, propertyMetadata[i]);
                }
                else {

                  // The metadata is new. Add to the vector
                  v.addElement(propertyMetadata[i]);
                }
              }
            }

            // Convert the vector back into a property array
            propertyMetadata = new String[v.size()][][];

            for (int i = 0; i < propertyMetadata.length; i++) {
              propertyMetadata[i] = (String[][]) v.elementAt(i);
            }
          }

        }

  /**
    * Get the 'Name' attribute of the given metadata
    *
    * @param metadata Property metadata
    * @return The name attribute of the property or null if not found
    */
  protected String findName(String metadata[][])
    {
      String name = null;

      // Look for the 'Name' attribute
      for (int i = 0; i < metadata.length; i++) {

        if (metadata[i].length > 1) {
          if (metadata[i][0].equals(NAME)) {
            name = metadata[i][1];
            break;
          }
        }
      }
      return name;
    }

  /**
    * Find the named property in the given vector of metadata. Returns -1 if
    * the property is not found
    *
    * @param name Property name
    * @param v Vector of metadata
    * @return Index into the vector or -1 if not found
    */
  protected int findProperty(String name, java.util.Vector v)
    {
      int index = -1;

      // Search the vector and look for the named property
      for (int i = 0; i < v.size(); i++) {
        String[][] metadata = (String[][]) v.elementAt(i);

        // Find the property name of this metadata
        String name2 = findName(metadata);

        if (name2 != null) {
          // Found a name; does it match?
          if (name2.equals(name)) {
            index = i;
            break;
          }
        }
      }
      return index;
    }

  /**
    * Merge the given metadata together.
    *
    * @param v Vector containing the base (or default) metadata
    * @param index Index into the vector
    * @param metadata Additional property metadata
    */
  protected void mergeMetadata(java.util.Vector v, int index, String[][] metadata)
    {
      // Get the default metadata
      String[][] md = (String[][]) v.elementAt(index);

      // Put all of the default metadata into a vector
      java.util.Vector p = new java.util.Vector();
      for (int i = 0; i < md.length; i++) {
        p.addElement(md[i]);
      }

      // Loop through the new metadata and merge. Only add to the vector if this
      // is a new property
      for (int i = 0; i < metadata.length; i++) {

        // Make sure we have named metadata
        if (metadata[i].length < 1) {
          continue;
        }

        boolean hasValue = false;
        
        // Attempt to find the same value in the default metadata
        for (int j = 0; j < p.size(); j++) {
          String[] def = (String[]) p.elementAt(j);

          // Make sure it's named
          if (def.length < 1) {
            continue;
          }

          if (def[0].equals(metadata[i][0])) {
            hasValue = true;
            break;
          }
        }

        // If we did not have a value, set it
        if (!hasValue) {
          p.addElement(metadata[i]);
        }
      }

      // Convert the vector back into a String array
      md = new String[p.size()][];

      for (int i = 0; i < md.length; i++) {
        md[i] = (String[]) p.elementAt(i);
      }
      
      // Set the new metadata
      v.setElementAt(md, index);
    }
}
