// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

/**
 * An Introspector class which returns additional SAS IDE bean info.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @support sasspl
 */

package com.sas.beans;

import com.sas.GlobalInfoTable;
import com.sas.collection.Pair;
import com.sas.security.BaseSecuritySupport;
import com.sas.util.Util;
import com.sas.util.Classes;

import java.io.IOException;
import java.io.InputStream;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import java.beans.BeanInfo;
import java.beans.SimpleBeanInfo;
import java.beans.ParameterDescriptor;
import java.beans.MethodDescriptor;
import java.beans.EventSetDescriptor;
import java.beans.IntrospectionException;
import java.beans.PropertyVetoException;
import java.beans.PropertyDescriptor;
import java.beans.IndexedPropertyDescriptor;
import java.beans.BeanDescriptor;

import java.text.MessageFormat;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;
import java.util.Set;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.WeakHashMap;
import java.util.HashSet;

public class Introspector
{
	public static final String RB_KEY = "Introspector.";

	public static final String DEFAULT_WIDTH = "defaultWidth";
	public static final String DEFAULT_HEIGHT = "defaultHeight";
	public static final String DEFAULT_INPUT_PROPERTY = "defaultInputProperty";
	public static final String DEFAULT_OUTPUT_PROPERTY = "defaultOutputProperty";

	/**
	 * This is a cache for BeanInfo that has already been introspected.
	 * There is currently no way to clear the cache.
	 */
	private static WeakHashMap 	cache = new WeakHashMap();

	private static Map			defaultSizeMap;
	private static Map 			extendedBeanInfoMap;

	/**
	 * Private constructor to prevent creation of this object...
	 */
	private Introspector()
	{
	}

	/*
	 * Get the bean info for a class.
	 * @see java.beans.BeanInfo(java.lang.Class)
	 */
	public static java.beans.BeanInfo getBeanInfo(Class beanClass)
		throws IntrospectionException
	{
		return getBeanInfo(beanClass, null);
	}

	/*
	 * Get the bean info for a class.
	 * @see java.beans.BeanInfo(java.lang.Class,java.lang.Class)
	 */
	public static BeanInfo getBeanInfo(Class beanClass, Class stopClass)
		throws IntrospectionException
	{
		/**
		 * Look the BeanInfo up in the cache
		 */
		Pair index = new Pair(beanClass, stopClass);
		BeanInfo beanInfo = (BeanInfo) cache.get(index);
		if (null == beanInfo)
		{
			// Introspect the BeanInfo if it's not in the cache

			beanInfo = introspect(beanClass, stopClass);
			cache.put(index, beanInfo);
		}
		return beanInfo;
	}


    /**
     * Flush all of the Introspector's internal caches.  This method is
     * not normally required.  It is normally only needed by advanced
     * tools that update existing "Class" objects in-place and need
     * to make the Introspector re-analyze existing Class objects.
     */

	public static void flushCaches()
	{
		cache.clear();
		extendedBeanInfoMap = null; // this will force the resources to get reloaded.
	}

	/*
	 * Decapitalize a string
	 * @see java.beans.Decapitalize(java.lang.String)
	 */
	public static java.lang.String decapitalize(String string)
	{
		return java.beans.Introspector.decapitalize(string);
	}

	/*
	 * Get the bean info search path
	 * @see java.beans.getBeanInfoSearchPath()
	 */
	public static String[] getBeanInfoSearchPath()
	{
		return java.beans.Introspector.getBeanInfoSearchPath();
	}

	/*
	 * Get the bean info search path
	 * @see java.beans.setBeanInfoSearchPath(java.lang.String[])
	 */
	public static void setBeanInfoSearchPath(String path[])
	{
		java.beans.Introspector.setBeanInfoSearchPath(path);
	}

	/**
	* Perform introspection on the bean classes to find their bean
	* info. This uses java.beans.Introspector to do most of the work,
	* then if there are SAS IDE component properties which are registered to send
	* PropertyChangeEvents's on modification  then we construct additional event
	* set descriptors for that event.  Currently, we do not save the
	* set of property names. There's no place to put them.
	*
	* @param beanClass the bean class from which we look up properties.
	* @param stopClass Do not return properties defined at or above this class.
	*                  Pass null to not exclude any properties.
	* @return BeanInfo a new BeanInfo which contains the new event set descriptors.
	* @exception java.beans.IntrospectionException if there is an introspection error.
	*/
	protected static BeanInfo introspect(Class beanClass, Class stopClass)
		throws IntrospectionException
	{
		validateStopClass(beanClass, stopClass); // may throw IntrospectionException

		// get the default bean info, including the default event set descriptors
		BeanInfo beanInfo = com.sas.security.BaseSecuritySupport.securitySupport.getBeanInfo(beanClass, stopClass);

		EventSetDescriptor descriptors[] = beanInfo.getEventSetDescriptors();

		ExtendedBeanInfo ebi = new ExtendedBeanInfo();
		getExtendedBeanInfo( beanClass, ebi );

		// Declare storage for extended bean information

		java.beans.PropertyDescriptor[] instanceProperties = beanInfo.getPropertyDescriptors();
		java.beans.PropertyDescriptor[] staticProperties = introspectStaticProperties( beanClass, beanInfo );
		int defaultInputPropertyIndex = beanInfo.getDefaultPropertyIndex();
		int defaultOutputPropertyIndex = beanInfo.getDefaultPropertyIndex();
		java.awt.Image icon = beanInfo.getIcon(java.beans.BeanInfo.ICON_COLOR_16x16);

		// Extract extended bean information if available

		if ( ebi != null )
		{
			// Annotate the properties based on the ExtendedBeanInfo
			annotateProperties( beanClass, instanceProperties, ebi );
			annotateProperties( beanClass, staticProperties, ebi );

			// Lookup the default input and output property indexes
			if ( ebi.defaultInputPropertyName != null && ebi.defaultInputPropertyName.length() > 0 )
				defaultInputPropertyIndex = lookupPropertyIndex( instanceProperties, defaultInputPropertyIndex, ebi.defaultInputPropertyName );
			if ( ebi.defaultOutputPropertyName != null && ebi.defaultOutputPropertyName.length() > 0 )
				defaultOutputPropertyIndex = lookupPropertyIndex( instanceProperties, defaultInputPropertyIndex, ebi.defaultOutputPropertyName );
		} // if		

		BeanDescriptor beanDescriptor = beanInfo.getBeanDescriptor();
		if ( beanDescriptor == null )
		{
			// Make a default beanDescriptor
			beanDescriptor = new com.sas.beans.BeanDescriptor( beanClass );
		}
		else if ( beanDescriptor instanceof com.sas.beans.BeanDescriptor == false )
		{
			com.sas.beans.BeanDescriptor newBeanDescriptor = new com.sas.beans.BeanDescriptor( beanDescriptor.getBeanClass(), beanDescriptor.getCustomizerClass() );

			// Copy the information out of the old BeanDescriptor into the new one.
			newBeanDescriptor.setShortDescription( beanDescriptor.getShortDescription() );

			beanDescriptor = newBeanDescriptor;
		}

		// Fill in the short description from the ExtendedBeanInfo

		if (ebi.shortDescription.length() > 0 )
		{
			beanDescriptor.setShortDescription(ebi.shortDescription);
		}

		// Annotate the BeanDescriptor

		//System.out.println("defaultInputPropertyIndex = " + defaultInputPropertyIndex );
		//System.out.println("defaultOutputPropertyIndex = " + defaultOutputPropertyIndex );

		String defaultInputProperty = null;
		String defaultOutputProperty = null;
		if ( defaultInputPropertyIndex != -1 )
		{
			defaultInputProperty = instanceProperties[defaultInputPropertyIndex].getName();
		}
		if ( defaultOutputPropertyIndex != -1 )
		{
			defaultOutputProperty = instanceProperties[defaultOutputPropertyIndex].getName();
		}

		annotateBeanDescriptor( beanClass, beanDescriptor,
			defaultInputProperty,
			defaultOutputProperty );

		// Create and return the BeanInfo
		return new com.sas.beans.SimpleBeanInfo(beanDescriptor,
			descriptors,
			beanInfo.getDefaultEventIndex(),
			instanceProperties,
			beanInfo.getDefaultPropertyIndex(),
			beanInfo.getMethodDescriptors(),
			staticProperties,
			defaultInputPropertyIndex,
			defaultOutputPropertyIndex,
			icon );
	}

	/**
	 * 
	 */
	protected static void annotateBeanDescriptor( Class beanClass, BeanDescriptor beanDescriptor, String defaultInputProperty, String defaultOutputProperty )
	{
		//		System.out.println("annotateBeanDescriptor" );
		try
		{
			Class [] paramTypes = new Class[0];
 			java.lang.reflect.Method getDefaultWidthMethod = com.sas.security.BaseSecuritySupport.securitySupport.getMethod( beanClass, "getDefaultWidth", paramTypes );
			Integer defaultWidth = (Integer) com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( getDefaultWidthMethod, null, null );
			beanDescriptor.setValue( DEFAULT_WIDTH, defaultWidth );

			java.lang.reflect.Method getDefaultHeightMethod = com.sas.security.BaseSecuritySupport.securitySupport.getMethod( beanClass, "getDefaultHeight", paramTypes );
			Integer defaultHeight = (Integer) com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( getDefaultHeightMethod, null, null );
			beanDescriptor.setValue( DEFAULT_HEIGHT, defaultHeight );

			beanDescriptor.setValue( DEFAULT_INPUT_PROPERTY, defaultInputProperty );
			beanDescriptor.setValue( DEFAULT_OUTPUT_PROPERTY, defaultOutputProperty );
		} // try
		catch ( Exception e)
		{
			//			e.printStackTrace();
			// Look in the defaultSizeMap for the values
			loadEBIMap();

			if ( defaultSizeMap == null )
				return;

			try
			{
				String defaultWidthPropertyValue = (String)defaultSizeMap.get(beanClass.getName() + "_defaultWidth" );
				if ( defaultWidthPropertyValue != null )
					beanDescriptor.setValue( DEFAULT_WIDTH, new Integer( defaultWidthPropertyValue ) );
			}
			catch ( Exception e1 )
			{
				e1.printStackTrace();
			}

			try
			{
				String defaultHeightPropertyValue = (String)defaultSizeMap.get(beanClass.getName() + "_defaultHeight" );
				if ( defaultHeightPropertyValue != null )
					beanDescriptor.setValue( DEFAULT_HEIGHT, new Integer( defaultHeightPropertyValue ) );
			}
			catch ( Exception e2 )
			{
				e2.printStackTrace();
			}
		}
	}

	protected static void loadEBIMap()
	{
		if ( extendedBeanInfoMap == null ||
			 defaultSizeMap == null )
		{
			try
			{
				Properties ebiPropertiesMap = new Properties();
				java.io.InputStream inputStream = com.sas.security.BaseSecuritySupport.securitySupport.getResourceAsStream( com.sas.beans.Introspector.class, "ExtendedBeanInfoMap.properties" );
				ebiPropertiesMap.load( inputStream );
				defaultSizeMap = new HashMap();
				extendedBeanInfoMap = parseEBIMap( ebiPropertiesMap );
			}
			catch( java.io.IOException e )
			{
				System.out.println( " IOException " + e );
				e.printStackTrace();
			}
		}
	}

	// Format of properties file:
	//  ExtendedBeanInfo for properties:
	//   <class name>.class.<property>.<property attribute>=<value>
	//
	//  other ExtendedBeanInfo values:
	//   <class name>.class.default<Input|Output>PropertyName=<property name>
	//  
	//  default size values:
	//   <class name>.class.default<Width|Height>=<value>
	//
	// Format of extendedBeanInfoMap:
	//  key: className
	//  value: ExtendedBeanInfo for class of type className
	//
	// Format of defaultSizeMap:
	//  key: className+(_defaultHeight|_defaultWidth)
	//  value: defaultHeight or defaultWidth for objects of the className type
	private static Map parseEBIMap( Map ebiPropMap )
	{
		// Possible metadata values
		Map metaDataMap = new HashMap();
		metaDataMap.put( "NAME", "Name");
		metaDataMap.put( "DEFAULT_VALUE", "Default value");
		metaDataMap.put( "SHORT_DESCRIPTION", "Short description");
		metaDataMap.put( "DESCRIPTION", "Description");
		metaDataMap.put( "READ_METHOD", "Read method");
		metaDataMap.put( "WRITE_METHOD", "Write method");
		metaDataMap.put( "BOUND", "Bound");
		metaDataMap.put( "LINKABLE", "Linkable");
		metaDataMap.put( "COMPARE_VALUE", "Compare value");
		metaDataMap.put( "PERSIST", "Persist");
		metaDataMap.put( "FIELD_NAME", "Field name");
		metaDataMap.put( "CONSTRAINED", "Constrained");
		metaDataMap.put( "PROPERTY_EDITOR_CLASS", "Property editor class");
		metaDataMap.put( "EXPERT", "Expert");
		metaDataMap.put( "HIDDEN", "Hidden");
		metaDataMap.put( "MASKED", "Masked");
		metaDataMap.put( "SEND_EVENT", "Send event");
		metaDataMap.put( "STATIC", "Static");
		metaDataMap.put( "CATEGORY", "Category");

		// classMap contents:
		//  key: className
		//  value: Map of ExtendedBeanInfo for objects of type className
		//         key: _ExtendedBeanInfo
		//          value: ExtendedBeanInfo object initialized with correct default data
	    //         key: property name
	    //          value: Map containing metadata 
		//		      key: metadata name
		//            value: metadata value for specified property name
		Map classMap = new HashMap();
		String classToken = ".class";
		Set keySet = ebiPropMap.keySet();
		Iterator iterator = keySet.iterator();
		while ( iterator.hasNext() )
		{
			String key = (String)iterator.next();
			Object value = ebiPropMap.get( key );

			try
			{
				// Check if correct token used to identify classname.
				int index = key.indexOf( classToken );
				if ( index == -1 )
					continue;

				String className = key.substring( 0, index );
				String remainingName = key.substring( index + classToken.length() + 1 );
				index = remainingName.indexOf( '.' );

				// Fill defaultSizeMap
				if ( remainingName.equals("defaultHeight") ||
					 remainingName.equals("defaultWidth") )
				{
					String sizeKey = className + "_" + remainingName;
					defaultSizeMap.put( sizeKey, value );
					continue;
				}

				// Create Map containing ExtendedBeanInfo and propertyMetadata information
				Map classPropMap = (Map)classMap.get( className );
				if ( classPropMap == null )
				{
					classPropMap = new HashMap();
					classMap.put( className, classPropMap );
				}

				// Create ExtendedBeanInfo and
				// set the default properties on it
				if ( index == -1 )
				{
					ExtendedBeanInfo ebi = (ExtendedBeanInfo)classPropMap.get( "_ExtendedBeanInfo" );
					if ( ebi == null )
					{
						ebi = new ExtendedBeanInfo();
						classPropMap.put( "_ExtendedBeanInfo", ebi );
					}

					if ( remainingName.equals( "defaultInputPropertyName" ) )
						ebi.defaultInputPropertyName = (String)value;
					else if ( remainingName.equals( "defaultOutputPropertyName" ) )
						ebi.defaultOutputPropertyName = (String)value;
					else if ( remainingName.equals( "shortDescription" ) )
						ebi.shortDescription = (String)value;
				}
				// Create and use a map of the properties and their metadata information
				else
				{
					String propName = remainingName.substring( 0, index );
					String attribName = remainingName.substring( index + 1 );
					Map propNameMap = (Map)classPropMap.get( propName );
					if ( propNameMap == null )
					{
						propNameMap = new HashMap();
						propNameMap.put( "Name", propName );
						classPropMap.put( propName, propNameMap );
					}

					String metaDataName = (String)metaDataMap.get( attribName );
					if ( metaDataName != null )
						propNameMap.put( metaDataName, value );
				}
			}
			catch( Exception e )
			{
				String msgTemplate = RB.getStringResource( RB_KEY, "errorParsingKey.txt" );
				System.out.println( MessageFormat.format( msgTemplate, new Object[] {key, value} ) );
			}
		}

		Map ebiMap = new HashMap();
		Set classNameSet = classMap.keySet();
		iterator = classNameSet.iterator();
		while ( iterator.hasNext() )
		{
			String className = (String)iterator.next();
			Map classPropMap = (Map)classMap.get( className );

			ExtendedBeanInfo ebi = (ExtendedBeanInfo)classPropMap.get( "_ExtendedBeanInfo" );
			if ( ebi == null )
				ebi = new ExtendedBeanInfo();
			else
				classPropMap.remove( "_ExtendedBeanInfo" );

			ebi.propertyMetadata = new String[classPropMap.size()][][];
			Set classPropSet = classPropMap.keySet();
			Iterator classPropSetIterator = classPropSet.iterator();
			// Create propertyMetadata String[][][]
			for( int i = 0; classPropSetIterator.hasNext(); i++ )
			{
				String propName = (String)classPropSetIterator.next();
				Map propNameMap = (Map)classPropMap.get( propName );

				String [][]metaData = new String[propNameMap.size()][2];
				Set metaDataSet = propNameMap.keySet();
				Iterator metaDataIterator = metaDataSet.iterator();
				for ( int j = 0; metaDataIterator.hasNext(); j++ )
				{
					metaData[j][0] = (String)metaDataIterator.next(); 			// Attribute Name
					metaData[j][1] = (String)propNameMap.get( metaData[j][0] );	// Property Name
				}

				ebi.propertyMetadata[i] = metaData;
			}

			ebiMap.put( className, ebi );
		}

		return ebiMap;
	}

	protected static void getExtendedBeanInfo( Class beanClass, ExtendedBeanInfo compositeEBI )
	{
		if ( beanClass.getSuperclass() != null )
		{
			getExtendedBeanInfo( beanClass.getSuperclass(), compositeEBI );
		}

		// Ask the Bean for an ExtendedBeanInfo object using getExtendedBeanInfo.
		// If calling getExtendedBeanInfo throws an exception, catch it and do
		// not try to use extended bean info.

		ExtendedBeanInfo ebi = null;
		try
		{
			Class [] paramTypes = new Class[0];
			java.lang.reflect.Method gebiMethod = com.sas.security.BaseSecuritySupport.securitySupport.getMethod( beanClass, "getExtendedBeanInfo", paramTypes );
			ebi = (ExtendedBeanInfo) com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( gebiMethod, null, null );
		} // try
		catch ( Exception e)
		{
			//			System.out.println( e.toString() );

			// XXX: We might want to be more specific about the exceptions we catch here...
			// Don't do anything here.
		} // catch

		// Use ExtendedBeanInfoMap.properties to load ExtendedBeanInfo for classes that 
		// have no default ExtendedBeanInfo.
		if ( ebi == null )
		{
			loadEBIMap();
			if ( extendedBeanInfoMap != null )
				ebi = (ExtendedBeanInfo)extendedBeanInfoMap.get( beanClass.getName() );
		}

		if ( ebi != null )
			compositeEBI.composite( ebi );


		// See the the class being introspected is a ROCF class.  If it is,
		// find the interface class for the ROCF class and get it's
		// ExtendedBeanInfo and composite it with the already found EBI.

		Class introspectInterfaceForBeanInfoClass = GlobalInfoTable._loadClass( "com.sas.beans.IntrospectInterfaceForBeanInfo" );
		if ( introspectInterfaceForBeanInfoClass.isAssignableFrom( beanClass )  ||
			beanClass.isInterface() )
		{
			//			System.out.println( "Introspecting Remote Class : " + beanClass.getName() );

			// Break apart the class name and package name

			String className = getRemoteClassBaseName( beanClass.getName() );

			try
			{
				Class interfaceBeanInfoClass = GlobalInfoTable.loadClass( className + "BeanInfo" );

				//				System.out.println( "Found remote BeanInfo class : " + className + "BeanInfo" );
				getExtendedBeanInfo( interfaceBeanInfoClass, compositeEBI );
			}
			catch (ClassNotFoundException e)
			{
				//				System.out.println( "Caught exception looking for remote bean info for : " + className + "BeanInfo" );
				// Ignore exception
			}
		} // if
		//		System.out.println( "compositeBeanInfo shortDescription : " + compositeEBI.shortDescription );
		//		Thread.dumpStack();
	}

	protected static String getRemoteClassBaseName( String className )
	{
		String packageName = "";
		if ( className.indexOf( '.' ) != -1 )
		{
			// Include the dot on the end of the package name
			packageName = className.substring( 0, className.lastIndexOf( '.' ) + 1 );
			className = className.substring( className.lastIndexOf( '.' ) + 1 );
		}

		// XXX: Strip off the ROCF classname prefixs

		if ( className.startsWith( "Final"/*I18nOK:EMS*/ ) )
			className = className.substring( 5 );
		if ( className.startsWith( "_rprx" ) )
			className = className.substring( 5 );
		if ( className.startsWith( "J2" ) )
			className = className.substring( 2 );
		if ( className.startsWith( "Remote"/*I18nOK:EMS*/ ) )
			className = className.substring( 6 );
		if ( className.startsWith( "DesignTime" ) )
			className = className.substring( 10 );

		className = packageName + className/*I18nOK:COS*/;
		if ( className.endsWith( "Interface"/*I18nOK:EMS*/ ) == false )
		{
			className += "Interface"/*I18nOK:LINE*/;
		}

		return className;
	}

	protected static void annotateProperties( Class beanClass, java.beans.PropertyDescriptor[] properties, ExtendedBeanInfo ebi ) throws IntrospectionException
	{
		Hashtable metadataTable = PropertyMetadata.create( ebi.propertyMetadata );

		// Annotate the non-static properties

		for ( int i = 0; i < properties.length; i++ )
		{
			if ( metadataTable.containsKey( properties[i].getName() ) )
			{
				PropertyMetadata metadata = (PropertyMetadata) metadataTable.get( properties[i].getName() );

				// Annotate the PropertyDescriptor with information from the PropertyMetadata object

				com.sas.beans.PropertyDescriptor pd = null;

				if ( properties[i] instanceof com.sas.beans.PropertyDescriptor )
				{
					pd = (com.sas.beans.PropertyDescriptor) properties[i];
				} // if
				else
				{
					pd = new com.sas.beans.PropertyDescriptor( properties[i] );
					properties[i] = pd;
				}

				// Convert the metadata to PropertyDescriptor information

				if ( metadata.defaultValue_set )
					pd.setDefaultValue( metadata.defaultValue );
				if ( metadata.linkable_set )
					pd.setLinkable( metadata.linkable );
				if ( metadata.description_set )
					pd.setDescription( metadata.description );
				if ( metadata.persist_set )
					pd.setPersisted( metadata.persist );

				// Override introspected items with metadata values if specified

				if ( metadata.shortdesc_set )
					pd.setShortDescription( metadata.shortDescription );
				if ( metadata.bound_set )
					pd.setBound( metadata.bound );
				if ( metadata.constrained_set )
					pd.setConstrained( metadata.constrained );
				if ( metadata.expert_set )
					pd.setExpert( metadata.expert );
				if ( metadata.hidden_set )
					pd.setHidden( metadata.hidden );
				if ( metadata.masked_set )
					pd.setMasked( metadata.masked );

				// Set the property editor class if specified in the metadata

				pd.setPropertyEditorClass( metadata.propertyEditorClass );

				// Set the default value for the default value

				if ( metadata.defaultValue.length() == 0 )
				{
					Class propType = pd.getPropertyType();

					if ( propType == java.lang.String.class )
					{
						pd.setDefaultValue( "\"\"" );
					} // if
					else if ( propType == boolean.class )
					{
						pd.setDefaultValue( "false" );
					} // else if
					else if ( propType == int.class  ||
						propType == short.class ||
						propType == long.class  ||
						propType == double.class  ||
						propType == float.class  ||
						propType == byte.class  ||
						propType == char.class  )
					{
						pd.setDefaultValue( "0" );
					} // else if
					else
					{
						// The type must be an Object

						pd.setDefaultValue( "null" );
					} // else

				} // if

				// Change the read method if metadata was specified

				if ( metadata.readMethodName != null )
				{
					try
					{
						Method getMethod = com.sas.security.BaseSecuritySupport.securitySupport.getMethod( beanClass, metadata.readMethodName, null );
						pd.setReadMethod( getMethod );
					}
					catch ( NoSuchMethodException e )
					{
						String message = RB.getStringResource( RB_KEY, "annotateProperties.ex.txt" );
						System.out.println( MessageFormat.format( message, new Object[] {e.getMessage()} ) );
						e.printStackTrace();
					}
				} // if

				// Change the write method if metadata was specified

				if ( metadata.writeMethodName != null )
				{
					Class[] params = new Class[1];
					params[0] = pd.getPropertyType();

					try
					{
						Method setMethod = com.sas.security.BaseSecuritySupport.securitySupport.getMethod( beanClass, metadata.writeMethodName, params );
						pd.setWriteMethod( setMethod );
					}
					catch ( NoSuchMethodException e )
					{
						String message = RB.getStringResource( RB_KEY, "annotateProperties.ex.txt" );
						System.out.println( MessageFormat.format( message, new Object[] {e.getMessage()} ) );
						e.printStackTrace();
					}
				} // if
			} // if
		} // for

		// Hook up PropertyEditors for String[] and int[]

		String[] stringArray = new String[1];
		int[] intArray = new int[1];
		for ( int i = 0; i < properties.length; i++ )
		{
			if ( properties[i].getPropertyEditorClass() == null )
			{
				if ( properties[i].getPropertyType() == stringArray.getClass() )
				{
					//					System.out.println("String array property found: " + properties[i].getName() );

					com.sas.beans.PropertyDescriptor pd = null;

					if ( properties[i] instanceof com.sas.beans.PropertyDescriptor )
					{
						pd = (com.sas.beans.PropertyDescriptor) properties[i];
					} // if
					else
					{
						pd = new com.sas.beans.PropertyDescriptor( properties[i] );
						properties[i] = pd;
					}

					pd.setPropertyEditorClass( "com.sas.beans.editors.StringArrayEditor" );
				} // if
				else if ( properties[i].getPropertyType() == intArray.getClass() )
				{
					//					System.out.println("int array property found: " + properties[i].getName() );

					com.sas.beans.PropertyDescriptor pd = null;

					if ( properties[i] instanceof com.sas.beans.PropertyDescriptor )
					{
						pd = (com.sas.beans.PropertyDescriptor) properties[i];
					} // if
					else
					{
						pd = new com.sas.beans.PropertyDescriptor( properties[i] );
						properties[i] = pd;
					}

					pd.setPropertyEditorClass( "com.sas.beans.editors._IntArrayEditor" );
				} // else if
			} // if
		} // for
	}

	protected static PropertyDescriptor[] introspectStaticProperties( Class beanClass, java.beans.BeanInfo beanInfo ) throws IntrospectionException
	{
		boolean propertyChangeSource = false;

		// properties maps from String names to PropertyDescriptors
		java.util.Hashtable properties = new java.util.Hashtable();

		// Apply some reflection to the current class.

		// First get an array of all the beans methods at this level
		Method methodList[] = com.sas.security.BaseSecuritySupport.securitySupport.getDeclaredMethods( beanClass );

		// Now analyze each method.
		for (int i = 0; i < methodList.length; i++)
		{
			Method method = methodList[i];
			int modifier = com.sas.security.BaseSecuritySupport.securitySupport.getMethodModifiers( method );

			// skip non-public and non-static methods.
			if ( Modifier.isPublic(modifier) == false  ||
				Modifier.isStatic(modifier) == false )
			{
				continue;
			}

			String name = com.sas.security.BaseSecuritySupport.securitySupport.getMethodName( method );
			Class argTypes[] = com.sas.security.BaseSecuritySupport.securitySupport.getMethodParameterTypes( method );
			Class resultType = com.sas.security.BaseSecuritySupport.securitySupport.getMethodReturnType( method );

			PropertyDescriptor pd = null;

			boolean getMethod = false;
			boolean setMethod = false;
			boolean isMethod = false;
			if ( name == null )
				continue;
			else if ( name.length() > 3 )
			{
				if ( name.startsWith("get") )
					getMethod = true;
				else if ( name.startsWith("set") )
					setMethod = true;
			}
			else if ( name.startsWith("is") &&
				name.length() > 2 )
				isMethod = true;

			if ( !getMethod && !setMethod && !isMethod )
				continue;

			if ( getMethod && argTypes.length == 0 )
			{
				// Simple getter
				pd = new com.sas.beans.PropertyDescriptor(decapitalize(name.substring(3)), method, null);
			}
			else if ( getMethod &&
				argTypes.length == 1  &&
				argTypes[0] == Integer.TYPE)
			{
				pd = new IndexedPropertyDescriptor( decapitalize(name.substring(3)),
					null, null,
					method,	null);
			}
			else if (setMethod && argTypes.length == 1 &&
				resultType == Void.TYPE)
			{
				// Simple setter
				pd = new com.sas.beans.PropertyDescriptor(decapitalize(name.substring(3)),
					null, method);

				if (throwsException(method, PropertyVetoException.class))
				{
					pd.setConstrained(true);
				}
			}
			else if (setMethod  &&
				argTypes.length == 2  &&
				argTypes[0] == Integer.TYPE)
			{
				pd = new IndexedPropertyDescriptor(
					decapitalize(name.substring(3)),
					null, null,
					null, method);
				if (throwsException(method, PropertyVetoException.class))
				{
					pd.setConstrained(true);
				}
			}
			else if (isMethod &&
				argTypes.length == 0  &&
				resultType == Boolean.TYPE)
			{
				// Boolean getter
				pd = new com.sas.beans.PropertyDescriptor(decapitalize(name.substring(2)), method, null);
			} // else if

			if (pd != null)
			{
				// If this class or one of its base classes is a PropertyChange
				// source, then we assume that any properties we discover are "bound".
				if (propertyChangeSource)
				{
					pd.setBound(true);
				}
				addProperty(properties, pd);
			}
		} // for

		// Allocate and populate the result array.
		PropertyDescriptor result[] = new PropertyDescriptor[properties.size()];
		java.util.Enumeration elements = properties.elements();
		for (int i = 0; i < result.length; i++)
		{
			result[i] = (PropertyDescriptor)elements.nextElement();
		}

		return result;
	}

	static void addProperty(Hashtable properties, PropertyDescriptor pd) throws IntrospectionException
	{
		String name = pd.getName();
		PropertyDescriptor old = (PropertyDescriptor)properties.get(name);
		if (old == null)
		{
			properties.put(name, pd);
			return;
		}

		// If the property type has changed, use the new descriptor.
		Class opd = old.getPropertyType();
		Class npd = pd.getPropertyType();
		if (opd != null && npd != null && opd != npd)
		{
			properties.put(name, pd);
			return;
		}

		PropertyDescriptor composite;
		if (old instanceof IndexedPropertyDescriptor  ||  pd instanceof IndexedPropertyDescriptor)
		{
			//System.out.println("old="+old);
			//System.out.println("pd=" +pd);

			// XXX: Needs work

			//			composite = null;
			Method readMethod = old.getReadMethod();
			Method writeMethod = old.getWriteMethod();

			if ( pd.getReadMethod() != null )
				readMethod = pd.getReadMethod();

			if ( pd.getWriteMethod() != null )
				writeMethod = pd.getWriteMethod();

			composite = new com.sas.beans.IndexedPropertyDescriptor(old.getName(), readMethod, writeMethod);
		}
		else
		{
			Method readMethod = old.getReadMethod();
			Method writeMethod = old.getWriteMethod();

			if ( pd.getReadMethod() != null )
				readMethod = pd.getReadMethod();

			if ( pd.getWriteMethod() != null )
				writeMethod = pd.getWriteMethod();

			composite = new com.sas.beans.PropertyDescriptor( old.getName(), readMethod, writeMethod );
		}
		properties.put(name, composite);
	} // addProperty


	// (The following should have been a public method in
	// java.beans.Introspector but lack of insight made it private.)

	/**
	* Validate that the stop class is an ancestor class of the bean class.
	* @param beanClass the bean class from which we look up properties.
	* @param stopClass Do not return properties defined at or above this class.
	*                  Pass null to not exclude any properties.
	* @exception java.beans.IntrospectionException if stopClass is not
	*            an ancestor of beanClass.
	*/
	protected static void validateStopClass(Class beanClass, Class stopClass)
		throws IntrospectionException
	{
		if (stopClass == null)
			return;
		String stopClassName = stopClass.getName();
		for (Class c = beanClass.getSuperclass(); c != null; c = c.getSuperclass())
			// java.beans.Introspector uses name equality rather than ==
			// on the class references. I'm not sure why (perhaps since
			// different jar files can have different class loaders? trouble!)
			// Anyway, to be safe, we will do the same. We'll do it a bit more
			// more efficiently, tho.
			if (stopClassName.equals(c.getName()))
				return;

		String message = RB.getStringResource( RB_KEY, "validateStopClass.ex.txt" );
		throw new IntrospectionException( MessageFormat.format( message, new Object[] {stopClass.getName(), beanClass.getName()} ) );
	}

	/**
	 * Return true iff the given method throws the given exception.
	 */
	private static boolean throwsException(Method method, Class exception)
	{
		Class exs[] = com.sas.security.BaseSecuritySupport.securitySupport.getMethodExceptionTypes( method );
		for (int i = 0; i < exs.length; i++)
		{
			if (exs[i] == exception)
			{
				return true;
			}
		}

		return false;
	}

	protected static int lookupPropertyIndex( PropertyDescriptor[] properties, int defaultIndex, String propertyName )
	{
		for ( int i = 0; i < properties.length; i++ )
		{
			if ( properties[i].getName().equals( propertyName ) )
			{
				return i;
			} // if
		} // for

		return defaultIndex;
	}


	/**
	 * This is a helper method which returns a property descriptor given a BeanInfo and a property name.
	 * null is returned if the specified property is not found.
	 */
	public static java.beans.PropertyDescriptor getPropertyDescriptor( String propertyName, java.beans.BeanInfo beanInfo )
	{
		java.beans.PropertyDescriptor[] properties = beanInfo.getPropertyDescriptors();

		for ( int i = 0; i < properties.length; i++ )
		{
			if ( properties[i].getName().equals( propertyName ) )
			{
				return properties[i];
			} // if
		} // for

		return null;
	}

	//*************************************************************************
	// Utility Methods
	//*************************************************************************

	/**
	 * getInterfaces returns a list of interfaces supported the specified class
	 * and all of it's superclasses.  The method is used instead of java.lang.Class.getInterfaces()
	 * since there is a discrepency in the behavior of Sun's and MS's versions of the
	 * methods.
	 * @param objClass The class to return a list of interfaces for.
	 * @return An array of interface references.
	 */
	public static Class[] getInterfaces( Class objClass )
	{
		// Count the number of interfaces that we will return

		int numInterfaces = 0;
		Class tmpClass = objClass;
		while ( tmpClass != null )
		{
			numInterfaces += tmpClass.getInterfaces().length;
			tmpClass = tmpClass.getSuperclass();
		} // while

		// Allocate an array for the interfaces

		Class[] interfaces = new Class[ numInterfaces ];

		// Add the interfaces to the array

		int index = 0;
		tmpClass = objClass;
		while ( tmpClass != null )
		{
			Class[] oneClassIFList = tmpClass.getInterfaces();

			for ( int i = 0; i < oneClassIFList.length; i++ )
			{
				interfaces[index++] = oneClassIFList[i];
			} // for

			tmpClass = tmpClass.getSuperclass();
		} // while

		return interfaces;
	} // getInterfaces

}
