package com.sas.beans;

import java.beans.PropertyChangeListener;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Collection;
import java.util.EventListener;
import java.util.Collections;
import java.util.List;
import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;


/**
 * This subclass of java.beans.PropertyChangeSupport is identical in functionality.  It adds a method to 
 * process a PropertyChangeEvent without checking the old and new values.  It also adds a method to get
 * all of the listeners that have been added to the PropertyChangeSupport.  The overridden methods were
 * required because the java.beans.PropertyChangeSupport doesn't provide any accessors for its instance data.
 *
 */
public class PropertyChangeSupport 
	extends java.beans.PropertyChangeSupport
{

    /**
     * "listeners" lists all the generic listeners.
     *
     *  This is transient - its state is written in the writeObject method.
     */
    transient protected java.util.List listeners;

    /** 
     * Map for managing listeners for specific properties.
     * Maps property names to PropertyChangeSupport objects.
     * @serial 
     */
    protected java.util.Map children;

    /** 
     * The object to be provided as the "source" for any generated events.
     * @serial
     */
    protected Object source;

    /**
     * Serialization version ID
     */
	static final long serialVersionUID = -6042039294865660725L;

    /**
     * Constructs a <code>PropertyChangeSupport</code> object.
     *
     * @param sourceBean  The bean to be given as the source for any events.
     */
    public PropertyChangeSupport( Object sourceBean ) 
	{
		super( sourceBean );
		if ( sourceBean == null )
		    throw new IllegalArgumentException();

		source = sourceBean;
    }

    /**
     * Add a PropertyChangeListener to the listener list.
     * The listener is registered for all properties.
     *
     * @param listener  The PropertyChangeListener to be added
     */
    public synchronized void addPropertyChangeListener( PropertyChangeListener listener ) 
	{
		if ( listeners == null )
		    listeners = new java.util.ArrayList();

		listeners.add(listener);
    }

    /**
     * Remove a PropertyChangeListener from the listener list.
     * This removes a PropertyChangeListener that was registered
     * for all properties.
     *
     * @param listener  The PropertyChangeListener to be removed
     */
    public synchronized void removePropertyChangeListener( PropertyChangeListener listener )
	{
		if ( listeners == null )
		    return;

		listeners.remove(listener);
    }

    /**
     * Add a PropertyChangeListener for a specific property.  The listener
     * will be invoked only when a call on firePropertyChange names that
     * specific property.
     *
     * @param propertyName  The name of the property to listen on.
     * @param listener  The PropertyChangeListener to be added
     */
    public synchronized void addPropertyChangeListener( String propertyName, PropertyChangeListener listener )
	{
		if ( children == null )
		    children = new java.util.HashMap();

		PropertyChangeSupport child = (PropertyChangeSupport)children.get(propertyName);
		if ( child == null ) 
		{
		    child = new PropertyChangeSupport(source);
		    children.put(propertyName, child);
		}

		child.addPropertyChangeListener(listener);
    }

    /**
     * Remove a PropertyChangeListener for a specific property.
     *
     * @param propertyName  The name of the property that was listened on.
     * @param listener  The PropertyChangeListener to be removed
     */
    public synchronized void removePropertyChangeListener( String propertyName, PropertyChangeListener listener) 
	{
		if ( children == null )
		    return;

		PropertyChangeSupport child = (PropertyChangeSupport)children.get(propertyName);
		if ( child == null )
		    return;

		child.removePropertyChangeListener(listener);
    }

    /**
     * Report a bound property update to any registered listeners.
     * No event is fired if old and new are equal and non-null.
     *
     * @param propertyName  The programmatic name of the property
     *		that was changed.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    public void firePropertyChange(String propertyName, Object oldValue, Object newValue) 
	{
		if ( oldValue != null && 
			 newValue != null && 
			 oldValue.equals(newValue) ) 
	    	return;

	    java.beans.PropertyChangeEvent evt = new java.beans.PropertyChangeEvent(source, propertyName, oldValue, newValue);
		processPropertyChange( evt );
    }

    /**
     * Fire an existing PropertyChangeEvent to any registered listeners.
     * No event is fired if the given event's old and new values are
     * equal and non-null.
     * @param evt  The PropertyChangeEvent object.
     */
    public void firePropertyChange(java.beans.PropertyChangeEvent evt) 
	{
		Object oldValue = evt.getOldValue();
		Object newValue = evt.getNewValue();
		if ( oldValue != null && 
			 newValue != null && 
			 oldValue.equals(newValue) )
		    return;

		processPropertyChange( evt );
	}

    /**
     * Fire an existing PropertyChangeEvent to any registered listeners.
	 * The event is fired regardless of what the new and old values are.
     * @param evt  The PropertyChangeEvent object.
     */
    public void processPropertyChange( java.beans.PropertyChangeEvent evt ) 
	{
		java.util.List targets = null;
		PropertyChangeSupport child = null;
		synchronized (this) 
		{
		    if ( listeners != null )
			{
		        targets = new java.util.ArrayList( listeners.size() );
				targets.addAll( listeners );
			}

			String propertyName = evt.getPropertyName();
			if ( children != null && propertyName != null )
				child = (PropertyChangeSupport)children.get(propertyName);
		}

		if ( targets != null ) 
		{
		    for ( int i = 0; i < targets.size(); i++ ) 
			{
		        PropertyChangeListener target = (PropertyChangeListener)targets.get(i);
		        target.propertyChange(evt);
		    }
		}

		if ( child != null )
		    child.processPropertyChange(evt);
    }

    /**
     * Check if there are any listeners for a specific property.
     *
     * @param propertyName  the property name.
     * @return true if there are ore or more listeners for the given property
     */
    public synchronized boolean hasListeners( String propertyName ) 
	{
	    // there is a generic listener
		if ( listeners != null && !listeners.isEmpty() )
		    return true;

		if (children != null) 
		{
		    PropertyChangeSupport child = (PropertyChangeSupport)children.get(propertyName);
		    if ( child != null )
//				return !child.listeners.isEmpty();
				return child.hasListeners( propertyName );
		}

		return false;
    }

	public synchronized EventListener[] getListeners()
	{
		if (( listeners == null || listeners.isEmpty() ) &&
			( children == null || children.isEmpty() ))
			return new EventListener[0];

		int listenerSize = 0;
		if ( listeners != null )
			listenerSize = listeners.size();

		int childrenSize = 0;
		if ( children != null )
			childrenSize = children.size();

		List listenerList = new ArrayList( listenerSize + childrenSize );
		if ( listeners != null && !listeners.isEmpty() )
			listenerList.addAll( listeners );

		if ( children != null && !children.isEmpty() )
		{
			Collection childrenList = children.values();
			java.util.Iterator childrenIterator = children.values().iterator();
			while ( childrenIterator.hasNext() )
			{
				PropertyChangeSupport child = (PropertyChangeSupport)childrenIterator.next();
				listenerList.addAll( Arrays.asList( child.getListeners() ) );
			}
		}

		EventListener[] listenerArray = new EventListener[listenerList.size()];
		listeners.toArray( listenerArray );

		return listenerArray;
	}

    /**
     * @serialData Null terminated list of <code>PropertyChangeListeners</code>.
     * <p>
     * At serialization time we skip non-serializable listeners and
     * only serialize the serializable listeners.
     *
     */
    private void writeObject(ObjectOutputStream s) throws IOException 
	{
        s.defaultWriteObject();

		java.util.List v = null;
		synchronized (this) 
		{
		    if (listeners != null)
			{
		        v = new java.util.ArrayList();
				v.addAll( listeners );
			}
		}

		if (v != null) 
		{
		    for (int i = 0; i < v.size(); i++) 
			{
		        PropertyChangeListener l = (PropertyChangeListener)v.get(i);
		        if (l instanceof Serializable)
		            s.writeObject(l);
            }
        }

        s.writeObject(null);
    }


    private void readObject(ObjectInputStream s) throws ClassNotFoundException, IOException 
	{
        s.defaultReadObject();
      
        Object listenerOrNull;
        while ( null != (listenerOrNull = s.readObject()) ) 
		{
			addPropertyChangeListener((PropertyChangeListener)listenerOrNull);
        }
    }
}
