// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.beans;

import com.sas.GlobalInfoTable;
import com.sas.PublicClonable;

import java.beans.IntrospectionException;
import java.lang.reflect.Method;
import java.text.MessageFormat;

/**
 * Propertydescriptor extends java.beans.PropertyDescriptor to provide more information about
 * properties for use within the webAF environment.
 */
public class PropertyDescriptor extends java.beans.PropertyDescriptor implements PublicClonable
{
	/**
	 * The name of the property editor class.
	 */
	protected String propertyEditorClassName;

    /**
    * Whether the attribute can be linked with addLink().
    * Checked in addLink() to ensure the link is allowed
    * before actually setting it.
    */
    protected boolean linkable;

	/**
	* The default value of the attribute. (optional)<P>
	* The default property value is not used at this time.
	*/
	protected String defaultValue;

    /**
    * Whether to compare the property as a reference or compare the contents of the object.
    * compareValue is not used at this time.
    */
    protected boolean compareValue;

    /**
    * Whether to persist the value of the property when the component is persisted
    * persisted is not used at this time.
    */
    protected boolean persisted;

    /**
    * The name of the Java field used to hold the value of this property.
    * fieldName is not used at this time.
    */
    protected String fieldName;

	/**
	 * This is a description of the property.
	 */
	protected String description;

	/**
	 * Whether the property value should be displayed "masked" in the property sheet.  
	 * This is used for password type fields which want to obscure the display of the value.
	 */
	protected boolean masked;
		
	/**
	 * This is the category of the property.
	 * The category is currently unused.
	 */
	protected String category;


	/**
	 * This field is used by the overridden getReadMethod() and setReadMethod() methods.
	 */
	Method readMethod;

	/**
	 * This field is used by the overridden getWriteMethod() and setWriteMethod() methods.
	 */
	Method writeMethod;

	/**
	 * Constructor used to construct a com.sas.beans.PropertyDescriptor from either a
	 * java.beans.PropertyDescriptor or a com.sas.beans.PropertyDescriptor.
	 * @param property A java.beans.PropertyDescriptor to use to create this property descriptor.
	 * @exception java.beans.IntrospectionException Exception thrown by java.beans.PropertyDescriptor.
	 */
	public PropertyDescriptor( java.beans.PropertyDescriptor property ) throws IntrospectionException
	{
		super( property.getName(), property.getReadMethod(), property.getWriteMethod() );

		commonInit();

		// Record these values for later use since we override {get|set}{Read|Write}Method.

		setReadMethod( property.getReadMethod() );
		setWriteMethod( property.getWriteMethod() );

		// Copy com.sas.bean.PropertyDescriptor information if one is passed in.

		if ( property instanceof com.sas.beans.PropertyDescriptor )
		{
			com.sas.beans.PropertyDescriptor pd = (com.sas.beans.PropertyDescriptor) property;
			
			setDescription( pd.getDescription() );
			setLinkable( pd.isLinkable() );
			setDefaultValue( pd.getDefaultValue() );
			setComparesValue( pd.comparesValue() );
			setFieldName( pd.getFieldName() );
			setPersisted( pd.isPersisted() );
			setMasked( pd.isMasked() );
			setCategory( pd.getCategory() );
		} // if
	}

	/**
	 * Constructor to create a property descriptor based on the property name and the get and set methods.
	 * @exception java.beans.IntrospectionException Exception thrown by java.beans.PropertyDescriptor.
	 */
	public PropertyDescriptor( String propertyName, Method getter, Method setter ) throws IntrospectionException
	{
		super( propertyName, getter, setter );

		commonInit();

		// Record these values for later use since we override {get|set}{Read|Write}Method.

		setReadMethod( super.getReadMethod() );
		setWriteMethod( super.getWriteMethod() );
	}
	
	/**
	 * Constructor to create a PropertyDescriptor based on a property name and class.
	 * @exception java.beans.IntrospectionException Exception thrown by java.beans.PropertyDescriptor.
	 */
	public PropertyDescriptor( String propertyName, Class beanClass ) throws IntrospectionException
	{
		super( propertyName, beanClass );

		commonInit();

		// Record these values for later use since we override {get|set}{Read|Write}Method.

		setReadMethod( super.getReadMethod() );
		setWriteMethod( super.getWriteMethod() );
	}

	/**
	 * Constructor to create a PropertyDescriptor based on a property name, class, and accessor method names.
	 * @exception java.beans.IntrospectionException Exception thrown by java.beans.PropertyDescriptor.
	 */
	public PropertyDescriptor( String propertyName, Class beanClass, String getterName, String setterName ) throws IntrospectionException
	{
		super( propertyName, beanClass, getterName, setterName );

		commonInit();

		// Record these values for later use since we override {get|set}{Read|Write}Method.

		setReadMethod( super.getReadMethod() );
		setWriteMethod( super.getWriteMethod() );
	}

	/**
	 * Utility method to perform common operations for constructors.
	 */
	protected void commonInit()
	{
		linkable = true;
		compareValue = true;
	}

	/**
	 * Return the PropertyEditor class.  If the class is stored as a string, the class will be
	 * loaded an returned.  null will be returned if the class cannot be loaded or was not set.
	 * @return the PropertyEditor class or null.
	 */
	public Class getPropertyEditorClass()
	{
		if ( propertyEditorClassName == null )
		{
			return super.getPropertyEditorClass();
		}

		Class propertyEditorClass = null;
		try
		{
			propertyEditorClass = GlobalInfoTable.loadClass( propertyEditorClassName );
		}
		catch ( ClassNotFoundException cnfe )
		{
			String message = RB.getStringResource( "Introspector.annotateProperties.invalidPropertyEditorClass.ex.txt" );
//			com.sas.util.SystemLog.logWarning( MessageFormat.format( message, new Object[] {getName()} ) );

			// Fall through and return null
		}
		return propertyEditorClass;
	}

	/**
	 * Set the name of the PropertyEditor class.
	 * @param className The name of the PropertyEditor class.
	 */
	public void setPropertyEditorClass( String className )
	{
		propertyEditorClassName = className;
	}

	/**
	 * Set the PropertyEditor class.
	 * <P>
	 * This method exists so that calls to it will clear out the field used to store the name
	 * of the PropertyEditor class since calls to setPropertyEditorClass( Class ) override
	 * previous calls to setPropertyEditorClass( String ).
	 * @param propertyEditorClass The PropertyEditor class.
	 */
	public void setPropertyEditorClass( Class propertyEditorClass )
	{
		propertyEditorClassName = null;
		super.setPropertyEditorClass( propertyEditorClass );
	}

	/**
	 * Returns the read accessor method for the property.
	 */
	public Method getReadMethod()
	{
		return readMethod;
	}

	/**
	 * Sets the read accessor method for the property.
	 */
	public void setReadMethod( Method readMethod )
	{
		this.readMethod = readMethod;
		
		// TO-DO:(SPL): Note that we don't check/change the type or anything if the method does not take the
		// same type of param as the current type.  We should also check to make sure the method has
		// only one param and is void.
	}

	/**
	 * Returns the write accessor method for the property.
	 */
	public Method getWriteMethod()
	{
		return writeMethod;
	}

	/**
	 * Sets the write accessor method for the property.
	 */
	public void setWriteMethod( Method writeMethod )
	{
		this.writeMethod = writeMethod;

		// TO-DO:(SPL): Note that we don't check/change the type or anything if the method does not return the
		// same type of param as the current type.  We should also check to make sure the method has
		// only no params and returns the current type.
	}

	/**
	 * Returns the property description.
	 * @return the property description.
	 */
	public String getDescription()
	{
		return description;
	}

	/**
	 * Sets the property description.
	 * @param description the property description.
	 */
	public void setDescription( String description )
	{
		this.description = description;
	}

	/**
	 * isLinkable returns whether the property can be linked to.
	 * This is used by addLink() to check if a link is valid.
	 * @returns Whether the property can be linked to.
	 */
	public boolean isLinkable()
	{
		return linkable;
	}

	/**
	 * setLinkable() sets whether the property can be linked to.
	 * @param linkable Whether the property can be linked to.
	 */
	public void setLinkable( boolean linkable )
	{
		this.linkable = linkable;
	}


	/**
	 * getDefaultValue returns the default value for the property.
	 * The default value of the property is currently unused.
	 * @returns The default value for the property.
	 */
	public String getDefaultValue()
	{
		return defaultValue;
	}

	/**
	 * setDefaultValue sets the default value of the attribute.
	 * The default value of the property is currently unused.
	 * @param defaultValue The default value of the attribute.
	 */
	public void setDefaultValue( String defaultValue )
	{
		this.defaultValue = defaultValue;
	}

	/**
	 * comparesValue returns whether the property value is compared
	 * in the set accessor for the property.
	 * @returns Whether the property value is compared in the set accessor for the property.
	 */
	public boolean comparesValue()
	{
		return compareValue;
	}

	/**
	 * Sets whether the propert value is compared.
	 */
	public void setComparesValue( boolean compareValue )
	{
		this.compareValue = compareValue;
	}
	

	/**
	 * This gets the field associated with the property.
	 */
	public String getFieldName()
	{
		return fieldName;
	}

	/**
	 * This sets the field associated with the property.  The field will contain the value of the
	 * property and be of the same type as the property.
	 */
	public void setFieldName( String fieldName )
	{
		this.fieldName = fieldName;
	}


	/**
	 * Returns whether the property value should be persisted.
	 */
	public boolean isPersisted()
	{
		return persisted;
	}

	/**
	 * Sets whether the value of this property should be persisted.
	 */
	public void setPersisted( boolean persisted )
	{
		this.persisted = persisted;
	}


	/**
	 * Determines if the property is writable.
	 * @return true if the property is writable
	 */
	public boolean isWritable()
	{
		if ( getWriteMethod() != null )
			return true;
		else
			return false;
	}

	/**
	 * Gets whether the property value should be displayed "masked" in the property sheet.  
	 * This is used for password type fields which want to obscure the display of the value.
	 * @return true if the property value should be masked in a property sheet
	 */
	public boolean isMasked()
	{
		return masked;
	}

	/**
	 * Sets whether the property value should be displayed "masked" in the property sheet.  
	 * This is used for password type fields which want to obscure the display of the value.
	 * @param masked whether the property value should be masked in a property sheet
	 */
	public void setMasked( boolean masked )
	{
		this.masked = masked;
	}

	/**
	 * This gets the category for the property (currently unused).
	 * @return the category for the property
	 */
	public String getCategory()
	{
		return category;
	}

	/**
	 * This sets the category for the property (currently unused).  
	 * @param category the category for the property
	 */
	public void setCategory( String category )
	{
		this.category = category;
	}


	/**
	* Makes a copy of the PropertyDescriptor object.
	* @return A copy of the PropertyDescriptor object.
	*/
	public Object clone()
	{
        try
        {
			PropertyDescriptor clone = (PropertyDescriptor) super.clone();
			
			return clone;
        }
        catch (CloneNotSupportedException cnse)
        {
			// This should never happen

			System.err.println("PropertyDescriptor.clone(): " + cnse.getMessage() );
        }
        return null;
	} // clone

//DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG 
//	public void setBound( boolean bound )
//	{
//		System.out.println( this.getName() + " : b=" + bound ); 
//		super.setBound(bound);
//	}
//DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG 
}
