// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513
package com.sas.beans;

import com.sas.GlobalInfoTable;

/**
 * PropertyEditor is an abstract class which provides a default implementation
 * for the methods of the java.beans.PropertyEditor interface.  PropertyEditors
 * for webAF components should extend this class and override any methods
 * needed to get their desired behavior.
 */
public abstract class PropertyEditor 
	implements java.beans.PropertyEditor, 
			   java.beans.PropertyChangeListener,
			   com.sas.beans.ParentWindowRequestor
{
	protected PropertyChangeListenerList propertyChangeListeners = new PropertyChangeListenerList();
	protected Object value;
	protected java.awt.Window parentWindow;

	/**
	 * Sets the Object to be editted by this PropertyEditor.  Subclasses
	 * may override this method and perform any processing needed when a
	 * new Object is set.
	 * <P>
	 * Sublasses which are editing mutable objects should call super.setValue()
	 * if they override this method to have the value updated properly.
	 */
	public void setValue(java.lang.Object value)
	{
		// Stop listening to PropertyChangeEvents from the old value
		
		if ( this.value != null )
		{
			com.sas.beans.Util.removePropertyChangeListener( this.value, this );
		} // if
		
		this.value = value;

		// Start listening to PropertyChangeEvents from the new value

		if ( value != null )
		{
			com.sas.beans.Util.addPropertyChangeListener( value, this );
		} // if

		// Tell everyone that the value changed
		
		firePropertyChange();
	}
	
	/**
	 * Returns the Object currently being editted by this PropertyEditor.
	 * null may be returned.
	 */
	public java.lang.Object getValue()
	{
		return value;
	}
	
	/**
	 * Subclasses should override this method if they support painting the
	 * property value.
	 * @see java.beans.PropertyEditor#isPaintable
	 * @see java.beans.PropertyEditor#paintValue
	 */
	public boolean isPaintable()
	{
		return false;
	}
	
	/**
	 * Subclasses should override this method if they support painting the
	 * property value.
	 * @see java.beans.PropertyEditor#isPaintable
	 * @see java.beans.PropertyEditor#paintValue
	 */
	public void paintValue(java.awt.Graphics g, java.awt.Rectangle rect )
	{
	}
	
	/**
	 * Subclasses should override this method and return the appropriate
	 * snippet of Java code.
	 * @see java.beans.PropertyEditor#getJavaInitializationString
	 */
	public java.lang.String getJavaInitializationString()
	{
		return "";
	}
	
	/**
	 * Subclasses which support getting and setting the property value
	 * as text should override this method and return the value represented
	 * as a String.
	 * @see java.beans.PropertyEditor#getAsText
	 */
	public java.lang.String getAsText()
	{
		return null;
	}
	
	/**
	 * Subclasses which support getting and setting the property value
	 * as text should override this method and set the value based on the
	 * String passed in.
	 * <P>
	 * Subclasses may need to call firePropertyChange() from the setAsText
	 * method if they are not propogating PropertyChangeEvents from the
	 * property "Value".
	 * @see java.beans.PropertyEditor#setAsText
	 */
	public void setAsText(java.lang.String text)
	{
	}
	
	/**
	 * Subclasses which have a small finite set of values for the property
	 * should override this method and return a list of strings representing
	 * the values.
	 * @see java.beans.PropertyEditor#getTags
	 */
	public java.lang.String[] getTags()
	{
		return null;
	}

	/**
	 * Look for a Customizer associated with the PropertyEditor.  If the 
	 * PropertyEditor is called <<ComponentName>>Editor, the customizer
	 * should be called <<ComponentName>>Customizer.  If not, the subclass
	 * of PropertyEditor should override this method.
	 */
	protected Class findCustomEditorClass()
	{
		Class customizerClass = null;
		if ( value != null )
			customizerClass = com.sas.beans.BaseCustomizerUtil.getCustomizerClass( value );

		if ( customizerClass == null )
		{
			// No value determine editor based on editor name.
			// Determine if a custom editor is available.
			// We assume the custom editor is called <<ComponentName>>Customizer.		
			String className = getClass().getName();

			if ( className.endsWith( "Editor"/*I18nOK:EMS*/ ) )
			{
				String withoutEditor = className.substring(0, className.length() - 6 );
				String customizerName = withoutEditor + "Customizer"/*I18nOK:LINE*/;
				try
				{
					Class beansCustomizerClass = java.beans.Customizer.class;
					customizerClass = GlobalInfoTable.loadClass( customizerName );
					if ( beansCustomizerClass.isAssignableFrom( customizerClass ) )
						return customizerClass;
				} // try
				catch ( Exception e )
				{
					// Class not found.  Fall through and return null...
				} // catch
			} // if
		}

		return customizerClass;
	}

	/**
	 * This is a utility method used by getCustomEditor to take a customizer
	 * class and instanciate it and initialize it.
	 * <P>
	 * This may be used by subclasses of the PropertyEditor class to initialize
	 * custom editors in a standard way.
	 */
	public static java.awt.Component createCustomEditor( Class customizerClass, Object value )
	{
		try
		{
			try
			{
				java.awt.Component customEditor = (java.awt.Component) customizerClass.newInstance();

				if ( customEditor instanceof com.sas.InitializeInterface )
				{
					((com.sas.InitializeInterface)customEditor).initialize();
				} // if

				if ( customEditor instanceof java.beans.Customizer )
				{
					((java.beans.Customizer)customEditor).setObject( value );
				} // if
				
				return customEditor;
			} // try
			catch (InstantiationException ie )
			{
				// Unable to create custom editor, fall through and return null
			} // catch
		} // try
		catch (IllegalAccessException iae )
		{
			// Unable to create custom editor, fall through and return null
		} // catch
		
		return null;
	}

	/**
	 * getCustomEditor looks up the custom editor associated with the PropertyEditor.
	 * The custom editor is named *Customizer where * is the PropertyEditor class
	 * name without the Editor suffix.
	 * <P>
	 * If the custom editor implements the Customizer interface, setObject is called on
	 * the custom editor with the value from the PropertyEditor.
	 * @see #findCustomEditorClass
	 * @see java.beans.PropertyEditor#getCustomEditor
	 */
	public java.awt.Component getCustomEditor()
	{
		Class customizerClass = findCustomEditorClass();
		if ( customizerClass != null )
		{
			return createCustomEditor( customizerClass, getValue() );
		}

		return null;		
	}

	public java.awt.Window getParentWindow()
	{
		return parentWindow;
	}
	
	public void setParentWindow( java.awt.Window w )
	{
		parentWindow = w;
	}

	/**
	 * @see #findCustomEditorClass
	 * @see java.beans.PropertyEditor#supportsCustomEditor
	 */
	public boolean supportsCustomEditor()
	{
		Class customizerClass = findCustomEditorClass();
		if ( customizerClass != null )
			return true;

		return false;		
	}
	
	/**
	 * @see java.beans.PropertyEditor#addPropertyChangeListener
	 */
	public void addPropertyChangeListener(java.beans.PropertyChangeListener listener)
	{
		propertyChangeListeners.add(listener);
	}
	
	/**
	 * @see java.beans.PropertyEditor#removePropertyChangeListener
	 */
	public void removePropertyChangeListener(java.beans.PropertyChangeListener listener)
	{
		propertyChangeListeners.remove(listener);
	}


	public void firePropertyChange()
	{
		propertyChangeListeners.propertyChange( new PropertyChangeEvent( this, null, null, null ) );
	}

	public void propertyChange( java.beans.PropertyChangeEvent e )
	{
		firePropertyChange();
	}
}