// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.beans;

import java.util.Hashtable;


/**
 * PropertyMetadata is a utility class used by com.sas.beans.Introspector
 * to parse the metadata array returned in the ExtendedBeanInfo object.
 * 
 */
public class PropertyMetadata
{
	/**
	* The name of the property (required).
	*/
	public String name;
	
	/**
	* The default value of the property. (required)<P>
	* Used only in code generation.  Code generation will make a
	* setDefaultValues method that will look up the default value
	* and call the appropriate set method.  We could go ahead
	* and make such a method in the components now but it would
	* involve being able to convert from the string default value
	* to its appropriate type.  The setDefaultValues method can be
	* called from the design time and default constructors.
	*/
	public String defaultValue;
	
	/**
	* Short description of property meaning (suggested)
	*/
	public String shortDescription;
	
	/**
	* Description of property meaning (suggested)
	*/
	public String description;
	
	/**
	* Method used to get property value
	*/
	public String readMethodName;
	
	/**
	* Method used to set property value
	*/
	public String writeMethodName;
	
	/**
	* Whether to send an event when the property value changes
	* Checked in firePropertyChange before the
	* event is actually sent
	*/
	public boolean sendEvent;
	
	/**
	* Whether the property can be linked with addLink()
	* Checked in addLink() to ensure the link is allowed
	* before actually setting it.
	*/
	public boolean linkable;                      
	
	/**
	* Whether to compare the property as a reference or compare the contents of the object
	* unused
	*/
	public boolean compareValue;
	
	/**
	* Whether to persist the value of the property when the component is persisted
	* unused
	*/
	public boolean persist;
	
	/**
	* The name of the Java field used to hold the value of this property.
	* unused
	*/
	public String fieldName;
	
	/**
	* Whether the property is an class (true) or instance (false) property.
	*/
	public boolean staticProperty;
	
	/**
	* TO-DO:(SPL): For future use
	*/
	public String category;

	/**
	 * Whether the property write method sends PropertyChangeEvents
	 */
	public boolean bound;

	/**
	 * Whether the property write method sends VetoablePropertyChangeEvents
	 */
	public boolean constrained;

	/**
	 * The class the use to edit the property value in a property sheet
	 */
	public String propertyEditorClass;

	/**
	 * Whether the property is likely to only be used by expert users
	 */
	public boolean expert;

	/**
	 * Whether the property should not be shown in the property sheet
	 */
	public boolean hidden;

	/**
	 * Whether the property value should be masked in the property sheet
	 */
	public boolean masked;


	/**
	 * Whether the "Bound" attribute was set in the metadata array
	 */
	public boolean bound_set;
	/**
	 * Whether the "Constrained" attribute was set in the metadata array
	 */
	public boolean constrained_set;
	/**
	 * Whether the "Expert" attribute was set in the metadata array
	 */
	public boolean expert_set;
	/**
	 * Whether the "Hidden" attribute was set in the metadata array
	 */
	public boolean hidden_set;
	/**
	 * Whether the "Short description" attribute was set in the metadata array
	 */
	public boolean shortdesc_set;
	/**
	 * Whether the "Linkable" attribute was set in the metadata array
	 */
	public boolean linkable_set;
	/**
	 * Whether the "Default value" attribute was set in the metadata array
	 */
	public boolean defaultValue_set;
	/**
	 * Whether the "Persist" attribute was set in the metadata array
	 */
	public boolean persist_set;
	/**
	 * Whether the "Description" attribute was set in the metadata array
	 */
	public boolean description_set;
	/**
	 * Whether the "Masked" attribute was set in the metadata array
	 */
	public boolean masked_set;


	/**
	 * Constructor.  Sets default values.
	 */
	public PropertyMetadata()
	{
		name = "";
		defaultValue = "";
		readMethodName = null;
		writeMethodName = null;
		linkable = true;
		sendEvent= true;
		compareValue = true;
		persist = true;
		shortDescription = "";
		description = "";
		fieldName = "";
		staticProperty = false;
		category = null;

		bound = true;
		constrained = false;
		propertyEditorClass = null;
		expert = false;
		hidden = false;
		masked = false;

	     bound_set = true;
	}


	/**
	* createOne is a helper method for the PropertyMetadata class.
	* It reads a string array and creates the appropriate PropertyMetadata object
	* based on the contents of the array.  This is called by create().
	* @param attrArray String array of named property information.
	* @return An PropertyMetadata object based on the contents of the input array.
	* @see #create
	*/
	protected static PropertyMetadata createOne( String[][] attrArray )
	{
		boolean name_set = false;
		boolean get_accessor_set = false;
		boolean set_accessor_set = false;
		
		PropertyMetadata propInfo = new PropertyMetadata();
		
		for (int i = 0; i < attrArray.length; i++ )
		  {
			String tag = attrArray[i][0];
			String value = attrArray[i][1];

		    if ( tag.equals( ExtendedBeanInfo.NAME ) )
		      {
		        propInfo.name = value;		
		        name_set = true;
		      } // if
		    else if ( tag.equals(ExtendedBeanInfo.DEFAULT_VALUE) )
		      {
		        propInfo.defaultValue = value;
		        propInfo.defaultValue_set = true;
		      } // else if
		    else if ( tag.equals(ExtendedBeanInfo.SHORT_DESCRIPTION) )
		      {
		        propInfo.shortDescription = value;		
		        propInfo.shortdesc_set = true;
		      } // else if
		    else if ( tag.equals(ExtendedBeanInfo.DESCRIPTION) )
		      {
		        propInfo.description = value;		
		        propInfo.description_set = true;
		      }
		    else if ( attrArray[i][0].equals(ExtendedBeanInfo.READ_METHOD) )
		      {
		        propInfo.readMethodName = attrArray[i][1];
		
		        get_accessor_set = true;
		      } // else if
		    else if ( attrArray[i][0].equals(ExtendedBeanInfo.WRITE_METHOD) )
		      {
		        propInfo.writeMethodName = attrArray[i][1];
		
		        set_accessor_set = true;
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.SEND_EVENT) )
	          {
	            propInfo.sendEvent = Boolean.valueOf(attrArray[i][1]).booleanValue();
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.LINKABLE) )
	          {
	            propInfo.linkable = Boolean.valueOf(attrArray[i][1]).booleanValue();
//System.out.println("Linkable: " + propInfo.linkable);
	            propInfo.linkable_set = true;
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.COMPARE_VALUE) )
	          {
	            propInfo.compareValue = Boolean.valueOf(attrArray[i][1]).booleanValue();
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.PERSIST) )
	          {
	            propInfo.persist = Boolean.valueOf(attrArray[i][1]).booleanValue();
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.FIELD_NAME) )
	          {
	            propInfo.fieldName = attrArray[i][1];
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.STATIC) )
	          {
	            propInfo.staticProperty = Boolean.valueOf(attrArray[i][1]).booleanValue();
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.BOUND) )
	          {
	            propInfo.bound = Boolean.valueOf(attrArray[i][1]).booleanValue();
	            propInfo.bound_set = true;
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.CONSTRAINED) )
	          {
	            propInfo.constrained = Boolean.valueOf(attrArray[i][1]).booleanValue();
	            propInfo.constrained_set = true;
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.PROPERTY_EDITOR_CLASS) )
	          {
	            propInfo.propertyEditorClass = attrArray[i][1];
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.EXPERT) )
	          {
	            propInfo.expert = Boolean.valueOf(attrArray[i][1]).booleanValue();
	            propInfo.expert_set = true;
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.HIDDEN) )
	          {
	            propInfo.hidden = Boolean.valueOf(attrArray[i][1]).booleanValue();
	            propInfo.hidden_set = true;
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.MASKED) )
	          {
	            propInfo.masked = Boolean.valueOf(attrArray[i][1]).booleanValue();
	            propInfo.masked_set = true;
	          } // else if
	        else if ( attrArray[i][0].equals(ExtendedBeanInfo.CATEGORY) )
	          {
	            propInfo.category = attrArray[i][1];
	          } // else if
	        else
	        {
//XXX(spl)		System.out.println("PropertyMetadata::createOne(): Unused metadata warning : " +  attrArray[i][0] );
//	          throw new ComponentException("PropertyMetadata::createOne(): Invalid property name specified : " + attrArray[i][0] );
			}
	
	      } // for
	
		// Make sure Name is set.
		
		if ( name_set == false )
		{
			throw new com.sas.ComponentException( RB.getStringResource( "PropertyMetadata.createOne.ex.txt" ) );
		}
		
		// Set default values for items not specified
		
		if ( propInfo.description_set == false )
		  {
		    propInfo.description = propInfo.name;
		  } // if
		
		if ( propInfo.shortdesc_set == false )
		  {
		    propInfo.shortDescription = propInfo.description;
		  } // if

		return propInfo;
	} // createOne


	/**
	 * Method used by the Introspector to create a Hashtable of PropertyMetadata objects
	 * (hashed on the property name).
	 * @param namedMetadata the array of metadata to parse into PropertyMetadata objects
	 * @return a Hashtable of PropertyMetadata objects
	 * @see #createOne
	 */
	public static Hashtable create( String[][][] namedMetadata )
	{
		Hashtable metadata = new Hashtable();

		if ( namedMetadata != null )
		{
			for ( int i = 0; i < namedMetadata.length; i++ )
			{
				PropertyMetadata propertyInfo = createOne( namedMetadata[i] );
	
				metadata.put( propertyInfo.name, propertyInfo );
			} // for
		} // if
		
		return metadata;
	}
}