// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.beans;

import java.text.MessageFormat;
import java.util.WeakHashMap;
import java.util.Map;

import com.sas.GlobalInfoTable;

/**
 * Util is a utility class for JavaBeans.  It has a bunch of common methods which are used by
 * the Introspector, the PropertyEditors, the Customizers, and various other parts of the
 * component library. Methods in this class relate to:
 * <P>
 * <UL>Getting information about properties (descriptor, type)
 * <LI>Getting/setting property values on an object using reflection
 * <LI>Adding/removing PropertyChangeListeners from a JavaBean
 * <LI>Adding/removing generic event listeners from a JavaBean
 * <LI>Adding to the PropertyEditor and BeanInfo search paths
 * <LI>Getting/setting the JavaBeans design time state for the current Thread
 * </UL>
 */
public class Util
{
	public static java.beans.PropertyDescriptor getPropertyDescriptor( Object bean, String property )
	{
		Class cl = bean.getClass();
		return getPropertyDescriptor(cl,property);
	} // getPropertyDescriptor

	public static java.beans.PropertyDescriptor getPropertyDescriptor( Class cl, String property )
	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo(cl);
			java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			return propertyDescriptor;
		} // try
		catch ( Exception e )
		{
			// TO-DO: Log message
			String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
			System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
			e.printStackTrace();
		}

		return null;
	} // getPropertyDescriptor



	public static Class getPropertyType( Object bean, String property )
	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( bean.getClass() );
			java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			return propertyDescriptor.getPropertyType();
		} // try
		catch ( Exception e )
		{
			// TO-DO: Log message
			String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
			System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
			e.printStackTrace();
		}

		return null;
	} // getPropertyType

	public static Class getIndexedPropertyType( Object bean, String property, int index )
	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( bean.getClass() );
			java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			if ( propertyDescriptor instanceof java.beans.IndexedPropertyDescriptor )
			{
				return ((java.beans.IndexedPropertyDescriptor)propertyDescriptor).getIndexedPropertyType();
			} // if
			else
			{
				return null;
			} // else
		} // try
		catch ( Exception e )
		{
			// TO-DO: Log message
			String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
			System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
			e.printStackTrace();
		}

		return null;
	} // getIndexedPropertyType


	public static Object getPropertyValue( Object bean, String property )
	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( bean.getClass() );
			java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			java.lang.reflect.Method readMethod;
			readMethod = propertyDescriptor.getReadMethod();

			if ( readMethod == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetReadMethod.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			return com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( readMethod, bean, null );

		} // try
		catch ( com.sas.ComponentException ce )
		{
			// Don't do anything, fall through and return null...
		}
		catch ( Exception e )
		{
			// XXX
			if ( e instanceof java.lang.reflect.InvocationTargetException )
			{
				Throwable realException = ((java.lang.reflect.InvocationTargetException)e).getTargetException();
				// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {realException.toString()} ) );
				realException.printStackTrace();
			}
			else
			{
				// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
				e.printStackTrace();
			}
		}

		return null;
	} // getPropertyValue

	public static Object getIndexedPropertyValue( Object bean, String property, int index )
	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( bean.getClass() );
			java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			java.lang.reflect.Method readMethod;
			if ( propertyDescriptor instanceof java.beans.IndexedPropertyDescriptor )
			{
				readMethod = ((java.beans.IndexedPropertyDescriptor)propertyDescriptor).getIndexedReadMethod();

				Object[] params = new Object[1];
				params[0] = new Integer(index);
				return com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( readMethod, bean, params );
			} // if

			String errorMessage = RB.getStringResource( "Util.unableToGetIndexedReadMethod.ex.txt" );
			throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
		} // try
		catch ( Exception e )
		{
			// XXX
			if ( e instanceof java.lang.reflect.InvocationTargetException )
			{
				Throwable realException = ((java.lang.reflect.InvocationTargetException)e).getTargetException();
			// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {realException.toString()} ) );
				realException.printStackTrace();
			}
			else
			{
				// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
				e.printStackTrace();
			}
		}

		return null;
	} // getPropertyValue


	public static void setPropertyValue( Object bean, String property, Object value )
	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( bean.getClass() );
			java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}
			if ( propertyDescriptor.getWriteMethod() == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetWriteMethod.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			Object params[] = new Object[1];
			params[0] = value;
			com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( propertyDescriptor.getWriteMethod(), bean, params );

		} // try
		catch ( Exception e )
		{
			String errorMessage2 = RB.getStringResource( "Util.setPropertyValue.ex.txt" );
			System.out.println( MessageFormat.format( errorMessage2, new Object[] {property} ) );

			// XXX
			if ( e instanceof java.lang.reflect.InvocationTargetException )
			{
				Throwable realException = ((java.lang.reflect.InvocationTargetException)e).getTargetException();
				// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {realException.toString()} ) );
				realException.printStackTrace();
			}
			else
			{
				// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
				e.printStackTrace();
			}
		}
	} // setPropertyValue


	public static void setIndexedPropertyValue( Object bean, String property, Object value, int index )
	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( bean.getClass() );
			java.beans.IndexedPropertyDescriptor propertyDescriptor = (java.beans.IndexedPropertyDescriptor) com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}
			if ( propertyDescriptor.getIndexedWriteMethod() == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetIndexedWriteMethod.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			Object params[] = new Object[2];
			params[0] = new Integer(index);
			params[1] = value;
			com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( propertyDescriptor.getIndexedWriteMethod(), bean, params );

		} // try
		catch ( Exception e )
		{
			// XXX
			if ( e instanceof java.lang.reflect.InvocationTargetException )
			{
				Throwable realException = ((java.lang.reflect.InvocationTargetException)e).getTargetException();
				// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {realException.toString()} ) );
				realException.printStackTrace();
			}
			else
			{
				// TO-DO: Log message
				String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
				System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
				e.printStackTrace();
			}
		}
	} // setIndexedPropertyValue


	public static Class getPropertyEditorClass( Object bean, String property )

	{
		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( bean.getClass() );
			java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( property, sourceBeanInfo );
			if ( propertyDescriptor == null )
			{
				String errorMessage = RB.getStringResource( "Util.unableToGetPropertyDescriptor.ex.txt" );
				throw new com.sas.ComponentException( MessageFormat.format( errorMessage, new Object[] {property} ) );
			}

			return propertyDescriptor.getPropertyEditorClass();
		} // try
		catch ( Exception e )
		{
			// TO-DO: Log message
			String errorMessage = RB.getStringResource( "Util.errorLookingUpPropertyInformation.ex.txt" );
			System.out.println( MessageFormat.format( errorMessage, new Object[] {e.toString()} ) );
			e.printStackTrace();
		}

		return null;
	} // getPropertyEditorClass


	/**
	 * addPropertyChangeListener calls addPropertyChangeListener on the specified
	 * object and adds the specified listener.  It uses reflection to do this since there
	 * is not a java.beans.PropertyChangeSource interface.
	 * <P>
	 * If the object passed in implements the com.sas.beans.PropertyChangeSource interface,
	 * addPropertyChangeListener uses the interface to call the method instead of reflection.
	 * <P>
	 * @param object The object to add the listener to.
	 * @param listener The listener to add.
	 * @returns true if the listener was added, false if the listener was not.
	 */
	public static boolean addPropertyChangeListener( Object object, java.beans.PropertyChangeListener listener )
	{
		if ( object instanceof com.sas.beans.PropertyChangeSource )
		{
			com.sas.beans.PropertyChangeSource source = (com.sas.beans.PropertyChangeSource) object;
			source.addPropertyChangeListener( listener );

			return true;
		} // if

		// See if the Object is a java.lang Object, they don't send PropertyChangeEvents

		if ( object.getClass().getName().startsWith("java.lang.") )
		{
			return false;
		}

		return addEventListener( java.beans.PropertyChangeListener.class, "addPropertyChangeListener", object, listener );
	}

	/**
	 * removePropertyChangeListener calls removePropertyChangeListener on the specified
	 * object and removes the specified listener.  It uses reflection to do this since there
	 * is not a java.beans.PropertyChangeSource interface.
	 * <P>
	 * If the object passed in implements the com.sas.beans.PropertyChangeSource interface,
	 * removePropertyChangeListener uses the interface to call the method instead of reflection.
	 * <P>
	 * @param object The object to remove the listener from.
	 * @param listener The listener to remove.
	 * @returns true if the listener was removed, false if the listener was not.
	 */
	public static boolean removePropertyChangeListener( Object object, java.beans.PropertyChangeListener listener )
	{
		if ( object instanceof com.sas.beans.PropertyChangeSource )
		{
			com.sas.beans.PropertyChangeSource source = (com.sas.beans.PropertyChangeSource) object;
			source.removePropertyChangeListener( listener );

			return true;
		} // if

		// See if the Object is a java.lang Object, they don't send PropertyChangeEvents

		if ( object.getClass().getName().startsWith("java.lang.") )
		{
			return false;
		}

		return removeEventListener( java.beans.PropertyChangeListener.class, "removePropertyChangeListener", object, listener );
	}

	/**
	 * addeventListener calls <code>addListenerMethod</code> on the specified
	 * object and adds the specified listener.  It uses reflection to do this since there
	 * are not source interfaces for all event types.
	 * <P>
	 * @param listenerClass The listener class (ie. java.beans.PropertyChangeListener)
	 * @param addListenerMethod The method to call to add the listener (ie. addPropertyChangeListener).
	 * @param object The object to add the listener to.
	 * @param listener The listener to add.
	 * @returns true if the listener was added, false if the listener was not.
	 */
	public static boolean addEventListener( Class listenerClass, String addListenerMethod, Object object, java.beans.PropertyChangeListener listener )
	{
		Object[] params = new Object[1];
		params[0] = listener;

		try
		{
			Class[] argTypes = new Class[]{listenerClass};
			java.lang.reflect.Method method = com.sas.security.BaseSecuritySupport.securitySupport.getMethod( object.getClass(), addListenerMethod, argTypes );
			com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( method, object, params );

			return true;
		}
		catch (Exception e)
		{
          	// Ignore any exceptions and fall through to return false.
		}

		return false;
	}

	/**
	 * removeEventListener calls <code>removeListenerMethod</code> on the specified
	 * object and removes the specified listener.  It uses reflection to do this since there
	 * are not source interfaces for all event types.
	 * <P>
	 * @param listenerClass The listener class (ie. java.beans.PropertyChangeListener)
	 * @param removeListenerMethod The method to call to remove the listener (ie. removePropertyChangeListener).
	 * @param object The object to remove the listener from.
	 * @param listener The listener to remove.
	 * @returns true if the listener was removed, false if the listener was not.
	 */
	public static boolean removeEventListener( Class listenerClass, String removeListenerMethod, Object object, java.beans.PropertyChangeListener listener )
	{
		Object[] params = new Object[1];
		params[0] = listener;

		try
		{
			Class[] argTypes = new Class[]{listenerClass};
			java.lang.reflect.Method method = com.sas.security.BaseSecuritySupport.securitySupport.getMethod( object.getClass(), removeListenerMethod, argTypes );
			com.sas.security.BaseSecuritySupport.securitySupport.invokeMethod( method, object, params );

			return true;
		}
		catch (Exception e)
		{
          	// Ignore any exceptions and fall through to return false
		}

		return false;
	}

	/**
	 * prependToPropertyEditorSearchPath puts a new search package in the
	 * PropertyEditor search path maintained by the PropertyEditorManager.
	 * @param path the package to add to the beginning of the search path.
	 */
	public static void prependToPropertyEditorSearchPath( String path )
	{   
		//set modelEditor for swing components
		specifyModelEditor();

		// Add com.sas.beans.editors to the property editor search path

		String[] searchPath = java.beans.PropertyEditorManager.getEditorSearchPath();

		for ( int i = 0; i < searchPath.length; i++ )
		{
			if ( searchPath[i].equals( path ) )
			{
				return;
			} // if
		} // for

		String[] newSearchPath = new String[ searchPath.length + 1 ];
		System.arraycopy( searchPath, 0, newSearchPath, 1, searchPath.length );
		newSearchPath[0] = path;
		final String[] newSearchPathFinal = newSearchPath;
		java.security.AccessController.doPrivileged( new java.security.PrivilegedAction(){
			public Object run(){
				java.beans.PropertyEditorManager.setEditorSearchPath(newSearchPathFinal);
				return null;
			}
		});
	}

	/**
	 * prependToBeanInfoSearchPath puts a new search package in the
	 * search path for beanifnos.
	 * @param path the package to add to the beginning of the search path.
	 */
	public static void prependToBeanInfoSearchPath( String path )
	{   
		
		// Add com.sas.beans.beaninfos to the beaninfo search path

		String[] searchPath = java.beans.Introspector.getBeanInfoSearchPath();

		for ( int i = 0; i < searchPath.length; i++ ){
			if ( searchPath[i].equals( path ) )
					return;
			
		} // for

		String[] newSearchPath = new String[ searchPath.length + 1 ];
		System.arraycopy( searchPath, 0, newSearchPath, 1, searchPath.length );
		newSearchPath[0] = path;
		final String[] newSearchPathFinal = newSearchPath;
		java.security.AccessController.doPrivileged( new java.security.PrivilegedAction(){
			public Object run(){
				java.beans.Introspector.setBeanInfoSearchPath(newSearchPathFinal);
				return null;
			}
		});
	}



	// Specify the modelEditor for swingComponents. Prepend the path for the 
	// swing beaninfos
	public static void specifyModelEditor()
	{
 
	 //prepend path for subclassed swing beaninfos
	 prependToBeanInfoSearchPath("com.sas.beans.beaninfos");

	 try
	 {
	 	Class swingClasses[] = {GlobalInfoTable.loadClass("javax.swing.JButton"),
							   GlobalInfoTable.loadClass("javax.swing.JCheckBox"),
							   GlobalInfoTable.loadClass("javax.swing.JList"),
							   GlobalInfoTable.loadClass("javax.swing.JComboBox"),
							   GlobalInfoTable.loadClass("javax.swing.JProgressBar"),
							   GlobalInfoTable.loadClass("javax.swing.JRadioButton"),
							   GlobalInfoTable.loadClass("javax.swing.JScrollBar"),
							   GlobalInfoTable.loadClass("javax.swing.JSlider"),
							   GlobalInfoTable.loadClass("javax.swing.JToggleButton"),
							   GlobalInfoTable.loadClass("javax.swing.JTree"),
			                   GlobalInfoTable.loadClass("javax.swing.JTabbedPane"),
							   GlobalInfoTable.loadClass("javax.swing.JTable")};
		Class propertyType = null;
		String modelEditor = "com.sas.beans.editors._javaxSwingModelEditor";
		for (int i=0; i < swingClasses.length; i++){
	 		java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo(swingClasses[i] );
	 		java.beans.PropertyDescriptor propertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor("model", sourceBeanInfo );

	 		if ( propertyDescriptor != null ){
	   			propertyType = propertyDescriptor.getPropertyType();
	    		java.beans.PropertyEditorManager.registerEditor(propertyType, GlobalInfoTable.loadClass(modelEditor));
	 	  
     		}//if
    	} //for
    } //try
	catch (Exception e)
	{
	 
    }
   } //specifyModelEditor


	/**
	* designTimeTable is used by sisDesignTime() and setDesignTime() to
	* store and retrieve the design time flag value on a per-thread basis.
	* Only threads which have called setDesignTime() will have data in
	* designTimeTable.
	* @see #isDesignTime
	* @see #setDesignTime
	*/
	private static Map designTimeTable;

	/**
	* setDesignTime() sets the design time flag value for the currently executing
	* thread.
	* This method is used by webAF since the JavaBeans method
	* <code>java.beans.Beans.setDesignTime()</code>
	* does not work well in a multithreaded development environment.
	* @param designTime The new value for the designTime flag for the current thread.
	* @see #isDesignTime
	*/
	public static synchronized void setDesignTime( boolean designTime )
	{
		if ( designTimeTable == null )
			designTimeTable = new WeakHashMap();

		designTimeTable.put( Thread.currentThread(), designTime ? Boolean.TRUE : Boolean.FALSE );
	} // beansSetDesignTime

	/**
	* isDesignTime() returns the design time flag value for the currently executing
	* thread.
	* This method is used by webAF since the JavaBeans method
	* <code>java.beans.Beans.isDesignTime()</code>
	* does not work well in a multithreaded development environment.
	* @return ssDesignTime() returns the design time flag value for the currently executing
	* thread.
	* @see #setDesignTime
	*/
	public static synchronized boolean isDesignTime()
	{
		if ( designTimeTable == null )
			return false;

		Boolean designTimeBoolean = (Boolean) designTimeTable.get( Thread.currentThread() );

		if ( designTimeBoolean == null )
		{
			return false;
		} // if

		return designTimeBoolean.booleanValue();
	} // beansIsDesignTime

} // Util
