
package com.sas.collection;

import com.sas.util.Comparator;
import com.sas.util.CompositeComparator;
import java.io.Serializable;
import com.sas.PublicClonable;
import java.util.NoSuchElementException;
import java.util.Enumeration;
import java.util.Vector;
import com.sas.util.Enumerable;
import java.util.Stack;
// import com.sas.DeepCloneSupport;
// import com.sas.DeepClonable;

// See also implementation notes/credits at the bottom

/**
 * A node for AVL tree.
 * This class represents the nodes within an AVL tree,
 * named after Adelson-Velskii and Landis.
 * The nodes are <em>threaded</em> which means nodes
 * which do not have a left or right child have links to
 * the previous or next subtree; this allows for easy
 * forward or backward enumeration of the items in the tree.
 * Each node has either a left child or a reference to the
 * subtree containing the previous node.
 * Each node also has either a right child or a reference to the a
 * subtree containing the next node.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @since 1.2
 * @see AVLTree
 */

public class AVLNode
       implements Serializable,
                  PublicClonable,
                    // DeepClonable,
                  Enumerable
{
  public static final String RB_KEY = "AVLNode.";

  static final long serialVersionUID = 3784226262729130827L;

  //  private static final boolean DEBUG = false;
  //  private static String indent = ""; // for debugging

  private static long counter;       //for generating unique id's

  // the value stored in the node
  private Object        obejct;

  // left right are transient here so that
  // writeObject()/readObject() can stream them out in
  // a specific order so that we can ensure the order relation of
  // each node's counter is preserved correctly.
  private transient AVLNode  left;
  private transient AVLNode  right;
  // combines balance and 2 bits for left/right threading
  private byte               state;

  // Unique id, generated from static counter.
  // See init() which is also called from readObject
  // Since a pure AVLTree does not support duplicates, the ID is used
  // to disambiguate items for which a Comparator.compare() returns EQUALS,
  private transient long id;

  /**
   * Construct a new AVLNode.
   */
  public AVLNode()
    {
      this(null);
    }

  /**
   * Construct a new AVLNode which contains an object
   */
  public AVLNode(Object object)
    {
      init();
      setLeftThread(null);
      setRightThread(null);
      setEvenBalance();
      setObject(object);
    }

  // Give the node a unique ID for comparisons.
  private void init()
    {
      id = counter++;
    }

  /**
   * Set the object stored at this node.
   * This object is used for comparing new items
   * being inserted or in item searches.
   * @param object the object stored at this node.
   */
  public void setObject(Object object)
    {
      this.obejct = object;
    }

  /**
   * Return the object stored at this node.
   * @return the object stored at this node.
   */
  public Object getObject()
    {
      return obejct;
    }


  // -------------------------------- State maintanance -------------------------------- //

  // If LEFT_THREAD bit is set in state, then left is a thread pointer to
  // left nodes (or null), else left is a pointer to the left child
  // If RIGHT_THREAD bit is set in state, then right is a thread pointer to
  // right nodes (or null), else right is a pointer to the right child
  private final static int LEFT_THREAD      = 0x1;
  private final static int RIGHT_THREAD     = 0x2;
  private final static int THREAD_MASK      = LEFT_THREAD | RIGHT_THREAD;

  // if _binstate & NEGATIVE_BALANCE then node has balance of -1
  // if _binstate & POSITIVE_BALANCE then node has balance of +1
  private static final int    NEGATIVE_BALANCE = 0x4;
  private static final int    POSITIVE_BALANCE = 0x8;
  private static final int    BALANCE_MASK     = NEGATIVE_BALANCE | POSITIVE_BALANCE;


  // -------------------------------- Balance -------------------------------- //

  /**
   * Set a node's balance to negative, even, or positive
   * In many published algorithms, this is denoted by <code>node.balance = a</code>
   * Positive balance means the depth of the right subtree is greater than
   * the depth of the left subtree.
   * Negative balance means the depth of the left subtree is greater than
   * the depth of the rightsubtree.
   * @param balance -1 for negative balance, 0 for even balance, 1 for positive balance
   */
  protected void setBalance(int balance)
    {
      if (balance == 0)
        setEvenBalance();
      else if (balance == -1)
        setNegativeBalance();
      else if (balance == +1)
        setPositiveBalance();
      else
        throw new IllegalArgumentException("balance not in {-1, 0, +1}");
    }

  /**
   * Return a node's balance
   * @return the current node balance; -1 means negative, 0 is balanced, 1 is positive balance.
   * Positive balance means the depth of the right subtree is greater than
   * the depth of the left subtree.
   * Negative balance means the depth of the left subtree is greater than
   * the depth of the rightsubtree.
   */
  protected int balance()
    {
      if (isNegativelyBalanced())
        return -1;
      else if (isPositivelyBalanced())
        return +1;
      else
        return 0;
      // return (state >> 2);
    }

  /**
   * Mark the node as having negative balance.
   * In many published algorithms, this is denoted by <code>node.balance == -1</code>
   * Calling <code>setNegativeBalance</code> is equivalent to setting <code>node.balance = -1</code>
   * @see #setBalance(int)
   */
  protected void setNegativeBalance()
    {
      state = (byte) ((state & THREAD_MASK) | NEGATIVE_BALANCE);
    }

  /**
   * Mark the node as having positive balance.
   * In many published algorithms, this is denoted by <code>node.balance == +1</code>
   * but for space reasons, the balance is embedded within the <var>state</var>
   * variable. Calling <code>setPositiveBalance</code> is equivalent to setting <code>node.balance = +1</code>
   * @see #setBalance(int)
   */
  protected void setPositiveBalance()
    {
      state = (byte) ((state & THREAD_MASK) | POSITIVE_BALANCE);
    }

  /**
   * Mark the node as having even balance.
   * In many published algorithms, this is denoted by <code>node.balance == 0</code>
   * but for space reasons, the balance is embedded within the <var>state</var>
   * variable. Calling <code>setEvenBalance</code> is equivalent to setting <code>node.balance = 0</code>
   * @see #setBalance(int)
   */
  protected void setEvenBalance()
    {
      state = (byte) (state & THREAD_MASK);
    }

  /**
   * Return true if the node has negative balance.
   */
  protected boolean isNegativelyBalanced()
    {
      return (state & NEGATIVE_BALANCE) == NEGATIVE_BALANCE;
    }

  /**
   * Return true if the node has positive balance.
   */
  protected boolean isPositivelyBalanced()
    {
      return (state & POSITIVE_BALANCE) == POSITIVE_BALANCE;
    }

  /**
   * Return true if the node has even balance.
   */
  protected boolean isEvenlyBalanced()
    {
      return (state & BALANCE_MASK) == 0;
    }


 // --------------------------------  Left Child/Previous Node  -------------------------------- //


  /**
   * Get the left link. If the node is left threaded, this
   * is the previous node in the tree. If the node is not
   * left threaded, this is the left child.
   * @return the left link
   */
  public AVLNode leftLink()
    {
      return left;
    }

  /**
   * Assign the left child node or the left thread if child is null
   * @param child the node to assign as the left child node.
   * @param thread the node to assign as the left thread, iff child == null
   */
  protected void setLeftChild(AVLNode child, AVLNode thread)
    {
//      if (DEBUG && child == null && thread == null)
//        {
//          System.out.println(indent + "NOTE: " + getObject() + ".left child/thread = null/null");
//          // new Throwable("traceback").printStackTrace();
//        }
      if (child == null)
        setLeftThread(thread);
      else
        setLeftChild(child);
    }

  /**
   * Assign the left child node.
   * @param prev the node to assign as the left child node.
   */
  protected void setLeftChild(AVLNode child)
    {
      state &= ~LEFT_THREAD;
      // ASSERT //       if (ASSERT && child == null)
      // ASSERT //         {
      // ASSERT //           System.out.println(indent + "NOTE: setLeftChild(null)" );
      // ASSERT //           setLeftThread(null);
      // ASSERT //           // new Throwable().printStackTrace();
      // ASSERT //         }
      left = child;
    }

  /**
   * Get the left child.
   * @return the left child, or null if there is no left child.
   */
  public AVLNode getLeftChild()
    {
      if (hasLeftThread())
        return null;
      else
        return leftLink();
    }


  /**
   * Get the left thread
   * @return the left thread, or null if there is no left thread
   */
  public AVLNode getLeftThread()
    {
      if (hasLeftThread())
        return leftLink();
      else
        return null;
    }

  /**
   * Assign the previous node. This node will no longer have a left child.
   * This node is marked as left threaded.
   * @param prev the node to assign as the previous node.
   */
  protected void setLeftThread(AVLNode prev)
    {
      state |= LEFT_THREAD;
      left = prev;
    }

  /**
   * Return true if the node is left threaded, that is the leftLink() is a link to
   * a previous node in the enumeration. Note that this may be true when
   * the left thread is null.
   * This method is really the equivalent to !(hasLeftChild())
   */
  protected boolean hasLeftThread()
    {
      return (state & LEFT_THREAD) != 0;
    }

  /**
   * Return true if the node has a left child
   * @return true if the node has a left child
   */
  public final boolean hasLeftChild()
    {
      return !hasLeftThread();
    }

 // --------------------------------  Right Child/Next Node  -------------------------------- //


  /**
   * Get the right link. If the node is right threaded, this
   * is the next node in the tree. If the node is not
   * left threaded, this is the left child.
   * @return the right link
   */
  public AVLNode rightLink()
    {
      return right;
    }


  /**
   * Get the right child.
   * @return the right child, or null if there is no right child.
   */
  public AVLNode getRightChild()
    {
      if (hasRightThread())
        return null;
      else
        return rightLink();
    }

  /**
   * Get the right thread
   * @return the right thread, or null if there is no right thread
   */
  public AVLNode getRightThread()
    {
      if (hasRightThread())
        return rightLink();
      else
        return null;
    }

  /**
   * Assign the next node. This node will no longer have a right child.
   * This node is marked as right threaded.
   * @param next the node to assign as the next node.
   */
  protected void setRightThread(AVLNode next)
    {
      state |= RIGHT_THREAD;
      right = next;
    }

  /**
   * Assign the right child node, or the right thread if child is null.
   * Mark the node as not right threaded. The right link is not a link to
   * a next node in the enumeration.
   * @param child the new right child node
   * @param thread the node to assign as the right thread, iff child == null
   */
  protected void setRightChild(AVLNode child, AVLNode thread)
    {
// DEBUG //      if (DEBUG && child == null && thread == null)
// DEBUG //        {
// DEBUG //          System.out.println(indent + "NOTE: " + getObject() + ".right child/thread = null/null");
// DEBUG //          // new Throwable("traceback").printStackTrace();
// DEBUG //        }
      if (child == null)
        setRightThread(thread);
      else
        setRightChild(child);
    }

  /**
   * Assign the right child node.
   * Mark the node as not right threaded. The right link is not a link to
   * a next node in the enumeration.
   * @param child the new right child node
   */
  protected void setRightChild(AVLNode child)
    {
      state &= ~RIGHT_THREAD;
      // ASSERT //       if (ASSERT && child == null)
      // ASSERT //         {
      // ASSERT //           System.out.println(indent + "NOTE: setRightChild(null)" );
      // ASSERT //           setRightThread(null);
      // ASSERT //           // new Throwable().printStackTrace();
      // ASSERT //         }
      right = child;
    }

  /**
   * Return true if the node is right threaded, that is the rightLink() is a link to
   * a next node in the enumeration. Note: this right thread link may be null.
   * This method is really the equivalent to !(hasRightChild())
   */
  protected boolean hasRightThread()
    {
      return (state & RIGHT_THREAD) != 0;
    }


  /**
   * Return true if the node has a right child.
   * @return true if the node has a right child.
   */
  public final boolean hasRightChild()
    {
      return ! hasRightThread();
    }


  /**
   * Get the next node. If the node has a right child,
   * we find its letftmost descendent.
   * @return the next node, or null if there is no next node node.
   */
  public AVLNode getNextNode()
    {
      AVLNode node = rightLink();
      if (hasRightThread())
        return node;
      return node.leftMostDescendent();
    }

  /**
   * Return the left-most descendent
   * @return the left-most descendent
   */
  public AVLNode leftMostDescendent()
    {
      AVLNode node = this;

      /*-- go all the way left --*/
      while ( node.hasLeftChild() )
         node = node.getLeftChild();

      return node;
    }


  /**
   * Get the previous node. If the node has a left child,
   * we find its rightmost descendent.
   * @return the previous node, or null if there is no previous node.
   */
  public AVLNode getPreviousNode()
    {
      AVLNode node = leftLink();
      if (hasLeftThread())
        return node;
      return node.rightMostDescendent();
    }

  /**
   * Return the right-most descendent
   * @return the right-most descendent
   */
  public AVLNode rightMostDescendent()
    {
      AVLNode node = this;

      /*-- go all the way right --*/
      while ( node.hasRightChild() )
         node = node.getRightChild();

      return node;
    }

 // -------------------------------- Search --------------------------------

  /**
   * Search for an <var>object</var> in the tree rooted by <var>node</var>,
   * using the <var>comparator</var>.
   * @param the object to search for
   * @param node the root node to start the search
   * @param comparator the comparator to use. The <var>object</var>
   * is passed as the first argument to the {@see com.sas.util.Comparator#compare(Object, Object)}
   * method; the element from the AVL tree nodes is passed as the second.
   * @return the AVLNode that contains an item matcing the <var>object</var>,
   * or null if no item was found.
   */
  public static AVLNode search(Object object, AVLNode node, Comparator comparator)
  {
    if (node == null)
      return null;

    int compare = comparator.compare(object, node.getObject());
    switch (compare)
      {
      case (Comparator.EQUALS):
        return node;

      case (Comparator.LESS_THAN):
        return search(object, node.getLeftChild(), comparator);

      case (Comparator.GREATER_THAN):
        return search(object, node.getRightChild(), comparator);

      default:
        return null;
      }
  }

 // -------------------------------- add -------------------------------- //

  /**
   * Add the <var>object</var> into a tree rooted at <var>root</var>,
   * using the comparator.
   * @param object the object to insert into the tree
   * @param root the root node. This may be null
   * @param comparator the Comparator to use when comparing items.
   * @param allowDuplicates if true, you can insert equal items into the tree;
   *        a unique identifier is assigned each node and that is used to
   *        distinguish items.
   * @return the new root, or null if a item that equals node.object already
   *         exists in the tree and allowDuplicates is false
   */
  public static AVLNode add(Object object, AVLNode root, Comparator comparator, boolean allowDuplicates)
    {
      return add(new AVLNode(object), root, comparator, allowDuplicates);
    }


  /**
   * Add a AVL <var>node</var> into a tree rooted at <var>root</var>,
   * using the comparator.
   * @param node an AVL tree node
   * @param root the root node. This may be null
   * @param comparator the Comparator to use when comparing items.
   * @param allowDuplicates if true, you can insert equal items into the tree;
   *        a unique identifier is assigned each node and that is used to
   *        distinguish items.
   * @return the new root, or null if a item that equals node.object already
   *         exists in the tree and allowDuplicates is false
   */
  public static AVLNode add(AVLNode node, AVLNode root,
                               Comparator comparator, boolean allowDuplicates)
    {

      if ( root == null )
        {
          //-- first node
          root = node;
          node.setLeftThread(null);
          node.setRightThread(null);
          node.setEvenBalance();
          return root;
        }

      //-- a1: initialize

      AVLNode   p       = root;
      AVLNode   q       = null;
      AVLNode   s       = root;
      AVLNode   t       = null;

      /*-------------------------------------------------------------------*/
      /*-- cmap is the comparison map that records the path from the     --*/
      /*-- nearest unbalanced parent to the new node. '0' is a left link     --*/
      /*-- move and '1' is a right link move.                                 --*/
      /*-------------------------------------------------------------------*/
      long    cmap = 0L;  //-- comparison map --
      long    gt   = 1L;  //-- greater-than mask --

      //-- search/insert loop --

      for ( ; ; )
        {
          // a2: compare and insert

          int compare = comparator.compare(node.getObject(), p.getObject());
          if (   (   compare == Comparator.EQUALS
                  || compare == Comparator.INCOMPARABLE)
              && allowDuplicates)
            {
              compare = (  (node.id < p.id) ? Comparator.LESS_THAN
                         : (  (node.id > p.id)
                            ? Comparator.GREATER_THAN
                            : Comparator.EQUALS));
              if ( compare == 0 )
                throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "duplicatesNotAllowed.ex.txt"));
            }

          // System.out.println("compare: " + node.object + ", " + p.object + " ==> " + compare);

          if (compare == Comparator.LESS_THAN)
            // node preceeds p
            {
              AVLNode ll = p.leftLink();
              if ( p.hasLeftThread() )
                {
                  //-- at leaf node, transfer thread --
                  node.setLeftThread(ll);
                  node.setRightThread(p);
                  p.setLeftChild(node);
                  break;
                }

              q = ll;
            }
          else if (compare == Comparator.GREATER_THAN)
            {
              // node follows p

              cmap |= gt;
              AVLNode rl = p.rightLink();

              if ( p.hasRightThread() )
                {
                  //-- at leaf node, transfer thread --
                  node.setRightThread(rl);
                  node.setLeftThread( p );
                  p.setRightChild(node);
                  break;
                }

              q = rl;
            }
          else // INCOMPARABLE or EQUALS
            {
              return null;
            }

          //-- advance markers to nearest unbalanced parent --

          // ll = q.leftLink();
          // rl = q.rightLink();

          if ( ! q.isEvenlyBalanced() )
            {
              s = q;
              t = p;
              //-- new unbalanced parent, restart cmap --
              cmap = 0L;
              gt = 1L;
            }
          else
            {
              gt <<= 1;
            }

          p = q;
        }

      //-- If we get here, "new" was added. Finish init.--

      node.setEvenBalance();

      //-- a6: adjust balancing factors --

      int a;
      if ( (cmap & 1) != 0 )
        {
          a = 1;
          p = s.rightLink();
          q = p;
        }
      else
        {
          a = -1;
          p = s.leftLink();
          q = p;
        }

      for ( gt = 2L; p != node; gt <<= 1 )
        {
          //-- follow the cmap down to the new node --
          if ( (cmap & gt) != 0 )
            {
              p.setPositiveBalance();
              // ASSERT( !IS_THREAD( p.rightLink() ) );
              p = p.rightLink();
            }

          //---- Java has 64-bit longs, so this check from C is not needed
          //
          // else if ( gt == 0 )
          //   {
          //     /*-------------------------------------------------------------*/
          //     /*-- gt has been shifted out to zero which means that cmap   --*/
          //     /*-- has overflowed and we have lost our way because there   --*/
          //     /*-- are more than bits-in-s-long levels between the last    --*/
          //     /*-- unbalanced parent and "node". This is not a total        --*/
          //     /*-- disaster, it just means that the tree is a bit          --*/
          //     /*-- unbalanced and will not perform quite as well. To reach --*/
          //     /*-- this state, the tree must have on the order of tens of  --*/
          //     /*-- millions (if not hundreds of millions) of nodes on a    --*/
          //     /*-- 32-bit long machine. Before we give-up check to see if  --*/
          //     /*-- we are only "off-by-one".                               --*/
          //     /*-------------------------------------------------------------*/
          //     if ( p.rightLink() == node )
          //       p.setPositiveBalance();
          //     else if ( p.leftLink() == node )
          //       p.setNegativeBalance();
          //
          //
          //     /*-------------------------------------------------------------*/
          //     /*-- NOTE: Only print message for debug case since a         --*/
          //     /*-- function call here would prevent dsa-less linkage on    --*/
          //     /*-- OS/390.                                                 --*/
          //     /*-------------------------------------------------------------*/
          //     else if (DEBUG)
          //       System.err.println( indent + "WARNING: CMAP overflow, AVL tree has lost balance." );
          //
          //     break;
          //   }
          else
            {
              p.setNegativeBalance();
              // ASSERT( !IS_THREAD( p.leftLink() ) );
              p = p.leftLink();
            }

          // ASSERT( p );
        }

      //-- a7: balancing act --

      if ( s.isEvenlyBalanced() )
        {
          s.setBalance(a);
          return root;
        }

      if ( s.balance() == -a )
        {
          s.setEvenBalance();
          return root;
        }

      s.setEvenBalance();


      if ( q.balance() == a )
        {
          //-- a8: single rotation --

          q.setEvenBalance();
          p = q;

          if ( a > 0 )
            {
              //-- rotate left --
              if ( q.hasLeftThread() )
                s.setRightThread(q);
              else
                s.setRightChild(q.leftLink());

              q.setLeftChild(s);
            }
          else
            {
              //-- rotate right --
              if ( q.hasRightThread() )
                s.setLeftThread( q );
              else
                s.setLeftChild(q.rightLink());

              q.setRightChild(s);
            }
        }
      else
        {
          //-- a9: double rotation --

          q.setEvenBalance();

          if ( a > 0 )
            {
              //-- rotate left twice --
              p = q.leftLink();

              if ( p.hasRightThread() )
                q.setLeftThread( p );
              else
                q.setLeftChild(p.rightLink());

              p.setRightChild(q);

              if ( p.hasLeftThread( ) )
                s.setRightThread( p );
              else
                s.setRightChild(p.leftLink());

              p.setLeftChild(s);
            }
          else
            {
              //-- rotate right twice --
              p = q.rightLink();

              if ( p.hasLeftThread( ) )
                q.setRightThread( p );
              else
                q.setRightChild(p.leftLink());

              p.setLeftChild(q);

              if ( p.hasRightThread( ) )
                s.setLeftThread( p );
              else
                s.setLeftChild(p.rightLink());

              p.setRightChild(s);
            }

          if ( p.balance() == a )
            s.setBalance(-a);
          else if ( ! p.isEvenlyBalanced() )
            q.setBalance(a);

          p.setEvenBalance();
        }

      //-- a10: finishing touch --

      if ( t == null )
        root = p;
      else if ( t.rightLink() == s )
        t.setRightChild(p);
      else
        t.setLeftChild(p);

      return root;
    }


  /**
   * Return an enumeration of the items in the tree rooted at this node.
   */
  public Enumeration getItems()
    {
      return getItems(false);
    }

  /**
   * Get an enumeration of the tree contents, or a subrange of items.
   * @param start the enumeration after <var>start</var> items.
   * @param end the enumeration after <var>end</var> items.
   */
  public Enumeration getItems(int start, int end)
    {
      return new IndexedEnumeration(getItems(false), start, end);
    }

  /**
   * Get an enumeration of the tree contents.
   * @param backwards if true, the enumeration
   * traverses the tree backwards, from the end to the beginning.
   */
  public Cursor getItems(boolean reversedOrder)
    {
      return new Cursor(this, reversedOrder);
    }

  /**
   * Cursor is an Enumeration for AVLNodes in an AVLTree.
   * It works like an Enumeration but also provides backward
   * enumeration.
   */
  public static class Cursor implements Enumeration
  {
    private AVLNode node;
    private boolean reversedOrder;

    /**
     * Construct an AVLNode cursor.
     * @param reversedOrder if true, start at the rightmost
     * descendent and move backwards, else start
     * at the leftmost descendent and move forwards.
     */
    public Cursor(AVLNode root, boolean reversedOrder)
    {
      this.reversedOrder = reversedOrder;
      if (root == null)
        node = null;
      else
        if (reversedOrder)
          {
            node = new AVLNode();
            node.setLeftThread(root.rightMostDescendent());
          }
        else
          {
            node = new AVLNode();
            node.setRightThread(root.leftMostDescendent());
          }
    }

    /**
     * Return true if this enumeration has more elements.
     */
    public boolean hasMoreElements()
    {
      if (reversedOrder)
        return hasBackwardElement();
      else
        return hasForwardElement();
    }

    /**
     * Return the next element in this enumeration.
     * @throws NoSuchElementException if there are no more items.
     * @see #hasMoreElements
     */
    public Object nextElement()
    {
      if (reversedOrder)
        return backwardElement();
      else
        return forwardElement();
    }

    /**
     * Return true if there are more elements in the forward direction.
     */
    public boolean hasForwardElement()
    {
      return node != null && node.right != null;
    }

    /**
     * Return the element in the forward direction.
     */
    public Object forwardElement()
    {
      try
        {
          node = node.getNextNode();
          return node.getObject();
        }
      catch (NullPointerException e)
        {
          e.printStackTrace();
          throw new NoSuchElementException(e.getMessage());
        }
    }

    /**
     * Return true if there are more elements in the backward direction.
     */
    public boolean hasBackwardElement()
    {
      return node != null && node.left != null;
    }

    /**
     * Return the element in the backward direction.
     */
    public Object backwardElement()
    {
      try
        {
          node = node.getPreviousNode();
          return node.getObject();
        }
      catch (NullPointerException e)
        {
          e.printStackTrace();
          throw new NoSuchElementException(e.getMessage());
        }
    }

  }


/*-------------------------------------------------------------------*
 * This insert implementation was taken from the SAS V8 AVL tree
 * implementation by Robert Ray. This implementation is nice
 * because it maintains threading -- all the nodes are linked
 * together, which makes linear traveral (enumeration) easy.
 * I changed the representation of threaded nodes, since we
 * can't to bit twiddling with pointers in Java. So, the state
 * byte stores both the balance and the flags for whether
 * the leftLink/rightLinks are thread links or child links.
 * See:
 *
 *     /sas/m800/base/src/sfmavl.c
 *     /sas/m800/base/h/sfmavl.h
 *     /sas/m800/base/h/sfmavla.h
 *     /sas/m800/base/h/sfmavlf.h
 *
 * Also, Lewis Church has an XOB AVL tree implementation
 * which supports node deletion:
 *
 *     /sas/m800/hostcm/src/wobcavl.c
 *     /sas/m800/hostcm/h/wobcavl.h
 *
 * sasdjb 31Mar99
 *-------------------------------------------------------------------*/


/*-------------------------------------------------------------------*
 * COPYRIGHT (C) 1996 by SAS Institute Inc, Cary NC 27511 USA
 *
 * NAME:     sfmavla.h
 * DATE:     1996/02/08
 * AUTHOR:   Robert Ray
 * SUPPORT:  sasrsr - Robert Ray
 * STYLE:    sasrsr
 *
 * PURPOSE: Common src code to build right and left in-order threaded
 *          AVL tree insertion functions with inlined code to do
 *          comparisons of a single key type.  Based on ztrebld (which
 *          was taken from Knuth vol 3).  With these trees, it is
 *          possible to have multiple simultanious traversals and also
 *          to do a "find" followed by a "next" or "prev"
 *          operation. In other words, in-order traversals may start
 *          with any node. The only context that needs to maintained
 *          during a traversal is the current tree node ptr.
 *
 *          Also added for this version the "cmap" comparison caching
 *          code which allows the balancing phase of the code to avoid
 *          re-comparing keys that have already been compared during
 *          the initial search loop.
 *
 * CREDITS: Adelson-Velsky & Landis: The original height-balanced
 *             binary tree algorithm.
 *          Perlis & Thorton: Originators of binary tree threading.
 *          Brian Schellenberger: Original SAS version of portable C
 *             AVL routins.
 *          Tom Truscott: Code simplifications, bit packing of thread
 *             info into links.
 *
 * See sfmavl.h for typedefs and other macros.  See also sfmavlf.h for
 * src code to find nodes in this tree.  For examples of using this
 * code, see afmavl.c which also contains two utility functions
 * avl2list and next_node for working with the resulting trees.
 *
 *-------------------------------------------------------------------*/

 // -------------------------------- Node Removal --------------------------------//

  // Node deletion was adapted from from _Algorithms + Date Structures = Programs_
  // by Nicklaus Wirth, pp223-225, Prentice Hall 1976

  // Because Java does not support output parameters (Pascal's var parameters),
  // we need to modifiy the code slightly, since Wirth's code uses two
  // output parameters for several methods. Instead, we pass a structure
  // object, a Env instance, which contains the 'var' parameters.
  // This is passed as the environment 'env' which alsi serves to simulate
  // nested Pascal's procedure environments.
  // The methods can update the p, q and h fields to emulate updating
  // the p and h output parameters.

  // I also added code to maintain the left/right thread pointers
  // which allows fast traversal/enumeration.
  // sasdjb 07Apr99

  /**
   * Env keeps track of local state while navigating a tree for deletion
   * it tracks balance and temporary nodes.
   */
  static class Env
  {
    /** the node to operate on */
    AVLNode p;
    /** the height of a subtree has been reduced */
    boolean h;
    /** comparator to use */
    Comparator comparator;
    /** true if remove() deleted an item */
    boolean deleted;
    /** specific node id */
    long id;
    Env(AVLNode node, boolean h, Comparator comparator, long id)
      {
        this.p = node;
        this.h = h;
        this.comparator = comparator;
        this.deleted = false;
        this.id = id;
      }
  }


  /**
   * Remove an object <var>x</var> from the AVL tree rooted at <var>root</var>
   * using the specified comparator.
   */
  static AVLNode[] remove(AVLNode root, Object x, Comparator comparator, long id)
    {
      Env env = new Env(root, false, comparator, id);
      remove(x, env, null, null);
      if (env.deleted)
        return new AVLNode[] { env.p };
      else
        return null;
    }

  /**
   * Remove the node containing the object (env.object)
   * starting at the root node, env.p.
   * This leaves env.p pointing at the new root.
   */
  static void remove(Object x, Env env, AVLNode leftThread, AVLNode rightThread) // delete(x, AVLNode &p, boolean &h)
    {
      // try { if (DEBUG) enter("remove \"" + x + "\"", env);
      AVLNode p;
      if (env.p == null)
        {
          env.h = false;
          return;
        }
      int result = env.comparator.compare(x, env.p.getObject());
      // if (DEBUG) System.out.println(indent + "compare(" + x + "," + env.p.getObject() + ") result: " + result);
      if (result < 0)
        {
          p = env.p;
          try { // Convert pass by reference into pass by env block.
            env.p = env.p.getLeftChild();
            remove(x, env, leftThread, p); // delete(p.left, h)
            AVLNode newRoot = env.p;
            p.setLeftChild(newRoot, leftThread);
            if (newRoot != null && newRoot.hasRightThread())
              newRoot.setRightThread(p);
          }
          finally { env.p = p; }
          if (env.h)
            balance1(env);
        }
      else if (result > 0)
        {
          p = env.p;
          try { // Convert pass by reference into pass by env block.
            env.p = env.p.getRightChild();
            remove(x, env, p, rightThread); // delete(p.right, h)
            AVLNode newRoot = env.p;
            p.setRightChild(newRoot, rightThread);
            if (newRoot != null && newRoot.hasLeftThread())
              newRoot.setLeftThread(p);
          }
          finally { env.p = p; }
          if (env.h)
            balance2(env);
        }
      else // delete p
        {
          env.deleted = true;
          // if (DEBUG) System.out.println(indent + "remove p " + env.p);
          AVLNode q = env.p;
          if ( ! q.hasRightChild() )
            {
              env.p = q.getLeftChild();
              if (env.p != null)
                // added thread maintenance. q.left child cannot
                // have a right child (since q does not have a right child,
                // q.left has a maximum height of 1 due to balance)
                env.p.setRightThread(rightThread);
              env.h = true;
            }
          else if ( ! q.hasLeftChild() )
            {
              env.p = q.getRightChild();
              if (env.p != null)
                // added thread maintenance. q.right child cannot
                // have a left child (since q does not have a left child,
                // q.right has a maximum height of 1 due to balance)
                env.p.setLeftThread(leftThread);
              env.h = true;
            }
          else
            {
              p = env.p;
              try {  // Convert pass by reference into pass by env block.
                env.p = q.getLeftChild();
                del(q, rightThread, env); // del(q.left, h)
                q.setLeftChild(env.p, leftThread);
              }
              finally { env.p = p; }
              if (env.h)
                balance1(env);
            }
          // dispose(q)
        }
      // } finally { if (DEBUG) leave("remove \"" + x + "\"", env); }
    }

  private static void del(AVLNode q, AVLNode rightThread, Env env) // del(AVLNode &p, boolean &h)
    { // env.h == false
      // try { if (DEBUG) enter("del", env);
      AVLNode p = env.p;
      AVLNode right = p.getRightChild();
      if (right != null)
        {
          try {
            env.p = right;
            del(q, rightThread, env);  // del(r.right, h)
            p.setRightChild(env.p, q); // maintain thread
          }
          finally { env.p = p; }
          if (env.h)
            balance2(env);
        }
      else
        {
          q.setObject(env.p.getObject());
          // q.setCount(env.r.count());
          AVLNode leftChild = env.p.getLeftChild();
          if (leftChild != null)
            leftChild.setRightThread(q);
          env.p = leftChild;
          env.h = true;
        }
      // } finally { if (DEBUG) leave("del", env); }
    }


  private static void balance1(Env env) // balance(AVLNode &p, boolean &h)
    { // env.h == true, left branch has become less high
      // try { if (DEBUG) enter("balance1", env);
      AVLNode p1, p2;
      int b1, b2;
      switch (env.p.balance())
        {
        case -1:
          env.p.setEvenBalance();     // p.bal = 0
          break;
        case 0:
          env.p.setPositiveBalance(); // p.bal = +1
          env.h = false;
          break;
        case 1:
          // rebalance
          p1 = env.p.getRightChild();
          b1 = p1.balance();
          if (b1 >= 0)
            { // single RR rotation
              // p1, p1 are non-null
              env.p.setRightChild(p1.getLeftChild(), p1);
              p1.setLeftChild(env.p); // no threading necessary; env.p != null
              if (b1 == 0)
                {
                  env.p.setPositiveBalance(); // p.bal  = +1
                  p1.setNegativeBalance();    // p1.bal = -1
                  env.h = false;
                }
              else
                {
                  env.p.setEvenBalance(); // p.bal = 0
                  p1.setEvenBalance();    // p1.bal = 0
                }
              env.p = p1;
            } // end else
          else
            { // double RL rotation
              // p1, p1, p2 are non-null
              p2 = p1.getLeftChild();
              b2 = p2.balance();
              p1.setLeftChild(p2.getRightChild(), p2);
              p2.setRightChild(p1); // no threading necessary; p1 != null
              env.p.setRightChild(p2.getLeftChild(), p2);
              p2.setLeftChild(env.p); // no threading necessary; p != null
              if (b2 == +1)
                env.p.setNegativeBalance();   // p.bal = -1
              else
                env.p.setEvenBalance();       // p.bal = +1
              if (b2 == -1)
                p1.setPositiveBalance();      // p1.bal = +1
              else
                p1.setEvenBalance();          // p1.bal = 0
              env.p = p2;
              p2.setEvenBalance();            // p2.bal = 0
            }
        }
      // } finally { if (DEBUG) leave("balance1", env); }
    } // balance1

  private static void balance2(Env env) // balance2(AVLNode &p, boolean &h)
    {
      // try { if (DEBUG) enter("balance2", env);
      // env.h == true, right branch has become less high
      AVLNode p1, p2;
      int b1, b2;
      switch (env.p.balance())
        {
        case 1:
          env.p.setEvenBalance(); // p.bal = 0
          break;
        case 0:
          env.p.setNegativeBalance(); // p.bal = -1
          env.h = false;
          break;
        case -1:
          // rebalance
          p1 = env.p.getLeftChild();
          b1 = p1.balance();
          if (b1 <= 0)
            { // single LL rotation
              // p, p1 are non-null
              env.p.setLeftChild(p1.getRightChild(), p1);
              p1.setRightChild(env.p); // no threading necessary; p != null
              if (b1 == 0)
                {
                  env.p.setNegativeBalance(); // p.bal = -1
                  p1.setPositiveBalance();    // p1.bal = +1
                  env.h = false;
                }
              else
                {
                  env.p.setEvenBalance(); // p.bal = 0
                  p1.setEvenBalance();    // p1.bal = 0
                }
              env.p = p1;
            }
          else
            { // double LR rotation
              // p, p1, p2 are non-null
              p2 = p1.getRightChild();
              b2 = p2.balance();
              p1.setRightChild(p2.getLeftChild(), p2);
              p2.setLeftChild(p1); // no threading necessary; p != null
              env.p.setLeftChild(p2.getRightChild(), p2);
              p2.setRightChild(env.p); // no threading necessary; p != null
              if (b2 == -1)
                env.p.setPositiveBalance();  // p.bal = +1
              else
                env.p.setEvenBalance();      // p.bal =  0
              if (b2 == +1)
                p1.setNegativeBalance();     // p1.bal = -1
              else
                p1.setEvenBalance();         // p1.bal =  0
              env.p = p2;
              p2.setEvenBalance();           // p2.bal = 0
            }
        }
      // } finally { if (DEBUG) leave("balance2", env); }
    }

  // for DEBUG
// DEBUG //  private static void indent()
// DEBUG //    {
// DEBUG //      if (DEBUG)
// DEBUG //        indent += "  ";
// DEBUG //    }
// DEBUG //
// DEBUG //  // for DEBUG
// DEBUG //  private static void dedent()
// DEBUG //    {
// DEBUG //      if (DEBUG)
// DEBUG //        indent = indent.substring(2, indent.length());
// DEBUG //    }
// DEBUG //
// DEBUG //  private static void enter(String label, Env env)
// DEBUG //      {
// DEBUG //        if (DEBUG)
// DEBUG //          {
// DEBUG //            indent();
// DEBUG //            System.out.println(indent + "Enter: " + label + "(root: " + (env.p == null ? "null" : env.p.getObject()) + ", h: " + env.h + ")");
// DEBUG //            // DEBUG // printTree(env.p);
// DEBUG //          }
// DEBUG //      }
// DEBUG //
// DEBUG //  private static void leave(String label, Env env)
// DEBUG //      {
// DEBUG //        if (DEBUG)
// DEBUG //          {
// DEBUG //            dedent();
// DEBUG //            System.out.println(indent + "Leave: " + label + "(root: " + (env.p == null ? "null" : env.p.getObject()) + ", h: " + env.h + ")");
// DEBUG //            // DEBUG // printTree(env.p);
// DEBUG //          }
// DEBUG //      }

// DEBUG //   private static void printTree(AVLNode root)
// DEBUG //     {
// DEBUG //       if ( DEBUG )
// DEBUG //         {
// DEBUG //           if (root == null)
// DEBUG //             {
// DEBUG //               System.out.println(indent + "null");
// DEBUG //               return;
// DEBUG //             }
// DEBUG //           Vector v = TestAVLTree.printHorizontalTree(root, false, false);
// DEBUG //           for (int i = 0, vLen = v.size(); i < vLen; i++)
// DEBUG //             {
// DEBUG //               System.out.println(indent + v.elementAt(i));
// DEBUG //             }
// DEBUG //           System.out.println();
// DEBUG //         }
// DEBUG //     }


  // -------------------------------- Search for a node -------------------------------- //


  /**
   * Search for and return a value in the AVL tree
   * @param root the root of the tree
   * @param object The object to search for
   * @param comparator comparison function
   * @return The first node whose <code>getObject()</code>
   *         equals the <var>object</var>, or null if none was found
   */
  public static AVLNode find( AVLNode    root,
                              Object     object,
                              Comparator comparator )
    {
      while ( root != null )
        {
          int result = comparator.compare( object, root.getObject() );

          if ( result < 0 )       /*-- object < root.getObject()  --*/
            root = root.getLeftChild();
          else if ( result > 0 )  /*-- object > root.getObject()  --*/
            root = root.getRightChild();
          else
            break;               /*-- object == root.getObject() --*/
        }
      return root;
    }


  // -------------------------------- Default Object methods -------------------------------- //


  /**
   * Return the string representation of this node.
   * This calls {@link #toString(boolean,boolean) toString(false, false)}
   * @return the string representation of this node.
   */
  public String toString()
    {
      return toString(false, false);
    }

  /**
   * Return the string representation of this node.
   * @param printValue if true, include <code>getObject().toString()</code>
   * @param showThreads if true, show left/right threading information
   * @return the string representation of this node.
   */
  public String toString(boolean printValue, boolean showThreads)
    {
      StringBuffer b = new StringBuffer();
      if (printValue)
        b.append(getObject());
      b.append("#");
      b.append(id);
      switch (balance())
        {
        case -1: b.append("[-]"); break;
        case  0: b.append("[ ]"); break;
        case  1: b.append("[+]"); break;
        default: b.append("[****");
                 b.append(balance());
                 b.append("****]"); break;
        }
      if (showThreads)
        {
          if (hasLeftThread())
            {
              b.append(" left:");
              b.append(getLeftThread().toString(printValue, false));
            }
          if (hasRightThread())
            {
              b.append(" right:");
              b.append(getRightThread().toString(printValue, false));
            }
        }
      return b.toString()/*I18nOK:MTH*/;
    }

  /**
   * Clone this AVL tree node.
   * The subtrees or the object held in this node are not cloned.
   * @return a clone of this node.
   */
  public Object clone() throws CloneNotSupportedException
    {
      AVLNode clone = (AVLNode) super.clone();
      clone.init();
      return clone;
    }

  /**
   * Clone this AVL tree node.
   * This method clones using depth first traversal of the tree.
   * @param <var>nodes</var>
   * <var>nodes</var> is used for two purposes. As an input parameter,
   * <var>nodes</var>[0] contains the root of the cloned left sibling tree.
   * the leftmost descendent clone of this current node will
   * set it's left link to thi sibling.
   * <var>nodes</var> is also used on output: after cloning a node,
   * <var>nodes</var>[0] points to the clone's leftmost descendent and
   * <var>nodes</var>[0] points to the clone's rightmost descendent.
   * Thus, the cloned subtrees can hook up theirgh left/right threads
   * as we progress.
   * <p>
   * The caller must pass a valid AVLNode noreRef array
   * of two elements.
   *
   * @return a clone of this node.
   * @see #clone()
   */
  AVLNode clone(AVLNode nodes[]) throws CloneNotSupportedException
    {
      AVLNode clone = (AVLNode) clone();
      if (hasLeftChild())
        {
          nodes[1] = clone;
          AVLNode leftClone = getLeftChild().clone(nodes);
          nodes[0] = leftClone;
          clone.setLeftChild(leftClone);
          nodes[1].setRightThread(clone);
        }
      else
        {
          clone.setLeftThread(nodes[0]);
          nodes[0] = clone;
        }

      if (hasRightChild())
        {
          AVLNode temp = nodes[0];
          nodes[0] = clone;
          AVLNode rightClone = getRightChild().clone(nodes);
          clone.setRightChild(rightClone);
          nodes[0] = temp;
        }
      else
        {
          nodes[1] = clone;
        }
      return clone;
    }

  // initial deep clone implementation. Kind of a memory/time hog due
  // to use of DeepCloneSupport object (which uses LongKeyDictionary)

//  public Object clone(DeepCloneSupport support, boolean deeply) throws CloneNotSupportedException
//  {
//    AVLNode clone = (AVLNode) clone();
//    if (support != null)
//      support.register(this, clone); // do this before recursion
//    if (deeply)
//      {
//        try
//          {
//            // Get the clone of the left & right links. (They may have been
//            // cloned already.)
//            if (left != null)
//              clone.left = (AVLNode) support.clone(left, deeply);
//            if (right != null)
//              clone.right = (AVLNode) support.clone(right, deeply);
//          }
//        catch (NullPointerException e)
//          { throw new CloneNotSupportedException("support == null"); }
//      }
//    return clone;
//  }


  // Serialization
  private void readObject(java.io.ObjectInputStream stream)
    throws java.io.IOException, ClassNotFoundException
    {
      stream.defaultReadObject();
      // read left node first, before calling init(),
      // so that the left node is assigned a smaller id
      // than this node. This ensures correct ordering of
      // items whose getObject() values are equal.
      left = (AVLNode) stream.readObject();
      init();
      right = (AVLNode) stream.readObject();
    }

  // Serialization
  private void writeObject(java.io.ObjectOutputStream stream)
    throws java.io.IOException
    {
      stream.defaultWriteObject();
      // write left link, then right link.
      // See readObject for details.
      stream.writeObject(left);
      stream.writeObject(right);
    }

      // DEBUG //   public static void main(String argv[])
      // DEBUG //     {
      // DEBUG //       String test = "insertDelete";
      // DEBUG //       if (System.getProperty("TEST") != null)
      // DEBUG //         test = System.getProperty("TEST");
      // DEBUG //       com.sas.test.Test.main(new String[] { "com.sas.collection.TestAVLTree", test });
      // DEBUG //     }

}
