
package com.sas.collection;

// This implementation was taken from the SAS V8 AVL tree
// implementation by Robert Ray. This implementation is nice
// because it maintains threading -- all the nodes are linked
// together, which makes linear traveral (enumeration) easy.

import com.sas.util.Comparator;
import com.sas.util.CompositeComparator;
import java.io.Serializable;
import com.sas.PublicClonable;
import java.util.NoSuchElementException;
import java.util.Enumeration;
import com.sas.util.Enumerable;
import com.sas.util.Countable;
// import com.sas.DeepCloneSupport;

/**
 * An AVL tree, named after Adelson-Velskii and Landis.
 * An AVL Tree consists of a root AVLNode and
 * a comparator which is used to add new items or to search
 * for values.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @since 1.2
 */
public class AVLTree
       implements Serializable,
                  PublicClonable,
                  Enumerable,
                  Countable
{
  static final long serialVersionUID = 90136725763567088L;

  public static final String RB_KEY = "AVLTree.";

  private AVLNode root;
  private AVLNode here;
  private int index;
  private Comparator comparator;
  private boolean duplicatesAllowed;
  private int count;

  public AVLTree()
    {
      this(null);
    }

  public AVLTree(Comparator comparator)
    {
      this(comparator, true);
    }

  public AVLTree(Comparator comparator, boolean duplicatesAllowed)
    {
      setComparator(comparator);
      setDuplicatesAllowed(duplicatesAllowed);
      reset();
    }

  /**
   * Return the number of nodes in the tree.
   * @return the number of nodes in the tree.
   */
  public int count()
    {
      return count;
    }


  /**
   * Set the comparator. This will also empty the tree
   * and add the items back in.
   * @param newComparator the new comparator.
   */
   public void setComparator(Comparator newComparator)
   {
     if (com.sas.util.Util.equal(comparator, newComparator))
        return;
     comparator = newComparator;
     if (count() > 0)
       {
         Enumeration e = getItems();
         removeAll();
         while (e.hasMoreElements())
           add(e.nextElement());
       }
   }

  /**
   * Remove all items from the tree.
   */
  public void removeAll()
    {
      root = null;
      count = 0;
      reset();
    }

  private void reset()
    {
      here = null;
      index = -1;
    }

//  public synchronized Object get(int i)
//    {
//      if (i < 0)
//        throw new IndexOutOfBoundsException(i + " < 0");
//      if ( i > count() )
//        throw new IndexOutOfBoundsException(i + " > " + count());
//
//      int dist1 = i - 0;        // from first node, forward
//      int dist2 = count() - i; // from last node, backward
//      int dist3;               // from index to node, forward or backward as may be.
//      if (index < i)
//        dist3 = i - index;
//      else
//        dist3 = index - i;
//      if (dist1 <= dist2 && dist1 <= dist3)
//        {
//          index = 0;
//          for (here = firstNode(); index < i ; index++)
//            here = here.getNextNode();
//        }
//      else if (dist2 <= dist1 && dist2 <= dist3)
//        {
//          index = count() - 1;
//          for (here = lastNode(); index > i; index--)
//            here = here.getPreviousNode();
//        }
//      else if (index <= i)
//        {
//          for (; index < i ; index++)
//            here = here.getNextNode();
//        }
//      else // (index > i)
//        {
//          for (; i < index ; index--)
//            here = here.getPreviousNode();
//        }
//      return here.getObject();
//    }

  public AVLNode root()
    {
      return root;
    }

  public AVLNode firstNode()
    {
      AVLNode node = root();
      if ( node == null )
        return null;
      return node.leftMostDescendent();
    }

  public AVLNode lastNode()
    {
      AVLNode node = root();
      if ( node == null )
        return null;
      return node.rightMostDescendent();
    }

   /**
    * Return the comparator used to compare items.
    * @return the tree's comparator.
    */
    public Comparator getComparator()
    {
        return comparator;
    }

  /**
   * Set the whether the tree allows duplicate values to be inserted.
   * @param duplicatesAllowed if false, you cannot add an item
   * that already exists to an AVLTree.
   */
   public void setDuplicatesAllowed(boolean duplicatesAllowed)
   {
     this.duplicatesAllowed = duplicatesAllowed;
   }

   /**
    * Return true if duplicate values may be added to the AVL tree.
    * @return true if duplicate values may be added to the AVL tree.
    */
    public boolean isDuplicatesAllowed()
    {
        return duplicatesAllowed;
    }

  /**
   * Add the <var>object</var> into a tree rooted at <var>root</var>,
   * using the comparator.
   * @param object the object to add into the tree
   * @param root the root node. This may be null
   * @param comparator the Comparator to use when comparing items.
   * @return the new root, or null if a item that equals node.key already
   *         exists in the tree and duplicatesAllowed is false
   */
  public void add(Object object)
    {
      AVLNode newRoot = AVLNode.add(new AVLNode(object), root, getComparator(), isDuplicatesAllowed());
      if (newRoot != null)
        {
          root = newRoot;
          count ++;
        }
    }

  /**
   * Test if an object exists in the tree.
   * @param object a value to search for. The tree's comparator
   * is used; the <var>object</var> will be the first parameter.
   * @return true if the object exists.
   */
  public boolean contains(Object object)
  {
    return get(object) != null;
  }

  /**
   * Return the object in the tree which is equal to <var>query</var>
   * @param query a value to search for. The tree's comparator
   * is used; the <var>query</var> object will be the first parameter to
   * the comparator's compare method; the item in the tree will be the second parameter.
   * @return a non-null value if an object equal to <var>query</var> exists
   */
  public Object get(Object query)
  {
    return AVLNode.find(root, query, comparator);
  }

  /**
   * Remove an object from the tree.
   * @param object a value to remove. The tree's comparator
   * is used; the <var>object</var> will be the first parameter.
   * @return true if the object was removed.
   */
  public synchronized boolean remove(Object object)
    {
      if (root == null)
        return false;
      AVLNode del[] = AVLNode.remove(root, object, getComparator(), -1);
      boolean deleted = del != null;
      if (deleted)
        {
          root = del[0];
          count--;
        }
      return deleted;
    }

  /**
   * Get an enumeration of the tree contents
   */
  public Enumeration getItems()
    {
      return getItems(false);
    }

  /**
   * Get an enumeration of the tree contents
   */
  public Enumeration getItems(int start, int end)
    {
      if (root == null)
        // create an empty enumeration
        return new java.util.Vector().elements();
      else if (start == 0 && end == count())
        return root.getItems(false);
      else
        return root.getItems(start, end);
    }

  /**
   * Get an enumeration of the tree contents.
   * @param backwards if true, the enumeration
   * traverses the tree backwards, from the end to the beginning.
   */
  public AVLNode.Cursor getItems(boolean backwards)
    {
      if (root != null)
        return root.getItems(backwards);
      else
        return new AVLNode.Cursor(null, backwards);
    }

//  /**
//   * Get an enumeration of the tree contents.
//   * @param backwards if true, the enumeration
//   * traverses the tree backwards, from the end to the beginning.
//   * @param start Start the enumeration after <var>start</var> items.
//   * If <var>backwards</var> is true, this is <var>start</var> from the
//   * end of the tree, else it is counted from the beginning.
//   * @param end End the enumeration after <var>end</var> items.
//   * If <var>backwards</var> is true, this is <var>end</var> from the
//   * end of the tree, else it is counted from the beginning.
//   * For example, <code>getItems(0, tree.count(), true)</code>returns all
//   * items in reverse order.  To return the last 5 items in
//   * reverse order, use <code>getItems(0, 5, true)</code> and to get the
//   * first 5 items in reverse order, use <code>getItems(count()-5, count(), true)</code>
//   */
//  public Enumeration getItems(int start, int end, boolean backwards)
//    {
//      if (root == null)
//        // create an empty enumeration
//        return new java.util.Vector().elements();
//      return root.getItems(start, end, backwards);
//    }

  /**
   * Clone this AVL tree node.
   * The subtrees or the object held in this node are not cloned.
   * @return a clone of this node.
   * @see com.sas.collection.AVLNode#clone()
   */
  public Object clone() throws CloneNotSupportedException
    {
      AVLTree clone = (AVLTree) super.clone();
      if (root != null)
//        clone.root = (AVLNode) clone.root.clone(new DeepCloneSupport(), true);
        {
          AVLNode nodeRef[] = new AVLNode[2];
          clone.root = root.clone(nodeRef);
        }
      clone.reset();
      return clone;
    }

}
