package com.sas.collection;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.io.Serializable;
/**
 * An Enumeration which filters another Enumeration
 * by integer indices.
 * IndexedEnumeration filters an input Enumeration
 * such that only the items with idices greater than or
 * equal to a <var>start</var> index and less than a <var>end</var>
 * index are returned.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @since 1.2
 * @version 1.0
 */

public class IndexedEnumeration
  implements Enumeration,
             Serializable
{
  static final long serialVersionUID = 426674497607469026L;

  /**
   * The actual <code>Enumeration</code> that this <code>IndexedEnumeration</code> filters.
   */
  private Enumeration e;

  /**
   * Becomes true after first call to skipToStart, after skipping the first <var>start</var>
   * values from the original Enumeration.
   */
  private boolean started;
  private int start, index, end;

  /**
   * Construct an IndexedEnumeration from another IndexedEnumeration
   * such that only items from the input Enumeration e are returned
   * if their absolute index, starting at 0, is greater than or equal
   * to <var>start</var> and less than <var>end</var>.
   */
  public IndexedEnumeration(Enumeration e, int start, int end)
  {
    this.e = e;
    this.start = start;
    this.end = end;
    this.index = 0;
    this.started = false;
  }

  /**
   * Return the current index into the original
   * enumeration
   */
  public int getIndex()
  {
    if (started)
      return index;
    else
      return start;
  }

  /**
   * Return true if there are more items in the enumeration.
   * This is true if the index is in the range <code>[<var>start</var>, <var>end</var>)</code>
   */
  public boolean hasMoreElements()
  {
    skipToStart();
    return (start <= index) && (index < end) && e.hasMoreElements();
  }

  /**
   * Returns the next element from the enumeration.
   * @return the next element from the enumeration.
   */
  public Object nextElement()
  {
    if (! hasMoreElements() )
      throw new NoSuchElementException();
    index++;
    return e.nextElement();
  }

  /**
   * Skip over the first <var>start</var> returned values from nextElement()
   */
  private void skipToStart()
  {
    if (started)
      return;
    started = true;

    while (index < start && e.hasMoreElements())
      {
        index++;
        Object nextElement = e.nextElement();
      }
  }

  /**
   * Return the underlying Enumeration.
   */
  public Enumeration getEnumeration()
  {
    return e;
  }

  /**
   * Return the original start index
   */
  public int getStart()
  {
    return start;
  }

  /**
   * Return the original end index
   */
  public int getEnd()
  {
    return end;
  }
}
