// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package  com.sas.collection;

import java.util.NoSuchElementException;
import java.io.Serializable;
import java.io.IOException;

import com.sas.PublicClonable;
import com.sas.util.Util;

/**
 * <code>LongKeyDictionary</code> stores and retrieve objects by a long integer
 * key. It has lookup behavior like a hashtable, but keys must
 * be long integers, not Objects. Note that the <code>LongKeyDictionary</code>
 * can only contain one item per long key.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public class LongKeyDictionary
       implements PublicClonable,
                  Serializable
{
  static final long serialVersionUID = 7961991170118102709L;

  /**
   * The NO_SUCH_ELEMENT object is a useful item to use as a default
   * object in get(long key, Object default) calls as a way
   * of indicating that no item with the specified key was found.
   * You could use two method calls:
   * <pre>
   *    if (dict.containsKey(key)) {
   *       Object object = dict.get(key);
   *       <em>process the object</em>
   *       }
   *    else
   *       <em>object not found</em>
   * </pre>
   * or you can use get(key, NO_SUCH_ELEMENT) and use a less
   * expensive comparison to NO_SUCH_ELEMENT to indicate
   * the value does not exist:
   * <pre>
   * Object object = dict.get(key, PropertyBag.NO_SUCH_ELEMENT);
   * if (object == dict.NO_SUCH_ELEMENT)
   *    <em>object not found</em>
   * else
   *    <em>process the object</em>
   * </pre>
   */
  public static final Object NO_SUCH_ELEMENT = new String("NO_SUCH_ELEMENT");

  // items is an array Entry objects, each items[slot] representing a chain of
  // key/value pairs of objects with the same hash code, i.e. items
  // with equal values of slot = hashSlot(long key).
  // The first item in the list is a chain head that does not contain values;
  // the empty value allows the Entry previous(...) methods to return the preceeding
  // item in the chain, allowing easy removal of entries without requiring doubly linked
  // lists.
  //
  // items[] is serialized manually, to avoid recursive serialization of the linked list of Entry objects.
  private transient Entry items[];

  /**
   * The default load factor.
   */
  public static float DEFAULT_LOAD_FACTOR = 0.75f;

  // this dictionary's load factor
  private float loadFactor;
  // integer value of (capacity * loadFactor)
  private transient int max;
  // the number of items in the table.
  private transient int count;
  // table of prime numbers for optimal table sizes
  private static final int tableSizes[] =
   { 101, 211, 401, 607, 1009, 2003, 4001, 8009,
     16001, 19997, 40009, 80021, 160001, 199967,
     499973, 999983};
  private int itemsLength;  // items.size

  // compute a slot index from a key
  private int hashSlot(long key)
    {
      int slot = (int) key % itemsLength;
      if (slot < 0) slot = -slot;  // faster that Math.abs(); this is a critical method
      return slot;
    }

  /**
   * Construct an empty LongKeyDictionary.
   */
  public LongKeyDictionary()
    {
      this(tableSizes[0], DEFAULT_LOAD_FACTOR);
    }

  /**
   * Construct a <code>LongKeyDictionary</code> with an initial capacity and
   * load factor. When the <code>count()</code> exceeds <code>(<var>capacity</var> * <var>loadFactor</var>)</code>,
   * the dictionary grows to accommodate more items.
   * @param capacity the minimum capacity of the dictionary.
   * @param loadFactor a number greater than 0; when this dictionary
   * grows to beyond <code>(<var>capacity</var> * <var>loadFactor</var>)</code>, the
   * dictionary automatically grows to accommodate more items.
   */
  public LongKeyDictionary(int capacity, float loadFactor)
    {
      if (capacity <= 0 || loadFactor <= 0.0f)
        throw new IllegalArgumentException();
      this.loadFactor = loadFactor;
      itemsLength = optimalSize(capacity);
      items = new Entry[itemsLength];
      max = (int) (itemsLength * loadFactor);
    }

  /**
   * Clone a LongKeyDictionary.
   * @return a clone of this dictionary.
   * @exception java.lang.CloneNotSupportedException if the
   *            clone fails.
   */
   public Object clone() throws CloneNotSupportedException
    {
      LongKeyDictionary clone = (LongKeyDictionary) super.clone();
      clone.items = new Entry[itemsLength];
      for (int i = 0; i < itemsLength; i++)
        {
          Entry entry = items[i];
          if (entry != null)
            clone.items[i] = (Entry) entry.clone();
        }
      return clone;
    }

  /**
   * Compute the optimal size, based on the desired size and the tableSizes
   * @param desiredSize desired (minimum) capacity
   * @return the optimal size to accommodate the minimum capacity.
   * There is a limit, though, as defined by the tableSizes array.
   * The internal vector never exceeds <code>tableSizes[tableSizes.length - 1]</code>
   */
  private int optimalSize(int desiredSize)
    {
      for (int i = 0; i < tableSizes.length - 1; i++)
        {
          if (desiredSize <= tableSizes[i])
            return tableSizes[i];
        }
      return tableSizes[tableSizes.length - 1];
    }

  /**
   * Return the Entry which preceeds the Entry containg value.
   * This is useful, since it allows constant time access to the
   * value, as well as constant time removal of the value.
   * @param value the value to search for. This method uses the
   *        <code>equals</code> method to compare value to
   *        values in this dictionary.
   * @return the Entry which preceeds the Entry containing the value
   * The Entry objects are chained together. Returns null if the
   * value is not found.
   * @see #previous(java.lang.Object, boolean)
   */
  private Entry previous(Object value)
    {
      return previous(value, true);
    }


  /**
   * Return the Entry which preceeds the Entry containg value.
   * This is useful, since it allows constant time access to the
   * value, as well as constant time removal of the value.
   * @param value the value to search for
   * @param whenEquals if true, use </code>equals()</code>
   *        to compare values, else use <code>==</code>
   * @return the Entry which preceeds the Entry containing the value
   * The Entry objects are chained together. Returns null if the
   * value is not found.
   */
  private Entry previous(Object value, boolean whenEquals)
    {
      for (int i = 0; i < itemsLength; i++)
        {
          Entry head = items[i], entry;
          if (head != null)
            {
              entry = head.previous(value, whenEquals);
              if (entry != null)
                return entry; // moveNextToFront(entry, head)
            }
        }
      return null;
    }

//   /**
//    * Move the following item to the head of the list. This
//    * is a small optimization which allows the item to be found
//    * more quickly the next time, since most uses are operations
//    * like <code> if (dict.contains(value)) x = dict.remove(value)
//    * @param previous the Entry preceeding the entry we wish to
//    * move to the front.
//    * @param head the Entry at the head of the chain. When done,
//    * head.next will point to previous.next
//    * @return head
//    */
//   private Entry moveNextToFront(Entry previous, Entry head)
//     {
//       if (previous == null || previous == head)
//         return previous;
//       Entry current = previous.next;
//       previous.next = current.next;
//       current.next = head.next;
//       head.next = current;
//       return head;
//     }

  /**
   * Return the Entry which preceeds the Entry containg value.
   * This is useful, since it allows constant time access to the
   * value, as well as constant time removal of the value.
   * @param value the value to search for
   * @return the Entry which preceeds the Entry containing the value
   * The Entry objects are chained together. Returns null if the
   * value is not found.
   */
  private Entry previous(long key)
    {
      final int slot = hashSlot(key);
      Entry head = items[slot], entry;
      if (head != null)
        {
          entry = head.previous(key);
          return entry; // moveNextToFront(entry, head)
        }
      return null;
    }

  /**
   * Remove the value and its key from this dictionary.
   * This only removes the first occurrence of value.
   * @param value a value to remove.
   * @return <code>true</code> if the item was found and removed,
   *         <code>false</code> if it does not exist.
   * @see #removeAll(java.lang.Object)
   */
   public boolean remove(Object value)
    {
      Entry prev = previous(value);
      if (prev != null)
        {
          prev.next = prev.next.next;
          count--;
          return true;
        }
      return false;
    }

  /**
   * Remove the key and item associated with the key from this dictionary.
   * This method does nothing if there is no such value.
   * @param key remove the item associated with this key
   * @return <code>true</code> if the item was found and removed,
   *         <code>false</code> if it does not exist.
   */
   public boolean removeAt(long key)
    {
      Entry prev = previous(key);
      if (prev != null)
        {
          prev.next = prev.next.next;
          count--;
          return true;
        }
        return false;
    }

  /**
   * Store a new value in this dictionary.
   * @param key the key that is used to lookup the value
   * @param value the value to store in this dictionary, associated
   *        with the <var>key</var>. If there is already a value associated
   *        with the <var>key</var>, it is replaced.
   */
   public void set(long key, Object value)
    {
      Entry prev = previous(key);
      if (prev != null)
        prev.next.value = value;
      else
        {
          count ++;
          if (count >= max)
            rehash();
          int slot = hashSlot(key);
          if (items[slot] == null)
            items[slot] = new Entry(0, null, new Entry(key, value, null));
          else
            items[slot].next = new Entry(key, value, items[slot].next);
        }
    }

  /**
   * rehash the table if it is "full"
   */
  private void rehash()
    {
      int newSize = optimalSize(count + (count / 10));
      if (newSize <= itemsLength)
        return;
      // System.out.println( "LongKeyDictionary.rehash(): resize from " + itemsLength + " to " + newSize);

      max = (int) (newSize * loadFactor);
      Entry oldItems[] = items;
      items = new Entry[newSize];
      // chain existing list heads together
      Entry freeChain = null;
      for (int i = 0; i < itemsLength; i++)
        {
          Entry head = oldItems[i];
          if (head != null)
            {
              oldItems[i] = head.next;  // remove list head from array;
                                        // it is not used in the loop below
              head.next = freeChain;    // and chain it to the freeChain
              freeChain = head;
            }
        }
      // now move items into the new array, reusing the freeChain
      // of existing heads until it is exhausted, then allocating
      int len = itemsLength;
      // must set itemsLength before calling hashSlot() since that is based on itemsLength
      itemsLength = newSize;
      for (int i = 0; i < len; i++)
        {
          Entry entry = oldItems[i];
          while (entry != null)
            {
              Entry next = entry.next;
              int slot = hashSlot(entry.key);
              if (items[slot] == null) // no chain yet for this slot
                {
                  if (freeChain != null) // can pop a chain head item off the freeChain
                    {
                      items[slot] = freeChain;
                      freeChain = freeChain.next;
                      items[slot].next = null;
                    }
                  else // freeChain is exhausted; allocate a new chain head
                    { items[slot] = new Entry();}
                }
              // add item to the chain
              entry.next = items[slot].next;
              items[slot].next = entry;
              entry = next;
            }
        }
    }

  /**
   * Remove all key/value pairs from this dictionary.
   */
   public void removeAll()
    {
      for (int i = 0; i < itemsLength; i++)
        items[i] = null;
      count = 0;
    }


  /**
   * Remove from this dictionary all key/value pairs matching this value.
   * This is based on </code>equals()</code> item equality, not <code>==</code>.
   * @param value the value to remove.
   * @return the number of times the item was found and removed.
   */
   public int removeAll(Object value)
    {
      return removeAll(value, true);
    }

  /**
   * Remove all occurrences of value from this dictionary.
   * @param value the value to remove.
   * @param whenEquals if <code>true</code>, use </code>equals()</code>
   *        to compare values, else use <code>==</code>
   * @return the number of items removed.
   */
   public int removeAll(Object value, boolean whenEquals)
    {
      Entry previous;
      int nRemoved = 0;
      while ((previous = previous(value, whenEquals)) != null)
        {
          previous.next = previous.next.next;
          count--;
          nRemoved ++;
        }
      return nRemoved;
    }

  /**
   * Check if this dictionary contains an item associated with this key
   * @param key a key
   * @return <code>true</code> if there is a value associated with the key, else <code>false</code>.
   */
   public boolean containsKey(long key)
    {
      return previous(key) != null;
    }

  /**
   * Check if this dictionary contains an item.
   * Note: this is not a quick method. It basically requires
   * sequentially searching the entire dictionary.
   * @param value an item to look for. The test is done with the <code>eqauals()</code> method.
   * @return <code>true</code> if the <var>value</var> exists in this dictionary, else <code>false</code>
   * @see #contains(java.lang.Object, boolean)
   */
   public boolean contains(Object value)
    {
      return contains(value, true);
    }

   /**
   * Test if this dictionary contains an item.
   * Note: this is not a quick method; it basically requires
   * sequentially searching the entire dictionary.
   * @param value an item to look for
   * @param whenEquals if true, use </code>equals()</code>
   *        to compare values, else use <code>==</code>
   * @return <code>true</code> if the value exists in this dictionary, else <code>false</code>.
   */
   public boolean contains(Object value, boolean whenEquals)
    {
      return previous(value, whenEquals) != null;
    }

  /**
   * Return the item associated with this key.
   * @param key the value key
   * @return the value associated with this key.
   * @exception java.lang.NoSuchElementException if there
   * is no value associated with the <var>key</var>. Note that it is possible
   * to store null values in this dictionary, so this method
   * cannot return null to indicate that there is no value for
   * the <var>key</var>. However, it is very efficient to call
   * <pre>
   * if (dict.containsKey(key))
   *    value = dict.get(key);
   * </pre>
   * @see #get(long, java.lang.Object)
   */
   public Object get(long key)
    {
      Entry previous = previous(key);
      if (previous != null)
        return previous.next.value;
      throw new NoSuchElementException();
    }

  /**
   * Return the item associated with this key, or a default value
   * if the key does not exist.
   * @param key the value key
   * @param defaultValue a value to return if
   *        this dictionary does not contain a value for
   *        the <var>key</var>.
   * @return the value associated with this key,
   *         or <var>defaultValue</var> if there is no item
   *         assocuated with the <var>key</var>.
   * @see #get(long)
   */
   public Object get(long key, Object defaultValue)
    {
      Entry previous = previous(key);
      if (previous != null)
        return previous.next.value;
      return defaultValue;
    }

  /**
   * Return the key associated with the value.
   * Note: this is not a quick method; it basically requires
   * sequentially searching the entire dictionary.
   * @param value a value
   * @return key the key which maps to this value.
   * @exception java.lang.NoSuchElementException if the
   * item does not exist in this dictionary
   * @see #getKey(java.lang.Object, long)
   */
   public long getKey(Object value)
    {
      Entry previous = previous(value);
      if (previous != null)
        return previous.next.key;
      throw new NoSuchElementException();
    }

  /**
   * Return the key associated  with the value.
   * Note: this is not a quick method; it basically requires
   * sequentially searching the entire dictionary.
   * @param value a value
   * @param defaultKey the key to return if the value does
   *        not exist in this dictionary.
   * @return key the key which maps to this value,
   *         or defaultValue if the value does not exist.
   */
   public long getKey(Object value, long defaultKey)
    {
      Entry previous = previous(value);
      if (previous != null)
        return previous.next.key;
      return defaultKey;
    }

  /**
   * Return an array of the keys in the table,
   * in no specified order.
   * @return an array of the keys in the table
   */
   public long[] getKeys()
    {
      long result[] = new long[count];
      int pos = 0;
      for (int i = 0; i < itemsLength; i++)
        {
          Entry entry = items[i];
          if (entry != null)
            pos += entry.fill(result, pos);
        }
      return result;
    }

  /**
   * Return an array of the values in the table,
   * in no specified order.
   * @return an array of the values in the table
   */
   public Object[] getValues()
    {
      Object result[] = new Object[count];
      int pos = 0;
      for (int i = 0; i < itemsLength; i++)
        {
          Entry entry = items[i];
          if (entry != null)
            pos += entry.fill(result, pos);
        }
      return result;
    }

  /**
   * Return the number of key/value pairs in this dictionary.
   * @return the number of key/value pairs in this dictionary.
   */
   public int count()
    {
      return count;
    }

  // Serialization
  private void readObject(java.io.ObjectInputStream stream)
    throws IOException, ClassNotFoundException
    {
      stream.defaultReadObject();
      // read the number of items written to the stream. count is transient so that
      // it is 0 after defaultReadObject.  NOTE: the set method updates count.
      int n = stream.readInt();
      itemsLength = optimalSize(n);
      items = new Entry[itemsLength];
      max = (int) (itemsLength * loadFactor);
      for (;n > 0; n--)
        {
          final long k = stream.readLong();
          // System.out.println("read key: " + k);
          Object o = stream.readObject();
          set(k, o);
        }
    }

  // Serialization
  private void writeObject(java.io.ObjectOutputStream stream)
    throws IOException
    {
      stream.defaultWriteObject();
      // write the number of items we will write
      stream.writeInt(count);
      for (int i = 0; i < itemsLength; i++)
        {
          Entry entry = items[i];
          // skip list head
          if (entry != null)
             entry = entry.next;
          while (entry != null)
            {
              // System.out.println("write key: " + entry.key);
              stream.writeLong(entry.key);
              stream.writeObject(entry.value);
              entry = entry.next;
            }
        }
    }

}

/**
 * Entry is used for the internal storage of a LongKeyDictionary;
 * it is used for the chain of values for all items that
 * hash to the same slot.
 * each slot in the dictionary is a chain of
 * key/value pairs of objects with the same hash code, i.e. items
 * with equal values of slot = hashSlot(long key).
 * The first item in the list is a chain head (an Entry) that does not contain values;
 * the empty value allows the Entry previous(...) methods to return the preceeding
 * item in the chain, allowing easy removal of entries without requiring doubly linked
 * lists.
 */
class Entry implements PublicClonable
{
  /** the item's key */
  long key;
  /** the value associated with the key */
  Object value;
  /** next Entry in the chain */
  Entry next;

  /**
   * Construct a default Entry
   */
  public Entry()
    {
    }

  /**
   * Construct a Entry with values
   * @param key the item key
   * @param value the value associated with the key
   * @param next the next Entry in the chain
   */
  public Entry(long key, Object value, Entry next)
    {
      this.key = key;
      this.value = value;
      this.next = next;
    }

  /**
   * Find the item preceeding the lookup key
   * @param lookup the key to find
   * @return the Entry whose .next field contains the lookup key,
   *         or null if none is found.
   */
  Entry previous(long lookup)
    {
      Entry previous = this;
      Entry nextItem = next;
      while (nextItem != null)
        {
        if (nextItem.key == lookup)
          return previous;
        previous = nextItem;
        nextItem = previous.next;
        }
      return null;
    }

  /**
   * Find the item preceeding the lookup value
   * @param item the item to search for.
   * @param whenEquals if true, compare with <code>equals()</code>
   *                   else compare with <code>==</code>
   * @return the Entry whose .next field contains the lookup value
   *         or null if none is found.
   */
  Entry previous(Object item, boolean whenEquals)
    {
      Entry previous = this;
      Entry nextItem = next;
      if (whenEquals)
        while (nextItem != null)
          {
            if (Util.equal(nextItem.value, item))
              return previous;
            previous = nextItem;
            nextItem = previous.next;
          }
      else
        while (nextItem != null)
          {
            if (nextItem.value ==  item)
              return previous;
            previous = nextItem;
            nextItem = previous.next;
          }
      return null;
    }

  /**
   * Fill the array with the values from this Entry and the
   * remaining Entries in the chain,
   * @param array an array in which to store the values
   * @param index start storing values at <code>array[index]</code>
   * @return the number of items stored in the array
   */
  int fill(Object array[], int index)
    {
      int count = 0;
      Entry item = next;
      while (item != null)
        {
          array[index++] = item.value;
          count++;
          item = item.next;
        }
      return count;
    }

  /**
   * Fill the array with the keys from this Entry and the
   * remaining Entries in the chain,
   * @param array an array in which to store the keys
   * @param index start storing keys at <code>array[index]</code>
   * @return the number of keys stored in the array
   */
  int fill(long array[], int index)
    {
      int count = 0;
      Entry item = next;
      while (item != null)
        {
          array[index++] = item.key;
          count++;
          item = item.next;
        }
      return count;
    }

  /**
   * Clone the object
   * @return a clone of this entry, and all remaining
   * items in the chain
   */
  public Object clone()
    {
      // Simplest clone support; does not support subclass,
      // but then this is not a public class...
      return new Entry(key, value, (next == null) ? null : (Entry) next.clone());

      // try { Entry clone = (Entry) super.clone(); }
      // catch (CloneNotSupportedException e)
      //    { }
      // clone.next = (Entry) next.clone();
      // return clone;
    }

}
