// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.collection;

import java.io.PrintStream;
import java.util.Enumeration;
import com.sas.PublicClonable;
import com.sas.util.Countable;
import com.sas.util.transforms.TransformInterface;
import com.sas.util.transforms.TransformException;
import com.sas.collection.Nil;
import java.io.Serializable;

/**
 * A <var>Pair</var> is an object which refers to two other objects.
 * This provides a very flexible mechanism for building lists, trees,
 * and other data structures.
 * <p>
 * Those familiar with Lisp will
 * find pairs very familiar -- a <code>Pair</code> is basically a <var>cons</var> cell.
 * The power of a <code>Pair</code> object comes when you build recursive
 * structures. A simple list is created when you make a <code>Pair</code>
 * whose first element is some data and the second element
 * is another <code>Pair</code>.  A tree can be built by making both the first
 * and second items be leaves or other <code>Pairs</code>.
 * <p>
 * This package provides a <code>nil</code> value which represents
 * an empty list. Most lists are sequences of <code>Pair</code> objects,
 * terminated with a <code>nil</code> value for the second item of the
 * final pair. Such lists are printed as <code>(item1 item2 [...] item<var>n</var>)</code>
 * but is really more accurately described as
 * <code>(item1 . (item2 . ... (item<var>n</var> . nil) ) )</code>
 * <p>
 * A Pair can also be a "dotted pair", which occurs when the
 * second element of the pair is not another <code>Pair</code>. Such lists
 * are printed as (item1 . item2). Dotted pairs are really the
 * exception, not the rule. Most lists should end with nil. Many methods
 * will raise a run time exception if you access a dotted pair
 * when a correct list (one whose second element is a <code>Pair</code>)
 * is expected.
 * <p>
 * A list of one item is a <code>Pair</code> with <code>nil</code> for the
 * second item and is printed as <code>(item1)</code> and not as
 * <code>(item1 . nil)</code> or <code>(item1 . ())</code>.
 *
 * @author David J. Biesack
 * @version 1.0
 */

public class Pair
       implements PublicClonable, Countable, Serializable
{
  static final long serialVersionUID = 6595897507134640222L;

  static final String RB_KEY = "Pair.";

  /**
   * <code>nil</code> is the empty list. It is unique: you cannot create
   * other empty lists. It is immutable: you cannot change
   * its elements.
   */
  public static final Pair nil = new Nil();

  // these fields are transient and written explicitly in writeObject,
  // so that nil is handled correctly (it's a singleton)

  /**
   * _car holds the first element of the Pair
   */
  transient protected Object _car;
  /**
   * _cdr holds the second element of the Pair
   */
  transient protected Object _cdr;

  /**
   * construct a default Pair. Both the first and second elements are nil.
   */
  public Pair()
  {
    _car = nil;
    _cdr = nil;
  }

  /**
   * Construct a Pair from the contents of another Pair.
   * Note: do not confuse this constructor with
   * new </code>Pair(Object)</code>
   * @param other another pair. The new pair
   * contains the same elements as the <var>other</var>
   * @see #Pair(java.lang.Object)
   */
  public Pair(Pair other)
  {
    this(other.getFirst(), other.getSecond());
  }

  /**
   * construct a new Pair from two objects. This is a 'dotted list'
   * if the second argument is not another Pair. If it is a Pair, then
   * this Pair becomes a proper list.
   * This constructor replaces null parameters with nil.
   * @param first the first element, retrievable with getFirst()
   * @param second the second element, retrievable with getSecond()
   * @see #getFirst
   * @see #getSecond
   * @see #setFirst
   * @see #setSecond
   */
  public Pair(Object first, Object second)
  {
    _car = (first == null) ? nil : first;
    _cdr = (second == null) ? nil : second;
  }

  /**
   * Construct a new Pair from one object.
   * This constructor replaces a null parameter with nil.
   * The default 'second' is nil.
   * Note: do not confuse this constructor with
   * new </code>Pair(Pair)</code>
   * @param first the first element, retrievable with first()
   * @see #Pair(com.sas.collection.Pair)
   */
  public Pair(Object first)
  {
    _car = (first == null) ? nil : first;
    _cdr = nil;
  }

  /**
   * Construct a list of <em>count</em> objects. That is, create
   * <em>count</em> Pair objects, chained together, with the last Pair
   * terminated with a nil Pair.
   * This constructor replaces null parameters with nil.
   * @param count the number of items in the list.
   * @param initial the value to place in the first element
   *        of each Pair.
   * @exception java.lang.IllegalArgumentException if count < 1
   */
  public Pair(int count, Object initial)
  {
    if (count < 1)
      throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "cannotMakeEmpty.ex.txt"));
    _car = initial;
    _cdr = (  (count > 1)
            ? _cdr = new Pair(count-1, initial)
              : nil);
  }

  /**
   * Construct a list of <em>count</em> nil items.
   * This constructor is equivalent to Pair(count, nil)
   * @param count the number of items in the list.
   * @param initial the value to place in the first element
   *        of each Pair.
   * @exception java.lang.IllegalArgumentException if count < 1
   */
  public Pair(int count)
  {
    _car = nil;
    if (count < 1)
      throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "cannotMakeEmpty.ex.txt"));
    if (count > 1)
      _cdr = new Pair(count-1);
    else
    _cdr = nil;
  }

  /**
   * Construct a list from an array of objects
   * @param elements an array of objects.
   */
  public Pair(Object elements[])
  {
    Pair last = this;
    if (elements.length < 1)
      throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "cannotMakeEmpty.ex.txt"));
    last._car = elements[0];
    last._cdr = nil;
    for (int i = 1;
         i < elements.length;
         i++)
      {
        Pair cons = new Pair(elements[i]);
        last._cdr = cons;
        last = cons;
      }
  }

  /**
   * Construct a list from an enumeration of objects
   * @param elements an enumeration of objects.
   */
  public Pair(Enumeration elements)
  {
    Pair last = this;
    if (! elements.hasMoreElements() )
      throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "cannotMakeEmpty.ex.txt"));
    last._car = elements.nextElement();
    last._cdr = nil;
    while (elements.hasMoreElements())
      {
        Pair cons = new Pair(elements.nextElement());
        last._cdr = cons;
        last = cons;
      }
  }


  public int hashCode()
  {
    return ( _car == nil ? 0 : _car.hashCode() ) +
           ( _cdr == nil ? 0 : _cdr.hashCode() );
  }

  /**
   * Return the first element of a Pair.
   * Functions like the Lisp car function.
   * @return the first element.
   */
  public final Object car()
  {
    return this._car;
  }

  /**
   * Return the first element of a Pair.
   * This is an alias for the car() method
   * but is named to match the "Pair"-ness of Pair objects.
   * @return the first element.
   */
  public Object getFirst()
  {
    return car();
  }

  /**
   * Return the second element of a Pair.
   * Functions like the Lisp cdr function.
   * @return the second element.
   */
  public final Object cdr()
  {
    return this._cdr;
  }

  /**
   * Return the second element of a Pair.
   * This is an alias for the cdr() method
   * but is named to match the "Pair"-ness of Pair objects.
   * @return the first element.
   */
  public Object getSecond()
  {
    return cdr();
  }

  /**
   * Return the <var>n</var><sup>th</sup> second element from a list of pairs, performed by
   * calling getSecond() n times (on subsequent results).
   * Thus, getSecond(1) is like calling getSecond(), and
   * getSecond(2) is like calling getSecond().getSecond().
   * This is like the Lisp <code>(NTHCDR list)</code> function.
   * @param n specifies which Pair to get from the list.
   * @return the <var>n</var><sup>th</sup> , zero based.
   * @exception java.lang.IllegalArgumentException if the
   *            list does not contain n Pairs.
   */
  public Object getSecond(int n)
  {
    // predecriment so we call tail n-1 times
    Pair p = this.getTail(n-1);
    return p.cdr();
  }

  /**
   * Return the <var>n</var><sup>th</sup> Pair from a list, performed by
   * calling getTail() n times (on subsequent results).
   * This is like getSecond except that the return type is a Pair.
   * This is like the Lisp <code>(NTHCDR list)</code> function.
   * @param n specifies which Pair to get from the list.
   * @return the <var>n</var><sup>th</sup> , zero based.
   * @exception java.lang.IllegalArgumentException if the
   *            list does not contain n Pairs.
   */
  public Pair getTail(int n)
  {
    Pair p = this;
    while (n-- > 0)
      p = p.getTail();
    return p;
  }

  /**
   * Return the second element of a Pair as another Pair
   * This is like getSecond except that the return type is a Pair.
   * @return the tail of this list, as another Pair
   * @exception java.lang.IllegalArgumentException  if the
   *            tail is not another Pair.
   */
  public final Pair getTail()
  {
    if (this._cdr instanceof Pair)
      return (Pair) this._cdr;
    else
      throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "invalidTail.ex.txt"));
  }

  /**
   * Replace the first element of a Pair.
   * This is an alias for the setcar() method
   * but is named to match the list-like properties of Pair objects.
   * @param head the new first element (i.e. new head of the list)
   */
  public void setFirst(Object head)
  {
    setcar(head);
  }

  /**
   * Replace the first element of a Pair. For Lisp fans
   * Null is not allowed; if car == null, nil is used instead.
   * @param car the new first element.
   */
  public void setcar(Object car)
  {
    this._car = (car == null ? nil : car);
  }

  /**
   * Replace the tail of a Pair list.
   * This is an alias for the setcar() method
   * but is named to match the "Pair"-ness of Pair objects.
   * @param tail the new list tail element. Note it does
   *        not have to be a Pair. If it is not a Pair, then
   *        the Pair becomes a dotted pair. If tail is
   *        null, then the tail is set to nil instead.
   */
  public void setSecond(Object tail)
  {
    setcdr(tail);
  }

  /**
   * Replace the second element of a Pair list. For Lisp fans.
   * @param cdr the new cdr of the Pair. Note it does
   *        not have to be a Pair. If it is not a Pair, then
   *        the Pair becomes a dotted pair. If tail is
   *        null, then the tail is set to nil instead.
   */
  public void setcdr(Object cdr)
  {
    this._cdr = (cdr == null) ? nil : cdr;
  }

  /*
   * Construct and return a reversed copy of this list.
   * This raises an error if the list ends in a dotted pair.
   * Note: reversing the result again does not return the
   * original list.
   * @return a new list whose elements are the same
   *         as the input list, only reversed.
   * @exception java.lang.IllegalArgumentException if the list
   *            is terminated by a dotted pair and
   *            not a nil terminated Pair, or if the list
   *            list cannot be reversed.
   */
  public Pair reverse()
  {
    try
      {
        Pair copy = (Pair) clone();
        return copy.reverseInPlace();
      }
    catch(CloneNotSupportedException c)
      {
        throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "cannotReverse.ex.txt"));
      }
  }

  /**
   * Return the length of the list.
   * @return the number of items in the list.
   * @exception java.lang.IllegalArgumentException if the list
   *            is terminated by a dotted pair and
   *            not a nil terminated Pair.
   */
  public int count()
  {
    int len = 1;
    Pair cdr = getTail();
    while (cdr != nil)
      {
        len++;
        cdr = cdr.getTail();
      }
    return len;
  }

  /**
   * Print a list to System.out
   */
  public void print()
  {
    this.print(System.out);
  }

  /**
   * Print a list to the specified output print stream.
   * @param the print stream on which the list is printed.
   */
  protected void print(PrintStream s)
  {
    this.print(s, 0);
  }

  /**
   * The maximum length of a printed list.
   * The <code>{@link #print()}</code> methods stop printing when the number
   * of items printed exceeds this. This is meant to
   * prevent infinite loops when printing lists with cycles.
   * If this value is less than 0, then no check is performed
   * for cyclical lists but you will probably halt with a
   * Java stack overflow due to <code>print()</code> recursion.
   * <p>
   * An example of a cyclical list is:
   * <pre>
   *   Pair list = new Pair(nil, nil);
   *   list.setcdr(list);
   * </pre>
   */
  public static int printLength = 65536;

  /**
   * The maximum depth of a printed list.
   * The {@link #print()} methods stop printing when the number
   * of nesting levels exceeds this. This is meant to
   * prevent infinite loops when printing lists with mutual recursion.
   * If this value is less than 0, then no check is performed
   * for recursive lists, but you will probably halt with a
   * Java stack overflow due to <code>print()</code> recursion.
   * <p>
   * An example of a recursive list is:
   * <pre>
   *   Pair list = new Pair(nil, nil);
   *   list.setcar(list);
   * </pre>
   */
  public static int printDepth = 256;

  private void printCar(PrintStream s, int depth)
  {
    if (_car instanceof Pair)
      ((Pair)_car).print(s, depth+1);
    else
      s.print(_car.toString());
  }

  /**
   * Print a list, keeping track of depth and length
   * @param s the print stream
   * @param depth the current print depth. This is increased
   *        by 1 on recursive calls.
   */
  protected void print(PrintStream s, int depth)
  {
    Object p = cdr();
    int length = 1;

    if (printDepth > 0 && depth > printDepth)
      {
        s.print("(***)");
        return;
      }

    s.print("(");
    printCar(s, depth+1);

    while ( (p != nil) && (p instanceof Pair) )
      {
        Pair list = (Pair) p;
        if (++length > printLength)
          {
            s.print(" ...");
            break;
          }
        s.print(" ");
        list.printCar(s, depth+1);
        p = list.getTail();
      }

    if (p != nil)
      {
        s.print(". ");
        s.print(p.toString());
      }
    s.print(")\n");
  }


  /**
   * Apply a transformation to all items in the list,
   * returning a new list with the transformed items.
   * @param t a transformation -- a function which
   *        is applied to all items.
   * @return a new list consisting of the transformed items.
   */
  public Pair transform(TransformInterface t)
  {
    Object tFirst = null;
    Object tRest  = null;
    try { tFirst = t.transform(car()); }
    catch (TransformException te) { }
    tRest = cdr();
    if (tRest instanceof Pair)
      tRest = ((Pair) tRest).transform(t);
    Pair result = new Pair(tFirst, tRest);
    return result;
  }

  /**
   * Clone the Pair, returning a new <code>Pair</code>.
   * @return a new list which contains a copy of the
   *         input list. Each item is also copied
   *         recursively, if it is a <code>Pair</code>. Note that
   *         this deep copy behavior differs from most
   *         object clone methods. Use <code>new Pair(pair)</code>
   *         to make a shallow copy.
   * @exception java.lang.CloneNotSupportedException if the clone fails.
   */
  public Object clone()
                throws CloneNotSupportedException
    {
      Pair newPair = (Pair) super.clone();
      Object newCar = _car;
      Object newCdr = _cdr;

      if (this == nil)
        return nil;

      if (_car instanceof Pair)
        newCar = ((Pair)_car).clone();
      if (_cdr instanceof Pair)
        newCdr = ((Pair)_cdr).clone();
      newPair.setcar(newCar);
      newPair.setcdr(newCdr);
      return newPair;
    }

  /**
   * Return the difference between this list and another list.
   * That is, copy each <code>Pair</code> until the Pair which equals (==) the
   * sublist is found.
   * @param sublist a <code>Pair</code> which should be one of the cdr
   *        parts of the input list.
   * @return a copy of the original list, with the
   *         tail part matching the input sublist removed.
   */
  public Pair difference(Pair sublist)
  {
    if (_cdr == nil)
      return new Pair(_car, _cdr);
    else if (_cdr == sublist)
      return new Pair(_car);
    else if (_cdr instanceof Pair)
      return new Pair(_car, getTail().difference(sublist));
    else
      return new Pair(_car, _cdr);
  }

  /**
   * Reverse a list in place, swapping first and second elements
   * of each <code>Pair</code>.
   * Note that this alters the <code>Pair</code> and any other <code>Pair</code>
   * objects traversed
   * via the tail sections (recursively). This method is useful
   * when constructing new lists: repeatedly create new <code>Pair</code>
   * objects:
   * <pre>
   *  list = new Pair(item, list);
   * </pre>
   * When done, perform a <code>list = list.reverseInPlace()</code>. Lisp fans will recognize this
   * as <code>(NREVERSE list)</code>.
   * @return the last Pair in the original list, which is now the
   *         head of the reversed list.
   * @exception java.lang.IllegalArgumentException if a dotted pair is
   *            encountered.
   */
  public Pair reverseInPlace()
  {
    Pair p = this, last = nil, next = nil;
    while (p != nil)
      {
        next = p.getTail();
        p._cdr = last;
        last = p;
        p = next;
      }
    return last;
  }

  /**
   * Test if another list has the same structure/contents.
   * This is a recursive operation. This is just a convenience method.
   * @param that the left hand side list.
   * @param other the right hand side list.
   */
  private static boolean equals(Object that, Object other)
  {
    if (that == nil || other == nil)
      return that == other;
    else if (that instanceof Pair && other instanceof Pair)
      return ((Pair)that).equals((Pair) other);
    else
      return that.equals(other);
  }

  /**
   * Test if another list has the same structure/contents.
   * @param other the other list to compare. May not be null.
   */
  public boolean equals (Pair other)
  {
    if (other == nil)
      return false;
    else
      return equals(_car, other.car()) && equals(_cdr, other._cdr);
  }

  /**
   * Test if another list has the same structure/contents.
   * @param other the other list to compare.
   */
  public boolean equals (Object otherObject)
  {
    if ( otherObject instanceof Pair )
    {
      return equals( (Pair)otherObject );
    }
    else
       return false;
  }

  /**
   * Return the <var>n</var><sup>th</sup> item from a list.
   * @param n specifies which item to get from the list.
   * @return the <var>n</var><sup>th</sup> element, zero based.
   * @exception java.lang.IllegalArgumentException if the
   *            list does not contain n Pairs.
   */
  public Object getItem(int n)
  {
    Pair p = getTail(n);
    return p.car();
  }

  /**
   * Return the last <code>Pair</code> in a list. Works on dotted pairs.
   * Like the Lisp <code>(LAST lisp)</code> function.
   * @return the last <code>Pair</code> in a sequence of pairs.
   */
  public Pair getLast()
  {
    Pair p = this;
    while (p != nil && (p._cdr instanceof Pair))
      p = p.getTail();
    return p;
  }

  /**
   * Return true if the <code>Pair</code> is empty (it is the <code>nil Pair</code>)
   * @return <code>true</code> if the list is <code>nil</code>, <code>false</code> otherwise.
   */
  public boolean isNull(Pair p)
  {
    return false;
  }

  public String toString()
    {
      return "(" + getFirst() + " . " + getSecond() + ")";
    }


  static {
    /* initialize nil. */
    nil._car = nil;
    nil._cdr = nil;
  }


  // Serialization
  private void readObject(java.io.ObjectInputStream stream)
    throws java.io.IOException, ClassNotFoundException
    {
      stream.defaultReadObject();
      // Handle nil values by reading a boolean written in writeObject
      if (stream.readBoolean())
        _car = stream.readObject();
      else
        _car = nil;
      if (stream.readBoolean())
        _cdr = stream.readObject();
      else
        _cdr = nil;

      // other stuff
    }


  // Serialization
  private void writeObject(java.io.ObjectOutputStream stream)
    throws java.io.IOException
    {
      stream.writeBoolean(_car != nil);
      if (_car != nil)
        stream.writeObject(_car);
      stream.writeBoolean(_cdr != nil);
      if (_cdr != nil)
        stream.writeObject(_cdr);
      stream.defaultWriteObject();
      // other stuff
    }


}
