// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.collection;

import com.sas.PublicClonable;
import com.sas.util.Comparator;
import com.sas.util.ConstantComparator;
import com.sas.util.ConstantPredicate;
import com.sas.util.PredicateInterface;

import com.sas.collection.Pair;

/**
 * A <code>PairFirst</code> object provides a way to select items from a <code>Pair</code>
 * when performing a predicate or comparison operation
 * on <code>Pair</code> objects. For example, this is useful when enumerating
 * or sorting items in an <code>AssociationList</code>.  The
 * {@link com.sas.collection.AssociationList#sort(com.sas.util.Comparator) <code>sort(Comparator)</code>}
 * method of <code>AssociationList</code> passes <code>Pair</code> objects (containing
 * the key and value) to the predicate or comparator.
 * If you already have a predicate or comparator which will work on
 * the keys, then you can wrap your predicate or comparator
 * with a <code>PairFirst</code> object and it will extract the key from each
 * <code>Pair</code> and pass just the keys to your function.
 *
 * @author <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @see com.sas.collection.Pair
 * @see com.sas.collection.PairFirst
 * @see com.sas.collection.AssociationListInterface
 * @see com.sas.collection.AssociationList
 */

public class PairFirst
             implements Comparator,
                        PredicateInterface,
                        java.io.Serializable,
                        PublicClonable
{
  static final long serialVersionUID = 7372277745352503439L;

  Comparator comparator;
  PredicateInterface predicate;

  /**
   * Default Constructor.
   * This creates an object which used as a Comparator always returns
   * {@link com.sas.util.Comparator#EQUALS Comparator.EQUALS},
   * and when used as a predicate always returns <code>false</code>.
   */
  public PairFirst()
    {
      this.comparator = ConstantComparator.COMPARE_EQUALS;
      this.predicate = ConstantPredicate.FALSE;
    }

  /**
    * Create an comparator/predicate which applies the predicate and
    * comparator to the first items of Pair objects.
    * This <code>PairFirst</code> object, when used as a <code>Comparator</code>,
    * compares the <code>{@link Pair#getFirst()}</code> elements
    * of two <code>Pair</code> objects using a specific comparator.
    * When used as a predicate always returns <code>false</code>.
    * @param comparator compare the <code>getFirst()</code> elements of
    *        two pair objects using the comparator.
    */
  public PairFirst(Comparator comparator)
    {
      this.comparator = comparator;
      this.predicate = ConstantPredicate.FALSE;
    }

  /**
    * Create an comparator which applies the predicate to the first
    * item of Pair objects. This <code>PairFirst</code> object, when used as a Comparator,
    * always returns {@link com.sas.util.Comparator#EQUALS Comparator.EQUALS}.
    * When used as a predicate, returns the value of
    * the specific predicate applied to the <code>getFirst()</code>
    * element of the <code>Pair</code>.
    * @param predicate when used as a PredicateInterface,
    * this object will retrieve the getFirst() element
    * of a Pair and apply this predicate to that value.
    */
  public PairFirst(PredicateInterface predicate)
    {
      this.comparator = ConstantComparator.COMPARE_EQUALS;
      this.predicate = predicate;
    }

  /**
    * Create an object which, when used as a Comparator,
    * compares the <code>{@link Pair#getFirst()}</code> elements
    * of two Pair objects using a specific comparator, and
    * when used as a predicate, returns the value of
    * the specific predicate applied to the <code>getFirst()</code>
    * element of the Pair.
    * @param comparator compare the <code>getFirst()</code> elements of
    * two pair objects using the comparator.
    * @param predicate when used as a <code>PredicateInterface</code>,
    * this object will retrieve the <code>getFirst()</code> element
    * of a <code>Pair</code> and apply this predicate to that value.
    */
  public PairFirst(Comparator comparator, PredicateInterface predicate)
    {
      this.comparator = comparator;
      this.predicate = predicate;
    }

  /**
    * Assume the <var>object</var> is a <code>Pair</code> and apply the initial
    * predicate (see the constructors) to that <code>Pair</code>'s <code>{@link Pair#getFirst()}</code>
    * value.
    * @param object an object whish must be a <code>Pair</code>.
    */
  public boolean predicate(Object object)
    {
      Pair pair = (Pair) object;
      Object o = pair.getFirst();
      return predicate(o);
    }

  /**
    * Assume the objects are <code>Pair</code> objects and apply the initial
    * comparator (see the constructors) to the <code>Pair</code>s' <code>{@link Pair#getFirst()}</code>
    * values.
    * @param a an object whish must be a <code>Pair</code>. It's <code>getFirst()</code>
    * value is used for the real comparison.
    * @param b an object whish must be a <code>Pair</code>. It's <code>getFirst()</code>
    * value is used for the real comparison.
    */
  public int compare(Object a, Object b)
    {
      Pair pairA = (Pair) a;
      Pair pairB = (Pair) b;
      Object objectA = pairA.getFirst();
      Object objectB = pairB.getFirst();
      return comparator.compare(objectA, objectB);
    }

  /**
   * Clone a <code>PairFirst</code> object.
   * @return a clone of this object
   * @exception CloneNotSupportedException if the clone fails.
   */
  public Object clone()
                throws CloneNotSupportedException
    {
      // I'm not sure why I clone this way...
      // should be sufficient to just return super.clone()...
      PairFirst clone = new PairFirst();
      clone.comparator = comparator;
      clone.predicate = predicate;
      return (Object) clone;
    }

}


