package com.sas.io;

import java.io.*;

//
// History:
//
// Date    Name            Remarks
// 11Oct99 B Beers         S0067555: Fixed superclass member variable settings in setBytes.
// 07Feb00 B Beers         Fix bad links in doc comments.
/**
 * <code>BytesInputStream</code> is a {@link java.io.ByteArrayInputStream} which
 * allows you to replace the bytes on which the stream is based.
 * You can use a <code>BytesInputStream</code> if you wish to divide a larger
 * set of bytes (for example, regions of bytes read from a {@link java.io.DataInputStream})
 * into smaller, independent substreams, without creating a bew
 * <code>ByteArrayInputStream</code> for each such subregion.
 * <p>
 * This is particularly useful when
 * you wish to use a {@link java.io.InputStreamReader} on such subregions of a <code>DataInputStream</code>
 * in order to efficiently convert bytes into strings using a non-default encoding.
 * (The default <code>InputStreamReader</code> uses buffering techniques which perform read ahead.
 * If you ask to read <var>n</var> characters from an <code>InputStreamReader</code>, it will
 * read 8192 bytes.)
 * <p>
 * The following example shows how to read an input stream, mixing <code>DataInputStream</code>
 * operations to read integers and shorts, and using a <code>InputStreamReader</code> to
 * convery byte array subranges of the input stream <var>inputStream</var>
 * into characters using some character encoding, <var>ENCODING</var>,
 * from which new Strings are made.
 * <pre>
 * // Construct the DataInputStream for reading numbers
 * DataInputStream dataIn = new DataInputStream(inputStream);
 * // Make a BytesInputStream which will be used as an input stream
 * // for the InputStreamReader
 * BytesInputStream encodedIn = new BytesInputStream(new byte[0]);
 * InputStreamReader reader = new InputStreamReader(encodedIn, ENCODING);
 * // reusable input data buffers:
 * byte bytes[] = new byte[80];
 * char chars[] = new char[80];
 * // read the number of strings in the incoming buffer:
 * int nStrings = dataIn.readInt();
 * for (int i = 0; i < nStrings; i++)
 *   {
 *     // read the length of the next incoming String
 *     int strLen = dataIn.readShort();
 *     if (strLen > chars.length) // grow character buffer if necessary
 *       chars = new char[strLen];
 *
 *     // read number of bytes used to encode the string
 *     int nBytes = dataIn.readShort();
 *     if (nBytes > bytes.length) // grow byte buffer if necessary
 *         bytes = new byte[nBytes];
 *     // read the array of bytes from the DataInputStream
 *     nBytes = dataIn.read(bytes, 0, nBytes);
 *     // Put those bytes into the BytesInputStream
 *     encodedIn.setBytes(bytes, 0, nBytes);
 *     // Read the characters using the InputStreamReader's encoding:
 *     strLen = reader.read(chars, 0, strLen);
 *     // and construct a String from the characters.
 *     String string = new String(chars, 0, strLen);
 *     vector.addElement(string);
 *   }
 * reader.close();
 * dataIn.close();
 * </pre>
 * Without BytesInputStream, you would have to pass the encoding to
 * the String constructor, which results in a table lookup for
 * each string constructed. Using a <code>BytesInputStream</code> yields a
 * large performance benefit when reading many strings.
 *
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @since 2.0
 */
public class BytesInputStream
       extends ByteArrayInputStream
{
  private static byte EMPTY_BYTE_ARRAY[] = new byte[0];

  // A "safe" byte array wrapper:
  private static byte[] bytes(byte bytes[])
    {
      return bytes == null ? EMPTY_BYTE_ARRAY : bytes;
    }

  // A "safe" byte array length function
  private static int length(byte bytes[])
    {
      return bytes == null ? 0 : bytes.length;
    }

  /**
   * Construct a BytesInputStream which will read bytes from from an
   * array of bytes.
   * @param bytes the array of bytes to read from.
   * This may be null, if the bytes will be assigned later using {@link #setBytes(byte[], int, int)}
   */
  public BytesInputStream(byte bytes[])
    {
      this(bytes(bytes), 0, length(bytes));
    }

  /**
   * Construct a BytesInputStream which will read bytes from from
   * a subrange of array of bytes.
   * The <var>bytes</var> array is used directly, rather than
   * simply copying the contents.
   * @param bytes the array of bytes to read from.
   * @param offset the first byte within <var>bytes</var> to start reading.
   * @param length the number of bytes from the <var>bytes</var> array
   * that should be considered part of the input stream.
   */
  public BytesInputStream(byte bytes[], int offset, int length)
    {
      super(bytes(bytes), offset, length);
    }

  /**
   * Replace the bytes which this BytesInputStream reads.
   * Any mark previously set on the buffer is lost.
   * This calls <code>setBytes(<var>bytes</var>, 0, <var>bytes</var>.length)</code>
   * @param bytes the array of bytes to read from.
   */
  public final void setBytes(byte bytes[])
    {
      setBytes(bytes(bytes), 0, length(bytes));
    }

  /**
   * Replace the bytes which this BytesInputStream reads.
   * Any mark previously set on the buffer is lost.
   * The <var>bytes</var> array is used directly, rather than
   * simply copying the contents.
   * @param bytes the array of bytes to read from.
   * @param offset the first byte within <var>bytes</var> to start reading.
   * @param length the number of bytes from the <var>bytes</var> array
   * that should be considered part of the input stream.
   */
  public void setBytes(byte bytes[], int offset, int length)
    {
      buf = bytes(bytes);
      count = offset + length;
      pos = offset;
      mark = 0;
    }
}
