package com.sas.io;

import java.io.Reader;
import java.io.InputStreamReader;
import java.io.Writer;
import java.io.ByteArrayOutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.IOException;
import java.io.DataInput;
import java.io.DataOutputStream;
import java.io.UTFDataFormatException;
import java.io.UnsupportedEncodingException;
import java.util.Vector;

//
// History:
//
// Date    Name            Remarks
// 18Oct99 D Biesack       S0068047: isValid returns false when checking Cp1047.
// 07Feb00 B Beers         Fix bad links in doc comments.
// 04Aug00 B Beers         S0092352: Use alias 8859_1 for ISO8859_1
// 04Dec00 B Beers         S0100429: Fix length of UTF-8 data.
// 18Dec00 D Biesack       S0105475: Fix memory leak with Cp1047 encoder.
// 27Sep01 B Beers         S0127846: Make thread safe.
// 08Apr03 B Beers         S0193963: Prevent overallocation of charBuffer.
//
/**
 * Encoder provides a simple and efficient mechanism to encode
 * many different byte arrays into Strings or character arrays, using the same byte to character encoding.
 * This class provides much more efficient encoding than many calls to
 * {@link java.lang.String#String(byte[], String) new String(byte bytes[], String encoding)}
 * because it can avoid repeated encoding table lookups and byte/character array allocations normally done with
 * <code>new String(byte bytes[], String encoding)</code>.
 * You can also use Encoder to convert in the other direction: from character arrays or strings into byte arrays.
 * <p>
 * This class also provides maintenance of a pool of encoders.
 * Use {@link #getInstance(String encoding)} to get an encoder from a pool of available
 * encoders. When done, return the allocated encoder with {@link #returnInstance(EncoderInterface)}
 * This is useful in situations where you need to reuse encoders for one or two encodings
 * but want to avoid reallocating an Encoder each time one is needed.
 * @version 1.0
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @since 1.0
 */

public class Encoder
       implements EncoderInterface
{
  /** Used to read bytes as Strings from an InputStreamReader with an encoding. */
  protected BytesInputStream iStream;
  /** Used to write Strings to byte arrays using an OutputStreamWriter with an encoding. */
  protected ByteArrayOutputStream oStream;
  /** A Reader wrapped around the BytesInputStream {@link #iStream} which is used to read
   * bytes, converting them to Strings using an encoding. */
  protected InputStreamReader reader;
  /** A Writer wrapped around the ByteArrayOutputStream {@link #oStream} which is used to write
   * Strings, converting them to byte arrays using an encoding. */
  protected OutputStreamWriter writer;
  /** The encoding name. */
  protected String encoding;
  /** Internal array to be reused for each String -> byte array encoding */
  protected char byteByffer[];
  /** Internal array to be reused for each byte array -> String encoding */
  protected char charBuffer[];
  private char[] charArray;

  private static final int MAX_POOL_SIZE = 64;
  public static boolean DEBUG = false;

  /** A blank string "" */
  protected final static String EmptyString = "";

  /**
   * Construct a new Encoder with which you can efficiently
   * convert between Strings and byte arrays.
   * @param encoding the name of the byte to character encoding to use. You can change
   *        this later with {@link #setEncoding(String)}.
   *        If null, this Encoder will use the default encoding.
   * @exception  UnsupportedEncodingException
   *             If the named encoding is not supported
   */
  public Encoder(String encoding)
         throws UnsupportedEncodingException
    {
      setEncoding(encoding);
      // System.out.println("new Encoder(" + encoding + ")" );
    }

  /**
   * Construct a new Encoder with which you can efficiently
   * convert between Strings and byte arrays. The default
   * Encoder uses the default encoding, although you can change
   * this later with {@link #setEncoding(String)}.
   * @exception  UnsupportedEncodingException
   *             If the named encoding is not supported
   */
  public Encoder()
    {
      try
        { setEncoding(null); }
      catch (UnsupportedEncodingException ignore)
        { }
      // System.out.println("new Encoder(" + null + ")" );
    }

  /**
   * Return the reader used to read characters from a byte array.
   * Encoder constructs an InputStream around a BytesInputStream,
   * so that the byte to character converter lookup is done only once,
   * and we can avoid constructing new InputStreams for each conversion.
   */
  protected Reader getReader()
    {
      return reader;
    }

  /**
   * Make an InputStreamReader using the current encoding.
   * The reader uses an instance of {@link com.sas.io.BytesInputStream}
   * for efficiency. Each call to {@link #toString(byte[])} for example
   * can simply call {@link com.sas.io.BytesInputStream#setBytes(byte[], int offset, int length)}
   * so that the InputStreamReader can read from the new byte array, instead
   * of making a new ByteArrayInputString/InputStreamReader pair for
   * each byte array (or instead of using <code>new String(byte[], String encoding)</code>
   * which performs an encoding lookup.  This way, the encoding lookup
   * is only used once and we greatly reduce the number of new object allocations.
   */
  protected void createReader() throws UnsupportedEncodingException
    {
      if (iStream == null)
        iStream = new BytesInputStream(null);
      // API doc for InputStreamReader does not say if new InputStreamReader(iStream, null)
      // will use the default encoding, so we have to use a different constructor if null
      String enc = getEncoding();

      if (enc == null)
      {
        reader = new InputStreamReader(iStream);
      }
      else
      {
        if (enc.equals("ISO8859_1"))
        {
          // Use an alias for ISO8859_1. Both the Microsoft VM and
          // the JavaSoft VM understand this alias, but only the
          // JavaSoft VM understands "ISO8859_1"
          //------------------------------------------------------
          enc = "8859_1";
        }

        reader = new InputStreamReader(iStream,enc);
      }
    }

  /**
   * Return the Writer used to convert characters to bytes.
   */
  protected Writer getWriter()
    {
      return writer;
    }

  /**
   * Create a Writer which we can reuse when converting from
   * character arrays or Strings to byte arrays. We reuse the
   * same OutputStreamWriter over and over, resetting the
   * ByteArrayOutputStream oStream on each conversion.
   * This eliminates redundant encoding lookups done with
   * String.toByteArray(String encoding) and eliminates
   * new object allocation.
   */
  protected void createWriter() throws UnsupportedEncodingException
    {
      if (oStream == null)
        oStream = new ByteArrayOutputStream();
      // API doc for OutputStreamWriter does not say if new OutputStreamWriter(oStream, null)
      // will use the default encoding, so we have to use a different constructor if null
      String enc = getEncoding();

      if (enc == null)
      {
        writer = new OutputStreamWriter(oStream);
      }
      else
      {
        if (enc.equals("ISO8859_1"))
        {
          // Use an alias for ISO8859_1. Both the Microsoft VM and
          // the JavaSoft VM understand this alias, but only the
          // JavaSoft VM understands "ISO8859_1"
          //------------------------------------------------------
          enc = "8859_1";
        }

        writer = new OutputStreamWriter(oStream,enc);
      }
    }

  /**
   * Set the name of the string encoding used to encode bytes to Strings.
   * @param encoding the new string encoding. If null, this Encoder
   *        will use the default encoding.
   * @exception  UnsupportedEncodingException
   *             If the named encoding is not supported.
   */
   public void setEncoding(String encoding)
         throws UnsupportedEncodingException
   {
     // this.encoding = patchEncoding(encoding);
     this.encoding = encoding;
     createReader();
     createWriter();
   }

   /**
    * Return the value of the character to byte encoding.
    */
    public String getEncoding()
    {
      return encoding;
    }

  /**
   * Convert an array of bytes into a string, using the current encoding.
   * @param bytes an array of bytes to encode as a String
   * @return a new String constructed from the array of bytes, using the encoding.
   * This is like {@link java.lang.String#String(byte[], String) new String(bytes, getEncoding)}
   * except that it is more efficient.
   */
  public final String toString(byte bytes[])
    {
      return toString(bytes, 0, bytes.length);
    }

  /**
   * Convert an array of bytes into a String, using the current encoding.
   * @param bytes an array of bytes to encode as a String
   * @param offset the byte offset to start encoding. (Ignore the first <var>offset</var> bytes.)
   * @param length the number of bytes to read from the <var>bytes</var> array
   * @return a String constructed from encoding the bytes using this Encoder's current
   *         encoding, or null if there is an IOException.
   */
  public synchronized String toString(byte bytes[], int offset, int length)
    {
      // it saves an array allocation & copy to call encode() here
      // instead of returning new String(toCharArray(bytes, offset, length))
      if (length == 0)
        {
          // System.out.println("encode 0 characters (empty string): ");
          return "";
        }
      try
        {
          //System.out.print("encode " + bytes.length + " bytes:"); for (int i = offset; i < length; i++) System.out.print(bytes[i] + " "); System.out.println();
          int nChars = encode(bytes, offset, length);
          // System.out.print("encode " + nChars + "characters: "); for (int i = 0; i < nChars; i++) System.out.print((char) charBuffer[i]); System.out.println();
          String string = new String(charBuffer, 0, nChars);
          // Get rid of absurdly large char buffer right away
          if (charBuffer.length >= 1024 * 1024)
            {
              charBuffer = null;
            }
          return string;
        }
      catch (IOException e)
        {
          return null;
        }

    }

  /*
   * Encode an array of bytes into the internal {@link #charBuffer} buffer.
   * @return the number of characters that were encoded, or -1
   *         if there was
   * @param offset the byte offset to start encoding. (Ignore the first <var>offset</var> bytes.)
   * @param length the number of characters read from the <var>bytes</var> array
   * @exception IOException if the butes could not be encoded
   */
  protected int encode(byte bytes[], int offset, int length)
    throws IOException
    {
      iStream.setBytes(bytes, offset, length);
      Reader reader = getReader();
      ensureBuffer(length);
      length = reader.read(charBuffer, 0, length);
      return length;
    }

  /**
   * Convert an array of bytes into an array of characters, using the current encoding.
   * @param bytes an array of bytes to encode as a String
   * @return a new character array, constructed from the array of bytes, using the encoding.
   * This is like {@link java.lang.String#String(byte[], String) new String(bytes, getEncoding)}
   * except that it is more efficient.
   */
  public final char[] toCharArray(byte bytes[])
    {
      return toCharArray(bytes, 0, bytes.length);
    }

  /**
   * Convert an array of bytes into a string, using the current encoding.
   * @param bytes an array of bytes to encode as a String
   * @param offset the number to offset by
   * @param length then of the returned char array
   * @return the char array
   */
  public synchronized char[] toCharArray(byte bytes[], int offset, int length)
    {
      try
        {
          int nChars = encode(bytes, offset, length);
          char charArray[] = new char[nChars];
          System.arraycopy(charBuffer, 0, charArray, 0, nChars);
          return charArray;
        }
      catch (IOException e)
        {
          return null;
        }
    }

  /**
   * Convert the string to a byte array using this Encoder's current encoding.
   * @param string a string to convert to a byte array
   * @return the encoded bytes
   */
  public final byte[] toByteArray(String string)
    {
      return toByteArray(string, 0, string.length());
    }

  /**
   * Convert a substring of a string to a byte array using this Encoder's current encoding.
   * @param string a string to convert to a byte array
   * @param offset the index of the first character in the input which should be encoded.
   * @param length the number of characters to extract from the string
   * @return the encoded bytes
   */
  public synchronized final byte[] toByteArray(String string, int offset, int length)
    {
      ensureBuffer(length);
      string.getChars(offset, offset+length, charBuffer, 0);
      byte bytes[] = toByteArray(charBuffer, 0, length);
      // System.out.print(bytes.length + " bytes:"); for (int i = 0; i < bytes.length; i++) System.out.print(bytes[i] + " "); System.out.println();
      return bytes;
    }

  /**
   * Ensure that the internal charBuffer has room for at least <var>size</var>
   * characters.
   */
  protected void ensureBuffer(int size)
    {
      if (charBuffer == null || charBuffer.length < size)
        charBuffer = new char[size];
    }

  /**
   * Convert a character array to a byte array using this Encoder's current encoding.
   * @param chars an array of characters to convert to a byte array
   * @return the encoded bytes
   */
  public final byte[] toByteArray(char chars[])
    {
      return toByteArray(chars, 0, chars.length);
    }

  /**
   * Convert region of a character array to a byte array using this Encoder's current encoding.
   * @param string a string to convert to a byte array
   * @param offset the index of the first character in the input which should be encoded.
   * @param length the number of characters to encode
   * @return the encoded bytes
   */
  public synchronized byte[] toByteArray(char chars[], int offset, int length)
    {
      Writer writer = getWriter();
      oStream.reset();
      try
        {
          writer.write(chars, offset, length);
          writer.flush();
        }
      catch (IOException ioe)
        {
          System.out.println(ioe.getMessage());
          ioe.printStackTrace();
        }
      return oStream.toByteArray();
    }

  private byte[] byteArray;
  /**
   * Provide access to a reusable byte array
   * which may be filled then used as input to the
   * {@link #toString(byte[], int, int)} methods.
   * This provides the user of this Encoder a place
   * to read/write bytes without having to manage them.
   * @param nBytes the minimum number of bytes needed.
   * @return an array of bytes with at least <var>nBytes</var>
   * bytes in it; there may be more.
   */
  public byte[] getByteArray(int nBytes)
    {
      if (byteArray == null || byteArray.length <= nBytes)
        byteArray = new byte[nBytes];
      return byteArray;
    }

  /**
   * Provide access to a reusable char array
   * which may be filled then used as input to the
   * {@link #toByteArray(char[], int, int)} methods.
   * This provides the user of this Encoder a place
   * to read/write chars without having to manage them.
   * @param nChars the minimum number of chars needed.
   * @return an array of chars with at least <var>nChars</var>
   * characterss in it; there may be more.
   */
  public char[] getCharArray(int nChars)
    {
      if (charArray == null || charArray.length <= nChars)
        charArray = new char[nChars];
      return charArray;
    }


  /**
   * Read a modified UTF-8 string from a DataInput stream.
   * This method reads an unsigned short <code>utflen</code>
   * then calls {@link #readUTF(DataInput in, int utflen)}
   * @param in a DataInput stream from which a UTF-8 string is read.
   * @return a String read from the stream. Returns EmptyString
   *         if the length is 0, else constructs a new string.
   */
  public final static String readUTF(DataInput in) throws IOException
    {
      int utflen = in.readUnsignedShort();
      return readUTF(in, utflen);
    }

  /**
   * Read a modified UTF-8 string from a DataInput stream.
   * This uses the same algorithm for modified UTF-8 as
   * DataInput, but is slightly more efficient in that
   * it makes a special case for 0 length strings (this readUTF returns
   * EmptyString if <var>utflen</var> is 0). Also, this method works
   * with an integer that has already been read.
   */
  public final static String readUTF(DataInput in, int utflen) throws IOException
  {
      if (utflen == 0)
        return EmptyString;
      // To-Do: Reuse a static array with locked/synchronized access???
      char str[] = new char[utflen];
      int count = 0;
      int strlen = 0;
      while (count < utflen) {
        int c = in.readUnsignedByte();
        int char2, char3;
        switch (c >> 4) {
        case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
          // 0xxxxxxx
          count++;
          str[strlen++] = (char)c;
          break;
        case 12: case 13:
          // 110x xxxx   10xx xxxx
          count += 2;
          if (count > utflen)
            throw new UTFDataFormatException();
          char2 = in.readUnsignedByte();
          if ((char2 & 0xC0) != 0x80)
            throw new UTFDataFormatException();
          str[strlen++] = (char)(((c & 0x1F) << 6) | (char2 & 0x3F));
          break;
        case 14:
          // 1110 xxxx  10xx xxxx  10xx xxxx
          count += 3;
          if (count > utflen)
            throw new UTFDataFormatException();
          char2 = in.readUnsignedByte();
          char3 = in.readUnsignedByte();
          if (((char2 & 0xC0) != 0x80) || ((char3 & 0xC0) != 0x80))
            throw new UTFDataFormatException();
          str[strlen++] = (char)(((c & 0x0F) << 12) |
                                 ((char2 & 0x3F) << 6) |
                                 ((char3 & 0x3F) << 0));
          break;
        default:
          // 10xx xxxx,  1111 xxxx
          UTFDataFormatException e = new UTFDataFormatException("" + ((byte) c)/*I18nOK:COS*/);
          throw e;
        }
      }
      String s = new String(str, 0, strlen);
//      if (log != null) { log.print("\""); log.print(s); log.println("\""); }
      return s;
    }

  /**
   * Write a String to a DataOutput stream using modified UTF-8
   * @param string the String to write; it may be null, in which case
   * the empty string is written instead.
   * @param stream the stream to write using modified UTF-8.
   * @see java.io.DataOutput#writeUTF(String string)
   */
  public static void writeUTF(String string, DataOutputStream stream) throws IOException
  {
    string = string == null ? EmptyString : string;
    stream.writeUTF(string);
  }

    /**
     * Write a String to a DataOutput stream using modified UTF-8.
     * This is like the UTF-8 written by DataOutputStream.writeUTF
     * except that the length is written as a 4 byte integer.
     * The length (the number of bytes to follow, not the number of characters)
     * is written as a 32 bit integer as by DataOutputStream.writeInt
     * @param string The String to write; it may be null, in which case
     * the empty string is written instead.
     * @param stream the stream to write using modified UTF-8.
     * @see java.io.DataOutput#writeUTF(String string)
     * @see #utfByteCount(String)
     */
    public static void writeUTF32(String string, DataOutputStream stream) throws IOException
    {
      string = string == null ? EmptyString : string;
      stream.writeInt(utfByteCount(string));
      writeUTFData(string, stream);
    }

    /**
     * Compute the number of bytes of modified UTF-8 data to write
     * for a String.
     * @param string a non-null String
     * @return the number of bytes of data. For each character <var>c</var>
     * in the input <var>string</var> that is
     * in the range 0x0001 to 0x007F inclusive, add 1 byte.
     * For each character <var>c</var> &gt; 0x07FF, add 3 bytes, else
     * add 2 bytes. Return the sum.
     */
    static public int utfByteCount(String string)
    {
      int utfLength = 0;
      for (int i = 0, length = string.length() ; i < length ; i++)
        {
          int c = string.charAt(i);
          if ((c >= 0x0001) && (c <= 0x007F))
            utfLength++;
          else if (c > 0x07FF)
            utfLength += 3;
          else
            utfLength += 2;
        }
      return utfLength;
    }

    /**
     * Writes to the <var>stream</var>
     * the string characters from <var>string</var>
     * in modified UTF-8 format. This method assumes the
     * number of bytes to be written to the stream has already been written.
     * @param string The String to write; it may be null, in which case
     * the empty string is written instead.
     * @param stream the stream to write the modified UTF-8.
     * @see java.io.DataOutput#writeUTF(String string)
     */
    public static void writeUTFData(String string, DataOutputStream stream) throws IOException
    {
      string = string == null ? EmptyString : string;
      int length = string.length();
      int utfLength = utfByteCount(string);

      if (utfLength > 65535)
          throw new UTFDataFormatException();

      // It's a bit faster to blast all characters unconditionally
      // if there are no double byte/triple byte sequences.
      // This accounts for many (most?) strings so is worthwhile.
      if (length == utfLength)
        {
          for (int i = 0 ; i < length ; i++)
            {
              int c = string.charAt(i);
              stream.write(c);
            }
        }
      else
        {
          for (int i = 0 ; i < length ; i++)
            {
              int c = string.charAt(i);
              if ((c >= 0x0001) && (c <= 0x007F))
                stream.write(c);
              else if (c > 0x07FF)
                {
                  stream.write(0xE0 | ((c >> 12) & 0x0F));
                  stream.write(0x80 | ((c >>  6) & 0x3F));
                  stream.write(0x80 | ((c      ) & 0x3F));
                }
              else
                {
                  stream.write(0xC0 | ((c >>  6) & 0x1F));
                  stream.write(0x80 | ((c      ) & 0x3F));
                }
            }
        }
      }
        
  /**
   * Reset this Encoder. This removes references to internal structures
   * and reallocates them based on the current encoding.
   */
  public void reset()
  {
    charArray = null;
    reader = null;
    writer = null;
    iStream = null;
    oStream = null;
    try { setEncoding(encoding); }
    catch (UnsupportedEncodingException e) { }
  }

  /**
   * A list of Encoders. The keys are the encodings,
   * the values are tne Encoders, (There may be more than
   * one Encoder per encoding.)
   */
  protected static Vector encoders = new Vector();

  /**
   * Get an encoder for the encoding.
   * This will return an encoder from the internal
   * pool of available encoders. When you are done with
   * the encoder, return it to the pool with the
   * method {@link #returnInstance(EncoderInterface)}
   * These methods allow efficient reuse of Encoder
   * objects in high turnover situations.
   * @parm encoding an encoding name
   * @return an encoder for that encoding, taken from
   * the pool of available encoders, else a new encoder.
   * @see #returnInstance(EncoderInterface)
   */
  public static EncoderInterface getInstance(String encoding)
    throws UnsupportedEncodingException
  {
    if (DEBUG)
      {
        System.out.println("_DEBUG_:Looking for encoder " + encoding/*I18nOK:COS*/);
      }
    synchronized(encoders)
      {
        EncoderInterface encoder;
        for (int i = 0, encodersLen = encoders.size(); i < encodersLen; i++)
          {
            EncoderInterface enc = (EncoderInterface) encoders.elementAt(i);
            if (encoding != null && encoding.equals(enc.getEncoding()))
              {
                encoders.removeElementAt(i);
                if (DEBUG)
                  {
                    System.out.println("_DEBUG_:Found encoder in pool"/*I18nOK:COS*/);
                  }
                return enc;
              }
          }

        if (encoding.equalsIgnoreCase("cp1047")
            && needsCp1047Patch())
          encoder = new Cp1047Encoder();
        else
          encoder = new Encoder(encoding);
        if (DEBUG)
          System.out.println("_DEBUG_:Creating new encoder " + encoder.getClass().getName()/*I18nOK:COS*/);
        return encoder;
      }
  }

     /**
      * Return an encoder to the pool of available encoders
      * if the encoder pool is not full.
      * @param encoder the encoder to return to the pool.
      * @see #getInstance(String)
      */
     public static void returnInstance(EncoderInterface encoder)
     {
          if (DEBUG)
          {
               System.out.println("_DEBUG_:Returning encoder to pool: " + encoder.getEncoding()/*I18nOK:COS*/);
          }

          synchronized (encoders)
          {
               if ( encoders.size() < MAX_POOL_SIZE )
               {
                    try
                    {
                         ((Encoder)encoder).charBuffer = null;
                    }
                    catch (ClassCastException cce)
                    {
                    }
                    encoders.addElement(encoder);
               }
               else if (DEBUG)
               {
                    System.out.println("_DEBUG_:No room in pool; discarding encoder");
               }
          }
     }

  static final byte AA[] = { 65, (byte) 0xC1 };  // 'A' in ASCII, 'A' in EBCDIC

  public static boolean isValid(String encoding)
  {
    try
      {
        if (encoding.equals("Cp1047"/*I18nOK:EMS*/) && needsCp1047Patch() )
          return true;

        String s = new String(AA, encoding);
        // System.out.println("Note: Encoding " + encoding + " is valid");
        return true;
      }
    catch (UnsupportedEncodingException e)
      {
//*//        if (encoding.equals(encoding.toUpperCase()))
//*//          System.out.println("ERROR: Encoding " + encoding + " is invalid");
//*//        else
//*//          return isValid(encoding.toUpperCase());
      }
    return false;
  }

  /**
   * Return the String representation of the object
   */
  public String toString()
  {
    return encoding;
  }

  /**
   * Compare for equality to another instance
   * @param other another instance if Encoder
   * @return true if this object equals the <var>other</var>
   */
  public boolean equals(Encoder enc)
  {
    if (enc == null)
      return false;
    return enc.getEncoding().equals(getEncoding());
  }

  /**
   * Compare for equality to another instance
   * @param other another object
   * @return true if this object equals the <var>other</var>
   */
  public boolean equals(Object object)
  {
    if (object instanceof Encoder)
      return equals((Encoder) object);
    return false;
  }

  /**
   * Return a hash code for this object.
   * @return a hash code for this object.
   */
  public int hashCode()
  {
    return encoding != null ? encoding.hashCode() : 0;
  }

  private static Boolean needsCp1047Patch = null;
  private static boolean needsCp1047Patch()
  {
    if ( needsCp1047Patch == null )
      {
        needsCp1047Patch = Boolean.TRUE;
        try
          {
            if (new String(AA, "Cp1047"/*I18nOK:COS*/).charAt(1) == 'A')
              needsCp1047Patch = Boolean.FALSE;
          }
        catch (UnsupportedEncodingException e) { }
      }
    return needsCp1047Patch.booleanValue();
  }

//  private static String Cp1047_Replacement = null;
//   public static String patchEncoding(String encoding)
//   {
//     // Cp1047 is IBM Open Edition US EBCDIC
//     // (see http://www.s390.ibm.com/products/oe/bpxqs11.html)
//     // but JavaSoft got it wrong and thinks Cp1046 is US EBCDIC
//     // (see http://dpdweb.pc.sas.com/Java/jdk1.1.7/docs/guide/intl/encoding.doc.html )
//     // But Cp1046 is really an ASCII encoding for Arabic
//     // (see http://www.austin.ibm.com/doc_link/en_US/a_doc_lib/aixkybd/kybdtech/Arabic2.htm )
//     //
//     // We test character 1 which is the character encoded from 0xC1
//     // and if it == 'A', then the JavaSoft defect has been fixed
//     // and we can use the encoding.
//     // If it is not 'A', then the Cp1047 encoder is not encoding EBCDIC,
//     // so substitute Cp037 instead (if it works.)
//     //
//     if ( ! "Cp1047".equalsIgnoreCase(encoding) )
//       return encoding;
//
//     // Already tried a replacement for Cp1047?
//     if (Cp1047_Replacement != null)
//       return Cp1047_Replacement;
//
//     try
//       {
//         String s = new String(AA, encoding);
//         if (s.charAt(1) == 'A')
//           {
//             Cp1047_Replacement = encoding;
//           }
//         else
//           {
//             // if Cp037 is not supported, this will throw an exception
//             s = new String(AA, "Cp037");
//             Cp1047_Replacement = "Cp037";
//             encoding = Cp1047_Replacement;
//           }
//       }
//     catch (UnsupportedEncodingException e) { }
//     return encoding;
//   }


}

