//Copyright (c) 2001-2003 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.rmi.RemoteException;

/**
 * Provides a serializabe extension to InputSteam, allowing it to be used in 
 * both local and remote environments.  
 *
 * @author <a href="mailto:rob.owen@sas.com">Rob Owen</a>
 * @version 1.1 3-Mar-2003
 * @since   1.1
 * 
 * @history
 *    3-Mar-2003 ro Created.
 *
 */
public class InputStreamAdapter extends InputStream implements Serializable {

    static final long serialVersionUID = 4866449630932653718L;
    
    /**
     * Private copy of the remote stream to which we are delegating.
     */
    private RemoteInputStreamInterface rin;

    /**
     * Creates an input stream adapter from an existing input stream.
     * @param in The existing input stream
     * @throws RemoteException in the case of a network failure
     */
    public InputStreamAdapter( InputStream in ) throws RemoteException {
        this( new RemoteInputStream( in ) );
    }

	/**
	 * Creates an input stream adapter from a remote input stream.
	 * @param in The existing remote input stream
	 */
    public InputStreamAdapter( RemoteInputStreamInterface rin ) {
        this.rin = rin;
    }

	/**
	 * Returns the number of bytes that can be read (or skipped over) from
	 * this input stream without blocking by the next caller of a method for
	 * this input stream.  The next caller might be the same thread or or
	 * another thread.
	 *
	 * @return the number of bytes that can be read from this input stream
	 *         without blocking.
	 * @throws IOException if an I/O error occurs.
	 * @see java.io.InputStream#available
	 */
    public int available() throws IOException {
        return rin.available();
    }

	/**
	 * Closes this input stream and releases any system resources associated
	 * with the stream.
	 *
	 * @throws IOException if an I/O error occurs.
	 * @see java.io.InputStream#close
	 */
    public void close() throws IOException {
        rin.close();
    }

	/**
	 * Marks the current position in this input stream. A subsequent call to
	 * the <code>reset</code> method repositions this stream at the last marked
	 * position so that subsequent reads re-read the same bytes.
	 *
	 * <p> The <code>readlimit</code> argument tells this input stream to
	 * allow that many bytes to be read before the mark position gets
	 * invalidated.
	 *
	 * <p> The general contract of <code>mark</code> is that, if the method
	 * <code>markSupported</code> returns <code>true</code>, the stream somehow
	 * remembers all the bytes read after the call to <code>mark</code> and
	 * stands ready to supply those same bytes again if and whenever the method
	 * <code>reset</code> is called.  However, the stream is not required to
	 * remember any data at all if more than <code>readlimit</code> bytes are
	 * read from the stream before <code>reset</code> is called.
	 *
	 * @param readlimit the maximum limit of bytes that can be read before
	 *                  the mark position becomes invalid.
	 * @throws RemoteException if a network I/O error occurs.
	 * @see java.io.InputStream#close
	 */
    public synchronized void mark( int readlimit ) {
        try {
            rin.mark( readlimit );
        } catch ( RemoteException e ) {
            throw new java.lang.reflect.UndeclaredThrowableException( e );
        }
    }

	/**
	 * Tests if this input stream supports the <code>mark</code> and
	 * <code>reset</code> methods. Whether or not <code>mark</code> and
	 * <code>reset</code> are supported is an invariant property of a
	 * particular input stream instance.
	 *
	 * @return <code>true</code> if this stream instance supports the mark
	 *         and reset methods; <code>false</code> otherwise.
	 * @throws RemoteException if a network I/O error occurs.
	 * @see java.io.InputStream#markSupported
	 */
    public boolean markSupported() {
        try {
            return rin.markSupported();
        } catch ( RemoteException e ) {
            throw new java.lang.reflect.UndeclaredThrowableException( e );
        }
    }

	/**
	 * Reads the next byte of data from the input stream. The value byte is
	 * returned as an <code>int</code> in the range <code>0</code> to
	 * <code>255</code>. If no byte is available because the end of the stream
	 * has been reached, the value <code>-1</code> is returned. This method
	 * blocks until input data is available, the end of the stream is detected,
	 * or an exception is thrown.
	 *
	 * @return the next byte of data, or <code>-1</code> if the end of the
	 *         stream is reached.
	 * @throws IOException if an I/O error occurs.
	 * @see java.io.InputStream#read()
	 */
    public int read() throws IOException {
        return rin.read();
    }

	/**
	 * Reads bytes from the input stream and returns them in the
	 * byte array b and returns the actual number of bytes read. This method 
	 * blocks until input data is available, end of file is detected, or an 
	 * exception is thrown.
	 *
	 * @param b The byte array into which the bytes are read
	 * @return int The number of bytes read into b
	 * @throws IOException if an I/O error occurs.
	 * @see java.io.InputStream#read(byte[])
	 */
    public int read( byte[] b ) throws IOException {
        byte tempbuf[] = rin.read( b.length );

        /* check for end of file */
        if (tempbuf == null)
            return -1;

        /* otherwise copy tempbuf into b */
        System.arraycopy( tempbuf, 0, b, 0, tempbuf.length );
        return tempbuf.length;
    }

	/**
	 * Reads bytes from the input stream and returns them in the
	 * byte array b, starting at offset <code>off</code> and a maximum 
	 * of <code>len</code> bytes. This method blocks until input data is 
	 * available, end of file is detected, or an exception is thrown.
	 *
	 * @param b The byte array into which the bytes are read
	 * @param off The offset into the byte array where the information will be stored
	 * @param len The maximum number of bytes that can be read
	 * @return int The actual number of bytes read into b
	 * @throws IOException if an I/O error occurs.
	 * @see java.io.InputStream#read(byte[],int,int)
	 */
    public int read( byte[] b, int off, int len ) throws IOException {
        byte tempbuf[] = rin.read(len);

        /* check for end of file */
        if ( tempbuf == null ) return -1;

        /* otherwise copy tempbuf into b */
        System.arraycopy( tempbuf, 0, b, off, tempbuf.length );

        return tempbuf.length;
    }

	/**
	 * Repositions this stream to the position at the time the
	 * <code>mark</code> method was last called on this input stream.
	 *
	 * <p> The general contract of <code>reset</code> is:
	 *
	 * <p><ul>
	 *
	 * <li> If the method <code>markSupported</code> returns
	 * <code>true</code>, then:
	 *
	 *     <ul><li> If the method <code>mark</code> has not been called since
	 *     the stream was created, or the number of bytes read from the stream
	 *     since <code>mark</code> was last called is larger than the argument
	 *     to <code>mark</code> at that last call, then an
	 *     <code>IOException</code> might be thrown.
	 *
	 *     <li> If such an <code>IOException</code> is not thrown, then the
	 *     stream is reset to a state such that all the bytes read since the
	 *     most recent call to <code>mark</code> (or since the start of the
	 *     file, if <code>mark</code> has not been called) will be resupplied
	 *     to subsequent callers of the <code>read</code> method, followed by
	 *     any bytes that otherwise would have been the next input data as of
	 *     the time of the call to <code>reset</code>. </ul>
	 *
	 * <li> If the method <code>markSupported</code> returns
	 * <code>false</code>, then:
	 *
	 *     <ul><li> The call to <code>reset</code> may throw an
	 *     <code>IOException</code>.
	 *
	 *     <li> If an <code>IOException</code> is not thrown, then the stream
	 *     is reset to a fixed state that depends on the particular type of the
	 *     input stream and how it was created. The bytes that will be supplied
	 *     to subsequent callers of the <code>read</code> method depend on the
	 *     particular type of the input stream. </ul></ul>
	 *
	 * @throws IOException if an I/O error occurs.
	 * @see java.io.InputStream#reset
	 */
    public synchronized void reset() throws IOException {
        rin.reset();
    }

	/**
	 * Skips over and discards <code>n</code> bytes of data from this input
	 * stream. The <code>skip</code> method may, for a variety of reasons, end
	 * up skipping over some smaller number of bytes, possibly <code>0</code>.
	 * This may result from any of a number of conditions; reaching end of file
	 * before <code>n</code> bytes have been skipped is only one possibility.
	 * The actual number of bytes skipped is returned.  If <code>n</code> is
	 * negative, no bytes are skipped.
	 *
	 * @param n the number of bytes to be skipped.
	 * @return the actual number of bytes skipped.
	 * @throws IOException if an I/O error occurs.
	 * @see java.io.InputStream#skip(long)
	 */
    public long skip( long n ) throws IOException {
        return rin.skip( n );
    }

}
