// Copyright (c) 2002 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Map;


/**
 * Provides the capability to read header information from an
 * com.sas.iom.InputStream IOM InputStream.
 *
 * @author <a href="mailto:bryan.wolfe@sas.com">Bryan Wolfe</a>
 * @author <a href="mailto:tony.dean@sas.com">Tony Dean</a>
 * @version 1.0
 * @since 1.0
 *
 * @history
 *    28Aug2002 td  Created.
 *
 */
public interface InputStreamHeaderInterface extends Remote {

   /**
    * Returns the encoding of the response.<p>
    * The encoding is specified by the <code>charset</code>
    * attribute of the <code>content-type</code> header field.
    *
    * @return the character encoding of the response stream, or
    *         <code>null</code> if not known.
    *
    * @throws IOException if an I/O error occurs.
    * @throws RemoteException if a network problem occurs.
    */
   public String getCharacterEncoding( ) throws IOException, RemoteException;

   /**
    * Returns the value of the response <code>content-encoding</code> header field.
    *
    * @return the content encoding of the response stream, or <code>null</code>
    *          if not known.
    *
    * @throws IOException if an I/O error occurs.
    * @throws RemoteException if a network problem occurs.
    */
   public String getContentEncoding( ) throws IOException, RemoteException;

   /**
    * Returns the value of the response <code>content-type</code> header field.
    *
    * @return the content type of the response stream, or <code>null</code>
    *          if not known.
    *
    * @throws IOException if an I/O error occurs.
    * @throws RemoteException if a network problem occurs.
    */
   public String getContentType( ) throws IOException, RemoteException;

   /**
    * Returns the value for the named response header field parsed as a
    * date.  {@link java.text.DateFormat} is used to obtain a
    * date formatter (with the default formatting style for the
    * default locale) to parse the field.  The result is the
    * number of milliseconds since January 1,
    * 1970 GMT represented by the named field.  It returns the
    * <code>Default</code> value if the field does not exist.
    *
    * @return the value of the named header field as a date
    * @param name the name of a header field.
    * @param Default the default value.
    *
    * @throws IOException if an I/O error occurs.
    * @throws RemoteException if a network problem occurs.
    */
   public long getHeaderFieldDate( String name, long Default ) throws IOException, RemoteException;

   /**
    * Returns the value for the named response header field parsed as a
    * number. It returns the <code>Default</code> value if the field
    * does not exist.  {@link java.lang.Integer#parseInt(String)} is used
    * to parse the field.
    *
    * @return the value of the named header field as in int.
    * @param name the name of a header field.
    * @param Default the default value.
    *
    * @throws IOException if an I/O error occurs.
    * @throws RemoteException if a network problem occurs.
    */
   public int getHeaderFieldInt( String name, int Default ) throws IOException, RemoteException;

   /**
    * Returns response headers as a Map.
    *
    * @return the list of all response headers.
    * @throws IOException if an I/O error occurs.
    * @throws RemoteException if a network problem occurs.
    */
   public Map getHeaders( ) throws IOException, RemoteException;

   /**
    * Tests whether or not the header is available for this input stream.<p>
    * The other header methods will block until the header is available.
    * This method will not block.  The header is considered available once the
    * stream writer has written any data.
    *
    * @return <code>true</code> if the header is available; <code>false</code> otherwise.
    * @throws RemoteException if a network problem occurs.
    */
   public boolean headerAvailable() throws RemoteException;

}
