// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.InputStream;
import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;

/**
 * An object which provides input streams which may be remote.   Classes
 * implement this interface if they provide one ore more input streams.
 *
 * @author  Russell Gonsalves
 * @version 1.0 3Dec01
 */

public interface InputStreamProviderInterface extends Remote {
    /**
    * Check if getInputStream will succeed.  See {@link #getInputStream} to
    * determine whether this method should be used.
    *
    * @return non-null if getInputStream() would succeed; throws UnmarshalException if it would fail
    * @throws UnmarshalException if invoked on a remote object
    * @throws RemoteException if a network I/O error occurs.
    */
	Object hasInputStream() throws UnmarshalException, RemoteException;

    /**
    * Gets an InputStream.  If this method is invoked on a remote object, a
    * UnmarshalException exception is thrown if the InputStream was successfully
    * created since InputStream is not serializable.  However, the InputStream
    * will have been created.  If
    * this is an issue {@link #hasInputStream} may be invoked to determine
    * if getInputStream would succeed.  If an InputStream is needed in a
    * remote context
    * getRemoteInputStream should be invoked.
    *
    * @param streamKey is used to identify a specific input stream
    * when a provider supports multiple input streams.  It is implementation
    * specific and is typically null if only one stream is provided.
    *
    * @return an InputStream from the provider, as identified by
    * the <var>streamKey</var>; null if the provider does not provide the stream.
    *
    * @throws UnmarshalException if invoked on a remote object
    * @throws IOException if an exception occurred while construction the InputStream
    * @throws RemoteException if a network I/O error occurs.
    */
	InputStream getInputStream(Object streamKey) throws UnmarshalException, IOException, RemoteException;

    /**
    * Gets a remote input stream.  This method is invoked on a remote
    * object or if the returned InputStream needs to be remoted.
    *
    * @param streamKey is used to identify a specific input stream
    * when a provider supports multiple input streams.  It is implementation
    * specific and is typically null if only one stream is provided.
    *
    * @return an InputStream from the provider, as identified by the
    * <var>streamKey</var>
    *
    * @throws UnmarshalException if invoked on a remote object
    * @throws IOException if an exception occurred while construction the RemoteInputStreamInterface
    * @throws RemoteException if a network I/O error occurs.
    */
	RemoteInputStreamInterface getRemoteInputStream(Object streamKey) throws IOException, RemoteException;
}
