package com.sas.io;

//
// History:
//
// Date    Name            Remarks
// 18Dec00 D Biesack       S0105475: Fix memory leak with Cp1047 encoder.
//
/**
 * LookupEncoder provides a simple way of converting or encoding text
 * between Unicode and a single byte code page.
 * The encoding is done by looking up a character code in a table.
 * LookupEncoder supports two tables: one for mapping from
 * Unicode to the single byte code page, and one for going the other direction.
 * For example, the character 'C' is at position 67 in the Unicode code table.
 * The encoding in the lookup table is determined by indexing
 * character [67] in the <var>fromUnicode</var> table.
 * A byte from a byte array is converted to Unicode by indexing that
 * byte value in the <var>toUnicode</var> table.
 * <p>
 * <strong>Note:</strong>
 * This class does not handle Unicode characters outside the range [0..255]
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @since 1.2
 */
public class LookupEncoder
      implements EncoderInterface
{

  private String encoding;
  private char[] toUnicode;
  private char[] fromUnicode;
  private final static int CODE_TABLE_SIZE = 256;
  public static boolean DEBUG = false;

  /**
   * Construct a new lookup encoder for the encoding <var>name</var>,
   * which uses the lookup tables.
   * @param toUnicode the lookup table to use when encoding an array of bytes
   *        as Unicode. If null, the inverse of the <var>fromUnicode</var>
   *        table is used.
   * @param fromUnicode the lookup table to use when encoding a String
   *        as an array of bytes. If null, the inverse of the <var>toUnicode</var>
   *        table is used.
   */
  public LookupEncoder(String name, char[] toUnicode, char[] fromUnicode)
  {
    this.encoding = name;
    setUnicodeArrays(fromUnicode, toUnicode);
  }

  /**
   * Set the <var>toUnicode</var> lookup table.
   * If the <var>fromUnicode</var> table is null,
   * is is generated as the inverse of the <var>toUnicode</var> table.
   * @param toUnicode the lookup table to use to convert
   *        from single bytes to Unicode characters.
   *        If this table is less than 256 characters long,
   *        an identify relation is used whn encoding.
   *        (i.e. bytes n through 255 will map to Unicode characters n through 255)
   */
  protected void setToUnicodeArray(char toUnicode[])
  {
    this.toUnicode = toUnicode;
    if (fromUnicode == null)
       fromUnicode = invert(toUnicode);
  }

  /**
   * Set the <var>fromUnicode</var> lookup table.
   * If the <var>toUnicode</var> table is null,
   * is is generated as the inverse of the <var>fromUnicode</var> table.
   * @param fromUnicode the lookup table to use to convert
   *        from Unicode characters to single bytes.
   *        If this table is less than 256 characters long,
   *        an identify relation is used whn encoding.
   *        (i.e. Unicode characters n through 255 will map to bytes n through 255)
   */
  protected void setFromUnicodeArray(char fromUnicode[])
  {
    this.fromUnicode = fromUnicode;
    this.toUnicode = invert(fromUnicode);
  }

  private char[] invert(char chars[])
  {
    int max = 0;
    for (int i = 0; i < chars.length; i++)
    {
        int c = chars[i] & 0xff;
        max = Math.max(max, c);
    }
    char inverse[] = new char[max+1];
    for (int i = max - 1; i>=0; i--)
      inverse[i] = (char) i;
    for (int i = 0; i < max; i++)
      {
        int c = chars[i] & 0xff;
        inverse[c] = (char) i;
      }
    return inverse;
  }

  protected void setUnicodeArrays(char fromUnicode[], char toUnicode[])
  {
    if (fromUnicode == null && toUnicode == null)
      throw new IllegalArgumentException();
    if (fromUnicode == null)
      setToUnicodeArray(toUnicode);
    else if (toUnicode == null)
      setFromUnicodeArray(fromUnicode);
    else
      {
        this.fromUnicode = fromUnicode;
        this.toUnicode = toUnicode;
      }
  }


  /**
   * Convert an array of bytes into a String.
   * @param bytes an array of bytes to encode as a String
   * @param offset the byte offset to start encoding. (Ignore the first <var>offset</var> bytes.)
   * @param length the number of bytes to read from the <var>bytes</var> array
   * @return a String constructed from encoding the bytes using this Encoder's current
   *         encoding, or null if there is an IOException.
   */
  public String toString(byte bytes[], int offset, int length)
  {
    return new String(toCharArray(bytes, offset, length));
  }

  /**
   * Convert an array of bytes into a string.
   * @param bytes an array of bytes to encode as a String
   * @param offset the number to offset
   * @param length the length of the array	
   * @return the char array
   */
  public char[] toCharArray(byte bytes[], int offset, int length)
  {
    char chars[] = new char[length];
    for (int there = offset, here = 0, end = offset+length; there < end; here++, there++)
      chars[here] = toChar(bytes[there]);
    return chars;
  }

  /**
   * Convert a byte value to a character.
   * @param b a byte value
   * @retun the Unicode character <var>toUnicode[b]</var>
   */
  public char toChar(byte b)
  {
    int index = ((int)b & 0xff); // necessary to prevent sign extension
    char c;
    if (index <= toUnicode.length)
      c = toUnicode[index];
    else
      c = (char) index;
    if ( DEBUG )
      System.out.println(getClass().getName() + "_DEBUG_: byte " + /*I18nOK:LINE*/
                         Integer.toHexString(((int) b) & 0xff) + " ==> char " + /*I18nOK:LINE*/
                         Integer.toHexString(((int) c) & 0xff) + " == '" + ((char) c) + "'" /*I18nOK:LINE*/
                         );/*I18nOK:LINE*/
    return c;
  }

  /**
   * Convert a Unicode character to a byte value
   * @param c a character
   * @retun the byte <var>fromUnicode[c]</var>
   */
  public byte toByte(char c)
  {
    int index = ((int)c & 0xff);
    byte b;
    if (index <= fromUnicode.length)
      b = (byte) fromUnicode[index];
    else
      b = (byte) index;
    if ( DEBUG )
      System.out.println(getClass().getName() + "_DEBUG_: char " + /*I18nOK:LINE*/
                         Integer.toHexString(((int) c) & 0xff) + " == '" + ((char) c) + "'" + /*I18nOK:LINE*/
                         " ==> byte " + Integer.toHexString(((int) b) & 0xff) /*I18nOK:LINE*/
                          );/*I18nOK:LINE*/
    return b;
  }

  /**
   * Convert a substring of a string to a byte array using this Encoder's current encoding.
   * @param string a string to convert to a byte array
   * @param offset the index of the first character in the input which should be encoded.
   * @param length the number of characters to extract from the string
   * @return the encoded bytes
   */
  public final byte[] toByteArray(String string, int offset, int length)
  {
    return toByteArray(string.substring(0, length).toCharArray(), 0, length);
  }

  /**
   * Convert region of a character array to a byte array using this Encoder's current encoding.
   * @param string a string to convert to a byte array
   * @param offset the index of the first character in the input which should be encoded.
   * @param length the number of characters to encode
   * @return the encoded bytes
   */
  public byte[] toByteArray(char chars[], int offset, int length)
  {
    byte bytes[] = new byte[length];
    for (int there = offset, here = 0, end = offset+length; there < end; here++, there++)
      bytes[here] = toByte(chars[there]);
    return bytes;
  }

  private byte[] byteArray;
  /**
   * Provide access to a reusable byte array
   * which may be filled then used as input to the
   * {@link #toString(byte[], int, int)} methods.
   * This provides the user of this Encoder a place
   * to read/write bytes without having to manage them.
   * @param nBytes the minimum number of bytes needed.
   * @return an array of bytes with at least <var>nBytes</var>
   * bytes in it; there may be more.
   */
  public byte[] getByteArray(int nBytes)
    {
      if (byteArray == null || byteArray.length <= nBytes)
        byteArray = new byte[nBytes];
      return byteArray;
    }

  /**
   * Return the name of the character encoding
   * @return the name of the character encoding
   */
  public String getEncoding()
  {
    return encoding;
  }

  /**
   * Return the String representation of the object
   */
  public String toString()
  {
    return "(" + getEncoding() + ")";
  }

  /**
   * Compare for equality to another instance
   * @param other another instance if ThisClas
   * @return true if this object equals the <var>other</var>
   */
  public boolean equals(LookupEncoder other)
  {
    if (other == null)
      return false;
    if (getEncoding() == other.getEncoding()
        || (getEncoding() != null
            && getEncoding().equals(other.getEncoding())))
      {
        char otherTo[] = other.toUnicode;
        char otherFrom[] = other.fromUnicode;
        return (   arrayEqual(otherTo, toUnicode)
                && arrayEqual(otherFrom, fromUnicode));
      }
    return false;
  }

  private static boolean arrayEqual(char[] charArrayA, char[] charArrayB)
    {
      if (charArrayA.length != charArrayB.length)
        return false;
      else
        {
          for (int i = charArrayA.length-1; i >= 0; i--)
            {
              if (charArrayA[i] != charArrayB[i])
                return false;
            }
        }
      return true;
    }


  /**
   * Compare for equality to another instance
   * @param other another object
   * @return true if this object equals the <var>other</var>
   */
  public boolean equals(Object object)
  {
    if (object instanceof LookupEncoder)
      return equals((LookupEncoder) object);
    return false;
  }

  /**
   * Return a hash code for this object.
   * @return a hash code for this object.
   */
  public int hashCode()
  {
    return ((getEncoding() != null)
            ? (getEncoding().hashCode())
            : 1963);
  }

  /**
   * Clone the object.
   * @return a clone of the object.
   */
  public Object clone() throws CloneNotSupportedException
    {
      LookupEncoder clone = (LookupEncoder) super.clone();
      return clone;
    }


}
