// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.io;

import java.io.Serializable;
import java.io.StreamTokenizer;
import java.io.Reader;
import java.io.IOException;
import java.text.NumberFormat;
import java.text.ParsePosition;

import com.sas.util.NumberUtil;

/**
 * ObjectTokenizer is a StreamTokenizer which also parses some
 * simple object values, such as integers, longs, doubles,
 * and strings. Unquoted strings are allowed, but double quotes
 * are necessary for items which contains embedded quotes
 * or numbers, for example. Character objects are recognized with single quotes.
 * You can configure what types of numbers
 * the parser returns; the default is integer data is returned
 * as Integer object is in the Integer range, else Long, and
 * floating point values are returned as Doubles.
 *
 * <strong>Note</strong> There is a known defect in the JDK java.io.StreamTokenizer
 * (as of 1.1.5) in that parseNumbers() causes it to parse numeric values, but it
 * does not parse scientific notation, such as 1.0e-6. This class will parse
 * such numbers when you call readObject(), but not when you call parseNumbers(); nextToken();
 * and get back a TT_NUMBER token.
 *
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * <!--
 * @support sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @history
 * <dl>
 * <dt>Wed Sep 11 1996  <dd>Created
 * </dl>
 * -->
 */

public class ObjectTokenizer
  extends StreamTokenizer
{

  public static final int TT_OBJECT = 100 * (TT_EOF + TT_NUMBER + TT_WORD); // unique enough?

  protected Object lastObject;

  private boolean parseCharacters;
  private boolean parseBooleans;
  private boolean parseBytes;
  private boolean parseShorts;
  private boolean parseIntegers;
  private boolean parseLongs;
  private boolean parseFloats;
  private boolean parseDoubles;
  private int token;
  private StringBuffer buffer;
  // a double formatted using the current Locale
  // which is used to determine the set of characters
  // in a numeric token. However, if the Locale's
  private String localeFormattedNumber;


  /**
   * Construct an ObjectTokenizer from an input Reader.
   * The tokeinzer state is initialized to
   * parse strings and numbers.
   * The default state is to parse booleans, integers, longs,
   * and doubles along with String objects.
   * @param reader an input reader which we tokenize.
   */
  public ObjectTokenizer(Reader reader)
    {
      super(reader);
      resetSyntax();
      quoteChar('\"');  // so that strings are read as tokens
      // no special whitespace characters
      ordinaryChars(0, ' ');

      // parseNumbers();

    }

  public void resetSyntax()
    {
      super.resetSyntax();
      ordinaryChars(0, ' ');
      parseCharacters();
      parseBytes(false);
      parseShorts(false);
      parseIntegers(false);
      parseLongs(false);
      parseFloats(false);
      parseDoubles(false);
      parseBooleans(false);
    }

  protected boolean parsingNumbers()
    {
      // note: Using | instead of || here is ok.
      return (parseBytes | parseShorts | parseIntegers | parseLongs | parseFloats | parseDoubles);
    }

  private boolean tokenIsNumeric(String delimiters)
    {
      // note: Using | instead of || here, and using & instead of && here is ok.
      // return (token == '-' | token == '.' | (token >= '0' & token <= '9')) && parsingNumbers();

      // I18N correct version:
      if (delimiters.indexOf((char) token) >= 0)
        return false;
      if (localeFormattedNumber == null)
        {
          // format a simple double using the current Locale
          // Use -1.5 because it can be represented/formatted as a duuble exactly
          NumberFormat format = NumberFormat.getInstance();
          localeFormattedNumber = format.format(-1.5D);
          // System.out.println("localeFormattedNumber == \"" + localeFormattedNumber + "\"");
        }
      return Character.isDigit((char) token) || localeFormattedNumber.indexOf(token) >= 0;
    }

  private void resetStringBuffer()
    {
      if (buffer == null)
        buffer = new StringBuffer(16);
      buffer.setLength(0);
    }

  /**
   * Read an object from the token stream.
   * <strong>Note: This method resets the stream syntax
   * so that numbers and quoted strings may be parsed.
   * It does <em>not</em> restore the tokenizer syntax state.
   * (for example, it is left with no word chars, no
   * whitespace chars), no comment chars, etc.</strong>
   * @param delimiters is a set of characters which
   * mark object boundaries, so that unquoted strings may be
   * tokenized until an EOF or a delimiter is hit. Note that
   * single quotes '\'' and double quotes '\"' are used for
   * character and string delimiters, and numeric characters
   * (digits, '.' or ',', '+', '-', 'e', 'E') are used for parsing
   * numeric values and therefore are not recognized as
   * delimiters when parsing numbers.
   * @param allowNull if true, recognize the special word "null"
   * and set the last object read to null instead of the String "null"
   * @return a token: TT_EOF if end of file was found, or TT_OBJECT
   * @see com.sas.io.ObjectTokenizer#getLastObject() to
   * get the value of the object that was parsed, if TT_OBJECT.
   * @exception java.io.IOException if there is an IO exception
   * on the reader.
   */
  public int nextObject(String delimiters, boolean allowNull)
          throws IOException
    {
      lastObject = null;
      parseObjects();
      skipWhitespace(delimiters);
      token = nextToken();
      if (token == TT_EOF)
        return TT_EOF;

      if (token == '\'' && sval.length() == 1 && parseCharacters)
        // have parsed a character literal
        {
          lastObject = new Character(sval.charAt(0));
        }

      // note: Using | instead of || here is ok.
      else if (token == '\"' | token == '\'')
        // have parsed a string literal
        {
          lastObject = sval;
        }

      else if (delimiters.indexOf(token) != -1)
        return token;

      else
        parseAnObject(delimiters, allowNull);
      // System.out.println("token: " + (object == null? "" : object.getClass().toString()) + " " + object);

      skipWhitespace(delimiters);
      return TT_OBJECT;
    }

  private void parseAnObject(String delimiters, boolean allowNull) throws IOException
    {
      resetStringBuffer();
      boolean numeric = tokenIsNumeric(delimiters);
      if (numeric)
        {
          while (tokenIsNumeric(delimiters) || token == 'e' || token == 'E')
            {
              buffer.append((char) token);
              token = nextToken();
            }
          int len = buffer.length();

          skipWhitespace(delimiters, true); // skip whitespace, but possibly preserve it.

          // if have end of file, or a delimiter, then the text is complete
          // and we can try to parse as a number. If there is any other token
          // here before the EOF or delimiter, then we have a non-numeric
          // value such as "4/5" or "10:15pm" or "1 2 3" etc.
          if (token == TT_EOF || delimiters.indexOf(token) != -1)
            {
              pushBack();
              try
                {
                  buffer.setLength(len);
                  nval = NumberUtil.doubleValueOf(buffer.toString()/*I18nOK:LSM*/); 
                  // System.out.println("Found a number \"" + buffer.toString() + "\" == " + d + ", " + nval);
                  if (      parseBytes             && nval >= Byte.MIN_VALUE    && nval <= Byte.MAX_VALUE    && (nval == (byte)  nval))
                    lastObject = new Byte((byte)nval);
                  else if ( parseShorts            && nval >= Short.MIN_VALUE   && nval <= Short.MAX_VALUE   && (nval == (short) nval))
                    lastObject = new Short((short)nval);
                  else if ( parseIntegers          && nval >= Integer.MIN_VALUE && nval <= Integer.MAX_VALUE && (nval == (int)   nval))
                    lastObject = new Integer((int)nval);
                  else if ( parseLongs             && nval >= Long.MIN_VALUE    && nval <= Long.MAX_VALUE    && (nval == (long)  nval))
                    lastObject = new Long((long)nval);
                  else if ( parseFloats            && nval >= Float.MIN_VALUE   && nval <= Float.MAX_VALUE)
                    lastObject = new Float((float)nval);
                  else if ( parseDoubles )
                    lastObject = new Double(nval);
                  else
                    lastObject = sval;
                  // System.out.println("ObjectTokenizer parsed double nval= " + nval + " from \"" + buffer + "\"");
                  return;
                }
              catch (Exception e)
                {
                  // System.out.println("ObjectTokenizer parsed string sval from \"" + buffer + "\"");
                }
            }
        }
      resetSyntax();
      // no special whitespace characters
      ordinaryChars(0, ' ');
      while (token != TT_EOF && delimiters.indexOf(token) == -1)
        {
          if (sval == null)
            buffer.append((char)token);
          else
            buffer.append(sval);
          token = nextToken();
        }
      pushBack();

      parseBooleans(true);

      // NOTE: String.trim() assumes anything <= 0x020 is whitespace.
      String itemText = buffer.toString()/*I18nOK:LSM*/;
      String trimmedText = itemText.trim();
      // System.out.println("ObjectTokenizer: trimmedText=\"" + trimmedText + "\", parseBooleans = " + parseBooleans);
      if (allowNull && trimmedText.equals("null"))
        lastObject = null;
      else if (parseBooleans && trimmedText.equals("true"))
        lastObject = Boolean.TRUE;
      else if (parseBooleans && trimmedText.equals("false"))
        lastObject = Boolean.FALSE;
      else
        lastObject = trimmedText;

      // This is necessary, even though we do this at the top
      // of the next nextObject() call:
      parseObjects();
      return;
    }

  /**
   * Return the value of the last object parsed via nextObject()
   * @return the last object parsed by nextObject(), or null
   * if nextObject has not parsed an object.
   * @see com.sas.io.ObjectTokenizer#nextObject
   */
  public Object getLastObject()
    {
      return lastObject;
    }

  /**
   * Specify that the parser should recognize character literal values
   * and return instances of java.lang.Character
   * @param state if true, recognize 'x' - like character literals
   * tokens and return java.lang.Character objects
   */
  public void parseCharacters()
  {
    quoteChar('\'');  // so that character literals are read as tokens
    parseCharacters = true;
  }

  /**
   * Specify that the parser should recognize boolean values
   * and return instances of java.lang.Boolean
   * @param state if true, recognize true and false
   * tokens and return Boolean.TRUE or Boolean.FALSE, respectively.
   */
  public void parseBooleans(boolean state)
  {
    parseBooleans = state;
  }

  protected void parseObjects()
    {
      resetSyntax();
      parseNumbers();
      parseBooleans(true);
      // no special whitespace characters
      ordinaryChars(0, ' ');
      quoteChar('\"');  // so that strings are read as tokens
    }

  /**
   * Specify that the parser should recognize byte values
   * and return instances of java.lang.Byte
   * @param state if true, create Byte objects for integer tokens
   * between Byte.MIN_VALUE Byte.MAX_VALUE. If false, integer
   * values in this range are handled by larger sized integers
   * or floating point types.
   */
  public void parseBytes(boolean state)
    {
      parseBytes = state;
    }

  /**
   * Specify that the parser should recognize short values
   * and return instances of java.lang.Short
   * @param state if true, create Short objects for integer tokens
   * between Short.MIN_VALUE Short.MAX_VALUE that are not already
   * handled as bytes if parsing bytes. If false, integer
   * values in this range are handled by larger sized integers
   * or floating point types.
   */
  public void parseShorts(boolean state)
    {
      parseShorts = state;
    }

  /**
   * Specify that the parser should recognize int values
   * and return instances of java.lang.Integer
   * @param state if true, create Integer objects for integer tokens
   * between Integer.MIN_VALUE Integer.MAX_VALUE that are not already
   * handled as bytes/shorts if parsing those types. If false, integer
   * values in this range are handled by larger sized integers
   * or floating point types.
   */
  public void parseIntegers(boolean state)
    {
      parseIntegers = state;
    }

  /**
   * Specify that the parser should recognize long values
   * and return instances of java.lang.Long
   * @param state if true, create Byte objects for integer tokens
   * between Long.MIN_VALUE Long.MAX_VALUE that are not already
   * handled as bytes/shorts/ints if parsing those types. If false, integer
   * values in this range are handled by floating point types.
   */
  public void parseLongs(boolean state)
    {
      parseLongs = state;
    }

  /**
   * Specify that the parser should recognize float values
   * and return instances of java.lang.Float
   * @param state if true, create Byte objects for integer tokens
   * between Float.MIN_VALUE Float.MAX_VALUE that are not already
   * handled as integers if parsing integer types. If false,
   * values in this range are handled as Doubles.
   */
  public void parseFloats(boolean state)
    {
      parseFloats = state;
    }

  /**
   * Specify that the parser should recognize floating point values
   * and return instances of java.lang.Double
   * @param state if true, create Double objects for numeric tokens
   * that are not already handled by smaller types.
   */
  public void parseDoubles(boolean state)
    {
      parseDoubles = state;
    }

  public void parseNumbers()
    {
      parseBytes(false);
      parseShorts(false);
      parseIntegers(true);
      parseLongs(true);
      parseFloats(false);
      parseDoubles(true);
    }

  /**
   * Skip whitespace. Whitespace characters may be
   * significant in strings/nextObject, so the tokenizer has
   * no whitespace characters set. This call explicitly
   * skips whitespace.
   * @param delimiters Don't count a character as whitespace if it is in the
   * delimiters. For example, when your delimiter is a space or newline,
   * don't skip over them.
   * @exception IOException if there is an error on the input reader.
   */
  public void skipWhitespace(String delimiters) throws IOException
    {
      token = nextToken();
      while (token != TT_EOF && (delimiters.indexOf(token) == -1) && Character.isWhitespace((char) token))
        token = nextToken();
      pushBack();
    }

  /**
   * Skip whitespace, copying into the buffer. Whitespace characters may be
   * significant in strings/nextObject, so the tokenizer has
   * no whitespace characters set. This call explicitly
   * skips whitespace.
   * @param append if true, copy whitespace into buffer.
   * @exception IOException if there is an error on the input reader.
   */
  public void skipWhitespace(String delimiters, boolean append) throws IOException
    {
      while (token != TT_EOF && (delimiters.indexOf(token) == -1)  && Character.isWhitespace((char) token))
        {
          if (append)
            buffer.append((char) token);
          token = nextToken();
        }
    }

}
