// Copyright (c) 2002 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.OutputStream;
import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;

/**
 * An object which provides output streams which may be remote.   Classes
 * implement this interface if they provide one ore more output streams.
 *
 * @author  Bryan Wolfe
 * @version 1.0 28Aug02
 */

public interface OutputStreamProviderInterface extends Remote {
    /**
    * Check if getOutputStream will succeed.  See {@link #getOutputStream} to
    * determine whether this method should be used.
    *
    * @return non-null if getOutputStream() would succeed; throws UnmarshalException if it would fail
    * @throws UnmarshalException if invoked on a remote object
    * @throws RemoteException if a network I/O error occurs.
    */
	Object hasOutputStream() throws UnmarshalException, RemoteException;

    /**
    * Gets an OutputStream.  If this method is invoked on a remote object, a
    * UnmarshalException exception is thrown if the OutputStream was successfully
    * created since OutputStream is not serializable.  However, the OutputStream
    * will have been created.  If
    * this is an issue {@link #hasOutputStream} may be invoked to determine
    * if getOutputStream would succeed.  If an OutputStream is needed in a
    * remote context
    * getRemoteOutputStream should be invoked.
    *
    * @param streamKey is used to identify a specific output stream
    * when a provider supports multiple output streams.  It is implementation
    * specific and is typically null if only one stream is provided.
    *
    * @return an OutputStream from the provider, as identified by
    * the <var>streamKey</var>; null if the provider does not provide the stream.
    *
    * @throws UnmarshalException if invoked on a remote object
    * @throws IOException if an exception occurred while construction the OutputStream
    * @throws RemoteException if a network I/O error occurs.
    */
	OutputStream getOutputStream(Object streamKey) throws UnmarshalException, IOException, RemoteException;

    /**
    * Gets a remote output stream.  This method is invoked on a remote
    * object or if the returned OutputStream needs to be remoted.
    *
    * @param streamKey is used to identify a specific output stream
    * when a provider supports multiple output streams.  It is implementation
    * specific and is typically null if only one stream is provided.
    *
    * @return an OutputStream from the provider, as identified by the
    * <var>streamKey</var>
    *
    * @throws UnmarshalException if invoked on a remote object
    * @throws IOException if an exception occurred while construction the RemoteOutputStreamInterface
    * @throws RemoteException if a network I/O error occurs.
    */
	RemoteOutputStreamInterface getRemoteOutputStream(Object streamKey) throws IOException, RemoteException;
}
