// Copyright (c) 2001-2003 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.rmi.RemoteException;

/**
 * A proxy wrapper for access to a 
 * <code>java.io.OutputStream</code> on a remote
 * virtual machine.  The class extends and reimplements 
 * <code>java.io.OutputStream<code> on an encapsulated
 * <code>RemoteOutputStreamInterface</code> object.  
 *
 * @see     com.sas.io.RemoteOutputStream
 * @author  Bryan Wolfe
 * @version 1.0 09/28/01
 *          1.1 3-Mar-2003 Rob Owen 
 */
public class OutputStreamProxy extends java.io.OutputStream implements Serializable {

    static final long serialVersionUID = -2078894499165416890L;

	/**
	 * Private copy of the remote stream to which we are delegating.
	 */
    private RemoteOutputStreamInterface out;

	/**
	 * Creates an <code>OutputStreamProxy</code> from an existing output stream.
	 * @param in The existing output stream
	 * @throws RemoteException in the case of a network failure
	 */
    public OutputStreamProxy( OutputStream out ) throws RemoteException {
        this( new RemoteOutputStream(out) );
    }

    /**
     * Constructs an <code>OutputStreamProxy</code> 
     * encapsulating the specified remote output stream.
     * @param out Thhe stream to be wrapped by this object.
     */
    public OutputStreamProxy( RemoteOutputStreamInterface out ) {
        this.out = out;
    }

    /**
     * Writes the specified byte to the output stream.
     * @param b - The byte to be written (low order byte of int)
     * @throws IOException - If an I/O error occurs. 
     */
    public void write( int b ) throws IOException {
        out.write( b );
    }

    /**
     * Writes b.length bytes from the specified byte array to the output stream.
     * @param b The byte array to be written to the output stream 
     * @throws IOException - If an I/O error occurs. 
     */
    public void write( byte[] b ) throws IOException {
        out.write( b );
    }

	/**
	 * Writes <code>len</code> bytes from the specified byte array, starting at an offset
	 * <code>off</code> to the output stream.
	 * @param b The byte array containing the bytes to be written to the stream
	 * @param off The offset into the byte array at which the bytes to be written start
	 * @param len The number of bytes to write to the output stream
	 * @throws IOException - If an I/O error occurs. 
	 */
    public void write( byte[] b, int off, int len ) throws IOException {
        out.write( b, off, len );
    }

	/**
	 * Flushes this output stream and forces any buffered output bytes to be written out.
	 * @throws IOException - If an I/O error occurs.
	 */
    public void flush() throws IOException {
        out.flush();
    }

    /**
     * Closes the output stream and releases any system resources associated with the stream.
	 * @throws IOException - If an I/O error occurs.
     */
    public void close() throws IOException {
        out.close();
    }
}