//Copyright (c) 2001-2003 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.io.Reader;
import java.io.Serializable;
import java.rmi.RemoteException;

import com.sas.util.UncheckedException;

/**
 * Provides a serializabe extension to Reader, allowing it to be used in 
 * both local and remote environments.  
 *
 * @author <a href="mailto:rob.owen@sas.com">Rob Owen</a>
 * @version 1.1 3-Mar-2003
 * @since   1.1
 * 
 * @history
 *    3-Mar-2003 ro Created.
 *
 */
public class ReaderAdapter extends java.io.Reader implements Serializable
{
	static final long serialVersionUID = -6501975424007685167L;

	/**
	 * Private copy of the remote reader to which we are delegating.
	 */
	private RemoteReaderInterface reader;

	/**
	 * Creates a reader adapter from an existing reader.
	 * @param in The existing reader
	 * @throws RemoteException in the case of a network failure
	 */
	public ReaderAdapter( Reader reader ) throws RemoteException {
		this( new RemoteReader( reader ) );
	}
	
	/**
	 * Creates an reader adapter from a remote reader.
	 * @param in The existing remote reader
	 */
	public ReaderAdapter(RemoteReaderInterface reader)
	{
		this.reader = reader;
	}

	/**
	 * Tell whether this reader is ready to be read.
	 * @return boolean. True if the next read() is guaranteed not to block for input, false otherwise. 
	 * @throws IOException - If an I/O error occurs
	 */
	public boolean ready() throws IOException {
		return reader.ready();
	}

	/**
	 * Close the reader.
	 */  
	public void close() throws IOException {
		reader.close();
	}

	/**
	 * Mark the present position in the stream. Subsequent calls to reset() will attempt to 
	 * reposition the stream to this point. Not all character-input streams support the 
	 * mark() operation.
	 * @param readAheadLimit - Limit on the number of characters that may be read while still preserving the mark. 
	 * After reading this many characters, attempting to reset the stream may fail.
	 * @throws IOException - in the case of an I/O error
	 */
	public synchronized void mark(int readlimit) throws IOException
	{
		reader.mark(readlimit);
	}

	/**
	 * Does this stream support the mark() operation. Default is false.
	 * @return boolean - whether the stream supports the mark() operation
	 */
	public boolean markSupported()
	{
		try {
			return reader.markSupported();
		}
		catch (RemoteException e) {
			throw new UncheckedException(e, RB.getStringResource("RCOM.wrapCheck.ex.txt"));
		}
	}

	/**
	 * Read a single character.
	 * @return int - The character read, as an integer in the range 0 to 65535 (0x00-0xffff), 
	 * or -1 if the end of the stream has been reached
	 * @throws IOException - If an I/O error occurs
	 */
	public int read() throws IOException {
		return reader.read();
	}

	/**
	 * Read characters into an array.
	 * @param cbuf - Destination buffer
	 * @return int - The number of characters read, or -1 if the end of the stream has been reached
	 * @throws IOException - If an I/O error occurs
	 */
	public int read( char[] cbuf ) throws IOException {
		char tempbuf[] = reader.read( cbuf.length );

		/* check for end of file */
		if ( tempbuf == null ) return -1;

		/* otherwise copy tempbuf into b */
		System.arraycopy( tempbuf, 0, cbuf, 0, tempbuf.length );
		return tempbuf.length;
	}

	/**
	 * Read characters into a portion of an array. 
	 * @param cbuf - Destination buffer
	 * @param off - Offset at which to start storing characters
	 * @param len - Maximum number of characters to read
	 * @return int - The number of characters read, or -1 if the end of the stream has been reached
	 * @throws IOException - If an I/O error occurs
	 */
	public int read(char[] cbuf, int off, int len) throws IOException {
		char tempbuf[] = reader.read(len);

		/* check for end of file */
		if ( tempbuf == null ) return -1;

		/* otherwise copy tempbuf into b */
		System.arraycopy( tempbuf, 0, cbuf, off, tempbuf.length );

		return tempbuf.length;
	}

	/**
	 * Reset the stream.  
	 * @throws IOException - If an I/O error occurs 
	 */
	public synchronized void reset() throws IOException {
		reader.reset();
	}

	/**
	 * Skip characters. This method will block until some characters are available, 
	 * an I/O error occurs, or the end of the stream is reached.
	 * @param n The number of characters to skip
	 * @return long The actual number of characters skipped
	 * @throws IOException - If an I/O error occurs 
	 */
	public long skip(long n) throws IOException {
		return reader.skip(n);
	}
}

