// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.Reader;
import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;

/**
 * A interface which provides input streams which may be remote.  Remote
 * interfaces typically subclass this interface if they provide one
 * or more input streams.
 *
 * @see InputStreamProviderInterface
 * @see RemoteReaderInterface
 * @see java.io.Reader
 *
 * @author  Russell Gonsalves
 * @version 1.0 3Dec01
 */

public interface ReaderProviderInterface extends Remote {
    /**
    * Check if getReader will succeed.  See {@link #getReader} to
    * determine whether this method should be used.
    *
    * @return non-null if getReader() would succeed; throws UnmarshalException if it would fail
    * @throws UnmarshalException if invoked on a remote object
    * @throws RemoteException if a network I/O error occurs.
    */
	Object hasReader() throws UnmarshalException, RemoteException;

    /**
    * Gets an Reader.  If this method is invoked on a remote object, a
    * UnmarshalException exception is thrown if the Reader was successfully
    * created since Reader is not serializable.  However, the Reader
    * will have been created.  If
    * this is an issue {@link #hasReader} may be invoked to determine
    * if getReader would succeed.  If an Reader is needed in a
    * remote context getRemoteReader should be invoked.
    *
    * @param streamID is used to identify a specific input stream
    * when a provider supports multiple input streams.  It is implementation
    * specific and is typically null if only one stream is provided.
    *
    * @return a Reader from the provider, as identified by
    * the <var>streamID</var>; null if the provider does not provide the reader.
    *
    * @throws UnmarshalException if invoked on a remote object
    * @throws IOException if an exception occurred while construction the Reader
    * @throws RemoteException if a network I/O error occurs.
    */

	Reader getReader(Object streamID) throws UnmarshalException, IOException, RemoteException;

    /**
    * Gets a RemoteReaderInterface.  This method is invoked on a remote
    * object or if the returned Reader needs to be remoted.
    *
    * @param streamID is used to identify a specific input stream
    * when a provider supports multiple input streams.  It is implementation
    * specific and is typically null if only one stream is provided.
    *
    * @return The Reader requested
    * @throws UnmarshalException if invoked on a remote object
    * @throws IOException if an exception occurred while construction the RemoteReaderInterface
    * @throws RemoteException if a network I/O error occurs.
    */
	RemoteReaderInterface getRemoteReader(Object streamID) throws IOException, RemoteException;
}
