// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.InputStream;
import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

/**
 * Encapsulates a <code>java.io.InputStream</code> to
 * allow access from a remote virtual machine.
 * <code>RemoteInputStream</code> implements
 * <code>RemoteInputStreamInterface</code>.  Method
 * signatures and behavior
 * of this class are identical to the encapsulated
 * <code>InputStream</code> except for the addition of
 * remote exception handling where necessary and a
 * modification of the read method to make it compatible
 * with a remotable interface.
 *
 *
 *
 * @see     com.sas.io.InputStreamProviderInterface
 * @author  Bryan Wolfe
 * @version 1.0 09/28/01
 */

/*
 * Stolen from experimental work by Chris Bailey.
 */

public class RemoteInputStream extends UnicastRemoteObject
	implements RemoteInputStreamInterface
{
	transient private InputStream is;

	/**
	 * Constructs a <code>RemoteInputStream</code>
	 * encapsulating the specified input stream.
	 *
	 * @param is the stream to be wrapped by this object.
	 * @throws RemoteException for any rmi network error.
	 */
	public RemoteInputStream(InputStream is)
		throws RemoteException
	{
		this.is = is;
	}

	public int available()
		throws IOException
	{
		return is.available();
	}

	public void close()
		throws IOException
	{
		is.close();
	}

	public synchronized void mark(int readlimit)
	{
		is.mark(readlimit);
	}

	public boolean markSupported()
	{
		return is.markSupported();
	}

	public int read()
		throws IOException
	{
		return is.read();
	}

	/* The following read method replaces the read( byte[] ) and
	 * read(byte[], int, int) methods of InputStream because remote
	 * interfaces do not support update parameters.
	 */
	public byte[] read(int maxread)
		throws IOException
	{
		byte b[] = new byte[maxread];
		int nread = is.read(b);

		/* if at end of file, return null */
		if (nread<0) return null;

		/* if we read less than requested, reallocate the buffer */
		if (nread<maxread) {
			byte newb[] = new byte[nread];
			for (int i=0; i<nread; i++) newb[i] = b[i];
			b = newb;
		}
		return b;
	}

	public synchronized void reset()
		throws IOException
	{
		is.reset();
	}

	public long skip(long n)
		throws IOException
	{
		return is.skip(n);
	}

    /**
     * Gets a local or remote input stream.  It first tries to get a local
     * input stream from the <var>provider</var> and if unsuccessful
     * attempts to get a remote input stream.
     *
     * @param provider the input stream provider
     * @param streamID identifes the stream
     *
     * @throws IOException  If the provider's getInputStream / getRemoteInputStream throw such an exception
     * @throws RemoteException  If a network anomaly is encountered.
     */
	static public InputStream getInputStream(InputStreamProviderInterface provider, Object streamID) throws IOException, RemoteException {
		InputStream f;
		try {
			f = provider.getInputStream(streamID);
		} catch (java.rmi.UnmarshalException e) {
			RemoteInputStreamInterface ris = provider.getRemoteInputStream(streamID);
			f = new Proxy(ris);
		}
		return f;
	}

	static private class Proxy  extends java.io.InputStream
	{
		transient private RemoteInputStreamInterface is;


		public Proxy(RemoteInputStreamInterface is)
		{
			this.is = is;
		}

		public int available()
			throws IOException
		{
			return is.available();
		}

		public void close()
			throws IOException
		{
			is.close();
		}

		public synchronized void mark(int readlimit)
		{
			try {
				is.mark(readlimit);
			}
			catch (RemoteException e) {
				throw new java.lang.reflect.UndeclaredThrowableException(e);
			}
		}

		public boolean markSupported()
		{
			try {
				return is.markSupported();
			}
			catch (RemoteException e) {
				throw new java.lang.reflect.UndeclaredThrowableException(e);
			}
		}

		public int read()
			throws IOException
		{
			return is.read();
		}

		public int read(byte[] b)
			throws IOException
		{
			byte tempbuf[] = is.read(b.length);

			/* check for end of file */
			if ( tempbuf == null ) return -1;

			/* otherwise copy tempbuf into b */
			for ( int i=0; i<tempbuf.length; i++ )
				b[i] = tempbuf[i];
			return tempbuf.length;
		}

		public int read(byte[] b, int off, int len)
			throws IOException
		{
			byte tempbuf[] = is.read(len);

			/* check for end of file */
			if ( tempbuf == null ) return -1;

			/* otherwise copy tempbuf into b */
			for ( int i=0; i<tempbuf.length; i++ )
				b[off++] = tempbuf[i];

			return tempbuf.length;
		}

		public synchronized void reset()
			throws IOException
		{
			is.reset();
		}

		public long skip(long n)
			throws IOException
		{
			return is.skip(n);
		}

	}

}
