// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;

/**
 * Remotable interface corresponding to
 * <code>java.io.InputStream</code>.  This interface duplicates
 * the methods of <code>java.io.InputStream</code> with minor
 * changes for RMI compatability.
 *
 * Objects implementing this interface may be accessed remotely
 * directly using this interface or through the
 * <code>RemoteInputStreamProxy</code>.  <code>RemoteInputStreamProxy</code>
 * extends <code>java.io.InputStream</code> and thus may be used by
 * <code>java.io.Reader</code> objects and other consumers of
 * <code>java.io.InputStream</code>
 *
 * @see java.io.InputStream
 *
 * @author  Bryan Wolfe, Russell Gonsalves
 * @version 1.0 09/28/01
 */

public interface RemoteInputStreamInterface
    extends Remote
{

/**
 * Returns the number of bytes that can be read (or skipped over) from
 * this input stream without blocking by the next caller of a method for
 * this input stream.  The next caller might be the same thread or or
 * another thread.
 *
 * @return the number of bytes that can be read from this input stream
 *         without blocking.
 * @throws IOException if an I/O error occurs.
 * @see java.io.InputStream#available
 */
    int available() throws IOException;

/**
 * Closes this input stream and releases any system resources associated
 * with the stream.
 *
 * @throws IOException if an I/O error occurs.
 * @see java.io.InputStream#close
 */
    void close() throws IOException;

/**
 * Marks the current position in this input stream. A subsequent call to
 * the <code>reset</code> method repositions this stream at the last marked
 * position so that subsequent reads re-read the same bytes.
 *
 * <p> The <code>readlimit</code> argument tells this input stream to
 * allow that many bytes to be read before the mark position gets
 * invalidated.
 *
 * <p> The general contract of <code>mark</code> is that, if the method
 * <code>markSupported</code> returns <code>true</code>, the stream somehow
 * remembers all the bytes read after the call to <code>mark</code> and
 * stands ready to supply those same bytes again if and whenever the method
 * <code>reset</code> is called.  However, the stream is not required to
 * remember any data at all if more than <code>readlimit</code> bytes are
 * read from the stream before <code>reset</code> is called.
 *
 * @param readlimit the maximum limit of bytes that can be read before
 *                  the mark position becomes invalid.
 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.InputStream#close
 */
    void mark(int readlimit) throws RemoteException;

/**
 * Tests if this input stream supports the <code>mark</code> and
 * <code>reset</code> methods. Whether or not <code>mark</code> and
 * <code>reset</code> are supported is an invariant property of a
 * particular input stream instance.
 *
 * @return <code>true</code> if this stream instance supports the mark
 *         and reset methods; <code>false</code> otherwise.
 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.InputStream#markSupported
 */
    boolean markSupported() throws RemoteException;

/**
 * Reads the next byte of data from the input stream. The value byte is
 * returned as an <code>int</code> in the range <code>0</code> to
 * <code>255</code>. If no byte is available because the end of the stream
 * has been reached, the value <code>-1</code> is returned. This method
 * blocks until input data is available, the end of the stream is detected,
 * or an exception is thrown.
 *
 * @return the next byte of data, or <code>-1</code> if the end of the
 *         stream is reached.
 * @throws IOException if an I/O error occurs.
 * @see java.io.InputStream#read()
 */
 int read() throws IOException;

/**
 * Reads some number of bytes from the input stream and returns them as
 * a byte array.  The length of the array indicates the number of bytes
 * actually read.  This method blocks until input data is
 * available, end of file is detected, or an exception is thrown.
 *
 * <p> If <code>maxread</code> is negative, an
 * <code>IndexOutOfBoundsException</code> is thrown.  If
 * <code>maxread</code> is zero, then no bytes are read and an array
 * of zero length is returned; otherwise, there is an attempt to read at
 * least one byte. If no byte is available because the stream is at end
 * of file, <code>null</code> is returned.
 *
 * <p> If the first byte cannot be read for any reason other than end of
 * file, then an <code>IOException</code> is thrown. In particular, an
 * <code>IOException</code> is thrown if the input stream has been closed.
 *
 * <p> This method provides the same functionality as the
 * <code>read( byte[] )</code> and <code>read( byte[], int, int )</code>
 * methods of <code>java.io.InputStream</code>.  Thexe methods could not
 * be implemented directly in a <code>Remote</code> interface because
 * they use input/output parameters.
 *
 * @param maxread the buffer into which the data is read.
 * @return the byte array containing the read data, or
 *         <code>null</code> is there is no more data because the end of
 *         the stream has been reached.  Note: the size of the returned
 *         array will match the number of bytes read (whereas InputStream.read
 *         returns the int directly).
 * @throws IOException if an I/O error occurs.
 * @see java.io.InputStream#read(byte[])
 * @see java.io.InputStream#read(byte[], int, int)
 */

/* This method replaces the read( byte[] ) and read(byte[], int, int)
 * methods of InputStream because remote interfaces do not support
 * update parameters.
 */
 byte[] read(int maxread) throws IOException;

/**
 * Repositions this stream to the position at the time the
 * <code>mark</code> method was last called on this input stream.
 *
 * <p> The general contract of <code>reset</code> is:
 *
 * <p><ul>
 *
 * <li> If the method <code>markSupported</code> returns
 * <code>true</code>, then:
 *
 *     <ul><li> If the method <code>mark</code> has not been called since
 *     the stream was created, or the number of bytes read from the stream
 *     since <code>mark</code> was last called is larger than the argument
 *     to <code>mark</code> at that last call, then an
 *     <code>IOException</code> might be thrown.
 *
 *     <li> If such an <code>IOException</code> is not thrown, then the
 *     stream is reset to a state such that all the bytes read since the
 *     most recent call to <code>mark</code> (or since the start of the
 *     file, if <code>mark</code> has not been called) will be resupplied
 *     to subsequent callers of the <code>read</code> method, followed by
 *     any bytes that otherwise would have been the next input data as of
 *     the time of the call to <code>reset</code>. </ul>
 *
 * <li> If the method <code>markSupported</code> returns
 * <code>false</code>, then:
 *
 *     <ul><li> The call to <code>reset</code> may throw an
 *     <code>IOException</code>.
 *
 *     <li> If an <code>IOException</code> is not thrown, then the stream
 *     is reset to a fixed state that depends on the particular type of the
 *     input stream and how it was created. The bytes that will be supplied
 *     to subsequent callers of the <code>read</code> method depend on the
 *     particular type of the input stream. </ul></ul>
 *
 * @throws IOException if an I/O error occurs.
 * @see java.io.InputStream#reset
 */
    void reset() throws IOException;

/**
 * Skips over and discards <code>n</code> bytes of data from this input
 * stream. The <code>skip</code> method may, for a variety of reasons, end
 * up skipping over some smaller number of bytes, possibly <code>0</code>.
 * This may result from any of a number of conditions; reaching end of file
 * before <code>n</code> bytes have been skipped is only one possibility.
 * The actual number of bytes skipped is returned.  If <code>n</code> is
 * negative, no bytes are skipped.
 *
 * @param n the number of bytes to be skipped.
 * @return the actual number of bytes skipped.
 * @throws IOException if an I/O error occurs.
 * @see java.io.InputStream#skip(long)
 */
    long skip(long n) throws IOException;

}
