// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;

/**
 * Remoteable interface corresponding to 
 * <code>java.io.OutputStream</code>.  This interface duplicates 
 * the methods of <code>java.io.OutputStream</code> with minor
 * changes for RMI compatability.
 *
 * Objects implementing this interface may be accessed remotely
 * directly using this interface or through the 
 * <code>RemoteOutputStreamProxy</code>.  <code>RemoteOutputStreamProxy</code>
 * extends <code>java.io.OutputStream</code> and thus may be used by
 * <code>Writer</code> objects and other consumers of 
 * <code>OutputStream</code>
 *
 * @see java.io.OutputStream
 *
 * @author  Bryan Wolfe
 * @version 1.0 09/28/01
 */

public interface RemoteOutputStreamInterface
    extends Remote

{
/**
 * Writes the specified byte to this output stream. The general 
 * contract for <code>write</code> is that one byte is written 
 * to the output stream. The byte to be written is the eight 
 * low-order bits of the argument <code>b</code>. The 24 
 * high-order bits of <code>b</code> are ignored.
 *
 * @param b the <code>byte</code>.
 * @throws IOException if an I/O error occurs. In particular, 
 *           an <code>IOException</code> may be thrown if the 
 *           output stream has been closed.
 * @see java.io.OutputStream#write(int)
 */
    public void write(int b)
        throws IOException;

/**
 * Writes <code>b.length</code> bytes from the specified byte array 
 * to this output stream. The general contract for <code>write(b)</code> 
 * is that it should have exactly the same effect as the call 
 * <code>write(b, 0, b.length)</code>.
 *
 * @param b the data.
 * @throws IOException if an I/O error occurs. In particular, 
 *           an <code>IOException</code> may be thrown if the 
 *           output stream has been closed.
 * @see java.io.OutputStream#write(byte[])
 */
 public void write(byte[] b) throws IOException;

/**
 * Writes <code>len</code> bytes from the specified byte array 
 * starting at offset <code>off</code> to this output stream. 
 * The general contract for <code>write(b, off, len)</code> is that 
 * some of the bytes in the array <code>b</code> are written to the 
 * output stream in order; element <code>b[off]</code> is the first 
 * byte written and <code>b[off+len-1]</code> is the last byte written 
 * by this operation.
 * <p>
 * The <code>write</code> method of <code>OutputStream</code> calls 
 * the write method of one argument on each of the bytes to be 
 * written out. Subclasses are encouraged to override this method and 
 * provide a more efficient implementation. 
 * <p>
 * If <code>b</code> is <code>null</code>, a 
 * <code>NullPointerException</code> is thrown.
 * <p>
 * If <code>off</code> is negative, or <code>len</code> is negative, or 
 * <code>off+len</code> is greater than the length of the array 
 * <code>b</code>, then an <tt>IndexOutOfBoundsException</tt> is thrown.
 *
 * @param b the data.
 * @param off the start offset in the data.
 * @param len the number of bytes to write.
 * @throws IOException if an I/O error occurs. In particular, 
 *           an <code>IOException</code> may be thrown if the 
 *           output stream has been closed.
 * @see java.io.OutputStream#write(byte[], int, int)
 */
    public void write(byte[] b, int off, int len)
        throws IOException;

/**
 * Flushes this output stream and forces any buffered output bytes 
 * to be written out. The general contract of <code>flush</code> is 
 * that calling it is an indication that, if any bytes previously 
 * written have been buffered by the implementation of the output 
 * stream, such bytes should immediately be written to their 
 * intended destination.
 *
 * @throws IOException if an I/O error occurs. In particular, 
 *           an <code>IOException</code> may be thrown if the 
 *           output stream has been closed.
 * @see java.io.OutputStream#flush
 */
    public void flush()
        throws IOException;

/**
 * Closes this output stream and releases any system resources 
 * associated with this stream. The general contract of <code>close</code> 
 * is that it closes the output stream. A closed stream cannot perform 
 * output operations and cannot be reopened.
 *
 * @throws IOException if an I/O error occurs. In particular, 
 *           an <code>IOException</code> may be thrown if the 
 *           output stream has been closed.
 * @see java.io.OutputStream#close
 */
    public void close()
        throws IOException;
}
