// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import com.sas.text.Message;
import com.sas.util.UncheckedException;
import java.io.Reader;
import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;


/**
 * Encapsulates a <code>java.io.Reader</code> to
 * allow access from a remote virtual machine.
 * <code>RemoteReader</code> implements
 * <code>RemoteReaderInterface</code>.  Method
 * signatures and behavior
 * of this class are identical to the encapsulated
 * <code>Reader</code> except for the addition of
 * remote exception handling where necessary and a
 * modification of the read method to make it compatible
 * with a remotable interface.
 *
 *
 * @see     com.sas.io.ReaderProviderInterface
 * @author  Russell Gonsalves (based on Bryan Wolfe's RemoteInputStream)
 * @version 1.0 28Nov2001
 */


public class RemoteReader extends UnicastRemoteObject implements RemoteReaderInterface
{
	private Reader is;

	/**
	 * Constructs a <code>RemoteReader</code>
	 * encapsulating the specified input stream.
	 *
	 * @param is the stream to be wrapped by this object.
	 * @throws RemoteException for any rmi network error.
	 */
	public RemoteReader(Reader is) throws RemoteException
	{
		this.is = is;
	}

	public boolean ready() throws IOException
	{
		return is.ready();
	}

	public void close() throws IOException
	{
		is.close();
	}

	public synchronized void mark(int readlimit) throws IOException
	{
		is.mark(readlimit);
	}

	public boolean markSupported()
	{
		return is.markSupported();
	}

	public int read() throws IOException
	{
		return is.read();
	}

	/* The following read method replaces the read( byte[] ) and
	 * read(byte[], int, int) methods of Reader because remote
	 * interfaces do not support update parameters.
	 */
	public char[] read(int maxread) throws IOException
	{
		if (maxread < 0) {
			String msg = Message.format(RB.getResources(), "RR.negArg.ex.fmt.txt", Integer.toString(maxread));
			throw new IllegalArgumentException(msg);
		}
		char b[] = new char[maxread];
		int nread = is.read(b);

		/* if at end of file, return null */
		if (nread < 0) return null;

		/* if we read less than requested, reallocate the buffer */
		if (nread < maxread) {
			char newb[] = new char[nread];
			for (int i=0; i<nread; i++) newb[i] = b[i];
			b = newb;
		}
		return b;
	}

	public synchronized void reset() throws IOException
	{
		is.reset();
	}

	public long skip(long n) throws IOException
	{
		return is.skip(n);
	}

    /**
     * Gets a local or remote reader.  It first tries to get a local reader
     * from the <var>provider</var> and if unsuccessful attempts to get a
     * remote reader.
     *
     * @param provider the reader provider
     * @param streamID identifes the stream
     */
	static public Reader getReader(ReaderProviderInterface provider, Object streamID) throws IOException, RemoteException {
		Reader r;
		try {
			r = provider.getReader(streamID);
		} catch (java.rmi.UnmarshalException e) {
			RemoteReaderInterface rr = provider.getRemoteReader(streamID);
			r = new Proxy(rr);
		}
		return r;
	}

	static private class Proxy extends java.io.Reader
	{
		private RemoteReaderInterface reader;

		public Proxy(RemoteReaderInterface reader)
		{
			this.reader = reader;
		}

		public boolean ready() throws IOException {
			return reader.ready();
		}

		public void close() throws IOException {
			reader.close();
		}

		public synchronized void mark(int readlimit) throws IOException
		{
			reader.mark(readlimit);
		}

		public boolean markSupported()
		{
			try {
				return reader.markSupported();
			}
			catch (RemoteException e) {
				throw new UncheckedException(e, RB.getStringResource("RCOM.wrapCheck.ex.txt"));
			}
		}

		public int read() throws IOException {
			return reader.read();
		}

		public int read(char[] b) throws IOException {
			char tempbuf[] = reader.read(b.length);

			/* check for end of file */
			if ( tempbuf == null ) return -1;

			/* otherwise copy tempbuf into b */
			for ( int i=0; i<tempbuf.length; i++ )
				b[i] = tempbuf[i];
			return tempbuf.length;
		}

		public int read(char[] b, int off, int len) throws IOException {
			char tempbuf[] = reader.read(len);

			/* check for end of file */
			if ( tempbuf == null ) return -1;

			/* otherwise copy tempbuf into b */
			for ( int i=0; i<tempbuf.length; i++ )
				b[off++] = tempbuf[i];

			return tempbuf.length;
		}

		public synchronized void reset() throws IOException {
			reader.reset();
		}

		public long skip(long n) throws IOException {
			return reader.skip(n);
		}
	}

}
