// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;

/**
 * Remotable interface corresponding to
 * <code>java.io.Reader</code>.  This interface duplicates
 * the methods of <code>java.io.Reader</code> with minor
 * changes for RMI compatability.
 *
 * Objects implementing this interface may be accessed remotely
 * directly using this interface or through the
 * <code>RemoteReaderProxy</code>.  <code>RemoteReaderProxy</code>
 * extends <code>java.io.Reader</code> and thus may be used by
 * consumers of
 * <code>java.io.Reader</code>
 *
 * @see java.io.Reader
 *
 * @author  Russell Gonsalves
 * @version 1.0 28Nov2001
 */

public interface RemoteReaderInterface
    extends Remote
{

/**
 * Tell whether this stream is ready to be read.
 *
 * @return True if the next read() is guaranteed not to block for input,
 * false otherwise. Note that returning false does not guarantee
 * that the next read will block.
 *
 * @throws RemoteException if a network I/O error occurs.
 * @throws IOException If an I/O error occurs
 */
    boolean ready() throws IOException, RemoteException;

/**
 * Close the stream. Once a stream has been closed, further read(), ready(),
 * mark(), or reset() invocations will throw an IOException. Closing a
 * previously-closed stream, however, has no effect.
 *
 * @throws IOException if an I/O error occurs.
 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.Reader#close
 */
    void close() throws IOException, RemoteException;

/**
 * Mark the present position in the stream. Subsequent calls to reset()
 * will attempt to reposition the stream to this point. Not all
 * character-input streams support the mark() operation.
 *
 * @param readlimit Limit on the number of characters that may be read
 * while still preserving the mark. After reading this many characters,
 * attempting to reset the stream may fail.
 *
 * @throws IOException if the stream does not support mark(),
 * or if some other I/O error occurs.
 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.Reader#close
 */
    void mark(int readlimit) throws IOException, RemoteException;

/**
 * Tell whether this stream supports the mark() operation.
 * The default implementation always returns false.
 * Subclasses should override this method.
 *
 * @return true if and only if this stream supports the mark operation.

 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.Reader#markSupported
 */
    boolean markSupported() throws RemoteException;

/**
 * Read a single character. This method will block until a character is
 * available, an I/O error occurs, or the end of the stream is reached.
 *
 * <p>Subclasses that intend to support efficient single-character input
 * should override this method.
 *
 * @return The character read, as an integer in the range
 * 0 to 65535 (0x00-0xffff),
 * or -1 if the end of the stream has been reached
 *
 * @throws IOException if an I/O error occurs.
 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.Reader#read()
 */
 int read() throws IOException, RemoteException;

/**
 * Reads some number of bytes from the input stream and returns them as
 * a byte array.  The length of the array indicates the number of bytes
 * actually read.  This method blocks until input data is
 * available, end of file is detected, or an exception is thrown.
 *
 * <p> If <code>maxread</code> is negative, an
 * <code>IllegalArgumentException</code> is thrown.  If
 * <code>maxread</code> is zero, then no bytes are read and an array
 * of zero length is returned; otherwise, there is an attempt to read at
 * least one byte. If no byte is available because the stream is at end
 * of file, <code>null</code> is returned.
 *
 * <p> If the first byte cannot be read for any reason other than end of
 * file, then an <code>IOException</code> is thrown. In particular, an
 * <code>IOException</code> is thrown if the input stream has been closed.
 *
 * <p> This method provides the same functionality as the
 * <code>read( byte[] )</code> and <code>read( byte[], int, int )</code>
 * methods of <code>java.io.Reader</code>.  These methods could not
 * be implemented directly in a <code>Remote</code> interface because
 * they use input/output parameters.
 *
 * @param maxread the buffer into which the data is read.
 *
 * @return the byte array containing the read data, or
 *         <code>null</code> is there is no more data because the end of
 *         the stream has been reached. The size of the returned array
 *         will match the number of bytes read (whereas Reader.read
 *         returns the int directly).
 *
 * @throws IllegalArgumentException If n is negative.
 * @throws IOException if an I/O error occurs.
 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.Reader#read(char[])
 * @see java.io.Reader#read(char[], int, int)
 */

/* This method replaces the read( byte[] ) and read(byte[], int, int)
 * methods of Reader because remote interfaces do not support
 * update parameters.
 */
 char[] read(int maxread) throws IOException, RemoteException;

/**
 * Reset the stream. If the stream has been marked, then attempt to reposition
 * it at the mark. If the stream has not been marked, then attempt to reset
 * it in some way appropriate to the particular stream, for example by
 * repositioning it to its starting point. Not all character-input streams
 * support the reset() operation, and some support reset() without supporting
 * mark().
 *
 * @throws IOException If the stream has not been marked,
 * @throws RemoteException if a network I/O error occurs.
 * or if the mark has been invalidated,
 * or if the stream does not support reset(),
 * or if some other I/O error occurs.
 * @see java.io.Reader#reset
 */
    void reset() throws IOException, RemoteException;

/**
 * Skip characters. This method will block until some characters are
 * available, an I/O error occurs, or the end of the stream is reached.
 *
 * @param n The number of characters to skip.
 * @return The number of characters actually skipped
 * @throws IllegalArgumentException If n is negative.
 * @throws IOException If an I/O error occurs
 * @throws RemoteException if a network I/O error occurs.
 * @see java.io.Reader#skip(long)
 */
    long skip(long n) throws IOException, RemoteException;

}
