// Copyright (c) 2001-2003 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.io.Writer;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

/**
 * Encapsulates a <code>java.io.Writer</code> to
 * allow access from a remote virtual machine.
 * <code>RemoteWriter</code> implements
 * <code>RemoteWriterInterface</code>.  Method
 * signatures and behavior
 * of this class are identical to the encapsulated
 * <code>Writer</code> except for the addition of
 * remote exception handling where necessary
 *
 * @author <a href="mailto:rob.owen@sas.com">Rob Owen</a>
 * @version 1.1 3-Mar-2003
 * @since   1.1
 * 
 * @history
 *    3-Mar-2003 ro Created.
 */
public class RemoteWriter extends UnicastRemoteObject implements RemoteWriterInterface {
    private Writer out;

    /**
     * Constructs a <code>RemoteWriter</code>
     * encapsulating the specified writer.
     * @param out the writer to be wrapped by this object.
     * @throws RemoteException for any rmi network error.
     */
    public RemoteWriter( Writer out ) throws RemoteException {
        this.out = out;
    }

	/**
	 * Write a single character. The character to be written is contained in the 16 low-order bits
	 * of the given integer value; the 16 high-order bits are ignored.
	 * @param c The character to be written
	 * @throws IOException - If an I/O error occurs
	 */ 
    public void write( int b ) throws IOException {
        out.write( b );
    }

	/**
	 * Write an array of characters. 
	 * @param cbuf The array of characters to be written
	 * @throws IOException - If an I/O error occurs
	 */
    public void write( char[] cbuf ) throws IOException {
        out.write( cbuf );
    }

	/**
	 * Write a portion of an array of characters. 
	 * @param cbuf The array containing the characters to be written
	 * @param off The offset into cbuf from where to start writing
	 * @param len The number of characters to write 
	 * @throws IOException - If an I/O error occurs
	 */
    public void write( char[] cbuf, int off, int len ) throws IOException {
        out.write( cbuf, off, len );
    }

	/**
	 * Write a string.
	 * @param str The string containing the characters to be written
	 * @throws IOException - If an I/O error occurs
	 */
    public void write( String str ) throws IOException {
        out.write( str );
    }

	/**
	 * Write a portion of a string. 
	 * @param str The string containing the characters to be written
	 * @param off The offset into str from where to start writing
	 * @param len The number of characters to write 
	 * @throws IOException - If an I/O error occurs
	 */
    public void write( String str, int off, int len ) throws IOException {
        out.write( str, off, len );
    }

	/**
	 * Flush the stream. 
	 * @throws IOException - If an I/O error occurs
	 */
    public void flush() throws IOException {
        out.flush();
    }

	/**
	 * Close the stream, flushing it first.
	 * @throws IOException - If an I/O error occurs
	 */
    public void close() throws IOException {
        out.close();
    }

}
