// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.rmi.Remote;

/**
 * Remoteable interface corresponding to 
 * <code>java.io.Writer</code>.  This interface duplicates 
 * the methods of <code>java.io.Writer</code> with minor
 * changes for RMI compatability.
 *
 * Objects implementing this interface may be accessed remotely
 * directly using this interface or through the 
 * <code>WriterProxy</code>.  <code>WriterProxy</code>
 * extends <code>java.io.Writer</code> and thus may be used by
 * consumers of <code>Writer</code>
 *
 * @see java.io.Writer
 *
 * @author <a href="mailto:rob.owen@sas.com">Rob Owen</a>
 * @version 1.1 3-Mar-2003
 * @since   1.1
 * 
 * @history
 *    3-Mar-2003 ro Created.
 */
public interface RemoteWriterInterface extends Remote {
    /**
     * Writes the specified char to this writer. The general 
     * contract for <code>write</code> is that one char is written 
     * to the writer. 
     * 
     * @param c the <code>character</code>.
     * @throws IOException if an I/O error occurs. In particular, 
     *           an <code>IOException</code> may be thrown if the 
     *           writer has been closed.
     * @see java.io.Writer#write(int)
     */
    public void write(int c) throws IOException;

	/**
	 * Writes <code>cbuf.length</code> chars from the specified char array 
	 * to this writer. The general contract for <code>write(cbuf)</code> 
	 * is that it should have exactly the same effect as the call 
	 * <code>write(cbuf, 0, cbuf.length)</code>.
	 *
	 * @param cbuf the character data.
	 * @throws IOException if an I/O error occurs. In particular, 
	 *           an <code>IOException</code> may be thrown if the 
	 *           writer has been closed.
	 * @see java.io.Writer#write(char[])
	 */
	public void write( char[] cbuf ) throws IOException;

    /**
     * Writes <code>len</code> chars from the specified char array 
     * starting at offset <code>off</code> to this writer. 
     * The general contract for <code>write(cbuf, off, len)</code> is that 
     * some of the chars in the array <code>cbuf</code> are written to the 
     * writer in order; element <code>cbuf[off]</code> is the first 
     * char written and <code>cbuf[off+len-1]</code> is the last char written 
     * by this operation.
     * <p>
     * The <code>write</code> method of <code>Writer</code> calls 
     * the write method of one argument on each of the chars to be 
     * written out. Subclasses are encouraged to override this method and 
     * provide a more efficient implementation. 
     * <p>
     * If <code>cbuf</code> is <code>null</code>, a 
     * <code>NullPointerException</code> is thrown.
     * <p>
     * If <code>off</code> is negative, or <code>len</code> is negative, or 
     * <code>off+len</code> is greater than the length of the array 
     * <code>cbuf</code>, then an <tt>IndexOutOfBoundsException</tt> is thrown.
     *
     * @param c the data.
     * @param off the start offset in the data.
     * @param len the number of chars to write.
     * @throws IOException if an I/O error occurs. In particular, 
     *           an <code>IOException</code> may be thrown if the 
     *           writer has been closed.
     * @see java.io.Writer#write(char[], int, int)
     */
    public void write( char[] cbuf, int off, int len) throws IOException;

	/**
	 * Writes <code>str.length()</code> chars from the specified string 
	 * to this writer. The general contract for <code>write(str)</code> 
	 * is that it should have exactly the same effect as the call 
	 * <code>write(str, 0, str.length())</code>.
	 *
	 * @param str the data.
	 * @throws IOException if an I/O error occurs. In particular, 
	 *           an <code>IOException</code> may be thrown if the 
	 *           writer has been closed.
	 * @see java.io.Writer#write(String)
	 */
	public void write( String str ) throws IOException;

	/**
	 * Writes <code>len</code> chars from the specified string 
	 * starting at offset <code>off</code> to this writer. 
	 * The general contract for <code>write(str, off, len)</code> is that 
	 * some of the chars in the string <code>cbuf</code> are written to the 
	 * writer in order
	 * <p>
	 * The <code>write</code> method of <code>Writer</code> calls 
	 * the write method of one argument on each of the chars to be 
	 * written out. Subclasses are encouraged to override this method and 
	 * provide a more efficient implementation. 
	 * <p>
	 * If <code>cbuf</code> is <code>null</code>, a 
	 * <code>NullPointerException</code> is thrown.
	 * <p>
	 * If <code>off</code> is negative, or <code>len</code> is negative, or 
	 * <code>off+len</code> is greater than the length of the string 
	 * <code>cbuf</code>, then an <tt>IndexOutOfBoundsException</tt> is thrown.
	 *
	 * @param str the data.
	 * @param off the start offset in the data.
	 * @param len the number of chars to write.
	 * @throws IOException if an I/O error occurs. In particular, 
	 *           an <code>IOException</code> may be thrown if the 
	 *           writer has been closed.
	 * @see java.io.Writer#write(String, int, int)
	 */
	public void write( String str, int off, int len ) throws IOException;

    /**
     * Flushes this output stream and forces any buffered output bytes 
     * to be written out. The general contract of <code>flush</code> is 
     * that calling it is an indication that, if any bytes previously 
     * written have been buffered by the implementation of the output 
     * stream, such bytes should immediately be written to their 
     * intended destination.
     *
     * @throws IOException if an I/O error occurs. In particular, 
     *           an <code>IOException</code> may be thrown if the 
     *           output stream has been closed.
     * @see java.io.OutputStream#flush
     */
    public void flush() throws IOException;

    /**
     * Closes this output stream and releases any system resources 
     * associated with this stream. The general contract of <code>close</code> 
     * is that it closes the output stream. A closed stream cannot perform 
     * output operations and cannot be reopened.
     *
     * @throws IOException if an I/O error occurs. In particular, 
     *           an <code>IOException</code> may be thrown if the 
     *           output stream has been closed.
     * @see java.io.OutputStream#close
     */
    public void close() throws IOException;
}
