// Copyright (c) 2001 by SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.FilterInputStream;
import java.io.PushbackInputStream;
import java.io.IOException;
//
// History:
//
// Date    Name            Remarks
// 16Mar01 B Beers         S0110897: Improve performance for large parameters.
//
/**
 * An input stream filter that scans the stream for occurrences of a given
 * sequence of bytes and, if found, replaces the sequence with another given
 * sequence of bytes. The stream is not rescanned after a replacement is made.
 * @author Biff Beers
 */
public class ReplacementInputStream extends FilterInputStream
{
     /**
      * Construct an input stream that scans the stream for
      * occurrences of the targetSequence and replaces them with
      * replacementSequence.
      * @param in the underlying input stream. This must be a PushbackInputStream because
      * the filter must sometimes scan ahead while searching for the target sequence. These
      * lookahead bytes are pushed back to the underlying stream if they do not contain the
      * target sequence. The size of the pushback buffer should be at least as big as
      * the size of the target sequence.
      * @param targetSequence the byte sequence to search for and replace in the
      * input stream. This array must contain at least one byte.
      * @param replacementSequence the byte sequence to insert in the stream whenever
      * an occurrence of the target sequence is found and removed. This array may be null or empty.
      */
     public ReplacementInputStream(PushbackInputStream in,byte[] targetSequence,byte[] replacementSequence)
     {
          super(in);

          if (targetSequence == null || targetSequence.length == 0)
          {
               throw new IllegalArgumentException();
          }

          this.pStream = in;
          this.targetSequence = targetSequence;
          this.replacementSequence = replacementSequence;
          buffer = new byte[targetSequence.length];
     }

     /**
      * Read enough bytes from the underlying input stream to determine if
      * the target sequence has occurred, and then return one byte from the
      * stream after making any necessary replacements.
      * @return one byte from the underlying stream after making any necessary
      * replacements.
      * @exception java.io.IOException if there is an exception in the underlying
      * stream.
      */
     public synchronized int read() throws IOException
     {
          int b = -1;

          do
          {
               if (replaced)
               {
                    if (index < replacementSequence.length)
                    {
                         b = replacementSequence[index++] & 0xFF;
                    }
                    else
                    {
                         replaced = false;
                    }
               }
               else
               {
                    int i = 0;

                    do
                    {
                         int j = pStream.read(buffer,i,buffer.length-i);

                         if (j < 0)
                         {
                              break;
                         }

                         i += j;
                    }
                    while (i < buffer.length);

                    if (i == 0)
                    {
                         return -1;
                    }

                    if (i == buffer.length && arraysEqual(buffer,targetSequence))
                    {
                         replaced = true;
                         index = 0;
                    }
                    else
                    {
                         b = buffer[0] & 0xFF;
                         pStream.unread(buffer,1,i-1);
                    }
               }
          }
          while (b == -1);

          return b;
     }

     /**
      * Calls this.read() continuously until l bytes have been read and
      * stored in array b starting at offset o. If EOF is encountered
      * before l bytes have been read, but after at least on byte has
      * been read, then the number of bytes read before EOF was encountered
      * will be returned. If no bytes have been read before when EOF is
      * encountered, -1 will be returned. If l is 0, then 0 will be
      * returned.
      * @param b a byte array to store read bytes in.
      * @param o the offset in array b to begin storing read bytes.
      * @param l the maximum number of bytes to store in array b.
      * @return the number of bytes stored in array b
      * @exception java.io.IOException if there is an exception reading
      * from the underlying stream.
      */
     public synchronized int read(byte[] b,int o,int l) throws IOException
     {
          for (int i=0;i<l;i++)
          {
               int e = read();

               if (e < 0)
               {
                    return (i>0)?i:-1;
               }
               else
               {
                    b[o+i] = (byte)e;
               }
          }

          return l;
     }

     /**
      * Calls this.read() continuously and discards the output until l bytes 
      * have been read and discarded. If EOF is encountered
      * before n bytes have been read, then the number of bytes read before EOF was encountered
      * will be returned.
      * @param l the maximum number of bytes to read and discard.
      * @return the number of bytes read and discarded
      * @exception java.io.IOException if there is an exception reading
      * from the underlying stream.
      */
     public synchronized long skip(long n) throws IOException
     {
          for (long i=0;i<n;i++)
          {
               int e = read();

               if (e < 0)
               {
                    return i;
               }
          }

          return n;
     }

     /**
      * Estimates the number of bytes that can be returned to callers
      * of this input stream without blocking. The estimate will always
      * be less than or equal to the actual number of bytes that can be
      * returned, and it could be 0 even when bytes are available to
      * be returned.
      * @return an estimate of the number of bytes that can be returned to callers
      * of this input stream without blocking
      * @exception java.io.IOException if there is an exception reading
      * from the underlying stream.
      */
     public synchronized int available() throws IOException
     {
          final int i = pStream.available();
          final int targetLength = targetSequence.length;
          final int replacementLength = (replacementSequence==null)?0:replacementSequence.length;

          if (replacementLength < targetLength)
          {
               return ((i / targetLength) * replacementLength) + (i % targetLength);
          }

          return i;
     }

     private PushbackInputStream pStream = null;
     private int index = 0;
     private boolean replaced = false;
     private byte[] buffer = null;
     private byte[] targetSequence = null;
     private byte[] replacementSequence = null;

     private static boolean arraysEqual(byte[] b1,byte[] b2)
     {
          for (int i=0;i<b1.length;i++)
          {
               if (b1[i] != b2[i])
               {
                    return false;
               }
          }

          return true;
     }
}