// Copyright (c) 2001 by SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
//
// History:
//
// Date    Name            Remarks
// 16Mar01 B Beers         S0110897: Improve performance for large parameters.
//
/**
 * An output stream filter that scans the stream for occurrences of a given
 * sequence of bytes and, if found, replaces the sequence with another given
 * sequence of bytes. The stream is not rescanned after a replacement is made.
 * @author Biff Beers
 */
public class ReplacementOutputStream extends FilterOutputStream
{
     /**
      * Construct an output stream that scans the stream for
      * occurrences of the targetSequence and replaces them with
      * replacementSequence.
      * @param out the underlying output stream
      * @param targetSequence the byte sequence to search for and replace in the
      * output stream. This array must contain at least one byte.
      * @param replacementSequence the byte sequence to insert in the stream whenever
      * an occurrence of the target sequence is found and removed. This array may be null or empty.
      */
     public ReplacementOutputStream(OutputStream out,byte[] targetSequence,byte[] replacementSequence)
     {
          super(out);

          if (targetSequence == null || targetSequence.length == 0)
          {
               throw new IllegalArgumentException();
          }

          this.replacementSequence = replacementSequence;

          if (targetSequence.length > 1)
          {
               this.targetSequence = targetSequence;
               bufferLength = targetSequence.length;
               buffer = new byte[bufferLength];
          }
          else
          {
               target = targetSequence[0];
          }
     }

     /**
      * Writes a byte at the end of an internal buffer and, if the buffer contains
      * enough bytes, compares the buffer with the target sequence. If the
      * buffer matches the target sequence, the buffer is cleared and the
      * replacement sequence is written to the underlying output stream.
      * Otherwise, one byte is removed from beginning of the buffer and  written
      * to the underlying output stream.
      * @param b a byte  to add to the internal buffer and eventually write to
      * the underlying output stream
      * @exception java.io.IOException if there is an exception when writing to
      * the underlying output stream.
      */
     public synchronized void write(int b) throws IOException
     {
          if (targetSequence == null)
          {
               if (target == (byte)b)
               {
                    if (replacementSequence != null)
                    {
                         out.write(replacementSequence);
                    }
               }
               else
               {
                    out.write(b);
               }
          }
          else
          {
               buffer[index] = (byte)b;
               dataLength++;
               index = (index + 1) % bufferLength;

               if (dataLength == bufferLength)
               {
                    if (arraysEqual(buffer,index,targetSequence))
                    {
                         if (replacementSequence != null)
                         {
                              out.write(replacementSequence);
                         }

                         index = 0;
                         dataLength = 0;
                    }
                    else
                    {
                         out.write(buffer[index]);
                         dataLength--;
                    }
               }
          }
     }

     /**
      * Writes l bytes from array b starting at offset o to the underlying output
      * stream, filtering as necessary. If the target sequence that this filter is
      * scanning for is one byte long, and none of the bytes from array a that are to
      * be written equals that byte, then all bytes that are to be written from array
      * a are written to the underlying output stream at once by calling out.write(byte[],int,int).
      * Otherwise, the bytes are written one at a time by calling this.write(int) for each
      * byte.
      * @param b the byte array containing bytes to be filtered and written to the underlying
      * output stream.
      * @param o the index of the first byte in the array to be filtered and written
      * to the underlying output stream.
      * @param o the number of bytes in the array to be filtered and written
      * to the underlying output stream.
      * @exception java.io.IOException if there is an exception when writing to
      * the underlying output stream.
      */
     public synchronized void write(byte[] b,int o,int l) throws IOException
     {
          if (targetSequence == null)
          {
               int i = 0;

               for (;i<l;i++)
               {
                    if (target == b[i+o])
                    {
                         break;
                    }
               }

               if (i == l)
               {
                    out.write(b,o,l);
                    return;
               }
          }

          for (int i=0;i<l;i++)
          {
               write(b[i+o]);
          }
     }

     /**
      * Writes a byte to the underlying output stream without first adding it to
      * the internal buffer and scanning for the target sequence. The internal buffer
      * is flushed to the underlying stream before this byte is written to the
      * underlying stream.
      * @param b the byte to write to the underlying output stream.
      * @exception java.io.IOException if there is an exception when writing to
      * the underlying output stream.
      */
     public synchronized void writeNoReplacement(int b) throws IOException
     {
          flush();
          out.write(b);
     }

     /**
      * Calls this.writeNoReplacement(b,0,b.length).
      * @param b the byte array to write to the underlying output stream.
      * @exception java.io.IOException if there is an exception when writing to
      * the underlying output stream.
      */
     public void writeNoReplacement(byte[] b) throws IOException
     {
          writeNoReplacement(b,0,b.length);
     }

     /**
      * Writes l bytes of array b starting at offset o to the underlying output 
      * stream without first adding them to
      * the internal buffer and scanning for the target sequence. The internal buffer
      * is flushed to the underlying stream before these bytes are written to the
      * underlying stream.
      * @param b the byte array to write to the underlying output stream.
      * @exception java.io.IOException if there is an exception when writing to
      * the underlying output stream.
      */
     public synchronized void writeNoReplacement(byte[] b,int o,int l) throws IOException
     {
          flush();
          out.write(b,o,l);
     }

     /**
      * Writes the contents of the internal buffer to the underlying output stream and then
      * flushed that stream.
      * @exception java.io.IOException if there is an exception when writing to
      * the underlying output stream.
      */
     public synchronized void flush() throws IOException
     {
          if (dataLength > 0)
          {
               if (index > dataLength)
               {
                    out.write(buffer,0,dataLength);
               }
               else
               {
                    for (int i=0;i<dataLength;i++)
                    {
                         out.write(buffer[(index+i+1)%bufferLength]);
                    }
               }

               index = 0;
               dataLength = 0;
          }

          out.flush();
     }

     private byte[] targetSequence = null;
     private byte[] replacementSequence = null;
     private byte[] buffer = null;
     private int bufferLength = 0;
     private int dataLength = 0;
     private int index = 0;
     private byte target = 0;

     protected static boolean arraysEqual(byte[] a1,int offset,byte[] a2)
     {
          final int l = a2.length;

          for (int i=0;i<l;i++)
          {
               if (a1[(i+offset)%l] != a2[i])
               {
                    return false;
               }
          }

          return true;
     }
}