// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.io;

import java.io.IOException;
import java.io.Writer;
import java.io.Serializable;
import java.rmi.RemoteException;

/**
 * A proxy wrapper for access to a 
 * <code>java.io.Writer</code> on a remote
 * virtual machine.  The class extends and reimplements 
 * <code>java.io.Writer<code> on an encapsulated
 * <code>RemoteWriterInterface</code> object.  
 *
 * @see     com.sas.io.RemoteWriter
 * @author  Bryan Wolfe
 * @version 1.0 09/28/01
 *          1.1 3-Mar-2003 Rob Owen 
 */
public class WriterProxy extends java.io.Writer implements Serializable {

    static final long serialVersionUID = 3046303106738714603L;

	/**
	 * Private copy of the remote writer to which we are delegating.
	 */
    private RemoteWriterInterface rout;

	/**
	 * Creates an <code>WriterProxy</code> from an existing writer.
	 * @param in The existing writer
	 * @throws RemoteException in the case of a network failure
	 */
    public WriterProxy( Writer out ) throws RemoteException {
        this( new RemoteWriter(out) );
    }

    /**
     * Constructs a <code>WriterProxy</code> 
     * encapsulating the specified remote output stream.
     * @param is the stream to be wrapped by this object.
     */
    public WriterProxy( RemoteWriterInterface rout ) {
        this.rout = rout;
    }

	/**
	 * Write a single character. The character to be written is contained in the 16 low-order bits
	 * of the given integer value; the 16 high-order bits are ignored.
	 * @param c The character to be written
	 * @throws IOException - If an I/O error occurs
	 */ 
    public void write( int c ) throws IOException {
        rout.write( c );
    }

    /**
     * Write an array of characters. 
     * @param cbuf The array of characters to be written
	 * @throws IOException - If an I/O error occurs
     */
    public void write( char[] cbuf ) throws IOException {
        rout.write( cbuf );
    }

    /**
     * Write a portion of an array of characters. 
     * @param cbuf The array containing the characters to be written
     * @param off The offset into cbuf from where to start writing
     * @param len The number of characters to write 
	 * @throws IOException - If an I/O error occurs
     */
    public void write( char[] cbuf, int off, int len ) throws IOException {
        rout.write( cbuf, off, len );
    }

    /**
     * Write a string.
     * @param str The string containing the characters to be written
	 * @throws IOException - If an I/O error occurs
     */
    public void write( String str ) throws IOException {
        rout.write( str );
    }

	/**
	 * Write a portion of a string. 
	 * @param str The string containing the characters to be written
	 * @param off The offset into str from where to start writing
	 * @param len The number of characters to write 
	 * @throws IOException - If an I/O error occurs
	 */
    public void write( String str, int off, int len ) throws IOException {
        rout.write( str, off, len );
    }

	/**
	 * Flush the stream. 
	 * @throws IOException - If an I/O error occurs
	 */
	public void flush() throws IOException {
        rout.flush();
    }

    /**
     * Close the stream, flushing it first.
	 * @throws IOException - If an I/O error occurs
     */
    public void close() throws IOException {
        rout.close();
    }

}