/* $Id: FileCleanup.java,v 1.4 2003/12/05 15:23:43 majohn Exp $ */
/*
 *
 * SAS Institute, Inc.
 *
 * Copyright (c) 2003 SAS Institute, Inc.  All rights reserved.
 *
 */

package com.sas.io.util;

import java.io.File;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * This class provides <b>singleton</b> access to a FileCleanupInterface
 * implementation. It also provides a mechanism to replace the existing (and
 * running!) implementation with a different implementation.
 * <p>
 * This class also provides a default implementation of FileCleanupInterface.
 * That implementation uses the three retention constants defined in
 * FileCleanupInterface, with default retention periods of 15 minutes, 2 hours,
 * and 12 hours. The three retention periods can be modified by setting the
 * following properties:
 * <ol>
 * <li><b>sas.filecleanup.short.retention.minutes</b></li>
 * <li><b>sas.filecleanup.medium.retention.minutes</b></li>
 * <li><b>sas.filecleanup.long.retention.minutes</b></li>
 * </ol>
 * The default implementation uses a daemon thread to perform its cleanup work.
 * By default, the thread is activated every five minutes. This can be tuned by
 * setting the property <b>sas.filecleanup.thread.sleep.minutes</b>.
 * <p>
 * The default implementation's <b>newTemporaryFile</b> methods create files in
 * the directory specified by the <b>sas.filecleanup.temp.directory</b>
 * property and by default creates them in a directory called
 * <b>sas.core.tmpfiles</b> in the directory specified by the Java property
 * <b>java.io.tmpdir</b>. On startup, the directory it uses for
 * <b>newTemporaryFile</b> calls is purged of its contents.
 * <p>
 * <b>Applications are encouraged to explicitly invoke
 * <u>FileCleanup.getInstance().deleteAllFilesNow()</u> as they exit.</b>
 *
 * @author Marc Johnson
 */

public class FileCleanup
{
    private static FileCleanupInterface _instance = null;

    /**
     * Get the current implementation of the FileCleanupInterface service.
     * Users should not cache the implementation, but should request the
     * current implementation each time they need to interact with the
     * FileCleanupInterface service.
     *
     * @return the current FileCleanupInterface service.
     */
    public static synchronized FileCleanupInterface getInstance()
    {
        if (_instance == null)
        {
            _instance = new FileCleanupImpl();
        }
        return _instance;
    }

    /**
     * Register an implementation of FileCleanupInterface to replace the
     * current implementation. This is typically done when the default
     * FileCleanupInterface implementation provided by this class is replaced
     * by an implementation tailored to the caller's application.
     * <p>
     * The <b>deleteAllFilesNow</b> method of the pre-existing implementation
     * of FileCleanupInterface is called as a side effect of invoking this
     * method.
     *
     * @param instance the new FileCleanupInterface implementation
     *
     * @throws NullPointerException if instance is null.
     */
    public static synchronized void registerImplementation(FileCleanupInterface instance)
    {
        if (instance == null)
        {
            throw new NullPointerException();
        }
        FileCleanupInterface oldInstance = _instance;
        _instance = instance;
        if (oldInstance != null)
        {
            oldInstance.deleteAllFilesNow();
        }
    }

    /* ******* START private implementation of FileCleanupInterface ******* */

    private static class FileCleanupImpl implements FileCleanupInterface
    {
        /**
         * Timing constants: retention periods and thread wakeup interval
         */
        private long _cleanupRetentionShort;
        private long _cleanupRetentionMedium;
        private long _cleanupRetentionLong;
        private long _cleanupThreadWakeupInterval;

        /**
         * Constants for temporary files
         */
        private File _tmpFileDirectory;

        // Sorted map of lists of files to be cleaned up, keyed by termination
        // dates. This object is also the lock object, synchronized on whenever
        // our view of what's to be cleaned up is being manipulated.
        private SortedMap _cleanupMap;
        // A reverse, of sorts, of the _cleanupMap. The keys are File objects,
        // and the values are termination dates for the associated files. This
        // map is manipulated at the same time as _cleanupMap, and manipulation
        // is synchronized on _cleanupMap.
        private Map _filesToCleanup;
        // our daemon worker thread
        private StoppableThread _cleanupThread;

        /**
         * Private constructor. Purges the temporary file directory,
         * initializes the maps, and creates and starts the daemon thread that
         * does the cleanup work.
         */
        private FileCleanupImpl()
        {
            long msPerSec = 1000;
            long secPerMin = 60;
            long msPerMin = msPerSec * secPerMin;
            long minPerHour = 60;

            _cleanupRetentionShort =
                msPerMin
                    * Math.max(
                        1,
                        getPropertyAsLong(
                            "sas.filecleanup.short.retention.minutes",
                            15));
            _cleanupRetentionMedium =
                msPerMin
                    * Math.max(
                        1,
                        getPropertyAsLong(
                            "sas.filecleanup.medium.retention.minutes",
                            2 * minPerHour));
            _cleanupRetentionLong =
                msPerMin
                    * Math.max(
                        1,
                        getPropertyAsLong(
                            "sas.filecleanup.long.retention.minutes",
                            12 * minPerHour));
            _cleanupThreadWakeupInterval =
                msPerMin
                    * Math.max(
                        1,
                        getPropertyAsLong(
                            "sas.filecleanup.thread.sleep.minutes",
                            5));
            String property =
                getProperty("sas.filecleanup.temp.directory", null);
            _tmpFileDirectory =
                property == null
                    ? new File(
                        new File(System.getProperty("java.io.tmpdir")),
                        "sas.core.tmpfiles")
                    : new File(property);
            try
            {
                if (_tmpFileDirectory.exists())
                {
                    prune(_tmpFileDirectory);
                }
                // note: _tmpFileDirectory will be created when needed
                _cleanupMap = new TreeMap();
                _filesToCleanup = new HashMap();
                _cleanupThread = createCleanupThread();
            }
            catch (Exception ignored)
            {
            }
        }

        /**
         * Add a file to be deleted by this service. While temporary files
         * created by this service are automatically scheduled for deletion,
         * this method allows the service user to take files that it has
         * created and schedule them for deletion.
         * <p>
         * <b>file</b> may be a directory. Directories will be deleted
         * regardless of content.
         * <p>
         * Implementations of this method are never to throw any exceptions.
         * <p>
         * This implementation is equivalent to
         * <b>registerFileForDeletion(file,
         * FileCleanupInterface.RETENTION_MEDIUM)</b>.
         *
         * @param file the file to be deleted in the future.
         */
        public void registerFileForDeletion(File file)
        {
            registerFileForDeletion(
                file,
                FileCleanupInterface.RETENTION_MEDIUM);
        }

        /**
         * Add a file to be deleted by this service. While temporary files
         * created by this service are automatically scheduled for deletion,
         * this method allows the service user to take files that it has
         * created and schedule them for deletion. A retention period is
         * suggested, which the implementation is free to interpret as it sees
         * fit.
         * <p>
         * <b>file</b> may be a directory. Directories will be deleted
         * regardless of content.
         * <p>
         * Implementations of this method are never to throw any exceptions.
         *
         * @param file the file to be deleted in the future.
         * @param retentionPeriod a value that provides a hint to the
         *                        implementation as to how long the file will
         *                        be retained.
         */
        public void registerFileForDeletion(File file, int retentionPeriod)
        {
            try
            {
                if (file == null)
                {
                    return;
                }
                long interval = _cleanupRetentionMedium;
                if (retentionPeriod <= FileCleanupInterface.RETENTION_SHORT)
                {
                    interval = _cleanupRetentionShort;
                }
                else if (
                    retentionPeriod >= FileCleanupInterface.RETENTION_LONG)
                {
                    interval = _cleanupRetentionLong;
                }
                setUpFileForDeletion( file, interval);
            }
            catch (RuntimeException ignored)
            {
            }
        }

        /**
         * Create a File that will automatically be deleted later.
         * <p>
         * Implementations of this method are never to throw any exceptions.
         * <p>
         * This implementation is equivalent to
         * <b>newTemporaryFile(FileCleanupInterface.RETENTION_MEDIUM)</b>
         *
         * @return the newly created File.
         */
        public File newTemporaryFile()
        {
            return newTemporaryFile(FileCleanupInterface.RETENTION_MEDIUM);
        }

        /**
         * Create a File that will automatically be deleted later. A retention
         * period is suggested, which the implementation is free to interpret
         * as it sees fit.
         * <p>
         * Implementations of this method are never to throw any exceptions.
         *
         * @param retentionPeriod a value that provides a hint to the
         *                        implementation as to how long the file will
         *                        be retained.
         *
         * @return the newly created File.
         */
        public File newTemporaryFile(int retentionPeriod)
        {
            return newTemporaryFile(retentionPeriod, "tmp", "file");
        }

        /* (non-Javadoc)
         * @see com.sas.io.util.FileCleanupInterface#newTemporaryFile(int, java.lang.String, java.lang.String)
         */
        public File newTemporaryFile(
            int retentionPeriod,
            String prefix,
            String suffix)
        {
            File temp = null;
            try
            {
                // make sure deleteAllFilesNow isn't stomping on us, and vice
                // versa
                synchronized (_tmpFileDirectory)
                {
                    if (!_tmpFileDirectory.exists())
                    {
                        _tmpFileDirectory.mkdirs();
                    }
                    if (prefix == null || prefix.length() < 3)
                    {
                        prefix = "tmp";
                    }
                    temp =
                        File.createTempFile(prefix, suffix, _tmpFileDirectory);
                }
            }
            catch (IOException ignored)
            {
            }
            registerFileForDeletion(temp, retentionPeriod);
            return temp;
        }

        /* (non-Javadoc)
         * @see com.sas.io.util.FileCleanupInterface#deleteSoon(java.io.File)
         */
        public void deleteSoon(File file)
        {
            try
            {
                if (file == null)
                {
                    return;
                }
                long interval = 1;
                setUpFileForDeletion(file, interval);
            }
            catch (RuntimeException ignored)
            {
            }
        }

        private void setUpFileForDeletion(File file, long interval)
        {
            synchronized (_cleanupMap)
            {
                // deregister the file if it's already registered
                Date oldTermination = getTerminationDate(file);
                if (oldTermination != null)
                {
                    List oldList = (List) _cleanupMap.get(oldTermination);
                    // should never be null, but being careful
                    if (oldList != null)
                    {
                        oldList.remove(file);
                    }
                    _filesToCleanup.remove(file);
                }
                Date termination =
                    new Date(
                        calculateTermination(
                            new Date().getTime(),
                            interval));
                List list = (List) _cleanupMap.get(termination);
                if (list == null)
                {
                    list = new ArrayList();
                    _cleanupMap.put(termination, list);
                }
                list.add(file);
                _filesToCleanup.put(file, termination);
            }
        }

        /**
         * Delete a file immediately. If the file was already scheduled for
         * deletion, take care of any bookkeeping needed to keep the schedule
         * straight, but delete it NOW.
         * <p>
         * Implementations of this method are never to throw any exceptions.
         *
         * @param file the File to be deleted.
         */
        public void deleteNow(File file)
        {
            if (file != null)
            {
                try
                {
                    if (!prune(file))
                    {
                        if (file.exists())
                        {
                            // someone must be using it ... try scheduling it
                            // real soon, and we'll try again later.
                            registerFileForDeletion(
                                file,
                                FileCleanupInterface.RETENTION_SHORT);
                        }
                    }
                }
                catch (RuntimeException ignored)
                {
                }
            }
        }

        /**
         * Immediately delete all of the files that the implementation intends
         * to delete, even if they were not yet scheduled for deletion.
         * <p>
         * Implementations of this method are never to throw any exceptions.
         */
        public void deleteAllFilesNow()
        {
            try
            {
                // this will quickly get rid of most of the files we're
                // managing
                synchronized (_tmpFileDirectory)
                {
                    prune(_tmpFileDirectory);
                }
                // clean up anything else we held
                synchronized (_cleanupMap)
                {
                    for (Iterator iter = _filesToCleanup.keySet().iterator();
                        iter.hasNext();
                        )
                    {
                        File file = (File) iter.next();
                        if (file.exists())
                        {
                            // ignore return ... nothing we can do if it fails
                            prune(file);
                        }
                    }
                }
            }
            catch (RuntimeException ignored)
            {
            }
        }

        private static long getPropertyAsLong(
            String propertyName,
            long defaultValue)
        {
            long rval = defaultValue;
            try
            {
                rval = Long.parseLong(System.getProperty(propertyName).trim());
            }
            catch (Exception ignored)
            {
            }
            return rval;
        }

        private static String getProperty(
            String propertyName,
            String defaultValue)
        {
            String rval = null;
            try
            {
                rval = System.getProperty(propertyName);
            }
            catch (Exception ignored)
            {
            }
            if ((rval == null) || (rval.trim().length() == 0))
            {
                rval = defaultValue;
            }
            return rval;
        }

        private StoppableThread createCleanupThread()
        {
            StoppableThread rval =
                new StoppableThread("sas.filecleanup.default.implementation");
            rval.setDaemon(true);
            rval.start();
            return rval;
        }

        /**
         * Find out when a file is scheduled for termination.
         *
         * @param file a File to check.
         *
         * @return the Date representing its termination; null if the file is
         *         not scheduled for termination.
         */
        private Date getTerminationDate(File file)
        {
            Date result = null;
            // all access to _filesToCleanup is synchronized on _cleanupMap
            synchronized (_cleanupMap)
            {
                result = (Date) _filesToCleanup.get(file);
            }
            return result;
        }

        /**
         * Calculate the termination time for a file.
         *
         * @param currentTime the current time in milliseconds.
         * @param delay       the amount of time a file should exist before
         *                    cleanup, relative to currentTime, in
         *                    milliseconds. Must be at least 1.
         *
         * @return the earliest time at which the associated file should be
         *         deleted, in milliseconds, conforming to the algorithmic
         *         constraints described in this method's javadoc.
         */
        private long calculateTermination(long currentTime, long delay)
        {
            return currentTime + delay;
        }

        private boolean prune(File file)
        {
            if (file.isDirectory())
            {
                String[] contents = file.list();
                if (contents != null)
                {
                    for (int k = 0; k < contents.length; k++)
                    {
                        prune(new File(file, contents[k]));
                    }
                }
            }
            return file.delete();
        }

        /**
         * This does the work inside our thread ...
         */
        private void executeCleanup()
        {
            if (_cleanupMap.size() > 0)
            {
                List eligible_files = new ArrayList();
                Date now = new Date();
                synchronized (_cleanupMap)
                {
                    for (Iterator iter = _cleanupMap.keySet().iterator();
                        iter.hasNext();
                        )
                    {
                        Date then = (Date) iter.next();
                        if (now.after(then))
                        {
                            List files = (List) _cleanupMap.get(then);
                            eligible_files.addAll(files);
                            iter.remove();
                        }
                        else
                        {
                            // keys are sorted, so no more entries
                            break;
                        }
                    }
                    _filesToCleanup.keySet().removeAll(eligible_files);
                }
                if (!eligible_files.isEmpty())
                {
                    for (Iterator iter = eligible_files.iterator();
                        iter.hasNext();
                        )
                    {
                        File f = (File) iter.next();
                        boolean status = prune(f);
                        if (!status)
                        {
                            // we failed to delete the file - before we get
                            // bent out of shape, maybe it got deleted out from
                            // under us?
                            status = !f.exists();
                        }
                        if (!status)
                        {
                            // try again ...
                            registerFileForDeletion(
                                f,
                                FileCleanupInterface.RETENTION_SHORT);
                        }
                    }
                }
            }
        }

        /**
         * StoppableThread borrowed from Joshua Bloch's "Effective Java"
         */
        private class StoppableThread extends Thread
        {
            private boolean _stopRequested = false;

            StoppableThread(String name)
            {
                super(name);
            }

            public void run()
            {
                while (!stopRequested())
                {
                    try
                    {
                        Thread.sleep(_cleanupThreadWakeupInterval);
                        if (!stopRequested())
                        {
                            executeCleanup();
                        }
                    }
                    catch (Exception ignored)
                    {
                    }
                }
            }

            public synchronized void requestStop()
            {
                _stopRequested = true;
                interrupt();
            }

            private synchronized boolean stopRequested()
            {
                return _stopRequested;
            }
        }

    } // end private FileCleanupImpl

    /* *******  END  private implementation of FileCleanupInterface ******* */

} // end public class FileCleanup
