/* $Id: FileCleanupInterface.java,v 1.3 2003/12/05 15:23:47 majohn Exp $ */
/*
 *
 * SAS Institute, Inc.
 *
 * Copyright (c) 2003 SAS Institute, Inc.  All rights reserved.
 *
 */

package com.sas.io.util;

import java.io.File;

/**
 * This interface defines behavior for a file cleanup service. The service
 * provides for the creation of temporary files and the automatic deletion of
 * those files at some later time without the user of the temporary files
 * having to keep track of those files. The intent is that files would be
 * retained for a few hours (and this would probably be configurable) before
 * deletion. The retention period may vary if the implementation chooses to
 * implement variable retention periods. Constants are defined by this
 * interface for three different retention periods, but implementations are not
 * required to honor those values.
 * <p>
 * A typical implementation strategy is to spin a daemon thread to perform the
 * work of deleting the files that the implementation is managing. This thread
 * cannot interfere with the lifecycle of the VM - it must not artificially
 * keep the VM running.
 *
 * @author Marc Johnson
 */

public interface FileCleanupInterface
{
    /**
     * Short retention time suggested, on the order of 15 minutes
     */
    public static final int RETENTION_SHORT = 0;

    /**
     * Medium retention time suggested, on the order of 2 hours
     */
    public static final int RETENTION_MEDIUM = 1;

    /**
     * Long retention time suggested, on the order of 12 hours
     */
    public static final int RETENTION_LONG = 2;

    /**
     * Add a file to be deleted by this service. While temporary files created
     * by this service are automatically scheduled for deletion, this method
     * allows the service user to take files that it has created and schedule
     * them for deletion.
     * <p>
     * <b>file</b> may be a directory. Directories will be deleted regardless
     * of content.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     *
     * @param file the file to be deleted in the future.
     */
    public void registerFileForDeletion(File file);

    /**
     * Add a file to be deleted by this service. While temporary files created
     * by this service are automatically scheduled for deletion, this method
     * allows the service user to take files that it has created and schedule
     * them for deletion. A retention period is suggested, which the
     * implementation is free to interpret as it sees fit.
     * <p>
     * <b>file</b> may be a directory. Directories will be deleted regardless
     * of content.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     *
     * @param file the file to be deleted in the future.
     * @param retentionPeriod a value that provides a hint to the
     *                        implementation as to how long the file will be
     *                        retained. This is expected, but not required, to
     *                        be one of the constants defined in this
     *                        interface.
    
     */
    public void registerFileForDeletion(File file, int retentionPeriod);

    /**
     * Create a File that will automatically be deleted later.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     *
     * @return the newly created File.
     */
    public File newTemporaryFile();

    /**
     * Create a File that will automatically be deleted later. A retention
     * period is suggested, which the implementation is free to interpret as it
     * sees fit.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     *
     * @param retentionPeriod a value that provides a hint to the
     *                        implementation as to how long the file will be
     *                        retained. This is expected, but not required, to
     *                        be one of the constants defined in this
     *                        interface.
     *
     * @return the newly created File.
     */
    public File newTemporaryFile(int retentionPeriod);

    /**
     * Create a File that will automatically be deleted later. A retention
     * period is suggested, which the implementation is free to interpret as it
     * sees fit. The file will be created with the specified prefix and suffix.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     *
     * @param retentionPeriod a value that provides a hint to the
     *                        implementation as to how long the file will be
     *                        retained. This is expected, but not required, to
     *                        be one of the constants defined in this
     *                        interface.
     * @param prefix the prefix for the file. If null or too short, a suitable
     *               prefix (e.g., "tmp") will be silently substituted
     * @param suffix the suffix for the file.
     *
     * @return the newly created File.
     * 
     * @see java.io.File#createTempFile(java.lang.String, java.lang.String)
     */
    public File newTemporaryFile(
        int retentionPeriod,
        String prefix,
        String suffix);

    /**
     * Delete a file immediately. If the file was already scheduled for
     * deletion, take care of any bookkeeping needed to keep the schedule
     * straight, but delete it NOW.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     *
     * @param file the File to be deleted. Can be a directory.
     */
    public void deleteNow(File file);

    /**
     * Delete a file soon - somewhere between now and RETENTION_SHORT,
     * inclusive.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     *
     * @param file File to be deleted. Can be a directory.
     */
    public void deleteSoon(File file);

    /**
     * Immediately delete all of the files that the implementation intends to
     * delete, even if they were not yet scheduled for deletion.
     * <p>
     * Implementations of this method are never to throw any exceptions.
     */
    public void deleteAllFilesNow();
} // end public interface FileCleanupInterface
