/* $Id$ */
// Copyright (c) 2002 SAS Institute Inc., Cary, NC 27513
package com.sas.iom;

import java.io.IOException;
import java.util.Map;
import java.util.HashMap;
import com.sas.iom.SASCache.ICacheReader;
import com.sas.iom.SASCache.ICacheService;
import com.sas.iom.SASIOMDefs.GenericError;
import com.sas.iom.SASIOMDefs.OctetSeqHolder;
import com.sas.iom.SASIOMDefs.StringSeqHolder;
import com.sas.io.InputStreamHeaderInterface;
import java.text.MessageFormat;


/**
 * Provides the capability to read input from an IOM InputStream.
 *
 * @author <a href="mailto:bryan.wolfe@sas.com">Bryan Wolfe</a>
 * @author <a href="mailto:tony.dean@sas.com">Tony Dean</a>
 * @version 1.0
 * @since 1.0
 * 
 * @history
 *    28Aug2002 td  Created.
 *    23Oct2002 td  Only return response headers prefixed by HTTP:
 *    26Oct2002 td  Added closeStream flag to getData() and getDataAsString() so
 *                  stream can be left open to read headers if desired.
 *    23Nov2004 td  S0276823: read() now returns less than requested length if it is
 *                  determined that requesting more will cause io to block.
 *
 */
public class InputStream extends java.io.InputStream {

   private ICacheReader _iCacheReader = null;
   private ICacheService _iCacheService = null;
   
   private byte[] _buffer = null;
   private int _offsetOfNextAvailableInBuffer = 0;
   private int _lengthAvailableInBuffer = 0;
   private boolean _eof = false;
   private boolean _headerAvailable = false;
   
   private Map _responseHeaders = null;
   
   private InputStreamHeader _header = null;
  
   
   public InputStream(ICacheReader iCacheReader, ICacheService iCacheService) {
      _iCacheReader = iCacheReader;
      _iCacheService = iCacheService;
   }


   /**
    * Get the header interface for this InputStream.
    * 
    * @return header InputStream header.
    * @see com.sas.io.InputStreamHeaderInterface
    */
   public InputStreamHeaderInterface getHeader() {

      if( _header == null )
      {
         try {
            _header = new InputStreamHeader(this);
         }
         catch( java.rmi.RemoteException re ) {
            ;//return null header
         }
      }
      
      return _header;
   }
   
   /**
    * Get the data of this InputStream.<p>
    * This convenience method returns the entire contents
    * of the InputStream as a byte array.<p>
    * The InputStream will be automatically closed when
    * all the data has been read.
    * 
    * @return data associated with this InputStream.
    */
   public byte[] getData() throws IOException {
      return getData(true);
   }
   
   /**
    * Get the data of this InputStream.<p>
    * This convenience method returns the entire contents
    * of the InputStream as a byte array.
    * 
    * @param closeStream set to true to cause the stream to be closed; otherwise false.
    * @return data associated with this InputStream.
    */
   public byte[] getData(boolean closeStream) throws IOException {

      byte[] data = new byte[0];
      
      int BUFFER_LENGTH = 4096;

      java.io.BufferedInputStream bis = new java.io.BufferedInputStream(this, BUFFER_LENGTH);
         
      if( bis.available() > 0 )
      {         
         byte[] buffer = new byte[BUFFER_LENGTH];

         while( true )
         {
            int numBytes = bis.read(buffer, 0, BUFFER_LENGTH);
            if( numBytes < 0 )
               break;
               
            byte[] newData = new byte[data.length+numBytes];
            
            // copy old data
            System.arraycopy(data, 0, newData, 0, data.length);
               
            // copy new data
            for( int i=0; i<numBytes; i++ )
               newData[data.length+i] = buffer[i];
               
            data = newData;
         }
         
         // we're done so close it based on flag
         if( closeStream )
            bis.close();

      }
               
      return data;
   }

   /**
    * Get the data of this InputStream.<p>
    * This convenience method returns the entire contents
    * of the InputStream as a String.<p>
    * The default charset is used.<p>
    * The InputStream will be automatically closed when
    * all the data has been read.
    * 
    * @return data associated with this InputStream.
    */
   public String getDataAsString() throws IOException {

      return getDataAsString(null, true);
   }
   
   /**
    * Get the data of this InputStream.<p>
    * This convenience method returns the entire contents
    * of the InputStream as a String.
    * 
    * @param encoding Name of a supported charset.
    * @param closeStream set to true to cause the stream to be closed; otherwise false. 
    * @return data associated with this InputStream.
    */
   public String getDataAsString(String encoding, boolean closeStream) throws IOException {

      StringBuffer data = new StringBuffer();
      
      int BUFFER_LENGTH = 4096;

      java.io.BufferedReader br = null;
      if( encoding == null )
         br = new java.io.BufferedReader(new java.io.InputStreamReader(this));
      else
      {
         try {
            br = new java.io.BufferedReader(new java.io.InputStreamReader(this, encoding), BUFFER_LENGTH);
         }
         catch( java.io.UnsupportedEncodingException ue ) {
            throw new IOException(RB.getStringResource("InputStream.unsupportedEncodingException.ex.txt"/*I18nOK:EMS*/));
         }
      }
      
      char[] buffer = new char[BUFFER_LENGTH];
         
      while( true )
      {
         int numChars = br.read(buffer, 0, BUFFER_LENGTH);
         if( numChars < 0 )
            break;

         data.append(buffer, 0, numChars);
      }

      // we're done so close it based on flag
      if( closeStream )
         br.close();

      return data.toString();
   } 

   public long skip(long n) throws IOException
   {
      return read(null, 0, (int)n);
   }

   public int read() throws IOException
   {
      final byte[] b = new byte[1];
      if(read(b, 0, 1) > 0)
      {
        return b[0] & 0x000000ff;
      }
      return -1;
   }

   public int read(byte[] b) throws IOException
   {
      return read(b, 0, b.length);
   }

   public int read(byte[] b, int off, int len) throws IOException
   {
      // here's the guts
      // only copy into b if b != null
      // remember to return only the number of bytes copied into b
      // this may be less than len.
      // return -1 if there are no more bytes
      for (int i = 0; i < len; i++)
      {
         if ( _offsetOfNextAvailableInBuffer >= _lengthAvailableInBuffer )
         {

            // handle eof case
            if ( _eof )
            {
               // return the number of bytes read even if eof...
               // the next call will return -1 in this case
               if( i==0 )
                  return -1;
               else
                  return i;
            }
            else
            if (i > 0 && (available() <= 0)) {
               // return buffered data if getting
               // more will cause us to block...
               return i;
            }

            getMore();

            // handle case where no bytes were read
            if ( _lengthAvailableInBuffer == 0 )
            {
               if( _eof && i==0 )
                  return -1;
               else
                  return i;
            }

         }

         if ( b != null )
         {
            b[i+off] = _buffer[_offsetOfNextAvailableInBuffer];
         }
         
         _offsetOfNextAvailableInBuffer++;
         
      }
     
      return len;
   }

   public int available() throws IOException
   {
      int bytesOnServer = 0;
      
      try {
         bytesOnServer = _iCacheReader.Available();
      }
      catch( GenericError e ) {
         throw new IOException(e.getMessage());
      }

      return _lengthAvailableInBuffer - _offsetOfNextAvailableInBuffer + bytesOnServer;
   }

   public void close() throws IOException
   {
      try {
         if( _iCacheReader != null )
            _iCacheReader.Close();
      }
      catch( GenericError e ) {
         throw new IOException(MessageFormat.format(RB.getStringResource("InputStream.closeException.fmt.ex.txt"/*I18nOK:EMS*/),
                                   new Object[] { e.getMessage() }));
      }
   }


   // Helpers
   
   private void getMore() throws IOException
   {       
      _lengthAvailableInBuffer = 0;
      _offsetOfNextAvailableInBuffer = 0;

      try {
         OctetSeqHolder binaryData = new OctetSeqHolder();
         int ret = 0;
         if( _iCacheReader != null )
            ret = _iCacheReader.Read(4096, binaryData);

         // always check buffer for data
         // ret=0  means data was read successfully
         // ret=-1 means data may be present, but eof is for sure
         _buffer = binaryData.value;
         _lengthAvailableInBuffer = _buffer.length;
         
         if( _lengthAvailableInBuffer > 0 )
            _headerAvailable = true;
            
         if( ret == -1 )
           _eof = true;

      }
      catch( GenericError e) {
         throw new IOException(MessageFormat.format(RB.getStringResource("InputStream.readException.fmt.ex.txt"/*I18nOK:EMS*/),
                                   new Object[] { e.getMessage() }));
      }

   }   


   // Handle delegation from InputStreamHeader

   /**
    * Get response header properties from server.<p>
    * This call will block until data is read from the stream.
    * <p>
    * The Cache may have properties other than HTTP header properties
    * so as a rule only return properties that are prefixed by "HTTP:".
    * 
    * @return map of response header properties.
    */
   Map getResponseHeaders() throws IOException {
      
      if( _responseHeaders == null )
      {
         StringSeqHolder names = new StringSeqHolder();
         StringSeqHolder values = new StringSeqHolder();      

         try {
            if( _iCacheService != null && _iCacheReader != null )
            {
               if( !_headerAvailable )
               {
                  // block until data is written to the cache
                  // zero-length read hack
                  OctetSeqHolder binaryData = new OctetSeqHolder();
                  _iCacheReader.Read(0, binaryData);
                  
                  _headerAvailable = true;
               }
               
               _iCacheService.GetCacheProperties(_iCacheReader.GetUUID(), names, values);
            }
         }
         catch (GenericError e) {
            throw new IOException(MessageFormat.format(RB.getStringResource("InputStream.getPropertiesException.fmt.ex.txt"/*I18nOK:EMS*/),
                                      new Object[] { e.getMessage() }));
         }
         
         _responseHeaders = new HashMap();
      
         String [] nameArray = names.value;
         String [] valuesArray = values.value;
         for (int i = 0; i < nameArray.length; i++)
         {
            if( nameArray[i].startsWith("HTTP:"/*I18nOK:EMS*/) )
               _responseHeaders.put(nameArray[i].substring(5), valuesArray[i]);
         }
      }

      return _responseHeaders;
      
   }
   
   /**
    * Get header availability.<p>
    * Tests whether or not the input stream header is available
    * without blocking.
    * 
    * @return <code>true</code> if available; otherwise <code>false</code>.
    */
   boolean headerAvailable() {
      
      return _headerAvailable;
   }
   
   
}
