/* $Id$ */
// Copyright (c) 2002 SAS Institute Inc., Cary, NC 27513
package com.sas.iom;

import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.text.DateFormat;
import java.text.ParseException;

import com.sas.io.InputStreamHeaderInterface;


/**
 * Provides the capability to read header information from an
 * {@link com.sas.iom.InputStream IOM InputStream}.
 *
 * @author <a href="mailto:bryan.wolfe@sas.com">Bryan Wolfe</a>
 * @author <a href="mailto:tony.dean@sas.com">Tony Dean</a>
 * @version 1.0
 * @since 1.0
 * 
 * @history
 *    28Aug2002 td  Created.
 *    02Oct2002 td  Remove dependency on JavaMail.
 *    22Dec2003 td  s0227513: clean up catch/throwable try blocks.
 *
 */
public class InputStreamHeader extends UnicastRemoteObject
   implements InputStreamHeaderInterface {
      

   private InputStream _inputStream = null;
   
   public InputStreamHeader(InputStream inputStream) throws RemoteException {
      _inputStream = inputStream;
   }
   
   public String getCharacterEncoding( ) throws IOException, RemoteException {

     String charset = null;
           
     String ctString = getContentType();
     
     if( ctString != null )
     {         
         // iteratate over ContentType parameters:
         // Content-Type: text/html; parm1=value1; parm2=value2
         StringTokenizer tokenizer = new StringTokenizer(ctString, ";"/*I18nOK:EMS*/);
         while( tokenizer.hasMoreTokens() )
         {
             String strParm = tokenizer.nextToken();

             int index = strParm.indexOf("="/*I18nOK:EMS*/);

             if( index == -1 )
                 continue;
               
             String strParmName = strParm.substring(0,index).trim();
             String strParmValue = strParm.substring(index+1).trim();
            
             if( strParmName.equalsIgnoreCase("charset"/*I18nOK:EMS*/) )
             {
                 charset = strParmValue;
                 break;
             }
         }
     }
     
     return charset;
      
   }


   public String getContentEncoding( ) throws IOException, RemoteException {
      
      return getHeaderField("Content-Encoding"/*I18nOK:EMS*/);
   }


   public String getContentType( ) throws IOException, RemoteException {
      
      return getHeaderField("Content-Type"/*I18nOK:EMS*/);
   }


   public long getHeaderFieldDate( String name, long Default ) throws IOException, RemoteException {
      
      long value = Default;
      
      String field = getHeaderField(name);
      if( field != null ) {
          try {
              value = DateFormat.getDateInstance().parse(field).getTime();
          }
          catch( ParseException pe ) {
              ;//ignore
          }
      }

      return value;
   }


   public int getHeaderFieldInt( String name, int Default ) throws IOException, RemoteException {

      int value = Default;
      
      String field = getHeaderField(name);
      if( field != null ) {
          try {
              value = Integer.parseInt(field);
          }
          catch( NumberFormatException nfe ) {
              ;//ignore
          }
      }
      
      return value;
   }


   /**
    * Get response headers by delegating to the InputStream.
    */
   public Map getHeaders() throws IOException, RemoteException {

      Map responseHeaders = null;
      
      if( _inputStream != null )
         responseHeaders = _inputStream.getResponseHeaders();
      
      return responseHeaders;
   }


   /**
    * Determine whether or not the header is available by delegating to the InputStream.
    */
   public boolean headerAvailable() throws RemoteException {

      boolean available = false;
      
      if( _inputStream != null )
         available = _inputStream.headerAvailable();
               
      return available;
   }
   
   
   // Helper methods
   
   private String getHeaderField(String name) throws IOException {

      String s = null;

      if( _inputStream != null )
      {
         Map responseHeaders = _inputStream.getResponseHeaders();

         //try {
         //    s = (String) responseHeaders.get(name);
         //}
         //catch (java.util.NoSuchElementException e) {
            ;//ok
         //}

         // search for case insensitive name instead
         String key;
         
         Iterator iter = responseHeaders.keySet().iterator();
         while( iter.hasNext() )
         {
            key = (String) iter.next();

            if( name.equalsIgnoreCase(key) )
            {
               s = (String) responseHeaders.get(key);
               break;
            }
         }
      }
      
      return s;
   }


}
