// Copyright (c) 2002 SAS Institute Inc., Cary, NC 27513
package com.sas.iom;

import java.io.IOException;
import java.util.Map;
import java.util.Iterator;
import com.sas.iom.SASCache.ICacheWriter;
import com.sas.iom.SASCache.ICacheService;
import com.sas.iom.SASIOMDefs.GenericError;
import com.sas.io.OutputStreamHeaderInterface;
import java.text.MessageFormat;


/**
 * Provides the capability to write output to an IOM OutputStream.
 *
 * @author <a href="mailto:bryan.wolfe@sas.com">Bryan Wolfe</a>
 * @author <a href="mailto:tony.dean@sas.com">Tony Dean</a>
 * @version 1.0
 * @since 1.0
 * 
 * @history
 *    28Aug2002 td  Created.
 *
 */
public class OutputStream extends java.io.OutputStream {

   private ICacheWriter _iCacheWriter = null;
   private ICacheService _iCacheService = null;
   private boolean _dataWritten = false;
   
   private OutputStreamHeader _header = null;
   
   
   public OutputStream(ICacheWriter iCacheWriter, ICacheService iCacheService) {
      _iCacheWriter = iCacheWriter;
      _iCacheService = iCacheService;
   }


   /**
    * Get the header interface for this OutputStream.
    * 
    * @return header OutputStream header.
    * @see com.sas.io.OutputStreamHeaderInterface
    */
   public OutputStreamHeaderInterface getHeader() {

      if( _header == null )
      {
         try {
            _header = new OutputStreamHeader(this);
         }
         catch( java.rmi.RemoteException re ) {
            ;//return null header
         }
      }
      
      return _header;
   }
   
   /**
    * Put data to this OutputStream.
    * 
    * @param data Data to write.
    */
   public void putData(byte[] data) {
      
      StreamData runnable = new StreamData(this, data);
      Thread t = new Thread(runnable);
      t.start();
   }

   /**
    * Put data to this OutputStream.
    * 
    * @param data Data to write.
    */
   public void putData(String data) {

      putData(data, null);
   }
      
   /**
    * Put data to this OutputStream.
    * 
    * @param data Data to write.
    * @param encoding Name of a supported charset.
    */
   public void putData(String data, String encoding) {
      
      StreamData runnable = new StreamData(this, data, encoding);
      Thread t = new Thread(runnable);
      t.start();
   }   
   
   public void write(int b) throws IOException {

      byte[] buffer = new byte[1];
      buffer[0] = (byte) b;
      
      writeBuffer(buffer);
            
      _dataWritten = true;

   }

   public void write(byte[] b) throws IOException {

      writeBuffer(b);
      
      _dataWritten = true;
      
   }

   public void write(byte[] b, int off, int len) throws IOException {

      byte[] buffer = new byte[len];
      for( int i=0; i<len; i++ )
      {
         buffer[i] = b[off+i];
      }

      writeBuffer(buffer);      
      
      _dataWritten = true;

   }

   public void flush() throws IOException {
   }

   public void close() throws IOException {

      try {
         if( _iCacheWriter != null )
            _iCacheWriter.Close();
      }
      catch( com.sas.iom.SASIOMDefs.GenericError e ) {
         throw new IOException(MessageFormat.format(RB.getStringResource("OutputStream.closeException.fmt.ex.txt"/*I18nOK:EMS*/),
                                   new Object[] { e.getMessage() }));
      }
      
   }
   
   
   // Helpers
   
   private void writeBuffer(byte[] buffer) throws IOException {
      
      try {
         if( _iCacheWriter != null )
            _iCacheWriter.Write(buffer);
      }
      catch( com.sas.iom.SASIOMDefs.GenericError e ) {
         throw new IOException(MessageFormat.format(RB.getStringResource("OutputStream.writeException.fmt.ex.txt"/*I18nOK:EMS*/),
                                   new Object[] { e.getMessage() }));
      }
      
   }


   // Handle delegation from OutputStreamHeader
   
   /**
    *  Set a header property for the output stream.
    * 
    * @param name  header property name.
    * @param value header property value.
    * 
    * @throws IOException if a network I/O error occurs.
    * @throws IllegalStateException if called after data is written to the output stream.
    */
   void setHeader( String name, String value ) throws IOException, IllegalStateException {

      if (_dataWritten)
        throw new IllegalStateException(RB.getStringResource("OutputStream.illegalWrite.ex.txt"/*I18nOK:EMS*/));

      try {
         if( _iCacheWriter != null )
            _iCacheWriter.SetProperty( name, value );
      }
      catch( com.sas.iom.SASIOMDefs.GenericError e ) {
         throw new IOException(MessageFormat.format(RB.getStringResource("OutputStream.setPropertyException.fmt.ex.txt"/*I18nOK:EMS*/),
                                   new Object[] { e.getMessage() }));
      }
     
   }

   /**
    *  Set header properties for the output stream.
    * 
    * @param headers map of header properties.
    * 
    * @throws IOException if a network I/O error occurs.
    * @throws IllegalStateException if called after data is written to the output stream.
    */
   void setHeaders( Map headers ) throws IOException, IllegalStateException {
      if (_dataWritten)
        throw new IllegalStateException(RB.getStringResource("OutputStream.illegalWrite.ex.txt"/*I18nOK:EMS*/));

      Iterator i = headers.keySet().iterator();
      while ( i.hasNext() ) {
         String key = (String)i.next();
         try {
            if( _iCacheWriter != null )
               _iCacheWriter.SetProperty( key, (String)headers.get(key) );
         }
         catch( com.sas.iom.SASIOMDefs.GenericError e ) {
            throw new IOException(MessageFormat.format(RB.getStringResource("OutputStream.setPropertyException.fmt.ex.txt"/*I18nOK:EMS*/),
                                      new Object[] { e.getMessage() }));
         }
      }
   }
   
   /**
    * Utility class that streams output data.
    */
   class StreamData implements Runnable {

      OutputStream _outputStream = null;
      Object _data = null;
      String _encoding = null;
            
      StreamData(OutputStream outputStream, Object data) {
         this(outputStream, data, null);
      }
      
      StreamData(OutputStream outputStream, Object data, String encoding) {
         _outputStream = outputStream;
         _data = data;
         _encoding = encoding;
      }
         
      
      public void run()
      {
         if( _data instanceof String )
         {
            String data = (String) _data;
               
            java.io.BufferedWriter bw = null;
            if( _encoding == null )
               bw = new java.io.BufferedWriter(new java.io.OutputStreamWriter(_outputStream));
            else
            {
               try {
                  bw = new java.io.BufferedWriter(new java.io.OutputStreamWriter(_outputStream, _encoding));
               }
               catch( java.io.UnsupportedEncodingException ue ) {
                  ue.printStackTrace();
               }
            }
            
            if( bw != null )
            {
               try {
                  bw.write(data, 0, data.length());
                  bw.flush();
               }
               catch( IOException ioe ) {
                  ioe.printStackTrace();
               }
            }
         }
         else
         {
            byte[] data = (byte[]) _data;
            
            java.io.BufferedOutputStream bos = new java.io.BufferedOutputStream(_outputStream);
            
            try {
               bos.write(data, 0, data.length);
               bos.flush();
            }
            catch( IOException ioe ) {
               ioe.printStackTrace();
            }
         }

         // close the stream after writing all of the data
         try {
            _outputStream.close();
         }
         catch( Throwable t ) {
            ;//ignore
         }

      }

   }

}
