// Copyright (c) 2000 by SAS Institute Inc., Cary, NC 27513
package com.sas.iom;

import com.sas.iom.SAS.IWorkspace;
import com.sas.services.connection.ConnectionInterface;
//
// History:
//
// Date    Name            Remarks
// 23Jul00 B Beers         S0090931: Integrate WorkspacePoolDriver and WorkspaceFactory.
// 02Aug00 B Beers         S0091771: Reimplement getServersByLogicalName.
// 22Aug00 B Beers         S0092375: Scope sasMinSize and sasMinAvail to puddles.
// 23Aug00 B Beers         S0094285: Close connector when it becomes unreferenced.
// 06Sep00 B Beers         S0094267: Enclose thread creation in privileged action blocks.
// 08Sep00 B Beers         S0095633: Prevent close of failed connector in finalization.
// 12Oct00 B Beers         S0099056: Add sas-ServerShutdownAfter and sas-ServerRunForever.
// 16Oct00 B Beers         S0099898: Add getWorkspaceConnectorByUUID to WorkspaceFactory.
// 26Mar01 B Beers         S0112178: Resolve conflict with remote object reference finalizers.
// 10Apr01 B Beers         S0112631: Handle stale connections more robustly.
// 09Apr02 B Beers         S0141718: Remove dependencies on netutil.jar.
// 03Sep02 B Beers         S0164497: Reimplement over com.sas.services.connection.
//
/**
 * This class is the interface between pooled connection to IWorkspace objects
 * and the workspace factory. Users may obtain a workspace by calling
 * <CODE>getWorkspace()</CODE>, and they may notify the workspace factory that they
 * are done with the workspace by calling <CODE>close()</CODE>.
 * @author Biff Beers
 */
public class WorkspaceConnector
{
     protected WorkspaceConnector(WorkspacePool pool,ConnectionInterface cx,IWorkspace iWorkspace)
     {
          this.pool = pool;
          this.cx = cx;
          this.iWorkspace = iWorkspace;
     }

     /**
      * Gets an IWorkspace object. This method returns the same instance of the
      * IWorkspace object each time it is called. The behavior of this method is
      * undefined when it is called after a call to <CODE>close()</CODE>.
      * @return an IOM workspace object
      */
     public synchronized IWorkspace getWorkspace()
     {
          return iWorkspace;
     }

     /**
      * Notify the workspace factory that you are done with the IWorkspace
      * object obtained from this connector. After you call this method, the
      * connection to the IWorkspace object will be returned to the pool
      * and possible used by another user.
      * You do not have to call <CODE>Close()</CODE>
      * on the IWorkspace object before calling this method, but it does not cause
      * problems if you do. You may call this method multiple time without causing
      * problems, but the behavior is undefined if you call either <CODE>getWorkspace()</CODE>
      * or any methods on the IWorkspace object after calling this method.
      */
     public synchronized void close()
     {
          if (!closed)
          {
               cx.close();
               closed = true;
               pool.removeConnector(this);
          }
     }

     private WorkspacePool pool = null;
     private ConnectionInterface cx = null;
     private IWorkspace iWorkspace = null;
     private boolean closed = false;

     protected void finalize() throws Throwable
     {
          if (!closed)
          {
               pool.removeConnector(this);
          }
     }
}