// Copyright (c) 2000 by SAS Institute Inc., Cary, NC 27513
package com.sas.iom;

import java.io.PrintWriter;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.naming.CommunicationException;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import org.omg.CORBA.SystemException;
import com.sas.iom.SAS.IWorkspace;
import com.sas.iom.SAS.IWorkspaceHelper;
import com.sas.services.connection.BridgeServer;
import com.sas.services.connection.ConnectionFactoryAdminInterface;
import com.sas.services.connection.ConnectionFactoryConfiguration;
import com.sas.services.connection.ConnectionFactoryException;
import com.sas.services.connection.ConnectionFactoryInterface;
import com.sas.services.connection.ConnectionFactoryManager;
import com.sas.services.connection.ConnectionInterface;
import com.sas.services.connection.Credential;
import com.sas.services.connection.FailoverCluster;
import com.sas.services.connection.ManualConnectionFactoryConfiguration;
import com.sas.services.connection.PasswordCredential;
import com.sas.services.connection.Puddle;
import com.sas.services.connection.Server;
import com.sas.services.connection.jndi.JNDIConnectionFactoryConfiguration;
import com.sas.text.Message;
import com.sas.util.SasPasswordEncodingException;
import com.sas.util.SasPasswordString;
//
// History:
//
// Date    Name            Remarks
// 23Jul00 B Beers         S0090931: Integrate WorkspacePoolDriver and WorkspaceFactory.
// 24Jul00 B Beers         S0091011: Restore comma-separated host list feature.
// 02Aug00 B Beers         S0091771: Reimplement getServersByLogicalName.
// 22Aug00 B Beers         S0092375: Scope sasMinSize and sasMinAvail to puddles.
// 08Sep00 B Beers         S0095302: Prevent NullPointerException when shutting down non-pooling factory.
// 04Oct00 B Beers         S0098564: Add getFactory to WorkspaceFactory.
// 05Oct00 B Beers         S0098327: Add destroy method to WorkspaceFactory.
// 12Oct00 B Beers         S0099056: Add sas-ServerShutdownAfter and sas-ServerRunForever.
// 16Oct00 B Beers         S0099898: Add getWorkspaceConnectorByUUID to WorkspaceFactory.
// 24Jan01 B Beers         S0107599: Add equals method.
// 01Apr02 B Beers         S0131022: Fix encryption support for LDAP server definitions created with ITAdmin.
// 09Apr02 B Beers         S0141718: Remove dependencies on netutil.jar.
// 03Sep02 B Beers         S0164497: Reimplement over com.sas.services.connection.
// 15Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 21Feb03 B Beers         S0185859: Change handling of factory name to prevent factory proliferation.
//
/**
 * WorkspaceFactory provides a factory for creating and connecting to a remote object
 * on an IOM Server that implements com.sas.iom.SAS.IWorkspace.
 * Information necessary to create a connection can be read out of a JNDI repository
 * (such as LDAP), or it may be provided directly by the caller. The necessary
 * information may be associated with an instance of the factory at the time the factory
 * is instantiated, or it may be provided on a per-method-call basis. If the connection
 * information is provided at the time the factory is instantiated, the factory will
 * provide connection pooling.
 * <P>
 * If connection pooling is used, then this factory will maintain a configurable number
 * of active connections spread across one or more similar servers. When
 * a caller accesses the factory to request a connection to a workspace object, the
 * factory will allocate a connection from the pool if one is available. Otherwise,
 * the factory will create a new connection for the caller. A connection pool has
 * a configurable maximum service limit. If a caller accesses a factory to request a
 * connection to a workspace object, and the factory has reached its maximum service
 * limit, then the factory will not create a new connection, and it will either throw an
 * exception or it will delay service of the request until another factory caller
 * releases a connection and returns it to the pool.
 * <P>
 * Connection pooling requires a two-tiered security model because
 * the factory will need to create connections before the ultimate users of the
 * connections can present credentials that authorize them to use the connections.
 * The two security roles that correspond to each tier are called <EM>pool
 * administrator</EM> and <EM>pool user</EM>. The administrator is responsible for
 * instantiating the factory and, thus, starting the pool. The user must establish
 * that it is authorized to use connections from the factory before it can call a method
 * on the factory to obtain a connection. If the administrator directly provided the
 * factory with its pooling information upon instantiation, then
 * authorization is granted to any user that merely holds an object reference for the
 * factory. The factory is serializable to allow the administrator to control access to
 * the factory by storing it in a protected repository.
 * <P>
 * If the information necessary to start and run the connection pool came from
 * a JNDI respository, then additional security requirements may be placed on the
 * administrator and the user. To successfully instantiate a factory that reads
 * connection information from a JNDI repository, the administrator must provide
 * JNDI credentials that authorize it to read the appropriate server definitions
 * and login definitions in the repository. To successfully
 * obtain a connection from the factory, a user must provide JNDI credentials that
 * are listed in the value of the <CODE>sasAllowedClientDN</CODE> attribute of the 
 * appropriate login definition. The administrator
 * has the option of making this user authorization step unnecessary.
 * <P>
 * If the factory tries to connect to a server in the pool and fails to connect, the
 * factory will remove the server from the pool and will not attempt to make any
 * further connections to that server.
 * @author Murali Srinivasan
 * @author Biff Beers
 */
public class WorkspaceFactory implements Serializable
{
     /**
      * JNDI POOLING CONSTRUCTOR - constructs a factory that reads the information
      * necessary to create a pool of live connections to remote workspace objects
      * from a JNDI repository. The connections can then be accessed quickly using
      * this factory's getWorkspaceConnector() methods. The caller of this constructor
      * must either
      * provide JNDI credentials that allow read access to the server definitions
      * and login definitions in the repository or it must provide an
      * existing instance of javax.naming.directory.DirContext. The caller
      * may configure this factory such that further access through the
      * <CODE>getWorkspaceConnector()</CODE>, <CODE>shutdown()</CODE>, and
      * <CODE>destroy()</CODE> methods also require authorization against the JNDI
      * repository.
      * <P>
      * @param jndiInformation an instance of <STRONG>either</STRONG> <CODE>java.util.Hashtable</CODE>
      * <STRONG>or</STRONG> <CODE>javax.naming.directory.DirContext</CODE>. If a 
      * <CODE>java.util.Hashtable</CODE> object, then it is a table of values that
      * can be used to initialize a <CODE>javax.naming.directory,DirContext</CODE> object.
      * The information in the table identifies the JNDI repository
      * that contains the values necessary to start a connection pool and provides
      * the distinguished name and password of an entity that has read access to that
      * information. The keys for the values are defined in javax.naming.Context, and
      * the required keys are listed below.
      * <P>
      * <UL>
      * <LI><CODE>javax.naming.Context.INITIAL_CONTEXT_FACTORY</CODE>
      * <LI><CODE>javax.naming.Context.PROVIDER_URL</CODE>
      * <LI><CODE>javax.naming.Context.SECURITY_AUTHENTICATION</CODE>
      * <LI><CODE>javax.naming.Context.SECURITY_PRINCIPAL</CODE>
      * <LI><CODE>javax.naming.Context.SECURITY_CREDENTIALS</CODE>
      * </UL>
      * <P>
      * If <EM>jndiInformation</EM> is not an instance of <CODE>java.util.Hashtable</CODE>,
      * then it must be an existing instance of <CODE>javax.naming.directory.DirContext</CODE>
      * that can be used as described above.
      * @param baseDN a partial JNDI distinguished name that defines the section within
      * the repository naming hierarchy that contains the pooling information.
      * @param logicalName the value of the <CODE>sasLogicalName</CODE> attribute for all login
      * definitions and server definitions (relative to the baseDN) in the JNDI
      * repository that contain information for this factory.
      * @param userAuthorizationRequired true if user access to this factory through
      * the getWorkspaceConnector() method should
      * require authorization against the JNDI repository.
      * @param adminAuthorizationRequired true if administative access to this factory through
      * the <CODE>shutdown()</CODE> and <CODE>destroy()</CODE> methods should
      * require authorization against the credentials used to construct this factory.
      * @param logWriter a print stream to write log messages to. If <CODE>null</CODE>, then
      * no log message will be written.
      * @exception com.sas.iom.WorkspaceFactoryException if the administrator credentials
      * are invalid, the pooling information is incomplete, or a connection to a server
      * fails.
      */
     public WorkspaceFactory(Object jndiInformation,String baseDN,String logicalName,boolean userAuthorizationRequired,boolean adminAuthorizationRequired,PrintWriter logWriter) throws WorkspaceFactoryException
     {
          DirContext ctx = null;

          if (jndiInformation instanceof Hashtable)
          {
               try
               {
                    ctx = new InitialDirContext((Hashtable)jndiInformation);
               }
               catch (NamingException ne)
               {
                    throw new WorkspaceFactoryException(ne);
               }

               closeContext = true;
          }
          else
          {
               ctx = (DirContext)jndiInformation;
          }

          ConnectionFactoryConfiguration cxfConfig = new JNDIConnectionFactoryConfiguration(ctx,baseDN,logicalName,"",userAuthorizationRequired,adminAuthorizationRequired,false);
          cookie = WorkspacePool.startPool(cxfConfig,logWriter);

          if (closeContext)
          {
               WorkspacePool.stashContext(cookie,ctx);
          }
     }

     /**
      * AD HOC POOLING CONSTRUCTOR - constructs a factory that uses the provided
      * information to create a pool of live connections to remote workspace objects. The
      * connections can then be accessed quickly using this factory's
      * <CODE>getWorkspaceConnector()</CODE> methods. Access to this factory through the
      * <CODE>getWorkspaceConnector()</CODE>, <CODE>shutdown()</CODE>, and 
      * <CODE>destroy()</CODE> methods do not require further
      * authorization so the caller of this constructor should take care to properly
      * control access to the factory.  The factory is serializable to allow the
      * caller to control access to the factory by storing it in a protected
      * repository.
      * @param connectionDescriptors an array of tables that describe how to connect to
      * the servers that participate in the connection pool. The required values for each
      * table are listed below.
      * <P>
      * <DL>
      * <DT><EM>host</EM><DD>The IP name or address of the machine to connect to
      * <DT><EM>port</EM><DD>The TCP port number to connect to
      * <DT><EM>userName</EM><DD>a user name defined on the host machine
      * <DT><EM>password</EM><DD>the password for <EM>userName</EM>
      * </DL>
      * <P>
      * The optional values for each table are listed below.
      * <P>
      * <DL>
      * <DT><EM>encryptionPolicy</EM><DD><CODE>none</CODE>, <CODE>optional</CODE>, or
      * <CODE>required</CODE>. The default value is <CODE>none</CODE>. If the specified
      * value is not <CODE>none</CODE>, then the client and server will attempt to
      * agree upon an encryption algorithm and what should be encrypted. If that
      * negotiation fails and the value is <CODE>required</CODE>, then the connection
      * will fail. If the value is <CODE>optional</CODE>, then the connection will
      * continue with no encryption unless the server is configured to require
      * encryption. In that case, the connection will fail.
      * <DT><EM>encryptionAlgorithms</EM><DD>a comma-separated list of encryption
      * algorithms that the client would like to use to secure this connection in order
      * of preference. Possible values
      * are <CODE>sasproprietary</CODE>, <CODE>rc2</CODE>, <CODE>rc4</CODE>,
      * <CODE>des</CODE>, and <CODE>tripledes</CODE>. To use any algorithm other than
      * <CODE>sasproprietary</CODE>, SAS/SECURE software must be installed on both the
      * client and server. Furthermore, use of <CODE>des</CODE>, and
      * <CODE>tripledes</CODE> may be restricted outside the USA due to export law.
      * The connection will use the first algorithm on this list that is also
      * supported by the server. If no value is specified, the connection will use the
      * first value suggested by the server that the client is capable of supporting.
      * <DT><EM>encryptionContent</EM><DD> specifies which messages between the client
      * and server should be encrypted if encryption is used. Possible values are
      * <CODE>all</CODE> (the default; encrypt all messages) and
      * <CODE>authentication</CODE> (encrypt only message that contain user name and
      * password information). The default is <CODE>all</CODE>. If the server requires a
      * higher level of encryption than is specified by this value, the connection
      * will fail.
      * <DT><EM>sasMaxPerWorkspacePool</EM><DD>the maximum number of connections
      * this factory should have to this server at any given time. The default value 
      * is 10.
      * <DT><EM>sas-RecycleActivationLimit</EM><DD>the maximum number of times a
      * connection can be reused before it is disconnected (and reconnected if
      * necessary). The value must be greater than or equal to <CODE>0</CODE>.
      * If the value is <CODE>0</CODE>, then the connection can be reused any
      * number of times. The default value is <CODE>0</CODE>.
      * <DT><EM>sas-ServerRunForever</EM><DD>The value must be either <CODE>true</CODE>
      * or <CODE>false</CODE>. If the value is <CODE>false</CODE>, then unallocated live
      * connections will be disconnected after a period of time (determined by the value
      * of <EM>sas-ServerShutdownAfter</EM>) unless they are allocated to a user before
      * that period of time passes. Otherwise, unallocated live connections will remain
      * alive indefinitely. The default value is <CODE>true</CODE>.
      * <DT><EM>sas-ServerShutdownAfter</EM><DD>The period of time, in minutes, that an
      * unallocated live connection will remain alive waiting to be allocated to a user.
      * This attribute is ignored if the value of <EM>sas-ServerRunForever</EM> is
      * <CODE>true</CODE>. The value must not be less than <CODE>0</CODE>, and it must
      * not be greater than <CODE>1440</CODE> (the number of minutes in a day). The
      * default value is <CODE>3</CODE>. If the value is <CODE>0</CODE>, then a
      * connection returned to a pool by a user will be disconnected immediately unless
      * another user is waiting for a connection from the pool.
      * </DL>
      * @param poolParameters a table of values that describe how the connection pool
      * should behave. Possible values are listed below. All values are optional.
      * This parameter may be <CODE>null</CODE> if the default is desired for all
      * values.
      * <P>
      * <DL>
      * <DT><EM>sasMinSize</EM><DD>The minimum number of live connections that must be
      * associated with this factory at any given time unless the factory has been shut
      * down or destroyed. This number includes both connections
      * that have been allocated to users and unallocated connections. The default value
      * is 0.
      * <DT><EM>sasMinAvail</EM><DD>The minimum number of unallocated live connections
      * that must be associated with this factory at any given time unless the factory
      * has been shut down or destroyed. The default value is 0.
      * </DL>
      * @param logWriter a print stream to write log messages to. If <CODE>null</CODE>, then
      * no log message will be written.
      * @exception com.sas.iom.WorkspaceFactoryException if a connection descriptor or
      * login descriptor is invalid or if a connection to a server fails
      */
     public WorkspaceFactory(Properties[] connectionDescriptors,Properties poolParameters,PrintWriter logWriter) throws WorkspaceFactoryException
     {
          if (connectionDescriptors == null || connectionDescriptors.length == 0)
          {
               throw new IllegalArgumentException("\"connectionDescriptors\" cannot be null or empty.");
          }

          // collate all the servers by their respective logins
          // keep a running tally of maxClients for each puddle and the pool as a whole
          Map puddleMap = new HashMap();
          long maxClientsPool = 0L;

          for (int i=0;i<connectionDescriptors.length;i++)
          {
               Properties cd = connectionDescriptors[i];
               Credential iomLogin = propertiesToLogin(cd);
               List newClusterList = propertiesToClusterList(cd);
               Iterator newClusterItor = newClusterList.iterator();
               long maxClientsPuddle = 0L;

               while (newClusterItor.hasNext())
               {
                    FailoverCluster iomCluster = (FailoverCluster)newClusterItor.next();
                    int maxClients = iomCluster.getMaxClients();
                    maxClientsPuddle += maxClients;
                    maxClientsPool += maxClients;
               }

               List curClusterList = (List)puddleMap.get(iomLogin);

               if (curClusterList == null)
               {
                    curClusterList = new LinkedList();
                    curClusterList.add(Zero);
                    curClusterList.add(iomLogin);
                    puddleMap.put(iomLogin,curClusterList);
               }

               Long CurMaxClients = (Long)curClusterList.get(0);
               long newMaxClients = CurMaxClients.longValue() + maxClientsPuddle;
               Long NewMaxClients = new Long(newMaxClients);
               curClusterList.set(0,NewMaxClients);
               curClusterList.addAll(newClusterList);

          }

          // dump all the "puddles" into a list and sort the list by maxClients
          Set puddleSet = puddleMap.entrySet();
          int puddleCount = puddleSet.size();
          List puddleList = new ArrayList(puddleCount);
          Iterator puddleItor = puddleSet.iterator();

          while (puddleItor.hasNext())
          {
               Map.Entry puddleEntry = (Map.Entry)puddleItor.next();
               List iomClusterList = (List)puddleEntry.getValue();
               puddleList.add(iomClusterList);
          }

          Comparator comparator = new Comparator()
          {
               public int compare(Object o1,Object o2)
               {
                    List l1 = (List)o1;
                    List l2 = (List)o2;

                    if (l1 == l2)
                    {
                         return 0;
                    }

                    Long M1 = (Long)l1.get(0);
                    Long M2 = (Long)l2.get(0);
                    int result = M1.compareTo(M2);
                    return result;
               }
          };

          Collections.sort(puddleList,comparator);

          // determine what portion of minSize and minAvail each puddle should get
          // round DOWN for each puddle - fix rounding errors later
          int[][] minimums = new int[puddleCount][2];
          int minSize = 0;
          int minAvail = 0;

          if (poolParameters != null)
          {
               String minSizeString = poolParameters.getProperty("sasMinSize","0");
               minSize = Integer.parseInt(minSizeString);
               String minAvailString = poolParameters.getProperty("sasMinAvail","0");
               minAvail = Integer.parseInt(minAvailString);
          }

          puddleItor = puddleList.iterator();
          int minSizeTally = 0;
          int minAvailTally = 0;
          int i = 0;

          while (puddleItor.hasNext())
          {
               List puddleInfo = (List)puddleItor.next();
               Long MaxClients = (Long)puddleInfo.get(0);
               Credential iomLogin = (Credential)puddleInfo.get(1);
               double ratio = MaxClients.doubleValue() / maxClientsPool;
               int ms = (int)Math.floor(minSize * ratio);
               int ma = (int)Math.floor(minAvail * ratio);
               minimums[i][0] = ms;
               minimums[i][1] = ma;
               minSizeTally += ms;
               minAvailTally += ma;
               i++;
          }

          // go back and fix rounding errors
          int[] errs = new int[2];
          errs[0] = minSize - minSizeTally;
          errs[1] = minAvail - minAvailTally;

          for (int j=0;j<2;j++)
          {
               for (i=0;i<errs[j];i++)
               {
                    minimums[i][j]++;
               }
          }

          puddleItor = puddleList.iterator();
          List iomPuddleList = new ArrayList(puddleCount);
          i = 0;

          while (puddleItor.hasNext())
          {
               List puddleInfo = (List)puddleItor.next();
               Credential iomLogin = (Credential)puddleInfo.get(1);
               List iomClusterList = puddleInfo.subList(2,puddleInfo.size());
               int iomClusterCount = iomClusterList.size();
               FailoverCluster[] iomClusterArray = new FailoverCluster[iomClusterCount];
               iomClusterArray = (FailoverCluster[])iomClusterList.toArray(iomClusterArray);
               Puddle iomPuddle = new Puddle(iomClusterArray,iomLogin);
               iomPuddle.setMinSize(minimums[i][0]);
               iomPuddle.setMinAvail(minimums[i][1]);
               i++;
               iomPuddleList.add(iomPuddle);
          }

          Puddle[] iomPuddleArray = new Puddle[puddleCount];
          iomPuddleArray = (Puddle[])iomPuddleList.toArray(iomPuddleArray);
          ConnectionFactoryConfiguration cxfConfig = new ManualConnectionFactoryConfiguration(iomPuddleArray);
          cookie = WorkspacePool.startPool(cxfConfig,logWriter);
     }

     /**
      * NON-POOLING CONSTRUCTOR - constructs a factory that creates a new connection
      * each time createWorkspaceByLogicalName() or createWorkspaceByServer() is called.
      */
     public WorkspaceFactory()
     {
     }

     private WorkspaceFactory(WorkspacePoolCookie cookie)
     {
          this.cookie = cookie;
     }

     /**
      * Compares an object with this factory for equality. Two non-pooling
      * factories are considered equal only if they are the same object
      * (i.e. their object references are equal).
      * @param that the object to compare with this object.
      * @return true if this object is equal to that object.
      */
     public boolean equals(Object that)
     {
          if (this == that)
          {
               return true;
          }

          if (cookie == null || that == null || !that.getClass().getName().equals(getClass().getName()))
          {
               return false;
          }

          WorkspaceFactory thatFactory = (WorkspaceFactory)that;
          WorkspacePoolCookie thatCookie = thatFactory.getCookie();

          if (!cookie.equals(thatCookie))
          {
               return false;
          }

          return true;
     }

     public int hashCode()
     {
          int hc = 0;

          if (cookie == null)
          {
               hc = super.hashCode();
          }
          else
          {
               hc = getClass().getName().hashCode();
               hc ^= cookie.hashCode();
          }

          return hc;
     }

     /**
      * Searches for a factory that any user can use to obtain a
      * connection to a server described in the JNDI repository identified by
      * <EM>repositoryProperties</EM> under the given base DN and logical name.
      * If such a factory has not been created yet, then this method returns null.
      * @param repositoryProperties
      * The information in the table identifies the JNDI repository
      * that contains the values necessary to start a connection pool. The keys for the
      * values are defined in javax.naming.Context, and
      * the required keys are listed below.
      * <P>
      * <UL>
      * <LI><CODE>javax.naming.Context.INITIAL_CONTEXT_FACTORY</CODE>
      * <LI><CODE>javax.naming.Context.PROVIDER_URL</CODE>
      * </UL>
      * <P>
      * Other values that are not required but are significant when searching are
      * <UL>
      * <LI><CODE>javax.naming.Context.DNS_URL</CODE>
      * <LI><CODE>javax.naming.Context.OBJECT_FACTORIES</CODE>
      * <LI><CODE>javax.naming.Context.URL_PKG_PREFIXES</CODE>
      * </UL>
      * <P>
      * Values that describe a user identity for security purposes including
      * <CODE>javax.naming.Context.SECURITY_AUTHENTICATION</CODE>,
      * <CODE>javax.naming.Context.SECURITY_PRINCIPAL</CODE>, and
      * <CODE>javax.naming.Context.SECURITY_CREDENTIALS</CODE>, are not necessary
      * because this method only searches for factories that all users can use.
      * @param baseDN a partial JNDI distinguished name that defines the section within
      * the repository naming hierarchy that contains the pooling information.
      * @param logicalName the value of the <CODE>sasLogicalName</CODE> attribute for all login
      * definitions and server definitions (relative to the baseDN) in the JNDI
      * repository that contain information for starting a connection pool.
      */
     public static WorkspaceFactory getFactory(Hashtable repositoryProperties,String baseDN,String logicalName)
     {
          return getFactory(null,null,repositoryProperties,baseDN,logicalName);
     }

     /**
      * Searches for a factory that the given user can use to obtain a
      * connection to a server described in the JNDI repository identified by
      * <EM>repositoryProperties</EM> under the given base DN and logical name.
      * If such a factory has not been created yet, then this method returns null.
      * @param userDN the distinguished name of some object in the given JNDI directory
      * that has a <CODE>userPassword</CODE> attribute. If this method finds a factory
      * that meets all the other search criteria specified in the arguments, and that
      * factory has set <EM>userAuthorizationRequired</EM> to <CODE>true</CODE>, then
      * This object's DN must be a value of (or be a member of a group whose DN is a
      * value of) the
      * <CODE>sasAllowedClientDN</CODE> attribute of one of
      * the login definitions used to create that factory. If the factory does not
      * require user authorization, then this parameter is ignored.
      * @param userPassword the value of the <CODE>userPassword</CODE> attribute for the
      * object identified with <EM>userDN</EM>. This value may be in plain text or it
      * may be base64 encoded. If it is base64 encoded, it must begin with the tag
      * <CODE>{base64}</CODE>.
      * @param repositoryProperties
      * The information in the table identifies the JNDI repository
      * that contains the values necessary to start a connection pool. The keys for the 
      * values are defined in javax.naming.Context, and
      * the required keys are listed below.
      * <P>
      * <UL>
      * <LI><CODE>javax.naming.Context.INITIAL_CONTEXT_FACTORY</CODE>
      * <LI><CODE>javax.naming.Context.PROVIDER_URL</CODE>
      * </UL>
      * <P>
      * Other values that are not required but are significant when searching are
      * <UL>
      * <LI><CODE>javax.naming.Context.DNS_URL</CODE>
      * <LI><CODE>javax.naming.Context.OBJECT_FACTORIES</CODE>
      * <LI><CODE>javax.naming.Context.URL_PKG_PREFIXES</CODE>
      * </UL>
      * <P>
      * Values that describe a user identity for security purposes including
      * <CODE>javax.naming.Context.SECURITY_AUTHENTICATION</CODE>,
      * <CODE>javax.naming.Context.SECURITY_PRINCIPAL</CODE>, and
      * <CODE>javax.naming.Context.SECURITY_CREDENTIALS</CODE>, are not necessary
      * because security information is provided with the <EM>userDN</EM> and
      * <EM>userPassword</EM> parameters.
      * @param baseDN a partial JNDI distinguished name that defines the section within
      * the repository naming hierarchy that contains the pooling information.
      * @param logicalName the value of the <CODE>sasLogicalName</CODE> attribute for all login
      * definitions and server definitions (relative to the baseDN) in the JNDI
      * repository that contain information for starting a connection pool.
      */
     public static WorkspaceFactory getFactory(String userDN,String userPassword,Hashtable repositoryProperties,String baseDN,String logicalName)
     {
          WorkspacePoolCookie cookie = WorkspacePool.matchPool(userDN,userPassword,repositoryProperties,baseDN,logicalName);

          if (cookie == null)
          {
               return null;
          }

          return new WorkspaceFactory(cookie);
     }

     /**
      * Allocates a connection from an
      * ad hoc pool or a JNDI pool with user authorization turned off. If all live
      * connections in the pool have been allocated to other users, then this factory
      * will create a new connection for this user. If this factory
      * requires user authorization through JNDI, then this method will throw an
      * exception. If an administrator has shut down or destroyed this factory, then
      * this method will throw an exception.
      * @param waitMilliseconds used if this factory has reached its maximum
      * use limit at the time of this method call. If that is the case, then this method
      * will behave in one of the three following ways based on the value of
      * <CODE>waitMilliseconds</CODE>:
      * <P>
      * <DL>
      * <DT>greater than zero<DD>this method will wait up to <EM>N</EM> milliseconds for
      * factory usage to fall below the limit (where <EM>N</EM> is the value of
      * <CODE>waitMilliseconds</CODE>) before either returning a workspace connector or
      * throwing an exception.
      * <DT>zero<DD>this method will wait indefinitely for factory usage to fall below
      * the maximum usage limit, and will return a workspace connector once it does.
      * <DT>less than zero<DD>this method will throw an exception immediately.
      * </DL>
      * @return a connector to a workspace
      * @exception com.sas.iom.WorkspaceFactoryException if this factory requires user
      * authorization through JNDI, if the factory has been shut down or destroyed,
      * if the factory has reached its maximum usage limit, or there is a problem
      * creating new connections.
      */
     public WorkspaceConnector getWorkspaceConnector(long waitMilliseconds) throws WorkspaceFactoryException
     {
          if (cookie == null)
          {
               // {0}: The connection pooling features are not available on this instance of WorkspaceFactory.
               String messageIndex = "IOM.S206.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          WorkspacePool pool = WorkspacePool.lookupPool(cookie);

          if (pool == null)
          {
               // {0}: The requested pool has already been shutdown.
               String messageIndex = "IOM.S208.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          return pool.getConnector(null,null,waitMilliseconds);
     }

     /**
      * Allocates a connection from  a JNDI pool with user authorization
      * turned on. User credentials will be authorized against the same JNDI repsoitory
      * that was used to instantiate the factory. If the user does not
      * have authority to use the factory, then this method will throw an exception.
      * If all live
      * connections in the pool have been allocated to other users, then this factory
      * will create a new connection for this user. If an administrator has shut down
      * or destroyed this factory, then this method will throw an exception.
      * @param userDN the distinguished name of some object in the JNDI directory used
      * to instantiate this factory that has a <CODE>userPassword</CODE> attribute. This
      * object's DN must be a value of (or be a member of a group whose DN is a value of)
      * the <CODE>sasAllowedClientDN</CODE> attribute of one of
      * the login definitions used to create this factory. If this factory does not
      * require user authorization, then this parameter is ignored.
      * @param userPassword the value of the <CODE>userPassword</CODE> attribute for the
      * object identified with <EM>userDN</EM>. This value may be in plain text or it
      * may be base64 encoded. If it is base64 encoded, it must begin with the tag
      * <CODE>{base64}</CODE>. If this factory does not require user
      * authorization, then this parameter is ignored.
      * @param waitMilliseconds used if this factory has reached its maximum
      * use limit at the time of this method call. If that is the case, then this method
      * will behave in one of the three following ways based on the value of
      * <CODE>waitMilliseconds</CODE>:
      * <P>
      * <DL>
      * <DT>greater than zero<DD>this method will wait up to <EM>N</EM> milliseconds for
      * factory usage to fall below the limit (where <EM>N</EM> is the value of
      * <CODE>waitMilliseconds</CODE>) before either returning a workspace connector or
      * throwing an exception.
      * <DT>zero<DD>this method will wait indefinitely for factory usage to fall below
      * the maximum usage limit, and will return a workspace connector once it does.
      * <DT>less than zero<DD>this method will throw an exception immediately.
      * </DL>
      * @return a connector to a workspace
      * @exception com.sas.iom.WorkspaceFactoryException if the user is not authorized to
      * use this factory, if the factory has been shut down or destroyed,
      * if the factory has reached its maximum usage limit, or there is a problem
      * creating new connections.
      */
     public WorkspaceConnector getWorkspaceConnector(String userDN,String userPassword,long waitMilliseconds) throws WorkspaceFactoryException
     {
          if (cookie == null)
          {
               // {0}: The connection pooling features are not available on this instance of WorkspaceFactory.
               String messageIndex = "IOM.S207.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          WorkspacePool pool = WorkspacePool.lookupPool(cookie);

          if (pool == null)
          {
               // {0}: The requested pool has already been shutdown.
               String messageIndex = "IOM.S209.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          return pool.getConnector(userDN,userPassword,waitMilliseconds);
     }

     /**
      * Returns a reference to a previously created workspace connector that encapsulates a
      * workspace object with the given UUID. The UUID for a workspace object can be found
      * by calling its <CODE>UniqueIdentifier()</CODE> method. If no such workspace connector
      * has been created,
      * or if such a connector has been created but has already been closed, this method
      * returns <CODE>null</CODE>.
      * @param uuid the UUID of the desired workspace
      * @return a reference to the workspace connector that encapsulates the desired workspace
      */
     public static WorkspaceConnector getWorkspaceConnectorByUUID(String uuid)
     {
          return getWorkspaceConnectorByUUID(null,null,uuid);
     }

     /**
      * Returns a reference to a previously created workspace connector that encapsulates a
      * workspace object with the given UUID. The UUID for a workspace object can be found
      * by calling its <CODE>UniqueIdentifier()</CODE> method. If no such workspace connector
      * has been created,
      * or if such a connector has been created but has already been closed, this method
      * returns <CODE>null</CODE>.
      * @param userDN the distinguished name of some object in the given JNDI directory
      * that has a <CODE>userPassword</CODE> attribute. If this method finds a factory
      * that meets all the other search criteria specified in the arguments, and that
      * factory has set <EM>userAuthorizationRequired</EM> to <CODE>true</CODE>, then
      * This object's DN must be a value of (or be a member of a group whose DN is a
      * value of) the
      * <CODE>sasAllowedClientDN</CODE> attribute of one of
      * the login definitions used to create that factory. If the factory does not
      * require user authorization, then this parameter is ignored.
      * @param userPassword the value of the <CODE>userPassword</CODE> attribute for the
      * object identified with <EM>userDN</EM>. This value may be in plain text or it
      * may be base64 encoded. If it is base64 encoded, it must begin with the tag
      * <CODE>{base64}</CODE>.
      * @param uuid the UUID of the desired workspace
      * @return a reference to the workspace connector that encapsulates the desired workspace
      * @exception com.sas.iom.WorkspaceFactoryException if the user is not authorized to
      * use the desired workspace
      */
     public static WorkspaceConnector getWorkspaceConnectorByUUID(String userDN,String userPassword,String uuid) throws WorkspaceFactoryException
     {
          if (uuid == null)
          {
               return null;
          }

          return WorkspacePool.getConnectorByUUID(userDN,userPassword,uuid);
     }

     /**
      * Creates a new connection to a workspace object using LDAP information.
      * Server definitions are located in the LDAP repository by searching
      * (relative to <EM>baseDN</EM>) for sasServer objects that have
      * the given <EM>logicalName</EM> as an attribute. Login definitions are located in
      * the LDAP repository by searching (relative to <EM>baseDN</EM>) for sasLogin
      * objects that have the given <EM>sasReferenceDN</EM> as an attribute.
      * Each server definition is then paired with a login definition by matching
      * objects with like values for the <CODE>sasDomainName</CODE> attribute.
      *
      * <p>
      * @param logicalName A valid logical name of the server.
      * @param sasReferenceDN A reference DN for the user to match with the domain name
      * to get a login definition.
      * @param protocol Protocol to communicate with the server. Currently, the only
      * valid value is <CODE>bridge</CODE>.
      * @param ldapProviderUrl A url specifying the location of an LDAP provider.
      * This value typically takes the form <CODE>ldap://<EM>host</EM>:<EM>port</EM></CODE>
      * @param ldapProperties A Properties object containing the following LDAP
      * properties:
      * <p>
      * <dl>
      * <dt><code>baseDN</code>
      * <dd>Container that will contain all sasserver and saslogin definitions.
      * <p>
      * <dt><code>ldapUserName</code>
      * <dd>DN of a user to use to connect to LDAP.
      * <p>
      * <dt><code>ldapPassword</code>
      * <dd>Password that corresponds with ldapUserName.
      * <p>
      * <dt><code>authentication</code>
      * <dd>LDAP authentication value. This is an optional field. The authentication values are NONE or SIMPLE.
      * </dl>
      *
      * @return A workspace object.
      * @exception WorkspaceFactoryException This exception is thrown if an error occurs.
      */
     public IWorkspace createWorkspaceByLogicalName(String logicalName,String sasReferenceDN,String protocol,String ldapProviderUrl, Properties ldapProperties ) throws WorkspaceFactoryException
     {
          String baseDN = null;

          if (ldapProperties != null)
          {
               baseDN = ldapProperties.getProperty("baseDN");
          }

          Hashtable env = new Hashtable();
          DirContext ctx = createContext(env,ldapProviderUrl,ldapProperties,baseDN);
          ConnectionFactoryConfiguration cxfConfig = new JNDIConnectionFactoryConfiguration(ctx,baseDN,logicalName,"",true,false,true);
          ConnectionFactoryManager cxfManager = new ConnectionFactoryManager();
          ConnectionFactoryInterface cxf = null;

          try
          {
               cxf = cxfManager.getFactory(cxfConfig);
          }
          catch (ConnectionFactoryException cfe)
          {
               throw new WorkspaceFactoryException(cfe);
          }

          List domainList = cxf.getDomains();
          List iomLoginList = findLogins(ctx,env,domainList,sasReferenceDN,baseDN);
          Iterator iomLoginItor = iomLoginList.iterator();
          WorkspaceFactoryException wfe = null;

          while (iomLoginItor.hasNext())
          {
               Credential iomLogin = (Credential)iomLoginItor.next();
               ConnectionInterface cx = null;

               try
               {
                    cx = cxf.getConnection(iomLogin);
               }
               catch (ConnectionFactoryException cfe)
               {
                    if (wfe == null)
                    {
                         wfe = new WorkspaceFactoryException(cfe);
                    }

                    continue;
               }

               try
               {
                    org.omg.CORBA.Object obj = cx.getObject();
                    IWorkspace iWorkspace = IWorkspaceHelper.narrow(obj);
                    return iWorkspace;
               }
               catch (SystemException se)
               {
                    cx.close();

                    if (wfe == null)
                    {
                         wfe = new WorkspaceFactoryException(se);
                    }
               }
          }

          throw wfe;
     }

     /**
      * Creates a new IWorkspace object based on Server definitions.
      *
      * @param serverInfo A properties object containing the following server definitions and a valid
      * username and password :
      * <p>
      * <dt><code>host</code>
      * <dd>A comma delimited String of valid host names. This definition is required. This maps to machineDNSName in LDAP.
      * <p>
      * <dt><code>port</code>
      * <dd>TCP/IP port String to connect with bridge. Do not specify this definition if you specify serviceName.
      * <p>
      * <dt><code>protocol</code>
      * <dd>A valid protocol String to communicate with the SAS server (bridge or corba). If you do not specify a value
      * <dd>this property will be defaulted to "bridge".
      * <p>
      * <dt><code>userName</code>
      * <dd>A valid username String for the MVA server.
      * <p>
      * <dt><code>password</code>
      * <dd>The password String that corresponds to the username you specify with the <code>userName</code> property.
      * <DT><I>encryptionPolicy</I><DD>whether or not JavaBridge should attempt to negotiate with
      * the server over which encryption algorithm to use and what to do if the negotiations fail.
      * Possible values:
      * <DL>
      * <DT>none<DD>do not use encryption. This is the default.
      * <DT>optional<DD>attempt to use encryption but, if algorithm negotiation fails, continue with an
      * unencrypted session
      * <DT>required<DD>attempt to use encryption but, if algorithm negotiation fails, fail the connection
      * </DL>
      * <p>
      * <DT>encryptionContent<DD>specifies which messages should be encrypted if encryption is used. Possible values:
      * <DL>
      * <DT>all<DD>encrypt all messages. This is the default.
      * <DT>authentication<DD>encrypt only messages that contain user name and password information.
      * </DL>
      * <p>
      * <DT>encryptionAlgorithms<DD>the list of algorithms you are willing to use in order of
      * preference. Values in the list should be separated by commas and chosen from sasproprietary,
      * rc2, rc4, des, or tripledes. If no value is specified, then one of the server's favorite algorithms
      * will be used.
      * <P>
      * Note that the server's encryption policy is, by default, optional so no special action is needed
      * to prepare an IOM BRIDGE server to use encryption. However, it is possible to set up the server
      * so that encryption is required. Consult IOM BRIDGE server documentation for details.
      * <p>
      * <dt><code>name</code>
      * <dd>Name of the server definition. This definition is optional.
      * <p>
      * <dt><code>description</code>
      * <dd>A description of the server definition. This definition is optional.
      * <p>
      * <dt><code>transcodingPreference</code>
      * <dd>This property controls where character data will be transcoded. There are two possible
      * values:
      * <p>
      * <dl>
      * <dt><code>client</code>
      * <dd>Transcode character data on the client using Java character encoding converters to convert
      * between Unicode and encoding appropriate for SAS specific locale.
      * <dt><code>server</code>
      * <dd>Transcode character data on the server using SAS character encoding converters to convert
      * betweeen some intermediate encoding (which both client and server is capable of converting to/from)
      * and encoding appropriate for SAS specific locale. The client also trancodes character data using
      * the Java character encoding converters to convert between the intermediate encoding (described
      * above) and Unicode.
      * <p>
      * Note : The only time user sets the value to server is when he/she knows that the Java environment
      * does not have a converter that converts between Unicode and the encoding appropriate for SAS
      * specific locale.
      * </dl>
      * <p>
      * The default value is <code>client</code>. If the Java environment is not capable of performing
      * the required conversions, then the transcoding responsibilities will automatically be shifted to
      * SAS.
      * <p>
      * </dl>
      *
      * @return A workspace object.
      * @exception WorkspaceFactoryException This exception is thrown if an error occurs.
      */
     public IWorkspace createWorkspaceByServer(java.util.Properties serverInfo) throws WorkspaceFactoryException
     {
          // validation - we need at least a host and port to continue
          if (serverInfo == null)
          {
               // {0}: WorkspaceFactory could not create a connection to a server because no server information was provided.
               String messageIndex = "IOM.S202.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          String hosts = serverInfo.getProperty("host");

          if (hosts == null || hosts.length() == 0)
          {
               // {0}: WorkspaceFactory could not create a connection to a server because the server information provided did not contain a value for the host attribute.
               String messageIndex = "IOM.S203.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          String portString = serverInfo.getProperty("port");

          if (portString == null || portString.length() == 0)
          {
               // {0}: WorkspaceFactory could not create a connection to a server because the server information provided did not contain a value for the port attribute.
               String messageIndex = "IOM.S204.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          int port = Integer.parseInt(portString);

          // map encryption properties if necessary
          String encryptionPolicy = serverInfo.getProperty("encryptionPolicy");
          String encryptionContent = serverInfo.getProperty("encryptionContent");
          String encryptionAlgorithms = serverInfo.getProperty("encryptionAlgorithms");

          if (encryptionPolicy == null && encryptionContent == null && encryptionAlgorithms == null)
          {
               String sasRequiredEncryptionLevel = serverInfo.getProperty("sasrequiredencryptionlevel");

               if (sasRequiredEncryptionLevel != null)
               {
                    sasRequiredEncryptionLevel = sasRequiredEncryptionLevel.toLowerCase();

                    if (sasRequiredEncryptionLevel.equals("authentication") || sasRequiredEncryptionLevel.equals("credentials"))
                    {
                         encryptionPolicy = BridgeServer.ENCRYPTION_POLICY_REQUIRED;
                         encryptionContent = BridgeServer.ENCRYPTION_CONTENT_AUTHENTICATION;
                    }
                    else if (sasRequiredEncryptionLevel.equals("all") || sasRequiredEncryptionLevel.equals("everything"))
                    {
                         encryptionPolicy = BridgeServer.ENCRYPTION_POLICY_REQUIRED;
                         encryptionContent = BridgeServer.ENCRYPTION_CONTENT_ALL;
                    }
               }

               encryptionAlgorithms = serverInfo.getProperty("sasclientencryptionalgorithm");
          }

          // parse the list of hosts
          StringTokenizer hostTok = new StringTokenizer(hosts,",");
          int hostCount = hostTok.countTokens();
          List iomServerList = new ArrayList(hostCount);

          while (hostTok.hasMoreTokens())
          {
               String host = hostTok.nextToken().trim();

               if (host.length() > 0)
               {
                    BridgeServer iomServer = new BridgeServer(Server.CLSID_SAS,host,port);

                    if (encryptionPolicy != null)
                    {
                         iomServer.setEncryptionPolicy(encryptionPolicy);
                    }

                    if (encryptionContent != null)
                    {
                         iomServer.setEncryptionContent(encryptionContent);
                    }

                    if (encryptionAlgorithms != null)
                    {
                         iomServer.setEncryptionAlgorithms(encryptionAlgorithms);
                    }

                    iomServerList.add(iomServer);
               }
          }

          final int iomServerListSize = iomServerList.size();

          if (iomServerListSize == 0)
          {
               // {0}: WorkspaceFactory could not create a connection to a server because the server information provided did not contain a value for the host attribute.
               String messageIndex = "IOM.S205.ex.txt";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new WorkspaceFactoryException(message);
          }

          // create a factory configuration
          Server[] iomServerArray = new Server[iomServerListSize];
          iomServerArray = (Server[])iomServerList.toArray(iomServerArray);
          FailoverCluster iomCluster = new FailoverCluster(iomServerArray);
          ConnectionFactoryConfiguration cxfConfig = new ManualConnectionFactoryConfiguration(iomCluster);

          // set up logging
          PrintWriter logWriter = (PrintWriter)serverInfo.get("logWriter");
          String debugLevelString = serverInfo.getProperty("debugLevel","0");
          int debugLevel = Integer.parseInt(debugLevelString);

          if (debugLevel > 0 && logWriter == null)
          {
               logWriter = new PrintWriter(System.out);
          }

          ConnectionFactoryManager cxfManager = new ConnectionFactoryManager();
          String loggerName = cxfManager.getFactoryLoggerName(cxfConfig);
          WorkspacePool.configLogger(loggerName,logWriter,debugLevel);

          // get a connection
          String userName = serverInfo.getProperty("userName");
          String password = serverInfo.getProperty("password");
          ConnectionInterface cx = null;

          try
          {
               ConnectionFactoryInterface cxf = cxfManager.getFactory(cxfConfig);
               cx = cxf.getConnection(userName,password);
          }
          catch (ConnectionFactoryException cfe)
          {
               throw new WorkspaceFactoryException(cfe);
          }

          // get a workspace and return it
          try
          {
               org.omg.CORBA.Object obj = cx.getObject();
               IWorkspace iWorkspace = IWorkspaceHelper.narrow(obj);
               return iWorkspace;
          }
          catch (SystemException se)
          {
               throw new WorkspaceFactoryException(se);
          }
     }

     /**
      * Gets a list of server definitions matching the logicalName.
      *
      * @param logicalName A valid logical name of the server.
      * @param protocol Protocol to communicate with the server. Valid values are bridge or corba.
      * @param ldapProviderUrl A valid url specifying and LDAP server and its port to make connection of the form :
      * "ldap://host:port".
      * @param ldapProperties A Properties object containing the following LDAP properties :
      * <p>
      * <dl>
      * <dt><code>baseDN</code>
      * <dd>Container that will contain all sasserver and saslogin definitions.
      * <p>
      * <dt><code>ldapUserName</code>
      * <dd>DN of a user to use to connect to LDAP.
      * <p>
      * <dt><code>ldapPassword</code>
      * <dd>Password that corresponds with ldapUserName.
      * <p>
      * <dt><code>authentication</code>
      * <dd>LDAP authentication value. This is an optional field. The authentication values are NONE or SIMPLE.
      * <p>
      * </dl>
      *
      * @return An array of Properties object containing server definitions matching the logical name.
      * <p>
      * The properties returned contains the following server definitions.
      * <p>
      * <dl>
      * <dt><code>host</code>
      * <dd>A comma delimited String  of valid host names. This definition is required. This maps to machineDNSName in LDAP.
      * <p>
      * <dt><code>port</code>
      * <dd>TCP/IP port to connect with bridge.
      * <p>
      * <dt><code>protocol</code>
      * <dd>Protocol to communicate with the SAS server. Valid values are bridge or corba.
      * <p>
      * <dt><code>name</code>
      * <dd>Name of the server definition.
      * <p>
      * <dt><code>description</code>
      * <dd>A description of the server definition.
      * </dl>
      * @exception WorkspaceFactoryException This exception is thrown if an error occurs.
      */
     public Properties[] getServersByLogicalName(String logicalName,String protocol,String ldapProviderUrl,Properties ldapProperties) throws WorkspaceFactoryException
     {
          String baseDN = null;

          if (ldapProperties != null)
          {
               baseDN = ldapProperties.getProperty("baseDN");
          }

          Hashtable env = new Hashtable();
          DirContext ctx = createContext(env,ldapProviderUrl,ldapProperties,baseDN);
          Properties[] serverArray = findServers(ctx,env,baseDN,logicalName);
          return serverArray;
     }

     /**
      * Disconnects all unallocated live connections to remote workspace objects. Allocated
      * connections and connections in progress are not immediately affected, but,
      * as they are deallocated, the connections will be closed instead of returned
      * to the pool. After this method has been called, all subsequent calls to
      * <CODE>shutdown()</CODE> will do nothing, but calls
      * <CODE>getWorkspaceConnector()</CODE> will result in an exception.
      * @exception com.sas.iom.WorkspaceFactoryException if this factory requires
      * administrators to request authorization to use this method.
      */
     public void shutdown() throws WorkspaceFactoryException
     {
          shutdown(null,null);
     }

     /**
      * Disconnects all unallocated live connections to remote workspace objects. Allocated
      * connections and connections in progress are not immediately affected, but,
      * as they are deallocated, the connections will be closed instead of returned
      * to the pool. After this method has been called, all subsequent calls to
      * <CODE>shutdown()</CODE> will do nothing, but calls
      * <CODE>getWorkspaceConnector()</CODE> will result in an exception.
      * @param adminDN the distinguished name of some object in the JNDI directory used
      * to instantiate this factory that has a <CODE>userPassword</CODE> attribute. This
      * object must have read access to the pool parameters object that was
      * named when this factory was instantiated. If this factory does not require administrator
      * authorization, then this parameter is ignored.
      * @param adminPassword the value of the <CODE>userPassword</CODE> attribute for the
      * object identified with <EM>adminDN</EM>. This value may be in plain text or it
      * may be base64 encoded. If it is base64 encoded, it must begin with the tag
      * <CODE>{base64}</CODE>. If this factory does not require administrator
      * authorization, then this parameter is ignored.
      * @exception com.sas.iom.WorkspaceFactoryException if the administrator credentials
      * provided are invalid.
      */
     public void shutdown(String adminDN,String adminPassword) throws WorkspaceFactoryException
     {
          if (cookie != null)
          {
               WorkspacePool pool = WorkspacePool.lookupPool(cookie);

               if (pool != null)
               {
                    pool.shutdown(adminDN,adminPassword);
               }

               closeCtx();
          }
     }

     /**
      * Immediately disconnects all live connections to remote workspace object including
      * allocated connections and connections in progress. Any attempt to use a connection
      * allocated from this factory after this method has been called will result in an
      * exception. After this method has been called, all subsequent calls to
      * <CODE>shutdown()</CODE> and <CODE>destroy()</CODE> will do nothing, but calls
      * <CODE>getWorkspaceConnector()</CODE> will result in an exception.
      * @exception com.sas.iom.WorkspaceFactoryException if this factory requires
      * administrators to request authorization to use this method.
      */
     public void destroy() throws WorkspaceFactoryException
     {
          destroy(null,null);
     }

     /**
      * Immediately disconnects all live connections to remote workspace object including
      * allocated connections and connections in progress. Any attempt to use a connection
      * allocated from this factory after this method has been called will result in an
      * exception. After this method has been called, all subsequent calls to
      * <CODE>shutdown()</CODE> and <CODE>destroy()</CODE> will do nothing, but calls
      * <CODE>getWorkspaceConnector()</CODE> will result in an exception.
      * @param adminDN the distinguished name of some object in the JNDI directory used
      * to instantiate this factory that has a <CODE>userPassword</CODE> attribute. This
      * object must have read access to the pool parameters object that was
      * named when this factory was instantiated. If this factory does not require administrator
      * authorization, then this parameter is ignored.
      * @param adminPassword the value of the <CODE>userPassword</CODE> attribute for the
      * object identified with <EM>adminDN</EM>. This value may be in plain text or it
      * may be base64 encoded. If it is base64 encoded, it must begin with the tag
      * <CODE>{base64}</CODE>. If this factory does not require administrator
      * authorization, then this parameter is ignored.
      * @exception com.sas.iom.WorkspaceFactoryException if the administrator credentials
      * provided are invalid.
      */
     public void destroy(String adminDN,String adminPassword) throws WorkspaceFactoryException
     {
          if (cookie != null)
          {
               WorkspacePool pool = WorkspacePool.lookupPool(cookie);

               if (pool != null)
               {
                    pool.destroy(adminDN,adminPassword);
               }

               closeCtx();
          }
     }

     static final long serialVersionUID = 5794004025788357148L;
     private static final Long Zero = new Long(0L);
     private static SearchControls searchControls = null;
     private WorkspacePoolCookie cookie = null;
     private boolean closeContext = false;

     static
     {
          searchControls = new SearchControls();
          searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
     }
     
     protected WorkspacePoolCookie getCookie()
     {
          return cookie;
     }

     private void closeCtx()
     {
          if (cookie != null)
          {
               if (closeContext)
               {
                    DirContext ctx = WorkspacePool.removeContext(cookie);

                    if (ctx != null)
                    {
                         try
                         {
                              ctx.close();
                         }
                         catch (NamingException ne)
                         {
                         }
                    }
               }
          }
     }

     private static Credential propertiesToLogin(Properties p)
     {
          String userName = p.getProperty("userName","");
          String password = p.getProperty("password","");
          Credential iomLogin = new PasswordCredential(userName,password);
          return iomLogin;
     }

     private static List propertiesToClusterList(Properties p)
     {
          String hosts = p.getProperty("host");

          if (hosts == null)
          {
               throw new IllegalArgumentException("\"host\" cannot be unspecified in a connection descriptor.");
          }

          String portString = p.getProperty("port");

          if (portString == null)
          {
               throw new IllegalArgumentException("\"port\" cannot be unspecified in a connection descriptor.");
          }

          int port = Integer.parseInt(portString);
          String encryptionPolicy = p.getProperty("encryptionPolicy");
          String encryptionContent = p.getProperty("encryptionContent");
          String encryptionAlgorithms = p.getProperty("encryptionAlgorithms");
          String maxClientsString = p.getProperty("sasMaxPerWorkspacePool","10");
          int maxClients = Integer.parseInt(maxClientsString);
          String reactivationLimitString = p.getProperty("sas-RecycleActivationLimit","0");
          int reactivationLimit = Integer.parseInt(reactivationLimitString);
          String runForeverString = p.getProperty("sas-ServerRunForever","true");
          Boolean runForever = new Boolean(runForeverString);
          int shutdownAfter = -1;

          if (!runForever.booleanValue())
          {
               String shutdownAfterString = p.getProperty("sas-ServerShutdownAfter","3");
               shutdownAfter = Integer.parseInt(shutdownAfterString);
          }

          StringTokenizer hostTok = new StringTokenizer(hosts,",");
          int hostCount = hostTok.countTokens();

          if (hostCount == 0)
          {
               throw new IllegalArgumentException("\"host\" cannot be empty in a connection descriptor.");
          }

          List iomClusterList = new ArrayList(hostCount);

          while (hostTok.hasMoreTokens())
          {
               String host = hostTok.nextToken();
               host = host.trim();
               BridgeServer iomServer = new BridgeServer(Server.CLSID_SAS,host,port);

               if (encryptionPolicy != null)
               {
                    iomServer.setEncryptionPolicy(encryptionPolicy);
               }

               if (encryptionContent != null)
               {
                    iomServer.setEncryptionContent(encryptionContent);
               }

               if (encryptionAlgorithms != null)
               {
                    iomServer.setEncryptionAlgorithms(encryptionAlgorithms);
               }

               iomServer.setMaxClients(maxClients);
               iomServer.setReactivationLimit(reactivationLimit);
               iomServer.setShutdownAfterMinutes(shutdownAfter);
               FailoverCluster iomCluster = new FailoverCluster(iomServer);
               iomClusterList.add(iomCluster);
          }

          return iomClusterList;
     }

     private static DirContext createContext(Hashtable env,String providerURL,Properties ldapProperties,String baseDN) throws WorkspaceFactoryException
     {
          String securityAuthentication = null;
          String securityPrincipal = null;
          String securityCredentials = null;

          if (ldapProperties != null)
          {
               securityAuthentication = ldapProperties.getProperty("authentication","none").toLowerCase();
               securityPrincipal = ldapProperties.getProperty("ldapUserName");
               securityCredentials = ldapProperties.getProperty("ldapPassword");
          }

          if ("simple".equals(securityAuthentication))
          {
               env.put(Context.SECURITY_PRINCIPAL,securityPrincipal);
               env.put(Context.SECURITY_CREDENTIALS,securityCredentials);
          }

          env.put(Context.SECURITY_AUTHENTICATION,securityAuthentication);
          env.put(Context.PROVIDER_URL,providerURL);
          env.put(Context.INITIAL_CONTEXT_FACTORY,"com.sun.jndi.ldap.LdapCtxFactory");
          DirContext ctx = null;

          try
          {
               ctx = new InitialDirContext(env);
          }
          catch (NamingException ne)
          {
               if ("none".equals(securityAuthentication))
               {
                    throw new WorkspaceFactoryException(ne);
               }

               StringBuffer oldStylePrincipalBuffer = new StringBuffer();
               oldStylePrincipalBuffer.append("cn=").append(securityPrincipal);

               if (baseDN != null)
               {
                    oldStylePrincipalBuffer.append(",").append(baseDN);
               }

               String oldStylePrincipal = oldStylePrincipalBuffer.toString();
               env.put(Context.SECURITY_PRINCIPAL,oldStylePrincipal);

               try
               {
                    ctx = new InitialDirContext(env);
               }
               catch (NamingException ne2)
               {
                    throw new WorkspaceFactoryException(ne2);
               }
          }

          return ctx;
     }

     private static Properties[] findServers(DirContext ctx,Hashtable env,String baseDN,String logicalName) throws WorkspaceFactoryException
     {
          List serverList = new LinkedList();

          try
          {
               StringBuffer stringBuffer = new StringBuffer();
               stringBuffer.append("(&(objectclass=sasServer)(sasProtocol=bridge)(sasLogicalName=");
               String filter = stringBuffer.append(logicalName).append("))").toString();
               NamingEnumeration serverEnumeration = null;

               try
               {
                    serverEnumeration = ctx.search(baseDN,filter,searchControls);
               }
               catch (CommunicationException ce)
               {
                    ctx = reconnect(ctx,env,ce);
                    serverEnumeration = ctx.search(baseDN,filter,searchControls);
               }

               while (serverEnumeration.hasMore())
               {
                    Properties cd = new Properties();
                    Attribute attr = null;
                    NamingEnumeration ne = null;
                    SearchResult serverResult = (SearchResult)serverEnumeration.next();
                    Attributes serverAttributes = serverResult.getAttributes();
                    attr = serverAttributes.get("sasMachineDNSName");

                    if (attr != null)
                    {
                         ne = attr.getAll();
                         StringBuffer nameListBuffer = new StringBuffer();
                         boolean hasMore = ne.hasMore();

                         while (hasMore)
                         {
                              nameListBuffer.append((String)ne.next());

                              if (hasMore = ne.hasMore())
                              {
                                   nameListBuffer.append(",");
                              }
                         }

                         cd.put("host",nameListBuffer.toString());
                    }

                    // The strange use of an enumeration here is
                    // for compatibility with an earlier version.
                    //-------------------------------------------
                    attr = serverAttributes.get("sasPort");

                    if (attr != null)
                    {
                         ne = attr.getAll();

                         while (ne.hasMore())
                         {
                              cd.put("port",ne.next());
                         }
                    }

                    attr = serverAttributes.get("sasdomainname");

                    if (attr != null)
                    {
                         ne = attr.getAll();

                         while (ne.hasMore())
                         {
                              cd.put("sasdomainname",ne.next());
                         }
                    }

                    attr = serverAttributes.get("sasrequiredencryptionlevel");

                    if (attr != null)
                    {
                         ne = attr.getAll();

                         while (ne.hasMore())
                         {
                              cd.put("sasrequiredencryptionlevel",ne.next());
                         }
                    }

                    attr = serverAttributes.get("sasclientencryptionalgorithm");

                    if (attr != null)
                    {
                         ne = attr.getAll();

                         while (ne.hasMore())
                         {
                              cd.put("sasclientencryptionalgorithm",ne.next());
                         }
                    }

                    attr = serverAttributes.get("saslogicalname");

                    if (attr != null)
                    {
                         ne = attr.getAll();

                         while (ne.hasMore())
                         {
                              cd.put("saslogicalname",ne.next());
                         }
                    }

                    attr = serverAttributes.get("sasProtocol");

                    if (attr != null)
                    {
                         ne = attr.getAll();

                         while (ne.hasMore())
                         {
                              cd.put("protocol",ne.next());
                         }
                    }

                    serverList.add(cd);
               }
          }
          catch (NamingException e0)
          {
               throw new WorkspaceFactoryException(e0);
          }

          final int serverListSize = serverList.size();
          Properties[] serverArray = new Properties[serverListSize];
          serverArray = (Properties[])serverList.toArray(serverArray);
          return serverArray;
     }

     private static List findLogins(DirContext ctx,Hashtable env,List domainList,String referenceDN,String baseDN) throws WorkspaceFactoryException
     {
          int domainListSize = domainList.size();
          List iomLoginList = new ArrayList(domainListSize);
          Iterator domainItor = domainList.iterator();

          try
          {
               while (domainItor.hasNext())
               {
                    String domain = (String)domainItor.next();
                    StringBuffer stringBuffer = new StringBuffer("(&(objectclass=sasLogin)(sasReferenceDN=").append(referenceDN).append(")");

                    if (domain != null)
                    {
                         stringBuffer.append("(sasDomainName=").append(domain).append(")");
                    }
                    else
                    {
                         stringBuffer.append("(!(sasDomainName=*))");
                    }

                    String filter = stringBuffer.append(")").toString();
                    NamingEnumeration loginEnumeration = null;

                    try
                    {
                         loginEnumeration = ctx.search(baseDN,filter,searchControls);
                    }
                    catch (CommunicationException ce)
                    {
                         ctx = reconnect(ctx,env,ce);
                         loginEnumeration = ctx.search(baseDN,filter,searchControls);
                    }

                    if (!loginEnumeration.hasMore())
                    {
                         filter = stringBuffer.append("objectclass=top)").toString();

                         try
                         {
                              loginEnumeration = ctx.search(baseDN,filter,searchControls);
                         }
                         catch (CommunicationException ce)
                         {
                              ctx = reconnect(ctx,env,ce);
                              loginEnumeration = ctx.search(baseDN,filter,searchControls);
                         }
                    }

                    if (!loginEnumeration.hasMore())
                    {
                         Credential iomLogin = new PasswordCredential("","",domain);
                         iomLoginList.add(iomLogin);
                    }
                    else
                    {
                         while (loginEnumeration.hasMore())
                         {
                              SearchResult searchResult = (SearchResult)loginEnumeration.next();
                              Attributes loginAttributes = searchResult.getAttributes();
                              Attribute userNameAttr = loginAttributes.get("sasLoginName");
                              String userName = (String)userNameAttr.get();
                              Attribute passwordAttr = loginAttributes.get("sasUserPassword");
                              String codedPassword = (String)passwordAttr.get();
                              String password = SasPasswordString.decode(codedPassword);
                              Credential iomLogin = new PasswordCredential(userName,password,domain);
                              iomLoginList.add(iomLogin);
                         }
                    }
               }
          }
          catch (NamingException e0)
          {
               throw new WorkspaceFactoryException(e0);
          }
          catch (SasPasswordEncodingException e1)
          {
               throw new WorkspaceFactoryException(e1);
          }

          return iomLoginList;
     }

     private static DirContext reconnect(DirContext ctx,Hashtable env,CommunicationException ce) throws NamingException
     {
          try
          {
               ctx.close();
          }
          catch (NamingException e)
          {
          }

          try
          {
               ctx = new InitialDirContext(env);
          }
          catch (NamingException e)
          {
               throw ce;
          }

          return ctx;
     }
}
