// Copyright (c) 2000 by SAS Institute Inc., Cary, NC 27513
package com.sas.iom;

import java.io.PrintWriter;
import java.lang.ref.WeakReference;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import org.omg.CORBA.SystemException;
import com.sas.iom.SAS.IWorkspace;
import com.sas.iom.SAS.IWorkspaceHelper;
import com.sas.net.util.WriterHandler;
import com.sas.services.connection.ConnectionFactoryAdminInterface;
import com.sas.services.connection.ConnectionFactoryConfiguration;
import com.sas.services.connection.ConnectionFactoryException;
import com.sas.services.connection.ConnectionFactoryInterface;
import com.sas.services.connection.ConnectionFactoryManager;
import com.sas.services.connection.ConnectionInterface;
import com.sas.services.connection.jndi.JNDIConnectionFactoryConfiguration;
//
// History:
//
// Date    Name            Remarks
// 23Jul00 B Beers         S0090931: Integrate WorkspacePoolDriver and WorkspaceFactory.
// 02Aug00 B Beers         S0091771: Reimplement getServersByLogicalName.
// 22Aug00 B Beers         S0092375: Scope sasMinSize and sasMinAvail to puddles.
// 08Sep00 B Beers         S0095633: Prevent close of failed connector in finalization.
// 04Oct00 B Beers         S0098564: Add getFactory to WorkspaceFactory.
// 05Oct00 B Beers         S0098327: Add destroy method to WorkspaceFactory.
// 12Oct00 B Beers         S0099056: Add sas-ServerShutdownAfter and sas-ServerRunForever.
// 16Oct00 B Beers         S0099898: Add getWorkspaceConnectorByUUID to WorkspaceFactory.
// 23Jan01 B Beers         S0107539: Prevent getFactory from returning shutdown factories.
// 10Apr01 B Beers         S0112631: Handle stale connections more robustly.
// 13Apr01 B Beers         S0112632: Improve handling of sasAllowedClientDN.
// 09Apr02 B Beers         S0113902: Account for pending connections when deciding if pool is full.
// 09Apr02 B Beers         S0141718: Remove dependencies on netutil.jar.
// 03Sep02 B Beers         S0164497: Reimplement over com.sas.services.connection.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 21Feb03 B Beers         S0185859: Change handling of factory name to prevent factory proliferation.
//
/**
 * @author Biff Beers
 */
public final class WorkspacePool
{
     private WorkspacePool(WorkspacePoolCookie cookie,ConnectionFactoryConfiguration cxfConfig,ConnectionFactoryInterface cxf)
     {
          this.cookieRef = new WeakReference(cookie);
          this.cxfConfig = cxfConfig;
          this.cxf = cxf;
     }

     private static Map poolTable = new HashMap();
     private static Map poolTable2 = new WeakHashMap();
     private static Map contextTable = new HashMap();
     WeakReference cookieRef = null;
     ConnectionFactoryConfiguration cxfConfig = null;
     ConnectionFactoryInterface cxf = null;
     Set connectorRefSet = new HashSet();

     protected WorkspacePoolCookie getCookie()
     {
          return (WorkspacePoolCookie)cookieRef.get();
     }

     protected ConnectionFactoryConfiguration getConfig()
     {
          return cxfConfig;
     }

     protected Set getConnectorRefSet()
     {
          return connectorRefSet;
     }

     protected static WorkspacePoolCookie startPool(ConnectionFactoryConfiguration cxfConfig,PrintWriter logWriter) throws WorkspaceFactoryException
     {
          synchronized (poolTable)
          {
               ConnectionFactoryManager cxfManager = new ConnectionFactoryManager();
               String loggerName = cxfManager.getFactoryLoggerName(cxfConfig);
               configLogger(loggerName,logWriter,0);
               int hc = cxfConfig.hashCode();
               WorkspacePoolCookie cookie = new WorkspacePoolCookie(hc);
               ConnectionFactoryInterface cxf = null;

               try
               {
                    cxf = cxfManager.getFactory(cxfConfig);
               }
               catch (ConnectionFactoryException cfe)
               {
                    throw new WorkspaceFactoryException(cfe);
               }

               WorkspacePool pool = new WorkspacePool(cookie,cxfConfig,cxf);
               poolTable.put(cookie,pool);
               poolTable2.put(cookie,pool);
               return cookie;
          }
     }

     protected static WorkspacePoolCookie matchPool(String userDN,String userPassword,Hashtable repositoryProperties,String baseDN,String logicalName)
     {
          synchronized (poolTable)
          {
               Set poolSet = poolTable.entrySet();
               Iterator poolItor = poolSet.iterator();

               while (poolItor.hasNext())
               {
                    Map.Entry poolEntry = (Map.Entry)poolItor.next();
                    WorkspacePool pool = (WorkspacePool)poolEntry.getValue();
                    ConnectionFactoryConfiguration thatConfig = pool.getConfig();

                    if (thatConfig instanceof JNDIConnectionFactoryConfiguration)
                    {
                         JNDIConnectionFactoryConfiguration jndiConfig = (JNDIConnectionFactoryConfiguration)thatConfig;

                         if (matchConfigs(jndiConfig,repositoryProperties,baseDN,logicalName))
                         {
                              if (jndiConfig.isUserAuth())
                              {
                                   try
                                   {
                                        WorkspaceConnector connector = pool.getConnector(userDN,userPassword,0L);
                                        connector.close();
                                   }
                                   catch (WorkspaceFactoryException wfe)
                                   {
                                        continue;
                                   }
                              }

                              WorkspacePoolCookie cookie = pool.getCookie();
                              return cookie;
                         }
                    }
               }

               return null;
          }
     }

     protected static WorkspacePool lookupPool(WorkspacePoolCookie cookie) throws WorkspaceFactoryException
     {
          return (WorkspacePool)poolTable2.get(cookie);
     }

     protected static WorkspaceConnector getConnectorByUUID(String userDN,String userPassword,String uuid) throws WorkspaceFactoryException
     {
          synchronized (poolTable)
          {
               Set poolSet = poolTable.entrySet();
               Iterator poolItor = poolSet.iterator();

               while (poolItor.hasNext())
               {
                    Map.Entry poolEntry = (Map.Entry)poolItor.next();
                    WorkspacePool pool = (WorkspacePool)poolEntry.getValue();
                    Set connectorRefSet = pool.getConnectorRefSet();

                    synchronized (connectorRefSet)
                    {
                         Iterator connectorRefItor = connectorRefSet.iterator();

                         while (connectorRefItor.hasNext())
                         {
                              WeakReference connectorRef = (WeakReference)connectorRefItor.next();
                              Object obj = connectorRef.get();

                              if (obj != null)
                              {
                                   WorkspaceConnector connector = (WorkspaceConnector)obj;
                                   IWorkspace iWorkspace = connector.getWorkspace();
                                   String ui = iWorkspace.UniqueIdentifier();

                                   if (uuid.equals(ui))
                                   {
                                        ConnectionFactoryConfiguration cxfConfig = pool.getConfig();

                                        if (cxfConfig instanceof JNDIConnectionFactoryConfiguration)
                                        {
                                             if (((JNDIConnectionFactoryConfiguration)cxfConfig).isUserAuth())
                                             {
                                                  try
                                                  {
                                                       WorkspaceConnector wc = pool.getConnector(userDN,userPassword,0L);
                                                       wc.close();
                                                  }
                                                  catch (WorkspaceFactoryException wfe)
                                                  {
                                                       return null;
                                                  }
                                             }
                                        }

                                        return connector;
                                   }
                              }
                         }
                    }
               }
          }

          return null;
     }

     protected WorkspaceConnector getConnector(String userDN,String password,long waitMilliseconds) throws WorkspaceFactoryException
     {
          ConnectionInterface cx = null;

          try
          {
               cx = cxf.getConnection(userDN,password,waitMilliseconds);
          }
          catch (ConnectionFactoryException cfe)
          {
               throw new WorkspaceFactoryException(cfe);
          }

          IWorkspace iWorkspace = null;

          try
          {
               org.omg.CORBA.Object obj = cx.getObject();
               iWorkspace = IWorkspaceHelper.narrow(obj);
          }
          catch (SystemException se)
          {
               throw new WorkspaceFactoryException(se);
          }

          WorkspaceConnector connector = new WorkspaceConnector(this,cx,iWorkspace);

          synchronized (connectorRefSet)
          {
               connectorRefSet.add(new WeakReference(connector));
          }

          return connector;
     }

     protected static void stashContext(WorkspacePoolCookie cookie,DirContext ctx)
     {
          synchronized (contextTable)
          {
               contextTable.put(cookie,ctx);
          }
     }

     protected static DirContext removeContext(WorkspacePoolCookie cookie)
     {
          synchronized (contextTable)
          {
               return (DirContext)contextTable.remove(cookie);
          }
     }

     protected void removeConnector(WorkspaceConnector connector)
     {
          synchronized (connectorRefSet)
          {
               Iterator connectorRefItor = connectorRefSet.iterator();

               while (connectorRefItor.hasNext())
               {
                    WeakReference connectorRef = (WeakReference)connectorRefItor.next();
                    Object obj = connectorRef.get();

                    if (obj == connector)
                    {
                         connectorRefItor.remove();
                         return;
                    }
               }
          }
     }

     protected void shutdown(String adminDN,String adminPassword) throws WorkspaceFactoryException
     {
          synchronized (poolTable)
          {
               try
               {
                    ConnectionFactoryAdminInterface cxfAdmin = cxf.getAdminInterface(adminDN,adminPassword);
                    cxfAdmin.shutdown();
               }
               catch (ConnectionFactoryException cfe)
               {
                    throw new WorkspaceFactoryException(cfe);
               }

               WorkspacePoolCookie cookie = getCookie();
               poolTable.remove(cookie);
          }
     }

     protected void destroy(String adminDN,String adminPassword) throws WorkspaceFactoryException
     {
          synchronized (poolTable)
          {
               try
               {
                    ConnectionFactoryAdminInterface cxfAdmin = cxf.getAdminInterface(adminDN,adminPassword);
                    cxfAdmin.destroy();
               }
               catch (ConnectionFactoryException cfe)
               {
                    throw new WorkspaceFactoryException(cfe);
               }

               WorkspacePoolCookie cookie = getCookie();
               poolTable.remove(cookie);
               poolTable2.remove(cookie);
          }
     }

     protected static void configLogger(String loggerName,final PrintWriter logWriter,int debugLevel)
     {
          final Logger logger = Logger.getLogger(loggerName);
          final int dl = (debugLevel > 3) ? 3 : debugLevel;

          PrivilegedAction action = new PrivilegedAction()
          {
               public Object run()
               {
                    logger.setUseParentHandlers(false);

                    if (logWriter == null)
                    {
                         logger.setLevel(Level.OFF);
                    }
                    else
                    {
                         WriterHandler handler = new WriterHandler(logWriter);
                         logger.addHandler(handler);
                         Level level = null;

                         switch (dl)
                         {
                              case 1:level=Level.FINE;break;
                              case 2:level=Level.FINER;break;
                              case 3:level=Level.FINEST;break;
                              default:level=Level.INFO;break;
                         }

                         logger.setLevel(level);
                    }

                    return null;
               }
          };

          try
          {
               AccessController.doPrivileged(action);
          }
          catch (SecurityException se)
          {
               // oh well, we tried.
          }
     }

     private static boolean matchConfigs(JNDIConnectionFactoryConfiguration thatConfig,Hashtable env,String baseDN,String logicalName)
     {
          Hashtable thatEnv = null;

          try
          {
               DirContext thatCtx = thatConfig.getContext();
               thatEnv = thatCtx.getEnvironment();
          }
          catch (NamingException ne)
          {
               return false;
          }

          if (env != thatEnv)
          {
               if (env == null || thatEnv == null)
               {
                    return false;
               }

               String thisFact = (String)env.get(Context.INITIAL_CONTEXT_FACTORY);
               String thatFact = (String)thatEnv.get(Context.INITIAL_CONTEXT_FACTORY);

               if (thisFact == null ? thatFact != null : !thisFact.equals(thatFact))
               {
                    return false;
               }

               String thisProv = (String)env.get(Context.PROVIDER_URL);
               String thatProv = (String)thatEnv.get(Context.PROVIDER_URL);

               if (thisProv == null ? thatProv != null : !thisProv.equals(thatProv))
               {
                    return false;
               }

               String thisDNS = (String)env.get(Context.DNS_URL);
               String thatDNS = (String)thatEnv.get(Context.DNS_URL);

               if (thisDNS == null ? thatDNS != null : !thisDNS.equals(thatDNS))
               {
                    return false;
               }

               String thisObjFact = (String)env.get(Context.OBJECT_FACTORIES);
               String thatObjFact = (String)thatEnv.get(Context.OBJECT_FACTORIES);

               if (thisObjFact == null ? thatObjFact != null : !thisObjFact.equals(thatObjFact))
               {
                    return false;
               }

               String thisPre = (String)env.get(Context.URL_PKG_PREFIXES);
               String thatPre = (String)thatEnv.get(Context.URL_PKG_PREFIXES);

               if (thisPre == null ? thatPre != null : !thisPre.equals(thatPre))
               {
                    return false;
               }
          }

          String thatBaseDN = thatConfig.getBaseDN();

          if (baseDN != thatBaseDN)
          {
               if (baseDN == null || thatBaseDN == null)
               {
                    return false;
               }

               String normThis = JNDIConnectionFactoryConfiguration.normalizeDN(baseDN);
               String normThat = JNDIConnectionFactoryConfiguration.normalizeDN(thatBaseDN);

               if (!normThis.equals(normThat))
               {
                    return false;
               }
          }

          String thatLogicalName = thatConfig.getLogicalName();

          if (logicalName == null ? thatLogicalName != null : !logicalName.equals(thatLogicalName))
          {
               return false;
          }

          return true;
     }
}