// Copyright (c) 1999 by SAS Institute Inc., Cary, NC 27513
package com.sas.iom;

import java.io.PrintWriter;
import java.util.Hashtable;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
//
// History:
//
// Date    Name            Remarks
// 24Nov99 B Beers         Created.
// 23Jul00 B Beers         S0090931: Integrate WorkspacePoolDriver and WorkspaceFactory.
// 27Jul00 B Beers         S0091360: Fix ClassCastException in setLog.
// 05Oct00 B Beers         S0098327: Add destroy method to WorkspaceFactory.
// 25Jan01 B Beers         S0108231: Added doc comment and deprecation tag.
// 03Sep02 B Beers         S0164497: Reimplement over com.sas.services.connection.
//
/**
 * The old interface for connection pooling.
 * @author Biff Beers
 * @deprecated The methods in this class are ambiguous. Use com.sas.iom.WorkspaceFactory instead.
 * It is not always clear when calling <CODE>getPooledWorkspace</CODE> if the intent is to create a
 * new connection pool or to get a connection from an existing pool. Furthermore, calls to
 * <CODE>closePool</CODE> may match more than one pool, and the caller has no way to control which
 * pools should be closed.
 * @see com.sas.iom.WorkspaceFactory
 */
public class WorkspacePoolDriver
{
     /**
      * This constructor is private so that no one can instantiate the class. All methods
      * on the class are static.
      */
     private WorkspacePoolDriver()
     {
     }

     /**
      * Get a workspace from an appropriate pool.
      * @param credentials information that identifies a JNDI directory and specifies how to bind to it. Behavior is undefined
      * if the value of this parameter is <CODE>null</CODE>.
      * @param searchContext the starting point in the JNDI directory for searches. Behavior is undefined
      * if the value of this parameter is <CODE>null</CODE>.
      * @param sasLogicalName a logical name that identifies a pool or a
      * set of pools. Behavior is undefined
      * if the value of this parameter is <CODE>null</CODE>.
      * @param userDN the distinguished name of a user or group of users that is used to determine which, if any, of the
      * the pools identified by <CODE>sasLogicalName</CODE> can be used to fulfill this request. If the value is
      * <CODE>null</CODE>, then any of the pools identified by <CODE>sasLogicalName</CODE> can be used to fulfill this
      * request.
      * @param waitMilliseconds used if the workspace pool has reached its maximum
      * use threshold at the time of this method call. If that is the case, then this method will behave in
      * one of the three following ways based on the value of <CODE>waitMilliseconds</CODE>:
      * <P>
      * <DL>
      * <DT>greater than zero<DD>this method will wait up to <EM>N</EM> milliseconds for workspace pool usage
      * to fall below the threshold (where <EM>N</EM> is the value of <CODE>waitMilliseconds</CODE>) before either
      * returning a workspace connector or throwing an exception.
      * <DT>zero<DD>this method will wait indefinitely for workspace pool usage to fall below the maximum
      * usage threshold, and will return a workspace connector once it does.
      * <DT>less than zero<DD>this method will throw an exception immediately.
      * </DL>
      * @return a connector to a workspace
      * @exception com.sas.iom.WorkspacePoolException if the given logical name and user DN do not identify any
      * workspace pools or if the given credentials do not provide sufficient authority
      * to access the workspace pool information in JNDI or if the pool cannot be initialized.
      */
     public static WorkspaceConnector getPooledWorkspace(Hashtable credentials,String searchContext,String sasLogicalName,String userDN,long waitMilliseconds) throws WorkspacePoolException
     {
          String password = findPassword(credentials,userDN);
          WorkspaceConnector connector = null;

          try
          {
               WorkspaceFactory wf = WorkspaceFactory.getFactory(userDN,password,credentials,searchContext,sasLogicalName);

               if (wf == null)
               {
                    wf = new WorkspaceFactory(credentials,searchContext,sasLogicalName,(userDN!=null),false,logWriter);
               }

               connector = wf.getWorkspaceConnector(userDN,password,waitMilliseconds);
          }
          catch (WorkspaceFactoryException wfe)
          {
               throw new WorkspacePoolException(wfe);
          }

          return connector;
     }

     /**
      * Close all unused workspace objects in the appropriate pool(s). Workspace objects in use at the time of this
      * call will not be closed until they are released by their users. Multiples calls to this method per pool do
      * no harm. Requests for workspaces that occur after the pool that would service the request has been closed
      * result in the pool being reinitialized. Exceptions that occur while closing workspace objects will be
      * reported in the log but will not result in an exception being thrown from this method.
      * @param credentials information that identifies a JNDI directory and specifies how to bind to it. Behavior is undefined
      * if the value of this parameter is <CODE>null</CODE>.
      * @param searchContext the starting point in the JNDI directory for searches. Behavior is undefined
      * if the value of this parameter is <CODE>null</CODE>.
      * @param sasLogicalName a logical name that identifies a pool or a
      * set of pools. Behavior is undefined
      * if the value of this parameter is <CODE>null</CODE>.
      * @param userDN the distinguished name of a user or group of users that is used to determine which, if any, of the
      * the pools identified by <CODE>sasLogicalName</CODE> should be closed. If the value is
      * <CODE>null</CODE>, then all of the pools identified by <CODE>sasLogicalName</CODE> will be closed.
      * @exception com.sas.iom.WorkspacePoolException if the given logical name and user DN do not identify any
      * workspace pools or if the given credentials do not provide sufficient authority
      * to access the workspace pool information in JNDI.
      */
     public static void closePool(Hashtable credentials,String searchContext,String sasLogicalName,String userDN) throws WorkspacePoolException
     {
          String password = findPassword(credentials,userDN);

          try
          {
               WorkspaceFactory wf = WorkspaceFactory.getFactory(userDN,password,credentials,searchContext,sasLogicalName);

               if (wf != null)
               {
                    wf.shutdown();
               }
          }
          catch (WorkspaceFactoryException wfe)
          {
               throw new WorkspacePoolException(wfe);
          }
     }

     /**
      * Messages about all pools operating under this driver will be printed on this writer.
      * @param log writer to print messages to.
      */
     public static void setLog(PrintWriter l)
     {
          logWriter = l;
     }

     /**
      * Messages about all pools operating under this driver are printed on this writer.
      * @return writer to print messages to.
      */
     public static PrintWriter getLog()
     {
          return logWriter;
     }

     private static String findPassword(Hashtable credentials,String userDN) throws WorkspacePoolException
     {
          if (userDN == null)
          {
               return null;
          }

          String password = null;

          try
          {
               DirContext ctx = new InitialDirContext(credentials);
               Attributes attrs = ctx.getAttributes(userDN);
               Attribute attr = attrs.get("Password");
               password = (String)attr.get();
               ctx.close();
          }
          catch (NamingException ne)
          {
               throw new WorkspacePoolException(ne);
          }

          return password;
     }

     private static PrintWriter logWriter = null;
}