package com.sas.lang;

import java.security.PrivilegedAction;
import com.sas.security.BaseSecuritySupport;
//
// History:
//
// Date    Name            Remarks
// 27Feb02 B Beers         S0141718: Move to sas.core.jar.
//
/**
 * This class provides a solution to the classic readers-writers thread
 * synchronization problem. In this problem, several threads compete
 * for possesion of a lock simultaneously. Some threads
 * (readers) are allowed non-exclusive possesion while others (writers) are
 * allowed only exclusive possesion. The problem is named as it is because
 * several threads can safely read a file or record cuncurrently, but
 * only one thread can safely write on a file or record at a time.
 * <P>
 * This implementation solves two variants of the readers-writers
 * problem. In the first variant, readers are allowed immediate
 * access to the critical section as long as no writer currently
 * has it locked even if writers are waiting. In the second variant,
 * no reader will be allowed access to the critical section while
 * a writer is waiting for access. Note that, in the first variant,
 * writers may starve (i.e. never gain possesion of the lock)
 * while, in the second variant, readers may starve.
 * <P>
 * This class is intended to behave as much as possible like
 * the Java language synchronized contruct. The same care
 * that a programmer must take to avoid deadlocks for synchronized
 * code must be taken here.
 * @author Biff Beers
 */
public abstract class NonExclusiveLock
{
     /**
      * Flag which indicates that requests for non-exclusive possesion
      * of the lock will <STRONG>not</STRONG> be granted while a
      * request for exclusive possesion is pending.
      */
     public static final int PRIORITY_EXCLUSIVE = 0;

     /**
      * Flag which indicates that requests for exclusive possesion
      * of the lock will <STRONG>not</STRONG> be granted while a
      * request for non-exclusive possesion is pending.
      */
     public static final int PRIORITY_NON_EXCLUSIVE = 1;

     /**
      * Get a NonExclusiveLock instance. This has the same effect
      * and <CODE>getLock(PRIORITY_EXCLUSIVE)</CODE>.
      */
     public static NonExclusiveLock getInstance()
     {
          return getInstance(PRIORITY_EXCLUSIVE);
     }

     /**
      * Get a NonExclusiveLock instance with the specified priority variant.
      * @param priorityVariant the priority variant
      * @see NonExclusiveLock#PRIORITY_EXCLUSIVE
      * @see NonExclusiveLock#PRIORITY_NON_EXCLUSIVE
      */
     public static NonExclusiveLock getInstance(int priorityVariant)
     {
          switch (priorityVariant)
          {
               case PRIORITY_EXCLUSIVE:     return new NonExclusiveLock.PriorityExclusive();
               case PRIORITY_NON_EXCLUSIVE: return new NonExclusiveLock.PriorityNonExclusive();
               default:throw new IllegalArgumentException(Integer.toString(priorityVariant));
          }
     }

     /**
      * Get the priority flag.
      * @return the priority flag
      * @see NonExclusiveLock#PRIORITY_EXCLUSIVE
      * @see NonExclusiveLock#PRIORITY_NON_EXCLUSIVE
      */
     public abstract int getPriority();

     /**
      * Obtain exclusive possesion of this lock and execute the specified task.
      */
     public abstract void executeExclusive(Runnable task);

     /**
      * Obtain non-exclusive possesion of this lock and execute the specified task.
      */
     public abstract void executeNonExclusive(Runnable task);

     protected int readersWaiting = 0;
     protected int readersExecuting = 0;
     protected int writersWaiting = 0;
     protected int writersExecuting = 0; // should be 0 or 1
     protected int[] lock = new int[0];

     protected static void execute(Runnable task,boolean interrupted)
     {
          if (interrupted)
          {
               // If a thread is interrupted while it waits on a
               // synchronization lock, it just keeps waiting.
               // However, the interrupted flag is set for the
               // thread so it needs to be set before executing
               // the caller's task.
               //-----------------------------------------------
               PrivilegedAction action = new PrivilegedAction()
               {
                    public Object run()
                    {
                         Thread.currentThread().interrupt();
                         return null;
                    }
               };

               BaseSecuritySupport.securitySupport.doThreadPrivileged(action);
          }

          task.run();
     }

     // implement the execute* methods with writers getting priority
     //-------------------------------------------------------------
     private static class PriorityExclusive extends NonExclusiveLock
     {
          public int getPriority()
          {
               return PRIORITY_EXCLUSIVE;
          }

          public void executeExclusive(Runnable task)
          {
               boolean interrupted = false;

               synchronized (lock)
               {
                    writersWaiting++;

                    while (writersExecuting > 0 || readersExecuting > 0)
                    {
                         try
                         {
                              lock.wait();
                         }
                         catch (InterruptedException e)
                         {
                              interrupted = true;
                         }
                    }

                    writersWaiting--;
                    writersExecuting++;
               }

               try
               {
                    execute(task,interrupted);
               }
               finally
               {
                    synchronized (lock)
                    {
                         writersExecuting--;
                         lock.notifyAll();
                    }
               }
          }

          public void executeNonExclusive(Runnable task)
          {
               boolean interrupted = false;

               synchronized (lock)
               {
                    readersWaiting++;

                    while (writersWaiting > 0 || writersExecuting > 0)
                    {
                         try
                         {
                              lock.wait();
                         }
                         catch (InterruptedException e)
                         {
                              interrupted = true;
                         }
                    }

                    readersWaiting--;
                    readersExecuting++;
               }

               try
               {
                    execute(task,interrupted);
               }
               finally
               {
                    synchronized (lock)
                    {
                         readersExecuting--;
                         lock.notifyAll();
                    }
               }
          }
     }

     // implement the execute* methods with readers getting priority
     //-------------------------------------------------------------
     private static class PriorityNonExclusive extends NonExclusiveLock
     {
          public int getPriority()
          {
               return PRIORITY_NON_EXCLUSIVE;
          }

          public void executeExclusive(Runnable task)
          {
               boolean interrupted = false;

               synchronized (lock)
               {
                    writersWaiting++;

                    while (writersExecuting > 0 || readersExecuting > 0 || readersWaiting > 0)
                    {
                         try
                         {
                              lock.wait();
                         }
                         catch (InterruptedException e)
                         {
                              interrupted = true;
                         }
                    }

                    writersWaiting--;
                    writersExecuting++;
               }

               try
               {
                    execute(task,interrupted);
               }
               finally
               {
                    synchronized (lock)
                    {
                         writersExecuting--;
                         lock.notifyAll();
                    }
               }
          }

          public void executeNonExclusive(Runnable task)
          {
               boolean interrupted = false;

               synchronized (lock)
               {
                    readersWaiting++;

                    while (writersExecuting > 0)
                    {
                         try
                         {
                              lock.wait();
                         }
                         catch (InterruptedException e)
                         {
                              interrupted = true;
                         }
                    }

                    readersWaiting--;
                    readersExecuting++;
               }

               try
               {
                    execute(task,interrupted);
               }
               finally
               {
                    synchronized (lock)
                    {
                         readersExecuting--;
                         lock.notifyAll();
                    }
               }
          }
     }
}