// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513
package com.sas.lang;

import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.List;
import com.sas.security.BaseSecuritySupport;
//
// History:
//
// Date    Name            Remarks
// 24Nov97 B Beers         Created.
// 23Dec97 B Beers         Adapted for SimpleCommand interface.
// 10Dec98 B Beers         Implemented bank teller queue model.
// 22Dec98 B Beers         Moved from com.sas.util.
// 29Jan99 B Beers         Made service threads protected.
// 14Dec99 B Beers         S0072009: Remove spurious stack traces.
// 06Sep00 B Beers         S0094267: Enclose thread creation in privileged action blocks.
// 27Feb01 B Beers         S0110393: Add bind and server flow features.
// 17Aug01 B Beers         S0124013: Enclose thread.setDaemon in privileged action blocks.
// 27Feb02 B Beers         S0141718: Move to sas.core.jar.
// 09Apr02 B Beers         S0146665: Fix NullPointerException with null thread name.
// 10Apr02 B Beers         S0146908: Restored daemon threads.
//
/**
 * RunnableQueue implements a bank teller queue model. There is one
 * line of bank customers (each implementing 
 * <CODE>java.lang.Runnable</CODE>)
 * and 1 or more tellers (or service threads waiting to call
 * <CODE>run()</CODE> on the <CODE>Runnable</CODE>
 * interface. Whenever a teller is available, it serves the
 * first customer in line.
 * @author Biff Beers
 */
public class RunnableQueue implements Runnable
{
     // constructors
     //-------------------------------------------------------------
     /**
      * Construct a RunnableQueue with the {@link RunnableQueue#defaultCount default}
      * number of service threads.
      * This constructor has the same effect as
      * <CODE>RunnableQueue(null,null,defaultCount)</CODE>.
      */
     public RunnableQueue()
     {
          this(null,null,defaultCount);
     }

     /**
      * Construct a RunnableQueue with
      * <PARAM>count</PARAM> service threads.
      * This constructor has the same effect as
      * <CODE>RunnableQueue(null,null,<PARAM>count</PARAM>)</CODE>.
      * @param count the number of service threads
      */
     public RunnableQueue(int count)
     {
          this(null,null,count);
     }

     /**
      * Construct a RunnableQueue with
      * the {@link RunnableQueue#defaultCount default} number of named service threads.
      * <P>
      * Each service thread will have a name based on <PARAM>name</PARAM>.
      * If <PARAM>name</PARAM> is <CODE>null</CODE>, then a name will be
      * generated for each service thread.
      * <P>
      * This constructor has the same effect as
      * <CODE>RunnableQueue(null,<PARAM>name</PARAM>,defaultCount)</CODE>.
      * @param name the base name for the service threads
      */
     public RunnableQueue(String name)
     {
          this(null,name,defaultCount);
     }

     /**
      * Construct a RunnableQueue with
      * <PARAM>count</PARAM> named service threads.
      * <P>
      * Each service thread will have a name based on <PARAM>name</PARAM>.
      * If <PARAM>name</PARAM> is <CODE>null</CODE>, then a name will be
      * generated for each service thread.
      * <P>
      * This constructor has the same effect as
      * <CODE>RunnableQueue(null,<PARAM>name</PARAM>,<PARAM>count</PARAM>)</CODE>.
      * @param name the base name for the service threads
      * @param count the number of service threads
      */
     public RunnableQueue(String name,int count)
     {
          this(null,name,count);
     }

     /**
      * Construct a RunnableQueue with
      * the {@link RunnableQueue#defaultCount default} number of grouped service threads.
      * <P>
      * Each service thread be a member of the group specified with <PARAM>group</PARAM>.
      * If <PARAM>group</PARAM> is <CODE>null</CODE>, then the current
      * thread's thread group will be used.
      * <P>
      * This constructor has the same effect as
      * <CODE>RunnableQueue(<PARAM>group</PARAM>,null,defaultCount)</CODE>.
      * @param group the thread group to contain the service threads
      */
     public RunnableQueue(ThreadGroup group)
     {
          this(group,null,defaultCount);
     }

     /**
      * Construct a RunnableQueue with the {@link RunnableQueue#defaultCount default}
      * number of named and grouped service threads.
      * <P>
      * Each service thread be a member of the group specified with <PARAM>group</PARAM>.
      * If <PARAM>group</PARAM> is <CODE>null</CODE>, then the current
      * thread's thread group will be used.
      * <P>
      * This constructor has the same effect as
      * <CODE>RunnableQueue(<PARAM>group</PARAM>,null,<PARAM>count</PARAM>)</CODE>.
      * @param group the thread group to contain the service threads
      * @param count the number of service threads
      */
     public RunnableQueue(ThreadGroup group,int count)
     {
          this(group,null,count);
     }

     /**
      * Construct a RunnableQueue with
      * the {@link RunnableQueue#defaultCount default} number of named and grouped service threads.
      * <P>
      * Each service thread will have a name based on <PARAM>name</PARAM>.
      * If <PARAM>name</PARAM> is <CODE>null</CODE>, then a name will be
      * generated for each service thread.
      * <P>
      * Each service thread be a member of the group specified with <PARAM>group</PARAM>.
      * If <PARAM>group</PARAM> is <CODE>null</CODE>, then the current
      * thread's thread group will be used.
      * <P>
      * This constructor has the same effect as
      * <CODE>RunnableQueue(<PARAM>group</PARAM>,<PARAM>name</PARAM>,defaultCount)</CODE>.
      * @param group the thread group to contain the service threads
      * @param name the base name for the service threads
      */
     public RunnableQueue(ThreadGroup group,String name)
     {
          this(group,name,defaultCount);
     }

     /**
      * Construct a RunnableQueue with
      * <PARAM>count</PARAM> named and grouped service threads.
      * <P>
      * Each service thread will have a name based on <PARAM>name</PARAM>.
      * If <PARAM>name</PARAM> is <CODE>null</CODE>, then a name will be
      * generated for each service thread.
      * <P>
      * Each service thread be a member of the group specified with <PARAM>group</PARAM>.
      * If <PARAM>group</PARAM> is <CODE>null</CODE>, then the current
      * thread's thread group will be used.
      * @param group the thread group to contain the service threads
      * @param name the base name for the service threads
      * @param count the number of service threads (must be >= 1)
      */
     public RunnableQueue(ThreadGroup group,String name,int count)
     {
          if (count < 1)
          {
               throw new IllegalArgumentException();
          }

          if (group == null)
          {
               group = Thread.currentThread().getThreadGroup();
          }

          this.group = group;
          taskList = new ArrayList();
          serviceThreads = new Thread[count];
          StringBuffer nameBuffer = new StringBuffer();
          BaseSecuritySupport securitySupport = BaseSecuritySupport.securitySupport;

          for (int i=0;i<count;i++)
          {
               // need modifyThreadGroup permission to create a thread
               final Thread serviceThread = securitySupport.createThread(this);
               serviceThreads[i] = serviceThread;
               final String threadName = (name == null) ? null : nameBuffer.append(name).append(" ").append(i).toString();/*I18nOK:LINE*/
               nameBuffer.setLength(0);

               // need modifyThread permission to change thread attributes
               PrivilegedAction action = new PrivilegedAction()
               {
                    public Object run()
                    {
                         if (threadName != null)
                         {
                              serviceThread.setName(threadName);
                         }

                         serviceThread.setDaemon(true);
                         return null;
                    }
               };

               securitySupport.doThreadPrivileged(action);
               serviceThread.start();
          }

          threadCensus = new int[1];
          threadCensus[0] = count;

          shutDownTask = new Runnable()
          {
               public void run()
               {
                    shutDown = true;
               }
          };
     }

     // implement Runnable
     //-------------------------------------------------------------
     /**
      * Implements java.lang.Runnable. Used by java.lang.Thread
      */
     public void run()
     {
          Runnable task = null;
          boolean done = false;

          do
          {
               synchronized (taskList)
               {
                    try
                    {
                         while (taskList.size() == 0 && !shutDown)
                         {
                              taskList.wait();
                         }

                         if (!shutDown)
                         {
                              task = (Runnable)taskList.remove(0);
                         }
                    }
                    catch (InterruptedException e)
                    {
                    }
               }

               if (task != null)
               {
                    try
                    {
                         task.run();
                    }
                    catch (ThreadDeath td)
                    {
                         done = true;
                    }
                    catch (Throwable t)
                    {
                         group.uncaughtException(Thread.currentThread(),t);
                    }
                    finally
                    {
                         task = null;
                    }
               }
               else
               {
                    done = true;
               }
          }
          while (!done);

          synchronized (threadCensus)
          {
               threadCensus[0]--;
               threadCensus.notifyAll();
          }
     }
     
     // public stuff
     //-------------------------------------------------------------
     /**
      * Default number of service threads.
      */
     public static final int defaultCount = 1;

     /**
      * Add a task to execute.
      * @param task a task to execute
      * @return <CODE>true</CODE> if the task will execute or
      * <CODE>false</CODE> if the RunnableQueue has already been shutdown
      */
     public boolean enqueueTask(Runnable task)
     {
          if (isShutDown())
          {
               return false;
          }
          
          synchronized (taskList)
          {
               taskList.add(task);
               taskList.notifyAll();
          }

          return true;
     }

     /**
      * Shut down the RunnableQueue. This is the same as calling
      * <CODE>shutDown(true,false)</CODE>.
      */
     public void shutDown()
     {
          shutDown(true,false);
     }

     /**
      * Shut down the RunnableQueue. This allows the currently executing commands
      * to finish, but does
      * not execute any queued up commands that have not yet begun to execute.
      * This is the same as calling
      * <CODE>shutDown(<PARAM>waitForCompletion</PARAM>,false)</CODE>.
      * @param waitForCompletion if true, then this method will not return until
      * all currently executing commands have completed. Otherwise, this method will
      * return immediately even though some commands may still be executing.
      */
     public void shutDown(boolean waitForCompletion)
     {
          shutDown(waitForCompletion,false);
     }

     /**
      * Shut down the command queue. This allows the currently executing commands
      * to finish, and, if <PARAM>executePendingCommands</PARAM> is true, then it allows any queued
      * up commands to execute.
      * @param waitForCompletion if true, then this method will not return until
      * all currently executing and, if executePendingCommands is true, all pending 
      * commands have completed. Otherwise, this method will
      * return immediately even though some commands may still be executing.
      * @param executePendingCommands if true, then this method will execute all
      * commands currently in the queue before shutting down. Otherwise, only the
      * currently executing commands will complete before the queue is shut down.
      */
     public void shutDown(boolean waitForCompletion,boolean executePendingCommands)
     {
          if (executePendingCommands)
          {
               enqueueTask(shutDownTask);
          }
          else
          {
               synchronized (taskList)
               {
                    shutDown = true;
                    taskList.notifyAll();
               }
          }

          if (waitForCompletion)
          {
               try
               {
                    synchronized (threadCensus)
                    {
                         while (threadCensus[0] > 0)
                         {
                              threadCensus.wait();
                         }
                    }
               }
               catch (InterruptedException e)
               {
               }
          }
     }

     /**
      * Set the daemon attribute of all the service threads.
      * @param daemon the value for the daemon attribute of each service thread
      */
     public void setDaemon(final boolean daemon)
     {
          for (int i=0;i<serviceThreads.length;i++)
          {
               final Thread serviceThread = serviceThreads[i];

               PrivilegedAction action = new PrivilegedAction()
               {
                    public Object run()
                    {
                         serviceThread.setDaemon(daemon);
                         return null;
                    }
               };

               BaseSecuritySupport.securitySupport.doThreadPrivileged(action);
          }
     }

     // internal stuff
     //-------------------------------------------------------------
     private ThreadGroup group = null;
     private Thread[] serviceThreads = null;
     private int[] threadCensus = null;
     private List taskList = null;
     private boolean shutDown = false;
     private Runnable shutDownTask = null;

     private boolean isShutDown()
     {
          synchronized (taskList)
          {
               return shutDown || taskList.contains(shutDownTask);
          }
     }
}