// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.measures;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Toolkit;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.awt.GraphicsEnvironment;

import com.sas.text.Message;

/**
 * <p>
 * <code>Distance</code> provides an object which represents a linear measurement in one
 * of several different units.  The units are broken up into three types:
 * <code>{@link #FIXED_UNITS}</code>, <code>{@link #FONT_UNITS}</code> and <code>{@link #DEVICE_UNITS}</code>.
 * Currently only <code>FIXED_UNITS</code> and <code>DEVICE_UNITS</code> units are supported.
 * </p>
 * The units, grouped by type with full name followed by abbreviation, are as follows:
 * <br>
 * <ul>
 * <li><code>{@link #FIXED_UNITS}</code>
 * <ul>
 * <li> angstrom - A
 * <li> nanometer - nm
 * <li> micrometer - um
 * <li> millimeter - mm
 * <li> point - pt
 * <li> pica - pica
 * <li> centimeter - cm
 * <li> decimeter - dm
 * <li> meter - m
 * <li> dekameter - dam
 * <li> hectometer - nm
 * <li> kilometer - km
 * <li> inch - in
 * <li> foot - ft
 * <li> yard - yd
 * <li> rod - rod
 * <li> furlong - fur
 * <li> mile - mi
 * <li> nautical mile - nmi
 * <li> league - lg
 * <li> link - li
 * <li> chain - ch
 * <li> astronomical unit - au
 * <li> parsec - pc
 * </ul>
 * <li><code>{@link #FONT_UNITS}</code>
 * <ul>
 * <li> l-space - el
 * <li> m-space - em
 * <li> n-space - en
 * <li> x-space - ex
 * <li> figure width - fg
 * <li> space - sp
 * <li> character cell - cc
 * <li> height - ht (height of tallest character in font)
 * <li> depth - dp (depth of deepest character in font)
 * <li> line - ln (line space based on font size)
 * </ul>
 * <li><code>{@link #DEVICE_UNITS}</code>
 * <ul>
 * <li> x-pixel - px
 * <li> y-pixel - py
 * </ul>
 * </ul>
 *
 * @author Ryan Norris
 * @version 1.0
 */
public class BaseDistance extends Measure
{
  public static final String RB_KEY = "Distance.";

   static final long serialVersionUID = 103288383102555192L;

   //////////////////////////////////////////////////////////////////////////////////
   // Units conversion constants ...
   //////////////////////////////////////////////////////////////////////////////////

   private static final double INCHES_PER_METER  = 39.37;
   private static final double INCHES_PER_FOOT   = 12.0;
   private static final double INCHES_PER_POINT  = (1.0 / 72.27);

   private static final double MM_PER_METER      = 1000.0;
   private static final double MM_PER_POINT      = ((INCHES_PER_POINT / INCHES_PER_METER) * MM_PER_METER);
   private static final double MM_PER_ANGSTROM   = 0.0000001;
   private static final double MM_PER_NM         = 0.000001;
   private static final double MM_PER_UM         = 0.001;
   private static final double MM_PER_PICA       = (MM_PER_POINT * 12.0);

   private static final double METER_PER_MM      = 0.001;
   private static final double METER_PER_CM      = 0.01;
   private static final double METER_PER_DM      = 0.1;
   private static final double METER_PER_DAM     = 10.0;
   private static final double METER_PER_HM      = 100.0;
   private static final double METER_PER_KM      = 1000.0;

   private static final double FEET_PER_INCH     = (1.0 / INCHES_PER_FOOT);
   private static final double FEET_PER_YARD     = 3.0;
   private static final double FEET_PER_ROD      = 16.5;
   private static final double FEET_PER_FURLONG  = 660.0;
   private static final double FEET_PER_MILE     = 5280.0;
   private static final double FEET_PER_NMILE    = 6076.11549;
   private static final double FEET_PER_LEAGUE   = 15840.0;
   private static final double FEET_PER_LINK     = 0.66;
   private static final double FEET_PER_CHAIN    = 66.0;

   private static final double MILES_PER_AU      = 92900000.0;
   private static final double AU_PER_PARSEC     = 206265.0;

   private static final double FEET_PER_METER    = (INCHES_PER_METER / INCHES_PER_FOOT);
   private static final double METER_PER_FOOT    = (INCHES_PER_FOOT / INCHES_PER_METER);

   private static final double FEET_PER_MM       = (FEET_PER_METER * METER_PER_MM);
   private static final double MM_PER_FOOT       = (MM_PER_METER * METER_PER_FOOT);

   private static final double FEET_PER_AU       = (FEET_PER_MILE * MILES_PER_AU);
   private static final double METER_PER_AU      = (METER_PER_FOOT * FEET_PER_AU);
   private static final double MM_PER_AU         = (MM_PER_METER * METER_PER_AU);

   private static final double MILES_PER_FOOT    = (1.0 / FEET_PER_MILE);
   private static final double MILES_PER_METER   = (FEET_PER_METER / FEET_PER_MILE);
   private static final double AU_PER_FOOT       = (MILES_PER_FOOT / MILES_PER_AU);
   private static final double AU_PER_METER      = (MILES_PER_METER / MILES_PER_AU);
   private static final double AU_PER_MM         = (AU_PER_METER * MM_PER_METER);

   private static final double DEFAULT_DPI       = 96.0;

   //////////////////////////////////////////////////////////////////////////////////
   // Supported units constants ...
   //////////////////////////////////////////////////////////////////////////////////

   private static final int UM_ANGSTROM          = 0;   /* angstrom   (.0000001mm) 10^-10 meters */
   private static final int UM_NANOMETER         = 1;   /* nanometer  (.000001mm)  10^-9 meters */
   private static final int UM_MICROMETER        = 2;   /* micrometer (.001mm) 10^-16 meters */
   private static final int UM_MILLIMETER        = 3;   /* millimeter (.001m) 10^-3 meters */
   private static final int UM_POINT             = 4;   /* point (.013837in, .3514598mm) */
   private static final int UM_PICA              = 5;   /* pica  (12pt, 4.2175176mm) */

   private static final int UM_MILLIMETERS  = UM_PICA;   /* store in terms of millimeters */

   private static final int UM_CENTIMETER        = 6;   /* centimeter (.01m) */
   private static final int UM_DECIMETER         = 7;   /* decimeter (.1m) */
   private static final int UM_METER             = 8;   /* meter (39.37in) */
   private static final int UM_DEKAMETER         = 9;   /* dekameter   (10m) */
   private static final int UM_HECTOMETER        = 10;  /* hectometer (100m) */
   private static final int UM_KILOMETER         = 11;  /* kilometer (1000m) */

   private static final int UM_METERS  = UM_KILOMETER;  /* store in terms of meters */

   private static final int UM_INCH              = 12;  /* inch (1/12 ft) */
   private static final int UM_F00T              = 13;  /* foot (39.37/12 m) */
   private static final int UM_YARD              = 14;  /* yard (3ft) */
   private static final int UM_ROD               = 15;  /* rod (5.5yd, 16.5ft) */
   private static final int UM_FUR               = 16;  /* furlong (40rd, 660ft) */
   private static final int UM_MILE              = 17;  /* statute or land mile (5,280ft) */
   private static final int UM_NAUTICAL_MILE     = 18;  /* international nautical mile (6,076.11549ft) */
   private static final int UM_LEAGUE            = 19;  /* league (3mi, 15,840ft) */
   private static final int UM_LINK              = 20;  /* link (7.92in, 0.66ft) */
   private static final int UM_CHAIN             = 21;  /* chain (100li, 66ft) */

   private static final int UM_FEET   = UM_CHAIN;  /* store in terms of feet */

   private static final int UM_ASTRONOMICAL_UNIT = 22;  /* astronomical unit (92.9 x 10^6 mi) */
   private static final int UM_PARSEC            = 23;  /* parsec (206,265AU) */

   private static final int UM_AUS    = UM_PARSEC;  /* store in terms of AUs */

   private static final int FIXED  = UM_PARSEC;  /* fixed units of measure above here */

   private static final int UM_L_SPACE           = 24;  /* l-space */
   private static final int UM_M_SPACE           = 25;  /* m-space */
   private static final int UM_N_SPACE           = 26;  /* n-space */
   private static final int UM_X_HEIGHT          = 27;  /* x-height */
   private static final int UM_FIGURE_WIDTH      = 28;  /* figure width */
   private static final int UM_SPACE             = 29;  /* space */
   private static final int UM_CHARACTER_CELL    = 30;  /* character cell (width of widest character in font) */
   private static final int UM_HT     = 31;  /* height of tallest character in font */
   private static final int UM_DP     = 32;  /* depth of deepest character in font */
   private static final int UM_LN     = 33;  /* line space (font size) */

   private static final int FONT   = UM_LN ;  /* font units: store in terms of decipoints */

   private static final int UM_X_PIXEL           = 34;  /* x-pixels */
   private static final int UM_Y_PIXEL           = 35;  /* y-pixels */

   private static final int DEVICE = UM_Y_PIXEL;  /* device units */

   /**
    * Unit of measure is fixed.
    */
   public static int FIXED_UNITS  = 0x00000001;
   /**
    * Unit of measure is based on a font.
    */
   public static int FONT_UNITS   = 0x00000002;

   // unused 0x00000004;

   /**
    * Unit of measure is based on a device.
    */
   public static int DEVICE_UNITS = 0x00000008;


   //////////////////////////////////////////////////////////////////////////////////
   // Units normalization table ...
   //////////////////////////////////////////////////////////////////////////////////

   /* ordered by id */
   private final static UnitInfo[] dunits =
   {
      new UnitInfo("A",     UM_ANGSTROM,           MM_PER_ANGSTROM,   RB.getStringResource(RB_KEY, "A.txt")),
      new UnitInfo("nm",    UM_NANOMETER,          MM_PER_NM,         RB.getStringResource(RB_KEY, "nm.txt")),
      new UnitInfo("um",    UM_MICROMETER,         MM_PER_UM,         RB.getStringResource(RB_KEY, "um.txt")),
      new UnitInfo("mm",    UM_MILLIMETER,         1.0,               RB.getStringResource(RB_KEY, "mm.txt")),
      new UnitInfo("pt",    UM_POINT,              MM_PER_POINT,      RB.getStringResource(RB_KEY, "pt.txt")),
      new UnitInfo("pica",  UM_PICA,               MM_PER_PICA,       RB.getStringResource(RB_KEY, "pica.txt")),

      // UM_MILLIMETERS

      new UnitInfo("cm",    UM_CENTIMETER,         METER_PER_CM,      RB.getStringResource(RB_KEY, "cm.txt")),
      new UnitInfo("dm",    UM_DECIMETER,          METER_PER_DM,      RB.getStringResource(RB_KEY, "dm.txt")),
      new UnitInfo("m",     UM_METER,              1.0,               RB.getStringResource(RB_KEY, "m.txt")),
      new UnitInfo("dam",   UM_DEKAMETER,          METER_PER_DAM,     RB.getStringResource(RB_KEY, "dam.txt")),
      new UnitInfo("hm",    UM_HECTOMETER,         METER_PER_HM,      RB.getStringResource(RB_KEY, "hm.txt")),
      new UnitInfo("km",    UM_KILOMETER,          METER_PER_KM,      RB.getStringResource(RB_KEY, "km.txt")),

      // UM_METERS

      new UnitInfo("in",    UM_INCH,               FEET_PER_INCH,     RB.getStringResource(RB_KEY, "in.txt")),
      new UnitInfo("ft",    UM_F00T,               1.0,               RB.getStringResource(RB_KEY, "ft.txt")),
      new UnitInfo("yd",    UM_YARD,               FEET_PER_YARD,     RB.getStringResource(RB_KEY, "yd.txt")),
      new UnitInfo("rod",   UM_ROD,                FEET_PER_ROD,      RB.getStringResource(RB_KEY, "rod.txt")),
      new UnitInfo("fur",   UM_FUR,                FEET_PER_FURLONG,  RB.getStringResource(RB_KEY, "fur.txt")),
      new UnitInfo("mi",    UM_MILE,               FEET_PER_MILE,     RB.getStringResource(RB_KEY, "mi.txt")),
      new UnitInfo("nmi",   UM_NAUTICAL_MILE,      FEET_PER_NMILE,    RB.getStringResource(RB_KEY, "nmi.txt")),
      new UnitInfo("lg",    UM_LEAGUE,             FEET_PER_LEAGUE,   RB.getStringResource(RB_KEY, "lg.txt")),
      new UnitInfo("li",    UM_LINK,               FEET_PER_LINK,     RB.getStringResource(RB_KEY, "li.txt")),
      new UnitInfo("ch",    UM_CHAIN,              FEET_PER_CHAIN,    RB.getStringResource(RB_KEY, "ch.txt")),

      // UM_FEET

      new UnitInfo("au",    UM_ASTRONOMICAL_UNIT,  1.0,               RB.getStringResource(RB_KEY, "au.txt")),
      new UnitInfo("pc",    UM_PARSEC,             AU_PER_PARSEC,     RB.getStringResource(RB_KEY, "pc.txt")),

      // UM_AUS




      // FIXED

      new UnitInfo("el",    UM_L_SPACE,            0.0,               RB.getStringResource(RB_KEY, "el.txt")),
      new UnitInfo("em",    UM_M_SPACE,            0.0,               RB.getStringResource(RB_KEY, "em.txt")),
      new UnitInfo("en",    UM_N_SPACE,            0.0,               RB.getStringResource(RB_KEY, "en.txt")),
      new UnitInfo("ex",    UM_X_HEIGHT,           0.0,               RB.getStringResource(RB_KEY, "ex.txt")),
      new UnitInfo("fg",    UM_FIGURE_WIDTH,       0.0,               RB.getStringResource(RB_KEY, "fg.txt")),
      new UnitInfo("sp",    UM_SPACE,              0.0,               RB.getStringResource(RB_KEY, "sp.txt")),
      new UnitInfo("cc",    UM_CHARACTER_CELL,     0.0,               RB.getStringResource(RB_KEY, "cc.txt")),
      new UnitInfo("ht",    UM_HT,                 0.0,               RB.getStringResource(RB_KEY, "ht.txt")),
      new UnitInfo("dp",    UM_DP,                 0.0,               RB.getStringResource(RB_KEY, "dp.txt")),
      new UnitInfo("ln",    UM_LN,                 0.0,               RB.getStringResource(RB_KEY, "ln.txt")),

	  // DEVICE UNITS

      // UM_LN

      new UnitInfo("px",    UM_X_PIXEL,            1.0,               RB.getStringResource(RB_KEY, "px.txt")),
      new UnitInfo("py",    UM_Y_PIXEL,            1.0,               RB.getStringResource(RB_KEY, "py.txt")),
   };

   // ordering by most commonly used units, for faster lookup in parseUnit
   private static int mostCommonlyUsed[] =
   {
      UM_MILLIMETER,
      UM_CENTIMETER,
      UM_METER,
      UM_INCH,
      UM_F00T,
      UM_X_PIXEL,
      UM_Y_PIXEL,
      UM_M_SPACE,
      UM_N_SPACE,
      UM_POINT,
      UM_PICA,
      UM_SPACE,
      UM_L_SPACE,
      UM_CHARACTER_CELL,
      UM_HT,
      UM_LN,
      UM_FIGURE_WIDTH,
      UM_DP,
      UM_X_HEIGHT,
      UM_KILOMETER,
      UM_DECIMETER,
      UM_MICROMETER,
      UM_MILE,
      UM_YARD,
      UM_ASTRONOMICAL_UNIT,
      UM_PARSEC,
      UM_NAUTICAL_MILE,
      UM_DEKAMETER,
      UM_HECTOMETER,
      UM_ANGSTROM,
      UM_NANOMETER,
      UM_LEAGUE,
      UM_LINK,
      UM_CHAIN,
      UM_ROD,
      UM_FUR,
   };

   // index into table above for initialize unit
   protected int unitIndex;

   /**
    * Constructs a <code>Distance</code> object from the specified magnitude and unit of measure.
    * @param magnitude numeric value of the measurement
    * @param unit unit of measure
    * @exception IllegalArgumentException if the unit is not a supported unit of measure
    */
   public BaseDistance(double magnitude, String unit)
   {
      super(magnitude, unit);
   }

   /**
    * Constructs a <code>Distance</code> object from a string consisting of
    * a numeric value and unit of measure.
    * @param text a string consisting of a number,
    * some whitespace, and a supported units token, such as
    * <code>"10 mm"</code> or <code>"2.5e3 nm"</code>
    * @param text string containing the magnitude and unit of measure
    * @exception IllegalArgumentException if the unit is not a supported unit of measure
    */
   public BaseDistance(String text)
   {
      super(text);
   }

   //////////////////////////////////////////////////////////////////////////////////
   // Public methods ...
   //////////////////////////////////////////////////////////////////////////////////

   /**
    * Returns an equivalent magnitude of the current measurement in terms of the
    * specified unit of measure. For example, to convert from kilometers
    * to millimeters, you can use
    * <code>new BaseDistance(1, "km").getMagnitude("mm")</code>
    * @param unit the unit of measure to convert to
    * @return the value of the current magnitude converted into the specified unit of measure
    * @exception IllegalArgumentException if the unit is not a supported unit of measure
    */
   public double getMagnitude(Object unit)
   {
      int du_index = parseUnit(unit);
      if (du_index == -1)
         throw new IllegalArgumentException(Message.format(RB.getStringResource("invalidUnit.ex.txt"), unit));

      return internalGetEquivalent(dunits[unitIndex], dunits[du_index], magnitude, null);
   }

   /**
    * Returns an equivalent magnitude of the current measurement in terms of the
    * specified unit of measure.  This version of <code>getMagnitude</code> is to be used
    * when one of the 2 units of measure is of type <code>{@link #FONT_UNITS}</code>.
    * @param unit the unit of measure to get the equivalent of
    * @param font the Font object to measure with
    * @return the value of the current magnitude converted into the specified unit of measure
    * @exception IllegalArgumentException if the unit is not a supported unit of measure
    */
   public double getMagnitude(Object unit, Font font)
   {
      int du_index = parseUnit(unit);
      if (du_index == -1)
         throw new IllegalArgumentException(Message.format(RB.getStringResource("invalidUnit.ex.txt"), unit));

      System.out.println(RB.getStringResource(RB_KEY, "methodNotImplemented"));
      return internalGetEquivalent(dunits[unitIndex], dunits[du_index], magnitude, font);
   }

   /**
  * Returns the abbreviated name of the unit property.  For example, <code>"in"</code> for inch
  * and <code>"cm"</code> for centimeter.
  * @return unit of measure of the measurement
  */
   public Object getUnit()
   {
      return dunits[unitIndex].name;
   }

   /**
    * Returns the full name of the unit property.  This is the full name of the
    * unit of measure
    * @return unit of measure of the measurement
    */
   public String getUnitName()
   {
      return dunits[unitIndex].longName;
   }

   /**
    * Returns the type of the unit.  The type returned can be one of the following:
    * <ul>
    * <li> <code>{@link #FIXED_UNITS}</code>
    * <li> <code>{@link #FONT_UNITS}</code>
    * <li> <code>{@link #DEVICE_UNITS}</code>
    * </ul>
    * @return the type of the unit
    */
   public int getUnitType()
   {
      if (dunits[unitIndex].id <= FIXED)
         return FIXED_UNITS;
      else if (dunits[unitIndex].id <= FONT)
         return FONT_UNITS;
      else
         return DEVICE_UNITS;
   }

   /**
    * Initialize a unit. This is a framework method.  This is called by the superclass when a new instance
    * object is being constructed.  This method should only be called while
    * an object is being constructed.  This is called to initialize a unit.
    * @param unit the unit to initialize
    * @exception IllegalArgumentException if the unit is not a supported unit of measure
    */
   protected void initializeUnit(Object unit)
   {
      unitIndex = parseUnit(unit);
      if (unitIndex == -1)
         throw new IllegalArgumentException(Message.format(RB.getStringResource("invalidUnit.ex.txt"), unit));
   }

   /**
    * Returns true if the specified unit is a supported unit of measure, false otherwise.
    * @param unit the unit of measure to validate
    */
   public boolean validateUnit(String unit)
   {
      return (parseUnit(unit) != -1);
   }

   //////////////////////////////////////////////////////////////////////////////////
   // Private methods ...
   //////////////////////////////////////////////////////////////////////////////////

   /**
    * Returns the index of the UnitInfo structure for the specified unit.
    * Returns -1 if not found.
    */
   private int parseUnit(Object unit)
   {
      for (int index = 0; index < dunits.length; index++)
      {
         int i = mostCommonlyUsed[index];
         String dunit = dunits[i].name;
         if (dunit == unit || dunit.equals(unit))
            return i;
         dunit = dunits[i].longName;
         if (dunit == unit || dunit.equals(unit))
            return i;
      }

      return -1;
   }

   private double internalGetEquivalent(UnitInfo srcInfo, UnitInfo dstInfo, double magnitude, Font font)
   {
      if (dstInfo.name.equals(srcInfo.name))
         return magnitude;

      double mag = magnitude * srcInfo.normf;

      if (dstInfo.id <= UM_MILLIMETERS)
      {
         if (srcInfo.id <= UM_MILLIMETERS)
            mag = mag / dstInfo.normf;
         else if (srcInfo.id <= UM_METERS)
            mag = (mag * MM_PER_METER) / dstInfo.normf;
         else if (srcInfo.id <= UM_FEET)
            mag = (mag * MM_PER_FOOT) / dstInfo.normf;
         else if (srcInfo.id <= FIXED)
            mag = (mag * MM_PER_AU) / dstInfo.normf;
         else if (srcInfo.id < FONT)
         {
            System.out.println( RB.getStringResource( RB_KEY, "fontUnitConversionNotImplemented.txt" ) );
         }
         else /* if srcInfo.id <= DEVICE) */
         {
            Toolkit toolkit = Toolkit.getDefaultToolkit();
            if (toolkit == null)
            {
               System.out.println( RB.getStringResource( RB_KEY, "toolkitIsNull.txt" ) );
               return magnitude;
            }

            double dotsPerInch = DEFAULT_DPI;
			if (!GraphicsEnvironment.isHeadless())
				dotsPerInch = (double)toolkit.getScreenResolution();
            mag *= (1/dotsPerInch) * FEET_PER_INCH;

            mag = (mag * MM_PER_FOOT) / dstInfo.normf;
         }
      }
      else if (dstInfo.id <= UM_METERS)
      {
         if (srcInfo.id <= UM_MILLIMETERS)
            mag = (mag * METER_PER_MM) / dstInfo.normf;
         else if (srcInfo.id <= UM_METERS)
            mag = mag / dstInfo.normf;
         else if (srcInfo.id <= UM_FEET)
            mag = (mag * METER_PER_FOOT) / dstInfo.normf;
         else if (srcInfo.id <= FIXED)
            mag = (mag * METER_PER_AU) / dstInfo.normf;
         else if (srcInfo.id <= FONT)
         {
            System.out.println( RB.getStringResource( RB_KEY, "fontUnitConversionNotImplemented.txt" ) );
         }
         else /* if srcInfo.id <= DEVICE) */
         {
            Toolkit toolkit = Toolkit.getDefaultToolkit();
            if (toolkit == null)
            {
               System.out.println( RB.getStringResource( RB_KEY, "toolkitIsNull.txt" ) );
               return magnitude;
            }

            double dotsPerInch = DEFAULT_DPI;
			if (!GraphicsEnvironment.isHeadless())
				dotsPerInch = (double)toolkit.getScreenResolution();
            mag *= (1/dotsPerInch) * FEET_PER_INCH;

            mag = (mag * METER_PER_FOOT) / dstInfo.normf;
         }
      }
      else if (dstInfo.id <= UM_FEET)
      {
         if (srcInfo.id <= UM_MILLIMETERS)
            mag = (mag * FEET_PER_MM) / dstInfo.normf;
         else if (srcInfo.id <= UM_METERS)
            mag = (mag * FEET_PER_METER) / dstInfo.normf;
         else if (srcInfo.id <= UM_FEET)
            mag = mag / dstInfo.normf;
         else if (srcInfo.id <= FIXED)
            mag = (mag * FEET_PER_AU) / dstInfo.normf;
         else if (srcInfo.id < FONT)
         {
            System.out.println( RB.getStringResource( RB_KEY, "fontUnitConversionNotImplemented.txt" ) );
         }
         else /* if srcInfo.id <= DEVICE) */
         {
            Toolkit toolkit = Toolkit.getDefaultToolkit();
            if (toolkit == null)
            {
               System.out.println( RB.getStringResource( RB_KEY, "toolkitIsNull.txt" ) );
               return magnitude;
            }

            double dotsPerInch = DEFAULT_DPI;
			if (!GraphicsEnvironment.isHeadless())
				dotsPerInch = (double)toolkit.getScreenResolution();
            mag *= (1/dotsPerInch) * FEET_PER_INCH;

            mag = mag / dstInfo.normf;
         }
      }
      else if (dstInfo.id <= FIXED)
      {
         if (srcInfo.id <= UM_MILLIMETERS)
            mag = (mag * AU_PER_MM) / dstInfo.normf;
         else if (srcInfo.id <= UM_METERS)
            mag = (mag * AU_PER_METER) / dstInfo.normf;
         else if (srcInfo.id <= UM_FEET)
            mag = (mag * AU_PER_FOOT) / dstInfo.normf;
         else if (srcInfo.id <= FIXED)
            mag = mag / dstInfo.normf;
         else if (srcInfo.id < FONT)
         {
            System.out.println( RB.getStringResource( RB_KEY, "fontUnitConversionNotImplemented.txt" ) );
         }
         else /* if srcInfo.id <= DEVICE) */
         {
            Toolkit toolkit = Toolkit.getDefaultToolkit();
            if (toolkit == null)
            {
               System.out.println( RB.getStringResource( RB_KEY, "toolkitIsNull.txt" ) );
               return magnitude;
            }

            double dotsPerInch = DEFAULT_DPI;
			if (!GraphicsEnvironment.isHeadless())
				dotsPerInch = (double)toolkit.getScreenResolution();

            mag *= (1/dotsPerInch) * FEET_PER_INCH;

            mag = (mag * AU_PER_FOOT) / dstInfo.normf;
         }
      }
      else if (dstInfo.id <= FONT)
      {
         System.out.println( RB.getStringResource( RB_KEY, "fontUnitConversionNotImplemented.txt" ) );
      }
      else /* if (dstInfo.id <= DEVICE) */
      {
         Toolkit toolkit = Toolkit.getDefaultToolkit();
         if (toolkit == null)
         {
            System.out.println( RB.getStringResource( RB_KEY, "toolkitIsNull.txt" ) );
            return magnitude;
         }

        double dotsPerInch = DEFAULT_DPI;
		if (!GraphicsEnvironment.isHeadless())
			dotsPerInch = (double)toolkit.getScreenResolution();

         if (dstInfo.id == UM_X_PIXEL)
         {
            if (srcInfo.id <= UM_MILLIMETERS)
            {
               double XPIXEL_PER_MM = dotsPerInch * INCHES_PER_FOOT * FEET_PER_MM;
               mag = (mag * XPIXEL_PER_MM) / dstInfo.normf;
            }
            else if (srcInfo.id <= UM_METERS)
            {
               double XPIXEL_PER_METER = dotsPerInch * INCHES_PER_FOOT * FEET_PER_METER;
               mag = (mag * XPIXEL_PER_METER) / dstInfo.normf;
            }
            else if (srcInfo.id <= UM_FEET)
            {
               double XPIXEL_PER_FOOT = dotsPerInch * INCHES_PER_FOOT;
               mag = (mag * XPIXEL_PER_FOOT) / dstInfo.normf;
            }
            else if (srcInfo.id <= FIXED)
            {
               double XPIXEL_PER_AU = dotsPerInch * INCHES_PER_FOOT * FEET_PER_AU;
               mag = (mag * XPIXEL_PER_AU) / dstInfo.normf;
            }
            else if (srcInfo.id <= FONT)
            {
               System.out.println( RB.getStringResource( RB_KEY, "fontUnitConversionNotImplemented.txt" ) );
            }
            else /* if srcInfo.id <= DEVICE) */
            {
               if (srcInfo.id == UM_X_PIXEL)
                  return mag / dstInfo.normf;
               else /* if (srcInfo.id == UM_Y_PIXEL) */
                  return mag / dstInfo.normf;
            }
         }
         else if (dstInfo.id == UM_Y_PIXEL)
         {
            if (srcInfo.id <= UM_MILLIMETERS)
            {
               double YPIXEL_PER_MM = dotsPerInch * INCHES_PER_FOOT * FEET_PER_MM;
               mag = (mag * YPIXEL_PER_MM) / dstInfo.normf;
            }
            else if (srcInfo.id <= UM_METERS)
            {
               double YPIXEL_PER_METER = dotsPerInch * INCHES_PER_FOOT * FEET_PER_METER;
               mag = (mag * YPIXEL_PER_METER) / dstInfo.normf;
            }
            else if (srcInfo.id <= UM_FEET)
            {
               double YPIXEL_PER_FOOT = dotsPerInch * INCHES_PER_FOOT;
               mag = (mag * YPIXEL_PER_FOOT) / dstInfo.normf;
            }
            else if (srcInfo.id <= FIXED)
            {
               double YPIXEL_PER_AU = dotsPerInch * INCHES_PER_FOOT * FEET_PER_AU;
               mag = (mag * YPIXEL_PER_AU) / dstInfo.normf;
            }
            else if (srcInfo.id < FONT)
            {
               System.out.println( RB.getStringResource( RB_KEY, "fontUnitConversionNotImplemented.txt" ) );
            }
            else /* if srcInfo.id <= DEVICE) */
            {
               if (srcInfo.id == UM_X_PIXEL)
                  return mag / dstInfo.normf;
               else /* if (srcInfo.id == UM_Y_PIXEL) */
                  return mag / dstInfo.normf;
            }
         }

      }
      return mag;
   }

   /**
    * Returns a collection of equivalent different units with scaled magnitude values.
    * Possible values for id are: <code>{@link #FIXED_UNITS}</code>,
    * <code>{@link #FONT_UNITS}</code> and
    * <code>{@link #DEVICE_UNITS}</code> which can be
    * combined in an OR relationship to specify all.
    * @param id Specifies in which units the values should be returned.
    */
   public static Map makeUnitsMap(int id)
   {
      Map aMap = new HashMap();
      BaseDistance meter = new BaseDistance(1.0, "m");

      if ((id & FIXED_UNITS) != 0)
      {
         for (int i = 0, l = dunits.length; i <= FIXED && i < l; i++)
         {
            UnitInfo info = dunits[i];
            double d = meter.getMagnitude(info.name);
            aMap.put(info.longName, new Double(1/d));
         }
      }

      if ((id & FONT_UNITS) != 0)
      {
         for (int i = (FIXED+1), l = dunits.length; i <= FONT && i < l; i++)
         {
            UnitInfo info = dunits[i];
            double d = meter.getMagnitude(info.name);
            aMap.put(info.longName, new Double(1/d));
         }
      }

      if ((id & DEVICE_UNITS) != 0)
      {
         for (int i = (FONT+1), l = dunits.length; i <= DEVICE && i < l; i++)
         {
            UnitInfo info = dunits[i];
            double d = meter.getMagnitude(info.name);
            aMap.put(info.longName, new Double(1/d));
         }
      }
      return aMap;
   }

   /**
    * Returns a collection of the specified units.
    * Possible values for id are: <code>{@link #FIXED_UNITS}</code>,
    * <code>{@link #FONT_UNITS}</code> and
    * <code>{@link #DEVICE_UNITS}</code> which can be
    * combined in an OR relationship to specify all.
    * @param id Specifies which units should be in the collection.
    */
   public static List makeUnitsJavaList(int id)
   {
      List list = new ArrayList();

      if ((id & FIXED_UNITS) != 0)
      {
         for (int i = 0, l = dunits.length; i <= FIXED && i < l; i++)
         {
            UnitInfo info = dunits[i];
            list.add(info.longName);
         }
      }

      if ((id & FONT_UNITS) != 0)
      {
         for (int i = (FIXED+1), l = dunits.length; i <= FONT && i < l; i++)
         {
            UnitInfo info = dunits[i];
            list.add(info.longName);
         }
      }

      if ((id & DEVICE_UNITS) != 0)
      {
         for (int i = (FONT+1), l = dunits.length; i <= DEVICE && i < l; i++)
         {
            UnitInfo info = dunits[i];
            list.add(info.longName);
         }
      }
      return list;
   }

}
