/* $Id$ */
// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.measures;

import com.sas.PublicClonable;
import com.sas.util.NumberUtil;
import com.sas.util.Util;

import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.ParsePosition;
import java.util.StringTokenizer;

/**
 * <code>Measure</code> is a class which is used to contain any type of measurement.  The
 * measurement consists of a numeric value plus a unit of measurement.
 * This class does not dictate what valid units are. A subclass may enforce
 * units to be within a discrete set.  <code>Measure</code> does not
 * store the unit value.  This
 * is done to allow subclasses to store the unit in whichever manner they choose.
 *
 * @author David J. Biesack
 * @author Ryan Norris
 * @version 1.0
 */
public abstract class Measure
   implements MeasureInterface,
              PublicClonable,
              java.io.Serializable
{
   /* Serial Version UID */
   static final long serialVersionUID = -6504648673403471770L;

   public final static String RB_KEY = "Measure.";

   protected double magnitude;

   /**
    * Construct a <code>Measure</code> object from the specified magnitude and unit of measure.
    * @param magnitude numeric value of the measurement
    * @param unit unit of measure
    */
   protected Measure(double magnitude, Object unit)
   {
      initializeUnit(unit);
      this.magnitude = magnitude;
   }

   /**
    * Construct a <code>Measure</code> object from a string consisting of
    * a numeric value and unit of measure.
    * @param text a string consisting of a number,
    *    some whitespace, and a units token, such as
    *    <code>"10 mm"</code> or <code>"2.5e3 nm"</code>
    * @param text string containing the magnitude and unit of measure
    */
   protected Measure( String text )
   {
      setText(text);
   }

   //////////////////////////////////////////////////////////////////////////////////
   // Public methods ...
   //////////////////////////////////////////////////////////////////////////////////

   /**
    * Returns the value of the magnitude property.  This is the numeric value of
    * the unit of measure.
    * @return magnitude numeric value of the measurement
    */
   public double getMagnitude()
   {
      return magnitude;
   }

  /**
   * Returns an equivalent magnitude of the current measurement in terms of the
   * specified unit of measure. For example, to convert from kilometers
   * to millimeters, you can use the <code>{@link Distance}</code> class:
   * <code>new Distance(1, "km").getMagnitude("mm")</code>
   * @param unit the unit of measure to convert to
   * @return an equivalent magnitude of the current measurement in terms of the
   *    specified unit of measure
   */
   public abstract double getMagnitude(Object unit);

  /**
   * Returns the unit of measuremnt of the measurement.  Typically this will be
   * the abbreviated name of the unit property.  For example, <code>"in"</code> for inch and
   * <code>"cm"</code> for centimeter.
   * @return unit of measure of the measurement
   */
   public abstract Object getUnit();

   /**
    * Returns the full name of the unit property.  For example, <code>"inch"</code> and <code>"centimeter"</code>.
    * @return unit of measure of the measurement
    */
   public abstract String getUnitName();

   /**
    * Calls to subclasses to initializes a unit.  This method is called on a unit
    * whenever a <code>Measure</code> is constructed with a given measurement.  Since the unit
    * is not stored in <code>Measure</code> this gives subclasses a chance to store the unit
    * in which ever way they find appropriate.
    * @param unit the unit to initialize
    */
   protected abstract void initializeUnit(Object unit);

   /**
    * Returns the a string representation of this <code>Measure</code> object.
    * The string returned contains the magnitude and the abbreviated
    * name of the unit.
    */
   public String toString()
   {
      return NumberUtil.toString(getMagnitude()) + " " + getUnit();
   }

   /**
    * Returns a clone of this <code>Measure</code> object.
    */
   public Object clone()
    {
      try { return super.clone(); }
      catch (CloneNotSupportedException e) { return null; }
    }

   /**
    * Compare the current measure with another to see if they are identical.
    * They are considered equal if they have the same unit and magnitude.
    *
    * @param measure the measure object to compare
    * @return true if <var>measure</var> is equal to the current measure,
    *         otherwise false
    */
   public boolean equals(MeasureInterface measure)
   {
      return (measure.getMagnitude() == getMagnitude() &&
         Util.equal(measure.getUnit(), getUnit()));
   }

   /**
    * Compare to another object.
    *
    * @param o another object
    * @return if the other object implements <code>MeasureInterface</code>,
    *         return <code>equals((MeasureInterface)o)</code>, else
    *         return false
    */
   public final boolean equals(Object o)
   {
      if (o instanceof MeasureInterface)
         return equals((MeasureInterface)o);
      return false;
   }

   public int hashCode()
   {
      int code = 0;
      code += getMagnitude();
      if (getUnit() != null)
         code += getUnit().hashCode();
      return code;
   }

   //////////////////////////////////////////////////////////////////////////////////
   // Private methods ...
   //////////////////////////////////////////////////////////////////////////////////

   private void setText(String text)
   {
      try
      {
         // first trim the text
         text = text.trim();

         NumberFormat formatter = NumberFormat.getInstance();
         ParsePosition pos = new ParsePosition(0);

         // get the magnitude
         Number number = formatter.parse(text, pos);
         if (number == null || pos.getIndex() == 0)
            throw new Exception();
         double v = number.doubleValue();

         // get the unit and validate it
         initializeUnit(text.substring(pos.getIndex()).trim());

         // finally set the magnitude if the unit was ok
         this.magnitude = v;
      }
      catch (Exception e)
      {
         RuntimeException rte = new IllegalArgumentException(
            MessageFormat.format(
               RB.getStringResource(RB_KEY, "invalidArgument.ex.txt"), new Object[] {text} ));
		 rte.initCause( e );
		 throw rte;
      }
   }
}
