// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;

/**
 * BufferHints is used when specifying hints to an object
 * implementing the BufferHintsInterface. This object
 * conveys the hint information, as well as
 * a link to additional hints.
 * The view passes this information to the model via the model's
 * <code>setBufferHints(BufferHints)<code> method
 * Note: this object has public members, for ease of use and convenience.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public class BufferHints
{

  //  Potential values for BufferHints.what :

  /**
   * A cache access type which indicates the cache user
   * will primarily access data from the cache in sequential
   * manner. The cache may use this hint to perform block
   * look-ahead reads from the secondary cache.
   */
  public static final int SEQUENTIAL_ACCESS  = 0x1000;
  /**
   * A cache access type which indicates the cache user
   * will primarily access data from the cache in a non-sequential
   * or "random" manner. The cache may use this hint to avoid block
   * look-ahead reads from the secondary cache.
   */
  public static final int RANDOM_ACCESS      = 0x2000;

  /**
   * A BufferHint.what value, indicating what direction the
   * cache is reading; use either READ_FORWARD or READ_BACKWARD
   * for the corresponding value for the buffer hint 'how'
   * @since version 2.0
   */
  public static final int READ_DIRECTION = 0x4000;

  /**
   * A buffer hint {@link #how} value for a READ_DIRECTION, indicating forward read ahead
   * use this when reading.
   * @since version 2.0
   */
  public static final int READ_FORWARD   = 0x00100;
  /**
   * A buffer hint {@link #how} value for a READ_DIRECTION, indicating backward read ahead
   * use this when reading.
   * @since version 2.0
   */
  public static final int READ_BACKWARD  = 0x00200;

  /**
   * A buffer hint {@link #how} value, indicates read access
   */
  public static final int READ_ACCESS        = 0x0001;

  /**
   * A buffer hint {@link #how} value, indicates write access
   */
  public static final int WRITE_ACCESS       = 0x0002;

  /**
   * A buffer hint {@link #how} value, indicating the cache owner
   * cannot specify if how it will access data.
   * Mutually exclusive with READ_ACCESS and WRITE_ACCESS
   * @see #READ_ACCESS
   * @see #WRITE_ACCESS
   */
  public static final int UNKNOWN_ACCESS     = 0x0004;


  /** A buffer hint {@link #how} value, combines SEQUENTIAL_ACCESS with READ_ACCESS  */
  public static final int SEQUENTIAL_READ_ACCESS  = SEQUENTIAL_ACCESS | READ_ACCESS ;
  /** A buffer hint {@link #how} value, combines RANDOM_ACCESS with READ_ACCESS */
  public static final int RANDOM_READ_ACCESS      = RANDOM_ACCESS     | READ_ACCESS ;
  /** A buffer hint {@link #how} value, combines SEQUENTIAL_ACCESS with WRITE_ACCESS */
  public static final int SEQUENTIAL_WRITE_ACCESS = SEQUENTIAL_ACCESS | WRITE_ACCESS ;
  /** A buffer hint {@link #how} value, combines RANDOM_ACCESS with WRITE_ACCESS */
  public static final int RANDOM_WRITE_ACCESS     = RANDOM_ACCESS     | WRITE_ACCESS ;

  // a bug in the JDK Java compiler:
  // Even though all the constructors initialize
  // the fields, if this file is compiled indirectly when compiling
  // another, I get an error. So, be sure to compile this file
  // stand-alone

   /**
    * What will be accessed; see items in BufferHintsInterface
    */
   public final int what;

   /**
    * How many items will be accessed.
    * Should be a positive integer or BufferHintsInterface.DO_ALL
    */
   public final int count;

   /**
    * How the items will be accessed.
    * One of BufferHints.READ_ACCESS or BufferHints.WRITE_ACCESS
    */
   public final int how;

   /**
    * Additional data which the viewer can associate with this hint.
    */
   public final Object data;

   /**
    * Additional hints, allowing BufferHintsInterface.useBufferHints
    * to pas an entire chain of hints.
    */
   public final BufferHints moreHints;

  public BufferHints(int iWhat, int iHow, int iCount)
    {
      // JDK javac barfs if I try:
      // this(iWhat, iHow, iCount, null, null);

      this.what = iWhat;
      this.how = iHow;
      this.count = iCount;
      this.data = null;
      this.moreHints = null;
    }
  public BufferHints(int iWhat, int iHow, int iCount, Object iData)
    {
      // JDK javac barfs if I try:
      // this(iWhat, iHow, iCount, iData, null);

      this.what = iWhat;
      this.how = iHow;
      this.count = iCount;
      this.data = iData;
      this.moreHints = null;
    }
  public BufferHints(int iWhat, int iHow, int iCount, Object iData, BufferHints iMoreHints)
    {
      this.what = iWhat;
      this.how = iHow;
      this.count = iCount;
      this.data = iData;
      this.moreHints = iMoreHints;
    }

  public String toString()
    {
      return "<BufferHints what: 0x" + Integer.toHexString(what)
        + ", how: 0x" + Integer.toHexString(how)
        + ", count: " + count
        + ", data: " + data
        + ((moreHints == null) ? "" : moreHints.toString())
        + ">";
    }
}
