// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;

/**
 * BufferHintsInterface is used for communicating data buffering hints
 * between by viewers and models. A viewer object which reads data
 * from or writes data to a model can provide hints to the model to
 * describe the view's intended read or write access patterns, such as
 * claiming it will be reading sequential rows of data,
 * or reading arbitrary data objects in a non-sequential manner.
 * These hints allow the models, especially those which perform I/O
 * including network I/O to satisfy their data requests, to cache
 * the remote data locally or to perform look-ahead reads.
 *
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public interface BufferHintsInterface
{

  /**
   * Ask a model to use a set of buffering hints. This
   * is a way for a viewer to inform the model how the viewer
   * intends to access the model's data.
   */
  public void useBufferHints(BufferHints hints);

  /** A value for BufferHints.what, indicating access by objects */
  public static final int BY_OBJECT = 0;

  /** A value for BufferHints.what, indicating access by cell */
  public static final int BY_CELL   = 1;

  /** A value for BufferHints.what, indicating access by row */
  public static final int BY_ROW    = 2;

  /** A value for BufferHints.what, indicating access by column */
  public static final int BY_COLUMN = 3;

  /** A value for BufferHints.what, indicating access by byte */
  public static final int BY_BYTE   = 4;


  /*
   * A value for BufferHints.what, used to set the buffered read
   * size.
   * <p>
   * When reading data from
   * a (remote) model, a model that implements BufferHintsInterface
   * may supprt reading a block at a time,
   * to reduce the number of separate network method calls.
   * The data read from the remote model are stored locally,
   * usually until they are purged from the internal cache.
   * This parameter controls how many values are read when
   * doing this look-ahead reading.
   * If it is critical that the model not perform look-ahead
   * reads, set this to 1. However, this may degrade performance.
   */
  static final int LOOKAHEAD_SIZE = 0x1000;

  /**
   * A value for BufferHints.what, used to set the model's cache size.
   * <p>
   * The number of values which the client side object maintains
   * in a local memory cache. The cache size allows more efficient
   * back-and-forth reading behavior from the model, for example,
   * scrolling a table view displaying a table model in a window of rows
   * and columns. With a cache size setting larger than a window, the
   * table model can keep more than one page of data locally, so that
   * moving back a forth between two or three pages does not require
   * rereading the data from the remote model.  This is especially
   * critical when the data model cannot
   * backed up to a previously read observation without rewinding to
   * the beginning and reading all preceeding data, as may be the
   * case with certain JDBC or other data tables.
   */
  static final int CACHE_SIZE = 0x2000;

  /** A value for BufferHints.count, indicating the view will access all
      items at once. */
  public static final int DO_ALL = -1;

}
