// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;

/**
 * This is an astract base class for cache objects.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @see com.sas.models.StaticCacheInterface
 * @see com.sas.models.CacheInterface
 * @see com.sas.models.SequentialCache
 * @see com.sas.models.SparseCache
 */

public abstract class CacheBase
       implements CacheInterface
{
  // Cache configuration

  // Hmm, need to run some benchmarks to optimize these defaults...
  public static final int DEFAULT_MAX_SIZE = 128;
  public static final int DEFAULT_LOOKAHEAD_SIZE = 16;
  public static final int DEFAULT_THRESHOLD_VALUE = 25;
  public static final int DEFAULT_THRESHOLD_TYPE  = THRESHOLD_PERCENT;

  /**
   * The secondary cache this cache references on cache misses
   */
  protected StaticCacheInterface secondaryCache;
  /**
   * The cache size; the cache will store up to <var>maxSize</var> elements
   */
  protected int maxSize;
  /**
   * The number of items in the cache
   */
  protected int count;
  /**
   * The value for the threshold.
   */
  protected int threshold;
  /**
   * The threshold type
   */
  protected int thresholdType;
  /**
   * how the cache is accessed
   */
  protected int accessType;
  /**
   * true if the access is sequential
   */
  protected boolean sequential;

  /**
   * On a cache miss, lookAheadSize is how large of a block read to perform.
   */
  protected int lookAheadSize;

  /**
   * What is the current read direction? Possible values
   * include but are not limited to BufferHintsInterface.READ_FORWARD
   * and BufferHintsInterface.READ_BACKWARD
   * @see com.sas.models.BufferHints#READ_FORWARD
   * @see com.sas.models.BufferHints#READ_BACKWARD
   */
  protected int direction;

  protected int directionChangeThreshold;
  protected int leaning;


  /**
   * Construct a cache.
   * @param secondaryCache the secondary cache this cache references on cache misses
   */
  public CacheBase(StaticCacheInterface secondaryCache)
    {
      this(secondaryCache, DEFAULT_MAX_SIZE, DEFAULT_LOOKAHEAD_SIZE, DEFAULT_THRESHOLD_TYPE, DEFAULT_THRESHOLD_VALUE);
    }

  /**
   * Construct a cache. The default is direction READ_FORWARD.
   * @param secondaryCache the secondary cache this cache references on cache misses
   * @param maxSize the cache size; the cache will store up to <var>maxSize</var> elements;
   *                use -1 to not change the max size.
   * @param lookAheadSize the size of the look-ahead buffer
   *                use -1 to not change the look-ahead buffer size
   * @param thresholdType the threshold type
   *                use -1 to not change the threshold type
   * @param thresholdValue the value for the threshold.
   *                use -1 to not change the threshold value
   * @see com.sas.models.StaticCacheInterface#configureCache(int, int, int, int, int)
   * @see com.sas.models.BufferHints#READ_FORWARD
   */
  public CacheBase(StaticCacheInterface secondaryCache,
               int maxSize, int lookAheadSize, int thresholdType, int thresholdValue)
    {
      this.secondaryCache = secondaryCache;
      if (maxSize != -1)
        this.maxSize = maxSize;
      if (lookAheadSize != -1)
        this.lookAheadSize = Math.max(1,lookAheadSize);
      if (threshold != -1)
        this.threshold = threshold;
      if (thresholdValue != -1)
        this.thresholdType = thresholdType;
      setDirection(BufferHints.READ_FORWARD);
      directionChangeThreshold = 2;
    }

  /**
   * Return the cache size: the number of items in the cache
   * @return the cache size
   */
  public int count()
    {
      return count;
    }


  /**
   * Controls whether to try lookahead reads. Lookahead is incremented
   * by disableLookahead() and re-enabled when enableLookAhead()
   * decrements this back to 0.
   * @see #enableLookAhead
   * @see #disbleLookAhead
   * @see #lookAheadEnabled
   */
  private int lookAheadState;

  /**
   * Check if lookahead reads are enabled.
   * @return <code>true</code> if lookahead reads are enabled, else return
   * false.
   */
  public boolean lookAheadEnabled()
    {
      return lookAheadState == 0;
    }

  /**
   * Enable lookahead if possible.
   * If disableLookAhead is called three times,
   * enableLookAhead must be called three times
   * before lookahead reads are actually used.
   * @see #disableLookAhead
   */
  public void enableLookAhead()
    {
      if (lookAheadState > 0)
        lookAheadState--;
      // System.out.println("enableLookAhead " + lookAheadState);
    }

    /**
     * Disable lookahead if possible.
     * If disableLookAhead is called three times,
     * enableLookAhead must be called three times
     * before lookahead reads are actually used.
     * @see #enableLookAhead
     */
  public boolean disableLookAhead()
    {
      lookAheadState++;
      // System.out.println("disableLookAhead " + lookAheadState);
      return true;
    }

  private static Object noObjects[];

  /**
   * Utility method to return an Object array of no elements
   */
  public static Object[] nullObjectArray()
    {
      if (noObjects == null)
        noObjects = new Object[0];
      return noObjects;
    }


   /**
    * Ask the local data set object to use a set of buffering hints.
    * Use this to configure the number of rows to read ahead
    * when doing buffered reads, or how many rows to maintain
    * in the local client-side cache.
    * <p>
    * The DataSet object
    * currently supports only caching of sequential reads,
    * and row-based caching.
    * @param hints a (linked list) of buffer hints.
    */
    public void useBufferHints(BufferHints hints)
    {
      while (hints != null)
        {
          try
            {
              // we don't handle WRITE_ACESS buffering, just READ
              if ((hints.how & BufferHints.READ_ACCESS) == BufferHints.READ_ACCESS)
                {
                  if (hints.what == BufferHintsInterface.CACHE_SIZE)
                    configureCache(hints.count, -1, -1, -1, -1);
                  else if  (hints.what == BufferHintsInterface.LOOKAHEAD_SIZE)
                    configureCache(-1, hints.count, -1, -1, -1);
                }
              else if ( hints.what == BufferHints.READ_DIRECTION )
                setDirection(hints.how);
            }
          catch (CacheException ce) { /* IGNORE EXCEPTION */ }

          hints = hints.moreHints;
        }
    }


  /**
   * Set the read ahead direction.
   * @param newDirection the new value for the direction property,
   * one of READ_FORWARD or READ_BACKWARD, although subclasses
   * may define additional direction settings.
   * @see com.sas.models.BufferHints#READ_FORWARD
   * @see com.sas.models.BufferHints#READ_BACKWARD
   */
   public void setDirection(int newDirection)
   {
     direction = newDirection;
   }

   /**
    * Return the look-ahead direction, nominally one of
    * READ_FORWARD or READ_BACKWARD.
    * @return the look-ahead direction.
    */
    public int getDirection()
    {
      return direction;
    }

  /**
   * Configures the cache size, buffering factor and how the cache
   * responds to multiple-item read requests when more than one item
   * in the request is not in the cache and the cache must read values
   * from the sondary cache.
   * @param maxCacheSize the max cache size.
   * @param bufferLookAhead the look-ahead buffer size.
   *        <strong>Note:</strong> you will probably get poor performance
   *        if <var>bufferLookAhead</var> is larger than
   *        <var>maxCacheSize</var>
   * @param thresholdType the threshold type
   * @param thresholdValue the value for the threshold.
   * @param accessType a hint about how data will be read and accessed
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#configureCache
   */
  public synchronized void configureCache(int maxCacheSize, int bufferLookAhead,
                                          int thresholdType, int thresholdValue,
                                          int accessType) throws CacheException
    {
      if (bufferLookAhead != -1)
        lookAheadSize = bufferLookAhead;

      if (thresholdType != -1)
        {
          this.thresholdType = thresholdType;
          this.threshold     = thresholdValue;
        }

      if (this.accessType != -1)
        this.sequential = (accessType & BufferHints.SEQUENTIAL_ACCESS) != 0 ;

      if (maxCacheSize != -1)
        maxSize = maxCacheSize;
    }

  /**
   * Returns an item from the cache, based on a cache key <var>key</var>.
   * @param key the item's key
   * @return the item associated with the key.
   * @exception com.sas.models.CacheException if the cache or its secondary
   * cache does not contain the item.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long)
   */
   abstract public Object getFromCache(long key) throws CacheException;

  /**
   * Return the values associated with the <var>keys</var>.
   * @param keys an array of value keys
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result
   * @return an array of values from the cache, with values corresponding
   * to each key.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long[], boolean)
   */
   abstract public Object[] getFromCache(long keys[], boolean readFully) throws CacheException;

  /**
   * Gets and returns a set of values associated with the range of keys.
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result
   * @return an array of values from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long, long, boolean)
   */
   abstract public Object[] getFromCache(long startKey, long endKey, boolean readFully) throws CacheException;

  /**
   * Computes the complete set of keys between the <var>startKey</var> key and the
   * <var>endKey</var> key, inclusive.
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param size the number of items to read if a look-ahead read is possible.
   * @return an array of keys for the items between the start and end keys,
   * inclusive.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#computeCacheKeyRange(long, long, int)
   */
   abstract public long[] computeCacheKeyRange(long start, long end, int size) throws CacheException;

  /**
   * Returns an array containing the keys from the cache.
   * @return an array containing the keys for all items in the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   abstract public long[] getCacheKeys() throws CacheException;

  /**
   * Tests if the cache contains an item for the <var>key</var>
   * @param key a potential key for an item in the cache.
   * @return <code><b>true</b></code> if the cache contains a value for
   * the key <var>key</var>, <code><b>false</b></code> otherwise.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   abstract public boolean cacheContainsKey(long key) throws CacheException;

  /**
   * Assigns the secondary cache. A previously assigned secondary cache may be
   * discarded, or the cache may cascade its caches.
   * @param cache the secondary cache to assign to this cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.CacheInterface#setSecondaryCache(com.sas.models.StaticCacheInterface)
   */
   public void setSecondaryCache(StaticCacheInterface cache) throws CacheException
    {
      this.secondaryCache = secondaryCache;
    }

  /**
   * Return a reference to the secondary cache which feeds this one.
   * The secondary cache can be either an actual data source,
   * or a second level cache which is attached to a data source.
   * @return the secondary cache which this cache reads from.
   */
  public StaticCacheInterface getSecondaryCache()
    {
      return secondaryCache;
    }

  /**
   * Tests if the cache contains an item for the <var>key</var>
   * @return an integer describing how the cache interprets a range of keys,
   * one of <CODE>USES_SEQUENTIAL_KEYS</CODE> or <CODE>USES_SPARSE_KEYS</CODE>.
   * @see com.sas.models.StaticCacheInterface#getCacheKeyUse()
   */
   public int getCacheKeyUse()
    {
      return secondaryCache.getCacheKeyUse();
    }

  /**
   * Removes all items from the cache.
   */
   abstract public void invalidateCache();

  /**
   * removes from the cache the item associated with the key <var>key</var>.
   * @param key the key for the item to remove from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   abstract public void invalidateCache(long key) throws CacheException;

  /**
   * Stores the value <var>value</var> in the cache so that it may be
   * accessed by the key <var>key</var>. This does <strong>not</strong> write the item
   * through to the secondary cache.
   * @param key the item's key
   * @param value the value to store in the cache; it is associated with
   * the key. If a previous value is in the cache for this key,
   * it is dropped from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   abstract public void storeInCache(long key, Object value) throws CacheException;

  /**
   * Test if the cache is at capacity
   * @return true if the cache is full, else return false
   */
  protected boolean isFull()
    {
      return count == maxSize;
    }

  /**
   * Return the cache capacity.
   */
  public int getCacheCapacity()
    {
      return maxSize;
    }

}
