// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;

/**
 * A cache monitor is an object that observes
 * a cache and is notified when the cache removes an item.
 * Normally the cache monitor is the object that constructs and owns the cache.
 * The cache notifies the monitor when it is about to invalidate an object
 * from the cache, allowing the monitor to override the removal. Else,
 * the cache informs the monitor when it removes an object from the cache,
 * or when the cache fills up.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @see com.sas.models.StaticCacheInterface
 * @see com.sas.models.CacheInterface
 */

public interface CacheMonitorInterface
{

  /**
   * Ask the monitor if it is OK to remove an item from the cache.
   * The cache runs this check when the cache is at a near full
   * situation and wishes remove items. The monitor can veto
   * the removal request if it is necessary to keep the item
   * in the cache.
   * <p>
   * A invalidateCache() call does not call isCacheItemRemovable for any cache
   * elements, since it is assumed invalidateCache() is called
   * from the cache owner only.
   * @param item the item the cache wishes to remove from the cache.
   * @param key the item's key
   * @param cache the cache that wishes to remove the item
   * @return true if the monitor agrees to removing the item;
   *         the cache will then remove the item but not call
   *         noteCacheItemRemoved. Return
   *         false if the monitor wants the cache to keep the item
   *         in the cache.
   */
  public boolean isCacheItemRemovable (Object item, long key, StaticCacheInterface cache);

  /**
   * Notify the monitor that an item has been removed from
   * the cache, allowing the monitor to respond to the
   * removal in a reasonable manner.  This method will be called
   * if <code>isCacheItemRemovable()</code>was just called and return true.
   * It will also be called during an <code>invalidateCache()</code>method,
   * even though <code>invalidateCache()</code> does not call
   * <code>isCacheItemRemovable()</code>
   * @param item the item the cache has just removed.
   * @param key the item's key
   * @param cache the cache that removed the item
   */
  public void noteCacheItemRemoved (Object item, long key, StaticCacheInterface cache);

  /**
   * Notify the monitor that the cache is full and all
   * items are deemed unremovable as indicated
   * by the monitor's responses to the isCacheItemRemovable method.
   * The cache owner is
   * responsible for ensuring the cache has room for unremovable
   * as well as additional items. The monitor may wish to grow
   * the cache.
   * @param capacity the current cache capacity.
   * @param cache the cache that is at capacity
   * @return true if the cache may ignore isCacheItemRemovable and
   *         unconditionally remove items to make room for more recent
   *         cache entries. Return false if the cache should not
   *         remove such items.  A monitor should take some
   *         corrective action, such as enlarging the cache
   *         or marking items as removavle, before returning false.
   */
  public boolean noteCacheFull(int capacity, StaticCacheInterface cache);
}
