// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;
import java.io.Serializable;
import com.sas.PublicClonable;
import com.sas.util.Util;

/**
 * A class for maintaining a queue of long integer values.
 * Values are placed in the queue with the {@link #queue(long)} method and
 * removed by the {@link #dequeue()} method. Dequeued values are returned
 * in a first in first out order.
 * A <code>LongQueue</code> can also serve as
 * a stack of long integers, via the {@link #push(long)} and {@link #pop()}
 * methods; the <code>pop()</code> method returns values in a first in, last out order.
 * (Note that although not elegant, you can use (abuse?)
 * <code>LongQueue</code> to store double values by calling
 * {@link java.lang.Double#doubleToLongBits(double) <code>doubleToLongBits(double d)</code>} to store a double,
 * and calling {@link java.lang.Double#longBitsToDouble(long) <code>Double.longBitsToDouble(long l)</code>}
 * to decode a long into a double.)
 * <p>
 * This class is not a Component; it is meant for use in the internal
 * implementation of other components.
 * <p>
 * Note: none of the methods in <code>LongQueue</code> are synchronized.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */
public class LongQueue
       implements PublicClonable,
                  Serializable
{
  public static final String RB_KEY = "LongQueue.";

  static final long serialVersionUID = -8085230761683380277L;

  // array to hold the queue contents
  private long[] queue;
  // the number of items in the queue.
  private int size;
  // index of next free slot in the array.
  private int in;
  // is the queue growable or not?
  private boolean growable;

  /**
   * Construct a LongQueue.
   * The default queue has a fixed size and is not growable.
   * @param maxElements the maximum number of elements the queue
   *        can hold.
   */
  public LongQueue(int maxElements)
    {
      this(maxElements, false);
    }

  /**
   * Construct a LongQueue, specifying an initial size
   * and indicating if the queue can grow in size.
   * @param numElements the number of elements the queue
   *        can hold.
   * @param growable if true, the queue will grow when
   *        enqueing an item when the queue is full.
   *        If false, enqueing an item when the queue is full
   *        throws an <code>IllegalStateException</code>.
   */
  public LongQueue(int maxElements, boolean growable)
    {
      if (maxElements <= 0)
        throw new IllegalArgumentException("maxElements <= 0");
      queue = new long[maxElements];
      size = 0;
      in = 0;
      this.growable = growable;
    }

  /**
   * Put another value in the queue.
   * @param value the value to store in the queue.
   * @exception java.lang.IllegalStateException if the
   * queue is full and not growable.
   * @see #dequeue()
   * @see #isFull()
   * @see #setGrowable(boolean)
   */
  public void queue(long value)
    {
      if (isArrayFull())
        if (isGrowable())
          grow();
        else
          throw new IllegalStateException(RB.getStringResource(RB_KEY, "queueFull.ex.txt"));
      queue[in] = value;
      in = (in + 1) % queue.length;
      size++;
    }

  /**
   * Put another value in the queue. This method is an alias for
   * <code>{@link #queue(long)}</code>,
   * for use when you wish to treat the queue as a stack
   * @param value the value to store in the queue.
   * @exception java.lang.IllegalStateException if the
   * queue is full and not growable.
   * @see #pop()
   * @see #queue(long)
   * @see #isFull()
   * @see #setGrowable(boolean)
   */
  public final void push(long value)
    {
      queue(value);
    }

  /**
   * Grow the queue. Called when the queue is full.
   */
  private void grow()
    {
      long newQueue[] = new long[1 + queue.length * 3 / 2];
      int i = 0;
      while ( ! isEmpty() )
        newQueue[i++] = dequeue();
      queue = newQueue;
      in = i;
      size = i;
    }

  /**
   * Remove an item from the queue.
   * @return the least recently queued item.
   * @see #queue(long)
   * @exception java.lang.IllegalStateException if the queue is empty.
   */
  public long dequeue()
    {
      long result = peek();
      size--;
      return result;
    }

  /**
   * Remove an item from the back of the queue. This allows you to
   * treat the queue as a stack: After a {@link #push(long) <code>push(<var>value</var>)</code>} call,
   * <code>pop()</code> will return <var>value</var>
   * @return the most recently queued item.
   * @see #push(long)
   * @see #dequeue()
   * @exception java.lang.IllegalStateException if the queue is empty.
   */
  public long pop()
    {
      if ( isEmpty() )
        throw new IllegalStateException(RB.getStringResource(RB_KEY, "queueEmpty.ex.txt"));
      in = (in + queue.length - 1) % queue.length;
      size--;
      return queue[in];
    }

  /**
   * Return the item at the head of the queue, without dequeueing it. This is the item
   * that <code>{@link #dequeue dequeue()}</code> would return.
   * @return the least recently queued item.
   * @see #isEmpty()
   * @exception java.lang.IllegalStateException if the queue is empty.
   */
  public long peek()
    {
      if ( isEmpty() )
        throw new IllegalStateException(RB.getStringResource(RB_KEY, "queueEmpty.ex.txt"));
      int out = (in + queue.length - size) % queue.length;
      return queue[out];
    }

  /**
   * Dequeue all items from the queue.
   * This is faster than doing
   * <pre>
   * while ( ! q.isEmpty() )
   *    q.dequeue();
   * </pre>
   */
  public void removeAll()
    {
      size = 0;
      in = 0;
    }

  /**
   * Checks if a queue is full.
   * @return <code>true</code> if the queue is full.
   * Always returns false for a growable queue.
   * @see #isEmpty()
   */
  public boolean isFull()
    {
      return ( !isGrowable() && isArrayFull());
    }


  /**
   * Checks if a queue is full.
   * @return <code>true</code> if the queue is full.
   */
  private boolean isArrayFull()
    {
      return ( size == queue.length );
    }

  /**
   * Checks is a queue is empty.
   * @return <code>true</code> if and only if the queue is empty
   * @see #isFull()
   */
  public boolean isEmpty()
    {
      return (size == 0);
    }


  /**
   * Set whether or not the queue can grow when full.
   * If the growable property is true, then queueing an item
   * into a full queue will cause the queue to grow automatically
   * to make room for the new item (and additional items).
   * @param growable if <code>true</code>, the queue can grow when the internal
   *        capacity is reached. If <code>false</code>, the queue cannot grow.
   */
   public void setGrowable(boolean growable)
   {
     this.growable = growable;
   }

   /**
    * Check if the queue can grow to accommodate more items when the initial capacity is reached.
    * @return <code>true</code> if the queue us growable, <code>false</code> if not.
    */
    public boolean isGrowable()
    {
        return growable;
    }

  /**
   * Clone the LongQueue.
   * @return a clone
   * @exception java.lang.CloneNotSupportedException
   * if the object cannot be cloned.
   */
  public Object clone() throws CloneNotSupportedException
    {
      LongQueue clone = (LongQueue) super.clone();
      clone.queue = new long[queue.length];
      System.arraycopy(queue, 0, clone.queue, 0, queue.length);
      return clone;
    }
}
