// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;

import java.util.BitSet;
import com.sas.util.Util;
import com.sas.text.Message;

/**
 * A Cache implementation which keeps values in a range of sequential keys.
 * This cache is useful when reading from a primarily sequential
 * source (i.e. one whose keys are primarily consecutive monotonically
 * increasing integers, such as record/row/observation numbers from
 * a data table or database.)
 * <p>
 * SequentialCache uses an internal array for the data values; the
 * item key is mapped to an aray index offset based on the minimum key.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @see com.sas.models.StaticCacheInterface
 * @see com.sas.models.CacheInterface
 */

public class SequentialCache
       extends CacheBase
{
  public static final String RB_KEY = "SequentialCache.";
  private static boolean DEBUG = false;

  /**
   * The values are stored in this array of Objects.
   * The {@link #offset} is added to each key to yield
   * an index into this array.
   */
  protected Object values[];

  /**
   * The offset to add to a cache key to yield an integer
   * index into the {@link #values} array.
   */
  protected long offset;

  /**
   * A bit vector which marks which of the elememts of the {@link #values}
   * array are in use. Test with bit indices (key + {@link #offset})
   */
  protected BitSet inUse;

  /**
   * Construct a SequentialCache that uses default values.
   * Note that you should assign a secondary cache before
   * trying to access data.
   */
  public SequentialCache()
    {
      this(null);
    }

  /**
   * Construct a SequentialCache that uses a secondary cache and default configuration values.
   * @param secondaryCache the secondary cache this cache references on cache misses
   */
  public SequentialCache(StaticCacheInterface secondaryCache)
    {
      this(secondaryCache, DEFAULT_MAX_SIZE, DEFAULT_LOOKAHEAD_SIZE,
           DEFAULT_THRESHOLD_TYPE, DEFAULT_THRESHOLD_VALUE);
    }

  /**
   * Construct a SequentialCache that uses a secondary cache
   * for cache misses, specifying initial cache parameters.
   * @param secondaryCache the secondary cache this cache references on cache misses
   * @param maxSize the cache size; the cache will store up to <var>maxSize</var> elements
   * @param lookAheadSize the size of the look-ahead buffer
   */
  public SequentialCache(StaticCacheInterface secondaryCache,
                         int maxSize, int lookAheadSize)
    {
      this(secondaryCache, maxSize, lookAheadSize, DEFAULT_THRESHOLD_TYPE, DEFAULT_THRESHOLD_VALUE);
    }

  /**
   * Construct a SequentialCache that uses a secondary cache
   * for cache misses, specifying initial cache parameters.
   * @param secondaryCache the secondary cache this cache references on cache misses
   * @param maxSize the cache size; the cache will store up to <var>maxSize</var> elements
   * @param lookAheadSize the size of the look-ahead buffer
   * @param thresholdType the threshold type
   * @param thresholdValue the value for the threshold.
   */
  public SequentialCache(StaticCacheInterface secondaryCache,
                         int maxSize, int lookAheadSize, int thresholdType, int thresholdValue)
    {
      super(secondaryCache, maxSize, lookAheadSize, thresholdType, thresholdValue);
      init(maxSize);
    }

  private void init(int size)
    {
      inUse = new BitSet(size);
      values = new Object[size];
      offset = 0;
    }


  /**
   * Convert a cache item key into an internal index into the values array.
   * The key is <strong>not</strong> validated against the current cache bounds
   * @param key a cache key
   * @return an index into the values array, or a bit index into the inUse BitSet
   */
   protected final int index(long key)
     {
       return (int) (key + offset);
     }


  /**
   * Return the size of of the cache internal structures.
   * @return the size of of the cache internal structures.
   */
  protected final int size()
    {
      return maxSize;
    }

  /**
   * Convert an internal cache item index into a cache key.
   * The index is <strong>not</strong> validated against the current cache bounds
   * @param index an index into the <var>values</var> array
   * @return the key corresponding to that index.
   */
   protected final long key(int index)
     {
       return (index - offset);
     }

  /**
   * Check if the cache contains an item at the internal index.
   * @param index the item index within the internal values array/inUse bitset
   * @see #index(long)
   * @see #key(int)
   */
   protected final boolean inCache(int index)
     {
       return (index >= 0) && (index < size()) && inUse.get(index);
     }

  /**
   * Returns an item from the cache, based on a cache key <var>key</var>.
   * @param key the item's key
   * @return the item associated with the key.
   * @exception com.sas.models.CacheException if the cache or its secondary
   * cache does not contain the item.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long)
   */
   public synchronized Object getFromCache(long key) throws CacheException
     {
       int index = index(key);
       if (inCache(index))
         return values[index];

       inferReadDirection(key);

       if (getDirection() == BufferHints.READ_FORWARD && inCache(index + 1))
         {
           try  // don't do lookahead if next item is in cache
             {
               disableLookAhead();
               return getFromSecondaryCache(key);
             }
           finally
             { enableLookAhead(); }
         }
       else if (getDirection() == BufferHints.READ_BACKWARD && inCache(index - 1))
         {
           try  // don't do lookahead if previous item is in cache
             {
               disableLookAhead();
               return getFromSecondaryCache(key);
             }
           finally
             { enableLookAhead(); }
         }
       else
         return getFromSecondaryCache(key);

     }

  /**
   * Get a value from the secondary cache, trying a lookahead read.
   * This has a side effect of reading a look-ahead block of data from the
   * secondary cache, if lookahead is enabled, and storing
   * those values in this cache (which in turn may remove
   * some other items already in the cache.)
   * @param key a cache key
   * @return the value from the secondary cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
  protected synchronized Object getFromSecondaryCache(long key) throws CacheException
    {
      inferReadDirection(key);
      if (lookAheadEnabled())
        {
          boolean disabled = false;
          try
            {
              disabled = disableLookAhead();
              // perform lookahead.
              long keys[] = computeCacheKeyRange(key, key, lookAheadSize);
              if (DEBUG) System.out.println("secondaryCache.getFromCache(" + toString(keys) + ")");
              lastFetchStart = keys[0];
              lastFetchEnd   = keys[keys.length - 1];
              Object newValues[] = secondaryCache.getFromCache(keys, false);
              storeInCache(keys, newValues);

              // expect first item in result to match, but that depends
              // on behavior of computeCacheKeyRange, so check all values to make sure.
              for (int i = 0, nKeys = Math.min(keys.length, newValues.length); i < nKeys; i++)
                if (keys[i] == key)
                  return newValues[i];
              // Hmm, secondary cache failed to return a value for this key.
              throw new CacheException(Message.format(RB.getStringResource("keyNotInCache.ex.txt"), new Long(key)));
            }
          finally
            {
              if (disabled)
                enableLookAhead();
            }
        }
      else
        {
          if (DEBUG) System.out.println("secondaryCache.getFromCache(" + key + ")");
          Object value = secondaryCache.getFromCache(key);
          lastFetchStart = lastFetchEnd   = key;
          storeInCache(key, value);
          return value;
        }
    }

  long lastFetchStart = -Long.MAX_VALUE;
  long lastFetchEnd = Long.MAX_VALUE;

  protected void inferReadDirection(long key)
    {
      int index = index(key);
      if (DEBUG)
          System.out.println("inferReadDirection(" + key + "), index is " + index + ", leaning is " + leaning);
      if (index < 0 || index < firstInCache()) // lean towards backward read
        {
        leanLeft();
	    }
      else if (index > size() || index > lastInCache())
        leanRight();
      else if (key < lastFetchStart)
        {
        leanLeft();
	    }
      else if (key > lastFetchStart)
        leanRight();
      // else leaning = 0;
      // else, not enough to infer enough of a direction change yet
      // may be just a small backup.  Only change direction when
      // reading data outside the cache.
    }

  protected void leanRight()
  {
    leaning ++;
    if ((leaning - directionChangeThreshold) >= 0)
      {
        setDirection(BufferHints.READ_FORWARD);
        leaning = 0;
      }
  }
  protected void leanLeft()
  {
    leaning --;
    if ((leaning + directionChangeThreshold) <= 0)
      {
        setDirection(BufferHints.READ_BACKWARD);
        leaning = 0;
      }
  }

  // TO-DO: update firstInCache() and lastInCache() to use fields
  // and modify other cache methods to keep those fields correct,
  // so we can avoid the loop/search
  protected int firstInCache()
  {
    for (int i = 0; i < maxSize; i++)
      if (inUse.get(i))
        return i;
    return 0;
  }

  protected int lastInCache()
  {
    for (int i = maxSize - 1; i >= 0; i--)
      if (inUse.get(i))
        return i;
    return 0;
  }

  /**
   * Set the read direction.
   * @since version 2.0
   */
  public void setDirection(int newDirection)
    {
      int actual = getDirection();
      if (newDirection != actual)
        {
          if (DEBUG) System.out.println("setDirection " + (newDirection == BufferHints.READ_BACKWARD
                                                           ? "backwards"
                                                           : "forwards"));
          super.setDirection(newDirection);
          secondaryCache.useBufferHints(new BufferHints(BufferHints.READ_DIRECTION,
                                                        getDirection(), 0));
        }
    }

  /**
   * Return the values associated with the <var>keys</var>.
   * If all the values for the <var>keys</var> are in the local cache,
   * return them. If one or more items are not in the local cache,
   * access them from the secondary cache.
   * @param keys an array of value keys
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result, i.e. an
   * array whose length is less than <code>keys.length</code>
   * @return an array of values from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long[], boolean)
   */
   public synchronized Object[] getFromCache(long keys[], boolean readFully) throws CacheException
     {
       int nFound = 0;
       int nKeys = keys.length;
       Object result[];
       if (nKeys == 0) return nullObjectArray(); // special case: see division by nKeys below
       // loop over all keys, saving items that exist in the cache,
       // counting those not in the cache.
       int blockStart = 0, blockEnd = -1;
       boolean haveSubBlock = false;
       int nMisses = 0;
       for (int i = 0; i < nKeys; i++)
         {
           int index = index(keys[i]);
           if ( inCache(index))
             nFound++;
           else
             {
               nMisses ++;
               if ( ! haveSubBlock )
                 {
                   // mark the beginning of the first subblock of cache misses
                   blockStart = i;
                   blockEnd = i; // 0; saskxn: should be i to be continous block
                   haveSubBlock = true;
                 }
               else if (i == blockEnd + 1)
                 // this cache miss is contiguous with the subblock, so extend the subblock
                 blockEnd = i;
             }
         }

       // if some items not found, then loop again, and fetch those values
       // from the secondary cache, and also return those
       // values in the result. These individual item fetches
       // may (should) perform lookahead fetches.

       boolean surpassedThreshold;
       haveSubBlock &= ((blockEnd - blockStart + 1) == nMisses) && (nMisses > 1) && (nMisses < nKeys);
       if (nMisses > 0 && !haveSubBlock)
         {
           // See if we've surpassed the threshold but don't have
           // a simple sequential subblock. If so, reread the entire set.
           if (thresholdType == THRESHOLD_ABSOLUTE)
             surpassedThreshold = (nKeys - nFound) > threshold;
           else // (thresholdType == THRESHOLD_PERCENT)
             surpassedThreshold = (100 * (nKeys - nFound) / nKeys) > threshold;

           if (surpassedThreshold)
             {
               if (DEBUG) System.out.println("secondaryCache.getFromCache(" + toString(keys) + "," + readFully + ")");
               result = secondaryCache.getFromCache(keys, readFully);
               lastFetchStart = keys[0];
               lastFetchEnd   = keys[keys.length - 1];
               storeInCache(keys, result);
               if (DEBUG)
                 for (int i=0; i < result.length; i++)
                   System.out.println("result[i]0-----------------------" +result[i]);/*I18nOK:LINE*/
               return result;
             }
         }
       else
         {
         lastFetchStart = keys[0];
         lastFetchEnd   = keys[keys.length - 1];
	     }


       // Else, threshold was not surpassed. Try subblock or individual reads.
       // First, get all values that are already in the cache,
       // since doing secondary reads for the missed values may invalidate
       // some items currently in the cache.
       result = new Object[keys.length];
       for (int i = 0, len = keys.length; i < len; i++)
         {
           int index = index(keys[i]);
           if ( inCache(index) )
             result[i] = values[index];
         }
       if (DEBUG)
         for (int i=0; i < result.length; i++)
           System.out.println("result[i]1-----------------------" +result[i]);/*I18nOK:LINE*/
       if (nFound == keys.length)
         return result;

       // now go back and read missed values.
       if (haveSubBlock)
         {
           long subKeys[] = new long[ nMisses ];
           System.arraycopy(keys, blockStart, subKeys, 0, nMisses);
           Object subResult[] = getFromCache(subKeys, readFully);
           if (subResult.length < nMisses) // may get partial result, if readFully is false
             result = Util.resizeArray(result, blockStart + subResult.length);
           System.arraycopy(subResult, 0, result, blockStart, subResult.length);
         }
       else // Don't have a simple subblock. Re-read individual items.
         {
           int i = 0;
           // This uses getFromSecondaryCache() which will
           // try lookahead block reads (which may invalidate
           // some cache values, and read/add a block of values)
           try
             {
               for (int len = keys.length; i < len; i++)
                 {
                   int index = index(keys[i]);
                   if ( ! inCache(index) )
                     result[i] = getFromSecondaryCache(keys[i]);
                 }
             }
           catch (CacheException e)
             {
               if ( readFully )
                 throw e;
               result = Util.resizeArray(result, i - 1);
             }
         }
       if (DEBUG)
         for (int i=0; i < result.length; i++)
           System.out.println("result[i]2-----------------------" +result[i]);/*I18nOK:LINE*/
       return result;
     }

  /**
   * Gets and returns a set of values associated with the range of keys.
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result, i.e. an
   * array whose length is less than <code>keys.length</code>
   * @return an array of values from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long, long, boolean)
   */
   public synchronized Object[] getFromCache(long startKey, long endKey, boolean readFully) throws CacheException
     {
       if (DEBUG)
         System.out.println(" primaryCache: startKey:" +startKey+ " endKey:" +endKey);/*I18nOK:LINE*/
       // one requested item, and it's in the cache? return it without
       // overhead of computing cache keys.
       if (startKey == endKey)
         {
           int index = index(startKey);
           if ( inCache(index) )
             return new Object[] {values[index]};
         }

       // Make sure we do a sufficient lookahead, if possible.
       // However, don't try extra look-ahead if readFully is true,
       // because the look-ahead may cause the entire read to fail
       // when a read without look-ahead might not fail.
       int size = readFully ? 0 : lookAheadSize;
       long keys[] = computeCacheKeyRange(startKey, endKey, size);
       // Did we get back the exact keys we asked for?
//       System.out.println("size = " + size);
//       System.out.println("keys.length = " + keys.length);
       boolean same = true;
		  long s = startKey;

       int len = keys.length;
       for (int i = 0; same && (i < len); i++, s++)
         {
           same = (keys[i] == s);
          if (DEBUG)
             System.out.println("keys[i]:" +keys[i]+ " i:" +i+ " s:" +s);/*I18nOK:LINE*/
         }
       Object results[] = getFromCache(keys, readFully);


       if (same && len > 0 && keys[len-1] >= endKey)
         {
         return results;
	     }

       // ok, we've been had. The secondary cache is not playing by the
       // book and computes cache keys slightly differently (for instance,
       // the MD table computes *all* keys to cache the entire table!)
       // That's ok, we'll adjust.

       // The secondary cache should have returned an array of keys
       // which is a superset of the requested keys.
       boolean found = false;
       //long seqKeys[] = computeSequentialKeyRange(startKey, endKey, 0);
       //startKey = seqKeys[0];
       len = keys.length;//seqKeys.length;
       int i;
       for (i = 0; (i < len); i++)
         if (found = (keys[i] == startKey))	//(seqKeys[i] == startKey))
           break;

       if (found)
		 {
         // a subarray was found. We assume sequential values from here.
         // i is the index of the start key within the array of keys.
         // len is the number of items asked for, so i+len would be (one past) the
         // the end of the subarray. However, the results may be les than
         // that.

         return (Object[]) Util.copyArray(results, i, Math.min(results.length, i + len));
		 }
       else
         // Really messed up. This should probably be a CacheException,
         // but I'll try to be forgiving.
         //return (seqKeys, readFully);
         throw new CacheException(Message.format(RB.getStringResource("keyNotInCache.ex.txt"), new Long(startKey)));
     }


  /**
   * Computes the complete set of keys between the <var>startKey</var> key and the
   * <var>endKey</var> key, inclusive, by calling <code>secondaryCache.computeCacheKeyRange(startKey, endKey, size)</code>
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param size the number of items to read if a look-ahead read is possible.
   * @return an array of keys for the items between the start and end keys,
   *         inclusive. The array may contain fewer items if the secondary
   *         cache does not have the items.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#computeCacheKeyRange(long, long, int)
   */
   public synchronized long[] computeCacheKeyRange(long start, long end, int size) throws CacheException
     {
       inferReadDirection(start);
       if (DEBUG) System.out.println("secondaryCache.computeCachekeyRange(" + start + ", " + end + ", " + size + ")");
       return secondaryCache.computeCacheKeyRange(start, end, size);
     }

  public String toString(long keys[])
    {
      StringBuffer b = new StringBuffer();
      b.append("{");
      for (int i = 0, keysLen = keys.length; i < keysLen; i++)
        {
          long key = keys[i];
          b.append(key);
          b.append(",");
        }
      b.append("}");
      return b.toString();/*I18nOK:MTH*/
    }

  /**
   * Computes the complete set of <var>size</var> keys starting
   * or ending at <var>key</var>, depending on the <var>direction</var>.
   * @param key the key for the first or last item to read from the cache.
   * @param size the number of items to read if a look-ahead read is possible.
   * @param direction either BufferHints.READ_FORWARD or BufferHints.READ_BACKWARD.
   *        If a forward read, the returned keys will range from
   *        <var>key</var> to <var>key</var> + <var>size</var> - 1,
   *        else if reading backwards, the returned keys will range from
   *        <var>key</var> - <var>size</var> + 1 to <var>key</var>,
   * @return an array of up to <var>size</var>keys, either starting or
   *         ending at <var>key</var>, depending on
   *         <var>direction</var>
   *         The array may contain fewer items if the secondary
   *         cache does not have the items.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @since version 2.0
   */
   public static synchronized long[] computeDirectionalCacheKeyRange(long key, int size, int direction) throws CacheException
     {
//       /*if (DEBUG) */System.out.println("computeDirectionalCacheKeyRange(" + key + ", " + size + ", " + direction + ")");
       if (direction == BufferHints.READ_BACKWARD)
         return computeSequentialKeyRange(key - size + 1, key, size);
       else
         return computeSequentialKeyRange(key, key + size - 1, size);
     }

  /**
   * Computes the complete set of keys between the <var>startKey</var> key and the
   * <var>endKey</var> key, inclusive.
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param size the number of items to read if a look-ahead read is possible.
   * @return an array of keys for the items between the start and end keys,
   * inclusive.
   */
   public static long[] computeSequentialKeyRange(long start, long end, int size)
     {
       int nKeys = Math.max((int) (end - start) + 1, size);
       long keys[] = new long[nKeys];
       for (int i = nKeys - 1; i >= 0; i--)
         keys[i] = start + i;
       return keys;
     }

  /**
   * Computes the complete set of keys between the <var>startKey</var> key and the
   * <var>endKey</var> key, inclusive.
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @return an array of keys for the items between the start and end keys,
   * inclusive.
   */
   public static long[] computeSequentialKeyRange(long start, long end)
     {
       return computeSequentialKeyRange(start, end, 0);
     }

  /**
   * Returns an array containing the keys from the cache.
   * @return an array containing the keys for all items in the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized long[] getCacheKeys() throws CacheException
     {
       int n = 0;
       for (int index = size() - 1; index >= 0; index--)
         {
           if (inUse.get(index))
             n++;
         }
       long keys[] = new long[n];
       for (int index = 0, size = size(), j = 0; index < size; index++)
         {
           if (inUse.get(index))
             keys[j++] = key(index);
         }
       return keys;
     }

  /**
   * Returns <code><b>true</b></code> if the cache contains a value for
   * the key <var>key</var>, <code><b>false</b></code> otherwise.
   * @param key a potential key for an item in the cache.
   * @return <code><b>true</b></code> if the cache contains a value for
   * the key <var>key</var>, <code><b>false</b></code> otherwise.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized boolean cacheContainsKey(long key) throws CacheException
     {
       return inCache(index(key));
     }

  /**
   * Removes all items from the cache.
   */
   public synchronized void invalidateCache()
     {
       for (int index = size() - 1; index >= 0; index--)
         {
           inUse.clear(index);
           values[index] = null;
         }
       count = 0;
     }

  /**
   * removes from the cache the item associated with the key <var>key</var>.
   * @param key the key for the item to remove from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized void invalidateCache(long key) throws CacheException
     {
       int index = index(key);
       if ( inUse.get(index))
         {
           count--;
           inUse.clear(index);
           values[index] = null;
         }
     }

  /**
   * Stores the value <var>value</var> in the cache so that it may be
   * accessed by the key <var>key</var>.
   * @param key the item's key
   * @param value the value to store in the cache; it is associated with
   * the key. If a previous value is in the cache for this key,
   * it is dropped from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized void storeInCache(long key, Object value) throws CacheException
     {
       int index = index(key);
       int s = size();
       if ((index < 0) || (index >= s)) // key is outside the range of current cache keys
         {
           accommodateKeys(key - s/2, key);
           index = index(key);  // must recompute the index after shifting the cache.
         }
       storeInCacheQuickly(index, value);
     }

   /**
    * Ensure the cache can accommodate keys between a range.
    * The range of values to accommodate must be <= count.
    * @param startKey the first key in the range
    * @param endKey the last key in the range
    * @exception com.sas.models.CacheException
    * if range is greater than count. (It's a situation that will
    * cause thrashing.)
    */
   protected void accommodateKeys(long startKey, long endKey) throws CacheException
     {
       long startIndex = index(startKey);
       long endIndex = index(endKey);
       int size = size();
       if (startIndex >= 0 && endIndex < size)
         return;

       // This should not be necessary: all calls to accommodateKeys
       // in this source file ensure their startKey/endkey are valid.
       // But a subclass may violate that, so leave this in to be safe.
       if ((endKey - startKey) >= size)
         throw new CacheException(Message.format(RB.getStringResource("cacheOverflow.ex.txt"), new Long(endKey - startKey)));

       // This combines an arraycopy and a BitSet copy, shifting
       // from the current inUse BitSet and values array
       // to the new BitSet and new values array.
       // Note: there is a heavy bias towards increasing sequential keys:
       // shift so the startKey is the first key of the new cache.
       long firstKey = key(0); // firstKey is the actual key of first item (index == 0) in the cache.

       // test if there is any overlap between the existing and the requested cache key range :
       //      existing cache key range is [firstKey..firstKey+maxSize]
       //      requested cache key range is [startKey..endKey]
       // existing range [firstKey..firstKey+maxSize] is less than all keys in new range?
       // Or all keys in new range are less than existing key range [firstKey..firstKey+maxSize] ?
       boolean noOverlap = (firstKey + maxSize < startKey) || (endKey < firstKey);

       // set up limits and parameters for the loop, based on
       // current cache key range and the desired new cache key range
       int delta = (int) Math.floor( ( maxSize - (endKey - startKey + 1) ) / 2 );
       int from;   // index in existing Object array we are copying from
       int to;     // index in new Object array we are copying to.
       int n;      // number of items to move
       long newOffset;

       if ( noOverlap )
         { // no overlap from old cache contents to new cache key range.
           invalidateCache();
           offset = -startKey;
           return;
         }
       else if (startKey > firstKey)
         {
           // Shifting the cache to the right in the range of indices.
           //
           // Definition of the pictograph below:
           // C represents the range of keys in the current cache
           // F represents the range of keys we are fetching/trying to accommodate
           // X are items to be copied from the old cache (preserve some values for backward scrolling)
           // N represents the new cache which contains the X and F values
           //
           //    firstKey
           //    -offset      from            lastKey
           //    |            |               |
           //    CCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
           //                 XXXXXXX
           //
           //                        startKey        endKey
           //                        |               |
           //                        FFFFFFFFFFFFFFFFF
           //                 |------| == delta
           //                 XXXXXXX
           //                 NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN
           //                 |
           //                 -newOffset
           //
           // For example: offset = -40; fistKey = 40; maxSize = 60
           // startKey = 90, endKey = 110
           // delta = ( 60 - (110 - 90 + 1) ) / 2 == 19
           from       = (int) (index(startKey) - delta); // for example, index(90) - 19 == (90+-40) - 19 == 31
           to         = 0;                               // for example, 0
           n          = delta;                           // for example, 19
           newOffset = offset - from;                    // for example, -40 - 19 == -59
           // so we'll move 19 items starting at index 31, to the new Object[] starting at 0.
         }
       else
         {
           // Shifting the cache to the left:
           //
           //                        firstKey
           //                        -offset                      lastKey
           //                        |                            |
           //                        CCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
           //                             XXXXXX
           //                             |
           //                             from
           //
           //            startKey        endKey
           //            |               |
           //            FFFFFFFFFFFFFFFFF
           //    |-------| == delta       XXXXXX
           //    NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN
           //    |
           //    -newOffset
           //
           // For example: offset = -100; fistKey = 100; maxSize = 60
           // startKey = 90, endKey = 110
           // delta = ( 60 - (110 - 90 + 1) ) / 2 == 19
           from  = (int) (endKey - firstKey + 1);                   // for example, 110 - 100 + 1 == 11
           to    = (int) (delta + (endKey - startKey) + 1);         // for example, 19 + (110 - 90) + 1 == 40
           n     = (int) (maxSize - (endKey - startKey + 1) - delta);   // for example, 60 - (110 - 90 + 1) - 19 = 18
           newOffset = (int) (delta - startKey);                    // for example, 19 - 90 == -71
           // for example, move 19 items from position 11 to the new Object[] array starting at index 41
         }

       // TO-DO: Rework this so I don't reallocate a new Object[] I'd
       // like to reuse the BitSet as well, but the API does not allow
       // it. We could write our own that adds shift operations.
       // Still, it not bad that we go through storeInCacheQuickly
       if (DEBUG)
         {
           System.out.println("accommodateKeys(startKey=" + startKey + ", endKey=" + endKey + ")");/*I18nOK:LINE*/
           System.out.println("accommodateKeys: offset was " + offset + ", becomes " + newOffset + ", size = " + maxSize);/*I18nOK:LINE*/
           System.out.println("accommodateKeys: move " + n + " items from " + from + " to " + to);/*I18nOK:LINE*/
         }
       BitSet oldBitSet = inUse;
       Object oldValues[] = values;
       offset = newOffset;
       inUse  = new BitSet(size);
       values = new Object[size];
       for (; n > 0; from++, to++, n--)
         if ( oldBitSet.get(from) )
           {
             if (DEBUG) System.out.println("accommodateKeys:\tmove [" + oldValues[from] + "] from " + from + " to " + to);/*I18nOK:LINE*/
             storeInCacheQuickly(to, oldValues[from]);
           }
     }

  /**
   * Store a value in the cache without testing the
   * cache boundaries. Use this only after calling accommodateKeys
   * @param index the value's index, assumed to be in the valid
   * If there is not already an item at that index, the item count
   * is incremented.
   * range of [0..maxSize-1]
   * @param value to store at that index.
   */
   protected void storeInCacheQuickly(int index, Object value) throws CacheException
     {
       if ( ! inUse.get(index))
         count++;
       inUse.set(index);
       values[index] = value;
     }

  /**
   * Stores the values in the cache so that they may be
   * accessed by the keys.  If <var>keys</var><code>.length&nbsp;&gt;&nbsp;size()</code>,
   * then only the first <var>n</var> items are stored.
   * @param keys an array of item keys
   * @param newValues an array of values the value to store in the cache;
   * each <code>newValues[i]</code> is associated with <code>key[i]</code>.
   * If a previous value is in the cache for this key,
   * it is dropped from the cache. If the array sizes exceed the <var>maxSize</var>,
   * only <var>maxSize</var> items are stored.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   protected synchronized void storeInCache(long keys[], Object newValues[]) throws CacheException
     {
       // Assume we want the last n items
       if (newValues.length == 0)
         return;

       int  firstIndex;
       int  nKeys      = newValues.length;
       int  lastIndex  = nKeys - 1;
       long lastKey    = keys[nKeys-1];
       if (direction == BufferHints.READ_BACKWARD) // assume BufferHints.READ_FORWARD unless different.
         {
           // find the first index firstIndex such that all values between
           // keys[firstIndex] and keys[newValues.length] will fit in the cache;
           // we have to ignore others.  We must do this to find a valid
           // range of keys before calling accommodateKeys.
           long minKey = lastKey - size() + 1;
           firstIndex = nKeys-1;
           for (int index = nKeys-2; index >= 0; index--)
             {
               if (keys[index] > lastKey || keys[index] < minKey)
                 break;
               firstIndex = index;
             }
           accommodateKeys(keys[firstIndex], lastKey);
           for (int index = firstIndex; index < nKeys; index++)
             storeInCacheQuickly( index(keys[index]), newValues[index] );

         }
       else // default is BufferHints.READ_FORWARD
         {
           // Scan for the maximum number of key/value pairs to add,
           // starting at the beginning of the array of keys,
           // but ensure sequential keys within the range
           long firstKey = keys[0];
           long maxKey   = firstKey + size() - 1;
           lastIndex     = 0;
           firstIndex    = 0;
           for (int index = 1;
                (index < keys.length) && (keys[index] <= lastKey) && (keys[index] <= maxKey);
                index++)
             lastIndex = index;
           accommodateKeys(keys[0], keys[lastIndex]);
           for (int index = firstIndex; index <= lastIndex; index++)
             storeInCacheQuickly( index(keys[index]), newValues[index] );
         }
     }

   /**
    * Ask the local data set object to use a set of buffering hints.
    * Use this to configure the number of rows to read ahead
    * when doing buffered reads, or how many rows to maintain
    * in the local client-side cache.
    * <p>
    * The DataSet object
    * currently supports only caching of sequential reads,
    * and row-based caching.
    * @param hints a (linked list) of buffer hints.
    */
    public void useBufferHints(BufferHints hints)
    {
      while (hints != null)
        {
          try
            {
              // we don't handle WRITE_ACESS buffering, just READ
              if ((hints.how & BufferHints.READ_ACCESS) == BufferHints.READ_ACCESS)
                {
                  if (hints.what == BufferHintsInterface.CACHE_SIZE)
                    configureCache(hints.count, -1, -1, -1, -1);
                  else if (hints.what == BufferHintsInterface.LOOKAHEAD_SIZE)
                    configureCache(-1, hints.count, -1, -1, -1);
                }
              super.useBufferHints(hints);
            }
          catch (CacheException ce) { /* IGNORE EXCEPTION */ }

          hints = hints.moreHints;
        }
    }

  /**
   * Configures the cache size, buffering factor and how the cache
   * responds to multiple-item read requests when more than one item
   * in the request is not in the cache and the cache must read values
   * from the sondary cache.
   * @param maxCacheSize the max cache size.
   * @param bufferLookAhead the look-ahead buffer size
   *        <strong>Note:</strong> you will probably get poor performance
   *        if <var>bufferLookAhead</var> is larger than
   *        <var>maxCacheSize</var>
   * @param thresholdType the threshold type
   * @param thresholdValue the value for the threshold.
   * @param accessType a hint about how data will be read and accessed; ignored.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#configureCache(int, int, int, int, int)
   */
  public synchronized void configureCache(int maxCacheSize, int bufferLookAhead,
                                          int thresholdType, int thresholdValue,
                                          int accessType) throws CacheException
    {
      if (   (maxCacheSize > values.length)
          || ((maxCacheSize - values.length) > values.length/2))
        {
          BitSet newBitSet = new BitSet(maxCacheSize);
          int newCount = 0;
          // Hmm, which values to keep from the old cache?
          // Choose simplest and leats error prone: those
          // with indices 0..min(oldSize, newSize)
          for (int i = 0, max = Math.min(maxSize, maxCacheSize); i < max; i++)
            if (inUse.get(i))
              {
                newCount++;
                newBitSet.set(i);
              }
          values = Util.resizeArray(values, maxCacheSize);
          inUse  = newBitSet;
          count  = newCount;
        }
      super.configureCache(maxCacheSize, bufferLookAhead,
                           thresholdType, thresholdValue,
                           accessType);
    }

}
