// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;

import java.util.BitSet;
import java.util.NoSuchElementException;

import com.sas.util.Util;
import com.sas.text.Message;
import com.sas.collection.LongKeyDictionary;

/**
 * A Cache implementation which keeps items with keys that may have
 * random didtsributions. That is, keys may not be related to each other
 * in any way.
 * <p>
 * The RandomCache may be constructed with CacheMonitorInterface, which
 * allows an outside agent (usually the object maintaining the cache)
 * to monitor the cache and be notified when the cache removes elements.
 * @see com.sas.models.SequentialCache
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public class SparseCache
       extends CacheBase
{
  public static final String RB_KEY = "SparseCache.";

  /**
   * FIFO queue of object id's, for invalidating when the cache fills up.
   */
  protected LongQueue idQueue;

  /**
   * Dictionary (like a hashtable) which maps long int ids to object values.
   */
  protected LongKeyDictionary dictionary;

  /**
   * the object monitoring the queue.
   */
  protected CacheMonitorInterface monitor;

  /**
   * Construct a SparseCache that uses default values.
   * Note that you should assign a secondary cache before
   * trying to access data.
   */
  public SparseCache()
    {
      this(null, null);
    }

  /**
   * Construct a SparseCache that uses default values
   * and a secondary cache.
   * @param secondaryCache the secondary cache this cache references on cache misses
   */
  public SparseCache(StaticCacheInterface secondaryCache)
    {
      this(secondaryCache, null);
    }

  /**
   * Construct a SparseCache that uses default values,
   * a secondary cache, and a cache monitor
   * @param secondaryCache the secondary cache this cache references on cache misses
   * @param monitor an object which monitors cache item removal
   *
   */
  public SparseCache(StaticCacheInterface secondaryCache, CacheMonitorInterface monitor)
    {
      this(secondaryCache, monitor,
           DEFAULT_MAX_SIZE, DEFAULT_LOOKAHEAD_SIZE,
           DEFAULT_THRESHOLD_TYPE, DEFAULT_THRESHOLD_VALUE);
    }


  /**
   * Construct a SparseCache that uses default values
   * and a secondary cache.
   * @param secondaryCache the secondary cache this cache references on cache misses
   * @param monitor an object which monitors cache item removal
   * @param maxSize the cache size; the cache will store up to <var>maxSize</var> elements
   * @param lookAheadSize the size of the look-ahead buffer
   */
  public SparseCache(StaticCacheInterface secondaryCache, CacheMonitorInterface monitor,
                         int maxSize, int lookAheadSize)
    {
      this(secondaryCache, null, maxSize, lookAheadSize, DEFAULT_THRESHOLD_TYPE, DEFAULT_THRESHOLD_VALUE);
      this.monitor = monitor;
    }

  /**
   * Construct a SparseCache that uses default values
   * and a secondary cache.
   * @param secondaryCache the secondary cache this cache references on cache misses
   * @param monitor an object which monitors cache item removal
   * @param maxSize the cache size; the cache will store up to <var>maxSize</var> elements
   * @param lookAheadSize the size of the look-ahead buffer
   * @param thresholdType the threshold type
   * @param thresholdValue the value for the threshold.
   */
  public SparseCache(StaticCacheInterface secondaryCache, CacheMonitorInterface monitor,
                         int maxSize, int lookAheadSize, int thresholdType, int thresholdValue)
    {
      super(secondaryCache, maxSize, lookAheadSize, maxSize, lookAheadSize);
      this.monitor = monitor;
      dictionary = new LongKeyDictionary(maxSize, 0.75f);
      idQueue = new LongQueue(maxSize);
    }


  /**
   * Set the object which monitors items removed from the cache.
   * @param newMonitor the new cache monitor
   * @exception java.lang.IllegalStateException if the cache
   * already has a monitor and newMonitor is not null.
   * The cache only supports one monitor at a time, so
   * you should call <code>setMonitor(null)</code>
   * before replacing the monitor.
   */
   public void setMonitor(CacheMonitorInterface newMonitor)
   {
     if (monitor != null && newMonitor != null)
        throw new IllegalStateException(RB.getStringResource(RB_KEY, "oneMonitor.ex.txt"));
     monitor = newMonitor;
   }

   /**
    * Return the object currently monitoring the cache.
    * @return the current cache monitor
    */
    public CacheMonitorInterface getMonitor()
    {
        return monitor;
    }

  /**
   * Check if the cache contains an item at the internal index.
   * @param key the item key
   */
   protected final boolean inCache(long key)
     {
       return dictionary.containsKey(key);
     }

  /**
   * Returns an item from the cache, based on a cache key <var>key</var>.
   * @param key the item's key
   * @return the item associated with the key.
   * @exception com.sas.models.CacheException if the cache or its secondary
   * cache does not contain the item.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long)
   */
   public synchronized Object getFromCache(long key) throws CacheException
     {
       if (inCache(key))
         return dictionary.get(key);
       else
         return getFromSecondaryCache(key);
     }

  /**
   * Get a value from the secondary cache, trying a lookahead read.
   * This has a side effect of reading a look-ahead block of data from the
   * secondary cache, if lookahead is enabled, and storing
   * those values in this cache (which in turn may remove
   * some other items already in the cache.)
   * @param key a cache key
   * @return the value from the secondary cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
  protected synchronized Object getFromSecondaryCache(long key) throws CacheException
    {
      if (lookAheadEnabled())
        {
          boolean disabled = false;
          try
            {
              disabled = disableLookAhead();
              // perform lookahead.
              long keys[] = computeCacheKeyRange(key, key, lookAheadSize);
              Object newValues[] = secondaryCache.getFromCache(keys, false);
              storeInCache(keys, newValues);

              // expect first item in result to match, but that depends
              // on behavior of computeCacheKeyRange, so check all values to make sure.
              for (int i = 0, nKeys = Math.min(keys.length, newValues.length); i < nKeys; i++)
                if (keys[i] == key)
                  return newValues[i];
              // Hmm, secondary cache failed to return a value for this key.
              throw new CacheException(Message.format(RB.getStringResource("keyNotInCache.ex.txt"), new Long(key)));
            }
          finally
            {
              if (disabled)
                enableLookAhead();
            }
        }
      else
        {
          Object newValue = secondaryCache.getFromCache(key);
          storeInCache(key, newValue);
          return newValue;
        }
    }

  /**
   * Return the values associated with the <var>keys</var>.
   * If all the values for the <var>keys</var> are in the local cache,
   * return them. If one or more items are not in the local cache,
   * access them from the secondary cache.
   * @param keys an array of value keys
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result, i.e. an
   * array whose length is less than <code>keys.length</code>
   * @return an array of values from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long[], boolean)
   */
   public synchronized Object[] getFromCache(long keys[], boolean readFully) throws CacheException
     {
       int nFound = 0;
       int nKeys = keys.length;
       Object result[];
       if (nKeys == 0) return nullObjectArray(); // special case: see division by nKeys below
       // loop over all keys, saving items that exist in the cache,
       // counting those not in the cache.
       for (int i = 0; i < nKeys; i++)
         {
           long key = keys[i];
           if ( inCache(key) )
             nFound++;
         }

       // if some items not found, then loop again, and fetch those values
       // from the secondary cache, and also return those
       // values in the result. These individual item fetches
       // may (should) perform lookahead fetches.

       boolean surpassedThreshold;
       int nMissing = (int) nKeys - nFound;
       if (thresholdType == THRESHOLD_ABSOLUTE)
         surpassedThreshold = nMissing > threshold;
       else // (thresholdType == THRESHOLD_PERCENT)
         surpassedThreshold = (100 * nMissing / nKeys) > threshold;

       if (surpassedThreshold)
         {
           result = secondaryCache.getFromCache(keys, readFully);
           storeInCache(keys, result);
           return result;
         }

       // else, threshold not surpassed. Try individual reads.
       // first, get all values already in the cache,
       // since doing secondary reads may invalidate
       // some items currently in the cache.
       result = new Object[keys.length];
       for (int i = 0, len = keys.length; i < len; i++)
         {
           long key = keys[i];
           if ( inCache(key) )
             result[i] = dictionary.get(key);
         }

       if (nFound == keys.length)
         return result;

       // now go back and read missed values with a subread
       long subKeys[] = new long[nMissing];
       int indices[] = new int[nMissing];
       Object subResult[]  = new Object[nMissing];
       //fix for S0190663 - array index out of bounds
       //don't increment j in the for loop statement, and check that j < nMissing before setting subKeys[j] and indices[j]
       //because nMissing is the length of both those arrays
       for (int i = 0, len = keys.length, j = 0; i < len; i++)
         {
           long key = keys[i];
           if ( ! inCache(key) && j < nMissing)
             {
               subKeys[j] = key;
               indices[j] = i;
               j++;
             }
         }
       subResult = secondaryCache.getFromCache(subKeys, readFully);
       storeInCache(subKeys, subResult);
       for (int j = 0; j < nMissing; j++)
         {
           result[ indices[j] ] = subResult[j];
         }
       return result;
     }

  /**
   * Gets and returns a set of values associated with the range of keys.
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result, i.e. an
   * array whose length is less than <code>keys.length</code>
   * @return an array of values from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#getFromCache(long, long, boolean)
   */
   public synchronized Object[] getFromCache(long startKey, long endKey, boolean readFully) throws CacheException
     {
       // one requested item, and it's in the cache? return it without
       // overhead of computing cache keys.
       if ( (startKey == endKey) && inCache(startKey) )
         return new Object[] { dictionary.get(startKey) };

       // Make sure we do a sufficient lookahead, if possible.
       // However, don't try extra look-ahead if readFully is true,
       // because the look-ahead may cause the entire read to fail
       // when a read without look-ahead might not fail.
       long keys[] = computeCacheKeyRange(startKey, endKey, readFully ? 0 : lookAheadSize);
       Object newValues[] = null;
       try {
         if (keys[0] == startKey && keys[keys.length-1] == endKey)
           {
             newValues = getFromCache(keys, readFully);
             return newValues;
           }
         int start = find(startKey, keys, 0);
         int end = find(endKey, keys, start);
         if (start == -1 || end == -1)
           {
             long newKeys[] = computeCacheKeyRange(startKey, endKey, 0);
             newValues = getFromCache(newKeys, readFully);
           }
         else
           {
             Object subArray[] = new Object[ end - start + 1 ];
			     newValues = getFromCache(keys, readFully);
             System.arraycopy(newValues, start, subArray, 0, end - start + 1);
             newValues = subArray;
           }
       }
//       catch (ArrayIndexOutOfBoundsException e)
		catch (Exception e)
        { throw new CacheException(e.getMessage()); }

       return newValues;

//       int nKeys = keys.length;
//       Object newValues[] = new Object[nKeys];
//       int i = 0;
//       for (; i < nKeys; i++)
//         {
//           try
//             {
//               newValues[i] = getFromCache(keys[i]);
//             }
//           catch (CacheException ce)
//             {
//               if (readFully || i == 0)
//                 throw ce;
//               // else, partial result. We failed on item i, so
//               // values 0 through i-1 are valid; there are i of them
//               newValues = Util.resizeArray(newValues, i);
//               return newValues;
//             }
//         }
//       return newValues;

     }

  private int find(long key, long keys[], int start)
    {
      for (int i = Math.max(start, 0), len = keys.length; i < len; i++)
        if (keys[i] == key)
          return i;
      return -1;
    }

  /**
   * Computes the complete set of keys between the <var>startKey</var> key and the
   * <var>endKey</var> key, inclusive, by calling <code>computeSequentialKeyRange(startKey, endKey, size)</code>
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param size the number of items to read if a look-ahead read is possible.
   * @return an array of keys for the items between the start and end keys,
   * inclusive.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#computeCacheKeyRange(long, long, int)
   */
   public synchronized long[] computeCacheKeyRange(long start, long end, int size) throws CacheException
     {
       return secondaryCache.computeCacheKeyRange(start, end, size);
     }

  /**
   * Returns an array containing the keys from the cache.
   * @return an array containing the keys for all items in the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized long[] getCacheKeys() throws CacheException
     {
       return dictionary.getKeys();
     }

  /**
   * Return an array of the values in the cache
   * in no specified order.
   * @return an array of the values in the table
   */
   public synchronized Object[] getCacheValues()
    {
      return dictionary.getValues();
    }

  /**
   * Returns <code><b>true</b></code> if the cache contains a value for
   * the key <var>key</var>, <code><b>false</b></code> otherwise.
   * @param key a potential key for an item in the cache.
   * @return <code><b>true</b></code> if the cache contains a value for
   * the key <var>key</var>, <code><b>false</b></code> otherwise.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized boolean cacheContainsKey(long key) throws CacheException
     {
       return inCache(key);
     }

  private transient boolean suppressMonitorQuery;

  /**
   * Removes all items from the cache.
   */
   public synchronized void invalidateCache()
     {
       // System.out.println("SparseCache@" + Integer.toHexString(hashCode()) + "invalidateCache()");
       if ( (monitor != null) && ! idQueue.isEmpty() )
         {
           // copy the long queue, in case the monitor
           // keeps adding items back in. This avoids
           // an infinite loop
           LongQueue q = idQueue;
           try { q = (LongQueue) idQueue.clone(); }
           catch (Exception e) {};
           while (!q.isEmpty())
             try
               { invalidateCache(q.dequeue()); }
             catch (CacheException e)
               { }
             // debugging post condition:
             // if ( count != 0 || dictionary.count() != 0)
             //  throw new IllegalStateException();
         }
       // just to be sure.
       count = 0;
       idQueue.removeAll();
       dictionary.removeAll();
     }

  /**
   * Removes from the cache the item associated with the key <var>key</var>.
   * The monitor is not queried before removing the item.
   * @param key the key for the item to remove from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized void invalidateCache(long key) throws CacheException
     {
       // System.out.println("SparseCache@" + Integer.toHexString(hashCode()) + "invalidateCache(" + key + ")");
       if (cacheContainsKey(key))
         {
           removeFromIdQueue(key);
           invalidateCacheQuickly(key);  // decrements count
         }
     }

  /**
   * Remove a key from the idQueue.
   * @param key the key to remove from the idQueue
   */
  protected void removeFromIdQueue(long key)
    {
      for (int i = count; i > 0; i--)
        {
          long k = idQueue.dequeue();
          if (k != key)
            idQueue.queue(k);
        }
    }

  /**
   * Removes from the cache the item associated with the key <var>key</var>.
   * Unlike <code>invalidateCache(key)</code>, this version assumes <var>key</var>
   * is valid and represents an item in the <var>cache</var>.
   * The monitor is not queried before removing the item.
   * @param key the key for the item to remove from the cache.
   *        It is assumed the key has already been removed from the idQueue.
   */
   protected synchronized void invalidateCacheQuickly(long key)
    {
      // System.out.println("SparseCache@" + Integer.toHexString(hashCode()) + "invalidateCache(" + key + ")");
      try
        {
          if (monitor != null)
            {
              Object item = dictionary.get(key);
              dictionary.removeAt(key);
              count--;
              monitor.noteCacheItemRemoved(item, key, this);
            }
          else
            {
              dictionary.removeAt(key);
              count--;
            }
        }
      // should not happen, but just in case:
      catch (NoSuchElementException e) { }
    }

  /**
   * Stores the value <var>value</var> in the cache so that it may be
   * accessed by the key <var>key</var>.
   * @param key the item's key
   * @param value the value to store in the cache; it is associated with
   * the key. If a previous value is in the cache for this key,
   * it is dropped from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   public synchronized void storeInCache(long key, Object value) throws CacheException
     {
       if ( cacheContainsKey(key) )
         {
           if (monitor != null)
             {
               Object item = dictionary.get(key);
               if (item == value)
                 return;
               dictionary.removeAt(key);
               // Leave count as is; item at same key is being replaced.
               // Don't ask monitor if it is ok to remove the item, just notify.
               // we kind of have to remove it on a replace request
               monitor.noteCacheItemRemoved(item, key, this);
             }
           // leave count as is; item at same key is being replaced
           dictionary.set(key, value);
         }
       else
         {
           if (monitor == null && isFull())
             removeOneUnconditionally(idQueue.dequeue());
           else
             // there is a monitor. Find an unremovable item, if one exists,
             // and remove it. Notify the monitor of the full condition;
             // the monitor may clean up or grow the cache or or change object states, and we
             // may be able to find a removable item; if so, remove it.
             // Otherwise, if the monitor says we can unconditionally remove an item,
             // even if it is "unremovable", then remove one.
             while (isFull())
               {
                 int n = findRemovable(true);
                 boolean ignore = false;
                 if (n == 0)
                   // can the cache ignore monitor's responses to CacheMonitorInterface.isCacheItemRemovable() ?
                   ignore = monitor != null && monitor.noteCacheFull(count, this);
                 if (isFull()) // monitor did nothing to clean up/grow the cache
                   {
                     // if there are removable items, move the first to the head of
                     // the queue. But if ignore is true, then the monitor has given
                     // us permission to remove any items.
                     if ( ! ignore && n != 0 )
                       findRemovable(true);
                     removeOneUnconditionally(idQueue.dequeue());
                   }
               }
           dictionary.set(key, value);
           count++;
           idQueue.queue(key);
         }
     }

  /**
   * Remove the item from the cache, without querying the monitor.
   * @param key the key to remove
   * It is assumed the key has already been removed from the idQueue
   */
   private void removeOneUnconditionally(long key)
     {
       boolean prev = suppressMonitorQuery;
       try
         {
           suppressMonitorQuery = true;
           invalidateCacheQuickly(key); // decrements count
         }
       finally
         { suppressMonitorQuery = prev; }
     }

  /**
   * Count/find a removable items.
   * @param stopWhenFound if true, stop counting when a removable
   *        item is found, and leave it at the head of the queue.
   * @return the number of removable items. If stopWhenFound is
   *         true, this will be 1 if a removable item is found, else 0
   */
   private int findRemovable(boolean stopWhenFound)
     {
       int n = 0; // count the number of unremovable items
       for (int i = 0; i < count(); i++)
         {
           long key = idQueue.peek();
           Object item = dictionary.get(key);
           if (monitor == null || monitor.isCacheItemRemovable(item, key, this))
             {
               if (stopWhenFound) return 1;
               n++;
             }
           // dequeue then requeue the id
           idQueue.queue(idQueue.dequeue());
         }
       return n;
     }

  /**
   * Stores the values in the cache so that they may be
   * accessed by the keys
   * @param keys an array of item keys
   * @param newValues an array of values the value to store in the cache;
   * each <code>newValues[i]</code> is associated with <code>key[i]</code>.
   * If a previous value is in the cache for this key,
   * it is dropped from the cache. If the array sizes exceed the <var>maxSize</var>,
   * only <var>maxSize</var> items are stored.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
   protected synchronized void storeInCache(long keys[], Object newValues[]) throws CacheException
     {
       if (newValues.length == 0)
         return;

       int nKeys = keys.length;
       // removed items if necessary. Do this *before* storing
       // the new items, since storing the new items
       // may cause items to be removed... and if the set getting
       // popped off the id queue ovrelaps with the incoming
       // set of new items, we may add an item, only to get it removed
       // when index reaches the end of the array count.
       int size = Math.min(maxSize, keys.length);
       ensureCapacity(size, keys);
       for (int index = 0, count; index < size; index++)
         storeInCache( keys[index], newValues[index] );
     }

  private static final boolean debug = false;

  /**
   * Make sure there is room to store <var>size</var> items.
   * This does not guarantee enought items are removed (the
   * monitor may prevent some from being remove)
   * but it tries to remove enough items so that size
   * items may be put into the cache.
   */
  protected void ensureCapacity(int size, long keys[]) throws CacheException
    {
      if (roomFor(size))
        return;

      if (debug) System.out.println("ensureCapacity kicked in... \n" + status(size) );

      // first, go down the list of keys being added and flush them when necessary
      if (monitor == null)
        for (int i = keys.length - 1; i >= 0 && !roomFor(size); i--)
          {
            long key = keys[i];
            if (cacheContainsKey(key))
              {
                if (debug) System.out.println("ensureCapacity: kick out " + key);
                removeFromIdQueue(key);
                invalidateCacheQuickly(key);
              }
          }
      else
        for (int i = keys.length - 1; i >= 0 && !roomFor(size); i--)
          {
            long key = keys[i];
            if ( cacheContainsKey(key)
                && monitor.isCacheItemRemovable(dictionary.get(key), key, this))
              {
                if (debug) System.out.println("ensureCapacity: kick out " + key);
                removeFromIdQueue(key);
                invalidateCacheQuickly(key);
              }
          }

      // did we free up enough room? if so, we're done
      if (roomFor(size))
        {
          if (debug) System.out.println("ensureCapacity kicked in... first pass successful\n" + status(size));
          return;
        }

      if (debug) System.out.println("ensureCapacity kicked in... after pass 1:\n"  + status(size));

      // try to free up a few more...
      int numToRemove = (count() + size) - maxSize;
      if (monitor != null)
        for (; numToRemove > 0; numToRemove--)
          {
            int n = findRemovable(true);
            if (n != 0)
              {
                long key = idQueue.dequeue();
                if (debug) System.out.println("ensureCapacity: kick out " + key);
                removeOneUnconditionally(key);
              }
          }
      else
        for (; numToRemove > 0; numToRemove--)
          {
            long key = idQueue.dequeue();
            if (debug) System.out.println("ensureCapacity: kick out " + key);
            removeOneUnconditionally(key);
          }

      if (debug)
        if (roomFor(size))
          System.out.println("ensureCapacity kicked in... second pass successful.\n" + status(size));
        else
          System.out.println("ensureCapacity kicked in... second pass failed also\n"
                             + status(size));


      // that's all we can do. cross your fingers if there's
      // still not room. Not a complete disaster; we may
      // thrash a bit
    }

  private String status(int size) throws CacheException
    {
      return "ensure " + size + ", contains " + count() + "/" + maxSize + ", "/*I18nOK:LINE*/
        + (maxSize - count) + " free, need " +/*I18nOK:LINE*/
        (((maxSize - count) < size)
         ? (size - (maxSize - count))
         : 0) + " more, next key in cache: "/*I18nOK:LINE*/
        + (idQueue.isEmpty()
           ? "empty"/*I18nOK:LINE*/
           : idQueue.peek() + "... in cache? " + cacheContainsKey(idQueue.peek())/*I18nOK:LINE*/
           + " monitored? " + (monitor != null));/*I18nOK:LINE*/
    }

  private boolean roomFor(int numItems)
    {
      return (count() + numItems) <= maxSize;
    }

  /**
   * Configures the cache size, buffering factor and how the cache
   * responds to multiple-item read requests when more than one item
   * in the request is not in the cache and the cache must read values
   * from the sondary cache.
   * @param maxCacheSize the max cache size.
   * @param bufferLookAhead the look-ahead buffer size
   * @param thresholdType the threshold type
   * @param thresholdValue the value for the threshold.
   * @param accessType a hint about how data will be read and accessed
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see com.sas.models.StaticCacheInterface#configureCache
   */
  public synchronized void configureCache(int maxCacheSize, int bufferLookAhead,
                                          int thresholdType, int thresholdValue,
                                          int accessType) throws CacheException
    {
      if ((maxCacheSize > -1) && (maxCacheSize != maxSize))
        {
          int numToRemove; // number of items we need to remove from the cache, when shrinking
          if (maxCacheSize >= count())
            numToRemove = 0; // don't have to remove any items
          else
            {
              if (monitor != null)
                {
                  // there is a monitor. Count how many items in the cache
                  // may be removed. From that, we have a minimum number
                  // of items that cannot be removed. Don't set the maxCacheSize
                  // less than the mimimum.
                  int numRemovable = findRemovable(false);
                  int minCacheSize = count() - numRemovable;
                  maxCacheSize = Math.max(minCacheSize, maxCacheSize);
                }
              // only remove n items which will bring the count
              // down to (the possibly adjusted) maxCacheSize
              numToRemove = count() - maxCacheSize;
            }

          if (numToRemove > 0 || maxCacheSize != maxSize)
            {
              LongQueue newQueue = new LongQueue(maxCacheSize);
              LongQueue q = idQueue;
              try { q = (LongQueue) idQueue.clone(); }
              catch (Exception e) {};
              while (!q.isEmpty())
                {
                  boolean remove = (numToRemove > 0);
                  if (remove && monitor != null)
                    {
                      long key = q.peek(); // dequeue below if can remove
                      Object item = dictionary.get(key);
                      remove = monitor.isCacheItemRemovable(item, key, this);
                    }
                  if (remove)
                    {
                      removeOneUnconditionally(q.dequeue());
                      numToRemove--;
                    }
                  else
                    newQueue.queue(q.dequeue());
                }
              idQueue = newQueue;
            }
        }
      super.configureCache(maxCacheSize, bufferLookAhead, thresholdType, thresholdValue, accessType);
    }


}
