// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.models;
import com.sas.util.Countable;

/**
 * An interface for a read-only data cache, where value keys are long integers.
 * Caches are object which hold a limited number of references to
 * other objects, often used for improving performance in distributed
 * applications. For example, a cache may store some items recently
 * read from a remote server.
 * However, there is non-trivial overhead with network communication when the Java client
 * is communicating with the server. The most general purpose (and necessary)
 * data access methods are not the most efficient. It is desirable for the
 * models to buffer and cache data from the remote server. For example, when
 * reading a row from a SAS data set, the model may prefetch
 * <var>n</var> rows (where <var>n</var>&ft;1), so that the access uses only
 * 1 instead of <var>n</var> round trips to the server.  The cache will then
 * store those items locally on the client, and when subsequent requests
 * for the data items are made, the cache can return them if they exist
 * locally, thereby reducing network traffic can be drastically.
 * <p>
 * A cache may have a secondary cache associated with it.
 * The secondary cache can be either an actual data source,
 * (which implements the StaticCacheInterface), or an adapter
 * class which forwards read requests to the data source, or
 * it may be a true second level cache which is attached to a data source.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public interface StaticCacheInterface
       extends Countable,
               BufferHintsInterface
{
  /**
   * Returns an item from the cache, based on a cache key <var>key</var>.
   * @param key the item's key
   * @return the item associated with the key.
   * If the cache does not contain the item
   * but has a secondary cache, the cache will read the item from the secondary
   * cache, store it locally by <var>key</var>, and return it.
   * Otherwise, if the item is not found,
   * throws a CacheException.
   * @exception com.sas.models.CacheException if the cache or its secondary
   * cache does not contain the item.
   */
  public Object getFromCache(long key) throws CacheException;

  /**
   * Return the values associated with the <var>keys</var>.
   * If all the values for the <var>keys</var> are in the local cache,
   * return them. If one or more items are not in the local cache,
   * access them from the secondary cache. The cache has two options
   * in such a case: fetch the missing items individually, or refetch
   * all the items from the secondary cache; this may be more
   * efficient in some cases, especially if the secondary cache is a remote model. If
   * the number of items missing, <var>m</var>, is 1, only that item
   * will be re-read from the secondary cache.
   * Otherwise,
   * the cache configuration will be used to determine whether to
   * read the <var>m</var> missing items from the secondary cache individually, or
   * whether to re-read the entire block of values.
   * @param keys an array of value keys
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result, i.e. an
   * array whose length is less than <code>keys.length</code>
   * @return an array of values from the cache, with values corresponding
   * to each key.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
  public Object[] getFromCache(long keys[], boolean readFully) throws CacheException;

  /**
   * Gets and returns a set of values associated with the range of keys.
   * If the cache has a secondary cache, this calls <code>nKeys =
   * countKeys(long startKey, long endKey)</code> and if <code>nkeys</code> exceeds
   * the maximum cache size, simply calls
   * <code>secondaryCache.getFromCache(startKey, endKey)</code>,
   * else it calls
   * <code>secondaryCache.getFromCache(computeKeys(long startKey, long endKey))</code>.
   * <p>
   * If the cache has no secondary cache, it simply performs
   * <code>getFromCache(computeKeys(long startKey, long endKey))</code>.
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param readFully if true, try to read items for all the keys.
   * If false, getFromCache may return a partial result, i.e. an
   * array whose length is less than <code>(endKey - startKey + 1)</code>
   * @return an array of values from the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
  public Object[] getFromCache(long startKey, long endKey, boolean readFully) throws CacheException;

  /**
   * Computes the complete set of keys between the <var>startKey</var> key and the
   * <var>endKey</var> key, inclusive. If the cache has a secondary SPARSE cache,
   * it defers this call to the secondary cache (which may be a data model
   * such as an MDDB model). Otherwise, (either there is no secondary cache,
   * or it is a SEQUENTIAL cache) the value is the consecutive
   * long integer values from <var>startKey</var> to <var>endKey</var>, inclusive.
   * The value of <var>startKey</var> must be less than or equal to <var>end</var>.
   * <p>
   * If the cache/model supports extrapolating the <var>endKey</var> in order
   * to do look-ahead reads, use <var>size</var> for this extrapolation.
   * For example, a sequential cache may generate the keys for
   * <code>startKey .. max(endKey,startKey+size)</code>, while a sparse
   * cache for a 100 by 100 cell data source may interpret
   * <code>getFromCache(100001, 500020, 200)</code> as a request to generate keys
   * for 100 cells (1,1) through (5,20) plus 100 more cells. It may
   * extrapolate by rows and compute keys for cells (1,1) through (10,20),
   * @param startKey the key for the first item to read from the cache, inclusive
   * @param endKey the key for the last item to read from the cache, inclusive
   * @param size the number of items to read if a look-ahead read is possible.
   * A cache can
   * ignore this parameter; it serves as a hint only and not a requirement.
   * @return an array of keys for the items between the start and end keys,
   * inclusive.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
  public long[] computeCacheKeyRange(long startKey, long endKey, int size) throws CacheException;

  /**
   * Returns an array containing the keys from the cache.
   * @return an array containing the keys for all items in the cache.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
  public long[] getCacheKeys() throws CacheException;

  /**
   * Returns <code><b>true</b></code> if the cache contains a value for
   * the key <var>key</var>, <code><b>false</b></code> otherwise.
   * @param key a potential key for an item in the cache.
   * @return <code><b>true</b></code> if the cache contains a value for
   * the key <var>key</var>, <code><b>false</b></code> otherwise.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   */
  public boolean cacheContainsKey(long key) throws CacheException;

  /**
   * Describes how the cache computes keys within a range.
   * @return an integer describing how the cache interprets a range of keys,
   * one of <CODE>USES_SEQUENTIAL_KEYS</CODE> or <CODE>USES_SPARSE_KEYS</CODE>.
   * @see #USES_SEQUENTIAL_KEYS
   * @see #USES_SPARSE_KEYS
   */
  public int getCacheKeyUse();

  /**
   * Configures the cache size, buffering factor and how the cache
   * responds to multiple-item read requests when more than one item
   * in the request is not in the cache and the cache must read values
   * from the sondary cache.  Not all caches implement all configuration
   * options. For example, some cache implementations cannot change
   * the cache access type from sequential to random.
   * @param maxCacheSize is the
   * maximum number of items the cache will hold at any time; -1 means
   * do not change the current value.
   * @param bufferLookAhead
   * is how many items the cache will read in look-ahead reads from
   * the secondary cache; this must be less that maxCacheSize; -1
   * means do not change the current value.
   * @param thresholdType defines how the cache handles multiple
   * reads when (<var>m</var> &gt; 1) items are not in the cache and it must revert to
   * reading more than one value from the secondary cache.
   * <var>thresholdType</var> is one of:
   * <dl>
   *     <dt>THRESHOLD_ABSOLUTE
   *     <dd>If <code>m >= thresholdValue</code>, that is the number of missing
   *     items exceeds an absolute amount, then re-read the entire set
   *     from the secondary cache.
   *     <dt><code>THRESHOLD_PERCENT</code>
   *     <dd>If the percentage of items
   *     <code>100*<var>m</var>/request.length</code> (a value from from 0 to 100)
   *     exceeds <var>thresholdValue</var>, then re-read the entire set
   *     from the secondary cache.
   *     <dt>-1<dd> means do not change the threshold; <var>thresholdValue</var> is ignored.
   * </dl>
   * @param accessType a hint about how data will be read and accessed,
   * one of the values:
   * <ul>
   * <li>-1 ignore the <var>accessType</var>
   * <li>BufferHints.SEQUENTIAL_ACCESS
   * <li>BufferHints.RANDOM_ACCESS
   * <li>BufferHints.UNKNOWN_ACCESS
   * </ul>
   * @param thresholdValue the value for the threshold.
   * @exception com.sas.models.CacheException if the request cannot
   * be satisfied.
   * @see #THRESHOLD_ABSOLUTE
   * @see #THRESHOLD_PERCENT
   */
  public void configureCache(int maxCacheSize, int bufferLookAhead,
                             int thresholdType, int thresholdValue,
                             int accessType) throws CacheException;


  /**
   * Return a reference to the secondary cache which feeds this one.
   * The secondary cache can be either an actual data source,
   * or a second level cache which is attached to a data source.
   * @return the secondary cache which this cache reads from.
   */
  public StaticCacheInterface getSecondaryCache();

  /**
   * The value <CODE>USES_SEQUENTIAL_KEYS</CODE>
   * means a a request to <code>computeCacheKeyRange(<var>startKey,endKey</var>)</code>
   * should return the trivial result consisting of the consecutive integers
   * from <var>startKey</var> to <var>endKey</var>;
   * @see #getCacheKeyUse()
   */
  public static final int USES_SEQUENTIAL_KEYS = 0;

  /**
   * <CODE>USES_SPARSE_KEYS</CODE>
   * indicates that the cache uses a non-trivial sparse key computation algorithm.
   * If a cache has a secondary cache, it defers this call to the secondary cache.
   * The implementation of the CacheInterface within the MDDB model
   * will use <CODE>USES_SPARSE_KEYS</CODE>, while the implementation
   * within the dataset model will use <CODE>USES_SEQUENTIAL_KEYS</CODE>
   * @see #getCacheKeyUse()
   */
  public static final int USES_SPARSE_KEYS = 1;

  /**
   * A cache configuration value for the threshold type, indicating a threshold
   * percentage (from 1 to 100) beyond which a cache can disregard locally cached items.
   * For example, when calling getFromCache(long[]), if the percentage
   * of keys not found in the local cache exceeds this threshold,
   * the cache will simply re-read <em>all</em> items from
   * its secondary cache.
   */
  public static final int THRESHOLD_PERCENT   = 0;

  /**
   * A cache configuration value for the threshold type, indicating an absolute threshold
   * beyond which a cache can disregard locally cached items.
   * For example, when calling getFromCache(int[]), if the number
   * of keys not found in the local cache exceeds this threshold,
   * the cache will simply re-read <em>all</em> items from
   * its secondary cache.
   */
  public static final int THRESHOLD_ABSOLUTE   = 1;

}
