package com.sas.net;

import com.sas.util.Util;

import java.applet.Applet;
import java.io.File;

/**
 * The AppletResourceLocator is used to locate any type of resources relative to an applet, 
 * whether the resource wanted is an image file or a webEIS document. By default, the applet 
 * the resource is relative to is the applet returned from {@link com.sas.GlobalInfoTable#getApplet() GlobalInfoTable.getApplet()} 
 * and is relative to the code base of the applet. The resource name should be the relative path 
 * and filename of the resource. 
 * <br> 
 * <br>
 * <b><u>Creating a AppletResourceLocator</b></u>
 * <p>
 * A valid instance is created by instantiating an object of AppletResourceLocator followed by setting
 * relative resource name.
 * <pre>
 * AppletResourceLocator appletResourceLocator1 = new AppletResourceLocator();
 * appletResourceLocator1.setResourceName("apple.gif"); //assumes apple.gif is in the applet's direcotry
 * </pre>
 * <p>
 * If the resource is in a sub-directory of from the applet, then you will need to pass in the sub-directory 
 * plus the filename name to the setResourceName call
 * <pre>
 * AppletResourceLocator appletResourceLocator1 = new AppletResourceLocator();
 * appletResourceLocator1.setResourceName("resource\images\apple.gif"); 
 * </pre>
 * If the resource is relative to the document base of the applet, then you will 
 * need to call setDocumentBaseRelative(true).
 * <pre>
 * AppletResourceLocator appletResourceLocator1 = new AppletResourceLocator();
 * appletResourceLocator1.setDocumentBaseRelative(true);
 * appletResourceLocator1.setResourceName("images\apple.gif"); 
 * </pre>
 * <P>
 * A call to getResource() will return the {@link java.net.URL URL} of the resource if it can be found.
 * <pre>
 * try {
 * 	java.net.URL url = appletResourceLocator1.getResource(); 
 * } catch (java.net.MalformedURLException ex)
 * {
 * //an error occurred while forming the URL, handle as needed
 * }
 * </pre>
 * <P>
 * @author Drew Anderson
 * @version 1.0
 * @history
 * <dl>
 * <dt>Sept 18 2000  <dd>Created
 * </dl>
 **/
public class AppletResourceLocator extends BaseResourceLocator implements ResourceLocatorInterface
{
   static final long serialVersionUID = -765029301747160495L;
	/** The applet the resource is relative to, null if the applet cannot be found. **/
	protected Applet _applet;
	/** Whether to find the resource relative to the document base that is set on the applet **/
	protected boolean _isDocumentBaseRelative;
	/** Whether to find the resource relative to the code base that is set on the applet **/
	protected boolean _isCodeBaseRelative;

	/**
	 * Default Constructor. Sets the AppletResourceLocator to look for the resource relative 
	 * to the code base.
	 **/
	public AppletResourceLocator()
	{
		super();
		_isCodeBaseRelative = true;
	}

	/**
	 * Sets the name of the resource. The resourceName should be a realtive path to the resource.
	 * Sets the AppletResourceLocator to look for the resource relative to the code base.
	 * @param resourceName the relative path and name of the resource to locate
	 **/
	public AppletResourceLocator(String resourceName)
	{
		super(resourceName);
		_isCodeBaseRelative = true;
	}

	/**
	 * Returns the applet that the resource is relative to. This method calls 
	 * GlobalInfoTable.getObject() to find the applet, unless the user has set the applet
	 * themselves. 
	 * @return the applet the resource is relative to, by default, it is relative to
	 * GlobalInfoTable.getObject(). The applet is only set on the GlobalInfoTable when the applet
	 * is running. At design-time, the applet will not be set on the GlobalInfoTable.
	 * @see com.sas.GlobalInfoTable#getObject()
	 * @see #setApplet(Applet)
	 **/
	public Applet getApplet()
	{
		if (_applet != null)
			return _applet;

		return (java.applet.Applet)com.sas.GlobalInfoTable.getObject();
	}

	/**
	 * Sets the applet to locate the resource relative to. If this method is called, the 
	 * AppletResourceLocator will not use the applet returned from GlobalInfoTable.getApplet()
	 * to locate the resource. 
	 * @param newValue the applet the resource is relative to
	 * @see com.sas.GlobalInfoTable#getApplet()
	 * @see #getApplet()
	 **/
	public void setApplet(Applet newValue)
	{
		if (!Util.equal(_applet, newValue))
		{
			Applet oldValue = _applet;
			_applet = newValue;

			firePropertyChange("applet", oldValue, newValue);
		}
	}

	/**
	 * Whether to locate the resource relative to the document base of the applet. The 
	 * AppletResourceLocator can only be document base relative or code base relative, not both.
	 * @return true if the resource is relative to the document base of the applet, false otherwise
	 * @see #setDocumentBaseRelative(boolean)
	 * @see #isCodeBaseRelative()
	 **/
	public boolean isDocumentBaseRelative()
	{
		return _isDocumentBaseRelative;
	}

	/**
	 * Sets whether the AppletResourceLocator should try to locate the resource relative to the document
	 * base or not. If true is passed in, a call to setCodeBaseRelative(false) will be made as the 
	 * AppletResourceLocator cannot be both relative to the document base and code base.
	 * @param true if the resource is relative to the document base of the applet, false otherwise
	 * @see #isDocumentBaseRelative()
	 * @see #setCodeBaseRelative(boolean)
	 **/
	public void setDocumentBaseRelative(boolean newValue)
	{
		if (_isDocumentBaseRelative != newValue)
		{
			_isDocumentBaseRelative = newValue;
			if (_isDocumentBaseRelative == true)
				setCodeBaseRelative(false);

			firePropertyChange("documentBaseRelative", new Boolean(!newValue), new Boolean(newValue));
		}
	}
	/**
	 * Whether to locate the resource relative to the code base of the applet. The 
	 * AppletResourceLocator can only be code base relative or document base relative, not both.
	 * @return true if the resource is relative to the code base of the applet, false otherwise
	 * @see #setCodeBaseRelative(boolean)
	 * @see #isDocumentBaseRelative()
	 **/
	public boolean isCodeBaseRelative()
	{
		return _isCodeBaseRelative;
	}

	/**
	 * Sets whether the AppletResourceLocator should try to locate the resource relative to the code
	 * base or not. If true is passed in, a call to setDocumentBaseRelative(false) will be made as the 
	 * AppletResourceLocator cannot be both relative to the code base and document base.
	 * @param true if the resource is relative to the code base of the applet, false otherwise
	 * @see #isCodeBaseRelative()
	 * @see #setDocumentBaseRelative(boolean)
	 **/
	public void setCodeBaseRelative(boolean newValue)
	{
		if (_isCodeBaseRelative != newValue)
		{
			_isCodeBaseRelative = newValue;
			if (_isCodeBaseRelative == true)
				setDocumentBaseRelative(false);

			firePropertyChange("codeBaseRelative", new Boolean(!newValue), new Boolean(newValue));
		}
	}

	/**
	 * Returns a {@link java.net.URL URL} of the resource. A null URL will be returned if the resource
	 * cannot be found relative to the applet or if an error occurs. A {@link java.net.MalformedURLException MalformedURLException}
	 * will be thrown if the construction of the URL fails.
	 * @return the url of the resource, or null if an error occurred or the resource cold not be found
	 * @see java.net.URL
	 * @see java.net.MalformedURLException
	 **/
	public java.net.URL getResource() throws java.net.MalformedURLException
	{
		if (getResourceName() == null)
			return null;
		Applet applet = getApplet();
		if (applet == null)
		{
			String projDir = getProjectDirectory();
			String fileLocation = null;
			if (projDir != null)
				fileLocation = projDir + java.io.File.separatorChar + getResourceName();/*I18nOK:COS*/
			else
				fileLocation = getResourceName();

			try
			{
				File file = new File(fileLocation);
				//needs to check if the file exists, otherwise any path can be converted to a url and the
				//resource will never be located at design-time.
				if (file.exists())
					return file.toURL();
			}
			catch (java.security.AccessControlException ex)
			{
			}
			
			return null;
		}
		else //RUNTIME MODE
		{
			ClassResourceLocator clrl = new ClassResourceLocator();
			clrl.setResourceName(getResourceName());
			clrl.setRelativeClass(applet.getClass());
			java.net.URL url = clrl.getResource();
			if (url != null)
				return url;

			if (!(isDocumentBaseRelative() || isCodeBaseRelative()))
				return null;

			ContextResourceLocator crl = new ContextResourceLocator();
			crl.setResourceName(getResourceName());
			if (isDocumentBaseRelative())
				crl.setContext(applet.getDocumentBase());
			else   // document and enforce that both can't be true!
			{
				String codeBase = applet.getCodeBase().toString();
				codeBase = removeHTMFileName(codeBase);

				if (codeBase != null)
					crl.setContext(new java.net.URL(codeBase));
			}
			return crl.getResource();
		}
	}

	/**
	 * Returns a string representation of the AppletResourceLocator that can be used with decode(String)
	 * to create a new AppletResourceLocator with the same resource name and Applet. 
	 * @return a string formatted "Applet: ~<code, document, or neither>~ ~<applet>~ ~<resource name>~", 
	 * where the resource name is a path relative to the resource. The Applet is the applet the
	 * resource is relative to. The "code, document, or neither" tells whether the resource is relative
	 * based on code base, document base, or neither.
	 * @see #decode(String)
	 * @see java.net.URL
	 **/
	public String encode()
	{
		String codeOrDocumentBase = null;
		if (isCodeBaseRelative())
			codeOrDocumentBase = "code";
		else if (isDocumentBaseRelative())
			codeOrDocumentBase = "document";
		else
			codeOrDocumentBase = "neither";

		return new String("Applet: ~<"+codeOrDocumentBase+">~ ~<"+getApplet()+">~ ~<"+getResourceName()+">~");/*I18nOK:LINE*/
	}

	/**
	 * Creates a AppletResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). 
	 * @param resource a string the should be in the format 
	 * "Applet: ~<code, document, or neither>~ ~<applet>~ ~<resource name>~", 
	 * where the resource name is a path relative to the resource. The Applet is the applet the
	 * resource is relative to. The "code, document, or neither" tells whether the resource is relative
	 * based on code base, document base, or neither.
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 * @see java.net.URL
	 **/
	public static ResourceLocatorInterface decode(String resource)
	{
		AppletResourceLocator arl = new AppletResourceLocator();
		if (resource.startsWith("Applet: "/*I18nOK:EMS*/))
		{
			int index = resource.indexOf(">~"/*I18nOK:EMS*/);
			if (index != -1)
			{
				String temp = resource.substring(10, index);
				if (Util.equal(temp, "code"/*I18nOK:EMS*/))
					arl.setCodeBaseRelative(true);
				else if (Util.equal(temp, "document"/*I18nOK:EMS*/))
					arl.setDocumentBaseRelative(true);

				int indexCopy = index;
				index = resource.indexOf(">~"/*I18nOK:EMS*/, index + 2);
				if (index != -1)
				{
					temp = resource.substring(indexCopy + 5, index - 1);
					//temp has applet info, not yet implemented
					arl.setResourceName(resource.substring(index + 5, resource.length() - 2));
				}
			}

			return arl;
		}
		else
			return null;
	}

	/**
	 * Creates a AppletResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). This method just calls the static method decode(Strng).
	 * @param resource a string the should be in the format 
	 * "Applet: ~<code, document, or neither>~ ~<applet>~ ~<resource name>~", 
	 * where the resource name is a path relative to the resource. The Applet is the applet the
	 * resource is relative to. The "code, document, or neither" tells whether the resource is relative
	 * based on code base, document base, or neither.
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 * @see #decode(String)
	 **/
	protected ResourceLocatorInterface createInstance(String resource)
	{
		return decode(resource);
	}

	/**
	 * Convenience method, returns the resource name. Will return the same value as {@link com.sas.net.ResourceLocatorInterface#getResource() getResource()}.
	 * @return the relative path to the resource
	 **/
	public String getRelativePath()
	{
		return getResourceName();
	}

	/**
	 * Returns the absolute path of the resource. This method will create the full path to the resource.
	 * It calls {@link com.sas.net.AppletResourceLocator#getProjectDirectory() getProjectDirectory()} to determine where the resource
	 * is. 
	 * @return the absolute file path to the resource
	 * @see #getProjectDirectory()
	 **/
	public String getAbsolutePath()
	{
		String absolutePath = getProjectDirectory();
		
		if (absolutePath != null && !absolutePath.endsWith(java.io.File.separator))
			absolutePath = absolutePath + java.io.File.separator;/*I18nOK:COS*/

		absolutePath = absolutePath + getResourceName()/*I18nOK:COS*/;

		return absolutePath;
	}

	/**
	 * Returns the directory where the applet is located, so that the resource can be found relative to it.
	 * @return a string representation of the directory where the applet is located. Used to find the resource relative to 
	 * the directory this method returns.
	 **/
	public String getProjectDirectory()
	{
		int mode = getMode();

		if (mode == WEBAF)
			return System.getProperty("webaf.project.directory");
		else if (mode == WEBEIS)
			return System.getProperty("webeis.document.directory");
		else
		{
			String base = null;
			Applet applet = getApplet();
			if (applet != null)
			{
				if (isDocumentBaseRelative())
					base = applet.getDocumentBase().toString(); 
				else
					base = applet.getCodeBase().toString();

				base = removeHTMFileName(base);
				base = removeFileString(base);
				return base;
			}
		}
		return null;
	}

	private String removeHTMFileName(String path)
	{
		String base = path;
		if (base != null)
		{
			if (System.getProperty("java.vendor").indexOf("Sun"/*I18nOK:EMS*/) >= 0) {
			   // appletviewer and hotjava include the html file as part of the
			   // codeBase!!
			   int index = base.lastIndexOf(".htm");
			   if (index > 0)
			       base = base.substring(0, base.lastIndexOf("/") + 1);
			}
		}
		return base;
	}

	//removes the "file:/" part from teh URL
	private String removeFileString(String path)
	{
		String base = path;
		if (base != null)
		{
			if (base.startsWith("file:"))
			{
				base = base.substring(6);
			}
		}
		return base;
	}
}