package com.sas.net;

import com.sas.util.Util;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.event.SwingPropertyChangeSupport;

/**
 * An implementation of ResourceLocatorInterface that takes a string representation of a URL as 
 * the resource name. The resource name can be any valid representation of a java.net.URL, e.g.
 * "http://www.somedomain.com/image1.jpg" representation or a "file:/c:/images/image1.gif". 
 * <br> 
 * <br>
 * <b><u>Creating a BaseResourceLocator</b></u>
 * <p>
 * A valid instance is created by instantiating an object of BaseResourceLocator followed by setting
 * the resource name.
 * <pre>
 * BaseResourceLocator baseResourceLocator1 = new BaseResourceLocator();
 * baseResourceLocator1.setResourceName("http://www.somedomain.com/images/apple.gif");
 * </pre>
 * <P>
 * A call to getResource() will return the {@link java.net.URL URL} of the resource if it can be found.
 * <pre>
 * try {
 * 	java.net.URL url = baseResourceLocator1.getResource(); 
 * } catch (java.net.MalformedURLException ex)
 * {
 * //an error occurred while forming the URL, handle as needed
 * }
 * </pre>
 * <P>
 * @author Drew Anderson
 * @version 1.0
 * @history
 * <dl>
 * <dt>Sept 18 2000  <dd>Created
 * </dl>
 **/
public class BaseResourceLocator implements ResourceLocatorInterface, java.io.Serializable, java.lang.Cloneable
{
   static final long serialVersionUID = -4196827828100369237L;

	/** The BaseResourceLocator is currently being used in WebAF at design-time" */
	public static final int WEBAF = 1;
	/** The BaseResourceLocator is currently being used in WebEIS at design-time" */
	public static final int WEBEIS = 2;
	/** The BaseResourceLocator is currently being used at run-time" */
	public static final int RUNTIME = 3;

	/** The resource name */
	protected String _resource;
	protected transient SwingPropertyChangeSupport _pcs;

	/**
	 * Default constructor, creates a BaseResourceLocator with no resource name set yet.
	 **/
	public BaseResourceLocator()
	{
		_pcs = new SwingPropertyChangeSupport(this);
	}

	/**
	 * Creates a BaseResourceLocator with the resource name being the string passed in.
	 * If the length of the string is 0, the resource name will be set to null.
	 **/
	public BaseResourceLocator(String resourceName)
	{
		this();
      setResourceName(resourceName);
	}

	/**
	 * Returns the name of the resource. Whatever the resource name is set to is what gets returned.
	 * @return the name of the resouce
	 * @see #setResourceName(String)
	 * @see com.sas.net.ResourceLocatorInterface#setResourceName(String)
	 **/
	public String getResourceName()
	{		
		return _resource;
	}

	/**
	 * Sets the name of the resource. Whenever getResource() is called, it needs to be able to
	 * interpret the resource name as a valid URL. If the new value has a length of 0, the resource
	 * name will be set to null.
	 * @param newValue the name of the resource
	 * @see #getResourceName
	 * @see com.sas.net.ResourceLocatorInterface#setResourceName(String)
	 **/
	public void setResourceName(String newValue)
	{
		if (newValue != null && (newValue.length() == 0 || newValue.equals("null")))
			newValue = null;
		if (!Util.equal(_resource, newValue))
		{
			String oldValue = _resource;
			_resource = newValue;

			firePropertyChange("resourceName", oldValue, newValue);
		}
	}

	/**
	 * Returns a {@link java.net.URL URL} of the resource. A null URL will be returned if the resource
	 * cannot be found based on the resource name or if an error occurs. A {@link java.net.MalformedURLException MalformedURLException}
	 * will be thrown if the construction of the URL fails.
	 * @return the url of the resource, or null if an error occurred or the resource could not be found
	 * @see java.net.URL
	 * @see java.net.MalformedURLException
	 **/
	public java.net.URL getResource() throws java.net.MalformedURLException
	{
		if (getResourceName() != null)
		{
			return new java.net.URL(getResourceName());
		}
		else
			return null;
	}

	/** 
	 * Returns a string representation of the ResourceLocatorInterface. It will be formatted
	 * with the class name followed by the URL to the resource.
	 * @return a string representation of the ResourceLocatorInterface
	 **/
   public String toString()
   {
      String retValue = getClass().getName();
		java.net.URL resource = null;
      try
      {
			resource = getResource();
      }
      catch (java.net.MalformedURLException ex)
      {
         // do nothing
      }
		if (resource != null)
      	retValue += ": " + resource.toExternalForm();/*I18nOK:LINE*/
      return retValue;
   }

	/**
	 * Determines if the object is equal to the current instance of ResourceLocatorInterface.
	 * It compares the result of getResource() and the class names.
	 * @param obj the Object to compare the existing ResourceLocatorInterface to. The obj should
	 * implement the ResourceLocatorInterface.
	 * @return true if the objects are equal, false otherwise
	 **/
	public boolean equals(Object obj)
	{
		if (!(obj instanceof ResourceLocatorInterface))
			return false;

		if (getClass() != obj.getClass())
			return false;

		try
		{
			java.net.URL url1 = getResource();
			java.net.URL url2 = ((ResourceLocatorInterface)obj).getResource();

			if (url1 != null && url2 != null)
				return Util.equal(url1.toExternalForm(), url2.toExternalForm());
			else
				return Util.equal(url1, url2);
		}
		catch (java.net.MalformedURLException ex)
		{
			return false;
		}
	}

	/**
	 * Returns the environment that the current instance of ResourceLocatorInterface is in.
	 * If there is a system property called "webaf.project.directory", then WEBAF mode will
	 * be returned. If there is a system property called "webeis.document.directory", then WEBEIS
	 * mode will be returned. If neither of these system properties are present, then RUNTIME will
	 * be returned. If access to either of the properties is denied, then RUNTIME will also be returned.
	 * @return either WEBAF, WEBEIS, or RUNTIME depending on the current environment that the 
	 * ResourceLocatorInterface is being used in
	 **/
	public int getMode()
	{
		try
		{
			String webAFdir = System.getProperty("webaf.project.directory");
			if (webAFdir != null && webAFdir.length() != 0)
				return WEBAF;
		}
		catch (java.security.AccessControlException ex)
		{
		}
		try
		{
			String webEISdir = System.getProperty("webeis.document.directory");
			if (webEISdir != null && webEISdir.length() != 0)
				return WEBEIS;
		}
		catch (java.security.AccessControlException ex)
		{
		}
		return RUNTIME;
	}

	/**
	 *	If getMode() retunrs WEBAF, then the system property "webaf.project.directory" will be returned.
	 *	If getMode() retunrs WEBEIS, then the system property "webeis.document.directory" will be returned.
	 *	If getMode() retunrs RUNTIME, then the system property "user.dir" will be returned.
	 * This method will need to be overrode in subclasses if a different behavior is desired.
	 * @return the directory where the project/document is located
	 **/
	public String getProjectDirectory()
	{
		int mode = getMode();

		if (mode == WEBAF)
			return System.getProperty("webaf.project.directory");
		else if (mode == WEBEIS)
			return System.getProperty("webeis.document.directory");
		else
			return System.getProperty("user.dir");
	}

	/**
	 * Returns a string representation of the BaseResourceLocator that can be used with decode(String)
	 * to create a new BaseResourceLocator with the same resource name. 
	 * @return a string formatted "URL: ~<resource name>~", where the resource name is a valid URL
	 * @see #decode(String)
	 **/
	public String encode()
	{
		return new String("URL: ~<"+getResourceName()+">~"/*I18nOK:EMS*/);
	}

	/**
	 * Creates a BaseResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). 
	 * @param resource a string the should be in the format "URL: ~<resource name>~", where the resource
	 * name should be a valid URL
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 **/
	public static ResourceLocatorInterface decode(String resource)
	{
		BaseResourceLocator brl = new BaseResourceLocator();
		if (resource.startsWith("URL: "))
		{
			brl.setResourceName(resource.substring(7, resource.length() - 2));
			return brl;
		}
		else
			return null;
	}

	/**
	 * Creates a BaseResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). This method just calls the static method decode(Strng).
	 * @param resource a string the should be in the format "URL: ~<resource name>~", where the resource
	 * name should be a valid URL
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 * @see #decode(String)
	 **/
	protected ResourceLocatorInterface createInstance(String resource)
	{
		return decode(resource);
	}

	/**
	 * Creates a copy (clone) of the existing instance of ResourceLocatorinterface.
	 * @return a copy of the existing instance of ResourceLocatorInterface
	 * @see java.lang.Cloneable
	 **/ 
	public Object clone()
	{
		return createInstance(encode());
	}

	/**
	 * Returns an instance of ResourceLocatorInterface if one can be created for the
	 * given code. The code should be a string that was created from calling encode() on
	 * an instance of ResourceLocatorInterface. If a ResourceLocatorInterface cannot be
	 * created, null will be returned.
	 * @return an instance of a ResourceLocatorInterface if one can be created, null if
	 * one cannot be created.
	 * @see #encode
	 **/
	public static ResourceLocatorInterface getResourceLocator(String code)
	{
		ResourceLocatorInterface rl = BaseResourceLocator.decode(code);
		if (rl == null)
		{
			rl = AppletResourceLocator.decode(code);
			if (rl == null)
			{
				rl = ClassResourceLocator.decode(code);
				if (rl == null)
				{
					rl = ContextResourceLocator.decode(code);
					if (rl == null)
					{
						rl = FileResourceLocator.decode(code);
					}
				}
			}
		}

		return rl;
	}

	private void readObject (java.io.ObjectInputStream stream)
		throws java.io.IOException, java.lang.ClassNotFoundException
	{
		try
		{
			stream.defaultReadObject();
		}

		finally
		{
			_pcs = new SwingPropertyChangeSupport(this);
		}
	}

	////////////////////////////////////////////////
	//
	// PropertyChange Methods
	//
	////////////////////////////////////////////////

	public void addPropertyChangeListener(PropertyChangeListener listener)
	{
		_pcs.addPropertyChangeListener(listener);
	}

	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener)
	{
		_pcs.addPropertyChangeListener(propertyName, listener);
	}

	public void removePropertyChangeListener(PropertyChangeListener listener)
	{
		_pcs.removePropertyChangeListener(listener);
	}

	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener)
	{
		_pcs.removePropertyChangeListener(propertyName, listener);
	}

	/**
	 * Report a bound property update to any registered listeners.
	 * No event is fired if old and new are equal and non-null.
	 * @param propertyName  The programmatic name of the property that was changed.
	 * @param oldValue  The old value of the property.
	 * @param newValue  The new value of the property.
	 */
	public void firePropertyChange(String propertyName, Object oldValue, Object newValue)
	{
		_pcs.firePropertyChange(propertyName, oldValue, newValue);
	}

	/**
	 * Fire an existing PropertyChangeEvent to any registered listeners.
	 * No event is fired if the given event's old and new values are equal and non-null.
	 * @param evt  The PropertyChangeEvent object.
	 */
	public void firePropertyChange(PropertyChangeEvent evt)
	{
		_pcs.firePropertyChange(evt);
	}
}