package com.sas.net;

import com.sas.util.Util;

/**
 * The ClassResourceLocator is used to locate any type of resources relative to a class,
 * whether the resource wanted is an image file or a webEIS document. The resource name should 
 * be the path and filename of the resource relative to the given relative class. 
 * <br> 
 * <br>
 * <b><u>Creating a ClassResourceLocator</b></u>
 * <p>
 * A valid instance is created by instantiating an object of ClassResourceLocator followed by setting
 * relative resource name and class.
 * <pre>
 * ClassResourceLocator classResourceLocator1 = new ClassResourceLocator();
 * classResourceLocator1.setRelativeClass(com.sas.visuals.ImageView.class); //assumes apple.gif is in the Class's direcotry
 * classResourceLocator1.setResourceName("res/Folder.gif");
 * </pre>
 * <P>
 * A call to getResource() will return the {@link java.net.URL URL} of the resource if it can be found.
 * <pre>
 * try {
 * 	java.net.URL url = classResourceLocator1.getResource(); 
 * } catch (java.net.MalformedURLException ex)
 * {
 * //an error occurred while forming the URL, handle as needed
 * }
 * </pre>
 * <P>
 * @author Drew Anderson
 * @version 1.0
 * @history
 * <dl>
 * <dt>Sept 18 2000  <dd>Created
 * </dl>
 **/
public class ClassResourceLocator extends BaseResourceLocator implements ResourceLocatorInterface
{
   static final long serialVersionUID = 2407231415924293430L;
	/** The class the resource is relative to **/
	protected Class _relativeClass;

	/**
	 * Default Constructor. The relative class and resource name still need to be set.
	 **/
	public ClassResourceLocator()
	{
		super();
	}

	/**
	 * Sets the name of the resource. The resourceName should be a relative path from the
	 * class to the resource. The class that the resource is relative to still needs to be set.
	 * @param resourceName the relative path and name of the resource to locate
	 **/
	public ClassResourceLocator(String resourceName)
	{
		super(resourceName);
	}

	/**
	 * Sets the relative class. The user will still need to set the resource name that
	 * is relative to this class.
	 * @param relClass the class the resource is relative to
	 **/
	public ClassResourceLocator(Class relClass)
	{
		this();
		_relativeClass = relClass;
	}

	/**
	 * Sets the name of the resource and relative class. The resourceName should be a relative path from the
	 * class to the resource.
	 * @param resourceName the relative path and name of the resource to locate
	 * @param relClass the class the resource is relative to
	 **/
	public ClassResourceLocator(String resourceName, Class relClass)
	{
		this(resourceName);
		_relativeClass = relClass;
	}

	/**
	 * Sets the name of the resource. The resource needs to be relative to whatever class that
	 * gets set.
	 * @param newValue the name of the resource
	 * @see #getResourceName
	 * @see #setRelativeClass(Class)
	 * @see com.sas.net.ResourceLocatorInterface#setResourceName(String)
	 **/
	public void setResourceName(String newValue)
	{
		if (newValue != null)
			newValue = newValue.replace('\\', '/');

		super.setResourceName(newValue);
	}

	/**
	 * Returns the class that the resource is relative to. Default value is null.
	 * @return the class the resource is relative to, null if it has not been set yet
	 * @see #setRelativeClass(Class)
	 **/
	public Class getRelativeClass()
	{
		return _relativeClass;
	}

	/** 
	 * Sets the class the resource is to be relative to.
	 * @param the class the resource is relative to
	 * @see #getRelativeClass
	 **/
	public void setRelativeClass(Class newValue)
	{
		if (!Util.equal(_relativeClass, newValue))
		{
			Class oldValue = _relativeClass;
			_relativeClass = newValue;

			firePropertyChange("relativeClass", oldValue, newValue);
		}
	}

	/**
	 * Returns a {@link java.net.URL URL} of the resource. A null URL will be returned if the resource
	 * cannot be found relative to the class or if an error occurs. A {@link java.net.MalformedURLException MalformedURLException}
	 * will be thrown if the construction of the URL fails. 
	 * @return the url of the resource, or null if an error occurred or the resource cold not be found
	 * @see java.net.URL
	 * @see java.net.MalformedURLException
	 **/
	public java.net.URL getResource() throws java.net.MalformedURLException
	{
		java.net.URL url = (java.net.URL)java.security.AccessController.doPrivileged(
			new java.security.PrivilegedAction()
			{
				public Object run()
				{
					java.net.URL newURL = null;
					try
					{
						if (getRelativeClass() != null)
						{
							ClassLoader classLoader = getRelativeClass().getClassLoader();
							if (classLoader != null && getResourceName() != null)
							{
								newURL = classLoader.getResource(getResourceName());
								if (newURL != null)
									return newURL;
							}
						}
					}
					catch (SecurityException ex)
					{
						// quietly ignore; just treat as not found
					}

					if (getResourceName() != null && getRelativeClass() != null)
					{
						newURL = getRelativeClass().getResource(getResourceName());
					}

					return newURL;	
				}
			});

		return url;
	}

	/**
	 * Returns a string representation of the ClassResourceLocator that can be used with decode(String)
	 * to create a new ClassResourceLocator with the same resource name and relative class. 
	 * @return a string formatted "Class: ~<class>~ ~<resource name>~", where the resource name 
	 * is a valid resource relative to the class.
	 * @see #decode(String)
	 **/
	public String encode()
	{
		return new String("Class: ~<"+getRelativeClass()+">~ ~<"+getResourceName()+">~");/*I18nOK:LINE*/
	}

	/**
	 * Creates a ClassResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). 
	 * @param resource a string the should be in the format "Class: ~<class>~ ~<resource name>~", 
	 * where the resource name is a valid resource relative to the class.
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 **/
	public static ResourceLocatorInterface decode(String resource)
	{
		ClassResourceLocator crl = new ClassResourceLocator();
		if (resource.startsWith("Class: "/*I18nOK:EMS*/))
		{
			int index = resource.indexOf(">~"/*I18nOK:EMS*/);
			if (index != -1)
			{
				try
				{
					//removes the "Class: ~<class " string from the start of the resource
					crl.setRelativeClass(Class.forName(resource.substring(15, index)));
				}
				catch (java.lang.ClassNotFoundException ex)
				{
				}
				crl.setResourceName(resource.substring(index + 5, resource.length() - 2));
			}

			return crl;
		}
		else
			return null;
	}

	/**
	 * Creates a ClassResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). This method just calls the static method decode(Strng).
	 * @param resource a string the should be in the format "Class: ~<class>~ ~<resource name>~", 
	 * where the resource name is a valid resource relative to the class.
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 * @see #decode(String)
	 **/
	protected ResourceLocatorInterface createInstance(String resource)
	{
		return decode(resource);
	}
}