package com.sas.net;

import com.sas.util.Util;

/**
 * The ContextResourceLocator is used to locate any type of resources relative to a context (resource base),
 * whether the resource wanted is an image file or a webEIS document. The resource name should 
 * be the path and filename of the resource relative the given context. 
 * <br> 
 * <br>
 * <b><u>Creating a ContextResourceLocator</b></u>
 * <p>
 * A valid instance is created by instantiating an object of ContextResourceLocator followed by setting
 * relative resource name and context.
 * <pre>
 * ContextResourceLocator contextResourceLocator1 = new ContextResourceLocator();
 * contextResourceLocator1.setContext(someURL); 
 * contextResourceLocator1.setResourceName("images/apple.gif");
 * </pre>
 * <P>
 * A call to getResource() will return the {@link java.net.URL URL} of the resource if it can be found.
 * <pre>
 * try {
 * 	java.net.URL url = contextResourceLocator1.getResource(); 
 * } catch (java.net.MalformedURLException ex)
 * {
 * //an error occurred while forming the URL, handle as needed
 * }
 * </pre>
 * <P>
 * @author Drew Anderson
 * @version 1.0
 * @history
 * <dl>
 * <dt>Sept 18 2000  <dd>Created
 * </dl>
 **/
public class ContextResourceLocator extends BaseResourceLocator implements ResourceLocatorInterface
{
   static final long serialVersionUID = -507629717298299238L;
	/** The context or resource base the resource is relative to, must be a valid java.net.URL **/
	protected java.net.URL _context;

	/**
	 * Default Constructor. The context and resource name still need to be set.
	 **/
	public ContextResourceLocator()
	{
		super();
	}

	/**
	 * Sets the name of the resource. The resourceName should be a relative path from the
	 * context to the resource. The context that the resource is relative to still needs to be set.
	 * @param resourceName the relative path and name of the resource to locate
	 **/
	public ContextResourceLocator(String resourceName)
	{
		super(resourceName);
	}

	/**
	 * Returns the context (resource base) that the resource is relative to. Default value is null.
	 * @return the context the resource is relative to, null if it has not been set yet.
	 * @see #setContext
	 * @see java.net.URL
	 **/
   public java.net.URL getContext()
	{
		return _context;
	}

	/** 
	 * Sets the context the resource is to be relative to. This should be a valid URL. The context
	 * could be a URL in the form of "http://", or "file:/", or if the context is a jar or zip file,
	 * the context will have a URL in the form of "jar:(file or http)://". The jar or zip format can be
	 * quite useful if the user stores all their images files in a zip file on their webserver. 
	 * @param the class the resource is relative to
	 * @see #getContext
	 * @see java.net.URL
	 **/
   public void setContext(java.net.URL newValue)
	{
		if (!Util.equal(_context, newValue))
		{
			java.net.URL oldValue = _context;
			_context = newValue;

			firePropertyChange("context", oldValue, newValue);
		}
	}

	/**
	 * Returns a {@link java.net.URL URL} of the resource. A null URL will be returned if the resource
	 * cannot be found relative to the context or if an error occurs. A {@link java.net.MalformedURLException MalformedURLException}
	 * will be thrown if the construction of the URL fails. 
	 * @return the url of the resource, or null if an error occurred or the resource cold not be found
	 * @see java.net.URL
	 * @see java.net.MalformedURLException
	 **/
   public java.net.URL getResource() throws java.net.MalformedURLException 
	{
      return new java.net.URL(getContext(), getResourceName());
   }

	/**
	 * Returns a string representation of the ContextResourceLocator that can be used with decode(String)
	 * to create a new ContextResourceLocator with the same resource name and context. 
	 * @return a string formatted "Context: ~<context>~ ~<resource name>~", where the resource name 
	 * is a path relative to the context. The context should be a valid URL.
	 * @see #decode(String)
	 * @see java.net.URL
	 **/
	public String encode()
	{
		return new String("Context: ~<"+getContext()+">~ ~<"+getResourceName()+">~");/*I18nOK:LINE*/
	}

	/**
	 * Creates a ContextResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). 
	 * @param resource a string the should be in the format "Context: ~<context>~ ~<resource name>~", 
	 * where the resource name should be a relative path based on the context. Context should be a
	 * valid URL
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 * @see java.net.URL
	 **/
	public static ResourceLocatorInterface decode(String resource)
	{
		ContextResourceLocator crl = new ContextResourceLocator();
		if (resource.startsWith("Context: "/*I18nOK:EMS*/))
		{
			int index = resource.indexOf(">~"/*I18nOK:EMS*/);
			if (index != -1)
			{
				try
				{
					//removes the "Context: " string from the start of the resource
					crl.setContext(new java.net.URL(resource.substring(11, index)));
				}
				catch (java.net.MalformedURLException ex)
				{
				}
				crl.setResourceName(resource.substring(index + 5, resource.length() - 2));
			}

			return crl;
		}
		else
			return null;
	}

	/**
	 * Creates a ContextResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). This method just calls the static method decode(Strng).
	 * @param resource a string the should be in the format "Context: ~<context>~ ~<resource name>~", 
	 * where the resource name should be a relative path based on the context. Context should be a
	 * valid URL
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 * @see #decode(String)
	 **/
	protected ResourceLocatorInterface createInstance(String resource)
	{
		return decode(resource);
	}
}  