package com.sas.net;

import com.sas.util.Util;

/**
 * The FileResourceLocator is used to locate any type of resources relative to a {@link java.io.File File},
 * whether the resource wanted is an image file or a webEIS document. The resource name should 
 * be the path and filename of the resource. 
 * <br> 
 * <br>
 * <b><u>Creating a FileResourceLocator</b></u>
 * <p>
 * A valid instance is created by instantiating an object of FileResourceLocator followed by setting
 * resource name (path to file).
 * <pre>
 * FileResourceLocator fileResourceLocator1 = new FileResourceLocator();
 * fileResourceLocator1.setResourceName("C:/Misc/images/apple.gif");
 * </pre>
 * <P>
 * A call to getResource() will return the {@link java.net.URL URL} of the resource if it can be found.
 * <pre>
 * try {
 * 	java.net.URL url = fileResourceLocator1.getResource(); 
 * } catch (java.net.MalformedURLException ex)
 * {
 * //an error occurred while forming the URL, handle as needed
 * }
 * </pre>
 * <P>
 * @author Drew Anderson
 * @version 1.0
 * @history
 * <dl>
 * <dt>Sept 18 2000  <dd>Created
 * </dl>
 **/
public class FileResourceLocator extends BaseResourceLocator implements ResourceLocatorInterface
{
   static final long serialVersionUID = -1793970847181329099L;

	/**
	 * Default Constructor. The resource name (path to file) still needs to be set.
	 **/
	public FileResourceLocator()
	{
		super();
	}

	/**
	 * Sets the name of the resource. The resourceName should be a path to the
	 * resource.
	 * @param resourceName the path and filename of the resource
	 **/
	public FileResourceLocator(String resourceName)
	{
		super(resourceName);
	}

	/**
	 * Returns a {@link java.net.URL URL} of the resource. A null URL will be returned if the resource
	 * cannot be found if an error occurs. A {@link java.net.MalformedURLException MalformedURLException}
	 * will be thrown if the construction of the URL fails. 
	 * @return the url of the resource, or null if an error occurred or the resource cold not be found
	 * @see java.net.URL
	 * @see java.io.File
	 * @see java.net.MalformedURLException
	 **/
	public java.net.URL getResource() throws java.net.MalformedURLException
	{
		if (getResourceName() != null)
		{
			java.io.File file = new java.io.File(getResourceName());
			return file.toURL();
		}
		else
			return null;
	}

	/**
	 * Returns a string representation of the FileResourceLocator that can be used with decode(String)
	 * to create a new FileResourceLocator with the same resource name. 
	 * @return a string formatted "File: ~<resource name>~", where the resource name is a valid file path.
	 * @see #decode(String)
	 **/
	public String encode()
	{
		return new String("File: ~<"+getResourceName()+">~");/*I18nOK:LINE*/
	}

	/**
	 * Creates a FileResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). 
	 * @param resource a string the should be in the format "File: ~<resource name>~", where the resource
	 * name should be a valid file path.
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 **/
	public static ResourceLocatorInterface decode(String resource)
	{
		FileResourceLocator frl = new FileResourceLocator();
		if (resource.startsWith("File: "))
		{
			frl.setResourceName(resource.substring(8, resource.length() - 2));
			return frl;
		}
		else
			return null;
	}

	/**
	 * Creates a FileResourceLocator from the string passed in. The string should in the same format
	 * as the the string returned from encode(). This method just calls the static method decode(Strng).
	 * @param resource a string the should be in the format "File: ~<resource name>~", where the resource
	 * name should be a valid file path.
	 * @return an instance of a ResourceLocatorInterface if one can be created, null otherwise
	 * @see #encode()
	 * @see #decode(String)
	 **/
	protected ResourceLocatorInterface createInstance(String resource)
	{
		return decode(resource);
	}
}