package com.sas.net;

/**
 * {@link java.net.URLEncoder} only supports 8 bit encodings.
 * This class allows you to encode characters > 0xff
 * as &#x5c;uxxxx. The resulting string can then be passed through
 * {@link java.net.URLEncoder} to encode it for a URL,
 * and conversely, strings pulled from a URL, after passing
 * through {@link java.net.URLDecoder}, can be decided
 * back into their Unicode.
 * <h2>Example</h2>
 * To encode a Unicode string for a URL, use:
 * <pre>
 * <blockquote>
 * String encoded = com.sas.net.UnicodeURLEncoder.encode(string);
 * encoded = java.net.URLEncoder.encode(encoded);
 * </blockquote>
 * </pre>
 * <p>
 * To decode a Unicode string from a URL, use:
 * <pre>
 * <blockquote>
 * String string = java.net.URLEncoder.decode(encoded);
 * string = com.sas.net.UnicodeURLDecoder.decode(string);
 * </blockquote>
 * </pre>
 * @author sasdjb <A HREF="http://sww.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 */
public class UnicodeURLEncoder
{
  /* hex characters as a char[] array */
  static private final char hex[] = "0123456789ABCDEF".toCharArray();
  /*
   * array used to map hex nibbles characters back to their unicode.
   * This is a generated lookup table that maps or '0' to 48,
   * 'a' to 65, and 'A' to 65, or example. Handles both lowercase and uppercase
   * nibbles.
   */
  static private final char decode[] = new char['f'+1];
  static {
    for (int i = 0; i < hex.length; i++)
      {
        decode[                      hex[i] ] = (char) i;
        decode[Character.toLowerCase(hex[i])] = (char) i;
      }
  }

  public static final UnicodeURLEncoder defaultInstance = new UnicodeURLEncoder();

  /**
   * Construct a UnicodeURLEncoder.
   * @deprecated You probably do not need to construct
   * a UnicodeURLEncoder; you can use either the default
   * instance {@link #defaultInstance}, or the static methods {@link #decode(String)}
   * and {@link #encode(String)}
   */
  public UnicodeURLEncoder()
  {
  }

  /**
   * Encode a string using the UnicodeURLEncoding.
   * This replaces each Unicode character that is
   * greater than 0xff with a string &#x5c;uxxxx.
   * The '&#x5c;' character is replaced also.
   * Other characters less than 0xff are left as is;
   * you can use {@link java.net.URLEncoder#encode(String)}
   * to further encode such strings for URLs.
   */
  public static final String encode(String string)
  {
    return defaultInstance.encodeString(string);
  }

  /**
   * Decode a string using the UnicodeURLEncoding.
   * This replaces each occurrence of &#x5c;uxxxx
   * in the input string with the corresponding Unicode character.
   */
  public static final String decode(String string)
  {
    return defaultInstance.decodeString(string);
  }

  /**
   * Encode a string using the UnicodeURLEncoding.
   * This replaces each Unicode character that is
   * greater than 0xff with a string &#x5c;uxxxx.
   * The '&#x5c;' character is replaced also.
   * Other characters less than 0xff are left as is;
   * you can use {@link java.net.URLEncoder#encode(String)}
   * to further encode such strings for URLs.
   */
  public String encodeString(String in)
  {
    int len = 0;
    for (int i = 0, l = in.length(); i < l; i++)
      {
        char ch = in.charAt(i);
        if (ch > 0xff)
          len += 6; // can represent as \uabcd
        else
          len++;
      }
    StringBuffer buffer = new StringBuffer(len);
    for (int i = 0, l = in.length(); i < l; i++)
      {
        char ch = in.charAt(i);
        if (ch > 0xff || ch == '\\')
          buffer.append((char)'\\')
            .append((char)'u')
            .append((char)hex[(ch >> 12)&0xf])
            .append((char)hex[(ch >>  8)&0xf])
            .append((char)hex[(ch >>  4)&0xf])
            .append((char)hex[(ch     )&0xf]);
        else
          buffer.append((char) ch); // can represent with itself
      }
    String s = buffer.toString();/*I18nOK:MTH*/
    //System.out.println( s.length() + "==" + len + "? " + (s.length() == len));
    return s;
  }

  /**
   * Decode a string using the UnicodeURLEncoding.
   * This replaces each occurrence of &#x5c;uxxxx
   * in the input string with the corresponding Unicode character.
   */
  public String decodeString(String in)
  {
    try
      {
        int len = 0;
        for (int i = 0, l = in.length(); i < l; i++)
          {
            char ch = in.charAt(i);
            if (ch == '\\')
              i+=5;
            len++;
          }

        StringBuffer buffer = new StringBuffer(len);

        for (int i = 0, l = in.length(); i < l; i++)
          {
            char ch = in.charAt(i);
            if (ch == '\\')
              {
                ch = in.charAt(++i); // this should be 'u'; ignore.
                char a = in.charAt(++i);
                char b = in.charAt(++i);
                char c = in.charAt(++i);
                char d = in.charAt(++i);
                ch = (char) ((decode[a] << 12) | (decode[b] << 8) | (decode[c] << 4) | decode[d]);
              }
            buffer.append((char) ch);
          }
        String s = buffer.toString();/*I18nOK:MTH*/
        //System.out.println( s.length() + "==" + len + "? " + (s.length() == len));
        return s;
      }
    catch (IndexOutOfBoundsException e)
      {
        throw new IllegalArgumentException(in);
      }
  }


  /**

  public static void main(String argv[])
  {
    int rc = test("A Unicode '\ucafe' \"\ubabe\" string. \udbdb");
    for (int i = 0; i < argv.length; i++)
      {
        rc += test(argv[i]);
      }
    System.exit(rc);
  }

  public static int test(String arg)
  {
    int rc;
    String encoded = UnicodeURLEncoder.encode(arg);
    String uEncoded = java.net.URLEncoder.encode(encoded);
    String decoded = UnicodeURLEncoder.decode(java.net.URLDecoder.decode(uEncoded));
    System.out.print(     "\nEncode:            " + arg
                        + "\nUnicodeURLEncoded: " + encoded
                        + "\nURL Encoded:       " + uEncoded
                        + "\nDecoded:            " + decoded);
    if (arg.equals(decoded))
      {
        System.out.println( "  (ok)");
        rc = 0;
      }
    else
      {
        System.out.println( "  (Encoding/Decoding error!!)");
        rc = 1;
      }

    arg = arg.toLowerCase();
    encoded = encoded.toLowerCase();
    decoded = UnicodeURLEncoder.decode(encoded);
    System.out.print(     "\nEncode:            " + arg
                        + "\nUnicodeURLEncoded: " + encoded
                        + "\nDecoded:            " + decoded);
    if (arg.equals(decoded))
      {
        System.out.println( "  (ok)");
        rc = 0;
      }
    else
      {
        System.out.println( "  (Encoding/Decoding error!!)");
        rc = 1;
      }
    return rc;
  }

  **/
}
