// Copyright (c) 2001 by SAS Institute Inc., Cary, NC 27513
package com.sas.net.brg;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Iterator;
import java.util.Properties;
import org.omg.CORBA.BAD_PARAM;
import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.SystemException;
import com.sas.net.brg.orb.BrgOrb;
import com.sas.net.brg.orb.ProtocolServerInterface;
import com.sas.text.Message;
//
// History:
//
// Date    Name            Remarks
// 28Feb01 B Beers         S0110393: Add bind and server flow features.
// 10Apr02 B Beers         S0141718: Remove dependency on netutil.jar.
//
/**
 * @author Biff Beers
 */
public class BridgeProtocolServer implements ProtocolServerInterface,Runnable
{
     public BridgeProtocolServer(BrgOrb orb,Properties properties)
     {
          this.orb = orb;
          this.properties = (Properties)properties.clone();

          try
          {
               debugLevel = Integer.parseInt(this.properties.getProperty("debugLevel","0"));
          }
          catch (NumberFormatException e)
          {
          }

          Object logFile = this.properties.get("logFile");

          if (logFile != null)
          {
               if (logFile instanceof String)
               {
                    try
                    {
                         logWriter = new PrintWriter(new FileWriter((String)logFile),true);
                    }
                    catch (IOException e)
                    {
                    }
               }
               else
               {
                    logWriter = (PrintWriter)logFile;
               }
          }

          if (logWriter == null)
          {
               logWriter = new PrintWriter(System.out,true);
          }

          this.properties.put("logFile",logWriter);
          String portString = this.properties.getProperty("port","5307");

          try
          {
               port = Integer.parseInt(portString);
          }
          catch (NumberFormatException e)
          {
               // {0}: Invalid server socket port number: {1}
               String messageIndex = "brg.B600.ex.msg";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex,portString);
               throw new BAD_PARAM(message);
          }

          try
          {
               socket = new ServerSocket(port);
          }
          catch (IOException e)
          {
               // {0}: An exception of type {1} was thrown while creating a server socket: {2}
               String messageIndex = "brg.B601.ex.msg";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex,e.getClass().getName(),e.getMessage());
               throw new BAD_PARAM(message);
          }

          listenThread = new Thread(this);
          listenThread.start();
     }

     public void run()
     {
          if (debugLevel > 0)
          {
               logWriter.print("Listening for bridge protocol connections on port ");
               logWriter.println(Integer.toString(port));
          }

          while (true)
          {
               try
               {
                    Socket s = socket.accept();
                    int baseClientID = orb.getNextClientID();
                    BridgeProtocolEngine engine = new BridgeProtocolEngine(orb,properties,s,baseClientID);
                    orb.addEngine(engine);
               }
               catch (IOException e)
               {
                    break;
               }
               catch (SystemException se)
               {
               }
          }

          synchronized (this)
          {

               if (socket != null)
               {
                    try
                    {
                         socket.close();
                    }
                    catch (IOException e)
                    {
                    }

                    socket = null;
               }

               stopped = true;
               notifyAll();
          }
     }

     public synchronized void close(boolean waitForCompletion)
     {
          if (socket != null)
          {
               try
               {
                    socket.close();
               }
               catch (IOException e)
               {
               }

               socket = null;
          }

          if (waitForCompletion)
          {
               while (!stopped)
               {
                    try
                    {
                         wait();
                    }
                    catch (InterruptedException e)
                    {
                         return;
                    }
               }
          }
     }

     private BrgOrb orb = null;
     private Properties properties = null;
     private int debugLevel = 0;
     private int port = 0;
     private ServerSocket socket = null;
     private boolean stopped = false;
     private Thread listenThread = null;
     private PrintWriter logWriter = null;
}