// Copyright (c) 2001 by SAS Institute Inc., Cary, NC 27513
package com.sas.net.brg.orb;

import java.applet.Applet;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import org.omg.CORBA.Any;
import org.omg.CORBA.BAD_INV_ORDER;
import org.omg.CORBA.BAD_PARAM;
import org.omg.CORBA.DynamicImplementation;
import org.omg.CORBA.Environment;
import org.omg.CORBA.ExceptionList;
import org.omg.CORBA.INITIALIZE;
import org.omg.CORBA.INV_OBJREF;
import org.omg.CORBA.NamedValue;
import org.omg.CORBA.NVList;
import org.omg.CORBA.ORB;
import org.omg.CORBA.UNKNOWN;
import org.omg.CORBA.StructMember;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.UnionMember;
import org.omg.CORBA.portable.Delegate;
import org.omg.CORBA.portable.ObjectImpl;
import org.omg.CORBA.portable.OutputStream;
import org.omg.CORBA.ORBPackage.InvalidName;
import com.sas.net.brg.BridgeProtocolEngine;
import com.sas.net.brg.BridgeProtocolServer;
import com.sas.text.Message;
//
// History:
//
// Date    Name            Remarks
// 06Jan99 B Beers         Created.
// 26Mar99 B Beers         Added documentation comments.
// 06May99 B Beers         S0053125: Marshaled input stream exceptions to threads waiting for input.
// 26May99 B Beers         S0053449: Improved object lifecycle management.
// 01Jun99 B Beers         S0055544: Moved string literals into resource file.
// 30Jun99 B Beers         S0057860: Improved object lifecycle management (Part II).
// 01Jul99 B Beers         S0058989: Catch unchecked exceptions in events and outcalls.
// 30Aug99 B Beers         S0063163: Map system exceptions.
// 20Sep99 B Beers         S0066345: Encode stringified object reference.
// 15Dec99 B Beers         S0067137: Ensure client will close when server does.
// 07Jan00 B Beers         S0073387: Prevent NullPointerException when closing.
// 27Mar00 B Beers         S0080909: Fix bad conditional statement on "enter as writer".
// 13Apr00 B Beers         S0081930: Allow 0x0 2D arrays.
// 18Oct00 B Beers         S0100131: Prevent GC from releasing previously closed remote object.
// 07Nov00 B Beers         S0102085: Allow null object references again.
// 28Feb01 B Beers         S0110393: Add bind and server flow features.
// 25Apr01 B Beers         S0114669: Move POA classes to orb.jar.
// 23Aug01 B Beers         S0124519: Split socket driver code into SocketAccessMethod.
// 12Oct01 B Beers         S0129352: Remove connection sharing logic.
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 10Apr02 B Beers         S0141718: Remove dependency on netutil.jar.
// 10Apr03 B Beers         S0193775: Bypass singleton ORB for factory methods (except TypeCode).
//
/**
 * This is the root class of JavaBridge, a standard, non-interoperable CORBA ORB.
 * It is standard because it implements the CORBA ORB specification (more
 * specifically, it extends <CODE>org.omg.CORBA.ORB</CODE>), but it is not
 * interoperable because it can only communicate with an IOM BRIDGE server. The
 * benefit of communicating with the IOM BRIDGE server through a standard CORBA
 * ORB is that it will support programs written to communicate with an IOM CORBA
 * server through an interoperable CORBA ORB without modification or even
 * recompilation.
 * <P>
 * To indicate to one of the <CODE>org.omg.CORBA.ORB.init()</CODE> method that
 * you would
 * like to use this ORB specify the following properties in the
 * <CODE>java.util.Properties</CODE> object required by those methods.
 * <P>
 * <table border="1">
 * <tr>
 * <th>Name</th>
 * <th>Value</th>
 * </tr>
 * <tr>
 * <td><code>org.omg.CORBA.ORBClass</CODE></td>
 * <td><code>com.sas.net.brg.orb.BrgOrb</code></td>
 * </tr>
 * <tr>
 * <td><code>org.omg.CORBA.ORBSingletonClass</CODE></td>
 * <td><code>com.sas.net.omgidl.orb.ORB</code></td>
 * </tr>
 * </table>
 * <P>
 * It is also possible to instantiate this ORB directly using its constructor.
 * <P>
 * JavaBridge supports a variety of encryption algoritms for encrypting messages
 * sent to and received from the IOM BRIDGE server. If encryption is used, then
 * the authentication message which contains a user name and password will be
 * encrypted. More information about control of encryption options is available
 * at the <CODE>string_to_object</CODE> method. The following algorithms are
 * supported.
 * <DL>
 * <DT>sasproprietary<DD>a stream cipher developed at SAS Institute.
 * <DT>rc2<DD>a block cipher developed by RSA Data Security
 * <DT>rc4<DD>a stream cipher developed by RSA Data Security
 * <DT>des<DD>a block cipher called Data Encryption Standard
 * <DT>tripledes<DD>des applied three times with separate keys
 * </DL>
 * Support for rc2, rc4, des, and tripledes is only available if SAS/SECURE
 * has been licensed on the IOM BRIDGE server and if the Java component of
 * SAS/SECURE (sasecjav.zip) has been installed in the client code base.
 * Furthermore, tripledes is generally only available in the United States
 * and Canada due to U.S. Government export restrictions.
 *
 * @author Biff Beers
 */
public class BrgOrb extends org.omg.CORBA_2_3.ORB
{
     // constructors
     //---------------------------------------------------------
     public BrgOrb()
     {
          engineSet = new HashSet();
          serverSet = new HashSet();
          activeObjectMap = new HashMap();
     }

     public BrgOrb(Properties properties)
     {
          this();
          initialize(properties);
     }

     // super-class methods
     //---------------------------------------------------------
     protected void set_parameters(java.applet.Applet applet,Properties properties)
     {
          initialize(properties);
     }

     protected void set_parameters(String[] argv,Properties properties)
     {
          initialize(properties);
     }

     public void shutdown(boolean waitForCompletion)
     {
          synchronized (serverSet)
          {
               Iterator serverIterator = serverSet.iterator();

               while (serverIterator.hasNext())
               {
                    ProtocolServerInterface server = (ProtocolServerInterface)serverIterator.next();
                    server.close(waitForCompletion);
               }

               serverSet.clear();
          }

          synchronized (engineSet)
          {
               Iterator engineIterator = engineSet.iterator();

               while (engineIterator.hasNext())
               {
                    ProtocolEngineInterface engine = (ProtocolEngineInterface)engineIterator.next();
                    engine.close(waitForCompletion);
               }

               engineSet.clear();
          }

          synchronized (this)
          {
               if (rootPOA != null)
               {
                    rootPOA.close(waitForCompletion);
               }
          }
     }

     public synchronized String[] list_initial_services()
     {
          if (initialServices == null)
          {
               initialServices = new String[1];
               initialServices[0] = "rootPOA";
          }

          return initialServices;
     }

     public org.omg.CORBA.Object resolve_initial_references(String name) throws InvalidName
     {
          if ("rootPOA".equals(name))
          {
               synchronized (this)
               {
                    if (rootPOA == null)
                    {
                         rootPOA = new POAImpl(this,"rootPOA",null,activeObjectMap);
                    }

                    return rootPOA;
               }
          }

          throw new InvalidName(name);
     }

     public String object_to_string(org.omg.CORBA.Object obj)
     {
          Properties properties = objectToProperties(obj);
          return propertiesToReference(properties);
     }

     /**
      * This method causes the server to instantiate an object
      * and sets up a remote reference to it on the client. If
      * object instantiation fails, then this method returns null,
      * and a reason will be printed in stderr.
      * <P>The
      * identity and location of the desired object are specified
      * in a specially formatted URL.
      * <BLOCKQUOTE>
      * bridge://<I>host</I>:<I>port</I>/<I>classId</I>?userName=<I>user_name</I>&password=<I>password</I>&<I>otherNames=otherValues</I>
      * </BLOCKQUOTE>
      * <P>
      * <DL>
      * <DT><I>host</I><DD>the IP name or address of the machine running an IOM BRIDGE server or spawner.
      * <DT><I>port</I><DD>the TCP port that the IOM BRIDGE server or spawner is listening to.
      * <DT><I>classId</I><DD>the UUID for the class you want to instantiate. The UUID for the
      * SASAutomation class is <CODE>440196D4-90F0-11D0-9F41-00A024BB830C</CODE>.
      * <DT><I>userName</I><DD>a valid user name on the IOM BRIDGE server host.
      * <DT><I>password</I><DD>the password corresponding to <I>userName</I>.
      * </DL>
      * <P>
      * The name/value pairs listed above are all required. Other name/value pairs may be appended to
      * the URL following the same format as <I>userName</I> and <I>password</I> to control optional
      * functionality.
      * <DL>
      * <DT><I>encryptionPolicy</I><DD>whether or not JavaBridge should attempt to negotiate with
      * the server over which encryption algorithm to use and what to do if the negotiations fail.
      * Possible values:
      * <DL>
      * <DT>none<DD>do not use encryption. This is the default.
      * <DT>optional<DD>attempt to use encryption but, if algorithm negotiation fails, continue with an
      * unencrypted session
      * <DT>required<DD>attempt to use encryption but, if algorithm negotiation fails, fail the connection
      * </DL>
      * <DT>encryptionContent<DD>specifies which messages should be encrypted if encryption is used. Possible values:
      * <DL>
      * <DT>all<DD>encrypt all messages. This is the default.
      * <DT>authentication<DD>encrypt only messages that contain user name and password information.
      * </DL>
      * <DT>encryptionAlgorithms<DD>the list of algorithms you are willing to use in order of
      * preference. Values in the list should be separated by commas and chosen from sasproprietary,
      * rc2, rc4, des, or tripledes. If no value is specified, then one of the server's favorite algorithms
      * will be used.
      * </DL>
      * <P>
      * Note that the server's encryption policy is, by default, optional so no special action is needed
      * to prepare an IOM BRIDGE server to use encryption. However, it is possible to set up the server
      * so that encryption is required. Consult IOM BRIDGE server documentation for details.
      * <P>
      * Note also that name/value pairs containing <CODE>&</CODE>, <CODE>=</CODE>, or <CODE>%</CODE>
      * need to be
      * encoded since these characters are reserved in the URL format. Use <CODE>java.net.URLEncoder</CODE>
      * to encode each name and each value individually.
      * @param url a specially formatted URL indicating the identity and location of a remote object.
      * @return a reference to the remote object
      */
     public org.omg.CORBA.Object string_to_object(String url)
     {
          // parse the url
          Properties properties = referenceToProperties(url);
          return propertiesToObject(properties);
     }

     /**
      * Create a name/value pair list.
      * @param count the initial capacity of the list
      * @return a name/value pair list
      */
     public NVList create_list(int count)
     {
          return new com.sas.net.brg.orb.NVList(this,count);
     }

     /**
      * Create a name/value pair.
      * @param s the name
      * @param any the value
      * @param flags argument direction. One of <CODE>ARG_IN</CODE>, <CODE>ARG_OUT</CODE>
      * or <CODE>ARG_INOUT</CODE>
      * @return a name/value pair
      */
     public NamedValue create_named_value(String s,Any any,int flags)
     {
          return new com.sas.net.brg.orb.NamedValue(s,any,flags);
     }

     /**
      * Create a list of exceptions raised by an operation.
      * @return a list of exceptions raised by an operation
      */
     public ExceptionList create_exception_list()
     {
          return new com.sas.net.brg.orb.ExceptionList();
     }

     /**
      * Create a list of Context objects. If BrgOrb is the
      * singleton orb, then this method will throw
      * org.omg.CORBA.NO_IMPLEMENT.
      * @return a list of Context objects
      */
     public org.omg.CORBA.ContextList create_context_list()
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               throw new org.omg.CORBA.NO_IMPLEMENT();
          }
          else
          {
               return orb.create_context_list();
          }
     }

     /**
      * Create a Context object. If BrgOrb is the
      * singleton orb, then this method will throw
      * org.omg.CORBA.NO_IMPLEMENT.
      * @return a Context object
      */
     public org.omg.CORBA.Context get_default_context()
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               throw new org.omg.CORBA.NO_IMPLEMENT();
          }
          else
          {
               return orb.get_default_context();
          }
     }

     /**
      * Creates an environment which holds the exception raised
      * by an operation invocation.
      * @return an environment.
      */
     public Environment create_environment()
     {
          return new com.sas.net.brg.orb.Environment();
     }

     /**
      * Create a stream for serializing parameters.
      * @return a stream for serializing parameters
      */
     public org.omg.CORBA.portable.OutputStream create_output_stream()
     {
          return new OutputStreamImpl(this);
     }

     /**
      * Throws NO_IMPLEMENT.
      */
     public void send_multiple_requests_oneway(org.omg.CORBA.Request[] req)
     {
          throw new org.omg.CORBA.NO_IMPLEMENT();
     }

     /**
      * Throws NO_IMPLEMENT.
      */
     public void send_multiple_requests_deferred(org.omg.CORBA.Request[] req)
     {
          throw new org.omg.CORBA.NO_IMPLEMENT();
     }

     /**
      * Throws NO_IMPLEMENT.
      */
     public boolean poll_next_response()
     {
          throw new org.omg.CORBA.NO_IMPLEMENT();
     }

     /**
      * Throws NO_IMPLEMENT.
      */
     public org.omg.CORBA.Request get_next_response()
     {
          throw new org.omg.CORBA.NO_IMPLEMENT();
     }

     /**
      * Get the TypeCode object representing a particular primitive type.
      * @param enumerated value representing the desired primitive type.
      * @return a TypeCode object
      */
     public TypeCode get_primitive_tc(TCKind tcKind)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return com.sas.net.brg.orb.TypeCode.getPrimitiveTypeCode(tcKind);
          }
          else
          {
               return orb.get_primitive_tc(tcKind);
          }
     }

     /**
      * Create a TypeCode object representing a struct.
      * @param id interface repository ID
      * @param name name of struct
      * @param members the struct members
      * @return a TypeCode object representing a struct.
      */
     public TypeCode create_struct_tc(String id,String name,StructMember[] members)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(org.omg.CORBA.TCKind._tk_struct,id,name,members);
          }
          else
          {
               return orb.create_struct_tc(id,name,members);
          }
     }

     /**
      * Create a TypeCode object representing a union.
      * @param id interface repository ID
      * @param name name of union
      * @param discriminator_type the discriminator type
      * @param members the union members
      * @return a TypeCode object representing a union.
      */
     public TypeCode create_union_tc(String id,String name,TypeCode discriminator_type,UnionMember[] members)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(id,name,discriminator_type,members);
          }
          else
          {
               return orb.create_union_tc(id,name,discriminator_type,members);
          }
     }

     /**
      * Create a TypeCode object representing an enum.
      * @param id interface repository ID
      * @param name name of enum
      * @param members the enum members
      * @return a TypeCode object representing an enum.
      */
     public TypeCode create_enum_tc(String id,String name,String[] members)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(id,name,members);
          }
          else
          {
               return orb.create_enum_tc(id,name,members);
          }
     }

     /**
      * Create a TypeCode object representing another TypeCode object.
      * @param id interface repository ID
      * @param name name of alias type
      * @param original_type the original TypeCode object
      * @return a TypeCode object representing another TypeCode object.
      */
     public TypeCode create_alias_tc(String id,String name,TypeCode original_type)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(id,name,original_type);
          }
          else
          {
               return orb.create_alias_tc(id,name,original_type);
          }
     }

     /**
      * Create a TypeCode object representing an exception.
      * @param id interface repository ID
      * @param name name of exception
      * @param members the exception members
      * @return a TypeCode object representing a exception.
      */
     public TypeCode create_exception_tc(String id,String name,StructMember[] members)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(org.omg.CORBA.TCKind._tk_except,id,name,members);
          }
          else
          {
               return orb.create_exception_tc(id,name,members);
          }
     }

     /**
      * Create a TypeCode object representing an interface.
      * @param id interface repository ID
      * @param name name of interface
      * @return a TypeCode object representing an interface.
      */
     public TypeCode create_interface_tc(String id,String name)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(id,name);
          }
          else
          {
               return orb.create_interface_tc(id,name);
          }
     }

     /**
      * Create a TypeCode object representing a string.
      * @param bound maximum size of string, or 0 to indicate unbounded string
      * @return a TypeCode object representing a string.
      */
     public TypeCode create_string_tc(int bound)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(org.omg.CORBA.TCKind._tk_string,bound);
          }
          else
          {
               return orb.create_string_tc(bound);
          }
     }

     /**
      * Create a TypeCode object representing a wide string.
      * @param bound maximum size of string, or 0 to indicate unbounded string
      * @return a TypeCode object representing a wide string.
      */
     public TypeCode create_wstring_tc(int bound)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(org.omg.CORBA.TCKind._tk_wstring,bound);
          }
          else
          {
               return orb.create_wstring_tc(bound);
          }
     }

     /**
      * Create a TypeCode object representing a sequence.
      * @param bound maximum size of sequence, or 0 to indicate unbounded sequence
      * @param element_type the element type
      * @return a TypeCode object representing a sequence.
      */
     public TypeCode create_sequence_tc(int bound,TypeCode element_type)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(org.omg.CORBA.TCKind._tk_sequence,bound,element_type);
          }
          else
          {
               return orb.create_sequence_tc(bound,element_type);
          }
     }

     /**
      * Create a TypeCode object representing a recursive sequence.
      * Throws org.omg.CORBA.NO_IMPLEMENT if the singleton orb is BrgOrb.
      * @param bound maximum size of sequence, or 0 to indicate unbounded sequence
      * @param offset offset within sequence
      * @deprecated
      */
     public org.omg.CORBA.TypeCode create_recursive_sequence_tc(int bound,int offset)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               throw new org.omg.CORBA.NO_IMPLEMENT();
          }
          else
          {
               return orb.create_recursive_sequence_tc(bound,offset);
          }
     }

     /**
      * Create a TypeCode object representing an array.
      * @param length size of array
      * @param element_type the element type
      * @return a TypeCode object representing an array.
      */
     public TypeCode create_array_tc(int length,TypeCode element_type)
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               return new com.sas.net.brg.orb.TypeCode(org.omg.CORBA.TCKind._tk_array,length,element_type);
          }
          else
          {
               return orb.create_array_tc(length,element_type);
          }
     }

     /**
      * Get a Current object
      * If the singleton orb is BrgOrb, then this method throws
      * org.omg.CORBA.NO_IMPLEMENT.
      * @return a Current object
      * @deprecated
      */
     public org.omg.CORBA.Current get_current()
     {
          ORB orb = ORB.init();

          if (orb instanceof BrgOrb)
          {
               throw new org.omg.CORBA.NO_IMPLEMENT();
          }
          else
          {
               return orb.get_current();
          }
     }

     /**
      * Create an Any object.
      * @return an Any object
      */
     public org.omg.CORBA.Any create_any()
     {
          return new AnyImpl(this);
     }

     // public methods
     //---------------------------------------------------------
     public synchronized int getNextClientID()
     {
          // this might not be atomic so use synchronization lock
          return clientIDCounter++ << 20;
     }

     public void addEngine(ProtocolEngineInterface engine)
     {
          synchronized (engineSet)
          {
               engineSet.add(engine);
          }
     }

     public GenericObject getGenericObject()
     {
          return new GenericObject();
     }

     public void returnGenericObject(GenericObject obj)
     {
     }

     public DelegateBase getDelegateBase()
     {
          return new DelegateBase(this);
     }

     public void returnDelegateBase(DelegateBase delegateBase)
     {
     }

     public Map getActiveObjectMap()
     {
          return activeObjectMap;
     }

     public ActiveObjectMapKey getActiveObjectMapKey()
     {
          return new ActiveObjectMapKey();
     }

     public void returnActiveObjectMapKey(ActiveObjectMapKey activeObjectMapKey)
     {
     }

     public org.omg.CORBA.Object propertiesToObject(Properties properties)
     {
          Properties copy = (Properties)properties.clone();
          return propertiesToObjectNoCopy(copy);
     }

     public org.omg.CORBA.Object propertiesToObjectNoCopy(Properties properties)
     {
          String clsid = properties.getProperty("clsid");

          if (clsid == null)
          {
               // {0}: The request for a remote object does not specify a class ID.
               String messageIndex = "orb.O000.ex.msg";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex);
               throw new INV_OBJREF(message);
          }

          // get a protocol engine
          ProtocolEngineInterface engine = null;
          String protocol = properties.getProperty("protocol");

          if ("bridge".equals(protocol))
          {
               engine = new BridgeProtocolEngine(this,properties);
          }
          else
          {
               // {0}: The stringified object reference contains unknown protocol: {1}.
               String messageIndex = "orb.O001.ex.msg";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex,protocol);
               throw new INV_OBJREF(message);
          }

          // store the protocol engine so it can later be shut down
          synchronized (engineSet)
          {
               engineSet.add(engine);
          }

          // get the requested object
          return engine.newRemoteObject(clsid);
     }

     public Properties objectToProperties(org.omg.CORBA.Object obj)
     {
          DelegateBase delegate = (DelegateBase)((ObjectImpl)obj)._get_delegate();
          ProtocolEngineInterface engine = delegate.getProtocolEngine();
          return engine.getProperties();
     }

     // These methods decode the URL format used in
     // string_to_object and convert it to a
     // java.util.Properties object.
     //--------------------------------------------
     public static Properties referenceToProperties(String reference)
     {
          String url = reference;
          Properties properties = new Properties();
          int index0 = url.indexOf("://");

          if (index0 < 0)
          {
               // protocol (that's all there is to the url)
               properties.put("protocol",url);
               return properties;
          }

          String s = url.substring(0,index0);
          int index2 = s.indexOf(":");

          if (index2 < 0)
          {
               properties.put("protocol",s);
          }
          else
          {
               properties.put("protocol",s.substring(0,index2));
               properties.put("subprotocol",s.substring(index2+1));
          }

          index0 += 3;
          int index1 = url.indexOf("/",index0);

          // protocol://host:port/clsid?name1=value1&name2=value2
          //            0        1
          //-----------------------------------------------------
          if (index1 < 0)
          {
               // protocol://host:port
               s = url.substring(index0);
          }
          else
          {
               // protocol://host:port/...
               s = url.substring(index0,index1);
          }

          if (s.length() > 0)
          {
               index2 = s.indexOf(":");

               // protocol://host:port/clsid?name1=value1&name2=value2
               //            0   2    1
               //-----------------------------------------------------
               if (index2 < 0)
               {
                    // protocol://host/...
                    properties.put("host",s);
               }
               else
               {
                    // protocol://host:port/...
                    properties.put("host",s.substring(0,index2));
                    properties.put("port",s.substring(index2+1));
               }
          }

          if (index1 < 0)
          {
               // protocol://host:port
               return properties;
          }

          index0 = index1 + 1;
          final int length = url.length();

          if (index0 >= length)
          {
               // protocol://host:port/
               return properties;
          }

          index1 = url.indexOf("?",index0);

          // protocol://host:port/clsid?name1=value1&name2=value2
          //                      0    1
          //-----------------------------------------------------
          if (index1 < 0)
          {
               // protocol://host:port/clsid
               properties.put("clsid",url.substring(index0));
               return properties;
          }

          properties.put("clsid",url.substring(index0,index1));
          index0 = index1 + 1;

          while (index0 < length)
          {
               index1 = url.indexOf("=",index0);

               // protocol://host:port/clsid?name1=value1&name2=value2
               //                            0    1
               //-----------------------------------------------------
               if (index1 < 0)
               {
                    // protocol://host:port/clsid?name1
                    properties.put(decode(url.substring(index0)),"");
                    return properties;
               }

               String name = url.substring(index0,index1);
               index0 = index1 + 1;

               if (index0 < length)
               {
                    index1 = url.indexOf("&",index0);

                    // protocol://host:port/clsid?name1=value1&name2=value2
                    //                                  0     1
                    //-----------------------------------------------------
                    if (index1 < 0)
                    {
                         // protocol://host:port/clsid?name1=value1
                         properties.put(decode(name),decode(url.substring(index0)));
                         return properties; // this is where we typically return from
                    }

                    properties.put(decode(name),decode(url.substring(index0,index1)));
                    index0 = index1 + 1;
               }
               else
               {
                    // protocol://host:port/clsid?name1=
                    properties.put(decode(name),"");
                    return properties;
               }
          }

          // protocol://host:port/clsid?name1=value1&
          return properties;
     }

     public static String propertiesToReference(Properties properties)
     {
          String host = properties.getProperty("host","localhost");
          String port = properties.getProperty("port","5312");
          String clsid = properties.getProperty("clsid","440196D4-90F0-11D0-9F41-00A024BB830C");
          StringBuffer buffer = new StringBuffer("bridge://");
          buffer.append(host);
          buffer.append(":");
          buffer.append(port);
          buffer.append("/");
          buffer.append(clsid);

          StringBuffer query = null;
          Enumeration propertyNames = properties.propertyNames();

          while (propertyNames.hasMoreElements())
          {
               String propertyName = (String)propertyNames.nextElement();

               if (!propertyName.equals("host") && !propertyName.equals("port") && !propertyName.equals("clsid"))
               {
                    if (query == null)
                    {
                         query = new StringBuffer();
                    }
                    else
                    {
                         query.append("&");
                    }

                    query.append(URLEncoder.encode(propertyName));
                    query.append("=");
                    query.append(URLEncoder.encode(properties.getProperty(propertyName)));
               }
          }

          if (query != null)
          {
               buffer.append("?");
               buffer.append(query.toString());
          }

          return buffer.toString();
     }

     // internal stuff
     //---------------------------------------------------------
     private Set engineSet = null;
     private Set serverSet = null;
     private Map activeObjectMap = null;
     private int clientIDCounter = 1;
     private String[] initialServices = null;
     private POAImpl rootPOA = null;

     private static String decode(String s)
     {
          if (s.indexOf('+') == -1 && s.indexOf('%') == -1)
          {
               return s;
          }

          try
          {
               return URLDecoder.decode(s);
          }
          catch (Exception e)
          {
               // {0}: An element of the stringified object reference was not properly URL encoded: {1}.
               String messageIndex = "orb.O002.ex.msg";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex,s);
               throw new org.omg.CORBA.BAD_PARAM(message);
          }
     }

     private void initialize(Properties properties)
     {
          if (properties == null)
          {
               return;
          }

          String serverNames = properties.getProperty("serverNames");

          if (serverNames != null && serverNames.length() > 0)
          {
               Set keys = properties.keySet();
               StringTokenizer nameTokenizer = new StringTokenizer(serverNames,",");

               while (nameTokenizer.hasMoreTokens())
               {
                    String serverName = nameTokenizer.nextToken().trim();
                    String prefix = serverName + ".";
                    final int prefixLength = prefix.length();
                    Properties serverProperties = new Properties();
                    Iterator keysIterator = keys.iterator();

                    while (keysIterator.hasNext())
                    {
                         String key = (String)keysIterator.next();

                         if (key.startsWith(prefix) && key.length() > prefixLength)
                         {
                              String value = properties.getProperty(key);
                              serverProperties.put(key.substring(prefixLength),value);
                         }
                    }

                    String serverProtocol = serverProperties.getProperty("protocol");

                    if (serverProtocol == null)
                    {
                         // {0}: A server name not associated with a server protocol was encountered: {1}.
                         String messageIndex = "orb.O003.ex.msg";
                         String messagePattern = RB.getStringResource(messageIndex);
                         String message = Message.format(messagePattern,messageIndex,serverName);
                         throw new org.omg.CORBA.BAD_PARAM(message);
                    }

                    ProtocolServerInterface server = null;

                    if (serverProtocol.equals("bridge"))
                    {
                         server = new BridgeProtocolServer(this,serverProperties);
                    }
                    else
                    {
                         // {0}: An unknown server protocol was encountered: {1}.
                         String messageIndex = "orb.O004.ex.msg";
                         String messagePattern = RB.getStringResource(messageIndex);
                         String message = Message.format(messagePattern,messageIndex,serverProtocol);
                         throw new org.omg.CORBA.BAD_PARAM(message);
                    }

                    serverSet.add(server);
               }
          }
     }
}