package com.sas.net.brg.sam;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.util.ArrayList;
import java.util.List;
import org.omg.CORBA.BooleanHolder;
import org.omg.CORBA.IntHolder;
import com.sas.net.brg.ParameterHandleInterface;
//
// History:
//
// Date    Name            Remarks
// 23Dec97 B Beers         Created.
// 23Dec97 B Beers         Adapted for SimpleCommand interface.
// 05Feb98 B Beers         Updated for changes in protocol.
// 02Apr98 B Beers         Added support for long arrays.
// 10Sep98 B Beers         Added cryptography support.
// 07Dec98 B Beers         Added event- and request-by-name.
// 06Jan99 B Beers         Moved SimpleCommand from com.sas.util.
// 28Feb01 B Beers         S0110393: Add bind and server flow features.
// 16Mar01 B Beers         S0110897: Improve performance for large parameters.
// 23Aug01 B Beers         S0124519: Split socket driver code into SocketAccessMethod.
// 10Apr02 B Beers         S0141718: Remove dependency on netutil.jar.
//
/**
 * This class is a specialization of Packet for packets sent by the server
 * to the client. It contains code designed to handle inbound packets.
 * @author Biff Beers
 */
public class InboundPacket extends Packet implements Runnable,ParameterHandleInterface
{
     protected InboundPacket(SocketAccessMethod accessMethod)
     {
          super();
          this.accessMethod = accessMethod;
     }

     public String toString()
     {
          String lineSeparator = System.getProperty("line.separator");
          StringBuffer buf = new StringBuffer(super.toString());

          if (payload != null && payloadSize > 0)
          {
               List dosParameters = new ArrayList(mapCount);
               int[] dosMaps = accessMethod.reconcileDosParameters(this,dosParameters);
               int parmCount = dosParameters.size();

               for (int i=0;i<parmCount;i++)
               {
                    buf.append(SocketAccessMethod.dosTypeToString(dosMaps[i]));

                    switch (dosMaps[i])
                    {
                         case SocketAccessMethod.DOS_MP_UCHAR:
                         {
                              String a = (String)dosParameters.get(i);

                              if (a.length() > 128)
                              {
                                   a = a.substring(0,128);
                              }

                              buf.append(" ").append(a);
                              break;
                         }

                         case SocketAccessMethod.DOS_MP_BINARY:
                         {
                              byte[] a = (byte[])dosParameters.get(i);
                              int l = (a == null) ? 0 : a.length;

                              for (int j=0;j<16;j++)
                              {
                                   buf.append(" ");

                                   if (j < l)
                                   {
                                        int b = a[j];

                                        if (b >= 0x00 && b <= 0x0F)
                                        {
                                             buf.append("0");
                                        }

                                        buf.append(Integer.toHexString((b>=0)?b:b+256));
                                   }
                                   else
                                   {
                                        buf.append("  ");
                                   }
                              }

                              if (l >= 16)
                              {
                                   buf.append(" (").append(l).append(" elements)");
                              }

                              break;
                         }

                         case SocketAccessMethod.DOS_MP_SHORT:
                         {
                              short[] a = (short[])dosParameters.get(i);
                              int count = a.length;

                              for (int j=0;j<count&&j<16;j++)
                              {
                                   buf.append(" ").append(Short.toString(a[j]));
                              }

                              if (count >= 16)
                              {
                                   buf.append(" (").append(count).append(" elements)");
                              }

                              break;
                         }

                         case SocketAccessMethod.DOS_MP_UINT:
                         case SocketAccessMethod.DOS_MP_LONG:
                         case SocketAccessMethod.DOS_MP_ULONG:
                         {
                              int[] a = (int[])dosParameters.get(i);
                              int count = a.length;

                              for (int j=0;j<count&&j<16;j++)
                              {
                                   buf.append(" ").append(Integer.toString(a[j]));
                              }

                              if (count >= 16)
                              {
                                   buf.append(" (").append(count).append(" elements)");
                              }

                              break;
                         }

                         case SocketAccessMethod.DOS_MP_DOUBLE:
                         {
                              double[] a = (double[])dosParameters.get(i);
                              int count = a.length;

                              for (int j=0;j<count&&j<16;j++)
                              {
                                   buf.append(" ").append(Double.toString(a[j]));
                              }

                              if (count >= 16)
                              {
                                   buf.append(" (").append(count).append(" elements)");
                              }

                              break;
                         }
                    }

                    buf.append(lineSeparator);
               }
          }

          return buf.toString();
     }

     public void run()
     {
          switch (getType())
          {
               case Packet.REPLY:
               {
                    accessMethod.handleReply(this);
                    break;
               }

               case Packet.INTERNAL:
               {
                    accessMethod.handleInternal(this);
                    break;
               }

               case Packet.EVENT_BY_NAME:
               {
                    accessMethod.handleEvent(this);
                    break;
               }

               case Packet.REQUEST:
               case Packet.REQUEST_BY_NAME:
               {
                    accessMethod.handleRequest(this);
                    break;
               }
          }
     }

     public synchronized int incrementReferenceCount()
     {
          return ++referenceCount;
     }

     public synchronized int decrementReferenceCount()
     {
          return --referenceCount;
     }

     public synchronized int getReferenceCount()
     {
          return referenceCount;
     }

     public boolean isEmpty()
     {
          switch (getType())
          {
               case Packet.EVENT_BY_NAME:
               case Packet.REQUEST_BY_NAME:
               {
                    ByteArrayInputStream inputBuffer = new ByteArrayInputStream(payload,0,payloadSize);
                    DataInputStream inputFilter = new DataInputStream(inputBuffer);
                    IntHolder typeHolder = new IntHolder();
                    IntHolder repFactorHolder = new IntHolder();
                    BooleanHolder unsignedHolder = new BooleanHolder();
                    int maps = accessMethod.readMap(inputFilter,typeHolder,repFactorHolder,unsignedHolder);
                    return (mapCount == maps);
               }

               default:
               {
                    return (mapCount == 0);
               }
          }
     }

     private static final int minimumPayloadArraySize = 1024;
     private static final int maximumPayloadArraySize = 32767;
     private SocketAccessMethod accessMethod = null;
     private byte[] payload = null;
     private int payloadSize = 0;
     private int mapCount = 0;
     private int referenceCount = 0;

     protected byte[] getPayload()
     {
          return payload;
     }

     protected void setPayloadSize(int payloadSize)
     {
          this.payloadSize = payloadSize;
     }

     protected int getPayloadSize()
     {
          return payloadSize;
     }

     protected void ensurePayloadCapacity(int capacity)
     {
          int target = Math.max(capacity,minimumPayloadArraySize);

          if (payload == null || payload.length < target)
          {
               payload = new byte[target];
          }
     }

     protected void setMapCount(int mapCount)
     {
          this.mapCount = mapCount;
     }

     protected int getMapCount()
     {
          return mapCount;
     }

     protected void reset()
     {
          payloadSize = 0;
          mapCount = 0;

          if (payload != null && payload.length > maximumPayloadArraySize)
          {
               payload = null;
          }
     }
}