// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.net.brg.sam;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import org.omg.CORBA.NVList;

//
// History:
//
// Date    Name            Remarks
// 23Dec97 B Beers         Created.
// 07Dec98 B Beers         Added event- and request-by-name.
// 28Feb01 B Beers         S0110393: Add bind and server flow features.
// 23Aug01 B Beers         S0124519: Split socket driver code into SocketAccessMethod.
// 11Sep01 B Beers         S0126305: Prevent NullPointerException when sending a packet with no UUID.
//
/**
 * A packet consists of a 64-byte header and additional parameters. All
 * communication between the client and server (except the first message)
 * must be in the form of this packet. There are six kinds of packets:
 * <DL>
 * <DT>REQUEST<DD>a method call from one object to another for which a REPLY
 * is expected.
 * <DT>REQUEST_BY_NAME<DD>a method call from one object to another for which a REPLY
 * is expected. The name of the method is the first string parameter.
 * <DT>REPLY<DD>the results of a method call from one object to another
 * <DT>EVENT<DD>a method call from one object to zero or more listening objects
 * for which no REPLY is expected.
 * <DT>EVENT_BY_NAME<DD>a method call from one object to zero or more listening objects
 * for which no REPLY is expected. The name of the event is the first string parameter.
 * <DT>INTERNAL<DD>administrative information passed between the client and server.
 * </DL>
 * This class contains all the data that is common to packets sent by the client
 * to the server (outbound packets) and packets sent by the server to the client
 * (inbound packets). However, inbound and outbound packets have specialized code
 * that is not common, so this class is abstract and can only be instantiatied 
 * through one of its subclasses, InboundPacket and OutboundPacket.
 * @author Biff Beers
 */
public class Packet
{
     // constructors
     //-------------------------------------------------------------
     protected Packet()
     {
     }
     
     // overridden super-class methods
     //-------------------------------------------------------------
     public String toString()
     {
          String lineSeparator = System.getProperty("line.separator");
          StringBuffer b = new StringBuffer();
          b.append("Type: ").append(typeToString(getType())).append(lineSeparator);
          b.append("ClientId ").append(getClientId()).append(lineSeparator);
          b.append("ObjectId ").append(getObjectId()).append(lineSeparator);
          b.append("MethodId ");

          if (getType() == INTERNAL)
          {
               b.append(SocketAccessMethod.methodIDToString(getMethodId()));
          }
          else
          {
               b.append(getMethodId());
          }

          b.append(lineSeparator);
          b.append("ReplyId ").append(getReplyId()).append(lineSeparator);
          b.append("uuid ").append(getUuid()).append(lineSeparator);
          return b.toString();
     }
     
     // internal stuff
     //-------------------------------------------------------------
     protected static final int INTERNAL = 0;
     protected static final int REQUEST = 1;
     protected static final int REPLY = 2;
     protected static final int EVENT = 3;
     protected static final int REQUEST_BY_NAME = 4;
     protected static final int EVENT_BY_NAME = 5;
     protected static final int PACKET_TYPE_COUNT = 6;
     private static final String defaultUUID = "00000000-0000-0000-0000-000000000000";
     private int type = 0;
     private int clientId = 0;
     private int objectId = 0;
     private int methodId = 0;
     private int replyId = -1;
     private String uuid = defaultUUID;

     protected void reset()
     {
          type = 0;
          clientId = 0;
          objectId = 0;
          methodId = 0;
          replyId = -1;
          uuid = defaultUUID;
     }

     protected int getType()
     {
          return type;
     }

     protected int getClientId()
     {
          return clientId;
     }

     protected int getObjectId()
     {
          return objectId;
     }

     protected int getMethodId()
     {
          return methodId;
     }
     
     protected int getReplyId()
     {
          return replyId;
     }

     protected String getUuid()
     {
          return uuid;
     }

     protected void setType(int type)
     {
          this.type = type;
     }
     
     protected void setClientId(int clientId)
     {
          this.clientId = clientId;
     }
     
     protected void setObjectId(int objectId)
     {
          this.objectId = objectId;
     }
                                          
     protected void setMethodId(int methodId)
     {
          this.methodId = methodId;
     }
     
     protected void setReplyId(int replyId)
     {
          this.replyId = replyId;
     }
     
     protected void setUuid(String uuid)
     {
          this.uuid = (uuid != null) ? uuid : defaultUUID;
     }

     private static String typeToString(int type)
     {
          switch (type)
          {
               case INTERNAL:return "INTERNAL";
               case REQUEST:return "REQUEST";
               case REPLY:return "REPLY";
               case EVENT:return "EVENT";
               case REQUEST_BY_NAME:return "REQUEST_BY_NAME";
               case EVENT_BY_NAME:return "EVENT_BY_NAME";
               default:return "WHO_KNOWS?";
          }
     }
      
}