// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.net.crypto;

import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * An interface that provides access to the encryption and decryption
 * services of this cryptograph algorithm.
 * @author Biff Beers
 * @version
 * <DL>
 * <DT>version 1.00, 24Jun98, BFB<DD> Created.
 * </DL>
 */
public interface CipherInterface
{
     /**
      * Agree on a shared key with the server. All encryption and decryption
      * done by this cipher object will use the generated key.
      * @param iStream a stream for receiving messages from the server
      * @param oStream a stream for sending messages to the server
      * @exception java.io.IOException if there is an error sending or
      * receiving data
      * @exception com.sas.net.crypto.CryptoException if there is an error
      * in execution of the key exchange algorithm
      */
     public abstract void keyExchange(InputStream iStream,OutputStream oStream) throws IOException,CryptoException;

     /**
      * Get the exact length of the cipher text that would be generated by
      * encrypting plain text of a given length using this cryptography
      * algorithm.
      * @param plainTextLength length of plain text in bytes
      * @return the exact length of the cipher text that would be generated by
      * encrypting plain text of a given length using this cryptography
      * algorithm
      */
     public abstract int getCipherTextLength(int plainTextLength);

     /**
      * Get the approximate length of the plain text that would be generated by
      * decrypting cipher text of a given length using this cryptography
      * algorithm. The exact length is guaranteed to be less than or equal to
      * the value returned by this method. The difference between the predicted
      * length and the exact length is typically less than 16 bytes, but that
      * is not guaranteed.
      * @param cipherTextLength length of cipher text (in bytes)
      * @return the approximate length of the plain text that would be generated by
      * decrypting cipher text of a given length using this cryptography
      * algorithm
      */
     public abstract int getMaxPlainTextLength(int cipherTextLength);

     /**
      * Encrypt a buffer.
      * @param plainText the buffer to encrypt
      * @param offset the offset within the buffer where the plain text to be ecrypted begins.
      * @param length the length of the plain text to be encrypted
      * @param oStream a stream to write the cipher text to.
      * @exception java.io.IOException if there is an error writing cipher text to the
      * stream
      * @exception com.sas.net.crypto.CryptoException if there is an error encrypting
      * the plain text
      */
     public abstract int encrypt(byte[] plainText,int offset,int length,OutputStream oStream) throws IOException,CryptoException;

     /**
      * Decrypt a stream.
      * @param iStream a stream of cipher text
      * @param cipherTextLength the length of the cipher text to read from the stream (in bytes)
      * @param buffer a buffer to hold the plain text. Use getMaxPlainTextLength to predict the
      * buffer size needed.
      * @param offset the offset within the buffer to begin writing plain text
      * @exception java.io.IOException if there is an error reading cipher text from the
      * stream
      * @exception com.sas.net.crypto.CryptoException if there is an error decrypting
      * the cipher text
      */
     public abstract int decrypt(InputStream iStream,int cipherTextLength,byte[] buffer,int offset) throws IOException,CryptoException;
     
     /**
      * Close the cipher object. This method clears from memory any sensitive information
      * such as shared keys and random numbers.
      * @exception com.sas.net.crypto.CryptoException if there is an error.
      */
     public abstract void close() throws CryptoException;
}