// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.net.crypto;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

//
// History:
//
// Date    Name            Remarks
// 24Jun98 Biff Beers      Created.
// 31Jun98 Biff Beers      Removed spurious println.
// 10Sep98 Biff Beers      Added getSupportedNames.
// 21Jan99 Biff Beers      Removed spurious println.
// 01Apr99 Biff Beers      Changed search order for crypto providers.
//
/**
 * An abstract class for accessing the encryption/decryption services
 * of a particular cryptography algorithm.
 * @author Biff Beers
 */
public abstract class Crypto
{
     // public stuff
     //---------------------------------------------------------
     /**
      * Get a cipher object for a particular cryptography algorithm. Though
      * all cipher objects use the same cryptography algorithm,
      * each has its own set of keys.
      * @return a cipher object implementing this cryptography algorithm.
      * @exception com.sas.net.crypto.CryptoException if there is an error.
      */
     public abstract CipherInterface openCipher() throws CryptoException;

     /**
      * Get a list of cryptography algorithm names supported by this
      * cryptography object.
      * @return a list of cryptography algorithm names supported by this
      * cryptography object.
      */
     public abstract String getSupportedNames();

     /**
      * Get the name of this cryptography algorithm.
      * @return the name of this cryptography algorithm
      */
     public String getName()
     {
          return name;
     }

     /**
      * Set the name of this cryptography algorithm.
      * @param name the name of this cryptography algorithm
      */
     public void setName(String name)
     {
          this.name = name.toUpperCase()/*I18nOK:MTH*/;
     }

     /**
      * Get an instance of the desired crypto algorithm. Returns null if
      * the desired algorithm cannot be loaded.
      * @param properties user-specified options
      * @return instance of the desired crypto algorithm
      */
     public static Crypto getCrypto(String algorithm,Properties properties)
     {
          if (algorithm == null)
          {
               return null;
          }

          String providerList = properties.getProperty("encryptionProviders",defaultProviders);
          String providerName = null;
          int index = 0;
          Class[] argt = new Class[1];
          Object[] argv = new Object[1];

          do
          {
               providerName = getNameFromList(providerList,',',index);

               if (providerName != null)
               {
                    try
                    {
                         Class providerClass = Class.forName(providerName);
                         argt[0] = java.util.Properties.class;
                         Constructor providerConstructor = providerClass.getConstructor(argt);
                         argv[0] = properties;
                         Crypto provider = (Crypto)providerConstructor.newInstance(argv);
                         String supportedNames = provider.getSupportedNames();
                         String supportedName = null;
                         int j = 0;

                         do
                         {
                              supportedName = getNameFromList(supportedNames,',',j);

                              if (supportedName != null)
                              {
                                   if (algorithm.equalsIgnoreCase(supportedName))
                                   {
                                        provider.setName(supportedName);
                                        return provider;
                                   }
                              }

                              j++;
                         }
                         while (supportedName != null);
                    }
                    catch (ClassNotFoundException e)
                    {
                         if (!providerName.equals("com.sas.net.crypto.rsa.RsaCrypto"))
                         {
                              e.printStackTrace();
                         }
                    }
                    catch (NoSuchMethodException e)
                    {
                         e.printStackTrace();
                    }
                    catch (InstantiationException e)
                    {
                         e.printStackTrace();
                    }
                    catch (IllegalAccessException e)
                    {
                         e.printStackTrace();
                    }
                    catch (InvocationTargetException  e)
                    {
                         e.printStackTrace();
                    }
               }

               index++;
          }
          while (providerName != null);

          return null;
     }

     // internal stuff
     //---------------------------------------------------------
     private static final String defaultProviders = "com.sas.net.crypto.sas.SasCrypto,com.sas.net.crypto.rsa.RsaCrypto";
     private String name = null;

     private static String getNameFromList(String list,char delim,int index)
     {
          if (list == null)
          {
          	return null;
          }

          int offset = 0;
          int end = 0;
          int listLength = list.length();

          for (int i=0;i<index;i++)
          {
               for (;offset<listLength;offset++)
               {
                    char c = list.charAt(offset);

                    if (c == delim)
                    {
                         offset++;
                         break;
                    }
               }
          }

          if (offset >= listLength)
          {
               return null;
          }

          for (end=offset;end<listLength;end++)
          {
               char c = list.charAt(end);

               if (c == delim)
               {
                    break;
               }
          }

          if (end - offset == 0)
          {
               return null;
          }

          return list.substring(offset,end);
     }
}