// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.net.crypto.sas;

import java.io.EOFException;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Properties;
import com.sas.net.crypto.Crypto;
import com.sas.net.crypto.CryptoException;
import com.sas.net.crypto.CipherInterface;

//
// History:
//
// Date    Name            Remarks
// 24Jun98 Biff Beers      Created.
// 10Sep98 Biff Beers      Added getSupportedNames.
// 12Oct98 Biff Beers      Marked final to prevent overrides.
// 03Feb99 Biff Beers      Added flush call.
// 04Feb00 Biff Beers      S0074465: Handle 64K+ buffers.
// 10Jul03 Biff Beers      S0210021: Fix concurrent access of scratch buffers.
// 28Aug03 Biff Beers      S0218349: Bypass zmod4 where possible to improve performance.
//
/**
 * This class implements SAS Proprietary cryptography.
 * @author Biff Beers
 */
public final class SasCrypto extends Crypto implements CipherInterface
{
     // constructors
     //---------------------------------------------------------
     /**
      * Construct a SAS proprietary cryptography object.
      */
     public SasCrypto(Properties properties)
     {
          super();
     }

     // overridden super-class methods
     //---------------------------------------------------------
     /**
      * @see com.sas.net.crypto.Crypto#getSupportedNames
      */
     public String getSupportedNames()
     {
          return name;
     }

     /**
      * @see com.sas.net.crypto.Crypto#openCipher
      */
     public CipherInterface openCipher() throws CryptoException
     {
          return this;
     }

     // implement CipherInterface
     //---------------------------------------------------------
     /**
      * @see com.sas.net.crypto.CipherInterface#keyExchange
      */
     public void keyExchange(InputStream iStream,OutputStream oStream) throws IOException,CryptoException
     {
     }

     /**
      * @see com.sas.net.crypto.CipherInterface#getCipherTextLength
      */
	public int getCipherTextLength(int plainTextLength)
     {
          if (plainTextLength == 0)
          {
               return 0;
          }

          final int segments = plainTextLength / maxPlain;
          final int remainder = plainTextLength % maxPlain;
          return (segments * maxCipher) + ((remainder + 5) / 3) * 4;
     }

     /**
      * @see com.sas.net.crypto.CipherInterface#getMaxPlainTextLength
      */
	public int getMaxPlainTextLength(int cipherTextLength)
     {
          if (cipherTextLength < 4)
          {
               return 0;
          }

          final int segments = cipherTextLength / maxCipher;
          final int remainder = cipherTextLength % maxCipher;
          return (segments * maxPlain) + ((remainder * 3) / 4) - 5 + 3;
     }

     /**
      * @see com.sas.net.crypto.CipherInterface#encrypt
      */
     public int encrypt(byte[] plainText,int offset,int plainTextLength,OutputStream oStream) throws IOException,CryptoException
     {
          if (plainTextLength == 0)
          {
               return 0;
          }

          final int bufferLength = (plainTextLength > maxPlain) ? maxCipher : getCipherTextLength(plainTextLength);
          final int segments = plainTextLength / maxPlain;
          final int remainder = plainTextLength % maxPlain;
          int cipherTotalLength = 0;
          int cipherSegmentLength = 0;

          synchronized (encodeBuffer)
          {
               if (encodeBuffer.length < bufferLength)
               {
                    encodeBuffer = new byte[bufferLength];
               }

               for (int i=0;i<segments;i++)
               {
                    cipherSegmentLength = zscode(encryptPower,plainText,offset,maxPlain,false,encodeBuffer,0);
                    oStream.write(encodeBuffer,0,cipherSegmentLength);
                    cipherTotalLength += cipherSegmentLength;
                    offset += maxPlain;
                    plainTextLength -= maxPlain;
               }

               cipherSegmentLength = zscode(encryptPower,plainText,offset,plainTextLength,false,encodeBuffer,0);
               oStream.write(encodeBuffer,0,cipherSegmentLength);
               oStream.flush();
               cipherTotalLength += cipherSegmentLength;
          }

          return cipherTotalLength;
     }

     /**
      * @see com.sas.net.crypto.CipherInterface#decrypt
      */
     public int decrypt(InputStream iStream,int cipherTextLength,byte[] buffer,int offset) throws IOException,CryptoException
     {
          if (cipherTextLength < 4)
          {
               return 0;
          }

          for (int i=offset;i<buffer.length;i++)
          {
               buffer[i] = (byte)'x';
          }

          final int segments = cipherTextLength / maxCipher;
          final int remainder = cipherTextLength % maxCipher;
          int plainTotalLength = 0;
          int plainSegmentLength = 0;
          final int bufferLength = (cipherTextLength > maxCipher) ? maxCipher : cipherTextLength;

          synchronized (decodeBuffer)
          {
               if (decodeBuffer.length < bufferLength)
               {
                    decodeBuffer = new byte[bufferLength];
               }

               for (int i=0;i<segments;i++)
               {
                    readFully(iStream,decodeBuffer,0,maxCipher);
                    cipherTextLength -= maxCipher;
                    plainSegmentLength = zscode(decryptPower,decodeBuffer,0,maxCipher,true,buffer,offset);
                    plainTotalLength += plainSegmentLength;
                    offset += plainSegmentLength;
               }

               readFully(iStream,decodeBuffer,0,cipherTextLength);
               plainSegmentLength = zscode(decryptPower,decodeBuffer,0,cipherTextLength,true,buffer,offset);
               plainTotalLength += plainSegmentLength;
          }

          return plainTotalLength;
     }

     /**
      * @see com.sas.net.crypto.CipherInterface#close
      */
     public void close() throws CryptoException
     {
     }

     // internal stuff
     //---------------------------------------------------------
     private static final String name = "SASPROPRIETARY";
     private static final long intMask = 0x00000000FFFFFFFFL;
     private static final long shortMask = 0x000000000000FFFFL;
     private static final long byteMask = 0x00000000000000FFL;
     private static final int encryptPower = (int)0x4093b9ab;
     private static final int decryptPower = 3;
     private static final int maxPlain = 65535;
     private static final int maxCipher = ((maxPlain + 5) / 3) * 4;
     private static final int minBuffer = 1024;
     private byte[] encodeBuffer = new byte[minBuffer];
     private byte[] decodeBuffer = new byte[minBuffer];

     // This method only encodes data in chunks of 64K-1 bytes or less.
     private static int zscode(int power,byte[] si,int iOff,int iLen,boolean flag,byte[] so,int oOff)
     {
          long ukey = 0x60dfc3ffL;
          long upower = power & intMask;
          long v0 = ukey >>> 16;
          long c = v0 >>> 8;
          long cc = 1L;
		for (long cd=c>>>1;cd!= 0L;cd>>>=1,cc<<=1);
          cc -= 1L;

		// the first stuff is the string length
          int i = 0;
          long m = 0L;
          long ca = 0L;
          int n = 0;

          // scratch for zmod4 which is called a zillion times
          long[] pl = new long[2];
          long[] ql = new long[2];

          if (flag) // decode
          {
          	for (i=3,v0=0L;i>=0;i--)
               {
               	v0 = (v0 << 8) + (si[i+iOff] & byteMask);
               }

			c = zpowmod(v0,upower,ukey,pl,ql) & intMask;

			// recover the string length
               m = c & shortMask;
               ca = m & byteMask;
               i = iOff + 4;
               n = oOff;
          }
          else //encode
          {
               m = si[iOff+(iLen>>>1)] & byteMask;
          	v0 = (m << 16) + iLen;
			c = zpowmod(v0,upower,ukey,pl,ql) & intMask;
			m = ((iLen + 5L) / 3L) << 2;

			for(i=0;i<4;i++)
               {
               	so[i+oOff] = (byte)(c & byteMask);
                    c >>= 8;
               }

			ca = so[oOff] & byteMask;
               i = iOff;
               n = oOff + 4;
          }

          final int end = (int)m + oOff;
          byte[] a = new byte[4];

		for(;i<iOff+iLen;)
          {
          	// move byte string
               v0 = (flag) ? (si[i++] & byteMask) : cc & ca;

			for (int j=0;j<3;j++)
               {
               	v0 = (v0 << 8) + ((i < iOff + iLen) ? (si[i++] & byteMask) : 0L);
               }

			c = zpowmod(v0,upower,ukey,pl,ql); // scramble

			for(int j=0;j<4;j++)
               {
               	a[j] = (byte)(c & byteMask);
                    c >>= 8;
               }

			if (!flag)
               {
               	so[n++] =  a[3];
               }

			if (n < end)
               {
               	so[n++] = a[2];
               }

               if (n < end)
			{
               	so[n++] = a[1];
               }

			if (n < end)
               {
               	ca = a[0];
                    so[n++] = a[0];
               }
          }

		return (int)m;
	}

     // b^p mod m
     /*--------------------------------------------------------------+
      | There is the possibility of further optimization for this    |
      | method, but we don't have time to adequately test these      |
      | ideas right now. If m (the modulus) is less than 2^31 (and   |
      | it always is in this class), then (anything % m) will be     |
      | also be less than 2^31. Thus, all these checks to make sure  |
      | the zpow4 operands are small are necessary only on the first |
      | run through the loop.                                        |
      +--------------------------------------------------------------*/
	private static long zpowmod(long b,long p,long m,long[] pl,long[] ql)
     {
          long r = 1L;

          if (p == 0L)
          {
          	return(r);
          }

		do
          {
               if (b > 0x7FFFFFFFL)
               {
               	if ((p&1L) != 0L)
                    {
                    	r = zmod4(r,b,m,pl,ql);

     				if (p == 1L)
                         {
                         	break;
                         }
                    }

     			b = zmod4(b,b,m,pl,ql); // square b mod m
               }
               else
               {
               	if ((p&1L) != 0L)
                    {
                         if (r > 0x7FFFFFFFL)
                         {
                    	     r = zmod4(r,b,m,pl,ql);
                         }
                         else
                         {
                              r = (r * b) % m;
                         }

     				if (p == 1L)
                         {
                         	break;
                         }
                    }

     			b = (b * b) % m; // square b mod m
               }

			// divide p as an unsigned long by 2
               p /= 2L;
          }
          while (true);

		return r & intMask;
	}

     // a*b mod m
	private static long zmod4(long a,long b,long m,long[] pl,long[] ql)
	{
          for (int il=0;il<pl.length;il++)
          {
               pl[il] = 0;
          }

          for (int il=0;il<ql.length;il++)
          {
               ql[il] = 0;
          }

          long s = 0L;
		a = a & intMask;
          b = b & intMask;
          m = m & intMask;
          zprod4(a,b,pl);
          pl[0] = pl[0] & intMask;
          pl[0] %= m;  // prevent overflow below
          int shift = 0;

		for(;(m&0x80000000)==0;shift++,m<<=1)
          {
          	m = m & intMask;
               pl[0] <<= 1;
               pl[0] = pl[0] & intMask;

			if ((pl[1] & 0x80000000) != 0 )
               {
               	pl[0]++;
                    pl[0] = pl[0] & intMask;
               }

			pl[1] <<= 1;
               pl[1] = pl[1] & intMask;
          }

		//---compute remainder; take out a half word at a time
          for(int i=0;i<2;i++)
          {
          	s = pl[0];

			if (i == 1)
               {
               	s <<= 16;
                    s = s & intMask;
                    s += (pl[1] >>> 16);
                    s = s & intMask;
               }

			// 16 bits of trial divisor
      		a = s / ((m >>> 16) & intMask);

			if (i == 0)
               {
               	a <<= 16;
               }

			a = a & intMask;
			zprod4(a,m,ql);

			if (ql[1] > pl[1])
               {
               	pl[0]--;
                    pl[0] = pl[0] & intMask;
               }

			pl[1] -= ql[1];
               pl[1] = pl[1] & intMask;
               pl[0] -= ql[0];
               pl[0] = pl[0] & intMask;

			while ((pl[0] & 0x80000000L) != 0)
               {
               	pl[1] += m;
                    pl[1] = pl[1] & intMask;
                    
                    if (pl[1] < m)
                    {
                    	pl[0]++;
                         pl[0] = pl[0] & intMask;
                    }
               }
          }

     	return ((pl[1] >>> shift) & intMask);
     }

	// c = a*b (with c holding up to 64 bits)
     private static void zprod4(long a,long b,long c[])
     {
     	long s11 = 0L;
          long s10 = 0L;
          long s01 = 0L;
          long s00 = 0L;
          long b1 = 0L;
		b = b & intMask;

		//---split out 4 half precision pieces
          s00 = b >>> 16;
          b1 = b - ((s00 << 16) & intMask);
          b1 = b1 & intMask;
          s10 = a >>> 16;
		s01 = a - ((s10 << 16) & intMask);
		s11 = b1 * s01;  // a1 * b1
		s01 *= s00;      // a1 * b0
          s00 *= s10;      // a0 * b0
          s10 *= b1;       // a0 * b1
          s01 = s01 & intMask;
          s00 = s00 & intMask;
          s10 = s10 & intMask;
          s11 = s11 & intMask;

		// add with overlap to form double long result
          s10 = s10 + s01;
          s10 = s10 & intMask;
          
          if (s10 < s01)
          {
          	s00 += 1 << 16;
          }

		s00 = s00 & intMask;
          s01 = s10 >>> 16;
          b1 = s10 - ((s01 << 16) & intMask);
          b1 = b1 & intMask;
          b1 <<= 16;
          b1 = b1 & intMask;
          s11 += b1;
          s11 = s11 & intMask;
          
          if (s11 < b1 )
          {
          	s00++;
          }

		s00 = s00 & intMask;
          s00 += s01;
          s00 = s00 & intMask;
          
          // save into place
          c[0] = s00;
          c[1] = s11;
          c[0] = c[0] & intMask;
          c[1] = c[1] & intMask;
     }

     private static void readFully(InputStream iStream,byte[] buffer,int offset,int length) throws IOException
     {
          int bytesRead = 0;

          do
          {
               int i = iStream.read(buffer,offset+bytesRead,length-bytesRead);

               if (i < 0)
               {
                    throw new EOFException();
               }

               bytesRead += i;
          }
          while (bytesRead < length);
     }
}