package com.sas.net.util;

import java.util.Calendar;
import java.util.Date;

//
// History:
//
// Date    Name            Remarks
// 25Mar99 B Beers         Created.
/**
 * This class has methods that convert one date format to another.
 * A date in this case is some value that represents an
 * instant in time to some precision. 
 * <P>
 * Java has a detailed collection
 * of classes for representing dates. The java.util.Date class
 * keeps track of time as the number of milliseconds that have
 * elapsed since midnight on 01Jan1970 GMT. The java.util.Calendar
 * class contain a java.util.Date value but also contains adjustments
 * for time zone and Daylight Savings Time.
 * <P>
 * CORBA date values measure the elapsed time in units of 100 nanoseconds
 * since 15Oct1582 local time. That was the date that most places starting
 * using the Gregorian Calendar.
 * <P>
 * IOM DateTime values are in CORBA format.
 * @author Biff Beers
 */
public abstract class DateConverter
{
     /**
      * Convert a Java date value to a CORBA format.
      * @param localDate a Java date
      * @returns a CORBA formatted date
      */
     public static long javaToCorba(Calendar localDate)
     {
          long gmtJavaDate = localDate.getTime().getTime();
          Calendar offsetCalendar = (Calendar)localDate.clone();
          offsetCalendar.clear();
          offsetCalendar.set(1970,0,1);
          long offset = offsetCalendar.getTime().getTime();
          long stJavaDate = gmtJavaDate - offset;
          long dtJavaDate = stJavaDate;

          if (localDate.getTimeZone().inDaylightTime(localDate.getTime()))
          {
               dtJavaDate += (3600L * 1000L);
          }

          long corbaDateInMillis = dtJavaDate + conversionFactor;
          return corbaDateInMillis * 10000L;
     }

     /**
      * Convert a CORBA date to a Java date.
      * @param corbaDate a CORBA date
      * @param localCalendar a java date object intialized with the proper
      * locale and time zone. The appropriate date will be filled in.
      */
	public static void corbaToJava(long corbaDate,Calendar localCalendar)
     {
          long corbaDateInMillis = corbaDate / 10000L;
          long dtJavaDate = corbaDateInMillis - conversionFactor;
          Calendar offsetCalendar = (Calendar)localCalendar.clone();
          offsetCalendar.clear();
          offsetCalendar.set(1970,0,1);
          long offset = offsetCalendar.getTime().getTime();
          long gmtJavaDate = dtJavaDate + offset;
          Date date = new Date(gmtJavaDate);

          if (localCalendar.getTimeZone().inDaylightTime(date))
          {
               gmtJavaDate -= (3600L * 1000L);
               date = new Date(gmtJavaDate);
          }

          localCalendar.setTime(date);
     }

     // number of milliseconds between midnight 15Oct1582 and midnight 01Jan1970
     //-------------------------------------------------------------------------
     private static final long conversionFactor = (((1970L - 1583L) * 365L) + 78L + 94L) * 24L * 3600L * 1000L;
}