package com.sas.net.util;

 //
 //History Records:
 //
 // Date    Name         Remarks
 // 03Feb99 B Beers      Created.
 //
/**
 * Base class for exceptions that may contain
 * nested exceptions. When there is a nested
 * exception, toString() and printStackTrace(),
 * etc. defer to the nested exception.
 * <P>
 * If the nested exception is itself a WrapperException,
 * then there will be yet another layer of deferal.
 * <P>
 * This is for unchecked exceptions.
 * There is also a WrapperException to use as
 * a base class for checked exceptions.
 * <P>
 * Java has it's own notion of a wrapper exception
 * in <CODE>java.lang.reflect.InvocationTargetException</CODE>.
 * This class would not exist were it not for three
 * problems with InvocationTargetException.
 * <UL>
 * <LI>The semantics of InvocationTargetException
 * imply that a method was called through
 * reflection while the semantics here are more general.
 * <LI>There is not unchecked equivalent of
 * InvocationTargetException.
 * <LI>InvocationTargetException does not properly
 * defer to the nestedException until JDK 1.2.
 * </UL>
 * <P>
 * This base class is often useful in the following
 * situation where <CODE>AnException</CODE> extends
 * <CODE>WrapperException</CODE>:
 * <CODE>
 * public void aMethod() throws AnException
 * {
 *    try
 *    {
 *       doStuff();
 *    }
 *    catch (AnotherException e)
 *    {
 *       throw new AnException(e);
 *    }
 * }
 * </CODE>
 * @author Biff Beers
 */
public abstract class RuntimeWrapperException extends RuntimeException
{
     // constructors
     //---------------------------------------------------------
     /**
      * No-argument constructor.
      */
     public RuntimeWrapperException()
     {
          super();
     }

     /**
      * Constructor with message.
      * @param message a message adding detail to the exception
      */
     public RuntimeWrapperException(String message)
     {
          super(message);
     }

     /**
      * Wrapper constructor.
      * @param nestedException exception being wrapped
      */
     public RuntimeWrapperException(Throwable nestedException)
     {
          this.nestedException = nestedException;
     }

     /**
      * Wrapper constructor.
      * @param message a message to include in the exception.
      * @param nestedException an exception being wrapped.
      */
     public RuntimeWrapperException(String message,Throwable nestedException)
     {
          super(message);
          this.nestedException = nestedException;
     }

     // overridden super class methods
     //---------------------------------------------------------
     /**
      * @see java.lang.Throwable#getMessage
      */
     public String getMessage()
     {
          if (nestedException == null)
          {
               return super.getMessage();
          }
          else
          {
               String superMessage = super.getMessage();

               if (superMessage != null)
               {
                    return super.getMessage() + "\n" + nestedException.getMessage();
               }
               else
               {
                    return nestedException.getMessage();
               }
          }
     }

     /**
      * @see java.lang.Throwable#getLocalizedMessage
      */
     public String getLocalizedMessage()
     {
          if (nestedException == null)
          {
               return super.getLocalizedMessage();
          }
          else
          {
               String superMessage = super.getLocalizedMessage();

               if (superMessage != null)
               {
                    return super.getLocalizedMessage() + "\n" + nestedException.getLocalizedMessage();
               }
               else
               {
                    return nestedException.getLocalizedMessage();
               }
          }
     }

     /**
      * @see java.lang.Throwable#toString
      */
     public String toString()
     {
          if (nestedException == null)
          {
               return super.toString();
          }
          else
          {
               String superMessage = super.toString();

               if (superMessage != null)
               {
                    return super.toString() + "\n" + nestedException.toString();
               }
               else
               {
                    return nestedException.toString();
               }
          }
     }

     /**
      * @see java.lang.Throwable#printStackTrace
      */
     public void printStackTrace()
     {
          printStackTrace(System.err);
     }

     /**
      * @see java.lang.Throwable#printStackTrace
      */
     public void printStackTrace(java.io.PrintStream p)
     {
          if (nestedException == null)
          {
               super.printStackTrace(p);
          }
          else
          {
               p.print(getClass().getName());
               String message = super.getMessage();

               if (message != null && message.length() > 0)
               {
                    p.println(": "/*I18nOK:EMS*/ + message/*I18nOK:COS*/);
               }
               else
               {
                    p.println();
               }

               nestedException.printStackTrace(p);
          }
     }

     /**
      * @see java.lang.Throwable#printStackTrace
      */
     public void printStackTrace(java.io.PrintWriter p)
     {
          if (nestedException == null)
          {
               super.printStackTrace(p);
          }
          else
          {
               p.print(getClass().getName());
               String message = super.getMessage();

               if (message != null && message.length() > 0)
               {
                    p.println(": "/*I18nOK:EMS*/ + message/*I18nOK:COS*/);
               }
               else
               {
                    p.println();
               }

               nestedException.printStackTrace(p);
          }
     }

     /**
      * @see java.lang.Throwable#fillInStackTrace
      */
     public Throwable fillInStackTrace()
     {
          if (nestedException == null)
          {
               return super.fillInStackTrace();
          }
          else
          {
               return nestedException.fillInStackTrace();
          }
     }

     // public stuff
     //---------------------------------------------------------
	/**
      * Gets the nested exception.
      * @return The nested exception.
      */
     public Throwable getNestedException()
     {
          return nestedException;
     }

     // internal stuff
     //---------------------------------------------------------
     protected Throwable nestedException = null;
}