package com.sas.net.util;

import java.io.ByteArrayOutputStream;
import java.io.Writer;
import java.io.UnsupportedEncodingException;
import java.util.logging.ErrorManager;
import java.util.logging.Filter;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.SimpleFormatter;
import java.util.logging.Handler;
//
// History:
//
// Date    Name            Remarks
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
//
/**
 * A handler for logging messages to a {@link java.io.Writer writer}.
 * <P>
 * Each <CODE>WriterHandler</CODE> is initialized using the following
 * <CODE>LogManager</CODE> configuration properties.  If properties are not defined
 * then the specified default values are used.
 * <P>
 * <UL>
 * <LI>   com.sas.util.logging.WriterHandler.level
 *     specifies the default level for the handler
 *     (defaults to <CODE>Level.ALL</CODE>).
 * <LI>   com.sas.util.logging.WriterHandler.filter
 *     specifies the name of a filter class to use
 *      (defaults to no filter).
 * <LI>   com.sas.util.logging.WriterHandler.formatter
 *     specifies the name of a formatter class to use
 *     (defaults to <CODE>java.util.logging.SimpleFormatter</CODE>).
 * </UL>
 * <P>
 * Configuration of <CODE>WriterHandler</CODE> done using the above
 * properties and the constructors will <STRONG>not</STRONG> cause
 * a <CODE>SecurityException</CODE> to be thrown even if the caller 
 * does not have <CODE>LoggingPermission("control"))</CODE>. However,
 * configuration of <CODE>WriterHandler</CODE> done using the setters
 * does require the caller to have <CODE>LoggingPermission("control"))</CODE>.
 */
public class WriterHandler extends Handler
{
    /**
     * Creates a <CODE>WriterHandler</CODE> with the given writer,
     * <CODE>Level.ALL</CODE> for the level, no filter, and
     * <CODE>java.util.logging.SimpleFormatter</CODE> for the formatter.
     * @param writer     the target writer
     */
     public WriterHandler(Writer newWriter)
     {
          this(newWriter,null);
     }

    /**
     * Creates a <CODE>WriterHandler</CODE> with the given writer
     * and formatter, <CODE>Level.ALL</CODE> for the level, and no filter.
     * @param writer     the target writer
     * @param formatter the desired formatter
     */
     public WriterHandler(Writer newWriter,Formatter newFormatter)
     {
          super();
          sealed = false;
          configure();

          if (newWriter == null)
          {
               throw new NullPointerException("\"newWriter\" cannot be null.");
          }

          this.writer = newWriter;

          if (newFormatter != null)
          {
               setFormatter(newFormatter);
          }

          sealed = true;
          writeHead();
     }

     public synchronized void close()
     {
          checkState();
          checkAccess();

          try
          {
               Formatter formatter = getFormatter();
               String tail = formatter.getTail(this);

               if (tail != null && tail.length() > 0)
               {
                    writer.write(tail);
               }
          }
          catch (Exception e)
          {
               reportError("An exception was thrown while writing the tail.",e,ErrorManager.WRITE_FAILURE);
          }

          doFlush();

          try
          {
               writer.close();
          }
          catch (Exception e)
          {
               reportError(null,e,ErrorManager.CLOSE_FAILURE);
          }

          closed = true;
     }

     public synchronized void flush()
     {
          checkState();
          checkAccess();
          doFlush();
     }

     public synchronized void publish(LogRecord logRecord)
     {
          checkState();

          if (!isLoggable(logRecord))
          {
               return;
          }

          String message = null;

          try
          {
               Formatter formatter = getFormatter();
               message = formatter.format(logRecord);
          } 
          catch (Exception e)
          {
               reportError(null,e,ErrorManager.FORMAT_FAILURE);
               return;
          }

          try
          {
               writer.write(message);
          }
          catch (Exception e)
          {
               reportError(null,e,ErrorManager.WRITE_FAILURE);
               return;
          }

          // flush the writer in case it is not set up for auto-flush
          doFlush();
     }

     public void setLevel(Level newLevel)
     {
          checkAccess();

          if (newLevel == null)
          {
               throw new IllegalArgumentException("\"newLevel\" cannot be null.");
          }

          this.level = newLevel;
     }

     public Level getLevel()
     {
          return level;
     }

     public void setFilter(Filter newFilter)
     {
          checkAccess();

          if (newFilter == null)
          {
               throw new IllegalArgumentException("\"newFilter\" cannot be null.");
          }

          this.filter = newFilter;
     }

     public Filter getFilter()
     {
          return filter;
     }

     public void setFormatter(Formatter newFormatter)
     {
          checkAccess();

          if (newFormatter == null)
          {
               throw new IllegalArgumentException("\"newFormatter\" cannot be null.");
          }

          this.formatter = newFormatter;
     }

     public Formatter getFormatter()
     {
          return formatter;
     }

     /**
      * This method does nothing. The encoding, if needed,
      * should be encapsulated in the writer.
      * @param encoding ignored encoding name
      * @exception SecurityException never thrown.
      * @exception UnsupportedEncodingException never thrown.
      */
     public void setEncoding(String encoding) throws SecurityException,UnsupportedEncodingException
     {
     }

     /**
      * This method always returns <CODE>null</CODE>. The encoding, 
      * if needed, should be encapsulated in the writer.
      * @return <CODE>null</CODE>
      */
     public String getEncoding()
     {
          return null;
     }

     private Writer writer /* = null */;
     private Level level /* = null */;
     private Filter filter /* = null */;
     private Formatter formatter /* = null */;
     private boolean closed /* = false */;

     /**
      * When <CODE>false</CODE>, setters can be called without
      * triggering a check for <CODE>LoggingPermission("control"))</CODE>.
      */
     protected boolean sealed = true;

     private void configure()
     {
          LogManager manager = LogManager.getLogManager();
          String cname = WriterHandler.class.getName();
          String levelName = manager.getProperty(cname + ".level");

          if (levelName != null && levelName.length() > 0)
          {
               Level newLevel = Level.parse(levelName);
               setLevel(newLevel);
          }
          else
          {
              setLevel(Level.ALL);
          }

          String filterName = manager.getProperty(cname + ".filter");

          if (filterName != null && filterName.length() > 0)
          {
               Filter newFilter = (Filter)objectForName(filterName);
               setFilter(newFilter);
          }

          String formatterName = manager.getProperty(cname + ".formatter");

          if (formatterName != null && formatterName.length() > 0)
          {
               Formatter newFormatter = (Formatter)objectForName(formatterName);
               setFormatter(newFormatter);
          }
          else
          {
               setFormatter(new SimpleFormatter());
          }
     }

     private Object objectForName(String name)
     {
          try
          {
               Class clazz = Class.forName(name);
               return clazz.newInstance();
          }
          catch (Exception e)
          {
               IllegalArgumentException iae = new IllegalArgumentException("An exception was thrown while instantiating class " + name + ".");
               iae.initCause(e);
               throw iae;
          }
     }

     private void writeHead()
     {
          Formatter formatter = getFormatter();

          try
          {
               String head = formatter.getHead(this);

               if (head != null && head.length() > 0)
               {
                    writer.write(head);
               }
          }
          catch (Exception e)
          {
               reportError("An exception was thrown while writing the head.",e,ErrorManager.WRITE_FAILURE);
          }
     }

     private void checkState()
     {
          if (closed)
          {
               throw new IllegalStateException();
          }
     }

     private void checkAccess()
     {
          if (sealed)
          {
               LogManager.getLogManager().checkAccess();
          }
     }

     private void doFlush()
     {
          try
          {
               writer.flush();
          }
          catch (Exception e)
          {
               reportError(null,e,ErrorManager.FLUSH_FAILURE);
          }
     }
}