
package com.sas.print;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Iterator;

import java.util.List;
import java.util.ArrayList;

import com.sas.util.Util;

public class PrinterUtil
{
	public static final String RB_KEY = "PrinterUtil.";

	public static final String PAPER_SIZE_LETTER = "Letter";
	public static final String PAPER_SIZE_A4 = "A4";

	/**
	 * Constructor to prevent construction of utility class.
	 */
	private PrinterUtil()
	{
		// Intentionally left blank
	}

	/**
	 * Keys used to lookup the paper sizes
	 */
	public static final String[] PAPER_SIZE_KEYS = { PAPER_SIZE_LETTER, PAPER_SIZE_A4 };

	private static Map paperSizeToCountriesMap = null;

	/**
	 * This method returns a page size string (PAPER_SIZE_LETTER, PAPER_SIZE_A4) based on the locale
	 * passed in.  Only the country code from the locale is used to perform the lookup of the
	 * appropriate paper size.
	 * @param locale The locale to determine the default paper size for
	 * @param defaultPaperSize The paper size to use if unable to determine the default paper size for
	 * the country in the locale passed in
	 * @return The paper size key (PAPER_SIZE_LETTER, PAPER_SIZE_A4) or the defaultPaperSize
	 * if the country specified in the locale is not known.
	 */
	public static final synchronized String getDefaultPaperSize( Locale locale, String defaultPaperSize )
	{
//FilePrinter.println( "DEBUG: start " + locale );

		// Use the default locale if one is not passed in
		if ( locale == null )
		{
			locale = Locale.getDefault();
//FilePrinter.println( "  DEBUG: using default locale " + locale );
		}

		// Grab the country code, even if it is null (we'll handle that later)
		String countryCode = locale.getCountry();

		// Create the map if necesssary
		if ( paperSizeToCountriesMap == null )
		{
			paperSizeToCountriesMap = new HashMap();

			// Load all the resources so we can search for the country
			for ( int i=0; i < PAPER_SIZE_KEYS.length; i++ )
			{
				String countries = RB.getStringResource( RB_KEY, PAPER_SIZE_KEYS[i] + ".notrans" );
				paperSizeToCountriesMap.put( PAPER_SIZE_KEYS[i], countries );
			}
		}
//FilePrinter.println( "  DEBUG: checking Country code" );

		// Check for a null country code. If so, try to figure out the country or paper size from the
		// language.
		if ( countryCode == null  || countryCode.length() == 0 )
		{
//FilePrinter.println( "  DEBUG: Country code is null" );
			List possibleCountries = getCountriesForLanguage( locale.getLanguage() );
			if ( possibleCountries.size() == 0 )
			{
//FilePrinter.println( "  DEBUG: no countries for language : " + locale.getLanguage() );
				countryCode = locale.getCountry();
			}
			else if ( possibleCountries.size() == 1 )
			{
//FilePrinter.println( "  DEBUG: only one country for language : " + locale.getLanguage() );
				countryCode = (String) possibleCountries.get( 0 );
			}
			else
			{
//FilePrinter.println( "  DEBUG: Testing all countries with language : " + locale.getLanguage() );
				// See if all the countries have the same paper size
				String firstPaperSize = getPaperSize( (String)possibleCountries.get( 0 ), null );
				for ( int i = 1, n=possibleCountries.size(); i < n; i++ )
				{
					String paperSize = (String) possibleCountries.get( i );
					if ( !Util.equal( firstPaperSize, paperSize) )
					{
//FilePrinter.println( "  DEBUG: paper sizes differ for countries.  returning default : " + defaultPaperSize );
						// Return the default if they are not all the same
						return defaultPaperSize;
					}
				}

//FilePrinter.println( "  DEBUG: paper sizes same for countries.  returning : " + firstPaperSize );
				return firstPaperSize;
			}
		}

//FilePrinter.println( "  DEBUG: returning getPaperSize : " + countryCode + ", " + defaultPaperSize);
		return getPaperSize( countryCode, defaultPaperSize );
	}

	private static String getPaperSize( String countryCode, String defaultPaperSize )
	{
		// Search the paper sizes and return the one that contains the country code
		Iterator e = paperSizeToCountriesMap.keySet().iterator();
		while ( e.hasNext() )
		{
			String paperSizeKey = (String) e.next();
			String countries = (String) paperSizeToCountriesMap.get( paperSizeKey );
			if ( countries.indexOf( countryCode ) != -1 )
			{
//FilePrinter.println( "  DEBUG: returning paperSizeKey: " +  paperSizeKey );
				return paperSizeKey;
			}
		}

//FilePrinter.println( "  DEBUG: returning defaultPaperSize: " +  defaultPaperSize);
		return defaultPaperSize;
	}

	/**
	 * 
	 * @param paperSize Either PAPER_SIZE_LETTER or PAPER_SIZE_A4
	 * @param locale The locale that the localized label for the paper size should be returned in.
	 * @return A text label for the specified paper size
	 */
	public static String getPaperSizeLabel( String paperSize, Locale locale )
	{
		return RB.getStringResource( RB_KEY, paperSize + ".Label.txt", locale );
	}


	private static List getCountriesForLanguage( String language )
	{
//FilePrinter.println( "  DEBUG: getCountriesForLanguage " + language );
		List countries = new ArrayList();
		Locale[] allLocales = Locale.getAvailableLocales();
//FilePrinter.println( "    DEBUG: n=" + allLocales.length );
		for ( int i = 0; i < allLocales.length; i++ )
		{
//FilePrinter.println( "    DEBUG: Checking " + allLocales[i] );
			if ( allLocales[i].getCountry() != null  &&
				 allLocales[i].getCountry().length() != 0  &&
				 language.equals( allLocales[i].getLanguage() ) )
			{
//FilePrinter.println( "      DEBUG: Adding " + allLocales[i].getCountry() );
				countries.add( allLocales[i].getCountry() );
			}
		}

		return countries;
	}

//	/**
//	 * Attempted to determine a country for a given locale.
//	 * @param locale
//	 * @return
//	 */
//	private static String findCountryForLocale( Locale locale )
//	{
//		if ( locale.getCountry() != null )
//		{
//			return locale.getCountry();
//		}
//
//		// Try to see if only one country supports the language passed in
//		String language = locale.getLanguage();
//		String country = null;
//		Locale[] allLocales = Locale.getAvailableLocales();
//		for ( int i = 0; i < allLocales.length; i++ )
//		{
//			if ( language.equals( allLocales[i].getLanguage() ) )
//			{
//				if ( country == null )
//				{
//					// remember the country if it is not null
//					if ( allLocales[i].getCountry() != null )
//					{
//						country = allLocales[i].getCountry();
//					}
//				}
//				else
//				{
//					// return null to indicate unknown country
//					return null;
//				}
//			}
//		}
//		return country;
//	}
}

//class FilePrinter
//{
//	static java.io.PrintWriter pw = null;
//	static
//	{
//		try
//		{
//			pw = new java.io.PrintWriter( new java.io.FileOutputStream( new java.io.File("C:/tmp/tmp.txt" ), true ) );
//		}
//		catch (Exception e)
//		{
//			e.printStackTrace();
//		}
//	}
//
//	public static void println( String s )
//	{
//		pw.println( s );
//		pw.flush();
//	}
//}

/*
PrinterUtil.InvalidCountryCode.txt=Locale passed in must contain two letter country code

# These are the labels that would appear in a user interface for the two default paper sizes
PrinterUtil.Letter.Label.txt=Letter
PrinterUtil.A4.Label.txt=A4

# These are the non-translated maps of paper size to country codes
PrinterUtil.Letter.notrans=US,AR,BO,BR,CA,CB,CL,CO,CR,DO,EC,SV,GT,HN,JM,MX,NI,PA,PY,PE,PH,PR,VE
PrinterUtil.A4.notrans=AL,DZ,AU,AT,BH,BY,BE,BW,BG,CN,HR,CZ,DK,EG,EE,FO,FI,FR,DE,GR,GL,HK,HU,IS,IN,ID,IR,IQ,IE,IL,IT,JP,JO,KW,LV,LB,LY,LI,LT,LU,MO,MK,MY,MT,MA,NL,NZ,NO,OM,PL,PT,QA,RO,RU,SA,SG,SK,SI,ZA,KR,ES,SD,SE,CH,SY,TW,TH,TN,TR,UA,AE,GB,UY,VN,YE,YU,ZW
*/
