/* $Id$ */
/* Copyright (c)1998 SAS Institute Inc, Cary NC USA 27513
 * Unpublished- All Rights Reserved.
 */
package com.sas.rio;

import com.sas.iom.SAS.BaseType;
import java.io.Serializable;

//
//History Records:
//
// Date          Name                Remarks
// 07/07/99 M Srinivasan   - Updated javadoc. Fixed setColumnType.
// 08/03/99 M Srinivasan   - S0061395 : RIOColumnType -> BaseType change (IOM Hierarchy Change).
// 03/27/00 M Srinivasan   - S0080843 : Convert null Strings to "" since Proxies cannot handle null
//                           elements yet.
// 27Apr2004 damack        - S0222220 : initialize the string fields to be "", so that iom doesn't blow up.
//

/**
 * <P>This is the Column class which defines a columns' attributes.
 * A Column object array is passed as an argument while creating a Table.
 * The Column object array is also used while retrieving selected columns
 * from an existing Table.
 */
public class Column implements Serializable{
    private String columnName;
    private String columnLabel;
    private int columnType;
    private String columnTypeName;
    private int columnLength;
    private String columnFormatName;
    private int columnFormatLength;
    private int columnFormatWidth;
    private int columnFormatDecimal;
    private String columnInformatName;
    private int columnInformatLength;
    private int columnInformatWidth;
    private int columnInformatDecimal;
    private int vendorCode;
    private String messageIndex;
    String m_errorMsg;
    /**
     * Constructs a Column object.
     */
    public Column() 
    {
    	columnName = "";
    	columnLabel = "";
    	columnTypeName="";
    	columnFormatName="";
    	columnInformatName="";
    	messageIndex="";
    	m_errorMsg = "";
    }
    /**
     * Gets the name of the column.
     *
     * @return The name of the column.
     * @exception RIOException This exception is never thrown.
     */
    public String getColumnName() throws RIOException {
        return columnName;
    }
    /**
     * Gets the label of the column.
     *
     * @return The label of the column.
     * @exception RIOException This exception is never thrown.
     */
    public String getColumnLabel() throws RIOException {
        return columnLabel;
    }
    /**
     * Gets the SQL type of the column.
     *
     * @return The SQL type of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getColumnType() throws RIOException {
        return columnType;
    }
    /**
     * Gets the SQL type name of the column.
     *
     * @return The SQL type name of the column.
     * @exception RIOException This exception is never thrown.
     */
    public String getColumnTypeName() throws RIOException {
        return columnTypeName;
    }
    /**
     * Gets the length of the column.
     *
     * @return The length of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getMaxLength() throws RIOException {
        return columnLength;
    }
    /**
     * Gets the format name of the column.
     *
     * @return The format name of the column.
     * @exception RIOException This exception is never thrown.
     */
    public String getColumnFormatName() throws RIOException {
        return columnFormatName;
    }
    /**
     * Gets the informat name of the column.
     *
     * @return The informat name of the column.
     * @exception RIOException This exception is never thrown.
     */
    public String getColumnInformatName() throws RIOException {
        return columnInformatName;
    }
    /**
     * Gets the format length of the column.
     *
     * @return The format length of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getColumnFormatLength() throws RIOException {
        return columnFormatLength;
    }
    /**
     * Gets the informat length of the column.
     *
     * @return The informat length of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getColumnInformatLength() throws RIOException {
        return columnInformatLength;
    }
    /**
     * Gets the format width of the column.
     *
     * @return The format width of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getColumnFormatWidth() throws RIOException {
        return columnFormatWidth;
    }
    /**
     * Gets the informat width of the column.
     *
     * @return The informat width of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getColumnInformatWidth() throws RIOException {
        return columnInformatWidth;
    }
    /**
     * Gets the format decimal of the column.
     *
     * @return The format decimal of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getColumnFormatDecimal() throws RIOException {
        return columnFormatDecimal;
    }
    /**
     * Gets the informat decimal of the column.
     *
     * @return The informat decimal of the column.
     * @exception RIOException This exception is never thrown.
     */
    public int getColumnInformatDecimal() throws RIOException {
        return columnInformatDecimal;
    }
    /**
     * Sets the name of the column.
     *
     * @param columnName The name of the column. Column name cannot be NULL.
     * @exception RIOException This exception is thrown if the name of the column is NULL.
     */
    public void setColumnName(String columnName) throws RIOException {
        if(columnName == null) {
	    // VendorCode??
	    MVALocalizedError.throwLocalizedRIOException(VendorCode.Column_setColumnNameError );

        }
	this.columnName = columnName;
    }
    /**
     * Sets the label of the column.
     *
     * @param columnLabel The label of the column. Column label may be NULL.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnLabel(String columnLabel) throws RIOException {
        if(columnLabel == null) {
            this.columnLabel = "";
		}
        else {
        	this.columnLabel = columnLabel;
        }
	}
    /**
     * Sets the SQL type of the column.
     *
     * @param coluumnType The SQL type of the column. Should be one of the following values.
     * <p>
     * <dl>
     * <dt><code>com.sas.iom.sas.BaseType._BaseTypeCharacter</code>
     * <dd>Indicates that the column is to contain character data.
     * <p>
     * <dt><code>com.sas.iom.sas.BaseType._BaseTypeNumeric</code>
     * <dd>Indicates that the column is to contain eight byte numeric data.
	 * </dl>
     * @exception RIOException This exception is thrown if column type is not valid.
     */
    public void setColumnType(int columnType) throws RIOException {
        if((columnType != BaseType._BaseTypeNumeric) &&
            (columnType != BaseType._BaseTypeCharacter)) {
	    // VendorCode??
	    MVALocalizedError.throwLocalizedRIOException(VendorCode.General_invalidColumnType );

        }
	this.columnType = columnType;
    }
    /**
     * Sets the maximum length of the column. Corresponds to the size of the column
     *
     * @param maxLength The maximum length of the column. Corresponds to the size of the column.
     * @exception RIOException This exception is thrown if maximum length is less than 1.
     */
    public void setMaxLength(int maxLength) throws RIOException {
        if(maxLength < 1) {
	    // VendorCode??
	    MVALocalizedError.throwLocalizedRIOException(VendorCode.Column_setMaxLengthError );

	}
        this.columnLength = maxLength;
    }
    /**
     * Sets the format name of the column.
     *
     * @param formatName The format name of the column. Format name may be NULL.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnFormatName(String formatName) throws RIOException {
        if(formatName == null) {
            this.columnFormatName = "";
        }
        else {
        	this.columnFormatName = formatName;
        }
    }
    /**
     * Sets the informat name of the column.
     *
     * @param informatName The informat name of the column. Informat name may be NULL.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnInformatName(String informatName) throws RIOException {
        if(informatName == null) {
            this.columnInformatName = "";
        }
        else {
        	this.columnInformatName = informatName;
        }
    }

    /**
     * Sets the format length of the column.
     *
     * @param formatLength The format length of the column.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnFormatLength(int formatLength) throws RIOException {
        this.columnFormatLength = formatLength;
    }
    /**
     * Sets the informat length of the column.
     *
     * @param informatLength The informat length of the column.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnInformatLength(int informatLength) throws RIOException {
        this.columnInformatLength = informatLength;
    }
    /**
     * Sets the format width of the column.
     *
     * @param formatWidth The format width of the column. Specify
	 * com.sas.iom.sas.RIONoFmtWidth.value to indicate no format width needs to
	 * be persisted for the column.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnFormatWidth(int formatWidth) throws RIOException {
        this.columnFormatWidth = formatWidth;
    }
    /**
     * Sets the informat width of the column.
     *
     * @param informatWidth The informat width of the column.Specify
	 * com.sas.iom.sas.RIONoFmtWidth.value to indicate no informat width needs to
	 * be persisted for the column.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnInformatWidth(int informatWidth) throws RIOException {
        this.columnInformatWidth = informatWidth;
    }
    /**
     * Sets the format decimal of the column.
     *
     * @param formatDecimal The format decimal of the column. Specify
	 * com.sas.iom.sas.RIONoFmtDecimal.value to indicate no format decimal needs
	 * to be persisted for the column.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnFormatDecimal(int formatDecimal) throws RIOException {
        this.columnFormatDecimal = formatDecimal;
    }
    /**
     * Sets the informat decimal of the column.
     *
     * @param informatDecimal The informat decimal of the column. Specify
	 * com.sas.iom.sas.RIONoFmtDecimal.value to indicate no informat decimal needs
	 * to be persisted for the column.
     * @exception RIOException This exception is never thrown.
     */
    public void setColumnInformatDecimal(int informatDecimal) throws RIOException {
        this.columnInformatDecimal = informatDecimal;
    }

}



