/* $Id$ */
/* Copyright (c)1998 SAS Institute Inc, Cary NC USA 27513
 * Unpublished- All Rights Reserved.
 */
package com.sas.rio;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Savepoint;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;

import com.sas.iom.JDBCConstants;
import com.sas.iom.LDAPConstants;
import com.sas.iom.WorkspaceConstants;
import com.sas.iom.SAS.IDataService;
import com.sas.iom.SAS.IWorkspace;
//
//History Records:
//
// Date          Name                Remarks
// 05/04/99 M Srinivasan   - Fix to S0052966. Supported multiple librefs in
//                           Connection properties. Used IDataService.AssignLibref()
//                           instead of IDBMSConnection.ExecuteStatement().
// 06/03/99 M Srinivasan   - Added constructor to instantiate MVAConnection object
//                           using IDataService and Properties object.
// 06/07/99 M Srinivasan   - RIO related changes. Status and Warning removed from
//                           return parameters in ILibref.AssignLibref call.
// 07/13/99 M Srinivasan   - Added documentation for encryptionPolicy,content and algorithms.
// 08/13/99 M Srinivasan   - S0063145 : Do not call DeassignLibref to release librefs. It
//                           causes "library in use" problem for empty libraries. Let closing
//               workspace handle deassigning and releasing librefs.
// 12/02/99 M Srinivasan   - S0071131 :Enhancement to support LDAP access to get DatabaseMetaData
//                           information. Connection to LDAP is made by DatabaseMetaData object
//                           based on ldapDbmd property.
// 12/14/99 M Srinivasan   -S0072011 :Use Constants and avoid hardcoding property names.
// 12/21/99 M Srinivasan   -S0072466 :Fix to IDataService.ILibref call. Saved warnings in java.sql.SQLWarning.
// 03/27/00 M Srinivasan   - Make DebugLevel optional.
// 12/10/2001 D Mackie     - S0133836: fixed setting of ResultSet type and concurrency in createStatement,
//                           prepareStatement, and prepareCall to be in accordance with JDBC2.0 spec.
// 04/15/2003 damack       - S0195798: change XHRSQPRECS to 100 from 10.
//
// 26Jun2003 damack        - S0207780: correcting Javadoc to remove mention of the ldap classes.
// 21Jul2003 domelz        - S0205937: correct javadoc to eliminate breakiterator warning
// 26Apr2004 damack        - S0205649: removing ldap references
// 26Apr2004 damack        - S0189758: javadoc for non-implemented methods, cleanup.
// 14Jan2005 damack        - S0278967: added a properties object to call to DatabaseMetaData ctor, so that
//                                     coercetypes property can be properly propagated.
// 02Aug2005 damack        - S0308525: added method "statementClosed", which the statement method calls to signal
//                                     that it has closed. (So, statement is no longer calling "deleteStatement"
//                                     directly.
/**
 * <P>A Connection represents a session with a specific
 * database. Within the context of a Connection, SQL statements are
 * executed and results are returned.
 *
 * <P>A Connection's database is able to provide information
 * describing its tables, its supported SQL grammar, its stored
 * procedures, the capabilities of this connection, etc. This
 * information is obtained with the getMetaData method.
 *
 * <P><B>Note:</B> By default the Connection automatically commits
 * changes after executing each statement. If auto commit has been
 * disabled, an explicit commit must be done or database changes will
 * not be saved.
 *
 * @see java.sql.DriverManager#getConnection
 * @see Statement
 * @see ResultSet
 * @see DatabaseMetaData
 */
public class MVAConnection implements java.sql.Connection
{
	/*
	 *  S0195798: change default fetchsize to 100 instead of 10
	 */
	public static final int XHRSQPRECS = 100;
	private boolean connected = false;
	private boolean readOnly = false;
	private String catalog = null;
	private Vector statementList = new Vector();
	private Vector librefList = new Vector();
	private String[] libnames = new String[10];
	private String url = null;
	private SQLWarning warnings = null;
	private String userName = null;
	private String user = null;
	private String librefs = null;
	private boolean remarks = false;
	private Boolean undoPolicyNone = new Boolean(true);
	private String engine = null;
	private String options = null;
	private int vendorCode;
	private String messageIndex;
	private boolean fromSSI = false;
	private boolean fromDataService = false;
	protected IWorkspace sasWorkspace = null;
	protected IDataService sasDataService = null;
	protected int resultSetType = ResultSet.TYPE_FORWARD_ONLY;
	protected int resultSetConcurrency = ResultSet.CONCUR_READ_ONLY;
	protected String version_msg;
	private boolean ldapDbmd = false;
	private String baseDN = null;
	private String authentication = null;
	private String ldapUserName = null;
	private String ldapPassword = null;
	private String ldapProviderUrl = null;
	private Properties info = null;
	private XUtil xUtil = null;
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	//
	//        EXCEPTION METHODS
	//
	//  The following methods may potentially replace the existing error handling. Currently they
	//  are experimental and will not be plugged into the code until a final decision is made for
	//  exception handling in Java.
	//
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	//
	//      MAIN METHODS
	//
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * Constructs an MVAConnection object.
	 * @param url  initial input url containing host and port information.
	 * @param info  java.util.Properties object which contains:
	 * <p>
	 * <dl>
	 * <DT> userName<DD>MVA server access user name
	 * <p>
	 * <DT> password<DD>MVA server access password
	 * <p>
	 * <DT> librefs<DD>a semi-colon delimited list of 'name filepath'.
	 * <p>
	 * <DT>encryptionPolicy<DD>whether or not JavaBridge should attempt to negotiate with
	 * the server over which encryption algorithm to use and what to do if the negotiations fail.
	 * Possible values:
	 * <DL>
	 * <DT>none<DD>do not use encryption. This is the default.
	 * <DT>optional<DD>attempt to use encryption but, if algorithm negotiation fails, continue with an
	 * unencrypted session
	 * <DT>required<DD>attempt to use encryption but, if algorithm negotiation fails, fail the connection
	 * </DL>
	 * <p>
	 * <DT>encryptionContent<DD>specifies which messages should be encrypted if encryption is used. Possible values:
	 * <DL>
	 * <DT>all<DD>encrypt all messages. This is the default.
	 * <DT>authentication<DD>encrypt only messages that contain user name and password information.
	 * </DL>
	 * <p>
	 * <DT>encryptionAlgorithms<DD>the list of algorithms you are willing to use in order of
	 * preference. Values in the list should be separated by commas and chosen from sasproprietary,
	 * rc2, rc4, des, or tripledes. If no value is specified, then one of the server's favorite algorithms
	 * will be used.
	 * <p>
     * <DT><code>applyFormats</code><DD>specifies whether or not column formats are applied
     * when retrieving data. If the value is true, all data is returned as Strings. For any column, the
     * format used is the one defined on the column in the dataset. If "applyFormats" is set to false
     * (the default case) no formatting is applied. 
	 * </dl>
	 * <P>
	 * Note that the server's encryption policy is, by default, optional so no special action is needed
	 * to prepare an IOM BRIDGE server to use encryption. However, it is possible to set up the server
	 * so that encryption is required. Consult IOM BRIDGE server documentation for details.
	 * <p>
	 * @exception SQLException This exception is thrown if a ShareNetException is detected.
	 */
	public MVAConnection(String url, java.util.Properties info)
		throws SQLException
	{
		this.url = url;
		info = updateDriverProperties(info);
		try
		{
			setRemarks(info);
			//setLdapDbmd(info);
			setUser(info);
			info.put(WorkspaceConstants.HOST, getHost(url));
			info.put(WorkspaceConstants.PORT, getPort(url));
			info.put(WorkspaceConstants.USERNAME, getUser());
			String rmiURL = (String) info.get("rmiurl");
			this.info = info;
			if (rmiURL == null || rmiURL.length() == 0)
				xUtil = new RIOUtil(url, info);
			else
			{
				/*
				   
				   //???   
				   // In "real life", what we'll do is put the AppServer in a separate jar,
				   // (or use an existing EM/whatever appserver, and use reflection(whatever)
				   // to create the Xutil...
				   Remote aS = java.rmi.Naming.lookup(rmiURL);
				   Class foo = aS.getClass();
				   
				   // a pause for a moment of reflection:
				   Class[] interfaces = foo.getInterfaces();
				   boolean validAppServer = false;
				   for (int i = 0; i < interfaces.length; i++)
				   {
				   String iname = interfaces[i].getName();
				   if (iname.equals("com.sas.rio.AppServer"))
				   validAppServer = true;   
				   }
				   if (!validAppServer)
				   {
				   MVALocalizedError.throwLocalizedException(
				   VendorCode.MVAConnection_error);
				   
				   }
				   else
				   {
				   
				   Method getXUtil = null;
				   Class peanut = com.sas.rio.rmi.AppServerImpl.class;
				   Class[] parms = 
				   new Class[]{String.class, java.util.Properties.class};
				   Object[] arguments = new Object[]{url, info};
				           
				   getXUtil = foo.getMethod("getXUtil", parms);
				           
				   xUtil = (XUtil)getXUtil.invoke(aS, arguments);
				           
				*/
				com.sas.rio.AppServer aS =
					(com.sas.rio.AppServer) java.rmi.Naming.lookup(rmiURL);
				xUtil = aS.getXUtil(url, info);
			}
			connected = xUtil.isConnected();
			if (connected)
				executeUndoPolicyNoneStatement(info);
		}
		catch (Exception e)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_error,
				null,
				e);
		}
	}
	/**
	 * Constructs an MVAConnection object.
	 * @param sasDataService  IDataService object.
	 * @param info  java.util.Properties object which contains:
	 * <dl>
	 * <dt> userName<dd>MVA server access user name
	 * <dt> password<dd>MVA server access password
	 * <dt> librefs<dd>a semi-colon delimited list of 'name filepath'.
	 * </dl>
	 * @exception SQLException This exception is thrown if a ShareNetException is detected.
	 */
	public MVAConnection(
		IDataService sasDataService,
		java.util.Properties info)
		throws SQLException
	{
		this.sasDataService = sasDataService;
		info = updateDriverProperties(info);
		this.info = info;
		String rmiURL = (String) info.get("rmiurl");
		fromDataService = true;
		fromSSI = true;
		try
		{
			setRemarks(info);
			
			this.user = info.getProperty(WorkspaceConstants.USERNAME);
			if (rmiURL == null || rmiURL.length() == 0)
				xUtil = new RIOUtil(sasDataService, info);
			else
			{
				MVALocalizedError.throwLocalizedException(
					VendorCode.MVAConnection_error);
			}
			connected = xUtil.isConnected();
			if (connected)
				executeUndoPolicyNoneStatement(info);
		}
		catch (Exception e)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_error,
				null,
				e);
		}
	}
	/**
	  * Constructs an MVAConnection object.
	  * @param sasWorkspace  IWorkspace object.
	  * @param info  java.util.Properties object which contains:
	  * <dl>
	  * <dt> userName<dd>MVA server access user name
	  * <dt> password<dd>MVA server access password
	  * <dt> librefs<dd>a semi-colon delimited list of 'name filepath'.
	  * </dl>
	  * @exception SQLException This exception is thrown if a ShareNetException is detected.
	  */
	public MVAConnection(IWorkspace sasWorkspace, java.util.Properties info)
		throws SQLException
	{
		if (sasWorkspace == null)
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_error);
		info = updateDriverProperties(info);
		this.info = info;
		String rmiURL = (String) info.get("rmiurl");
		fromDataService = true;
		fromSSI = true;
		try
		{
			setRemarks(info);
			
			this.user = info.getProperty(WorkspaceConstants.USERNAME);
			if (rmiURL == null || rmiURL.length() == 0)
				xUtil = new RIOUtil(sasWorkspace, info);
			else
			{
				MVALocalizedError.throwLocalizedException(
					VendorCode.MVAConnection_error);
			}
			connected = xUtil.isConnected();
			if (connected)
				executeUndoPolicyNoneStatement(info);
		}
		catch (Exception e)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_error,
				null,
				e);
		}
	}
	public MVAConnection(XUtil x, java.util.Properties p) throws SQLException
	{
		info = updateDriverProperties(p);
		fromDataService = true;
		fromSSI = true;
		try
		{
			setRemarks(info);
			
			this.user = info.getProperty(WorkspaceConstants.USERNAME);
			xUtil = x;
			connected = xUtil.isConnected();
			if (connected)
				executeUndoPolicyNoneStatement(info);
		}
		catch (Exception e)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_error,
				null,
				e);
		}
	}
	java.util.Properties updateDriverProperties(java.util.Properties info)
	{
		String key;
		Object value;
		if (info == null)
			return null;
		/*
		 * gotta lowercase all the property keys...
		 */
		java.util.Properties duplo = new java.util.Properties();
		for (Enumeration a = info.keys(); a.hasMoreElements();)
		{
			key = (String) (a.nextElement());
			value = info.get(key);
			duplo.put(key.toLowerCase(), value);
		}
		return duplo;
	}
	/**
	 * Creates a Statement object.
	 *
	 * SQL statements without parameters are normally
	 * executed using Statement objects. If the same SQL statement
	 * is executed many times, it is more efficient to use a
	 * PreparedStatement
	 *
	 * JDBC 2.0
	 *
	 * Result sets created using the returned Statement will have
	 * forward-only type, and read-only concurrency, by default.
	 *
	 * @return a new Statement object
	 * @exception SQLException This exception is thrown if an MVAStatement object cannot be created.
	 */
	public synchronized Statement createStatement() throws SQLException
	{
		Statement statement;
		if (!isConnected())
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_createStatementError);
		}
		statement = new MVAStatement(this, info);
		statementList.addElement(statement);
		return (statement);
	}
	/**
	 * Creates a PreparedStatement object.
	 *
	 * A SQL statement with or without IN parameters can be
	 * pre-compiled and stored in a PreparedStatement object. This
	 * object can then be used to efficiently execute this statement
	 * multiple times.
	 *
	 *
	 * JDBC 2.0
	 *
	 * Result sets created using the returned PreparedStatement will have
	 * forward-only type, and read-only concurrency, by default.
	 *
	 * @param sql A SQL statement that may contain one or more '?' IN
	 * parameter placeholders.
	 * @return A new PreparedStatement object containing the
	 * pre-compiled statement.
	 * @exception SQLException This exception is thrown if a PreparedStatement object cannot be created.
	 */
	public synchronized PreparedStatement prepareStatement(String sql)
		throws SQLException
	{
		MVAPreparedStatement prepstatement;
		if (!isConnected())
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_prepareStatementError);
		}
		prepstatement = new MVAPreparedStatement(this, sql, info);
		statementList.addElement(prepstatement);
		return (prepstatement);
	}
	/**
	 * Creates a CallableStatement object.
	 *
	 * JDBC 2.0
	 *
	 * Result sets created using the returned CallableStatement will have
	 * forward-only type, and read-only concurrency, by default.
	 *
	 * @param sql A SQL statement that may contain one or more '?'
	 * parameter placeholders. Typically this  statement is a JDBC
	 * function call escape string.
	 * @return A new CallableStatement object containing the
	 * pre-compiled SQL statement
	 * @exception SQLException This exception is required by the interface but it is never thrown.
	 */
	public CallableStatement prepareCall(String sql) throws SQLException
	{
		return new MVACallableStatement();
	}
	/**
	 * Returns the input string. The SQL is not optimized.
	 * @param sql An input SQL string for optimization.
	 * @return An input string.
	 * @exception SQLException This exception is required by the interface, but it is never thrown.
	 */
	public String nativeSQL(String sql) throws SQLException
	{
		return sql;
	}
	/**
	 * Sets the auto-commit state.
	 * @param autoCommit This value must be True. SAS software does not support transactions.
	 * @exception SQLException This exception is thrown if the value for autoCommit is False.
	 */
	public void setAutoCommit(boolean autoCommit) throws SQLException
	{
		if (autoCommit == false)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_transactionError);
		}
	}
	/**
	 * SAS software does not support transactions. All statements are committed
	 * individually. The current auto-commit state is always True.
	 * @return True.
	 * @exception SQLException This exception is required by the interface, but it is never thrown.
	 */
	public boolean getAutoCommit() throws SQLException
	{
		return true;
	}
	/**
	 * Closes all the open Statement and PreparedStatement objects that are associated with this object.
	 * @exception SQLException This exception is thrown if an exception occurred while outstanding
	 * statements were being closed.
	 */
	public synchronized void commit() throws SQLException
	{
		MVAStatement statement;
		int count = statementList.size();
		for (int i = 0; i < count; i++)
		{
			statement = (MVAStatement) statementList.elementAt(0);
			//TODO:
			//if (statement instanceof MVAPreparedStatement) {
			statement.close();
			
			//}
		}
	}
	/**
	 * SAS software does not support transactions. This method is not supported.
	 * @exception SQLException This exception is always thrown.
	 */
	public void rollback() throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.MVAConnection_transactionError);
	}
	/**
	 * In some cases, it is desirable to immediately release a
	 * Connection's database and JDBC resources instead of waiting for
	 * them to be automatically released; the close method provides this
	 * immediate release.
	 *
	 * <P><B>Note:</B> A Connection is automatically closed when it is
	 * garbage collected. Certain fatal errors also result in a closed
	 * Connection.
	 *
	 * @exception SQLException This exception is thrown if an error condition is detected.
	 */
	public void close() throws SQLException
	{
		MVAStatement statement;
		int count = statementList.size();
		try
		{
			for (int i = 0; i < count; i++)
			{
				statement = (MVAStatement) statementList.elementAt(0);
				statement.close();
			}
			statementList = new Vector();
			if (isConnected())
			{
				xUtil.disconnect();
				/*
				                if (dbmsConnection != null)
				                {
				                    dbmsConnection.Disconnect();
				                    dbmsConnection._release();
				                    dbmsConnection = null;
				                }
				                if (sasDataService != null)
				                {
				   */
				/*
				for(int i=0;i<librefList.size();i++) {
				
				sasDataService.DeassignLibref(libnames[i]);
				System.out.println("DeAssign successful.");
				}
				 */
				/*     
				                    if (fromDataService == false)
				                    {
				                        sasDataService._release();
				                        sasDataService = null;
				                    }
				                }
				                ILibref sasLibref;
				                for (int i = 0; i < librefList.size(); i++)
				                {
				                    sasLibref = (ILibref) librefList.elementAt(0);
				                    if (sasLibref != null)
				                    {
				                        sasLibref._release();
				                        sasLibref = null;
				                    }
				                }
				
				                if (fromSSI == false)
				                {
				                    if (sasWorkspace != null)
				                    {
				                        sasWorkspace.Close();
				                        sasWorkspace._release();
				                        sasWorkspace = null;
				                    }
				                    if (sasAutomation != null)
				                    {
				                        sasAutomation._release();
				                        sasAutomation = null;
				                    }
				                }
				*/
				setConnected(false);
			}
		}
		catch (Exception e)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_closeError,
				null,
				e);
			/*SQLException se = new SQLException(error, null, VendorCode.VC_UNKNOWNERROR);
			if (e instanceof SQLException)
			  se.setNextException((SQLException)e);
			
			throw se;
			*/
		}
	}
	/**
	 * Tests to see if a Connection is closed.
	 *
	 * @return true if the connection is closed; false if it's still open
	 * @exception SQLException This exception is required by the interface but it is never thrown.
	 */
	public boolean isClosed() throws SQLException
	{
		return (!isConnected());
	}
	//======================================================================
	// Advanced features:
	/**
	 * A Connection's database is able to provide information
	 * describing its tables, its supported SQL grammar, its stored
	 * procedures, the capabilities of this connection, etc. This
	 * information is made available through a DatabaseMetaData
	 * object.
	 *
	 * @return a DatabaseMetaData object for this Connection
	 * @exception SQLException This exception is thrown if an DatabaseMetaData object cannot be created.
	 */
	public DatabaseMetaData getMetaData() throws SQLException
	{
		DatabaseMetaData data = null;
		if (!isConnected())
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_getMetadataError);
		}
		//TODO: should we check to make sure we can connect only to SQLVIEW databases.
		try
		{
			// get version, and depending on that, either call v8 or v9 constructor.
			String sasVersion = null;
			try
			{
				sasVersion = xUtil.getDBMSVersion();
				/*
				                    IWorkspace sasWorkspace = sasDataService.Parent();
				                    IUtilities sasUtilities = sasWorkspace.Utilities();
				                    com.sas.iom.SAS.IHostSystem hostSystem =
				                        sasUtilities.HostSystem();
				
				                    StringSeqHolder systemInfoHolder = new StringSeqHolder();
				                    VariableArray2dOfStringHolder machineInfoHolder =
				                        new VariableArray2dOfStringHolder();
				                    hostSystem.GetInfo(systemInfoHolder, machineInfoHolder);
				                    hostSystem._release();
				                    sasUtilities._release();
				                    sasWorkspace._release();
				
				                    String[] systemInfo = systemInfoHolder.value;
				                    sasVersion = systemInfo[0];
				*/
			}
			catch (Exception se)
			{
				throw new SQLException(se.getMessage());
			}
			if (sasVersion == null) // really really bad
			{
				// throw an exception
				MVALocalizedError.throwLocalizedException(
					VendorCode.MVAConnection_getMetadataError);
			}
			else
			{
				/*
				 * since the version is of the form "major.minor", we check the major, and
				 * then if necessary the minor.
				 */
			    /* S0278967: added a properties object to ctor for metadata.
			     * this allows me to pass down the coercenumerictypes property to the
			     * resultset for getColumns and getBestRowIdentifer, s.t. the proper types
			     * get returned.
			     */
				int major = Character.digit(sasVersion.charAt(0), 10);
				int minor = Character.digit(sasVersion.charAt(2), 10);
				if (major == 8)
				{
					if (minor == 2)
						data = new MVADatabaseMetaDatav82(this, info);
					else
						data = new MVADatabaseMetaData(this, info);
				}
				else if (major == 9)
					data = new MVADatabaseMetaDatav82(this, info);
				else // default version is MVADatabaseMetaData
					// no version info at all available for some reason.
					data = new MVADatabaseMetaData(this, info);
			}
		}
		catch (Exception e)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_getMetadataError,
				null,
				e);
		}
		return data;
	}
	/**
	 * SAS software does not support read-only connections. The input value
	 * is ignored.
	 * @param readOnly This parameter is ignored.
	 * @exception SQLException This exception is required by the interface, but it is never thrown.
	 */
	public void setReadOnly(boolean readOnly) throws SQLException
	{
		return;
	}
	/**
	 * SAS software does not support read-only connections. The readOnly
	 * value is always False.
	 * @return False.
	 * @exception SQLException This exception is required by the interface, but it is never thrown.
	 */
	public boolean isReadOnly() throws SQLException
	{
		return false;
	}
	/**
	 * A sub-space of this Connection's database may be selected by setting a
	 * catalog name. If the driver does not support catalogs it will
	 * silently ignore this request.
	 * NOTE: SAS software does not support catalogs and so the catalog string
	 * will be silently ignored.
	 *
	 * @exception SQLException This exception is required by the interface but it is never thrown.
	 */
	public void setCatalog(String catalog) throws SQLException
	{
		return;
	}
	/**
	 * Gets the Connection's current catalog name.
	 *
	 * @return NULL. SAS software does not support catalogs.
	 * @exception SQLException if a database-access error occurs.
	 */
	public String getCatalog() throws SQLException
	{
		return null;
	}
	/**
	 * SAS software does not support transactions. The only valid input value is
	 * java.sql.Connection.TRANSACTION_NONE.
	 * @param level The transaction level.
	 * @exception SQLException This exception is thrown if the value of level is not valid.
	 */
	public void setTransactionIsolation(int level) throws SQLException
	{
		if (level != Connection.TRANSACTION_NONE)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_noTransactionsError);
		}
	}
	/**
	 * SAS software does not support transactions.
	 * @return java.sql.Connection.TRANSACTION_NONE.
	 * @exception SQLException This exception is required by the interface, but it is never thrown.
	 */
	public int getTransactionIsolation() throws SQLException
	{
		return (Connection.TRANSACTION_NONE);
	}
	/**
	 * Gets the warning reported by calls on this Connection.
	 *
	 * <P><B>Note:</B> Subsequent warnings will be chained to this
	 * SQLWarning.
	 *
	 * @return The SQLWarning.
	 * @exception SQLException This exception is required by the interface but it is never thrown.
	 */
	public SQLWarning getWarnings() throws SQLException
	{
		return warnings;
	}
	/**
	 * Clears the warnings.
	 *
	 * @exception SQLException This exception is required by the interface but it is never thrown.
	 */
	public void clearWarnings() throws SQLException
	{
		warnings = null;
	}
	//--------------------------JDBC 2.0-----------------------------
	/**
	 * Same as createStatement() above, but allows the default result set
	 * type and result set concurrency type to be overridden. If an invalid type or
	 * concurrency is passed, the values are defaulted to TYPE_FORWARD_ONLY and
	 * CONCUR_READ_ONLY.
	 * If an invalid scrolling type or concurrency is requested, then a valid scrolling
	 * type (or concurrency) is substituted, and a warning is generated. 
	 *
	 * JDBC 2.0
	 *
	 * @param resultSetType a result set type, see ResultSet.TYPE_XXX
	 * @param resultSetConcurrency a concurrency type, see ResultSet.CONCUR_XXX
	 * @return a new Statement object
	 * @exception SQLException This exception is thrown if a Statement object cannot be created.
	 */
	public synchronized Statement createStatement(
		int resultSetType,
		int resultSetConcurrency)
		throws SQLException
	{
		if ((resultSetType != ResultSet.TYPE_SCROLL_INSENSITIVE)
			&& (resultSetType != ResultSet.TYPE_FORWARD_ONLY))
		{
			// invalid type.  check to see if it equals SCROLL_SENSITIVE
			if (resultSetType == ResultSet.TYPE_SCROLL_SENSITIVE)
			{
				generateWarning(VendorCode.MVAConnection_scrollTypesWarning);
				this.resultSetType = ResultSet.TYPE_SCROLL_INSENSITIVE;
			}
			else
			{
				generateWarning(VendorCode.MVAConnection_rsTypeWarning);
				this.resultSetType = ResultSet.TYPE_FORWARD_ONLY;
			}
		}
		else
		{
			this.resultSetType = resultSetType;
		}
		if ((resultSetConcurrency != ResultSet.CONCUR_UPDATABLE)
			&& (resultSetConcurrency != ResultSet.CONCUR_READ_ONLY))
		{
			//throw SQL warning here and set to default value, read only....
			generateWarning(VendorCode.MVAConnection_concurrencyTypesWarning);
			this.resultSetConcurrency = ResultSet.CONCUR_READ_ONLY;
		}
		else
		{
			this.resultSetConcurrency = resultSetConcurrency;
		}
		return createStatement();
	}
	/**
	 * Same as prepareStatement() above, but allows the default result set
	 * type and result set concurrency type to be overridden. If an invalid type or
	 * concurrency is passed, the values are defaulted to TYPE_FORWARD_ONLY and
	 * CONCUR_READ_ONLY.
	 * If an invalid scroll type or concurrency is requested, then the next closest
	 * scrolling/concurrency is substituted, and a warning is generated.
	 * 
	 * JDBC 2.0
	 *
	 * @param resultSetType a result set type, see ResultSet.TYPE_XXX
	 * @param resultSetConcurrency a concurrency type, see ResultSet.CONCUR_XXX
	 * @return a new PreparedStatement object containing the
	 * pre-compiled SQL statement
	 * @exception SQLException This exception is thrown if a PreparedStatement object cannot be created.
	 */
	public synchronized PreparedStatement prepareStatement(
		String sql,
		int resultSetType,
		int resultSetConcurrency)
		throws SQLException
	{
		if ((resultSetType != ResultSet.TYPE_FORWARD_ONLY)
			&& (resultSetType != ResultSet.TYPE_SCROLL_INSENSITIVE))
		{
			//throw SQL warning here...
			if (resultSetType == ResultSet.TYPE_SCROLL_SENSITIVE)
			{
				generateWarning(VendorCode.MVAConnection_scrollTypesWarning);
				this.resultSetType = ResultSet.TYPE_SCROLL_INSENSITIVE;
			}
			else
			{
				generateWarning(VendorCode.MVAConnection_rsTypeWarning);
				this.resultSetType = ResultSet.TYPE_FORWARD_ONLY;
			}
		}
		else
		{
			this.resultSetType = resultSetType;
		}
		if ((resultSetConcurrency != ResultSet.CONCUR_READ_ONLY)
			&& (resultSetConcurrency != ResultSet.CONCUR_UPDATABLE))
		{
			//throw SQL warning here...
			generateWarning(VendorCode.MVAConnection_concurrencyTypesWarning);
			this.resultSetConcurrency = ResultSet.CONCUR_READ_ONLY;
		}
		else
		{
			this.resultSetConcurrency = resultSetConcurrency;
		}
		return prepareStatement(sql);
	}
	/**
	 * Same as prepareCall() above, but allows the default result set
	 * type and result set concurrency type to be overridden. If an invalid type or
	 * concurrency is passed, the values are defaulted to TYPE_FORWARD_ONLY and
	 * CONCUR_READ_ONLY.
	 * If an invalid scroll type/concurrency is requested, the next closest scroll
	 * type/concurrency is substituted, and a warning is generated.
	 * 
	 * JDBC 2.0
	 * 
	 * 
	 * @param resultSetType a result set type, see ResultSet.TYPE_XXX
	 * @param resultSetConcurrency a concurrency type, see ResultSet.CONCUR_XXX
	 * @return a new CallableStatement object containing the
	 * pre-compiled SQL statement
	 * @exception SQLException This exception is thrown if a CallableStatement object cannot be created.
	 */
	public CallableStatement prepareCall(
		String sql,
		int resultSetType,
		int resultSetConcurrency)
		throws SQLException
	{
		if ((resultSetType != ResultSet.TYPE_FORWARD_ONLY)
			&& (resultSetType != ResultSet.TYPE_SCROLL_INSENSITIVE))
		{
			if (resultSetType == ResultSet.TYPE_SCROLL_SENSITIVE)
			{
				generateWarning(VendorCode.MVAConnection_scrollTypesWarning);
				this.resultSetType = ResultSet.TYPE_SCROLL_INSENSITIVE;
			}
			else
			{
				generateWarning(VendorCode.MVAConnection_rsTypeWarning);
				this.resultSetType = ResultSet.TYPE_FORWARD_ONLY;
			}
		}
		else
		{
			this.resultSetType = resultSetType;
		}
		if ((resultSetConcurrency != ResultSet.CONCUR_READ_ONLY)
			&& (resultSetConcurrency != ResultSet.CONCUR_UPDATABLE))
		{
			//throw SQL warning here...
			generateWarning(VendorCode.MVAConnection_concurrencyTypesWarning);
			this.resultSetConcurrency = ResultSet.CONCUR_READ_ONLY;
		}
		else
		{
			this.resultSetConcurrency = resultSetConcurrency;
		}
		return prepareCall(sql);
	}
	/**
	* Gets the type-map object associated with this connection.
	 * <P>
	* By default, the map returned is empty.
	 * Method not supported.
	 * </P>
	 * 
	 * JDBC 2.0
	 *
	 * @exception SQLException This exception is always thrown as SAS software does not support user defined types.
	*/
	public java.util.Map getTypeMap() throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotSupported);
		return null;
	}
	/**
	* Install a type-map object as the default type-map for  this connection.
	 * Method not supported.
	 * 
	 * JDBC 2.0
	 *
	 * @exception SQLException This exception is always thrown as SAS software does not support user defined types.
	*/
	public void setTypeMap(java.util.Map map) throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotSupported);
	}
	/**
	 * Deletes statements associated with this object.
	 * 
	 * @param statement statement object to be removed from the statementList
	 */
	protected synchronized void deleteStatement(Object statement)
	{        
        statementList.remove(statement);
	}
    /*
     * called by statement's close method to signal to the connection that the
     * statement has closed, and it's ok to delete it.
     */
    protected  void statementClosed(Object statement)
    {
        if (statement != null)
            deleteStatement(statement);
    }
	/*
	 * Sets user.
	 */
	private void setUser(java.util.Properties info)
	{
		String user = info.getProperty(JDBCConstants.USER.toLowerCase());
		if (user == null)
			user = info.getProperty(JDBCConstants.USER_NAME.toLowerCase());
		this.user = user;
	}
	/*
	 * Gets user associated to this object.
	 * @return String username associated with this object
	 */
	protected String getUser()
	{
		if (user == null)
			return "";
		else
			return user;
	}
	/*
	 * Gets url associated with this object.
	 * @return String url associated with this object
	 */
	protected String getUrl()
	{
		return url;
	}
	/*
	 * Sets remarks.
	 */
	private void setRemarks(java.util.Properties info)
	{
		boolean remarks = false;
		String temp = info.getProperty(JDBCConstants.REMARKS.toLowerCase());
		if (temp != null)
		{
			remarks = Boolean.valueOf(temp).booleanValue();
		}
		this.remarks = remarks;
	}
	/*
	 * Gets remarks.
	 * @return boolean remarks
	 */
	protected boolean getRemarks()
	{
		return remarks;
	}
	/*
	 * Sets flag that determines if LDAP to be used for DatabaseMetaData.
	 */
	private void setLdapDbmd(java.util.Properties info)
	{
		boolean ldapDbmd = false;
		String temp = info.getProperty(LDAPConstants.LDAP_DBMD);
		if (temp != null)
		{
			ldapDbmd = Boolean.valueOf(temp).booleanValue();
		}
		this.ldapDbmd = ldapDbmd;
	}
	/*
	 * Gets flag that determines if LDAP to be used for DatabaseMetaData.
	 */
	protected boolean getLdapDbmd()
	{
		return ldapDbmd;
	}
	/*
	 * Protected method
	 * @return String baseDN associated with this object
	 */
	protected String getBaseDN()
	{
		return baseDN;
	}
	/*
	 * Protected method
	 * @return String baseDN associated with this object
	 */
	protected String getAuthentication()
	{
		return authentication;
	}
	/*
	 * Protected method
	 * @return String baseDN associated with this object
	 */
	protected String getLDAPProviderUrl()
	{
		return ldapProviderUrl;
	}
	/*
	 * Protected method
	 * @return String baseDN associated with this object
	 */
	protected String getLDAPUserName()
	{
		return ldapUserName;
	}
	/*
	 * Protected method
	 * @return String baseDN associated with this object
	 */
	protected String getLDAPPassword()
	{
		return ldapPassword;
	}
	/*
	 * Gets the state of the connection.
	 * @return connected A boolean value indicating the state of the connection.
	 */
	protected boolean isConnected()
	{
		return connected;
	}
	/*
	 * Sets the state of the connection.
	 * @param connected A boolean value indicating the state of the connected.
	 */
	private void setConnected(boolean connected)
	{
		this.connected = connected;
	}
	/*
	 * Gets host from url.
	 */
	private String getHost(String url) throws SQLException
	{
		int hoststart = 14;
		int hostend;
		hostend = url.indexOf(':', hoststart);
		if (hostend == -1)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_urlMustSpecifyServer);
		}
		if (url.length() < (hostend + 3))
		{
			// NOTE: this looks like a port error, not that no url is specified.
			// investigate on next pass.
			// AND if it *is* a port error, then this isn't really an error for getHost()
			// to catch, is it??
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_urlMustSpecifyServer);
		}
		return url.substring(hoststart, hostend);
	}
	/*
	 * Gets port from url.
	 */
	private String getPort(String url) throws SQLException
	{
		int hoststart = 14;
		int hostend, portend;
		Integer portInt;
		hostend = url.indexOf(':', hoststart);
		if (hostend == -1)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_urlMustSpecifyServer);
		}
		if (url.length() < (hostend + 3))
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_urlMustSpecifyServer);
		}
		portend = url.indexOf('/', hostend);
		if (portend == -1)
		{
			portInt = new Integer(url.substring(hostend + 1));
		}
		else
		{
			portInt = new Integer(url.substring(hostend + 1, portend));
		}
		return portInt.toString();
	}
	void DumpProperties(Properties properties)
	{
		Enumeration e;
		String key;
		e = properties.propertyNames();
		while (e.hasMoreElements())
		{
			key = (String) e.nextElement();
			System.out.println(
				key + " - |" + properties.getProperty(key) + "|");
		}
	}
	/*
	 * Executes RESET UNDO_POLICY = NONE statement.
	 */
	private long executeUndoPolicyNoneStatement(java.util.Properties info)
		throws SQLException
	{
		boolean undoPolicyNone = true;
		boolean foreignDbms = false;
		String undo = "RESET UNDO_POLICY = NONE";
		long status = 0;
		try
		{
			String temp =
				info.getProperty(JDBCConstants.UNDO_POLICY_NONE.toLowerCase());
			if (temp != null)
			{
				undoPolicyNone = Boolean.valueOf(temp).booleanValue();
			}
			temp = info.getProperty(JDBCConstants.DBMS.toLowerCase());
			if (temp != null && temp.indexOf("SQLVIEW") == -1)
			{
				foreignDbms = true;
			}
			if ((undoPolicyNone == true) && (foreignDbms == false))
			{
				status = xUtil.executeUpdate(undo);
				SQLWarning w = xUtil.getWarning();
				if (w != null)
				{
					if (warnings == null)
						warnings = w;
					else
						warnings.setNextWarning(w);
				}
			}
		}
		catch (Exception e)
		{
			MVALocalizedError.throwLocalizedException(
				VendorCode.MVAConnection_executeUndoPolicyError,
				null,
				e);
			/* SQLException se = new SQLException(error, null, VendorCode.VC_UNKNOWNERROR);
			  if (e instanceof SQLException)
			    se.setNextException((SQLException)e);
			  throw se;
			*/
		}
		return status;
	}
	/*
	 * Sets vector containing list of assigned ILibref objects.
	 */
	private void setLibrefList(Vector librefList)
	{
		this.librefList = librefList;
	}
	/*
	 * Converts Properties object to url to get stringifiedObjectReference.
	 */
	private static String propertiesToUrl(Properties properties)
	{
		StringBuffer buffer = new StringBuffer("bridge://");
		buffer.append(properties.getProperty(WorkspaceConstants.HOST));
		buffer.append(":");
		buffer.append(properties.getProperty(WorkspaceConstants.PORT));
		buffer.append("/");
		buffer.append(WorkspaceConstants.SASAUTOMATIONID);
		StringBuffer query = null;
		Enumeration propertyNames = properties.propertyNames();
		while (propertyNames.hasMoreElements())
		{
			String propertyName = (String) propertyNames.nextElement();
			if (!propertyName.equals(WorkspaceConstants.HOST)
				&& !propertyName.equals(WorkspaceConstants.PORT))
			{
				if (query == null)
				{
					query = new StringBuffer();
				}
				else
				{
					query.append("&");
				}
				query.append(propertyName);
				query.append("=");
				query.append(properties.getProperty(propertyName));
			}
		}
		if (query != null)
		{
			buffer.append("?");
			buffer.append(query.toString());
		}
		return buffer.toString();
	}
	/*
	 * convenience routine for throwing MVALocalizedError warnings.
	 */
	private void generateWarning(int vc)
	{
		SQLWarning w =
			new SQLWarning(MVALocalizedError.getString(vc), null, vc);
		if (warnings == null)
			warnings = w;
		else
			warnings.setNextWarning(w);
	}
	protected XUtil getXUtil()
	{
		return xUtil;
	}
	/**
	 * Creates a Statement object that will generate ResultSet objects with the given type, concurrency, and holdability.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public Statement createStatement(
		int resultSetType,
		int resultSetConcurrency,
		int resultSetHoldability)
		throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
	/**
	 * Retrieves the current holdability of ResultSet objects created using this Connection object.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public int getHoldability() throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return 0;
	}
	/**
	 * Creates a CallableStatement object that will generate ResultSet objects with the given type and concurrency. 
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public CallableStatement prepareCall(
		String sql,
		int resultSetType,
		int resultSetConcurrency,
		int resultSetHoldability)
		throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
	/**
	 * Creates a default PreparedStatement object that has the capability to retrieve auto-generated keys.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public PreparedStatement prepareStatement(
		String sql,
		int autoGeneratedKeys)
		throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
	/**
	 * Creates a default PreparedStatement object capable of returning the auto-generated keys designated by the given array.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public PreparedStatement prepareStatement(String sql, int[] columnIndexes)
		throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
	/**
	 * Creates a PreparedStatement object that will generate ResultSet objects with the given type, concurrency, and holdability.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public PreparedStatement prepareStatement(
		String sql,
		int resultSetType,
		int resultSetConcurrency,
		int resultSetHoldability)
		throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
	/**
	 * Creates a default PreparedStatement object capable of returning the auto-generated keys designated by the given array.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public PreparedStatement prepareStatement(String sql, String[] columnNames)
		throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
	/**
	 * Removes the given Savepoint object from the current transaction.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public void releaseSavepoint(Savepoint savepoint) throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
	}
	/**
	 * Undoes all changes made after the given Savepoint object was set.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public void rollback(Savepoint savepoint) throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
	}
	/**
	 * Changes the holdability of ResultSet objects created using this Connection object to the given holdability.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public void setHoldability(int holdability) throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
	}
	/**
	 * Creates an unnamed savepoint in the current transaction and returns the new Savepoint object that represents it.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public Savepoint setSavepoint() throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
	/**
	 * Creates a savepoint with the given name in the current transaction and returns the new Savepoint object that represents it.
	* Empty method implementation for JDK1.4 compliance.
	* @exception SQLException This exception is always thrown to indicate the method is not implemented.
	*/
	public Savepoint setSavepoint(String name) throws SQLException
	{
		MVALocalizedError.throwLocalizedException(
			VendorCode.General_methodNotImplemented);
		return null;
	}
}
