/* $Id$ */
/* Copyright (c)1998 SAS Institute Inc, Cary NC USA 27513
 * Unpublished- All Rights Reserved.
 */

package com.sas.rio;

import com.sas.iom.JDBCConstants;
import com.sas.iom.WorkspaceConstants;
import com.sas.iom.SAS.IDataService;
import java.sql.Connection;
import java.sql.DriverPropertyInfo;
import java.sql.SQLException;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Enumeration;
//
//History Records:
//
// Date             Name                     Remarks
// 07/13/99    M Srinivasan             -Added documentation for encryptionPolicy,content and algorithms.
//                                  Added librefs,encryptionPolicy,encryptionAlgorithms,encryptionContent to
//                                  getPropertyInfo method.
// 10/25/99         M Srinivasan        -S0068741 :Change driver version to 1.1.
// 12/02/99     M Srinivasan        -S0071131 :Code change to support LDAP DBMD properties like ldapDbmd,
//                                   authentication and so on in the getPropertyInfo() call.
// 12/14/99     M Srinivasan        -S0072011 :Use Constants and avoid hardcoding property names.
// 10/27/00     B Beers             -S0100866 :Change driver version to 1.3.
//  02Feb01     Russell Gonsalves   S0102199: MVADRIVER.GETPROPERTYINFO() ALWAYS THROWS JAVA.LANG.ARRAYINDEXOUTOFBOUNDSEXCEPTION
//
// 15Feb2002    D Mackie            S0139820: made property names case insensitive and upcased the dbms property
// 5 Feb2003    D Mackie            S0182471: fixed issue with properties in the url blowing away passed in properties.
// 1 May2003    D Mackie            S0198893: removed applyInformats property from m_driverPropertyInfo.
// 29May2003    D Mackie            S0203447: removed hard-coded property descriptions, and substituted calls to
//                                            MVALocalizedError.getString() instead (in createDefaultProperties).
// 26Jun2003    D Mackie            S0207780: removed properties related to turning on the LDAP database metadata functionality.
//                                            This is the first step in totally removing this bit of functionality in 9.2.  
// 22Sep2003    D Mackie            S0221218: fixed property persistence issue, and changed getDriverPropertyInfo to return 
//                                            default property info (and whatever user passes in via Properties object)
// 26Apr2004    damack              S0189758: javadoc fix (changed dbms_options to dbmsOptions)
// 23Sep2004    damack              S0235627: added a new driver property to control sasify switching.
// 10Jan2005    damack              S0278967: addressing typing issue by adding a property to turn fix on and off.
/**
 * <P>MVADriver class implements the Driver interface.
 *
 * <P>The DriverManager will try to load as many drivers as it can
 * find and then for any given connection request, it will ask each
 * driver in turn to try to connect to the target URL.
 *
 * <P>When a Driver class is loaded, it should create an instance of
 * itself and register it with the DriverManager. This means that a
 * user can load and register a driver by doing
 * Class.forName("foo.bah.Driver").
 * @see MVAConnection
 */
public class MVADriver implements java.sql.Driver {
    public static final int majorVersion = 1;
    public static final int minorVersion = 5;
    Properties m_driverProperties = null;
    DriverPropertyInfo m_driverPropertyInfo[] = null;

    static {
        try {
            java.sql.Driver driver = (java.sql.Driver) Class.forName("com.sas.rio.MVADriver").newInstance();
            java.sql.DriverManager.registerDriver(driver);
        }
        catch(Exception e) {}
        catch(ClassFormatError e1) {}
    }
    /**
     * Constructs an MVADriver object and registers the driver with the java.sql.DriverManager.
     * @exception SQLException This exception is thrown if an error condition is detected.
     */
    public MVADriver() throws SQLException 
    {
     createDefaultProperties();
    }

    /**
     * Attempts to make a database connection to the server specified by the URL. The driver returns a NULL if it determines that it is the wrong kind
     * of driver to connect to the server. The driver raises an SQLException if it is the right driver to connect to the server, but cannot connect to
     * the database.
     * <p>
     * The connect method accepts a string representation of a URL and a java.util.Properties object as input.
     *
      * @param url The URL of the database to connect to. If the URL does not specify sasiom, a NULL is returned.
     * <p>
      * The URL must be of the form:
     * <pre>
     *      jdbc:sasiom://<i>hostname:portnumber</i>
     *     </pre>
     * where <code><i>hostname</i></code> is the name of the machine where the MVA server is running,
     * and  <code><i>portnumber</i></code> is the port the MVA server is configured to use.
     * <p>
     * @param info A list of connection arguments as a Properties object.
     *  The properties for the MVAConnection class are as follows:
     * <p>
     * <dl>
     * <dt><code>userName</code>
     * <dd>A valid username for the MVA server you specify in the URL.
     * <p>
     * <dt><code>password</code>
     * <dd>The password that corresponds to the username you specify with the <code>userName</code> property.
     * <p>
     * <dt><code>librefs</code>
     * <dd>A list of libnames and corresponding paths that specify the location of each data set your program will
     * access. The <code>librefs</code> property value must be specifies as follows:
     * <pre>
     *      (<i>libname</i> <i>'filepath'</i>; <i>libname</i> <i>'filepath'</i>;...)
     * </pre>
     * <p>
     * <dt><code>dbms</code>
     * <dd>The name of the DBMS your program will access; the value must be eight characters padded with blanks. The default is "SQLVIEW ", which specifies that
     * SAS data sets will be accessed. If you are not accessing SAS data, specify the name of the foreign DBMS as it is defined using
     * SAS/ACCESS software.
     * <p>
     * <dt><code>dbmsOptions</code>
     * <dd>Options for the foreign DBMS your program will access. Specify the options as you specify the database-definition arguments when using the
     * PROC SQL statement CONNECT TO.
     * <p>
     * <dt><code>remarks</code>
     * <dd>A Boolean value that determines whether the SAS/Integrated Object Model driver for JDBC returns the data
     * set label for each data set in the library that your program accesses. If the value is
     * True, the labels are returned; the default value is False.
     * <p>
     * Note: When this property is set to True, any operation that involves reading and returning data set labels can
     * result in poor performance because each data set must be opened in order to read the label. For example,
     * if this property is set to True, performance may be degraded when using the MVADatabaseMetaData.getTables()
     * method to retrieve a description of the tables in a catalog.
     * <p>
     * <dt><code>undoPolicyNone</code>
     * <dd>A Boolean value that determines whether changes to the data are undone when an UPDATE or INSERT statement
     * generates errors. When the value is set to True (the recommended setting), the changes are not undone; when
     * the value is set to False, the server attempts to undo the changes.
     * <p>
      * <DT><code>encryptionPolicy</code><DD>whether or not JavaBridge should attempt to negotiate with
     * the server over which encryption algorithm to use and what to do if the negotiations fail.
     * Possible values:
     * <DL>
     * <DT>none<DD>do not use encryption. This is the default.
     * <DT>optional<DD>attempt to use encryption but, if algorithm negotiation fails, continue with an
     * unencrypted session
     * <DT>required<DD>attempt to use encryption but, if algorithm negotiation fails, fail the connection
     * </DL>
     * <p>
     * <DT><code>encryptionContent</code><DD>specifies which messages should be encrypted if encryption is used. Possible values:
     * <DL>
     * <DT>all<DD>encrypt all messages. This is the default.
     * <DT>authentication<DD>encrypt only messages that contain user name and password information.
     * </DL>
     * <p>
     * <DT><code>encryptionAlgorithms</code><DD>the list of algorithms you are willing to use in order of
     * preference. Values in the list should be separated by commas and chosen from sasproprietary,
     * rc2, rc4, des, or tripledes. If no value is specified, then one of the server's favorite algorithms
     * will be used.
     * <p>
     * Note that the server's encryption policy is, by default, optional so no special action is needed
     * to prepare an IOM BRIDGE server to use encryption. However, it is possible to set up the server
     * so that encryption is required. Consult IOM BRIDGE server documentation for details.
    * <p>
     * <DT><code>applyFormats</code><DD>specifies whether or not column formats are applied
     * when retrieving data. If the value is true, all data is returned as Strings. For any column, the
     * format used is the one defined on the column in the dataset. If "applyFormats" is set to false
     * (the default case) no formatting is applied. 
    * </dl>
    * <p>
    * <b><i>Note</i></b>: Users have the ability to pass in properties embedded in the url.  The syntax
    * for this is "<i>url?[attr=value]&[...]</i>", where attr is a valid property name, and value is a valid name.  If users
    * make use of this functionality, then if properties are also passed in via the Properties object, the properties in 
    * the url will, in the case of conflicts, override those in the Properties object.  This means that if a user 
    * specifies a libref (for example) in both the URL and in the Properties object, the libref in the url will be 
    * the one to take precedence, and the one in the Properties object will be ignored.

     * @return A connection to the URL if the correct form of URL is used; otherwise, a NULL is returned.
     * @exception SQLException This exception is thrown if a RIOException is detected.
     */
    
/* S0182471 - fix problem with url properties blowing away the "Properties info" object */

    public synchronized Connection connect(String url, java.util.Properties info) throws SQLException 
    {
        Connection con = null;
        String temp;
        

        if(url.length() < 17)
            return null;

        /**
         * TODO:lower case all the names in the properties "info" object
         */

        /**
         * dev work
         */
        if (info == null) info = new java.util.Properties();
        
        int qi = url.indexOf("?");
        if (qi >= 0)
            {
                qi++; // get off the '?'
                String query = url.substring(qi, url.length());
                String tu = url.substring(0, qi - 1);
                url = tu;

                StringTokenizer pairOff = new StringTokenizer(query, "&", false);
                while (pairOff.hasMoreTokens())
                  {
                    String pair = pairOff.nextToken();
                    char last = pair.charAt(pair.length() - 1);
                    if (last == '\\')
                      {

                        String realPair = pair.substring(0, pair.length() - 1) + "&" + pairOff.nextToken();

                        pair = realPair;
                      }

                    // after this comment, you'd set the properties.
                    int pi = pair.indexOf("=");
                    String propN = pair.substring(0, pi);
                    String propV = pair.substring(pi + 1, pair.length());

                    info.put(propN, propV);
                  }
            }
        if(info != null) 
            {
                temp = info.getProperty(JDBCConstants.USER.toLowerCase());
                if(temp != null)
                    info.put(JDBCConstants.USER_NAME.toLowerCase(),temp);
            }
        
        /**
         * end dev work
         */
        java.util.Properties cprops = updateDriverProperties(info);
        if(url.regionMatches(0,"jdbc:sasiom://",0,14)) {
          con = new MVAConnection(url,cprops);
        }
        return con;
    }


    //NON-STANDARD JDBC METHOD
    //ALLOWS DIRECT CONNECTION TO BE MADE USING IOM'S SASDataService
    public synchronized Connection connect(IDataService sasDataService, java.util.Properties info) throws SQLException {
        Connection con = null;
        String temp;

        if(info != null) {
            temp = info.getProperty(JDBCConstants.USER.toLowerCase());
            if(temp != null)
                info.put(JDBCConstants.USER_NAME.toLowerCase(),temp);
        }


        java.util.Properties cprops = updateDriverProperties(info);
        con = new MVAConnection(sasDataService,cprops);

        return con;
    }


    void createDefaultProperties()
    {
      m_driverPropertyInfo = new DriverPropertyInfo[14];
      m_driverPropertyInfo[0] = new DriverPropertyInfo(JDBCConstants.USER_NAME.toLowerCase(), "");
      m_driverPropertyInfo[0].description = MVALocalizedError.getString(VendorCode.DESC_USER_NAME);

      m_driverPropertyInfo[1] = new DriverPropertyInfo(JDBCConstants.PASSWORD.toLowerCase(), "");
      m_driverPropertyInfo[1].description = MVALocalizedError.getString(VendorCode.DESC_PASSWORD);

      m_driverPropertyInfo[2] = new DriverPropertyInfo(JDBCConstants.DBMS.toLowerCase(), "SQLVIEW ");
      m_driverPropertyInfo[2].description = MVALocalizedError.getString(VendorCode.DESC_DBMS);

      m_driverPropertyInfo[3] = new DriverPropertyInfo(JDBCConstants.DBMS_OPTIONS.toLowerCase(), "");
      m_driverPropertyInfo[3].description = MVALocalizedError.getString(VendorCode.DESC_DBMS_OPTIONS);

      m_driverPropertyInfo[4] = new DriverPropertyInfo(JDBCConstants.REMARKS.toLowerCase(), "false");
      m_driverPropertyInfo[4].description = MVALocalizedError.getString(VendorCode.DESC_REMARKS);
      m_driverPropertyInfo[4].choices = new String[2];
      m_driverPropertyInfo[4].choices[0] = "true";
      m_driverPropertyInfo[4].choices[1] = "false";

      m_driverPropertyInfo[5] = new DriverPropertyInfo(JDBCConstants.UNDO_POLICY_NONE.toLowerCase(), "true");
      m_driverPropertyInfo[5].description = MVALocalizedError.getString(VendorCode.DESC_UNDO_POLICY_NONE);                                            
      m_driverPropertyInfo[5].choices = new String[2];
      m_driverPropertyInfo[5].choices[0] = "true";
      m_driverPropertyInfo[5].choices[1] = "false";

      m_driverPropertyInfo[6] = new DriverPropertyInfo(JDBCConstants.LIBREFS.toLowerCase(), "");
      m_driverPropertyInfo[6].description = MVALocalizedError.getString(VendorCode.DESC_LIBREFS);
      m_driverPropertyInfo[7] = new DriverPropertyInfo(WorkspaceConstants.ENCRYPTIONPOLICY.toLowerCase(), "");
      m_driverPropertyInfo[7].description = MVALocalizedError.getString(VendorCode.DESC_ENCRYPTIONPOLICY);

      m_driverPropertyInfo[7].choices = new String[3];
      m_driverPropertyInfo[7].choices[0] = "none";
      m_driverPropertyInfo[7].choices[1] = "optional";
      m_driverPropertyInfo[7].choices[2] = "required";


      m_driverPropertyInfo[8] = new DriverPropertyInfo(WorkspaceConstants.ENCRYPTIONCONTENT.toLowerCase(), "");
      m_driverPropertyInfo[8].description = MVALocalizedError.getString(VendorCode.DESC_ENCRYPTIONCONTENT);
      m_driverPropertyInfo[8].choices = new String[2];
      m_driverPropertyInfo[8].choices[0] = "all";
      m_driverPropertyInfo[8].choices[1] = "authentication";

      m_driverPropertyInfo[9] = new DriverPropertyInfo(WorkspaceConstants.ENCRYPTIONALGORITHMS.toLowerCase(), "");
      m_driverPropertyInfo[9].description = MVALocalizedError.getString(VendorCode.DESC_ENCRYPTIONALGORITHMS);

      m_driverPropertyInfo[10] = new DriverPropertyInfo(JDBCConstants.APPLY_FORMATS.toLowerCase(), "false");
      m_driverPropertyInfo[10].description = MVALocalizedError.getString(VendorCode.DESC_APPLY_FORMATS);

      m_driverPropertyInfo[11] = new DriverPropertyInfo("rmiurl", "");
      m_driverPropertyInfo[11].description=MVALocalizedError.getString(VendorCode.DESC_RMIURL);
      /** rmiUrl, if not blank, provides the url for the appserver where the RMI util and workspace will be. */
      
      /* 
       * new driver property, for fix S0235627 (and related defect S0266879)
       * sasifyTableNames specifies which sasify behavior will take place in statement for
       * the lifetime of the connection.
       * default - the behavior specified in the fix for defects S0235627/S0266879
       * v91 - the behavior prior to the fix (ie, what shipped in 9.1)
       * none - no tablename modification performed (no spaces around dots are stripped, and no
       *        3-level to 2-level name modification)
       */
      m_driverPropertyInfo[12] = new DriverPropertyInfo("sasifytablenames".toLowerCase(), "default");
      m_driverPropertyInfo[12].description = MVALocalizedError.getString(VendorCode.DESC_SASIFYTABLENAMES);
      m_driverPropertyInfo[12].choices = new String[3];
      m_driverPropertyInfo[12].choices[0] = "default";
      m_driverPropertyInfo[12].choices[1] = "v91";
      m_driverPropertyInfo[12].choices[2] = "none";
      
      m_driverPropertyInfo[13] = new DriverPropertyInfo(JDBCConstants.COERCE_NUMERIC_TYPES.toLowerCase(), "true");
      // insert description here
      m_driverPropertyInfo[13].description = MVALocalizedError.getString(VendorCode.DESC_COERCENUMERICTYPES);
      m_driverPropertyInfo[13].choices = new String[2];
      m_driverPropertyInfo[13].choices[0] = "true";
      m_driverPropertyInfo[13].choices[1] = "false";
      
      int i;

      m_driverProperties = new Properties();
      for(i = 0; i < m_driverPropertyInfo.length;i++){
        m_driverProperties.put(m_driverPropertyInfo[i].name,m_driverPropertyInfo[i].value);
      }
    }

    java.util.Properties updateDriverProperties(java.util.Properties info)
    {
      String key;
      Object value;

      /*
       * create a properties object containing all the defaults, as specified in 
       * m_driverPropertyInfo.
       */
      java.util.Properties conProps = new java.util.Properties();
      for (Enumeration b = m_driverProperties.keys(); b.hasMoreElements();)
      {
        key = (String)(b.nextElement());
        value = m_driverProperties.get(key);
        conProps.put(key.toLowerCase(), value); 
      }

      if(info == null)
        return conProps;

      /**
       * gotta lowercase all the property keys...
       * that are in info.
       * We go thru, updating conProps with the new values, lowercasing the key
       * as we go.
       */

      for (Enumeration a = info.keys(); a.hasMoreElements(); )
      {
        key = (String)(a.nextElement());
        value = info.get(key);
        conProps.put(key.toLowerCase(), value);
      }
       
      /*
       * check for "USER" and convert to "USER_NAME"
       */  
      String t = (String)conProps.get(JDBCConstants.USER.toLowerCase());
      if (t != null)
        conProps.put(JDBCConstants.USER_NAME.toLowerCase(), t);
      
      /*
       * check for a DBMS value, and upper case it, just to be safe
       */  
      t = (String)conProps.get(JDBCConstants.DBMS.toLowerCase());
      if (t != null)
      {
        conProps.put(JDBCConstants.DBMS.toLowerCase(), t.toUpperCase()); 
      }  
      return conProps;
    }


    /**
     * Returns True if the driver thinks that it can open a connection
     * to the given URL, else returns False.
     *
     * @param url The URL of the database.
     * @return True if this driver can connect to the given URL.
     * @exception SQLException This exception is required by the interface but it is never thrown.
     */
    public boolean acceptsURL(String url) throws SQLException {
        if(url.length() < 17)
            return false;
        if (url.regionMatches(0,"jdbc:sasiom://",0,14))
            return true;
        else
            return false;
    }
    /**
     * Gets a DriverPropertyInfo object that lists the properties
     * for the SAS/Integrated Object Model driver for JDBC. The properties are as follows:
     * <p>
     * <dl>
     * <dt><code>userName</code>
     * <dd>A valid username for the MVA server you specify in the URL.
     * <p>
     * <dt><code>password</code>
     * <dd>The password that corresponds to the username you specify with the <code>userName</code> property.
     * <p>
     * <dt><code>librefs</code>
     * <dd>A list of libnames and corresponding paths that specify the location of each data set your program will
     * access. The <code>librefs</code> property value must be specifies as follows:
     * <pre>
     *      (<i>libname</i> <i>'filepath'</i>; <i>libname</i> <i>'filepath'</i>;...)
     * </pre>
     * <p>
     * <dt><code>dbms</code>
     * <dd>The name of the DBMS your program will access; the value must be eight characters padded with blanks. The default is "SQLVIEW ", which specifies that
     * SAS data sets will be accessed. If you are not accessing SAS data, specify the name of the foreign DBMS as it is defined using
     * SAS/ACCESS software.
     * <p>
     * <dt><code>dbmsOptions</code>
     * <dd>Options for the foreign DBMS your program will access. Specify the options as you specify the database-definition arguments when using the
     * PROC SQL statement CONNECT TO.
     * <p>
     * <dt><code>remarks</code>
     * <dd>A Boolean value that determines whether the SAS/Integrated Object Model driver for JDBC returns the data
     * set label for each data set in the library that your program accesses. If the value is
     * True, the labels are returned; the default value is False.
     * <p>
     * Note: When this property is set to True, any operation that involves reading and returning data set labels can
     * result in poor performance because each data set must be opened in order to read the label. For example,
     * if this property is set to True, performance may be degraded when using the MVADatabaseMetaData.getTables()
     * method to retrieve a description of the tables in a catalog.
     * <p>
     * <dt><code>undoPolicyNone</code>
     * <dd>A Boolean value that determines whether changes to the data are undone when an UPDATE or INSERT statement
     * generates errors. When the value is set to True (the recommended setting), the changes are not undone; when
     * the value is set to False, the server attempts to undo the changes.
     * <p>
      * <DT><code>encryptionPolicy</code><DD>whether or not JavaBridge should attempt to negotiate with
     * the server over which encryption algorithm to use and what to do if the negotiations fail.
     * Possible values:
     * <DL>
     * <DT>none<DD>do not use encryption. This is the default.
     * <DT>optional<DD>attempt to use encryption but, if algorithm negotiation fails, continue with an
     * unencrypted session
     * <DT>required<DD>attempt to use encryption but, if algorithm negotiation fails, fail the connection
     * </DL>
     * <p>
     * <DT><code>encryptionContent</code><DD>specifies which messages should be encrypted if encryption is used. Possible values:
     * <DL>
     * <DT>all<DD>encrypt all messages. This is the default.
     * <DT>authentication<DD>encrypt only messages that contain user name and password information.
     * </DL>
     * <p>
     * <DT><code>encryptionAlgorithms</code><DD>the list of algorithms you are willing to use in order of
     * preference. Values in the list should be separated by commas and chosen from sasproprietary,
     * rc2, rc4, des, or tripledes. If no value is specified, then one of the server's favorite algorithms
     * will be used.
     * <p>
     * <DT><code>applyFormats</code><DD>specifies whether or not column formats are applied
     * when retrieving data. If the value is true, all data is returned as Strings. For any column, the
     * format used is the one defined on the column in the dataset. If "applyFormats" is set to false
     * (the default case) no formatting is applied. 
    * </dl>
     * <p>
      * @param url The URL of the database to connect to.
     * @param info A list of tag/value pairs.
     *
     * @return An array of DriverPropertyInfo objects.
     * @exception SQLException This exception is required by the interface but it is never thrown.
     */
    public DriverPropertyInfo[] getPropertyInfo(String url, java.util.Properties info)
                throws SQLException
    {
        int i;
        //java.util.Properties p = updateDriverProperties(info);
      DriverPropertyInfo[] driverPropInfo = new DriverPropertyInfo[m_driverPropertyInfo.length];
      for (i = 0; i < driverPropInfo.length; i++)
      {
        driverPropInfo[i] = new DriverPropertyInfo(m_driverPropertyInfo[i].name,
                                    m_driverPropertyInfo[i].value);
        driverPropInfo[i].choices = m_driverPropertyInfo[i].choices; 
        driverPropInfo[i].description = m_driverPropertyInfo[i].description;
        driverPropInfo[i].required = m_driverPropertyInfo[i].required;

      }
      if (info == null) return driverPropInfo;
      String key, value;

      for (Enumeration e = info.keys(); e.hasMoreElements(); )
      {
        key = (String)e.nextElement();
        value = (String)info.get(key); 
      
        // now update my copy, and return it.
        key = key.toLowerCase();
        for (i = 0; i < driverPropInfo.length; i++)
          {
            if (driverPropInfo[i].name.equals(key))
            {
                driverPropInfo[i].value = value;
                break;
            }
          }
      }      
      return driverPropInfo;
    }
    /**
     * Gets the driver's major version number.
     * @return the driver's major version number.
     */
    public int getMajorVersion() {
        return MVADriver.majorVersion;
    }
    /**
     * Gets the driver's minor version number.
     * @return the driver's minor version number.
     */
    public int getMinorVersion() {
        return MVADriver.minorVersion;
    }
    /**
     * Reports whether the Driver is a genuine JDBC COMPLIANT (tm) driver.
     * A driver may only report "true" here if it passes the JDBC compliance
     * tests, otherwise it is required to return false.
     *
     *
     * @return false.
     */
    public boolean jdbcCompliant() {
        return false;
    }
}

