/***********************************************************************/
/* Title - MVARow.java                                                 */
/* Author - Dave Mackie, SAS Institute                                 */
/* Date - 5 June 2002                                                  */
/* Purpose - first whack at trying to implement caching in the iom     */
/*           driver.                                                   */
/* S0194092: changed setMissing() code to use set instead of add       */
/* S0196790: changed m_missingData to use a boolean array instead of   */
/*           ArrayList.                                                */
/***********************************************************************/
package com.sas.rio;
import java.sql.Types;
import java.util.ArrayList;
public class MVARow
{
    /**
     * MVARow is intended to be a way to hide the ugly details of 
     * reading data with rio.  
     * That's why the bookmark info is getting stored on the row.
     */
    /**
     * the actual data, stored in an ArrayList so everything is 
     * treated as an object.
     */
    private ArrayList m_rowData = null;
    /**
     * a parallel "missing data" array
     */
    private boolean[] m_missingData = null;
    private byte[] m_bookmark = null;
    private int m_maxSize = -1;

    public MVARow(int columnCount)
    {
        m_rowData = new ArrayList(columnCount);
        m_missingData = new boolean[columnCount];
        for (int i = 0; i < columnCount; i++)
            m_missingData[i] = false;
        m_maxSize = columnCount;
    }
    public MVARow(MVARow row)
    {
        // we'll see if java lets you do the same trick that 
        // c++ lets you do.
        m_rowData = new ArrayList(row.m_rowData);
        m_missingData = new boolean[row.m_missingData.length];
	for (int i = 0; i < m_missingData.length; i++)
	    m_missingData[i] = row.m_missingData[i];

        m_maxSize = row.getMaxSize();
    }
    public void setBookmark(byte[] bookmark)
    {
        m_bookmark = bookmark;
    }
    public byte[] getBookmark()
    {
        return m_bookmark;
    }
    public int getBookmarkLength()
    {
        return m_bookmark.length;
    }
    public int size()
    {
        return m_rowData.size();
    }
    /** 
     * setData, add, getData 
     */
    public boolean setData(int column, Object data)
        throws IndexOutOfBoundsException
    {
        /** column is 1-based.
         *  arraylist positions are 0-based.
         * don't forget.
         */
        if (column < 0 || column >= m_maxSize) // bail, out of range
            throw new IndexOutOfBoundsException();
        //column = column - 1;				
        m_rowData.set(column, data);
        if (data != null)
            m_missingData[column] = false;
        else 
            m_missingData[column] = true;
        return true;
    }
    public boolean setData(Object data)
    {
        int newSize = m_rowData.size() + 1;
        if (newSize > m_maxSize)
            return false; // bail, out of range
        m_rowData.add(data);
        return true;
    }
    public boolean add(Object data)
    {
        int newSize = m_rowData.size() + 1;
        if (newSize > m_maxSize)
            return false; // bail, out of range
        m_rowData.add(data);
        return true;
    }
    public boolean add(int column, Object data)
        throws IndexOutOfBoundsException
    {
        /** column is 1-based.
         *  arraylist positions are 0-based.
         * don't forget.
         */
        if (column < 0 || column >= m_maxSize) // bail, out of range
            throw new IndexOutOfBoundsException();
        //	column = column - 1;				
        m_rowData.add(column, data);
        return true;
    }
    public Object getData(int column) throws IndexOutOfBoundsException
    {
        /** again, column is 1-based.
         */
        //int _column = column - 1;	
        if (column >= m_maxSize)
            throw new IndexOutOfBoundsException();
        if (column >= m_rowData.size())
            return null; // no data, but not a bad call
        return m_rowData.get(column);
    }
    public Object get(int column) throws IndexOutOfBoundsException
    {
        /** again, column is 1-based.
         */
        //int _column = column - 1;	
        if (column >= m_maxSize)
            throw new IndexOutOfBoundsException();
        if (column >= m_rowData.size())
            return null; // no data, but not a bad call
        return m_rowData.get(column);
    }
    /**
     * set/getMissing
     */
    public void setMissing(int column, boolean missing)
        throws IndexOutOfBoundsException
    {
        //    	int c = column - 1;
        if (column < 0 || column >= m_maxSize)
            throw new IndexOutOfBoundsException();
        /**
         * S0194092: changed call to add to set
         */
        m_missingData[column]= missing;
    }
    public boolean getMissing(int column) throws IndexOutOfBoundsException
    {
        //	int c = column - 1;
        if (column < 0 || column >= m_maxSize)
            throw new IndexOutOfBoundsException();
	return m_missingData[column];


    }
    public int getMaxSize()
    {
        return m_maxSize;
    }
}
