/* $Id$ */
/************************************************************************/
/* NAME: MVARowCache                                                     */
/* AUTHOR: Dave Mackie                                                   */
/* PURPOSE: help in implementing caching for ResultSets                  */
/* DATE: 7/10/2002, copyright SAS Institute, Cary NC USA 27513           */
/* COMMENTS                                                              */
/* S0160259 - fixed problems with cursor positioning.                    */
/* S0177195 - fixed issue with missing numerics being returned as NaN    */
/* S0127940 - added code to use RIO GetRecordCounts for SQL result sets  */
/* S0183030 - added code to check for null type designator in the column */
/*            types being passed over to xUtil.  If present, sub in the  */
/*            'actual' type.                                             */
/* S0184797 - added code to correctly count rows when "max rows" has been*/
/*            set.                                                       */
/* S0190257 - added call to refresh() at the end of a successful call to */
/*            insertRow (fixes issue where the inserted row isn't showing*/
/*            up in the cache when making an absolute positioning move,  */
/*            and the m_actual is already in the bound of the last cache */
/*            block.                                                     */
/* S0191832 - added code to updateRow to replace the Double.NaN and null */
/*            values in the m_updateRow with zero and "".  This syncs up */
/*            the expected results of a getXXXX after update with what   */
/*            you'd get after a refresh.                                 */
/* S0193089 - follow on fix to S0191832. Need to tweak getIsMissing,     */
/*            getData and updateData to make sure that the data is       */
/*            consistent regardless of when its retrieved from the RS.   */
/* S0195329 - fix for formatted updates. (Meaning, fixing the fact the   */
/*            if apply formats was turned on, updates didn't work        */
/*            properly.                                                  */
/* S0196790 - various performance fixes.  (changed the column type array */
/*            passed to readRecords so that it only gets created once,   */
/*            instead of every read.  changed readRecords' return type to*/
/*            DataHolder.  Changed all setMissing calls to use primitive */
/*            boolean.
/* S0214665 - for the beforeFirst method, I had a routine called "peek", */
/*            which would do a read to see if the rs is empty.  (if it   */
/*            is, then beforeFirst is supposed to have no effect).       */
/*            Unfortunately, unless your next forward move was a next()  */
/*            the peek flag that got set saying that the read was redun- */
/*            dant didn't get unset.  This had no effect if you used     */
/*            the absolute and relative positioning methods, but if you  */
/*            did an absolute and then a next, you were hosed.  The fix  */
/*            is to just check the peek flag in absolute, and if it's    */
/*            true, set it to false.  (i supposed we could just always   */
/*            set it to false).  This check/set is not necessary for the */
/*            other positioning methods because the peek flag should only*/
/*            get set to false when the cursor moves from the beforeFirst*/
/*            position, and none of the other positioning methods can do */
/*            that (if at all) w/o going thru next or absolute.          */
/* S0218630 - changed the exceptions from absolute()'s call to           */
/*            readRecords so that if an exception comes from rio, it gets*/
/*            propagated properly.  Changed count() so that the call to  */
/*            countRows will propagate the exception properly if one     */
/*            occurs.  Also, in refresh(), changed the try/catch and the */
/*            exception thrown to be proper.                             */
/* S0206209 - changed ctor for MVARowCache(int) so that cursor and actual*/
/*            get properly init-ed. Also, added check in next() s.t.     */
/*            if we're not reading, we don't go to the read section.     */
/* S0278967 - added ctors that take properties object to get type coerce */
/*            property to my level                                       */
/* S0308479 - added changes, at Barbara Walters' request, to improve     */
/*            performance in getColumnIndex()                            */
/*************************************************************************/
/* note to self, would it be smart to break out some of these metadata
   type methods to a different class, or would that just muddy the waters
   further?
*/
/**
 * note, a number of methods are marked as "protected".
 * this is because those methods should NEVER be called directly.
 * in fact, none of this class should ever be called by anyone outside
 * of com.sas.rio, and if they had a "protected" class, I'd use it.
 */
package com.sas.rio;
import java.util.ArrayList;
import java.sql.Types;
import java.sql.SQLException;
import com.sas.iom.SAS.IDataSet;
import com.sas.iom.SAS.NumericMissingUnderscore;
import com.sas.iom.SAS.IDataSetPackage.BindKeys;
import com.sas.iom.SAS.BaseType;
import com.sas.iom.SAS.NotMissing;
import java.sql.ResultSet;
/** public ? */
class MVARowCache
{
    boolean m_numericTypeCoercion = true;
    XUtil m_xUtil = null;
    int m_hashCode = -1;
    private boolean m_countValid = true;
    /**
     * below are statics that the original MVAResultSet was using.
     * I may ditch them later, but for now....
     */
    public static final int MAYBE_INTEGER = 5;
    public static final int INTEGER = 4;
    public static final int DATETIME = 3;
    public static final int TIME = 2;
    public static final int DATE = 1;
    public static final int NOSUBTYPE = 0;
    public static final int CHARACTER = 2;
    public static final int NUMERIC = 1;
    static final int MAXLEN = 32767;
    private String m_tableName = null;
    private final int EOF = -1;
    private final int BOF = -2;
    private final int AOK = -3;
    private int m_status = -2;
    /** ArrayList of MVARow objects */
    private Cache m_rows = null;
    /** type of each column is stored in this ArrayList */
    private ArrayList m_columnTypes = null;
    /** name of each column is stored in this ArrayList */
    private ArrayList m_columnNames = null;
    private ArrayList actualList = null;
    /** the lengths of each column is stored in this ArrayList.
     * Used by the metadata to get the column display size
     */
    int[] m_columnLengths = null;
    String[] m_columnLabels = null;
    String[] m_formatNames = null;
    short[] m_formatWidths = null;
    short[] m_formatDecimals = null;
    short[] m_formatLengths = null;
    String[] m_informatNames = null;
    short[] m_informatWidths = null;
    short[] m_informatDecimals = null;
    /**  this cursor refers to the position IN THE CACHE. */
    int m_cursor;
    /** this refers to the position in the RESULT SET. */
    private int m_actual;
    private int m_maxRows = 0;
    private boolean m_alreadyCounted = false;
    private boolean m_isDirect = false;
    boolean m_noreads = false;
    int m_cacheSize = MVAConnection.XHRSQPRECS; // 100
    int m_columnCount;
    boolean AFTERLAST = false;
    boolean BEFOREFIRST = false;
    boolean INSERTROW = false;
    private IDataSet m_dataSet = null;
    private boolean m_deleteFlag = false;
    // flag indicating that we are on a deleted row.
    private MVARow m_updateRow = null;
    int m_bookmarkLength = 0;
    private int m_bindKey = BindKeys._BindKeysUnformatted;
    private boolean m_cacheRefresh = false;
    private int m_rowCount = 0;
    private int m_rsType = ResultSet.TYPE_SCROLL_INSENSITIVE;
    private boolean m_peekFlag = false;
    private boolean m_cacheChangeFlag = false;
    private int m_newCache = 0;
    private int[] m_colt = null;

    /**
     * default constructor
     */
    public MVARowCache() throws SQLException
    {
        m_rows = new Cache(0);
        m_columnTypes = new ArrayList(0);
        m_columnNames = new ArrayList(0);
    }
    /**
     * default constructor
     */
    public MVARowCache(boolean typeCoercion) throws SQLException
    {
        m_rows = new Cache(0);
        m_columnTypes = new ArrayList(0);
        m_columnNames = new ArrayList(0);
        m_numericTypeCoercion = typeCoercion;
    }
    /**
     * constructor
     * @param columns number of columns in result set
     */
    public MVARowCache(int columns) throws SQLException
    {
        m_rows = new Cache(m_cacheSize);
        m_columnTypes = new ArrayList(columns);
        m_columnNames = new ArrayList(columns);
        m_columnLabels = new String[columns];
        m_columnLengths = new int[columns];
        m_formatLengths = new short[columns];
        m_formatNames = new String[columns];
        m_formatWidths = new short[columns];
        m_formatDecimals = new short[columns];
        m_informatNames = new String[columns];
        m_informatWidths = new short[columns];
        m_informatDecimals = new short[columns];
        m_colt = new int[columns];
        m_columnCount = columns;
     	/* S0206209: init cursor and actual properly */
      	m_cursor = -1;
      	m_actual = 0;
    }
    /**
     * constructor
     * @param sasDataSet IDataSet the dataset used to populate the cache.
     */
    public MVARowCache(int hashCode, XUtil xUtil) throws SQLException
    {
        m_xUtil = xUtil;
        m_hashCode = hashCode;
        /**
         * call init() to get the column information.
         * and set up the cache.
         */
        init();
        m_cursor = -1; // before the first row.
        m_actual = 0; // before the first row.
        BEFOREFIRST = true;
        m_rows = new Cache(m_cacheSize);
    }
    /**
     * constructor
     * @param sasDataSet IDataSet the dataset used to populate the cache.
     */
    public MVARowCache(int hashCode, XUtil xUtil, boolean typeCoercion) throws SQLException
    {
        m_xUtil = xUtil;
        m_hashCode = hashCode;
        m_numericTypeCoercion = typeCoercion;
        /**
         * call init() to get the column information.
         * and set up the cache.
         */
        init();
        m_cursor = -1; // before the first row.
        m_actual = 0; // before the first row.
        BEFOREFIRST = true;
        m_rows = new Cache(m_cacheSize);
    }
    /**
     * ctor to be used when users are calling "getTable" to get a result set.
     */
    public MVARowCache(
        XUtil _x,
        MVAStatement s,
        String tableName,
        String[] password)
        throws SQLException
    {
        m_xUtil = _x;
        /**
         * get a dataset first
         */
        openDataSet(s, tableName, password);
        /**
         * then call init as above
         */
        init();
        // how do we bail on a failure of getTable?
        if (m_hashCode == -1)
            return;
        m_cursor = -1; // before the first row.
        m_actual = 0; // before the first row.
        BEFOREFIRST = true;
        m_rows = new Cache(m_cacheSize);
    }
    /**
     * ctor to be used when users are calling "getTable" to get a result set.
     */
    public MVARowCache(
        XUtil _x,
        MVAStatement s,
        String tableName,
        String[] password,
        	boolean typeCoercion)
        throws SQLException
    {
        m_xUtil = _x;
        m_numericTypeCoercion = typeCoercion;
        /**
         * get a dataset first
         */
        openDataSet(s, tableName, password);
        /**
         * then call init as above
         */
        init();
        // how do we bail on a failure of getTable?
        if (m_hashCode == -1)
            return;
        m_cursor = -1; // before the first row.
        m_actual = 0; // before the first row.
        BEFOREFIRST = true;
        m_rows = new Cache(m_cacheSize);
    }
    /**
     * sets the cache size to a size other than the default.
     * @param size new cache size
     */
    public void setCacheSize(int size) throws SQLException
    {
        m_cacheSize = size;
        if (BEFOREFIRST || AFTERLAST)
        {
            m_rows.setSize(size);
            return; // my work is done here.
        }
        /**
         * exploring fixes for S0168579
         */
        else
        {
            if (m_rsType != ResultSet.TYPE_FORWARD_ONLY)
            {
                //MVARow r = (MVARow) m_rows.get(0);
                MVARow r = (MVARow) m_rows.get(m_cursor);
                byte[] startBookmark = r.getBookmark();
                m_rows.setSize(size);
                refresh(startBookmark);
                m_cursor = 0;
            }
            else
            {
                m_cacheChangeFlag = true;
                m_newCache = size;
            }
        }
    }
    public void setRSType(int type)
    {
        m_rsType = type;
    }
    /**
     * return the current cache size
     * @return int cache size
     */
    public int getCacheSize()
    {
        return m_cacheSize;
    }
    public void setMaxRows(int m)
    {
        m_maxRows = m;
    }
    public int getMaxRows()
    {
        return m_maxRows;
    }
    /**
     * gets the number of columns in the cache/resultset
     * @return int number of columns
     */
    public int getColumnCount()
    {
        return m_columnCount;
    }
    /**
     * set the number of columns in the cache.
     * used by the metadata classes.
     * @param int c number of columns
     */
    public void setColumnCount(int c)
    {
        m_columnCount = c;
    }
    /**
     * get the current position in the result set
     * @return int actual position
     */
    public int getActual()
    {
        if (BEFOREFIRST || AFTERLAST)
            return 0;
        return m_actual;
    }
    /**
     * query cache to see if cursor is off the resultset end.
     * @return boolean true means after last, false means we're not.
     */
    public boolean getAfterLast() throws SQLException
    {
        if (INSERTROW)
            return false;
        return AFTERLAST;
    }
    /**
     * query cache to see if cursor if before the beginning of the resultset
     * @return boolean
     */
    public boolean getBeforeFirst() throws SQLException
    {
        if (INSERTROW)
            return false;
        if (m_actual > 0)
            return BEFOREFIRST;
        if (m_alreadyCounted && m_rowCount > 0)
            return BEFOREFIRST;
        try
        {
            //            if (m_rsType == ResultSet.TYPE_FORWARD_ONLY)
            BEFOREFIRST = peek();
            //            if (count(m_actual) == 0)
            //                return false;
        }
        catch (Exception e)
        {
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_isBeforeFirstError);
        }
        return BEFOREFIRST;
    }
    /**
     * query to see if the cursor is on the first row.
     * @return boolean
     */
    public boolean getIsFirst()
    {
        if (BEFOREFIRST || AFTERLAST)
            return false;
        if (m_actual == 1)
            return true;
        return false;
    }
    /**
     * query to see if cursor is on the last row.
     * @return boolean
     */
    // NOTE, what if nothing's been read yet?
    public boolean getIsLast() throws SQLException
    {
        if (m_noreads)
        {
            if (m_actual == m_cacheSize)
                return true;
            else
                return false;
        }
        if (BEFOREFIRST || AFTERLAST)
            return false;
        if (m_status != EOF)
        {
            /**
             * I'd originally thought that if we weren't at EOF, then we were not on last,
             * but, as always, there's an exception.  If you have a cache of x rows, and the dataset
             * happens to also be x rows, then you're not going to get an EOF when you read x rows.
             * You'll get it when you attempt to make that next read.
             * So, we have to check to see if we've counted the rows, and if so, compare the actual pos
             * to the count, and use that as a check.  If the rows have not been counted, then I guess the
             * thing to do would be to either a: count the rows, and then check against the actual,
             * or just call read records directly, w/o storing the value, just to what status we get.
             */
            if (m_cursor < (m_rows.size() - 1))
                return false;
            int save = m_actual;
            boolean f = absolute(m_actual + 1);
            absolute(save);
            return !f;
        }
        if (m_cursor == (m_rows.size() - 1))
            return true;
        return false;
    }
    /**
     * return whether the current row has a missing value at the i-th column
     * @param int column
     * @return boolean
     */
    public boolean getIsMissing(int i)
    {
        i -= 1;
        if (!INSERTROW)
        {
            /**
             * S0193089: when calling getIsMissing, if the update row isn't null
             * (meaning an update is in progress), then check the update row for the
             * missing value.
             */
            if (m_updateRow != null)
            {
                return m_updateRow.getMissing(i);
            }
            /** end S0193089 */
            else
            {
                MVARow r = (MVARow) m_rows.get(m_cursor);
                return (r.getMissing(i));
            }
        }
        else
        {
            if (m_updateRow == null)
                /** by definition, if a row is null then it's missing
            */
                return true;
            return m_updateRow.getMissing(i);
        }
    }
    /**
     * bind key setter (applies to whether values will be returned
     * formatted or not from the dataset/resultset.
     * @param int bind the bindkey you wish
     */
    public void setBindkey(int bind)
    {
        /**
        if (bind == BindKeys._BindKeysAll)
            m_bindKey = BindKeys._BindKeysUnformatted;
        else
        */
        m_bindKey = bind;
    }
    /**
     * bind key getter
     * @return int bindkey
     */
    public int getBindkey()
    {
        return m_bindKey;
    }
    /**
     * column type setters
     * @param int column being affected
     * @param int type of data for the given column
     */
    public void setColumnType(int column, int type)
    {
        boolean f_a = false;
        int f_t = type;
        column -= 1;
        if (type == Types.NULL)
        {
            if (actualList == null)
                actualList = new ArrayList();
            ArrayList r = new ArrayList();
            r.add(new Integer(column));
            r.add((Integer) m_columnTypes.get(column));
            actualList.add(r);
            f_t = ((Integer) m_columnTypes.get(column)).intValue();
        }
        m_colt[column] = f_t;
        int s = m_columnTypes.size();
        if (s == 0)
            m_columnTypes.add(column, new Integer(type));
        else if (column > (s - 1))
            m_columnTypes.add(column, new Integer(type));
        else
            m_columnTypes.set(column, new Integer(type));
    }
    /**

    public void setColumnType(int type)
    {
    	m_columnTypes.add(new Integer(type));
    }
    */
    /**
     * column type getter
     * @param int column being queried
     * @return type of column
     */
    public int getColumnType(int column)
    {
        int c = column - 1;
        Integer i_type = (Integer) m_columnTypes.get(c);
        return i_type.intValue();
    }
    /**
     * metadata related methods
     */
    /**
     * get column length
     * @param int column
     * @return int length of column
     */
    public int getColumnLength(int column)
    {
        column -= 1;
        return m_columnLengths[column];
    }
    /**
     * get the label (not the name) of a given column
     * @param int c column being queried
     * @return String the column label
     */
    public String getColumnLabel(int c)
    {
        c -= 1;
        return m_columnLabels[c];
    }
    /**
     * get the column format
     * @param c column
     * @return String format of the column
     */
    public String getColumnFormat(int c)
    {
        c -= 1;
        return m_formatNames[c];
    }
    /**
     * get the column informat
     * @param int c column
     * @return String informat
     */
    public String getColumnInformat(int c)
    {
        c -= 1;
        return m_informatNames[c];
    }
    /**
     * get the width of the column format
     * @param int c column
     * @return short width
     */
    public short getFormatWidth(int c)
    {
        c -= 1;
        return m_formatWidths[c];
    }
    /**
     * get the format decimal
     * @param int c column
     * @return short the format decimal
     */
    public short getFormatDecimal(int c)
    {
        c -= 1;
        return m_formatDecimals[c];
    }
    /**
     * get the format length
     * @param int c column
     * @return short format length
     */
    public short getFormatLength(int c)
    {
        c -= 1;
        return m_formatLengths[c];
    }
    /**
     * get the informat width
     * @param int c column
     * @return short the informat width
     */
    public short getInformatWidth(int c)
    {
        c -= 1;
        return m_informatWidths[c];
    }
    /**
     * get informat decimal
     * @param int c column
     * @return short informat decimal
     */
    public short getInformatDecimal(int c)
    {
        c -= 1;
        return m_informatDecimals[c];
    }
    /**
     * set the name of a particular column
     * @param int column
     * @param String name
     */
    public void setColumnName(int column, String name)
    {
        column -= 1;
        m_columnNames.add(column, name);
    }
    /**
       public void setColumnName(String name)
       {
       m_columnNames.add(name);
       }
    */
    /**
     * get the name of a given column
     * @param int column
     * @return String the column name
     */
    public String getColumnName(int column)
    {
        column = column - 1;
        return (String) m_columnNames.get(column);
    }
    /**
     * set the column label (not the same thing as the name)
     * @param int c column
     * @param String s the label
     */
    public void setColumnLabel(int c, String s)
    {
        c -= 1;
        m_columnLabels[c] = s;
    }
    /**
     * set the column length
     * @param int c the column
     * @param int l the length
     */
    public void setColumnLength(int c, int l)
    {
        c -= 1;
        m_columnLengths[c] = l;
    }
    /**
     * set the format length
     * @param int c the column
     * @param short v the length
     */
    public void setFormatLength(int c, short v)
    {
        c -= 1;
        m_formatLengths[c] = v;
    }
    /**
     * set the format name
     * @param int c the column
     * @param String v the format name
     */
    public void setFormatName(int c, String v)
    {
        c -= 1;
        m_formatNames[c] = v;
    }
    /**
     * set the format width
     * @param int c the column
     * @param short v the width
     */
    public void setFormatWidth(int c, short v)
    {
        c -= 1;
        m_formatWidths[c] = v;
    }
    /**
     * set the format decimal
     * @param int c the column
     * @param short v the decimal
     */
    public void setFormatDecimal(int c, short v)
    {
        c -= 1;
        m_formatDecimals[c] = v;
    }
    /**
     * set the informat name
     * @param int c the column
     * @param String v the name
     */
    public void setInformatName(int c, String v)
    {
        c -= 1;
        m_informatNames[c] = v;
    }
    /**
     * set the informat width
     * @param int c the column
     * @param short v the informat width
     */
    public void setInformatWidth(int c, short v)
    {
        c -= 1;
        m_informatWidths[c] = v;
    }
    /**
     * set the informat decimal
     * @param int c the column
     * @param short v the decimal
     */
    public void setInformatDecimal(int c, short v)
    {
        c -= 1;
        m_informatDecimals[c] = v;
    }
    /**
     * use the column name to get the index of the column (1-based)
     * @param String columnName
     * @return int the index
     */
    public int getColumnIndex(String columnName)
    {
        /* S0308479: was using a different algorithm, but changed to this cleaner one
         * at mitchel's suggestion.
         * NOTE:  the names in m_columnNames were uppercased in .init().
         */
        int index = m_columnNames.indexOf(columnName.toUpperCase());
        if(index >= 0)
            return index+1;
        else
            return index;
    }

    /**
     * set the data in a given column
     * ?ever used?
     * @param int column
     * @param Object data
     */
    public void setData(int column, Object data)
    {
        column -= 1;
        MVARow m_row = (MVARow) m_rows.get(m_cursor);
        m_row.setData(column, data);
    }
    /**
       public void setData(Object data)
       {
       MVARow m_row = (MVARow)m_rows.get(m_cursor);
       m_row.setData(data);
       }
    */
    /**
     * get the data in a given column.  Since the underlying data struct is an arraylist,
     * everything is stored as an Object.  Therefore, that's what we always return.
     * means that the ResultSet getxxxx methods are essentially just casting the Object
     * (plus some other stuff for dates, for example)
     * @param int column
     * @return Object the data
     */
    public Object getData(int column) throws SQLException
    {
        if (!INSERTROW && (AFTERLAST || BEFOREFIRST))
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_invalidRowPosition);
        column -= 1;
        if (m_updateRow != null)
        {
            /**
             * S0193089: if the column value in the update row is not varchar and'is missing',
             * then return a 0.0 instead of NaN (which is what's internally stored.
             */
            if ((getColumnType(column + 1) != Types.VARCHAR)
                && (m_updateRow.getMissing(column)))
            {
                return new Double(0.0);
            }
            /** end S0193089 */
            else
            {
                return m_updateRow.getData(column);
            }
        }
        else
        {
            MVARow m_row = (MVARow) m_rows.get(m_cursor);
            if (m_row == null)
                return null;
            return m_row.getData(column);
        }
    }
    /**
       public void addRow()
       {
       MVARow m_row = new MVARow(m_columnCount);
       m_rows.add(m_row);
       }
    */
    /**
       public void dumpCache()
       {
       MVARow r = null;
       for (int i = 0; i < m_rows.size(); i++)
       {
       r = (MVARow)m_rows.get(i);
       System.out.print(r.get(0).toString());

       System.out.println("");
       }
       }
    */
    public void moveToInsertRow()
    {
        INSERTROW = true;
        /*
         * null out the previous
         * and create a new one
         */
        m_updateRow = new MVARow(m_columnCount);
        for (int i = 0; i < m_columnCount; i++)
        {
            switch (getColumnType(i + 1))
            {
                case Types.VARCHAR :
                    m_updateRow.add(i, null);
                    m_updateRow.setMissing(i, true);
                    break;
                default :
                    m_updateRow.add(i, new Double(Double.NaN));
                    m_updateRow.setMissing(i, true);
            }
        }
    }
    public void moveToCurrentRow()
    {
        INSERTROW = false;
        m_updateRow = null;
    }
    /*****************************************************************************/
    /* cursor movement routines.  most important part of class.                  */
    /*****************************************************************************/
    /*
     * m_cursor - cache cursor
     * m_actual - actual position in the result set.
     */
    /**
     * moves the cursor forward in the cache by 1, fetching more
     * records over the wire as necessary.
     * @return boolean true if positioned on a valid row, false if not
     */
    public boolean next() throws SQLException
    {
        if (m_peekFlag)
        {
            m_peekFlag = false;
            BEFOREFIRST = false;
            // have to set beforefirst to false, since peeking
            // can leave it at true.
            return true;
        }
        boolean success = false;
        moveToCurrentRow();
        if (AFTERLAST)
            return false;

        /**
         * have to do special, non-absolute processing in the case of a forward only
         * result set.
         */
        if (m_maxRows > 0)
        {
            /**
             * check to see if we might be going beyond the user specified
             * max rows.
             */
            if ((m_actual + 1) > m_maxRows)
            {
                m_actual = m_maxRows + 1;
                AFTERLAST = true;
                return false;
            }
        }
        int flags = 0;
        byte[] pos = null;
        int offset = 0;
       	/* S0206209: check for !m_noreads prior to going into the
       	 * read more stuff code.
       	 */
        if ((m_cursor < 0 && !m_noreads)|| m_cursor >= m_rows.size() - 1)
        {
            if (m_cursor > m_rows.size() - 1)
                m_cursor = m_rows.size() - 1;
            if (m_cursor < 0 && m_actual == 0)
                flags =
                    IDataSet.FirstRecord
                        | IDataSet.ReadMissing
                        | IDataSet.ReadBookmarks;
            else if (m_noreads)
            {
                AFTERLAST = true;
                success = false;
                return success;
            }
            else
            {
                flags = IDataSet.ReadMissing | IDataSet.ReadBookmarks;
                if (m_cursor > 0 && m_cursor < m_rows.size())
                    pos = ((MVARow) m_rows.get(m_cursor)).getBookmark();
                if (pos == null || pos.length == 0)
                    offset = 0;
                else
                    offset = 1;
            }
            try
            {

                if (m_rsType == ResultSet.TYPE_FORWARD_ONLY
                    && m_cacheChangeFlag)
                {
                    m_cacheSize = m_newCache;
                    m_newCache = 0;
                    m_cacheChangeFlag = false;
                }
                success =
                    readRecords(flags, m_bindKey, m_cacheSize, offset, pos);
            }
            catch (Exception e)
            {
                MVALocalizedError.throwLocalizedException(
                    VendorCode.MVAResultSet_nextError,
                    null,
                    e);
                return false;
            }
            if (success)
            {
                if (m_noreads)
                    m_cursor = (m_cursor + 1) % m_cacheSize;
                else
                    m_cursor = 0;
                m_actual++;
                BEFOREFIRST = false;
                AFTERLAST = false;
            }
            else
            {
                AFTERLAST = true;
                m_actual++;
                m_cursor++;
            }
        }
        else
        {
            m_cursor++;
            m_actual++;
            if (m_noreads)
            {
                if (m_actual <= m_cacheSize)
                {
                    success = true;
                    BEFOREFIRST = false;
                    AFTERLAST = false;
                }
            }
            else if (m_cursor < m_rows.size())
            {
                success = true;
                BEFOREFIRST = false;
                AFTERLAST = false;
            }
            else
            {
                AFTERLAST = true;
                BEFOREFIRST = false;
                success = false;
            }
        }

        return success;
    }
    /**
     * moves the cursor backwards in the cache. (called by rs.previous())
     * @return boolean success or failure
     */
    public boolean previous() throws SQLException
    {
        boolean success = false;
        moveToCurrentRow();
        int flags = 0;
        byte[] pos = null;
        int offset = 0;
        if (BEFOREFIRST)
            return false;
        if (m_cursor < 0 || m_actual == 0)
        {
            return false;
        }
        if (m_cursor == 0 && m_actual == 1)
        {
            BEFOREFIRST = true;
            m_actual--;
            m_cursor--;
            return false;
        }
        if (m_cursor == 0
            || (AFTERLAST && m_cursor <= 0)
            || (AFTERLAST && m_rows.size() == 0))
        {
            if (AFTERLAST && m_rows.size() == 0)
            {
                if (m_maxRows == 0)
                {
                    pos = null;
                    flags =
                        IDataSet.ReadMissing
                            | IDataSet.ReadBookmarks
                            | IDataSet.LastRecord;
                }
                else
                {
                    return absolute(-1);
                }
            }
            else if (AFTERLAST && m_cursor <= 0)
            {
                pos = null;
                flags =
                    IDataSet.ReadMissing
                        | IDataSet.ReadBookmarks
                        | IDataSet.LastRecord;
                offset = 0;
            }
            else
            {
                flags = IDataSet.ReadMissing | IDataSet.ReadBookmarks;
                pos = ((MVARow) m_rows.get(m_cursor)).getBookmark();
                offset = -1;
            }
            try
            {
                if (m_status == BOF)
                    success = false;
                else
                {
                    success =
                        readRecords(
                            flags,
                            m_bindKey,
                            -m_cacheSize,
                            offset,
                            pos);
                }
            }
            catch (Exception e)
            {
                MVALocalizedError.throwLocalizedException(
                    VendorCode.MVAResultSet_previousError,
                    null,
                    e);
                return false;
                // this won't get hit, pal
            }
            if (success)
            {
                m_cursor = m_rows.size() - 1;
                m_actual--;
                AFTERLAST = false;
            }
            else
            {
                BEFOREFIRST = true;
                m_actual--;
                m_cursor = -1; // ?? is this the right way?
            }
        }
        else
        {
            m_cursor--;
            m_actual--;
            if (m_actual > 0)
            {
                success = true;
                AFTERLAST = false;
            }
            else
            {
                success = false;
                BEFOREFIRST = true;
            }
        }
        /**
        if (AFTERLAST)
            success = absolute(-1);
        else
            success = relative(-1);
        */
        return success;
    }
    /**
     * move to an absolute position in the result set.
     * @param int destination row
     * @return boolean success or failure
     */
    public boolean absolute(int destinationRow) throws SQLException
    {
        moveToCurrentRow();
        boolean ableToCalc = true;
        int cacheStart = 0;
        int cacheEnd = 0;
        int move = 0;
        int offset = 0;
        int flags = 0;
        boolean success = false;
        if (m_maxRows > 0)
        {
            if (destinationRow > m_maxRows)
            {
                /**
                 * S0192835:
                 * move to the end of the resultset, just as if you had done
                 * a move to some point way off the resultset w/o having maxrows set.
                 */
                while (next());
                return false;
                /** end S0192835 */
            }
        }
        if (destinationRow == 0)
        {
            /* S0214665 */
            m_peekFlag = false;
            BEFOREFIRST = true;
            AFTERLAST = false;
            m_actual = 0;
            m_cursor = -1;
            return false;
        }
        if (destinationRow > 0)
        {
            if (BEFOREFIRST || AFTERLAST)
            {
                ableToCalc = false;
            }
            else
            {
                ableToCalc = true;
                cacheStart = m_actual - m_cursor;
                cacheEnd = cacheStart + m_rows.size() - 1;
                if (m_status == EOF && destinationRow > cacheEnd)
                {
                    /* S0214665 */
                    m_peekFlag = false;
                    AFTERLAST = true;
                    BEFOREFIRST = false;
                    m_actual = cacheEnd + 1;
                    return false;
                }
            }
            if (!ableToCalc
                || m_actual == 0
                || m_cursor < 0
                || (destinationRow < cacheStart || destinationRow > cacheEnd))
            {
                /**
                 * need to read over the wires
                 */
                move = destinationRow / m_cacheSize;
                offset = destinationRow % m_cacheSize;
                if (destinationRow < m_cacheSize)
                {
                    move = move * m_cacheSize;
                }
                else
                {
                    if (move > 0 && offset == 0)
                        move = (move - 1) * m_cacheSize;
                    else
                        move = move * m_cacheSize;
                }
                if (offset == 0)
                    offset = m_cacheSize;
                offset = offset - 1;
                flags =
                    IDataSet.FirstRecord
                        | IDataSet.ReadMissing
                        | IDataSet.ReadBookmarks;
                try
                {
                    success =
                        readRecords(flags, m_bindKey, m_cacheSize, move, null);
                }
                catch (Exception e)
                {
                    MVALocalizedError.throwLocalizedException(
                        VendorCode.MVAResultSet_absoluteRowError,
                        null, e);
                    return false;
                }
                if (success)
                {
                    /**
                     * we were able to read some rows, but not enough to move to the desired
                     * location
                     */
                    if (offset >= m_rows.size())
                    {
                        AFTERLAST = true;
                        success = false;
                        /**
                         * set the "actual" position to just past the end of the resultset
                         */
                        m_actual = move + m_rows.size() + 1;
                    }
                    else
                    {
                        m_cursor = offset;
                        m_actual = destinationRow;
                        BEFOREFIRST = AFTERLAST = false;
                    }
                    /* S0214665 */
                    m_peekFlag = false;
                }
                /**
                    else
                        AFTERLAST = true;
                 */
                else
                {
                    if (destinationRow > 0)
                    {
                        // need to move to the end of the resultset.
                        while (next());
                    }
                }
            }
            else
            {
                /**
                 * can stay in the cache
                 */
                /* S0214665 */
                m_peekFlag = false;
                m_cursor = destinationRow - cacheStart;
                m_actual = destinationRow;
                BEFOREFIRST = AFTERLAST = false;
                success = true;
            }
        }
        else
        {
            /**
             * destinationRow < 0, so moving backward
             */
            int cc = count(m_actual);
            /**
             * S0192835: if |destinationRow| > row count, then what was happening
             * was that the newpos was remaining negative, forcing another go round thru
             * here, and so on, until a positive number go calculated, which more than likely
             * would not be where you wanted to end up.
             * Ex: say you have 92 rows, and do abs(-1000);
             * |1000| mod 92 is not going to be 0, which is what it should be to get you at beforeFirst,
             * like it should.
             */
            if (Math.abs(destinationRow) > cc)
                success = absolute(0);
            /** end S0192835 */
            else
            {
                int newpos = (cc + 1) - Math.abs(destinationRow);
                success = absolute(newpos);
            }
        }
        return success;
    }
    /**
     * move to a position relative to the current position
     * @param int distance
     * @return boolean success/failure
     */
    public boolean relative(int distance) throws SQLException
    {
        moveToCurrentRow();
        if (AFTERLAST || BEFOREFIRST)
            return false;
        int newpos = m_actual + distance;
        /**
         * S0192825:
         * if you try to do a relative way off the resultset (like -1000 on a 11 row resultset)
         * you get a problem similar to the one in absolute when you attempt to move negatively
         * So, if calculated newpos is < 0, then set it to zero.  This will correctly put you before
         * first.
         */
        if (newpos < 0)
            newpos = 0;
        boolean success = false;
        success = absolute(newpos);
        return success;
    }
    /**
     * change teh data in the given column of the "update row" to be
     * the updateObj
     * @param int column
     * @param Object updateObj
     */
    /*
     * it;s the job of the caller to ensure that the ogbject is of the right
     * type for the column.  We're trying to keep stuff like that out of this
     * object.
     */
    public void updateData(int column, Object updateObj)
    {
        column -= 1;
        if (m_updateRow == null)
        {
            /** make a copy of the row the cursor is positioned on.
             */
            if (!INSERTROW && m_actual > 0 && m_cursor < m_rows.size())
                m_updateRow = new MVARow((MVARow) m_rows.get(m_cursor));
            else // updating a new row
                {
                m_updateRow = new MVARow(m_columnCount);
                for (int i = 0; i < m_columnCount; i++)
                {
                    switch (getColumnType(i + 1))
                    {
                        case Types.VARCHAR :
                            m_updateRow.add(i, null);
                            m_updateRow.setMissing(i, true);
                            break;
                        default :
                            m_updateRow.add(i, new Double(Double.NaN));
                            m_updateRow.setMissing(i, true);
                    }
                }
            }
        }
        /**
         * S0193089: if the updateObj passed in is null, then we check to see if it's a
         * varchar, and if so, then we set the updateObj to the empty string.
         */
        if (updateObj == null)
        {
            if (getColumnType(column + 1) == Types.VARCHAR)
                updateObj = "";
        }
        /** end S0193089 */
        /** easy peasy */
        if (column >= m_updateRow.size())
            m_updateRow.add(column, updateObj);
        else
            m_updateRow.setData(column, updateObj);
        /**
         * S0193089: after setting the data, check to see if the value was NaN, indicating
         * a missing value.  If so, set missing on the column.  This has to go here due to
         * a glitch in setData in MVARow.
         */
        if (updateObj instanceof Double)
        {
            if (Double.isNaN(((Double) updateObj).doubleValue()))
                m_updateRow.setMissing(column, true);
        }
        /** end S0193089 */
    }
    /**
     * abandon the update by nulling out the update row
     */
    public void cancelUpdate()
    {
        m_updateRow = null;
    }
    /**
     * close the result set
     */
    public void close() throws SQLException
    {
        try
        {
            if (m_xUtil != null)
                m_xUtil.close(m_hashCode);
            m_rows.clear();
        }
        catch (Exception e)
        {
            throw new SQLException(e.toString());
        }
    }
    /********************************************************************/
    /* protected methods:                                               */
    /********************************************************************/
    /**
     * JDBC 2.0
     *
     * Insert the contents of the insert row into the result set and
     * the database.  Must be on the insert row when this method is called.
     *
     * NOTE: SAS software does not support inserts, though inserts might be
     * implemented as append.
     * @exception SQLException This exception is always thrown.
     *
     * This function is available only on ResultSets generated from MVAStatement.getTable()
     */
    protected synchronized void insertRow() throws SQLException
    {
        if (!INSERTROW)
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_notOnInsertRowError);
        int x = getColumnCount();
        Object[] data = new Object[x];
        int[] columnTypes = new int[x];
        if (m_updateRow == null)
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_insertFailedNoDataError);
        for (int i = 0; i < x; i++)
        {
            columnTypes[i] = getColumnType(i + 1);
            data[i] = m_updateRow.get(i);
        }
        try
        {
            m_xUtil.insertRow(m_hashCode, data, m_bindKey, columnTypes);
        }
        catch (Exception se)
        {
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_insertFailedError,
                null,
                se);
        }
        // m_updateRow = null;
        m_countValid = false;
        /**
         * S0190257: call refresh() after successfully inserting the row
         */
        if (!BEFOREFIRST && !AFTERLAST)
        {
            if (m_status == EOF)
                refresh();
        }
    }
    /**
    	 * JDBC 2.0
    	 *
     * Update the underlying database with the new contents of the
     * current row.  Cannot be called when on the insert row.
     *
     * @exception SQLException if a database-access error occurs, or
     * if called when on the insert row
     *
     * This function is available only on ResultSets generated from MVAStatement.getTable()
     */
    protected synchronized void updateRow() throws SQLException
    {
        /**
         * S0195329: create a columntypes array to pass over to update.
         */
        int[] columnTypes = new int[getColumnCount()];
        /** S0195329 */
        if (INSERTROW)
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_insertRowUpdateError);
        Object[] data = new Object[getColumnCount()];
        if (m_updateRow == null) // no - op
            return;
        for (int i = 0; i < getColumnCount(); i++)
        {
            /**
             * S0195329: get the column type and store in the array
             */
            columnTypes[i] = getColumnType(i + 1);
            /** S0195329 */
            data[i] = m_updateRow.get(i);
            /**
             * fix for S0193089 (AND S0194092)
             * when you update, I wasn't checking the missing value status when sending
             * stuff over to the server.  This made the server update a missing value with
             * a zero. This showed up during final sanity check testing.
             */
            /**
             * S0195329: check the column type, rather than type in data[][]
             */
            if (columnTypes[i] != Types.VARCHAR)                /** S0195329 */
            {
                if (m_updateRow.getMissing(i))
                    data[i] = new Double(Double.NaN);
            }
        }
        byte[] bookmarksUpdate = null;
        MVARow f = (MVARow) m_rows.get(m_cursor);
        bookmarksUpdate = f.getBookmark();
        try
        {
            m_xUtil.updateRow(
                m_hashCode,
                data,
                bookmarksUpdate,
                m_bindKey,
                columnTypes);
        }
        catch (Exception e)
        {
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_updateFailed,
                null,
                e);
        }
        m_updateRow.setBookmark(bookmarksUpdate);
        /**
         * S0191832: When I set nulls on numerics, I put a
         * Double.NaN in the updateRow. (ArrayLists really don't like it when you try to set
         * a null).  Unfortunately, I'm not updating it properly.
         * This means that when you update, instead of getting 0 and wasNull = true, you
         * get NaN and wasNUll = false.  Not good.
         */
        for (int xx = 0; xx < getColumnCount(); xx++)
        {
            /**
             * S0195329: again, use teh columnTypes to check the type
             */
            if (columnTypes[xx] != Types.VARCHAR)
            {
                /**
                 * S0195329:
                 * with the possibility of formatted data coming thru here, you have to
                 * check the type, since a missing numeric value when formatted comes back
                 * as a blank, and thus there's no indication that it's missing.
                 */
                double tdv;
                if (m_bindKey == BindKeys._BindKeysFormatted)
                {
                    String tv = data[xx].toString();
                    if (tv.length() == 0 || tv.trim().equals(""))
                        tdv = Double.NaN;
                    else
                        tdv = Double.parseDouble(tv);
                }
                else
                {
                    tdv = ((Double) data[xx]).doubleValue();
                }
                /** S0195329 */
                if (Double.isNaN(tdv))
                {
                    m_updateRow.setData(xx, new Double(0.0));
                    m_updateRow.setMissing(xx, true);
                }
            }
            /**
             * do the same for null, so that the behavior is right for both
             */
            /** S0195329: again, check the column type instead of the object type */
            else if (columnTypes[xx] == Types.VARCHAR && data[xx] == null)                /** S0195329 */
            {
                m_updateRow.setData(xx, "");
            }
        }
        /** End Fix */
        m_rows.set(m_cursor, m_updateRow);
        m_updateRow = null;
    }
    protected void refresh() throws SQLException
    {
        /**
         * get the bookmark that starts the cache.
         */
        if (BEFOREFIRST || AFTERLAST) //barf
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_invalidRowPosition);
        MVARow r = (MVARow) m_rows.get(0);
        int offset = 0;
        byte[] startBookmark = r.getBookmark();
        int flags = IDataSet.ReadMissing | IDataSet.ReadBookmarks;
        boolean success = false;
        try
        {
            success =
                readRecords(
                    flags,
                    m_bindKey,
                    m_cacheSize,
                    offset,
                    startBookmark);
        }
        catch (Exception e)
        {
            MVALocalizedError.throwLocalizedException(VendorCode.MVAResultSet_refreshRowError,
                                                        null, e);
        }
        if (success)
            m_updateRow = null;
    }
    protected void refresh(byte[] bookmark) throws SQLException
    {
        int offset = 0;
        int flags = IDataSet.ReadMissing | IDataSet.ReadBookmarks;
        boolean success = false;
        //try
        //{
        success = readRecords(flags, m_bindKey, m_cacheSize, offset, bookmark);
        //}
        //catch (Exception e)
        //{
        //}
        if (success)
        {
            m_updateRow = null;
            // MVARow x = (MVARow) m_rows.get(m_cursor);
        }
    }
    /**
     * uses the 2d object array passed in to populate the cache.
     * this is used by the metadata result sets to populate the cache
     * with hard coded data (like types, etc)
     * @param repos 2d array of Objects
     */
    protected void replaceData(Object[][] repo)
    {
        m_cacheSize = repo.length;
        m_rows = new Cache(repo[0].length);
        for (int i = 0; i < repo.length; i++)
        {
            MVARow r = new MVARow(repo[0].length);
            for (int j = 0; j < repo[0].length; j++)
                r.add(repo[i][j]);
            m_rows.add(r);
        }
        m_noreads = true;
    }
    /**
     * JDBC 2.0
     *
     * Delete the current row from the result set and the underlying
     * database.  Cannot be called when on the insert row.
     *
     * @exception SQLException if a database-access error occurs, or if
     * called when on the insert row.
     */
    protected synchronized void deleteRow() throws SQLException
    {
        if (BEFOREFIRST || AFTERLAST)
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_invalidRowPosition);
        byte[] bookmarksDelete = null;
        try
        {
            MVARow r = (MVARow) m_rows.get(m_cursor);
            bookmarksDelete = r.getBookmark();
            m_xUtil.deleteRow(m_hashCode, bookmarksDelete);
            // ok, now update the cache by deleting the row the cursor is on
            // do *NOT* update the cursor. (it drops down to the next row in the RS)
            // this way, if you're on the last row, then you've just walked off the rs,
            // and you're still safe.
            //blocknum = fetchSize - 1;
            m_rows.remove(m_cursor);
            if (m_cursor == m_rows.size()) // need to tweak;
            {
                m_cursor--;
                m_actual--;
                next();
            }
            m_deleteFlag = true;
            m_countValid = false;
        }
        catch (Exception e)
        {
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_deleteFailed,
                null,
                e);
        }
    }
    /********************************************************************/
    /* private methods:                                                 */
    /********************************************************************/
    private int getActualType(int column)
    {
        if (actualList == null)
            return -1;
        int i = 0;
        ArrayList p = (ArrayList) actualList.get(i++);
        column -= 1;
        int c = 0;
        while (p != null && i < actualList.size())
        {
            c = ((Integer) p.get(0)).intValue();
            if (c == column)
            {
                return ((Integer) p.get(1)).intValue();
            }
            p = (ArrayList) actualList.get(i++);
        }
        return -1;
    }
    private boolean peek() throws SQLException
    {
        /**
         * just reads 1st row, and returns success/failure.
         * and sets a peek flag as being true if it's not empty,
         * to indicate that the next forward move should not really
         * move.
         */
        m_peekFlag = next();
        return m_peekFlag;
    }
    /**
     * count - a private helper method to count the rows in the result set.
     * unfortunately, you have to do something like this when you start doing
     * absolute positioning.
     */
    private int count(int prior) throws SQLException
    {
        if (!m_isDirect && m_alreadyCounted)
            return m_rowCount;
        else if (m_isDirect)
        {
            if (m_alreadyCounted && m_countValid)
                return m_rowCount;
            int lrc = 0;
            try
            {
                lrc = m_xUtil.countRecords(m_hashCode);
                m_rowCount = lrc;
            }
            catch (Exception e)
            {
                MVALocalizedError.throwLocalizedException(
                    VendorCode.MVAStatement_executeError, null, e);
            }
            m_alreadyCounted = true;
            m_countValid = true;
            /** S0184797 fix */
            if (m_maxRows > 0)
            {
                if (m_rowCount > m_maxRows)
                    m_rowCount = m_maxRows;
            }
            return m_rowCount;
        }
        else
        {
            if (m_alreadyCounted && m_countValid)
                return m_rowCount;
            int lrc = 0;
            try
            {
                lrc = m_xUtil.countRecords(m_hashCode);
                m_rowCount = lrc;
            }
            catch (Exception e)
            {
                /**
                 * if we get an exception, it's possible that it's because
                 * the SAS server is not the most current.  Therefore, the
                 * proper behavior is to attempt to brute force the count
                 */
                try
                {
                    while (next());
                    m_rowCount = m_actual - 1; //-1;??
                    absolute(prior);
                    if (m_rowCount < 0)
                        m_rowCount = 0;
                    m_alreadyCounted = true;
                    return m_rowCount;
                }
                catch (Exception ee)
                {
                    MVALocalizedError.throwLocalizedException(
                        VendorCode.MVAStatement_executeError, null, ee);
                }
            }
            m_alreadyCounted = true;
            m_countValid = true;
            /** S0184797 fix */
            if (m_maxRows > 0)
            {
                if (m_rowCount > m_maxRows)
                    m_rowCount = m_maxRows;
            }
            return m_rowCount;
        }
        /**
                if (AFTERLAST)
                {
                    m_rowCount = m_actual - 1;
                    return m_rowCount;
                }
        */
        /**
         * fix for S0127940: removed count code below, and remove the "is (m_isDirect)
         * check above, so that both direct and sql tables get counted the same way.
         */
        /**
        while (next());
        m_rowCount = m_actual - 1; //-1;??
        absolute(prior);
        if (m_rowCount < 0)
            m_rowCount = 0;
        m_alreadyCounted = true;
        return m_rowCount;
        */
    }
    /**
     * init gets the column information and other relevant data set metadata.
     * returns void.
     * throws SQLException, in case anything bad happens.
     */
    private void init() throws SQLException
    {
        int numberColumns = 0;
        String[] columnNames = null;
        int[] columnTypes = null;
        Object[][] hack = null;
        try
        {
            hack = m_xUtil.getColumnInfo(m_hashCode);
        }
        catch (Exception ee)
        {
            // rethrow.
            throw new SQLException(ee.toString());
        }
        /**
         * now get the stuff in the holders, and assign it to the
         * correct variables.
         */
        if (hack == null)
            return;
        numberColumns = hack[0].length;
        m_colt = new int[numberColumns];
        columnNames = new String[numberColumns];
        columnTypes = new int[numberColumns];
        m_columnLengths = new int[numberColumns];
        m_columnLabels = new String[numberColumns];
        m_formatNames = new String[numberColumns];
        m_formatWidths = new short[numberColumns];
        m_formatDecimals = new short[numberColumns];
        m_formatLengths = new short[numberColumns];
        m_informatNames = new String[numberColumns];
        m_informatWidths = new short[numberColumns];
        m_informatDecimals = new short[numberColumns];
        int i = 0;
        for (i = 0; i < numberColumns; i++)
        {
            /** names */
            columnNames[i] = (String) hack[0][i];
            /** types */
            columnTypes[i] = ((Integer) hack[1][i]).intValue();
            /** lengths */
            m_columnLengths[i] = ((Integer) hack[2][i]).intValue();
            /** labels */
            m_columnLabels[i] = (String) hack[3][i];
            /** format names */
            m_formatNames[i] = (String) hack[4][i];
            /** format widths */
            m_formatWidths[i] = ((Short) hack[5][i]).shortValue();
            /** format decimals */
            m_formatDecimals[i] = ((Short) hack[6][i]).shortValue();
            /** format lengths */
            m_formatLengths[i] = ((Short) hack[7][i]).shortValue();
            /** informat names */
            m_informatNames[i] = (String) hack[8][i];
            /** informat widths */
            m_informatWidths[i] = ((Short) hack[9][i]).shortValue();
            /** informat decimals */
            m_informatDecimals[i] = ((Short) hack[10][i]).shortValue();
        }
        /**
         * instantiate the types and column names
         */
        m_columnCount = columnTypes.length;
        m_columnTypes = new ArrayList(m_columnCount);
        m_columnNames = new ArrayList(columnNames.length);
        /**
         * stash the column names in the m_columnNames arraylist.
         * I have no idea why the previous author was hashing them.
         */
        for (int a = 0; a < columnNames.length; a++)
        {
            setColumnName(a + 1, columnNames[a].trim().toUpperCase());
        }
        /**
         * derive the format by using helper class Sqpformat
         */
        Sqpformat[] sqpformat = new Sqpformat[m_columnCount];
        int[] columnSubTypes = new int[m_columnCount];
        for (i = 0; i < numberColumns; i++)
        {
            sqpformat[i] =
                new Sqpformat(
                    m_formatNames[i],
                    m_formatLengths[i],
                    m_formatWidths[i],
                    m_formatDecimals[i]);
            /**
             * figure out the sql type of each column, and stash with
             * setColumnType
             */
            if (columnTypes[i] == BaseType._BaseTypeNumeric)
            {
                columnSubTypes[i] = sqpformat[i].findFormat(m_formatNames[i]);
                /*
                 * S0278967/S0260901:
                 * numericTypeCoercion is a boolean that determines whether "normal"
                 * typing occurs, or the "fix" typing occurs.
                 */
                if (m_numericTypeCoercion)
                {
	                switch (columnSubTypes[i])
	                {
	                    case DATE :
	                        setColumnType(i + 1, Types.DATE);
	                        break;
	                    case TIME :
	                        setColumnType(i + 1, Types.TIME);
	                        break;
	                    case DATETIME :
	                        setColumnType(i + 1, Types.TIMESTAMP);
	                        break;
	                    case INTEGER :
	                        setColumnType(i + 1, Types.INTEGER);
	                        break;
	                    default :
	                        setColumnType(i + 1, Types.DOUBLE);
	                        break;
	                }
                }
                else
                {
                    /*
                     * S0278967/S0260901:
                     * "fix" is that unless the type comes back as date/time/datetime,
                     * we just default to double.
                     */
                    switch (columnSubTypes[i])
	                {
	                    case DATE :
	                        setColumnType(i + 1, Types.DATE);
	                        break;
	                    case TIME :
	                        setColumnType(i + 1, Types.TIME);
	                        break;
	                    case DATETIME :
	                        setColumnType(i + 1, Types.TIMESTAMP);
	                        break;
	                    default :
	                        setColumnType(i + 1, Types.DOUBLE);
	                        break;
                }
                }
            }
            else
            {
                setColumnType(i + 1, Types.VARCHAR);
            }
            m_colt[i] = getColumnType(i + 1);
        }
    }
    /**
    	 * Opens a SAS dataset
    	 */
    private void openDataSet(
        MVAStatement stmt,
        String tableName,
        String[] password)
        throws SQLException
    {
        m_isDirect = true;
        if (m_xUtil == null)
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_openDataSetError);
        try
        {
            m_hashCode =
                m_xUtil.openDataSet(
                    stmt.libref,
                    tableName,
                    password,
                    stmt.options,
                    stmt.getResultSetConcurrency(),
                    stmt.getResultSetType());
        }
        catch (Exception e)
        {
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_openDataSetError,
                null,
                e);
        }
    }
    /**
     * readRecords - wrapper function around dataset method ReadRecords
     * @params - flags, bindKey(unused right now), rows to read, offset
     * @returns boolean - true/success, false/failure
     */
    private boolean readRecords(
        int flags,
        int bindKey,
        int rows,
        int rowsOffset,
        byte[] pos)
        throws SQLException
    {
        boolean forward = true;
        if (rows < 0)
            forward = false;
        if (m_xUtil == null)
            return false;
        // colt == an array containing the column types
        /**
         * PERF HINT #1: move this block to init, since we don't have binding, there's no reason
         * to do this over and over again.
         */
        /*
                  for (int i = 0; i < colt.length; i++)
                  {
                  colt[i] = getColumnType(i + 1);
                  // check for null ness.
                  // fix for S0183030 (attempting to read data from the server that
                  // has a null type designator can be problematic, since you need to be able
                  // to convert it to a "real" type to figure out if the data is in the
                  // character array or the numeric array.
                  if (colt[i] == Types.NULL)
                  colt[i] = getActualType(i + 1);
                  }
        */
        flags |= NumericMissingUnderscore.value;
        DataHolder data = null;
        try
        {
            //            bookmarkArray = new ArrayList();
            data =
                m_xUtil.readRecords(
                    m_hashCode,
                    flags,
                    bindKey,
                    rows,
                    rowsOffset,
                    pos,
                    m_colt);
            //   bookmarkArray);
        }
        catch (Exception e)
        {
            // rethrow
            MVALocalizedError.throwLocalizedException(
                VendorCode.MVAResultSet_readRecordsFailed,
                null,
                e);
        }
        boolean aok = false;
        if (data != null)
            aok = true;
        if (aok)
        {
            if (data.characterValues.length
                == 0 & data.numericValues.length
                == 0)
                return false;
            m_status = data.status;
            byte[] bookmarks = null;
            bookmarks = data.bookmarkArray;
            addRows(
                data.characterValues,
                data.numericValues,
                data.missingNumericValues,
                bookmarks,
                forward);
        }
        if (aok)
            return true;
        else
            return false;
    }
    /**
     * addRows takes the data obtained from readRecords, and stores it properly
     * in the cache
     * @param characterValues 2d String array
     * @param numericValues 2d double array
     * @param missingNumericValues 2d byte array, required since a missing double is
     *         scored as a zero.
     * @param bookmarks byte array noting the position of each record
     */
    private void addRows(
        String[][] characterValues,
        double[][] numericValues,
        byte[][] missingNumericValues,
        byte[] bookmarks,
        boolean direction)
    {
        int cindex = 0;
        int nindex = 0;
        m_rows.clear();
        int blen = 0;
        if (m_bookmarkLength == 0)
            blen = bookmarks.length / characterValues.length;
        else
            blen = m_bookmarkLength;
        if (direction)
        {
            for (int i = 0; i < characterValues.length; i++)
            {
                MVARow row = new MVARow(m_columnCount);
                int s = i * blen;
                int e = s + blen;
                byte[] b = new byte[blen];
                int k = 0;
                int j = s;
                for (j = s, k = 0; j < e; j++, k++)
                {
                    b[k] = bookmarks[j];
                }
                row.setBookmark(b);
                cindex = nindex = 0;
                if (m_bindKey == BindKeys._BindKeysUnformatted)
                {
                    for (k = 0; k < m_columnCount; k++)
                    {
                        switch (m_colt[k])
                        {
                            case Types.VARCHAR :
                                row.add(k, characterValues[i][cindex++]);
                                row.setMissing(k, false);
                                break;
                            case Types.NULL :
                                int at = getActualType(k + 1);
                                switch (at)
                                {
                                    case Types.VARCHAR :
                                        row.add(
                                            k,
                                            characterValues[i][cindex++]);
                                        row.setMissing(k, false);
                                        break;
                                    default :
                                        row.add(
                                            k,
                                            new Double(numericValues[i][nindex]));
                                        if (missingNumericValues[i][nindex]
                                            == NotMissing.value)
                                            row.setMissing(k, false);
                                        else
                                            row.setMissing(k, true);
                                        nindex++;
                                        break;
                                }
                                break;
                            default :
                                row.add(
                                    k,
                                    new Double(numericValues[i][nindex]));
                                if (missingNumericValues[i][nindex]
                                    == NotMissing.value)
                                    row.setMissing(k, false);
                                else
                                    row.setMissing(k, true);
                                nindex++;
                                break;
                        }
                    }
                }
                else // bindkeys._bindkeysformatted
                    {
                    for (k = 0; k < m_columnCount; k++)
                    {
                        switch (m_colt[k])
                        {
                            case Types.VARCHAR :
                                row.add(k, characterValues[i][cindex++]);
                                row.setMissing(k, false);
                                break;
                            default :
                                row.add(k, characterValues[i][cindex++]);
                                if (missingNumericValues[i][nindex]
                                    == NotMissing.value)
                                    row.setMissing(k, false);
                                else
                                    row.setMissing(k, true);
                                nindex++;
                        }
                    }
                }
                m_rows.add(row);
            }
        }
        else
        {
            int z = 0;
            int i = 0;
            for (i = characterValues.length - 1, z = 0; i >= 0; i--, z++)
            {
                MVARow row = new MVARow(m_columnCount);
                int s = i * blen; //z
                int e = s + blen;
                byte[] b = new byte[blen];
                int k = 0;
                int j = s;
                for (j = s, k = 0; j < e; j++, k++)
                {
                    b[k] = bookmarks[j];
                }
                row.setBookmark(b);
                cindex = nindex = 0;
                if (m_bindKey == BindKeys._BindKeysUnformatted)
                {
                    for (k = 0; k < m_columnCount; k++)
                    {
                        switch (m_colt[k])
                        {
                            case Types.VARCHAR :
                                row.add(k, characterValues[i][cindex++]);
                                row.setMissing(k, false);
                                break;
                            case Types.NULL :
                                int at = getActualType(k + 1);
                                switch (at)
                                {
                                    case Types.VARCHAR :
                                        row.add(
                                            k,
                                            characterValues[i][cindex++]);
                                        row.setMissing(k, false);
                                        break;
                                    default :
                                        row.add(
                                            k,
                                            new Double(numericValues[i][nindex]));
                                        if (missingNumericValues[i][nindex]
                                            == NotMissing.value)
                                            row.setMissing(k, false);
                                        else
                                            row.setMissing(k, true);
                                        nindex++;
                                        break;
                                }
                                break;
                            default :
                                row.add(
                                    k,
                                    new Double(numericValues[i][nindex]));
                                if (missingNumericValues[i][nindex]
                                    == NotMissing.value)
                                    row.setMissing(k, false);
                                else
                                    row.setMissing(k, true);
                                nindex++;
                                break;
                        }
                    }
                }
                else // bindkeys._bindkeysformatted
                    {
                    for (k = 0; k < m_columnCount; k++)
                    {
                        switch (m_colt[k])
                        {
                            case Types.VARCHAR :
                                row.add(k, characterValues[i][cindex++]);
                                row.setMissing(k, false);
                                break;
                            default :
                                row.add(k, characterValues[i][cindex++]);
                                if (missingNumericValues[i][nindex]
                                    == NotMissing.value)
                                    row.setMissing(k, false);
                                else
                                    row.setMissing(k, true);
                                nindex++;
                        }
                    }
                }
                m_rows.add(row);
            }
        }
    }
    /**
     * rmi version of addrows.
     * S0177195 - changed all checks of <number> == Double.NaN to
     * Double.isNaN(<number>)
     */
    private void addRows(Object[][] data, byte[] bookmarks, boolean direction)
    {
        m_rows.clear();
        int blen = 0;
        if (m_bookmarkLength == 0)
            blen = bookmarks.length / data.length;
        else
            blen = m_bookmarkLength;
        if (direction)
        {
            for (int i = 0; i < data.length; i++)
            {
                MVARow row = new MVARow(m_columnCount);
                int s = i * blen;
                int e = s + blen;
                byte[] b = new byte[blen];
                int k = 0;
                int j = s;
                for (j = s, k = 0; j < e; j++, k++)
                {
                    b[k] = bookmarks[j];
                }
                row.setBookmark(b);
                if (m_bindKey == BindKeys._BindKeysUnformatted)
                {
                    for (j = 0; j < m_columnCount; j++)
                    {
                        switch (m_colt[j])
                        {
                            case Types.VARCHAR :
                                row.add(j, data[i][j]);
                                row.setMissing(j, false);
                                break;
                            case Types.NULL :
                                int at = getActualType(j + 1);
                                switch (at)
                                {
                                    case Types.VARCHAR :
                                        row.add(j, data[i][j]);
                                        row.setMissing(j, false);
                                        break;
                                    default :
                                        double g =
                                            ((Double) data[i][j]).doubleValue();
                                        if (Double.isNaN(g))
                                        {
                                            Double v = new Double(0);
                                            row.add(j, v);
                                            row.setMissing(j, true);
                                        }
                                        else
                                        {
                                            row.add(j, data[i][j]);
                                            row.setMissing(j, false);
                                        }
                                        break;
                                }
                                break;
                            default :
                                double v = ((Double) data[i][j]).doubleValue();
                                if (Double.isNaN(v))
                                {
                                    Double d = new Double(0);
                                    row.add(j, d);
                                    row.setMissing(j, true);
                                }
                                else
                                {
                                    row.add(j, data[i][j]);
                                    row.setMissing(j, false);
                                }
                        }
                    }
                }
                else // bindkeys._bindkeysformatted
                    {
                    for (j = 0; j < m_columnCount; j++)
                    {
                        switch (m_colt[j])
                        {
                            case Types.VARCHAR :
                                row.add(j, data[i][j]);
                                row.setMissing(j, false);
                                break;
                            default :
                                Object oo = data[i][j];
                                if (oo instanceof Double)
                                {
                                    double dd = ((Double) oo).doubleValue();
                                    if (Double.isNaN(dd))
                                    {
                                        row.add(j, "");
                                        row.setMissing(j, true);
                                    }
                                    else
                                    {
                                        row.add(j, data[i][j]);
                                        row.setMissing(j, false);
                                    }
                                }
                                else
                                {
                                    row.add(j, data[i][j]);
                                    row.setMissing(j, false);
                                }
                        }
                    }
                }
                m_rows.add(row);
            } // end outer for-loop
        }
        else
        {
            for (int i = (data.length - 1); i >= 0; i--)
            {
                MVARow row = new MVARow(m_columnCount);
                int s = i * blen;
                int e = s + blen;
                byte[] b = new byte[blen];
                int k = 0;
                int j = s;
                for (j = s, k = 0; j < e; j++, k++)
                {
                    b[k] = bookmarks[j];
                }
                row.setBookmark(b);
                if (m_bindKey == BindKeys._BindKeysUnformatted)
                {
                    for (j = 0; j < m_columnCount; j++)
                    {
                        switch (m_colt[j])
                        {
                            case Types.VARCHAR :
                                row.add(j, data[i][j]);
                                row.setMissing(j, false);
                                break;
                            case Types.NULL :
                                int at = getActualType(j + 1);
                                switch (at)
                                {
                                    case Types.VARCHAR :
                                        row.add(j, data[i][j]);
                                        row.setMissing(j, false);
                                        break;
                                    default :
                                        double g =
                                            ((Double) data[i][j]).doubleValue();
                                        if (Double.isNaN(g))
                                        {
                                            Double v = new Double(0);
                                            row.add(j, v);
                                            row.setMissing(j, true);
                                        }
                                        else
                                        {
                                            row.add(j, data[i][j]);
                                            row.setMissing(j, false);
                                        }
                                        break;
                                }
                                break;
                            default :
                                double v = ((Double) data[i][j]).doubleValue();
                                if (Double.isNaN(v))
                                {
                                    Double d = new Double(0);
                                    row.add(j, d);
                                    row.setMissing(j, true);
                                }
                                else
                                {
                                    row.add(j, data[i][j]);
                                    row.setMissing(j, false);
                                }
                        }
                    }
                }
                else // formatted...
                    {
                    for (j = 0; j < m_columnCount; j++)
                    {
                        switch (m_colt[j])
                        {
                            case Types.VARCHAR :
                                row.add(j, data[i][j]);
                                row.setMissing(j, false);
                                break;
                            default :
                                Object oo = data[i][j];
                                if (oo instanceof Double)
                                {
                                    double dd = ((Double) oo).doubleValue();
                                    if (Double.isNaN(dd))
                                    {
                                        row.add(j, "");
                                        row.setMissing(j, true);
                                    }
                                    else
                                    {
                                        row.add(j, data[i][j]);
                                        row.setMissing(j, false);
                                    }
                                }
                                else
                                {
                                    row.add(j, data[i][j]);
                                    row.setMissing(j, false);
                                }
                        }
                    }
                }
                m_rows.add(row);
            } // end outer for-loop
        }
    }
}
