package com.sas.security;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.io.InputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.security.AccessControlContext;
import java.security.AccessControlException;
import java.security.AccessController;
import java.security.CodeSource;
import java.security.Permission;
import java.security.PermissionCollection;
import java.security.Permissions;
import java.security.Policy;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.security.ProtectionDomain;
import java.security.SecurityPermission;

/**
 * A security support class that runs under the Java 2 JVM (JDK 1.2).
 * This class is loaded by {@link com.sas.security.SecuritySupport}
 * when the java.vendor and java.version match the spec in SecurityClasses.properties
 *
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 */
class BaseJDK_1_2_SecuritySupport extends BaseSecuritySupport
{
  /**
   * Return a stream, as if calling <code>aClass.getResourceAsStream(resourceName)</code>.
   * Uses {@link java.security.AccessController#doPrivileged(java.security.PrivilegedAction)}.
   * See {@link java.lang.Class#getResourceAsStream(String)}
   */
  public InputStream getResourceAsStream(final Class aClass, final String resourceName)
  {
    return (InputStream) AccessController.doPrivileged( new PrivilegedAction()
      {
        public Object run()
        {
          // System.out.println("\n\t***\n\tJDK_1_2_SecuritySupport getResourceAsStream\n\t***\n\n");
          return aClass.getResourceAsStream(resourceName);
        }
      });
  }

  /*
   * Return a URL, as if calling <code>aClass.getResource(resourceName)</code>.
   * Uses {@link java.security.AccessController#doPrivileged(java.security.PrivilegedAction)}.
   * See {@link java.lang.Class#getResource(String)}
   */
  public URL getResource(final Class aClass, final String resourceName)
  {
    return (URL) AccessController.doPrivileged( new PrivilegedAction()
      {
        public Object run()
        {
          // System.out.println("\n\t***\n\tJDK_1_2_SecuritySupport getResourceAsStream\n\t***\n\n");
          return aClass.getResource(resourceName);
        }
      });
  }

	public Thread createThread(final Runnable p)
	{
		return (Thread)doThreadPrivileged(new java.security.PrivilegedAction(){
			public Object run()
			{
				return new Thread(p);
			}
		});
	}

	private static RuntimePermission modifyThreadPermission = null;
	private static RuntimePermission modifyThreadGroupPermission = null;

	public Object doThreadPrivileged(java.security.PrivilegedAction p)
	{
		// initialize the permission objects used in this method, if necessary
		synchronized (getClass())
		{
			if (modifyThreadPermission == null)
			{
				modifyThreadPermission = new RuntimePermission("modifyThread");
			}

			if (modifyThreadGroupPermission == null)
			{
				modifyThreadGroupPermission = new RuntimePermission("modifyThreadGroup");
			}
		}

		// check to see if this code already has the thread permissions
		boolean hasThreadPermissions = true;

		try
		{
			java.security.AccessController.checkPermission(modifyThreadPermission);
			java.security.AccessController.checkPermission(modifyThreadGroupPermission);
		}
		catch (java.security.AccessControlException ace)
		{
			hasThreadPermissions = false;
		}

		// if this code already has the thread permissions, then
		// there is no need to modify the access control context
		if (hasThreadPermissions)
		{
			Object obj = p.run();
			return obj;
		}

		// add the thread permissions to this code on the fly
		CodeSource codeSource = new CodeSource(null,null);
		Permissions newPermissions = new Permissions();
		newPermissions.add(modifyThreadGroupPermission);
		newPermissions.add(modifyThreadPermission);
		ProtectionDomain protectionDomain = new ProtectionDomain(codeSource,newPermissions);
		final ProtectionDomain[] domains = {protectionDomain};

		// get the current context
		final AccessControlContext acc = AccessController.getContext();
//		System.out.println("context: "+acc);
		java.security.AccessControlContext combined = acc;
		try
		{
			Class domainCombinerClass = Class.forName( "java.security.DomainCombiner" );
			// Create a new context that combines both contexts together
			combined = (java.security.AccessControlContext)java.security.AccessController.doPrivileged(new java.security.PrivilegedAction () 
			{
				public Object run()
				{
					// Create a context with just the new permissions
					return new AccessControlContext(acc, new PermissionCombiner(domains));
				}
			});
		}
		catch( ClassNotFoundException e )
		{
//			System.out.println( "java.security.DomainCombiner not found" );
		}

		return java.security.AccessController.doPrivileged(p, combined) ;
	}

	//***********************************************************************
	// Reflection method wrappers
	//***********************************************************************

	public Object invokeMethod( final Method method, final Object object, final Object[] args ) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException
	{
		try
		{
			return AccessController.doPrivileged( new PrivilegedExceptionAction()
				{
					public Object run() throws IllegalAccessException, IllegalArgumentException, InvocationTargetException
					{
						return method.invoke( object, args );
					}
				}
			);
		}
		catch (PrivilegedActionException  pae)
		{
//Thread.dumpStack();
			if ( pae.getException() instanceof IllegalAccessException )
				throw (IllegalAccessException) pae.getException();
			else if ( pae.getException() instanceof IllegalArgumentException )
				throw (IllegalArgumentException) pae.getException();
			else
				throw (InvocationTargetException) pae.getException();
		}
	}

	public Method getMethod( final Class cls, final String name, final Class[] paramTypes ) throws NoSuchMethodException, SecurityException
	{
		try
		{
			return (Method) AccessController.doPrivileged( new PrivilegedExceptionAction()
				{
					public Object run() throws NoSuchMethodException, SecurityException
					{
						return cls.getMethod( name, paramTypes );
					}
				}
			);
		}
		catch (PrivilegedActionException  pae)
		{
//pae.printStackTrace();
//Thread.dumpStack();
			if ( pae.getException() instanceof NoSuchMethodException )
				throw (NoSuchMethodException) pae.getException();
			else
				throw (SecurityException) pae.getException();
		}
//		return cls.getMethod( name, paramTypes );
	}

	public Method[] getDeclaredMethods( final Class cls ) throws SecurityException
	{
		try
		{
			return (Method[]) AccessController.doPrivileged( new PrivilegedExceptionAction()
				{
					public Object run() throws SecurityException
					{
						return cls.getDeclaredMethods();
					}
				}
			);
		}
		catch (PrivilegedActionException  pae)
		{
//Thread.dumpStack();
			throw (SecurityException) pae.getException();
		}
//	    return cls.getDeclaredMethods();
	}

	public Constructor getConstructor( final Class c, final Class[] args ) throws NoSuchMethodException
	{
		try
		{
			return (Constructor) AccessController.doPrivileged( new PrivilegedExceptionAction()
				{
					public Object run() throws NoSuchMethodException
					{
						return c.getConstructor(args);
					}
				}
			);
		}
		catch (PrivilegedActionException  pae)
		{
//Thread.dumpStack();
			throw (NoSuchMethodException) pae.getException();
		}
//		return c.getConstructor(args);
	}

	public int getMethodModifiers( final Member method )
	{
		Integer modifiers = (Integer) AccessController.doPrivileged( new PrivilegedAction()
			{
				public Object run()
				{
					return new Integer( method.getModifiers() );
				}
			}
		);
		return modifiers.intValue();
//		return method.getModifiers();
	}

	public String getMethodName( final Method method )
	{
		return (String) AccessController.doPrivileged( new PrivilegedAction()
			{
				public Object run()
				{
					return method.getName();
				}
			}
		);
//		return method.getName();
	}

	public Class[] getMethodParameterTypes( final Method method )
	{
		return (Class[]) AccessController.doPrivileged( new PrivilegedAction()
			{
				public Object run()
				{
					return method.getParameterTypes();
				}
			}
		);
//		return method.getParameterTypes();
	}

	public Class getMethodReturnType( final Method method )
	{
		return (Class) AccessController.doPrivileged( new PrivilegedAction()
			{
				public Object run()
				{
					return method.getReturnType();
				}
			}
		);
//		return method.getReturnType();
	}

	public Class[] getMethodExceptionTypes( final Method method )
	{
		return (Class[]) AccessController.doPrivileged( new PrivilegedAction()
			{
				public Object run()
				{
					return method.getExceptionTypes();
				}
			}
		);
//		return method.getExceptionTypes();
	}

	public BeanInfo getBeanInfo( final Class startClass, final Class stopClass ) throws IntrospectionException
	{
		try
		{
			return (BeanInfo) AccessController.doPrivileged( new PrivilegedExceptionAction()
				{
					public Object run() throws IntrospectionException
					{
						return Introspector.getBeanInfo(startClass, stopClass);
					}
				}
			);
		}
		catch (PrivilegedActionException  pae)
		{
//Thread.dumpStack();
			throw (IntrospectionException) pae.getException();
		}
//		return Introspector.getBeanInfo(startClass, stopClass);
	}

	public String getPrivilegedLocalHostName() throws java.net.UnknownHostException
	{
		try
		{
			return (String) AccessController.doPrivileged( new PrivilegedExceptionAction()
				{
					public Object run() throws java.net.UnknownHostException
					{
						return java.net.InetAddress.getLocalHost().getHostName();
					}
				}
			);
		}
		catch (PrivilegedActionException  pae)
		{
//Thread.dumpStack();
			throw (java.net.UnknownHostException)pae.getException();
		}
//		return java.net.InetAddress.getLocalHost().getHostName();
	}


	//************************************************************************
	// Socket control
	//************************************************************************
	public Socket socketAccept(final ServerSocket serverSocket) throws java.io.IOException
	{
		try
		{
			return (Socket) AccessController.doPrivileged( new PrivilegedExceptionAction()
			{
				public Object run() throws java.io.IOException
				{
					return serverSocket.accept();
				}
			});
		}
		catch (PrivilegedActionException pae)
		{
			throw (java.io.IOException) pae.getException();
		}
//		return serverSocket.accept();
	}
}
