// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.security;

import com.sas.GlobalInfoTable;
import com.sas.util.Strings;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.io.InputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.util.Enumeration;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.Vector;

/**
 * This class provides access to methods which are secure under one of the VM's
 * which we support.  Currently, the only two VM's supported are JavaSoft's and
 * Microsoft's.
 */
public class BaseSecuritySupport
{
  private static final boolean DEBUG = false;

	/**
	 * Users of the SecuritySupport class should not instantiate an instance
	 * of the class but rather go through the instance stored in the securitySupport
	 * field.  This will allow a subclass of the SecuritySupport class to be used
	 * for VM's other than JavaSoft's.
	 */
	public static BaseSecuritySupport securitySupport = null;

  // See SecurityClasses.properties for doc on these values.
  static final private String delimiterDefault = ",";
  // Note: this array is split so that the class names
  // appear as individual literals so that the webAF project
  // packager (spider) can infer the subclasses.
  static final private String SecurityClassDefaults[][] =
  {
    { "Sun Microsystems Inc."/*I18nOK:EMS*/, "1.4*", "com.sas.security.BaseJDK_1_2_SecuritySupport"},
    { "Sun Microsystems Inc."/*I18nOK:EMS*/, "1.3*", "com.sas.security.BaseJDK_1_2_SecuritySupport"},
    { "Sun Microsystems Inc."/*I18nOK:EMS*/, "1.2*", "com.sas.security.BaseJDK_1_2_SecuritySupport"},
  };

  // ****************************************************************
  // NOTE: SecurityClassDefaults must be declared before securitySupport
  // is initialized in the static { } block below
  // ****************************************************************

	/**
	 * Instantiate the appropriate SecuritySupport class and store the instance in
	 * in the securitySupport field.
	 */
	static
	{
      securitySupport = newBaseSecuritySupport();
    }


  static BaseSecuritySupport newBaseSecuritySupport()
  {
    Vector classes;
    if (securitySupport != null)
      return securitySupport;

    // Figure out what VM we're running in

    String vendor = null;
    String version = null;

    try { vendor = System.getProperty("java.vendor"); }
    catch (SecurityException e) { version = "unknown"; }
    try { version = System.getProperty("java.version"); }
    catch (SecurityException e) { version = "unknown"; }

    classes = parseSelections();

    securitySupport = newBaseSecuritySupport(classes, vendor, version);
    if (securitySupport == null)
      securitySupport = new BaseSecuritySupport();
    return securitySupport;
  }

  private static Vector parseSelections()
  {
    Vector list = new Vector();
    Properties classes = new Properties();
    try
      {
        ResourceBundle bundle = ResourceBundle.getBundle("com.sas.security.BaseSecurityClasses");
        int count = Integer.parseInt(bundle.getString("count"));
        String delimiter = bundle.getString("delimiter");

        list.addElement(delimiter);
        for (int i = 0; i < count; i++)
          {
            String key = Integer.toString(i);
            String spec = null;
            try { spec = bundle.getString(key); }
            catch (MissingResourceException e) { }
            if (spec != null)
              list.addElement(spec);
          }
      }
    catch (MissingResourceException e) { }

    // If could not read the properties file, then load some hard wired defaults.
    // Force parsing of SecurityClassDefaults: System.out.println(list); list.removeAllElements();
    if (list.size() <= 1)
      {
        list.removeAllElements();
        list.addElement(delimiterDefault);
        for (int i = 0; i < SecurityClassDefaults.length; i++)
          {
            String item = SecurityClassDefaults[i][0] + delimiterDefault  +/*I18nOK:COS*/
              SecurityClassDefaults[i][1] + delimiterDefault + SecurityClassDefaults[i][2];
            if (DEBUG) System.out.println("default: "/*I18nOK:EMS*/ + item/*I18nOK:COS*/);
            list.addElement(item);
          }
      }
    // System.out.println(list);
    return list;
  }

  private static BaseSecuritySupport newBaseSecuritySupport(Vector classes, String vendor, String version)
  {
    // Now, try to find a match for our current java.vendor+java.version,
    // optionally looking at broader version numbers.
    String vers = version;
    BaseSecuritySupport securitySupport = null;
    String delimiter = (String) classes.elementAt(0);
    classes.removeElementAt(0);
    for (int i = 0; i < classes.size(); i++)
      {
        String spec = (String) classes.elementAt(i);
        int pos = spec.lastIndexOf(delimiter);
        String pattern = spec.substring(0, pos);
        String config = vendor + delimiter + vers;/*I18nOK:COS*/
        if (Strings.match(config, pattern, true))
          {
            String className = spec.substring(pos + delimiter.length(), spec.length());
            if (DEBUG) System.out.println("NOTE: Security class is " + className);
            Class secClass = GlobalInfoTable._loadClass(className);
            if (secClass != null)
              try {
                securitySupport = (BaseSecuritySupport) secClass.newInstance();
                return securitySupport;
              }
              catch (ClassCastException e) { }
              catch (InstantiationException e) { }
              catch (IllegalAccessException e) { }
          }
      }
    // no matches? deturn default.
    return new BaseSecuritySupport();
  }

	//***********************************************************************
	// Reflection method wrappers
	//***********************************************************************

	public Object invokeMethod( Method method, Object object, Object[] args ) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException
	{
		return method.invoke( object, args );
	}

	public Method getMethod( Class cls, String name, Class[] paramTypes ) throws NoSuchMethodException, SecurityException
	{
		return cls.getMethod( name, paramTypes );
	}

	public Method[] getDeclaredMethods( Class cls ) throws SecurityException
	{
	    return cls.getDeclaredMethods();
	}

	public Constructor getConstructor( Class c, Class[] args ) throws NoSuchMethodException
	{
		return c.getConstructor(args);
	}

	public int getMethodModifiers( Member method )
	{
		return method.getModifiers();
	}

	public String getMethodName( Method method )
	{
		return method.getName();
	}

	public Class[] getMethodParameterTypes( Method method )
	{
		return method.getParameterTypes();
	}

	public Class getMethodReturnType( Method method )
	{
		return method.getReturnType();
	}

	public Class[] getMethodExceptionTypes( Method method )
	{
		return method.getExceptionTypes();
	}

	public BeanInfo getBeanInfo( Class startClass, Class stopClass ) throws IntrospectionException
	{
		return Introspector.getBeanInfo(startClass, stopClass);
	}

	public String getPrivilegedLocalHostName() throws java.net.UnknownHostException
	{
		return java.net.InetAddress.getLocalHost().getHostName();
	}

	public String getLocalHostName() throws java.net.UnknownHostException
	{
		return java.net.InetAddress.getLocalHost().getHostName();
	}

	//***********************************************************************
	// Property access wrappers
	//***********************************************************************

    public String getProperty(String propertyName)
    {
      return System.getProperty(propertyName);
    }

    public void setProperties(Properties properties)
    {
      System.setProperties(properties);
    }

	//***********************************************************************
	// File I/O method wrappers
	//***********************************************************************

  public InputStream getResourceAsStream(Class aClass, String resourceName)
  {
    return aClass.getResourceAsStream(resourceName);
  }

  public URL getResource(Class aClass, String resourceName)
  {
    return aClass.getResource(resourceName);
  }

  //************************************************************************
  // Thread control
  //************************************************************************
	public Thread createThread(final Runnable p)
	{
		if ( p instanceof Thread )
			return (Thread)p;
		else
			return new Thread( p );
	}

	public Object doThreadPrivileged(java.security.PrivilegedAction p)
	{
		return p.run();
	}

  public void stop(Thread thread)
  {
    thread.stop();/*I18nOK:MTH*/
  }

  public void interrupt(Thread thread)
  {
    thread.interrupt();
  }

	//************************************************************************
	// Socket control
	//************************************************************************
	public Socket socketAccept(ServerSocket serverSocket) throws java.io.IOException
	{
		return serverSocket.accept();
	}

}
