/* $Id$ */
package com.sas.services.connection;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.omg.CORBA.SystemException;
import org.omg.CORBA.portable.Delegate;
import org.omg.CORBA.portable.ObjectImpl;

import com.sas.net.brg.orb.DelegateBase;
import com.sas.text.Message;
//
// History:
//
// Date    Name            Remarks
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 30Apr02 B Beers         S0149492: Prevent extra server round-trip when narrowing unshared object.
// 30Apr02 B Beers         S0149671: Resolve some garbage collection problems.
// 01May02 B Beers         S0149671: Don't let shared connections get garbage collected.
// 01May02 B Beers         S0149671: Don't do shared connections from this kernel.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 29May02 B Beers         S0139445: Logging service integration (happy 9th anniversary Biff & Julie).
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 21Feb03 B Beers         S0185859: Change handling of factory name to prevent factory proliferation.
// 14Mar03 B Beers         S0183680: Shut down factory when connection to repository is lost.
// 21Mar03 B Beers         S0174314: Close logger handlers when factory is destroyed.
// 15May03 B Beers         S0200797: Add API to control application monitoring.
// 11Dec03 B Beers         S0230810: Check for capacity < minSize.
// 29Jan04 B Beers         S0233579: Use logger associated with user that created factory.
/**
 * Implementation class for ConnectionFactoryKernel which manages non-pooled connections.
 * @author Biff Beers
 */
public final class AggregationKernel extends ConnectionFactoryKernel
{
     /**
      * Construct an aggregation kernel.
      * @param manager the factory manager causing this factory to be constructed
      * @param cxfConfig the factory configuration for this factory
      * @param loggingImpl the log
      * @param clusters the clusters to which this factory will connect
      * @throws java.lang.IllegalArgumentException if clusters is null, empty, or 
      * has null elements, or if two clusters have the same domain
      */
     public AggregationKernel(ConnectionFactoryManager manager,ConnectionFactoryConfiguration cxfConfig,Logger logger,Cluster[] clusters)
     {
          super(manager,cxfConfig,logger);

          if (clusters == null || clusters.length == 0)
          {
               throw new IllegalArgumentException("\"clusters\" cannot be null or empty.");
          }

          domains = new ArrayList(clusters.length);
          clusterMap = new HashMap(clusters.length,1.0F);

          for (int i=0;i<clusters.length;i++)
          {
               if (clusters[i] == null)
               {
                    throw new IllegalArgumentException("Element " + i + " of \"clusters\" cannot be null.");
               }

               String domain = clusters[i].getDomain();

               if (!domains.add(domain))
               {
                    StringBuffer msgBuffer = new StringBuffer("Each cluster must have a unique domain. The domain for element ");
                    msgBuffer.append(i).append(" (").append(domain).append(") has been used by another cluster.");
                    throw new IllegalArgumentException(msgBuffer.toString());
               }

               clusterMap.put(domain,clusters[i]);
          }

          cxSet = new HashSet();
     }

     public synchronized void shutdown()
     {
          if (isShutdown() || isDestroyed())
          {
               return;
          }

          ConnectionFactoryEvent event = new ConnectionFactoryEvent(ConnectionFactoryEvent.TYPE_SHUTDOWN);
          shutdownKernel(event);
     }

     public synchronized void destroy()
     {
          if (isDestroyed())
          {
               return;
          }

          Logger logger = getLogger();
          logger.info("commencing factory destruction");
          Iterator cxIterator = cxSet.iterator();

          while (cxIterator.hasNext())
          {
               WeakReference ref = (WeakReference)cxIterator.next();
               cxIterator.remove();
               ConnectionEnvelope cxEnv = (ConnectionEnvelope)ref.get();

               if (cxEnv != null)
               {
                    destroyConnection(cxEnv,MonitorRecord.CAUSE_DESTROY);
               }
          }

          destroyKernel();
     }

     private List domains = null;
     private Map clusterMap = null;
     private Set cxSet = null;

     protected synchronized ConnectionInterface getConnection(Credential userLogin,long waitMilliseconds,Logger cxLogger) throws ConnectionFactoryException
     {
          // this factory kernel ignores waitMilliseconds
          if (userLogin == null)
          {
               throw new IllegalArgumentException("\"userLogin\" cannot be null.");
          }

          if (isShutdown() || isDestroyed())
          {
               throw new IllegalStateException("\"getConnection()\" cannot be called on a factory after \"shutdown()\" or \"destroy()\" has been called.");
          }

          Logger logger = (cxLogger == null) ? getLogger() : cxLogger;
          logger.info("connection request received with " + userLogin.toString());
          String domain = userLogin.getDomain();
          Cluster cluster = (Cluster)clusterMap.get(domain);

          if (cluster == null)
          {
               // {0}: This factory does not support connections under the authentication domain "{1}".
               String messageIndex = "cxf.000.ex.msg"/*I18NOK:EMS*/;
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex,domain);
               throw new LoginException(message);
          }

          ConnectionEnvelope cxEnv = cluster.getConnection(userLogin,logger);
          
          if (cxEnv.isBogus())
          {
               Throwable thr = cxEnv.getCause();
               
               if (thr instanceof FatalConnectionFactoryException)
               {
                    FatalConnectionFactoryException ex = (FatalConnectionFactoryException)thr;
                    ConnectionFactoryEvent event = null;
                    event = new ConnectionFactoryEvent(ConnectionFactoryEvent.TYPE_FATAL_EXCEPTION,ex);
                    shutdownKernel(event);
                    throw ex;
               }
               
               if (thr instanceof RuntimeException)
               {
                    throw (RuntimeException)thr;
               }
               
               if (thr instanceof Error)
               {
                    throw (Error)thr;
               }
               
               ConnectionFactoryException cfe = null;
               
               if (thr instanceof ConnectionFactoryException)
               {
                    cfe = (ConnectionFactoryException)thr;
               }
               else
               {
                    cfe = new ConnectionFactoryException(thr);
               }
               
               throw cfe;
          }
          else
          {
               cxEnv.setUserCredential(userLogin);
          }

          // Don't keep a strong reference to the connection (via the
          // envelope) in the factory. The connection has only one user, and,
          // if he drops his reference to it, it should go away.
          //-----------------------------------------------------------------
          cxSet.add(new WeakReference(cxEnv));
          ConnectionInterface cx = new ConnectionImpl(cxEnv,this,0);
          ObjectImpl obj = (ObjectImpl)cxEnv.getRootObject();
          Delegate delegate = obj._get_delegate();

          if (delegate instanceof DelegateBase)
          {
               ((DelegateBase)delegate).addParent(cx);
          }

          logger.info("request served by unshared connection #" + cxEnv.getConnectionNumber());
          cutServerStatusRecord(cxEnv,MonitorRecord.CAUSE_REQUEST);
          return cx;
     }

     protected ConnectionFactoryAdminInterface getAdminInterface(Credential adminLogin) throws ConnectionFactoryException
     {
          return this;
     }

     protected List getDomains()
     {
          return ConnectionFactoryUtil.copyList(domains);
     }

     protected synchronized void reactivateConnection(ConnectionEnvelope cxEnv,int clientNumber,int reason)
     {
          if (isDestroyed())
          {
               return;
          }

          String whom = null;
          String cause = null;

          if (reason == ConnectionFactoryKernel.REACT_CX_CLOSED)
          {
               whom = "user";
               cause = MonitorRecord.CAUSE_RETURN;
          }
          else
          {
               whom = "garbage collector";
               cause = MonitorRecord.CAUSE_GC;
          }
          
          int cxNumber = cxEnv.getConnectionNumber();
          Logger logger = getLogger(cxEnv);
          logger.info("connection #" + cxNumber + " returned to factory by " + whom);
          Iterator cxIterator = cxSet.iterator();

          while (cxIterator.hasNext())
          {
               WeakReference ref = (WeakReference)cxIterator.next();

               if (cxEnv.equals(ref.get()))
               {
                    cxIterator.remove();
                    break;
               }
          }

          destroyConnection(cxEnv,cause);

          if (isShutdown() && cxSet.isEmpty())
          {
               destroy();
          }
     }

     private void destroyConnection(ConnectionEnvelope cxEnv,String cause)
     {
          ObjectImpl obj = (ObjectImpl)cxEnv.getRootObject();

          try
          {
               obj._orb().shutdown(true);
          }
          catch (SystemException e)
          {
          }

          Logger logger = getLogger(cxEnv);
          int cxNumber = cxEnv.getConnectionNumber();
          logger.info("connection #" + cxNumber + " destroyed");
          cutServerStatusRecord(cxEnv,cause);
     }

     protected void shutdownKernel(ConnectionFactoryEvent event)
     {
          Logger logger = getLogger();

          if (event.getType() == ConnectionFactoryEvent.TYPE_SHUTDOWN)
          {
               logger.info("factory shutdown commencing");
          }
          else
          {
               logger.info("fatal error detected; factory shutting down");
          }

          super.shutdownKernel(event);

          if (cxSet.isEmpty())
          {
               destroy();
          }
     }

     /*
      *
      */
     private void cutServerStatusRecord(ConnectionEnvelope cxEnv,String cause)
     {
          Logger monitor = getMonitor();

          if (monitor == null)
          {
               return;
          }

          ConnectionFactoryManager cxfManager = getManager();
          MonitorRecord record = new MonitorRecord();
          int clusterIndex = cxEnv.getClusterIndex();
          record.setClusterIndex(clusterIndex);
          int serverIndex = cxEnv.getServerIndex();
          record.setServerIndex(serverIndex);
          ConnectionFactoryConfiguration cxfConfig = getConfig();
          String loggerName = cxfManager.getFactoryLoggerName(cxfConfig);
          record.setFactoryName(loggerName);
          Server server = cxEnv.getServer();
          int reactivationLimit = server.getReactivationLimit();

          if (server instanceof TCPIPServer)
          {
               String serverName = ((TCPIPServer)server).getNormalizedHost();
               record.setServerName(serverName);
          }

          StringBuffer sb = new StringBuffer();
          Iterator cxItor = cxSet.iterator();
          int connections = 0;

          while (cxItor.hasNext())
          {
               WeakReference ref = (WeakReference)cxItor.next();
               ConnectionEnvelope tEnv = (ConnectionEnvelope)ref.get();
               Server tServer = (tEnv != null) ? tEnv.getServer() : null;

               if (server.equals(tServer))
               {
                    connections++;
               }
          }

          record.setConnections(connections);
          record.setIdle(0);
          record.setActive(connections);
          record.setCause(cause);
          monitor.log(record);
     }
}
