/* $Id$ */
package com.sas.services.connection;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import com.sas.text.Message;
//
// History:
//
// Date    Name            Remarks
// 26Nov01 B Beers         S0133039: Created.
// 03Dec01 B Beers         S0133638: Add JDNI configuration.
// 02Jan02 B Beers         S0135536: Add admin auth feature.
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 20May02 B Beers         S0152574: Add load balancing to puddle selection.
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 26Mar03 B Beers         S0185307: New metadata model for connection pools.
// 06Jun03 B Beers         S0204214: Streamline auth checking for pooling factories.
// 25Jun03 B Beers         S0208706: Remove remote calls from synch'd code in getConnection().
// 25Jun03 B Beers         S0208706: Fixed NullPointerException caused by previous fix.
// 04Sep03 B Beers         S0218036: Added support for trusted login.
// 10Sep03 B Beers         S0219412: Allow case-insensitive lookup of privileged user's login.
// 24Feb04 B Beers         S0215591: Remove identity from context ONLY for authentication failure.
// 04Aug05 B Beers         S0306929: Get cx from correct puddle in multi-puddle pool.
/**
 * Basic authentication and authorization services needed by the connection service.
 * @author Biff Beers
 */
public class AuthServer
{
     protected AuthServer(PuddleEnvelope[] envs,boolean findDomain)
     {
          this(envs,findDomain,false);
     }

     protected AuthServer(PuddleEnvelope[] envs,boolean findDomain,boolean cacheUsers)
     {
          if (findDomain)
          {
               Set domainSet = new HashSet();

               for (int i=0;i<envs.length;i++)
               {
                    Set userCache =  envs[i].getUserCache();

                    if (userCache != null)
                    {
                         synchronized (userCache)
                         {
                              Iterator userCacheIterator = userCache.iterator();

                              while (userCacheIterator.hasNext())
                              {
                                   Credential login = (Credential)userCacheIterator.next();
                                   domainSet.add(login.getDomain());
                              }
                         }
                    }
               }

               domains = new ArrayList(domainSet.size());
               domains.addAll(domainSet);
          }

          this.envs = envs;
          this.cacheUsers = cacheUsers;
     }

     private PuddleEnvelope[] envs /* = null */;
     private boolean cacheUsers /* = null */;
     private List domains /* = null */;

     protected final int[] authorizeUser(Credential userLogin) throws ConnectionFactoryException
     {
          userLogin = normalize(userLogin);
          boolean cached = false;
          boolean[] scoreCard = new boolean[envs.length];

          // first check the cache to see if we already know about this user
          for (int i=0;i<envs.length;i++)
          {
               PuddleEnvelope env = envs[i];
               Set userCache = env.getUserCache();

               if (userCache == null)
               {
                    scoreCard[i] = true;
                    cached = true;
               }
               else
               {
                    synchronized (userCache)
                    {
                         if (userCache.contains(userLogin))
                         {
                              scoreCard[i] = true;
                              cached = true;
                         }
                         else
                         {
                              scoreCard[i] = false;
                         }
                    }
               }
          }

          if (!cached)
          {
               String domain = userLogin.getDomain();
               List domainList = getDomains(userLogin);

               if (!domainList.contains(domain))
               {
                    // {0}: This factory does not support connections under the authentication domain "{1}".
                    String messageIndex = "cxf.102.ex.msg";
                    String messagePattern = RB.getStringResource(messageIndex);
                    String message = Message.format(messagePattern,messageIndex,domain);
                    throw new LoginException(message);
               }

               // try to authenticate the login
               Object identity = authenticate(userLogin);

               if (identity == null)
               {
                    // {0}: User could not be authenticated in domain "{1}".
                    String messageIndex = "cxf.100.ex.msg";
                    String messagePattern = RB.getStringResource(messageIndex);
                    String message = Message.format(messagePattern,messageIndex,userLogin.getDomain());
                    throw new InvalidCredentialsException(message);
               }

               boolean authorized = false;

               try
               {
                    // see if any of the puddles have authorized this user
                    for (int i=0;i<envs.length;i++)
                    {
                         PuddleEnvelope env = envs[i];
                         Puddle puddle = env.getPuddle();

                         if (authorizeUser(identity,puddle))
                         {
                              if (cacheUsers)
                              {
                                   env.addUserToCache(userLogin);
                              }

                              scoreCard[i] = true;
                              authorized = true;
                         }
                    }
               }
               finally
               {
                    closeIdentity(identity);
               }

               if (!authorized)
               {
                    // {0}: User is not authorized to use any puddles in this pool.
                    String messageIndex = "cxf.101.ex.msg";
                    String messagePattern = RB.getStringResource(messageIndex);
                    String message = Message.format(messagePattern,messageIndex);
                    throw new LoginException(message);
               }
          }

          int authPuddleCount = 0;

          for (int i=0;i<envs.length;i++)
          {
               if (scoreCard[i])
               {
                    authPuddleCount++;
               }
          }

          int[] authPuddles = new int[authPuddleCount];

          for (int i=0,j=0;i<envs.length;i++)
          {
               if (scoreCard[i])
               {
                    authPuddles[j++] = i;
               }
          }

          return authPuddles;
     }

     protected void authorizeAdmin(Credential adminLogin) throws ConnectionFactoryException
     {
     }

     protected Credential normalize(Credential login)
     {
          return login;
     }

     protected Object authenticate(Credential login) throws ConnectionFactoryException
     {
          return this;
     }

     protected boolean authorizeUser(Object identity,Puddle puddle) throws ConnectionFactoryException
     {
          return false;
     }

     protected void closeIdentity(Object identity)
     {
     }

     // called by PoolingKernel
     protected List getDomains()
     {
          return domains;
     }

     // subclasses may want to stash temporary domains in the login
     // only called from subclasses
     protected List getDomains(Credential login)
     {
          return domains;
     }

     protected final PuddleEnvelope[] getPuddleEnvelopes()
     {
          return envs;
     }

     protected final int findBestPuddle(int[] authPuddles)
     {
          double bestLoad = Double.POSITIVE_INFINITY;
          int bestIndex = 0;

          for (int i=0;i<authPuddles.length;i++)
          {
               PuddleEnvelope env = envs[authPuddles[i]];
               double load = env.getLoad();

               if (load < bestLoad)
               {
                    bestLoad = load;
                    bestIndex = authPuddles[i];
               }
          }

          return bestIndex;
     }

     protected void removeUserFromCache(Credential user)
     {
          Credential login = normalize(user);

          for (int i=0;i<envs.length;i++)
          {
               PuddleEnvelope env = envs[i];
               env.removeUserFromCache(login);
          }
     }

     protected void setDomains(List domains)
     {
          this.domains = domains;
     }
}