/* $Id: BridgeServer.java,v 1.6.16.1 2005/10/11 17:14:12 sasbfb Exp $ */
package com.sas.services.connection;

import java.util.Properties;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sas.net.brg.orb.BrgOrb;
//
// History:
//
// Date    Name            Remarks
// 26Nov01 B Beers         S0133039: Created.
// 16Jan02 B Beers         S0136628: Fixed one-off problem for reactivationLimit.
// 21Jan02 B Beers         S0136966: Fixed NullPointerException in getNormalizedEncryptionAlgorithms().
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 20May02 B Beers         S0152574: Add load balancing to puddle selection.
// 29May02 B Beers         S0139445: Logging service integration (happy 9th anniversary Biff & Julie).
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 13Jan03 B Beers         S0172209: Fixed spelling of sasproprietary.
// 05Apr03 B Beers         S0193189: Blot out password in FINER/FINEST log messages.
// 12Nov03 B Beers         S0227109: Remove stub finalizers.
// 11Oct05 B Beers         S0324947: Added serialVersionUID.
/**
 * A description of an IOM server.
 * @author Biff Beers
 */
public class BridgeServer extends TCPIPServer
{
     /**
      * Construct an object describing an IOM server.
      * @param classID the ID of the class to instantiate on the IOM server.
      * @param host the IP name or address of the machine hosting the IOM server.
      * @param port the TCP port number where the server is listening for connections.
      */
     public BridgeServer(String classID,String host,int port)
     {
          super(classID,host,port);
     }

     /**
      * Set the encryption policy. Valid values are specified
      * with the <CODE>ENCRYPTION_POLICY_*</CODE> fields. The
      * default value is <CODE>ENCRYPTION_POLICY_NONE</CODE>
      * @param encryptionPolicy the encryption policy
      * @throws java.lang.IllegalArgumentException if the value is
      * not one of the <CODE>ENCRYPTION_POLICY_*</CODE> fields
      * @see #ENCRYPTION_POLICY_REQUIRED
      * @see #ENCRYPTION_POLICY_OPTIONAL
      * @see #ENCRYPTION_POLICY_NONE
      */
     public void setEncryptionPolicy(String encryptionPolicy)
     {
          if (!ENCRYPTION_POLICY_REQUIRED.equals(encryptionPolicy) &&
              !ENCRYPTION_POLICY_OPTIONAL.equals(encryptionPolicy) &&
              !ENCRYPTION_POLICY_NONE.equals(encryptionPolicy))
          {
               String msg = "\"encryptionPolicy\" must be " +
                            ENCRYPTION_POLICY_REQUIRED + ", " +
                            ENCRYPTION_POLICY_OPTIONAL + ", or " +
                            ENCRYPTION_POLICY_NONE + ".";
               throw new IllegalArgumentException(msg);
          }

          this.encryptionPolicy = encryptionPolicy;
     }

     /**
      * Get the encryption policy.
      * @return the encryption policy
      * @see #ENCRYPTION_POLICY_REQUIRED
      * @see #ENCRYPTION_POLICY_OPTIONAL
      * @see #ENCRYPTION_POLICY_NONE
      * @see #setEncryptionPolicy
      */
     public String getEncryptionPolicy()
     {
          return encryptionPolicy;
     }

     /**
      * @param encryptionContent Valid values are as follows:
      * <P>
      * <DL>
      * </DL>
      * <P>
      *
      * Set the encryption content. This value is ignored if the
      * encryption policy is <CODE>ENCRYPTION_POLICY_NONE</CODE>. 
      * Valid values are specified
      * with the <CODE>ENCRYPTION_CONTENT_*</CODE> fields. The default value
      * is <CODE>ENCRYPTION_CONTENT_ALL</CODE>.
      * @param encryptionContent the encryption content
      * @throws java.lang.IllegalArgumentException if the value is
      * not one of the <CODE>ENCRYPTION_CONTENT_*</CODE> fields
      * @see #ENCRYPTION_CONTENT_ALL
      * @see #ENCRYPTION_CONTENT_AUTHENTICATION
      */
     public void setEncryptionContent(String encryptionContent)
     {
          if (!ENCRYPTION_CONTENT_ALL.equals(encryptionContent) &&
              !ENCRYPTION_CONTENT_AUTHENTICATION.equals(encryptionContent))
          {
               String msg = "\"encryptionContent\" must be " +
                            ENCRYPTION_CONTENT_ALL + ", or" +
                            ENCRYPTION_CONTENT_AUTHENTICATION + ".";
               throw new IllegalArgumentException(msg);
          }

          this.encryptionContent = encryptionContent;
     }

     /**
      * Get the encryption content.
      * @return the encryption content
      * @see #setEncryptionContent
      * @see #ENCRYPTION_CONTENT_ALL
      * @see #ENCRYPTION_CONTENT_AUTHENTICATION
      */
     public String getEncryptionContent()
     {
          return encryptionContent;
     }

     /**
      * Set the encryption algorithms. To use any
      * algorithm other than <CODE>ENCRYPTION_ALGORITHM_SASPROPRIETARY</CODE>, 
      * SAS/SECURE software must be installed on both the
      * client and server. Furthermore, use of <CODE>ENCRYPTION_ALGORITHM_DES</CODE>, and
      * <CODE>ENCRYPTION_ALGORITHM_TRIPLEDES</CODE> may be restricted outside the USA due to export law.
      * <P>
      * The value for this attribute may be a comma-separated list of more than
      * one algorithm.
      * The connection will use the first algorithm on this list that is also
      * supported by the server. If no value is specified, the connection will use the
      * first value suggested by the server that the client is capable of supporting.
      * This value is ignored if the encryption policy is <CODE>ENCRYPTION_POLICY_NONE</CODE>.
      * <P>
      * The value of this attribute is not checked for validity because it is possible to
      * extend the list of supported algorithms in the field although no such extensions
      * are currently available.
      * @param encryptionAlgorithms the encryption algorithms
      * @see #ENCRYPTION_ALGORITHM_SASPROPRIETARY
      * @see #ENCRYPTION_ALGORITHM_RC2
      * @see #ENCRYPTION_ALGORITHM_RC4
      * @see #ENCRYPTION_ALGORITHM_DES
      * @see #ENCRYPTION_ALGORITHM_TRIPLEDES
      */
     public void setEncryptionAlgorithms(String encryptionAlgorithms)
     {
          origEncryptionAlgorithms = encryptionAlgorithms;
          normEncryptionAlgorithms = normalizeEncryptionAlgorithms(encryptionAlgorithms);
     }

     /**
      * Get the encryption algorithms. This method may
      * return <CODE>null</CODE>.
      * @return the encryption algorithms
      * @see #setEncryptionAlgorithms
      * @see #ENCRYPTION_ALGORITHM_SASPROPRIETARY
      * @see #ENCRYPTION_ALGORITHM_RC2
      * @see #ENCRYPTION_ALGORITHM_RC4
      * @see #ENCRYPTION_ALGORITHM_DES
      * @see #ENCRYPTION_ALGORITHM_TRIPLEDES
      */
     public String getEncryptionAlgorithms()
     {
          return origEncryptionAlgorithms;
     }

     /**
      * Get the normalized encryption algorithm list.
      * The normalized encryption algorithm list is
      * all lowercase, and all whitespace around the
      * algorithm names is removed. This method may
      * return <CODE>null</CODE>.
      * @return the normalized encryption algorithm list
      */
     public String getNormalizedEncryptionAlgorithms()
     {
          return normEncryptionAlgorithms;
     }

     public boolean equals(Object that)
     {
          if (!super.equals(that))
          {
               return false;
          }

          BridgeServer thatBridgeServer = (BridgeServer)that;
          String thatEncryptionPolicy = thatBridgeServer.getEncryptionPolicy();

          if (!encryptionPolicy.equals(thatEncryptionPolicy))
          {
               return false;
          }

          String thatEncryptionContent = thatBridgeServer.getEncryptionContent();

          if (!encryptionContent.equals(thatEncryptionContent))
          {
               return false;
          }

          String thoseNormEncryptionAlgorithms = thatBridgeServer.getNormalizedEncryptionAlgorithms();

          if (normEncryptionAlgorithms == null ? thoseNormEncryptionAlgorithms != null : 
                                                 !normEncryptionAlgorithms.equals(thoseNormEncryptionAlgorithms))
          {
               return false;
          }

          return true;
     }

     public int hashCode()
     {
          int hc = super.hashCode();
          hc ^= encryptionPolicy.hashCode();
          hc ^= encryptionContent.hashCode();

          if (normEncryptionAlgorithms != null)
          {
               hc ^= normEncryptionAlgorithms.hashCode();
          }

          return hc;
     }

     /**
      * A value for the encryption policy attribute indicating that communication with 
      * the server will be encrypted if
      * the server supports an encryption algorithm supported by the client. Otherwise
      * the connection will fail.
      * @see #setEncryptionPolicy
      * @see #getEncryptionPolicy
      */
     public static final String ENCRYPTION_POLICY_REQUIRED = "required"/*I18NOK:EMS*/;

     /**
      * A value for the encryption policy attribute indicating that communication
      * with the server will be encrypted if
      * the server supports an encryption algorithm supported by the client. Otherwise
      * no communication with the server will be encrypted.
      * @see #setEncryptionPolicy
      * @see #getEncryptionPolicy
      */
     public static final String ENCRYPTION_POLICY_OPTIONAL = "optional"/*I18NOK:EMS*/;

     /**
      * A value for the encryption policy attribute indicating that no
      * communication with the server will be encrypted.
      * If the server requires encryption, the connection will fail.
      * If this value is used for the encryption policy, then the values for
      * <CODE>encryptionContent</CODE> and <CODE>encryptionAlgorithms</CODE> will
      * be ignored.
      * @see #setEncryptionPolicy
      * @see #getEncryptionPolicy
      */
     public static final String ENCRYPTION_POLICY_NONE = "none"/*I18NOK:EMS*/;

     /**
      * A value for the encryption content attribute indicating that
      * all communication is encrypted.
      * @see #setEncryptionContent
      * @see #getEncryptionContent
      */
     public static final String ENCRYPTION_CONTENT_ALL = "all"/*I18NOK:EMS*/;

     /**
      * A value for the encryption content attribute indicating that
      * only user name and password information is encrypted.
      * @see #setEncryptionContent
      * @see #getEncryptionContent
      */
     public static final String ENCRYPTION_CONTENT_AUTHENTICATION = "authentication"/*I18NOK:EMS*/;

     /**
      * A value for the encryption algorithms attribute indicating that
      * sasproprietary encryption may be used. 
      * Encryption using sasproprietary
      * is generally weaker and slower than other supported algorithms,
      * but it requires no special licensing.
      */
     public static final String ENCRYPTION_ALGORITHM_SASPROPRIETARY = "sasproprietary"/*I18NOK:EMS*/;

     /**
      * A value for the encryption algorithms attribute indicating that
      * RC2 encryption may be used. RC2 encryption can only be used if
      * SAS/SECURE software is installed on both the client and server.
      */
     public static final String ENCRYPTION_ALGORITHM_RC2 = "rc2"/*I18NOK:EMS*/;

     /**
      * A value for the encryption algorithms attribute indicating that
      * RC4 encryption may be used. RC4 encryption can only be used if
      * SAS/SECURE software is installed on both the client and server.
      */
     public static final String ENCRYPTION_ALGORITHM_RC4 = "rc4"/*I18NOK:EMS*/;

     /**
      * A value for the encryption algorithms attribute indicating that
      * DES encryption may be used. DES encryption can only be used if
      * SAS/SECURE software is installed on both the client and server.
      */
     public static final String ENCRYPTION_ALGORITHM_DES = "des"/*I18NOK:EMS*/;

     /**
      * A value for the encryption algorithms attribute indicating that
      * TRIPLEDES encryption may be used. TRIPLEDES encryption can only be used if
      * SAS/SECURE software is installed on both the client and server.
      */
     public static final String ENCRYPTION_ALGORITHM_TRIPLEDES = "tripledes"/*I18NOK:EMS*/;

     /**
      * Serial version ID. 
      * This value should be changed if this version is not compatible with the previous version.
      */
     private static final long serialVersionUID = 7325217805525243858L;
     private String encryptionPolicy = ENCRYPTION_POLICY_NONE;
     private String encryptionContent = ENCRYPTION_CONTENT_ALL;
     private String origEncryptionAlgorithms = null;
     private String normEncryptionAlgorithms = null;

     private String getNormEncryptionAlgorithms()
     {
          return normEncryptionAlgorithms;
     }

     private static String normalizeEncryptionAlgorithms(String encryptionAlgorithms)
     {
          if (encryptionAlgorithms == null)
          {
               return null;
          }

          StringTokenizer t = new StringTokenizer(encryptionAlgorithms,","/*I18NOK:EMS*/);
          StringBuffer b = new StringBuffer();

          while (t.hasMoreTokens())
          {
               String alg = t.nextToken();

               if (alg != null)
               {
                    alg = alg.trim();

                    if (alg.length() > 0)
                    {
                         if (b.length() > 0)
                         {
                              b.append(","/*I18NOK:EMS*/);
                         }

                         b.append(alg.toLowerCase());
                    }
               }
          }

          if (b.length() == 0)
          {
               return null;
          }

          return b.toString();
     }

     protected org.omg.CORBA.Object connect(Credential credential,Logger logger,int cxNumber)
     {
          String userName = null;
          String password = null;

          if (credential != null)
          {
               if (credential instanceof PasswordCredential)
               {
                    PasswordCredential login = (PasswordCredential)credential;
                    userName = login.getUserName();
                    password = login.getPassword();
               }
               else
               {
                    String msg = getClass().getName() + " does not support authorization for credential type " +
                                 credential.getClass().getName() + ".";
                    throw new IllegalArgumentException(msg);
               }
          }

          Properties p = new Properties();
          p.put("cxNumber",Integer.toString(cxNumber));
          p.put("protocol","bridge"); /*I18NOK:LINE*/
          String temp = getHost();
          p.put("host",temp); /*I18NOK:LINE*/
          temp = Integer.toString(getPort());
          p.put("port",temp); /*I18NOK:LINE*/
          temp = getClassID();
          p.put("clsid",temp); /*I18NOK:LINE*/
          temp = getEncryptionPolicy();
          p.put("encryptionPolicy",temp); /*I18NOK:LINE*/
          temp = getEncryptionContent();
          p.put("encryptionContent",temp); /*I18NOK:LINE*/
          temp = getEncryptionAlgorithms();

          if (temp != null && temp.length() > 0)
          {
               p.put("encryptionAlgorithms",temp); /*I18NOK:LINE*/
          }

          if (password != null)
          {
               p.put("password",password); /*I18NOK:LINE*/
          }

          if (userName != null)
          {
               p.put("userName",userName); /*I18NOK:LINE*/
          }

          p.put("logFile",logger); /*I18NOK:LINE*/

          if (logger.isLoggable(Level.INFO))
          {
               Properties q = (Properties)p.clone();

               if (q.containsKey("password"))
               {
                    q.put("password","xxxxxxxx");
               }

               logger.info("properties for new connection: " + q.toString());
          }

          BrgOrb orb = new BrgOrb();
          org.omg.CORBA.Object obj = orb.propertiesToObjectNoCopy(p);
          return obj;
     }
}