/* $Id: Cluster.java,v 1.9.2.1.14.1 2005/10/11 17:14:12 sasbfb Exp $ */
package com.sas.services.connection;

import java.io.Serializable;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.NO_PERMISSION;
import org.omg.CORBA.SystemException;
import com.sas.text.Message;
//
// History:
//
// Date    Name            Remarks
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 27Sep02 B Beers         S0166941: Disallow illegal log level values.
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 29Oct02 B Beers         S0156049: Added support for expired credentials.
// 26Nov02 B Beers         S0151508: Revise xml error message handling.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 23Jan03 B Beers         S0156049: Add granularity to authentication errors.
// 04Mar03 B Beers         S0187720: Put LoginException in log.
// 04Apr03 B Beers         S0193912: Improve error messages for connection failure.
// 01May03 B Beers         S0198710: Handle LB timeout.
// 12Nov03 B Beers         S0227109: Remove stub finalizers.
// 11Dec03 B Beers         S0230810: Check for capacity < minSize.
// 11Oct05 B Beers         S0324947: Added serialVersionUID.
/**
 * A base class representing a cluster of IOM servers.
 * @author Biff Beers
 */
public abstract class Cluster implements Cloneable,Serializable
{
     /**
      * Construct and return a cluster of IOM servers.
      * @param servers the servers to add to the cluster.
      * @param clusterType the cluster type
      * @see #TYPE_FAILOVER
      * @see #TYPE_LOAD_BALANCING
      */
     public static Cluster getInstance(Server[] servers,int clusterType)
     {
          if (clusterType == TYPE_FAILOVER)
          {
               return new FailoverCluster(servers);
          }
          else if (clusterType == TYPE_LOAD_BALANCING)
          {
               return new LoadBalancingCluster(servers);
          }
          else
          {
               throw new IllegalArgumentException("\"clusterType\" must be either TYPE_FAILOVER or TYPE_LOAD_BALANCING.");
          }
     }

     /**
      * Construct a cluster of IOM servers.
      * @param servers the servers to add to the cluster.
      */
     protected Cluster(Server[] servers)
     {
          if (servers == null || servers.length == 0)
          {
               throw new IllegalArgumentException("\"servers\" cannot be null or empty.");
          }

          for (int i=0;i<servers.length;i++)
          {
               if (servers[i] == null)
               {
                    throw new IllegalArgumentException("Element " + i + " of \"servers\" cannot be null.");
               }
          }

          this.servers = servers;
          init();
     }

     /**
      * Set the array of servers in the cluster.
      * @param serverArray the array of servers to add to the cluster
      */
     public void setServerArray(Server[] serverArray)
     {
          this.servers = serverArray;
     }

     /**
      * Get the array of servers that have been added to the cluster.
      * @return the array of servers that have been added to the cluster
      */
     public Server[] getServerArray()
     {
          return servers;
     }

     /**
      * Get the cluster type.
      * @return the cluster type
      * @see #TYPE_FAILOVER
      * @see #TYPE_LOAD_BALANCING
      */
     public abstract int getClusterType();

     /**
      * Get the class ID of the servers in this cluster.
      * All servers in a cluster must have the same class ID.
      * @return the class ID of the servers in this cluster
      */
     public String getClassID()
     {
          return classID;
     }

     /**
      * Get the domain of the servers in this cluster.
      * All servers in a cluster must have the same domain.
      * @return the domain of the servers in this cluster
      */
     public String getDomain()
     {
          return domain;
     }

     /**
      * Get the maximum number of clients this cluster
      * can support at one time. If this is a load balancing
      * cluster, then this value is the sum of maxClients
      * for all servers in the cluster. If this is a fail
      * over cluster, then this value is equal to the
      * smallest value of maxClients for all servers in
      * the cluster.
      * @return the maximum number of clients this cluster
      * can support at one time
      */
     public int getMaxClients()
     {
          return maxClients;
     }

     public boolean equals(Object that)
     {
          if (this == that)
          {
               return true;
          }

          if (that == null || !getClass().getName().equals(that.getClass().getName()))
          {
               return false;
          }

          Cluster thatCluster = (Cluster)that;
          Server[] thoseServers = thatCluster.getServerArray();

          if (servers.length != thoseServers.length)
          {
               return false;
          }

          for (int i=0;i<servers.length;i++)
          {
               if (!servers[i].equals(thoseServers[i]))
               {
                    return false;
               }
          }

          return true;
     }

     public int hashCode()
     {
          int hc = getClass().getName().hashCode();

          for (int i=0;i<servers.length;i++)
          {
               hc ^= servers[i].hashCode();
          }

          return hc;
     }

     public Object clone()
     {
          Cluster cluster = null;

          try
          {
               cluster = (Cluster)super.clone();
          }
          catch (CloneNotSupportedException e)
          {
               throw new InternalError("A Cloneable object threw CloneNotSupportedException.");
          }

          Server[] serverArray = cluster.getServerArray();
          serverArray = ConnectionFactoryUtil.copyServerArray(serverArray);
          cluster.setServerArray(serverArray);
          return cluster;
     }

     /**
      * Flag for load-balancing clusters. A load-balancing cluster
      * is implemented with a IOM Object Spawner that monitors the
      * load of a set of redundant servers and directs requests for new
      * connections to the least loaded server.
      */
     public static final int TYPE_LOAD_BALANCING = 0;

     /**
      * Flag for failover clusters. A failover cluster is a set of
      * redundant servers. The connection factory will attempt to
      * connect to one of them, and, if that server is not available
      * it will try another server in the cluster. The factory will
      * raise an exception only if none of the servers in the cluster
      * are available at the time of a connection request.
      */
     public static final int TYPE_FAILOVER = 1;

     /**
      * Serial version ID. 
      * This value should be changed if this version is not compatible with the previous version.
      */
     private static final long serialVersionUID = -8799220558539020978L;
     private Server[] servers = null;
     private String classID = null;
     private String domain = null;
     private int maxClients = 0;
     private int serverIndex = 0;

     protected ConnectionEnvelope getConnection(Credential credential,Logger logger)
     {
          org.omg.CORBA.Object obj = null;
          int index = serverIndex;
          ConnectionEnvelope cxEnv = new ConnectionEnvelope();
          int cxNumber = cxEnv.getConnectionNumber();
          
          do
          {
               Server server = servers[index];

               try
               {
                    obj = server.connect(credential,logger,cxNumber);
               }
               catch (NO_PERMISSION np)
               {
                    String msg = np.getMessage();
                    LoginException le = null;

                    if ((np.minor & 0x00000FFF) == 3)
                    {
                         le = new ExpiredCredentialsException(msg,np);
                    }
                    else if ((np.minor & 0x00000FFF) == 4)
                    {
                         le = new InvalidCredentialsException(msg,np);
                    }
                    else
                    {
                         le = new LoginException(msg,np);
                    }

                    logger.log(Level.SEVERE,msg,le);
                    cxEnv.setBogus(true,le);
                    return cxEnv;
               }
               catch (SystemException se)
               {
                    if (se instanceof COMM_FAILURE && (se.minor & 0x00000FFF) == 7)
                    {
                         String msg = se.getMessage();
                         ConnectionTimeoutException cte = new ConnectionTimeoutException(msg,se);
                         logger.log(Level.SEVERE,msg,cte);
                         cxEnv.setBogus(true,cte);
                         return cxEnv;
                    }

                    logger.log(Level.SEVERE,"new connection failed:",se);
                    index = (index + 1) % servers.length;

                    if (index == serverIndex) // we tried 'em, and they all failed
                    {
                         logger.severe("no more servers in cluster");

                         if (servers.length == 1)
                         {
                              String message = se.getMessage();
                              cxEnv.setBogus(true,new FatalConnectionFactoryException(message,se));
                         }
                         else
                         {
                              // {0}: An exception was thrown while connecting to a server in a cluster.
                              String messageIndex = "cxf.400.ex.msg";
                              String messagePattern = RB.getStringResource(messageIndex);
                              String message = Message.format(messagePattern,messageIndex);
                              cxEnv.setBogus(true,new FatalConnectionFactoryException(message,se));
                         }
                         
                         return cxEnv;
                    }
                    else
                    {
                         logger.severe("trying next server in cluster");
                    }
               }
          }
          while (obj == null);

          cxEnv.setRootObject(obj);
          cxEnv.setServer(servers[index]);
          cxEnv.setServerIndex(index);
          cxEnv.setOwnerCredential(credential);
          return cxEnv;
     }

     protected abstract int computeMaxClients();

     private void init()
     {
          maxClients = computeMaxClients();
          domain = servers[0].getDomain();

          for (int i=1;i<servers.length;i++)
          {
               if (!domain.equals(servers[i].getDomain()))
               {
                    StringBuffer msgBuffer = new StringBuffer("All servers in a cluster must have the same domain. ");
                    msgBuffer.append("Server 0 has domain \"").append(domain).append("\", but server ");
                    msgBuffer.append(i).append(" has domain \"").append(servers[i].getDomain()).append("\".");
                    throw new IllegalArgumentException(msgBuffer.toString());
               }
          }

          classID = servers[0].getNormalizedClassID();

          for (int i=1;i<servers.length;i++)
          {
               if (!classID.equals(servers[i].getNormalizedClassID()))
               {
                    StringBuffer msgBuffer = new StringBuffer("All servers in a cluster must have the same class ID. ");
                    msgBuffer.append("Server 0 has class ID \"").append(classID).append("\", but server ");
                    msgBuffer.append(i).append(" has classID \"").append(servers[i].getClassID()).append("\".");
                    throw new IllegalArgumentException(msgBuffer.toString());
               }
          }
     }
}