package com.sas.services.connection;

import java.util.Set;
//
// History:
//
// Date    Name            Remarks
// 26Nov01 B Beers         S0133039: Created.
// 02Jan02 B Beers         S0135536: Add admin auth feature.
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 03May02 B Beers         S0149086: Removed obsolete methods.
//
/**
 * User interface for connection factories. The factory
 * controls the lifetime of a connection. When a request for
 * a connection arrives, the factory will return an existing
 * pooled connection if one is available, or it will create a new connection
 * if necessary. Users must notify the factory when they are done with the
 * connection so it can be returned to the pool or destroyed. The
 * {@link com.sas.services.connection.ConnectionInterface connection handle}
 * provides the necessary notification method.
 * <P>
 * A factory may have a limit on the number of connections it
 * is allowed to create and manage at a time. If a factory allocates all the connections
 * it is allowed to managed to users, and
 * a new request for a connection arrives, the factory will not be able
 * to serve the request immediately. The various <CODE>getConnection()</CODE>
 * methods provide flags and defaults that allow the caller to specify how long he or
 * she is willing to wait for another user to return a connection to the factory's pool.
 *
 * @author Biff Beers
 */
public interface ConnectionFactoryInterface extends ConnectionFactoryShellInterface
{
     /**
      * Get a connection to a server that does not require a user name and password for
      * connections. If the factory has allocated all the connections it is allowed to manage to users
      * when this call is made, then this call will block until another user returns a connection
      * to the factory's pool. The server must belong to an authentication domain with name ""
      * (empty string).
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the factory has been shutdown
      */
     ConnectionInterface getConnection() throws ConnectionFactoryException;

     /**
      * Get a connection to a server that does not require a user name and password for
      * connections. If the factory has allocated all the connections it is allowed to manage to users
      * when this call is made, then this call will block until another user returns a connection
      * to the factory's pool. The server must belong to an authentication domain with the given name.
      * @param domain the authentication domain for the server
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the factory has been shutdown
      */
     ConnectionInterface getConnection(String domain) throws ConnectionFactoryException;

     /**
      * Get a connection to a server that does not require a user name and password for
      * connections. If the factory has allocated all the connections it is allowed to manage to users
      * when this call is made, then this call will behave as instructed by the value of
      * <PARAM>waitMilliseconds</PARAM>. The server must belong to an authentication domain with name ""
      * (empty string).
      * @param waitMilliseconds a flag indicating how the call should behave if the factory has allocated
      * all the connections it is allowed to manage to other users.
      * <P>
      * <DL>
      * <DT><CODE>&lt;0</CODE><DD>throw an exception immediately
      * <DT><CODE>0</CODE><DD>wait until another user returns a connection
      * to the factory's pool
      * <DT><CODE>&gt;0</CODE><DD>wait up to <PARAM>waitMilliseconds</PARAM> milliseconds for another user
      * to return a connection to the factory's pool and throw an exception if a connection is not returned
      * in the alloted time.
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the request times out or if the factory has been shutdown
      */
     ConnectionInterface getConnection(long waitMilliseconds) throws ConnectionFactoryException;

     /**
      * Get a connection to a server that does not require a user name and password for
      * connections. If the factory has allocated all the connections it is allowed to manage to users
      * when this call is made, then this call will behave as instructed by the value of
      * <PARAM>waitMilliseconds</PARAM>. The server must belong to an authentication domain with the given name.
      * @param domain the authentication domain for the server
      * @param waitMilliseconds a flag indicating how the call should behave if the factory has allocated
      * all the connections it is allowed to manage to other users.
      * <P>
      * <DL>
      * <DT><CODE>&lt;0</CODE><DD>throw an exception immediately
      * <DT><CODE>0</CODE><DD>wait until another user returns a connection
      * to the factory's pool
      * <DT><CODE>&gt;0</CODE><DD>wait up to <PARAM>waitMilliseconds</PARAM> milliseconds for another user
      * to return a connection to the factory's pool and throw an exception if a connection is not returned
      * in the alloted time.
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the request times out or if the factory has been shutdown
      */
     ConnectionInterface getConnection(String domain,long waitMilliseconds) throws ConnectionFactoryException;

     /**
      * Get a connection to a server. If the factory has allocated all the connections it is allowed
      * to manage to users
      * when this call is made, then this call will block until another user returns a connection
      * to the factory's pool. The server must belong to an authentication domain with name ""
      * (empty string).
      * @param userName the name of someone allowed to use this factory
      * @param password the password that authenticates <PARAM>userName</PARAM>
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the factory has been shutdown or
      */
     ConnectionInterface getConnection(String userName,String password) throws ConnectionFactoryException;

     /**
      * Get a connection to a server. If the factory has allocated all the connections it is allowed
      * to manage to users
      * when this call is made, then this call will block until another user returns a connection
      * to the factory's pool. The server must belong to an authentication domain with the given name.
      * @param userName the name of someone allowed to use this factory
      * @param password the password that authenticates <PARAM>userName</PARAM>
      * @param domain the authentication domain for <PARAM>userName</PARAM> and <PARAM>password</PARAM>
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the factory has been shutdown
      */
     ConnectionInterface getConnection(String userName,String password,String domain) throws ConnectionFactoryException;

     /**
      * Get a connection to a server. If the factory has allocated all the connections it is
      * allowed to manage to users
      * when this call is made, then this call will behave as instructed by the value of
      * <PARAM>waitMilliseconds</PARAM>. The server must belong to an authentication domain with name "" (empty string).
      * @param userName the name of someone allowed to use this factory
      * @param password the password that authenticates <PARAM>userName</PARAM>
      * @param waitMilliseconds a flag indicating how the call should behave if the factory has allocated
      * all the connections it is allowed to manage to other users.
      * <P>
      * <DL>
      * <DT><CODE>&lt;0</CODE><DD>throw an exception immediately
      * <DT><CODE>0</CODE><DD>wait until another user returns a connection
      * to the factory's pool
      * <DT><CODE>&gt;0</CODE><DD>wait up to <PARAM>waitMilliseconds</PARAM> milliseconds for another user
      * to return a connection to the factory's pool and throw an exception if a connection is not returned
      * in the alloted time.
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the request times out or if the factory has been shutdown
      */
     ConnectionInterface getConnection(String userName,String password,long waitMilliseconds) throws ConnectionFactoryException;

     /**
      * Get a connection to a server. If the factory has allocated all the connections it is
      * allowed to manage to users
      * when this call is made, then this call will behave as instructed by the value of
      * <PARAM>waitMilliseconds</PARAM>. The server must belong to an authentication domain with the given name.
      * @param userName the name of someone allowed to use this factory
      * @param password the password that authenticates <PARAM>userName</PARAM>
      * @param domain the authentication domain for <PARAM>userName</PARAM> and <PARAM>password</PARAM>
      * @param waitMilliseconds a flag indicating how the call should behave if the factory has allocated
      * all the connections it is allowed to manage to other users.
      * <P>
      * <DL>
      * <DT><CODE>&lt;0</CODE><DD>throw an exception immediately
      * <DT><CODE>0</CODE><DD>wait until another user returns a connection
      * to the factory's pool
      * <DT><CODE>&gt;0</CODE><DD>wait up to <PARAM>waitMilliseconds</PARAM> milliseconds for another user
      * to return a connection to the factory's pool and throw an exception if a connection is not returned
      * in the alloted time.
      * @return a connection
      * @throws com.sas.services.connection.ConnectionFactoryException if the factory cannot create a
      * connection to satisfy this request or if the request times out or if the factory has been shutdown
      */
     ConnectionInterface getConnection(String userName,String password,String domain,long waitMilliseconds) throws ConnectionFactoryException;

     /**
      * Get the factory administrator's interface. The administrator's interface allows you to manipulate factory logging
      * and to shutdown or destroy the factory. Actions performed on the administrator's interface may impact other
      * users of this factory.
      * @return the factory administrator's interface
      */
     ConnectionFactoryAdminInterface getAdminInterface() throws ConnectionFactoryException;

     /**
      * This method is provided for compatibility with previous releases, and it is otherwise rarely used.
      */
     ConnectionFactoryAdminInterface getAdminInterface(String adminName,String password) throws ConnectionFactoryException;
}