/* $Id$ */
package com.sas.services.connection;

import java.security.PrivilegedAction;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import org.omg.CORBA.portable.Delegate;
import org.omg.CORBA.portable.ObjectImpl;

import com.sas.net.brg.orb.DelegateBase;
import com.sas.net.brg.orb.ProtocolEngineInterface;
import com.sas.security.BaseSecuritySupport;
//
// History:
//
// Date    Name            Remarks
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 30Apr02 B Beers         S0149671: Resolve some garbage collection problems.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 29May02 B Beers         S0139445: Logging service integration (happy 9th anniversary Biff & Julie).
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 17Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring (phase 2).
// 14Mar03 B Beers         S0183680: Shut down factory when connection to repository is lost.
// 21Mar03 B Beers         S0174314: Close logger handlers when factory is destroyed.
// 31Mar03 B Beers         S0184610: Give threads useful names.
// 15May03 B Beers         S0200797: Add API to control application monitoring.
// 06Jun03 B Beers         S0204214: Streamline auth checking for pooling factories.
// 25Jun03 B Beers         S0208706: Remove remote calls from synch'd code in getConnection().
// 29Jan04 B Beers         S0233579: Use logger associated with user that created factory.
/**
 * The base class for connection managers.
 * @author Biff Beers
 */
public abstract class ConnectionFactoryKernel implements ConnectionFactoryAdminInterface
{
     protected ConnectionFactoryKernel(ConnectionFactoryManager manager,ConnectionFactoryConfiguration cxfConfig,Logger logger)
     {
          this.manager = manager;
          this.cxfConfig = cxfConfig;
          this.logger = logger;
     }

     public final Logger getLogger()
     {
          return logger;
     }

     protected static final int REACT_CX_CLOSED  = 0;
     protected static final int REACT_CX_GC      = 1;
     protected static final int REACT_CX_REFRESH = 2;
     private Logger logger /* = null */;
     private Logger monitor /* = null */;
     private ConnectionFactoryConfiguration cxfConfig /* = null */;
     private Set listenerSet = new HashSet();
     private ConnectionFactoryManager manager /* = null */;
     private boolean shutdown /* = null */;
     private boolean destroyed /* = null */;

     protected ConnectionFactoryManager getManager()
     {
          return manager;
     }

     protected ConnectionFactoryConfiguration getConfig()
     {
          return cxfConfig;
     }

     protected void addConnectionFactoryEventListener(ConnectionFactoryEventListener listener)
     {
          synchronized (listenerSet)
          {
               listenerSet.add(listener);
          }
     }

     protected void removeConnectionFactoryEventListener(ConnectionFactoryEventListener listener)
     {
          synchronized (listenerSet)
          {
               listenerSet.remove(listener);
          }
     }

     protected void broadcastEvent(final ConnectionFactoryEvent event)
     {
          synchronized (listenerSet)
          {
               // Make a copy of set so listeners can call 
               // removeConnectionFactoryEventListener from 
               // their event handlers without confusing the iterator.
               // Also notify listeners on a separate thread so
               // that the thread synchronization wont be defeated.
               final int listenerSetSize = listenerSet.size();

               if (listenerSetSize > 0)
               {
                    final Set copy = new HashSet(listenerSetSize);
                    copy.addAll(listenerSet);

                    Runnable r = new Runnable()
                    {
                         public void run()
                         {
                              Iterator listenerIterator = copy.iterator();

                              while (listenerIterator.hasNext())
                              {
                                   ConnectionFactoryEventListener listener = (ConnectionFactoryEventListener)listenerIterator.next();
                                   listener.handleConnectionFactoryEvent(event);
                              }
                         }
                    };

                    final Thread t = BaseSecuritySupport.securitySupport.createThread(r);

                    PrivilegedAction action = new PrivilegedAction()
                    {
                         public Object run()
                         {
                              t.setName("com.sas.services.connection.ConnectionFactoryKernel.event_broadcaster"/*I18NOK:EMS*/);
                              return null;
                         }

                    };

                    BaseSecuritySupport.securitySupport.doThreadPrivileged(action);
                    t.start();
               }
          }
     }

     protected void shutdownKernel(ConnectionFactoryEvent event)
     {
          shutdown = true;
          ConnectionFactoryManager cxfManager = getManager();
          cxfManager.removeKernel(this);
          broadcastEvent(event);
     }

     public void destroyKernel()
     {
          destroyed = true;

          if (!shutdown)
          {
               ConnectionFactoryManager cxfManager = getManager();
               cxfManager.removeKernel(this);
          }

          ConnectionFactoryEvent event = new ConnectionFactoryEvent(ConnectionFactoryEvent.TYPE_DESTROYED);
          broadcastEvent(event);

          synchronized (listenerSet)
          {
               listenerSet.clear();
          }

          ConnectionFactoryConfiguration cxfConfig = getConfig();
          cxfConfig.kernelDestroyed(this);
     }

     protected boolean isShutdown()
     {
          return shutdown;
     }

     protected boolean isDestroyed()
     {
          return destroyed;
     }

     protected Logger getMonitor()
     {
          return monitor;
     }

     protected void setMonitor(Logger monitor)
     {
          this.monitor = monitor;
     }

     protected void removeUserFromCache(Credential user)
     {
     }
     
     protected Logger getLogger(ConnectionEnvelope cxEnv)
     {
          Logger logger = null;
          ObjectImpl root = (ObjectImpl)cxEnv.getRootObject();
          
          if (root != null)
          {
               Delegate delegate = root._get_delegate();
          
               if (delegate instanceof DelegateBase)
               {
                    ProtocolEngineInterface engine = ((DelegateBase)delegate).getProtocolEngine();
                    logger = engine.getLogger();     
               }
          }
          
          if (logger == null)
          {
               logger = getLogger();
          }
          
          return logger;         
     }

     protected abstract ConnectionInterface getConnection(Credential userLogin,long waitMilliseconds,Logger cxLogger) throws ConnectionFactoryException;

     protected abstract ConnectionFactoryAdminInterface getAdminInterface(Credential adminLogin) throws ConnectionFactoryException;

     protected abstract List getDomains();

     protected abstract void reactivateConnection(ConnectionEnvelope cxEnv,int clientNumber,int reason);
}