package com.sas.services.connection;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.logging.Logger;
//
// History:
//
// Date    Name            Remarks
// 26Nov01 B Beers         S0133039: Created.
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 29May02 B Beers         S0139445: Logging service integration (happy 9th anniversary Biff & Julie).
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 17Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring (phase 2).
// 21Feb03 B Beers         S0185859: Change handling of factory name to prevent factory proliferation.
// 24Feb03 B Beers         S0186157: Fix bad conditional in getFactoryLoggerName.
// 14Mar03 B Beers         S0183680: Shut down factory when connection to repository is lost.
// 15May03 B Beers         S0200797: Add API to control application monitoring.
//
/**
 * A factory of connection factories. This class creates
 * {@link com.sas.services.connection.ConnectionFactoryInterface connection factories} that match
 * a desired {@link com.sas.services.connection.ConnectionFactoryConfiguration configuration}.
 * @author Biff Beers
 */
public class ConnectionFactoryManager extends LoggerFactory
{
     /**
      * Construct a new connection factory manager.
      */
     public ConnectionFactoryManager()
     {
     }

     /**
      * Get the name of the logger that will be used by the factory
      * created by this manager using the specified configuration.
      * @param cxfConfig the desired configuration. Cannot be null.
      * @return the name of the logger that will be used by the factory
      * created by this manager using the specified configuration
      */
     public final String getFactoryLoggerName(ConnectionFactoryConfiguration cxfConfig)
     {
          synchronized (kernelMap)
          {
               String loggerName = (String)loggerMap.get(cxfConfig);

               if (loggerName != null)
               {
                    return loggerName;
               }

               int hc = cxfConfig.hashCode();

               if (!loggerMap.isEmpty())
               {
                    while (collision(hc))
                    {
                         hc++;
                    }
               }

               StringBuffer loggerNameBuffer = new StringBuffer(PARENT_LOGGER_NAME.length() + 25);
               loggerNameBuffer.append(PARENT_LOGGER_NAME).append(".f");
               int mhc = hashCode();

               if (mhc < 0)
               {
                    loggerNameBuffer.append(0);
                    mhc = -mhc;
               }

               loggerNameBuffer.append(mhc).append("-");

               if (hc < 0)
               {
                    loggerNameBuffer.append(0);
                    hc = -hc;
               }

               loggerNameBuffer.append(hc);
               loggerName = loggerNameBuffer.toString();
               loggerMap.put(cxfConfig,loggerName);
               return loggerName;
          }
     }

     /**
      * Creates a connection factory that matches a desired configuration. The factory will be managed by this
      * connection factory manager. If such a connection
      * factory already exists in this manager's factory set, and if it has not been shutdown
      * or destroyed, a reference to it will be returned instead.
      * @param cxfConfig the desired configuration. Cannot be null.
      * @return a connection factory matching the desired configuration
      * @throws com.sas.services.connection.ConnectionFactoryException if the configuration is invalid
      */
     public final ConnectionFactoryInterface getFactory(ConnectionFactoryConfiguration cxfConfig) throws ConnectionFactoryException
     {
          ConnectionFactory shell = new ConnectionFactory();
          activateShell(shell,cxfConfig,this);
          return shell;
     }

     /**
      * Get the name of the logger that will be used by the factory
      * created by the default manager using the specified configuration.
      * @param cxfConfig the desired configuration. Cannot be null.
      * @return the name of the logger that will be used by the factory
      * created by the default manager using the specified configuration
      */
     public static String getConnectionFactoryLoggerName(ConnectionFactoryConfiguration cxfConfig)
     {
          return defaultManager.getFactoryLoggerName(cxfConfig);
     }

     /**
      * Creates a connection factory that matches a desired configuration. The factory will be managed by the
      * default connection factory manager. If such a connection
      * factory already exists in the default manager's factory set, and if it has not been shutdown
      * or destroyed, a reference to it will be returned instead.
      * @param cxfConfig the desired configuration. Cannot be null.
      * @return a connection fatory matching the desired configuration
      * @throws com.sas.services.connection.ConnectionFactoryException if the configuration is invalid
      */
     public static ConnectionFactoryInterface getConnectionFactory(ConnectionFactoryConfiguration cxfConfig) throws ConnectionFactoryException
     {
          return defaultManager.getFactory(cxfConfig);
     }

     /**
      * The name of the logger that is the parent of all connection
      * factory loggers.
      */
     public static final String PARENT_LOGGER_NAME = "com.sas.services.connection"/*I18NOK:EMS*/;

     /**
      * The default connection factory manager.
      */
     public static ConnectionFactoryManager defaultManager = new ConnectionFactoryManager();
     private Map kernelMap = new HashMap();
     private Map loggerMap = new WeakHashMap();

     protected final void activateShell(ConnectionFactoryShell shell,ConnectionFactoryConfiguration cxfConfig,LoggerFactory loggerFactory) throws ConnectionFactoryException
     {
          ConnectionFactoryKernel kernel = null;

          synchronized (kernelMap)
          {
               kernel = (ConnectionFactoryKernel)kernelMap.get(cxfConfig);

               if (kernel == null)
               {
                    Logger logger = loggerFactory.getLogger(cxfConfig);
                    kernel = cxfConfig.getKernel(this,logger);
                    kernelMap.put(cxfConfig,kernel);
               }
          }

          shell.setKernel(kernel);
     }

     protected final void removeKernel(ConnectionFactoryKernel kernel)
     {
          synchronized (kernelMap)
          {
               ConnectionFactoryConfiguration cxfConfig = kernel.getConfig();
               kernelMap.remove(cxfConfig);
          }
     }

     protected Logger getLogger(ConnectionFactoryConfiguration cxfConfig) throws ConnectionFactoryException
     {
          String loggerName = getFactoryLoggerName(cxfConfig);
          Logger logger = Logger.getLogger(loggerName);
          return logger;
     }

     private boolean collision(int hc)
     {
          Set cSet = loggerMap.keySet();
          Iterator cItor = cSet.iterator();

          while (cItor.hasNext())
          {
               ConnectionFactoryConfiguration c = (ConnectionFactoryConfiguration)cItor.next();
               int chc = c.hashCode();

               if (hc == chc)
               {
                    return true;
               }
          }

          return false;
     }
}