package com.sas.services.connection;

import java.util.logging.Logger;
//
// History:
//
// Date    Name            Remarks
// 26Nov01 B Beers         S0133039: Created.
// 03Dec01 B Beers         S0133638: Add JDNI configuration.
// 02Jan02 B Beers         S0135536: Add admin auth feature.
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 07May02 B Beers         S0150375: Added Info Service configuration.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 29May02 B Beers         S0139445: Logging service integration (happy 9th anniversary Biff & Julie).
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 18Oct02 B Beers         S0168379: Removed deprecated methods for previous change.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 21Feb03 B Beers         S0185859: Change handling of factory name to prevent factory proliferation.
//
/**
 * A connection factory configuration in which the servers, clusters, and puddles
 * that the factory is to connect to are explicitly described.
 * @author Biff Beers
 */
public final class ManualConnectionFactoryConfiguration extends ConnectionFactoryConfiguration
{
     /**
      * Create a configuration for an connection factory that creates
      * and manages connections to the specified server. The server object
      * is cloned immediately, and all connection factory operations
      * use the internal copy. Changes to the original server object
      * after construction of this configuration will not affect the
      * behavior of the factory.
      * @param server the server you want to make connections to
      */
     public ManualConnectionFactoryConfiguration(Server server)
     {
          this(new FailoverCluster(server));
     }

     /**
      * Create a configuration for an connection factory that creates
      * and manages connections to the specified server cluster. The cluster object
      * is cloned immediately, and all connection factory operations
      * use the internal copy. Changes to the original cluster object
      * after construction of this configuration will not affect the
      * behavior of the factory.
      * @param cluster the server cluster you want to make connections to
      */
     public ManualConnectionFactoryConfiguration(Cluster cluster)
     {
          this(new Cluster[]{cluster});
     }

     /**
      * Create a configuration for an connection factory that creates
      * and manages connections to the specified server clusters. The cluster array
      * is cloned immediately, and all connection factory operations
      * use the internal copy. Changes to the original cluster array
      * after construction of this configuration will not affect the
      * behavior of the factory.
      * @param clusters the server clusters you want to make connections to
      */
     public ManualConnectionFactoryConfiguration(Cluster[] clusters)
     {
          if (clusters == null || clusters.length == 0)
          {
               throw new IllegalArgumentException("\"clusters\" cannot be null or empty.");
          }

          for (int i=0;i<clusters.length;i++)
          {
               if (clusters[i] == null)
               {
                    throw new IllegalArgumentException("Element " + i + " of \"clusters\" cannot be null.");
               }
          }

          this.clusters = ConnectionFactoryUtil.copyClusterArray(clusters);
     }

     /**
      * Create a configuration for an connection factory that creates
      * and manages connections to the specified puddle. The puddle object
      * is cloned immediately, and all connection factory operations
      * use the internal copy. Changes to the original puddle object
      * after construction of this configuration will not affect the
      * behavior of the factory.
      * @param puddle the puddle you want to make connections to
      */
     public ManualConnectionFactoryConfiguration(Puddle puddle)
     {
          this(new Puddle[]{puddle});
     }

     /**
      * Create a configuration for an connection factory that creates
      * and manages connections to the specified puddles. The puddle array
      * is cloned immediately, and all connection factory operations
      * use the internal copy. Changes to the original puddle array
      * after construction of this configuration will not affect the
      * behavior of the factory.
      * @param puddles the puddles you want to make connections to
      */
     public ManualConnectionFactoryConfiguration(Puddle[] puddles)
     {
          if (puddles == null || puddles.length == 0)
          {
               throw new IllegalArgumentException("\"puddles\" cannot be null or empty.");
          }

          for (int i=0;i<puddles.length;i++)
          {
               if (puddles[i] == null)
               {
                    throw new IllegalArgumentException("Element " + i + " of \"puddles\" cannot be null.");
               }
          }

          this.puddles = ConnectionFactoryUtil.copyPuddleArray(puddles);
     }

     /**
      * Get the server clusters that factories with this
      * configuration connect to.
      * @return the server clusters that factories with this
      * configuration connect to
      */
     public Cluster[] getClusterArray()
     {
          return ConnectionFactoryUtil.copyClusterArray(clusters);
     }

     /**
      * Get a copy of the puddles that factories with this
      * configuration connect to.
      * @return a copy of the puddles that factories with this
      * configuration connect to
      */
     public Puddle[] getPuddleArray()
     {
          return ConnectionFactoryUtil.copyPuddleArray(puddles);
     }

     public boolean equals(Object that)
     {
          if (!super.equals(that))
          {
               return false;
          }

          ManualConnectionFactoryConfiguration thatConfig = (ManualConnectionFactoryConfiguration)that;
          Cluster[] thoseClusters = thatConfig.getClustersNoCopy();

          if (clusters != thoseClusters)
          {
               if (clusters == null || thoseClusters == null || clusters.length != thoseClusters.length)
               {
                    return false;
               }

               for (int i=0;i<clusters.length;i++)
               {
                    if (!clusters[i].equals(thoseClusters[i]))
                    {
                         return false;
                    }
               }
          }

          Puddle[] thosePuddles = thatConfig.getPuddlesNoCopy();

          if (puddles != thosePuddles)
          {
               if (puddles == null || thosePuddles == null || puddles.length != thosePuddles.length)
               {
                    return false;
               }

               for (int i=0;i<puddles.length;i++)
               {
                    if (!puddles[i].equals(thosePuddles[i]))
                    {
                         return false;
                    }
               }
          }

          return true;
     }

     public int hashCode()
     {
          int hc = super.hashCode();

          if (clusters != null)
          {
               for (int i=0;i<clusters.length;i++)
               {
                    hc ^= clusters[i].hashCode();
               }
          }
          else
          {
               for (int i=0;i<puddles.length;i++)
               {
                    hc ^= puddles[i].hashCode();
               }
          }

          return hc;
     }

     private Cluster[] clusters = null;
     private Puddle[] puddles = null;

     protected ConnectionFactoryKernel getKernel(ConnectionFactoryManager manager,ConnectionFactoryConfiguration cxfConfig,Logger logger) throws ConnectionFactoryException
     {
          if (clusters != null)
          {
               ConnectionFactoryKernel kernel = new AggregationKernel(manager,cxfConfig,logger,clusters);
               return kernel;
          }
          else
          {
               PuddleEnvelope[] envs = new PuddleEnvelope[puddles.length];

               for (int i=0;i<puddles.length;i++)
               {
                    envs[i] = new PuddleEnvelope(puddles[i]);
               }

               AuthServer authServer = new AuthServer(envs,true);
               return new PoolingKernel(manager,cxfConfig,logger,envs,authServer);
          }
     }

     protected Cluster[] getClustersNoCopy()
     {
          return clusters;
     }

     protected Puddle[] getPuddlesNoCopy()
     {
          return puddles;
     }
}