/* $Id$ */
package com.sas.services.connection;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.omg.CORBA.portable.Delegate;
import org.omg.CORBA.portable.ObjectImpl;

import com.sas.iom.SAS.IWorkspace;
import com.sas.iom.SAS.IWorkspaceHelper;
import com.sas.net.brg.orb.DelegateBase;
import com.sas.net.brg.orb.ProtocolEngineInterface;
//
// History:
//
// Date    Name            Remarks
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 30Apr02 B Beers         S0149671: Resolve some garbage collection problems.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 14Apr02 B Beers         S0149671: Resolve garbage collection problem for pooling.
// 20May02 B Beers         S0152574: Add load balancing to puddle selection.
// 29May02 B Beers         S0139445: Logging service integration (happy 9th anniversary Biff & Julie).
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 29Jan03 B Beers         S0175552: Check for minSize > sum of maxClients.
// 14Mar03 B Beers         S0183680: Shut down factory when connection to repository is lost.
// 21Mar03 B Beers         S0174314: Close logger handlers when factory is destroyed.
// 06Jun03 B Beers         S0204214: Streamline auth checking for pooling factories.
// 25Jun03 B Beers         S0208706: Remove remote calls from synch'd code in getConnection().
// 29Jan04 B Beers         S0233579: Use logger associated with user that created factory.
// 18Feb04 B Beers         S0238566: Clear reference to ConnectionImpl when returning connection.
/**
 * Implementation class for ConnectionFactoryKernel which manages pooled connections.
 * @author Biff Beers
 */
public final class PoolingKernel extends ConnectionFactoryKernel
{
     /**
      * Construct a pooling kernel.
      * @param manager the factory manager causing this factory to be constructed
      * @param cxfConfig the factory configuration for this factory
      * @param loggingImpl the log
      * @param envs the envelopes wrapping the puddles to which this factory will connect
      * @param authServer an object that handles authentication and authorization for the factory
      * @throws java.lang.IllegalArgumentException if envs is null, empty, or
      * has null elements, or if authServer is null
      */
     public PoolingKernel(ConnectionFactoryManager manager,ConnectionFactoryConfiguration cxfConfig,Logger logger,PuddleEnvelope[] envs,AuthServer authServer)
     {
          super(manager,cxfConfig,logger);

          if (envs == null || envs.length == 0)
          {
               throw new IllegalArgumentException("\"envs\" cannot be null or empty.");
          }

          for (int i=0;i<envs.length;i++)
          {
               PuddleEnvelope env = envs[i];

               if (env == null)
               {
                    throw new IllegalArgumentException("Element " + i + " of \"envs\" cannot be null.");
               }

               Puddle puddle = env.getPuddle();
               int minSize = puddle.getMinSize();
               Cluster[] clusters = puddle.getClusterArray();
               int globalMaxClients = 0;

               for (int j=0;j<clusters.length;j++)
               {
                    Cluster cluster = clusters[j];
                    int maxClients = cluster.getMaxClients();
                    globalMaxClients += maxClients;
               }

               if (globalMaxClients < minSize)
               {
                    throw new IllegalArgumentException("\"minSize\" of puddle " + i + " must be less than the sum of \"maxClients\" over all clusters in the puddle.");
               }

               env.setLock(this);
               env.setPuddleIndex(i);
               env.setLogger(logger);
          }

          if (authServer == null)
          {
               throw new IllegalArgumentException("\"authServer\" cannot be null.");
          }

          domains.addAll(authServer.getDomains());
          this.envs = envs;
          this.authServer = authServer;

          for (int i=0;i<envs.length;i++)
          {
               envs[i].init();
          }
     }

     public synchronized void shutdown()
     {
          if (isShutdown() || isDestroyed())
          {
               return;
          }

          ConnectionFactoryEvent event = new ConnectionFactoryEvent(ConnectionFactoryEvent.TYPE_SHUTDOWN);
          shutdownKernel(event);
     }

     public synchronized void destroy()
     {
          if (isDestroyed())
          {
               return;
          }

          Logger logger = getLogger();
          logger.info("commencing factory destruction");

          for (int i=0;i<envs.length;i++)
          {
               envs[i].destroy();
          }

          destroyKernel();
     }

     private PuddleEnvelope[] envs = null;
     private AuthServer authServer = null;
     private List domains = new ArrayList();
     private int requestCounter = 0;

     protected ConnectionInterface getConnection(Credential userLogin,long waitMilliseconds,Logger cxLogger) throws ConnectionFactoryException
     {
          if (userLogin == null)
          {
               throw new IllegalArgumentException("\"userLogin\" cannot be null.");
          }

          int[] authPuddles = authServer.authorizeUser(userLogin);
          ConnectionEnvelope cxEnv = null;
          int clientNumber = 0;
          Logger logger = (cxLogger == null) ? getLogger() : cxLogger;

          synchronized (this)
          {
               if (isShutdown() || isDestroyed())
               {
                    throw new IllegalStateException("\"getConnection()\" cannot be called on a factory after \"shutdown()\" or \"destroy()\" has been called.");
               }

               int rq = requestCounter++;
               logger.info("connection request rq#" + rq + " received with " + userLogin.toString());
               int puddleIndex = authServer.findBestPuddle(authPuddles);
               logger.info("rq#" + rq + " routed to puddle pd#" + puddleIndex);
               PuddleEnvelope env = envs[puddleIndex];

               try
               {
                    cxEnv = env.getConnection(waitMilliseconds,cxLogger);
                    cxEnv.setUserCredential(userLogin);
                    cxEnv.setPuddleIndex(puddleIndex);
               }
               catch (FatalConnectionFactoryException e)
               {
                    ConnectionFactoryEvent event = null;
                    event = new ConnectionFactoryEvent(ConnectionFactoryEvent.TYPE_FATAL_EXCEPTION,e);
                    shutdownKernel(event);
                    throw e;
               }

               clientNumber = cxEnv.getNextClientNumber();
               logger.info("rq#" + rq + " served by pooled/shared cx#" + cxEnv.getConnectionNumber() + " from pd#" + puddleIndex);
          }

          // reference management to get the garbage collector to behave
          ConnectionImpl cx = new ConnectionImpl(cxEnv,this,clientNumber);
          ObjectImpl obj = (ObjectImpl)cxEnv.getUserObject();
          Delegate delegate = obj._get_delegate();

          if (delegate instanceof DelegateBase)
          {
               DelegateBase delegateBase = (DelegateBase)delegate;
               delegateBase.addParent(cx);
               ProtocolEngineInterface engine = delegateBase.getProtocolEngine();
               engine.setLogger(logger);
          }
                    
          org.omg.CORBA.Object rootObject = cxEnv.getRootObject();

          if (obj == null || obj == rootObject)
          {
               try
               {
                    IWorkspace rootWorkspace = IWorkspaceHelper.narrow(rootObject);

                    if (rootWorkspace != rootObject)
                    {
                         cxEnv.setRootObject(rootWorkspace);
                    }

                    IWorkspace userWorkspace = rootWorkspace.GetNewWorkspace();
                    cxEnv.setUserObject(userWorkspace);
               }
               catch (Exception e)
               {
                    reactivateConnection(cxEnv,clientNumber,ConnectionFactoryKernel.REACT_CX_REFRESH);
                    return getConnection(userLogin,waitMilliseconds,cxLogger);
               }
          }
                    
          return cx;
     }

     protected ConnectionFactoryAdminInterface getAdminInterface(Credential adminLogin) throws ConnectionFactoryException
     {
          authServer.authorizeAdmin(adminLogin);
          return this;
     }

     protected List getDomains()
     {
          return ConnectionFactoryUtil.copyList(domains);
     }

     protected void reactivateConnection(ConnectionEnvelope cxEnv,int clientNumber,int reason)
     {
          // close the user object
          org.omg.CORBA.Object userObject = cxEnv.getUserObject();

          if (userObject != null && userObject != cxEnv.getRootObject())
          {
               cxEnv.setUserObject(null);
               IWorkspace userWorkspace = IWorkspaceHelper.narrow(userObject); // this is usually a no-op

               try
               {
                    userWorkspace.Close();
               }
               catch (Exception e)
               {
               }
          }
          
          synchronized (this)
          {
               if (isDestroyed())
               {
                    return;
               }

               PuddleEnvelope env = cxEnv.getPuddleEnvelope();
               env.reactivateConnection(cxEnv,clientNumber,reason);
               ObjectImpl root = (ObjectImpl)cxEnv.getRootObject();
               Delegate delegate = root._get_delegate();
          
               if (delegate instanceof DelegateBase)
               {
                    DelegateBase delegateBase = (DelegateBase)delegate;
                    ProtocolEngineInterface engine = delegateBase.getProtocolEngine();               
                    Logger logger = getLogger();
                    engine.setLogger(logger);
                    delegateBase.clearParents();
               }

               if (isShutdown() && !isDestroyed())
               {
                    int totalClients = 0;

                    for (int i=0;i<envs.length;i++)
                    {
                         totalClients += envs[i].getClients();
                    }

                    if (totalClients == 0)
                    {
                         destroy();
                    }
               }
          }
     }

     protected synchronized void shutdownKernel(ConnectionFactoryEvent event)
     {
          Logger logger = getLogger();

          if (event.getType() == ConnectionFactoryEvent.TYPE_SHUTDOWN)
          {
               logger.info("factory shutdown commencing");
          }
          else
          {
               logger.info("fatal error detected; factory shutting down");
          }

          for (int i=0;i<envs.length;i++)
          {
               envs[i].shutdown();
          }

          super.shutdownKernel(event);

          if (!isDestroyed())
          {
               int totalClients = 0;

               for (int i=0;i<envs.length;i++)
               {
                    totalClients += envs[i].getClients();
               }

               if (totalClients == 0)
               {
                    destroy();
               }
          }
     }

     protected void removeUserFromCache(Credential user)
     {
          authServer.removeUserFromCache(user);
     }
}