/* $Id: Puddle.java,v 1.3.16.1 2005/10/11 17:41:25 sasbfb Exp $ */
package com.sas.services.connection;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
//
// History:
//
// Date    Name            Remarks
// 26Nov01 B Beers         S0133039: Created.
// 03Dec01 B Beers         S0133638: Add JDNI configuration.
// 02Jan02 B Beers         S0135536: Add admin auth feature.
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 20May02 B Beers         S0152574: Add load balancing to puddle selection.
// 10Jun02 B Beers         S0155031: Correct load balancing for requests on separate threads.
// 27Sep02 B Beers         S0166941: Disallow illegal log level values.
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 15Oct02 B Beers         S0168667: Complete implementation of clone.
// 04Dec03 B Beers         S0228761: Tighten up user authorization queries.
// 11Oct05 B Beers         S0324947: Added serialVersionUID.
/**
 * A description of a puddle of IOM clusters. A puddle is an association of one or more
 * IOM clusters with exactly one IOM login and, optionally, with a list of users allowed
 * to use connections to this puddle of clusters.
 * @author Biff Beers
 */
public class Puddle implements Cloneable,Serializable
{
     /**
      * Construct a puddle with one IOM server.
      * This constructor is a shortcut for the following more
      * complicated code:
      * <PRE>
      * Puddle(new Cluster[]{new FailoverCluster(server)},serverCredential)
      * </PRE>
      * @param server the server in the puddle
      * @param serverCredential the credential to use when connecting to the server
      */
     public Puddle(Server server,Credential serverCredential)
     {
          this(new FailoverCluster(server),serverCredential);
     }

     /**
      * Construct a puddle with one IOM server cluster.
      * This constructor is a shortcut for the following more
      * complicated code:
      * <PRE>
      * Puddle(new Cluster[]{cluster},credential)
      * </PRE>
      * @param cluster the server cluster in the puddle
      * @param serverCredential the credential to use when connecting to a server in
      * the cluster
      */
     public Puddle(Cluster cluster,Credential serverCredential)
     {
          this(new Cluster[]{cluster},serverCredential);
     }

     /**
      * Construct a puddle of IOM server clusters.
      * @param clusters the server clusters in the puddle
      * @param serverCredential the credential to use when connecting to a server in
      * one of the clusters
      */
     public Puddle(Cluster[] clusters,Credential serverCredential)
     {
          if (clusters == null || clusters.length == 0)
          {
               throw new IllegalArgumentException("\"clusters\" cannot be null or empty.");
          }

          final int len = clusters.length;

          for (int i=0;i<len;i++)
          {
               if (clusters[i] == null)
               {
                    throw new IllegalArgumentException("Element " + i + " of \"clusters\" cannot be null.");
               }
          }

          classID = clusters[0].getClassID();

          for (int i=1;i<len;i++)
          {
               if (!classID.equals(clusters[i].getClassID()))
               {
                    StringBuffer msgBuffer = new StringBuffer("All clusters in a puddle must have the same class ID. ");
                    msgBuffer.append("Cluster 0 has class ID \"").append(classID).append("\", but cluster ");
                    msgBuffer.append(i).append(" has classID \"").append(clusters[i].getClassID()).append("\".");
                    throw new IllegalArgumentException(msgBuffer.toString());
               }
          }

          if (serverCredential != null)
          {
               String credentialDomain = serverCredential.getDomain();

               for (int i=0;i<len;i++)
               {
                    String serverDomain = clusters[i].getDomain();

                    if (!credentialDomain.equals(serverDomain))
                    {
                         throw new IllegalArgumentException("The domain for element " + i + " of \"clusters\": \"" + serverDomain + "\" does not match the domain \"" + credentialDomain +"\" of the server credential.");
                    }
               }
          }

          this.clusters = clusters;
          this.serverCredential = serverCredential;
     }

     /**
      * Set the server clusters.
      * @param clusterArray the clusters
      */
     public void setClusterArray(Cluster[] clusterArray)
     {
          this.clusters = clusterArray;
     }

     /**
      * Get the server clusters.
      * @return the clusters
      */
     public Cluster[] getClusterArray()
     {
          return clusters;
     }

     /**
      * Set the credential for the clusters.
      * @param serverCredential the credential for the clusters
      */
     public void setServerCredential(Credential serverCredential)
     {
          this.serverCredential = serverCredential;
     }

     /**
      * Get the credential for the clusters.
      * @return the credential for the clusters
      */
     public Credential getServerCredential()
     {
          return serverCredential;
     }

     /**
      * Set the minimum number of connections (idle or in use).
      * @param minSize the minimum number of connections (idle or in use)
      */
     public void setMinSize(int minSize)
     {
          this.minSize = minSize;
     }

     /**
      * Get the minimum number of connections (idle or in use).
      * @return the minimum number of connections (idle or in use)
      */
     public int getMinSize()
     {
          return minSize;
     }

     /**
      * Set the minimum number of idle connections.
      * @param minAvail the minimum number of idle connections
      */
     public void setMinAvail(int minAvail)
     {
          this.minAvail = minAvail;
     }

     /**
      * Get the minimum number of idle connections.
      * @return the minimum number of idle connections
      */
     public int getMinAvail()
     {
          return minAvail;
     }

     /*
      * Set the set of credentials representing users that are allowed to use this puddle.
      * @param userCredentials the set of credentials representing users
      * that are allowed to use this puddle
      */
     public void setUserCredentials(Set userCredentials)
     {
          Iterator userCredentialsItor = userCredentials.iterator();

          while (userCredentialsItor.hasNext())
          {
               Object o = userCredentialsItor.next();

               if (!(o instanceof Credential))
               {
                    throw new IllegalArgumentException("All elements of \"userCredentials\" must be instances of com.sas.services.connection.Credential.");
               }
          }

          this.userCredentials = userCredentials;
     }

     /*
      * Get the set of credentials representing users that are allowed to use this puddle.
      * If both this method and {@link #getAuthorizedUsers()
      * getAuthorizedUsers()} return <CODE>null</CODE>, then any
      * user will be allowed to use this puddle.
      * @return the set of credentials representing users that are allowed to use this puddle
      */
     public Set getUserCredentials()
     {
          return userCredentials;
     }

     /*
      * Set the set of users that are allowed to use this puddle.
      * @param users the set of users that are allowed to use this puddle
      */
     public void setUsers(Set users)
     {
          this.users = users;
     }

     /*
      * Get the set of users that are allowed to use this puddle.
      * If both this method and {@link #getUserCredentials()
      * getUserCredentials()} return <CODE>null</CODE>, then any
      * user will be allowed to use this puddle.
      * @return the set of users that are allowed to use this puddle
      */
     public Set getUsers()
     {
          return users;
     }

     /**
      * Set a string that identifies the puddle to the
      * {@link com.sas.services.connection.AuthServer authentication/authorization
      * server} implementation provided to the
      * {@link com.sas.services.connection.ConnectionFactoryConfiguration
      * connection factory configuration} for the purpose of authorizing factory
      * administrators. Factory administration is a feature provided only for
      * compatibility with previous releases. It is rarely used. If <CODE>null</CODE>,
      * then this puddle will play no part in determining if a user is authorized
      * to administer a factory.
      * @param puddleID a string that identifies the puddle
      */
     public void setPuddleID(String puddleID)
     {
          this.puddleID = puddleID;
     }

     /**
      * Get a string that identifies the puddle to the
      * {@link com.sas.services.connection.AuthServer authentication/authorization
      * server} implementation provided to the
      * {@link com.sas.services.connection.ConnectionFactoryConfiguration
      * connection factory configuration} for the purpose of authorizing factory
      * administrators. Factory administration is a feature provided only for
      * compatibility with previous releases. It is rarely used. If <CODE>null</CODE>,
      * then this puddle will play no part in determining if a user is authorized
      * to administer a factory.
      * @return a string that identifies the puddle
      */
     public String getPuddleID()
     {
          return puddleID;
     }

     /**
      * Set a string that names the puddle. This name is used by the
      * {@link com.sas.services.connection.AuthServer authentication/authorization
      * server} to determine if special authorization rules apply.
      * @param puddleName a string the names the puddle
      */
     public void setPuddleName(String puddleName)
     {
          this.puddleName = puddleName;
     }
     
     /**
      * Get a string that names the puddle. This name is used by the
      * {@link com.sas.services.connection.AuthServer authentication/authorization
      * server} to determine if special authorization rules apply.      * @return a string the names the puddle
      */
     public String getPuddleName()
     {
          return puddleName;
     }
     
     /**
      * Get the classID of the clusters in this puddle.
      * All clusters must have the same classID.
      * @return the classID of the clusters in this puddle
      */
     public String getClassID()
     {
          return classID;
     }

     public boolean equals(Object that)
     {
          if (this == that)
          {
               return true;
          }

          if (that == null || !getClass().getName().equals(that.getClass().getName()))
          {
               return false;
          }

          Puddle thatPuddle = (Puddle)that;

          if (minSize != thatPuddle.getMinSize() || minAvail != thatPuddle.getMinAvail())
          {
               return false;
          }

          Cluster[] thoseClusters = thatPuddle.getClusterArray();

          if (clusters != thoseClusters)
          {
               if (clusters.length != thoseClusters.length)
               {
                    return false;
               }

               for (int i=0;i<clusters.length;i++)
               {
                    if (!clusters[i].equals(thoseClusters[i]))
                    {
                         return false;
                    }
               }
          }

          Credential thatServerCredential = thatPuddle.getServerCredential();

          if (serverCredential == null ? thatServerCredential != null : !serverCredential.equals(thatServerCredential))
          {
               return false;
          }

          Set thoseUserCredentials = thatPuddle.getUserCredentials();

          if (userCredentials == null ? thoseUserCredentials != null : !userCredentials.equals(thoseUserCredentials))
          {
               return false;
          }

          Set thoseUsers = thatPuddle.getUsers();

          if (users == null ? thoseUsers != null : !users.equals(thoseUsers))
          {
               return false;
          }

          String thatPuddleID = thatPuddle.getPuddleID();

          if (puddleID == null ? thatPuddleID != null : !puddleID.equals(thatPuddleID))
          {
               return false;
          }

          return true;
     }

     public int hashCode()
     {
          int hashCode = getClass().getName().hashCode();

          for (int i=0;i<clusters.length;i++)
          {
               hashCode ^= clusters[i].hashCode();
          }

          if (serverCredential != null)
          {
               hashCode ^= serverCredential.hashCode();
          }

          hashCode ^= minSize;
          hashCode ^= minAvail;

          if (userCredentials != null)
          {
               hashCode ^= userCredentials.hashCode();
          }

          if (users != null)
          {
               hashCode ^= users.hashCode();
          }

          if (puddleID != null)
          {
               hashCode ^= puddleID.hashCode();
          }

          return hashCode;
     }

     public Object clone()
     {
          Puddle puddle = null;

          try
          {
               puddle = (Puddle)super.clone();
          }
          catch (CloneNotSupportedException e)
          {
               throw new InternalError("A Cloneable object threw CloneNotSupportedException.");
          }

          Cluster[] clusterArray = puddle.getClusterArray();
          clusterArray = ConnectionFactoryUtil.copyClusterArray(clusterArray);
          puddle.setClusterArray(clusterArray);
          Credential credential = puddle.getServerCredential();
          credential = (Credential)credential.clone();
          puddle.setServerCredential(credential);

          if (userCredentials != null)
          {
               final int size = userCredentials.size();
               Set newSet = new HashSet(size);
               Iterator userCredItor = userCredentials.iterator();

               while (userCredItor.hasNext())
               {
                    Credential cred = (Credential)userCredItor.next();
                    Object newCred = cred.clone();
                    newSet.add(newCred);
               }

               puddle.setUserCredentials(newSet);
          }

          if (users != null)
          {
               // don't need to copy elements of users set
               // because they are all immutable strings.
               final int size = users.size();
               Set newSet = new HashSet(size);
               newSet.addAll(users);
               puddle.setUsers(newSet);
          }

          return puddle;
     }

     /**
      * Serial version ID. 
      * This value should be changed if this version is not compatible with the previous version.
      */
     private static final long serialVersionUID = -9062810559912378857L;
     private Cluster[] clusters /* = null */;
     private Credential serverCredential /* = null */;
     private int minSize /* = 0 */;
     private int minAvail /* = 0 */;
     private Set userCredentials /* = null */;
     private Set users /* = null */;
     private String puddleID /* = null */;
     private String puddleName /* = null */;
     private String classID /* = null */;
}