/* $Id: PuddleEnvelope.java,v 1.11.4.3 2004/02/19 16:49:46 sasbfb Exp $ */
package com.sas.services.connection;

import java.lang.ref.WeakReference;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;
import org.omg.CORBA.ORB;
import org.omg.CORBA.portable.ObjectImpl;
import com.sas.iom.SAS.IWorkspace;
import com.sas.iom.SAS.IWorkspaceHelper;
import com.sas.security.BaseSecuritySupport;
import com.sas.text.Message;
//
// History:
//
// Date    Name            Remarks
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 26Nov02 B Beers         S0151508: Revise xml error message handling.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 21Feb03 B Beers         S0185859: Change handling of factory name to prevent factory proliferation.
// 14Mar03 B Beers         S0183680: Shut down factory when connection to repository is lost.
// 26Apr03 B Beers         S0198152: Delay factory destruction until pending connections resolve.
// 15May03 B Beers         S0200797: Add API to control application monitoring.
// 06Jun03 B Beers         S0204214: Streamline auth checking for pooling factories.
// 25Jun03 B Beers         S0208706: Remove remote calls from synch'd code in getConnection().
// 11Dec03 B Beers         S0230810: Check for capacity < minSize.
// 29Jan04 B Beers         S0233579: Use logger associated with user that created factory.
// 19Feb04 B Beers         S0238765: Clear cx from cxSet when destroying cx.
/**
 * A utility class for connection factory configurations.
 * @author Biff Beers
 */
public class PuddleEnvelope
{
     /**
      * Construct an envelope for a puddle.
      * @param puddle the puddle for which to make an envelope
      */
     public PuddleEnvelope(Puddle puddle)
     {
          Set userCredentials = puddle.getUserCredentials();

          if (userCredentials != null)
          {
               userCache = new HashSet(userCredentials.size());
               userCache.addAll(userCredentials);
          }
          else
          {
               if (puddle.getUsers() != null)
               {
                    userCache = new HashSet();
               }
          }

          this.puddle = puddle;
          String classID = puddle.getClassID();

          if (classID.equals(Server.CLSID_SAS) || classID.equals(Server.CLSID_SAS_1_1))
          {
               delegate = new PoolingDelegate();
          }
          else
          {
               //delegate = new SharingDelegate();
               throw new IllegalArgumentException("All servers in a puddle must be SAS Workspace servers.");
          }

          Cluster[] clusters = puddle.getClusterArray();
          clusterStatus = new boolean[clusters.length];

          for (int i=0;i<clusterStatus.length;i++)
          {
               clusterStatus[i] = true;
          }
     }

     public void addUserToCache(Credential user)
     {
          if (userCache != null)
          {
               synchronized (userCache)
               {
                    userCache.add(user);
               }
          }
     }

     public Puddle getPuddle()
     {
          return puddle;
     }

     public Set getUserCache()
     {
          return userCache;
     }

     private Puddle puddle /* = null */;
     private Set userCache /* = null */;
     private ConnectionFactoryKernel lock /* = null */;
     private int puddleIndex /* = 0 */;
     private String puddleName /* = null */;
     private Logger logger /* = null */;
     private int clusterIndex /* = 0 */;
     private boolean[] clusterStatus /* = null */;
     private Delegate delegate /* = null */;
     private int clients /* = 0 */;
     private int waiting /* = 0 */;
     private boolean shutdown /* = false */;

     private abstract class Delegate
     {
          protected abstract int capacity();
          protected abstract int available();
          protected abstract int pending();
          protected abstract int chooseCluster();
          protected abstract void createConnection(String cause);
          protected abstract ConnectionEnvelope getAvailableConnection(Logger cxLogger);
          protected abstract void reactivateConnection(ConnectionEnvelope cxEnv,int clientNumber,int reason);
          protected abstract void checkMinimums();
          protected abstract Throwable getCause();
          protected abstract void shutdown();
          protected abstract void destroy();
          protected abstract void cutServerStatusRecord(ConnectionEnvelope cxEnv,String cause);
     }

     private class PoolingDelegate extends Delegate
     {
          private PoolingDelegate()
          {
               Cluster[] clusters = puddle.getClusterArray();
               clusterUsage = new int[clusters.length];
          }

          private Set cxSet = new HashSet();
          private Set trackingSet = new HashSet();
          private List avail = new ArrayList();
          private int pend /* = 0 */;
          private int[] clusterUsage /* = null */;
          private Thread shutdownTimer /* = null */;
          private boolean shutdownTimerDone /* = false */;
          private Throwable cause /* = null */;

          private class ShutdownTimer implements Runnable
          {
               public void run()
               {
                    synchronized (lock)
                    {
                         while (!shutdownTimerDone)
                         {
                              long nowMillis = System.currentTimeMillis();
                              long nextCheck = Long.MAX_VALUE;
                              Iterator availItor = avail.iterator();

                              while (availItor.hasNext())
                              {
                                   ConnectionEnvelope cxEnv = (ConnectionEnvelope)availItor.next();
                                   long shutdownTime = cxEnv.getShutdownTime();

                                   if (shutdownTime < nowMillis)
                                   {
                                        availItor.remove();
                                        logger.info(puddleName + ": shutdown timer for cx#" + cxEnv.getConnectionNumber() + " has expired");
                                        destroyConnection(cxEnv,MonitorRecord.CAUSE_TIMEOUT);
                                   }
                                   else if (shutdownTime < nextCheck)
                                   {
                                        nextCheck = shutdownTime;
                                   }
                              }

                              try
                              {
                                   lock.wait(nextCheck - nowMillis);
                              }
                              catch (InterruptedException ie)
                              {
                                   return;
                              }
                         }
                    }
               }
          }

          // always called from synchronized code
          protected int capacity()
          {
               int cap = 0;
               Cluster[] clusters = puddle.getClusterArray();

               for (int i=0;i<clusters.length;i++)
               {
                    if (clusterStatus[i])
                    {
                         int maxClients = clusters[i].getMaxClients();

                         if (Integer.MAX_VALUE - maxClients < cap)
                         {
                              return Integer.MAX_VALUE;
                         }

                         cap += maxClients;
                    }
               }

               return cap;
          }

          // always called from synchronized code
          protected int available()
          {
               return avail.size();
          }

          // always called from synchronized code
          protected int pending()
          {
               return pend;
          }

          // always called from synchronized code
          protected void createConnection(final String cause)
          {
               final int clusterIndex = chooseCluster();
               final Cluster[] clusters = puddle.getClusterArray();
               final Credential serverCredential = puddle.getServerCredential();
               clusterUsage[clusterIndex]++;
               pend++;

               Runnable r = new Runnable()
               {
                    public void run()
                    {
                         ConnectionEnvelope cxEnv = clusters[clusterIndex].getConnection(serverCredential,logger);
                         
                         if (cxEnv.isBogus())
                         {
                              logger.info(puddleName + ": cx#" + cxEnv.getConnectionNumber() + " is invalid and will not be allocated to a user");
                         }
                         else
                         {
                              cxEnv.setClusterIndex(clusterIndex);
                              cxEnv.setPuddleEnvelope(PuddleEnvelope.this);
                              cxEnv.setOwnerCredential(serverCredential);
                              cxEnv.setPuddleIndex(puddleIndex);
                         }

                         synchronized (lock)
                         {
                              if (!cxEnv.isBogus())
                              {
                                   cxSet.add(cxEnv.getRootObject());
                                   trackingSet.add(new WeakReference(cxEnv));
                              }
                              else
                              {
                                   clusterStatus[clusterIndex] = false;
                              }

                              addAvail(cxEnv);
                              pend--;
                              cutServerStatusRecord(cxEnv,cause);
                         }
                    }
               };

               final Thread t = BaseSecuritySupport.securitySupport.createThread(r);

               PrivilegedAction action = new PrivilegedAction()
               {
                    public Object run()
                    {
                         t.setName("com.sas.services.connection.PuddleEnvelope.new_cx_creator"/*I18NOK:EMS*/);
                         return null;
                    }

               };

               BaseSecuritySupport.securitySupport.doThreadPrivileged(action);
               t.start();
          }

          // always called from synchronized code
          protected ConnectionEnvelope getAvailableConnection(Logger cxLogger)
          {
               while (avail.size() > 0)
               {
                    ConnectionEnvelope cxEnv = (ConnectionEnvelope)avail.remove(0);

                    if (cxLogger == null)
                    {
                         cxLogger = logger;                    
                    }
                    
                    cxLogger.info(puddleName + ": removing cx#" + cxEnv.getConnectionNumber() + " from available queue");

                    if (cxEnv.isBogus())
                    {
                         destroyConnection(cxEnv,null);
                         continue;
                    }

                    cxEnv.incrClientCount();
                    cxEnv.incrReactivationCount();
                    checkMinimums();
                    return cxEnv;
               }

               return null;
          }

          // always called from synchronized code
          protected void reactivateConnection(ConnectionEnvelope cxEnv,int clientNumber,int reason)
          {
               String whom = null;
               String cause = null;

               if (reason == ConnectionFactoryKernel.REACT_CX_CLOSED)
               {
                    whom = "user";
                    cause = MonitorRecord.CAUSE_RETURN;
               }
               else
               if (reason == ConnectionFactoryKernel.REACT_CX_REFRESH)
               {
                    whom = "factory (refresh failed)";
                    cause = MonitorRecord.CAUSE_REFRESH;
               }
               else
               {
                    whom = "garbage collector";
                    cause = MonitorRecord.CAUSE_GC;
               }

               Logger cxLogger = lock.getLogger(cxEnv);
               cxLogger.info(puddleName + ": cx#" + cxEnv.getConnectionNumber() + " returned to factory by " + whom);

               // close the user object if it hasn't already been closed (it probably has, but I'm paranoid)
               org.omg.CORBA.Object userObject = cxEnv.getUserObject();

               if (userObject != null && userObject != cxEnv.getRootObject())
               {
                    cxEnv.setUserObject(null);
                    IWorkspace userWorkspace = IWorkspaceHelper.narrow(userObject);

                    try
                    {
                         userWorkspace.Close();
                    }
                    catch (Exception e)
                    {
                    }
               }

               if (reason == ConnectionFactoryKernel.REACT_CX_REFRESH)
               {
                    destroyConnection(cxEnv,cause);
                    clients--;
                    return;
               }

               // check for factory shutdown
               if (shutdown)
               {
                    cxLogger.info(puddleName + ": cx#" + cxEnv.getConnectionNumber() + " will be destroyed due to factory shutdown");
                    destroyConnection(cxEnv,cause);
                    clients--;
                    return;
               }

               // check reactivation count
               int limit = cxEnv.getServer().getReactivationLimit();

               if (limit > 0)
               {
                    int reactCount = cxEnv.getReactivationCount();

                    if (reactCount >= limit)
                    {
                         logger.info(puddleName + ": cx#" + cxEnv.getConnectionNumber() + " has reached its reactivation limit");
                         destroyConnection(cxEnv,cause);
                         clients--;
                         checkMinimums();
                         return;
                    }
               }

               // put it back on the available queue
               clients--;
               cxEnv.decrClientCount();
               cxEnv.setUserCredential(null);
               addAvail(cxEnv);
               cutServerStatusRecord(cxEnv,cause);
          }

          // always called from synchronized code
          protected void checkMinimums()
          {
               int minSize = puddle.getMinSize();

               while (cxSet.size() + pend < minSize && cxSet.size() + pend < capacity())
               {
                    logger.info(puddleName + ": creating a connection to satisfy minSize");
                    createConnection(MonitorRecord.CAUSE_MIN_SIZE);
               }

               int minAvail = puddle.getMinAvail();

               while (avail.size() + pend < minAvail && cxSet.size() + pend < capacity())
               {
                    logger.info(puddleName + ": creating a connection to satisfy minAvail");
                    createConnection(MonitorRecord.CAUSE_MIN_AVAIL);
               }
          }

          protected Throwable getCause()
          {
               return cause;
          }

          // always called from synchronized code
          protected void shutdown()
          {
               shutdownTimerDone = true;
               lock.notifyAll();
               Iterator availItor = avail.iterator();

               while (availItor.hasNext())
               {
                    ConnectionEnvelope cxEnv = (ConnectionEnvelope)availItor.next();
                    availItor.remove();
                    destroyConnection(cxEnv,MonitorRecord.CAUSE_SHUTDOWN);
               }
          }

          // always called from synchronized code
          protected void destroy()
          {
               shutdownTimerDone = true;
               lock.notifyAll();
               int trackingSetSize = trackingSet.size();
               Set copy = new HashSet(trackingSetSize);
               copy.addAll(trackingSet);
               Iterator cxItor = copy.iterator();

               while (cxItor.hasNext())
               {
                    WeakReference ref = (WeakReference)cxItor.next();
                    ConnectionEnvelope cxEnv = (ConnectionEnvelope)ref.get();

                    if (cxEnv != null)
                    {
                         destroyConnection(cxEnv,MonitorRecord.CAUSE_DESTROY);
                    }
               }

               clients = 0;

               try
               {
                    while (pend > 0)
                    {
                         lock.wait();
                    }
               }
               catch (InterruptedException ie)
               {
               }
          }

          protected void cutServerStatusRecord(ConnectionEnvelope cxEnv,String cause)
          {
               Logger monitor = lock.getMonitor();

               if (monitor == null)
               {
                    return;
               }

               ConnectionFactoryManager cxfManager = lock.getManager();
               MonitorRecord record = new MonitorRecord();
               int puddleIndex = cxEnv.getPuddleIndex();
               record.setPuddleIndex(puddleIndex);
               int clusterIndex = cxEnv.getClusterIndex();
               record.setClusterIndex(clusterIndex);
               int serverIndex = cxEnv.getServerIndex();
               record.setServerIndex(serverIndex);
               ConnectionFactoryConfiguration cxfConfig = lock.getConfig();
               String loggerName = cxfManager.getFactoryLoggerName(cxfConfig);
               String factoryName = loggerName.substring(ConnectionFactoryManager.PARENT_LOGGER_NAME.length()+1);
               record.setFactoryName(factoryName);
               Server server = cxEnv.getServer();
               int reactivationLimit = server.getReactivationLimit();

               if (server instanceof TCPIPServer)
               {
                    String serverName = ((TCPIPServer)server).getNormalizedHost();
                    record.setServerName(serverName);
               }

               StringBuffer sb = new StringBuffer();
               Iterator cxItor = trackingSet.iterator();
               int connections = 0;
               int idle = 0;

               while (cxItor.hasNext())
               {
                    WeakReference ref = (WeakReference)cxItor.next();
                    ConnectionEnvelope tEnv = (ConnectionEnvelope)ref.get();
                    Server tServer = (tEnv != null) ? tEnv.getServer() : null;

                    if (server.equals(tServer))
                    {
                         connections++;
                         Credential userCred = tEnv.getUserCredential();

                         if (userCred == null)
                         {
                              idle++;
                         }
                    }
               }

               record.setConnections(connections);
               record.setIdle(idle);
               record.setActive(connections-idle);
               monitor.log(record);
          }

          // always called from synchronized code
          protected int chooseCluster()
          {
               int bestCluster = -1;
               double bestLoad = Double.POSITIVE_INFINITY;
               Cluster[] clusters = puddle.getClusterArray();

               for (int i=0;i<clusters.length;i++)
               {
                    if (clusterStatus[i])
                    {
                         double curClients = clusterUsage[i];
                         double maxClients = clusters[i].getMaxClients();
                         double load = ((curClients * 2.) + 1.) / maxClients;

                         if (load < bestLoad && load <= 2.)
                         {
                              bestLoad = load;
                              bestCluster = i;
                         }
                    }
               }

               return bestCluster;
          }


          // always called from synchronized code
          private void addAvail(ConnectionEnvelope cxEnv)
          {
               if (shutdown)
               {
                    destroyConnection(cxEnv,MonitorRecord.CAUSE_SHUTDOWN);
                    lock.notifyAll();
                    return;
               }

               if (!cxEnv.isBogus())
               {
                    int shutdownAfterMins = cxEnv.getServer().getShutdownAfterMinutes();

                    if (shutdownAfterMins >= 0)
                    {
                         BaseSecuritySupport secSup = BaseSecuritySupport.securitySupport;
                         long nowMillis = System.currentTimeMillis();
                         long offsetMillis = (shutdownAfterMins == 0) ? 500L : shutdownAfterMins * 60L * 1000L;
                         cxEnv.setShutdownTime(nowMillis + offsetMillis);

                         if (shutdownTimer == null)
                         {
                              shutdownTimer = secSup.createThread(new ShutdownTimer());

                              PrivilegedAction daemonAction = new PrivilegedAction()
                              {
                                   public Object run()
                                   {
                                        shutdownTimer.setName("com.sas.services.connection.PuddleEnvelope.shutdown_timer"/*I18NOK:EMS*/);
                                        shutdownTimer.setDaemon(true);
                                        return null;
                                   }
                              };

                              secSup.doThreadPrivileged(daemonAction);
                              shutdownTimer.start();
                         }
                    }
               }
               else
               {
                    cause = cxEnv.getCause();
               }

               avail.add(cxEnv);
               lock.notifyAll();
               Logger cxLogger = lock.getLogger(cxEnv);
               cxLogger.info(puddleName + ": putting cx#" + cxEnv.getConnectionNumber() + " on the available queue");
          }

          // always called from synchronized code
          private void destroyConnection(ConnectionEnvelope cxEnv,String cause)
          {
               if (!cxEnv.isBogus())
               {
                    clusterUsage[cxEnv.getClusterIndex()]--;
                    ObjectImpl obj = (ObjectImpl)cxEnv.getRootObject();
                    Iterator cxSetItor = cxSet.iterator();
                    
                    while (cxSetItor.hasNext())
                    {
                         org.omg.CORBA.Object t = (org.omg.CORBA.Object)cxSetItor.next();
                         
                         if (obj._is_equivalent(t))
                         {
                              cxSetItor.remove();
                              break;
                         }
                    }
                    
                    Iterator trackingSetItor = trackingSet.iterator();

                    while (trackingSetItor.hasNext())
                    {
                         WeakReference ref = (WeakReference)trackingSetItor.next();
                         Object reft = ref.get();

                         if (cxEnv.equals(reft))
                         {
                              trackingSetItor.remove();
                              break;
                         }
                    }

                    ORB orb = obj._orb();
                    orb.shutdown(true);
               }

               Logger cxLogger = lock.getLogger(cxEnv);
               cxLogger.info(puddleName + ": cx#" + cxEnv.getConnectionNumber() + " destroyed");
               cutServerStatusRecord(cxEnv,cause);
          }
     }
/*
     private class SharingDelegate extends Delegate
     {
          protected int capacity()
          {
               return 0;
          }

          protected int available()
          {
               return 0;
          }

          protected int pending()
          {
               return 0;
          }

          protected void createConnection()
          {
          }

          protected ConnectionEnvelope getAvailableConnection()
          {
               return null;
          }

          protected void reactivateConnection(ConnectionEnvelope cxEnv,int clientNumber,int reason)
          {
          }

          protected void checkMinimums()
          {
          }

          protected void shutdown()
          {
          }

          protected void destroy()
          {
          }
     }
*/
     protected void setLock(ConnectionFactoryKernel lock)
     {
          this.lock = lock;
     }

     protected void setLogger(Logger logger)
     {
          this.logger = logger;
     }

     protected void setPuddleIndex(int puddleIndex)
     {
          this.puddleIndex = puddleIndex;
          puddleName = "pd#"/*I18NOK:EMS*/ + puddleIndex;
     }

     protected void init()
     {
          synchronized (lock)
          {
               delegate.checkMinimums();
          }
     }

     // always called from synchronized code
     protected double getLoad()
     {
          int available = delegate.available();

          if (available > waiting)
          {
               return 0.;
          }

          int pending = delegate.pending();

          if (pending + available > waiting)
          {
               return Double.MIN_VALUE; // smallest strictly positive value
          }

          int capacity = delegate.capacity();
          return ((2. * clients) + 1.) / capacity;
     }

     protected int getClients()
     {
          return clients;
     }

     protected ConnectionEnvelope getConnection(long waitMilliseconds,Logger cxLogger) throws ConnectionFactoryException
     {
          ConnectionEnvelope cxEnv = null;

          synchronized (lock)
          {
               long timer = waitMilliseconds;

               do
               {
                    // First make sure enough servers are still in 
                    // play. When a server fails, the puddle tries
                    // to go on without it unless it was the last 
                    // server or it's capacity is necessary to 
                    // satisfy a request guaranteed to succeed due
                    // to minSize.
                    int cap = delegate.capacity();
                    int ms = puddle.getMinSize();
                    int clstr = delegate.chooseCluster();
                                           
                    if (cap == 0 || (cap < ms && clients + waiting >= cap))
                    {
                         // {0}: One or more servers in a puddle have failed,
                         // and the puddle can no longer meet its minimum 
                         // service requirements. The most recent server 
                         // failure is included as a cause of this exception. 
                         // The factory log may contain details of other 
                         // server failures.
                         Throwable cause = delegate.getCause();
                         String messageIndex = "cxf.300.ex.msg"/*I18NOK:EMS*/;
                         String messagePattern = RB.getStringResource(messageIndex);
                         String message = Message.format(messagePattern,messageIndex);
                         throw new FatalConnectionFactoryException(message,cause);
                    }

                    // make sure there is room in the inn
                    try
                    {
                         if (waitMilliseconds == 0L)
                         {
                              while (clients + waiting >= delegate.capacity() && !shutdown)
                              {
                                   lock.wait();
                              }
                         }
                         else
                         {
                              while (timer > 0L && clients + waiting >= delegate.capacity() && !shutdown)
                              {
                                   long start = System.currentTimeMillis();
                                   lock.wait(timer);
                                   long stop = System.currentTimeMillis();
                                   timer -= (stop - start);
                              }

                              if (clients + waiting >= delegate.capacity())
                              {
                                   // {0}: No connections became available in the time allotted.
                                   String messageIndex = "cxf.301.ex.msg"/*I18NOK:EMS*/;
                                   String messagePattern = RB.getStringResource(messageIndex);
                                   String message = Message.format(messagePattern,messageIndex);
                                   throw new ConnectionFactoryException(message);
                              }
                         }

                         if (shutdown)
                         {
                              // {0}: A factory administrator shut the factory down while this connection request was in process.
                              String messageIndex = "cxf.302.ex.msg"/*I18NOK:EMS*/;
                              String messagePattern = RB.getStringResource(messageIndex);
                              String message = Message.format(messagePattern,messageIndex);
                              throw new ConnectionFactoryException(message);
                         }
                    }
                    catch (InterruptedException ie)
                    {
                         // {0}: The connection factory was interrupted while waiting for a connection to become available.
                         String messageIndex = "cxf.303.ex.msg"/*I18NOK:EMS*/;
                         String messagePattern = RB.getStringResource(messageIndex);
                         String message = Message.format(messagePattern,messageIndex);
                         throw new ConnectionFactoryException(message);
                    }

                    // start a connection if necessary
                    while (waiting >= delegate.available() + delegate.pending())
                    {
                         delegate.createConnection(MonitorRecord.CAUSE_REQUEST);
                    }

                    // wait for a connection to become available
                    while (delegate.available() == 0 && !shutdown)
                    {
                         try
                         {
                              waiting++;
                              lock.wait();
                         }
                         catch (InterruptedException ie)
                         {
                              // {0}: The connection factory was interrupted while waiting for a connection to complete.
                              String messageIndex = "cxf.304.ex.msg"/*I18NOK:EMS*/;
                              String messagePattern = RB.getStringResource(messageIndex);
                              String message = Message.format(messagePattern,messageIndex);
                              throw new ConnectionFactoryException(message);
                         }
                         finally
                         {
                              waiting--;
                         }
                    }

                    if (shutdown)
                    {
                         // {0}: A factory administrator shut the factory down while this connection request was in process.
                         String messageIndex = "cxf.305.ex.msg"/*I18NOK:EMS*/;
                         String messagePattern = RB.getStringResource(messageIndex);
                         String message = Message.format(messagePattern,messageIndex);
                         throw new ConnectionFactoryException(message);
                    }

                    // this returns null if the available connections are stale
                    cxEnv = delegate.getAvailableConnection(cxLogger);
               }
               while (cxEnv == null);

               clients++;
          }

          return cxEnv;
     }

     protected void reactivateConnection(ConnectionEnvelope cxEnv,int clientNumber,int reason)
     {
          delegate.reactivateConnection(cxEnv,clientNumber,reason);
     }

     protected void shutdown()
     {
          shutdown = true;
          delegate.shutdown();
     }

     protected void destroy()
     {
          shutdown = true;
          delegate.destroy();
     }

     protected void cutServerStatusRecord(ConnectionEnvelope cxEnv,String cause)
     {
          delegate.cutServerStatusRecord(cxEnv,cause);
     }

     protected void removeUserFromCache(Credential user)
     {
          if (userCache != null)
          {
               synchronized (userCache)
               {
                    userCache.remove(user);
               }
          }
     }
}