/* $Id: Server.java,v 1.4.16.1 2005/10/11 17:14:12 sasbfb Exp $ */
package com.sas.services.connection;

import java.io.Serializable;
import java.util.logging.Logger;
//
// History:
//
// Date    Name            Remarks
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
// 16Jun03 B Beers         S0205937: Fix doc comments.
// 12Nov03 B Beers         S0227109: Remove stub finalizers.
// 11Oct05 B Beers         S0324947: Added serialVersionUID.
/**
 * A base class for describing IOM servers.
 * @author Biff Beers
 */
public abstract class Server implements Cloneable,Serializable
{
     /**
      * Construct an object describing an IOM server.
      * @param classID the class ID
      * @see #setClassID
      */
     public Server(String classID)
     {
          setClassID(classID);
     }

     /**
      * Set the class ID. The class ID identifies the type of IOM object
      * that the server is configured to instantiate. The fields of this class
      * named <CODE>CLSID_*</CODE> provide several possible values for
      * this attribute, but there may be other valid values in newer server
      * versions. If the server is not configured to instantiate IOM objects of
      * this type, then connections attempts to the server will fail.
      * @param classID the class ID
      * @see #CLSID_SAS
      * @see #CLSID_SAS_1_1
      * @see #CLSID_SASMDX
      * @see #CLSID_SASGMS
      * @see #CLSID_SASMQX
      * @see #CLSID_SASOMI
      * @see #CLSID_SASSTP
      * @throws java.lang.IllegalArgumentException if this attribute is set 
      * to <CODE>null</CODE>
      */
     public void setClassID(String classID)
     {
          if (classID == null)
          {
               throw new IllegalArgumentException("\"classID\" cannot be null.");
          }

          origClassID = classID;
          normClassID = classID.toLowerCase();
     }

     /**
      * Get the class ID.
      * @return the class ID
      * @see #setClassID
      */
     public String getClassID()
     {
          return origClassID;
     }

     /**
      * Get the normalized class ID.
      * A normalized class ID is all lowercase.
      * @return the normalized class ID
      * @see #setClassID
      */
     public String getNormalizedClassID()
     {
          return normClassID;
     }

     /**
      * Set the maximum number of clients. This attribute is ignored by
      * the connection factory unless the server is participating in a
      * connection pool. For connection pools, this attribute determines
      * the maximum number of connections a factory may have to the server
      * at any given time. The default value is 10.
      * @param maxClients the maximum number of clients
      * @throws java.lang.IllegalArgumentException if this attribute is set
      * to a value less than <CODE>1</CODE>
      */
     public void setMaxClients(int maxClients)
     {
          if (maxClients <= 0)
          {
               throw new IllegalArgumentException("\"maxClients\" cannot be less than 1.");
          }

          this.maxClients = maxClients;
     }

     /**
      * Get the maximum number of clients.
      * @return the maximum number of clients
      * @see #setMaxClients
      */
     public int getMaxClients()
     {
          return maxClients;
     }

     /**
      * Set the maximum number of reactivations. This attribute is ignored by
      * the connection factory unless the server is participating in a
      * connection pool. For connection pools, this attribute determines
      * the number of times a connection to this server will be refreshed and
      * reused by the factory to satisfy connection requests. Once the limit is
      * reached, the factory will destroy the connection and reconnect if necessary.
      * <P>
      * In theory, refreshing a connection should cause the server to release all
      * resources associated with the connection and return to its initial state.
      * However, if there are bugs in the server that allow resource use to accumulate
      * over several connection reuses, then the server may eventually run out of
      * that resource. This attribute allows the factory to periodically restart
      * servers and force all resources to be released.
      * <P>
      * If the value is <CODE>0</CODE>, then connections will be refreshed and reused indefinitely
      * (subject to limitations set by other attributes). The default value is <CODE>0</CODE>.
      * @param reactivationLimit the maximum number of reactivations
      * @throws java.lang.IllegalArgumentException if this attribute is set
      * to a value less than <CODE>0</CODE>
      */
     public void setReactivationLimit(int reactivationLimit)
     {
          if (reactivationLimit < 0)
          {
               throw new IllegalArgumentException("\"reactivationLimit\" cannot be less than 0.");
          }

          this.reactivationLimit = reactivationLimit;
     }

     /**
      * Get the maximum number of reactivations.
      * @return the maximum number of reactivations
      * @see #setReactivationLimit
      */
     public int getReactivationLimit()
     {
          return reactivationLimit;
     }

     /**
      * Set the number of minutes before an idle connection is destroyed. This attribute is ignored by
      * the connection factory unless the server is participating in a
      * connection pool. For connection pools, this attribute determines how long an unallocated
      * connection will be kept alive by the factory. This attribute takes precedence over the
      * {@link Puddle#setMinSize minSize} and {@link Puddle#setMinAvail minAvail} attributes in that
      * a puddle with a non-negative value for this attribute is allowed to shutdown all connections
      * after the required period of idleness even if it also has strictly positive values for minSize
      * or minAvail.
      * <P>
      * <CODE>-1</CODE> means idle connections will be kept alive until
      * connection factory shutdown. <CODE>0</CODE> means a connection returned to the factory will
      * be destroyed immediately unless another user is waiting for it. The default value is
      * <CODE>-1</CODE>.
      * @param shutdownAfterMinutes the number of minute before an idle connection is destroyed
      * @throws java.lang.IllegalArgumentException if this attribute is set
      * to a value less than <CODE>-1</CODE>
      */
     public void setShutdownAfterMinutes(int shutdownAfterMinutes)
     {
          if (shutdownAfterMinutes < -1)
          {
               throw new IllegalArgumentException("\"shutdownAfterMinutes\" cannot be less than -1");
          }

          this.shutdownAfterMinutes = shutdownAfterMinutes;
     }

     /**
      * Get the number of minutes before an idle connection is destroyed.
      * @return the number of minute before an idle connection is destroyed
      * @see #setShutdownAfterMinutes
      */
     public int getShutdownAfterMinutes()
     {
          return shutdownAfterMinutes;
     }

     /**
      * Set the authentication domain. The domain is a label that the factory
      * uses to group servers and credentials. When connecting through a factory, the
      * factory user must present a credential object with a domain that matches
      * the domain of at least one server managed by the factory.
      * <P>
      * Empty string and <CODE>null</CODE> are considered to be equivalent domains.
      * The default value is empty string.
      * @param domain the authentication domain
      */
     public void setDomain(String domain)
     {
          this.domain = (domain != null) ? domain : "";
     }

     /**
      * Get the authentication domain.
      * @return the authentication domain
      * @see #setDomain
      */
     public String getDomain()
     {
          return domain;
     }

     public boolean equals(Object that)
     {
          if (this == that)
          {
               return true;
          }

          if (that == null || !getClass().getName().equals(that.getClass().getName()))
          {
               return false;
          }

          Server thatServer = (Server)that;
          String thatNormClassID = thatServer.getNormalizedClassID();

          if (!normClassID.equals(thatNormClassID))
          {
               return false;
          }

          int thatMaxClients = thatServer.getMaxClients();

          if (maxClients != thatMaxClients)
          {
               return false;
          }

          int thatReactivationLimit = thatServer.getReactivationLimit();

          if (reactivationLimit != thatReactivationLimit)
          {
               return false;
          }

          int thatShutdownAfterMinutes = thatServer.getShutdownAfterMinutes();

          if (shutdownAfterMinutes != thatShutdownAfterMinutes)
          {
               return false;
          }

          String thatDomain = thatServer.getDomain();

          if (!domain.equals(thatDomain))
          {
               return false;
          }

          return true;
     }

     public int hashCode()
     {
          int hc = getClass().getName().hashCode();
          hc ^= normClassID.hashCode();
          hc ^= maxClients;
          hc ^= reactivationLimit;
          hc ^= shutdownAfterMinutes;
          hc ^= domain.hashCode();
          return hc;
     }

     public Object clone()
     {
          try
          {
               return super.clone();
          }
          catch (CloneNotSupportedException e)
          {
               throw new InternalError("A Cloneable object threw CloneNotSupportedException.");
          }
     }

     /**
      * The class ID of the SAS toplevel class.
      * @see com.sas.services.connection.Server#getClassID()
      */
     public static final String CLSID_SAS    = "440196d4-90f0-11d0-9f41-00a024bb830c";

     /**
      * The class ID of the SAS toplevel class (version 1.1).
      * @see com.sas.services.connection.Server#getClassID()
      */
     public static final String CLSID_SAS_1_1 = "cf7bc7e6-c7e8-11d5-87e3-00c04f38f9f6";

     /**
      * The class ID of the SASGMS toplevel class.
      * @see com.sas.services.connection.Server#getClassID()
      */
     public static final String CLSID_SASGMS = "9c186a64-83e4-11d2-b956-00c04f81993c";

     /**
      * The class ID of the SASMDX toplevel class.
      * @see com.sas.services.connection.Server#getClassID()
      */
     public static final String CLSID_SASMDX = "f3f46472-1e31-11d5-87c2-00c04f38f9f6";

     /**
      * The class ID of the SASMQX toplevel class.
      * @see com.sas.services.connection.Server#getClassID()
      */
     public static final String CLSID_SASMQX = "8d9ef489-8d1e-11d2-b94f-00c04f8198c0";

     /**
      * The class ID of the SASOMI toplevel class.
      * @see com.sas.services.connection.Server#getClassID()
      */
     public static final String CLSID_SASOMI = "2887e7d7-4780-11d4-879f-00c04f38f0db";

     /**
      * The class ID of the SASSTP toplevel class.
      * @see com.sas.services.connection.Server#getClassID()
      */
     public static final String CLSID_SASSTP = "15931e31-667f-11d5-8804-00c04f35ac8c";

     /**
      * Serial version ID. 
      * This value should be changed if this version is not compatible with the previous version.
      */
     private static final long serialVersionUID = 4922187406001339540L;
     private String origClassID = null;
     private String normClassID = null;
     private int maxClients = 10;
     private int reactivationLimit = 0;
     private int shutdownAfterMinutes = -1;
     private String domain = "";

     protected abstract org.omg.CORBA.Object connect(Credential credential,Logger logger,int cxNumber);
}