package com.sas.services.connection.jndi;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import javax.naming.CommunicationException;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import com.sas.services.connection.AuthServer;
import com.sas.services.connection.ConnectionFactoryException;
import com.sas.services.connection.Credential;
import com.sas.services.connection.LoginException;
import com.sas.services.connection.PasswordCredential;
import com.sas.services.connection.Puddle;
import com.sas.services.connection.PuddleEnvelope;
import com.sas.text.Message;
//
// History:
//
// Date    Name            Remarks
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 22Aug02 B Beers         S0163094: Initialize namingContexts when needed.
// 03Sep02 B Beers         S0164497: Reimplement over com.sas.services.connection.
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 26Mar03 B Beers         S0185307: New metadata model for connection pools.
// 10Sep03 B Beers         S0219412: Allow case-insensitive lookup of privileged user's login.
//
/**
 * Authentication and authorization services used internally by
 * {@link com.sas.services.connection.jndi.JNDIConnectionFactoryConfiguration}.
 * @author Biff Beers
 */
public class JNDIAuthServer extends AuthServer
{
     protected JNDIAuthServer(PuddleEnvelope[] envs,JNDIConnectionFactoryConfiguration cxfConfig)
     {
          super(envs,false,true);

          if (cxfConfig.isAdminAuth())
          {
               adminCache = new HashSet();
               adminCache.add(cxfConfig.getPrivLogin());
          }

          List domains = new ArrayList(1);
          String repositoryDomain = cxfConfig.getRepositoryDomain();
          domains.add(repositoryDomain);
          setDomains(domains);
          this.cxfConfig = cxfConfig;
     }

     private static SearchControls objectScopedSearchControls = null;
     private static SearchControls treeScopedSearchControls = null;
     private JNDIConnectionFactoryConfiguration cxfConfig = null;
     private Set adminCache = null;
     private Set namingContexts = null;

     static
     {
          objectScopedSearchControls = new SearchControls();
          objectScopedSearchControls.setSearchScope(SearchControls.OBJECT_SCOPE);
          treeScopedSearchControls = new SearchControls();
          treeScopedSearchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
     }

     protected void authorizeAdmin(Credential adminLogin) throws ConnectionFactoryException
     {
          // perhaps you think I care
          if (adminCache == null)
          {
               return;
          }

          // is this normal?
          adminLogin = normalize(adminLogin);

          // haven't I seen you here before?
          if (adminCache.contains(adminLogin))
          {
               return;
          }

          String domain = adminLogin.getDomain();
          List domainList = getDomains(adminLogin);

          if (!domainList.contains(domain))
          {
               // {0}: This factory does not support connections under the authentication domain "{1}".
               String messageIndex = "cxj.000.ex.msg";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex,domain);
               throw new LoginException(message);
          }

          // try to authenticate the login
          DirContext ctx = (DirContext)authenticate(adminLogin);

          if (ctx == null)
          {
               // {0}: Administrator could not be authenticated in domain "{1}".
               String messageIndex = "cxj.001.ex.msg";
               String messagePattern = RB.getStringResource(messageIndex);
               String message = Message.format(messagePattern,messageIndex,adminLogin.getDomain());
               throw new LoginException(message);
          }

          try
          {
               PuddleEnvelope[] envs = getPuddleEnvelopes();

               // see if this administrator can see all of the puddles' logins
               for (int i=0;i<envs.length;i++)
               {
                    Puddle puddle = envs[i].getPuddle();
                    String loginDN = puddle.getPuddleID();
                    Object o = null;

                    try
                    {
                         o = ctx.lookup(loginDN);
                    }
                    catch (CommunicationException ce)
                    {
                         try
                         {
                              ctx.close();
                         }
                         catch (NamingException e)
                         {
                         }

                         ctx = (DirContext)authenticate(adminLogin);

                         if (ctx != null)
                         {
                              try
                              {
                                   o = ctx.lookup(loginDN);
                              }
                              catch (NamingException e)
                              {
                              }
                         }
                    }
                    catch (NamingException ne)
                    {
                    }

                    if (o == null)
                    {
                         // {0}: Administrator is not authorized to administer this pool.
                         String messageIndex = "cxj.002.ex.msg";
                         String messagePattern = RB.getStringResource(messageIndex);
                         String message = Message.format(messagePattern,messageIndex);
                         throw new LoginException(message);
                    }
               }
          }
          finally
          {
               closeIdentity(ctx);
          }

          adminCache.add(adminLogin);
     }

     protected Credential normalize(Credential credential)
     {
          if (credential == null || !(credential instanceof PasswordCredential))
          {
               return credential;
          }

          PasswordCredential login = (PasswordCredential)credential;
          String dn = login.getUserName();
          String normDN = JNDIConnectionFactoryConfiguration.normalizeDN(dn);

          // result of IOMLogin.getUserName() cannot be null
          return (dn.equals(normDN)) ? login : new PasswordCredential(normDN,login.getPassword(),login.getDomain());
     }

     protected Object authenticate(Credential credential) throws ConnectionFactoryException
     {
          if (!(credential instanceof PasswordCredential))
          {
               throw new IllegalArgumentException("\"credential\" must be an instance of PasswordCredential.");
          }

          PasswordCredential login = (PasswordCredential)credential;
          Hashtable env = cxfConfig.getEnv();
          Hashtable authEnv = new Hashtable();
          authEnv.put(Context.SECURITY_PRINCIPAL,login.getUserName());
          authEnv.put(Context.SECURITY_CREDENTIALS,login.getPassword());
          authEnv.put(Context.SECURITY_AUTHENTICATION,"simple");
          authEnv.put(Context.PROVIDER_URL,env.get(Context.PROVIDER_URL));
          authEnv.put(Context.INITIAL_CONTEXT_FACTORY,env.get(Context.INITIAL_CONTEXT_FACTORY));

          try
          {
               return new InitialDirContext(authEnv);
          }
          catch (NamingException ne)
          {
               return null;
          }
     }

     protected boolean authorizeUser(Object identity,Puddle puddle) throws ConnectionFactoryException
     {
          try
          {
               Set users = puddle.getUsers();
               DirContext ctx = (DirContext)identity;
               Hashtable env = ctx.getEnvironment();
               String dn = (String)env.get(Context.SECURITY_PRINCIPAL);
               dn = JNDIConnectionFactoryConfiguration.normalizeDN(dn);

               if (users.contains(dn))
               {
                    return true;
               }

               String baseDN = findGroupBaseDN(dn);
               Set visitedGroups = new HashSet();
               return authorizeGroups(dn,baseDN,visitedGroups,users);
          }
          catch (NamingException ne)
          {
               throw new ConnectionFactoryException(ne);
          }
     }

     protected void closeIdentity(Object identity)
     {
          if (identity != null)
          {
               try
               {
                    ((DirContext)identity).close();
               }
               catch (NamingException ne)
               {
               }
          }
     }

     private String findGroupBaseDN(String personDN)
     {
          if (namingContexts == null)
          {
               namingContexts = new HashSet();
               findNamingContexts();
          }

          Iterator namingContextIterator = namingContexts.iterator();

          while (namingContextIterator.hasNext())
          {
               String namingContext = (String)namingContextIterator.next();

               if (personDN.indexOf(namingContext) >= 0)
               {
                    return namingContext;
               }
          }

          final int length = personDN.length();
          int lastComma = personDN.lastIndexOf(',');

          if (lastComma <= 0 || lastComma + 1 == length)
          {
               return personDN;
          }

          int secondToLastComma = personDN.lastIndexOf(',',lastComma-1);

          if (secondToLastComma < 0)
          {
               return personDN.substring(lastComma+1);
          }

          return personDN.substring(secondToLastComma+1);
     }

     private void findNamingContexts()
     {
          try
          {
               DirContext context = cxfConfig.getContext();
               NamingEnumeration results = null;

               try
               {
                    results = context.search("","(objectclass=*)",objectScopedSearchControls);
               }
               catch (CommunicationException ce)
               {
                    cxfConfig.reconnect(ce);
                    context = cxfConfig.getContext();
                    results = context.search("","(objectclass=*)",objectScopedSearchControls);
               }

               while (results.hasMore())
               {
                    SearchResult result = (SearchResult)results.next();
                    Attributes resultAttrs = result.getAttributes();
                    Attribute resultAttr = resultAttrs.get("namingContexts");

                    if (resultAttr != null)
                    {
                         NamingEnumeration values = resultAttr.getAll();

                         while (values.hasMore())
                         {
                              String value = (String)values.next();
                              value = JNDIConnectionFactoryConfiguration.normalizeDN(value);
                              namingContexts.add(value);
                         }
                    }
               }
          }
          catch (NamingException e)
          {
          }
     }

     private boolean authorizeGroups(String personDN,String groupBaseDN,Set visitedGroups,Set authorizedUserIdentities) throws ConnectionFactoryException
     {
          try
          {
               DirContext context = cxfConfig.getContext();
               StringBuffer buffer = new StringBuffer(128);
               buffer.append("(|(uniquemember=").append(personDN).append(")(member=");
               String filter = buffer.append(personDN).append("))").toString();
               buffer.setLength(0);
               NamingEnumeration results = null;

               try
               {
                    results = context.search(groupBaseDN,filter,treeScopedSearchControls);
               }
               catch (CommunicationException ce)
               {
                    cxfConfig.reconnect(ce);
                    context = cxfConfig.getContext();
                    results = context.search(groupBaseDN,filter,treeScopedSearchControls);
               }

               while (results.hasMore())
               {
                    SearchResult result = (SearchResult)results.next();
                    buffer.append(result.getName()).append(",").append(groupBaseDN);
                    String groupDN = JNDIConnectionFactoryConfiguration.normalizeDN(buffer.toString());
                    buffer.setLength(0);

                    if (authorizedUserIdentities.contains(groupDN))
                    {
                         return true;
                    }

                    if (visitedGroups.add(groupDN))
                    {
                         if (authorizeGroups(groupDN,groupBaseDN,visitedGroups,authorizedUserIdentities));
                         {
                              return true;
                         }
                    }
               }
          }
          catch (NamingException e)
          {
               throw new ConnectionFactoryException(e);
          }

          return false;
     }
}
