package com.sas.services.connection.jndi;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;
import javax.naming.CommunicationException;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import com.sas.services.connection.AggregationKernel;
import com.sas.services.connection.AuthServer;
import com.sas.services.connection.BridgeServer;
import com.sas.services.connection.Cluster;
import com.sas.services.connection.ConnectionFactoryConfiguration;
import com.sas.services.connection.ConnectionFactoryException;
import com.sas.services.connection.ConnectionFactoryKernel;
import com.sas.services.connection.ConnectionFactoryManager;
import com.sas.services.connection.FailoverCluster;
import com.sas.services.connection.PasswordCredential;
import com.sas.services.connection.PoolingKernel;
import com.sas.services.connection.Puddle;
import com.sas.services.connection.PuddleEnvelope;
import com.sas.services.connection.Server;
import com.sas.util.SasPasswordEncodingException;
import com.sas.util.SasPasswordString;
//
// History:
//
// Date    Name            Remarks
// 03Dec01 B Beers         S0133638: Add JDNI configuration.
// 02Jan02 B Beers         S0135536: Add admin auth feature.
// 14Feb02 B Beers         S0139873: Added User Service authorization.
// 09Apr02 B Beers         S0141718: Remove dependencies on netutil.jar.
// 26Apr02 B Beers         S0149086: Revise the connection factory interface.
// 07May02 B Beers         S0150375: Added Info Service configuration.
// 13May02 B Beers         S0149086: Reinstate advanced connection management.
// 29May02 B Beers         S0139445: Logging service integration (happy 9th anniversary Biff & Julie).
// 08Jul02 B Beers         S0156141: Added LDAP implementation to PlatformConnectionFactory.
// 11Oct02 B Beers         S0168379: Generalize connection service for other servers.
// 10Jan03 B Beers         S0177764: Use java.util.logging package and application monitoring.
//
/**
 * A connection factory configuration that describes a server or a pool of
 * servers to connect to using meta data in accessed through a JNDI interface.
 * @author Biff Beers
 */
public class JNDIConnectionFactoryConfiguration extends ConnectionFactoryConfiguration
{
     /**
      * Constructs a factory configuration that reads the information
      * necessary to create connections to remote IOM objects
      * from a JNDI repository.
      * @param context a connection to the root node of a JNDI repository
      * that contains the values necessary to start a connection factory. The connection
      * to the repository must have been made with
      * the distinguished name and password of an entity that has read access to those
      * values. This context is not closed when the factory is shutdown.
      * @param baseDN a partial JNDI distinguished name that defines the section within
      * the repository naming hierarchy that contains the pooling information.
      * @param logicalName the value of the <CODE>sasLogicalName</CODE> attribute for all login
      * definitions and server definitions under to the baseDN in the JNDI
      * repository that contain information for this factory.
      * @param repositoryDomain the name of the domain assigned to the JNDI repository. Connection
      * factories with this configuration will have this string in the list of domains
      * returned by the getDomains() method in 
      * {@link com.sas.services.connection.ConnectionFactoryInterface}.
      */
     public JNDIConnectionFactoryConfiguration(DirContext context,String baseDN,String logicalName,String repositoryDomain)
     {
          this(context,baseDN,logicalName,repositoryDomain,true,false,false);
     }

     /**
      * This constructor is provided for compatibility with a previous release. Otherwise,
      * it is rarely used.
      */
     public JNDIConnectionFactoryConfiguration(DirContext context,String baseDN,String logicalName,String repositoryDomain,boolean userAuth,boolean adminAuth,boolean noPool)
     {
          if (context == null)
          {
               throw new IllegalArgumentException("\"context\" cannot be null.");
          }

          if (baseDN == null)
          {
               throw new IllegalArgumentException("\"baseDN\" cannot be null.");
          }

          if (logicalName == null)
          {
               throw new IllegalArgumentException("\"logicalName\" cannot be null.");
          }

          this.context = context;
          this.baseDN = baseDN;
          this.logicalName = logicalName;
          this.repositoryDomain = (repositoryDomain != null) ? repositoryDomain : "";
          this.userAuth = userAuth;
          this.adminAuth = adminAuth;
          this.noPool = noPool;
     }

     public DirContext getContext()
     {
          return context;
     }

     public String getBaseDN()
     {
          return baseDN;
     }

     public String getLogicalName()
     {
          return logicalName;
     }

     public String getRepositoryDomain()
     {
          return repositoryDomain;
     }

     public boolean isUserAuth()
     {
          return userAuth;
     }

     public boolean isAdminAuth()
     {
          return adminAuth;
     }

     public boolean isNoPool()
     {
          return noPool;
     }

     public boolean equals(Object that)
     {
          if (!super.equals(that))
          {
               return false;
          }

          JNDIConnectionFactoryConfiguration thatConfig = (JNDIConnectionFactoryConfiguration)that;
          DirContext thatContext = thatConfig.getContext();

          if (context != thatContext)
          {
               Hashtable thisEnv = null;
               Hashtable thatEnv = null;

               try
               {
                    thisEnv = context.getEnvironment();
                    thatEnv = thatContext.getEnvironment();

                    if (thisEnv == null ? thatEnv != null : !thisEnv.equals(thatEnv))
                    {
                         return false;
                    }
               }
               catch (NamingException ne)
               {
                    return false;
               }
          }

          String thatBaseDN = thatConfig.getBaseDN();

          if (!baseDN.equals(thatBaseDN))
          {
               return false;
          }

          String thatLogicalName = thatConfig.getLogicalName();

          if (!logicalName.equals(thatLogicalName))
          {
               return false;
          }

          String thatRepositoryDomain = thatConfig.getRepositoryDomain();

          if (!repositoryDomain.equals(thatRepositoryDomain))
          {
               return false;
          }

          boolean thatUserAuth = thatConfig.isUserAuth();

          if (userAuth != thatUserAuth)
          {
               return false;
          }

          boolean thatAdminAuth = thatConfig.isAdminAuth();

          if (adminAuth != thatAdminAuth)
          {
               return false;
          }

          boolean thatNoPool = thatConfig.isNoPool();

          if (noPool != thatNoPool)
          {
               return false;
          }

          return true;
     }

     public int hashCode()
     {
          int hc = super.hashCode();

          try
          {
               Hashtable thisEnv = context.getEnvironment();
               hc ^= thisEnv.hashCode();
          }
          catch (NamingException ne)
          {
          }

          hc ^= baseDN.hashCode();
          hc ^= logicalName.hashCode();
          hc ^= repositoryDomain.hashCode();

          if (userAuth)
          {
               hc += 1;
          }

          if (adminAuth)
          {
               hc += 2;
          }

          if (noPool)
          {
               hc += 4;
          }

          return hc;
     }

     private static SearchControls searchControls /* = null */;
     private DirContext context /* = null */;
     private String baseDN /* = null */;
     private String logicalName /* = null */;
     private String repositoryDomain /* = null */;
     private boolean userAuth = true;
     private boolean adminAuth /* = false */;
     private boolean noPool /* = false */;
     private Hashtable env /* = null */;
     private PasswordCredential privLogin /* = null */;
     private Cluster[] clusters /* = null */;
     private Puddle[] puddles /* = null */;

     static
     {
          searchControls = new SearchControls();
          searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
     }

     protected ConnectionFactoryKernel getKernel(ConnectionFactoryManager manager,ConnectionFactoryConfiguration cxfConfig,Logger logger) throws ConnectionFactoryException
     {
          if (clusters == null && puddles == null)
          {
               try
               {
                    env = context.getEnvironment();
               }
               catch (NamingException ne)
               {
                    throw new ConnectionFactoryException(ne);
               }

               resolve();
          }

          if (puddles != null)
          {
               PuddleEnvelope[] envs = new PuddleEnvelope[puddles.length];

               for (int i=0;i<puddles.length;i++)
               {
                    envs[i] = new PuddleEnvelope(puddles[i]);

                    if (privLogin != null)
                    {
                         envs[i].addUserToCache(privLogin);
                    }
               }

               AuthServer authServer = new JNDIAuthServer(envs,this);
               return new PoolingKernel(manager,cxfConfig,logger,envs,authServer);
          }
          else
          {
               return new AggregationKernel(manager,cxfConfig,logger,clusters);
          }
     }

     private void resolve() throws ConnectionFactoryException
     {
          // look for servers with this logical name - sorted by domain
          Map clusterMap = resolveClusters();

          if (!noPool)
          {
               // look for logins and match them up by domain with the servers
               puddles = resolvePuddles(clusterMap);

               if (puddles != null)
               {
                    return;
               }
          }

          // No logins that matched the servers so just return the servers (grouped by domain).
          // The keys are domains. The values are sets. The elements of the sets are themselves
          // sets of IOMServer objects. The nested sets accommodate server objects in the
          // meta data that have multiple values for the host attribute (i.e. they define clusters).
          // This code flattens that structure so that all servers and clusters that share a domain
          // form a single failover cluster.
          //----------------------------------------------------------------------------------------
          Set mapEntries = clusterMap.entrySet();
          Iterator entryIterator = mapEntries.iterator();
          clusters = new Cluster[mapEntries.size()];
          int i = 0;

          while (entryIterator.hasNext())
          {
               Set serverSet = new HashSet();
               Set outerSet = (Set)((Map.Entry)entryIterator.next()).getValue();
               Iterator outerIterator = outerSet.iterator();

               while (outerIterator.hasNext())
               {
                    Set innerSet = (Set)outerIterator.next();
                    serverSet.addAll(innerSet);
               }

               Server[] servers = new Server[serverSet.size()];
               servers = (Server[])serverSet.toArray(servers);
               Cluster cluster = new FailoverCluster(servers);
               clusters[i++] = cluster;
          }

          return;
     }

     private Map resolveClusters() throws ConnectionFactoryException
     {
          try
          {
               Map clusterMap = new HashMap();
               StringBuffer filterBuffer = new StringBuffer(100);
               filterBuffer.append("(&(objectclass=sasServer)(sasProtocol=bridge)(sasLogicalName=");
               filterBuffer.append(logicalName).append("))");
               String filter = filterBuffer.toString();
               NamingEnumeration serverEnumeration = null;

               try
               {
                    serverEnumeration = context.search(baseDN,filter,searchControls);
               }
               catch (CommunicationException ce)
               {
                    reconnect(ce);
                    serverEnumeration = context.search(baseDN,filter,searchControls);
               }

               while (serverEnumeration.hasMore())
               {
                    SearchResult serverResult = (SearchResult)serverEnumeration.next();
                    Attributes serverAttributes = serverResult.getAttributes();
                    Attribute domainAttr = serverAttributes.get("sasDomainName");
                    String domain = (domainAttr != null) ? (String)domainAttr.get() : "";
                    Set outerSet = (Set)clusterMap.get(domain);

                    if (outerSet == null)
                    {
                         outerSet = new HashSet();
                         clusterMap.put(domain,outerSet);
                    }

                    outerSet.add(resolveServers(serverAttributes,domain));
               }

               return clusterMap;
          }
          catch (NamingException ne)
          {
               throw new ConnectionFactoryException(ne);
          }
     }

     private Puddle[] resolvePuddles(Map clusterMap) throws ConnectionFactoryException
     {
          // We'll want to cache the privileged login in all the puddles since we
          // know this login has permission to use all the puddles.
          //---------------------------------------------------------------------
          if (userAuth || adminAuth)
          {
               String privUserName = (String)env.get(Context.SECURITY_PRINCIPAL);
               String privPassword = (String)env.get(Context.SECURITY_CREDENTIALS);

               if (privUserName != null && privPassword != null)
               {
                    privLogin = new PasswordCredential(normalizeDN(privUserName),privPassword,repositoryDomain);
               }
          }

          try
          {
               // Find all the sasLogin objects that have the given sasLogicalName attribute.
               // I first try the search with (objectclass=top) in the filter. This is
               // to work around a bug in Netscape Directory Server. Unfortunately, not all
               // objects explicilty derive from top so, if the first search turns up
               // nothing, try again without (objectclass=top) in the filter.
               //----------------------------------------------------------------------------
               StringBuffer filterBuffer = new StringBuffer(100);
               filterBuffer.append("(&(objectclass=sasLogin)(objectclass=top)(sasLogicalName=");
               String filter = filterBuffer.append(logicalName).append("))").toString();
               filterBuffer.setLength(0);
               NamingEnumeration loginEnumeration = null;

               try
               {
                    loginEnumeration = context.search(baseDN,filter,searchControls);
               }
               catch (CommunicationException ce)
               {
                    reconnect(ce);
                    loginEnumeration = context.search(baseDN,filter,searchControls);
               }


               if (!loginEnumeration.hasMore())
               {
                    filter = filterBuffer.append("(&(objectclass=sasLogin)(sasLogicalName=").append(logicalName).append("))").toString();
                    filterBuffer.setLength(0);

                    try
                    {
                         loginEnumeration = context.search(baseDN,filter,searchControls);
                    }
                    catch (CommunicationException ce)
                    {
                         reconnect(ce);
                         loginEnumeration = context.search(baseDN,filter,searchControls);
                    }
               }

               if (!loginEnumeration.hasMore())
               {
                    return null;
               }

               Set puddleSet = new HashSet();

               while (loginEnumeration.hasMore())
               {
                    SearchResult searchResult = (SearchResult)loginEnumeration.next();
                    Attributes loginAttributes = searchResult.getAttributes();
                    Attribute domainAttr = loginAttributes.get("sasDomainName");
                    String domain = (domainAttr != null) ? (String)domainAttr.get() : "";
                    Set outerSet = (Set)clusterMap.get(domain);

                    if (outerSet != null)
                    {
                         puddleSet.add(resolvePuddle(searchResult.getName(),loginAttributes,domain,outerSet));
                    }
               }

               int puddleCount = puddleSet.size();

               if (puddleCount == 0)
               {
                    return null;
               }

               Puddle[] puddles = new Puddle[puddleCount];
               puddles = (Puddle[])puddleSet.toArray(puddles);
               return puddles;
          }
          catch (NamingException ne)
          {
               throw new ConnectionFactoryException(ne);
          }

     }

     private static Set resolveServers(Attributes serverAttributes,String domain) throws ConnectionFactoryException
     {
          try
          {
               // LDAP doesn't support anything but SAS Automation servers
               String classID = Server.CLSID_SAS;

               // sasPort = port
               int port = Integer.parseInt((String)(serverAttributes.get("sasPort").get()));

               // sasMaxPerWorkspacePool == maxClients
               Attribute maxPerPoolAttr = serverAttributes.get("sasMaxPerWorkspacePool");
               int maxClients = (maxPerPoolAttr != null) ? Integer.parseInt((String)maxPerPoolAttr.get()) : Integer.MIN_VALUE;

               // sas-RecycleActivationLimit == reactivationLimit
               Attribute recycleActivationLimitAttr = serverAttributes.get("sas-RecycleActivationLimit");
               int reactivationLimit = (recycleActivationLimitAttr != null) ? Integer.parseInt((String)recycleActivationLimitAttr.get()) : Integer.MIN_VALUE;

               // sas-ServerRunForever + sas-ServerShutdownAfter == shutdownAfterMinutes
               Attribute runForeverAttr = serverAttributes.get("sas-ServerRunForever");
               int shutdownAfterMinutes = Integer.MIN_VALUE;

               if (runForeverAttr != null && !Boolean.valueOf((String)runForeverAttr.get()).booleanValue())
               {
                    Attribute shutdownAfterAttr = serverAttributes.get("sas-ServerShutdownAfter");
                    shutdownAfterMinutes = (shutdownAfterAttr != null) ? Integer.parseInt((String)shutdownAfterAttr.get()) : 3;
               }

               // sasRequiredEncryptionLevel == encryptionPolicy + encryptionContent
               // leave encrytpionAlgorithms null so server will pick algorithm
               Attribute requiredEncryptionLevelAttr = serverAttributes.get("sasRequiredEncryptionLevel");
               String encryptionPolicy = null;
               String encryptionContent = null;

               if (requiredEncryptionLevelAttr != null)
               {
                    String requiredEncryptionLevel = ((String)requiredEncryptionLevelAttr.get()).toLowerCase();

                    if (requiredEncryptionLevel.equals("authentication") || requiredEncryptionLevel.equals("credentials"))
                    {
                         encryptionPolicy = BridgeServer.ENCRYPTION_POLICY_REQUIRED;
                         encryptionContent = BridgeServer.ENCRYPTION_CONTENT_AUTHENTICATION;
                    }
                    else if (requiredEncryptionLevel.equals("all") || requiredEncryptionLevel.equals("everything"))
                    {
                         encryptionPolicy = BridgeServer.ENCRYPTION_POLICY_REQUIRED;
                         encryptionContent = BridgeServer.ENCRYPTION_CONTENT_ALL;
                    }
               }

               // sasMachineDNSName == host
               Attribute dnsNames = serverAttributes.get("sasMachineDNSName");
               Set innerSet = new HashSet(dnsNames.size());
               NamingEnumeration dnsNameEnum = dnsNames.getAll();

               while (dnsNameEnum.hasMore())
               {
                    String host = (String)dnsNameEnum.next();
                    BridgeServer server = new BridgeServer(classID,host,port);

                    if (encryptionPolicy != null)
                    {
                         server.setEncryptionPolicy(encryptionPolicy);
                    }

                    if (encryptionContent != null)
                    {
                         server.setEncryptionContent(encryptionContent);
                    }

                    if (maxClients != Integer.MIN_VALUE)
                    {
                         server.setMaxClients(maxClients);
                    }

                    if (reactivationLimit != Integer.MIN_VALUE)
                    {
                         server.setReactivationLimit(reactivationLimit);
                    }

                    if (shutdownAfterMinutes != Integer.MIN_VALUE)
                    {
                         server.setShutdownAfterMinutes(shutdownAfterMinutes);
                    }

                    server.setDomain(domain);
                    innerSet.add(server);
               }

               return innerSet;
          }
          catch (NamingException ne)
          {
               throw new ConnectionFactoryException(ne);
          }
     }

     private Puddle resolvePuddle(String name,Attributes loginAttributes,String domain,Set outerSet) throws ConnectionFactoryException
     {
          try
          {
               String userName = (String)(loginAttributes.get("sasLoginName").get());
               String password = SasPasswordString.decode((String)(loginAttributes.get("sasUserPassword").get()));
               PasswordCredential login = new PasswordCredential(userName,password,domain);

               // sasMinSize == minSize
               Attribute sasMinSizeAttr = loginAttributes.get("sasMinSize");
               int minSize = (sasMinSizeAttr != null) ? Integer.parseInt((String)sasMinSizeAttr.get()) : 0;

               // sasMinAvail == minAvail
               Attribute sasMinAvailAttr = loginAttributes.get("sasMinAvail");
               int minAvail = (sasMinAvailAttr != null) ? Integer.parseInt((String)sasMinAvailAttr.get()) : 0;

               // sasReferenceDN + sasAllowedClientDN == authorizedUsers
               Set authorizedUsers = null;

               if (userAuth)
               {
                    authorizedUsers = new HashSet();
                    Attribute sasReferenceDNAttr = loginAttributes.get("sasReferenceDN");

                    if (sasReferenceDNAttr != null)
                    {
                         NamingEnumeration values = sasReferenceDNAttr.getAll();

                         while (values.hasMore())
                         {
                              authorizedUsers.add(normalizeDN((String)values.next()));
                         }
                    }

                    Attribute sasAllowedClientDNAttr = loginAttributes.get("sasAllowedClientDN");

                    if (sasAllowedClientDNAttr != null)
                    {
                         NamingEnumeration values = sasAllowedClientDNAttr.getAll();

                         while (values.hasMore())
                         {
                              authorizedUsers.add(normalizeDN((String)values.next()));
                         }
                    }
               }

               String puddleID = null;

               if (adminAuth)
               {
                    puddleID = name + "," + baseDN;
               }

               Cluster[] clusters = new Cluster[outerSet.size()];
               Iterator outerIterator = outerSet.iterator();
               int i = 0;

               while (outerIterator.hasNext())
               {
                    Set innerSet = (Set)outerIterator.next();
                    Server[] servers = new Server[innerSet.size()];
                    servers = (Server[])innerSet.toArray(servers);
                    clusters[i++] = new FailoverCluster(servers);
               }

               Puddle puddle = new Puddle(clusters,login);
               puddle.setMinSize(minSize);
               puddle.setMinAvail(minAvail);
               puddle.setUsers(authorizedUsers);
               puddle.setPuddleID(puddleID);
               return puddle;
          }
          catch (NamingException ne)
          {
               throw new ConnectionFactoryException(ne);
          }
          catch (SasPasswordEncodingException spee)
          {
               throw new ConnectionFactoryException(spee);
          }
     }

     protected Hashtable getEnv()
     {
          return env;
     }

     protected PasswordCredential getPrivLogin()
     {
          return privLogin;
     }

     protected void reconnect(CommunicationException ce) throws CommunicationException
     {
          try
          {
               context = new InitialDirContext(env);
          }
          catch (NamingException ne)
          {
               throw ce;
          }
     }

     // To normalize a dn, lowercase it, remove leading and trailing whitespace,
     // and remove whitespace before and after separators. Whitespace characters
     // are ' ', '\t', and '\n'. Separator characters are '=', ',' and ';'. Whitespace
     // is not removed from the inside of quotation marks ('"'), and quotation
     // marks can be escaped with a backslash ('\\'). Before changing anything,
     // run through the string to see if anything needs to be changed. If nothing
     // needs to be changed, return the input string. This is just a quick check
     // that ignores quotes so we may end up creating a new a string for a DN that
     // doesn't really need to be changed.
     public static String normalizeDN(String dn)
     {
          final int length = dn.length();

          if (length == 0)
          {
               return dn;
          }

          boolean normalize = false;
          int i = 1;
          int before = -1;
          int current = dn.charAt(0);
          int after = (length>1)?dn.charAt(1):-1;
          int last = dn.charAt(length-1);

          if (isWhite(current) || isWhite(last))
          {
               normalize = true;
          }

          while (current >= 0 && !normalize)
          {
               if (Character.isUpperCase((char)current))
               {
                    normalize = true;
               }
               else if ((current == '=' || current == ',' || current == ';') && (isWhite(before) || isWhite(after)))
               {
                    normalize = true;
               }
               else
               {
                    before = current;
                    current = after;
                    after = (++i<length)?dn.charAt(i):-1;
               }
          }

          if (!normalize)
          {
               return dn;
          }

          StringBuffer normalBuffer = new StringBuffer(length);
          i = 0;
          int state = 0;
          int white = 0;

          for (;i<length;i++)
          {
               char c = dn.charAt(i);

               switch (state)
               {
                    case 0: // skip whitespace
                    {
                         if (!isWhite(c))
                         {
                              normalBuffer.append(Character.toLowerCase(c));
                              state = (c=='"')?2:1;
                         }

                         break;
                    }

                    case 1: // count whitespace and rewind if c is a separator
                    {
                         if (isWhite(c))
                         {
                              white++;
                              normalBuffer.append((char)c);
                         }
                         else if (c == '=' || c == ',' || c == ';')
                         {
                              normalBuffer.setLength(normalBuffer.length()-white);
                              normalBuffer.append(c);
                              white = 0;
                              state = 0;
                         }
                         else if (c == '"')
                         {
                              normalBuffer.append(c);
                              white = 0;
                              state = 2;
                         }
                         else
                         {
                              normalBuffer.append(Character.toLowerCase(c));
                              white = 0;
                         }

                         break;
                    }

                    case 2: // look for ending quote
                    {
                         if (c == '"')
                         {
                              normalBuffer.append(c);
                              state = 1;
                         }
                         else if (c == '\\')
                         {
                              normalBuffer.append(c);
                              state = 3;
                         }
                         else
                         {
                              normalBuffer.append(Character.toLowerCase(c));
                         }

                         break;
                    }

                    case 3: // last character was escape
                    {
                         state = 2;
                         normalBuffer.append(Character.toLowerCase(c));
                         break;
                    }
               }
          }

          if (white > 0)
          {
               normalBuffer.setLength(normalBuffer.length()-white);
          }

          return normalBuffer.toString();
     }

     private static boolean isWhite(int c)
     {
          return (c ==  ' ' || c == '\t' || c == '\n');
     }


}