/* $Id$ */
// Copyright (c) 2001-2004 SAS Institute Inc., Cary, NC 27513
// All Rights Reserved.

package com.sas.services.webdav;

import java.io.InputStream;
import java.net.URL;
import java.net.MalformedURLException;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.httpclient.HttpStatus;
import org.apache.webdav.lib.methods.DepthSupport;

import com.sas.services.logging.LoggerInterface;
import com.sas.text.Message;

/**
 * This class represents a DAV collection and the operations expected of a collection. Collections can
 * contain other collections and/or resources (known as items).
 * <p>
 * @author  Rob Owen
 * @version 1.0 15-Nov-2001
 */
public class Collection extends Resource implements CollectionInterface
{

	/**
	 * Create a new collection in an existing collection. The model collection is used as the location in which to create
	 * the new collection. This constructor also reuses the model's connection for the new collection.
	 * @param model The collection in which to create the new collection and the connection to use for the new collection
	 * @param name The name of the collection
	 * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
	 */
	public static CollectionInterface createCollection( CollectionInterface model, String name ) throws DAVException, RemoteException {
		String _url = Utils.addSegment( model.getURL(), name );
		String path = getPath(_url);
	    if ( model.isSharedResource() ) {
	        SharedResourceInterface sr = model.getSharedResource();
	        if ( sr instanceof SharedResource ) {
		        try {
					DAVResourceInterface r = sr.getConnection();
					r.setPath( path );
					r.makeCollection();
					return new Collection( sr, _url );
				} finally {
				    sr.releaseConnection();
				}
	        } else {
	            return sr.createCollection(model, name);
	        }
	    } else {
			DAVResource r = ((Collection)model)._resource;
			synchronized(r) {
				r.setPath( path );
				r.makeCollection();
				return new Collection( r );
			}
	    }
	}
	
	/**
	 * Bind to a new resource in the existing model collection. This constructor also reuses the model's 
	 * connection for the new resource.
	 * @param model The collection in which the resource will reside
	 * @param name The name of the resource
	 * @return ResourceInterface The new resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
	 */
	public static ResourceInterface createResource( CollectionInterface model, String name ) throws DAVException, RemoteException {
		String _url = Utils.addSegment( model.getURL(), name );
		String path = getPath(_url);

        if ( model.isSharedResource() ) {
	        SharedResourceInterface sr = model.getSharedResource();
	        if ( sr instanceof SharedResource ) {
		        try {
					DAVResourceInterface r = sr.getConnection();
					r.setPath( path );
					return new Resource( sr, _url );
				} finally {
				    sr.releaseConnection();
				}
	        } else {
	            return sr.createResource(model, name);
	        }
	    } else {
			DAVResource r = ((Collection)model)._resource;
			synchronized(r) {
				r.setPath( path );
				return new Resource( r );
			}
	    }
	}

	/**
	 * Create a new collection in an existing collection. The model collection is used as the location in which to create
	 * the new collection. This constructor also reuses the model's connection for the new collection.
	 * @param model The collection in which to create the new collection and the connection to use for the new collection
	 * @param name The name of the collection
	 * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
 */
	public static CollectionInterface createCollection( SharedResourceInterface sr, CollectionInterface model, String name ) throws DAVException, RemoteException {
		String _url = Utils.addSegment( model.getURL(), name );
		String path = getPath(_url);
        try {
			DAVResourceInterface r = sr.getConnection();
			r.setPath( path );
			r.makeCollection();
			return new Collection( sr, _url );
		} finally {
		    sr.releaseConnection();
		}
	}
	
	/**
	 * Bind to a new resource in the existing model collection. This constructor also reuses the model's 
	 * connection for the new resource.
	 * @param model The collection in which the resource will reside
	 * @param name The name of the resource
	 * @return ResourceInterface The new resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
	 */
	public static ResourceInterface createResource( SharedResourceInterface sr, CollectionInterface model, String name ) throws DAVException, RemoteException {
		String _url = Utils.addSegment( model.getURL(), name );
		String path = getPath(_url);

        try {
			DAVResourceInterface r = sr.getConnection();
			r.setPath( path );
			return new Resource( sr, _url );
		} finally {
		    sr.releaseConnection();
		}
	}

    /**
     * Create a new collection using the shared resource provided
     * @param sr The shared resource on which to base this Collection object
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
	public Collection (SharedResourceInterface sr) throws DAVException, RemoteException
    {
        super(sr);
    }

    /**
     * Create a new collection using the shared resource provided and represents the WebDAV resource
     * at location urlStr
     * @param sr The shared resource on which to base this Collection object
     * @param urlStr The url of the WebDAV resource that this collection represents
     */
	public Collection (SharedResourceInterface sr, String urlStr)
    {
        super(sr, urlStr);
    }

    /**
     * Create a new collection in an existing collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
	public Collection (DAVResource r) throws DAVException, RemoteException
    {
        super(r);
    }

    /**
     * Create a new collection in an existing collection
     * @param collName The name of the collection to be created
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection( String collName )
        throws DAVException, RemoteException
    {
        return createCollection( collName, null, null, null );
    }

    /**
     * Create a new collection in the existing collection
     *
     * @param collName The name of the collection to be created
     * @param logger The logger to be utilized for reporting
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection( String collName, LoggerInterface logger )
        throws DAVException, RemoteException
    {
        return createCollection( collName, null, null, logger );
    }

    /**
     * Create a new collection in the existing collection
     *
     * @param collName The name of the collection to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection(String collName, String userName, String pwd)
        throws DAVException, RemoteException
    {
        return createCollection( collName, userName, pwd, null );
    }

    /**
     * Create a new collection in the existing collection
     *
     * @param collName The name of the collection to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @param logger The logger to be utilized for reporting
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection(String collName, String userName, String pwd, LoggerInterface logger)
        throws DAVException, RemoteException
    {
		if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
        String _url = _resource.getURL().toString();
        String fullUrl = Utils.addSegment( _url, collName );
        if (fullUrl.charAt(fullUrl.length()-1) != '/') fullUrl += "/";
        ConnectionInfo cInfo = new ConnectionInfo( fullUrl, userName, pwd );
        if ( _resource.proxyHost != null && _resource.proxyPort >= 0 ) {
            cInfo.setProxyHost( _resource.proxyHost );
            cInfo.setProxyPort( _resource.proxyPort );
        }
        DAVResource newcoll = new DAVResource( cInfo, true, logger);
        if ( !newcoll.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Collection.makecol.ex.fmt.txt",
                                                                      newcoll.getStatusMessage() ) );
        return new Collection( newcoll );
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource( String resName )
        throws DAVException, RemoteException
    {
        return createResource( resName, null, null, null );
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource(String resName, String userName, String pwd)
        throws DAVException, RemoteException
    {
        return createResource( resName, userName, pwd, null );
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @param logger The logger to be utilized for reporting
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource(String resName, LoggerInterface logger)
        throws DAVException, RemoteException
    {
        return createResource( resName, null, null, logger );
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @param logger The logger to be utilized for reporting
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource(String resName, String userName, String pwd, LoggerInterface logger)
        throws DAVException, RemoteException
    {
		if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
        String _url = _resource.getURL().toString();
        String fullUrl = Utils.addSegment( _url, resName );
        ConnectionInfo cInfo = new ConnectionInfo( fullUrl, userName, pwd );
        if ( _resource.proxyHost != null && _resource.proxyPort >= 0 ) {
            cInfo.setProxyHost( _resource.proxyHost );
            cInfo.setProxyPort( _resource.proxyPort );
        }
        DAVResource newres = new DAVResource( cInfo, false, logger);
        if ( !newres.SUCCEEDED() &&
             newres.getStatusCode() != HttpStatus.SC_NOT_FOUND )
             throw new DAVException( Message.format( RB.getResources(), "Collection.createres.ex.fmt.txt",
                                                                      newres.getStatusMessage() ) );
        return new Resource( newres );
    }

    /**
     * Get a list of the immediate items (depth=1) in the collection
     * @return List (of ResourceInterface/CollectionInterface) The list of items found in the collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public List getChildren() throws DAVException, RemoteException
    {
        List members = new ArrayList();
        org.apache.webdav.lib.Property[] pp = null;
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
                int savedDepth = r.getDepth();
                r.setDepth(DepthSupport.DEPTH_1);
                pp = r.getProperty("<D:resourcetype/>");
                if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Collection.getprop.ex.fmt.txt",
                                                                              r.getStatusMessage() ) );
                r.setDepth( savedDepth );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
	        int savedDepth = _resource.getDepth();
			if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
	        _resource.setDepth(DepthSupport.DEPTH_1);
	        pp = _resource.getProperty("<D:resourcetype/>");
	        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Collection.getprop.ex.fmt.txt",
	                                                                      _resource.getStatusMessage() ) );
	        _resource.setDepth( savedDepth );
        }
        for (int i=0; i<pp.length; i++)
        {
            String newurl = pp[i].getOwningURL();
            if (newurl.indexOf("://") < 0)
            {
                String tempurl; 
                try {
                    URL u = new URL(getURL());
                    tempurl = u.getProtocol()+"://"+u.getHost();
                    if (u.getPort() != -1 && u.getPort() != 80) tempurl += ':' + u.getPort();
                } catch (MalformedURLException e) {
                    tempurl = "";
                }
                newurl = tempurl + newurl; /*I18nOK:COS*/
            }
            if (!Utils.isSameUrl(newurl, getURL())) {
                ResourceInterface member;
                if (isSharedResource) {
                    member = ResourceFactory.bindResource(sharedResource, getPath(newurl));
                } else {
	                if (!_resource.isAuthReqd()) {
	                    member = ResourceFactory.bindResource(newurl);
	                }
	                else {
	                    member = ResourceFactory.bindResource( newurl, _resource.userName, _resource.password );
	                }
                }
                members.add(member);
            }
        }
        return members;
    }

    /**
     * Get all the properties of a collection's members (depth=1) in as single server request. This
     * does not bind to each one of the items in the collection and is a quick way to find out what
     * items are in the collection without fetching each one.
     * @return Map (maps urls to (Map of property name to value)) The property map for all the
     * items in this collection.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public Map getChildrenProperties() throws DAVException, RemoteException
    {
        HashMap _members        = new HashMap();
        String _newUrl          = null;
        String _currentUrl      = null;
        String _tempUrl         = null;
        PropertyMap _currentMap = null;

        org.apache.webdav.lib.Property[] pp = null;
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
                int savedDepth = r.getDepth();
                r.setDepth(DepthSupport.DEPTH_1);
                pp = r.getProperties();
                if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Collection.getprop.ex.fmt.txt",
                                                                              r.getStatusMessage() ) );
                r.setDepth( savedDepth );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
			int savedDepth = _resource.getDepth();
	        _resource.setDepth(DepthSupport.DEPTH_1);
	        pp = _resource.getProperties();
	        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Collection.getprop.ex.fmt.txt",
	                                                                      _resource.getStatusMessage() ) );
	        _resource.setDepth(savedDepth);
        }

        for (int i=0; i<pp.length; i++)
        {
            _newUrl = pp[i].getOwningURL();
            if (_newUrl.indexOf("://") < 0)
            {
                try {
                    URL u = new URL(getURL());
                    _tempUrl = u.getProtocol()+"://"+u.getHost();
                    if (u.getPort() != -1 && u.getPort() != 80) _tempUrl += ":" + u.getPort();
                } catch (MalformedURLException e) {
                    _tempUrl = "";
                }
                _newUrl = _tempUrl + _newUrl; /*I18nOK:COS*/
            }

            // PROPFIND will return properties for the parent collection too. Filter out these
            // properties for this method
            if ( Utils.isSameUrl( _newUrl, getURL() )) continue;


            if (!_newUrl.equals(_currentUrl))
            {
                _currentUrl = _newUrl;
                _currentMap = (PropertyMap) _members.get(_newUrl);
                if (_currentMap == null)
                {
                    _currentMap = new PropertyMap();
                    _members.put(_newUrl, _currentMap);
                }
            }
            if (pp[i].getNamespaceURI() != null)
                _currentMap.put('{'+pp[i].getNamespaceURI()+'}'+pp[i].getLocalName(), pp[i].getPropertyAsString());
            else
                _currentMap.put("{}"+pp[i].getLocalName(), pp[i].getPropertyAsString());
        }
        return _members;
    }

    /**
     * Returns null for the contents of a collection
     * @return null
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public InputStream getContents() throws DAVException, RemoteException {
		return null;
    }
    
    /**
     * Extract the path/query from the url. This is the part after the scheme, host and port.
     * @param urlStr URL
     * @return String The path/query portion of the url 
     */
    private static String getPath(String urlStr) {
		String path = null;
        int p = urlStr.indexOf("://"); /*I18nOK:EMS*/
        if ( p > 0 ) {
            p = urlStr.indexOf("/", p+3);
            path = (  p > 0 ) ? urlStr.substring(p) : "";
        } else path=urlStr;
        return path;
        
    }
}