package com.sas.services.webdav;

import java.net.URL;
/**
 * Identifies all the information required to make a connection to a DAV server. This information includes the url of the resource (containing the 
 * host, port and path), userid, password, proxy host, proxy port, proxy userid, proxy password. This object is then passed to a connection constructor.
 * For example, if one was using this to create a DAVResource, then use something similar to the following:
 * <pre>
 * ConnectionInfo c = new ConnectionInfo( "http://example.com/dav/resource", "user", "pwd" );
 * c.setproxyHost( "myproxy.com" );
 * c.setproxyPort( 3100 );
 * DAVResource r = new DAVResource( c, _logger );
 * </pre>
 *
 * @author Rob Owen
 * @version 1.1 30-Oct-2002
 * @since   1.1
 */
public class ConnectionInfo {
    
    private String _url       = null;
    private String _user      = null;
    private String _pw        = null;
    private String _proxyHost = null;
    private int    _proxyPort = -1;
    private String _proxyUser = null;
    private String _proxyPW   = null;
    
    public ConnectionInfo( String url ) {
        _url  = url;
    }
    
    public ConnectionInfo( String url, String user, String pw ) {
        _url  = url;
        _user = user;
        _pw   = pw;
    }
    
    public ConnectionInfo( ConnectionInfo connection ) {
        _url = connection.getUrl();
        _user = connection.getUser();
        _pw = connection.getPW();
        _proxyHost = connection.getProxyHost();
        _proxyPort = connection.getProxyPort();
        _proxyUser = connection.getProxyUser();
        _proxyPW = connection.getProxyPW();
    }
    
    /**
     * Returns the url from the connection information
     * @return String The URL in the connection information 
     */
    public String getUrl() {
        return _url;
    }
    
    /**
     * Sets the url in the connection information
     * @param url The url in the connection information
     */
    public void setUrl( String url ) {
        _url = url;
    }
        
    /**
     * Returns the user from the connection information
     * @return String The user in the connection information
     */
    public String getUser() {
        return _user;
    }
    
    /**
     * Sets the user in the connection information
     * @param user The user in the connection information
     */
    public void setUser( String user ) {
        _user = user;
    }
        
    /**
     * Returns the user password from the connection information
     * @return String The user password in the connection information
     */
    public String getPW() {
        return _pw;
    }
    
    /**
     * Sets the user password in the connection information
     * @param pw The user password in the connection information
     */
    public void setPW( String pw ) {
        _pw = pw;
    }
        
    /**
     * Returns the proxy host from the connection information
     * @return String The proxy host in the connection information
     */
    public String getProxyHost() {
        return _proxyHost;
    }
    
    /**
     * Sets the proxy host in the connection information
     * @param proxyHost The proxy host in the connection information
     */
    public void setProxyHost( String proxyHost ) {
        _proxyHost = proxyHost;
    }
        
    /**
     * Returns the proxy port from the connection information
     * @return String The proxy port in the connection information
     */
    public int getProxyPort() {
        return _proxyPort;
    }
    
    /**
     * Sets the proxy port in the connection information
     * @param proxyPort The proxy port in the connection information
     */
    public void setProxyPort( int proxyPort ) {
        _proxyPort = proxyPort;
    }
        
    /**
     * Sets the proxy in the connection information using a URL
     * @param proxyURL The proxy URL includes the host and port
     */
    public void setProxyURL( URL proxyURL ) {
        _proxyHost = proxyURL.getHost();
        _proxyPort = proxyURL.getPort();
    }
    
    /**
     * Returns the proxy user from the connection information
     * @return String The proxy user in the connection information 
     */
    public String getProxyUser() {
        return _proxyUser;
    }
    
    /**
     * Sets the proxy user in the connection information
     * @param proxyUser The proxy user in the connection information
     */
    public void setProxyUser( String proxyUser ) {
        _proxyUser = proxyUser;
    }
        
    /**
     * Returns the proxy password from the connection information
     * @return String The proxy password in the connection information
     */
    public String getProxyPW() {
        return _proxyPW;
    }
    
    /**
     * Sets the proxy password in the connection information
     * @param proxyPW The proxy password in the connection information
     */
    public void setProxyPW( String proxyPW ) {
        _proxyPW = proxyPW;
    }
        
}
