/* $Id$ */

// Copyright (c) 2001-2004 SAS Institute Inc., Cary, NC 27513
// All Rights Reserved.

package com.sas.services.webdav;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.rmi.RemoteException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.httpclient.Cookie;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HeaderElement;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.State;
import org.apache.util.HttpURL;
import org.apache.util.HttpsURL;
import org.apache.util.URIUtil;
import org.apache.util.WebdavStatus;
import org.apache.webdav.lib.Lock;
import org.apache.webdav.lib.Property;
import org.apache.webdav.lib.ResponseEntity;
import org.apache.webdav.lib.WebdavResource;
import org.apache.webdav.lib.WebdavState;
import org.apache.webdav.lib.methods.AclMethod;
import org.apache.webdav.lib.methods.CheckinMethod;
import org.apache.webdav.lib.methods.CheckoutMethod;
import org.apache.webdav.lib.methods.CopyMethod;
import org.apache.webdav.lib.methods.DepthSupport;
import org.apache.webdav.lib.methods.GetMethod;
import org.apache.webdav.lib.methods.LabelMethod;
import org.apache.webdav.lib.methods.LockMethod;
import org.apache.webdav.lib.methods.MkcolMethod;
import org.apache.webdav.lib.methods.MoveMethod;
import org.apache.webdav.lib.methods.OptionsMethod;
import org.apache.webdav.lib.methods.PropFindMethod;
import org.apache.webdav.lib.methods.PropPatchMethod;
import org.apache.webdav.lib.methods.PutMethod;
import org.apache.webdav.lib.methods.ReportMethod;
import org.apache.webdav.lib.methods.SearchMethod;
import org.apache.webdav.lib.methods.UncheckoutMethod;
import org.apache.webdav.lib.methods.UnlockMethod;
import org.apache.webdav.lib.methods.VersionControlMethod;
import org.apache.webdav.lib.methods.XMLResponseMethodBase;
import org.apache.webdav.lib.properties.LockDiscoveryProperty;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.sas.io.InputStreamAdapter;
import com.sas.services.logging.LoggerInterface;
import com.sas.services.session.SessionContextInterface;
import com.sas.services.util.XMLUtil;
import com.sas.text.Message;


/**
 *
 * A DAVResource represents a WEBDAV (Distributed Authoring and Versioning) resource as specified in IETF RFC 2518.
 * To create a DAVResource, use a ConnectionInfo object passed to the constructor. ConnectionInfo includes the url
 * of the resource (containing the host, port and path), userid, password, proxy host, proxy port, proxy userid,
 * proxy password. The DAVResource object can then be manipulated using the various methods defined in the
 * Resource interface. After any operation on a resource, check the status on the object
 * using getStatusCode() and getStatusMessage(). For example,
 * <pre>
 * try {
 *     ConnectionInfo c = new ConnectionInfo( "http://http-server/path-to-resource", "user", "pwd" );
 *     // proxy information, if required
 *     c.setProxyHost( "myproxy.com" );
 *     c.setProxyPort( 3100 );
 *
 *     DAVResource r1 = new DAVResource( c, logger );
 *     r1.setProperty(...);
 *     if ( r1.getStatusCode() != 200 ) out.println( r1.getStatusMessage() );
 * } catch (DAVException e) { e.printStackTrace(); }
 * </pre>
 *
 * In order to create a collection use the longer form of the constructor. For example,
 * <pre>
 * DAVResource r2 = new DAVResource( c, true, logger );
 * </pre>
 * this will ensure that r2 is a collection resource (it may contain other resources). Alternatively do this in two steps
 * <pre>
 * DAVResource r2 = new DAVResource( c, logger );
 * r2.makeCollection();
 * </pre>
 *
 * @author Rob Owen
 * @version 1.0 10-Sep-2001
 */
public class DAVResource implements DAVResourceInterface {

    private    HttpURL httpURL;
    // Note for using WebdavResource methods - when calling DAV methods on WebdavResource, pass the
    // proper requestURI (ie. escaped), as this is unescaped to call the actual DAV method. For calling
    // the actual DAV method directly, then do not pass an escaped string.
    // eg. assume http://host/dav/test(1)
    // webdavResource.lockMethod( "/test%281%29" );   // httpURL.getEscapedPath()
    // new LockMethod( "/test(1)" );                  // httpURL.getPath()
    private    WebdavResource webdavResource = null;
    private    boolean overwrite = false;
    private    int depth = DepthSupport.DEPTH_0;
    private    int debug = 0;
    private    boolean doneInitialPropfind = false;

    /**
     * Contains the user name used to connect to the server
     */
    protected  String userName = null;

    /**
     * Contains the password used to connect to the server
     */
    protected  String password = null;
    private    int status = 0;
    private    String[] allowedMethods = null;
    private    String[] davCapabilities = null;
    private    boolean authReqd = false;
    private    static LoggerInterface _logger = null;
    private    Map _multiStatus = new HashMap();
    private    SessionContextInterface _sessionContext;
    private static String DEBUG_OPT = System.getProperty("com.sas.services.webdav.debug");
    /**
     * Contains the name of the proxy host, if one is used
     */
    protected String proxyHost = null;
    /**
     * Contains the value of the proxy port, if one is used
     */
    protected int proxyPort = -1;

    private static final String LOGGING_CONTEXT = DAVResource.class.getName();
	private static final String PERFORMANCE_CONTEXT = DAVResource.class.getName()+".perf";
	private static File tempFileDir = null;
    private static final String USER_AGENT_HDR = "User-Agent"/*I18nOK:EMS*/;

    private static final String PUT       = "PUT"       /*I18nOK:EMS*/;
    private static final String GET       = "GET"       /*I18nOK:EMS*/;
    private static final String ACL       = "ACL"       /*I18nOK:EMS*/;
    private static final String PROPFIND  = "PROPFIND"  /*I18nOK:EMS*/;
    private static final String PROPPATCH = "PROPPATCH" /*I18nOK:EMS*/;
    private static final String REPORT    = "REPORT"    /*I18nOK:EMS*/;
    private static final String OPTIONS   = "OPTIONS"   /*I18nOK:EMS*/;
    private static final String LOCK      = "LOCK"      /*I18nOK:EMS*/;
    private static final String UNLOCK    = "UNLOCK"    /*I18nOK:EMS*/;
    private static final String COPY      = "COPY"      /*I18nOK:EMS*/;
    private static final String MOVE      = "MOVE"      /*I18nOK:EMS*/;
    private static final String MKCOL     = "MKCOL"     /*I18nOK:EMS*/;
    private static final String DELETE    = "DELETE"    /*I18nOK:EMS*/;
    private static final String SEARCH    = "SEARCH"    /*I18nOK:EMS*/;
    private static final String VERSIONCONTROL  = "VERSION-CONTROL"    /*I18nOK:EMS*/;
    private static final String CHECKIN         = "CHECKIN"            /*I18nOK:EMS*/;
    private static final String CHECKOUT        = "CHECKOUT"           /*I18nOK:EMS*/;
    private static final String UNCHECKOUT      = "UNCHECKOUT"         /*I18nOK:EMS*/;
    private static final String LABEL           = "LABEL"              /*I18nOK:EMS*/;
    private static final String UPDATE          = "UPDATE"             /*I18nOK:EMS*/;
    private static final String MERGE           = "MERGE"              /*I18nOK:EMS*/;
    private static final String BASELINECONTROL = "BASELINE-CONTROL"   /*I18nOK:EMS*/;
    private static final String MKWORKSPACE     = "MKWORKSPACE"        /*I18nOK:EMS*/;
    private static final String MKACTIVITY      = "MKACTIVITY"         /*I18nOK:EMS*/;

    private static final String CONTENTTYPE   = "Content-Type"      /*I18nOK:EMS*/;
    private static final String CONTENTLENGTH = "Content-Length"    /*I18nOK:EMS*/;



    /* --------------------------------- CONSTRUCTORS ----------------------------------*/

    /**
     * Create a DAVResource using a string URL
     *
     * @param url URL for the resource
     */
    public DAVResource(String url) throws DAVException {
        this( new ConnectionInfo(url), null, null );
    }

    /**
     * Create a DAVResource using a string URL. DAVResource will use the loggerinterface for logging.
     *
     * @param url URL for the resource
     * @param logger LoggerInterface to be used for logging
     */
    public DAVResource(String url, LoggerInterface logger) throws DAVException {
        this( new ConnectionInfo(url), logger, null );
    }

    /**
     * Create a DAVResource using a string URL, userName and password
     *
     * @param url URL for the resource
     * @param userName  authentication user name
     * @param password  password for authentication
     */
    public DAVResource(String url, String userName, String password) throws DAVException {
        this( new ConnectionInfo(url, userName, password), null, null );
    }

    /**
     * Create a DAVResource using a string URL, userName and password. The specified logger will be used to
     * log DAV activity.
     *
     * @param url URL for the resource
     * @param userName  authentication user name
     * @param password  password for authentication
     * @param logger The logger to be used by DAVResource
     */
    public DAVResource(String url, String userName, String password, LoggerInterface logger) throws DAVException {
        this( new ConnectionInfo(url, userName, password), logger, null );
    }

    /**
     * Create a DAVResource using a Java URL
     *
     * @param url URL for the resource
     */
    public DAVResource(java.net.URL url) throws DAVException {
        this( new ConnectionInfo(url.toString()), null, null );
    }

    /**
     * Create a DAVResource using a string URL and boolean flag to indicate whether a collection should be created or not
     *
     * @param url URL for the resource
     * @param createCollection The URL indicates a collection resource
     */
    public DAVResource(String url, boolean createCollection) throws DAVException {
        this(new ConnectionInfo(url), createCollection, null, null );
    }

    /**
     * Create a DAVResource using a URL and session context and domain. The user context contained within the session
     * context, along with the authentication domain, will be used to determine the credentials that will be used
     * when binding to the DAV server. The logger contained within the session context will be used by the DAVResource.
     *
     * @param url URL for the resource
     * @param sessionContext The session context
     * @param domain The authentication domain
     */

    /*
     public DAVResource(String url, SessionContextInterface sessionContext, String domain) throws DAVException
     {
     _sessionContext = sessionContext;
     _domain = domain;
     this( url, createCollection, "", "" );
     }
     */

    /**
     * Create a DAVResource using a string URL. This constructor can also create a collection if createCollection is true
     *
     * @param url URL for the resource
     * @param createCollection create a DAV collection if true
     * @param userName  authentication user name
     * @param password  password for authentication
     */
    public DAVResource(String url, boolean createCollection, String userName, String password) throws DAVException {
        this( new ConnectionInfo(url, userName, password), createCollection, null, null );
    }

    /**
     * Create a DAVResource using a string URL. This constructor can also create a collection if createCollection is true.
     * The logger is used for DAVResource logging.
     *
     * @param url URL for the resource
     * @param createCollection create a DAV collection if true
     * @param userName  authentication user name
     * @param password  password for authentication
     * @param logger The logger to be used
     */
    public DAVResource(String url, boolean createCollection, String userName, String password, LoggerInterface logger) throws DAVException {
        this( new ConnectionInfo(url, userName, password), createCollection, logger, null );
    }

    /**
     * Create a DAVResource using the connection information. The logger is used for DAVResource logging.
     *
     * @param cInfo The connection information for the resource
     * @param logger The logger to be used
     */
    public DAVResource( ConnectionInfo cInfo, LoggerInterface logger) throws DAVException {
        this( cInfo, false, logger, null );
    }

	/**
	 * Create a DAVResource using the connection information. The logger is used for DAVResource logging.
	 * For internal use only.
	 * @param cInfo The connection information for the resource
	 * @param logger The logger to be used
	 * @param session The session cookie to be used
	 */
	public DAVResource( ConnectionInfo cInfo, LoggerInterface logger, Cookie session ) throws DAVException {
		this( cInfo, false, logger, session );
	}

	/**
	 * Create a DAVResource using the connection information. The logger is used for DAVResource logging.
	 *
	 * @param cInfo The connection information for the resource
	 * @param createCollection create a DAV collection if true
	 * @param logger The logger to be used
	 */
	public DAVResource( ConnectionInfo cInfo, boolean createCollection, LoggerInterface logger ) throws DAVException {
		this( cInfo, createCollection, logger, null );
	}

    /**
     * Create a DAVResource using the connection information. The logger is used for DAVResource logging.
     * For internal use only
     * @param cInfo The connection information for the resource
     * @param createCollection create a DAV collection if true
     * @param logger The logger to be used
     * @param session The session cookie to be used
     */
    public DAVResource( ConnectionInfo cInfo, boolean createCollection, LoggerInterface logger, Cookie session ) throws DAVException {

		long start = System.currentTimeMillis();
        String url = cInfo.getUrl();
        _logger = logger;
        // _sessionContext = session;
        if (_logger != null && _logger.isDebugEnabled(LOGGING_CONTEXT) )
        {
        	debug = DAVConstants.DEFAULT_DEBUG_TRACE;
            if (cInfo.getUser() != null && !cInfo.getUser().equals("")) _logger.debug( Message.format( RB.getResources(), "DAVResource.constructor.fmt.txt", url, cInfo.getUser()), LOGGING_CONTEXT );
            else _logger.debug( Message.format( RB.getResources(), "DAVResource.constructornouser.fmt.txt", url), LOGGING_CONTEXT );
        }

        this.userName = cInfo.getUser();
        this.password = cInfo.getPW();
        this.proxyHost = cInfo.getProxyHost();
        this.proxyPort = cInfo.getProxyPort();

        try {
            // check this is a valid http url
            URL _url = null;
            String testUrl = url;

            // java.net.URL only supports http://
            // if (testUrl.startsWith("https://")) testUrl = "http://" + url.substring(8);
            try {
                _url = new URL(testUrl);
            } catch (MalformedURLException e) {
                throw new DAVException( Message.format(RB.getResources(), "DAVResource.invalidhttpurl.ex.fmt.txt", url), e);
            }

            // If it is valid, continue
            try {
                String userinfo = _url.getUserInfo();
                String user = null;
                String pw = null;
                int ptr = 0;

                if (userinfo != null) {
                    ptr = userinfo.indexOf(':');
                    if (ptr > 0 && ptr < userinfo.length() - 1) {
                        user = userinfo.substring(0, ptr);
                        pw = userinfo.substring(ptr + 1);
                        if ( userName == null ) userName = user;
                        if ( password == null ) password = pw;
                    }
                }

                int port = _url.getPort();
                String path = _url.getPath();
                if ( _url.getQuery() != null ) path += '?' + _url.getQuery();/*I18nOK:COS*/
                if ( _url.getRef() != null ) path += '#' + _url.getRef();/*I18nOK:COS*/
                try {
                    URI uri = new URI(null, null, path, null);
                    path = uri.toASCIIString();
                } catch ( URISyntaxException use ) {
                    // path is already set
                }
                if( "http".equalsIgnoreCase( _url.getProtocol() ) ) {
                    if ( port == -1) port = 80;
//                    httpURL = new HttpURL(userName, password, _url.getHost(), port, "/"+stripLeadingSlashes(path), null);
                    httpURL = new HttpURL( "http://"+
                                           ((userName == null) ? "" : URIUtil.escape(userName) +
                                           ((password == null) ? "" : ":" + URIUtil.escape(password)) + "@") +
                                           URIUtil.escape(_url.getHost(), URIUtil.hostReserved()) +
                                           ((port == HttpURL.port) ? "" : ":"+port) +
                                           "/"+Utils.stripLeadingSlashes(path));
                } else if( "https".equalsIgnoreCase( _url.getProtocol() ) ) {
                    if ( port == -1) port = 443;
//                    httpURL = new HttpsURL(userName, password, _url.getHost(), port, "/"+stripLeadingSlashes(path), null);
                    httpURL = new HttpsURL( "https://"+
                                            ((userName == null) ? "" : URIUtil.escape(userName) +
                                            ((password == null) ? "" : ":" + URIUtil.escape(password)) + "@") +
                                            URIUtil.escape(_url.getHost(), URIUtil.hostReserved()) +
                                            ((port == HttpsURL.port) ? "" : ":"+port) +
                                            "/"+Utils.stripLeadingSlashes(path));
                }
                webdavResource = new WebdavResource(httpURL, WebdavResource.NOACTION, DepthSupport.DEPTH_0 );
				if ( webdavResource != null ) {
					if ( _logger != null && _logger.isDebugEnabled(LOGGING_CONTEXT) ) {
						// Using interceptors one is able to trace the input and output streams
						webdavResource.retrieveSessionInstance().setStreamInterceptor(new DebugStreamInterceptor( _logger ));
						webdavResource.retrieveSessionInstance().setConnectionInterceptor(new DebugConnectionInterceptor( _logger ));
						webdavResource.setDebug( DAVConstants.DEFAULT_DEBUG_TRACE );
					} else if ( DEBUG_OPT != null && DEBUG_OPT.equalsIgnoreCase("true") ) {
						// Using interceptors one is able to trace the input and output streams
						webdavResource.retrieveSessionInstance().setStreamInterceptor(new DebugStreamInterceptor());
						webdavResource.retrieveSessionInstance().setConnectionInterceptor(new DebugConnectionInterceptor());
						webdavResource.setDebug( DAVConstants.DEFAULT_DEBUG_TRACE );
					}
				}
                if ( proxyHost != null && proxyPort >= 0 ) {
                    webdavResource.setProxy( proxyHost, proxyPort );
                    webdavResource.getSessionInstance( httpURL, true );
                }

//                if ( session != null ) webdavResource.retrieveSessionInstance().getState().addCookie( session );

                // Slide bug, for 401 Unauthorized they throw an IllegalStateException - rethrow as the correct HttpException
                try {
                    webdavResource.setProperties( WebdavResource.BASIC, DepthSupport.DEPTH_0 );
                    doneInitialPropfind = true;
                } catch ( IllegalStateException e ) { throw new HttpException( HttpStatus.SC_UNAUTHORIZED ); }
                status = webdavResource.getStatusCode();
                if ( status == HttpStatus.SC_UNAUTHORIZED ) throw new HttpException( status );

            } catch (HttpException we) {
                status = we.getStatusCode();
                // Check is it an authorization problem. If so, try again to retrieve the realm for authentication.
                if (we.getStatusCode() == HttpStatus.SC_UNAUTHORIZED) {
                    authReqd = true;
                    HttpClient client = null;
                    String realm = null;
                    try {
                        client = new HttpClient();
                        client.startSession( httpURL.toURL() );
                        PropFindMethod pf = new PropFindMethod( httpURL.getPath() );
                        pf.setDepth(DepthSupport.DEPTH_0);
                        client.executeMethod( pf );
                    } catch (HttpException ae) {
                        State state = client.getState();
                        String authToken = state.getAuthenticateToken();
                        if (authToken != null) {
                            int ptr = authToken.indexOf( "realm=" );
                            int ptr2 = authToken.indexOf(",", ptr);
                            if ( ptr != -1 ) {
                                if (ptr2 != -1) realm = authToken.substring( ptr+6, ptr2 );
                                else realm = authToken.substring( ptr+6 );
                            }
                            if (_logger != null  && _logger.isDebugEnabled(LOGGING_CONTEXT)) {
                                _logger.debug( Message.format( RB.getResources(),
                                               "DAVResource.authreqd.fmt.txt",
                                               authToken), LOGGING_CONTEXT );
                            }
                        }
                    }
                    httpURL = null;
                    throw new DAVException( Message.format( RB.getResources(),
                                                 "DAVResource.authenticationocredentials.txt",
                                                 realm ), we);
                } else {
                    httpURL = null;
                    throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                            Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
                }
            }
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        if (_logger != null  && _logger.isDebugEnabled(LOGGING_CONTEXT)) {
             _logger.debug( Message.format( RB.getResources(), "DAVResource.constructorstatus.fmt.txt",
                            httpURL.getPath(), Integer.toString(webdavResource.getStatusCode())), LOGGING_CONTEXT);
        }

		long end = System.currentTimeMillis();
		logTime( start, end, "DAVResource" );
        if (httpURL != null && createCollection) {
            try {
                MkcolMethod method = new MkcolMethod(httpURL.getPath());
                method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
                method.setDebug(debug);
                long t0, t1;
                t0 = System.currentTimeMillis();
                int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                status = statusCode;
                t1 = System.currentTimeMillis();
                logTime( t0, t1, MKCOL );
                if (!SUCCEEDED(statusCode)) {
					// HttpStatus.SC_METHOD_NOT_ALLOWED is returned when the collection already exists. I guess this
					// is one of two expected outcomes and as such will only be written to the DEBUG log, other error
					// conditions will be written to the WARNING log.
					if ( status == HttpStatus.SC_METHOD_NOT_ALLOWED ) {
                        logMethodMessage( MKCOL, status, LoggerInterface.DEBUG );
                    } else {
                        logMethodMessage( MKCOL, status, LoggerInterface.WARN );
                        processError(status, MKCOL, httpURL.getPath(), extractReason(method));
                    }
                }
            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                        Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }

        }
	}

    /* --------------------------------- DAV/HTTP METHODS -----------------------------*/

    /**
     * Make a collection. This method only works on a null resource (one that doesn't exist yet). One would normally
     * call this method immediately after creating the DAVResource.
     */
    public boolean makeCollection() throws DAVException, RemoteException {
        boolean success;

        try {
            // Remember, webdavResource allows code 405 (SC_METHOD_NOT_ALLOWED) as successful. This is usually because
            // the collection already exists, so in a way the makeCollection is a null operation (and successful)
            MkcolMethod method = new MkcolMethod(httpURL.getPath());
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setDebug(debug);

            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, MKCOL );
            if (!SUCCEEDED(statusCode)) {
            	// HttpStatus.SC_METHOD_NOT_ALLOWED is returned when the collection already exists. I guess this
            	// is one of two expected outcomes and as such will only be written to the DEBUG log, other error
            	// conditions will be written to the WARNING log.
            	if ( status == HttpStatus.SC_METHOD_NOT_ALLOWED ) {
                    logMethodMessage( MKCOL, status, LoggerInterface.DEBUG );
                } else {
                    logMethodMessage( MKCOL, status, LoggerInterface.WARN );
                    processError(status, MKCOL, httpURL.getPath(), extractReason(method));
                }
                success = false;
            } else success = true;
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return success;
    }

    /**
     * Copy resource to the target path on the same host as the source
     * @param targetPath The path to which current resource is to be copied.
     * @return If successful, then a new DAVResource representing the new destination is created and returned
     */
    public DAVResource copy(java.lang.String targetPath) throws DAVException, RemoteException {
        DAVResource retRes = null;

        if (httpURL != null) {
            try {
                CopyMethod method = new CopyMethod(httpURL.getPath(), targetPath, overwrite);

                method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
                method.setDebug(debug);
                long t0, t1;
                t0 = System.currentTimeMillis();
                int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                status = statusCode;
                t1 = System.currentTimeMillis();
                logTime( t0, t1, COPY );

                // processMultiStatus(method);
                if ( !SUCCEEDED(statusCode) ) {
                    logMethodMessage( COPY, status, LoggerInterface.WARN );
                    processError(status, COPY, httpURL.getPath(), extractReason(method));
                }
                else {
                    String resString = httpURL.getScheme() + "://" + httpURL.getHost();
                    if (httpURL.getPort() > 0) resString = resString + ":" + httpURL.getPort();
                    resString = resString + targetPath; /*I18nOK: COS*/
                    if (userName == null) retRes = new DAVResource(resString, _logger);
                    else retRes = new DAVResource( resString, false, userName, password, _logger );
                }

            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
        return retRes;
    }

    /**
     * Deletes the current resource
     *
     */
    public void delete() throws DAVException, RemoteException {
        if (httpURL != null) {
            try {
                long t0, t1;
                t0 = System.currentTimeMillis();
                boolean success = webdavResource.deleteMethod( httpURL.getEscapedPathQuery() );
                status = webdavResource.getStatusCode();
                t1 = System.currentTimeMillis();
                logTime( t0, t1, DELETE );
                if ( !success ) {
                    logMethodMessage( DELETE, status, LoggerInterface.WARN );
                    processError(status, DELETE, httpURL.getPath(), "");
                }
            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
    }

    public org.apache.webdav.lib.Lock getActiveLockFor(java.lang.String owner) throws DAVException {
        int ownerlock = -1;
        LockDiscoveryProperty ldp = webdavResource.getLockDiscovery();
        status = webdavResource.getStatusCode();

        if (ldp == null) return null;
        Lock[] locks = ldp.getActiveLocks();

        if ( locks != null ) {
        for (int i = 0; i < locks.length; i++) {
            if (_logger != null && _logger.isDebugEnabled(LOGGING_CONTEXT)) {
                _logger.debug( Message.format (RB.getResources(), "DAVResource.lockreport.fmt.txt",
                               new Integer(i+1), locks[i].toString()), LOGGING_CONTEXT );
            }
            if (owner.equals(locks[i].getOwner())) ownerlock = i;
        }
        }
        if (ownerlock >= 0) return locks[ownerlock];
        else return null;
    }

    /**
     * Returns the contents of the resource as a byte array. This method assumes there is a limit on the length of the
     * content and therefore uses memory to buffer the content. If the content is large or memory buffering is not
     * desirable, then use the other getContentxxx methods for streaming content.
     *
     * @return byte[] The contents of the resource returned as a byte array
     * @throws DAVException in the event of WebDAV errors
     * @throws RemoteException in the case of network errors
     */
    public byte[] getContents() throws DAVException, RemoteException {
        if (httpURL != null) {
            try {
                // Use memory for small downloads
                GetMethod method = new GetMethod(httpURL.getPath() );
                method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
                method.setUseDisk( false );
                method.setDebug( debug );
                long t0, t1;
                t0 = System.currentTimeMillis();
                int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                status = statusCode;
                t1 = System.currentTimeMillis();
                logTime( t0, t1, GET );
                if ( !SUCCEEDED(status) ) {
                    if ( status == HttpStatus.SC_NOT_FOUND ) logMethodMessage( GET, status, LoggerInterface.DEBUG );
                    else {
                        logMethodMessage( GET, status, LoggerInterface.WARN );
                        processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                    }
                    return new byte[0];
                } else {
	                byte[] buffer = new byte[4096];
	                ByteArrayOutputStream os = new ByteArrayOutputStream();
	                int nb = 0;
	                InputStream contents = method.getData();
	                while ((nb = contents.read(buffer)) != -1) {
	                    os.write(buffer, 0, nb);
	                }
	                contents.close();
	                return os.toByteArray();
                }
            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                e.printStackTrace();
                throw new DAVException(e.getMessage(), e);
            }
        }
        return null;
    }

    /**
     * Return the contents of the resource as a String
     * @return String The contents of the resource
     * @throws DAVException in the event of a WebDAV failure.
     * @throws RemoteException in the event of network failure.
     */
    public String getContentsAsString() throws DAVException, RemoteException {
        //byte[] contents = null;
        if (httpURL != null) {
            try {
                // Use memory for small downloads
                GetMethod method = new GetMethod(httpURL.getPath() );
                method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
                method.setUseDisk( false );

                long t0, t1;
                t0 = System.currentTimeMillis();
                int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                status = statusCode;
                t1 = System.currentTimeMillis();
                logTime( t0, t1, GET );
                if ( !SUCCEEDED(status) ) {
                    if ( status == HttpStatus.SC_NOT_FOUND ) logMethodMessage( GET, status, LoggerInterface.DEBUG );
                    else {
                        logMethodMessage( GET, status, LoggerInterface.WARN );
                        processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                    }
                    return "";
                }
                Header contentHeader = method.getHeader( CONTENTTYPE );
                String charset = null;
                if ( contentHeader != null ) {
                    HeaderElement values[] = HeaderElement.parse( contentHeader.getValue() );
                    // Should only be one header element
                    if (values.length == 1) {
                        NameValuePair[] nvpairs = values[0].getParameters();
                        if ( nvpairs != null ) for (int i=0; i<nvpairs.length; i++) {
                            if ( nvpairs[i].getName().equals("charset"/*I18nOK:EMS*/ ) ) {
                                charset = nvpairs[i].getValue();
                                break;
                            }
                        }
                    }
                    if ( charset == null ) charset = DAVConstants.DEFAULT_CONTENT_CHARSET;

                    InputStream content = method.getData();
                    byte[] buffer = new byte[4096];
                    ByteArrayOutputStream os = new ByteArrayOutputStream();
                    int nb = 0;

                    while ((nb = content.read(buffer)) != -1) {
                        os.write(buffer, 0, nb);
                    }
                    byte[] _content = os.toByteArray();
                    try {
                        return new String( _content, charset );
                    } catch ( UnsupportedEncodingException e ) {
                        if ( _logger != null )
                        	_logger.warn( Message.format( RB.getResources(),
                                          "DAVResource.unsuppcharencoding.fmt.txt",
                                          charset ),LOGGING_CONTEXT );
                        try {
                            return new String( _content, DAVConstants.DEFAULT_CONTENT_CHARSET );
                        } catch ( UnsupportedEncodingException e2 ) {
                            if ( _logger != null )
	                            _logger.warn( Message.format( RB.getResources(),
                                          "DAVResource.unsuppcharencoding.fmt.txt",
                                          DAVConstants.DEFAULT_CONTENT_CHARSET ), LOGGING_CONTEXT );
                            return new String( _content );
                        }
                    }
                }
            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
        return null;
    }

    /**
     * Returns the contents of the resource into the supplied file
     * @param fileData The file to which the contents of the resource should be written
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network failure
     */
    public void getContents(File fileData) throws DAVException, RemoteException {

        if (httpURL != null) {
            try {
                GetMethod method = new GetMethod(httpURL.getPath(), fileData);

                method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
                method.setDebug(debug);
                long t0, t1;
                t0 = System.currentTimeMillis();
                int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                status = statusCode;
                t1 = System.currentTimeMillis();
                logTime( t0, t1, GET );
                if ( !SUCCEEDED(status) ) {
                    if ( status == HttpStatus.SC_NOT_FOUND ) logMethodMessage( GET, status, LoggerInterface.DEBUG );
                    else {
                        logMethodMessage( GET, status, LoggerInterface.WARN );
                        processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                    }
                    FileOutputStream fout = null;
                    try {
                        fout = new FileOutputStream( fileData );
                    fout.write( new byte[0] );
                    } catch (IOException ioe) {
                        // ignore
                    } finally {
                        if ( fout != null ) {
                            try {
                    fout.close();
                            } catch (IOException ioe) {
                                // ignore
                            }
                        }
                    }
                }

            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
        return;
    }

    /**
     * Returns the contents of the resource as an InputStream
     * @return java.io.InputStream
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network failure
     */
    public InputStream getContentsInputStream() throws DAVException, RemoteException {
        InputStream rspStream = null;
        if (httpURL != null) {
            try {
                // Use of this method would indicate larger files and so don't use
                // memory for this type of download. Create a temporary file and then
                // return a stream that will delete the file when the stream is closed.
				File fileData = File.createTempFile( "pfs", ".tmp", tempFileDir );
				getContents( fileData );
                rspStream = new ReadOnceFileInputStream( fileData );
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
        return new InputStreamAdapter( rspStream );
    }

    /**
     * Return the content type of the content for this resource
     * @return String The content type for this resource
     * @throws DAVException in the event of a WebDAV failure.
     * @throws RemoteException in the event of network failure.
     */
    public String getContentType() throws DAVException, RemoteException {
        String contentType = webdavResource.getGetContentType();
        return contentType;
    }

    /**
     * Returns all the properties on the DAV resource
     *
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getProperties() throws DAVException, RemoteException {
        ArrayList list = new ArrayList();

        try {
            // No access to the XML response ... use our own PropFindMethod to get more control
            //Enumeration e = isCollection() ? webdavResource.propfindMethod(DepthSupport.DEPTH_INFINITY) :
            //								 webdavResource.propfindMethod(DepthSupport.DEPTH_0) ;

            //PropFindMethod method = new PropFindMethod(httpURL.getPath(),
            //                                           isCollection() ? DepthSupport.DEPTH_INFINITY :
            //                                                            DepthSupport.DEPTH_0);
            // To control the depth of the propfind ... we must also check for illegal depth values here
            PropFindMethod method = new PropFindMethod(httpURL.getPath());

            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setDepth(depth);
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPFIND );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                if ( status != HttpStatus.SC_NOT_FOUND ) {
                    logMethodMessage( PROPFIND, status, LoggerInterface.WARN );
                    processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                } else {
                    logMethodMessage( PROPFIND, status, LoggerInterface.DEBUG );
                }
            }

            Enumeration e = method.getResponses();
            while (e.hasMoreElements()) {
                XMLResponseMethodBase.Response r = (XMLResponseMethodBase.Response) e.nextElement();
                Enumeration e2 = r.getProperties();
                while (e2.hasMoreElements()) {
                    list.add(e2.nextElement());
                }
            }

        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return (org.apache.webdav.lib.Property[]) list.toArray(new org.apache.webdav.lib.Property[0]);
    }

    /**
     * Returns the property specified in the parameter on the DAV resource
     *
     * @param propName property name elements - specifying which property value should be returned
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getProperty(org.w3c.dom.Element propName) throws DAVException, RemoteException {
        // String strPropName = DOM2Writer.nodeToString(propName);
        String strPropName = XMLUtil.convertNodeToString( propName, true );
        return getProperty( strPropName );
    }

    /**
     * Returns all the properties specified in the parameter on the DAV resource
     *
     * @param propNames Array of property name elements - specifying which property values should be returned
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getProperties(org.w3c.dom.Element[] propNames) throws DAVException, RemoteException {

        StringBuffer propString = new StringBuffer();
        for (int i = 0; i < propNames.length; i++) {
            // propString = propString + DOM2Writer.nodeToString(propNames[i]); /*I18nOK:COS*/
            propString.append(XMLUtil.convertNodeToString( propNames[i], true )); /*I18nOK:COS*/
        }
        return getProperties( propString.toString() );/*I18nOK:LSM*/
    }

    /**
     * Returns a property for a given property name. It is possible to return multiple property objects
     * as the resouce might represent a collection.
     *
     * @param propName name of property for which value should be returned
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getProperty(PropertyName propName) throws DAVException, RemoteException {
        String strPropName = "";
        if ( propName.getPrefix() == null ) {
            strPropName = '<' + propName.getLocalName() + " xmlns=\"\"/>";/*I18nOK:COS*/
        } else {
            strPropName = '<' + propName.getPrefix() + ':' + propName.getLocalName() + " xmlns:"/*I18nOK:EMS*/ + propName.getPrefix() + "=\"" + propName.getNamespaceURI() + "\"/>";/*I18nOK:COS*/
        }
        return getProperty(strPropName);
    }

    /**
     * Returns all the properties specified in the parameter on the DAV resource
     *
     * @param propNames Array of property names, specifying which property values should be returned
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getProperties(PropertyName[] propNames) throws DAVException, RemoteException {
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < propNames.length; i++) {
            if ( propNames[i].getPrefix() == null ) {
                buffer.append('<' + propNames[i].getLocalName() + " xmlns=\"\"/>"); /*I18nOK:EMS*/
            } else {
                buffer.append('<' + propNames[i].getPrefix() + ':' + propNames[i].getLocalName() + " xmlns:"/*I18nOK:EMS*/ + propNames[i].getPrefix() + "=\"" + propNames[i].getNamespaceURI() + "\"/>");
        }
    }
        return getProperties(buffer.toString());/*I18nOK:LSM*/
    }

    /**
     * Returns a property for a given property name. It is possible to return multiple property objects
     * as the resouce might represent a collection.
     *
     * @param propName name of property for which value should be returned
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getProperty( String propName ) throws DAVException, RemoteException {
        // No namespace support in propfindMethod. Also generateQuery will not support namespaces, so we set
        // the query ... propfindMethod checks to see if this is set before generating its own query (which would
        // not support namespaces
        String pfQuery1 = "<?xml version=\"1.0\" encoding=\"utf-8\" ?><D:propfind xmlns:D=\"DAV:\"><D:prop>";
        String pfQuery2 = "</D:prop></D:propfind>";
        ArrayList list = new ArrayList();

        try {
            PropFindMethod method = new PropFindMethod(httpURL.getPath(), depth, PropFindMethod.NAMES);

            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setQuery(pfQuery1 + propName + pfQuery2); /*I18nOK:COS*/
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPFIND );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                if ( status != HttpStatus.SC_NOT_FOUND ) {
                    logMethodMessage( PROPFIND, status, LoggerInterface.WARN );
                    processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                } else {
                    logMethodMessage( PROPFIND, status, LoggerInterface.DEBUG );
                }
            }

            Enumeration e = method.getAllResponseURLs();
            while (e.hasMoreElements()) {
                String rurl = (String) e.nextElement();
                Enumeration e2 = method.getResponseProperties(rurl);

                // Some servers seem to return more than one property, although only one has been asked for (eg. Intraspect).
                // Assuming that the first property returned is the correct one, ignore the others
                if (e2.hasMoreElements()) list.add(e2.nextElement());
                //while (e2.hasMoreElements())
                //{
                //	list.add((org.apache.webdav.lib.Property)e2.nextElement());
                //}
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return (org.apache.webdav.lib.Property[]) list.toArray(new org.apache.webdav.lib.Property[0]);
    }

    /**
     * Returns properties for a given set of property names. Multiple properties from multiple resources might be returned
     * as the resouce might represent a collection.
     *
     * @param propNames names of properties for which values should be returned
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getProperties(String propNames) throws DAVException, RemoteException {
        // No namespace support in propfindMethod. Also generateQuery will not support namespaces, so we set
        // the query ... propfindMethod checks to see if this is set before generating its own query (which would
        // not support namespaces
        String pfQuery1 = "<?xml version=\"1.0\" encoding=\"utf-8\" ?><D:propfind xmlns:D=\"DAV:\"><D:prop>";
        String pfQuery2 = "</D:prop></D:propfind>";
        ArrayList list = new ArrayList();

        try {
            PropFindMethod method = new PropFindMethod(httpURL.getPath(), depth, PropFindMethod.NAMES);

            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setQuery(pfQuery1 + propNames + pfQuery2); /*I18nOK:COS*/
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPFIND );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                if ( status != HttpStatus.SC_NOT_FOUND ) {
                    logMethodMessage( PROPFIND, status, LoggerInterface.WARN );
                    processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                } else {
                    logMethodMessage( PROPFIND, status, LoggerInterface.DEBUG );
                }
            }

            Enumeration e = method.getAllResponseURLs();

            while (e.hasMoreElements()) {
                String rurl = (String) e.nextElement();
                Enumeration e2 = method.getResponseProperties(rurl);

                while (e2.hasMoreElements()) {
                    list.add(e2.nextElement());
                }
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return (org.apache.webdav.lib.Property[]) list.toArray(new org.apache.webdav.lib.Property[0]);
    }

    /**
     * Returns a list of property names for the resource
     *
     * @return Array of DAV properties
     */
    public org.apache.webdav.lib.Property[] getPropertyNames() throws DAVException, RemoteException {
        // No support for <D:propname/> in WebdavResource ... call PropFindMethod directly
        ArrayList list = new ArrayList();

        try {
            PropFindMethod method = new PropFindMethod(httpURL.getPath(), depth, PropFindMethod.NAMES);

            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPFIND );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                if ( status != HttpStatus.SC_NOT_FOUND ) {
                    logMethodMessage( PROPFIND, status, LoggerInterface.WARN );
                    processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                } else {
                    logMethodMessage( PROPFIND, status, LoggerInterface.DEBUG );
                }
            }

            Enumeration e = method.getAllResponseURLs();
            String rurl;

            while (e.hasMoreElements()) {
                rurl = (String) e.nextElement();
                Enumeration e2 = method.getResponseProperties(rurl);

                while (e2.hasMoreElements()) {
                    list.add(e2.nextElement());
                }
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return (org.apache.webdav.lib.Property[]) list.toArray(new org.apache.webdav.lib.Property[0]);
    }

    /**
     * Returns the status code for the last operation
     *
     * @return status code for DAV Resource
     */
    public int getStatusCode() throws DAVException, RemoteException {
    	if ( webdavResource != null && status == -1 ) {
    		try {
    			webdavResource.setProperties( WebdavResource.BASIC, DepthSupport.DEPTH_0 );
	    		status = webdavResource.getStatusCode();
				doneInitialPropfind = true;
    		} catch ( HttpException ioe ) {
    			throw new DAVException(ioe.getMessage(), ioe);
    		} catch ( IOException ioe ) {
				throw new DAVException(ioe.getMessage(), ioe);
			}
    	}

        return status;
        //if (webdavResource != null) return webdavResource.getStatusCode();
        //else if (status != 0) return status;
        //else return (-1);
    }

    /**
     * Returns the status message for the last operation. Text for statusCode.
     *
     * @return status message for DAV Resource
     */
    public String getStatusMessage() {
        String msg = WebdavStatus.getStatusText( status );
        if ( msg != null ) return msg;
        else return RB.getStringResource( "DAVResource.cantgetstatusmessage.txt");

    }

    /**
     * Returns the URL that this DAV Resource represents
     *
     * @return url for resource
     */
    public java.net.URL getURL() throws DAVException, RemoteException {
        URL url;

        try {
            url = httpURL.toURL();
        } catch (MalformedURLException e) {
            url = null;
        }
        return url;
    }

    /**
     * Returns whether this resource is a collection or not
     *
     * @return isCollection ?
     */
    public boolean isCollection() {
        if ( webdavResource != null ) {
        	if ( !doneInitialPropfind ) try {
        		webdavResource.setProperties(  WebdavResource.BASIC, DepthSupport.DEPTH_0 );
        		status = webdavResource.getStatusCode();
				doneInitialPropfind = true;
        	} catch ( IOException ioe ) {
        		return false;
        	} catch ( HttpException he ) {
        		return false;
        	}
        	return webdavResource.isCollection();
        }
        else return false;
    }

    /**
     * Returns whether this resource is locked or not
     *
     * @return isLocked ?
     */
    public boolean isLocked() throws DAVException, RemoteException {
        if ( webdavResource != null ) return webdavResource.isLocked();
        else return false;
    }

    /**
     * Locks the resource
     * @return String The lock token for the successfully applied lock
     */
    public String lock() throws DAVException, RemoteException {
        String lockToken = null;
        try {
            String owner = (httpURL.getUserName() != null) ? httpURL.getUserName() : "sas/webdav"/*I18nOK:EMS*/;
            short lockType = LockMethod.SCOPE_EXCLUSIVE;
            LockMethod method = new LockMethod(httpURL.getPath(), owner, lockType, 120);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            int depth = getDepth();
            if (depth != DepthSupport.DEPTH_0 && depth != DepthSupport.DEPTH_INFINITY) {
                depth = DepthSupport.DEPTH_INFINITY;
            }
            method.setDepth(depth);
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
			int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, LOCK );
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( LOCK, status, LoggerInterface.WARN );
                processError(status, LOCK, httpURL.getPath(), extractReason(method));
            } else {
            	WebdavState state = (WebdavState) webdavResource.retrieveSessionInstance().getState();
            	lockToken = state.getLock( httpURL.getPath() );
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return lockToken;

    }

    /**
     * Move this resource to another location (path) on the same server as the source.
     * @param targetPath The target path where the resource will be moved
     * @return If successful, then a new DAVResource representing the new destination is created and returned
     */
    public DAVResource move(java.lang.String targetPath) throws DAVException, RemoteException {
        DAVResource retRes = null;

        if (httpURL != null) {
            try {
				MoveMethod method = new MoveMethod(httpURL.getPath(), targetPath, overwrite);
				method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
				method.setDebug(debug);
                long t0, t1;
                t0 = System.currentTimeMillis();
				int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                status = statusCode;
                t1 = System.currentTimeMillis();
                logTime( t0, t1, MOVE );
                if ( !SUCCEEDED(status) ) {
                    logMethodMessage( MOVE, status, LoggerInterface.WARN );
                    processError(status, MOVE, httpURL.getPath(), extractReason(method));
                } else {
                    String resString = httpURL.getScheme() + "://" + httpURL.getHost();
                    if (httpURL.getPort() > 0) resString = resString + ":" + httpURL.getPort();
                    resString = resString + targetPath; /*I18nOK:COS*/
                    if (userName == null) retRes = new DAVResource( resString, _logger);
                    else retRes = new DAVResource( resString, false, userName, password, _logger );
                }
            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
        return retRes;
    }

    /**
     * Remove/delete properties from a resource
     *
     * @param props names of the properties to remove
     */
    public void removeProperties(com.sas.services.webdav.PropertyName[] props) throws DAVException, RemoteException {

        /**** Don't seem to be able to do this with webdavResource ****/

        /**** ... could set property to blank *****/
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            for (int i = 0; i < props.length; i++) {
                // Generate a false (unique) prefix so that addPropertytoSet doesn't get them mixed up - uses prefix
                // and name as a key
                method.addPropertyToRemove(props[i].getLocalName(), "ns"+i, props[i].getNamespaceURI());
            }
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // Possbile Status Codes => SC_OK
            // WebdavStatus.SC_FORBIDDEN, SC_CONFLICT, SC_LOCKED, 507
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Remove/delete properties from a resource
     *
     * @param props names of the properties to remove
     */
    public void removeProperties(org.w3c.dom.Element[] props) throws DAVException, RemoteException {

        /**** Don't seem to be able to do this with webdavResource ****/

        /**** ... could set property to blank *****/
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            for (int i = 0; i < props.length; i++) {
                method.addPropertyToRemove(props[i].getLocalName(), props[i].getPrefix(), props[i].getNamespaceURI());
            }
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // Possbile Status Codes => SC_OK
            // WebdavStatus.SC_FORBIDDEN, SC_CONFLICT, SC_LOCKED, 507
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Remove/delete properties from a resource
     *
     * @param prop property to remove
     */
    public void removeProperty(org.w3c.dom.Element prop) throws DAVException, RemoteException {

        /**** Don't seem to be able to do this with webdavResource ****/

        /**** ... could set property to blank *****/
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.addPropertyToRemove(prop.getLocalName(), prop.getPrefix(), prop.getNamespaceURI());
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // Possbile Status Codes => SC_OK
            // WebdavStatus.SC_FORBIDDEN, SC_CONFLICT, SC_LOCKED, 507
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Remove/delete properties from a resource
     *
     * @param prop property to remove
     */
    public void removeProperty(String prop) throws DAVException, RemoteException {

        /**** Don't seem to be able to do this with webdavResource ****/

        /**** ... could set property to blank *****/
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.addPropertyToRemove(prop, "", "");
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // Possbile Status Codes => SC_OK
            // WebdavStatus.SC_FORBIDDEN, SC_CONFLICT, SC_LOCKED, 507
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Remove/delete a property from a resource
     *
     * @param prop name of the property to remove
     */
    public void removeProperty(com.sas.services.webdav.PropertyName prop) throws DAVException, RemoteException {

        /**** Don't seem to be able to do this with webdavResource ****/

        /**** ... could set property to blank *****/
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.addPropertyToRemove(prop.getLocalName(), prop.getPrefix(), prop.getNamespaceURI());
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // Possbile Status Codes => SC_OK
            // WebdavStatus.SC_FORBIDDEN, SC_CONFLICT, SC_LOCKED, 507
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }

    }

    /**
     * The SEARCH method invokes one of the implemented search grammars on the server to evaluate the query.
     * <p>
     * @param searchString A string in the correct format for the grammar selected. This will be passed to the
     * server in the body of the of the SEARCH method.
     * @return Property[] An array of properties that satisfy the query
     */
     public Property[] search( String searchString ) throws DAVException, RemoteException {

        ArrayList list = new ArrayList();
        try {
            SearchMethod method = new SearchMethod( httpURL.getPath(), searchString );
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setDebug( debug );
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, SEARCH );
            if ( !SUCCEEDED(statusCode) ) { 
                logMethodMessage( SEARCH, status, LoggerInterface.WARN );
                processError(status, SEARCH, httpURL.getPath(), extractReason(method));
            }
            processMultiStatus(method);

            // Possbile Status Codes => SC_MULTI_STATUS
            // WebdavStatus.SC_BAD_REQUEST, WebdavStatus.SC_UNPROCESSABLE_ENTITY
            if ( status == HttpStatus.SC_MULTI_STATUS ) {
            	Enumeration e = method.getAllResponseURLs();
	            while (e.hasMoreElements()) {
	                String rurl = (String) e.nextElement();
	                Enumeration e2 = method.getResponseProperties(rurl);

	                while (e2.hasMoreElements()) {
	                    list.add(e2.nextElement());
	                }
	            }
            }

        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
     return (Property[])list.toArray( new Property[0] );
     }


    /**
     * The SEARCH method invokes one of the implemented search grammars on the server to evaluate the query.
     * <p>
     * @param searchQuery An XML representation of the query
     * @return Property[] An array of properties that satisfy the query
     */
     public Property[] search( Element searchQuery ) throws DAVException, RemoteException {

        // String searchString = DOM2Writer.nodeToString( searchQuery );
        String searchString = XMLUtil.convertNodeToString( searchQuery );
        return( search( searchString ) );
     }


    /**
     * Provide content for the resource. This will result in a PUT to the resource.
     *
     * @param contents byte array for the resource contents
     */
    public void setContents(byte[] contents) throws DAVException, RemoteException {
        if (httpURL != null) {
            try {
                long t0, t1;
                t0 = System.currentTimeMillis();
                webdavResource.putMethod( httpURL.getEscapedPathQuery(), contents );
                status = webdavResource.getStatusCode();
                t1 = System.currentTimeMillis();
                logTime( t0, t1, PUT );

                // For older servers, try again without expectation and ensure there is a
                // Content-Length header preventing chunked encoding as this probably isn't
                // supported correctly either.
                if ( status == HttpStatus.SC_EXPECTATION_FAILED ) {
                    logMethodMessage( PUT, status, LoggerInterface.DEBUG );
                    PutMethod method = getPutMethodWithoutExpectation( httpURL.getPathQuery() );
                    String contentType = getContentType();
                    if( contentType != null && !contentType.equals(""))
                        method.setHeader( CONTENTTYPE, contentType );
                    method.sendData( contents );
                    method.setHeader( CONTENTLENGTH, String.valueOf(contents.length) );
                    t0 = System.currentTimeMillis();
                    int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                    status = statusCode;
	                t1 = System.currentTimeMillis();
	                logTime( t0, t1, PUT );

                }

                if ( !SUCCEEDED(status) ) {
                    logMethodMessage( PUT, status, LoggerInterface.WARN );
                    processError(status, PUT, httpURL.getPath(), "");
                }

            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
    }

    /** Utility method to print out the method status message
     *
     * @param method The method name (eg. PUT, PROPPATCH)
     * @param status The status value
     * @param level The level of logging required (eg. DEBUG, WARN )
     */
    private void logMethodMessage( String method, int status, int level ) {
        if ( _logger == null ) return;
        switch ( level ) {
            case LoggerInterface.DEBUG:
	            if ( !_logger.isDebugEnabled(LOGGING_CONTEXT) ) return;
	            _logger.debug( Message.format (RB.getResources(),
	                           "DAVResource.methoderror.fmt.txt", method,
	                           Integer.toString( status),
	                           WebdavStatus.getStatusText( status ),
							   httpURL.getPath()),
	                           LOGGING_CONTEXT);
	            break;
	        case LoggerInterface.WARN:
		        _logger.warn( Message.format (RB.getResources(),
		                      "DAVResource.methodwarn.fmt.txt", method,
		                      Integer.toString( status),
		                      WebdavStatus.getStatusText( status ),
							  httpURL.getPath()),
		                      LOGGING_CONTEXT);
		        break;
		    default: break;
        }
    }

    /**
     * Create an instance of the PUT method, but override the needExpectation() method
     * setting it to false. This is used for servers that don't support the expectation
     * @param path The path for the PUT method
     * @return PutMethod The new PUT method
     */
    private PutMethod getPutMethodWithoutExpectation( String path ) {
	    PutMethod method = new PutMethod( httpURL.getPathQuery() ) {
	        public boolean needExpectation() { return false; }
	        };
	    return method;
    }

    /**
     * Set the content for the resource from an URL
     *
     * @param url The URL for the contents to be PUT to this resource
     */
    public void setContents( URL url ) throws DAVException, RemoteException {
        if (httpURL != null) {
            try {
                long t0, t1;
                t0 = System.currentTimeMillis();
	            webdavResource.putMethod( httpURL.getEscapedPathQuery(), url );
                status = webdavResource.getStatusCode();
                t1 = System.currentTimeMillis();
                logTime( t0, t1, PUT );

                // For older servers, try again without expectation and ensure there is a
                // Content-Length header preventing chunked encoding as this probably isn't
                // supported correctly either.
                if ( status == HttpStatus.SC_EXPECTATION_FAILED ) {
                    logMethodMessage( PUT, status, LoggerInterface.DEBUG );
                    PutMethod method = getPutMethodWithoutExpectation( httpURL.getPathQuery() );
                    String contentType = getContentType();
                    if( contentType != null && !contentType.equals(""))
                        method.setHeader( CONTENTTYPE, contentType );
                    InputStream is = url.openConnection().getInputStream();
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    byte[] buffer = new byte[4096];
                    while ( is.read(buffer) != -1 ) {
                        baos.write( buffer );
                    }
                    method.setHeader( CONTENTLENGTH, String.valueOf(baos.size()) );
                    method.sendData( baos.toByteArray() );
                    t0 = System.currentTimeMillis();
                    int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                    status = statusCode;
                    t1 = System.currentTimeMillis();
                    logTime( t0, t1, PUT );

                }

                if ( !SUCCEEDED(status) ) {
                    logMethodMessage( PUT, status, LoggerInterface.WARN );
                    processError(status, PUT, httpURL.getPath(), "");
                }

            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
    }

    /**
     * Set the contents of the resource, passing in a URL
     * @param content The string containing the content for this resource
     * @throws DAVException in the event of a WebDAV failure.
     * @throws RemoteException in the event of network failure.
     */
    public void setContents( String content ) throws DAVException, RemoteException {

        if (httpURL != null) {
            try {
                long t0, t1;
                String contentType = getContentType();
                int ptr=contentType.indexOf("charset=");
                String charset = null;
                byte[] contents = null;
                if ( ptr > 0 ) charset = contentType.substring( ptr+8 );
                if ( charset != null) contents = content.getBytes( charset );
                else contents = content.getBytes( DAVConstants.DEFAULT_CONTENT_CHARSET );

                t0 = System.currentTimeMillis();
                webdavResource.putMethod( httpURL.getEscapedPathQuery(), contents );
                status = webdavResource.getStatusCode();
                t1 = System.currentTimeMillis();
                logTime( t0, t1, PUT );

                // For older servers, try again without expectation and ensure there is a
                // Content-Length header preventing chunked encoding as this probably isn't
                // supported correctly either.
                if ( status == HttpStatus.SC_EXPECTATION_FAILED ) {
                    logMethodMessage( PUT, status, LoggerInterface.DEBUG );
                    PutMethod method = getPutMethodWithoutExpectation( httpURL.getPathQuery() );
                    if( contentType != null && !contentType.equals(""))
                        method.setHeader( CONTENTTYPE, contentType );
                    method.sendData( contents );
                    method.setHeader( CONTENTLENGTH, String.valueOf(contents.length));
                    t0 = System.currentTimeMillis();
                    int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                    status = statusCode;
                    t1 = System.currentTimeMillis();
                    logTime( t0, t1, PUT );
                }
                if ( !SUCCEEDED(status) ) {
                    logMethodMessage( PUT, status, LoggerInterface.WARN );
                    processError(status, PUT, httpURL.getPath(), "");
                }

            } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
    }

    /**
     * Set the content from an InputStream. As this information might be required more than
     * once (eg. authorization required or a network error), the contents of the stream will
     * be buffered in a temporary file, which will then be sent. For remote deployment
     * the InputStream should be wrapped in an {@link com.sas.io.InputStreamAdapter InputStreamAdapter}
     *
     * @param is The input stream that contains the content.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network failure
     */
    public void setContents( InputStream is ) throws DAVException, RemoteException {
        File fileData = null;
        if (httpURL != null) {
            try {
            	// Make the temporary filename
                fileData = File.createTempFile( "pfs", ".tmp", tempFileDir );
                fileData.deleteOnExit();

                // Transfer content from stream to file
                FileOutputStream fStream = new FileOutputStream( fileData );
		        int streamSize = 0;
		        byte[] buffer = new byte[4096];
		        int nb = 0;
		        while (true) {
		            nb = is.read(buffer);
		            if (nb == -1)
		                break;
		            fStream.write(buffer, 0, nb);
		            streamSize = streamSize + nb;
		        }
		        fStream.close();

            	// Put the content to the DAV resource
            	PutMethod method = new PutMethod( httpURL.getPathQuery() );
            	generateIfHeader(method);
            	String contentType = getContentType();
            	if( contentType != null && contentType.length() > 0 ) method.setHeader( CONTENTTYPE, contentType );
            	method.sendData( fileData );
                long t0, t1;
                t0 = System.currentTimeMillis();
	            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                status = statusCode;
                t1 = System.currentTimeMillis();
                logTime( t0, t1, PUT );

                // For older servers, try again without expectation and ensure there is a
                // Content-Length header preventing chunked encoding as this probably isn't
                // supported correctly either.
                if ( status == HttpStatus.SC_EXPECTATION_FAILED ) {
                    logMethodMessage( PUT, status, LoggerInterface.WARN );
                    method = getPutMethodWithoutExpectation( httpURL.getPathQuery() );
                    contentType = getContentType();
                    if( contentType != null && !contentType.equals(""))
                        method.setHeader( CONTENTTYPE, contentType );
                    method.setHeader( CONTENTLENGTH, String.valueOf(streamSize) );
                    method.sendData( fileData );
                    t0 = System.currentTimeMillis();
                    statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
                    status = statusCode;
                    t1 = System.currentTimeMillis();
                    logTime( t0, t1, PUT );
                }

                if ( !SUCCEEDED(status) ) {
                    logMethodMessage( PUT, status, LoggerInterface.WARN );
                    processError(status, PUT, httpURL.getPath(), "");
                }

            fileData.delete();
	        } catch (HttpException we) {
                throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
            } catch (IOException e) {
                throw new DAVException(e.getMessage(), e);
            }
        }
    }

    /**
     * Set the contents type for the content of the resource
     * @param contentType The content type for this resource
     * @throws DAVException in the event of a WebDAV failure.
     * @throws RemoteException in the event of network failure.
     */
    public void setContentType( String contentType ) throws DAVException, RemoteException {
        webdavResource.setContentType( contentType );
    }

    /**
     * Set properties on the resource
     *
     * @param elems an array of DOM2 elements that identify the properties to set on the resource
     */
    public void setProperties(org.w3c.dom.Element[] elems) throws DAVException, RemoteException {
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            for (int i = 0; i < elems.length; i++) {
                // String propAsString = DOM2Writer.nodeToString(elems[i]);
                String propAsString = XMLUtil.convertNodeToString( elems[i], true );

                // Property "value" is the serialised XML form minus the outer element
                int beginPtr = propAsString.indexOf('>');
                int endPtr = propAsString.lastIndexOf('<');

                if (beginPtr < endPtr && beginPtr > 0 && endPtr > 0) propAsString = propAsString.substring(beginPtr + 1, endPtr);

                    // Now add the property (using the components localname, prefix, namespace and value, although
                    // the property is the element. (We should check that the property name and the outer element
                    // of the property are the same ... this is assumed
                method.addPropertyToSet(elems[i].getLocalName(), propAsString, elems[i].getPrefix(), elems[i].getNamespaceURI());
            }
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // webdavResource.proppatchMethod(prop.getPrefix()+":"+prop.getLocalName(), elem);
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Set properties on the resource
     *
     * @param props property names
     * @param propValues property values (which will be escaped unless it is a CDATA section)
     */
    public void setProperties(com.sas.services.webdav.PropertyName[] props, String[] propValues) throws DAVException, RemoteException {
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            // Ensure that prefixes that are reused actually are for the same namespace. If not then
            // generate a new prefix for that namespace.
            String _ns, _prefix, _usedNS;
            Map _prefixMap = new HashMap();
            for (int i = 0; i < props.length; i++) {
                _ns = props[i].getNamespaceURI();
                _prefix = props[i].getPrefix();
                if ( _prefixMap.containsKey( _prefix ) ) {
                    _usedNS = (String)_prefixMap.get( _prefix );
                    if ( !_usedNS.equals( _ns ) ) {
                        _prefix = "nsZ"+i;
                        _prefixMap.put( _prefix, _ns );
                    }
                } else _prefixMap.put( _prefix, _ns );
                method.addPropertyToSet(props[i].getLocalName(), Utils.XMLESCAPE(propValues[i]), _prefix, props[i].getNamespaceURI());
            }
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
            int sc = processMultiStatus(method);
            if ( sc > -1 ) status = sc;
			if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // webdavResource.proppatchMethod(prop.getPrefix()+":"+prop.getLocalName(), elem);
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Set properties on the resource
     *
     * @param doc document fragment containing elements that represent the properties to be set on the resource
     */
    public void setProperties(org.w3c.dom.DocumentFragment doc) throws DAVException, RemoteException {
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            NodeList nl = doc.getChildNodes();

            for (int i = 0; i < nl.getLength(); i++) {
                Node n = nl.item(i);

                if (n instanceof org.w3c.dom.Element) {
                    Element e = (Element) n;
                    // String propAsString = DOM2Writer.nodeToString(e);
                    String propAsString = XMLUtil.convertNodeToString( e, true );

                    // Property "value" is the serialised XML form minus the outer element
                    int beginPtr = propAsString.indexOf('>');
                    int endPtr = propAsString.lastIndexOf('<');

                    if (beginPtr < endPtr && beginPtr > 0 && endPtr > 0) propAsString = propAsString.substring(beginPtr + 1, endPtr);
                    method.addPropertyToSet(e.getLocalName(), propAsString, e.getPrefix(), e.getNamespaceURI());
                } else if (_logger != null && _logger.isDebugEnabled( LOGGING_CONTEXT ))
                        _logger.debug( Message.format( RB.getResources(), "DAVResource.wrongnodetype.fmt.txt",
                                       Integer.toString(n.getNodeType())), LOGGING_CONTEXT);
            }
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) { 
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // webdavResource.proppatchMethod(prop.getPrefix()+":"+prop.getLocalName(), elem);
        } catch (HttpException we) {
            we.printStackTrace();
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            e.printStackTrace();
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Set a property on the resource
     *
     * @param prop name of the property
     * @param propValue value of the property (which will be escaped unless it is a CDATA section)
     */
    public void setProperty(com.sas.services.webdav.PropertyName prop, String propValue) throws DAVException, RemoteException {
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.addPropertyToSet(prop.getLocalName(), Utils.XMLESCAPE(propValue), prop.getPrefix(), prop.getNamespaceURI());
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
			if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                // For backward compatibility with EM, don't throw an exception for SC_METHOD_NOT_ALLOWED
                if (status != HttpStatus.SC_METHOD_NOT_ALLOWED && status != HttpStatus.SC_NOT_IMPLEMENTED) {
                    processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
                }
            }

            // webdavResource.proppatchMethod(prop.getPrefix()+":"+prop.getLocalName(), elem);
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Set a property on the resource
     *
     * @param elem an XML element that represents the property name and value
     */
    public void setProperty(org.w3c.dom.Element elem) throws DAVException, RemoteException {
        try {
            PropPatchMethod method = new PropPatchMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            // String propAsString = DOM2Writer.nodeToString(elem);
            String propAsString = XMLUtil.convertNodeToString( elem, true );

            // Property "value" is the serialised XML form minus the outer element
            int beginPtr = propAsString.indexOf('>');
            int endPtr = propAsString.lastIndexOf('<');

            if (beginPtr < endPtr && beginPtr > 0 && endPtr > 0) propAsString = propAsString.substring(beginPtr + 1, endPtr);
            else if ( beginPtr >= 0 ) propAsString = "";

                // Now add the property (using the components localname, prefix, namespace and value, although
                // the property is the element. (We should check that the property name and the outer element
                // of the property are the same ... this is assumed
            method.addPropertyToSet(elem.getLocalName(), propAsString, elem.getPrefix(), elem.getNamespaceURI());
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPPATCH );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( PROPPATCH, status, LoggerInterface.WARN );
                processError(status, PROPPATCH, httpURL.getPath(), extractReason(method));
            }

            // webdavResource.proppatchMethod(prop.getPrefix()+":"+prop.getLocalName(), elem);
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Unlock a resource
     *
     */
    public void unlock() throws DAVException, RemoteException {
        try {

            long t0, t1;
            t0 = System.currentTimeMillis();
            webdavResource.unlockMethod( httpURL.getEscapedPathQuery() );
            status = webdavResource.getStatusCode();
            t1 = System.currentTimeMillis();
            logTime( t0, t1, UNLOCK );
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( UNLOCK, status, LoggerInterface.WARN );
                processError(status, UNLOCK, httpURL.getPath(), "");
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Unlock a resource
     *
     */
    public void unlock( String lockToken ) throws DAVException, RemoteException {
        try {
            UnlockMethod method = new UnlockMethod( httpURL.getPath(), lockToken );
			method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
			method.setDebug(debug);

            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, UNLOCK );
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( UNLOCK, status, LoggerInterface.WARN );
                processError(status, LOCK, httpURL.getPath(), extractReason(method));
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Get value for overWrite
     *
     * @return overWrite
     */
    public boolean getOverwrite() {
        return overwrite;
    }

    /**
     * Set the value for overWrite
     *
     * @param overwrite
     */
    public void setOverwrite(boolean overwrite) {
        this.overwrite = overwrite;
    }

    /**
     * Set the value for userName. Used in authentication requests to the server
     *
     * @param userName user name
     */
    public void setUserName(String userName) {
        this.userName = userName;
    }

    /**
     * Set the value for password. Used in authentication requests to the server
     *
     * @param password password
     */
    public void setPassword(String password) {
        this.password = password;
    }

    /**
     * Set the value for debug.
     *
     * @param debug
     */
    public void setDebug(int debug) {
        this.debug = debug;
        webdavResource.setDebug(debug);
        //try {
        //	webdavResource.retrieveSessionInstance().setDebug(debug);
        //} catch (IOException ioex) { System.err.println("Error setting debug on HttpClient"); }
    }

    /**
     * Set the value for depth. This is the depth on which DAV operations will be based.
     * They can be one of three values:
     * <ul>
     *    <li>DepthSupport.DEPTH_0
     *    <li>DepthSupport.DEPTH_1
     *    <li>DepthSupport.DEPTH_INIFINITY
     * </ul>
     *
     * DepthSupport is in org.apache.webdav.lib.methods
     *
     * @param depth org.apache.webdav.lib.methods.DepthSupport
     */
    public void setDepth(int depth) throws DAVException, RemoteException {
        this.depth = depth;
        try {
            webdavResource.setProperties(depth, WebdavResource.NOACTION);
            // webdavResource.setProperties( depth );
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Get the value for depth. This is the depth on which DAV operations will be based.
     * They can be one of three values:
     * <ul>
     *    <li>DepthSupport.DEPTH_0
     *    <li>DepthSupport.DEPTH_1
     *    <li>DepthSupport.DEPTH_INIFINITY
     * </ul>
     *
     * DepthSupport is in org.apache.webdav.lib.methods
     *
     * @return org.apache.webdav.lib.methods.DepthSupport
     */
    public int getDepth() {
        return depth;
    }

    public java.util.Properties doOptions() throws DAVException, RemoteException {

        java.util.Properties headers = new java.util.Properties();
        String hdrText = null;

        try {
            OptionsMethod method = new OptionsMethod( httpURL.getPath() );
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setDebug( debug );
            long t0, t1;
            t0 = System.currentTimeMillis();
            webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = webdavResource.getStatusCode();
            t1 = System.currentTimeMillis();
            logTime( t0, t1, OPTIONS );
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( OPTIONS, status, LoggerInterface.WARN );
                processError(status, OPTIONS, httpURL.getPath(), extractReason(method));
            }

            Enumeration hdrs = method.getHeaders();
            while (hdrs.hasMoreElements()) {
                hdrText = hdrs.nextElement().toString();
                // Parse header text into name /value
                hdrText = hdrText.replace('\r', ' ');
                hdrText = hdrText.replace('\n', ' ');
                int indx = hdrText.indexOf(':');
                if (indx > 0) {
                    String hdr = hdrText.substring(0, indx).trim();
                    String val = hdrText.substring(indx + 1).trim();
                    headers.put( hdr, val );
                }
            }

            // Setup the allowed methods and DAV capabilities lists
            Enumeration e = method.getAllowedMethods();
            ArrayList arrList = new ArrayList();

            if (e != null) {
                while (e.hasMoreElements()) {
                    arrList.add(e.nextElement());
                }
                allowedMethods = (String[]) arrList.toArray(new String[0]);
            }
            e = method.getDavCapabilities();
            arrList.clear();
            if (e != null) {
                while (e.hasMoreElements()) {
                    arrList.add(e.nextElement());
                }
                davCapabilities = (String[]) arrList.toArray(new String[0]);
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return headers;
    }

    /**
     * Returns a list of methods allowed on this resource.
     * @return String[] - list of methods allowed on this resource
     */
    public String[] getAllowedMethods() {
        return allowedMethods;
    }

    /**
     * Returns a list of DAV capabilties. The DAV header provides information on what DAV functions
     * are supported by the server. Basic DAV support requires the DAV header and a value of "1", but
     * support for locking features would have a value "1,2" for the DAV header. Further capabilities
     * will be added in the future (eg. access-control). The DAV header is a comma delimited list of
     * capabilites, returned as items in a string array.
     * @return String[] - DAV capabilities supported by the server
     */
    public String[] getDavCapabilities() {
        return davCapabilities;
    }

    /**
     * Returns whether a password is required to connect to this server
     * @return boolean - true if authentication credentials are required for a connnection, false otherwise
     */
    protected boolean isAuthReqd() {
        return authReqd;
    }

    /**
     * Use the processed XML response body to generate a multistatus map, mapping each
     * response href to a propstat map (which maps individual properties to their status
     * values).
     * @param method The method that has just been executed
     */
    private int processMultiStatus( XMLResponseMethodBase method ) {
        // Check out the multistatus
        _multiStatus.clear();
        int accSC = -1;                  // "accumulated" status code
        if ( method.getStatusCode() == HttpStatus.SC_MULTI_STATUS ) {
            try {
                Enumeration e = method.getResponses();

                while (e.hasMoreElements()) {
                    ResponseEntity r = (ResponseEntity) e.nextElement();
                    Enumeration e2 = r.getProperties();
                    Map propStat = new HashMap();

                    while (e2.hasMoreElements()) {
                        Property p = (Property) e2.nextElement();

                        String _name = p.getLocalName();
                        if ( p.getNamespaceURI() != null ) {
                            _name = "{"+p.getNamespaceURI()+"}"+_name;
                        }
						int _sc = p.getStatusCode();
                        propStat.put( _name, Integer.toString( _sc ) );
                        // Need to document errors that are not just NOT_FOUND
                        if ( !SUCCEEDED(_sc) && _sc!=HttpStatus.SC_NOT_FOUND ) {
                        	if ( _logger != null && _logger.isDebugEnabled(LOGGING_CONTEXT) ) {
                        		String _msg = method.getName() + " - " + _name; /*I18nOK:LINE*/
								logMethodMessage( _msg, _sc, LoggerInterface.DEBUG );
                        	}
                        }
                        if ( accSC == -1 ) accSC = _sc;
                        else if ( accSC != _sc ) accSC = 0;

                    }
                    _multiStatus.put(r.getHref(), propStat);
                }
            } catch ( Exception e ) {
                if ( _logger != null && _logger.isDebugEnabled(LOGGING_CONTEXT) ) {
                    _logger.debug( RB.getStringResource("DAVResource.multistatus.err.txt"), LOGGING_CONTEXT );
                }
            }
			if ( accSC > 200 ) return accSC;
            else return -1;
        }
        return -1;
    }

    /**
     * Returns whether the last method was successful or not. The HTTP status codes in the range
     * 200-299 are successful. All others indicate some sort of unsuccessful attempt.
     * @return boolean - true if last method was successful, false otherwise
     */
    public boolean SUCCEEDED() {
        try {
            if (getStatusCode() < 200 || getStatusCode() >= 300) return false;
            else return true;
        } catch (DAVException de) {
        	// will return false
        } catch (RemoteException re) {
        	// will return false
        }
        return false;
    }

    /**
     * Returns whether the statusCode parameter represents success or not. The HTTP status codes in
     * the range 200-299 are successful. All others indicate some sort of unsuccessful attempt.
     * @param statusCode A status code to be checked for success or failure
     * @return boolean - true if <code>statusCode</code> is successful, false otherwise
     */
    protected static boolean SUCCEEDED(int statusCode) {
        if (statusCode < 200 || statusCode >= 300) return false;
        else return true;
    }

    /**
     * Set the logger to be used by DAVResource. A default logger can be set on the constructor, but the logger
     * can be changed using this method.
     *
     * @param logger The logger.
     */
    public void setLogger(LoggerInterface logger) {
        _logger = logger;
    }

    /**
     * Returns the logger currently being used by DAVResource.
     *
     * @return <CODE>LoggerInterface</CODE> The logger that is currently being used
     */
    public LoggerInterface getLogger() {
        return _logger;
    }

    /**
     *  Return the multistatus information for the last operation.
     *  @return Map (href to propstat) Multistatus response from the last DAV method.
     */
    public Map getMultiStatus() {
        return _multiStatus;
    }

    /**
     *  Close the connection used by this DAVResource
     */
    public void close() throws RemoteException {
        try {
            webdavResource.close();
        } catch ( IOException e ) {
            // nothing else to do
        }
    }

    /**
     * Sets the temp directory used for storing files. The default is set to the value
     * of the Java system variable <code>java.io.tmpdir</code>, but if required can be
     * changed for a particular instance using this method.
     * @param tempDir The directory to be used for temporary files
     */
    public void setTempDir( String tempDir )  throws RemoteException {
   		tempFileDir = new File(tempDir);
    }

    /**
     * Gets the current value of the directory used for temporary files. The initial
     * value of this field is taken from the Java system variable
     * <code>java.io.tmpdir</code>, but may have been changed.
     * @return String The directory used for temporary files
     */
    public String getTempDir()  throws RemoteException {
    	if ( tempFileDir != null ) return tempFileDir.getAbsolutePath();
    	else return null;
    }

    /**
     * Returns whether the resource exists or not
     * @return boolean true if resource exists, false otherwise
     */
    public boolean exists() throws RemoteException {
        if ( webdavResource == null ) return false;
		if ( !doneInitialPropfind ) try {
			webdavResource.setProperties(  WebdavResource.BASIC, DepthSupport.DEPTH_0 );
			status = webdavResource.getStatusCode();
			doneInitialPropfind = true;
		} catch ( IOException ioe ) {
			return false;
		} catch ( HttpException he ) {
			return false;
		}

        // Slide bug - sets existence to true even if multistatus response is 404 (NOT FOUND)
        if ( webdavResource.getStatusCode() == HttpStatus.SC_NOT_FOUND ) return false;
        return webdavResource.exists();
    }

    /**
     * Run a report on the current resource. The report body is the body of the
     * request sent to the server, using the REPORT method.
     * @param body An XML document string that represents the report
     * request
     * @return Property[]. A list of the results from the report.
     * @throws DAVException in the case of server failures
     * @throws RemoteException in the case of network failures
     */
    public Property[] report( String body ) throws DAVException, RemoteException {

        ArrayList list = new ArrayList();
        try {
            ReportMethod method = new ReportMethod( httpURL.getPath() );

            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setQuery( body );
            method.setDepth(depth);
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, REPORT );
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( REPORT, status, LoggerInterface.WARN );
                processError(status, REPORT, httpURL.getPath(), extractReason(method));
            }

            processMultiStatus(method);

            Enumeration e = method.getAllResponseURLs();
            while (e.hasMoreElements()) {
                String rurl = (String) e.nextElement();
                Enumeration e2 = method.getResponseProperties(rurl);

                while (e2.hasMoreElements()) {
                    list.add(e2.nextElement());
                }
            }
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return (org.apache.webdav.lib.Property[]) list.toArray(new org.apache.webdav.lib.Property[0]);
    }

    /**
     * Get the Access Control List for this resource.
     * @return String An XML serialized document describing the ACL for this resource
     * @throws DAVException in the case of server failures
     * @throws RemoteException in the case of network failures
     */
    public String getACL() throws DAVException, RemoteException {

        try {
            PropFindMethod method = new PropFindMethod( httpURL.getPath() );

            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setQuery( "<?xml version=\"1.0\" encoding=\"UTF-8\"?><D:propfind xmlns:D=\"DAV:\"><D:prop><D:acl/></D:prop></D:propfind>" );
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, PROPFIND );
            if ( !SUCCEEDED(status) ) {
                if ( status != HttpStatus.SC_NOT_FOUND ) {
                    logMethodMessage( PROPFIND, status, LoggerInterface.WARN );
                    processError(status, PROPFIND, httpURL.getPath(), extractReason(method));
                } else {
                    logMethodMessage( PROPFIND, status, LoggerInterface.DEBUG );
                }
            }
            return XMLUtil.node2String( method.getResponseDocument() );
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Set/Update the Access Control List for this resource.
     * @param acl An XML serialized document describing the ACL for this resource
     * @return Document The result document from the ACL method
     * @throws DAVException in the case of server failures
     * @throws RemoteException in the case of network failures
     */
    public String setACL( String acl ) throws DAVException, RemoteException {

        try {
            AclMethod method = new AclMethod( httpURL.getPath() );

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setQuery( acl );
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, ACL );
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( ACL, status, LoggerInterface.WARN );
                processError(status, ACL, httpURL.getPath(), extractReason(method));
            }
            //TODO Check to see if ACL can return a document
            Document doc = method.getResponseDocument();
            if ( doc != null ) return XMLUtil.node2String( method.getResponseDocument() );
            else return null;
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Logs some performance figures. Time are taken between two points and then this
     * function is called.
     * @param t0 Start time
     * @param t1 End time
     * @param text Text to accompany performance message, usually DAV method name
     */
    private void logTime( long t0, long t1, String text ) {
        if ( _logger != null && _logger.isDebugEnabled(PERFORMANCE_CONTEXT) ) {
            _logger.debug( Thread.currentThread().getName()+
                              "," +  /*I18nOK:COS*/
                              text + /*I18nOK:COS*/
                              "," +  /*I18nOK:COS*/
                              NumberFormat.getInstance().format((t1-t0)/1000.), /*I18nOK:MTH*/
                           PERFORMANCE_CONTEXT );
        } else if ( DEBUG_OPT != null && DEBUG_OPT.equalsIgnoreCase("true") ) {
			System.err.println( Thread.currentThread().getName()+
							  "," +  /*I18nOK:COS*/
							  text + /*I18nOK:COS*/
							  "," +  /*I18nOK:COS*/
							  NumberFormat.getInstance().format((t1-t0)/1000.) ); /*I18nOK:MTH*/
        }
    }

    public void setSessionContext( SessionContextInterface sessionContext ) throws RemoteException {
        _sessionContext = sessionContext;
    }


    /**
     * Returns the session cookie if there is one. Any cookie that has the string "session" will
     * be taken as a session cookie.
     * @return Cookie The session cookie
     * @throws RemoteException in the event of a network failure
     */
    public Cookie getSessionCookie() throws RemoteException {
    	Cookie thecookie = null;
    	if ( webdavResource != null ) {
    		try {
	    		java.util.Iterator cookies = webdavResource.retrieveSessionInstance().getState().getCookies().iterator();
	    		while ( cookies.hasNext() ) {
	    			Cookie cookie = (Cookie)cookies.next();
	    			String name = cookie.getName().toLowerCase();  /*I18nOK:LSM*/
	    			if ( name.indexOf("session") > -1 ) {
	    				thecookie = cookie;
	    				break;
	    			}
	    		}
    		} catch ( IOException ioe ) {}
    	}
    	return thecookie;
    }

	public void setPath( String path ) {
		try {
			if (webdavResource != null ) {
//				HttpURL httpURL = webdavResource.getHttpURL();
//	   		    httpURL.setPath( path );
//				webdavResource.setHttpURL( httpURL, WebdavResource.NOACTION, DepthSupport.DEPTH_0 );
				HttpURL _httpURL = webdavResource.getHttpURL();
		        if (!path.startsWith("/")) path = "/" + path;
		        String _path = null;
		        try {
		            URI _uri = new URI(null, null, path, null);
		            _path = _uri.toASCIIString();
		        } catch (URISyntaxException use) {
		            _path = path;
		        }
		        String URI = _httpURL.getEscapedScheme() + "://" + _httpURL.getEscapedAuthority()
		                     + _path;
		        if ( _httpURL instanceof HttpsURL) httpURL = new HttpsURL(URI);
		        else httpURL = new HttpURL(URI);

				webdavResource.setHttpURL( httpURL, WebdavResource.NOACTION, DepthSupport.DEPTH_0 );
				doneInitialPropfind = false;
                webdavResource.setContentType("");
				status = -1;
			}
		} catch ( HttpException he ) {
			// error setting the path
		} catch ( IOException ioe ) {
			// io error setting the path
		}
	}

	/**
	 * @deprecated
	 */
	// Originally created to change paths without causing a PROPFIND. A better solution
	// was found and this is used in setPath(path) which is the replacement for this method
	public void setPathNoAction( String path ) {
		setPath(path);
    }

    public String getPath() {
    	if ( webdavResource != null ) return webdavResource.getPath();
    	else return null;
    }

    private class ReadOnceFileInputStream extends FileInputStream {
	   private File _fileData;
	   public ReadOnceFileInputStream( File fileData ) throws FileNotFoundException {
		   super( fileData );
		   _fileData = fileData;
	   }

	   public void close() throws IOException {
		   try {
		       super.close();
		   } finally {
		   _fileData.delete();
	   }
   }
   }

    private void generateIfHeader(HttpMethod method) {

        try {
            HttpClient client = webdavResource.retrieveSessionInstance();
	        if (client == null) return;
	        if (method == null) return;

	        WebdavState state = (WebdavState) client.getState();
	        String lockToken = state.getLock(method.getPath());
	        if (lockToken == null || lockToken.length() == 0) return;
//	        String[] lockTokens = state.getAllLocks(method.getPath());
//	        if (lockTokens.length == 0) return;

	        StringBuffer ifHeaderValue = new StringBuffer();

//	        for (int i = 0; i < lockTokens.length; i++) {
//	            ifHeaderValue.append("(<").append(lockTokens[i]).append(">) ");
//	        }
	        ifHeaderValue.append("(<").append(lockToken).append(">) ");

	        method.setHeader("If" /*I18nOK:EMS*/, ifHeaderValue.toString()); /*I18nOK:LSM*/
        } catch ( IOException ioe ) {
			// don't generate an If header
        }

    }

    /**
     * Puts the resource until version control.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public void doControl() throws DAVException, RemoteException {
        try {
            VersionControlMethod method = new VersionControlMethod(httpURL.getPath());

            generateIfHeader(method);
            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, VERSIONCONTROL );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
			if ( !SUCCEEDED(status) ) {
			    logMethodMessage( VERSIONCONTROL, status, LoggerInterface.WARN );
                processError(status, VERSIONCONTROL, httpURL.getPath(), extractReason(method));
			}
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

//ControllableResource
//    public boolean getVersionControllable() throws RemoteException {
//        return false;
//    }
    
//ControllableResource
//    public boolean getIsCheckedOut() throws RemoteException {
//        return false;
//    }
    
    /**
     * Gets a list of previous version names for this resource.
     * @return List(String) The version names for this resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public List getPreviousVersionNames() throws DAVException, RemoteException {
        Version[] versions = getVersions();
        List names = new ArrayList(versions.length);
        for (int i=0; i< versions.length; i++) {
            names.add(versions[i].getName());
        }
        return names;
    }

    /**
     * Checks out this resource.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public void doCheckout() throws DAVException, RemoteException {
	    try {
	        CheckoutMethod method = new CheckoutMethod(httpURL.getPath());
	
	        generateIfHeader(method);
	        method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
	        method.setDebug(debug);
	        long t0, t1;
	        t0 = System.currentTimeMillis();
	        int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
	        status = statusCode;
	        t1 = System.currentTimeMillis();
	        logTime( t0, t1, CHECKOUT );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
			if ( !SUCCEEDED(status) ) {
			    logMethodMessage( CHECKOUT, status, LoggerInterface.WARN );
                processError(status, CHECKOUT, httpURL.getPath(), extractReason(method));
			}
	    } catch (HttpException we) {
	        throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
	                                Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
	    } catch (IOException e) {
	        throw new DAVException(e.getMessage(), e);
	    }
    }
    
    /**
     * Modify the version label on the resource (if this is supported by the server).
     * @param labelAction Specify whether the label should be set (specify 1), added (specify 2), removed (specify 3) 
     * @param label The value of the label
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public void doLabel(int labelAction, String label) throws DAVException, RemoteException {
        try {
            LabelMethod method = new LabelMethod(httpURL.getPath(), labelAction, label);

            generateIfHeader(method);
            method.setHeader(USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING);
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime(t0, t1, LABEL);
            int sc = processMultiStatus(method);
            if (sc > -1) status = sc;
            if (!SUCCEEDED(status)) {
                logMethodMessage(LABEL, status, LoggerInterface.WARN);
                processError(status, LABEL, httpURL.getPath(), extractReason(method));
            }
        } catch (HttpException we) {
            throw new DAVException(Message.format(RB.getResources(), "DAVResource.httpexception.ex.fmt.txt", Integer.toString(we.getStatusCode()),
                    HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
    }

    /**
     * Check in this resource.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public void doCheckin() throws DAVException, RemoteException {
	    String location = null;
	    try {
	        CheckinMethod method = new CheckinMethod(httpURL.getPath());
	
	        generateIfHeader(method);
	        method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
	        method.setDebug(debug);
	        long t0, t1;
	        t0 = System.currentTimeMillis();
	        int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
	        status = statusCode;
	        t1 = System.currentTimeMillis();
	        logTime( t0, t1, CHECKIN );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
			if ( !SUCCEEDED(status) ) { 
			    logMethodMessage( CHECKIN, status, LoggerInterface.WARN );
                processError(status, CHECKIN, httpURL.getPath(), extractReason(method));
			}
			Header locationHeader = method.getHeader("Location");
			if ( locationHeader != null ) location = locationHeader.getValue();
	    } catch (HttpException we) {
	        throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
	                                Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
	    } catch (IOException e) {
	        throw new DAVException(e.getMessage(), e);
	    }
//	    return location;
    }

    /**
     * Uncheckout this resource, cancelling a previous checkout.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public void doUncheckout() throws DAVException, RemoteException {
	    try {
	        UncheckoutMethod method = new UncheckoutMethod(httpURL.getPath());
	
	        generateIfHeader(method);
	        method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
	        method.setDebug(debug);
	        long t0, t1;
	        t0 = System.currentTimeMillis();
	        int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
	        status = statusCode;
	        t1 = System.currentTimeMillis();
	        logTime( t0, t1, UNCHECKOUT );
			int sc = processMultiStatus(method);
			if ( sc > -1 ) status = sc;
			if ( !SUCCEEDED(status) ) {
			    logMethodMessage( UNCHECKOUT, status, LoggerInterface.WARN );
                processError(status, UNCHECKOUT, httpURL.getPath(), extractReason(method));
			}
	    } catch (HttpException we) {
	        throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
	                                Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
	    } catch (IOException e) {
	        throw new DAVException(e.getMessage(), e);
	    }
        
    }
    
    /**
     * Retrieves the currently checked in resource.
     * @return Version An object describing the version that is checked in
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public Version getCheckedIn() throws DAVException, RemoteException {
        return null;        
    }

    /**
     * Retrieves the currently checked out resource.
     * @return Version An object describing the version that is checked out
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public Version getCheckedOut() throws DAVException, RemoteException {
        return null;        
    }

    /**
     * Gets a particular version of this resource
     * @param versionName Which version to retrieve for this resource
     * @return Version An object describing the version of this resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    public Version getVersion(String versionName) throws DAVException, RemoteException
    {
        Version[] versions = getVersions();
        for (int i=0; i<versions.length; i++) {
            if ( versionName.equalsIgnoreCase(versions[i].getName()) ) {
                return versions[i];
            }
        }
        return null;
    }
    
    private Version[] getVersions() throws DAVException {
        // version-tree report
        String request = "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" +
                 "<D:version-tree xmlns:D=\"DAV:\">"+
                    "<D:prop>" +
                       "<D:resourcetype/>"+
                       "<D:version-history/>"+
                       "<D:version-name/>"+
                       "<D:comment/>"+
                       "<D:checked-in/>"+
                       "<D:checked-out/>"+
                       "<D:creator-displayname/>"+
                       "<D:predecessor-set/>" +
                       "<D:successor-set/>"+
                       "<D:checkout-set/>"+
                       "<D:creationdate/>"+
                       "<D:getlastmodified/>"+
                       "<D:checkout-set/>"+
                    "</D:prop>" +
                "</D:version-tree>";

        ArrayList versionArray = new ArrayList();
        try {
            ReportMethod method = new ReportMethod( httpURL.getPath(), DepthSupport.DEPTH_0 );

            method.setHeader( USER_AGENT_HDR, DAVConstants.SASCLIENT_STRING );
            method.setQuery( request );
            method.setDebug(debug);
            long t0, t1;
            t0 = System.currentTimeMillis();
            int statusCode = webdavResource.executeHttpRequestMethod(webdavResource.retrieveSessionInstance(), method);
            status = statusCode;
            t1 = System.currentTimeMillis();
            logTime( t0, t1, REPORT );
            if ( !SUCCEEDED(status) ) {
                logMethodMessage( REPORT, status, LoggerInterface.WARN );
                processError(status, REPORT, httpURL.getPath(), extractReason(method));
            }

            processMultiStatus(method);
            //TODO Need to be able to process DeltaV responses here
            Document doc = method.getResponseDocument();
            // Get all the versions and print summary
            String comment, href, versionName, creationDate, lastModified, creatorDisplayName;
            NodeList nl = doc.getElementsByTagNameNS("DAV:", "response");
            for (int i=1; i<=nl.getLength(); i++) {
                comment = "";
                Element version = (Element)nl.item(i-1);
                Node _n = version.getElementsByTagNameNS("DAV:", "href").item(0);
                _n.normalize();
                href = _n.getFirstChild().getNodeValue();
                _n = version.getElementsByTagNameNS("DAV:", "version-name").item(0);
                _n.normalize();
                versionName = _n.getFirstChild().getNodeValue();
                _n = version.getElementsByTagNameNS("DAV:", "creationdate").item(0);
                _n.normalize();
                creationDate = _n.getFirstChild().getNodeValue();
                _n = version.getElementsByTagNameNS("DAV:", "getlastmodified").item(0);
                _n.normalize();
                lastModified = _n.getFirstChild().getNodeValue();
                _n = version.getElementsByTagNameNS("DAV:", "creator-displayname").item(0);
                _n.normalize();
                creatorDisplayName = _n.getFirstChild().getNodeValue();
                _n = version.getElementsByTagNameNS("DAV:", "comment").item(0);
                _n.normalize();
                if ( _n.hasChildNodes() ) comment = _n.getFirstChild().getNodeValue();
                _n = version.getElementsByTagNameNS("DAV:", "predecessor-set").item(0);
                NodeList hrefs = _n.getChildNodes();
                String p = "";
                for (int j=0; j<hrefs.getLength(); j++) {
                    if ( j > 0) p += ';';
                    p += hrefs.item(j).getFirstChild().getNodeValue();
                }
                
                Version v = new Version();
                v.setHref(href);
                v.setComment(comment);
                v.setName(versionName);
                v.setCreationDate(creationDate);
                v.setLastModifiedDate(lastModified);
                v.setCreatorDisplayName(creatorDisplayName);
                versionArray.add(v);
            }
              
        } catch (HttpException we) {
            throw new DAVException( Message.format( RB.getResources(), "DAVResource.httpexception.ex.fmt.txt",
                                    Integer.toString(we.getStatusCode()), HttpStatus.getStatusText(we.getStatusCode())), we);
        } catch (IOException e) {
            throw new DAVException(e.getMessage(), e);
        }
        return (Version[])versionArray.toArray(new Version[0]);

    }
    public static final int DONT_CARE = 0;
    public static final int MUST_EXIST = 1;
    public static final int MUST_NOT_EXIST = 2;
    private int existsRequirement = DONT_CARE;
    public void setExistsRequirement(int requirement) {
        if ( requirement >=0 && requirement < 3) existsRequirement = requirement;
        else existsRequirement = DONT_CARE;
    }

    public int getExistsRequirement() {
        return existsRequirement;
    }
    
    /**
     * Process the error that occurred. Exceptions are thrown for specific important errors.
     * Specifically 400 (Bad request), 403 (Forbidden), 409 (Conflict), and all 5xx responses.
     * @param status Status code from the WebDAV method
     * @param method The name of the WebDAV method
     * @param path The resource path
     * @throws DAVException Thrown for specific error conditions
     */
    private void processError(int status, String method, String path, String reason) throws DAVException {
        // Error on 400, 403, 409 and all 5xx response codes
        if (status == HttpStatus.SC_FORBIDDEN || 
            status == HttpStatus.SC_CONFLICT || 
            status == HttpStatus.SC_BAD_REQUEST ||
            status/100 == 5) {
            String msg = status+" ("+HttpStatus.getStatusText(status)+") during "+method+" on "+path;
            if (reason != null && reason.length() > 0) msg = msg + " " +reason;
            throw new DAVException(msg, new HttpException(status));
        }
    }
    
    /**
     * Extract the reason for the failure from the response body returned. This is supposed 
     * to be an <D:error> element with a pre or post condition. However sometimes it is HTML
     * and sometimes just plain text. Also, some methods swallow up the response body and in 
     * this case, no extra information can be returned from this method.
     * @param method The method that has just been executed
     * @return String A string with additional context for the error that occurred
     */
    private String extractReason(HttpMethod method) {
        String result = "";
        int sc = method.getStatusCode();
        if (sc == HttpStatus.SC_CONFLICT || sc == HttpStatus.SC_FORBIDDEN) {
            if ( method instanceof XMLResponseMethodBase) {
                Document doc = ((XMLResponseMethodBase)method).getResponseDocument();
                if (doc != null) {
                    Element docElem = doc.getDocumentElement();
                    if (docElem != null) {
                        if ( "DAV:".equals(docElem.getNamespaceURI()) &&
                             "error".equals(docElem.getLocalName()) ) {
                            result = XMLUtil.node2String(docElem);
                        } else {
                            result = XMLUtil.node2String(doc);
                        }
                    }
                }
            }
        }
        return result;
    }
    
}
