/* $Id$ */
// Copyright (c) 2001-2004 SAS Institute Inc., Cary, NC 27513
// All Rights Reserved.

package com.sas.services.webdav;

import java.io.File;
import java.io.InputStream;
import java.net.URL;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.List;

import org.apache.webdav.lib.Property;
import org.apache.commons.httpclient.Cookie;

/**
 * Provides the interface for a DAV Resource. A DAVResource is a WebDAV (Distributed Authoring and Versioning) resource
 * as specified in IETF RFC 2518.
 */
public interface DAVResourceInterface extends Remote
{
    /**
     * Delete the resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void delete() throws DAVException, RemoteException;

    /**
     * Get an active lock, if it exists, for the owner
     * @param owner The owner for which the lock is being sought
     * @return Lock The lock for this owner if it exists
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public org.apache.webdav.lib.Lock getActiveLockFor(java.lang.String owner) throws DAVException, RemoteException;

    /**
     * Get the content of the resource into a file specified as the parameter
     * @param fileData The file into which the contents of the resource should be written
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void getContents(File fileData) throws DAVException, RemoteException;

    /**
     * Return the contents of the resource as a byte array
     * @return byte[] The contents of the resource as a byte array
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public byte[] getContents() throws DAVException, RemoteException;

    /**
     * Return the contents of the resource as a String
     * @return String The contents of the resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public String getContentsAsString() throws DAVException, RemoteException;

    /**
     * Return the content type of the content for this resource
     * @return String The content type for this resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public String getContentType() throws DAVException, RemoteException;

    /**
     * Return the content of the resource as a stream
     * @return InputStream The stream representing the content of the resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public InputStream getContentsInputStream() throws DAVException, RemoteException;

    /**
     * Get the properties for the resource. This will attempt to find all the properties.
     * @return Property[] An array of properties for this resource. 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] getProperties() throws DAVException, RemoteException;

    /**
     * Get the property values for the properties specified in the parameter
     * @return Property[] The property values requested
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] getProperties(org.w3c.dom.Element[] propNames) throws DAVException, RemoteException;

    /**
     * Get the property values for the properties specified in the parameter
     * @return Property[] The property values requested
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] getProperty(org.w3c.dom.Element propName) throws DAVException, RemoteException;

    /**
     * Get the property values for the properties specified in the parameter
     * @return Property[] The property values requested
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] getProperties(com.sas.services.webdav.PropertyName[] propNames) throws DAVException, RemoteException;

    /**
     * Get the property value for the property specified in the parameter
     * @return Property[] The property values requested
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] getProperty(java.lang.String propName) throws DAVException, RemoteException;

    /**
     * Get the property value for the property specified in the parameter
     * @return Property The property value requested
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] getProperty(com.sas.services.webdav.PropertyName propName) throws DAVException, RemoteException;

    /**
     * Get all the property names on the resource
     * @return Property[] A property array containing the names of all the properties on the resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] getPropertyNames() throws DAVException, RemoteException;

    /**
     * Get the status code for the last operation/method
     * @return int The status code for the last request on the resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public int getStatusCode() throws DAVException, RemoteException;

    /**
     * Get the message associated with the status code for the last operation/method
     * @return String The status message for the last request on the resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public String getStatusMessage() throws DAVException, RemoteException;

    /**
     * Get the URL that the resource represents
     * @return URL The java.net.URL for this resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public java.net.URL getURL() throws DAVException, RemoteException;

    /**
     * Is this resource a collection?
     * @return boolean True if it is a collection, false otherwise
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public boolean isCollection() throws DAVException, RemoteException;

    /**
     * Is this resource locked?
     * @return boolean True if it is locked, false otherwise
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public boolean isLocked() throws DAVException, RemoteException;

    /**
     * Lock this resource
     * @return String The lock token if the resource is successfully locked
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public String lock() throws DAVException, RemoteException;

    /**
     * Move this resource to the location given in the parameter
     * @param targetUrl The destination URL for this copy
     * @return DAVResource A DAVResource that represents the new resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public com.sas.services.webdav.DAVResource move(java.lang.String targetUrl) throws DAVException, RemoteException;

    /**
     * Remove the properties specified in the parameter
     * @param propNames An array of property names that should be removed 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void removeProperties(com.sas.services.webdav.PropertyName[] propNames) throws DAVException, RemoteException;

    /**
     * Remove the properties specified in the parameter
     * @param propNames An array of property names that should be removed 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void removeProperties(org.w3c.dom.Element[] propNames) throws DAVException, RemoteException;

    /**
     * Remove the property specified in the parameter
     * @param propName The name of a property that should be removed from this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void removeProperty(org.w3c.dom.Element propName) throws DAVException, RemoteException;

    /**
     * Remove the property specified in the parameter
     * @param propName The name of a property that should be removed from this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void removeProperty(java.lang.String propName) throws DAVException, RemoteException;

    /**
     * Remove the property specified in the parameter
     * @param propName The name of a property that should be removed from this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void removeProperty(com.sas.services.webdav.PropertyName propName) throws DAVException, RemoteException;

    /**
     * Set the contents of the resource, passing in a byte array
     * @param contents The byte array containing the content for this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setContents(byte[] contents) throws DAVException, RemoteException;

    /**
     * Set the contents of the resource, passing in a URL
     * @param url A url that represents the content for this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setContents( URL url ) throws DAVException, RemoteException;

    /**
     * Set the contents of the resource, passing in a URL
     * @param content The string containing the content for this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setContents( String content ) throws DAVException, RemoteException;

    /**
     * Set the contents of the resource from an input stream. For remote deployment 
     * the InputStream should be wrapped in an {@link com.sas.io.InputStreamAdapter InputStreamAdapter}
     * @param is An input stream representing the content for this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setContents( InputStream is ) throws DAVException, RemoteException;

    /**
     * Set the contents type for the content of the resource
     * @param contentType The content type for this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setContentType( String contentType ) throws DAVException, RemoteException;

    /**
     * Set the values of properties, as given in the parameter
     * @param props An array of properties to be set on this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setProperties(org.w3c.dom.Element[] props) throws DAVException, RemoteException;

    /**
     * Set the values of properties, as given in the parameter
     * @param propNames An array of property names to be set on this resource 
     * @param props An array of property values to be set on this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setProperties(com.sas.services.webdav.PropertyName[] propNames, java.lang.String[] props) throws DAVException, RemoteException;

    /**
     * Set the values of properties, as given in the parameter
     * @param props An XML document fragment that defines the properties to be set on this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setProperties(org.w3c.dom.DocumentFragment props) throws DAVException, RemoteException;

    /**
     * Set the value of a property, as given in the parameter
     * @param propName An property name to be set on this resource 
     * @param prop A property value to be set on this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setProperty(com.sas.services.webdav.PropertyName propName, java.lang.String prop) throws DAVException, RemoteException;

    /**
     * Set the value of a property, as given in the parameter
     * @param prop An XML element that represents a property to be set on this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setProperty(org.w3c.dom.Element prop) throws DAVException, RemoteException;

    /**
     * Unlock the resource, if previously locked by the same principal
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void unlock() throws DAVException, RemoteException;

    /**
     * Unlock the resource, with the correct locktoken
     * @param lockToken The lock token for this resource 
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void unlock(java.lang.String lockToken) throws DAVException, RemoteException;

    /**
     * Get the overwrite flag - can this resource be overwritten
     * @return boolean True if the overwrite flag is set, false otherwise
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public boolean getOverwrite() throws DAVException, RemoteException;

    /**
     * Set the overwrite flag - can this resource be overwritten
     * @param overwrite The value of the overwrite flag to set on the resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setOverwrite(boolean overwrite) throws DAVException, RemoteException;

    /**
     * Set the depth that will be used for the following operations
     * DEPTH_0, DEPTH_1 and DEPTH_INFINITY from DepthSupport are valid values
     * @param depth The depth for operations on this resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public void setDepth(int depth) throws DAVException, RemoteException;

    /**
     * Get the current depth value
     * @return depth The depth for operations on this resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public int getDepth() throws DAVException, RemoteException;

    /**
     * Using the search query provided as a string, perform the search on this server
     * @param searchString A string containing an XML DASL search request
     * @return Property[] A list of properties from the resources that satisfied the search criteria
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] search( String searchString ) throws DAVException, RemoteException;

    /**
     * Using the XML representation of the search query, perform the search on the server
     * @param searchQuery An XML document containing a DASL search request
     * @return Property[] A list of properties from the resources that satisfied the search criteria
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] search( org.w3c.dom.Element searchQuery ) throws DAVException, RemoteException;

    /**
     * Close the socket used for this connection. This can be useful for servers
     * that have a limit on the number of connections (eg. IIS), and provides
     * a way to control network resources.
     * @throws RemoteException in the event of network failure. 
     */
    public void close() throws RemoteException;

    /**
     * Provides a way to set a directory for temporary files. The
     * <code>java.io.tmp</code> system property can be used to set this value
     * at JVM invocation. If the system property is not set "/temp/" is the
     * default, but can be overridden with this method.
     * @param tempDir Temp directory for temporary files
     * @throws RemoteException in the event of network failure. 
     */
    public void setTempDir( String tempDir ) throws RemoteException;

    /**
     * Returns the current setting for the directory for temporary files
     * @return String The temp directory for temporary files
     * @throws RemoteException in the event of network failure. 
     */
    public String getTempDir() throws RemoteException;
    
    /**
     * Returns whether the resource exists or not
     * @return boolean true if resource exists, false otherwise
     * @throws RemoteException in the event of network failure. 
     */
    public boolean exists() throws RemoteException;
    
    /**
     * Run a report on the current resource. The report body is the body of the
     * request sent to the server, using the REPORT method.
     * @param reportBody An XML document string that represents the report
     * request
     * @return Property[]. A list of the results from the report.
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public Property[] report( String reportBody ) throws DAVException, RemoteException;

    /**
     * Get the Access Control List for this resource. 
     * @return String An XML serialized document describing the ACL for this resource
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public String getACL() throws DAVException, RemoteException;

    /**
     * Set/Update the Access Control List for this resource.
     * @param acl An XML serialized document describing the ACL for this resource
     * @return Document The result document from the ACL method
     * @throws DAVException in the event of a WebDAV failure. 
     * @throws RemoteException in the event of network failure. 
     */
    public String setACL( String acl ) throws DAVException, RemoteException;
    
    /**
     * Set the session context for this DAVResource. 
     * @param sessionContext The session context for this DAVResource
     * @throws RemoteException in the event of network failure. 
    */
    void setSessionContext( com.sas.services.session.SessionContextInterface sessionContext ) throws RemoteException;
    
    /**
     * Retrieves a session cookie from the connection, if one exists. This method is for internal use only.
     * @return Cookie The session cookie associated with this resource connection
     * @throws RemoteException in the event of a network failure
     */
    Cookie getSessionCookie() throws RemoteException;
    
    /**
     * Get the path for the WebDAV resource that is currently represented by this DAVResource
     * @return String The current path for this WebDAV resource
     * @throws RemoteException in the event of a network failure
     */
    String getPath() throws RemoteException;
	
	/**
	 * Set the path for this resource. If the resource was originally created as http://server/origpath, then calling
	 * this method will change the resource path to http://server/path. This allows the same DAVResource object to be
	 * used for any WebDAV resource.
	 * @param path The new path for the WebDAV resource
	 * @throws RemoteException in the event of a network failure
	 */ 
    void setPath( String path ) throws RemoteException;

	/**
	 * Set the path for this resource. If the resource was originally created as http://server/origpath, then calling
	 * this method will change the resource path to http://server/path. This allows the same DAVResource object to be
	 * used for any WebDAV resource.
	 * @param path The new path for the WebDAV resource
	 * @throws RemoteException in the event of a network failure
	 */ 
	void setPathNoAction( String path ) throws RemoteException;

    /**
     * Make a collection. This method only works on a null resource (one that doesn't exist yet). One would normally
     * call this method immediately after creating the DAVResource.
     * @return boolean True is make collection successful, false otherwise
     * @throws DAVException in the case of a WebDAV error
     */
	boolean makeCollection() throws DAVException, RemoteException;

    /**
     * Returns whether the last method was successful or not. The HTTP status codes in the range
     * 200-299 are successful. All others indicate some sort of unsuccessful attempt.
     * @return boolean - true if last method was successful, false otherwise
     */
	boolean SUCCEEDED() throws RemoteException;
    
    /**
     * Returns a properties for a given set of property names. 
     * @param propNames names of properties for which values should be returned
     * @return Array of DAV properties
     */
	Property[] getProperties(String propNames) throws DAVException, RemoteException;
    
    /**
     * Copy resource to the target destination
     *
     * @param targetUrl location to which current resource is to be copied.
     * @return If successful, then a new DAVResource representing the new destination is created and returned
     */
	DAVResource copy(String targetUrl) throws DAVException, RemoteException;

    /**
     * Puts the resource until version control.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
	void doControl() throws DAVException, RemoteException;

    /**
     * Modify the version label on the resource (if this is supported by the server).
     * @param labelAction Specify whether the label should be set (specify 1), added (specify 2), removed (specify 3) 
     * @param label The value of the label
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
	void doLabel(int labelAction, String label) throws DAVException, RemoteException;
//    boolean getVersionControllable() throws DAVException, RemoteException;

    /**
     * Gets a list of previous version names for this resource.
     * @return List(String) The version names for this resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    List getPreviousVersionNames() throws DAVException, RemoteException;

    /**
     * Checks out this resource.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    void doCheckout() throws DAVException, RemoteException;

    /**
     * Check in this resource.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    void doCheckin() throws DAVException, RemoteException;

    /**
     * Uncheckout this resource, cancelling a previous checkout.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    void doUncheckout() throws DAVException, RemoteException;

    /**
     * Gets a particular version of this resource
     * @param versionName Which version to retrieve for this resource
     * @return Version An object describing the version of this resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    Version getVersion(String versionName) throws DAVException, RemoteException;

    /**
     * Retrieves the currently checked in resource.
     * @return Version An object describing the version that is checked in
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    Version getCheckedIn() throws DAVException, RemoteException;

    /**
     * Retrieves the currently checked out resource.
     * @return Version An object describing the version that is checked out
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of network failures
     */
    Version getCheckedOut() throws DAVException, RemoteException;
    
//    boolean getIsCheckedOut() throws DAVException, RemoteException;
    void setExistsRequirement(int requirement) throws RemoteException;
    int getExistsRequirement() throws RemoteException;
}
