/* $Id$ */
// Copyright (c) 2002 SAS Institute Inc., Cary, NC 27513
package com.sas.services.webdav;

import org.w3c.dom.Element;

/**
 * This class represents a qualified property name. A name consists of three parts,
 * viz. the local name, the prefix and the namespace. This corresponds to the
 * information in an XML element that represents the property's name.
 * <p>
 * e.g. <ns0:PreferredColour xmlns:ns0="http://some.namespace"/> would result in a
 * local name of "PreferredColour", a prefix of "ns0" and a namespace of
 * "http://some.namespace". Although, in general, the prefix can be generated
 * automatically, it is assumed that this process has already occurred and the
 * prefix is valid part of the property name.
 * <p>
 * @author Rob Owen
 * @version 1.0 5-Feb-2002
 */
public class PropertyName
{
    private String localName;
    private String prefix;
    private String namespaceURI;

    public PropertyName()
    {}

    /**
     * Construct a PropertyName from an existing Property object. Extract the relevant
     * information for the name of the property from the object that represents the
     * property.
     * <p>
     * @param p A Property object
     */
    public PropertyName (org.apache.webdav.lib.Property p)
    {
        this.localName = p.getLocalName();
        this.namespaceURI = p.getNamespaceURI();
        String pname = p.getName();
        String prefix = pname.substring(0,pname.indexOf(':'));
        this.prefix = prefix;
    }

    /**
     * Construct a PropertyName from the three independent parts, a local name, a
     * prefix and a namespace.
     * <p>
     * @param localName A string representing the local name of the property
     * @param prefix A string representing the prefix for the property name
     * @param namespaceURI A string representing a valid namespace uri
     */
    public PropertyName(String localName, String prefix, String namespaceURI)
    {
        this.localName = localName;
        this.prefix = prefix;
        this.namespaceURI = namespaceURI;
        // leave parsing value to element until needed
    }

    /**
     * Construct a PropertyName from an XML element that represents the property.
     * <p>
     * @param propElement An XML property element
     */
    public PropertyName( Element propElement )
    {
        this.localName = propElement.getLocalName();
        this.namespaceURI = propElement.getNamespaceURI();
        this.prefix = propElement.getPrefix();
    }

    /**
     * Construct a PropertyName from a string representing the property name. The
     * string might have an embedded namespace - in this case the string has the
     * format "{namespace}localname" and a prefix is generated for the PropertyName.
     * <p>
     * @param name A string representation of the property name
     */
    public PropertyName( String name )
    {
        String pName, prefix, namespace;

        // Does the name have a namespace (separator is '{namespace}')
        name = name.trim();
        int sep = name.indexOf('}');
        if (name.charAt(0) == '{' && sep > 0) {
            namespace = name.substring(1, sep);
            pName     = name.substring(sep+1);
        } else {
            namespace = "";
            pName     = name;
        }

        if ( !namespace.equals("") ) prefix = "ns0"/*I18nOK:EMS*/;
        else prefix = "";
        this.localName    = pName;
        this.prefix       = prefix;
        this.namespaceURI = namespace;
    }


    /**
     * Get the local name of the property.
     * <p>
     * @return String The local name of the property
     */
    public String getLocalName()
    {
        return localName;
    }

    /**
     * Get the prefix defined for the property.
     * <p>
     * @return String The prefix defined for the property
     */
    public String getPrefix()
    {
        return prefix;
    }

    /**
     * Get the namepsace associated with the property name.
     * <p>
     * @return String The namepsace associated with the property
     */
    public String getNamespaceURI()
    {
        return namespaceURI;
    }

    /**
     * Convert the property name to an XML string representing the property name
     * element using the format <code>&lt;prefix:localname xmlns:prefix="namespaceuri"/&gt;.
     * <p>
     * @return String String XML representation of the property name element
     */
    public String toString()
    {
        if ( prefix != null && prefix.length() > 0) {
            return "<"+prefix+":"+localName+" xmlns:"+prefix+"=\""+namespaceURI+"\"/>"; /*I18nOK:LINE*/
        } else {
            return "<"+localName+"/>"; /*I18nOK:LINE*/
        }
    }
}