/* $Id: RemoteCollection.java,v 1.1.4.1 2005/03/29 19:24:27 eurrow Exp $ */


// Copyright (c) 2001-2004 SAS Institute Inc., Cary, NC 27513
// All Rights Reserved.

package com.sas.services.webdav;

import java.io.InputStream;
import java.rmi.RemoteException;
import java.util.List;
import java.util.Map;

import com.sas.services.logging.LoggerInterface;

/**
 * This class represents a remotable DAV collection and the operations expected of a collection. Collections can
 * contain other collections and/or resources (known as items).
 * <p>
 * @author  Rob Owen
 * @version 1.0 4-Mar-05
 * @since   1.0
 */
public class RemoteCollection extends RemoteResource implements CollectionInterface
{
    
    /**
     * Keep a handle to the real collection and delegate to it for all method calls
     */
    private Collection collection;

	/**
	 * Create a new collection in an existing collection. The model collection is used as the location in which to create
	 * the new collection. This constructor also reuses the model's connection for the new collection.
	 * @param model The collection in which to create the new collection and the connection to use for the new collection
	 * @param name The name of the collection
	 * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
	 */
	public static CollectionInterface createCollection( RemoteCollection model, String name ) throws DAVException, RemoteException {
	    Collection c = (Collection)Collection.createCollection(model.collection, name);
	    return new RemoteCollection(c);
	}
	
	/**
	 * Bind to a new resource in the existing model collection. This constructor also reuses the model's 
	 * connection for the new resource.
	 * @param model The collection in which the resource will reside
	 * @param name The name of the resource
	 * @return ResourceInterface The new resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
	 */
	public static ResourceInterface createResource( RemoteCollection model, String name ) throws DAVException, RemoteException {
		Resource r = (Resource)Collection.createResource(model.collection, name);
		return new RemoteResource(r);
	}

	public RemoteCollection (SharedResourceInterface sr) throws DAVException, RemoteException
    {
        super(sr);
        collection = new Collection(sr);
    }

	public RemoteCollection (SharedResourceInterface sr, String urlStr) throws DAVException, RemoteException
    {
        super(sr, urlStr);
        collection = new Collection(sr, urlStr);
    }

	public RemoteCollection (Collection c) throws DAVException, RemoteException
    {
        super(c);
        collection = c;
    }

    /**
     * Create a new collection in an existing collection
     *
     * @param collName The name of the collection to be created
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection( String collName )
        throws DAVException, RemoteException
    {
        return createCollection( collName, null, null, null );
    }

    /**
     * Create a new collection in the existing collection
     *
     * @param collName The name of the collection to be created
     * @param logger The logger to be utilized for reporting
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection( String collName, LoggerInterface logger )
        throws DAVException, RemoteException
    {
        return createCollection( collName, null, null, logger );
    }

    /**
     * Create a new collection in the existing collection
     *
     * @param collName The name of the collection to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection(String collName, String userName, String pwd)
        throws DAVException, RemoteException
    {
        return createCollection( collName, userName, pwd, null );
    }

    /**
     * Create a new collection in the existing collection
     *
     * @param collName The name of the collection to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @param logger The logger to be utilized for reporting
     * @return CollectionInterface The newly created collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public CollectionInterface createCollection(String collName, String userName, String pwd, LoggerInterface logger)
        throws DAVException, RemoteException
    {
		Collection c = (Collection)collection.createCollection(collName, userName, pwd, logger);
		return new RemoteCollection(c);
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource( String resName )
        throws DAVException, RemoteException
    {
        return createResource( resName, null, null, null );
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource(String resName, String userName, String pwd)
        throws DAVException, RemoteException
    {
        return createResource( resName, userName, pwd, null );
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @param logger The logger to be utilized for reporting
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource(String resName, LoggerInterface logger)
        throws DAVException, RemoteException
    {
		Resource r = (Resource)collection.createResource(resName, logger);
		return new RemoteResource(r);
    }

    /**
     * Create a new resource in the existing collection
     *
     * @param resName The name of the resource to be created
     * @param userName The user name/id for authentication purposes
     * @param pwd The password for authentication purposes
     * @param logger The logger to be utilized for reporting
     * @return ResourceInterface The newly created resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface createResource(String resName, String userName, String pwd, LoggerInterface logger)
        throws DAVException, RemoteException
    {
        return new RemoteResource(resName, userName, pwd, logger);
    }

    /**
     * Get a list of the immediate items (depth=1) in the collection
     *
     * @return List (of ResourceInterface/CollectionInterface) The list of items found in the collection
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public List getChildren() throws DAVException, RemoteException
    {
        return collection.getChildren();
    }

    /**
     * Get all the properties of a collection's members (depth=1) in as single server request. This
     * does not bind to each one of the items in the collection and is a quick way to find out what
     * items are in the collection without fetching each one.
     *
     * @return Map (maps urls to (Map of property name to value)) The property map for all the
     * items in this collection.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public Map getChildrenProperties() throws DAVException, RemoteException
    {
        return collection.getChildrenProperties();
    }

    /**
     * Returns null for the contents of a collection
     * @return null
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public InputStream getContents() throws DAVException, RemoteException {
		return null;
    }
    
    Collection getCollection() {
        return collection;
    }
    
}