/* $Id: RemoteResource.java,v 1.1.4.2 2005/06/22 15:15:35 eurrow Exp $ */
// Copyright (c) 2001-2004 SAS Institute Inc., Cary, NC 27513
// All Rights Reserved.

package com.sas.services.webdav;

import java.io.InputStream;
import java.net.URL;
import java.rmi.RemoteException;

import com.sas.services.logging.LoggerInterface;
import com.sas.services.session.SessionContextInterface;

/**
 * This class represents a remotable DAV resource and the operations expected of a resource. 
 * <p>
 * @author  Rob Owen
 * @version 1.0 4-Mar-05
 * @since   1.0
 */
public class RemoteResource extends java.rmi.server.UnicastRemoteObject implements ResourceInterface
{
    /**
     * Keep a handle to the real collection and delegate to it for all method calls
     */
    protected Resource resource = null;
    protected String _urlString = null;
    
    public RemoteResource() throws RemoteException {
        // nothing specific to do here
    }

    /**
     * Creates a RemoteResource from an existing Resource.
     * @param r The existing resource object
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource(Resource r) throws RemoteException {
        this.resource = r;
    }

    /**
     * Creates a RemoteResource given a URL
     * @param url The URL for the DAV resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource ( String url ) throws DAVException, RemoteException
    {
        resource = new Resource(url);
    }

    /**
     * Creates a RemoteResource given a URL, userid and password.
     * @param url The URL for the DAV resource
     * @param userid The user name for authentication
     * @param pw The password for authentication
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource ( String url, String userid, String pw ) throws DAVException, RemoteException
    {
        resource = new Resource(url, userid, pw);
    }

    /**
     * Creates a RemoteResource given a URL and a logger.
     * @param url The URL for the DAV resource
     * @param logger The logger to be utilized by DAVResource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource ( String url, LoggerInterface logger ) throws DAVException, RemoteException
    {
        resource = new Resource(url, logger);
    }

    /**
     * Creates a RemoteResource given a URL, userid, password and logger.
     * @param url The URL for the DAV resource
     * @param userid The user name for authentication
     * @param pw The password for authentication
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource ( String url, String userid, String pw, LoggerInterface logger ) throws DAVException, RemoteException
    {
        resource = new Resource(url, userid, pw, logger);
    }

    /**
     * Creates a RemoteResource given a URL, session context and domain.
     * @param url The URL for the DAV resource
     * @param sessionContext The context in which the user's credentials can be obtained
     * @param domain The domain to use to find user credentials for the server
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource ( String url, SessionContextInterface sessionContext, String domain ) throws DAVException, RemoteException
    {
        resource = new Resource(url, sessionContext, domain);
    }
    /**
     * Creates a RemoteResource given a DAVResource.
     * @param r A DAVResource can be used to create a Resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource (DAVResource r) throws DAVException, RemoteException
    {
        resource = new Resource(r);
    }

    /**
     * Creates a RemoteResource given a SharedResource.
     * @param sr A shared resource that can be used
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource (SharedResourceInterface sr) throws DAVException, RemoteException
    {
        resource = new Resource(sr);
    }
    /**
     * Creates a RemoteResource given a SharedResource and a URL.
     * @param sr A shared resource that can be used
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public RemoteResource (SharedResourceInterface sr, String urlStr) throws DAVException, RemoteException
    {
        resource = new Resource(sr, urlStr);
    }

    /**
    * Get the URL for this resource
    * @return String A string representation of the URL that the resource represents
    * @throws DAVException in the case of a WebDAV error
    * @throws RemoteException in the case of a network error
    */
    public String getURL() throws DAVException, RemoteException
    {
        return resource.getURL();
    }

    /**
    * Get the name for this resource
    * @return String The name of this resource. This is the item name in a collection for example.
    * @throws DAVException in the case of a WebDAV error
    * @throws RemoteException in the case of a network error
    */
    public String getName() throws DAVException, RemoteException
    {
    	return resource.getName();
    }

    /**
     * Delete the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void delete() throws DAVException, RemoteException
    {
        resource.delete();
    }

    /**
     * Lock the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void lock() throws DAVException, RemoteException
    {
        resource.lock();
    }

    /**
     * Unlock the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void unlock() throws DAVException, RemoteException
    {
        resource.unlock();
    }

    /**
     * Copy the item to destination
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface copy(String destination) throws DAVException, RemoteException
    {
        return resource.copy(destination);
    }

    /**
     * Move an item to destination
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface move(String destination) throws DAVException, RemoteException
    {
        return resource.move(destination);
    }

    /**
     * Compare two items - are they equal
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public boolean equals(ResourceInterface r) throws DAVException, RemoteException
    { 
        return resource.equals(r);
    }


    /**
     * Remove properties from an item
     * @param pMap PropertyMap with all the property names to be removed from the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void removeProperties(PropertyMap pMap) throws DAVException, RemoteException
    {
        resource.removeProperties(pMap);
    }


    /**
     * Set the properties of an item
     * @param pMap PropertyMap with all the properties to be set on the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setProperties(PropertyMap pMap) throws DAVException, RemoteException
    {
        resource.setProperties(pMap);
	    return;
    }

    /**
     * Get the properties of a resource.
     * @return PropertyMap (maps property names to values) The property map contains the set of DAV properties that
     * exist on the resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public PropertyMap getProperties()  throws DAVException, RemoteException {

        return resource.getProperties();
    }

    /**
     * Set the contents of the item.
     * @param input A byte array with the contents for the resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setContents(byte[] input) throws DAVException, RemoteException
    {
        resource.setContents(input);
    }

    /**
     * Set the contents of the item from an input stream. For remote deployment 
     * the InputStream should be wrapped in an {@link com.sas.io.InputStreamAdapter InputStreamAdapter}
     * @param is An input stream that contains the contents
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setContents( InputStream is ) throws DAVException, RemoteException
    {
        resource.setContents(is);
    }

    /**
     * Set the contents of the item.
     * @param url The url for the content for this resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setContents( URL url ) throws DAVException, RemoteException
    {
        resource.setContents(url);
    }

    /**
     * Get the contents of the item.
     * @return InputStream An input stream that can be used to retrieve the content of this item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public InputStream getContents() throws DAVException, RemoteException
    {
        return resource.getContents();
    }

    /**
    * Set the debug level for the underlying DAVResource
    * @param level The level of debug - this can be any integer value
    */
    protected void setDebug (int level) {
        resource.setDebug(level);
    }

    /**
     * Checks whether the WebDAV resource exists.
     * @return boolean WebDAV resource exists or not
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public boolean exists() throws DAVException, RemoteException {
        return resource.exists();
    }
    
    /**
     * Close the underlying DAVResource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void close() throws DAVException, RemoteException {
        resource.close();

    }

    /**
     * Is this Resource using a SharedResource?
     * @return boolean True if it is using a SharedResource, otherwise not
     */
    public boolean isSharedResource() {
        return resource.isSharedResource();
    }
    
    /**
     * Returns the SharedResource, if one is being used
     * @return SharedResourceInterface The SharedResource being used
     */
    public SharedResourceInterface getSharedResource() {
        return resource.getSharedResource();
    }
}