/* $Id$ */
// Copyright (c) 2001-2004 SAS Institute Inc., Cary, NC 27513
// All Rights Reserved.

package com.sas.services.webdav;

import java.io.File;
import java.io.InputStream;
import java.net.URL;
import java.rmi.RemoteException;
import java.security.PrivilegedExceptionAction;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import com.sas.services.logging.LoggerInterface;
import com.sas.services.session.SessionContextInterface;
import com.sas.services.user.UserContextInterface;
import com.sas.services.user.UserIdentityInterface;
import com.sas.text.Message;

/**
 * This class represents a non-collection resource.
 * <p>
 * @author  Rob Owen
 * @version 1.0 15-Nov-2001
 */
public class Resource implements ResourceInterface
{
    protected DAVResource _resource                   = null;
    protected PropertyMap _properties                 = null;
    private   SessionContextInterface _sessionContext = null;
    private   LoggerInterface _logger                 = null;
    private   String _domain                          = null;
    private   String _urlString                       = null;
    protected String _path                            = null;
    private   String _name                            = null;
    protected SharedResourceInterface sharedResource  = null;
    protected boolean isSharedResource = false;

    /**
     * Creates a DAV resource.
     *
     * @param url The URL for the DAV resource
     *
     */
    public Resource ( String url ) throws DAVException
    {
        _resource = new DAVResource( new ConnectionInfo(url), null );
		_urlString = url;
		_path = getPath( url );
		File f = new File( _urlString );
		_name = f.getName();
    }

    /**
     * Creates a DAV resource.
     *
     * @param url The URL for the DAV resource
     * @param userid The user name for authentication
     * @param pw The password for authentication
     *
     */
    public Resource ( String url, String userid, String pw ) throws DAVException
    {
        _resource = new DAVResource( new ConnectionInfo(url, userid, pw), null );
		_urlString = url;
		_path = getPath( url );
		File f = new File( _urlString );
		_name = f.getName();
    }

    /**
     * Creates a DAV resource.
     *
     * @param url The URL for the DAV resource
     * @param logger The logger to be utilized by DAVResource
     *
     */
    public Resource ( String url, LoggerInterface logger ) throws DAVException
    {
        _logger = logger;
        _resource = new DAVResource( new ConnectionInfo(url), _logger );
		_urlString = url;
		_path = getPath( url );
		File f = new File( _urlString );
		_name = f.getName();
    }

    /**
     * Creates a DAV resource.
     *
     * @param url The URL for the DAV resource
     * @param userid The user name for authentication
     * @param pw The password for authentication
     *
     */
    public Resource ( String url, String userid, String pw, LoggerInterface logger ) throws DAVException
    {
        _logger = logger;
        _resource = new DAVResource( new ConnectionInfo(url, userid, pw), _logger );
		_urlString = url;
		_path = getPath( url );
		File f = new File( _urlString );
		_name = f.getName();
    }

    /**
     * Creates a DAV resource.
     *
     * @param url The URL for the DAV resource
     * @param sessionContext The context in which the user's credentials can be obtained
     * @param domain The domain to use to find user credentials for the server
     *
     */
    public Resource ( String url, SessionContextInterface sessionContext, String domain ) throws DAVException, RemoteException
    {
        _sessionContext = sessionContext;
        _domain = domain;
        _urlString = url;
		_path = getPath( url );
		File f = new File( _urlString );
		_name = f.getName();
        try {
            if (_sessionContext != null) {
                _logger = _sessionContext.getLogger();
            }
        } catch (RemoteException e) {
            throw new DAVException( RB.getStringResource("Resource.getloggerexception.ex.txt"), e );
        }
        
        ConnectionInfo cInfo = new ConnectionInfo( url );
        if (_sessionContext != null && _domain != null) {
            final UserContextInterface _userContext = _sessionContext.getUserContext();
            if (_userContext != null) {
				PrivilegedExceptionAction action = new PrivilegedExceptionAction()
				{
				   public Object run() throws Exception
				   {
					   return _userContext.getIdentityByDomain( _domain );
				   }
				};

				UserIdentityInterface _identity    = null;
				try
				{
					_identity = (UserIdentityInterface)java.security.AccessController.doPrivileged(action);               
				}
				catch (java.security.PrivilegedActionException pae)
				{
					Throwable t = pae.getCause();
					if (t instanceof RemoteException) {
						throw (RemoteException)t;
					} else {
						throw new DAVException( RB.getStringResource("Resource.privop.txt"),pae);
					} 
				}   

                if ((_userContext != null) && (_identity != null) ) {
                    cInfo.setUser( (String)_identity.getPrincipal() );
                    cInfo.setPW( (String)_identity.getCredential() );
                } 
            }
		}
        _resource = new DAVResource( cInfo, _logger );
    }
    /**
     * Creates a DAV resource.
     *
     * @param r A DAVResource can be used to create a Resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     *
     */
    public Resource (DAVResource r) throws DAVException, RemoteException
    {
        _resource = r;
		_urlString = _resource.getURL().toString();
		_path = getPath( _urlString );
		File f = new File( _urlString );
		_name = f.getName();
    }

    /**
     * Creates a DAV resource that uses a shared resource.
     * @param sr A shared resource that can be used
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public Resource (SharedResourceInterface sr) throws DAVException, RemoteException
    {
        sharedResource = sr;
        String urlStr = null;
        DAVResourceInterface r= null;
        try {
            r = sharedResource.getConnection();
            urlStr = r.getURL().toString();
        } finally {
            sharedResource.releaseConnection(r);
            r = null;
        }
        isSharedResource = true;
        sharedResource = sr;
        _urlString = urlStr;
		_path = getPath( urlStr );
		_name = new File(urlStr).getName();
        
    }
    /**
     * Creates a DAV resource using a shared resource and represents the location specified by the urlStr. 
     * @param sr A shared resource that can be used
     * @param urlStr The location that the new Resource should represent
     */
    public Resource (SharedResourceInterface sr, String urlStr) 
    {
        isSharedResource = true;
        sharedResource = sr;
        _urlString = urlStr;
		_path = getPath( urlStr );
		_name = new File(urlStr).getName();
    }

    /**
     * Get the URL for this resource
     * @return String A string representation of the URL that the resource represents
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
    */
    public String getURL() throws DAVException, RemoteException
    {
        return _urlString;
    }

    /**
     * Get the name for this resource
     * @return String The name of this resource. This is the item name in a collection for example.
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public String getName() throws DAVException, RemoteException
    {
    	return _name;
    }

    /**
     * Delete the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void delete() throws DAVException, RemoteException
    {
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
                r.delete();
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        _resource.delete();
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.delres.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
		        return;
			}
        }
    }

    /**
     * Lock the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void lock() throws DAVException, RemoteException
    {
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.lock();
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        _resource.lock();
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.lockres.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
		        return;
			}
        }
    }

    /**
     * Unlock the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void unlock() throws DAVException, RemoteException
    {
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.unlock();
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        _resource.unlock();
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.unlockres.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
		        return;
			}
        }
    }

    /**
     * Copy the item to destination
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface copy(String destination) throws DAVException, RemoteException
    {
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.copy(destination);
                return new Resource(sharedResource, destination);
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        DAVResource copied = _resource.copy(destination);
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.copyres.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
		        return new Resource(copied);
			}
        }
    }

    /**
     * Move an item to destination
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public ResourceInterface move(String destination) throws DAVException, RemoteException
    {
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.move(destination);
                return new Resource(sharedResource, destination);
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        DAVResource moved = _resource.move(destination);
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.moveres.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
		        return new Resource(moved);
			}
        }
    }

    /**
     * Compare two items - are they equal
     * @return boolean True if they represent the same resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public boolean equals(ResourceInterface r) throws DAVException, RemoteException
    {
        if ( !isSharedResource ) {
	        synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        return _resource.getURL().toString().equals(r.getURL().toString());
	        }
		} else {
		    return false;
		}
    }


    /**
     * Remove properties from an item
     * @param pMap PropertyMap with all the property names to be removed from the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void removeProperties(PropertyMap pMap) throws DAVException, RemoteException
    {
        String namespace;
        String pName;
        int nsCounter = 0;
        ArrayList pNames = new ArrayList();
        Iterator entries = pMap.entrySet().iterator();
        HashMap usedPrefixes = new HashMap();
        while (entries.hasNext())
        {
            Map.Entry pm = (Map.Entry)entries.next();
            String name = ((String) pm.getKey()).trim();
            // Does the name have a namespace (separator is '{namespace}')
            int sep = name.indexOf('}');
            if (name.charAt(0) == '{' && sep > 0) {
                namespace = name.substring(1, sep);
                pName     = name.substring(sep+1);
            } else {
                namespace = "";
                pName     = name;
            }
            String prefix = (String)usedPrefixes.get(namespace);
            if (prefix == null) {
                prefix = "ns"+nsCounter;
                nsCounter += 1;
                usedPrefixes.put(namespace, prefix);
            }
            pNames.add(new PropertyName(pName, prefix, namespace));
        }

        // Convert to array PropertyName[]
        PropertyName[] pNameArray = (PropertyName[]) pNames.toArray(new PropertyName[0]);
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.removeProperties(pNameArray);
		        if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.remprop.ex.fmt.txt",
                        r.getStatusMessage() ) );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        _resource.removeProperties(pNameArray);
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.remprop.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
		        return;
			}
        }
    }


    /**
     * Set the properties of an item
     * @param pMap PropertyMap with all the properties to be set on the item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setProperties(PropertyMap pMap) throws DAVException, RemoteException
    {
        String namespace;
        String pName;
        int nsCounter = 0;
        ArrayList pNames = new ArrayList();
        ArrayList pVals  = new ArrayList();
        Iterator entries = pMap.entrySet().iterator();
        HashMap usedPrefixes = new HashMap();
        while (entries.hasNext())
        {
            Map.Entry pm = (Map.Entry)entries.next();
            String name = (String) pm.getKey();
            if (name == null || name.length() == 0) continue;
            name = name.trim();

            // Does the name have a namespace (separator is '{namespace}')
            int sep = name.indexOf('}');
            if (name.charAt(0) == '{' && sep > 0) {
                namespace = name.substring(1, sep);
                pName     = name.substring(sep+1);
            } else {
                namespace = "";
                pName     = name;
            }
            String prefix = (String)usedPrefixes.get(namespace);
            if (prefix == null && !namespace.equals("")) {
                prefix = "ns"+nsCounter;
                nsCounter += 1;
                usedPrefixes.put(namespace, prefix);
            }
            pNames.add(new PropertyName(pName, prefix, namespace));
            pVals.add(pm.getValue());
        }

        // Convert to arrays PropertyName[] pp1Array = pp1.toArray(PropertyName[]);
        PropertyName[] pNameArray = (PropertyName[]) pNames.toArray(new PropertyName[0]);
        String[] pValArray = (String[]) pVals.toArray(new String[0]);
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setProperties(pNameArray, pValArray);
		        if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setprop.ex.fmt.txt",
                        r.getStatusMessage() ) );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        _resource.setProperties(pNameArray, pValArray);
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setprop.ex.fmt.txt",
	                                                                              _resource.getStatusMessage() ) );
			}
        }
	    return;
    }

    /**
     * Get the properties of a resource.
     * @return PropertyMap (maps property names to values) The property map contains the set of DAV properties that
     * exist on the resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public PropertyMap getProperties()  throws DAVException, RemoteException {

        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
		        if (_properties == null) {
		            _properties = new PropertyMap();
		            org.apache.webdav.lib.Property[] pp = r.getProperties();
		            if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.getprop.ex.fmt.txt",
		                                                                                  r.getStatusMessage() ) );
		           _properties = Utils.convertPropertyArray( pp );
		        }
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        if (_properties == null) {
		            _properties = new PropertyMap();
		            org.apache.webdav.lib.Property[] pp = _resource.getProperties();
		            if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.getprop.ex.fmt.txt",
		                                                                                  _resource.getStatusMessage() ) );
		           _properties = Utils.convertPropertyArray( pp );
		        }
			}
        }
        return _properties;
    }

    /**
     * Set the contents of the item.
     * @param input A byte array with the contents for the resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setContents(byte[] input) throws DAVException, RemoteException
    {

        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
    	        if (r.isCollection()) return;
    	        r.setContents(input);
    	        if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setcontent.ex.fmt.txt",
    	                                                                              r.getStatusMessage() ) );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        if (_resource.isCollection()) return;
		        _resource.setContents(input);
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setcontent.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
			}
        }
        return;
    }

    /**
     * Set the contents of the item from an input stream. For remote deployment 
     * the InputStream should be wrapped in an {@link com.sas.io.InputStreamAdapter InputStreamAdapter}
     * @param is An input stream that contains the contents
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setContents( InputStream is ) throws DAVException, RemoteException
    {

        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
    	        if (r.isCollection()) return;
    	        r.setContents(is);
    	        if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setcontent.ex.fmt.txt",
    	                                                                              r.getStatusMessage() ) );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        if (_resource.isCollection()) return;
		        _resource.setContents( is );
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setcontent.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
			}
        }
    }

    /**
     * Set the contents of the item.
     * @param url The url for the content for this resource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void setContents( URL url ) throws DAVException, RemoteException
    {

        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
    	        if (r.isCollection()) return;
    	        r.setContents(url);
    	        if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setcontent.ex.fmt.txt",
    	                                                                              _resource.getStatusMessage() ) );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        if (_resource.isCollection()) return;
		        _resource.setContents( url );
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.setcontent.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
			}
        }
    }

    /**
     * Get the contents of the item.
     * @return InputStream An input stream that can be used to retrieve the content of this item
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public InputStream getContents() throws DAVException, RemoteException
    {
        InputStream _result = null;
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
    	        _result = r.getContentsInputStream();
    	        if ( !r.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.getcontent.ex.fmt.txt",
    	                                                                              r.getStatusMessage() ) );
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
			synchronized(_resource) {
		        if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        _result = _resource.getContentsInputStream();
		        if ( !_resource.SUCCEEDED() ) throw new DAVException( Message.format( RB.getResources(), "Resource.getcontent.ex.fmt.txt",
		                                                                              _resource.getStatusMessage() ) );
			}
        }
        return _result;
    }

    /**
    * Set the debug level for the underlying DAVResource
    * @param level The level of debug - this can be any integer value
    */
    protected void setDebug (int level)
    {
        if ( !isSharedResource ) _resource.setDebug(level);
    }

    /**
     * Get the underlying implementation object.
     * @return DAVResource
     */
    public DAVResource getDAVResource()
    {
        if ( !isSharedResource ) {
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        return _resource;
			}
        } else return null;
    }
    
    /**
     * Checks whether the WebDAV resource exists.
     * @return boolean WebDAV resource exists or not
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public boolean exists() throws DAVException, RemoteException {
        if (isSharedResource) {
            DAVResourceInterface r= null;
            try {
                r = sharedResource.getConnection();
                r.setPath(_path);
    	        return r.exists();
            } finally {
                sharedResource.releaseConnection(r);
                r = null;
            }
        } else {
            if ( _resource == null ) return false;
			synchronized(_resource) {
				if ( !_path.equals( _resource.getPath() ) ) _resource.setPath( _path );
		        return _resource.exists();
			}
        }
    }
    
    /**
     * Extract the path/query from the url. This is the part after the scheme, host and port.
     * @param url
     * @return String The path/query portion of the url 
     */
    private String getPath( String url ) {
		String path = null;
        int p = url.indexOf("://");
        if ( p > 0 ) {
            p = url.indexOf("/", p+3);
            path = (p > 0) ? url.substring(p) : "";
        } else path=url;
		return path;
    }

    /**
     * Close the underlying DAVResource
     * @throws DAVException in the case of a WebDAV error
     * @throws RemoteException in the case of a network error
     */
    public void close() throws DAVException, RemoteException {
        if (!isSharedResource) {
            if ( _resource != null ) _resource.close();
        }
    }
    
    /**
     * Is this Resource using a SharedResource?
     * @return boolean True if it is using a SharedResource, otherwise not
     */
    public boolean isSharedResource() {
        return isSharedResource;
    }
    
    /**
     * Returns the SharedResource, if one is being used
     * @return SharedResourceInterface The SharedResource being used
     */
    public SharedResourceInterface getSharedResource() {
        if ( isSharedResource ) return sharedResource;
        else return null;
    }
}