/* $Id$ */
// Copyright (c) 2001 SAS Institute Inc., Cary, NC 27513
package com.sas.services.webdav;

import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import org.apache.commons.httpclient.Credentials;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.webdav.lib.methods.OptionsMethod;
import org.apache.webdav.lib.methods.SearchMethod;

import com.sas.services.logging.LoggerInterface;
import com.sas.text.Message;

/**
 * Factory interface for getting the root collections - setup from initialisation files. There
 * may be multiple root collections (multiple storage locations). Access to a root collection
 * allows one to create further collections and resources (elements). It is also possible to
 * "bind" to an existing resource using the url
 * @author  Rob Owen
 * @version 1.0 15-Nov-2001
 */
public class ResourceFactory
{
    /**
     * Search for content satifying the search filter on item metadata
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static List search (String serverUrl, String searchString /*SearchFilter search*/,
                                   String userName,  String pwd) throws DAVException, RemoteException
    {
        ArrayList results = new ArrayList();

        // Check this is a valid url
        URL serverURL = null;
        try {
            serverURL = new URL(serverUrl);
        } catch (MalformedURLException mue) {
            System.out.println( Message.format( RB.getResources(), "ResourceFactory.invalidurl.fmt.txt", serverUrl));
            return null;
        }

        HttpClient client = new HttpClient();
        client.startSession(serverURL);
        if (userName != null && !userName.equals("")) client.setCredentials(new Credentials(userName, pwd));

        // Get supported methods and DAV capabilities from the server
        OptionsMethod optMethod = new OptionsMethod(serverURL.getPath());
        try {
            client.executeMethod(optMethod);
        } catch (Exception ex) {
            ex.printStackTrace();
            System.out.println(ex.getMessage());
        }
        if (!optMethod.isAllowed("SEARCH")) {
            System.out.println( RB.getStringResource("ResourceFactory.nosearch.txt") );
            return results;
        }

        // Ok, now do the search
        SearchMethod amethod = new SearchMethod(serverURL.getPath(), searchString);
        //amethod.setDebug(Integer.MAX_VALUE);
        try {
            client.executeMethod(amethod);
        }
        catch (HttpException e) {
            throw new DAVException( RB.getStringResource("ResourceFactory.searchhttperr.ex.txt"), e);
        }
        catch (java.io.IOException e) {
            throw new DAVException( RB.getStringResource("ResourceFactory.searchioerr.ex.txt"), e);
        }

        Enumeration e = amethod.getAllResponseURLs();
        while (e.hasMoreElements()) {
            String rurl = (String) e.nextElement();
            Enumeration e2 = amethod.getResponseProperties(rurl);
            while (e2.hasMoreElements()) {
                results.add( e2.nextElement());
            }
        }
        return results;
    }

    /**
     * Search for content satifying the search filter on item metadata
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static List search ( ConnectionInfo cInfo, String searchString) throws DAVException, RemoteException
    {
        ArrayList results = new ArrayList();

        // Check this is a valid url
        try {
            new URL( cInfo.getUrl() );
        } catch (MalformedURLException mue) {
            System.out.println( Message.format( RB.getResources(), "ResourceFactory.invalidurl.fmt.txt", cInfo.getUrl()));
            return null;
        }

        DAVResource r = new DAVResource( cInfo , null );
        r.doOptions();

        // Get supported methods and DAV capabilities from the server
        String[] methods = r.getAllowedMethods();
        boolean searchOk = false;
        for (int i=0; i<methods.length; i++) {
            if ( "SEARCH".equalsIgnoreCase(methods[i]) ) {
                searchOk = true;
                break;
            }
        }
        if ( !searchOk ) {
            System.out.println( RB.getStringResource("ResourceFactory.nosearch.txt") );
            return results;
        }


        // Ok, now do the search
        org.apache.webdav.lib.Property pp[];
        try {
            pp = r.search( searchString );
        }
        catch (java.io.IOException e) {
            throw new DAVException( RB.getStringResource("ResourceFactory.searchioerr.ex.txt"), e);
        }

        for (int i=0; i<pp.length; i++) {
            results.add( pp[i] );
        }
        return results;
    }

    /**
     * Bind the url to a resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static ResourceInterface bindResource(String url) throws DAVException, RemoteException
    {
        return bindResource( new ConnectionInfo(url), null );
    }

    /**
     * Bind the url to a resource and set the logger on the resource
     * @param url The URL for the resource
     * @param logger The logger to be used by the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static ResourceInterface bindResource(String url, LoggerInterface logger) throws DAVException, RemoteException
    {
        return bindResource( new ConnectionInfo(url), logger );
    }

    /**
     * Bind the url to a resource with authentication
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static ResourceInterface bindResource(String url, String user, String password) throws DAVException, RemoteException
    {
        return bindResource( new ConnectionInfo(url, user, password), null );
    }

    /**
     * Bind the url to a resource with authentication, and set the logger on the resource
     * @param url The URL for the resource
     * @param user The userid to be used for authentication
     * @param password The password for authentication
     * @param logger The logger to be used by the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static ResourceInterface bindResource(String url, String user, String password, LoggerInterface logger) throws DAVException, RemoteException
    {
        return bindResource( new ConnectionInfo(url, user, password), logger );
    }

    /**
     * Bind the url to a resource with authentication, and set the logger on the resource
     * @param cInfo The connection information for the resource
     * @param logger The logger to be used by the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static ResourceInterface bindResource( ConnectionInfo cInfo, LoggerInterface logger) throws DAVException, RemoteException
    {
        DAVResource r = new DAVResource( cInfo, logger);

        // If no error, or not found (null resource)
        if (r.SUCCEEDED() || r.getStatusCode() == HttpStatus.SC_NOT_FOUND) {
            if (r.isCollection()) return new Collection(r);
            else return new Resource(r);
        }
        return null;
    }
    
	/**
	 * Bind to a resource using the model resource and path. The new resource uses the same connection as the model.
	 * @param model The model resource
	 * @param path The new path for the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
	 */
    public static ResourceInterface bindResource( ResourceInterface model, String path ) throws DAVException, RemoteException
    {
        if ( model.isSharedResource() ) {
        	SharedResourceInterface sr = model.getSharedResource();
        	if ( sr instanceof SharedResource) {
	    	    DAVResourceInterface r = null;
	    	    String urlStr = null;
	        	try {
	        	    r = sr.getConnection();
	        	    r.setPath(path);
	        		urlStr = r.getURL().toString();
					if (r.SUCCEEDED() || r.getStatusCode() == HttpStatus.SC_NOT_FOUND) {
						if (r.isCollection()) return new Collection(sr, urlStr);
						else return new Resource(sr, urlStr);
					}
	        	} finally {
	        	    sr.releaseConnection(r);
	        	    r = null;
	        	}
        	} else return model.getSharedResource().bindResource(path);
        } else {
	    	DAVResource r = null;
            if ( model instanceof Collection ) {
		    	r = ((Collection)model)._resource;
            } else if (model instanceof Resource) {
		    	r = ((Resource)model)._resource;
            }
	    	synchronized (r ) {
	    		r.setPath( path );
				// If no error, or not found (null resosurce)
				if (r.SUCCEEDED() || r.getStatusCode() == HttpStatus.SC_NOT_FOUND) {
					if (r.isCollection()) return new Collection(r);
					else return new Resource(r);
				}
	    	}
        }
		return null;
    }

	/**
	 * Bind to a resource using the model resource and path. The new resource uses the same connection as the model. 
	 * The resource type is found from the property map instead of having to actually get the information from 
	 * the server.
	 * @param model The model resource
	 * @param path The new path for the resource
	 * @param p The property map for the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
	 */
    public static ResourceInterface bindResource(ResourceInterface model, String path, PropertyMap p) throws DAVException, RemoteException {
        String collection = (String) p.get(DAVConstants.DAV_RESOURCETYPE);
        if (model.isSharedResource()) {
            SharedResourceInterface sr = model.getSharedResource();
            if (sr instanceof SharedResource) {
                DAVResourceInterface r = null;
                String urlStr = null;
                try {
                    r = sr.getConnection();
                    r.setPath(path);
                    urlStr = r.getURL().toString();
                } finally {
                    sr.releaseConnection(r);
                    r = null;
                }
                if (DAVConstants.DAV_COLLECTION.equals(collection)) return new Collection(sr, urlStr);
                else return new Resource(sr, urlStr);
            } else return model.getSharedResource().bindResource(path, p);
        } else {
            DAVResource r = null;
            if (model instanceof Collection) {
                r = ((Collection) model)._resource;
                r.setPath(path);
            } else if (model instanceof Resource) {
                r = ((Resource) model)._resource;
                r.setPath(path);
            }
            if (DAVConstants.DAV_COLLECTION.equals(collection)) return new Collection(r);
            else return new Resource(r);
        }
    }

    /**
	 * Bind to a resource using the shared resource and path. The new resource uses the same shared resource. 
	 * @param sr The shared resource
	 * @param path The new path for the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
	 */
    public static ResourceInterface bindResource( SharedResourceInterface sr, String path ) throws DAVException, RemoteException
    {
        if ( sr instanceof SharedResource ) {
		    DAVResourceInterface r = null;
	    	try {
	    	    r = sr.getConnection();
	    		r.setPath( path );
	    		String urlStr = r.getURL().toString();
				// If no error, or not found (null resosurce)
				if (r.SUCCEEDED() || r.getStatusCode() == HttpStatus.SC_NOT_FOUND) {
					if (r.isCollection()) return new Collection(sr, urlStr);
		        	else return new Resource(sr, urlStr);
				}
				return null;
	    	} finally {
	    	    sr.releaseConnection(r);
	    	    r = null;
	    	}
        } else {
            return sr.bindResource(path);
        }
    }
    
    /**
	 * Bind to a resource using the shared resource and path. The new resource uses the same shared resource. 
	 * @param sr The shared resource
	 * @param path The new path for the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
	 */
    public static ResourceInterface bindResource( SharedResourceInterface sr, String path, PropertyMap p ) throws DAVException, RemoteException
    {
        if ( sr instanceof SharedResource ) {
		    DAVResourceInterface r = null;
	    	try {
	    	    r = sr.getConnection();
	    		r.setPath( path );
	    		String urlStr = r.getURL().toString();
				// If no error, or not found (null resosurce)
				if (r.SUCCEEDED() || r.getStatusCode() == HttpStatus.SC_NOT_FOUND) {
					if (r.isCollection()) return new Collection(sr, urlStr);
					else return new Resource(sr, urlStr);
				}
				return null;
	    	} finally {
	    	    sr.releaseConnection(r);
	    	    r = null;
	    	}
        } else {
            return sr.bindResource(path,p);
        }
    }
    /**
     * Create a new collection
     * @param url The URL for the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static CollectionInterface createCollection(String url)
        throws DAVException, RemoteException
    {
        return createCollection( new ConnectionInfo(url), null );
    }

    /**
     * Create a new collection and set the logger
     * @param url The URL for the resource
     * @param logger The logger to be used by the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static CollectionInterface createCollection(String url, LoggerInterface logger)
        throws DAVException, RemoteException
    {
        return createCollection( new ConnectionInfo(url), logger );
    }

    /**
     * Create a new collection with authentication
     * @param url The URL for the resource
     * @param user The user name for authentication
     * @param password The password for authentication
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static CollectionInterface createCollection(String url, String user, String password)
        throws DAVException, RemoteException
    {
        return createCollection( new ConnectionInfo(url, user, password), null );
    }

    /**
     * Create a new collection with authentication and set the logger
     * @param url The URL for the resource
     * @param user The user name for authentication
     * @param password The password for authentication
     * @param logger The logger to be used by the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static CollectionInterface createCollection(String url, String user, String password, LoggerInterface logger)
        throws DAVException, RemoteException
    {
        return createCollection( new ConnectionInfo(url, user, password), logger );
    }

    /**
     * Create a new collection with authentication and set the logger
     * @param cInfo The connection information for the resource
     * @param logger The logger to be used by the resource
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
     */
    public static CollectionInterface createCollection( ConnectionInfo cInfo, LoggerInterface logger)
        throws DAVException, RemoteException
    {
            DAVResource r = new DAVResource( cInfo, logger);
            r.makeCollection();
            if (r.getStatusCode() == HttpStatus.SC_METHOD_NOT_ALLOWED) {

                // Collection already exists, lets delete it and try again
                if (logger != null && logger.isDebugEnabled( ResourceFactory.class.getName() ))
                    logger.debug( Message.format( RB.getResources(), "ResourceFactory.deletecollection.fmt.txt", cInfo.getUrl() ));
                r.delete();
                r.makeCollection();
            }
            if (r.SUCCEEDED()) return new Collection(r);
            else throw new DAVException( Message.format( RB.getResources(), "ResourceFactory.mkcolerr.ex.fmt.txt",
                                         String.valueOf(r.getStatusCode()), cInfo.getUrl() ) );
    }
    
	/**
	 * Create a new collection using a model resource and a new path. The new collection uses the same connection as the model.
	 * @param model The resource model
	 * @param urlString The path to the new collection
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
	 */
    public static CollectionInterface createCollection( ResourceInterface model, String urlString ) throws DAVException, RemoteException 
    {
        String path = getPath(urlString);
    	int sc;
    	if ( model.isSharedResource() ) {
    	    SharedResourceInterface sr = model.getSharedResource();
    	    if ( sr instanceof SharedResource ) {
	            DAVResourceInterface r = null;
	        	try {
	        	    r = sr.getConnection();
	        		r.setPath( path );
	    	    	r.makeCollection();
	    			if (r.getStatusCode() == HttpStatus.SC_METHOD_NOT_ALLOWED) {
	    				r.delete();
	    				r.makeCollection();
	    			}
	    			sc = r.getStatusCode();
	        	} finally {
	        	   sr.releaseConnection(r);
	        	   r = null;
	        	}
	    		if (DAVResource.SUCCEEDED(sc)) return new Collection(sr);
	    		else throw new DAVException( Message.format( RB.getResources(), "ResourceFactory.mkcolerr.ex.fmt.txt",
	    									 String.valueOf( sc ), urlString ) );
    	    } else return sr.createCollectionRF(model, urlString);
    	} else {
	    	DAVResource r = null;
            if ( model instanceof Collection ) {
		    	r = ((Collection)model)._resource;
            } else if (model instanceof Resource) {
		    	r = ((Resource)model)._resource;
            }
	    	synchronized(r) {
	    		r.setPath( path );
		    	r.makeCollection();
				if (r.getStatusCode() == HttpStatus.SC_METHOD_NOT_ALLOWED) {
					r.delete();
					r.makeCollection();
				}
				sc = r.getStatusCode();
	    	}
			if (DAVResource.SUCCEEDED(sc)) return new Collection(r);
			else throw new DAVException( Message.format( RB.getResources(), "ResourceFactory.mkcolerr.ex.fmt.txt",
										 String.valueOf( sc ), urlString ) );
    	}
    	
    }
    
	/**
	 * Create a new collection using a shared resource and a new path. The new collection uses the same shared
	 * resource.
	 * @param sr The shared resource whose connection will be used to create the collection
	 * @param urlString The path to the new collection
	 * @throws DAVException in the case of a WebDAV error 
     * @throws RemoteException in the case of a network error
	 */
    public static CollectionInterface createCollection( SharedResourceInterface sr, String urlString ) throws DAVException, RemoteException 
    {
        String path = getPath(urlString);
    	int sc;
        DAVResourceInterface r = null;
    	try {
    	    r = sr.getConnection();
    		r.setPath( path );
	    	r.makeCollection();
			if (r.getStatusCode() == HttpStatus.SC_METHOD_NOT_ALLOWED) {
				r.delete();
				r.makeCollection();
			}
			sc = r.getStatusCode();
    	} finally {
    	   sr.releaseConnection(r);
    	   r = null;
    	}
		if (DAVResource.SUCCEEDED(sc)) return new Collection(sr);
		else throw new DAVException( Message.format( RB.getResources(), "ResourceFactory.mkcolerr.ex.fmt.txt",
									 String.valueOf( sc ), urlString ) );
    }
    
    /**
     * Extract the path/query from the url. This is the part after the scheme, host and port.
     * @param urlStr
     * @return String The path/query portion of the url 
     */
    private static String getPath(String urlStr) {
    	String path = null;
        int p = urlStr.indexOf("://");
        if ( p > 0 ) {
            p = urlStr.indexOf("/", p+3);
            path = (  p > 0 ) ? urlStr.substring(p) : "";
        }
        return path;
    }
}