/* $Id: SharedResource.java,v 1.1.4.2 2005/04/20 22:21:51 eurrow Exp $ */
// Copyright (c) 2002-2004 SAS Institute Inc., Cary, NC 27513
// All Rights Reserved.
package com.sas.services.webdav;

import EDU.oswego.cs.dl.util.concurrent.Mutex;
import java.rmi.server.*;
import java.rmi.*;

/**
 * This class represents a shared resource, like a connection to a server. It controls access to the shared 
 * resource through a mutex. 
 * <p>
 * @author  Rob Owen
 * @version 1.0 27-Feb-05
 * @since   1.0
 */
public class SharedResource extends UnicastRemoteObject implements SharedResourceInterface {
    
    private Mutex _resourceMutex = new Mutex();
    private DAVResourceInterface resource;
    
	/**
	 * Create a new shared resource using the connection information provided.
	 * @param c The connection information provided to set up the new shared resource
     * @throws RemoteException in the case of a network error
	 */
    public SharedResource(ConnectionInfo c) throws RemoteException {
        try {
            resource = new DAVResource(c, null);
        } catch (DAVException e) {
            e.printStackTrace();
        }
    }
    
	/**
	 * Create a new shared resource using the DAVResource provided.
	 * @param r The existing DAVResource to use for the new shared resource
     * @throws RemoteException in the case of a network error
	 */
    public SharedResource(DAVResourceInterface r) throws RemoteException {
        resource = r;
    }
    
	/**
	 * Retrieves the underlying DAVResource. An attempt is made to access the resource using the mutex. Once
	 * the retrieval is successful, no one else can gain access to the underlying DAVResource until it is 
	 * released. See @see #releaseConnection(DAVResourceInterface).  
	 * @return DAVResourceInterface The underlying shared resource that is not remoteable
     * @throws RemoteException in the case of a network error
	 */
    public DAVResourceInterface getConnection() throws RemoteException {

        boolean success = false;
        try {
            success = _resourceMutex.attempt(90000);
        } catch (InterruptedException ie) {
            ie.printStackTrace();
        }
        if (!success) return null;
        return resource;

    }
    
	/**
	 * Releases the underlying DAVResource.  
	 * @param r The underlying shared resource that was previously retrieved
     * @throws RemoteException in the case of a network error
	 */
    public void releaseConnection(DAVResourceInterface r)  throws RemoteException {
        _resourceMutex.release();
    }
    
	/**
	 * Releases the underlying DAVResource.  
     * @throws RemoteException in the case of a network error
	 */
    public void releaseConnection()  throws RemoteException {
        _resourceMutex.release();
    }
    
	/**
	 * Close the underlying DAVResource.  
     * @throws RemoteException in the case of a network error
	 */
    public void close()  throws RemoteException {
        try {
            resource.close();
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }
    
	/**
	 * Calls the Collection.createCollection method for this instance and returns a remote collection. 
	 * @see Collection#createCollection(SharedResourceInterface, CollectionInterface, String)
	 * @param model 
	 * @param name  
     * @throws RemoteException in the case of a network error
	 */
    public CollectionInterface createCollection(CollectionInterface model, String name) throws RemoteException {
        try {
            Object o = Collection.createCollection(this, model, name);
            CollectionInterface c = new RemoteCollection((Collection)o);
            return c;
        } catch (DAVException e) {
            return null;
        }
    }
    
	/**
	 * Calls the Collection.createResource method for this instance and returns a remote resource. 
	 * @see Collection#createResource(SharedResourceInterface, CollectionInterface, String)
	 * @param model 
	 * @param name  
     * @throws RemoteException in the case of a network error
	 */
    public ResourceInterface createResource(CollectionInterface model, String name) throws RemoteException {
        try {
            Object o = Collection.createResource(this, model, name);
            ResourceInterface r = new RemoteResource((Resource)o);
            return r;
        } catch (DAVException e) {
            return null;
        }
    }
    
	/**
	 * Calls the ResourceFactory.createCollection method for this instance and returns a remote collection. 
	 * @see ResourceFactory#createCollection(SharedResourceInterface, String)
	 * @param model 
	 * @param urlString  
     * @throws RemoteException in the case of a network error
	 */
    public CollectionInterface createCollectionRF(ResourceInterface model, String urlString) throws RemoteException {
        try {
            Object o = ResourceFactory.createCollection(this, urlString);
            CollectionInterface c = new RemoteCollection((Collection)o);
            return c;
        } catch (DAVException e) {
            return null;
        }
    }
    
	/**
	 * Calls the ResourceFactory.bindResource method for this instance and returns either a remote collection
	 * or a remote resource. 
	 * @see ResourceFactory#bindResource(SharedResourceInterface, String)
	 * @param path
     * @throws RemoteException in the case of a network error
	 */
    public ResourceInterface bindResource(String path) throws RemoteException {
        try {
            Object o = ResourceFactory.bindResource(this, path);
            ResourceInterface ri;
            if ( o instanceof CollectionInterface) ri = new RemoteCollection((Collection)o);
            else ri = new RemoteResource((Resource)o);
            return ri;
        } catch (IllegalStateException e) {
           throw e;
        } catch (Exception e) {
            throw new RemoteException("DAVException: "+e.getMessage(), e);
        }
    }
    
	/**
	 * Calls the ResourceFactory.bindResource method for this instance and returns a remote collection. 
	 * @see ResourceFactory#bindResource(SharedResourceInterface, String, PropertyMap)
	 * @param path 
	 * @param p  
     * @throws RemoteException in the case of a network error
	 */
    public ResourceInterface bindResource(String path, PropertyMap p) throws RemoteException {
        try {
            Object o = ResourceFactory.bindResource(this, path, p);
            ResourceInterface ri;
            if ( o instanceof CollectionInterface) ri = new RemoteCollection((Collection)o);
            else ri = new RemoteResource((Resource)o);
            return ri;
        }catch (IllegalStateException e) {
           throw e;
        } catch (Exception e) {
            throw new RemoteException("DAVException: "+e.getMessage(), e);
        }
    }
}
