/*
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.rmi.AccessException;
import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.helper.ProjectHelper2;
import org.apache.tools.ant.taskdefs.MacroInstance;

import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.errors.BRMInternalError;
import com.sas.solutions.brm.errors.BRMUserError;
import com.sas.solutions.brm.logging.BRMBuildListener;
import com.sas.solutions.brm.logging.RemoteLoggingClient;
import com.sas.solutions.brm.logging.RemoteLoggingInterface;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.storedProcess.StoredProcDirFinder;
import com.sas.solutions.brm.util.BRMUtil;
import com.sas.solutions.brm.util.ConfigFile;


public class BRMAgent extends UnicastRemoteObject
    implements BRMAgentInterface
{
    public static final long serialVersionUID = 5l;
    private static final Logger LOG = Logger.getLogger(BRMAgent.class);
    private static transient final Logger PROGRESS = Logger.getLogger(
            "progress.agent.brm");
    private static final Logger REMOTE_ANT_LOG = Logger.getLogger(
            "remote.agentAnt");
    
    private int port = 0;
    private boolean shutdownAllowed = false;
    private Registry registry = null;

    private Properties installProps = null;
    private Properties agentConfig = null;
    private ConfigFile controlConfig = null;

    private String machineName = "";
    private String component = "";

    String SSCMountPoint = ""; // i.e. C:/SAS
    String SSC ="";   // i.e. SASSolutionsConfig
    String level = "";    // i.e. Lev1
    String applicationDirectory = "";   // i.e. SASMain
    // So full path to main application directory is (for example):
    // C:/SAS/SASSolutionsConfig/Lev1/SASMain

    /**
     * @throws AccessException
     * @throws RemoteException
     * @throws AlreadyBoundException
     */
    public BRMAgent() throws RemoteException,BRMConfigurationException
    {
        this(Constants.AGENT_DEFAULT_PORT);
    }
    
    /**
     * Construct the BRM Agent. Call init() and then listen() to run rmi 
     * server.
     * 
     * @param port
     * @throws AccessException
     * @throws RemoteException
     * @throws AlreadyBoundException
     */
    public BRMAgent(int port) throws RemoteException,BRMConfigurationException
    {
        this.port = port;
    }

    //////////////////////////////////////////////////////////////////////
    // Remote published interface

    // Called once before any other remote methods are called
    public void setup(String machineName, ConfigFile controlConfig)
        throws BRMUserError, BRMInternalError, RemoteException
    {
        this.machineName = machineName;
        this.controlConfig = controlConfig;
        this.component = null;
        renameThread();
        this.installProps = BRMUtil.loadInstallPropertyFiles();
        getInstallDirsFromProps();
    }

    // Called to enable remote logging
    public void setRemoteLogger(RemoteLoggingInterface remoteLogger)
        throws RemoteException
    {
        RemoteLoggingClient.setLogServer(remoteLogger);
    }

    // Called when config settings change on the controller (mainly after
    // discovered stored process directories)
    public void updateControlConfig(ConfigFile controlConfig)
        throws RemoteException
    {
        this.controlConfig = controlConfig;
    }
    
    // Called once before backup() is called for this agent
    public void backupConfigFiles()
        throws BRMUserError, BRMInternalError, RemoteException
    {
        String backupLoc = getBackupLocation();
        File configBackupDir = new File(backupLoc, "config-" +
                this.machineName);
        
        String SSCdir = BRMUtil.getSSCDir();
        REMOTE_ANT_LOG.debug("SSCdir: "+SSCdir);
        for (int i=0; i<Constants.INSTALL_PROPERTIES_FILES.length; ++i)
        {
            File configFile = new File(SSCdir, 
                    Constants.INSTALL_PROPERTIES_FILES[i]);
            if (!configFile.isFile())
            {
                continue;
            }
            backupConfigFile(configFile.getPath(), configBackupDir);
        }
        for (int i=0; i<Constants.CONFIG_FILES_WARNIFNOTTHERE.length; ++i)
        {
            backupConfigFile(Constants.CONFIG_FILES_WARNIFNOTTHERE[i], 
                    configBackupDir);
        }
        for (int i=0; i<Constants.CONFIG_FILES.length; ++i)
        {
            backupConfigFile(Constants.CONFIG_FILES[i],
                    configBackupDir, true);
        }
    }

    // Called before each component is started/stopped/backed up/restored
    public void setComponent(String component)
        throws BRMUserError, BRMInternalError, RemoteException
    {
        this.component = component;
        renameThread();
    }

    // Called on the Agent running on the metadata server before backup
    // tasks are run
    public Collection discoverSASLibraryDirs()
        throws BRMUserError, BRMInternalError, RemoteException
    {
        REMOTE_ANT_LOG.debug("discoverSASLibraryDirs");
        OMRConnection omr = BRMUtil.getNewOMRConnection(this.controlConfig,
                this.installProps);
        try
        {
            return StoredProcDirFinder.findStoredProcDirs(omr, 
                    this.installProps);
        }
        finally
        {
            omr.close();
        }
    }

    public boolean backup()
        throws BRMUserError, BRMInternalError, RemoteException
    {
        Project p = getAntProject();
        boolean result = executeTarget("backup_" + this.component, p);
        if (!result)
            return result;

        if (Constants.STORED_PROCESS_COMPONENT.equals(this.component)
            && BRMUtil.isTrueStr(
                this.controlConfig.get("backup.sas.library.dirs")))
        {
            result = backupSASLibraryDirs(p);
        }

        return result;
    }

    public boolean restore()
        throws BRMUserError, BRMInternalError, RemoteException
    {
        Project p = getAntProject();
        boolean result = executeTarget("restore_prepare_" + this.component,
                p, true);
        if (!result)
            return result;

        result = executeTarget("restore_" + this.component, p);

        if (result)
        {
            // case: the restore worked, now clean up
            if (!executeTarget("restore_clean_" + this.component, p, true))
            {
                REMOTE_ANT_LOG.warn("Restore worked, but couldn't clean up.");
            }
        }
        else
        {
            // case: restore failed, attempt recovery
            executeTarget("restore_recover_" + this.component, p, true);
        }
 
        return result;
    }

    public boolean startServices() 
        throws BRMUserError, BRMInternalError, RemoteException
    {
        return executeTarget("start_" + this.component, getAntProject());
    }
    
    public boolean stopServices() 
        throws BRMUserError, BRMInternalError, RemoteException
    {
        return executeTarget("stop_" + this.component, getAntProject());
    }

    public boolean executeCommand(String command)
    {
        if (!command.endsWith("_command"))
        {
            command += "_command";
        }
        return executeTarget(command, getAntProject());
    }

    public void teardown()
        throws AccessException, RemoteException, NotBoundException
    {
        if (this.shutdownAllowed)
        {
            Registry registry = LocateRegistry.getRegistry(this.port);
            registry.unbind (BRMAgentInterface.RMICLIENT_REGISTRY_NAME);
            REMOTE_ANT_LOG.debug("Teardown");
            UnicastRemoteObject.unexportObject(this, true);
            //UnicastRemoteObject.unexportObject(registry,true);
        }
        else
        {
            REMOTE_ANT_LOG.info(
              "Agent not started with --allow-shutdown, *not* shutting down");
        }
    }

    //////////////////////////////////////////////////////////////////////
    // Non-remote public methods

    public void init() throws BRMConfigurationException
    {
        this.agentConfig = new Properties();
        BRMUtil.loadPropFileAsResource(Constants.AGENT_CONFIG_FILE,
                this.agentConfig);
    }

    public void listen()
    throws AccessException, RemoteException, AlreadyBoundException,
                      BRMConfigurationException
    {
        int port = this.port;
        LOG.debug("BRMAgent.listen: RMI port was previously set to: " + port);
        if (port == 0)
        {
            String s = this.agentConfig.getProperty(
                    Constants.AGENT_PORT_PROPERTY);
            if (s != null)
                port = Integer.parseInt(s);
            if (port == 0)
                throw new BRMConfigurationException(
                        "No " + Constants.AGENT_PORT_PROPERTY + 
                        " property specified");
        }
        LOG.debug("RMI port = " + port);
        this.registry = LocateRegistry.createRegistry(port);
        try
        {
            PROGRESS.info(
                    "BRMAgent is listening for requests on port " + port);
            registry.bind(RMICLIENT_REGISTRY_NAME, this);
        }
        catch (AlreadyBoundException e)
        {
            LOG.warn("A BRMAgent was already bound. Using that one.");
        }
    }

    //////////////////////////////////////////////////////////////////////
    // Private utility methods

    private boolean executeTarget(String targetName, Project p)
    {
        return executeTarget(targetName, p, false);
    }

    private boolean executeTarget(String targetName, Project p, 
            boolean okIfNotDefined)
    {
        Hashtable targetTable = p.getTargets();
        if (targetTable.get(targetName) == null)
        {
            if (okIfNotDefined)
            {
                REMOTE_ANT_LOG.debug("Skipping Ant target \"" + 
                        targetName + "\"");
                return true;
            }
            else
            {
                REMOTE_ANT_LOG.error("Ant target \"" + 
                        targetName + "\" is not defined");
                return false;
            }
        }
        try
        {
            p.executeTarget(targetName);
        }
        catch (Exception e)
        {
            REMOTE_ANT_LOG.error("Error executing target: " + targetName, e);
            return false;
        }
        return true;
    }
    
    private void loadPropertiesIntoProject(Project proj, Properties props)
    {
        if(proj == null)
        {
            REMOTE_ANT_LOG.error("Project was null.");
            return;
        }
        if(props == null)
        {
            REMOTE_ANT_LOG.error("Properties were null.");
            return;
        }
        
        for (Iterator i = props.keySet().iterator(); i.hasNext(); )
        {
            String key = (String)i.next();
            if (key != null && props.get(key) != null && 
                    !props.get(key).equals(""))
            {
                proj.setProperty(key, props.get(key).toString());
            }
        }
    }
    
    private void loadConfigIntoProject(Project proj, ConfigFile config)
    {
        for (Iterator i=config.orderedKeyIterator(); i.hasNext(); )
        {
            String key = (String)i.next();
            proj.setProperty(key, config.get(key));
        }
    }
    
    private void renameThread()
    {
        String newThreadName = "";
        if (this.component != null)
            newThreadName = this.machineName + "[" + this.component + "]";
        else
            newThreadName = this.machineName;
        Thread.currentThread().setName(newThreadName);
    }

    private void backupConfigFile(String resource, File destinationDir)
    {
        backupConfigFile(resource, destinationDir, false);
    }
    
    private void backupConfigFile(String resource, File destinationDir,
            boolean okIfNotFound)
    {
        try
        {
            // logic here:
            // is it a file? if so, use that path.
            String fromString = "";
            if (new File(resource).exists())
            {
                fromString = resource;
            }
            else
            {
                // it isn't a relative or absolute path. Is it on the classpath?
                URL url = ClassLoader.getSystemResource(resource); 
                // if it is on the classpath, the URL won't be null or blank
                if (url!=null && !url.equals(""))
                {
                    fromString = url.getFile();
                }
                else
                {
                    // Can't find it.
                    throw new FileNotFoundException(resource +
                            " wasn't relative, absolute, or on the classpath");
                }
            }
            REMOTE_ANT_LOG.debug("Backing up config file: " + resource);
            File fromFile = new File(fromString);
            if (!destinationDir.isDirectory())
            {
                destinationDir.mkdirs();
            }
            File toFile = new File(destinationDir, fromFile.getName());
            InputStream in = new FileInputStream(fromFile);
            OutputStream out = new FileOutputStream(toFile);
            byte[] buf = new byte[1024];
            int len;
            while ((len = in.read(buf)) > 0) {
                out.write(buf, 0, len);
            }
            in.close();
            out.close();
        }
        catch (FileNotFoundException e)
        {
            if (okIfNotFound)
            {
                REMOTE_ANT_LOG.debug(
                        "Config file \"" + resource + "\" not found");
            }
            else
            {
                REMOTE_ANT_LOG.warn(
                        "Could not copy config file " + resource + ": " +
                        e.getMessage());
            }
        }
        catch (IOException e)
        {
            REMOTE_ANT_LOG.warn(
                    "Could not copy config file (I/O Exception): " +
                    resource, e);
        }
    }

    private boolean backupSASLibraryDirs(Project project)
    {
        String storedProcDirs = this.controlConfig.get(
                Constants.STORED_PROC_DIRS_PROP);
        if (null == storedProcDirs || "".equals(storedProcDirs))
        {
            REMOTE_ANT_LOG.info(
                "After filtering out the default data directories, there are " +
                "no additional stored process/data directories to back up.");
            return true;
        }
        String[] paths = storedProcDirs.split(";");
        
        // ${SSCMountPoint}/${SSC}/${level}/Data
        String commonPrefix = project.getProperty("SSCMountPoint") + 
            project.getProperty("SSC") +
            project.getProperty("level") + 
            "Data";
        commonPrefix = BRMUtil.normPath(commonPrefix);
        
        for (int i=0; i<paths.length; i++)
        {
            if (BRMUtil.normPath(paths[i]).startsWith(commonPrefix))
            {
                REMOTE_ANT_LOG.debug(
                        "Skipping path because it is already backed up: " +
                        paths[i]);
                continue;
            }
            String path = paths[i];
            REMOTE_ANT_LOG.debug("External path: "+path);
            if (!(new File(path).isFile()||new File(path).isDirectory()))
            {
                // The path wasn't on this machine
                REMOTE_ANT_LOG.debug(
                        "File or directory didn't exist on this machine:" +
                        path);
                continue;
            }
            REMOTE_ANT_LOG.debug("Backing up external path: " + path);
            MacroInstance tagz = (MacroInstance)project.createTask(
                    "tarAndGzip");
            String backupLocation = getBackupLocation();
            File storedProcBackupDir = new File(backupLocation,
                    Constants.STORED_PROCESS_COMPONENT);
            String fusedPath = "root-" + BRMUtil.fusePath(path);
            REMOTE_ANT_LOG.debug("fusedPath: " + fusedPath);
            File destFile = new File(storedProcBackupDir, fusedPath);
            REMOTE_ANT_LOG.debug("... to archive file: " + destFile.getPath());
            tagz.setDynamicAttribute("dest", destFile.getPath());
            tagz.setDynamicAttribute("base", path);
            tagz.execute();
        }
        return true;
    }

    private Project getAntProject()
    {
        // Create ant Project from ant XML file
        ////////////////////////////////////////////////////
        Project ant = new Project();
        ant.init();

        // Set up the project helper
        ProjectHelper2 helper = new ProjectHelper2();
        ant.addReference("ant.projectHelper", helper);

        // Create the build listener
        BRMBuildListener logger = new BRMBuildListener(this.component);
        ant.addBuildListener(logger);

        // Load the agent Ant command file
        URL uri = ClassLoader.getSystemResource(Constants.AGENT_COMMAND_FILE);
        if (uri == null)
        {
            throw new BRMConfigurationException("Missing config file: " +
                    Constants.AGENT_COMMAND_FILE);
        }
        File agentAnt = new File(uri.getFile());
        helper.parse(ant, agentAnt);
        
        // Setup Properties in Project
        ////////////////////////////////////////////////////
        loadConfigIntoProject(ant, this.controlConfig);
        loadPropertiesIntoProject(ant, this.installProps);
        loadPropertiesIntoProject(ant, this.agentConfig);

        ant.setProperty("BRM_SOLUTIONS_CONFIG_DIR", BRMUtil.getSSCDir());
        ant.setProperty("SSCMountPoint", this.SSCMountPoint);
        ant.setProperty("SSC", this.SSC);
        ant.setProperty("level", this.level);
        ant.setProperty("applicationDirectory", this.applicationDirectory);

        return ant;
    }

    private void getInstallDirsFromProps() throws BRMConfigurationException
    {
        this.SSCMountPoint = "UNKNOWN";
        this.SSC = "UNKNOWN";
        this.level = "UNKNOWN";
        this.applicationDirectory = "UNKNOWN";

        String dirPropName = "SOLUTIONS_SAS_SERVER_DIR";
        String baseDirProp = this.installProps.getProperty(dirPropName);
        if (baseDirProp == null)
        {
            // The SOLUTIONS_SAS_SERVER_DIR property is not set, indicating
            // that we are on a MySQL-only machine (no soltnsdata.properties
            // nor soltnsmid.properties, just mysqldbs.properties)
            dirPropName = "SOLUTIONS_SAS_UTILITIES_DIR";
            baseDirProp = this.installProps.getProperty(dirPropName);
        }
        if (baseDirProp == null)
        {
            String msg = "Could not find key install properties: " + 
                    "SOLUTIONS_SAS_SERVER_DIR, SOLUTIONS_SAS_UTILITIES_DIR";
            REMOTE_ANT_LOG.error(msg);
            throw new BRMConfigurationException(msg);
        }
        REMOTE_ANT_LOG.debug("property for examination: " + dirPropName + "=" + 
                baseDirProp);

        baseDirProp = BRMUtil.normPath(baseDirProp);
        String normSSCDir = BRMUtil.normPath(BRMUtil.getSSCDir());
        REMOTE_ANT_LOG.debug("normalized " + dirPropName + "=" +
                baseDirProp);
        REMOTE_ANT_LOG.debug("normalized Solutions Config Dir=" + normSSCDir);

        List parts = BRMUtil.splitPath(baseDirProp);
        List sscParts = BRMUtil.splitPath(normSSCDir);

        if (parts.size() == 0)
        {
            throw new BRMConfigurationException(
                    "Missing SOLUTIONS_SAS_SERVER_DIR or SOLUTIONS_SAS_UTILITIES_DIR");
        }
        if (sscParts.size() == 0)
        {
            throw new BRMConfigurationException(
                    "Missing Solutions Config Dir path");
        }

        this.SSCMountPoint = BRMUtil.joinPath(sscParts.subList(0,
                    sscParts.size() - 1));
        this.SSC = (String)sscParts.get(sscParts.size() - 1);

        if (BRMUtil.compareListPrefixes(parts, sscParts)==sscParts.size())
        {
            // Good - the Solutions configuration directory is a proper
            // prefix of the server/utilities directory.
            if (parts.size() > sscParts.size())
            {
                this.level = (String)parts.get(sscParts.size());
            }
            else
            {
                this.level = "";
            }
            if (parts.size() > (sscParts.size() + 1) &&
                    dirPropName.equals("SOLUTIONS_SAS_SERVER_DIR"))
            {
                this.applicationDirectory = BRMUtil.joinPath(
                    parts.subList(sscParts.size() + 1, parts.size()));
            }
            else
            {
                this.applicationDirectory = "";
            }
        }
        else
        {
            // Bad - the Solutions configuration directory as determined by
            // BRM has no relation to the values in the install config
            // files. So we have to do a best effort guess, working from the
            // back. Fortunately, this case will be rare or nonexistent.
            if (dirPropName.equals("SOLUTIONS_SAS_SERVER_DIR"))
            {
                REMOTE_ANT_LOG.warn("Solutions config directory does not match SOLUTIONS_SAS_SERVER_DIR");
                if (parts.size() > 3)
                {
                    this.SSCMountPoint = BRMUtil.joinPath(parts.subList(0,
                            parts.size() - 3));
                    this.SSC = (String)parts.get(parts.size() - 3);
                    this.level = (String)parts.get(parts.size() - 2);
                    this.applicationDirectory = (String)parts.get(
                            parts.size() - 1);
                }
                else
                {
                    this.SSCMountPoint = (String)parts.get(0);
                    if (parts.size() > 1)
                        this.SSC = (String)parts.get(1);
                    else
                        this.SSC = "";
                    if (parts.size() > 2)
                        this.level = (String)parts.get(2);
                    else
                        this.level = "";
                    this.applicationDirectory = "";
                }
            }
            else // dirPropName.equals("SOLUTIONS_SAS_UTILITIES_DIR")
            {
                // We can't possibly guess the level directory
                // But it may be all right because we may not need it on
                // this tier.
                REMOTE_ANT_LOG.warn("SOLUTIONS_SAS_SERVER_DIR property not available, can't determine Level directory");
            }
        }

        REMOTE_ANT_LOG.debug("SSCMountPoint = " + this.SSCMountPoint);
        REMOTE_ANT_LOG.debug("SSC = " + this.SSC);
        REMOTE_ANT_LOG.debug("level = " + this.level);
        REMOTE_ANT_LOG.debug("applicationDirectory = " + this.applicationDirectory);

        REMOTE_ANT_LOG.info("Application Dir: " +
                this.SSCMountPoint +
                "/" + this.SSC +
                "/" + this.level +
                "/" + this.applicationDirectory);
    }

    private String getBackupLocation()
        throws BRMConfigurationException
    {
        String backupLoc = this.controlConfig.get("backup.location");
        if (backupLoc == null || backupLoc.length() == 0)
            throw new BRMConfigurationException(
                    "Missing backup.location property");
        File backupLocFile = new File(backupLoc);
        if (!backupLocFile.isDirectory())
        {
            backupLocFile.mkdirs();
        }
        // Try to fix trailing slash questions.
        return (new File(backupLoc)).getAbsolutePath();
    }
    
    private void parseArguments(String[] args)
    {
        for (int i =0; i<args.length; i++)
        {
            if (args[i].equalsIgnoreCase("-p"))
            {
                this.port = Integer.parseInt(args[i+1]);
                //consuming the arg after -p, so advance one
                i++;
            }
            else if (args[i].equalsIgnoreCase("--allow-shutdown"))
            {
                this.shutdownAllowed = true;
            }
            else
            {
                System.err.println("Unknown option: " + args[i]);
            }
        }
    }

    private OMRConnection getNewOMRConnection() 
        throws BRMConfigurationException
    {
        return BRMUtil.getNewOMRConnection(this.controlConfig, 
                this.installProps);
    }

    //////////////////////////////////////////////////////////////////////
    // Main method

    public static void main(String[] args)
    throws Exception
    {
        Thread.currentThread().setName("agent");
        args = BRMUtil.filterCmdLineProps(args);
        BRMAgent agent = new BRMAgent();
        agent.parseArguments(args);
        agent.init();
        agent.listen();
    }
    
}
