/*
 * Created on Apr 28, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm;

import java.io.IOException;
import java.rmi.ConnectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.errors.BRMError;
import com.sas.solutions.brm.errors.BRMInternalError;
import com.sas.solutions.brm.errors.BRMUserError;
import com.sas.solutions.brm.errors.DefaultConfigUsedException;
import com.sas.solutions.brm.logging.RemoteLoggingInterface;
import com.sas.solutions.brm.logging.RemoteLoggingServer;
import com.sas.solutions.brm.messages.Messages;
import com.sas.solutions.brm.migrate.MainMigrator;
import com.sas.solutions.brm.util.BRMUtil;
import com.sas.solutions.brm.util.ConfigFile;

import com.sas.solutions.brm.omr.OMRConnection;

public class BRMController
{
    private static transient final Logger LOG = Logger.getLogger(
            BRMController.class);
    private static transient final Logger PROGRESS = Logger.getLogger(
            "progress.controller.brm");
    
    private ConfigFile controlConfig = null;
    private Properties agentConfigProps = null;

    // List of components to process, in order
    List componentList = null;
    
    // Map of host name to task runner (which contains an rmi connection)
    Map hostToTaskRunnerMap = null;

    // Callback object that enables remote logging (exported to rmi)
    RemoteLoggingInterface logServer = null;

    // Map of host name to agent interface (rmi connection to remote agent)
    Map hostToAgentMap = null;
    
    public BRMController() throws BRMConfigurationException, IOException
    {
        super();
        readProperties();
    }

    ///////////////////////////////////////////////////////////////////////
    // Public interface
    
    public void doBackup()
        throws BRMConfigurationException
    {
        PROGRESS.info("Starting backup");
        LOG.info("Starting backup");
        LOG.info("backup.location: " +
                this.controlConfig.get("backup.location"));
        ensureRequiredProperties();

        // Try to fetch the stored process locations.
        if (BRMUtil.isTrueStr(
                    this.controlConfig.get("backup.sas.library.dirs")))
            discoverSASLibraryDirs();
        
        runTasks(new BackupTaskRunnerFactory(this));

        LOG.info("Finished backup");
        PROGRESS.info("Finished backup");
    }

    private void discoverSASLibraryDirs()
        throws BRMConfigurationException
    {
        String host = this.controlConfig.get(Constants.METADATA_COMPONENT);
        if (host == null || host.trim().length() == 0)
        {
            LOG.warn("Can't discover Stored Process directories because " +
                    Constants.METADATA_COMPONENT + 
                    " is not set in the config file.");
            return;
        }
        BRMAgentInterface agent = getRemoteAgent(host);
        Collection storedProcs = null;
        try
        {
            storedProcs = agent.discoverSASLibraryDirs();
        }
        catch (BRMError e)
        {
            throw new BRMConfigurationException(
                    "Could not discover Stored Process directories", e);
        }
        catch (RemoteException e)
        {
            throw new BRMConfigurationException(
                    "Could not discover Stored Process directories", e);
        }
        StringBuffer sb = new StringBuffer();
        for (Iterator i = storedProcs.iterator(); i.hasNext(); )
        {
            Object current = i.next();
            LOG.info("Found stored proc path in metadata: "+current);
            sb.append(current.toString());
            // Use a semicolon as a separator b/c you can have file:///blah
            sb.append(";");
        }
        LOG.debug("Setting " + Constants.STORED_PROC_DIRS_PROP + "=" +
                sb.toString());
        this.controlConfig.set(
                Constants.STORED_PROC_DIRS_PROP,
                sb.toString());
    }
    
    public void doRestore()
        throws BRMConfigurationException
    {
        PROGRESS.info("Starting restore");
        LOG.info("Starting restore");
        ensureRequiredProperties();

        runTasks(new RestoreTaskRunnerFactory(this));

        LOG.info("Finished restore");
        PROGRESS.info("Finished restore");
    }
    
    public void doMigrate() throws BRMConfigurationException
    {
        OMRConnection omr = getNewOMRConnection();
        try
        {
            migrate(omr);
        }
        finally
        {
            omr.close();
        }
    }
    
    public void doReadOnlyMigrate() throws BRMConfigurationException
    {
        OMRConnection omr = getNewOMRConnection();
        LOG.info("Setting OMR connection to read-only");
        omr.setReadOnly(true);
        try
        {
            migrate(omr);
        }
        finally
        {
            omr.close();
        }
    }
    
    public void doDumpOMR() throws Exception
    {
        OMRConnection omr = getNewOMRConnection();
        try
        {
            omr.dump(System.out);
        }
        finally
        {
            omr.close();
        }
    }

    /**
      * Clean up and disconnect from all resources used by the Controller.
      * This must be run before exiting the program, to avoid hanging on to
      * any open RMI connections or other resources.
      */
    public void close()
    {
        closeTaskRunners();
        if (this.logServer != null)
        {
            try
            {
                java.rmi.server.UnicastRemoteObject.unexportObject(
                        this.logServer, 
                        true);
            }
            catch (Throwable t)
            {
                t.printStackTrace();
            }
            this.logServer = null;
        }
        if (this.hostToAgentMap != null)
        {
            for (Iterator
                    agentIter=this.hostToAgentMap.values().iterator();
                    agentIter.hasNext(); )
            {
                BRMAgentInterface agent = (BRMAgentInterface)agentIter.next();
                try
                {
                    agent.teardown();
                }
                catch (Throwable t)
                {
                    LOG.error("Error shutting down agent interface", t);
                }
            }
            this.hostToAgentMap = null;
        }
    }

    ///////////////////////////////////////////////////////////////////////
    // Public "getter" methods

    public int getAgentPort()
    {
        String s = this.agentConfigProps.getProperty(
                Constants.AGENT_PORT_PROPERTY);
        if (s == null || s.length() == 0)
        {
            return Constants.AGENT_DEFAULT_PORT;
        }
        return Integer.parseInt(s);
    }

    /**
      * Return the Controller configuration properties. Please don't change
      * them!
      */
    public ConfigFile getConfig()
    {
        return this.controlConfig;
    }

    public String getConfigProp(String propName)
    {
        return this.controlConfig.get(propName);
    }

    public RemoteLoggingInterface getLogServer()
    {
        if (this.logServer == null)
        {
            this.logServer = new RemoteLoggingServer();
            try
            {
                java.rmi.server.UnicastRemoteObject.exportObject(
                        this.logServer);
            }
            catch (Exception e)
            {
                throw new Error(e);
            }
        }
        return this.logServer;
    }

    ///////////////////////////////////////////////////////////////////////
    // Helper methods
    
    private void runTasks(TaskRunnerFactory factory)
    {
        PROGRESS.info("Running tasks...");
        List components = new ArrayList();
        components.addAll(this.componentList);
        String component;
        Iterator componentIter;
        for (componentIter=components.iterator(); componentIter.hasNext(); )
        {
            component = (String)componentIter.next();
            PROGRESS.info(component);
            String host = this.controlConfig.get(component);
            TaskRunner tr = getTaskRunnerForHost(host, factory);
            tr.stop(component);
            tr.runTask(component);
        }
        PROGRESS.info("Restarting...");
        Collections.reverse(components);
        for (componentIter=components.iterator(); componentIter.hasNext(); )
        {
            component = (String)componentIter.next();
            PROGRESS.info(component);
            String host = this.controlConfig.get(component);
            TaskRunner tr = getTaskRunnerForHost(host, factory);
            tr.start(component);
        }
        closeTaskRunners();
    }

    private TaskRunner getTaskRunnerForHost(String host, 
            TaskRunnerFactory factory)
    {
        String lcHost = host.toLowerCase();
        if (this.hostToTaskRunnerMap == null)
            this.hostToTaskRunnerMap = new HashMap();
        TaskRunner tr = (TaskRunner)this.hostToTaskRunnerMap.get(lcHost);
        if (tr != null)
            return tr;
        tr = factory.create(host);
        this.hostToTaskRunnerMap.put(lcHost, tr);
        return tr;
    }
    
    private void closeTaskRunners()
    {
        if (this.hostToTaskRunnerMap == null)
            return;
        Collection taskRunners = this.hostToTaskRunnerMap.values();
        for (Iterator taskRunnerIter=taskRunners.iterator(); 
            taskRunnerIter.hasNext();)
        {
            TaskRunner tr = (TaskRunner)taskRunnerIter.next();
            tr.close();
        }
        this.hostToTaskRunnerMap = null;
    }

    private OMRConnection getNewOMRConnection() 
        throws BRMConfigurationException
    {
        Properties installProps = BRMUtil.loadInstallPropertyFiles();
        return BRMUtil.getNewOMRConnection(this.controlConfig, installProps);
    }

    private void ensureRequiredProperties()
        throws BRMConfigurationException
    {
        ensureProperty("backup.location");
    }
    
    private void ensureProperty(String property) 
        throws BRMConfigurationException
    {
        String value = this.controlConfig.get(property);
        if (value==null || value.equals(""))
        {
            throw new BRMConfigurationException(
                    Messages.getString("missing.property")+" "+property);
        }
    }
    
    private void readProperties() 
        throws BRMConfigurationException, IOException
    {
        LOG.debug("Reading property files");
        this.controlConfig = new ConfigFile();
        this.controlConfig.loadResource(Constants.BRM_CONFIG_FILE);
        this.agentConfigProps = new Properties();
        BRMUtil.loadPropFileAsResource(Constants.AGENT_CONFIG_FILE,
                this.agentConfigProps);

        // Set up the list of components to process, in the correct order
        // (the order that they originally appear in the config file, except
        // that the metadata server always comes last.)
        this.componentList = new ArrayList();
        boolean includeMetadata = false;
        for (Iterator keyIter=this.controlConfig.orderedKeyIterator();
                keyIter.hasNext(); )
        {
            String key = keyIter.next().toString();
            key = key.toLowerCase();
            if (!key.endsWith("_component"))
                continue;
            String component = key;
            String hostname = this.controlConfig.get(key);
            hostname = hostname.toLowerCase();
            if (hostname.equals("default.host.com"))
            {
                throw new DefaultConfigUsedException(
                        "Detected \"default.host.com\"... " +
                        "Did you edit your cmdLine.properties file?");
            }
            // This ".component.ignore" stuff is purely for backwards
            // compatibility with old config files. Going forward I 
            // recommend commenting out unwanted _component lines with "#".
            String ignoreComponent = this.controlConfig.get(
                    component + ".component.ignore");
            if (ignoreComponent != null && ignoreComponent.equals("true"))
                continue;
            if (component.equals(Constants.METADATA_COMPONENT))
            {
                includeMetadata = true;
            }
            else
            {
                this.componentList.add(component);
            }
        }
        if (includeMetadata)
            this.componentList.add(Constants.METADATA_COMPONENT);
    }

    private void setConfigProp(String key, String value)
    {
        this.controlConfig.set(key, value);
    }

    private static void migrate(OMRConnection omr)
        throws BRMConfigurationException
    {
        MainMigrator mm = new MainMigrator(omr);
        mm.migrate();
    }

    public boolean commandDefined(String command)
    {
        if (!command.endsWith("_command"))
        {
            command += "_command";
        }
        return this.controlConfig.containsKey(command);
    }

    public void doCommand(String command)
    {
        LOG.debug("doCommand(" + command + ")");
        if (!command.endsWith("_command"))
        {
            command += "_command";
        }
        String hostSpec = this.controlConfig.get(command);
        if (hostSpec == null)
        {
            throw new Error("Command not defined: " + command);
        }
        List hostList = getHostList(hostSpec);
        LOG.debug("Running command on hostList=" + hostList);
        for (Iterator hostIter=getHostList(hostSpec).iterator();
                hostIter.hasNext(); )
        {
            String host = (String)hostIter.next();
            BRMAgentInterface agent = getRemoteAgent(host);
            try
            {
                agent.executeCommand(command);
            }
            catch (RemoteException e)
            {
                throw new BRMError(e);
            }
        }
    }

    private List getHostList(String hostSpec)
    {
        hostSpec = hostSpec.trim();
        hostSpec = hostSpec.toLowerCase();
        List result = new ArrayList();
        if ("*".equals(hostSpec))
        {
            // Gather all of the hostnames for components
            Set resultSet = new HashSet();
            for (Iterator keyIter=this.controlConfig.orderedKeyIterator();
                    keyIter.hasNext(); )
            {
                String key = (String)keyIter.next();
                key = key.toLowerCase();
                if (key.endsWith("_component"))
                {
                    String host = this.controlConfig.get(key);
                    String hostlc = host.toLowerCase();
                    if (!resultSet.contains(hostlc))
                    {
                        result.add(host);
                        resultSet.add(hostlc);
                    }
                }
            }
        }
        else
        {
            // Assume hostSpec is a single host name
            result.add(hostSpec);
        }
        return result;
    }

    private List getCommandList()
    {
        List result = new ArrayList();
        for (Iterator keyIter=this.controlConfig.orderedKeyIterator();
                keyIter.hasNext(); )
        {
            String key = (String)keyIter.next();
            key = key.toLowerCase();
            if (key.endsWith("_command"))
            {
                result.add(key);
            }
        }
        return result;
    }

    public BRMAgentInterface getRemoteAgent(String host)
    {
        host = host.toLowerCase();
        host = host.trim();
        if (this.hostToAgentMap == null)
            this.hostToAgentMap = new HashMap();
        BRMAgentInterface agent = 
            (BRMAgentInterface)this.hostToAgentMap.get(host);
        if (agent != null)
        {
            // Make sure the Agent has the latest config values
            try
            {
                agent.updateControlConfig(this.controlConfig);
            }
            catch (RemoteException e)
            {
                throw new BRMError("Error updating config on agent on " + host,
                        e);
            }
            return agent;
        }

        int port = getAgentPort();
        LOG.debug("Connecting to agent at " + host + ":" + port);
        try
        {
            Registry registry = LocateRegistry.getRegistry(host, port);
            agent = (BRMAgentInterface)registry.lookup(
                         BRMAgent.RMICLIENT_REGISTRY_NAME);
        }
        catch (NotBoundException e)
        {
            throw new BRMError("Couldn't find an agent on " + host, e);
        }
        catch (ConnectException e)
        {
            throw new BRMError("Could not connect to agent on " + host + 
                    ".  Is it running?", e);
        }
        catch (RemoteException e)
        {
            throw new BRMError("Error connecting to agent on " + host, e);
        }

        try
        {
            RemoteLoggingInterface rli = getLogServer();
            agent.setRemoteLogger(rli);
            agent.setup(host, getConfig());
        }
        catch (BRMUserError e)
        {
            throw new BRMError(e);
        }
        catch (BRMInternalError e)
        {
            throw new BRMError(e);
        }
        catch(RemoteException e)
        {
            throw new BRMError(e);
        }

        this.hostToAgentMap.put(host, agent);
        return agent;
    }

    public StringBuffer getUsageMessage()
    {
        StringBuffer msg = new StringBuffer();
        msg.append("Usage:\n");
        msg.append("    brmCmdLine command ... [backup.location]\n");
        msg.append("Where command is one of:\n");
        msg.append("    backup\n");
        msg.append("    restore\n");
        msg.append("    migrate\n");
        for (Iterator commandIter=getCommandList().iterator();
                commandIter.hasNext(); )
        {
            String command = (String)commandIter.next();
            command = command.toLowerCase();
            if (command.endsWith("_command"))
            {
                command = command.substring(0, command.length() - 8);
            }
            msg.append("    ");
            msg.append(command);
            msg.append('\n');
        }
        msg.append("and backup.location is optional.\n");
        return msg;
    }

    final static String BACKUP = "backup";
    final static String RESTORE = "restore";
    final static String MIGRATE = "migrate";
    final static String READONLYMIGRATE = "readonlymigrate";
    final static String DUMPOMR = "dumpomr";
    public final static String[] STATIC_COMMAND_LIST = {
        BACKUP, RESTORE, MIGRATE, READONLYMIGRATE, DUMPOMR,
        };

    public void run(String[] commandArray) throws Exception
    {
        if (commandArray.length < 1)
        {
            StringBuffer msg = new StringBuffer();
            msg.append("Command not specified\n");
            msg.append(getUsageMessage());
            throw new BRMUserError(msg.toString());
        }

        // Check the last parameter on the line.  If the last parameter
        // is not a command name, treat it as a backup directory.
        int commandCount = commandArray.length;
        if (commandCount >= 2)
        {
            String lastArg = commandArray[commandArray.length - 1];
            lastArg = lastArg.trim();
            lastArg = lastArg.toLowerCase();
            if (
              ((Arrays.asList(STATIC_COMMAND_LIST).indexOf(lastArg) < 0)
                && !commandDefined(lastArg)))
            {
                String backupLocation = commandArray[commandArray.length - 1];
                commandCount -= 1;
                LOG.info("Setting backup.location from command-line: " + 
                        backupLocation);
                setConfigProp("backup.location", backupLocation);
            }
        }

        for (int i=0; i<commandCount; ++i)
        {
            String command = commandArray[i];
            command = command.trim();
            command = command.toLowerCase();
            LOG.debug("About to execute command: " + command);
            PROGRESS.info("Running command: " + command);
            
            if (BACKUP.equals(command))
            {
                doBackup();
            }
            else if (RESTORE.equals(command))
            {
                doRestore();
            }
            else if (MIGRATE.equals(command))
            {
                doMigrate();
            }
            else if (READONLYMIGRATE.equals(command))
            {
                doReadOnlyMigrate();
            }
            else if (DUMPOMR.equals(command))
            {
                doDumpOMR();
            }
            else
            {
                if (!commandDefined(command))
                {
                    StringBuffer msg = new StringBuffer();
                    msg.append("Unknown command: ");
                    msg.append(command);
                    msg.append('\n');
                    msg.append(getUsageMessage());
                    throw new BRMUserError(msg.toString());
                }
                doCommand(command);
            }
        }
    }

    public static void main(String[] args)
        throws Exception
    {
        Thread.currentThread().setName("controller");
        args = BRMUtil.filterCmdLineProps(args);
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        LOG.info("BRMController starting");
        BRMController controller = new BRMController();
        try
        {
            controller.run(args);
        }
        catch (DefaultConfigUsedException e)
        {
            LOG.error(e.getMessage());
        }
        catch (BRMUserError e)
        {
            LOG.error(e.getMessage());
        }
        catch (Exception e)
        {
            LOG.error("Exception while running BRM Controller", e);
        }
        catch (Error e)
        {
            LOG.error("Error while running BRM Controller", e);
        }
        finally
        {
            controller.close();
            controller = null;
        }
    }
}
