// JExec.java
// Created by David Handy  19 June 2007
// Copyright (c) 2007 by SAS Institute Inc., Cary, NC 27513
// Contains code derived from Ant 1.6.5 -- see LICENSE.apache.ant.txt and NOTICE

package com.sas.solutions.brm.jexec;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;

import org.apache.log4j.Logger;

/**
 * Used by <code>Execute</code> to handle input and output stream of
 * subprocesses.
 *
 * @since Ant 1.2
 */
/*public*/ interface ExecuteStreamHandler {

    /**
     * Install a handler for the input stream of the subprocess.
     *
     * @param os output stream to write to the standard input stream of the
     *           subprocess
     */
    void setProcessInputStream(OutputStream os) throws IOException;

    /**
     * Install a handler for the error stream of the subprocess.
     *
     * @param is input stream to read from the error stream from the subprocess
     */
    void setProcessErrorStream(InputStream is) throws IOException;

    /**
     * Install a handler for the output stream of the subprocess.
     *
     * @param is input stream to read from the error stream from the subprocess
     */
    void setProcessOutputStream(InputStream is) throws IOException;

    /**
     * Start handling of the streams.
     */
    void start() throws IOException;

    /**
     * Stop handling of the streams - will not be restarted.
     */
    void stop();
}

public class JExec
{
    private static final Logger LOG = Logger.getLogger(JExec.class);

    public static void main(String[] args) throws Exception
    {
        if (args.length == 0)
        {
            System.out.println("Use: java JExec [cmd] ...");
            LOG.warn("No command given.");
            return;
        }
        StringBuffer sb = new StringBuffer();
        sb.append("running [");
        for (int i=0; i<args.length; ++i)
        {
            if (i > 0)
                sb.append(", ");
            sb.append("\"" + args[i] + "\"");
        }
        sb.append("]");
        LOG.info(sb.toString());
        JExec jexec = new JExec();
        jexec.setCommandline(args);
        int rc = jexec.execute();
        if (rc == 0)
            LOG.info("exit code: " + rc);
        else
            LOG.warn("exit code: " + rc);
    }

    // Code below is scaffolding to adapt copied Ant code to a stand-alone
    // test environment.

    private static class Project
    {
        public static String MSG_DEBUG = "debug";
        public void log(String s1, String s2) { }
    }

    private static class Commandline
    {
        public static String describeCommand(String[] cmdLine)
        {
            return cmdLine.toString();
        }
    }

    private static class Task
    {
        Project project;
        public Task() { project = new Project(); }
        public void log(String s1, String s2) { }
        public Project getProject() { return project; }
    }

    public static class BuildException extends RuntimeException
    {
        public BuildException(String msg)
        {
            super(msg);
        }

        public BuildException(String msg, Throwable t)
        {
            super(msg, t);
        }
    }

    public static class ExecuteWatchdog
    {
        public void start(Process p) { }
        public void stop() { }
        public void checkException() { }
    }

    public static class ProcessDestroyer extends ArrayList { }

    public static Process launch(Project project, String[] command,
                                 String[] env, File dir, boolean useVM)
        throws IOException //, java.lang.NoSuchMethodException
    {
        assert (useVM);
        CommandLauncher launcher = null;
        try
        {
            launcher = new Java13CommandLauncher();
        }
        catch (java.lang.NoSuchMethodException e)
        {
            throw new IOException(e.toString());
        }
        Process process = launcher.exec(
                project,
                command,
                env,
                dir);
        return process;
    }

    Project project = null;
    String[] cmdLine = null;
    String[] getCommandline() { return this.cmdLine; }
    String[] getEnvironment() { return null; }
    File workingDirectory = null;
    boolean useVMLauncher = true;
    private ExecuteStreamHandler streamHandler = new PumpStreamHandler(
            System.out, System.err, System.in);
    private ExecuteWatchdog watchdog = null;
    /** Used to destroy processes when the VM exits. */
    private static ProcessDestroyer processDestroyer = new ProcessDestroyer();
    public static final int INVALID = Integer.MAX_VALUE;
    private int exitValue = INVALID;

    void setCommandline(String[] cmdLine) { this.cmdLine = cmdLine; }
    void setExitValue(int exitValue) { this.exitValue = exitValue; }


    // Code below copied out of Ant 1.6.5 Execute.java

    /**
     * Runs a process defined by the command line and returns its exit status.
     *
     * @return the exit status of the subprocess or <code>INVALID</code>
     * @exception java.io.IOException The exception is thrown, if launching
     *            of the subprocess failed
     */
    public int execute() throws IOException {
        if (workingDirectory != null && !workingDirectory.exists()) {
            throw new BuildException(workingDirectory + " doesn't exist.");
        }
        final Process process = launch(project, getCommandline(),
                                       getEnvironment(), workingDirectory,
                                       useVMLauncher);

        try {
            streamHandler.setProcessInputStream(process.getOutputStream());
            streamHandler.setProcessOutputStream(process.getInputStream());
            streamHandler.setProcessErrorStream(process.getErrorStream());
        } catch (IOException e) {
            process.destroy();
            throw e;
        }
        streamHandler.start();

        try {
            // add the process to the list of those to destroy if the VM exits
            //
            processDestroyer.add(process);

            if (watchdog != null) {
                watchdog.start(process);
            }
            waitFor(process);

            if (watchdog != null) {
                watchdog.stop();
            }
            streamHandler.stop();
            closeStreams(process);

            if (watchdog != null) {
                watchdog.checkException();
            }
            return getExitValue();
        } catch (ThreadDeath t) {
            // #31928: forcibly kill it before continuing.
            process.destroy();
            throw t;
        } finally {
            // remove the process to the list of those to destroy if the VM exits
            //
            processDestroyer.remove(process);
        }
    }

    /**
     * A command launcher for a particular JVM/OS platform.  This class is
     * a general purpose command launcher which can only launch commands in
     * the current working directory.
     */
    private static class CommandLauncher {
        /**
         * Launches the given command in a new process.
         *
         * @param project       The project that the command is part of
         * @param cmd           The command to execute
         * @param env           The environment for the new process.  If null,
         *                      the environment of the current process is used.
         * @throws IOException  if attempting to run a command in a specific directory
         */
        public Process exec(Project project, String[] cmd, String[] env)
             throws IOException {
            if (project != null) {
                project.log("Execute:CommandLauncher: "
                    + Commandline.describeCommand(cmd), Project.MSG_DEBUG);
            }
            return Runtime.getRuntime().exec(cmd, env);
        }

        /**
         * Launches the given command in a new process, in the given working
         * directory.
         *
         * @param project       The project that the command is part of
         * @param cmd           The command to execute
         * @param env           The environment for the new process.  If null,
         *                      the environment of the current process is used.
         * @param workingDir    The directory to start the command in.  If null,
         *                      the current directory is used
         * @throws IOException  if trying to change directory
         */
        public Process exec(Project project, String[] cmd, String[] env,
                            File workingDir) throws IOException {
            if (workingDir == null) {
                return exec(project, cmd, env);
            }
            throw new IOException("Cannot execute a process in different "
                + "directory under this JVM");
        }
    }

    /**
     * A command launcher for JDK/JRE 1.3 (and higher).  Uses the built-in
     * Runtime.exec() command
     */
    private static class Java13CommandLauncher extends CommandLauncher {
        public Java13CommandLauncher() throws NoSuchMethodException {
            // Locate method Runtime.exec(String[] cmdarray,
            //                            String[] envp, File dir)
            myExecWithCWD = Runtime.class.getMethod("exec",
                new Class[] {String[].class, String[].class, File.class});
        }

        /**
         * Launches the given command in a new process, in the given working
         * directory
         * @param project the ant project
         * @param cmd the command line to execute as an array of strings
         * @param env the environment to set as an array of strings
         * @param workingDir the working directory where the command should run
         * @throws IOException probably forwarded from Runtime#exec
         */
        public Process exec(Project project, String[] cmd, String[] env,
                            File workingDir) throws IOException {
            try {
                if (project != null) {
                    project.log("Execute:Java13CommandLauncher: "
                        + Commandline.describeCommand(cmd), Project.MSG_DEBUG);
                }
                Object[] arguments = {cmd, env, workingDir};
                return (Process) myExecWithCWD.invoke(Runtime.getRuntime(),
                                                     arguments);
            } catch (InvocationTargetException exc) {
                Throwable realexc = exc.getTargetException();
                if (realexc instanceof ThreadDeath) {
                    throw (ThreadDeath) realexc;
                } else if (realexc instanceof IOException) {
                    throw (IOException) realexc;
                } else {
                    throw new BuildException("Unable to execute command",
                                             realexc);
                }
            } catch (Exception exc) {
                // IllegalAccess, IllegalArgument, ClassCast
                throw new BuildException("Unable to execute command", exc);
            }
        }

        private Method myExecWithCWD;
    }

    /**
     * Query the exit value of the process.
     * @return the exit value or Execute.INVALID if no exit value has
     * been received
     */
    public int getExitValue() {
        return exitValue;
    }

    /**
     * wait for a given process
     *
     * @param process the process one wants to wait for
     */
    protected void waitFor(Process process) {
        LOG.debug("JExec.waitFor: enter");
        try {
            process.waitFor();
            setExitValue(process.exitValue());
        } catch (InterruptedException e) {
            process.destroy();
        }
        LOG.debug("JExec.waitFor: exit");
    }

    /**
     * Close the streams belonging to the given Process.
     * @param process   the <CODE>Process</CODE>.
     */
    public static void closeStreams(Process process) {
        try {
            process.getInputStream().close();
        } catch (IOException eyeOhEx) {
        }
        try {
            process.getOutputStream().close();
        } catch (IOException eyeOhEx) {
        }
        try {
            process.getErrorStream().close();
        } catch (IOException eyeOhEx) {
        }
    }
}

// end-of-file
