/*
 * Created on May 31, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import org.apache.log4j.Logger;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;

import com.sas.solutions.brm.ClassIds;
import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.errors.BRMInternalError;
import com.sas.solutions.brm.errors.BRMUserError;
import com.sas.solutions.brm.omr.GenericMetadataObject;
import com.sas.solutions.brm.omr.OMRPropertyFileLoader;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.util.BRMUtil;

public class MainMigrator
{
    public static final String OMR = MigrateConfig.OMR;
    public static final String MIDTIER = MigrateConfig.MIDTIER;
    public static final String WORKSPACE = MigrateConfig.WORKSPACE;
//    public static final String STOREDPROC = MigrateConfig.STOREDPROC;
    public static final String SPAWNER = MigrateConfig.SPAWNER;
    public static final String DAV = MigrateConfig.DAV;
    public static final String REMOTESERVICES = MigrateConfig.REMOTESERVICES;
    public static final String ODCS = MigrateConfig.ODCS;
    public static final String MYSQL = MigrateConfig.MYSQL;
    public static final String OLAP = MigrateConfig.OLAP;
    
    OMRConnection omr = null;
    MigrateConfig config = null;
    Map candidateObjCountMap = null;
    private static final Logger LOG = Logger.getLogger(MainMigrator.class);
    
    Map directoryComponentMap = null;

    public MainMigrator(OMRConnection omr) throws BRMConfigurationException
    {
        this.omr = omr;
        this.config = new MigrateConfig();
        this.candidateObjCountMap = new HashMap();
    }

    public MainMigrator(OMRConnection omr, String configFileName) 
        throws BRMConfigurationException
    {
        this.omr = omr;
        this.config = new MigrateConfig(configFileName);
        this.candidateObjCountMap = new HashMap();
    }
    
    static public MigrateAction[] MigrateActions = {
        // Metadata Type, Repository, Search Param, Method Name
        new MigrateAction("AttributeProperty", "Foundation", "", "migrateAttributeProperty"),
        new MigrateAction("Directory", "", "", "migrateDirectory"),
        new MigrateAction("Document", "Foundation", "*[@Name='SAS Setinit Location']", "migrateSetinit"),
        new MigrateAction("Login", "Foundation", "", "migrateLogin"),
        new MigrateAction("TCPIPConnection", "", "", "migrateTCPIPConnection"),
        new MigrateAction("ServerComponent", "Foundation", "", "migrateServerComponent"),
        new MigrateAction("ConfiguredComponent", "Foundation", "", "migrateConfiguredComponent"),
        new MigrateAction("TextStore", "Foundation", "", "migrateTextStore"),
        new MigrateAction("Property", "Foundation", "*[@PropertyName='Command'][AssociatedObject/ServerComponent[@Name?'- Workspace Server']]", "migrateCommandProperty", WORKSPACE),
        new MigrateAction("Property", "Foundation", "*[@PropertyName='Command'][AssociatedObject/ServerComponent[@Name?'- Stored Process Server']]", "migrateCommandProperty", MigrateConfig.STOREDPROC_SERVER),
        new MigrateAction("Property", "Foundation", "*[@PropertyName='Command'][AssociatedObject/ServerComponent[@Name?'- Connect Server']]", "migrateCommandProperty", MigrateConfig.CONNECT),
        new MigrateAction("Property", "Foundation", "*[@Name contains 'mysql']", "migrateMysqlProperty"),
        new MigrateAction("Property", null, "*[@Name='MySQL Server']","migrateMysqlProperty"),
        new MigrateAction("Document", null, "*[@Name='BIP Tree']","migrateURIHost",DAV),
        //TODO: These next 2 are probably migrating to the wrong logical component.
        new MigrateAction("SASClientConnection", "Foundation", "*[@ApplicationProtocol='SHARE']","migrateRemoteAddressHost",MigrateConfig.WORKSPACE),
        new MigrateAction("Property", "Foundation", "*[@PropertyName='HostName'][AssociatedObject/SASClientConnection[@ApplicationProtocol='SHARE']]","migrateDefaultValueHost",MigrateConfig.WORKSPACE),
        new MigrateAction("SASClientConnection", "Foundation", "*[@ApplicationProtocol='CONNECT']","migrateRemoteAddressHost",MigrateConfig.CONNECT),
        new MigrateAction("Property", "Foundation", "*[@PropertyName='workpath']","correctSlashes",null)
    };

    public void migrate()
    {
        LOG.info("Starting Migration at " + 
                GregorianCalendar.getInstance().getTime());
        int startNumCreates = this.omr.getNumCreates();
        int startNumUpdates = this.omr.getNumUpdates();
        int startNumDeletes = this.omr.getNumDeletes();
        boolean stopFlag = false;
        for (int actionIndex=0; 
             (!stopFlag) && actionIndex<MigrateActions.length; ++actionIndex)
        {
            MigrateAction action = MigrateActions[actionIndex];
            String typeName = action.getTypeName();
            String repositoryName = action.getRepositoryName();
            String searchParam = action.getSearchParam();
            String methodName = action.getMethodName();
            LOG.info("Processing Migration Action: " + typeName + ":" + 
                    repositoryName + ":" + searchParam + ":" + methodName);
            List repositoryNames = getRepositoryNames(repositoryName);
            for (Iterator reposNameIter=repositoryNames.iterator();
                 (!stopFlag) && reposNameIter.hasNext(); )
            {
                String reposName = (String)reposNameIter.next();
                LOG.debug("Searching repository " + reposName);
                this.omr.setRepository(reposName);
                List objects = this.omr.doOMRQuery(typeName, searchParam);
                LOG.debug(objects.size() + " objects found");
                for (Iterator objIter=objects.iterator(); objIter.hasNext(); )
                {
                    GenericMetadataObject gmo = 
                        (GenericMetadataObject)objIter.next();
                    String methodParam = action.getMethodParam();
                    if (methodParam == null)
                    {
                        if (!callMigrateHandler(methodName, gmo))
                        {
                            LOG.error("Stopping migration early due to fatal error");
                            stopFlag = true;
                            break;
                        }
                    }
                    else
                    {
                        if (!callMigrateHandler(methodName, gmo, methodParam))
                        {
                            LOG.error("Stopping migration early due to fatal error");
                            stopFlag = true;
                            break;
                        }
                    }
                    if (gmo.isModified())
                        this.omr.store(gmo);
                }
            }
        }
        LOG.info("Num. objects created during migration: " +
                (this.omr.getNumCreates() - startNumCreates));
        LOG.info("Num. objects updated during migration: " +
                (this.omr.getNumUpdates() - startNumUpdates));
        LOG.info("Num. objects deleted during migration: " +
                (this.omr.getNumDeletes() - startNumDeletes));
        if (!stopFlag)
        {
            loadInstallPropertyFiles();
        }
        LOG.debug("***************************************");
        LOG.debug("Numbers of candidate objects found");
        LOG.debug("=======================================");
        dumpCounts(this.candidateObjCountMap);
        LOG.debug("***************************************");
        LOG.info("Ending Migration at " + 
                GregorianCalendar.getInstance().getTime());
    }
    
    public void migrateAttributeProperty(GenericMetadataObject gmo)
    {
        String fromOMRHost = this.config.getMetadataFromHost();
        String defaultValue = gmo.getAttribute("DefaultValue");
        if (fromOMRHost.equals(defaultValue))
        {
            replaceHost(gmo, "DefaultValue", OMR);
        }
    }
    
    public void migrateDirectory(GenericMetadataObject gmo)
    {
        String name = gmo.getName();

        if (name.equalsIgnoreCase("Sample Stored Processes") ||
            name.equalsIgnoreCase("SP Source Directory"))
        {
            replaceStoredProcDirPrefix(gmo);
            return;
        }

        String component = lookupDirectoryMapping(name);
        if (component != null)
        {
            replacePathPrefix(gmo, "DirectoryName", component);
        }
    }

    public void replaceStoredProcDirPrefix(GenericMetadataObject gmo)
    {
        logCandidateObject(gmo);
        String directoryName = gmo.getAttribute("DirectoryName");
        String longestMatchingPrefix = null;
        String fromKey = null;
        String configSection = MigrateConfig.STOREDPROC_DIRS;
        Map storedProcDirs = this.config.getSection(configSection);
        for (Iterator it=storedProcDirs.entrySet().iterator();
                it.hasNext();)
        {
            Map.Entry entry = (Map.Entry)it.next();
            String key = (String)entry.getKey();
            String value = (String)entry.getValue();
            if (key.endsWith(".from.Prefix") &&
                    directoryName.startsWith(value))
            {
                if (longestMatchingPrefix == null ||
                    value.length() > longestMatchingPrefix.length())
                {
                    longestMatchingPrefix = value;
                    fromKey = key;
                }
            }
        }
        if (longestMatchingPrefix == null)
        {
            LOG.debug("_PATH NOT CHANGED_: " + directoryName);
            return;
        }
        String toKey = fromKey.substring(0, 
                fromKey.length() - ".from.Prefix".length()) + ".to.Prefix";
        String toPrefix = this.config.getValue(configSection, toKey);
        String result = BRMUtil.replacePathPrefix(directoryName,
                longestMatchingPrefix, toPrefix);
        if (result.equals(directoryName))
        {
            LOG.debug("_PATH NOT CHANGED_: " + directoryName);
        }
        else
        {
            LOG.debug("_PATH CHANGED_: " + directoryName + " -> " + result);
            gmo.setAttribute("DirectoryName", result);
        }
    }
    
    public void migrateSetinit(GenericMetadataObject gmo)
    {
        replacePathPrefix(gmo, "URI", MigrateConfig.SETINIT);
    }
    
    public void migrateLogin(GenericMetadataObject gmo)
    {
        String fromDomain = this.config.getFromValue("logins", "Domain");
        String toDomain = this.config.getToValue("logins", "Domain");
        replaceLoginHost(gmo, "Name", fromDomain, toDomain);
        replaceLoginHost(gmo, "UserID", fromDomain, toDomain);
        replaceLoginHost(gmo, "ChangeState", fromDomain, toDomain);
    }
    
    private TCPIPMigrateAction[] TCPIPMigrateActions = {
        // 
        // sourceClass, sourceName, sourceComponent, methodName, componentName
        //
        // For each row in the table, if each non-null value in (sourceClass
        // sourceName, sourceComponent) matches the current TCPIPConnection,
        // then the method identified by methodName is called with componentName
        // as the parameter.
        // 
        new TCPIPMigrateAction(ClassIds.CLSID_SAS_WORKSPACE, null, null, "migrateTCPIPHost", WORKSPACE),
        new TCPIPMigrateAction(ClassIds.CLSID_SASSTP_STOREDPROCESSSERVER, null, null, "migrateTCPIPHost", MigrateConfig.STOREDPROC_SERVER),
        new TCPIPMigrateAction("IOM Spawner", null, null, "migrateTCPIPHost", SPAWNER),
        new TCPIPMigrateAction(ClassIds.CLASSID_HTTP, null, null, "migrateTCPIPHost", DAV),
        new TCPIPMigrateAction(ClassIds.CLASSID_OLAP, null, null, "migrateTCPIPHost", OLAP),
        new TCPIPMigrateAction(null, "Platform Service Registry", null, "migrateTCPIPHostAndName", REMOTESERVICES),
        new TCPIPMigrateAction(null, "Event Broker Service", null, "migrateTCPIPHostAndName", REMOTESERVICES),
        new TCPIPMigrateAction(null, "Registered SAS Solution", "ODCS", "migrateTCPIPHost", ODCS),
        new TCPIPMigrateAction(null, "Registered SAS Solution", null, "migrateTCPIPHost", MIDTIER),
        new TCPIPMigrateAction(null, "SAS Presentation Theme", null, "migrateTCPIPHostAndName", MIDTIER),
        new TCPIPMigrateAction("ABC1212121212", null, null, "migrateTCPIPHost", REMOTESERVICES),
        new TCPIPMigrateAction("ABC1313131313", null, null, "migrateTCPIPHostAndName", REMOTESERVICES),
    };
    
    public void migrateTCPIPConnection(GenericMetadataObject gmo)
    {
        List objects = gmo.getAssociation("Source");
        if (objects == null || objects.size() == 0)
            return;
        GenericMetadataObject sourceDC = (GenericMetadataObject)objects.get(0);
        
        // Get full association info for the Source object
        sourceDC = this.omr.getObjectInfoById(sourceDC.getTypeName(), 
                sourceDC.getId());
        String curSourceClass = sourceDC.getAttribute("ClassIdentifier");
        String curSourceName = sourceDC.getName();
        if (curSourceName != null)
            curSourceName = curSourceName.toLowerCase();
        String curSourceComponent = null;
        objects = sourceDC.getAssociation("DescriptiveComponent");
        if (objects != null && objects.size() > 0)
        {
            curSourceComponent = (
                    (GenericMetadataObject)objects.get(0)).getName();
            if (curSourceComponent != null)
                curSourceComponent = curSourceComponent.toLowerCase();
        }
        for (int i=0; i<TCPIPMigrateActions.length; ++i)
        {
            TCPIPMigrateAction action = TCPIPMigrateActions[i];
            String sourceClass = action.getSourceClass();
            String sourceName = action.getSourceName();
            if (sourceName != null)
                sourceName = sourceName.toLowerCase();
            String sourceComponent = action.getSourceComponent();
            if (sourceComponent != null)
                sourceComponent = sourceComponent.toLowerCase();
            if ((sourceClass == null ||
                        sourceClass.equalsIgnoreCase(curSourceClass)) &&
                (sourceName == null || 
                        sourceName.indexOf(curSourceName) >= 0) &&
                (sourceComponent == null || 
                        sourceComponent.indexOf(curSourceComponent) >=0))
            {
                String methodName = action.getMethodName();
                String componentName = action.getComponentName();
                boolean ok = callMigrateHandler(methodName, gmo, componentName);
                if (!ok)
                {
                    throw new BRMInternalError(
                        "Ending TCPIPConnection migration early because of errors");
                }
                break;
            }
        }
    }
    
    public void migrateDefaultValueHost(GenericMetadataObject gmo, String component){
        replaceHost(gmo, "DefaultValue", component);
    }
    public void migrateRemoteAddressHost(GenericMetadataObject gmo, String component){
        replaceHost(gmo, "RemoteAddress", component);
    }
    public void migrateURIHost(GenericMetadataObject gmo, String component){
        replaceHost(gmo, "URI", component);
    }
    public void migrateTCPIPHost(GenericMetadataObject gmo, String component)
    {
        replaceHost(gmo, "HostName", component);
    }
    
    public void migrateTCPIPHostAndName(GenericMetadataObject gmo, String component)
    {
        replaceHost(gmo, "HostName", component);
        replaceHost(gmo, "Name", component);
    }
    
    public ServerComponentAction[] ServerComponentActions = {
            // Name, ClassIdentifier, Component
            // The first row where the name and class identifier match the 
            // ServerComponent is the row that supplies the Component
            new ServerComponentAction("HTTP Dav Server", null, DAV),
            ///FIXME: is the following the correct component to migrate?
            new ServerComponentAction(null, ClassIds.CLSID_SASSTP_STOREDPROCESSSERVER, MigrateConfig.STOREDPROC_SERVER),
            new ServerComponentAction(null, "IOM Spawner", SPAWNER),
            new ServerComponentAction(null, ClassIds.CLSID_SAS_WORKSPACE, WORKSPACE),
            new ServerComponentAction(null, "SSS_MYSQL_TIER", MYSQL),
            new ServerComponentAction(null, "MYSQL", MYSQL),
    };
    
    //FIXME:  This naming is really confusing.
    //I'm only using it now TO SEE IF IT WORKS
    public ServerComponentAction[] ConfiguredComponentActions = {
            new ServerComponentAction(null,"SSS_OMR_TIER",OMR),
            new ServerComponentAction(null,"SSS_MYSQL_TIER",MYSQL),
            //TODO: This might need to be the stored process tier
            new ServerComponentAction(null,"SSS_Data_TIER",WORKSPACE),
            new ServerComponentAction(null,"SSS_MIDDLE_TIER_PRIMARY",MIDTIER),
            new ServerComponentAction(null,ClassIds.CLASSID_OLAP,OLAP),
            new ServerComponentAction(null,"Connect Spawner",OLAP),
            new ServerComponentAction(null,ClassIds.CLASSID_CONNECT_SERVER,OLAP),
            new ServerComponentAction(null,"SHARE",OLAP)
    };
    
    public void migrateServerComponent(GenericMetadataObject gmo){
        migrateServerAndConfiguredComponent(ServerComponentActions,gmo);
    }
    public void migrateConfiguredComponent(GenericMetadataObject gmo) {
        migrateServerAndConfiguredComponent(ConfiguredComponentActions,gmo);
    }
    
    public void migrateServerAndConfiguredComponent(ServerComponentAction[] componentActions,
            GenericMetadataObject gmo)
    {
        for (int i=0; i<componentActions.length; ++i)
        {
            ServerComponentAction sca = componentActions[i];
            String name = sca.getName();
            String classIdentifier = sca.getClassIdentifier();
            if ((name == null || name.equalsIgnoreCase(gmo.getName())) &&
                (classIdentifier == null || 
                   classIdentifier.equalsIgnoreCase(
                           gmo.getAttribute("ClassIdentifier"))))
            {
                String component = sca.getComponent();
                String newMachineName = this.config.getToHost(component);
                if (newMachineName == null || newMachineName.length() == 0)
                {
                    throw new Error("Missing migration config property: " +
                            "[" + component + "] to.Host");
                }
                associateMachine(gmo, newMachineName);
            }
        }
    }
    
    public void migrateTextStore(GenericMetadataObject gmo)
    {
        // As per email from Nancy Rausch on 01 May 2008, make sure we do
        // not modify TextStore objects whose TextRole is TEMPLATE.
        // Related to customer issue:
        // http://sirius.na.sas.com/Sirius/GSTS/ShowTrack.aspx?trknum=7610088651
        // After dumping all text store objects on a couple of sample
        // systems and grepping them for keywords, I'm pretty sure that
        // TEMPLATE TextStore objects aren't what we want to modify, by our
        // current requirements.
        if (isTemplateTextStore(gmo))
            return;
        String storedTextBefore = gmo.getAttribute("StoredText");
        if (storedTextBefore == null || storedTextBefore.length() == 0)
            return;
        if (storedTextBefore.charAt(0) != '<')
            return;
        Document doc = buildDOMTree(storedTextBefore);
        if (doc == null)
            return;
        // Normalize stored text for more accurate change detection
        // Note: We MUST avoid false positives in testing for XML that
        //       needs to be changed! If we are not interested in this
        //       TextStore object, we must not modify it in any way, not
        //       even by normalizing the XML, because too much SAS code
        //       depends on abnormal XML forms being preserved.
        // Related to customer issue:
        // http://sirius.na.sas.com/Sirius/GSTS/ShowTrack.aspx?trknum=7610088651
        String normalizedStoredTextBefore = domTreeToString(doc);
        migrateInformationService(doc);
        migrateUserService(doc);
        migrateParam(doc);
        // If doc changed, store new StoredText value out
        String storedTextAfter = domTreeToString(doc);
        if (!normalizedStoredTextBefore.equals(storedTextAfter))
        {
            logCandidateObject(gmo);
            LOG.debug("StoredText before: " + storedTextBefore);
            LOG.debug("StoredText after : " + storedTextAfter);
            gmo.setAttribute("StoredText", storedTextAfter);
        }
    }

    public boolean isTemplateTextStore(GenericMetadataObject gmo)
    {
        String textRole = gmo.getAttribute("TextRole");
        if (textRole == null)
            return false;
        if (textRole.toUpperCase().equals("TEMPLATE"))
            return true;
        return false;
    }
   
    public void migrateInformationService(Document doc)
    {
        NodeList nodes = doc.getElementsByTagName("InformationService");
        for (int i=0; i<nodes.getLength(); ++i)
        {
            Element element = (Element)nodes.item(i);
            NodeList children = element.getChildNodes();
            for (int j=0; j<children.getLength(); ++j)
            {
                Node node = children.item(j);
                if (!(node instanceof Element))
                    continue;
                Element e = (Element)node;
                if (e.getNodeName().equals("ReposDef"))
                {
                    String protocol = e.getAttribute("protocol");
                    if (protocol == null)
                        continue;
                    if (protocol.equals("omi"))
                    {
                        e.setAttribute("host", this.config.getMetadataToHost());
                    }
                    else if (protocol.equals("dav"))
                    {
                        e.setAttribute("host", this.config.getToHost(DAV));
                    }
                }
            }
        }
    }
    
    public void migrateUserService(Document doc)
    {
        NodeList nodes = doc.getElementsByTagName("UserService");
        for (int i=0; i<nodes.getLength(); ++i)
        {
            Element element = (Element)nodes.item(i);
            NodeList children = element.getChildNodes();
            for (int j=0; j<children.getLength(); ++j)
            {
                Node node = children.item(j);
                if (!(node instanceof Element))
                    continue;
                Element e = (Element)node;
                if (e.getNodeName().equals("Profile"))
                {
                    String url = e.getAttribute("url");
                    if (url != null)
                    {
                        String newUrl = replaceHost(url, OMR);
                        e.setAttribute("url", newUrl);
                    }
                }
                else if (e.getNodeName().equals("UserContext"))
                {
                    String userid = e.getAttribute("userid");
                    if (userid != null && userid.length() > 0)
                    {
                        userid = replaceHost(userid, OMR);
                        e.setAttribute("userid", userid);
                    }
                }
            }
        }
    }
    
    public void migrateParam(Document doc)
    {
        NodeList nodes = doc.getElementsByTagName("param");
        for (int i=0; i<nodes.getLength(); ++i)
        {
            Element element = (Element)nodes.item(i);
            String name = element.getAttribute("name");
            if (name == null || name.length() == 0)
                continue;
            if (name.equals("file"))
            {
                String value = element.getAttribute("value");
                if (value != null && value.length() > 0)
                {
                    value = replacePathPrefix(value, MIDTIER);
                    element.setAttribute("value", value);
                }
            }
        }
    }
    
    public void correctSlashes(GenericMetadataObject gmo) {
        String defaultValue = gmo.getAttribute("DefaultValue");
        if(defaultValue==null || defaultValue.equals("")) {return;}
        //TODO: This should probably look at another tier.
        String flag = this.config.getToPrefix(MigrateConfig.STOREDPROC_DATA);
        if(flag.lastIndexOf('/')!=-1) {
            //going to unix
            // "\\" to the regexp engine, which matches \
            defaultValue = defaultValue.replaceAll("\\\\","/");
        }else{
            //going to windows
            defaultValue = defaultValue.replaceAll("/","\\");
        }
        gmo.setAttribute("DefaultValue",defaultValue);
    }
    
    public void migrateCommandProperty(GenericMetadataObject gmo, 
            String component)
    {
        replacePathPrefix(gmo, "DefaultValue", component);
        // Use the to.Command property to set the new
        // Command property value, if it exists in the migrate.ini file.
        String command = this.config.getToCommand(component);
        if (command != null && command.length() > 0)
        {
            String origCmdPath = gmo.getAttribute("DefaultValue");
            String newCmdPath = BRMUtil.changeFileNameInPath(origCmdPath,
                    command);
            gmo.setAttribute("DefaultValue", newCmdPath);
        }
    }
    
    public void migrateMysqlProperty(GenericMetadataObject gmo) {
        replaceHost(gmo, "DefaultValue", MYSQL);
    }
    
    /**
     * Attempt to locate and load the installation property files into the OMR.
     * This should repair settings that were clobbered due to a backup/restore
     * sequence of operations, especially if the backup and restore were done
     * on separate machines with potentially different operating systems,
     * numbers of servers, etc.
     */
    public void loadInstallPropertyFiles()
    {
        int startNumCreates = this.omr.getNumCreates();
        int startNumUpdates = this.omr.getNumUpdates();
        int startNumDeletes = this.omr.getNumDeletes();
        this.omr.setFoundationRepository();
        OMRPropertyFileLoader loader = new OMRPropertyFileLoader(this.omr);
        String SSCdir = BRMUtil.getSSCDir();
        LOG.debug("SAS Solutions Config dir: " + SSCdir);
        File soltnsdata = new File(SSCdir, Constants.SOLTNSDATA);
        File soltnsmid = new File(SSCdir, Constants.SOLTNSMID);
        File[] files = new File[]{soltnsdata, soltnsmid};
        for (int i=0; i<files.length; ++i)
        {
            if (files[i].exists())
            {
                LOG.info("Loading install properties: " + 
                        files[i].getPath());
                loader.loadPropertyFile(files[i]);
            }
            else
            {
                LOG.debug("Skipping loading install properties file " +
                         "(file not on this machine, but that's Ok): " + 
                        files[i].getPath());
            }
        }
        LOG.info("Num. objects created during property loading: " +
                (this.omr.getNumCreates() - startNumCreates));
        LOG.info("Num. objects updated during property loading: " +
                (this.omr.getNumUpdates() - startNumUpdates));
        LOG.info("Num. objects deleted during property loading: " +
                (this.omr.getNumDeletes() - startNumDeletes));
    }
    
    /////////////////////////////////////////////////////////////////////////
    // Helper classes and methods
    
    private String domTreeToString(Document doc)
    {
        StringWriter sw = new StringWriter();
        try
        {
              OutputFormat format = new OutputFormat(doc);
              XMLSerializer output = new XMLSerializer(sw, format);
              output.serialize(doc);
              return sw.toString();
        }
        catch (IOException e) {
            throw new BRMInternalError(e);
        }
    }
    
    public static Document buildDOMTree(String in)
    {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setValidating(false);
        try
        {
            DocumentBuilder db = dbf.newDocumentBuilder();
            return db.parse(new ReaderToInputStream(new StringReader(in)));
        }
        catch (ParserConfigurationException e)
        {
            throw new BRMInternalError(e);
        }
        catch (SAXException e)
        {
            // The input string was not well-formed XML
            return null;
        }
        catch (IOException e)
        {
            throw new BRMInternalError(e);
        }
    }
    
    static class StringReaderToInputStream extends InputStream {

        /* (non-Javadoc)
         * @see java.io.InputStream#read()
         */
        StringReader reader = null;
        public StringReaderToInputStream(StringReader in){
            reader = in;
        }
        public int read() throws IOException {
            return reader.read();
        }
        public int available(){
            try{
            if(reader.ready()){
                return 1;
            }else return 0;
            }catch(IOException e){return 0;}
        }

    }    
    
    private void associateMachine(GenericMetadataObject gmo, String machineName)
    {
        logCandidateObject(gmo);
        LOG.debug("associateMachine(" + gmo.getName() + ", " + machineName + ")");
        List objects = gmo.getAssociation("AssociatedMachine");
        GenericMetadataObject assocMachine = null;
        if (objects != null && objects.size() >= 1)
        {
            assocMachine = (GenericMetadataObject)objects.get(0);
            if (machineName.equalsIgnoreCase(assocMachine.getName()))
            {
                // Already associated to this machine
                return;
            }
        }
        // Search for machine of that name
        this.omr.setFoundationRepository();
        objects = this.omr.doOMRQuery("Machine", 
                "Machine[@Name='" + machineName + "']");
        if (objects != null && objects.size() == 1)
        {
            // Found the machine, make the association
            assocMachine = (GenericMetadataObject)objects.get(0);
        }
        else if (objects != null && objects.size() > 1)
        {
            throw new BRMUserError("Too many machines named '" + machineName + "', cannot make association");
        }
        else
        {
            // Need to create a Machine object with the right name
            assocMachine = new GenericMetadataObject();
            assocMachine.setTypeName("Machine");
            assocMachine.setName(machineName);
            assocMachine = this.omr.create(assocMachine);
        }
        gmo.setAssociation("AssociatedMachine", assocMachine);
    }
    
    class TCPIPMigrateAction
    {
        private String sourceClass;
        private String sourceName;
        private String sourceComponent;
        private String methodName;
        private String componentName;
        public TCPIPMigrateAction(
                String sourceClass, String sourceName, String sourceComponent,
                String methodName, String componentName)
        {
            this.sourceClass = sourceClass;
            this.sourceName = sourceName;
            this.sourceComponent = sourceComponent;
            this.methodName = methodName;
            this.componentName = componentName;
        }
        public String getComponentName()
        {
            return componentName;
        }
        public String getMethodName()
        {
            return methodName;
        }
        public String getSourceClass()
        {
            return sourceClass;
        }
        public String getSourceName()
        {
            return sourceName;
        }
        public String getSourceComponent()
        {
            return this.sourceComponent;
        }
    }
    
    class ServerComponentAction
    {
        private String name;
        private String classIdentifier;
        private String component;
        public ServerComponentAction(String name, String classIdentifier,
                String component)
        {
            this.name = name;
            this.classIdentifier = classIdentifier;
            this.component = component;
        }
        public String getClassIdentifier()
        {
            return classIdentifier;
        }
        public String getComponent()
        {
            return component;
        }
        public String getName()
        {
            return name;
        }
    }
    
    private static String getFirstDottedPart(String s)
    {
        int i = s.indexOf(".");
        if (i >= 0)
            return s.substring(0, i);
        else
            return s;
    }
    
    private List getRepositoryNames(String reposParam)
    {
        List repositoryNames = new ArrayList();
        if (reposParam == null || reposParam.length() == 0)
        {
            List repositories = this.omr.getRepositories();
            for (Iterator reposIter = repositories.iterator(); reposIter.hasNext(); )
            {
                GenericMetadataObject reposInfo = (GenericMetadataObject)reposIter.next();
                repositoryNames.add(reposInfo.getName());
            }
        }
        else
        {
            repositoryNames.add(reposParam);
        }
        return repositoryNames;
    }
    
    private boolean callMigrateHandler(String methodName,
            GenericMetadataObject gmo, Object[] params)
    {
        Class contextClass = this.getClass();
        Object[] fullParams = new Object[params.length + 1];
        fullParams[0] = gmo;
        for (int i=1; i<fullParams.length; ++i)
        {
            fullParams[i] = params[i - 1];
        }
        Class[] fullParamTypes = new Class[fullParams.length];
        for (int i=0; i<fullParams.length; ++i)
        {
            fullParamTypes[i] = fullParams[i].getClass();
        }
        Method method;
        try
        {
            method = contextClass.getMethod(methodName, fullParamTypes);
            method.invoke(this, fullParams);
            return true;
        }
        catch (SecurityException e)
        {
            LOG.error("Error in callMigrateHandler(" + methodName + ", " +
                    gmo + ")", e);
            return false;
        }
        catch (NoSuchMethodException e)
        {
            LOG.error("Error in callMigrateHandler(" + methodName + ", " +
                    gmo + ")", e);
            return false;
        }
        catch (IllegalAccessException e)
        {
            LOG.error("Error in callMigrateHandler(" + methodName + ", " +
                    gmo + ")", e);
            return false;
        }
        catch (InvocationTargetException e)
        {
            LOG.error("Error in callMigrateHandler(" + methodName + ", " +
                    gmo + ")", e);
            return false;
        }
    }
    
    private boolean callMigrateHandler(String methodName,
            GenericMetadataObject gmo)
    {
        return callMigrateHandler(methodName, gmo, new Object[]{});
    }
    
    private boolean callMigrateHandler(String methodName,
            GenericMetadataObject gmo, String s1)
    {
        return callMigrateHandler(methodName, gmo, new Object[]{s1});
    }
    
    /**
     * Replace occurrences of searchText in the value of the named attribute
     * on the object with replacementText. searchText is in the form of a 
     * machine name, possibly containing dots, and only appropriate whole
     * words will match.
     * 
     * For example, if the value of the "Name" attribute is "spmxyz", then
     * replaceHost(gmo, "Name", "spm", "abc") will *not* turn the Name into
     * "abcxyz". (This is an improvement over the previous version of BRM.)
     * 
     */
    private void replaceHost(GenericMetadataObject gmo, String attrName,
            String componentName)
    {
        String beforeValue = gmo.getAttribute(attrName);
        if (beforeValue == null)
        {
            // No such attribute on which to do search-and-replace
            return;
        }
        String s = replaceHost(beforeValue, componentName);
        logCandidateObject(gmo);
        LOG.debug(attrName + " = replaceHost(" + beforeValue + ", " +
                componentName + ") -> " + s);
        gmo.setAttribute(attrName, s);
    }
    
    //This is the inner replaceHost call.  
    private String replaceHost(String origText, String componentName)
    {
        String searchString = this.config.getFromHost(componentName);
        String replacement = this.config.getToHost(componentName);
        searchString = BRMUtil.escapeRegex(searchString);
        replacement = BRMUtil.escapeRepl(replacement);
        String s = origText.replaceAll("(?i)\\b" + searchString + "\\b",
                replacement);
        if(origText.equals(s)){
            LOG.debug("_HOST NOT CHANGED_:["+componentName+"] "+origText);
        }else {
            LOG.debug("_HOST CHANGED_: ["+componentName+"] "+origText+"->"+s );
        }
        return s;
    }
    
    public void replaceLoginHost(GenericMetadataObject gmo, String attrName,
            String searchText, String replacementText)
    {
        if (searchText == null || replacementText == null)
        {
            // Missing a vital setting, make no changes
            return;
        }
        String beforeValue = gmo.getAttribute(attrName);
        if (beforeValue == null)
        {
            // No value to be changed
            return;
        }
        if (searchText.length() == 0 && beforeValue.indexOf("@") >= 0)
        {
            // Don't disturb LDAP login
            return;
        }
        String result = null;
        if ("Name".equals(attrName))
        {
            // Modify user names of the form "sestest1.sasspusr.Login",
            // where "sestes1" happens to be the OMR server name. This is to
            // make sure that we get rid of all instances of the old machine
            // name in metadata, so that they don't show up in a search
            // later, but it has no functional effect.
            result = replacePrefix(beforeValue, searchText, replacementText,
                    ".");
        }
        else
        {
            result = replacePrefix(beforeValue, searchText, replacementText, 
                    "\\");
        }
        logCandidateObject(gmo);
        gmo.setAttribute(attrName, result);
        LOG.debug(attrName + " = replaceLogin(" + beforeValue + ", " +
                searchText + ", " + replacementText + ") -> " + result);
    }

    public static String replacePrefix(String beforeValue, String searchText,
            String replacementText, String separator)
    {
        String result = beforeValue;
        String beforePrefix = "";
        int separatorIndex = beforeValue.indexOf(separator);
        if (separatorIndex >= 0)
            beforePrefix = beforeValue.substring(0, separatorIndex);
        if (searchText.length() > 0 && replacementText.length() == 0)
        {
            // Remove existing prefix and separator if it matches
            if (beforePrefix.equalsIgnoreCase(searchText))
            {
                result = beforeValue.substring(separatorIndex + 
                        separator.length());
            }
        }
        else
        {
            // Replace existing prefix if it matches
            if (beforePrefix.equalsIgnoreCase(searchText))
            {
                if (separatorIndex >= 0)
                {
                    result = replacementText +
                        beforeValue.substring(separatorIndex);
                }
                else
                {
                    if (replacementText.length() > 0)
                        result = replacementText + separator + beforeValue;
                    else
                        result = beforeValue;
                }
            }
        }
        return result;
    }
    
    private void replacePathPrefix(GenericMetadataObject gmo, String attrName,
            String componentName)
    {
        logCandidateObject(gmo);
        String value = gmo.getAttribute(attrName);
        String result = replacePathPrefix(value, componentName);
        gmo.setAttribute(attrName, result);
    }
    
    private String replacePathPrefix(String origString, String componentName)
    {
        String fromPrefix = this.config.getFromPrefix(componentName);
        if (fromPrefix == null || fromPrefix.length() == 0)
        {
            throw new BRMUserError("Missing configuration in migrate.ini file:\n" +
                    "[" + componentName + "] section has no from.Prefix property.");
        }
        String toPrefix = this.config.getToPrefix(componentName);
        if (toPrefix == null || toPrefix.length() == 0)
        {
            throw new BRMUserError("Missing configuration in migrate.ini file:\n" +
                    "[" + componentName + "] section has no to.Prefix property.");
        }
        String result = BRMUtil.replacePathPrefix(origString, fromPrefix, toPrefix);
        if(origString.equals(result)){
            LOG.debug("_PATH NOT CHANGED_:["+componentName+"] "+result);
        }else {
            LOG.debug("_PATH CHANGED_: ["+componentName+"] "+origString+"->"+result );
        }
        return result;
    }

    private void logCandidateObject(GenericMetadataObject gmo)
    {
        incCount(this.candidateObjCountMap, gmo.getTypeName());
        LOG.debug("Change candidate: " + gmo);
    }
    
    private void incCount(Map countMap, String key)
    {
        Integer count = (Integer)countMap.get(key);
        if (count == null)
        {
            count = new Integer(1);
        }
        else
        {
            count = new Integer(count.intValue() + 1);
        }
        countMap.put(key, count);
    }
    
    private void dumpCounts(Map countMap)
    {
        Object[] keys = countMap.keySet().toArray();
        Arrays.sort(keys);
        for (int i=0; i<keys.length; ++i)
        {
            String key = (String)keys[i];
            LOG.debug(key + ": " + countMap.get(key));
        }
    }
    
    /**
     * Given a Directory OMR object's name, return a component name to use
     * for migrating that Directory object.
     * 
     * @param directoryName
     * @return a component name, or null if the directory name is not in the 
     * table.
     */
    private String lookupDirectoryMapping(String directoryName)
    {
        if (this.directoryComponentMap == null)
        {
            Map dcm = new HashMap();
            dcm.put("DDSData.Path", WORKSPACE);
            dcm.put("stagedds.Path", WORKSPACE);
            dcm.put("errorData.Path", WORKSPACE);
            dcm.put("ConformedDataMart.Path", WORKSPACE);
            dcm.put("SPMData.Path", WORKSPACE);
            dcm.put("FMSData.Path", WORKSPACE);
            dcm.put("HCMData.Path", WORKSPACE);
            dcm.put("maps.Path", MigrateConfig.SAS_EXE_DIR);
            this.directoryComponentMap = dcm;
        }
        return (String)this.directoryComponentMap.get(directoryName);
    }
    
    private static void testQuery(OMRConnection omr)
    {
        omr.setFoundationRepository();
        List objects = omr.doOMRQuery("Document", 
                "*[@Name='SAS Setinit Location']");
        System.out.println("Query returned: " + objects);
    }

    /**
     * @param args
     */
    public static void main(String[] args)
    {
        System.out.println("Test run of MainMigrator");
        if (args.length < 1)
        {
            System.err.println(
                "Need at least one parameter: metadata server hostname.\n" +
                "Add 'readonly' to do dry-run migration");
            return;
        }
        String omrHost = args[0];
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        OMRConnection omr = new OMRConnection(
                omrHost, "8561", "sasadm", "AdminAdmin1"
                //"d15724", "8561", "sasadm", "AdminAdmin1"
                //"ngl2xp1.emea.sas.com", "8561", "ngladm", "Ng1123"
                );
        omr.connect();
        LOG.info("Connected to OMR server: " + omr.getHost());
        if (args.length >= 2 && args[1].equalsIgnoreCase("readonly"))
        {
            LOG.info("Setting OMR connection to readonly");
            omr.setReadOnly(true);
        }
        if (false)
            testQuery(omr);
        try{
            MainMigrator migrator = new MainMigrator(omr);
            migrator.migrate();
        }catch(BRMConfigurationException e) {
            LOG.error("Could not migrate because there was a configuration exception: ",e);
        }
        omr.close();
    }
}
