// MetadataServerWaiter.java
// Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
// Created by David Handy  29 Jan 2007

package com.sas.solutions.brm.omr;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.util.BRMUtil;

public class MetadataServerWaiter
{
    private static final Logger LOG = Logger.getLogger(
            MetadataServerWaiter.class);
    
    private Properties installProps;

    public MetadataServerWaiter(Properties installProps)
    {
        this.installProps = installProps;
    }

    /**
     * Wait until the metadata server is no longer running.
     * This is determined by checking if the files in the repository manager
     * directory are locked. This only works on Windows.
     * 
     * @param timeoutSeconds  Maximum time to wait
     * @return true if server shutdown was detected, false if timeout.
     */
    public boolean waitTillServerIsStopped(final int timeoutSeconds)
    {
        LOG.info("Waiting up to " + timeoutSeconds + 
                " seconds for metadata server to shut down.");
        final long timeoutMilliSeconds = timeoutSeconds * 1000L;

        // C:\\SAS\\SASSolutionsConfig\\Lev1\\SASMain
        String n1 = this.installProps.getProperty("SOLUTIONS_SAS_SERVER_DIR");
        if (n1 == null || n1.length() == 0)
        {
            throw new BRMConfigurationException(
                "Missing property SOLUTIONS_SAS_SERVER_DIR " +
                "in " + Constants.SOLTNSDATA);
        }
        // MetadataServer/MetadataRepositories
        String n2 = this.installProps.getProperty("REPOS_FOLDER");
        if (n2 == null || n2.length() == 0)
        {
            throw new BRMConfigurationException(
                "Missing property REPOS_FOLDER " +
                "in " + Constants.SOLTNSDATA);
        }
        
        File repositoriesDir = new File(n1, n2);
        File metadataServerDir = repositoriesDir.getParentFile();
        File rposmgrDir = new File(metadataServerDir, "rposmgr");
        
        if (!rposmgrDir.isDirectory())
        {
            throw new Error("Missing repository manager directory: " + 
                    rposmgrDir);
        }
        
        boolean atLeastOneTimeThroughLoop = false;
        boolean allFilesUnlocked = false;
        long retryIntervalMilliSeconds = 2L * 1000L;
        for (long t0 = System.currentTimeMillis(); 
             (!atLeastOneTimeThroughLoop) || 
                 ((System.currentTimeMillis() - t0) < timeoutMilliSeconds); 
             delayMS(retryIntervalMilliSeconds))
        {
            atLeastOneTimeThroughLoop = true;
            allFilesUnlocked = true;
            String[] fileNames = rposmgrDir.list();
            for (int i=0; i<fileNames.length; ++i)
            {
                String fileName = fileNames[i];
                File file = new File(rposmgrDir, fileName);
                if (file.isDirectory())
                    continue;
                if (!file.canRead() && !file.canWrite())
                {
                    allFilesUnlocked = false;
                    break;
                }
                FileInputStream fis = null;
                try
                {
                    fis = new FileInputStream(file);
                }
                catch (FileNotFoundException e1)
                {
                    if (file.isFile())
                    {
                        // File *is* found, but it is locked
                        allFilesUnlocked = false;
                        break;
                    }
                    throw new Error(
                        "Unexpected error opening file " + file, e1);
                }
                try
                {
                    fis.close();
                }
                catch (IOException e2)
                {
                    throw new Error(
                        "Unexpected error closing file " + file, e2);
                }
            }
            if (allFilesUnlocked || timeoutMilliSeconds == 0)
                break;
        }
        if (!allFilesUnlocked)
        {
            LOG.warn("Timed out after " + timeoutSeconds + 
                    " seconds while waiting for metadata server to stop");
        }
        else
        {
            LOG.info("Metadata Server files are unlocked, done waiting");
        }
        return allFilesUnlocked;
    }
    
    private static void delayMS(long delayMilliSeconds)
    {
        try
        {
            Thread.sleep(delayMilliSeconds);
        }
        catch (InterruptedException e)
        {
            ; // ignore
        }
    }

    public static void main(String[] args)
    {
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        Properties installProps = new Properties();
        if (!BRMUtil.loadInstallPropertyFile(Constants.SOLTNSDATA,
                    installProps))
        {
            throw new BRMConfigurationException(
                "Missing " + Constants.SOLTNSDATA + 
                " install properties file, not found in " + 
                BRMUtil.getSSCDir());
        }
        int timeoutSeconds = 30 * 60;

        MetadataServerWaiter msw = new MetadataServerWaiter(installProps);
        msw.waitTillServerIsStopped(timeoutSeconds);
    }
}

// end-of-file
