/*
 * Created on Apr 10, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.omr;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.omg.CORBA.StringHolder;

import com.sas.iom.SASIOMDefs.GenericError;
import com.sas.meta.SASOMI.IOMI;
import com.sas.meta.SASOMI.IOMIHelper;
import com.sas.services.connection.ConnectionInterface;

/**
 * Connect to the SAS Solutions OMR server and grab configuration info.
 * 
 * @author davhan
 * 
 */
public class OMRConnection
{
    public static final String DEFAULT_OMR_NAMESPACE = "SAS";
    public static final String FOUNDATION_REPOSITORY_NAME = "Foundation";

    private String host;
    private String port;
    private String user;
    private String password;

    private String reposName;
    private String reposId;
    private String nameSpace;
    private List reposListCache;

    transient private ConnectionInterface cx;
    
    private boolean readOnly;

    // Statistics
    private int numCreates = 0;
    private int numQueries = 0;
    private int numUpdates = 0;
    private int numDeletes = 0;
    
    private static final Logger LOG = Logger.getLogger(OMRConnection.class);

    public OMRConnection(String host, String port, String user,
            String password)
    {
        this.host = host;
        this.port = port;
        this.user = user;
        this.password = password;
        this.cx = null; // set by connect()
        this.reposName = null; // set by setRepository()
        this.reposId = null; // set by setRepository()
        this.nameSpace = DEFAULT_OMR_NAMESPACE; // (re)set by setRepository()
        this.reposListCache = null;
        this.readOnly = false;
    }

    public void connect()
    {
        if (this.cx != null)
        {
            close();
        }
        try
        {
            this.cx = OMRConnector.makeConnection(host, port, user, password);
        }
        catch (OMRException e)
        {
            LOG.error("Failed to log into the metadata server.", e);
            throw new Error("Failed to log in to the metadata server.", e);
        }
    }
    
    public boolean isConnected()
    {
        return (this.cx != null);
    }

    /**
     * Close this OMR connection
     */
    public void close()
    {
        if (this.cx != null)
        {
            this.cx.close();
            this.cx = null;
        }
    }

    public String getHost()
    {
        return host;
    }

    public String getPort()
    {
        return port;
    }

    public String getUser()
    {
        return user;
    }

    public String getPassword()
    {
        return password;
    }
    
    public void setReadOnly(boolean readOnly)
    {
        this.readOnly = readOnly;
    }
    
    public boolean isReadOnly()
    {
        return this.readOnly;
    }

    /**
     * Set the current working respository.
     * @param reposName  Name of the respository
     * @throws  IllegalArgumentException if the repository does not exist,
     *          Error if some other error occurs.
     */
    public void setRepository(String reposName)
    {
        if (reposName == null)
            throw new IllegalArgumentException("reposName parameter is null");
        if (reposName.equals(this.reposName))
        {
            // Already set to this repository
            return;
        }
        GenericMetadataObject reposInfo = getRepositoryInfoByName(reposName);
        if (reposInfo == null)
            throw new IllegalArgumentException("No repository named " + 
                    reposName);
        this.reposName = reposName;
        this.reposId = reposInfo.getId();
        this.nameSpace = reposInfo.getAttribute("DefaultNS"); /*I18nOK:LINE*/
        if (this.reposId == null)
            throw new Error("Error getting Id of Repository " + 
                    reposName);
        if (this.nameSpace == null)
            this.nameSpace = DEFAULT_OMR_NAMESPACE;
    }
    
    /**
     * @return the name of the currently selected repository, or null if no
     * repository has been selected yet.
     */
    public String getRepositoryName()
    {
        return this.reposName;
    }
    
    public void setFoundationRepository()
    {
        setRepository(FOUNDATION_REPOSITORY_NAME);
    }
    
    /**
     * Get a list of information items for each Repository in the OMR.
     * Since the list of repositories ordinarily never changes during a
     * session, this method may use cached results. To ensure you have the
     * very latest list of repositories, use getRepositoriesNoCache().
     * @return a List of GenericMetadataObject instances, each with the 
     *      attributes:
     *          "Id", "Name", "DefaultNS"
     */
    public List getRepositories()
    {
        if (this.reposListCache != null)
            return this.reposListCache;
        return getRepositoriesNoCache();
    }

    /**
     * Get a list of information items for each Repository in the OMR.
     * This method never uses cached results, but it does set the cached
     * results so that the next call to getRepositories() will go faster.
     * @return a List of GenericMetadataObject instances, each with the 
     *      attributes:
     *          "Id", "Name", "DefaultNS"
     */
    public List getRepositoriesNoCache()
    {
        StringHolder queryResult = new org.omg.CORBA.StringHolder();
        int flags = 0;
        String options = "";
        try
        {
            int rc = getIOMI().GetRepositories(queryResult, flags, options);
            if (rc != 0)
            {
                throw new Error("GetRepositories non-zero return: " + rc);
            }
            this.numQueries += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
        ObjectListHandler listHandler = ObjectListHandler
                .parseXML(queryResult.value);
        queryResult = null;
        List result = new ArrayList();
        for (Iterator it = listHandler.getList().iterator(); it.hasNext();)
        {
            Map item = (Map) it.next();
            if (item == null)
                throw new Error("Repository info is null");
            String id = (String)item.get("Id");
            String name = (String)item.get("Name");
            String defaultNS = (String)item.get("DefaultNS");
            if (id == null || id.length() == 0)
                throw new Error("Repository info missing Id");
            if (name == null || name.length() == 0)
                throw new Error("Repository info missing Name");
            if (defaultNS == null || defaultNS.length() == 0)
                throw new Error("Repository info missing DefaultNS");
            result.add(new GenericMetadataObject(item));
        }
        this.reposListCache = result;
        return result;
    }

    /**
     * Get information about an OMR Repository.
     * 
     * @param reposName
     *            The name of the Repository to query
     * @return a GenericMetadataObject contining information about the 
     *         Repository, with the keys:
     *           {"Id", "Name", "DefaultNS"},
     *         or return null if no respository of that name is found.
     */
    public GenericMetadataObject getRepositoryInfoByName(String reposName)
    {
        if (reposName == null)
            throw new IllegalArgumentException("reposName parameter is null");
        for (Iterator it=getRepositories().iterator(); it.hasNext(); )
        {
            GenericMetadataObject reposInfo = (GenericMetadataObject)it.next();
            // getRepositories() should insure that reposInfo is not null
            String name = reposInfo.getName();
            // getRepositories() should insure that reposInfo had "Name" key
            if (reposName.equalsIgnoreCase(name))
                return reposInfo;
        }
        return null;
    }

    /**
     * This is really an internal method. Use doOMRQuery() instead.
     */
    List doXMLQuery(String typeName, String xmlQuery, int flags)
    {
        if (this.reposId == null)
            throw new IllegalStateException("Call setRepository() first");
        StringHolder queryResult = new StringHolder();
        try
        {
            int rc = getIOMI().GetMetadataObjects(this.reposId, typeName, 
                    queryResult, this.nameSpace, flags, xmlQuery);
            if (rc != 0)
            {
                throw new Error(
                        "GetMetadataObjects() returned non-zero code: " + rc);
            }
            this.numQueries += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
        List rawList = ObjectListHandler.parseXML(queryResult.value).getList();
        List result = new ArrayList();
        for (Iterator mapIter=rawList.iterator(); mapIter.hasNext(); )
        {
            result.add(new GenericMetadataObject((Map)mapIter.next()));
        }
        return result;
    }
    
    /**
     * Get the sub-types of a particular metadata type.
     * @param supertypeName  The name of a parent metadata type
     * @param allDescendents  Include sub-sub types, etc.
     * @return
     *      A List of GenericMetadata objects. For each object in the list:
     *      <ul>
     *      <li> getTypeName() returns "Type" </li>
     *      <li> getName() returns null </li>
     *      <li> getId() returns the name of the metadata type </li>
     *      <li> getAttribute("HasSubtypes") returns "0" or "1" </li>
     *      </ul>
     *
     */
    public List getSubtypes(String supertypeName, boolean allDescendents)
    {
        if (this.reposId == null)
            throw new IllegalStateException("Call setRepository() first");
        StringHolder returnInfo = new StringHolder();
        int flags = 0;
        if (allDescendents)
            flags = OMRConst.OMI_ALL_DESCENDANTS;
        String options = "";
        try
        {
            int rc = getIOMI().GetSubtypes(supertypeName, returnInfo, 
                    this.nameSpace, flags, options);
            if (rc != 0)
            {
                throw new Error(
                        "GetSubtypes() returned non-zero code: " + rc);
            }
            this.numQueries += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
        List rawList = ObjectListHandler.parseXML(returnInfo.value).getList();
        List result = new ArrayList();
        for (Iterator mapIter=rawList.iterator(); mapIter.hasNext(); )
        {
            result.add(new GenericMetadataObject((Map)mapIter.next()));
        }
        return result;
    }

    /**
     * Get the metadata types supported by the current repository.
     * @return
     *      A List of GenericMetadata objects. For each object in the list:
     *      <ul>
     *      <li> getTypeName() returns "Type" </li>
     *      <li> getName() returns null </li>
     *      <li> getId() returns the name of the metadata type </li>
     *      <li> getAttribute("HasSubtypes") returns "0" or "1" </li>
     *      </ul>
     */
    public List getTypes()
    {
        if (this.reposId == null)
            throw new IllegalStateException("Call setRepository() first");
        StringHolder returnInfo = new StringHolder();
        int flags = 0;
        String options = "";
        try
        {
            int rc = getIOMI().GetTypes(returnInfo, this.nameSpace, flags, 
                    options);
            if (rc != 0)
            {
                throw new Error(
                        "GetTypes() returned non-zero code: " + rc);
            }
            this.numQueries += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
        List rawList = ObjectListHandler.parseXML(returnInfo.value).getList();
        List result = new ArrayList();
        for (Iterator mapIter=rawList.iterator(); mapIter.hasNext(); )
        {
            result.add(new GenericMetadataObject((Map)mapIter.next()));
        }
        return result;
    }
    
    /**
     * @param typeName -- The name of an OMR type
     * @return a List of (String) names of associations for that type
     */
    public List getTypeAssociations(String typeName)
    {
        // We make sure that setRepository() has been called mainly so we have
        // a valid nameSpace.
        if (this.reposId == null)
            throw new IllegalStateException("Call setRepository() first");
        StringHolder queryResult = new StringHolder();
        try
        {
            int rc = getIOMI().GetTypeProperties(typeName, queryResult, 
                    this.nameSpace, 0, "");
            if (rc != 0)
            {
                throw new Error(
                        "GetTypeProperties() returned non-zero code: " + rc);
            }
            this.numQueries += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
        PropertyListHandler handler = 
            PropertyListHandler.parseXML(queryResult.value);
        return handler.getAssociationNames();
    }
    
    public GenericMetadataObject getObjectInfoById(String typeName, String id)
    {
        assert id != null;
        StringBuffer inMetadata = new StringBuffer();
        inMetadata.append("<");
        inMetadata.append(typeName);
        inMetadata.append(" Id=\"");
        inMetadata.append(OMRUtil.escapeXML(id));
        inMetadata.append("\" />");
        StringHolder outMetadata = new StringHolder();
        String ns = this.nameSpace;
        if (ns == null)
        {
            ns = DEFAULT_OMR_NAMESPACE;
        }
        int flags = OMRConst.OMI_ALL;
        String options = "";
        try
        {
            int rc = getIOMI().GetMetadata(inMetadata.toString(), outMetadata, 
                    ns, flags, options);
            if (rc != 0)
            {
                throw new Error(
                        "GetMetadata() returned non-zero code: " + rc);
            }
            this.numQueries += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
        return new GenericMetadataObject(
                ObjectInfoHandler.parseXML(outMetadata.value).getObjectInfo());
    }
    
    /**
     * Search the OMR using its "XPath"-like query language. Call 
     * setRepository(reposName) or setFoundationRepository() first to determine
     * which OMR repository will be searched.
     * 
     * @param typeName
     *      The name of type of the OMR Metadata object(s) to find.
     * @param searchParam
     *      The search parameter. Here are some examples:
     *      <pre>
     *      Machine[DeployedComponents/ConfiguredComponent[@name='SSS Data Tier']]
     *      *[DescriptiveComponent/SoftwareComponent[@name='SAS Solutions Installer']]
     *      </pre>
     *      Special XML characters in searchParam will be automatically escaped.
     * @return a List of GenericMetadataObject objects.
     *      The returned GenericMetadataObject instances contain the simple
     *      attribute names and values, but not the associations.
     * 
     */
    public List doOMRQuery(String typeName, String searchParam)
    {
        return doOMRQuery(typeName, searchParam, false);
    }
    
    public List doOMRQuery(String typeName, String searchParam, 
            boolean matchExactType)
    {
        // Build XML select command
        StringBuffer xmlQuery = new StringBuffer();
        xmlQuery.append("<XMLSelect Search=\"");
        xmlQuery.append(OMRUtil.escapeXML(searchParam));
        xmlQuery.append("\"/>");
        int flags = (OMRConst.OMI_GET_METADATA
                | OMRConst.OMI_ALL
                | OMRConst.OMI_SUCCINCT
                | OMRConst.OMI_XMLSELECT
                );
        if (!matchExactType)
        {
            flags |= OMRConst.OMI_INCLUDE_SUBTYPES;
        }
        return doXMLQuery(typeName, xmlQuery.toString(), flags);
    }

    /**
      * Like doOMRQuery() except that it returns objects with only the Id
      * and Name attributes, and without any of their associations (saves
      * memory, etc.)
      */
    public List doOMRQuerySimpleResult(String typeName, String searchParam)
    {
        return doOMRQuerySimpleResult(typeName, searchParam, false);
    }

    /**
      * Like doOMRQuery() except that it returns objects with only the Id
      * and Name attributes, and without any of their associations (saves
      * memory, etc.)
      */
    public List doOMRQuerySimpleResult(String typeName, String searchParam,
            boolean matchExactType)
    {
        // Build XML select command
        StringBuffer xmlQuery = new StringBuffer();
        xmlQuery.append("<XMLSelect Search=\"");
        xmlQuery.append(OMRUtil.escapeXML(searchParam));
        xmlQuery.append("\"/>");
        int flags = (OMRConst.OMI_GET_METADATA
                | OMRConst.OMI_SUCCINCT
                | OMRConst.OMI_XMLSELECT
                );
        if (!matchExactType)
        {
            flags |= OMRConst.OMI_INCLUDE_SUBTYPES;
        }
        return doXMLQuery(typeName, xmlQuery.toString(), flags);
    }
    
    public List getSolutionsComponents()
    {
        setFoundationRepository();
        return doOMRQuery("ConfiguredComponent", 
            "*[DescriptiveComponent/SoftwareComponent[@name='SAS Solutions Installer']]");
    }
    
    /**
     * Recursively get the info for each Metadata object identified by
     * idList, and each object associated with those objects. Each ID is looked
     * up in the resultAccumulator before a metadata query is performed, thus
     * avoiding infinite loops.
     * 
     * @param idList
     *      A Collection of GenericMetadataObject instances. The Id of each
     *      one will be used for searching.
     * @param resultAccumulator
     *      A Map holding the results of the search. Each key is a String
     *      containing a Metadata object ID, and the associated value is a
     *      GenericMetadataObject as returned by getObjectInfoById().
     */
    public void getObjectsFollowingAssociations(Collection idList, 
            Map resultAccumulator)
    {
        Map nextPass = new HashMap();
        for (Iterator idIter=idList.iterator(); idIter.hasNext(); )
        {
            GenericMetadataObject idObj = (GenericMetadataObject)idIter.next();
            String id = idObj.getId();
            String typeName = idObj.getTypeName();
            if (id == null)
                continue;
            if (resultAccumulator.containsKey(id))
                continue;
            GenericMetadataObject gmo = getObjectInfoById(typeName, id);
            resultAccumulator.put(id, gmo);
            Set newAssocSet = new HashSet();
            for (Iterator assocNameIter=gmo.getAssociationNames().iterator(); 
                 assocNameIter.hasNext(); )
            {
                String assocName = (String)assocNameIter.next();
                List assocList = gmo.getAssociation(assocName);
                for (Iterator assocIter=assocList.iterator(); 
                     assocIter.hasNext(); )
                {
                    Object assocO = assocIter.next();
                    GenericMetadataObject assoc = (GenericMetadataObject)assocO;
                    String assocId = assoc.getId();
                    if (!(resultAccumulator.containsKey(assocId) ||
                          nextPass.containsKey(assocId) ||
                          newAssocSet.contains(assocId)))
                    {
                        nextPass.put(assocId, assoc);
                        newAssocSet.add(assocId);
                    }
                }
            }
        }
        if (nextPass.size() == 0)
        {
            return;
        }
        getObjectsFollowingAssociations(nextPass.values(), resultAccumulator);
    }
    
    public GenericMetadataObject create(GenericMetadataObject gmo)
    {
        if (isReadOnly())
            return gmo;
        if (this.reposId == null)
            throw new IllegalStateException("Call setRepository() first");
        if (gmo.getTypeName() == null)
            throw new IllegalStateException("Call gmo.setTypeName() first");
        
        String inMetadata = gmo.getXML();
        StringHolder outMetadata = new StringHolder();
        String ns = this.nameSpace;
        if (ns == null)
        {
            ns = DEFAULT_OMR_NAMESPACE;
        }
        int flags = OMRConst.OMI_TRUSTED_CLIENT;
        String options = "";
        try
        {
            int rc = getIOMI().AddMetadata(inMetadata, this.reposId, 
                    outMetadata, ns, flags, options);
            if (rc != 0)
            {
                throw new Error("AddMetadata non-zero return: " + rc);
            }
            this.numCreates += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
        return new GenericMetadataObject(
                ObjectInfoHandler.parseXML(outMetadata.value).getObjectInfo());
    }
    
    /**
     * Store a (potentially modified) metadata object back into the OMR.
     * If isReadOnly() is true, then this method does nothing.
     * @param gmo
     */
    public void store(GenericMetadataObject gmo)
    {
        if (isReadOnly())
            return;
        if (gmo.getId() == null)
            throw new Error("Object has no Id");
        
        String inMetadata = gmo.getXML("Replace");
        StringHolder outMetadata = new StringHolder();
        String ns = this.nameSpace;
        if (ns == null)
        {
            ns = DEFAULT_OMR_NAMESPACE;
        }
        int flags = OMRConst.OMI_TRUSTED_CLIENT;
        String options = "";
        try
        {
            int rc = getIOMI().UpdateMetadata(inMetadata, outMetadata, ns, 
                    flags, options);
            if (rc != 0)
            {
                throw new Error("UpdateMetadata non-zero return: " + rc);
            }
            this.numUpdates += 1;
            gmo.setModified(false);
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
    }
    
    public void delete(GenericMetadataObject gmo)
    {
        if (isReadOnly())
            return;
        String id = gmo.getId();
        if (id == null)
            throw new Error("Can't delete object without Id");
        StringBuffer sb = new StringBuffer();
        sb.append("<");
        sb.append(gmo.getTypeName());
        sb.append(" Id=\"");
        sb.append(OMRUtil.escapeXML(id));
        sb.append("\"/>");
        String inMetadata = sb.toString();
        StringHolder outMetadata = new StringHolder();
        String ns = this.nameSpace;
        if (ns == null)
        {
            ns = DEFAULT_OMR_NAMESPACE;
        }
        int flags = (OMRConst.OMI_TRUSTED_CLIENT | 
                OMRConst.OMI_RETURN_LIST);
        String options = "";
        try
        {
            int rc = getIOMI().DeleteMetadata(inMetadata, outMetadata, ns, 
                    flags, options);
            if (rc != 0)
            {
                throw new Error("DeleteMetadata non-zero return: " + rc);
            }
            this.numDeletes += 1;
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
    }

    public void deleteList(List gmos)
    {
        if (isReadOnly())
            return;
        StringBuffer sb = new StringBuffer();
        for (Iterator gmoIter=gmos.iterator(); gmoIter.hasNext(); )
        {
            GenericMetadataObject gmo = (GenericMetadataObject)gmoIter.next();
            String id = gmo.getId();
            if (id == null)
                throw new Error("Can't delete object without Id");
            sb.append("<");
            sb.append(gmo.getTypeName());
            sb.append(" Id=\"");
            sb.append(OMRUtil.escapeXML(id));
            sb.append("\"/>");
        }
        String inMetadata = sb.toString();
        sb = null;
        StringHolder outMetadata = new StringHolder();
        String ns = this.nameSpace;
        if (ns == null)
        {
            ns = DEFAULT_OMR_NAMESPACE;
        }
        int flags = (OMRConst.OMI_TRUSTED_CLIENT | 
                OMRConst.OMI_RETURN_LIST);
        String options = "";
        try
        {
            int rc = getIOMI().DeleteMetadata(inMetadata, outMetadata, ns, 
                    flags, options);
            if (rc != 0)
            {
                throw new Error("DeleteMetadata non-zero return: " + rc);
            }
            this.numDeletes += gmos.size();
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    // Internal helper methods

    private IOMI getIOMI()
    {
        if (this.cx == null)
            throw new IllegalStateException("Call connect() first");
        org.omg.CORBA.Object obj = this.cx.getObject();
        IOMI iomiHandle = IOMIHelper.narrow(obj);
        if (iomiHandle == null)
        {
            throw new RuntimeException(
                    new OMRException("Could not get handle to OMR server"));
        }
        return iomiHandle;
    }

    ///////////////////////////////////////////////////////////////////////////
    // Test methods

    public void test()
    {
        StringHolder returnInfoFromOMI = new org.omg.CORBA.StringHolder();
        int rc = 0;
        String options = "";
        String nameSpace = "SAS";
        try
        {
            setFoundationRepository();
            System.out.println("Types: " + getTypes());

            System.out.println("GetSubtypes: ");
            rc = getIOMI().GetSubtypes("SoftwareComponent", returnInfoFromOMI,
                    nameSpace, OMRConst.OMI_ALL_DESCENDANTS, options);
            System.out.println("returnCodeFromOMI = " + rc);
            System.out
                    .println("returnInfoFromOMI = " + returnInfoFromOMI.value);

            System.out.println("SAS Solutions installed components: ");
            List components = getSolutionsComponents();
            for (Iterator it=components.iterator(); it.hasNext(); )
            {
                GenericMetadataObject comp = (GenericMetadataObject)it.next();
                String installationLocation = comp.getAttribute(
                        "InstallationLocation");
                String classIdentifier = comp.getAttribute("ClassIdentifier");
                String productName = comp.getAttribute("ProductName");
                String softwareVersion = comp.getAttribute("SoftwareVersion");
                String vendor = comp.getAttribute("Vendor");
                String desc = comp.getAttribute("Desc");
                String isLicensed = comp.getAttribute("IsLicensed");
                String platform = comp.getAttribute("Platform");
                String name = comp.getName();
                String id = comp.getId();
                System.out.println("Name: " + name);
                System.out.println("    Id: " + id);
                System.out.println("    Desc: " + desc);
                System.out.println("    ProductName: " + productName);
                System.out.println("    InstallationLocation: " + 
                        installationLocation);
                System.out.println("    ClassIdentifier: " + classIdentifier);
                System.out.println("    SoftwareVersion: " + softwareVersion);
                System.out.println("    Vendor: " + vendor);
                System.out.println("    IsLicensed: " + isLicensed);
                System.out.println("    Platform: " + platform);
                List machines = doOMRQuery("Machine", 
                    "Machine[DeployedComponents/ConfiguredComponent[@name='" + 
                    name + "']]");
                if (machines.size() > 0)
                {
                    GenericMetadataObject machineInfo =
                        (GenericMetadataObject)machines.get(0);
                    String host = machineInfo.getName();
                    System.out.println("    Host: " + host);
                }
            }
            
            String typeName = "Login";
            System.out.println("Getting TypeProperties for " + typeName);
            rc = getIOMI().GetTypeProperties(typeName, returnInfoFromOMI, 
                    nameSpace, 0, "");
            System.out.println("GetTypeProperties() returned " + rc);
            System.out.println(returnInfoFromOMI.value);
        }
        catch (GenericError e)
        {
            throw new Error(e);
        }
    }
    
    public void dump(java.io.PrintStream out) throws Exception
    {
        String prevReposName = getRepositoryName();
        List repositories = getRepositories();
        Collections.sort(repositories, GenericMetadataObject.getComparator());
        for (Iterator reposIter=repositories.iterator(); reposIter.hasNext(); )
        {
            GenericMetadataObject gmoRepos = 
                (GenericMetadataObject)reposIter.next();
            String reposName = gmoRepos.getName();
            out.println("REPOSITORY \"" + reposName + "\" {");
            setRepository(reposName);
            List types = getTypes();
            Collections.sort(types, GenericMetadataObject.getComparator());
            for (Iterator typeIter=types.iterator(); typeIter.hasNext(); )
            {
                GenericMetadataObject gmoType = 
                    (GenericMetadataObject)typeIter.next();
                // DON'T call gmoType.getName() to get the type name,
                // see the Javadoc for getTypes().
                String typeName = gmoType.getId();
                typeName = typeName.trim();
                if (typeName.length() == 0)
                    continue;
                List objects = null;
                try
                {
                    objects = doOMRQuery(typeName, "", true);
                }
                catch (Throwable t)
                {
                    System.err.println("OMR query failed for reponsName: " +
                            reposName + " typeName: " + typeName + " error: " + 
                            t);
                    continue;
                }
                if (objects == null)
                    continue;
                Collections.sort(objects, 
                        GenericMetadataObject.getComparator());
                for (Iterator objIter=objects.iterator(); objIter.hasNext(); )
                {
                    GenericMetadataObject gmo = 
                        (GenericMetadataObject)objIter.next();
                    out.println(gmo);
                }
            }
            out.println("}"); // end of repository
        }
        if (prevReposName != null)
            setRepository(prevReposName);
    }
    
    public void testCRUD()
    {
        System.out.println("testCRUD");
        setFoundationRepository();
        // Create a new string Property
        List objects = doOMRQuery("PropertyType", "*[@Name='String']");
        GenericMetadataObject stringType = (GenericMetadataObject)objects.get(0);
        List typeAssoc = new ArrayList();
        typeAssoc.add(stringType);
        GenericMetadataObject gmo1 = new GenericMetadataObject();
        gmo1.setTypeName("Property");
        gmo1.setName("zzTestCRUD");
        gmo1.setAttribute("DefaultValue", "ABC");
        gmo1.setAssociation("OwningType", typeAssoc);
        System.out.println(gmo1);
        GenericMetadataObject gmo2 = create(gmo1);
        System.out.println(gmo2);
        GenericMetadataObject gmo3 = null;
        // Change an attribute value
        gmo2.setAttribute("DefaultValue", "123");
        store(gmo2);
        gmo3 = getObjectInfoById(gmo2.getTypeName(), gmo2.getId());
        System.out.println(gmo3);
        assert gmo3.getAttribute("DefaultValue").equals("123");
        // Create a Note object
        GenericMetadataObject note = new GenericMetadataObject();
        note.setTypeName("TextStore");
        note.setName("zzTestNote");
        note.setAttribute("StoredText", "This is a test note on a property.");
        note = create(note);
        // Associate the Note with the Property
        gmo2.setAssociation("Notes", note);
        store(gmo2);
        gmo3 = getObjectInfoById(gmo2.getTypeName(), gmo2.getId());
        System.out.println(gmo3);
        objects = gmo3.getAssociation("Notes");
        GenericMetadataObject gmo4 = (GenericMetadataObject)objects.get(0);
        System.out.println(gmo4);
        // Clean up
        delete(gmo2);
        delete(gmo4);
        // Verify some more
        assert objects != null;
        assert objects.size() == 1;
        if (! gmo4.getId().equals(note.getId()))
        {
            throw new Error("Setting association failed");
        }
    }

    // Access statistics

    public int getNumCreates()
    {
        return this.numCreates;
    }

    public int getNumQueries()
    {
        return this.numQueries;
    }

    public int getNumUpdates()
    {
        return this.numUpdates;
    }

    public int getNumDeletes()
    {
        return this.numDeletes;
    }

    /**
     * @param args
     */
    public static void main(String[] args) throws Exception
    {
        String host = System.getProperty("omr.host", "localhost");
        String port = System.getProperty("omr.port", "8561");
        String user = System.getProperty("omr.user", "sasadm");
        String password = System.getProperty("omr.password", "AdminAdmin1");
        String command = "test";
        if (args.length >= 1)
            command = args[0];
        OMRConnection omr = new OMRConnection(host, port, user, password);
        System.out.println("Connecting to OMR server: " + omr.getHost() + 
                ":" + omr.getPort());
        omr.connect();
        System.out.println("Command: " + command);
        if (command.equalsIgnoreCase("test"))
        {
            omr.test();
        }
        else if (command.equalsIgnoreCase("testCRUD"))
        {
            omr.testCRUD();
        }
        else if (command.equalsIgnoreCase("dump"))
        {
            omr.dump(System.out);
        }
        else
        {
            System.err.println("Unknown command: " + command);
        }
        omr.close();
    }

}
