/*
 * Created on Apr 12, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.omr;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This SAX Handler class is for parsing the result of the IOMI 
 * GetTypeProperties() method.
 * @author davhan
 */
public class PropertyListHandler extends DefaultHandler
{
    private static final int STATE_INIT = 0;
    private static final int STATE_IN_LIST = 1;
    private static final int STATE_AFTER_LIST = 2;
    
    private int state;
    private int levelsDeep;
    private String listTag;
    private int beginLevel;
    private List propertyNames;
    private List associationNames;
    
    public PropertyListHandler()
    {
        this(null);
    }

    public PropertyListHandler(String listTag)
    {
        super();
        this.state = STATE_INIT;
        this.levelsDeep = 0;
        this.listTag = listTag;
        this.beginLevel = -1;
        this.propertyNames = new ArrayList();
        this.associationNames = new ArrayList();
    }

    public String getListTag()
    {
        return listTag;
    }
    
    public List getPropertyNames()
    {
        return this.propertyNames;
    }
    
    public List getAssociationNames()
    {
        return this.associationNames;
    }
    
    public void startElement(String uri, String localName, String qName,
            Attributes attributes) throws SAXException 
    {
        ++this.levelsDeep;
        if (this.state == STATE_INIT)
        {
            if (this.listTag == null)
            {
                // Auto-detected list tag
                this.listTag = qName;
            }
            else if (!this.listTag.equalsIgnoreCase(qName))
            {
                return;
            }
            int numAttributes = attributes.getLength();
            for (int i=0; i<numAttributes; ++i)
            {
                String propertyName = attributes.getQName(i);
                this.propertyNames.add(propertyName);
            }
            this.state = STATE_IN_LIST;
            this.beginLevel = this.levelsDeep;
            return;
        }
        else if (this.state == STATE_AFTER_LIST)
        {
            return;
        }
        else if (this.state == STATE_IN_LIST && 
                 this.levelsDeep == (this.beginLevel + 1))
        {
            // If we reach this point we must be looking for elements whose tag
            // names are association names.
            this.associationNames.add(qName);
        }
    }
    
    public void endElement(String uri, String localName, String qName)
            throws SAXException 
    {
        --this.levelsDeep;
        if (this.state == STATE_IN_LIST && this.levelsDeep < this.beginLevel)
        {
            this.state = STATE_AFTER_LIST;
            return;
        }
    }
    
    public static PropertyListHandler parseXML(String inputXML)
    {
        PropertyListHandler handler = new PropertyListHandler();
        parseXML(handler, inputXML);
        return handler;
    }

    public static void parseXML(PropertyListHandler handler, String inputXML) 
    {
       try 
       {
          //create a new sax parser
          SAXParserFactory parserFactory = SAXParserFactory.newInstance();
          parserFactory.setNamespaceAware(false);
          parserFactory.setValidating(true);        
          SAXParser parser = parserFactory.newSAXParser();
                      
          // parse the input xml file
          if ((inputXML != null) && (inputXML.length() > 0))
          {
             StringReader sr = new StringReader(inputXML);          
             org.xml.sax.InputSource is = new org.xml.sax.InputSource(sr);
             parser.parse(is, handler);
             parser = null;
          }
       }     
       catch (Exception e)
       { 
          //if the parsing fails, then the process should be aborted
          throw new Error(e);
       } 
    }

    /**
     * @param args
     */
    public static void main(String[] args)
    {
        String inputXML = "<Properties ChangeState=\"\" Desc=\"\" Id=\"\" LockedBy=\"\" MetadataCreated=\"\" MetadataUpdated=\"\" Name=\"\" Password=\"\" UserID=\"\"><AccessControls/><AssociatedIdentity/><Changes/><DataSources/><Documents/><Domain/><Extensions/><ExternalIdentities/><Groups/><Implementors/><Keywords/><Libraries/><LocalizedAttributes/><Notes/><PrimaryPropertyGroup/><Properties/><PropertySets/><ResponsibleParties/><SourceTransformations/><SpecSourceTransformations/><SpecTargetTransformations/><TargetTransformations/><Timestamps/><Trees/><UsedByPrototypes/><UsingPrototype/><Variables/></Properties>";
        PropertyListHandler listHandler = parseXML(inputXML);
        System.out.println("List tag: " + listHandler.getListTag());
        System.out.println("Properties: " + listHandler.getPropertyNames());
        System.out.println("Associations: " + listHandler.getAssociationNames());
    }

}
