/*
 * Created on Aug 21, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.storedProcess;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.omr.GenericMetadataObject;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.util.BRMUtil;

public class StoredProcDirFinder 
{
    private static final Logger LOG = Logger.getLogger(
            StoredProcDirFinder.class);

    /**
      * Find the list of Stored Process directories. The candidate stored
      * process directories are all of the Directory objects associated with
      * a File or with a SASLibrary, in all repositories. The original
      * version filtered to exclude paths that look as though they might be
      * under "Program Files/SAS 9.1". I presume this means that the actual
      * requirement is to exclude the installed sample stored processes.
      */
    public static Collection findStoredProcDirs(OMRConnection omr,
            Properties props)
    {
        LOG.info("Searching for potential stored process dirs");
        String savedReposName = omr.getRepositoryName();
        List repositories = omr.getRepositories();
        Collection result = new HashSet();
        for (Iterator reposIter=repositories.iterator();
                reposIter.hasNext(); )
        {
            GenericMetadataObject gmo =
                (GenericMetadataObject)reposIter.next();
            omr.setRepository(gmo.getName());
            List dirsWithFiles = omr.doOMRQuery("Directory", "*[Files/*]");
            List dirsWithSASLibraries = omr.doOMRQuery("Directory", 
                    "*[UsedByPackages/SASLibrary]");
            List allDirObjs = new ArrayList();
            allDirObjs.addAll(dirsWithFiles);
            allDirObjs.addAll(dirsWithSASLibraries);
            result.addAll(getDirNamesFromObjs(allDirObjs));
        }
        if (savedReposName != null)
            omr.setRepository(savedReposName);
        else
            omr.setFoundationRepository(); // default
        LOG.info("Found " + result.size() + " potential stored process dirs");
        result = filterDirList(result, props);
        LOG.info("Found " + result.size() + " filtered stored process dirs");
        return result;
    }

    private static Collection getDirNamesFromObjs(List dirObjs)
    {
        Collection result = new ArrayList();
        for (Iterator dirObjIter=dirObjs.iterator(); dirObjIter.hasNext(); )
        {
            GenericMetadataObject gmo =
                (GenericMetadataObject)dirObjIter.next();
            String dirName = gmo.getAttribute("DirectoryName");
            if (dirName != null)
            {
                result.add(dirName);
            }
            else
            {
                LOG.error("Not a valid directory object: " + gmo);
            }
        }
        return result;
    }

    /**
      * Filter the list of "stored process" directories to be backed up.
      * Currently excluded are directories that look like they are under the
      * SAS installation directory, or directories beginning with
      * "!SASROOT".
      * 
      * @param dirList  Collection of Strings containing directory names
      * @return the filtered collection
      */
    private static Collection filterDirList(Collection dirList, 
            Properties props)
    {
        String sasExeDir = props.getProperty(Constants.SAS_EXE_DIR);
        sasExeDir = BRMUtil.normPath(sasExeDir);
        if (sasExeDir == null)
        {
            LOG.warn(Constants.SAS_EXE_DIR + " property is not set");
        }
        String noFilterSAS = props.getProperty("BRM.no.filter.SAS");
        if (noFilterSAS != null)
        {
            LOG.debug("BRM.no.filter.SAS set, including sample stored procs");
            sasExeDir = null;
        }
        Collection result = new ArrayList();
        for (Iterator dirIter=dirList.iterator(); dirIter.hasNext(); )
        {
            String dirName = (String)dirIter.next();
            if (dirName == null || dirName.length() == 0)
                continue;
            if (dirName.toLowerCase().startsWith("!sasroot"))
            {
                LOG.debug("Excluding \"" + dirName + 
                        "\" because it starts with !SASROOT");
                continue;
            }
            if (sasExeDir != null)
            {
                if (BRMUtil.normPath(dirName).startsWith(sasExeDir))
                {
                    LOG.debug("Excluding \"" + dirName +
                        "\" because it starts with SAS_EXE_DIR");
                    continue;
                }
            }
            result.add(dirName);
        }
        return result;
    }
}
