/*
 * Created on May 23, 2008
 * Copyright (c) 2008 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.tar;

import java.io.File;

import org.apache.tools.ant.Project;

import com.sas.solutions.brm.tar.Tar;
import com.sas.solutions.brm.tar.Untar;

/**
  * The main class for running the BRM Ant Tar code from a script.
  */
public class ScriptMain
{
    public static final String help = (
"Create or extract \"tar\" format archive file compressed with gzip.\n" +
"Uses the BRM-modified Ant Java tar library, so is compatible with BRM.\n" +
"\n" +
"Use examples:\n" +
"\n" +
"java com.sas.solutions.brm.tar.ScriptMain cfz sample.tar.gz -C C:/temp sasdav\n" +
"    -> Creates gzipped tar file sample.tar.gz, containing the sasdav\n" +
"    subdirectory of the C:/temp directory, and all its contents.\n" +
"\n" +
"java com.sas.solutions.brm.tar.ScriptMain xfz sample.tar.gz -C C:/temp\n" +
"    -> Extracts contents of gzipped tar file to C:/temp directory\n" +
"\n" +
"java com.sas.solutions.brm.tar.ScriptMain tfz sample.tar.gz\n" +
"    -> Lists contents of gzipped tar file.\n" +
"       File names with non-ASCII characters will be printed with escapes\n" +
"       in the format \\uNNNN, where NNNN is the hex unicode number.\n" +
"\n" +
"This utility does not implement all of the Unix tar command options.\n" +
"Only the options demonstrated above may be relied upon.\n" +
"");

    public static void createTarGz(String tarfile, String basedir, 
            String[] includes) throws Exception
    {
        Project project = new Project();
        project.setBaseDir(new File(basedir));
        Tar task = new Tar();
        task.setProject(project);
        task.setDestFile(new File(tarfile));
        task.setBasedir(new File(basedir));
        Tar.TarLongFileMode mode = new Tar.TarLongFileMode();
        mode.setValue(Tar.TarLongFileMode.GNU);
        task.setLongfile(mode);
        if (includes.length > 0)
        {
            StringBuffer sb = new StringBuffer();
            boolean first = true;
            for (int i=0; i<includes.length; ++i)
            {
                if (!first)
                    sb.append(",");
                else
                    first = false;
                sb.append(includes[i]);
            }
            task.setIncludes(sb.toString());
        }
        task.execute();
    }

    public static void extractTarGz(String tarfile, String basedir,
            String[] includes) throws Exception
    {
        if (includes.length > 0)
        {
            System.err.println("Warning: extractTarGz: ignoring includes:" +
                    includes);
        }
        Project project = new Project();
        project.setBaseDir(new File(basedir));
        Untar task = new Untar();
        task.setProject(project);
        task.setSrc(new File(tarfile));
        task.setDest(new File(basedir));
        task.setOverwrite(true);
        task.execute();
    }

    public static void listTarGz(String tarfile)
    {
        String basedir = ".";
        Project project = new Project();
        project.setBaseDir(new File(basedir));
        Untar task = new Untar();
        task.setProject(project);
        task.setSrc(new File(tarfile));
        task.setDest(new File(basedir));
        task.setListmode(true);
        task.execute();
    }

    /**
     * @param args
     */
    public static void main(String[] args) throws Exception
    {
        if (args.length >= 1 && (
            args[0].equalsIgnoreCase("-h") ||
            args[0].equalsIgnoreCase("--help") ||
            args[0].equalsIgnoreCase("/?") ||
            args[0].equalsIgnoreCase("/h")))
        {
            System.out.println(help);
            System.exit(0);
        }
        if (args.length < 2)
        {
            System.err.println("Missing options. Use -h for help.");
            System.exit(1);
        }

        String options = args[0];
        String tarfile = args[1];
        String basedir = ".";

        int optionsEnd = 2;
        if (args.length >= (optionsEnd + 2) && args[optionsEnd].equals("-C"))
        {
            basedir = args[optionsEnd + 1];
            optionsEnd += 2;
        }

        String[] includes = new String[args.length - optionsEnd];
        for (int i=0; i<includes.length; ++i)
        {
            includes[i] = args[optionsEnd + i];
        }
        if (options.startsWith("c"))
        {
            createTarGz(tarfile, basedir, includes);
        }
        else if (options.startsWith("x"))
        {
            extractTarGz(tarfile, basedir, includes);
        }
        else if (options.startsWith("t"))
        {
            listTarGz(tarfile);
        }
        else
        {
           System.err.println("Unknown command option:" + options);
           System.exit(1);
        }
    }
}

// end-of-file
