/*
 * Copyright  2000,2002,2004-2005 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

/*
 * This package is based on the work done by Timothy Gerard Endres
 * (time@ice.com) to whom the Ant project is very grateful for his great code.
 */
/*
 * THIS FILE WAS DERIVED FROM APACHE ANT, BUT HAS CHANGED.
 * Wed 04 Jun 2008 David Handy - Added escapeNonASCII(), charToHex4()
 * Wed 04 Jun 2008 David Handy - Made getNameBytes() and parseName() use UTF-8
 * Fri 06 Jun 2008 David Handy - Added decodeInputStream()
 */
package com.sas.solutions.brm.tar;

import java.io.InputStream;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CodingErrorAction;

/**
 * This class provides static utility methods to work with byte streams.
 *
 */
public class TarUtils {

    public static int byteToUnsigned(byte b) {
        return (0x100 + (int)b) & 0xff;
    }

    public static byte toSignedByte(long i) {
        if (i < 0 || i >= 0x100)
            throw new RuntimeException("Byte value overflow");
        return (i > 0x7f) ? (byte)(i - 0x100) : (byte)i;
    }

    /**
      * Parse an integer field from a header buffer. If the first byte of
      * the field has the upper bit set, then the rest of the field is
      * treated as an unsigned big-endian binary integer according to GNU
      * extensions. Otherwise, this method delegates to parseOctal().
      */
    public static long parseInt(byte[] header, int offset, int length) {
        if ((byteToUnsigned(header[offset]) & 0x80) == 0)
        {
            return parseOctal(header, offset, length);
        }
        long result = byteToUnsigned(header[offset]) & 0x7f;
        for (int i = offset + 1; i < offset + length; ++i) {
            long prevResult = result;
            result <<= 8;
            result += byteToUnsigned(header[i]);
            if (result < prevResult)
                throw new RuntimeException("Overflow reading integer from field");
        }
        return result;
    }

    /**
     * Parse an octal string from a header buffer. This is used for the
     * file permission mode value.
     *
     * @param header The header buffer from which to parse.
     * @param offset The offset into the buffer from which to parse.
     * @param length The number of header bytes to parse.
     * @return The long value of the octal string.
     */
    public static long parseOctal(byte[] header, int offset, int length) {
        long    result = 0;
        boolean stillPadding = true;
        int     end = offset + length;

        for (int i = offset; i < end; ++i) {
            if (header[i] == 0) {
                break;
            }

            if (header[i] == (byte) ' ' || header[i] == '0') {
                if (stillPadding) {
                    continue;
                }

                if (header[i] == (byte) ' ') {
                    break;
                }
            }

            stillPadding = false;
            result = (result << 3) + (header[i] - '0');
        }

        return result;
    }

    /**
     * Parse an entry name from a header buffer. Converts bytes to
     * characters using UTF-8 encoding.
     *
     * @param header The header buffer from which to parse.
     * @param offset The offset into the buffer from which to parse.
     * @param length The number of header bytes to parse.
     * @return The header's entry name.
     */
    public static StringBuffer parseName(byte[] header, int offset, int length) {
        // Exclude trailing zero bytes
        while ( (length > 0) && (header[offset + length - 1] == 0) )
            --length;
        try
        {
            return new StringBuffer(
                    new String(header, offset, length, "UTF-8"));
        }
        catch (java.io.UnsupportedEncodingException e)
        {
            throw new RuntimeException(e);
        }
    }

    /**
     * Determine the number of bytes in an entry name. Converts characters
     * to bytes using UTF-8 encoding.
     *
     * WARNING: Truncates name bytes if they don't fit within allowed output
     * space!
     *
     * @param name The header name from which to parse.
     * @param buf The buffer from which to parse.
     * @param offset The offset into the buffer from which to parse.
     * @param length The number of header bytes to parse.
     * @return The number of bytes in a header's entry name.
     */
    public static int getNameBytes(StringBuffer name, byte[] buf, int offset, int length) {
        byte[] nameBytes;
        try
        {
            nameBytes = name.toString().getBytes("UTF-8");
        }
        catch (java.io.UnsupportedEncodingException e)
        {
            throw new RuntimeException(e);
        }
        int i;

        for (i = 0; i < length && i < nameBytes.length; ++i) {
            buf[offset + i] = nameBytes[i];
        }

        for (; i < length; ++i) {
            buf[offset + i] = 0;
        }

        return offset + length;
    }

    /** Put an integer into the header buffer.
     * Use this method for fields that must end with two padding bytes, a
     * space and a zero. If the value can be represented in the space
     * available as ASCII octal digits, then that is how they are written
     * out. Otherwise, following the GNU extensions, the high bit is set on
     * the first byte, and all of the rest of the bits in the field will be
     * written as an unsigned, big-endian binary number, with no trailing
     * pad bytes.
     *
     * @param value The header value
     * @param buf The buffer into which bytes will be stored
     * @param offset The offset into the buffer at which to store
     * @param length The number of header bytes in the field
     * @return The new buffer offset (past the end of the field)
     */
    public static int getIntBytes(long value, byte[] buf, int offset, int length) {
        if (value < 0)
            throw new RuntimeException("Field value cannot be less than zero");
        int usableFieldBits = (length - 2) * 3;
        if (usableFieldBits >= 63 || value < (1L << usableFieldBits))
            return getOctalBytes(value, buf, offset, length);
        return getBinaryBytes(value, buf, offset, length);
    }

    /** Put an integer into the header buffer.
     * This method is just like getIntBytes(), except that it only stores
     * one terminating byte, a space. (Similar to the difference between
     * getOctalBytes() and getLongOctalBytes().)
     * 
     * @param value The header value
     * @param buf The buffer into which bytes will be stored
     * @param offset The offset into the buffer at which to store
     * @param length The number of header bytes in the field
     * @return The new buffer offset (past the end of the field)
     */
    public static int getLongIntBytes(long value, byte[] buf, int offset, int length) {
        if (value < 0)
            throw new RuntimeException("Field value cannot be less than zero");
        int usableFieldBits = (length - 1) * 3;
        if (usableFieldBits >= 63 || value < (1L << usableFieldBits))
            return getLongOctalBytes(value, buf, offset, length);
        return getBinaryBytes(value, buf, offset, length);
    }

    public static int getBinaryBytes(long value, byte[] buf, int offset, int length) {
        if (value < 0)
            throw new RuntimeException("Field value cannot be less than zero");
        for (int i = offset + length - 1; i >= offset; --i) {
            buf[i] = toSignedByte(value & 0xff);
            value >>= 8;
        }
        buf[offset] |= 0x80;
        return offset + length;
    }

    /** * Parse an octal integer from a header buffer.
     * 
     *
     * @param value The header value
     * @param buf The buffer from which to parse.
     * @param offset The offset into the buffer from which to parse.
     * @param length The number of header bytes to parse.
     * @return The integer value of the octal bytes.
     */
    public static int getOctalBytes(long value, byte[] buf, int offset, int length) {
        int    idx = length - 1;

        buf[offset + idx] = 0;
        --idx;
        buf[offset + idx] = (byte) ' ';
        --idx;

        if (value == 0) {
            buf[offset + idx] = (byte) '0';
            --idx;
        } else {
            for (long val = value; idx >= 0 && val > 0; --idx) {
                buf[offset + idx] = (byte) ((byte) '0' + (byte) (val & 7));
                val = val >> 3;
            }
        }

        for (; idx >= 0; --idx) {
            buf[offset + idx] = (byte) ' ';
        }

        return offset + length;
    }

    /**
     * Parse an octal long integer from a header buffer.
     *
     * @param value The header value
     * @param buf The buffer from which to parse.
     * @param offset The offset into the buffer from which to parse.
     * @param length The number of header bytes to parse.
     * @return The long value of the octal bytes.
     */
    public static int getLongOctalBytes(long value, byte[] buf, int offset, int length) {
        byte[] temp = new byte[length + 1];

        getOctalBytes(value, temp, 0, length + 1);
        System.arraycopy(temp, 0, buf, offset, length);

        return offset + length;
    }

    /**
     * Parse the checksum octal integer from a header buffer.
     *
     * @param value The header value
     * @param buf The buffer from which to parse.
     * @param offset The offset into the buffer from which to parse.
     * @param length The number of header bytes to parse.
     * @return The integer value of the entry's checksum.
     */
    public static int getCheckSumOctalBytes(long value, byte[] buf, int offset, int length) {
        getOctalBytes(value, buf, offset, length);

        buf[offset + length - 1] = (byte) ' ';
        buf[offset + length - 2] = 0;

        return offset + length;
    }

    /**
     * Compute the checksum of a tar entry header.
     *
     * @param buf The tar entry's header buffer.
     * @return The computed checksum.
     */
    public static long computeCheckSum(byte[] buf) {
        long sum = 0;

        for (int i = 0; i < buf.length; ++i) {
            sum += 255 & buf[i];
        }

        return sum;
    }

    /**
      * Convert a (full unicode) string to a string containing only ASCII
      * characters, with all non-printable or non-ASCII characters converted
      * to unicode escape sequences. The backslash character is doubled.
      */
    public static String escapeNonASCII(String s)
    {
        char[] charArray = s.toCharArray();
        StringBuffer result = new StringBuffer();
        for (int i=0; i<charArray.length; ++i)
        {
            char c = charArray[i];
            if (c == '\\')
            {
                result.append("\\\\");
                continue;
            }
            int n = (int)c;
            if (n >= 32 && n < 127)
            {
                result.append(c);
                continue;
            }
            result.append("\\u");
            result.append(charToHex4(c));
        }
        return result.toString();
    }

    /**
      * Convert a single character to a 4-digit hexidecimal string.
      * @param c  The character to be converted.
      */
    public static String charToHex4(char c)
    {
        final int base = 16;
        final char[] digitArray = {
            '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 
            'A', 'B', 'C', 'D', 'E', 'F'};
        char[] result = new char[4];
        for (int digit=result.length-1; digit>=0; --digit)
        {
            result[digit] = digitArray[c % base];
            c /= base;
        }
        return new String(result);
    }

    /**
      * Convert bytes in stream to string via UTF-8 decoding.
      * Make bufSize at least 4 bytes, because the decoder needs to buffer
      * at least the maximum number of bytes in a single UTF-8 character.
      * (Usually make the buffer bigger, i.e. 4K, for read efficiency).
      * @param inputStream  The input stream from which bytes are read
      * @param bufSize  The size in bytes of buffer for reading and converting
      */
    public static StringBuffer decodeInputStream(InputStream inputStream, 
            int bufSize)
        throws java.io.IOException
    {
        if (bufSize < 4)
        {
            throw new IllegalArgumentException(
                "Invalid buffer size: " + bufSize + " (should be at least 4)");
        }
        Charset cs = Charset.forName("UTF-8");
        CharsetDecoder decoder = cs.newDecoder();
        decoder.onMalformedInput(CodingErrorAction.REPLACE);
        decoder.onUnmappableCharacter(CodingErrorAction.REPLACE);
        ReadableByteChannel rbc = Channels.newChannel(inputStream);
        ByteBuffer bb = ByteBuffer.allocateDirect(bufSize);
        // UTF-8 can generate at most 1 char per byte, so char buf size
        // is the same as byte buf size.
        CharBuffer cb = CharBuffer.allocate(bufSize);
        StringBuffer result = new StringBuffer();
        while (true)
        {
            int bytesRead = rbc.read(bb);
            if (bytesRead < 0)
                break; // End Of File
            if (bytesRead == 0)
            {
                if (bb.hasRemaining())
                    break; // Treat zero read with space available as EOF
                throw new java.io.IOException("Buffer of " + bufSize + 
                        " bytes is full. Read failed.");
            }
            bb.flip();
            decoder.decode(bb, cb, false);
            cb.flip();
            result.append(cb.toString());
            cb.clear();
            bb.compact();
        }
        // At this point, there may still be some bytes in the buffer
        bb.flip();
        decoder.decode(bb, cb, true);
        decoder.flush(cb);
        cb.flip();
        result.append(cb.toString());
        return result;
    }
}
