/*
 * Created on Jun 23, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.messages.Messages;
import com.sas.solutions.brm.omr.OMRConnection;

public class BRMUtil
{
    private static final Logger LOG = Logger.getLogger(BRMUtil.class);

    private BRMUtil()
    {
        // a class with nothing but static utility methods
    }

    /**
     * Prefix the characters that are special in a regex parameter, so they
     * are not interpreted with their special meaning but instead are used
     * as literal match characters.
     * @param s
     * @return the escaped version of s
     */
    public static String escapeRegex(String s)
    {
        // Yes, we really need all of these backslashes. Honest.
        s = s.replaceAll("\\\\", "\\\\\\\\");
        s = s.replaceAll("\\[", "\\\\[");
        s = s.replaceAll("\\.", "\\\\.");
        s = s.replaceAll("\\^", "\\\\^");
        s = s.replaceAll("\\$", "\\\\$");
        s = s.replaceAll("\\+", "\\\\+");
        s = s.replaceAll("\\{", "\\\\{");
        s = s.replaceAll("\\(", "\\\\(");
        s = s.replaceAll("\\?", "\\\\?");
        s = s.replaceAll("\\*", "\\\\*");
        s = s.replaceAll("\\|", "\\\\|");
        return s;
    }

    /**
     * Prefix the characters that are special in a replacement parameter, so 
     * they are not interpreted with their special meaning but instead are
     * replaced literally.
     * 
     * @param s
     * @return the escaped version of s
     */
    public static String escapeRepl(String s)
    {
        // Yes, we really need all of these backslashes. Honest.
        s = s.replaceAll("\\\\", "\\\\\\\\");
        s = s.replaceAll("\\$", "\\\\$");
        return s;
    }

    /**
     * Escape XML special characters
     */
    public static String escapeXML(String s)
    {
        s = s.replaceAll("&", "&amp;");
        s = s.replaceAll("\"", "&quot;");
        s = s.replaceAll("<", "&lt;");
        s = s.replaceAll(">", "&gt;");
        return s;
    }

    /**
     * Determine whether the string looks like a Unix-style file or directory
     * path.
     * @param s
     * @return true if the string contains at least one forward slash and no
     *              backslashes.
     */
    public static boolean isUnixPath(String s)
    {
        boolean hasForwardSlash = (s.indexOf("/") >= 0);
        boolean hasBackwardSlash = (s.indexOf("\\") >= 0);
        return (hasForwardSlash && !hasBackwardSlash);
    }

    /**
     * @param origString
     * @param fromPrefix
     * @param toPrefix
     * @return
     */
    public static String replacePathPrefix(String origString, 
            String fromPrefix, String toPrefix)
    {
        char fromPrefixLastChar = fromPrefix.charAt(fromPrefix.length() - 1);
        boolean fromPrefixEndsWithSlash = (fromPrefixLastChar == '/') ||
                                          (fromPrefixLastChar == '\\');
        boolean toPrefixIsUnixPath = isUnixPath(toPrefix);
        char toPrefixLastChar = toPrefix.charAt(toPrefix.length() - 1);
        boolean toPrefixEndsWithSlash = (toPrefixLastChar == '/') || 
                                        (toPrefixLastChar == '\\');
        if (fromPrefixEndsWithSlash && !toPrefixEndsWithSlash)
        {
            // Add a slash (forward or backward, *as appropriate*) to the end
            // of toPrefix. This will fix problems with disappearing directory
            // separator characters.
            if (toPrefixIsUnixPath)
                toPrefix = toPrefix + "/";
            else
                toPrefix = toPrefix + fromPrefixLastChar;
        }
        else if (!fromPrefixEndsWithSlash && toPrefixEndsWithSlash)
        {
            // Remove a trailing slash from toPrefix to avoid creating bogus
            // doubled slashes in the resulting paths.
            toPrefix = toPrefix.substring(0, toPrefix.length() - 1);
        }
        Pattern p = Pattern.compile("\\b" + escapeRegex(fromPrefix) + "\\b");
        Matcher m = p.matcher(origString);
        String result = null;
        if (m.find())
        {
            int prefixStart = m.start();
            int prefixEnd = m.end();
            StringBuffer sb = new StringBuffer();
            sb.append(origString.substring(0, prefixStart));
            sb.append(toPrefix);
            String tail = origString.substring(prefixEnd);
            // If toPrefix contains at least one forward slash, and contains no
            // backslashes, then convert the entire result part from the end of
            // the prefix till the end of the string from backslashes to 
            // forward slashes.
            if (toPrefixIsUnixPath)
            {
                tail = tail.replace('\\', '/');
            }
            sb.append(tail);
            result = sb.toString();
        }
        else
        {
            result = origString;
        }
        return result;
    }

    /**
     * Get the SAS Solutions Config directory. If the BRM.SSC.dir system 
     * property is defined, then that is used, else the config directory is
     * assumed to be two directories up from the current directory.
     * 
     * @return The absolute directory pathname
     */
    public static String getSSCDir()
    {
        String SSCdir = System.getProperty("BRM.SSC.dir");
        if (SSCdir == null || SSCdir.equals(""))
        {
            SSCdir = "../../";
        }
        SSCdir = new File(SSCdir).getAbsolutePath();
        return SSCdir;
    }

    /**
      * Load the local installation properties files. The returned
      * properties object contains all the accumulated property values from
      * all the files found. If the same property exists in more than one
      * file, the property value from the file loaded first is the one used.
      *
      * As of 27 Nov 2007, this method attempts to load the following files,
      * in this order:
      *     soltnsdata.properties
      *     soltnsmid.properties
      *     mysqldbs.properties
      *
      * @return a Properties object containing all properties from all files
      *         found.
      * @throws BRMConfigurationException if there is an error loading a
      *         file, or if no files are found at all. It is not an error if
      *         one or more files don't exist, because not all install
      *         properties files will be on all machines. But each machine
      *         (that we care about) should have at least one.
      */
    public static Properties loadInstallPropertyFiles()
        throws BRMConfigurationException
    {
        Properties configProperties = new Properties();
        boolean somePropertyFileFound = false;
        for (int i=0; i<Constants.INSTALL_PROPERTIES_FILES.length; ++i)
        {
            if (!loadInstallPropertyFile(Constants.INSTALL_PROPERTIES_FILES[i],
                        configProperties))
            {
                continue;
            }
            somePropertyFileFound = true;
        }
        if (!somePropertyFileFound)
        {
            throw new BRMConfigurationException(
                "No install properties files in SAS configuration directory: " 
                + BRMUtil.getSSCDir());
        }
        return configProperties;
    }

    /**
      * Load a local installation properties file. The properties in that
      * file are added to the Properties object passed as a parameter to
      * this method. If the property already exists in the properties
      * object, it is *not* overwritten.
      * @param propFileName  The name of a property file, without the path.
      * @param props  The Properties object into which the file will be loaded.
      * @return true if the config file is found, or false if that properties 
      *         file does not exist.
      * @throws BRMConfigurationException if there is an error loading the
      *         file.
      */
    public static boolean loadInstallPropertyFile(String propFileName,
            Properties props)
        throws BRMConfigurationException
    {
        String SSCdir = BRMUtil.getSSCDir();
        File configFile = new File(SSCdir, propFileName);
        if (!configFile.isFile())
        {
            // File does not exist
            return false;
        }
        loadPropFile(configFile, props);
        return true;
    }

    /**
      * Like loadPropFile(), except it loads a resource instead of a file.
      */
    public static void loadPropFileAsResource(String resourceName,
            Properties properties)
        throws BRMConfigurationException
    {
        URL url = ClassLoader.getSystemResource(resourceName);
        if (url == null)
        {
            throw new BRMConfigurationException(
                    "Resource " + resourceName + " not found");
        }
        File propFile = new File(url.getFile());
        loadPropFile(propFile, properties);
    }

    /**
      * Load a local properties file. The properties in that file are added
      * to the Properties object passed as a parameter to this method. If
      * the property already exists in the properties object, it is *not*
      * overwritten.
      * @param propFile The property file. It must exist.
      * @param properties The Properties object into which the file is loaded.
      * @throws BRMConfigurationException if there is an error loading the
      *         file.
      */
    public static void loadPropFile(File propFile, Properties properties)
        throws BRMConfigurationException
    {
        LOG.info("Loading " + propFile.getAbsolutePath());
        FileInputStream fis = null;
        Properties newProps = new Properties();
        try
        {
            fis = new FileInputStream(propFile);
            newProps.load(fis);
        }
        catch(IOException e)
        {
            LOG.error("Loading " + propFile, e);
            throw new BRMConfigurationException(
                    Messages.getString("cmd.file.not.found"));
        }
        finally
        {
            try
            {
                if (fis != null)
                    fis.close();
            }
            catch (Exception e)
            {
                // Ignore errors on closing stream
            }
        }
        // Merge the new property values into the existing Properties obj
        // If the property was previously set, do not overwrite it
        for (Enumeration e=newProps.propertyNames(); e.hasMoreElements(); )
        {
            String propName = (String)e.nextElement();
            if (properties.getProperty(propName) == null)
                properties.setProperty(propName,
                        newProps.getProperty(propName));
        }
    }

    /**
      * Filter an array of command-line arguments, removing and executing
      * options of the form "-Dname=value".
      * @return the command-line arguments minus the property settings
      */
    public static String[] filterCmdLineProps(String[] args)
    {
        List result = new ArrayList();
        for (int i=0; i<args.length; ++i)
        {
            if (args[i].startsWith("-D"))
            {
                String name = null;
                String value = null;
                int eqIndex = args[i].indexOf("=");
                if (eqIndex >= 0)
                {
                    name = args[i].substring(2, eqIndex);
                    value = args[i].substring(eqIndex + 1);
                }
                else
                {
                    name = args[i].substring(2);
                    value = "";
                }
                if (name.length() > 0)
                {
                    System.setProperty(name, value);
                    continue;
                }
            }
            result.add(args[i]);
        }
        String[] resultArray = new String[result.size()];
        int argIndex = -1;
        for (Iterator argIter=result.iterator(); argIter.hasNext(); )
        {
            resultArray[++argIndex] = (String)argIter.next();
        }
        return resultArray;
    }

    public static String stringArrayToString(String[] args)
    {
        StringBuffer sb = new StringBuffer();
        sb.append('[');
        for (int i=0; i<args.length; ++i)
        {
            if (i > 0)
            {
                sb.append(", ");
            }
            sb.append(args[i]);
        }
        sb.append(']');
        return sb.toString();
    }

    /**
      * Mangle a file or directory pathname such that it can be used as a
      * valid file name (but a user can still recognize the directory names
      * as parts of the file name).
      */
    public static String fusePath(String pathName)
    {
        String fusedPath = pathName.replaceAll("[\\\\/|:]", "-");
        fusedPath = fusedPath.replaceAll("\\s", "_");
        return fusedPath;
    }

    /**
      * Change the filename portion of a file path, leaving the rest alone.
      * This method should work regardless of what kind of slashes are used
      * in the pathName parameter.
      *
      * Examples:<br/>
      *   changeFileNameInPath("a/b/c", "d") returns "a/b/d"
      *   changeFileNameInPath("a/b/", "d") returns "a/b/d"
      *   changeFileNameInPath("/a", "d") returns "/d"
      *   changeFileNameInPath("a", "d") returns "d"
      */
    public static String changeFileNameInPath(String pathName, 
            String newFileName)
    {
        int i = pathName.lastIndexOf("/");
        if (i < 0)
            i = pathName.lastIndexOf("\\");
        // If still (i < 0), this algorithm replaces the whole string,
        // which makes as much sense as anything else.
        return pathName.substring(0, i + 1) + newFileName;
    }

    /**
      * @param s A String
      * @return true iff The string parameter is non-null and equal
      * (case-insensitive) to "true", "yes", or "on", compatible with the
      * Ant definition of truth.
      */
    public static boolean isTrueStr(String s)
    {
        if (s == null || s.length() == 0)
            return false;
        s = s.toLowerCase();
        if (s.equals("true"))
            return true;
        if (s.equals("yes"))
            return true;
        if (s.equals("on"))
            return true;
        return false;
    }

    public static void dumpObjInfo(Object obj, PrintStream out)
    {
        try
        {
            Class contextClass = obj.getClass();
            out.println(contextClass.toString());
            Method[] methods = contextClass.getDeclaredMethods();
            for (int methodIndex=0; methodIndex<methods.length; methodIndex++)
            {
                Method m = methods[methodIndex];
                out.print(m.toString());
                if (m.getName().startsWith("get") &&
                        m.getParameterTypes().length==0 &&
                        !(m.getReturnType().equals(void.class)))
                {
                    out.print("=");
                    out.print(m.invoke(obj, new Object[]{}));
                }
                out.println();
            }
        }
        catch (Exception e)
        {
            throw new Error(e);
        }
    }

    public static Object getDeclaredFieldValue(Object obj, String fieldName)
    {
        try
        {
            Class contextClass = obj.getClass();
            Field[] fields = contextClass.getDeclaredFields();
            for (int i=0; i<fields.length; ++i)
            {
                Field field = fields[i];
                if (field.getName().equals(fieldName))
                {
                    return field.get(obj);
                }
            }
            return null;
        }
        catch (Exception e)
        {
            throw new Error(e);
        }
    }

    public static OMRConnection getNewOMRConnection(ConfigFile controlConfig, 
            Properties installProps)
        throws BRMConfigurationException
    {
        String host = controlConfig.get(
                Constants.METADATA_COMPONENT);
        if (host == null || host.length() == 0)
            host = installProps.getProperty("OMI_HOST");
        String port = controlConfig.get("port");
        if (port == null || port.length() == 0)
            port = installProps.getProperty("OMI_PORT");
        String user = controlConfig.get("user");
        if (user == null || user.length() == 0)
            user = installProps.getProperty("OMI_USER_ID");
        String password = controlConfig.get("password");
        if (password == null || password.length() == 0)
            password = installProps.getProperty("OMI_USER_PASSWORD");
        OMRConnection omr = new OMRConnection(host, port, user, password);
        omr.connect();
        return omr;
    }

    /**
      * Return a file or directory path that has been converted to an
      * common format such that it can be compared with other normalized
      * paths. The normalized path is absolute and contains no "." or ".."
      * elements. On Windows, the path is converted to lower case.
      */
    public static String normPath(String pathName)
    {
        return normPath(pathName, File.separatorChar);
    }

    public static String normPath(String pathName, char separatorChar)
    {
        if (pathName == null)
            return null;
        if (separatorChar == '\\')
        {
            // Special case for Windows
            try
            {
                return ((new File(pathName)).getCanonicalPath()).toLowerCase();
            }
            catch (IOException e)
            {
                // Ignore the error in getCanonicalPath(), fall back to our
                // own algorithm after normalizing the case. (We call
                // getAbsolutePath() first so that when it gets called again
                // it won't add a capitalized drive letter.)
                pathName = (new File(pathName)).getAbsolutePath();
                pathName = pathName.toLowerCase();
                // The fallback algorithm can't handle UNC paths.
                if (pathName.length() >= 2 &&
                        pathName.charAt(0) == '\\' &&
                        pathName.charAt(1) == '\\')
                    throw new Error(e);
            }
        }
        // We don't use File.getCanonicalPath() here because on Unix it
        // dereferences symbolic links. We want to preserve symbolic links.
        // This "if" statement is for testing separatorChar='/' on Windows.
        if (pathName.length() == 0 || pathName.charAt(0) != separatorChar)
            pathName = (new File(pathName)).getAbsolutePath();
        List parts = splitPath(pathName);
        List result = new ArrayList();
        // This "if" statement is for testing separatorChar='/' on Windows.
        if (parts.size() > 0 && ((String)parts.get(0)).length() == 0)
            result.add("");
        for (Iterator it=parts.iterator(); it.hasNext(); )
        {
            String part = (String)it.next();
            if (part.length() == 0)
                continue;
            if (part.equals("."))
                continue;
            if (part.equals(".."))
            {
                if (result.size() > 1)
                    result.remove(result.size() - 1);
                continue;
            }
            result.add(part);
        }
        if (result.size() < 2)
            result.add("");
        return joinPath(result, separatorChar);
    }

    /**
      * Split a file or directory path into a list containing its component
      * elements without the separators. Both / and \ are treated as path
      * separators. The first element of a split path beginning with / or \
      * will be an empty string (this is a feature). splitPath() and
      * joinPath() are opposites of each other.
      */
    public static List splitPath(String path)
    {
        path = path.replace('\\', '/');
        List parts = new ArrayList();
        int i = 0;
        int j = 0;
        for ( ; i <= path.length(); i = j + 1)
        {
            j = path.indexOf('/', i);
            if (j < 0)
                j = path.length();
            parts.add(path.substring(i, j));
        }
        return parts;
    }

    /**
      * Join a list of path elements into a path string. Each element is
      * separated in the string by separatorChar.  If the list has only 0 or
      * 1 elements, then no separatorChar is added. splitPath() and
      * joinPath() are opposites of each other.
      */
    public static String joinPath(List parts, char separatorChar)
    {
        StringBuffer sb = new StringBuffer();
        boolean first = true;
        for (Iterator it=parts.iterator(); it.hasNext(); )
        {
            if (first)
                first = false;
            else
                sb.append(separatorChar);
            sb.append(it.next());
        }
        return sb.toString();
    }

    /**
      * Join a list of path elements by the default separator char.
      */
    public static String joinPath(List parts)
    {
        return joinPath(parts, File.separatorChar);
    }

    /**
      * Compare two lists and return the number of first elements that are
      * common between them. This implementation works most efficiently with
      * ArrayList instances.
      */
    public static int compareListPrefixes(List list1, List list2)
    {
        int min = list1.size();
        if (list2.size() < min)
            min = list2.size();
        int i = 0;
        for ( ; i<min; ++i)
        {
            Object o1 = list1.get(i);
            Object o2 = list2.get(i);
            if (!o1.equals(o2))
                break;
        }
        return i;
    }
}
