package com.sas.swing.util;

import com.sas.swing.util.Action;
import com.sas.swing.util.ActionList;

import java.util.ArrayList;
import java.util.Iterator;
import javax.swing.SwingUtilities;

/**
 * Miscellaneous utility methods for working with <code>Actions</code>,
 * including an iterator for traversing an <code>ActionList</code> hierarchy.
 *
 * @author Ryan Norris
 */
public final class ActionUtils
{

public static final String RB_KEY = "ActionUtils.";

   /**
    *  Returns an iterator over the <code>Actions</code> in a given
    *  <code>ActionList</code>.
    *  This iterator differs from the one returned by
    *  <code>ActionList.getActions().iterator()</code> in that it will
    *  iterate over nested <code>ActionLists</code>.
    *  @param actionList
    *     the <code>ActionList</code> to return an iterator for
    *  @return
    *     an iterator over the actions in the given list in proper sequence
    */
   public static Iterator iterator (ActionList actionList)
   {
      return new ActionListIterator (actionList);
   }

   /**
    * Asynchronously executes the specified action.  This method uses
    * <code>SwingUtilities.invokeLater(Runnable)</code> to asynchronously
    * execute the action.
    * @param action the action to execute
    * @see javax.swing.SwingUtilities#invokeLater(java.lang.Runnable)
    */
   public static void invokeLater(final Action action)
   {
      SwingUtilities.invokeLater(new Runnable() {
         public void run() {
            action.performAction();
         }
      });
   }


   // private constructor to prevent instantiation
   private ActionUtils ()
   {
   }


private static class ActionListIterator
   implements Iterator
{
   ActionNode _root;
   ActionNode _next;

   public ActionListIterator(ActionList list)
   {
      _root = new ActionNode(list);
      _next = _root;
   }

   public boolean hasNext()
   {
      return (_next != null);
   }

   public Object next()
   {
      ActionNode next = _next;
      _next = _next.getNext();
      return next.getAction();
   }

   public void remove()
   {
      throw new UnsupportedOperationException();
   }

   private class ActionNode
   {
      Action _action;
      ActionNode _parent;
      int _indexInParent;
      boolean _isSeparator;

      java.util.ArrayList _children;

      public ActionNode(ActionList root)
      {
         this(root, null, 0);
      }

      public ActionNode(Object action, ActionNode parent, int indexInParent)
      {
//		  System.out.println("***ActionUtils: ActionNode(): action = " + action );
//		  if ( action != null ) {
//			  Class aclass = action.getClass();
//			  System.out.println("******ActionUtils: ActionNode(): action class = " + aclass );
//			  if ( aclass != null ) {
//				  System.out.println("*****ActionUtils: ActionNode(): action class name = " + aclass.getName() );
//			  }
//		  }

         if (action instanceof Action)
            _action = (Action)action;
         else if (action == ActionList.SEPARATOR)
            _isSeparator = true;
         else
            throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "invalidAction.txt"));

         _parent = parent;
         _indexInParent = indexInParent;

         // delayed adding children until requested.  this is now done in getNext().
//         if (_action != null && _action instanceof ActionList)
//            addChildren((ActionList)_action);
      }

      public Object getAction()
      {
         if (_isSeparator)
            return ActionList.SEPARATOR;
         return _action;
      }

      private void addChildren(ActionList aList)
      {
         java.util.List list = aList.getActions();
         if (list != null && !list.isEmpty())
         {
            _children = new java.util.ArrayList();

            for (int i=0, cnt=list.size(); i<cnt; i++)
               _children.add(new ActionNode(list.get(i), this, i));
         }
      }

      // pre-order iteration
      public ActionNode getNext()
      {
         // add children if this action has any
         if (_children == null && _action != null && _action instanceof ActionList)
            addChildren((ActionList)_action);

         if (_children != null && !_children.isEmpty())
            return (ActionNode)_children.get(0);
         if (_parent != null)
            return _parent.getNext(this);
         return null;
      }

      public ActionNode getNext(ActionNode node)
      {
         if (node != null && !_children.isEmpty() && (node._indexInParent + 1) < _children.size())
            return (ActionNode)_children.get(node._indexInParent + 1);
         if (_parent != null)
            return _parent.getNext(this);
         return null;
      }
   }
}

}