package com.sas.table;

/**
 * Interface for implementing controllers to edit models for tables.
 * When specifying columnNames when using the EditControllerInterface,
 * it is best to implement the methods such that they use the actual
 * column name from the model and not the column label to avoid
 * conflicts between the two.
 *
 * @author sasccf Chad Ferguson
 * @date 10Jul2001
 */

public interface EditControllerInterface
{
	/**
	 * The index to used when inserting and the default location (extending
	 * the table).
	 */
	public static final int DEFAULT_INDEX = -1;

	/**
	 * Used to set the insert mode of the controller. Allow a new
	 * row to be inserted before each row in the table.
	 */
	public static final int INSERT = 1;

	/**
	 * Used to set the insert mode of the controller. Allow a new
	 * row to be inserted at the default location (usually the end)
	 * of the table.
	 */
	public static final int EXTEND = 2;

	/**
	 * Enables or disables editing on the TableView.
	 * @param edit whether to enable editing on the TableView
	 */
	public void setEditEnabled(boolean edit);

	/**
	 * Returns whether editing has been enabled for the TableView.
	 * @return whether editing has been enabled for the TableView
	 */
	public boolean isEditEnabled();

	/**
	 * Returns the columns in which editing is allowed. The
	 * <var>editableColumns</var> contains all the column names in which editing is
	 * allowed on the model, all other columns are to be non-editable. Returns
	 * null if all columns are allowed to be edited. An empty
	 * array is returned if no columns are to be edited.
	 * @return An array of column names in which editing is allowed.
	 */
	public String[] getEditableColumns();

	/**
	 * Sets the columns in which editing is allowed. The <var>editableColumns</var>
	 * contains all the column names in which editing is allowed on the model,
	 * all other columns are to be non-editable. If a null is passed, then all
	 * columns will be editable. If an empty array is passed, then no columns
	 * will be editable. The default is that all columns are allowed to be
	 * edited.
	 * @param columnNames An array of column names in which editing is allowed.
	 */
	public void setEditableColumns(String columnNames[]);

	/**
	 * Returns a boolean indicating whether the column is allowed
	 * to be edited. The default is that all columns are allowed to be edited.
	 * @param columnName The name of the column to query about whether it is
	 * allowed to be edited.
	 * @return A boolean indicating whether the column is
	 * allowed to be edited.
	 */
	public boolean isEditableColumn(String columnName);

	/**
	 * Returns the insert mode for the controller. Valid values are
	 * EditControllerInterface.INSERT and EditControllerInterface.EXTEND.
	 * If mode is EditControllerInterface.INSERT, then the controller will
	 * allow rows to be inserted before each row in the table. If mode is
	 * EditControllerInterface.EXTEND, then the controller will allow rows
	 * to be inserted at the default location (usually the end) of the
	 * table.
	 * @return the insert mode for the controller
	 */
	public int getInsertMode();

	/**
	 * Sets the insert mode for the controller. Valid values are
	 * EditControllerInterface.INSERT and EditControllerInterface.EXTEND.
	 * If mode is EditControllerInterface.INSERT, then the controller allows
	 * a new row to be inserted before each row in the table. If mode is
	 * EditControllerInterface.EXTEND, then the controller allow a new row
	 * to be inserted at the default location (usually the end) of the
	 * table.
	 * @param mode the type of inserting to allow on the controller
	 */
	public void setInsertMode(int mode);
}