package com.sas.table;

/**
 * An interface that adds the ability to get formatted data from the model. 
 * Viewers may want to display the data in the formatted state but have the raw
 * data to do actual calculations.
 * 
 * @author Drew Anderson
 * @version 1.0
 * @history
 * <dl>
 * <dt>6/02 <dd>Created.
 * </dl>
 */
public interface FormattedInterface
{
	/**
	 * Returns the format that is being used for the given column in the model. 
	 * Format could be anything from a SAS format to a JDBC format. User can use this
	 * to help format data on their own if they have the raw data. 
	 * @param columnIndex the column index in the model that the format is to be returned for
	 * @return the format that is applied to the raw data
	 */
	public String getFormat(int columnIndex);

	/**
	 * Returns the formatted data for the given cell coordinates. 
	 * @param rowIndex the row whose value is to be queried
	 * @param columnIndex the column whose value is to be queried
	 * @return the formatted value of the data at the specified cell
	 */
	public String getFormattedValueAt(int rowIndex, int columnIndex);

	/**
	* Returns the unformatted data for the given cell coordinates. 
	* @param rowIndex the row whose value is to be queried
	* @param columnIndex the column whose value is to be queried
	* @return the unformatted value of the data at the specified cell
	*/
	public Object getUnformattedValueAt(int rowIndex, int columnIndex);

	/**
	 * Sets whether the model is using formatted data or just the raw data. 
	 * A call to getFormattedVaueAt(int, int) should ignore this variable and
	 * always return the formatted data when possible. The getColumnClass(int) method
	 * should pay attention to this value and return the column class based on whether
	 * formatted or unformatted data is being used. 
	 * @param useFormattedData whether to use formatted data by default versus raw data
	 * @see #isFormattedDataUsed()
	 */
	public void setFormattedDataUsed(boolean useFormattedData);

	/**
	 * Returns whether the model is using formatted data or just the raw data. 
	 * A call to getFormattedVaueAt(int, int) should ignore this variable and
	 * always return the formatted data when possible. The getColumnClass(int) method
	 * should pay attention to this value and return the column class based on whether
	 * formatted or unformatted data is being used. 
	 * @return true if formatted data is being used or false otherwise
	 * @see #setFormattedDataUsed(boolean)
	 */
	public boolean isFormattedDataUsed();

	/**
	 * Returns the type of raw data that exists in the given column. This could be used for 
	 * justifying data left or right in a table even when all formatted data is being displayed.
	 * @param columnIndex the column whose raw data type is being queried
	 * @return the class type for the raw data of the given column, if the class is unknown, Object.class
	 * should probably be returned.
	 */
	public Class getUnformattedColumnClass(int columnIndex);

}