// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.table;

 /**
  * Defines an interface for accessing a table's row labels.
  * This will usually be implemented in tandem with the primary interface to 
  * tables, <A HREF="StaticTableInterface.html">StaticTableInterface</A>.
  * @author saskjp Ken Peressini
  * @version 1.0
  * @history
  * <dl>
  * <dt>Fri October 17, 1997  <dd>Split from RowLabelInterface (saskjp).
  * </dl>
  * @see com.sas.table.StaticTableInterface
  * @see com.sas.table.StaticColumnLabelInterface
  * @see com.sas.table.RowLabelInterface
  */

 public interface StaticRowLabelInterface
    {

    /**
      * Returns the label for a row.  The row is identified by a given index.
      * <p>
      * When isRowsLabelSupported() is true, <code>getRowLabel(0)</code> can
      * be used to retrieve the label for the entire set of row labels.
      * @param index 
      *    A one-based index that identifies the row.
      *    See note about use of zero above.
      * @return
      *    The specified row's label.
      * @exception TableException 
      *    Thrown if the query can not be satisfied (e.g. i/o-failure).
      * @exception IndexOutOfBoundsException 
      *    Thrown if <var>index</var> is negative, zero, or greater than
      *    the number of rows.  See note about use of zero above.
	  * @see #isRowLabelsSupported
      * @see #isRowsLabelSupported
     */
    public Object getRowLabel (int index) throws TableException;
    

    /**
      * Returns the labels for a range of rows.  The range is identified by a
      * given index which defines a starting row and by a given count
      * which defines the extent of the range.  The set of labels
      * is returned as an array with each element of the array 
      * corresponding to a label in the range. 
      * @param startIndex 
      *    A one-based index which identifies the start of the range.
      * @param count
      *    The number of labels to return (the extent of the range).
      * @return
      *    The specified rows' labels.
      * @exception TableException 
      *    Thrown if the query can not be satisfied (e.g. i/o-failure).
      * @exception IndexOutOfBoundsException 
      *    Thrown if <var>startIndex</var> or <var>count</var> is negative, zero, or 
      *    greater than the number of rows; except when getRowCount() has 
      *    returned -1 -- in that case, the implementation should return an array 
      *    whose length is less than <var>count</var> and the caller should be 
      *    prepared for this.
      * @see com.sas.table.StaticTableInterface#getRowCount
	  * @see #isRowLabelsSupported
     */
    public Object[] getRowLabels (int startIndex, int count) throws TableException;


    /**
      * Indicates whether getRowLabel() and getRowLabels() should be called.
      * Normally the act of implementing this interface indicates that labels are
      * available, but some implementations may not always have labels available.
      * @return
      *    <code>true</code> if labels are available, 
      *    and <code>false</code> otherwise.
      * @exception TableException 
      *    Thrown if the query can not be satisfied (e.g. i/o-failure).
	 */
	public boolean isRowLabelsSupported () throws TableException;


    /**
      * Indicates whether there is support for a row labels label, 
      * that is a label for the set of row labels taken as a whole.  
      * For example, a SAS dataset model might use "OBSERVATION".
      * If one is supported, <code>getRowLabel(0)</code> can be used to access it.
      * @return
      *    <code>true</code> if a row labels label is supported, 
      *    and <code>false</code> otherwise.
      * @exception TableException 
      *    Thrown if the query can not be satisfied (e.g. i/o-failure).
      * @see #getRowLabel
     */
    public boolean isRowsLabelSupported () throws TableException;

    }

