package com.sas.text;

import java.text.MessageFormat;
import java.util.ResourceBundle;

/**
 * Format a MessageFormat object.
 * <p>
 * This class provides a convenient way to format a MessageFormat
 * to a string, as well as a convenience for getting the
 * format pattern from a ResourceBundle.
 * <p>
 * There are four groups of methods provided here:
 * <ol>
 * <li>Constructors. For example,
 * <pre>
 * new Message("{0,number,integer} {1} cards selected")
 * new Message(resources, "MyClass.resultMessage", arg0, arg1)
 * new Message(resources, "MyClass.resultMessage", argArray)
 * </pre>
 * <li>methods to format a Message object into a String. For example,
 * <pre>
 * message.toString()
 * message.toString(arg0, new Integer(n), suit)
 * </pre>
 * <li>static methods to format using a String pattern, for example
 * <pre>
 * Message.format("{0,number,integer} {1} cards selected", new Integer(n), suit);
 * Message.format("{0,number,integer} {1} cards selected", argArray);
 * </pre>
 * <li>static methods to format using a String resource obtained from a ResourceBundle,
 * <pre>
 * Message.format(resources, "MyClass.resultMessage", new Integer(n), suit);
 * Message.format(resources, "MyClass.resultMessage", argArray);
 * </pre>
 * </ol>
 * Within these groups, the values to substitute into the format pattern can be
 * be passed as an Object array, or if you do not have the values in an
 * an array, there are six variants of each method, taking from one to six
 * Object parameters.  If you have more than six parameters, you must use
 * the form that takes an Object array. Use
 * <pre>new Object[] { arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7}
 * </pre>
 * if you have eight parameters, for example.
 * @author <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 *
 * <!--
 * @support sasdjb
 * @history Created                                    sasdjb 05Dec96
 * -->
 */
public final class Message
       extends MessageFormat
{
  Object args[];
  transient StringBuffer buffer;

  /**
   * Create a Message (MessageFormat) object with the pattern
   * @param pattern the MessageFormat pattern text
   */
  public Message(String pattern)
    {
      this(pattern, null);
    }


  /**
   * Set the args property.
   * @param newValues the new value for the args property.
   */
  public void setValues(Object newValues[])
    {
      args = newValues;
    }

  /**
   * Replace an argument that is supplied when formatting the message.
   * @param i an argument index, from 0 to getValues().length-1
   * @param value a new value to substitute for {i} in the message.
   */
  public void setValues(int i, Object arg)
    {
      args[i] = arg;
    }

  /**
   * Return the value of the args property.
   * @return the value of the args property.
   */
  public Object[] getValues()
    {
      return args;
    }

  /**
   * Get an argument that is substituted into the pattern when formatting.
   * @param i an argument index, from 0 to getValues().length-1
   * @return argument i
   */
  public Object getValues(int i)
    {
      return args[i];
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * and arguments.
   * @param pattern the MessageFormat pattern text
   * @param args an array of arguments to substitute into the pattern
   */
  public Message(String pattern, Object[] args)
    {
      super(pattern);
      this.args = args;
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * and arguments.
   * @param pattern the MessageFormat pattern text
   * @param arg0 first argument to substitute into the pattern
   */
  public Message(String pattern, Object arg0)
    {
      this(pattern, new Object[] { arg0 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * and arguments.
   * @param pattern the MessageFormat pattern text
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   */
  public Message(String pattern, Object arg0, Object arg1)
    {
      this(pattern, new Object[] { arg0, arg1 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * and arguments.
   * @param pattern the MessageFormat pattern text
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   */
  public Message(String pattern, Object arg0, Object arg1, Object arg2)
    {
      this(pattern, new Object[] { arg0, arg1, arg2 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * and arguments.
   * @param pattern the MessageFormat pattern text
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   * @param arg3 next argument to substitute into the pattern
   */
  public Message(String pattern, Object arg0, Object arg1, Object arg2, Object arg3)
    {
      this(pattern, new Object[] { arg0, arg1, arg2, arg3 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * and arguments.
   * @param pattern the MessageFormat pattern text
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   * @param arg3 next argument to substitute into the pattern
   * @param arg4 next argument to substitute into the pattern
   */
  public Message(String pattern, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4)
    {
      this(pattern, new Object[] { arg0, arg1, arg2, arg3, arg4 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * and arguments.
   * @param pattern the MessageFormat pattern text
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   * @param arg3 next argument to substitute into the pattern
   * @param arg4 next argument to substitute into the pattern
   * @param arg5 next argument to substitute into the pattern
   */
  public Message(String pattern, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4, Object arg5)
    {
      this(pattern, new Object[] { arg0, arg1, arg2, arg3, arg4, arg5 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and no arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   */
  public Message(ResourceBundle bundle, String key)
    {
      super(bundle.getString(key));
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   * @param args an array of arguments to substitute into the pattern
   */
  public Message(ResourceBundle bundle, String key, Object[] args)
    {
      super(bundle.getString(key));
      this.args = args;
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   * @param arg0 first argument to substitute into the pattern
   */
  public Message(ResourceBundle bundle, String key, Object arg0)
    {
      this(bundle.getString(key), new Object[] { arg0 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   */
  public Message(ResourceBundle bundle, String key, Object arg0, Object arg1)
    {
      this(bundle.getString(key), new Object[] { arg0, arg1 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   */
  public Message(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2)
    {
      this(bundle.getString(key), new Object[] { arg0, arg1, arg2 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   * @param arg3 next argument to substitute into the pattern
   */
  public Message(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2, Object arg3)
    {
      this(bundle.getString(key), new Object[] { arg0, arg1, arg2, arg3 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   * @param arg3 next argument to substitute into the pattern
   * @param arg4 next argument to substitute into the pattern
   */
  public Message(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4)
    {
      this(bundle.getString(key), new Object[] { arg0, arg1, arg2, arg3, arg4 });
    }

  /**
   * Create a Message (MessageFormat) object with the pattern
   * from a resource bundle and arguments.
   * @param bundle a resource bundle from which the pattern is obtained
   * @param key the key identifying a String in the bundle
   * @param arg0 first argument to substitute into the pattern
   * @param arg1 next argument to substitute into the pattern
   * @param arg2 next argument to substitute into the pattern
   * @param arg3 next argument to substitute into the pattern
   * @param arg4 next argument to substitute into the pattern
   * @param arg5 next argument to substitute into the pattern
   */
  public Message(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4, Object arg5)
    {
      this(bundle.getString(key), new Object[] { arg0, arg1, arg2, arg3, arg4, arg5 });
    }

  /**
   * Format a Message by applying the stored format args to the pattern.
   * The existing format args in the object are preserved
   * @param unformattedText the new raw format string
   * @return the formatted text
   */
  public synchronized String toString()
    {
      if (buffer == null)
        buffer = new StringBuffer();
      buffer.setLength(0);
      if (args == null)
        return toPattern();
      this.format(args, buffer, null);
      return buffer.toString()/*I18nOK:MTH*/;
    }

  /**
   * Format a Message by applying the format args to the pattern.
   * The existing format args in the object are replaced with
    * the new ones (unless a formatting exception is raised.)
   * @param unformattedText the new raw format string
   * @param args Array of values to be formatted with the message.
   * @return the formatted text
   */
  public synchronized String toString(Object args[])
    {
      Object[] oldArgs = this.args;
      try
        {
          this.args = args;
          return toString();
        }
      catch (RuntimeException e)
        {
          this.args = oldArgs;
          throw(e);
        }
    }

  /**
   * Format a Message by applying the arg to the pattern.
   * @param arg0 First value to be formatted with the message.
   * @return the formatted text
   */
  public String toString(Object arg0)
    {
      return toString(new Object[] { arg0 });
    }

  /**
   * Format a Message by applying the args to the pattern.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @return the formatted text
   */
  public String toString(Object arg0, Object arg1)
    {
      return toString(new Object[] { arg0, arg1 });
    }

  /**
   * Format a Message by applying the args to the pattern.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @return the formatted text
   */
  public String toString(Object arg0, Object arg1, Object arg2)
    {
      return toString(new Object[] { arg0, arg1, arg2 });
    }

  /**
   * Format a Message by applying the args to the pattern.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @return the formatted text
   */
  public String toString(Object arg0, Object arg1, Object arg2, Object arg3)
    {
      return toString(new Object[] { arg0, arg1, arg2, arg3 });
    }

  /**
   * Format a Message by applying the args to the pattern.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @return the formatted text
   */
  public String toString(Object arg0, Object arg1, Object arg2, Object arg3, Object arg4)
    {
      return toString(new Object[] { arg0, arg1, arg2, arg3, arg4 });
    }

  /**
   * Format a Message by applying the args to the pattern.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @param arg5 Next value to be formatted with the message.
   * @return the formatted text
   */
  public String toString(Object arg0, Object arg1, Object arg2, Object arg3, Object arg4, Object arg5)
    {
      return toString(new Object[] { arg0, arg1, arg2, arg3, arg4, arg5 });
    }

  /**
   * Format a message from a resource bundle by applying the args
   * to the pattern.
   * @param bundle a resource bundle from which the pattern is read
   * @param key the key identifying which pattern to fetch form the resource bundle.
   * @param args Array of values to be formatted with the message.
   * @return the formatted text
   * @exception java.util.MissingResourceException if the named pattern
   *            does not exist in the resource bundle
   */
  public static String format(ResourceBundle bundle, String key, Object args[])
    {
      return MessageFormat.format(bundle.getString(key), (Object []) args);
    }

  /**
   * Format a message from a resource bundle by applying the arg
   * to the pattern.
   * @param bundle a resource bundle from which the pattern is read
   * @param key the key identifying which pattern to fetch form the resource bundle.
   * @param arg0 First value to be formatted with the message.
   * @return the formatted text
   * @exception java.util.MissingResourceException if the named pattern
   *            does not exist in the resource bundle
   */
  public static String format(ResourceBundle bundle, String key, Object arg0)
    {
      return MessageFormat.format(bundle.getString(key), new Object[] {arg0});
    }

  /**
   * Format a message from a resource bundle by applying the args
   * to the pattern.
   * @param bundle a resource bundle from which the pattern is read
   * @param key the key identifying which pattern to fetch form the resource bundle.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @return the formatted text
   * @exception java.util.MissingResourceException if the named pattern
   *            does not exist in the resource bundle
   */
  public static String format(ResourceBundle bundle, String key, Object arg0, Object arg1)
    {
      return MessageFormat.format(bundle.getString(key), new Object[] {arg0,  arg1});
    }

  /**
   * Format a message from a resource bundle by applying the args
   * to the pattern.
   * @param bundle a resource bundle from which the pattern is read
   * @param key the key identifying which pattern to fetch form the resource bundle.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @return the formatted text
   * @exception java.util.MissingResourceException if the named pattern
   *            does not exist in the resource bundle
   */
  public static String format(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2)
    {
      return MessageFormat.format(bundle.getString(key), new Object[] {arg0,  arg1,  arg2});
    }

  /**
   * Format a message from a resource bundle by applying the args
   * to the pattern.
   * @param bundle a resource bundle from which the pattern is read
   * @param key the key identifying which pattern to fetch form the resource bundle.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @return the formatted text
   * @exception java.util.MissingResourceException if the named pattern
   *            does not exist in the resource bundle
   */
  public static String format(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2, Object arg3)
    {
      return MessageFormat.format(bundle.getString(key), new Object[] {arg0,  arg1,  arg2,  arg3});
    }

  /**
   * Format a message from a resource bundle by applying the args
   * to the pattern.
   * @param bundle a resource bundle from which the pattern is read
   * @param key the key identifying which pattern to fetch form the resource bundle.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @return the formatted text
   * @exception java.util.MissingResourceException if the named pattern
   *            does not exist in the resource bundle
   */
  public static String format(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4)
    {
      return MessageFormat.format(bundle.getString(key), new Object[] {arg0,  arg1,  arg2,  arg3,  arg4});
    }

  /**
   * Format a message from a resource bundle by applying the args
   * to the pattern.
   * @param bundle a resource bundle from which the pattern is read
   * @param key the key identifying which pattern to fetch form the resource bundle.
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @param arg5 Next value to be formatted with the message.
   * @return the formatted text
   * @exception java.util.MissingResourceException if the named pattern
   *            does not exist in the resource bundle
   */
  public static String format(ResourceBundle bundle, String key, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4, Object arg5)
    {
      return MessageFormat.format(bundle.getString(key), new Object[] {arg0,  arg1,  arg2,  arg3,  arg4,  arg5});
    }

  /**
   * Format a message by applying the arg to the pattern.
   * Note that static String formay(String pattern, args[])
   * is inherited from MessageFormat.
   * @param pattern the pattern touse for formatting the message
   * @param arg0 First value to be formatted with the message.
   * @return the formatted text
   */
  public static String format(String pattern, Object arg0)
    {
      return MessageFormat.format(pattern, new Object[] { arg0 });
    }

  /**
   * Format a message by applying the args
   * to the pattern.
   * @param pattern the pattern to use for formatting the message
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @return the formatted text
   */
  public static String format(String pattern, Object arg0, Object arg1)
    {
      return MessageFormat.format(pattern, new Object[] { arg0, arg1 });
    }

  /**
   * Format a message by applying the args
   * to the pattern.
   * @param pattern the pattern to use for formatting the message
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @return the formatted text
   */
  public static String format(String pattern, Object arg0, Object arg1, Object arg2)
    {
      return MessageFormat.format(pattern, new Object[] { arg0, arg1, arg2 });
    }

  /**
   * Format a message by applying the args
   * to the pattern.
   * @param pattern the pattern to use for formatting the message
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @return the formatted text
   */
  public static String format(String pattern, Object arg0, Object arg1, Object arg2, Object arg3)
    {
      return MessageFormat.format(pattern, new Object[] { arg0, arg1, arg2, arg3 });
    }

  /**
   * Format a message by applying the args
   * to the pattern.
   * @param pattern the pattern to use for formatting the message
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @return the formatted text
   */
  public static String format(String pattern, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4)
    {
      return MessageFormat.format(pattern, new Object[] { arg0, arg1, arg2, arg3, arg4 });
    }

  /**
   * Format a message by applying the args
   * to the pattern.
   * @param pattern the pattern to use for formatting the message
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @param arg5 Next value to be formatted with the message.
   * @return the formatted text
   */
  public static String format(String pattern, Object arg0, Object arg1, Object arg2, Object arg3, Object arg4, Object arg5)
    {
      return MessageFormat.format(pattern, new Object[] { arg0, arg1, arg2, arg3, arg4, arg5 });
    }


}

