// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

/**
 * An interface which allows the caller to
 * perform some action on an object.
 * ApplyInterface is normally used to apply some action to
 * all elements of a collection; see
 * {@link com.sas.collection.StaticCollectionInterface#apply(com.sas.util.ApplyInterface)}.
 * For example, to print the contents of a collection, create an instance of
 * a {@link com.sas.collection.PrintContents} and apply it to a collection:
 * <pre>
 *    StaticCollectionInterface c =  ...some collection...;
 *    PrintContents printer = new PrintContents();
 *    c.apply(printer);
 * </pre>
 * ApplyInterface can be used to alter all items in a collection, or to
 * search collections for items matching specific conditions, etc.
 * Ths operation can be done without knowledge of how to access
 * individual items in the collection, or without having to obtain
 * an {@link java.util.Enumeration Enumeration}, which requires extra
 * object allocations.
 * @version 1.0
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 */
public interface ApplyInterface
{
  /**
   * A return code from the {@link #apply(java.lang.Object)} method
   * which indicates that the action should continue for other objects.
   */
  public static final short CONTINUE = 0;

  /**
   * A return code from the {@link #apply(java.lang.Object)} method
   * which indicates that the action should not continue for other objects.
   * For example, if you implement an object to search a collection for
   * an item that matches some criteria, when the item is found, your
   * <code>apply</code> method can return <code>ABORT</code>
   * to indicate that the collection can abort the searh and not
   * apply your search object to the remaining items in the collection.
   */
  public static final short ABORT    = 1;

  /**
   * Does something with/to an object
   * @param object The object which should be used
   * @return one of <code>{@link #ABORT}</code> or <code>{@link #CONTINUE}</code>.
   * The <code>apply()</code> method should
   * return <code>ABORT</code> if it
   * wishes to abort or short circuit an
   * iteration/application of the <code>ApplyInterface</code> to items
   * in a collection.  Normally, this method would return <code>CONTINUE</code>
   * to indicate normal continuation of the iteration.
   */
  int apply(Object o);
}
