/* Copyright (c) 2002 by SAS Institute Inc., Cary, NC 27513 */

package com.sas.util;

import com.sas.util.Comparator;
import java.io.Serializable;
import java.text.Collator;

/**
 * BaseStringComparator is used to compare two string values. This is most often
 * done when sorting collections of strings.  You can also use BaseStringComparator
 * for non-string items; it invokes the toString() method of objects and
 * compares using the resulting strings.
 * <p>
 * BaseStringComparator provides a bit of extra smarts: you can specify
 * that the comparison start at a character offset in the strings, instead of
 * a default 0, or that case be ignored (although this is not always appropriate
 * for some character sets).
 * @version 1.0
 */

public class BaseStringComparator
       implements Comparator,
                  Serializable,
				  java.util.Comparator

{
  static final long serialVersionUID = 770854091247656479L;

  private boolean ascending;
  private boolean caseSensitive;
  private int startChar;
  private Collator collator;

  /**
   * If you wish to use a default <code>BaseStringComparator</code>, you can use
   * <code>defaultInstance</code> to avoid creating a new instance.
   * It is initialized to <code><code>new StringComparator()</code> and thus compares
   * strings in ascending order, honoring case, starting at character 0.
   * <strong>Note:</strong> Java does not provide a means to define const objects,
   * so nothing prevents you from calling one of the below set
   * methods on <code>defaultInstance</code>, but doing so is <strong>highly</strong> discouraged,
   * as other packages may be relying on <code>defaultInstance</code> to have
   * the default behavior.
   */
  public static final BaseStringComparator defaultInstance = new BaseStringComparator();

  /**
   * Default constructor. This comparator will perform an ascending
   * comparison of two objects using their <code>String</code> representation.
   */
  public BaseStringComparator()
  {
    this(true, true, 0);
  }

  /**
   * Construct a <code>BaseStringComparator</code> which uses a <code>Collator</code>
   * to perform the comparison.
   * @param collator a collator. You can use <code>{@link java.text.Collator#getInstance()}</code>
   * to construct the default Collator for the current <code>{@link java.util.Locale}</code><code>.
   */
  public BaseStringComparator(Collator collator)
  {
    this(true, true, 0);
    this.collator = collator;
  }

  /**
   * Construct a string based comparator.
   * This comparator will perform an ascending or descending
   * comparison of two objects using their <code>String</code> representation.
   * @param ascending if false, negate the default comparison.
   */
  public BaseStringComparator(boolean ascending)
  {
    this(ascending, true, 0);
  }

  /**
   * Construct a string based comparator. This comparator will perform an
   * ascending or descending comparison of two objects using their
   * <code>String</code> representation, optionally ignoring case differences.
   * @param ascending if <code>false</code>, negate the default comparison.
   * @param caseSensitive if <code>true</code>, the comparison ignores case.
   * Warning: case insensitive string comparisons are not possible
   * or correct in all languages/character encodings.
   */
  public BaseStringComparator(boolean ascending, boolean caseSensitive )
  {
    this(ascending, caseSensitive, 0);
  }

  /**
   * Construct a string based comparator.
   * This comparator will perform an ascending or descending
   * comparison of two objects using their <code>String</code> representation.
   * @param ascending if <code>false</code>, negate the default comparison.
   * @param caseSensitive if <code>true</code>, the comparison ignores case.
   * Warning: case insensitive string comparisons are not possible
   * or correct in all languages/character encodings.
   * @param startChar the starting character for comparisons.
   * characters before the startChar are ignored.
   */
  public BaseStringComparator(boolean ascending, boolean caseSensitive, int startChar)
  {
    setAscending(ascending);
    setStartChar(startChar);
    setCaseSensitive(caseSensitive);
  }

  /**
   * Return <code>true</code> if this <code>BaseStringComparator</code> is performing an ascending
   * comparison.
   * @return <code>true</code> if <code><code>compare("A", "B")</code> will return LESS</code>_THAN
   */
  public boolean isAscending()
    {
      return ascending;
    }

  /**
   * Set the <code>BaseStringComparator</code> to an ascending sort comparison (the default).
   * @param ascending <code>true</code> if you want an ascending string <code>{"A", "B", "C"}</code>;
   * <code>false</code> if you want a descending sort <code>{"C", "B", "A"}</code>, for example.
   */
  public void setAscending(boolean ascending)
    {
      if (this.ascending != ascending)
        {
          this.ascending = ascending;
        }
    }

  /**
   * Indicate if the comparison is case sensitive or not.
   * @return <code>true</code> if the comparison ignores case by comparing
   * only the <code>{@link java.lang.String#toLowerCase()}</code> version of strings.
   */
  public boolean isCaseSensitive()
    {
      return caseSensitive;
    }

  /**
   * Set the case sensitivity of the <code>StringComparator</code><code><code>.
   * @param caseSensitive if t</code>rue, the comparator will
   * honor the case of the strings; if f</code>alse, comparisons
   * are done on the <code>toLowerCase()</code> results. If the <code>StringComparator</code>
   * was constructed with a <code>Collator</code>, this method will set the
   * collator's strength to <code><code>{@link java.text.Collator#TERTIARY}</code> if <var>caseSensitive</var><code> is tr</code>ue,
   * and <code>{@link java.text.Collator#PRIMARY}</code> if fal</code>se. If you need finer granularity,
   * you can modify the collator directly.
   */
  public void setCaseSensitive(boolean caseSensitive)
    {
      if (collator != null)
        collator.setStrength(caseSensitive ? Collator.TERTIARY : Collator.PRIMARY );
      if (this.caseSensitive != caseSensitive)
        {
          this.caseSensitive = caseSensitive;
        }
    }

  /**
   * Set the collator used to compare strings.
   * @param newCollator the new value for the collator property.
   * If <code>null</code>, then simple string comparison is used.
   */
   public void setCollator(Collator newCollator)
   {
     if (com.sas.util.Util.equal(collator, newCollator))
        return;
     Collator oldCollator = collator;
     collator = newCollator;
   }

   /**
    * Return the collator used to compare the strings.
    * This may be <code>null</code> if no collator exists, in which case the
    * <code>StringComparator</code> uses default string comparison.
    * @return the string collator used to compare strings
    */
    public Collator getCollator()
    {
        return collator;
    }

  /**
   * Return the character at which string comparison begins.
   * @return the start character. The default is 0, which means
   * the entire string is considered when comparing.
   */
  public int getStartChar()
    {
      return startChar;
    }

  /**
   * Set the character at which string comparisons are performed.
   * The default is 0.
   * @param startChar the character index where comparison begins;
   * characters before this are ignored. If a string is less than
   * startChar characters long, it is compared as an empty
   * string; i.e. less than other strings in an ascending sort.
   */
  public void setStartChar(int startChar)
    {
      if (startChar < 0)
        throw new IllegalArgumentException("startChar < 0");
      if (this.startChar != startChar)
        {
          Integer oldVal = new Integer(this.startChar);
          Integer newVal = new Integer(startChar);
          this.startChar = startChar;
        }
    }

  /**
   * Compare two values. If the <code>BaseStringComparator</code> is constructed with <var>ascending</var><code> == <code>true</code>,
   * the <code>compare()</code> method returns <code>LESS_THAN</code> if the <code>toString()</code> value
   * of the first object compares less than the <code>toString()</code> value of
   * the second, else <code>EQUALS</code> or <code><code>GREATER_THAN</code>. A nu</code>ll item
   * is sorted <code>LESS_THAN</code> a non-null item in an ascending comparison.
   * @param first the left hand side of the comparison
   * @param second the right hand side of the comparison
   * @return
   * <table>
   * <tr><td>LESS_THAN     <td>the first object compares "less than" the second, meaning it
   *                           should appear before the second in the sorted result
   * <tr><td> EQUALS       <td>the objects are equal
   * <tr><td> GREATER_THAN <td>the first object compares greater than the first, meaning it
   *                           should appear after the second in the sorted result
   * </table>
   * (swap results if ascending == false)
   * @see com.sas.util.Comparator#LESS_THAN
   * @see com.sas.util.Comparator#EQUALS
   * @see com.sas.util.Comparator#GREATER_THAN
   */
  public int compare(Object first, Object second)
  {
    int result;

    if (first == second)
      return EQUALS;
    else if (first == null)
      return ascending ? LESS_THAN : GREATER_THAN;
    else if (second == null)
      return ascending ? GREATER_THAN : LESS_THAN;
    if ( !(first instanceof String) || !(second instanceof String))
      return compare(first.toString(), second.toString());

    String a = (String) first;
    String b = (String) second;

    if (startChar > 0)
      {
        if (a.length() < startChar)
          a = null;
        else
          a = a.substring(startChar);
        if (b.length() < startChar)
          b = null;
        else
          b = b.substring(startChar);
        if (a == b)
          return EQUALS;
        else if (a == null)
          return ascending ? LESS_THAN : GREATER_THAN;
        else if (b == null)
          return ascending ? GREATER_THAN : LESS_THAN;
      }

    if (collator != null)
      result = collator.compare(a, b);
    else if (caseSensitive)
      result = a.compareTo(b);
    else
      result = a.toLowerCase()/*I18nOK:MTH*/.compareTo(b.toLowerCase()/*I18nOK:MTH*/);

    if (result < 0)
      result = ascending ? LESS_THAN : GREATER_THAN;
    else if (result > 0)
      result = ascending ? GREATER_THAN : LESS_THAN;
    return result;
  }

  /**
   * Return a string representation of the comparator
   * @return a string representation of the comparator
   */
  public String toString()
    {
      return ("ascending: "/*I18nOK:EMS*/ + ascending
              + ", caseSensitive:" + caseSensitive
              + ", startChar:" + startChar
              + " StringComparator ");
    }
}
