/* $Id$ */
package com.sas.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.UndeclaredThrowableException;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.security.PrivilegedActionException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.ByteArrayOutputStream;
import com.sas.util.Classes;

/**
 * ChainedException is a base exception for exceptions which
 * &quot;wrap&quot; or chain other other exceptions.
 * The wrapped exception may be unchecked exceptions if desired.
 * <p>
 * You may use {@link #getRootException()} to get the
 * underlying or root cause from which this ChainedException
 * was constructed.
 * <p>
 * ChainedException contains some static methods for printing
 * the entire stack chain. You may print the exception chain
 * to either a {@link PrintStream}, {@link PrintWriter}, or a {@link StringBuffer}.
 * Be careful to not reimplement a ChainedException subclass by changing
 * one of it's {@link Throwable#printStackTrace()} methods to call
 * one of these static methods, since that will result in an infinite
 * look (because those methods call back to the base <code>printStackTrace()</code>
 * methods).
 * <p>
 * Exception classes can extend this base class in order to
 * implement ChainedExceptionInterface. Exception classes which
 * cannot extend this class because of other inheritance requirements
 * can still implement this interface quite trivially by adding:
 * <pre>
 * public class CustomException
 *        extends SpecialException
 *        implements ChainedExceptionInterface
 * {
 *   private Throwable chainedException;
 *   public CustomException()
 *   {
 *   }
 *   public CustomException(String message)
 *   {
 *     super(message);
 *   }
 *   public CustomException(Throwable chainedException, String message)
 *   {
 *     super(message);
 *     this.chainedException = chainedException;
 *   }
 *   // Implement {@link ChainedExceptionInterface#getTargetException()}:
 *   public Throwable getTargetException()
 *   {
 *     return chainedException;
 *   }
 * // ... other CustomException implementation here....
 * }
 * </pre>
 * Note that JDK 1.4 provides a <code>Throwable getCause()</code> method which
 * provides the same behavior as {@link #getTargetException()}, but the <code>getCause()</code> method
 * does not exist prior to 1.4.
 * @see UncheckedException
 * @see java.lang.reflect.InvocationTargetException
 * @see java.sql.SQLException
 * @see java.sql.SQLWarning
 * @author sasdjb <A HREF="http://sww.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */
public class ChainedException
       extends Exception
       implements ChainedExceptionInterface
{

  //////////////////////////////// Static Instance Variables ////////////////////////////////
  static final long serialVersionUID = -5552882514700119725L;


  //////////////////////////////// Instance Variables ////////////////////////////////

  private Throwable chainedException;


  //////////////////////////////// Constructors ////////////////////////////////

  /**
   * Create a ChainedException that contains no other exception
   * or exception message.
   */
  public ChainedException()
  {
  }

  /**
   * Create a ChainedException that is not based on another exception.
   * @param message the text describing the exception.
   * @see Throwable#getMessage()
   */
  public ChainedException(String message)
  {
    super(message);
  }


  //////////////////////////////// Instance Methods ////////////////////////////////
  /**
   * Create a ChainedException based on another exception.
   * This is the only way to set the target exception that
   * is returned by {@link #getRootException()}.
   * @param rootException An exception which this
   * ChainedException will chain to via {@link #getRootException}.
   * It may itself be
   * another ChainedException.
   * @param message the text describing the exception.
   * If null, then the <var>rootException</var>'s
   * message is used.
   * @see Throwable#getMessage()
   * @see #getTargetException()
   */
  public ChainedException(Throwable rootException, String message)
  {
    super(  (message == null && rootException == null)
          ? null
          : (  (message == null)
             ? rootException.getMessage()
             : message));
    this.chainedException = rootException;
  }

  /**
   * Return the exception with which this ChainedExceptionInterface was constructed
   * via {@link #ChainedException(Throwable, String)}.
   * Note that JDK 1.4 provides a <code>Throwable getCause()</code> method which
   * provides the same behavior as {@link #getTargetException()}, but the <code>getCause()</code> method
   * does not exist prior to 1.4.
   */
  public Throwable getTargetException()
  {
    return chainedException;
  }


  /**
   * Return the root cause of this ChainedException.
   * @param the underlying exception; i.e. the exception that
   * has no other chained/invocation/next exception.
   */
  public Throwable getRootException()
  {
    return getRootException(this);
  }

  /**
   * Return this <code>ChainedException</code> as a <code>String</code>
   * @return the value of <code>toStringBuffer(this, new StringBuffer()).toString()</code>
   */
  public String toString()
  {
    return toStringBuffer(this, new StringBuffer()).toString()/*I18nOK:MTH*/;
  }


  //////////////////////////////// Static Methods ////////////////////////////////

  /**
   * Helper method to determine if the input throwable/exception
   * has an chained exception.
   * If <var>throwable</var> imlements <code>ChainedExceptionInterface</code>
   * return its {@link ChainedExceptionInterface#getTargetException() getTargetException()}
   * If <var>throwable</var> is a <code>InvocationTargetException</code>
   * return its {@link InvocationTargetException#getTargetException() getTargetException()}.
   * If <var>throwable</var> is a <code>UndeclaredThrowableException</code>
   * return its {@link UndeclaredThrowableException#getUndeclaredThrowable()}.
   * If <var>throwable</var> is a <code>ClassNotFoundException</code>
   * return its {@link ClassNotFoundException#getException() getException()}.
   * If <var>throwable</var> is a <code>PrivilegedActionException</code>
   * return its {@link PrivilegedActionException#getException() getException()}.
   * If <var>throwable</var> is a <code>ExceptionInInitializerError </code>
   * return its {@link ExceptionInInitializerError#getException() getException()}.
   * If <var>throwable</var> is a <code>SQLException</code>
   * return its {@link SQLException#getNextException() getNextException()}.
   * If <var>throwable</var> is a <code>SQLWarning</code>
   * return its {@link SQLWarning#getNextWarning() getNextWarning()}.
   * @param throwable a Throwable
   * @return <var>t</var>'s chained exception, or null
   */
  public static Throwable getTargetException(Throwable throwable)
  {
    Throwable cause = null;

         if (throwable instanceof ChainedExceptionInterface)        cause = ((ChainedExceptionInterface) throwable).getTargetException();
    else if (throwable instanceof InvocationTargetException)        cause = ((InvocationTargetException) throwable).getTargetException();
    else if (throwable instanceof UndeclaredThrowableException)     cause = ((UndeclaredThrowableException) throwable).getUndeclaredThrowable();
    else if (throwable instanceof PrivilegedActionException)        cause = ((PrivilegedActionException) throwable).getException();
    else if (throwable instanceof ClassNotFoundException)           cause = ((ClassNotFoundException) throwable).getException();
    else if (throwable instanceof ExceptionInInitializerError)      cause = ((ExceptionInInitializerError) throwable).getException();
    else if (throwable instanceof SQLException)                     cause = ((SQLException) throwable).getNextException();
    else if (throwable instanceof SQLWarning)                       cause = ((SQLWarning) throwable).getNextWarning();

    if ( cause == null )
      cause = throwable.getCause(); // null;  //S0168152

    return cause;
  }

  /**
   * Print the exception chain to a PrintStream.
   * You may invoke this static method on any exception, not just a ChainedException.
   * If printChain is false, this method acts like {@link Throwable#printStackTrace(PrintStream)}.
   * If {@link #getTargetException(Throwable) getTargetException(<var>t</var>)} is not null,
   * print the next exceptions stack trace as well.
   * Then, print <var>throwable</var>'s stack trace.
   * @param throwable the exception to print. If null, return without printing anything.
   * @param stream The stream to print to. If null, print to System.err
   * @param printChain if false, do not print the rest of the chain, this <var>throwable</var>.
   */
  public static void printStackTrace(Throwable throwable, PrintStream stream, boolean printChain)
  {
    if (throwable == null)
      return;
    if (stream == null)
      stream = System.err;
    synchronized (stream)
      {
        if (printChain)
          printStackTrace( getTargetException(throwable), stream, true);
        throwable.printStackTrace(stream);
      }
  }

  /**
   * Print the exception chain to a PrintWriter.
   * You may invoke this static method on any exception, not just a ChainedException.
   * If {@link #getTargetException(Throwable) getTargetException(<var>t</var>)} is not null,
   * print the next exceptions stack trace as well.
   * Then, print <var>throwable</var>'s stack trace.
   * @param throwable the exception to print. If null, return without printing anything.
   * @param writer The PrintWriter to print to.
   * @param printChain if false, do not print the rest of the chain, this <var>throwable</var>.
   */
  public static void printStackTrace(Throwable throwable, PrintWriter writer, boolean printChain)
  {
    if (throwable == null)
      return;
    synchronized(writer)
      {
        if (printChain)
          printStackTrace( getTargetException(throwable), writer, true);
        throwable.printStackTrace(writer);
      }
  }


  /**
   * Print the exception chain to a <code>StringBuffer</code>.
   * You may invoke this static method on any exception, not just a ChainedException.
   * If {@link #getTargetException(Throwable) getTargetException(<var>t</var>)} is not null,
   * print the next exceptions stack trace as well.
   * Then, print <var>throwable</var>'s stack trace.
   * @param throwable the exception to print. If null, return <var>stringBuffer</var> without appending anything.
   * @param writer The <code>StringBuffer</code> to print to. if null, a new <code>StringBuffer</code>
   * is allocated and returned.
   * @param printChain if false, do not print the rest of the chain, this <var>throwable</var>.
   * @param the <var>stringBuffer</var>, if not null, else a new <code>StringBuffer</code>. The chained stack trace
   * will be appended to this <code>StringBuffer</code>.
   */
  public static StringBuffer printStackTrace(Throwable throwable, StringBuffer stringBuffer, boolean printChain)
  {
    if (throwable == null)
      return stringBuffer;
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    PrintWriter printWriter = new PrintWriter(baos);
    printStackTrace(throwable, printWriter, printChain);
    printWriter.flush();
    String s = baos.toString();
    if ( stringBuffer == null )
      stringBuffer = new StringBuffer(s);
    else
      stringBuffer.append(s);
    printWriter.close();
    return stringBuffer;
  }

  /**
   * Return the root cause of an exception.
   * For standard exceptions/throwables, it is the <var>throwable</var>
   * itself. For a ChainedExceptionInterface, InvocationTargetException,
   * or SQLException, it defined recusively as getRootException() of
   * the chained/invocation/next exceptions, if they are not null.
   * You may invoke this static method on any Throwable, not
   * just a ChainedException.
   * @param throwable a Throwable whose root exception/throwable
   * you want to obtain.
   * @param the underlying exception; i.e. the exception that
   * has no other chained/invocation/next exception.
   */
  public static Throwable getRootException(Throwable throwable)
  {
    Throwable next = getTargetException(throwable);
    if ( next != null )
      return getRootException(next);
    else
      return throwable;
  }

  /**
   * Helper method to print the name and message of a throwable to
   * a string. The {@link #toString()} method calls this.
   * The result contains the throwable's class name and the
   * message, if any. If the throwable has a next exception
   * as defined by {@link #getTargetException}, then also
   * append a " [ ", then recursively call <code>toStringBuffer</code>
   * on it, then append a " ] "
   * @param throwable print the class name, message, and exception chain for this Throwable
   * @param stringBuffer print the exception class names and messages to this <code>StringBuffer</code>
   *        if null, a new StringBuffer is allocated. This is then returned.
   * @return the <code>StringBuffer</code> stringBuffer, or a new <code>StringBuffer</code>
   */
  public static StringBuffer toStringBuffer(Throwable throwable, StringBuffer stringBuffer)
  {
    if ( stringBuffer == null )
      stringBuffer = new StringBuffer();
    stringBuffer.append(throwable.getClass().getName());
    if ( throwable.getMessage() != null)
      {
        stringBuffer.append(": ");
        stringBuffer.append(throwable.getMessage());
      }
    Throwable next = getTargetException(throwable);
    if ( next != null ) {
      stringBuffer.append(" [ ");
      toStringBuffer(next, stringBuffer);
      stringBuffer.append(" ] ");
    } // end of if ()
    return stringBuffer;
  }


  /**
   * Test if a Throwable or any Exception or Throwable in its exception chain
   * extends a specific Throwable class. For example, if you have caught
   * an {@link java.lang.reflect.InvocationTargetException}, you can see
   * if it was caused by an SQL exception with the following:
   * <pre>
   *  catch (java.lang.reflect.InvocationTargetException ite)
   *     {
   *        boolean sql = ChainedException.chainContains(ite, java.sql.SQLException.class)
   *        ...
   *     }
   * </pre>
   * You may invoke this static method on any Throwable, not
   * just a ChainedException.
   *
   * @param throwable any Throwable object, including any Exception.
   * @param throwableClass a class object of a Throwable or Exception class.
   * @return true if the <var>throwable</var> is an instance of the <var>throwableClass</var>
   * or a subclass of <var>throwableClass</var>, or, recursively, if
   * <code>chainContains(getTargetException(<var>throwable</var>), <var>throwableClass</var>)</code>
   * @see com.sas.util.Classes#extendsClass(Class,Class)
   */
  public static boolean chainContains(Throwable throwable, Class throwableClass)
  {
    try
      {
        return (   Classes.extendsClass(throwable.getClass(), throwableClass)
                   || chainContains(getTargetException(throwable), throwableClass));
      }
    catch (NullPointerException npe)
      { return false; }
  }

}
