// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import java.text.MessageFormat;

/**
 * <p>
 * CharBuffer is a character string management class.  It provides some
 * features that are present in java.lang.String and some features that
 * are present in java.lang.StringBuffer.  Additionally it provides some
 * features that are present in both of these classes and some features
 * that are not present in either.  CharBuffer provides general character
 * retrieval, insertion and removal.  The purpose of this class is to
 * reduce the number of object allocations required when performing heavy
 * text manipulation such as parsing and formatting.
 * </p>
 * <p>
 * For the methods that take a beginning index and an ending index as
 * arguments, -1 is allowed as the value of the ending index.
 * Whenever -1 is passed as the value of the ending index, the effective
 * range will be interpreted as all values from the starting index to
 * the end of the CharBuffer.
 * </p>
 *
 * @author Ryan Norris - sasrrn
 * @version 1.0
 */
public class CharBuffer
   extends java.lang.Object
   implements java.io.Serializable,
              com.sas.PublicClonable
{
   static final long serialVersionUID = 7996781090478772284L;

   public static final String RB_KEY = "CharBuffer.";
   private char m_data[];
   private int  m_count;

   /**
    * Constructs an empty CharBuffer object.  The initial capacity is 16.
    */
   public CharBuffer()
   {
	   this(16);
   }

   /**
    * Constructs an empty CharBuffer object with an initial capacity of
    * the specified size.
    * @param size the initial capacity
    */
   public CharBuffer(int size)
   {
	   m_data = new char[size];
	   m_count = 0;
   }

   /**
    * Constructs a CharBuffer object that is initialized with the contents
    * of the specified character array.  The initial capacity is 16.
    * @param chars initial value
    */
   public CharBuffer(char[] chars)
   {
	   this(chars.length + 16);
	   append(chars);
   }

  /**
    * Constructs a CharBuffer object that is initialized with the contents
    * of the specified String.  The initial capacity is string.length() + 16.
    * @param string initial value
    */
   public CharBuffer(String string)
   {
	   this(string.length() + 16);
	   append(string);
   }

  /**
    * Constructs a CharBuffer object that is initialized with the contents
    * of the specified StringBuffer.  The initial capacity is
    * stringBuf.length() + 16.
    * @param string initial value
    */
   public CharBuffer(StringBuffer stringBuf)
   {
	   this(stringBuf.length() + 16);
	   append(stringBuf);
   }

   /**
    * Constructs a CharBuffer object that is initialized with the contents
    * of the specified CharBuffer.  The initial capacity is
    * inBuf.length() + 16.
    * @param inBuf initial value
    */
   public CharBuffer(CharBuffer inBuf)
   {
      this(inBuf.m_count+16);
	   append(inBuf);
   }

   /**
    * Appends a character.
    * @param ch the character to append.
    */
   public void append(char ch)
   {
      ensureCapacity(m_count + 1);
      m_data[m_count++] = ch;
   }

   /**
    * Appends an array of characters.
    * @param chars array of characters to append
    */
   public void append(char[] chars)
   {
      int length = chars.length;
      ensureCapacity(m_count + length);
      System.arraycopy(chars, 0, m_data, m_count, length);
      m_count += length;
   }

   /**
    * Appends a sequence of characters from the specified CharBuffer.
    * @param buf sequence of characters to append.
    */
   public void append(CharBuffer buf)
   {
      int length = buf.m_count;
      ensureCapacity(m_count + length);
      System.arraycopy(buf.m_data, 0, m_data, m_count, length);
      m_count += length;
   }

   /**
    * Appends the specified String.
    * @param string sequence of characters to append.
    */
   public void append(String string)
   {
      int length = string.length();
      ensureCapacity(m_count + length);
      string.getChars(0, length, m_data, m_count);
      m_count += length;
   }

   /**
    * Appends the specified StringBuffer.
    * @param stringBuf sequence of characters to append.
    */
   public void append(StringBuffer stringBuf)
   {
      int length = stringBuf.length();
      ensureCapacity(m_count + length);
      stringBuf.getChars(0, length, m_data, m_count);
      m_count += length;
   }

   /**
    * Inserts the specified String at the specified index.
    * @param index the index of the CharBuffer at which to insert the string
    * @param string String to insert.
    */
   public void insert(int index, String string)
   {
	   int length = string.length();
	   ensureCapacity(m_count + length);
	   System.arraycopy(m_data, index, m_data, index + length, m_count - index);
	   string.getChars(0, length, m_data, index);
	   m_count += length;
   }

   /**
    * Inserts the specified StringBuffer at the specified index.
    * @param index the index of the CharBuffer at which to insert the string buffer
    * @param stringBuf StringBuffer to insert.
    */
   public void insert(int index, StringBuffer stringBuf)
   {
	   int length = stringBuf.length();
	   ensureCapacity(m_count + length);
	   System.arraycopy(m_data, index, m_data, index + length, m_count - index);
	   stringBuf.getChars(0, length, m_data, index);
	   m_count += length;
   }

   /**
    * Inserts the specified characters at the specified index.
    * @param index the index of the CharBuffer at which to insert the characters
    * @param chars characters to insert.
    */
   public void insert(int index, char[] chars)
   {
      ensureCapacity(m_count + chars.length);
      System.arraycopy(m_data, index, m_data, index + chars.length, m_count - index);
      System.arraycopy(chars, 0, m_data, index, chars.length);
      m_count += chars.length;
   }

   /**
    * Inserts the specified CharBuffer at the specified index.
    * @param index the index of the CharBuffer at which to insert
    * the specified CharBuffer
    * @param cbuf CharBuffer to insert.
    */
   public void insert(int index, CharBuffer cbuf)
   {
	   ensureCapacity(m_count + cbuf.m_count);
	   System.arraycopy(m_data, index, m_data, index + cbuf.m_count, m_count - index);
	   cbuf.getChars(0, cbuf.m_count, m_data, index);
	   m_count += cbuf.m_count;
   }

   /**
    * Inserts a character.
    * @param index the index of the CharBuffer at which to insert the character
    * @param ch character to insert.
    */
   public void insert(int index, char ch)
   {
      ensureCapacity(m_count + 1);
      shiftRight(index, 1);
      m_data[index] = ch;
      m_count++;
   }

   /**
    * Fills the CharBuffer from the specified position.  The character provided
    * will be replace the number of positions specified by count.
    * @exception IndexOutOfBoundsException if start is out of range [0, length()-1]
    */
   public void fill(char ch, int start, int count)
   {
      if ((start >= m_count) || (start < 0))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(start),
            new Integer(m_count-1)} ));

      if ((start + count) > m_count)
      {
         ensureCapacity(start + count);
         m_count = start + count;
      }

      for (int i=start; i < (start + count); i++)
         m_data[i] = ch;
   }

   /**
    * Fills the CharBuffer from the specified position.  The character
    * array provided will be inserted the number of times specified by
    * count overwriting any characters that are there.
    * @param chars the characters to insert
    * @param start the offset at which to start filling
    * @param count the number of times to fill with the specified characters
    * @exception IndexOutOfBoundsException if start is out of range [0, length()-1]
    */
   public void fill(char[] chars, int start, int count)
   {
      if ((start >= m_count) || (start < 0))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(start),
            new Integer(m_count-1)} ));

      int posToFill = chars.length * count;

      // if we are going to insert more characters than space allows,
      // we need to bump up the capacity
      if ((start + posToFill) > m_count)
      {
         ensureCapacity(start + posToFill);
         m_count = start + posToFill;
      }

      // insert the characters
      int offset = start;
      for (int i=0; i < count; i++)
      {
         // TO-DO:  use copy array instead
         for (int k=0; k<chars.length; k++)
            m_data[k + offset] = chars[k];
         offset += chars.length;
     }
   }

   // TO-DO:  all of the following fill methods funnel through fill(char[], ,).  Investigate
   //         a more efficient means

   /**
    * Fills the CharBuffer from the specified position.  The CharBuffer
    * provided will be inserted the number of times specified by count
    * overwriting any characters that are there.
    * @param cbuf the CharBuffer to insert
    * @param start the offset at which to start filling
    * @param count the number of times to fill with the specified CharBuffer
    * @exception IndexOutOfBoundsException if start is out of range [0, length()-1]
    */
   public void fill(CharBuffer cbuf, int start, int count)
   {
      fill(cbuf.toCharArray(), start, count);
   }

   /**
    * Fills the CharBuffer from the specified position.  The String provided
    * will be inserted the number of times specified by count overwriting
    * any characters that are there.
    * @param string the string to insert
    * @param start the offset at which to start filling
    * @param count the number of times to fill with the specified string
    * @exception IndexOutOfBoundsException if start is out of range [0, length()-1]
    */
   public void fill(String string, int start, int count)
   {
      fill(string.toCharArray(), start, count);
   }

   /**
    * Fills the CharBuffer from the specified position.  The StringBuffer
    * provided will be inserted the number of times specified by count
    * overwriting any characters that are there.
    * @param stringBuf the string buffer to insert
    * @param start the offset at which to start filling
    * @param count the number of times to fill with the specified string buffer
    * @exception IndexOutOfBoundsException if start is out of range [0, length()-1]
    */
   public void fill(StringBuffer stringBuf, int start, int count)
   {
      fill(stringBuf.toString()/*I18nOK:MTH*/.toCharArray(), start, count);
   }

   /**
    * Removes a range of characters.  The first character to be removed
    * is at index begin.  The last character to be removed is at index
    * end-1.  A value of -1 can be used for "end" to remove all
    * characters from "begin" to the end of the CharBuffer.
    * @param begin index of the first character to remove
    * @param end the index one part the last character to remove or -1 to
    * specify from "start" to the end of the CharBuffer
    * @exception IndexOutOfBoundsException if begin > end, if begin < 0, or if end > length()
    */
   public void remove(int begin, int end)
   {
      if (end == -1)
         end = m_count;

      if (begin < 0 || begin >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(m_count-1)} ));

      if (begin > end)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endBeforeBegin.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(end)} ));

      if (end > m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endOutOfRange.ex.txt"),
            new Object[] {new Integer(end),
            new Integer(m_count)} ));

      System.arraycopy(m_data, end, m_data, begin, m_count - end);
      m_count -= (end - begin);
   }

   /**
    * Removes a character.
    * @param index index of the character to remove
    * @exception IndexOutOfBoundsException if index is out of range [0, length()-1]
    */
   public void removeAt(int index)
   {
      if (index < 0 || index >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "indexOutOfRange.ex.txt"),
            new Object[] {new Integer(index),
            new Integer(m_count-1)} ));

      // if last character then just bump down counter
      if (index < (m_count-1))
         shiftLeft(index + 1, 1);

      m_count--;
   }

   /**
    * Empties the contents of the CharBuffer.
    */
   public void removeAll()
   {
      m_count = 0;
   }

   // TO-DO:  all of the following replace methods funnel through replace(char[],,).
   //         Investigate a more efficient means

   /**
    * Replaces a range of characters.  The characters in the specified
    * range are replaced by the characters in string.  If the length
    * of the specified String is less than or greater than the
    * amount of characters being replaced the characters from the index
    * end to the end of the CharBuffer are adjusted the number of
    * positions necessary to insert the new characters.  A value of -1
    * can be used for "end" to replace all characters from "begin" to
    * the end of the CharBuffer.
    *
    * @param string the String to insert
    * @param begin the index of the first character to be replaced
    * @param end the index 1 past the last character to be replaced
    * or -1 to specify from "start" to the end of the CharBuffer
    * @exception IndexOutOfBoundsException if begin > end,
    * if begin < 0, or if end > length()
    */
   public void replace(String string, int begin, int end)
   {
      char[] chars = string.toCharArray();
      replace(chars, begin, end);
   }

   /**
    * Replaces a range of characters.  The characters in the specified
    * range are replaced by the characters in string.  If the length
    * of the specified StringBuffer is less than or greater than the
    * amount of characters being replaced the characters from the index
    * end to the end of the CharBuffer are adjusted the number of positions
    * necessary to insert the new characters. A value of -1 can be used
    * for "end" to replace all characters from "begin" to the end of the
    * CharBuffer.
    * @param stringBuf the StringBuffer to insert
    * @param begin the index of the first character to be replaced
    * @param end the index 1 past the last character to be replaced
    * or -1 to specify from "start" to the end of the CharBuffer
    * @exception IndexOutOfBoundsException if begin > end,
    * if begin < 0, or if end > length()
    */
   public void replace(StringBuffer stringBuf, int begin, int end)
   {
      char[] chars = stringBuf.toString()/*I18nOK:MTH*/.toCharArray();
      replace(chars, begin, end);
   }

   /**
    * Replaces a range of characters.  The characters in the specified
    * range are replaced by the characters in cbuf.  If the length of
    * the specified CharBuffer is less than or greater than the
    * amount of characters being replaced the characters from the
    * index end to the end of the CharBuffer are adjusted the number
    * of positions necessary to insert the new characters.  A value
    * of -1 can be used for "end" to replace all characters from
    * "begin" to the end of the CharBuffer.
    *
    * @param cbuf the CharBuffer to insert
    * @param begin the index of the first character to be replaced
    * @param end the index 1 past the last character to be replaced
    * or -1 to specify from "start" to the end of the CharBuffer
    * @exception IndexOutOfBoundsException if begin > end,
    * if begin < 0, or if end > length()
    */
   public void replace(CharBuffer cbuf, int begin, int end)
   {
      char[] chars = cbuf.toCharArray();
      replace(chars, begin, end);
   }

   /**
    * Replaces a range of characters.  The characters in the specified
    * range are replaced by the characters in chars.  If the length of
    * the characters specified by chars is less than or greater than the
    * amount of characters being replaced the characters from the index
    * end to the end of the CharBuffer are adjusted the number of positions
    * necessary to insert the new characters. A value of -1 can be used
    * for "end" to replace all characters from "begin" to the end of the
    * CharBuffer.
    *
    * @param chars the characters to insert
    * @param begin the index of the first character to be replaced
    * @param end the index 1 past the last character to be replaced
    * or -1 to specify from "start" to the end of the CharBuffer
    * @exception IndexOutOfBoundsException if begin > end,
    * if begin < 0, or if end > length()
    */
   public void replace(char[] chars, int begin, int end)
   {
      if (end == -1)
         end = m_count;

      if (begin < 0 || begin >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(m_count-1)} ));

      if (begin > end)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endBeforeBegin.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(end)} ));

      if (end > m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endOutOfRange.ex.txt"),
            new Object[] {new Integer(end),
            new Integer(m_count)} ));

      int numRemoved = end - begin;

      // string is larger than range
      if (numRemoved < chars.length)
      {
         ensureCapacity(m_count - numRemoved + chars.length);
         shiftRight(end, chars.length - numRemoved);
         System.arraycopy(chars, 0, m_data, begin, chars.length);
         m_count += chars.length - numRemoved;
      }

      // string is smaller than range
      else if (numRemoved > chars.length)
      {
         shiftLeft(end, numRemoved - chars.length);
         System.arraycopy(chars, 0, m_data, begin, chars.length);
         m_count -= numRemoved - chars.length;
      }

      // string is same size as range
      else
      {
         System.arraycopy(chars, 0, m_data, begin, numRemoved);
      }
   }


   /**
    * Sets a character at the specified index.
    * @param index the at which to set the character
    * @param ch the character to set
    * @exception IndexOutOfBoundsException if index is out of range [0, length()-1]
    */
   public void setAt(int index, char ch)
   {
	   if (index >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "indexOutOfRange.ex.txt"),
            new Object[] {new Integer(index),
            new Integer(m_count-1)} ));

	   ensureCapacity(index);
	   m_data[index] = ch;
   }

   /**
    * Returns the number of characters currently contained in the CharBuffer.
    * @return the length of the CharBuffer in characters
    */
   public int length()
   {
      return m_count;
   }

   /**
    * Returns the current capacity of the CharBuffer.
    * @return the character capacity of the CharBuffer
    */
   public int capacity()
   {
      return m_data.length;
   }

   /**
    * Returns true if empty, otherwise false.
    * @return true if empty, otherwise false
    */
   public boolean isEmpty()
   {
      return (m_count > 0) ? false : true;
   }

   /**
    * Returns a character.
    * @param index the index of the character to return.
    * @return the character at specified index
    * @exception IndexOutOfBoundsException if index is out of range [0, length()-1]
    */
   public char charAt(int index)
   {
	   if (index >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "indexOutOfRange.ex.txt"),
            new Object[] {new Integer(index),
            new Integer(m_count-1)} ));

      return m_data[index];
   }

   /**
    * Copies the character contained in the CharBuffer into the destination
    * character array.  The characters copied will start with the character
    * at scrBegin and end with the character at srcEnd - 1. A value of -1
    * can be used for "end" to replace all characters from "begin" to
    * the end of the CharBuffer.
    * <br>
    * @param srcBegin the index of the first character to be copied
    * @param srcEnd the index 1 past the last character to be copied  or -1
    * to specify from "srcBegin" to the end of the CharBuffer
    * @param dst the destination character array
    * @param dstBegin the index of the first position to be assigned in
    * the destination character array
    * @exception IndexOutOfBoundsException if srcBegin > srcEnd,
    * if srcBegin < 0, or if srcEnd > length()
    * @exception ArrayIndexOutOfBoundsException if dstBegin > dst.length - 1
    */
   public void getChars(int srcBegin, int srcEnd, char[] dst, int dstBegin)
   {
      if (dstBegin >= dst.length)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "indexOutOfRange.ex.txt"),
            new Object[] {new Integer(dstBegin),
            new Integer(dst.length)} ));

      if (srcEnd == -1)
         srcEnd = m_count;

      if (srcBegin < 0 || srcBegin >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(srcBegin),
            new Integer(m_count-1)} ));

      if (srcBegin > srcEnd)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endBeforeBegin.ex.txt"),
            new Object[] {new Integer(srcBegin),
            new Integer(srcEnd)} ));

      if (srcEnd > m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endOutOfRange.ex.txt"),
            new Object[] {new Integer(srcEnd),
            new Integer(m_count)} ));

      // all is ok so copy the characters
      System.arraycopy(m_data, srcBegin, dst, dstBegin, srcEnd - srcBegin);
   }

   /**
    * Returns the characters contained in the CharBuffer.
    */
   public char[] toCharArray()
   {
      char newData[] = new char[m_count];
      System.arraycopy(m_data, 0, newData, 0, m_count);

	 return newData;
   }

   /**
    * Returns the specified range of characters contained in
    * the CharBuffer.  Specifying -1 as end will return all
    * of the characters from begin to the end of the Charbuffer.
    * A value of -1 can be used for "end" to replace all characters
    * from "begin" to the end of the CharBuffer.
    *
    * @param begin the index of the first character to return
    * @param end the index 1 past the last character to return or -1
    * to specify from "begin" to the end of the CharBuffer
    * @return the characters contained in the CharBuffer in the specified range
    * @exception IndexOutOfBoundsException if begin > end,
    * if begin < 0, or if end > length()
    */
   public char[] toCharArray(int begin, int end)
   {
      if (end == -1)
         end = m_count;

      if (begin < 0 || begin >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(m_count-1)} ));

      if (begin > end)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endBeforeBegin.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(end)} ));

      if (end > m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endOutOfRange.ex.txt"),
            new Object[] {new Integer(end),
            new Integer(m_count)} ));

      char newData[] = new char[end - begin];
      System.arraycopy(m_data, begin, newData, 0, end - begin);

      return newData;
   }

   /**
    * Returns the first "count" characters of the CharBuffer.
    * @param count the number of characters to return
    * @return the first "count" characters of the CharBuffer
    */
   public CharBuffer leftCharBuffer(int count)
   {
      return subCharBuffer(0, count);
   }

   /**
    * Returns the last "count" characters of the CharBuffer.
    * @param count the last number of characters to return
    * @return the last "count" characters of the CharBuffer
    */
   public CharBuffer rightCharBuffer(int count)
   {
      return subCharBuffer(m_count - count, m_count);
   }

   /**
    * Get a range of characters contained in the CharBuffer.
    * A value of -1 can be used for "end" to replace all
    * characters from "begin" to the end of the CharBuffer.
    * @param begin the beginning index
    * @param end the ending index or -1 to specify from "begin" to the
    * end of the CharBuffer
    * @return the specified range of characters as a CharBuffer.
    * @exception IndexOutOfBoundsException if begin > end, if begin < 0, or if end > length()
    */
   public CharBuffer subCharBuffer(int begin, int end)
   {
      if (end == -1)
         end = m_count;

      if (begin < 0 || begin >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(m_count-1)} ));

      if (begin > end)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endBeforeBegin.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(end)} ));

      if (end > m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endOutOfRange.ex.txt"),
            new Object[] {new Integer(end),
            new Integer(m_count)} ));

      CharBuffer retBuf = new CharBuffer(end - begin);

      if (end <= m_count)
      {
	     System.arraycopy(m_data, begin, retBuf.m_data, 0, end - begin);
	     retBuf.m_count = end - begin;
         return retBuf;
	 }
	 else
         return retBuf;
   }

   /**
    * Get a range of characters contained in the CharBuffer. A value of -1
    * can be used for "end" to replace all characters from "begin" to the
    * end of the CharBuffer.
    * @param begin the beginning index
    * @param end the ending index or -1 to specify from "begin" to the
    * end of the CharBuffer
    * @return the specified range of characters as a String.
    * @exception IndexOutOfBoundsException if begin > end,
    * if begin < 0, or if end > length()
    */
   public String substring(int begin, int end)
   {
      if (end == -1)
         end = m_count;

      if (begin < 0 || begin >= m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "startOutOfRange.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(m_count-1)} ));

      if (begin > end)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endBeforeBegin.ex.txt"),
            new Object[] {new Integer(begin),
            new Integer(end)} ));

      if (end > m_count)
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "endOutOfRange.ex.txt"),
            new Object[] {new Integer(end),
            new Integer(m_count)} ));

	   String s;
	   s = new String(m_data, begin, end - begin);
        return s;
   }

   /**
    * Trims whitespace from both ends of the CharBuffer.
    */
   public void trim()
   {
      trimLeft();
      trimRight();
   }

   /**
    * Trims whitespace from the beginning of the CharBuffer.
    */
   public void trimLeft()
   {
      int begin = 0;

      while ((begin < m_count) && Character.isWhitespace(m_data[begin]))
	    begin++;

      System.arraycopy(m_data, begin, m_data, 0, m_count - begin);
      m_count -= begin;
   }

   /**
    * Trims whitespace from the end of the CharBuffer.
    */
   public void trimRight()
   {
      int begin = m_count - 1;

      while ((begin > 0) && Character.isWhitespace(m_data[begin]))
	    begin--;

      // since we are equating begin to count we must add 1 back that we took away above
      m_count = begin + 1;
   }

   /**
    * Reverses the characters in the current CharBuffer.
    */
   public void reverse()
   {
      if (m_count <= 1)
         return;

      char ch;

      for (int i=0; i< m_count/2; i++)
      {
         ch = m_data[i];
         m_data[i] = m_data[(m_count - 1) - i];
         m_data[(m_count - 1) - i] = ch;
      }
   }

   /**
    * Returns true if the CharBuffer ends with the specified String, false otherwise.
    * @param string the string to determine that the CharBuffer ends with
    * @return true if the Charbuffer ends with the specified String, false otherwise
    */
   public boolean endsWith(String string)
   {
      return regionMatches(false, string, m_count - string.length(), m_count);
   }

   /**
    * Returns true if the CharBuffer starts with the specified String, false otherwise.
    * @param string the string to determine that the CharBuffer starts with
    * @return true if the CharBuffer starts with the specified String, false otherwise
    */
   public boolean startsWith(String string)
   {
      return regionMatches(false, string, 0, string.length());
   }

   /**
    * Returns the index of the specified character.  The search is started at
    * the location specified by offset.
    * @param ch the character being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the character if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int indexOf(char ch, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

	   // if find the character return the index
       for (int i = offset; i < (m_count - 1) ;i++)
	   {
	      if (m_data[i] == ch)
		    return i;
	   }

       // didn't find it
	   return -1;
   }

   /**
    * Returns the index of the specified character array.  The search is started at
    * the location specified by offset.
    * @param chars the character array being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the beginning character of the character array if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int indexOf(char[] chars, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

 	   int charCount, i, k = -1;

       // if the string is empty, return the startIndex
       if (chars.length == 0)
         return offset;

        char first  = chars[0];
	    charCount = m_count - chars.length;

        for (i=offset; i<=charCount; i++)
        {
           // we found an occurence of the first char
           if (first == m_data[i])
           {
              // check for the rest of the char's to match and bump up counter k
              for (k=1; k < chars.length; k++)
                 if (chars[k] != m_data[i+k])
                    break;

              // if k is equal to the length of the string then we have a match
              if (k == chars.length)
                 return i;
           }
        }

        // no match
        return -1;
   }

   /**
    * Returns the index of the specified CharBuffer.  The search is started at
    * the location specified by offset.
    * @param cbuf the CharBuffer being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the beginning character of the CharBuffer if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int indexOf(CharBuffer cbuf, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

 	   int charCount, i, k = -1;

       // if the string is empty, return the startIndex
       if (cbuf.m_count == 0)
         return offset;

        char first  = cbuf.m_data[0];
	    charCount = m_count - cbuf.m_count;

        for (i=offset; i<=charCount; i++)
        {
           // we found an occurence of the first char
           if (first == m_data[i])
           {
              // check for the rest of the char's to match and bump up counter k
              for (k=1; k < cbuf.m_count; k++)
                 if (cbuf.m_data[k] != m_data[i+k])
                    break;

              // if k is equal to the length of the string then we have a match
              if (k == cbuf.m_count)
                 return i;
           }
        }

        // no match
        return -1;
   }

   /**
    * Returns the index of the specified String.  The search is started at
    * the location specified by offset.
    * @param string the String being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the beginning character of the string if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int indexOf(String string, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

 	   int length, charCount, i, k = -1;

       // if the string is empty, return the startIndex
       if ((length = string.length()) == 0)
         return offset;

        char first  = string.charAt(0);
	    charCount = m_count - length;

        for (i=offset; i<=charCount; i++)
        {
           // we found an occurence of the first char
           if (first == m_data[i])
           {
              // check for the rest of the char's to match and bump up counter k
              for (k=1; k<length; k++)
                 if (string.charAt(k) != m_data[i+k])
                    break;

              // if k is equal to the length of the string then we have a match
              if (k == length)
                 return i;
           }
        }

        // no match
        return -1;
   }

   /**
    * Returns the index of the specified StringBuffer.  The search is started at
    * the location specified by offset.
    * @param stringBuf the StringBuffer being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the beginning character of the StringBuffer if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int indexOf(StringBuffer stringBuf, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

 	   int length, charCount, i, k = -1;

       // if the string is empty, return the startIndex
       if ((length = stringBuf.length()) == 0)
         return offset;

        char first  = stringBuf.charAt(0);
	    charCount = m_count - length;

        for (i=offset; i<=charCount; i++)
        {
           // we found an occurence of the first char
           if (first == m_data[i])
           {
              // check for the rest of the char's to match and bump up counter k
              for (k=1; k<length; k++)
                 if (stringBuf.charAt(k) != m_data[i+k])
                    break;

              // if k is equal to the length of the string then we have a match
              if (k == length)
                 return i;
           }
        }

        // no match
        return -1;
   }



   /**
    * Returns the index of a character contained in the specified String.  The search is started at
    * the location specified by offset.
    * @param string the String containing the characters being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the beginning character of the string if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int indexOfChar(String string, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

 	   int length, charCount, i, k = -1;
       char[] chars = string.toCharArray();

       // if the string is empty, return the startIndex
       if ((length = string.length()) == 0)
         return -1;

        // loop throught the characters starting at offset
        for (i=offset; i<m_count; i++)
        {
           // loop through the characters in the string and compare to the current character in m_data
           for (k=0; k<length; k++)
           {
              if (m_data[i] == chars[k])
                 return i;
           }
        }

        // no match
        return -1;
   }

   /**
    * Returns the last index of the specified character.  The search is performed backwards starting at
    * the location specified by offset.
    * @param ch the character being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the character if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int lastIndexOf(char ch, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

	   // if find the character return the index
       for (int i = offset; i >= 0; i--)
	   {
	      if (m_data[i] == ch)
		    return i;
	   }

       // didn't find it
	   return -1;
   }

    /**
    * Returns the last index of the specified character array.  The search is performed backwards starting at
    * the location specified by offset.
    * @param chars the character array being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the character array if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int lastIndexOf(char[] chars, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

      int index, lastIndex;
      index = lastIndex = indexOf(chars, 0);

      while ((index != -1) && (index <= offset))
      {
         lastIndex = index;
         index = indexOf(chars, index + 1);
      }

      if ((index == -1) || (index > offset))
         index = lastIndex;

      return index;
   }

  /**
    * Returns the last index of the specified CharBuffer.  The search is performed backwards starting at
    * the location specified by offset.
    * @param cbuf the CharBuffer being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the CharBuffer if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int lastIndexOf(CharBuffer cbuf, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

      int index, lastIndex;
      index = lastIndex = indexOf(cbuf, 0);

      while ((index != -1) && (index <= offset))
      {
         lastIndex = index;
         index = indexOf(cbuf, index + 1);
      }

      if ((index == -1) || (index > offset))
         index = lastIndex;

      return index;
   }

  /**
    * Returns the last index of the specified String.  The search is performed backwards starting at
    * the location specified by offset.
    * @param string the String being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the string if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int lastIndexOf(String string, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

      int index, lastIndex;
      index = lastIndex = indexOf(string, 0);

      while ((index != -1) && (index <= offset))
      {
         lastIndex = index;
         index = indexOf(string, index + 1);
      }

      if ((index == -1) || (index > offset))
         index = lastIndex;

      return index;
   }

  /**
    * Returns the last index of the specified StringBuffer.  The search is performed backwards starting at
    * the location specified by offset.
    * @param stringBuf the StringBuffer being searched for
    * @param offset the offset at which to begin searching
    * @return the index of the StringBuffer if found, -1 otherwise
    * @exception IndexOutOfBoundsException if offset is out of range [0, length()-1]
    */
   public int lastIndexOf(StringBuffer stringBuf, int offset)
   {
      if ((offset < 0) || (offset >= m_count))
         throw new IndexOutOfBoundsException(MessageFormat.format(
            RB.getStringResource(RB_KEY, "offsetOutOfRange.ex.txt"),
            new Object[] {new Integer(offset),
            new Integer(m_count-1)} ));

      int index, lastIndex;
      index = lastIndex = indexOf(stringBuf, 0);

      while ((index != -1) && (index <= offset))
      {
         lastIndex = index;
         index = indexOf(stringBuf, index + 1);
      }

      if ((index == -1) || (index > offset))
         index = lastIndex;

      return index;
   }

   /**
    * Test the equality of specified CharBuffer with the current CharBuffer.
    * @return true if the 2 CharBuffer objects are equal, false otherwise
    */
   public boolean equals(Object object)
   {
      if (object == null || !(object instanceof CharBuffer))
         return false;

      CharBuffer buf = (CharBuffer)object;
      if (m_count != buf.m_count)
         return false;

      for (int i=0; i<m_count; i++)
         if (m_data[i] != buf.m_data[i])
            return false;

      return true;
   }

   /**
    * Test the equality of specified CharBuffer with the current CharBuffer.  The case of any characters
    * is ignored during comparison.
    * @return true if the 2 CharBuffer objects are equal, false otherwise
    */
   public boolean equalsIgnoreCase(Object object)
   {
      if (object == null || !(object instanceof CharBuffer))
         return false;

      CharBuffer buf = (CharBuffer)object;
      if (m_count != buf.m_count)
         return false;

      // check both upper and lower case, if neither is true then unequality
      for (int i=0; i<m_count; i++)
         if ( (Character.toLowerCase(m_data[i]) != Character.toLowerCase(buf.m_data[i]))
               && (Character.toUpperCase(m_data[i]) != Character.toUpperCase(buf.m_data[i])) )
            return false;

      return true;
   }

   /**
    * Returns a hash code value for the CharBuffer.
    * @return a hash code value for the CharBuffer
    */
   public int hashCode()
   {
      return toString().hashCode();
   }

   /**
    * Returns the contents of the CharBuffer as a StringBuffer object.
    * @return the contents of the CharBuffer as a StringBuffer object
    */
   public StringBuffer toStringBuffer()
   {
	   StringBuffer retBuf = new StringBuffer(m_count);
	   retBuf.append(m_data, 0, m_count);
	   return retBuf;
   }

   /**
    * Returns the contents of the CharBuffer as a String object.
    * @return the contents of the CharBuffer as a String object
    */
   public String toString()
   {
	   return new String(m_data, 0, m_count);
   }

   //////////////////////////////
   //  PublicClonable method...
   //////////////////////////////

   /**
     * @see com.sas.PublicClonable
     */
   public Object clone() throws CloneNotSupportedException
   {
      CharBuffer cbuf = (CharBuffer)super.clone();
      cbuf.m_data = new char[m_data.length];
      System.arraycopy(m_data, 0, cbuf.m_data, 0, m_count);
      cbuf.m_count = m_count;

      return cbuf;
   }


   ///////////////////////////
   //  Private methods ...
   ///////////////////////////

   private void ensureCapacity(int capacity)
   {
      if (capacity > m_data.length)
      {
         int newSize = m_data.length * 2;

         if (capacity > newSize)
            newSize = capacity;

         char newData[] = new char[newSize];
         System.arraycopy(m_data, 0, newData, 0, m_count);
         m_data = newData;
      }
   }

   /**
     * Shifts the contents of the data buffer starting at the "begin" offset the "amount"
     * number of positions to the left.
     */
   private void shiftLeft(int begin, int amount)
   {
      // NOTE:  need to check for valid indexes
      int offset = begin - amount;
      if (offset < 0)
         offset = 0;

      System.arraycopy(m_data, begin, m_data, offset, m_count - begin);
   }

   /**
     * Shifts the contents of the data buffer starting at the "begin" offset the "amount"
     * number of positions to the right.
     */
   private void shiftRight(int begin, int amount)
   {
      // NOTE:  need to check for valid indexes
      int offset = begin + amount;

      System.arraycopy(m_data, begin, m_data, offset, m_count - begin);
   }

   /**
     * Determines if the current CharBuffer contains a region that matches a given string
     * in a certain range.
     */
   private boolean regionMatches(boolean ignoreCase, String string, int begin, int end)
   {
      if (begin > end || (end - begin) != string.length())
         return false;

      char cha, chb;
      for (int i=begin, j=0; i<end; i++, j++)
      {
         cha = m_data[i];
         chb = string.charAt(j);

         if (!ignoreCase)
            if (cha != chb)
               return false;
         else
         {
            if ( (Character.toLowerCase(cha) != Character.toLowerCase(chb))
                  && (Character.toUpperCase(cha) != Character.toUpperCase(chb)) )
                  return false;
         }
      }

      return true;
   }
}
