// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import com.sas.GlobalInfoTable;
/**
 * Auxiliary information about Java classes, etc.
 * @author <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public class Classes {


  // private constructor prevents instantiation
  private Classes()
  {
  }

  /**
   * Test if a class implements a specific interface, by name.
   * @param thisClass test if this class implements the named interface
   * @param interfaceName the fully qualified name of the interface.
   * @return <code>true</code> if the class implements the named interface. This is true
   *         if the class directly implements the interface or
   *         implements an interface which extends this interface,
   *         or if the class inherits from a class which implements the
   *         interface.
   */
  public static boolean implementsInterface(Class thisClass, String interfaceName)
  {
    Class interfaceClass = null;
    try
      {
//        ClassLoader systemClassLoader = ClassLoader.getSystemClassLoader();
//        interfaceClass = Class.forName(interfaceName, true, systemClassLoader);
		  interfaceClass = GlobalInfoTable.loadClass( interfaceName );
        return implementsInterface(thisClass, interfaceClass);
      }
    catch (Exception e) // ClassNotFoundException, NullPointerException, ....
      {
        return false;
      }
  }

  /**
   * Test if a class implements an interface.
   * @param thisClass test if this class implements the named interface
   * @param interfaceClass a interface class object
   * @return <code>true</code> if the class implements the interface. This is true
   *         if the class directly implements the interface or
   *         implements an interface which extends this interface,
   *         or if the class inherits from a class which implements the
   *         interface.
   */
  public static boolean implementsInterface(Class thisClass, Class interfaceClass)
  {
    return (   interfaceClass.isInterface()
            && implementsInterfaceInternal(thisClass, interfaceClass));
  }

  // private static String indent = "";
  private static boolean implementsInterfaceInternal(Class thisClass, Class interfaceClass)
  {
    // String originalIndent = indent;
    // indent += "  ";
    // System.out.println(indent + "implementsInternal(" + thisClass + ", " + interfaceClass);
    while (thisClass != null)
      {
        Class interfaces[] = thisClass.getInterfaces();
        // System.out.println(indent + thisClass + " implements " + interfaces.length + " interfaces");
        if (interfaces != null)
          for (int i = 0; i < interfaces.length; i++)
            {
              // System.out.println(indent + " interface["+ i + "] : " + interfaces[i]);
              if (   (interfaces[i] == interfaceClass)
                     || implementsInterfaceInternal(interfaces[i], interfaceClass))
                return true;
            }
        // System.out.print(indent + thisClass + ".getSuperclass()" );
        thisClass = thisClass.getSuperclass();
        // System.out.println("-> " + thisClass );
      }
    // indent = originalIndent;
    return false;
  }

  /**
   * Test if a class extends (that is is derived from) another class
   * @param thisClass test if this class extends the named class
   * @param className the name of a potential super class
   * @return <code>true</code> if <var>thisClass</var> extends the class named by <var>className</var>.
   */
  public static boolean extendsClass(Class thisClass, String className)
  {
    Class superClass = null;
    try
      {
//        ClassLoader systemClassLoader = ClassLoader.getSystemClassLoader();
//        superClass = Class.forName(className, true, systemClassLoader);
		superClass = GlobalInfoTable.loadClass( className );
        return extendsClass(thisClass, superClass);
      }
    catch (Exception e) // ClassNotFoundException, NullPointerException, ....
      {
        return false;
      }
  }


  /**
   * Test if a class extends (is derived from) another class
   * @param thisClass test if this class extends the named class
   * @param className a potential super class.
   * @return <code>true</code> if <var>thisClass</var> extends (i.e. is derived from) the <var>superClass</var>
   *         This is true if <code><var>thisClass</var> == <var>superClass</var></code> or if
   *         the <var>superClass</var> is a ancestor class of <var>thisClass</var>
   */
  public static boolean extendsClass(Class thisClass, Class superClass)
  {
    while (thisClass != null)
      if (thisClass == superClass)
        return true;
      else
        thisClass = thisClass.getSuperclass();
    return false;
  }

  /**
   * Test if an object is an array.
   * @param thisClass test if this class is an array class
   * @return <code>true</code> if and only if <var>object</var> is an array.
   */
  public static boolean isArray(Object object)
  {
    return object != null && object.getClass().isArray();
  }

  /**
   * If a class loader is passed, try loading the class from it.  If that fails 
   * and an object is passed, try loading a class from it.  
   *
   * @param className the well-formed name of a class to load.
   * @param loader a class loader to try loading the class from. May be null
   * @param object a object whose class loader we try loading the class from, if
   *        loading from the <var>loader</var> failed. May be null.
   * @return a class
   * @throws ClassNotFoundException if the class could not be loaded by any of these
   *         four attempts
   */
  public static Class loadClass(String className,
                                ClassLoader loader,
                                Object object) throws ClassNotFoundException
    {
      if ( loader != null )
        try { return loader.loadClass(className); }
        catch (ClassNotFoundException e) {  }

      if ( object != null ) {
         if (object instanceof Class)
            loader = ((Class)object).getClassLoader();
         else
            loader = object.getClass().getClassLoader();
         if (loader != null) {
            try { return loader.loadClass(className); }
           catch (ClassNotFoundException e) {  }
        }
      }

      try { return Class.forName(className); }
        catch (ClassNotFoundException e) {  }

      try { return Class.forName(className, true, ClassLoader.getSystemClassLoader()); }
      // this will be thrown if loader == null, object == null, and we are in the applet
      // class loader -- this can only occur using exploded classes
      catch (SecurityException e) { }


      throw new ClassNotFoundException(className);

    }

}
