// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import com.sas.lang.StringDataInterface;
import com.sas.util.SimpleCommand;
import com.sas.PublicClonable;


/**
 * A <code>Command</code> interface. <code>Command</code>  objects
 * extend <code>{@link com.sas.util.SimpleCommand}</code>
 * and adds command undo operations. The <code>Command</code> object can indicate if
 * it can be executed or not. This is useful if <code>Command</code> objects
 * are attached to menu items or buttons.  You can disable the menu items
 * or button if the command is not executable at the current time.
 * You can also check if a <code>Command</code> can be undone, and if so, you
 * can invoke the <code>{@link #undo()}</code> method to undo the action the command performed.
 * Thus, by maintaining a stack of <code>Command</code> objects, you can
 * implement an undo history. (The <code>Command</code> object is responsible
 * for maintaining internal state in order to implement <code>undo()</code>).
 * <p>
 * The <code>Command</code> interface extends
 * {@link com.sas.lang.StringDataInterface <code>StringDataInterface</code>},
 * through which the label for the command object may be obtained for display
 * in a user interface.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.1
 * @see com.sas.util.SimpleCommand
 */

public interface Command
       extends   PublicClonable,
                 SimpleCommand,
                 StringDataInterface
{

  /**
   * Test whether the <code>Command</code> object in in a state in which
   * it can execute. Note that
   * it should do no harm to actually call <code>{@link com.sas.util.SimpleCommand#execute(java.lang.Object) execute(object)}</code> even
   * if <code>isExecutable()</code> returns <code>false</code>. Basically, this test
   * can be used in GUI situations to gray out a GUI
   * component that would normally invoke the
   * <code>execute()</code> method when selected.
   * @param object the potential argument for the execute method.
   * @return <code>true</code> if it is valid to call the execute method,
   *         <code>false</code> if <code>execute()</code> should not be called.
   */
  public boolean isExecutable();

  /**
   * Test if the command can be undone.  This is normally
   * done after the command has been executed and placed
   * on an undo history. Not all commands can be undone.
   * @see #undo()
   */
  public boolean isUndoable();

  /**
   * Undo the command. This
   * should undo the work that <code>{@link com.sas.util.SimpleCommand#execute(java.lang.Object) execute(object)}</code> performed.
   * Note that <code>{@link #undo()}</code> should still "work"
   * (i.e. run without throwing an exception) even if it is called
   * when <code>{@link #isUndoable()}</code> returns false.
   * @see com.sas.util.SimpleCommand#execute(Object)
   */
  public void undo();

  /**
   * Commands are clonable. The framework which uses commands
   * may clone a <code>Command</code> object after invoking it, if it
   * wishes to store the command on an undo history or
   * otherwise maintain a reference to the Command object.
   * The clone should preserve the undo information so
   * that the cloned command stored on an undo history
   * may be undone.
   * @return the cloned object.
   * @exception java.lang.CloneNotSupportedException if
   *            the command cannot be cloned.
   */
  public Object clone() throws CloneNotSupportedException;

}
