// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

/**
 * An abstract comparator class which defines a comparison
 * method to be used for things like sorting.
 * <p>
 * Java does not support operator overloading, so for a class <code>C</code>
 * and two instances of <code>C</code> in
 * variables <var>x</var> and <var>y</var>, you cannot write relational expressions
 * like <code>x &lt; y</code> or </code>x &gt;= y</code>. Instead, you implement a Comparator
 * which works for objects of class <code>C</code>:
 * <pre>
 * public class CComparator implements com.sas.util.Comparator {
 *   public int compare(Object a, Object b) {
 *       C c1 = (C) a, c2 = (C) b;
 *       int result;
 *       // perform comparison based on c1, c2 instances of C
 *       return result;
 *       }
 * }
 * </pre>
 * <p>
 * To use the <code>Comparator</code>, create an instance:
 * <pre>CComparator comparator = new CComparator();</pre>
 * Then, to implement the various conditions on the left in the table below,
 * use the corresponding expression on the right:
 * <table  cols=2 frame=box rules=all border=2>
 * <tr><td>desired comparison:<td>use:</tr>
 * <tr><td>x &lt; y<td>comparator.compare(x, y) &lt; 0</tr>
 * <tr><td>x &lt;= y<td>comparator.compare(x, y) &lt;= 0</tr>
 * <tr><td>x == y<td>x.equals(y) or comparator.compare(x, y) == 0</tr>
 * <tr><td>x &gt;= y<td>comparator.compare(x, y) &gt;= 0</tr>
 * <tr><td>x &gt; y<td>comparator.compare(x, y) &gt; 0</tr>
 * </table>
 * That is, for any comparison operator <em>op</em>, for
 * <code>x <em>op</em> y</code>, use <code>comparator.compare(x,y) <em>op</em> 0)</code>
 * Note that for equality comparison, you normally will use the {@link
 * java.lang.Object#equals(java.lang.Object) equals(object)}
 * method. (This normally implies that you implement an equals method for your class.)
 * Still, a <code>Comparator</code> should return <code>{@link #EQUALS}</code> if <code>x.equals(y)</code>.
 * <p>
 * For a comparator to work when sorting a collection, it must be able to
 * handle all instance types found in the collection. Normally, you will
 * have a homogeneous collection and comparator designed specifically
 * for the types stored in the collection is used. Your comparator need only
 * cast the arguments to the expected type. See
 * {@link com.sas.collection.StringCollection} and {@link com.sas.util.StringComparator}, for example.
 * If you want to sort
 * homogenous collections, your <code>Comparator</code> must be prepared for the
 * operands to be of various types. You must also enforce a rule that if
 * <code>compare(x, y)</code> returns <code>{@link #LESS_THAN}</code>,
 * then <code>compare(y, x)</code> should return
 * <code>{@link #GREATER_THAN}</code>, and vice versa, and that if
 <code>compare(x, y)</code> returns <code>{@link #EQUALS}</code>,
 * then <code>compare(y, x)</code> returns <code>EQUALS</code> as well. Otherwise, a sort algorithm
 * may not sort correctly, or even loop infinitely.
 * <p>
 * A <code>Comparator</code> can return <code>{@link com.sas.util.Comparator#INCOMPARABLE}</code> if it does not know how to
 * comparate the items. You can avoid this in most cases by establishing
 * rules such as null is less than all non-null values, then
 * fall back on comparing the class names or the identity hash codes.
 * <h2>Tips</h2>
 * <p>
 * If you want to evaluate the various sorting algorithms in the
 * {@link com.sas.util.Sort Sort} class,
 * you can use a comparator which contains a <code>int count</code> field
 * and each time the compare method is invoked, increment the count.
 * Be sure to reset the count before sorting. After sorting, the
 * count will tell how many times the sort algorithm needed to
 * compare elements. You can use this information to fine-tune
 * your application by choosing the sort algorithm which
 * works best for your data.
 * <p>
 * Below is the definition of a comparator
 * which compares items by their identity hash code.
 * This is useful for making faster set operations
 * (such as set intersection/union) using array or collection
 * representations of sets. By sorting
 * the items first (in <em>O(n log n)</em> time), the intersection and union operations
 * can be done in <em>O(n)</em> time instead of <em>O(n<sup>2</sup>)</em> time for unsorted lists.
 * <p>
 * For example, in one case of scanning two collections of 2,000+ items for common
 * elements, two nested for loops consumed 16000ms, while sorting the two
 * loops using this <code>IdentityComparator</code> (&lt;200ms each) and then scanning the collections
 * with a single loop and comparing items consumed only 30ms,
 * for a total of &lt;430ms, an 37x performance improvement.
 * <pre>
 * public class IdentityComparator implements Comparator
 * {
 *   public int compare(Object a, Object b)
 *     {
 *       if (a == b)      return EQUALS;
 *       if (a == null)   return LESS_THAN;
 *       if (b == null)   return GREATER_THAN;
 *       // ensure positive long integers.
 *       long aId = ((long) System.identityHashCode(a)) & 0xffffL;
 *       long bId = ((long) System.identityHashCode(b)) & 0xffffL;
 *       // objects not ==, so identity hash codes can't be ==, only < or >
 *       return (aId < bId ? LESS_THAN : GREATER_THAN );
 *     }
 * }
 * </pre>
 * @author David Biesack, Ernest Pasour
 * @version 1.0
 * @see com.sas.util.StringComparator
 */


public interface Comparator extends java.util.Comparator
{
  /**
   * Return value indicating the first value compares less than the second.
   */
  static final int LESS_THAN    = -1; // used to be 2 (second item is greater)
  /**
   * Return value indicating the first value compares is equal to the second.
   */
  static final int EQUALS       = 0;
  /**
   * Return value indicating the first value compares greater than the second.
   */
  static final int GREATER_THAN = 1;
  /**
   * Return value indicating the the <code>Comparator</code> cannot compare the values.
   */
  static final int INCOMPARABLE = Integer.MAX_VALUE;

  /**
   * Compare two objects to see if the first is less than, equal to,
   * or greater than the second, or if they are simply incomparable.
   * @param first the first object to use in the comparison
   * @param second the second object to use in the comparison
   * @return
   * <table>
   * <tr><td>LESS_THAN     <td>the first object compares "less than" the second, meaning it
   *                           should appear before the second in the sorted result
   * <tr><td>EQUALS       <td>the objects are equal
   * <tr><td>GREATER_THAN <td>the first object compares greater than the first, meaning it
   *                           should appear after the second in the sorted result
   * <tr><td>IMCOMPARABLE <td>the data could not be compared.
   * </table>
   */
  int compare(Object first, Object second);
}
