/* Copyright (c) 1999 by SAS Institute Inc., Cary, NC 27513 */
 package com.sas.util;

 /**
   * A command which combines two other commands.
   * This is useful when you want to apply more than one command to a set of
   * items without traversing the set more than once.
   * For example,
   * <blockquote><code><pre>
   * import java.awt.Cursor;
   * import com.sas.awt.util.ChangeCursorCommand;
   * import com.sas.awt.util.ChangeEnabledStateCommand;
   * // create command to disable components and set their cursor "busy"
   * Cursor busyCursor = Cursor.getPredifinedCursor (Cursor.WAIT_CURSOR);
   * ChangeCursorCommand cmd1 = new ChangeCursorCommand (busyCursor);
   * ChangeEnabledStateCommand cmd2 = new ChangeEnabledStateCommand (false);
   * CompositeCommand cmd = new CompositeCommand (cmd1, cmd2);
   * // apply the command to our container and all its children
   * com.sas.awt.util.Util.apply (container, cmd);
   * // do some work
   * [...]
   * // restore the container and its children back to their original states
   * cmd1.undoAll ();
   * cmd2.undoAll ();
   * </blockquote></code></pre>
   * @author saskjp Ken Peressini
   * @history
   * <dl>		 
   * <dt>November 3, 1999<dd>Created.
   * </dl>
  */

 public class CompositeCommand implements Command, ApplyInterface
    {
    private Command _first;
    private Command _second;

 /**
   * Constructor.
   * @param first
   *    the command to execute first
   * @param second
   *    the command to execute second
   * @see #getFirstCommand
   * @see #getSecondCommand
   * @see #execute
  */
    public CompositeCommand (Command first, Command second)
       {
       if (first == null || second == null)
          throw new NullPointerException();
       _first = first;
       _second = second;
       }

 /**
   * Clones this command.
   * @return
   *    a new command equivalent to this one
   * @exception CloneNotSupportedException
   *    this class supports cloning, but a subclass may not
  */
    public Object clone () throws CloneNotSupportedException
       {
       return super.clone ();
       }

 /**
   * Returns the first command executed by <code>execute()</code>.
   * @return
   *    the first command; will never be <code>null</code>
  */
    public Command getFirstCommand ()
       {
       return _first;
       }

 /**
   * Returns the second command executed by <code>execute()</code>.
   * @return
   *    the second command; will never be <code>null</code>
  */
    public Command getSecondCommand ()
       {
       return _second;
       }

 /**
   * Returns <code>true</code> if either command is executable.
   * @return
   *    <code>true</code> if either of the commands is executable;
   *    <code>false</code> otherwise
   * @see #execute
  */
    public boolean isExecutable ()
       {
       return getFirstCommand().isExecutable() 
           || getSecondCommand().isExecutable();
       }

 /**
   * Executes both commands.
   * Executes the first command, then the second.
   * @param obj
   *    passed to the <code>execute(Object)</code> method of both commands
   * @see #undo
  */
    public void execute (Object obj)
       {
       getFirstCommand().execute (obj);
       getSecondCommand().execute (obj);
       }

 /**
   * Executes both commands, like <code>execute()</code>, but uses
   * <code>ApplyInterface</code> if available.
   * <p>
   * If the first command implements <code>ApplyInterface</code>, then
   * its apply method is called, otherwise its execute method is called.
   * If the apply method was called and did not return 
   * <code>ApplyInterface.CONTINUE</code>, then the second command will not
   * be executed and the value will be returned.
   * <p>
   * The second command is handled similarly.
   * @param obj
   *    passed to the <code>apply(Object)</code> or <code>execute(Object)</code>
   *    methods of both commands
   * @return
   *    <code>ApplyInterface.CONTINUE</code> or
   *    <code>ApplyInterface.ABORT</code>
   * @see #execute
  */
    public int apply (Object obj)
       {
       int rc = CONTINUE;

       Command first = getFirstCommand();
       if (first instanceof ApplyInterface)
          rc = ((ApplyInterface)first).apply (obj);
       else
          first.execute (obj);

       if (rc == CONTINUE)
          {
          Command second = getSecondCommand();
          if (second instanceof ApplyInterface)
             rc = ((ApplyInterface)second).apply (obj);
          else
             second.execute (obj);
          }

       return rc;
       }

 /**
   * Returns <code>true</code> if both commands are undoable.
   * @return
   *    <code>true</code> if both of the commands are undoable;
   *    <code>false</code> otherwise
   * @see #undo
  */
    public boolean isUndoable ()
       {
       return getFirstCommand().isUndoable() && getSecondCommand().isUndoable();
       }

 /**
   * Undoes both commands.
   * Undoes the second command, then the first.
   * @see #execute
  */
    public void undo ()
       {
       getSecondCommand().undo();
       getFirstCommand().undo();
       }

 /**
   * Placeholder to satisy interface requirements.
   * @return
   *    the empty string ("")
   * @see #setText
  */
    public String getText ()
       {
       return "";
       }

 /**
   * Placeholder to satisy interface requirements; does nothing.
   * @see #getText
  */
    public void setText (String text)
       {
       }

    }