// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;
import java.io.Serializable;

/**
 * A {@link Comparator} which combines two or more comparators. This is useful
 * when you want to sort with primary and secondary (or tertiary keys);
 * simply compose two or more keys with one or more <code>CompositeComparator</code>s.
 *
 * @author <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @see com.sas.util.ReverseComparator
 * @see com.sas.util.Comparator
 */
public class CompositeComparator
	implements Comparator,
               Serializable
{
  static final long serialVersionUID = -5673390108715472745L;
  // These used to be com.sas.util.Comparator but that iface
  // now implements java.util.Comparator
  java.util.Comparator primaryComparator, secondaryComparator;

  /**
   * Create a <code>CompositeComparator</code> which first compares items
   * using a <var>primaryComparator</var> (compare by a primary key),
   * then uses the <var>secondaryComparator</var> if the first returns
   * {@link Comparator#EQUALS}.
   * @param primaryComparator the primary comparison operation
   * @param secondaryComparator the primary comparison operation
   * The <var>secondaryComparator</var> is not used unless the <var>primaryComparator</var>
   * returns {@link Comparator#EQUALS}.
   * @since 2.0
   */
  public CompositeComparator(java.util.Comparator primaryComparator, java.util.Comparator secondaryComparator)
  {
    this.primaryComparator = safeComparator(primaryComparator);
    this.secondaryComparator = safeComparator(secondaryComparator);
  }

  /**
   * Create a <code>CompositeComparator</code> from two {@link com.sas.util.Comparator}s
   * Since {@link com.sas.util.Comparator} implements
   * {@link java.util.Comparator}, so the {@link #CompositeComparator(java.util.Comparator,java.util.Comparator)}
   * constructor should be used.
   * using a <var>primaryComparator</var> (compare by a primary key),
   * then uses the <var>secondaryComparator</var> if the first returns
   * {@link Comparator#EQUALS}.
   * @param primaryComparator the primary comparison operation
   * @param secondaryComparator the primary comparison operation
   * The <var>secondaryComparator</var> is not used unless the <var>primaryComparator</var>
   * returns {@link Comparator#EQUALS}.
   */
  public CompositeComparator(com.sas.util.Comparator primaryComparator, com.sas.util.Comparator secondaryComparator)
  {
    this((java.util.Comparator) primaryComparator, (java.util.Comparator) secondaryComparator);
  }

  /**
   * Create a <code>CompositeComparator</code> from an array of comparators.
   * This creates a chain of <code>CompositeComparator</code>
   * with the behavior like that outlined below:
   * <pre>
   * for (int i = 0; i &lt; comparators.length; i++) {
   *   int result = comparators[i].comparare(a,b);
   *   if (result != {@link Comparator#EQUALS})
   *      return result;
   *   }
   * // all comparators returned EQUALS
   * return Comparator.EQUALS;
   * </pre>
   * @param comparators an array of comparators.
   * The array may not be null and may not contain null values. The array length
   * must be at least 1.
   * @since 2.0
   */
  public CompositeComparator(java.util.Comparator comparators[])
  {
    this(comparators[0], comparator(comparators, 1, comparators.length - 1));
  }

  /**
   * Return a safe, non-null, comparator.
   * @param comparator a Comparator
   * @return <var>comparator</var> if <var>comparator</var>  is not null,
   *         else {@link ConstantComparator#COMPARE_EQUALS}.
   * @since 2.0
   */
  public static java.util.Comparator safeComparator(java.util.Comparator comparator)
  {
    return (  comparator == null
              ? ConstantComparator.COMPARE_EQUALS
              : comparator);
  }

  /**
   * Recursive helper for the <code>Comparator[]</code> constructor.
   * This creates a new CompositeConstructor(comparators[index], new CompositeConstructor(comparators, index+1, last)
   * @param comparators an array of comparators
   * @param index where in the array to select the next comparator.
   * @param last the index of the last element in <var>comparators</var>.
   *        if <code><var>index</var> == <var>last</var> - 1</code>, the
   *        secondaryComparator is set to null, else
   *        if <code><var>index</var> == <var>last</var></code>, the
   *        secondaryComparator is set to <code>comparators[index+1]</code>, else
   *        secondaryComparator is set to <code>new CompositeComparator(<var>comparators</var>, <var>index</var> +1, <var>last</var>);</code>
   * @since 2.0
   */
  protected static java.util.Comparator comparator(java.util.Comparator comparators[], int index, int last)
  {
    if (index > last)
      return null;
    else if (index == last)
      return comparators[last];
    else
      return new CompositeComparator(comparators[index],
                                     comparator(comparators, index+1, last));
  }

  /**
   * Compare two items using the primary and secondary comparators.
   * First, compare by the primary comparator. If that returns
   * either {@link Comparator#LESS_THAN}
   * or {@link Comparator#GREATER_THAN},
   * that is the result.
   * If it returns
   * {@link Comparator#EQUALS},
   * call the secondary comparator's {@link Comparator#compare(Object,Object)}
   * method and return that result.
   */
  public int compare(Object first, Object second)
  {
    int result = primaryComparator.compare(first, second);
    if (result == 0 && secondaryComparator != null)
      result = secondaryComparator.compare(first, second);
    return result;
  }

  public String toString()
  {
    return "(" + primaryComparator + ", " + secondaryComparator + ")";
  }

  public boolean equals(CompositeComparator other)
  {
    return (   other != null
               && Util.equal(primaryComparator, other.primaryComparator)
               && Util.equal(secondaryComparator, other.secondaryComparator));
  }

  public boolean equals(Object object)
  {
    return (object instanceof CompositeComparator
            && equals((CompositeComparator) object));
  }

}
