// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import com.sas.util.PredicateInterface;

/**
 * A predicate which always returns a constant value.
 * Two versions are available. <code>{@link #TRUE}</code> always returns <code>true</code> from
 * it's predicate; <code>{@link #FALSE}</code> always returns <code>false</code>.
 * @author <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public class ConstantPredicate
       implements PredicateInterface,
                  java.io.Serializable
{
  /*
   * serialVersionUID
   */
  static final long serialVersionUID = -2913500839241249572L;

  /** A predicate which always returns <code>true</code>. */
  public static final PredicateInterface TRUE = new ConstantPredicate(true);

  /** A predicate which always returns <code>false</code>. */
  public static final PredicateInterface FALSE = new ConstantPredicate(false);

  private boolean result;

  /**
    * Default constructor, equivalent to <code>ConstantPredicate(false)</code>.
    * <strong>Note:</strong>
    * You probably don't want to call this;
    * you are better off using <code>{@link #TRUE ConstantPredicate.TRUE}</code>
    * or <code>{@link #FALSE ConstantPredicate.FALSE}</code>, such as
    * <pre>
    * PredicateInterface pred = someCondition ? ConstantPredicate.TRUE : ConstantPredicate.FALSE
    * </pre>
    */
  public ConstantPredicate()
    {
      this(false);
    }

  /**
    * Value constructor.
    * <strong>Note:</strong>
    * You probably don't want to call this;
    * you are better off using <code>{@link #TRUE ConstantPredicate.TRUE}</code>
    * or <code>{@link #FALSE ConstantPredicate.FALSE}</code>, such as
    * <pre>
    * PredicateInterface pred = someCondition ? ConstantPredicate.TRUE : ConstantPredicate.FALSE
    * </pre>
    */
  public ConstantPredicate(boolean result)
    {
      this.result = result;
    }


  /**
   * Return the boolean value that this predicate always returns.
   * @return the boolean value that this predicate always returns.
   */
  public boolean getPredicateResult()
    {
      return result;
    }

  /**
   * Return the predetermined predicate result with which the
   * <code>ConstantPredicate</code> was constructed
   * @param object a predicate argument; it is always ignored.
   * @return the predetermined predicate result with which the
   * <code>ConstantPredicate</code> was constructed.
   * @see #ConstantPredicate(boolean)
   */
  public boolean predicate(Object object)
    {
      return result;
    }
}
