// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;
import java.io.PrintStream;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import java.util.MissingResourceException;
import com.sas.beans.PropertyEditor;
/**
 * The base enumeration class.
 * Concrete enumeration classes will extend this base class.
 * Some of the static methods provide the default implementation
 * of the enumerations.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @support sasdjb
 */

public abstract class Enum {

  /**
   * The integer value associated with the name.
   * The value is not really necessary, but many enumerations
   * are convenient if there is a numeric value associated
   * with it. Values do not have to be consecutive or even
   * unique values, however, there is no public method
   * for searching for duplicate values.
   */
  private int _value;
  /**
   * The next enum in the set, or null if this is the last.
   */
  protected Enum next;

  /**
   * An enum's name: an identifier, and it's print name
   * @see #toString
   */
  private String _name;

  /**
   * Convert an <code>Enum</code> to a string, for printing and other purposes.
   */
  public final String toString()
  {
    return toString(false);
  }

  /**
   * Get the next <code>Enum</code> in the sequence.
   * @return the next Enum, or null if this is the last in the sequence.
   */
  public abstract Enum getNextEnum();

  /**
   * Convert an <code>Enum</code> to a string, for printing and other purposes,
   * optionally with the fully qualified package name.
   * @param withPackageName if <code>true</code>, include the package
   *                        name of the enumeration's class.
   */
  public final String toString(boolean withPackageName)
  {
    if (withPackageName)
      return this.getClass().getName() + "." + _name;
    else
      return _name;
  }

  /**
   * Return the index, or ordinal position, of an enum. The
   * index is it's position relative to the others.
   * This is an <code>O(N)</code> sequential search. It could be made
   * <code>O(1)</code> at the expense of adding an <code>int index</code>
   * instance variable to the <code>Enum</code> class, but that would
   * be pretty space intensive and we anticipate that the
   * <code>getIndex()</code> method to be used infrequently, and N is
   * usually pretty small anyway. This static method provides
   * a common implementation for the instance method <code>getIndex()</code>
   * @param Return the index for this Enum.
   * @return the ordinal value of the enum.
   * @see #getIndex()
   */
  protected final int getIndex(Enum target, Enum first)
  {
    int i = 0;
    for (Enum enum = first; enum != null; enum = enum.getNextEnum(), i++)
      if (enum == target)
        return i;
    return -1;
  }

  /**
   * Return the index, or ordinal position, of an enum. The
   * index is it's position relative to the others.
   * @return the enum's index
   * @see #getIndex(Enum target, Enum first)
   */
  public abstract int getIndex();

  /**
   * Look up the <code>Enum</code> corresponding to a <code>int</code> value
   * @param value the value to search for.
   * @return the <code>Enum</code> which contains the specified
   *         value, or <code>null</code> if none match. Only the first <code>Enum</code>
   *         containing the value (as per ordinal ordering) is
   *         returned; there is no way to retrieve subsequent
   *         <code>Enum</code>s with the same value.
   */
  protected final static Enum getEnum(int value, Enum first)
  {
    for (Enum enum = first; enum != null; enum = enum.getNextEnum())
      {
        if (enum._value == value)
          return enum;
      }
    return null;
  }

  /**
   * Look up the <code>Enum</code> corresponding to a name.
   * @param name the name to search for.
   * @return the <code>Enum</code> which has the specified name, or <code>null</code>
   */
  public final static Enum get(String name, Enum first)
  {
    for (Enum enum = first; enum != null; enum = enum.getNextEnum())
      {
        if (enum._name.equals(name))
          return enum;
      }
    return null;
  }

  /**
   * Return up the <code>Enum</code> corresponding to an index
   * @param name the name to search for.
   * @param first the enum in the class sequence
   * @return the <code>Enum</code> which has the specified index, or <code>null</code>
   */
  public final static Enum get(int index, Enum first)
  {
    Enum enum = first;
    for (;((index > 0) && (enum != null)); enum = enum.getNextEnum(), index--)
      ; // null for loop body
    return enum;
  }

  /**
   * Look up the <code>Enum</code> corresponding to a name.
   * @param name the name to search for.
   * @return the Enum which has the specified name, or null
   * @deprecated use get(String name, first)
   */
  public final static Enum getAt(String name, Enum first)
  {
    for (Enum enum = first; enum != null; enum = enum.getNextEnum())
      {
        if (enum._name.equals(name))
          return enum;
      }
    return null;
  }

  /**
   * Return up the <code>Enum</code> corresponding to an index.
   * @param name the name to search for.
   * @param first the enum in the class sequence.
   * @return the <code>Enum</code> which has the specified index, or null
   * @deprecated use <code>get(int, com.sas.util.Enum)</code>
   */
  public final static Enum getAt(int index, Enum first)
  {
    Enum enum = first;
    for (;((index > 0) && (enum != null)); enum = enum.getNextEnum(), index--)
      ; // null for loop body
    return enum;
  }

  /**
   * Return the numeric value associated with an enumeration.
   * @return the the numeric value associated with an enumeration.
   */
  public int getValue()
  {
    return _value;
  }


  /**
   * A protected Constructor. Note there is no public constructor.
   * Subclasses should construct instances in static methods.
   * The order in which constructors are called determines the
   * sequence; the first constructed Enum should be returned by the
   * getFirst() method; it is the subclass' responsibility to do thid.
   * @param name the enum's name -- an identifier, returned by <var>toString()</var>
   *        and matched by the <var>enum(String name)</var> method
   * @param value the value associated with the Enum; returned by
   *        the <var>value()</var> method and matched by the getEnum(int
   *        value) method defined by the subclass.
   */
  protected Enum(String name, int value)
  {
    this._name     = name;
    this._value    = value;
  }

  /**
   * The <code>Enum.Editor</code> class is a base class for property editors for <code>Enum</code> subclasses.
   * It provides base functionality common to all *EnumEditor classes.
   */
  public static abstract class Editor
         extends PropertyEditor
  {

    /**
     * Should <code>getAsText()</code> and <code>getTags()</code> include the enum name?
     */
    protected boolean displayTags;
    /**
     * Should <code>getAsText()</code> and <code>getTags()</code> include the enum description from a resource bundle?
     */
    protected boolean displayDescriptions;
    /**
     * Should the editor <code>getTags()</code> include "null" ?
     */
    protected boolean listNull;

    /**
     * Should <code>getAsText()</code> and <code>getTags()</code> in the next created <code>Enum.Editor</code> include the enum name?
     */
    protected static boolean nextDisplayTags = true;
    /**
     * Should <code>getAsText()</code> and <code>getTags()</code> in the next created <code>Enum.Editor</code> include the enum description from a resource bundle?
     */
    protected static boolean nextDisplayDescriptions = false;
    /**
     * Should the editor <code>getTags()</code> in the next created <code>Enum.Editor</code> include <code>"null"</code> ?
     */
    protected static boolean nextListNull = true;

    /**
     * Default constructor. Make a property editor which
     * displays the tag and description.
     */
    public Editor()
      {
        setDisplayOptions(nextDisplayTags, nextDisplayDescriptions, nextListNull);
        // reset default so next editor is created with the prewired default
        // this is so editors created in property sheets get just tags;
        // editors created for
        setNextEnumEditorDisplayOptions(true, false, true);
      }

    /**
     * Return a resource bundle from which text descriptions
     * of each enum may be found.
     * @return a ResourceBundle
     */
    protected abstract ResourceBundle resources();

    /**
     * Return the resource property prefix by which text descriptions
     * are found in the resource bundle. This is normally the
     * base name of the class, with a <code>"."</code> appended, such as
     * <code>"MissingValues."</code> for the <code>class com.sas.MissingValues</code>
     */
    protected abstract String classKey();

    /**
     * Return the first <code>Enum</code> value in the class, from
     * which all the <code>Enum</code> values can be found.
     */
    protected abstract Enum first();

    /**
     * Set the <code>Enum.Editor</code> display option for <code>getAsText()</code> and <code>getTags()</code> to include
     * the enum element tag (that is it's name) and description.
     * @param displayTags if <code>true</code>, the enum text display includes the tag
     * If <code>false</code>, the description display is automatically turned on.
     * @param displayDescriptions if <code>true</code>, the enum text display includes the description.
     * If <code>false</code>, the tag display is automatically turned on.
     * @param listNull if <code>true</code>, the <code>Enum.Editor</code> will include <code>null</code>
     *             in <code>getTags()</code>
     * @see #resources()
     */
    public void setDisplayOptions(boolean displayTags, boolean displayDescriptions, boolean listNull)
      {
        // System.out.println("setDisplayOptions(" + displayTags + ", " + displayDescriptions + ")");
        this.displayTags = displayTags;
        this.displayDescriptions = displayDescriptions;
        if ( ! (displayTags | displayDescriptions) )
          displayTags = true;
        this.listNull = listNull;
      }

    /**
     * Set the <code>Enum.Editor</code> display option for the next created <code>Enum.Editor</code>.
     * @param displayTags if <code>true</code>, the next <code>Enum.Editor</code>
     * instance will include the tag
     * name in enum displays.
     * If <code>false</code>, the description display is automatically turned on.
     * @param displayDescriptions if true, the next <code>Enum.Editor</code> will include the description.
     * If <code>false</code>, the tag display is automatically turned on.
     * @param listNull if <code>true</code>, the next <code>Enum.Editor</code> will include <code>null</code> in <code>getTags()</code>
     * @see #resources()
     */
    public static void setNextEnumEditorDisplayOptions(boolean displayTags, boolean displayDescriptions, boolean listNull)
      {
        // System.out.println("setNextEnumEditorDisplayOptions(" + displayTags + ", " + displayDescriptions);
        nextDisplayTags = displayTags;
        nextDisplayDescriptions = displayDescriptions;
        if ( ! (displayTags || displayDescriptions))
          nextDisplayTags = true;
        nextListNull = listNull;
      }

    /**
     * Set the <code>Enum.Editor</code> display option for <code>getAsText()</code> and <code>getTags()</code> to include
     * the enum description from the package resource file (i.e. it's name)
     * @param displayDescriptions if <code>true</code>, the enum text display includes the description.
     * If <code>false</code>, the tag display is automatically turned on.
     * @see #resources()
     */
    void displayDescriptions(boolean displayDescriptions)
      {
        this.displayDescriptions = displayDescriptions;
        if ( ! displayDescriptions )
          this.displayTags = true;
        // System.out.println("displayDescriptions(displayDescriptions: " + displayDescriptions + ") displayTags: " + displayTags);
      }

    /**
     * Set the value of a property editor, based on text.
     * This method implements the <code>void setAsText(String text)</code>
     * for an <code>Enum</code> subclass' property editor.
     * @param text the string value the user types into the text field.
     *        Try to set the value of the editor from this text.
     *        We scan for the first word in the text and look up
     *        that word in the <code>Enum</code> list. The comparison is
     *        done ignoring case.
     * @param first the first <code>Enum</code> in the <code>Enum</code> subclass. This
     *        provides a list to the other enums.
     * @param editor the <code>PropertyEditor</code>
     * @exception java.lang.IllegalArgumentException if the text does
     *            not match any of the Enum values.
     * @see java.beans.PropertyEditor
     * @see com.sas.beans.PropertyEditor
     */
    public void setAsText(String text) throws IllegalArgumentException
      {
        if (   (text == null)
               || ((text = text.trim()).length() == 0)
               || text.equals("null"))
          {
            if (Editor.this.getValue() != null)
              {
                setValue(null);
                firePropertyChange();
              }
          }

        // If displayDescriptions is on, first scan enums to match by descriptions
        ResourceBundle bundle = resources();
        String desc = "";
        String resourcePrefix = classKey();
        for (Enum enum = first(); enum != null; enum = enum.getNextEnum())
          {
            desc = getDescription(enum, resourcePrefix, bundle);
            // System.out.println(enum + " desc: " + desc);
            // System.out.println(enum + "t ag: " + getTag(enum, resourcePrefix, bundle, displayTags, displayDescriptions));
            if (    null != desc
               && (   text.equalsIgnoreCase(desc)
                   || text.equalsIgnoreCase(getTag(enum, resourcePrefix, bundle, displayTags, displayDescriptions))))
            {
              if (Editor.this.getValue() != enum)
                {
                  setValue(enum);
                  firePropertyChange();
                }
              return;
            }
          }

        // Did not match on description or tag. get first word of the text
        // and try to match on that.
        String word = new StringTokenizer(text).nextToken();
        // System.out.println("Did not match with exact text. Try first word: " + word);
        for (Enum enum = first(); enum != null; enum = enum.getNextEnum())
          if (word.equalsIgnoreCase(enum.toString()))
            {
              if (Editor.this.getValue() != enum)
                {
                  setValue(enum);
                  firePropertyChange();
                }
              return;
            }
        // System.out.println("Did not find match for \"" + text + "\" in:");
        // for (Enum enum = first(); enum != null; enum = enum.getNextEnum())
        // System.out.println("\t\"" + getDescription(enum, resourcePrefix, bundle)  + "\"");
        throw new IllegalArgumentException();
      }

    /**
     * Make a list of property editor tags.  Given an enum with the name
     * <code>"XYZ"</code>, if the resource bundle contains the description resource for the key
     * <code>(resourcePrefix + "XYZ")</code>, then the corresponding tag will be
     * <code>"XYZ" + " : " + description</code> (if <var>displayTags</var> and <var>displayDescriptions</var> are <code>true</code>),
     * otherwise it will be simply "XYZ" if displayTags is true and displayDescriptions is false,
     * else it will be just the description if <var>displayTags</var> is <code>false</code> and <var>displayDescriptions</var> is <code>true</code>.
     * (one of these two flags is always true). This
     * allows the tags to be locale specific, and yet still use the <code>Enum</code> name.
     * @param first the first <code>Enum</code> in the <code>Enum</code> subclass. This
     *        provides a list to the other enums.
     * @return an array of tags. <code>tags[0]</code> is always <code>"null"</code> to allow selection/display
     * of a null value in the <code>PropertyEditor</code>. Remaining values are the tag text
     * for each enum starting with first.
     */
    public String[] getTags()
      {
        Enum enum = first();
        int count = 0;
        for (; enum != null; enum = enum.getNextEnum())
          count++;
        if (listNull)
          count++;
        String tags[] = new String[count];

        int i = 0;
        if (listNull)
          tags[i++] = "null";
        String classKey = classKey();
        ResourceBundle resources = resources();
        for (enum = first(); enum != null; enum = enum.getNextEnum())
          tags[i++] = getTag(enum, classKey, resources, displayTags, displayDescriptions);
        return tags;
      }

    /**
     * Make a property editor tag for an Enum, for use in a <code>PropertyEditor</code>
     * for an <code>Enum</code> subclass.
     * @param enum the Enum object we want a tag for.
     * @param resourcePrefix a string prefix for resource names in the resource bundle.
     * @param bundle a resource bundle.
     * @param displayTags true if the string should contain the enum name
     * @param displayDescriptions <code>true</code> if the string should contain the description
     * @return the tag for enum: its name and optionally the resource string
     * for the enum, taken from the resource bundle. The resource string has a
     * name constructed as <code>(<var>resourcePrefix</var> + tag + ".txt")</code>.
     */
    public static String getTag(Enum enum, String resourcePrefix, ResourceBundle bundle, boolean displayTags, boolean displayDescriptions)
      {
        String tag = "";
        // System.out.print("getTag(" + enum + ", \"" + resourcePrefix + "\", " + enum + ", displayTags=" +
        //                 displayTags + ", displayDescriptions=" + displayDescriptions + ")");
        if (displayTags)
          tag = enum.toString();
        if (displayDescriptions)
          {
            String desc = getDescription(enum, resourcePrefix, bundle);
            if (desc != null)
              tag += (displayTags ? " : "/*I18nOK:EMS*/ : "") + desc;
          }
        if ("".equals(tag)) // if desc fails when displayTags is false, revert to tag anyway
          tag = enum.toString();
        // System.out.println(" => " + tag);
        return tag;
      }

    /**
     * Read the enum description from the resource bundle.
     * @param resourcePrefix a prefix for the resource name (usually the
     * <code>Enum</code>'s class name
     * with a ".", such as <code>"Styles."</code>)
     * @param ResourceBundle the bundle from which the description is obtained.
     * @return the description string, or <code>null</code> of no such string exists in the bundle.
     */
    static String getDescription(Enum enum, String resourcePrefix, ResourceBundle bundle)
      {
        if (bundle != null)
          try
            {
              return bundle.getString(resourcePrefix + enum.toString() + ".txt");
            }
        catch (MissingResourceException e)
          { }
        return null;
      }

    /**
     * Get the text value for an <code>Enum</code>, for use in a <code>PropertyEditor</code>
     * @param enum an <code>Enum</code> object
     * @return the text value associated with the <code>Enum</code>.
     */
    public String getAsText()
      {
        Enum enum = (Enum) value;
        if ( enum == null )
          return null;
        else
          return getTag(enum, classKey(),  resources(), displayTags, displayDescriptions);
      }

    /**
     * Implement the <code>getJavaInitializationString</code> for a <code>PropertyEditor</code>.
     * @param enum the <var>Enum</var> value for intialization
     * @return <var>"null"</var> (if enum is null) or
     * enum.toString(true)
     * @see #toString(boolean)
     */
    public String getJavaInitializationString()
      {
        Enum enum = (Enum) value;
        if (enum == null)
          return "null";
        else
          return enum.toString(true);
      }
  }

  /**
  * Return an array of enumerations in an <code>Enum</code> subclass, starting at the
  * first enum.
  * @return an array of enumerations in an <code>Enum</code> subclass, starting at the
  * first enum.
  */
  public static Enum[] getEnums(Enum first)
    {
      Enum e = first;
      int count = 0;
      for (; e != null; e = e.getNextEnum())
        count++;
      Enum enums[] = new Enum[count];
      for (count = 0; e != null; e = e.getNextEnum())
        enums[count++] = e;
      return enums;
    }

}
