// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.util;

/**
 * An interface for suppressing or allowing generation of events.
 * This is useful if you wish to perform multiple changes to an object
 * and only want a single event (normally PropertyChangeEvent) sent when the group of changes
 * have finished, rather than having events sent on each change.
 * <p>
 * Note that you <em><strong>must</strong></em> use a try/finally block to allow
 * events to proceed.
 * For example:
 *
 * <pre>
 *   synchronized (object) {
 *      try {
 *          object.disableEvents();
 *          // make multiple changes to the object here
 *          object.multipleChanges();
 *        }
 *      finally {
 *          object.enableEvents();
 *        }
 *      }
 * </pre>
 * <p>
 * This ensures the <code>enableEvents()</code> method is called, even if
 * an exception is thrown in the body. If you don't use a try/finally block,
 * the <code>enableEvents()</code> method may not be called and the object will no
 * longer send events as expected.
 * <p>
 * It is the object's responsibility to allow nesting of
 * <code>disableEvents()/enableEvents()</code> calls. For example, if one of the methods called
 * in the changes section
 * above also does a <code>disableEvents()/enableEvents()</code>, the collection will only start
 * firing events when all <code>enableEvents()</code> calls have been made.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

public interface EventGateInterface
{
  /**
    * Ask an object to suppress delivery of events
    * on subsequent changes to the object.
    * <p>
    * This call should appear in a <code>try { }</code> block, with the matching <code>enableEvents()</code>
    * method in the <code>finally {}</code> block.
    */
  public void disableEvents();

  /**
    * Undo a <code>disableEvents()</code> call. This may or may not cause the
    * object to start sending events again, depending on the nesting
    * of <code>disableEvents()/enableEvents()</code> calls. If
    * there are no other pending disable/enable gates on the
    * object, the object will send a single event (normally a {@link
    * java.beans.PropertyChangeEvent PropertyChangeEvent})
    * and then subsequent changes to the object will also
    * fire events normally.
    * <p>
    * This call should always appear in a <code>finally {}</code> block;
    * the <code>disableEvents()</code> calls should be in the <code>try {}</code> block.
    */
  public void enableEvents();
}
