/* $Id: ExceptionUtil.java,v 1.1.2.2.4.1 2004/12/03 21:14:49 sasbcl Exp $ */
package com.sas.util;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.util.Locale;

import com.sas.util.ChainedException;

/** A utility class designed to help in logging exceptions with error codes */
public class ExceptionUtil
{
	private static String lineSeparator = System.getProperty("line.separator");

	/** Packages that should be excluded from stack traces */
	public static final String[] DEFAULT_EXCLUDED_PACKAGES =
	{
		"org.apache.",
		"javax.servlet.",
		"junit."
	};
	
	/**
	 * When generating a stack trace, log the top exception only
	 */
	public static final int GENERATESTACKTRACE_TOP = 0;
	
	/**
	 * When generating a stack trace, log the top exception and the 
	 * authoritative one, or the root one if no authoritative exception is found,
	 * providing the authoritative (or root) exception is different from the top exception
	 */
	public static final int GENERATESTACKTRACE_AUTHORITATIVE = 1;
	
	/**
	 * When generating a stack trace, log the top exception and the 
	 * root exception, providing the root exception is different from the top exception
	 */
	public static final int GENERATESTACKTRACE_ROOT = 2;
	
	/**
	 * When generating a stack trace, log the top exception, the
	 * authoritative exception, and the root exception, providing 
	 * none of them are the same.
	 */
	public static final int GENERATESTACKTRACE_ALL = 3;

	/** no error code found */
	//public static final int NO_CODE = -1;

	/**
	 * Generate a (possibly trimmed) stack trace for a Throwable
	 *
	 * @param t the Throwable
	 * @param buffer a buffer to write to. If null, a StringBuffer
	 *               will be created and written to.
	 * @param authoritative an integer telling what combination of exceptions to show.
	 * 			<ul><li>GENERATESTACKTRACE_TOP = 0 = Top exception only</li>
	 * 				<li>GENERATESTACKTRACE_AUTHORITATIVE = 1 = Top exception and authoritative exception if not same as top exception, 
	 * 						or root exception if no authoritative one found</li> 
	 * 				<li>GENERATESTACKTRACE_ROOT = 2 = Top exception and root exception if not same as top exception</li>
	 * 				<li>GENERATESTACKTRACE_ALL = 3 = Top exception, authoritative exception if one exists and if it is not the same as top 
	 * 						exception, and root exception if not same as other exceptions</li>
	 * 			</ul>
	 *
	 * @return the StringBuffer that the stack trace was written to.
	 */
	public static StringBuffer generateStackTrace( StringBuffer buffer, Throwable t, String[] excludedPackages, boolean trim)
	{
		StringBuffer buf = (buffer == null) ? new StringBuffer() : buffer;
		Throwable rootCause = locateAuthoritativeOrLastException(t);
		if ((rootCause != null) && (rootCause != t))
		{
			buf.append("Root Cause:   ");
			generateTrimmedStackTrace( buf, rootCause, excludedPackages, trim);
			buf.append("Which Caused: ");
		}
		generateTrimmedStackTrace( buf, t, excludedPackages, trim);
		return buf;
	}
	
	/**
	 * Generate a (possibly trimmed) custom stack trace for a Throwable
	 *
	 * @param t the Throwable
	 * @param buffer a buffer to write to. If null, a StringBuffer
	 *               will be created and written to.
	 * @param stackTraceLevel an integer telling what combination of exceptions to show.
	 * 			<ul><li>GENERATESTACKTRACE_TOP = 0 = Top exception only</li>
	 * 				<li>GENERATESTACKTRACE_AUTHORITATIVE = 1 = Top exception and authoritative exception if not same as top exception, 
	 * 						or root exception if no authoritative one found</li> 
	 * 				<li>GENERATESTACKTRACE_ROOT = 2 = Top exception and root exception if not same as top exception</li>
	 * 				<li>GENERATESTACKTRACE_ALL = 3 = Top exception, authoritative exception if one exists and if it is not the same as top 
	 * 						exception, and root exception if not same as other exceptions</li>
	 * 			</ul>
	 *
	 * @return the StringBuffer that the stack trace was written to.
	 */
	public static StringBuffer generateCustomStackTrace( StringBuffer buffer, Throwable t, String[] excludedPackages, boolean trim, int stackTraceLevel)
	{
		StringBuffer buf = (buffer == null) ? new StringBuffer() : buffer;
		Throwable rootCause = null;
		Throwable authoritativeCause = null;
		if (stackTraceLevel == GENERATESTACKTRACE_AUTHORITATIVE || stackTraceLevel == GENERATESTACKTRACE_ALL)
		{
			authoritativeCause = locateAuthoritativeOrLastException(t);
		}
		if (stackTraceLevel == GENERATESTACKTRACE_ROOT || stackTraceLevel == GENERATESTACKTRACE_ALL)
		{
			rootCause = ChainedException.getRootException(t);
		}
		if ((rootCause != null) && (rootCause != t))
		{
			buf.append("Root Cause:   ");
			generateTrimmedStackTrace( buf, rootCause, excludedPackages, trim);
			buf.append("Which Caused: ");
		}
		if ((authoritativeCause != null) && (authoritativeCause != t) && (authoritativeCause != rootCause))
		{
			generateTrimmedStackTrace( buf, authoritativeCause, excludedPackages, trim);
			buf.append("Which Caused: ");
		}
		generateTrimmedStackTrace( buf, t, excludedPackages, trim);
		return buf;
	}

	/**
	 * Generate a (possibly trimmed) stack trace for a Throwable
	 *
	 * @param t the Throwable
	 * @param buffer a buffer to write to.
	 */
	public static void generateTrimmedStackTrace(StringBuffer buffer, Throwable t, String[] excludedPackages, boolean trim )
	{
		buffer.append(t).append(lineSeparator);
		StackTraceElement[] elements = t.getStackTrace();
		boolean processElements = true;
		for (int j = 0; processElements && (j < elements.length); j++)
		{
			buffer.append("\tat ").append(elements[ j ]).append(lineSeparator);
			if (trim)
			{
				String classname = elements[ j ].getClassName();
				for (int k = 0; k < excludedPackages.length; k++)
				{
					if (classname.startsWith(excludedPackages[ k ]))
					{
						processElements = false;
						int unprocessedLineCount = elements.length - j - 1;
						if (unprocessedLineCount > 0)
						{
							buffer.append("\t[ ").append(unprocessedLineCount);
							buffer.append(" line");
							if (unprocessedLineCount != 1)
							{
								buffer.append("s");
							}
							buffer.append(" suppressed ]").append(lineSeparator);
						}
						break;
					}
				}
			}
		}
	} // generateTrimmedStackTrace

	/**
	 * Get the message from a given exception and resource bundle
	 *
	 * @param e the Exception containing the error code
	 * @param rb the ResourceBundle in which to look up the error code
	 *
	 * @return the String that corresponds to the error code in the resource bundle
	 */
	public static String getMessage(Exception e, ResourceBundle rbOverride, Locale locale)
	{
		String code;
		String message;
		if (e instanceof ExceptionCodeInterface )
		{
			code = ((ExceptionCodeInterface)e).getGenericExceptionCode();
			if (code != null && code.length()>0 && rbOverride != null)
				message = getResourceString(rbOverride, code);
			//return rbOverride.getString(code);
			else
			{
				message = ((ExceptionCodeInterface)e).getLocalizedMessage(locale);
			}
			return message;
		    //return rbOverride.getString(code);
		}
		else
			return e.getLocalizedMessage();
	}

	/**
	 * Get the authoritative message from a given exception and resource bundle
	 * If the exception passed in is not authoritative, go up the chain of exceptions
	 * and find the first one that is authoritative.  If no authoritative exception is found,
	 * return null.
	 *
	 * @param e the Exception containing the error code
	 * @param rbOverride the ResourceBundle in which to look up the error code
	 *
	 * @return the String that corresponds to the authoritative error code in the resource bundle
	 */
	public static String getAuthoritativeMessage(Exception e, ResourceBundle rbOverride, Locale locale)
	{
		String code = getAuthoritativeCode(e);
		String message = null;

		if (code != null && code.length()>0 && rbOverride != null)
			message = getResourceString(rbOverride, code);
		//return rbOverride.getString(code);
		else
		{
			ExceptionCodeInterface ec = locateAuthoritativeException(e);
			if (ec != null)
				message = ec.getLocalizedMessage(locale);
		}
		//return getResourceString(rbOverride, code);
		//else
			//return null;
		
		return message;
	}
	
	/**
	 * Get the authoritative message from a given exception and resource bundle
	 * If the exception passed in is not authoritative, go up the chain of exceptions
	 * and find the first one that is authoritative.  If no authoritative exception is found,
	 * return the root exception (the one at the end of the chain).
	 *
	 * @param e the Exception containing the error code
	 * @param rbOverride the ResourceBundle in which to look up the error code
	 *
	 * @return the String that corresponds to the authoritative error code in the resource bundle
	 */
	public static String getAuthoritativeOrLastMessage(Exception e, ResourceBundle rbOverride, Locale locale)
	{
		String code = getAuthoritativeCode(e);
		String message = null;

		if (code != null && code.length()>0 && rbOverride != null)
			message = getResourceString(rbOverride, code);
		//return rbOverride.getString(code);
		else
		{
			Throwable ex = locateAuthoritativeOrLastException(e);
			if (ex instanceof ExceptionCodeInterface)
			{
				message = ((ExceptionCodeInterface)ex).getLocalizedMessage(locale);
			}
			else
			{
				message = ex.getLocalizedMessage();
			}
			
		}
		//return getResourceString(rbOverride, code);
		//else
			//return null;
		
		return message;
	}
	

	/**
	 * Get the code from a given exception
	 *
	 * @param e the Exception containing the error code
	 *
	 * @return the error code or -1 if there is no error code
	 */
	public static String getCode(Exception e)
	{
		if (e instanceof ExceptionCodeInterface)
		{
			return ((ExceptionCodeInterface)e).getGenericExceptionCode();
		}
		else
			return "";
	}

	/**
	 * Get the authoritative code from a given exception
	 * If the exception passed in is not authoritative, go up the chain of exceptions
	 * and find the first one that is authoritative.  If no exception is authoritative, then
	 * return -1 for no code.
	 *
	 * @param e the Exception containing the error code
	 *
	 * @return the error code or -1 if there is no error code
	 */
	public static String getAuthoritativeCode(Exception e)
	{
		ExceptionCodeInterface ec = locateAuthoritativeException(e);
		if (ec != null)
			return ec.getGenericExceptionCode();
		else
			return "";
	}

	/**
	 * Scan through the possible chain of exceptions, locating the "first" which implements
	 * ExceptionCodeInterface, and that returns true for ExceptionCodeInterface.isAuthoritative.
	 * "First" would be the "topmost" exception, closer to this user rather than closer to the
	 * original exception.
	 *
	 * @return Exception as described above, will extend Exception, and will implement
	 * ExceptionCodeInterface. OFTEN null which means no "authoritative" exception
	 * could be located.
	 */
	public static ExceptionCodeInterface locateAuthoritativeException(Exception e)
	{
		if (e instanceof ExceptionCodeInterface && ((ExceptionCodeInterface)e).isAuthoritative())
			return (ExceptionCodeInterface)e;
		else
		{
			if (ChainedException.getTargetException(e) != null)
				return locateAuthoritativeException((Exception)ChainedException.getTargetException(e));
			else
				return null;
		}
	}
	
	/**
	 * Scan through the possible chain of exceptions, locating the "first" which implements
	 * ExceptionCodeInterface, and that returns true for ExceptionCodeInterface.isAuthoritative.
	 * "First" would be the "topmost" exception, closer to this user rather than closer to the
	 * original exception.  If no authoritative exception is found, return the last exception 
	 * in the chain.
	 *
	 * @return Exception as described above, will extend Exception, and will implement
	 * ExceptionCodeInterface. 
	 */
	public static Throwable locateAuthoritativeOrLastException(Throwable e)
	{
		if (e instanceof ExceptionCodeInterface && ((ExceptionCodeInterface)e).isAuthoritative())
			return e;
		else
		{
			if (ChainedException.getTargetException(e) != null)
				return locateAuthoritativeOrLastException(ChainedException.getTargetException(e));
			else
				return e;
		}
	}

	/**
	 * Scan through the possible chain of exceptions, returning the "lowermost" or "root" exception.
	 */
	public static Exception locateRootException(Exception e)
	{
		if (e != null && ChainedException.getTargetException(e) != null)
			return locateRootException((Exception)ChainedException.getTargetException(e));
		else
			return e;
	}
	
	public static String getResourceString(ResourceBundle rb, String key)
	{
        if (key == null)
            throw new NullPointerException();
        String rval = key;
        try
        {
            rval = rb.getString(key);
        }
        catch (MissingResourceException e)
        {
        }
        return rval;
	}

}
