package com.sas.util;

import java.io.Serializable;
import com.sas.MissingValues;
import com.sas.lang.StringDataInterface;
import com.sas.lang.DoubleDataInterface;
import com.sas.lang.FloatDataInterface;
import com.sas.lang.LongDataInterface;
import com.sas.lang.IntegerDataInterface;
import com.sas.lang.ShortDataInterface;
import com.sas.lang.CharacterDataInterface;
import com.sas.lang.ByteDataInterface;
import java.util.Date;
import java.util.Calendar;
import java.math.BigDecimal;

/**
 * <code>GenericComparator</code> is a {link Comparator} that can
 * compare {@link java.lang.String},
 * numeric ({@link java.lang.Number}), {@link java.util.Date}, {@link java.util.Calendar}, and {@link com.sas.MissingValues} values.
 * A <code>null</code> value
 * compares less than non null values. <code>MissingValues</code> compare less than
 * numeric or <code>String</code> values. <code>MissingValues</code> compare to each other according
 * to their normal ordering relation based on {@link
 * com.sas.MissingValues#getIndex() getIndex()}. String values are all
 * considered less than numeric values. Date and Calendar objects are converted to
 * long integer values, which represents the interval in milliseconds since the Epoch, January 1, 1970.
 * Numeric items use normal numeric comparison; strings use normal
 * string comparison.
 * <p>
 * <em>No conversions</em> are attempted between
 * types (such as comparing the string "13" to the number 13, or
 * comparing {@link java.lang.Double#NaN NaN} to {@link com.sas.MissingValues a missing value}.
 * <p>
 * Since GenericComparator is stateless,
 * this class also provides a default instance {@link GenericComparator#defaultInstance}
 * which you can use instead of creating a new instance each time.
 * <p>
 * If you wish to compare the numeric values of formatted String values, you should
 * use a {@link com.sas.util.TransformingComparator} constructed with a Format instance
 * and {@link GenericComparator#defaultInstance}.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 * @since 1.2
 * @see com.sas.collection.SortedCollection
 */

public class GenericComparator
	implements Comparator, 
			   Serializable,
			   java.util.Comparator
{
	static final long serialVersionUID = -6209322708319709038L;

	/**
	 * Compare two values.
	 * @return {@link Comparator#LESS_THAN} if <var>a</var> is less than <var>b</var>,
	 * {@link Comparator#GREATER_THAN} if <var>a</var> is greater than <var>b</var>,
	 * {@link Comparator#EQUALS}<var>a</var> equals <var>b</var>, and
	 * {@link Comparator#INCOMPARABLE}<var>a</var> cannot be compared to <var>b</var>
	 */
	public int compare(Object a, Object b)
	{
		if (a == b)
			return EQUALS;
		if (a == null)
			return LESS_THAN;
		if (b == null)
			return GREATER_THAN;
		if (a instanceof MissingValues && b instanceof MissingValues)
		{
			int missA = ((MissingValues) a).getIndex();
			int missB = ((MissingValues) b).getIndex();
			return ( (missA < missB)
				? LESS_THAN
				: ((missA > missB)
				? GREATER_THAN
				: EQUALS));
		}
		if (a instanceof MissingValues)
			return LESS_THAN;
		if (b instanceof MissingValues)
			return GREATER_THAN;
		if (a instanceof String && b instanceof Number)
			return LESS_THAN;
		if (a instanceof Number && b instanceof String)
			return GREATER_THAN;
		if (a instanceof String && b instanceof String)
			return BaseStringComparator.defaultInstance.compare(a, b);
		if (a instanceof Number && b instanceof Number)
		{
			if (isInteger(a) && isInteger(b))
				return compareIntegers((Number) a, (Number) b);
			else
				return compareFloat((Number) a, (Number) b);
		}
		Object compA = makeComparable(a);
		Object compB = makeComparable(b);
		if (compA != null && compB != null)
			return compare(compA, compB);
		return INCOMPARABLE;
	}

	/**
	 * Take a value <var>a</var> that is not a {@link Number}
	 * and coerce it to a value that can be compared numerically.
	 * The millisecond value of {@link java.util.Date} and {@link java.util.Calendar} objects is used
	 * if <var>a</var> is a Date or Calendar instance.
	 * If the object implements one of
	 * {@link com.sas.lang.DoubleDataInterface},
	 * {@link com.sas.lang.FloatDataInterface},
	 * {@link com.sas.lang.LongDataInterface},
	 * {@link com.sas.lang.IntegerDataInterface},
	 * {@link com.sas.lang.ShortDataInterface},
	 * {@link com.sas.lang.CharacterDataInterface}, or
	 * {@link com.sas.lang.ByteDataInterface},
	 * the numeric value is extracted and wrapped in the corresponding wrapper
	 * class from the <code>java.lang</code> package.
	 * @return an instance of a <code>java.lang</code> wrapper, if appropriate,
	 * or null if <var>a</var> does not implement one of the known
	 * data interfaces.
	 */
	protected static Object makeComparable(Object a)
	{
		if (a instanceof StringDataInterface)
			return ((StringDataInterface) a).getText();
		if (a instanceof DoubleDataInterface)
			return new Double(((DoubleDataInterface) a).getDoubleData());
		if (a instanceof FloatDataInterface)
			return new Float(((FloatDataInterface) a).getFloatData());
		if (a instanceof LongDataInterface)
			return new Long(((LongDataInterface) a).getLongData());
		if (a instanceof IntegerDataInterface)
			return new Integer(((IntegerDataInterface) a).getIntegerData());
		if (a instanceof ShortDataInterface)
			return new Short(((ShortDataInterface) a).getShortData());
		if (a instanceof CharacterDataInterface)
			return new Character(((CharacterDataInterface) a).getCharacterData());
		if (a instanceof ByteDataInterface)
			return new Byte(((ByteDataInterface) a).getByteData());
		if (a instanceof Date)
			return new Long(((Date) a).getTime());
		if (a instanceof Calendar)
			return new Long(((Calendar) a).getTime().getTime());
		return null;
	}

	/**
	 * Tests is an object is a non-floating point number.
	 * @return true if an object is a {@link java.lang.Number}
	 * but not a {@link java.lang.Float} or {@link java.lang.Double}
	 */
	public static boolean isInteger(Object a)
	{
		return  (    (a instanceof Number)
			&& ! (   (a instanceof Float)
			|| (a instanceof Double) || (a instanceof BigDecimal)));
	}

	/**
	 * Compare two non-floating point Numbers.
	 * @return {@link Comparator#LESS_THAN} if <var>a</var> &lt; <var>b</var>,
	 * {@link Comparator#EQUALS} if </code}<var>a</var> == <var>b</var>,
	 * {@link Comparator#GREATER_THAN} if <var>a</var> > <var>b</var>
	 */
	public static int compareIntegers(Number a, Number b)
	{
		long lA = a.longValue();
		long lB = b.longValue();
		return ( (lA < lB)
			? LESS_THAN
			: ((lA > lB)
			? GREATER_THAN
			: EQUALS));
	}

	/**
	 * Compare two floating point Numbers.
	 * @return {@link Comparator#LESS_THAN} if <var>a</var> &lt; <var>b</var>,
	 * {@link Comparator#EQUALS} if <var>a</var> == <var>b</var>,
	 * {@link Comparator#GREATER_THAN} if <var>a</var> > <var>b</var>
	 */
	public static int compareFloat(Number a, Number b)
	{
		double dA = a.doubleValue();
		double dB = b.doubleValue();
		return ( (dA < dB)
			? LESS_THAN
			: ((dA > dB)
			? GREATER_THAN
			: EQUALS));
	}

	/**
	 * A default instance of GenericComparator.
	 */
	public static final GenericComparator defaultInstance = new GenericComparator();

}
