// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


//package com.sas.collection;
package com.sas.util;

import java.util.List;

/**
 * Permutes the items in a Java collection according
 * to a index reordering. A permutation is an integer array
 * which shows where each item in the new ordering of items
 * should come from, based on the original order.
 * For example, a collection containing the string
 * values
 * <pre>
 * { "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L" }
 * </pre>
 * and permuted with the int array
 * <pre>
 * { 8, 9, 10, 11, 4, 5, 6, 7, 0, 1, 2, 3 }
 * </pre>
 * will result in the collection
 * <pre>
 * { "I", "J", "K", "L", "E", "F", "G", "H", "A", "B", "C", "D" }
 * </pre>
 * A permutation array must satisfy these constraints:
 * <ol>
 * <li>its length equals the collection size, determined
 *     by invoking collection's the count() method
 * <li>All values must be in the range [0, length-1]
 * <li>All values must be unique.
 * </ol>
 * @author Neal Smith
 * @version 1.0
 */

public class JavaCollectionPermuter
{

  private List collection;
  private int count;
  private boolean haveSize;
  private int permutation[];
  private List aList;

  ////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////

  /**
    * Default constructor.
    * Use <code>{@link #setCollection}</code> and <code>{@link #setPermutation}</code>
    * to initialize the <code>Permuter</code>.
    */
  public JavaCollectionPermuter()
    {
    }

  /**
    * Construct a <code>Permuter</code> for a collection and a permutation.
    * @param collection the collection to permute
    * @param permutation the purmutation array
    */
  public JavaCollectionPermuter(List collectionParm, int permutation[])
    {
      setCollection(collectionParm);
      count = -1; // unknown size
      setPermutation(permutation);
    }

  /**
    * Construct a Permuter with a permutation array.
    * Use <code>{@link #setCollection}</code>
    * to specifiy the collection to permute.
    * @param permutation the purmutation array
    */
  public JavaCollectionPermuter(int permutation[])
    {
      setPermutation(permutation);
    }

  /**
    * Construct a Permuter for a collection.
    * Use <code>{@link #setPermutation}</code>
    * to specifiy the permutation array.
    * @param collection the collection to permute
    */
  public JavaCollectionPermuter(List collectionParm)
    {
      setCollection(collectionParm);
    }

  /**
    * Construct a Permuter for a collection.
    * @param collection the collection to permute
    * @param size the collection size, used instead of
    *        <code>collection.count()</code>.
    */
  public JavaCollectionPermuter(List collectionParm, int size)
    {
      setCollection(collectionParm);
      this.count = size;
      haveSize = true;
    }

  ////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////

  /**
    * Return the collection that this <code>Permuter</code> will permute.
    * @return the collection that this <code>Permuter</code> will permute.
    */
  public List getCollection()
    {
      return collection;
    }

  /**
    * Set the collection which this <code>Permuter</code> will permute.
    * @param collection a ordered collection
    */
  public void setCollection(List collectionParm)
    {
      this.collection = collectionParm;
//	  if (collection instanceof AssociationListInterface)
//		aList = (AssociationListInterface) collection;
//	  else
//		aList = null;
    }

  /**
    * Set the collection which this <code>Permuter</code> will permute.
    * @param collection a Java collection
    * @param size the collection size, used instead of
    *        <code>collection.count()</code>.
    */
  public void setCollection(List collectionParm, int size)
    {
      this.collection = collectionParm;
      this.count = size;
      haveSize = true;
    }

  /**
    * Get the permutation for the collection.
    * @return the <code>int[]</code> permutation.
    */
  public int[] getPermutation()
    {
      return permutation;
    }

  /**
    * Set the permutation array.
    * @param an array of permutation indices.
    * A permutation array must satisfy the constraints
    * defined in <code>{@link #validatePermutation(int[])}</code>:
    */
  public void setPermutation(int permutation[])
    {
      validatePermutation(permutation);
      this.permutation = permutation;
    }

  /**
    * If the collection implements the {@link com.sas.util.EventGateInterface EventGateInterface},
    * use it to suppress events during a permute operation.
    */
//  protected void disableEvents()
//    {
//      if (collection instanceof EventGateInterface)
//        ((EventGateInterface) collection).disableEvents();
//    }

  /**
    * If the collection implements the {@link com.sas.util.EventGateInterface EventGateInterface},
    * use it to allow events after a permute operation.
    */
//  protected void enableEvents()
//    {
//      if (collection instanceof EventGateInterface)
//        ((EventGateInterface) collection).enableEvents();
//    }

  private int count()
    {
      return (  haveSize
              ? count
              : collection.size());
    }

  /**
    * Validate a permutation.
    * A permutation array must satisfy these constraints:
    * <ol>
    * <li>its length equals the collection.size()
    * <li>all values must be in the range [0, length-1]
    * <li>all values must be unique.
    * </ol>
    * @param permutation an array of indices to validate
    */
  protected void validatePermutation(int permutation[])
    {
      if (permutation == null)
        return;
      synchronized (collection)
        {
          synchronized (permutation)
            {
              int length = permutation.length;
              int count = count();
              if (length != count)
                throw new IllegalArgumentException("permutation.length {" + length +
                                                   "} != collection.count {" + count + "}");
              int permuted[] = new int[length];
              for (int i = 0; i < length; i++)
                permuted[i] = -1;
              for (int i = 0; i < length; i++)
                {
                  int j = permutation[i];
                  if (j < 0 || j >= length)
                    throw new IllegalArgumentException("permutation["/*I18nOK:EMS*/ + i +
                                                       "] == " + j + " ! in [0,"/*I18nOK:EMS*/ + (length-1) + "]");
                  if (permuted[j] != -1)
                    throw new IllegalArgumentException("permutation["/*I18nOK:EMS*/ + i + "] == permutation[" + permuted[j] + "]");

                  permuted[j] = i;
                }
            }
        }
    }

  /**
    * Permute the collection according to the permutation.
    * This reorders the items in collection.
//    * If the collection implements the {@link com.sas.util.EventGateInterface EventGateInterface},
//    * use it to suppress events during the permutation operation.
//    * so that an event is not fired for every item that is moved.
    */
  public synchronized void permute()
    {
      synchronized (permutation)
        {
          // find the range of permuted values, to minimize swapping.
          int start, end;
          int length = permutation.length;
          for (start = 0; start < (length - 1) && permutation[start] == start; start++)
            ;
          for (end = length-1; end > 0 && permutation[end] == end; end--)
            ;
          if ((end - start) < 1)
            return;
          permute(start, end+1);
        }
    }

  private void validateIndex(int index, int length)
    {
      if (index < 0 || index >= length)
        throw new IllegalArgumentException("index < 0 || index >= collection.count()"/*I18nOK:EMS*/);
    }

  private void validateStartEnd(int start, int end, int length)
    {
      if (start < 0 || start >= length)
        throw new IllegalArgumentException("start < 0 || start >= collection.count()"/*I18nOK:EMS*/);
      if (end < 0 || end > length)
        throw new IllegalArgumentException("end < - || end > collection.count()"/*I18nOK:EMS*/);
    }

  /**
    * Permute a subset of the collection according to the permutation.
    * Items outside the range [star, end-1] are not affected.
    * @param start the index of the first item to permute
    * @param end one past the index of the last item to permute.
    */
  private void permute(int start, int end)
    // This is protected so that clients cannot call it -- it is
    // dangerous, since they can violate the purmutation rules
    // if the values outside [start,end) do not match the
    // identify permutation.
    {
      synchronized (collection)
        {
          synchronized (permutation)
            {
              try
                {
                  int length = permutation.length;
                  validateStartEnd(start, end, length);
//                disableEvents();
                  int count = end - start;
                  Object objects[] = new Object[count];
				  if (aList != null)
					{
					  Object keys[] = new Object[count];
					  for (int i = start, j = 0; i < end; i++, j++)
						{
						  int k = permutation[i];
						  objects[j] = collection.get(k);
//						  keys[j] = aList.getKey(k);
						}
					  for (int i = start, j = 0; i < end; i++, j++)
						{
//						  aList.set(i, keys[j], objects[j]);
						}
					}
				  else
					{
					  for (int i = start, j = 0; i < end; i++, j++)
						{
						  int k = permutation[i];
						  objects[j] = collection.get(k);
						}
					  for (int i = start, j = 0; i < end; i++, j++)
						{
						  collection.set(i, objects[j]);
						}
					}
                }
              finally
                {
//                  enableEvents();
                }
            }
        }
    }

  /**
    * Set the permutation, then permutes the collection.
    */
  public void permute(int permutation[])
    {
      setPermutation(permutation);
      permute();
    }

  /**
    * A static method useful for a one-time permutation of a collection.
    * @param collection the collection to permite
    * @param permutation the permutation array
    */
  public static void permute(List collectionParm, int permutation[])
    {
      synchronized (collectionParm)
        {
          new JavaCollectionPermuter(collectionParm, permutation).permute();
        }
    }

  /**
    * A static method useful for reversing a (subrange) of items in a collection.
    * To reverse the entire collection, use
    * <pre>
    * reverse(collection, 0, collection.count());
    * </pre>
    * @param collection the collection to reverse
    * @param start the index of the first item in the range of items to reverse
    * @param end one past the index of the last item in the range of items to reverse.
    */
  public static void reverse(List collectionParm, int start, int end)
    {
      synchronized (collectionParm)
        {
          new JavaCollectionPermuter(collectionParm).reverse(start, end);
        }
    }

  /**
    * Reverse a (subrange) of the collection.
    * This does a <code>{@link #setPermutation}</code>.
    * To reverse the entire collection, use
    * <pre>
    * reverse(0, collection.count());
    * </pre>
    * @param start the index of the first item in the range of items to reverse
    * @param end one past the index of the last item in the range of items to reverse.
    */
  public void reverse(int start, int end)
    {
      synchronized (collection)
        {
          permute( newReversePermutation(start, end) );
        }
    }

  /**
    * Create a permutation array which reverses the items
    * between <var>start</var> and <var>end</var>
    * @param start the index of the first item in the range of items to reverse
    * @param end one past the index of the last item in the range of items to reverse.
    * @return a new int array for the permutation.
    */
  public int[] newReversePermutation(int start, int end)
    {
      int length = count();
      int indices[] = new int[length];
      for (int i = 0; i < start; i++)
        indices[i] = i;
      for (int i = start, j = end - 1; i < end; i++, j--)
        indices[i] = j;
      for (int i = end; i < length; i++)
        indices[i] = i;
      return indices;
    }

  /**
    * Convenience method to swap two items in an ordered collection.
    * This method handles the special case of swapping items
    * in an <code>{@link com.sas.collection.AssociationListInterface
    * AssociationListInterface</code>};
    * it swaps the key and item values at <var>i</var> and <var>j</var>
    * @param collection the collection in which to swap items
    * @param i index of one item to swap
    * @param j index of the other item to swap
    */
  public static void swap(List collection, int i, int j)
    {
      if (i == j)
        return;
      synchronized (collection)
        {
//           System.out.println("swap(" + i + "," + j + "), count=" + collection.count());
//
//          if ( collection instanceof AssociationListInterface )
//            {
//			  AssociationListInterface aList = (AssociationListInterface) collection;
//			  Object	tempValueI		= aList.get(i);
//			  Object	tempKeyI		= aList.getKey(i);
//			  aList.set( i, aList.getKey(j), aList.get(j) );
//			  aList.set( j, tempKeyI, tempValueI );
//            }
//          else
//            {
               Object temp = collection.get(i);
		       collection.set(i, collection.get(j));
		       collection.set(j, temp);
//            }
        }
    }

  /**
    * A static method useful for rotating a (subrange) of items in a collection.
    * To rotate the entire collection, use
    * <pre>
    * rotate(collection, 0, collection.count(), n);
    * </pre>
    * @param collection the collection to rotate
    * @param start the index of the first item in the range of items to rotate
    * @param end one past the index of the last item in the range of items to rotate
    * @param n the distance you want to rotate items.
    * @see #rotate(int, int, int)
    */
  public static void rotate(List collectionParm, int start, int end, int n)
    {
      new JavaCollectionPermuter(collectionParm).rotate(start, end, n);
    }

  /**
    * Rotate a (subrange) of items in a collection.
    * To rotate the entire collection, use
    * <pre>
    * rotate(0, collection.count(), n);
    * </pre>
    * @param collection the collection to rotate
    * @param start the index of the first item in the range of items to rotate.
    * @param end one past the index of the last item in the range of items to rotate.
    * @param n the distance you want to rotate items. If <code><var>n</var> &gt; 0</code>, items are rotated
    * right to left <var>n</var> places; if <code><var>n</var> &lt; 0</code>, items are rotated left to right <var>n</var> places.
    * For example, the collection
    * <pre>
    * { "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L" }
    * </pre>
    * rotated by 2 yields
    * <pre>
    * { "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "A", "B" }
    * </pre>
    */
  public void rotate(int start, int end, int n)
    {
      if (n == 0)
        return;
      // print("rotation", rotation);
      // print("permutation", permutation);
      setPermutation(newRotatePermutation(start, end, n));
      permute();
    }

  /**
    * Create a permutation array which rotates the items
    * between <var>start</var> and <var>end</var> by <var>n</var> items.
    * @param start the index of the first item in the range of items to rotate.
    * @param end one past the index of the last item in the range of items to rotate.
    * @param n the distance you want to rotate items. If <code><var>n</var> &gt; 0</code>, items are rotated
    * right to left <var>n</var> places; if <code><var>n</var> &lt; 0</code>, items are rotated left to right <var>n</var> places.
    * @return a new int array for the permutation.
    */
  public int[] newRotatePermutation(int start, int end, int n)
    {
      int count = count();
      validateStartEnd(start, end, count);
      int length = (end - start);
      if ( n < 0 )
        n = length+n;
      n = n % length;
      int indices[] = identityPermutation();
      if (n == 0)
        return indices;
      int rotation[] = new int[length];
      System.arraycopy(indices, start, rotation, length - n, n);
      System.arraycopy(indices, start + n, rotation, 0, length - n);
      System.arraycopy(rotation, 0, indices, start, length);
      return indices;
    }

   //   static void print(String name, int a[])
   //     {
   //       System.out.print(name + "{");
   //       for (int i = 0; i < a.length; i++)
   //         System.out.print(a[i] + ", ");
   //       System.out.println("}");
   //     }

  /**
    * Move an item in a collection by a specified offset.
    * @param collection the collection to permute.
    * @param index the index of the item you want to move.
    * @param offset the distance you want to move the item.
    * Move the item to position <var>index + offset</var>.
    * @see #move(int, int)
    */
  public static void move(List collectionParm, int index, int offset)
    {
      new JavaCollectionPermuter(collectionParm).move(index, offset);
    }

  /**
    * Move an item in the collection by a specified offset.
    * @param index the index of the item you want to move.
    * @param offset the distance you want to move the item.
    *        Negative numbers move the items closer to the beginning;
    *        positive numbers move the items towards the end.  Move
    *        the item to position <var>index + offset</var>.  All other
    *        items between <var>index</var> and <var>index+offset<var> are
    *        shifted. This therefore has the same effect as rotating
    *        the affected items left or right 1 item, depending on the
    *        sign of the offset.
    */
  public void move(int index, int offset)
    {
      if (offset == 0)
        return;

      synchronized (collection)
        {
          validateIndex(index, count());
          validateMove(index, offset);

          if (offset == 1 || offset == -1)
            {
              /*
               * ok, it's not clear what's best when we do a swap.
               * should I *not* disable events:
               *    swap will perform 2 set() calls and generate two events,
               *    but they may be efficiently handled with
               *    CollectionChangedInfo events;
               * or disable/enable events, so that enableEvents() will
               *    send a single event, but a generic one that
               *    listeners can't optimize?
               * Probably best to just allows the two events.
               *
               * try
               *  {
               * disableEvents();
               */

              swap(collection, index+offset, index);
              return;

              /*
               *
               * }
               * finally
               * {
               *   enableEvents();
               * }
               *
               */
            }

          // non trivial; requires block moves
          setPermutation(newMovePermutation(index, offset));
          permute();
        }
    }

  /**
    * Create a permutation which will
    * move an item in the collection by a specified offset.
    * @param index the index of the item you want to move.
    * @param offset the distance you want to move the item.
    *        Negative numbers move the items closer to the beginning;
    *        positive numbers move the items towards the end.
    * @return a new permutation array.
    */
  public int[] newMovePermutation(int index, int offset)
    {
      int indices[] = identityPermutation();
      moveIndex(indices, index, offset);
      return indices;
    }

  private void validateMove(int index, int offset)
    {
      int target = index + offset;
      if (target < 0 || target >= count())
        throw new IllegalArgumentException("index + offset < 0 || " +
                                           "index + offset >= length; " +
                                           "index == " + index +
                                           ", offset = " + offset);
    }

  /**
    * unchecked method to move an item; shifting all others.
    */
  private void moveIndex(int indices[], int index, int offset)
    {
      int target = index + offset;
      if (offset < 0)
        // shift items up
        System.arraycopy(indices, target, indices, target + 1, -offset);
      else
        // shift items down
        System.arraycopy(indices, index + 1, indices, index, offset);
      indices[target] = index;
    }

  /**
    * Move the items selected by an array of indices by an offset.
    * @param collection the collection to permute.
    * @param items an array of item indices. Each element
    *        must be a valid index within the collection.
    * @param offset how much to move the items.
    *        negative numbers move the items closer to the beginning;
    *        positive numbers move the items towards the end.
    * @see #move(int[], int)
    */
  public static void move(List collectionParm, int items[], int offset)
    {
      new JavaCollectionPermuter(collectionParm).move(items, offset);
    }

  /**
    * Move the items selected by an array of indices by an offset.
    * @param collection the collection to permute.
    * @param items an array of item indices. Each element
    *        must be a valid index within the collection.
    *        The items must be in ascending order and
    *        each index must be unique.
    * @param offset how much to move the items.
    *        negative numbers move the items closer to the beginning;
    *        positive numbers move the items towards the end.
    *
    */
  public void move(int items[], int offset)
    {
      if (offset == 0)
        return;
      permute(newMovePermutation(items, offset));
    }

  /**
    * Create a permutation which will
    * move by a specified offset the set of items specified by an array of item indeces.
    * @param items an array of indices identifying the items item you want to move.
    *        The array need not be contiguous, but the items must be in ascending order and
    *        each index must be unique.
    * @param offset the distance you want to move the items.
    *        negative numbers move the items closer to the beginning;
    *        positive numbers move the items towards the end.
    * @return a new permutation array.
    */
  public int[] newMovePermutation(int items[], int offset)
    {
      int length = count();
      int indices[] = identityPermutation();
      validateItems(items, offset, length);
      if (offset < 0)
        for (int i = 0; i < items.length; i++)
          {
            int index = items[i];
            moveIndex(indices, index, offset);
          }
      else
        for (int i = items.length - 1; i >= 0; i--)
          {
            int index = items[i];
            moveIndex(indices, index, offset);
          }
      return indices;
    }

  /**
   * Validate that all items are in the range <code>[0,length-1]</code>
   * may be moved by an offset.
   * @param items an array of permutation indices
   * @param offset how much to move the items.
   * @param length the permutation length.
   */
  protected void validateItems(int items[], int offset, int length)
    {
      for (int i = 0; i < items.length; i++)
        {
          int index = items[i];
          validateIndex(index, length);
          validateMove(index, offset);
          if (i > 0)
            {
              int previous = items[i - 1];
              if (previous >= index)
                throw new IllegalArgumentException("items[ + "/*I18nOK:EMS*/ + (i-1) + "] >= items[" + i + "]");
            }
        }
    }

  /**
    * Return an identity permutation.
    * @return an int array with values {0, 1, 2, ..., collection.count()-1}
    */
  public int[] identityPermutation()
    {
      int length = count();
      int permutation[] = new int[length];
      for (int i = 0; i < length; i++)
        permutation[i] = i;
      return permutation;
    }

  /**
    * Return an identity permutation.
    * @param length the size of the permutation array
    * @return an int array with values {0, 1, 2, ..., length-1}
    */
  public static int[] identityPermutation(int length)
    {
      int permutation[] = new int[length];
      for (int i = 0; i < length; i++)
        permutation[i] = i;
      return permutation;
    }
}
