package com.sas.util;


import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.RandomAccess;

/**
 * The purpose of this class is to provide a seemless method for allocating new
 * List objects, <b>especially</b> List objects which are specialized at being
 * sharable. There are two significant motivations for using this factory.<ol>
 *
 * <li>This factory isolates users from the actual List class being used,
 * allowing the decision of which class is appropriate to change over
 * time.  (Impl note: Currently the list returned is an internal implementation,
 * a COWList, COW - Copy On Write.)</li>
 *
 * <li>Usage patterns such as described below are excellent candidates for this class:
 *   <ul>
 *
 *     <li>A subsytem exports, via defensive copy, its internal list structure.</li>
 *
 *     <li>Sometimes, consumers mutate that list, and feed it back to the
 *     subsystem via an appropriate interface.  (Hence, consumers currently
 *     expect a mutable list, making exporting Collections.unmodifiableList()
 *     instead of a defensive copy not suitable).</li>
 *
 *     <li>At least one consumer calls this "get list" functionality MANY
 *     times.</li>
 *
 *     <li>The returned List needs to be mutable, but rarely is actually
 *     mutated.</li>
 *
 *     <li>As a result, many copies of List are made, at a non-trivial
 *     expense.</li>
 *
 *   </ul>
 * </li>
 *
 * The default List returned by this class has the following characteristics:
 * <ol>
 *   <li>Full List semantics</li>
 *   <li>Synchronized - it behaves just as a list returned by
 *   Collections.synchronizedList() would.</li>
 *   <li>Fast - almost as fast as a Collections.synchronizedList().</li>
 *   <li><b>Minimal memory use</b> - wherever possible multiple lists share the actual
 *   internal storage, but the is never done at the expense of correctness.
 *   This is achieved by carefully detecting any "mutate" operations, and
 *   ensuring the storage is not shared just prior to a mutate.  Operations
 *   such as iterators, and "fast fail" semantics are properly maintained, even
 *   across a mutate driven copy.</li>
 * </ol>
 * <p>
 *
 * For experimentation / performance tuning purposes, the class of the List
 * returned by the <code>newList()</code> functions can be influenced by
 * setting the Java property: "com.sas.util.ListFactory.strategy".  The default
 * behavior, and the only one adhering to the contract specified above, is
 * "COWList".  Other values are "ArrayList" which results in a simple ArrayList
 * being returned, and "SynchronizedList" which returns an ArrayList, wrapped
 * with a call to Collections.synchronizedList().
 *
 * @author Brian Thorstad
 */
public final class ListFactory
{

    // TODO (thorstad) - use real strategy object, not switch statements.

    /* These used to decide upon a strategy, see static code below.  */
    private static final String PROP_STRATEGY_KEY = "com.sas.util.ListFactory.strategy";
    private static final String PROP_COW_LIST   = "COWList";
    private static final String PROP_ARRAY_LIST = "ArrayList";
    private static final String PROP_SYNC_LIST  = "SynchronizedList";

    /** The default type of list this factory returns, and internally
     * implemented 'copy on write' list. */
    private static final int ALLOC_COW_LIST = 0;
    /** An override, recommended only for testing. Return a plain ArrayList.*/
    private static final int ALLOC_LIST = 1;
    /** An override, recommended only for testing. Return a synchronized
     * ArrayList. */
    private static final int ALLOC_SYNC_LIST = 2;

    /**
     * What kind of list should be allocated?  Intentionally non-final
     * (initially) allowing debugger to change, or whatever.
     */
    private static final int ALLOC_STRATEGY;
    static
    {
        int strategy = ALLOC_COW_LIST;
        try
        {
            String strategyRequested = System.getProperty(PROP_STRATEGY_KEY);

            if (strategyRequested != null && strategyRequested.length()>0)
            {
                System.out.println("*** ListFactory - Strategy OVERRIDE requested ("+strategyRequested+") ***"); /*I18nOK:Line*/
                if (strategyRequested.equalsIgnoreCase(PROP_COW_LIST))
                {
                    strategy = ALLOC_COW_LIST;
                }
                else if (strategyRequested.equalsIgnoreCase(PROP_ARRAY_LIST))
                {
                    strategy = ALLOC_LIST;
                }
                else if (strategyRequested.equalsIgnoreCase(PROP_SYNC_LIST))
                {
                    strategy = ALLOC_SYNC_LIST;
                }
                else
                {
                    System.out.println("*** ListFactory ignoring unrecognized option ("+strategyRequested+") ***"); /*I18nOK:Line*/
                }
            }
        }
        catch (RuntimeException e)
        {
            // ignore, go w/ default
        }
        ALLOC_STRATEGY = strategy;
    }

    /**
     * this class uses static fncs, disallow constructor.
     */
    private ListFactory()
    {
    }

    /**
     * This is the proper technique for creating a "copy" of a List.  The
     * returned list will be a completely independent (shallow) copy of the
     * supplied source collection.  It can be iterated over independently, can
     * mutate, ...  Like other collection copies, the collection is unique, but
     * the underlying objects are all the same.  That is, for every object in
     * source, that same object will also be in this returned List.
     *
     * <p>The motive for this function, however, is that if the underlying
     * source collection was allocated by this class (and hence is a COWList),
     * then internal implementation allows sharing of the storage area, until
     * one of the collections mutates.  See class comments for more
     * information.
     */
    public static List newList(Collection source)
    {
        switch (ALLOC_STRATEGY)
        {
          case ALLOC_LIST:
            return new ArrayList(source);
          case ALLOC_SYNC_LIST:
            return Collections.synchronizedList(new ArrayList(source));
          case ALLOC_COW_LIST:
          default:
            // Note: if source instanceof COWList - special behavior kicks
            // in...This is the point of all this effort.
            return new COWList(source);
        }
    }

    /**
     * Construct and return a new List - by default the inner class
     * implementation, a COWList.  See class comments.
     */
    public static List newList()
    {
        switch (ALLOC_STRATEGY)
        {
          case ALLOC_LIST:
            return new ArrayList();
          case ALLOC_SYNC_LIST:
            return Collections.synchronizedList(new ArrayList());
          case ALLOC_COW_LIST:
          default:
            return new COWList();
        }
    }

    /**
     * Construct and return a new List - by default the inner class
     * implementation, a COWList.  See class comments.
     *
     * @param initSize passed to the list's constructor as a hint for the
     * initial allocation of internal storage, of course, the List is free to
     * grow beyond this.
     */
    public static List newList(int initSize)
    {
        switch (ALLOC_STRATEGY)
        {
          case ALLOC_LIST:
            return new ArrayList(initSize);
          case ALLOC_SYNC_LIST:
            return Collections.synchronizedList(new ArrayList(initSize));
          case ALLOC_COW_LIST:
          default:
            return new COWList(initSize);
        }
    }

    /* *************** START Inner Class COWList *********************** */
    /* *************** START Inner Class COWList *********************** */


/**
 * This class is not a list of bovines to be processed, but instead mimics
 * ArrayList, while allowing for multiple independent COWLists to share the
 * same storage area, in a fashion much the same as String and StringBuffer do.
 * That is, with a <b>Copy On Write</b> strategy, also known as "COW".  The
 * motive to use such a list is that we are observing <em>many</em> occasions
 * where a defensive copy of a list is made to protect the internal structure
 * of the owner of the list, yet the consumer never intends to mutate the list.
 * In this scenario, passing a new ArrayList results in an extra Object[] being
 * created, and filled.  Performance testing shows this is happening thousands
 * of times for minor operations (and hints many more for "real" operations).
 *
 * <p>Implementation detail - this class borrowed rather heavily from
 * ArrayList.  ("borrow" - I used its source as the starting point, and
 * minimally reshapped it).  The only significant change was to place all
 * dynamic content in a static inner class (called Shared), and manage access
 * to that reference counting object.
 *
 * <p>Synchronization, thread safety, and Collections.SynchronizedList: This
 * class is fully synchronized and thread safe.  It does not need to be wrapped
 * in a SynchronizedList, in fact that would be counter productive because for
 * <code>COWList</code>s to effectively share storage, they need unfettered
 * access to eachother (which is why this is already fully synchronized in the
 * fashion of Collections.SynchronizedList().  All comments about how to
 * properly use a Synchronized collection (think Iterator) apply.
 *
 * <p>Synchronization between differnt instances of COWList which are currently
 * sharing storage is unnecessary.  Any coordination is handled properly,
 * including safely copying during the "copy on write" operation.  Iterators,
 * including "fast fail" semantics are handled correctly; furthermore, they
 * only "fail" for this list which has suffered a co-modification.
 *
 * <p>------- The remainder of the documentation is from ArrayList --------
 *
 * <p>Resizable-array implementation of the <tt>List</tt> interface.
 * Implements all optional list operations, and permits all elements, including
 * <tt>null</tt>.  In addition to implementing the <tt>List</tt> interface,
 * this class provides methods to manipulate the size of the array that is used
 * internally to store the list.  (This class is roughly equivalent to
 * <tt>Vector</tt>, except that it is unsynchronized.)<p>
 *
 * The <tt>size</tt>, <tt>isEmpty</tt>, <tt>get</tt>, <tt>set</tt>,
 * <tt>iterator</tt>, and <tt>listIterator</tt> operations run in constant
 * time.  The <tt>add</tt> operation runs in <i>amortized constant time</i>,
 * that is, adding n elements requires O(n) time.  All of the other operations
 * run in linear time (roughly speaking).  The constant factor is low compared
 * to that for the <tt>LinkedList</tt> implementation.<p>
 *
 * Each <tt>ArrayList</tt> instance has a <i>capacity</i>.  The capacity is
 * the size of the array used to store the elements in the list.  It is always
 * at least as large as the list size.  As elements are added an ArrayList,
 * its capacity grows automatically.  The details of the growth policy are not
 * specified beyond the fact that adding an element has constant amortized
 * time cost.<p>
 *
 * An application can increase the capacity of an <tt>ArrayList</tt> instance
 * before adding a large number of elements using the <tt>ensureCapacity</tt>
 * operation.  This may reduce the amount of incremental reallocation.<p>
 *
 * <strong>Note that this implementation is not synchronized.</strong> If
 * multiple threads access an <tt>ArrayList</tt> instance concurrently, and at
 * least one of the threads modifies the list structurally, it <i>must</i> be
 * synchronized externally.  (A structural modification is any operation that
 * adds or deletes one or more elements, or explicitly resizes the backing
 * array; merely setting the value of an element is not a structural
 * modification.)  This is typically accomplished by synchronizing on some
 * object that naturally encapsulates the list.  If no such object exists, the
 * list should be "wrapped" using the <tt>Collections.synchronizedList</tt>
 * method.  This is best done at creation time, to prevent accidental
 * unsynchronized access to the list:
 * <pre>
 *	List list = Collections.synchronizedList(new ArrayList(...));
 * </pre><p>
 *
 * The iterators returned by this class's <tt>iterator</tt> and
 * <tt>listIterator</tt> methods are <i>fail-fast</i>: if list is structurally
 * modified at any time after the iterator is created, in any way except
 * through the iterator's own remove or add methods, the iterator will throw a
 * ConcurrentModificationException.  Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the
 * future.
 *
 * <p>Note that the fail-fast behavior of an iterator cannot be guaranteed
 * as it is, generally speaking, impossible to make any hard guarantees in the
 * presence of unsynchronized concurrent modification.  Fail-fast iterators
 * throw <tt>ConcurrentModificationException</tt> on a best-effort basis.
 * Therefore, it would be wrong to write a program that depended on this
 * exception for its correctness: <i>the fail-fast behavior of iterators
 * should be used only to detect bugs.</i>
 *
 * @author  Brian Thorstad (borrowing heavily from Sun's Josh Bloch)
 * @version 1 (based on 1.4.1_01)
 * @see	    java.util.ArrayList
 */

    private static class COWList extends AbstractList
        implements List, RandomAccess, Cloneable, java.io.Serializable
    {
        private static final long serialVersionUID = 8683452581122892189L;

        private static final int  INITIAL_ARRAY_SIZE = 10;

        private transient Shared _shared; // guaranteed non-null after construction

        /**
         * Preferred constructor for sharing with another COWList.  The
         * resulting list will share the same storage as the input.
         */
        public COWList(COWList cowListToShare)
        {
            super();
            share(cowListToShare);
        }

        /**
         * Constructs an empty list with the specified initial capacity.
         *
         * @param   initialCapacity   the initial capacity of the list.
         * @exception IllegalArgumentException if the specified initial capacity
         *            is negative
         */
        public COWList(int initialCapacity) {
            super();
            if (initialCapacity < 0)
                throw new IllegalArgumentException("Illegal Capacity: "+/*I18nOK:LINE*/
                                                   initialCapacity);
            _shared = new Shared(initialCapacity);
        }

        /**
         * Constructs an empty list with an initial capacity of ten (five).
         */
        public COWList() {
            this(INITIAL_ARRAY_SIZE);  // 10 ?
        }

        /**
         * Constructs a list containing the elements of the specified
         * collection, in the order they are returned by the collection's
         * iterator.  The <tt>COWList</tt> instance has an initial capacity of
         * 110% the size of the specified collection.
         *
         * <p>Note: If c is an instanceof COWList the behavior is identical to
         * {@link #COWList(COWList)}, otherwise it is as described above.
         *
         * @param c the collection whose elements are to be placed into this list.
         * @throws NullPointerException if the specified collection is null.
         */
        public COWList(Collection c) {
            if (c instanceof COWList)
            {
                share((COWList)c);
            }
            else
            {
                _shared = new Shared(c);
            }
        }


        /* *********** START "sharing" Code Section ************************ */


        /* ------------ START Inner class holding my data (Shared) --------- */
        private static class Shared
        {
            /**
             * The array buffer into which the elements of the ArrayList are
             * stored.  The capacity of the ArrayList is the length of this array
             * buffer.
             *
             * <b>Note</b> that this array is potentially shared
             */
            Object _elementData[];
            /**
             * The size of the COWList (the number of elements it contains).
             *
             * @serial
             */
            int _size;

            int _useCount;

            Shared(int initialCapacity)
            {
                _elementData = new Object[initialCapacity];
                _useCount = 1;
            }

            /**
             * Make myself a shallow copy of the original, this is likely being
             * done because I am about to mutate myself.  When complete,
             * _useCount=1, _size=orig._size, and _elementData is a new
             * ObjectArray, with all the same references as orig._elementData.
             *
             * <p>This construtor is the heart of the "copy" for the "copy on
             * write" algorithm which this class was designed for.
             */
            Shared(Shared orig, boolean decrementOrigUseCount)
            {
                synchronized (orig)
                {
                    _size = orig._size;
                    _useCount = 1;
                    _elementData = new Object[_size];
                    System.arraycopy(orig._elementData, 0, _elementData, 0, _size);
                    if (decrementOrigUseCount)
                    {
                        orig._useCount--;
                    }
                }
            }

            /**
             * Constructor similar to (and logic borrowed from) {@link
             * ArrayList#ArrayList(Collection)}.
             */
            Shared(Collection c)
            {
                _size = c.size();
                // Allow 10% room for growth
                _elementData = new Object[
                    (int)Math.min((_size*110L)/100,Integer.MAX_VALUE)];
                c.toArray(_elementData);
                _useCount = 1;
            }

            /**
             * Constructor to use supplied array - intended for deserialization
             * @param elementData
             */
            private Shared(Object[] elementData)
            {
                _elementData = elementData;
                _size        = elementData.length;
                _useCount    = 1;
            }

            synchronized boolean isShared()
            {
                return _useCount > 1;
            }
        } // inner class - Shared
        /* ------------  END  Inner class holding my data (Shared) --------- */

        /**
         * "share" the parameter's internal list, abandoning my previous store.
         * This is likely only called from the "copy constructors.
         */
        private void share(COWList cowListToShare)
        {
            if (_shared != null)
            {
                _shared._useCount--;
                _shared = null;
            }

            synchronized (cowListToShare._shared)
            {
                _shared = cowListToShare._shared;
                _shared._useCount++;
            }
        } // share

        /**
         * Stop sharing, this is called just prior to a mutation event.  Typically,
         * nothing is done, as typically we are not "shared", and hence do not have
         * a responsibility to make a copy.  However, if we are "shared", we (the
         * mutating list) have a responsibility to "copy on write" - I must make a
         * private copy prior to mutating.
         */
        private void unshare()
        {
            synchronized (_shared)
            {
                if (_shared.isShared())
                {
                    // make private copy, decrement use count of original
                    _shared = new Shared(_shared, true);
                }
                // else - no need to bother, keep using _shared.
            }
        } // unshare

        /**
         * Similar to {@link #clear()} but this function goes farther in that it
         * releases the underlying array storage.  Further use of this list after
         * destroy() is invoked will fail, typically with NullPointerException.
         *
         * <p>Initially (and still?) this funcion is private, and only invoked by
         * the finalizer.  But, functionality is separate, allowing public access
         * if suitable reason for expanding beyond the List interface support
         * this. (It is potentially more effecient on the running system to invoke
         * this manually).
         *
         * <p>This function is idempotent - it can be invoked multiple times
         * without harm.
         */
        private void destroy()
        {
            // "unshare()" by hand here.
            if (_shared != null)
            {
                synchronized(_shared)
                {
                    _shared._useCount--;
                    _shared = null;  // WILL cause future NullPointerException
                }
            }
            modCount++;
        }

        /**
         * Finalizer needed to ensure destroy() is called.
         */
        protected void finalize() throws Throwable
        {
            // the form here follows Sun's recommended form for finalize()
            try
            {
                destroy();
            }
            finally
            {
                super.finalize();
            }
        }

        /* ***********  END  "sharing" Code Section ************************ */


        /**
         * Trims the capacity of this <tt>COWList</tt> instance to be the
         * list's current size.  An application can use this operation to minimize
         * the storage of an <tt>COWList</tt> instance.
         */
        public synchronized void trimToSize() {
            unshare();
            modCount++;
            int oldCapacity = _shared._elementData.length;
            if (_shared._size < oldCapacity) {
                Object oldData[] = _shared._elementData;
                _shared._elementData = new Object[_shared._size];
                System.arraycopy(oldData, 0, _shared._elementData, 0, _shared._size);
            }
        }

        /**
         * Increases the capacity of this <tt>COWList</tt> instance, if
         * necessary, to ensure  that it can hold at least the number of elements
         * specified by the minimum capacity argument.
         *
         * @param   minCapacity   the desired minimum capacity.
         */
        public synchronized void ensureCapacity(int minCapacity) {
            unshare();
            modCount++;
            int oldCapacity = _shared._elementData.length;
            if (minCapacity > oldCapacity) {
                Object oldData[] = _shared._elementData;
                int newCapacity = (oldCapacity * 3)/2 + 1;
                if (newCapacity < minCapacity)
                    newCapacity = minCapacity;
                _shared._elementData = new Object[newCapacity];
                System.arraycopy(oldData, 0, _shared._elementData, 0, _shared._size);
            }
        }

        /**
         * Returns the number of elements in this list.
         *
         * @return  the number of elements in this list.
         */
        public synchronized int size() {
            return _shared._size;
        }

        /**
         * Tests if this list has no elements.
         *
         * @return  <tt>true</tt> if this list has no elements;
         *          <tt>false</tt> otherwise.
         */
        public synchronized boolean isEmpty() {
            return _shared._size == 0;
        }

        /**
         * Returns <tt>true</tt> if this list contains the specified element.
         *
         * @param elem element whose presence in this List is to be tested.
         * @return  <code>true</code> if the specified element is present;
         *		<code>false</code> otherwise.
         */
        public synchronized boolean contains(Object elem) {
            return indexOf(elem) >= 0;
        }

        /**
         * Searches for the first occurence of the given argument, testing
         * for equality using the <tt>equals</tt> method.
         *
         * @param   elem   an object.
         * @return  the index of the first occurrence of the argument in this
         *          list; returns <tt>-1</tt> if the object is not found.
         * @see     Object#equals(Object)
         */
        public synchronized int indexOf(Object elem) {
            if (elem == null) {
                for (int i = 0; i < _shared._size; i++)
                    if (_shared._elementData[i]==null)
                        return i;
            } else {
                for (int i = 0; i < _shared._size; i++)
                    if (elem.equals(_shared._elementData[i]))
                        return i;
            }
            return -1;
        }

        /**
         * Returns the index of the last occurrence of the specified object in
         * this list.
         *
         * @param   elem   the desired element.
         * @return  the index of the last occurrence of the specified object in
         *          this list; returns -1 if the object is not found.
         */
        public synchronized int lastIndexOf(Object elem) {
            if (elem == null) {
                for (int i = _shared._size-1; i >= 0; i--)
                    if (_shared._elementData[i]==null)
                        return i;
            } else {
                for (int i = _shared._size-1; i >= 0; i--)
                    if (elem.equals(_shared._elementData[i]))
                        return i;
            }
            return -1;
        }

        /**
         * Returns a shallow copy of this <tt>COWList</tt> instance.  (The
         * elements themselves are not copied.)
         *
         * @return  a clone of this <tt>COWList</tt> instance.
         */
        public synchronized Object clone() {
            // this class has a special impl which makes the following the perfect
            // implementation.
            return new COWList(this);
        }

        /**
         * Returns an array containing all of the elements in this list
         * in the correct order.
         *
         * @return an array containing all of the elements in this list
         * 	       in the correct order.
         */
        public synchronized Object[] toArray() {
            Object[] result = new Object[_shared._size];
            System.arraycopy(_shared._elementData, 0, result, 0, _shared._size);
            return result;
        }

        /**
         * Returns an array containing all of the elements in this list in the
         * correct order; the runtime type of the returned array is that of the
         * specified array.  If the list fits in the specified array, it is
         * returned therein.  Otherwise, a new array is allocated with the runtime
         * type of the specified array and the size of this list.<p>
         *
         * If the list fits in the specified array with room to spare (i.e., the
         * array has more elements than the list), the element in the array
         * immediately following the end of the collection is set to
         * <tt>null</tt>.  This is useful in determining the length of the list
         * <i>only</i> if the caller knows that the list does not contain any
         * <tt>null</tt> elements.
         *
         * @param a the array into which the elements of the list are to
         *		be stored, if it is big enough; otherwise, a new array of the
         * 		same runtime type is allocated for this purpose.
         * @return an array containing the elements of the list.
         * @throws ArrayStoreException if the runtime type of a is not a supertype
         *         of the runtime type of every element in this list.
         */
        public synchronized Object[] toArray(Object a[]) {
            if (a.length < _shared._size)
                a = (Object[])java.lang.reflect.Array.newInstance(
                    a.getClass().getComponentType(), _shared._size);

            System.arraycopy(_shared._elementData, 0, a, 0, _shared._size);

            if (a.length > _shared._size)
                a[_shared._size] = null;

            return a;
        }

        // Positional Access Operations

        /**
         * Returns the element at the specified position in this list.
         *
         * @param  index index of element to return.
         * @return the element at the specified position in this list.
         * @throws    IndexOutOfBoundsException if index is out of range <tt>(index
         * 		  &lt; 0 || index &gt;= _size())</tt>.
         */
        public synchronized Object get(int index) {
            RangeCheck(index);

            return _shared._elementData[index];
        }

        /**
         * Replaces the element at the specified position in this list with
         * the specified element.
         *
         * @param index index of element to replace.
         * @param element element to be stored at the specified position.
         * @return the element previously at the specified position.
         * @throws    IndexOutOfBoundsException if index out of range
         *		  <tt>(index &lt; 0 || index &gt;= _size())</tt>.
         */
        public synchronized Object set(int index, Object element) {
            RangeCheck(index);
            // unshare() is special here, as COW *is* needed, but co-mod is not.

            Object oldValue = _shared._elementData[index];
            if (oldValue != element) // must be identity compare here
            {
                unshare();
                _shared._elementData[index] = element;
            }
            return oldValue;
        }

        /**
         * Appends the specified element to the end of this list.
         *
         * @param o element to be appended to this list.
         * @return <tt>true</tt> (as per the general contract of Collection.add).
         */
        public synchronized boolean add(Object o) {
            ensureCapacity(_shared._size + 1);  // Increments modCount!!, & unshare()
            _shared._elementData[_shared._size++] = o;
            return true;
        }

        /**
         * Inserts the specified element at the specified position in this
         * list. Shifts the element currently at that position (if any) and
         * any subsequent elements to the right (adds one to their indices).
         *
         * @param index index at which the specified element is to be inserted.
         * @param element element to be inserted.
         * @throws    IndexOutOfBoundsException if index is out of range
         *		  <tt>(index &lt; 0 || index &gt; size())</tt>.
         */
        public synchronized void add(int index, Object element) {
            if (index > _shared._size || index < 0)
                throw new IndexOutOfBoundsException(
                    "Index: "/*I18nOK:EMS*/+index+", Size: "+_shared._size);

            ensureCapacity(_shared._size+1);  // Increments modCount!! & unshare()
            System.arraycopy(_shared._elementData, index, _shared._elementData, index + 1,
                             _shared._size - index);
            _shared._elementData[index] = element;
            _shared._size++;
        }

        /**
         * Removes the element at the specified position in this list.
         * Shifts any subsequent elements to the left (subtracts one from their
         * indices).
         *
         * @param index the index of the element to removed.
         * @return the element that was removed from the list.
         * @throws    IndexOutOfBoundsException if index out of range <tt>(index
         * 		  &lt; 0 || index &gt;= size())</tt>.
         */
        public synchronized Object remove(int index) {
            RangeCheck(index);

            unshare();
            modCount++;
            Object oldValue = _shared._elementData[index];

            int numMoved = _shared._size - index - 1;
            if (numMoved > 0)
                System.arraycopy(_shared._elementData, index+1, _shared._elementData, index,
                                 numMoved);
            _shared._elementData[--_shared._size] = null; // Let gc do its work

            return oldValue;
        }

        /**
         * Removes all of the elements from this list.  The list will
         * be empty after this call returns.
         */
        public synchronized void clear() {
            // "unshare()" by hand here.
            if (_shared != null)
            {
                synchronized(_shared)
                {
                    _shared._useCount--;
                }
            }
            // abandon old _shared
            _shared = new Shared(0);

            modCount++;

        }

        /**
         * Appends all of the elements in the specified Collection to the end of
         * this list, in the order that they are returned by the
         * specified Collection's Iterator.  The behavior of this operation is
         * undefined if the specified Collection is modified while the operation
         * is in progress.  (This implies that the behavior of this call is
         * undefined if the specified Collection is this list, and this
         * list is nonempty.)
         *
         * @param c the elements to be inserted into this list.
         * @return <tt>true</tt> if this list changed as a result of the call.
         * @throws    IndexOutOfBoundsException if index out of range <tt>(index
         *		  &lt; 0 || index &gt; size())</tt>.
         * @throws    NullPointerException if the specified collection is null.
         */
        public synchronized boolean addAll(Collection c) {
            unshare();
            modCount++;
            int numNew = c.size();
            ensureCapacity(_shared._size + numNew);

            Iterator e = c.iterator();
            for (int i=0; i<numNew; i++)
                _shared._elementData[_shared._size++] = e.next();

            return numNew != 0;
        }

        /**
         * Inserts all of the elements in the specified Collection into this
         * list, starting at the specified position.  Shifts the element
         * currently at that position (if any) and any subsequent elements to
         * the right (increases their indices).  The new elements will appear
         * in the list in the order that they are returned by the
         * specified Collection's iterator.
         *
         * @param index index at which to insert first element
         *		    from the specified collection.
         * @param c elements to be inserted into this list.
         * @return <tt>true</tt> if this list changed as a result of the call.
         * @throws    IndexOutOfBoundsException if index out of range <tt>(index
         *		  &lt; 0 || index &gt; size())</tt>.
         * @throws    NullPointerException if the specified Collection is null.
         */
        public synchronized boolean addAll(int index, Collection c) {
            if (index > _shared._size || index < 0)
                throw new IndexOutOfBoundsException(
                    "Index: "/*I18nOK:EMS*/+index+", Size: "+_shared._size);

            int numNew = c.size();
            ensureCapacity(_shared._size + numNew);  // Increments modCount!! & unshare()

            int numMoved = _shared._size - index;
            if (numMoved > 0)
                System.arraycopy(_shared._elementData, index, _shared._elementData, index + numNew,
                                 numMoved);

            Iterator e = c.iterator();
            for (int i=0; i<numNew; i++)
                _shared._elementData[index++] = e.next();

            _shared._size += numNew;
            return numNew != 0;
        }

        /**
         * Removes from this List all of the elements whose index is between
         * fromIndex, inclusive and toIndex, exclusive.  Shifts any succeeding
         * elements to the left (reduces their index).
         * This call shortens the list by <tt>(toIndex - fromIndex)</tt> elements.
         * (If <tt>toIndex==fromIndex</tt>, this operation has no effect.)
         *
         * @param fromIndex index of first element to be removed.
         * @param toIndex index after last element to be removed.
         */
        protected void removeRange(int fromIndex, int toIndex) {
            unshare();
            modCount++;
            int numMoved = _shared._size - toIndex;
            System.arraycopy(_shared._elementData, toIndex, _shared._elementData, fromIndex,
                             numMoved);

            // Let gc do its work
            int newSize = _shared._size - (toIndex-fromIndex);
            while (_shared._size != newSize)
                _shared._elementData[--_shared._size] = null;
        }

        /**
         * Check if the given index is in range.  If not, throw an appropriate
         * runtime exception.
         */
        private void RangeCheck(int index) {
            if (index >= _shared._size || index < 0)
                throw new IndexOutOfBoundsException(
                    "Index: "/*I18nOK:EMS*/+index+", Size: "+_shared._size);
        }

        /**
         * Save the state of the <tt>COWList</tt> instance to a stream (that
         * is, serialize it).
         *
         * @serialData The length of the array backing the <tt>COWList</tt>
         *             instance is emitted (int), followed by all of its elements
         *             (each an <tt>Object</tt>) in the proper order.
         */
        private synchronized void writeObject(java.io.ObjectOutputStream s)
            throws java.io.IOException{
            // Write out element count, and any hidden stuff
            s.defaultWriteObject();

            // Write out array length
            s.writeInt(_shared._size);

            // Write out all elements in the proper order.
            for (int i=0; i<_shared._size; i++)
                s.writeObject(_shared._elementData[i]);
        }

        /**
         * Reconstitute the <tt>COWList</tt> instance from a stream (that is,
         * deserialize it).  Note, if two "shared" lists are serialized, they
         * will not be "shared" when read back in.  THis is rare enough to be just fine.
         */
        private synchronized void readObject(java.io.ObjectInputStream s)
            throws java.io.IOException, ClassNotFoundException {
            // Read in _size, and any hidden stuff
            s.defaultReadObject();

            // Read in array length and allocate array
            int size = s.readInt();
            Object[] elementData = new Object[size];
            
            // Read in all elements in the proper order.
            for (int i=0; i<size; i++)
            {
                elementData[i] = s.readObject();
            }
            _shared = new Shared(elementData);
        }
    } // inner class COWList

    /* ***************  END  Inner Class COWList *********************** */
    /* ***************  END  Inner Class COWList *********************** */
} // ListFactory
