package com.sas.util;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Collection;
import java.util.HashMap;

/**
 * The MapWithDefaults class allows you to create a map with default values assigned. If values are added
 * to the map with the same key as a default value the new value will be returned. An example of when you might
 * use a MapWithDefaults is when you have a set of default values that will be used the majority of the time, but
 * the need to bee overriden occasionaly.  In this case, the defaultMap should be a static map that is used by
 * all instances of a component.  The internalMap holds any changes to the defaultMap.  When using the MapWithDefaults
 * the get(String Key) method will look in the internalMap first, if the key is found in the internalMap the given object
 * is return otherwise the defaultMap will be searched and if the key is found in the defaultMap, the object is returned from
 * the defaultMap.  This is where caution needs to be taken, if the object is returned from the defaultMap which is static,
 * the object should not be modified directly, if this is done the default object will be changed for all instances of the objects
 * that are using the MapWithDefaults.  To override the default object a new object needs to be created and then set on the
 * MapWithDefaults using the put(String key) method.
 * @see java.util.Map
 * @author Tony Prier
 **/

public class MapWithDefaults implements java.util.Map
{

   public static final String RB_KEY = "MapWithDefaults.";

   Map defaultMap;
   Map internalMap;

   /**
    * Default Constructor
    **/
   public MapWithDefaults()
   {
   }

   /**
    *  Constructor
    *  @param defaultMap Map containing default values. This map is not edited.
    *  @param internalMap Map that contains any overriden or new new values in the
    *  default map.
    **/

   public MapWithDefaults(Map defaultMap)
   {   this();
       this.defaultMap=defaultMap;
   }

   /**
    * Returns Map with default values.
    * @return map with default values.
    **/
   public Map getDefaultMap()
   {
	   return defaultMap;
   }

   /**
    * Sets the map with default values
    * @param defaultMap map with default values.
    */
   public void setDefaultMap(Map defaultMap)
   {
	   this.defaultMap=defaultMap;
   }

   /**
    * Returns map with overridden values.
    * @return map with overridden values.
    **/
   public Map getInternalMap()
   {
	   return acquireInternalMap();
   }

   /**
    * Sets the internal map that contains overridden values
    * @param internalMap map that contains overridden values
    **/
   public void setInternalMap(Map internalMap)
   {
	   this.internalMap=internalMap;
   }

   /**
    * Clears all changes made the the default map.
    **/

   public void clear()
   throws UnsupportedOperationException
   {
	   if (internalMap != null)
		   internalMap.clear();
   }

   /**
    * Returns true if the Key exists in the Map, false otherwise.
    * @param key the key that is being searched for.
    **/

   public boolean containsKey(Object key)
   throws ClassCastException, NullPointerException
   {
	  try
	  {
         if (defaultMap != null && defaultMap.containsKey(key))
            return true;
         else if (internalMap != null && internalMap.containsKey(key))
            return true;
         else
            return false;
  	  }
  	  catch(ClassCastException cce)
  	  {
		  throw new ClassCastException(RB.getStringResource(RB_KEY, "invalidType.txt"));
      }
      catch(NullPointerException npe)
      {
		  throw new NullPointerException(RB.getStringResource(RB_KEY, "nullKey.txt"));
      }
   }

   /**
    * Returns true if the Object exists in the Map, false otherwise.
    * @param key the key that is being searched for.
    **/

   public boolean containsValue(Object value)
   {
      if (defaultMap != null && defaultMap.containsValue(value))
         return true;
      else if (internalMap != null && internalMap.containsValue(value))
         return true;
      else
         return false;

   }

   /**
    * Returns a set of all objects in the Map.
    * @return set of all objects in the Map.
    **/

   public Set entrySet()
   {
      if (defaultMap != null && internalMap != null)
      {
         HashMap merged = new HashMap(defaultMap);
         merged.putAll(internalMap);
         return merged.entrySet();
      }
      else if (defaultMap==null && internalMap != null)
         return internalMap.entrySet();
      else if (defaultMap!=null && internalMap == null)
         return defaultMap.entrySet();
      else
         return new HashSet();
   }

   /**
    *  Returns the Object associated with the specified key.Returns null if the map contains no mapping for this key.
    *  A return value of null does not necessarily indicate that the map contains no mapping for the key; it's also possible that the map explicitly maps the key to null.
    *  The containsKey operation may be used to distinguish these two cases.
    *  @param key key whose associated value is to be returned.
    *  @return the value to which this map maps the specified key, or null if the map contains no mapping for this key.
    */

   public Object get(Object key)
   throws ClassCastException, NullPointerException
   {
	  try
      {
		  Object o=null;
		  if (internalMap != null)
			o = internalMap.get(key);
		  if (o==null && defaultMap != null)
			o=defaultMap.get(key);
		  return o;
  	  }
  	  catch(ClassCastException cce)
	  {
	    throw new ClassCastException(RB.getStringResource(RB_KEY, "invalidType.txt"));
	  }
	  catch(NullPointerException npe)
	  {
	    throw new NullPointerException(RB.getStringResource(RB_KEY, "nullKey.txt"));
      }
   }

   /**
    * Returns true if this map contains no key-value mappings.
    * @return true if this map contains no key-value mappings.
    **/

   public boolean isEmpty()
   {
	 if (defaultMap != null && !defaultMap.isEmpty())
        return false;
     else if (internalMap != null && !internalMap.isEmpty())
        return false;
     else if ((internalMap != null && defaultMap != null) &&
              (internalMap.isEmpty() && defaultMap.isEmpty()))
        return true;
     else
        return true;

   }

   /**
    * Returns a set view of the keys contained in this map. The set is backed by the map,
    * so changes to the map are reflected in the set, and vice-versa. If the map is modified
    * while an iteration over the set is in progress, the results of the iteration are undefined.
    * The set supports element removal, which removes the corresponding mapping from the map,
    * via the Iterator.remove, Set.remove, removeAll retainAll,
    * and clear operations. It does not support the add or addAll operations
    * @return a set view of the keys contained in this map.
    **/

   public Set keySet()
   {

      if (defaultMap != null && internalMap != null)
      {
         HashMap merged = new HashMap(defaultMap);
         merged.putAll(internalMap);
         return merged.keySet();
      }
      else if (defaultMap != null && internalMap == null)
         return defaultMap.keySet();
      else if (defaultMap== null && internalMap != null)
         return internalMap.keySet();
      return new HashSet();
   }

   /**
    * Returns the number of key-value mappings in this map. If the map contains
    * more than Integer.MAX_VALUE elements, returns Integer.MAX_VALUE.
    * @return the number of key-value mappings in this map.
    */

   public int size()
   {
      if (defaultMap != null && internalMap != null)
      {
         HashMap merged = new HashMap(defaultMap);
         merged.putAll(internalMap);
         return merged.size();
      }
      else if (defaultMap != null)
         return defaultMap.size();
      else if (internalMap !=null)
         return internalMap.size();
      else
         return 0;
   }

   /**
    * Returns a collection view of the values contained in this map.
    * @return a collection view of the values contained in this map.
    **/

   public Collection values()
   {
      if (defaultMap != null && internalMap != null)
      {
	 	HashMap merged = new HashMap(defaultMap);
	 	merged.putAll(this);
	 	return merged.values();
      }
      else if (defaultMap != null)
	     return defaultMap.values();
	  else if (internalMap !=null)
         return internalMap.values();
      else return new HashSet();
   }

   /**
    *  Compares the specified object with this map for equality. Returns true if
    *  the given object is also a map and the two Maps represent the same mappings.
    *  More formally, two maps t1 and t2 represent the same mappings if
    *  t1.entrySet().equals(t2.entrySet()).
    *  @param o object to be compared for equality with this map.
    *  @return true if the specified object is equal to this map.
    */


   public boolean equals(Object o)
   {

	if (!(o instanceof Map))
	   return false;

   	if (defaultMap != null && internalMap != null)
    {
   	   HashMap merged = new HashMap(defaultMap);
   	   merged.putAll(internalMap);
   	   return merged.entrySet().equals(((Map)o).entrySet());
    }
    else if (defaultMap != null)
    {
		return defaultMap.entrySet().equals(((Map)o).entrySet());
	}
	else if (internalMap != null)
	{
		return internalMap.entrySet().equals(((Map)o).entrySet());
	}
	else return false;

   }

   /**
   * Returns the hash code value for this map. The hash code of a map is defined to be
   * the sum of the hashCodes of each entry in the map's entrySet view. This ensures that
   * t1.equals(t2) implies that t1.hashCode()==t2.hashCode() for any two maps t1 and t2,
   * as required by the general contract of Object.hashCode
   * @return the hash code value for this map.
   **/

   public int hashCode()
   {
    if (defaultMap != null && internalMap != null)
	{
	   HashMap merged = new HashMap(defaultMap);
	   merged.putAll(internalMap);
	   return merged.hashCode();
	}
	else if (defaultMap != null)
	{
		return defaultMap.hashCode();
	}
	else if (internalMap != null)
	{
		return internalMap.hashCode();
	}
	else return 0;

   }

   public String toString()
   {
       if (defaultMap != null && internalMap != null)
	   {
	      HashMap merged = new HashMap(defaultMap);
	      merged.putAll(internalMap);
	      return merged.toString();
	   }
	   else if (defaultMap != null)
	   {
	    	return defaultMap.toString();
	   }
	   else if (internalMap != null)
	   {
	   		return internalMap.toString();
	   }
	   else return "";
   }

   /**
    * Associates the specified value with the specified key in this map (optional operation).
    * If the map previously contained a mapping for this key, the old value is replaced.
    * @param key key with which the specified value is to be associated.
    * @param value value to be associated with the specified key.
    * @return previous value associated with specified key, or null if there was no mapping
    * for key. A null return can also indicate that the map previously associated null with
    * the specified key, if the implementation supports null values.
    **/

   public Object put(Object key, Object value)
   throws UnsupportedOperationException, ClassCastException,
          IllegalArgumentException, NullPointerException

   {
	   return acquireInternalMap().put(key,value);
   }

   /**
    *  Copies all of the mappings from the specified map to this map (optional operation).
    *  These mappings will replace any mappings that this map had for any of the keys currently in the specified map
    *  @param t Mappings to be stored in this map.
    **/
   public void putAll(Map t)
   throws java.lang.UnsupportedOperationException, ClassCastException,
          IllegalArgumentException, NullPointerException
   {
	   acquireInternalMap().putAll(t);
   }

   /**
    * Removes the mapping for this key from this map if present (optional operation).
    * @param key key whose mapping is to be removed from the map.
    * @return previous value associated with specified key, or null if there was no mapping
    * for key. A null return can also indicate that the map previously associated null with
    * the specified key, if the implementation supports null values.
    **/

   public Object remove(Object o)
   throws java.lang.UnsupportedOperationException
   {
   	  if (internalMap != null)
         return internalMap.remove(o);

	   return null;
   }

   private Map acquireInternalMap()
   {
	   if ( internalMap == null ) {
		   internalMap = new HashMap(8,1);
	   }
	   return internalMap;
   }

}