// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.ParsePosition;
import java.util.Locale;

/**
 * NumberUtil is a utility class that performs
 * operations on numbers such as converting to and from strings.
 *
 * @author Jennifer Cole
 * @version 1.0
 */

public class NumberUtil
{

	public static final String RB_KEY = "NumberUtil.";

	private static NumberFormat numFormatter;		// used by toString
	private static NumberFormat numberFormatter;	// used by valueOf
	private static Locale lastLocale;				// used by toString
	private static Locale numberFormatterLocale;	// used by valueOf


	/**
	 * toString converts the passed in Double object to a String properly formatted for the current locale.
	 * @param value The Double value to convert to a String.
	 * @return A string representing the double value.
	 */
	public static String toString( Double value )
	{
		if ( value == null )
		{
			return "null";
		}

		return toString( value.doubleValue() );
	}

	/**
	 * toString converts the passed in double primitive to a String properly formatted for the current locale.
	 * @param value The double value to convert to a String.
	 * @return A string representing the double value.
	 */
	public static String toString( double d )
	{
		return (toString(d, Locale.getDefault()));
	}

	/**
	 * toString converts the passed in double primitive to a String properly formatted for the given locale.
	 * @param value The double value to convert to a String.
	 * @param locale The locale used for formatting the string
	 * @return A string representing the double value.
	 */
	public static String toString( double d, Locale locale )
	{
     Locale loc = locale;
     if (numFormatter == null || lastLocale != loc)
      {
        numFormatter = DecimalFormat.getInstance(loc);
        // System.out.println("detected locale change from " + lastLocale + " to " + loc);
        lastLocale = loc;
      }
      DecimalFormat formatter = null;
      if (numFormatter instanceof DecimalFormat)
         formatter = (DecimalFormat) numFormatter;
      if (Double.isNaN(d) || formatter == null)
        return numFormatter.format(d);

      // Inspect the format of the value.
      DecimalFormatSymbols symbols = formatter.getDecimalFormatSymbols();

      // Format the locale independent version, then change the '.' to decimal,
      // and position the sign
      StringBuffer buffer = new StringBuffer();
      buffer.append(d < 0 ? formatter.getNegativePrefix() : formatter.getPositivePrefix());
      String formatted = Double.toString(Math.abs(d));
      for (int i = 0, len = formatted.length(); i < len; i++)
        {
         char c = formatted.charAt(i);
         buffer.append((c == '.') ? symbols.getDecimalSeparator() : c);
        }
      buffer.append(d < 0 ? formatter.getNegativeSuffix() : formatter.getPositiveSuffix());
      return buffer.toString() /*I18nOK:LMS*/;
	}


	/**
	 * toString converts the passed in Float object to a String properly formatted for the current locale.
	 * @param value The Float value to convert to a String.
	 * @return A string representing the float value.
	 */
	public static String toString( Float value )
	{
		if ( value == null )
		{
			return "null";
		}

		return toString( value.floatValue() );
	}

	/**
	 * toString converts the passed in float primitive to a String properly formatted for the current locale.
	 * @param value The float value to convert to a String.
	 * @return A string representing the float value.
	 */
	public static String toString( float d )
	{
		return (toString(d, Locale.getDefault()));
	}

	/**
	 * toString converts the passed in float primitive to a String properly formatted for the given locale.
	 * @param value The float value to convert to a String.
	 * @param locale The locale used for formatting the string
	 * @return A string representing the float value.
	 */
	public static String toString( float d, Locale locale )
	{
     Locale loc = locale;
     if (numFormatter == null || lastLocale != loc)
      {
        numFormatter = DecimalFormat.getInstance(loc);
        // System.out.println("detected locale change from " + lastLocale + " to " + loc);
        lastLocale = loc;
      }
      DecimalFormat formatter = null;
      if (numFormatter instanceof DecimalFormat)
         formatter = (DecimalFormat) numFormatter;
      if (Float.isNaN(d) || formatter == null)
        return numFormatter.format(d);

      // Inspect the format of the value.
      DecimalFormatSymbols symbols = formatter.getDecimalFormatSymbols();

      // Format the locale independent version, then change the '.' to decimal,
      // and position the sign
      StringBuffer buffer = new StringBuffer();
      buffer.append(d < 0 ? formatter.getNegativePrefix() : formatter.getPositivePrefix());
      String formatted = Float.toString(Math.abs(d));
      for (int i = 0, len = formatted.length(); i < len; i++)
        {
         char c = formatted.charAt(i);
         buffer.append((c == '.') ? symbols.getDecimalSeparator() : c);
        }
      buffer.append(d < 0 ? formatter.getNegativeSuffix() : formatter.getPositiveSuffix());
      return buffer.toString() /*I18nOK:LMS*/;
	}

	/**
	 * valueOf converts the passed in (locale specific) String to a float.
	 * @param floatString A String to convert to a float.
	 * @returns The floating point value of the String.
	 */
	public static float floatValueOf( String floatString )
	{
		return (floatValueOf(floatString, Locale.getDefault()));
	}

	/**
	 * valueOf converts the passed in (locale specific) String to a float.
	 * @param floatString A String to convert to a float.
	 * @param locale The locale of the string
	 * @returns The floating point value of the String.
	 */
	public static float floatValueOf( String floatString, Locale locale )
	{
		Locale loc = locale;
		if ( numberFormatter == null  ||  !numberFormatterLocale.equals( loc ) )
		{
			numberFormatterLocale = loc;
			numberFormatter = NumberFormat.getInstance(numberFormatterLocale);
		}

		ParsePosition parsePosition = new ParsePosition( 0 );
		Number number = numberFormatter.parse( floatString.toUpperCase() /*I18nOK:LMS*/, parsePosition );
//		System.out.println( "pp = " + parsePosition.getIndex() + ", length=" + floatString.length() + "number=" + number );
		if ( parsePosition.getIndex() != 0  &&  number != null )
		{
			return number.floatValue();
		}

		String message = RB.getStringResource(RB_KEY, "floatValueOf.invalidString.ex.txt");
		throw new NumberFormatException( MessageFormat.format( message, new Object[] {floatString} ) );
	}

	/**
	 * valueOf converts the passed in (locale specific) String to a double.
	 * @param doubleString A String to convert to a double.
	 * @returns The double floating point value of the String.
	 */
	public static double doubleValueOf( String doubleString )
	{
		return (doubleValueOf(doubleString, Locale.getDefault()));
	}

	/**
	 * valueOf converts the passed in (locale specific) String to a double.
	 * @param doubleString A String to convert to a double.
	 * @param locale The locale of the string
	 * @returns The double floating point value of the String.
	 */
	public static double doubleValueOf( String doubleString, Locale locale )
	{
		Locale loc = locale;
		if ( numberFormatter == null  ||  !numberFormatterLocale.equals( loc ) )
		{
			numberFormatterLocale = loc;
			numberFormatter = NumberFormat.getInstance(numberFormatterLocale);
		}

		ParsePosition parsePosition = new ParsePosition( 0 );
		Number number = numberFormatter.parse( doubleString.toUpperCase() /*I18nOK:LMS*/, parsePosition );
		if ( parsePosition.getIndex() != 0  &&  number != null )
		{
			return number.doubleValue();
		}

		String message = RB.getStringResource(RB_KEY, "doubleValueOf.invalidString.ex.txt");
		throw new NumberFormatException( MessageFormat.format( message, new Object[] {doubleString} ) );
	}
}

