/* $Id$ */
// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import com.sas.text.Message;
import com.sas.util.NumberUtil;
import com.sas.util.transforms.TransformInterface;
import com.sas.util.transforms.TransformException;

import java.util.Locale;
import java.util.Map;
import java.util.HashMap;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.util.Hashtable;

/**
 * <code>RBBase</code> is an abstract base class for all <code>RB</code> classes
 * in various packages. The <code>RB</code> in each package classes provide access to
 * a shared resource bundle for that package.
 * @author sasdjb <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */
public class RBBase
{
	public static final String RB_KEY = "RBBase.";

  private ResourceBundle resourceBundle; // lazy instantiation
  private Hashtable localeBundles; // lazy instantiation
  private static final String DEFAULT_RESOURCE_BUNDLE_NAME = "Resources" /*I18nOK:EMS*/;
  private String resourceBundleName;

  protected RBBase()
    {
      this(DEFAULT_RESOURCE_BUNDLE_NAME);
    }

  protected RBBase(String resourceBundleName)
    {
      this.resourceBundleName = resourceBundleName;
    }

  // Utility

  private String resourceName()
  {
    String className = getClass().getName();
    int dot = className.lastIndexOf('.');
    return (  (dot == -1)
	    ? resourceBundleName
	    : (className.substring(0, dot) + "." + resourceBundleName) );
  }

  /**
   * Return a string resource from the default resource bundle.
   * @param key a resource key
   * @return String a string resource obtained from the key, namely
   * <code>_getResources().getString(key)</code>
   * @exception java.util.MissingResourceException if there is no resource
   * for the key.
   */
  protected final String _getStringResource(String key)
  {
    return _getStringResource(key, Locale.getDefault());
  }

  /**
   * Return a string resource from the default resource bundle.
   * @param key a resource key
   * @return String a string resource obtained from the key, namely
   * <code>_getResources().getString(key)</code>
   * @exception java.util.MissingResourceException if there is no resource
   * for the key.
   */
  protected final String _getStringResource(String key, Locale locale)
  {
    try
      {
        return _getResources(locale).getString(key);
      }
    catch (MissingResourceException r)
      {
//        System.out.println("Missing resource " + key + " for " + resourceName());
        throw r;
      }
    catch (Exception e)
      {
//        System.out.println("Caught exception trying to load resource " + key);
        throw new MissingResourceException(e.getMessage(), resourceName(), key);
      }
  }

   /**
    * Get the default resource bundle for this package, which is defined
    * in the properties file, <code>"<em>&lt;package-path&gt;</em>/Resources.properties"</code>.
    * If a locale-specific
    * version of the properties file exists, such as
    * <code>"<em>&lt;package-path&gt;</em>/Resources_de_DE.properties"</code>,
    * then it and its base resources will be loaded and concatenated.
    * @return the default resource bundle.
    */
   protected final ResourceBundle _getResources()
   {
     return _getResources(Locale.getDefault());
   }

   protected final ResourceBundle _getResources(Locale locale)
   {
     try
       {
		 if ( locale == null )
			   locale = Locale.getDefault();

		 if ( locale.equals(Locale.getDefault()) )
           {
			 synchronized(this)
			 {
	             if (resourceBundle == null)
	               // Must use the current RB object's class loader since
	               // the RB class may be loaded from the application class loader
	               // (and thus it's Resources.properties only available from
	               // that class loader), whereas RBBase is in webAF.jar
	               // in the extensions class loader. If we did not pass
	               // a class loader, ResourceBundle would use the extensions
	               // class loader and not be able to find the Resources.properties
	               // file or the Resources.class resource bundle class.
	               resourceBundle = ResourceBundle.getBundle(resourceName(),
	                                                         locale,
	                                                         this.getClass().getClassLoader());
			 }
             return resourceBundle;
           }
         else
           {
             ResourceBundle localeBundle = null;
             if (localeBundles == null)
               localeBundles = new Hashtable();
             else
               localeBundle = (ResourceBundle) localeBundles.get(locale);
             if (localeBundle != null)
               return localeBundle;

			 synchronized(this)
			 {
	             // see note above about using the correct class loader
	             localeBundle = ResourceBundle.getBundle(resourceName(),
	                                                     locale,
	                                                     this.getClass().getClassLoader());
	             localeBundles.put(locale, localeBundle);
			 }
             return localeBundle;
           }
       }
     catch (MissingResourceException r)
       {
		 // NOTE!: (SPL) Do not load the message from the resource bundle since that just failed and
		 // may well lead to an infinite loop (see S0249956)
		 System.out.println( "Missing resource bundle " + resourceName() + " for locale " + locale ); /*I18nOK:LINE*/
         throw r;
       }
   }

  // Convenience methods for converting resources to primitive types

  public boolean getResourceAsBoolean( String resourceName )
  {
	  return getResourceAsBoolean( this, resourceName, Locale.getDefault() );
  }

  public boolean getResourceAsBoolean( String resourceName, Locale locale )
  {
	  return getResourceAsBoolean( this, resourceName, locale );
  }

  public byte getResourceAsByte( String resourceName )
  {
	  return getResourceAsByte( this, resourceName, Locale.getDefault() );
  }

  public byte getResourceAsByte( String resourceName, Locale locale )
  {
	  return getResourceAsByte( this, resourceName, locale );
  }

  public char getResourceAsChar( String resourceName )
  {
	  return getResourceAsChar( this, resourceName, Locale.getDefault() );
  }

  public char getResourceAsChar( String resourceName, Locale locale )
  {
	  return getResourceAsChar( this, resourceName, locale );
  }

  public double getResourceAsDouble( String resourceName )
  {
	  return getResourceAsDouble( this, resourceName, Locale.getDefault() );
  }

  public double getResourceAsDouble( String resourceName, Locale locale )
  {
	  return getResourceAsDouble( this, resourceName, locale );
  }

  public float getResourceAsFloat( String resourceName )
  {
	  return getResourceAsFloat( this, resourceName, Locale.getDefault() );
  }

  public float getResourceAsFloat( String resourceName, Locale locale )
  {
	  return getResourceAsFloat( this, resourceName, locale );
  }

  public int getResourceAsInt( String resourceName )
  {
	  return getResourceAsInt( this, resourceName, Locale.getDefault() );
  }

  public int getResourceAsInt( String resourceName, Locale locale )
  {
	  return getResourceAsInt( this, resourceName, locale );
  }

  public long getResourceAsLong( String resourceName )
  {
	  return getResourceAsLong( this, resourceName, Locale.getDefault() );
  }

  public long getResourceAsLong( String resourceName, Locale locale )
  {
	  return getResourceAsLong( this, resourceName, locale );
  }

  public short getResourceAsShort( String resourceName )
  {
	  return getResourceAsShort( this, resourceName, Locale.getDefault() );
  }

  public short getResourceAsShort( String resourceName, Locale locale )
  {
	  return getResourceAsShort( this, resourceName, locale );
  }

  public Object getResourceAsObject( String resourceName, Class targetClass )
  {
	  return getResourceAsObject( this, resourceName, targetClass, Locale.getDefault() );
  }

  public Object getResourceAsObject( String resourceName, Class targetClass, Locale locale )
  {
	  return getResourceAsObject( this, resourceName, targetClass, locale );
  }

  public Object getResourceAsObject( String resourceName, TransformInterface transform )
  {
	  return getResourceAsObject( this, resourceName, transform, Locale.getDefault() );
  }

  public Object getResourceAsObject( String resourceName, TransformInterface transform, Locale locale )
  {
	  return getResourceAsObject( this, resourceName, transform, locale );
  }


  // static Convenience methods for converting resources to primitive types

	public static boolean getResourceAsBoolean( RBBase rb, String resourceName )
	{
      return getResourceAsBoolean( rb, resourceName, Locale.getDefault() );
	}

	public static boolean getResourceAsBoolean( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		return Boolean.valueOf( stringValue ).booleanValue();
	}

	public static byte getResourceAsByte( RBBase rb, String resourceName )
	{
      return getResourceAsByte( rb, resourceName, Locale.getDefault() );
	}

	public static byte getResourceAsByte( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		return Byte.parseByte( stringValue );
	}

	public static char getResourceAsChar( RBBase rb, String resourceName )
	{
      return getResourceAsChar( rb, resourceName, Locale.getDefault() );
	}

	public static char getResourceAsChar( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		if ( stringValue.length() >= 1 )
		{
			return stringValue.charAt(0);
		}

		// To-Do(SPL): Should this be an exception???
		return (char)0;
	}

	public static double getResourceAsDouble( RBBase rb, String resourceName )
	{
      return getResourceAsDouble( rb, resourceName, Locale.getDefault() );
	}

	public static double getResourceAsDouble( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		return NumberUtil.doubleValueOf( stringValue );
	}

	public static float getResourceAsFloat( RBBase rb, String resourceName )
	{
      return getResourceAsFloat( rb, resourceName, Locale.getDefault() );
	}

	public static float getResourceAsFloat( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		return NumberUtil.floatValueOf( stringValue );
	}

	public static int getResourceAsInt( RBBase rb, String resourceName )
	{
      return getResourceAsInt( rb, resourceName, Locale.getDefault() );
	}

	public static int getResourceAsInt( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		return Integer.valueOf( stringValue ).intValue();
	}

	public static long getResourceAsLong( RBBase rb, String resourceName )
	{
      return getResourceAsLong( rb, resourceName, Locale.getDefault() );
	}

	public static long getResourceAsLong( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		return Long.valueOf( stringValue ).longValue();
	}

	public static short getResourceAsShort( RBBase rb, String resourceName )
	{
      return getResourceAsShort( rb, resourceName, Locale.getDefault() );
	}

	public static short getResourceAsShort( RBBase rb, String resourceName, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );
		return Short.valueOf( stringValue ).shortValue();
	}

	public static Object getResourceAsObject( RBBase rb, String resourceName, Class targetClass )
	{
      return getResourceAsObject( rb, resourceName, targetClass, Locale.getDefault() );
	}

	public static Object getResourceAsObject( RBBase rb, String resourceName, Class targetClass, Locale locale )
	{
		// Lookup the transform
		TransformInterface transform = getStringTransform( targetClass );

		// Let the other getResourceAsObject do the resource load and transform
		return getResourceAsObject( rb, resourceName, transform, locale );
	}

	public static Object getResourceAsObject( RBBase rb, String resourceName, TransformInterface transform )
	{
      return getResourceAsObject( rb, resourceName, transform, Locale.getDefault());
    }

	public static Object getResourceAsObject( RBBase rb, String resourceName, TransformInterface transform, Locale locale )
	{
		String stringValue = rb._getStringResource( resourceName, locale );

		// Run the transform
		Object transformedValue = stringValue;
		if ( transform != null )
		{
			try
			{
				transformedValue = transform.transform( transformedValue );
			}
			catch (TransformException te)
			{
				te.printStackTrace();
				// XXX: How to handle exception???
			}
		}
		return transformedValue;
	}

	private static Map transformMap;
	{
		transformMap = new HashMap();
		transformMap.put( "java.lang.Boolean", com.sas.util.transforms.StringToBooleanTransform.defaultInstance );
		transformMap.put( "java.lang.Byte", com.sas.util.transforms.StringToByteTransform.defaultInstance );
		transformMap.put( "java.lang.Double", com.sas.util.transforms.StringToDoubleTransform.defaultInstance );
		transformMap.put( "java.lang.Float", com.sas.util.transforms.StringToFloatTransform.defaultInstance );
		transformMap.put( "java.lang.Integer", com.sas.util.transforms.StringToIntegerTransform.defaultInstance );
		transformMap.put( "java.lang.Short", com.sas.util.transforms.StringToShortTransform.defaultInstance );
	}

	private static TransformInterface getStringTransform( Class outputClass )
	{
		return (TransformInterface) transformMap.get( outputClass.getName() );
	}

}
