/* $Id$ */
package com.sas.util;

import java.util.List;

/**
 * Defines a role, where a role is a set of items that have 
 * been grouped together for a common purpose, such as a set of columns in a 
 * table.
 * 
 * @see RoleProviderInterface
 */
public interface RoleInterface
   {

   /**
    *  Indicates that a role is optional, that is that it does not have to have 
    *  any items assigned to it.
    *  @see #getMinRoleItemCount
    */
   public static final int NO_MINIMUM = 0;

   /**
    *  Indicates that there is no limit to the number of items that may be
    *  assigned to a role.
    *  @see #getMaxRoleItemCount
    */
   public static final int NO_MAXIMUM = 0;

   /**
    *  Returns the role's name.
    *  @return the name of the role
    *  @see #getRoleLabel
    */
   public String getRoleName ();

   /**
    *  Returns the role's label.
    *  The label is descriptive text that can be used, for example, to 
    *  identify the role in a customizer window.
    *  <p>
    *  Note: implementations should return labels as localized strings.
    *  @return the label for the role
    *  @see #getRoleName
    *  @see #getRoleDescription
    */
    public String getRoleLabel ();

    /**
     *  Returns the role's detailed label.
     *  The detailed label is descriptive text that can be used, for example, to 
     *  identify the role in a customizer window. This text can be more descriptive
     *  than the label itself.
     *  <p>
     *  Note: implementations should return labels as localized strings.
     *  @return the detailed label for the role
     *  @see #getRoleName
     *  @see #getRoleLabel
     *  @see #getRoleDescription
     */
     public String getRoleDetailedLabel();

   /**
    *  Returns the role's description.
    *  The description is a longer or more detailed version of the label that
    *  is appropriate to use for accessibility or tooltip text.
    *  <p>
    *  Note: implementations should return descriptions as localized strings.
    *  @return the description for the role
    *  @see #getRoleLabel
    */
    public String getRoleDescription ();

   /** 
    *  Returns the minimum number of items that must be assigned to the role.
    *  @return the required minimum or <code>NO_MINIMUM</code>
    *  @see #NO_MINIMUM
    *  @see #getMaxRoleItemCount
    */
    public int getMinRoleItemCount ();
  
    /**
     *  Returns the maximum number of items that can be assigned to the role.
     *  @return the maximum or <code>NO_MAXIMUM</code>
     *  @see #NO_MAXIMUM
     *  @see #getMinRoleItemCount
     */
    public int getMaxRoleItemCount ();  

    /**
     *  Returns all the items that are candidates for assignment to the role.
     *  This list should only include an item that is already assigned if it 
     *  makes sense to assign the item multiple times.
     *  @return the available items
     *  @see #getRoleItems
     */
    public List getAvailableRoleItems ();

    /**
     *  Returns the items that have been assigned to the role.
     *  @return the assigned items
     *  @see #getAvailableRoleItems
     */
    public List getRoleItems ();

    /**
     *  Assigns the given item to the role.
     *  Should not fail if caller has honored the information provided by
     *  <code>getAvailableRoleItems()</code>, 
     *  <code>getMinRoleItemCount()</code>, and
     *  <code>getMaxRoleItemCount()</code>.
     *  After calling caller should recall these other methods since assigning 
     *  one item may make another invalid for assignment and/or otherwise change 
     *  the assignment rules.
     *  @param item  the item to assign
     *  @see #addRoleItems 
     *  @see #removeRoleItem
     */
    public void addRoleItem (Object item);

    /**
     *  Assigns the given items to the role.
     *  <p>This is a convenience method for consecutive calls to 
     *  <code>addRoleItem()</code>, but saves the caller the trouble of requerying
     *  <code>getAvailableRoleItems()</code>, 
     *  <code>getMinRoleItemCount()</code>, and
     *  <code>getMaxRoleItemCount()</code> and validating the returned results
     *  against the remaining items to be added.
     *  @param items  the items to assign
     *  @return <code>null</code> if all items added, otherwise a list of
     *  the items that could not be added
     *  @see #addRoleItem
     */
    public List addRoleItems (List items);

    /**
     *  Unassigns the given item from the role.
     *  After calling caller should recall
     *  <code>getAvailableRoleItems()</code>, 
     *  <code>getMinRoleItemCount()</code>, and
     *  <code>getMaxRoleItemCount()</code>
     *  since unassigning 
     *  one item may make another valid for assignment and/or otherwise change 
     *  the assignment rules.
     *  @param item  the item to unassign
     *  @see #addRoleItem 
     *  @see #removeAllRoleItems
     */
    public void removeRoleItem (Object item);

    /**
     *  Unassigns all items from the role.
     *  After calling caller should recall
     *  <code>getAvailableRoleItems()</code>, 
     *  <code>getMinRoleItemCount()</code>, and
     *  <code>getMaxRoleItemCount()</code>
     *  since unassigning 
     *  one item may make another valid for assignment and/or otherwise change 
     *  the assignment rules.
     *  @see #removeRoleItem
     */
    public void removeAllRoleItems ();

	/**
	 *  Moves the item in the list of selected items
	 * 
	 *  @param item The item in the selected items list to move
	 *  @param index The number of positions in the list to move the item.  This 
	 *  value can be positive or negative.
	 */
	public void moveRoleItem(Object item, int index);

	/**
	 * Returns the name of the role's opened icon.
	 * @return The name of the icon representing the open state.
     * @see #getRoleClosedIcon
     * @see #getRoleDisabledIcon
	 */
	public String getRoleOpenedIcon();

	/**
	 * Returns the name of the role's closed icon.
	 * @return The name of the icon representing the closed state.
     * @see #getRoleOpenedIcon
     * @see #getRoleDisabledIcon
	 */
	public String getRoleClosedIcon();

	/**
	 * Returns the name of the role's disabled icon.
	 * @return The name of the icon representing the disabled state.
     * @see #getRoleOpenedIcon
     * @see #getRoleClosedIcon
	 */
	public String getRoleDisabledIcon();

    }