/* Copyright (c)1999 SAS Institute Inc, Cary NC USA 27513
 * Unpublished- All Rights Reserved.
 */
package com.sas.util;
import java.io.UnsupportedEncodingException;
//
// History:
//
// Date    Name            Remarks
// 24Aug99 G Williams      Created.
// 27Feb02 B Beers         S0141718: Move to sas.core.jar.
// 27Jan03 B Beers         S0180119: Add sas001.
//
/**
 * This class has methods to encode and decode a sensitive string such 
 * as a password. Encoding a string
 * makes it less likely that the value of the string will be accidentally
 * revealed when it is, for example, transmitted across a public network
 * or stored in a file or repository. However, it should not be
 * considered adequate protection against malicious attempts to
 * discover the value of the string. Secure networks, file systems, 
 * repositories, etc. are necessary to guard against malicious attacks.
 * <P>
 * Encoded strings are prefixed with
 * <CODE>{<EM>tag</EM>}</CODE> where <EM>tag</EM> identifies the encoding
 * algorithm.
 * @author Gary Williams
 */
public class SasPasswordString
{
    /**
     * base64 encoding.
     * @see com.sas.util.BASE64
     */
    public static final String BASE64_ENCODING = "base64" /*I18nOK:EMS*/;

    /**
     * a proprietary SAS encoding.
     */
    public static final String SAS001_ENCODING = "sas001" /*I18nOK:EMS*/;

    /**
     * Decode a string. If the string was previously encoded by the
     * <CODE>encode</CODE> method on this class, then this method
     * will decode it. If the string was not previously encoded, then
     * it will be returned unchanged.
     * @param instring the string to decode
     * @return the decoded string
     * @throws com.sas.util.SasPasswordEncodingException if decoding fails.
     * @throws java.lang.NullPointerException if <CODE>instring</CODE>
     * is <CODE>null</CODE>.
     */
    public static String decode(String instring) throws SasPasswordEncodingException
    {
        // Find the encoding algorithm. If one is not found, return the
        // string as is.
        if (!instring.startsWith("{"))
        {
            return instring;
        }

        int ndx = instring.indexOf("}");

        if (ndx < 0)
        {
            return instring;
        }

        String tag = instring.substring(1,ndx);

        if (SAS001_ENCODING.equals(tag) || BASE64_ENCODING.equals(tag))
        {
            if (instring.length() < ndx + 1)
            {
                return "";
            }

            String data = instring.substring(ndx+1);
            byte[] utf8 = BASE64.decode(data);

            if (utf8 == null)
            {
                return instring;
            }

            try
            {
                String rs = new String(utf8,"UTF8");
                return rs;
            }
            catch (UnsupportedEncodingException ex)
            {
                SasPasswordEncodingException spee = new SasPasswordEncodingException(ex);
                try {spee.initCause(ex);}catch(IllegalStateException ise){}
                throw spee;
            }
        }

        return instring;
    }

    /**
     * Encode a string.
     * @param encoding the encoding to use
     * @param instring the string to encode
     * @return the encoded string
     * @throws com.sas.util.SasPasswordEncodingException if the specified encoding is
     * unknown or if the encoding process fails
     * @throws java.lang.NullPointerException if <CODE>instring</CODE>
     * is <CODE>null</CODE>.
     * @see #BASE64_ENCODING
     * @see #SAS001_ENCODING
     */
    public static String encode(String encoding,String instring) throws SasPasswordEncodingException
    {
        if (SAS001_ENCODING.equals(encoding) || BASE64_ENCODING.equals(encoding))
        {
            int instringLength = instring.length();
            int bufferLength = 2 + encoding.length() + ((instringLength / 3) * 4);
            StringBuffer buffer = new StringBuffer(bufferLength);
            buffer.append("{").append(encoding).append("}");

            try
            {
                if (instringLength > 0)
                {
                    byte[] utf8 = instring.getBytes("UTF8");
                    String base64 = BASE64.encode(utf8);
                    buffer.append(base64);
                }
            }
            catch (UnsupportedEncodingException ex)
            {
                SasPasswordEncodingException spee = new SasPasswordEncodingException(ex);
                try {spee.initCause(ex);}catch(IllegalStateException ise){}
                throw spee;
            }

            String code = buffer.toString() /*I18nOK:LMS*/;
            return code;
        }
        else
        {
            throw new SasPasswordEncodingException(encoding);
        }
    }
}