/* $Id: SimpleTemplate2.java,v 1.2.2.1.4.2 2004/12/22 19:53:11 chmast Exp $ */
// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import com.sas.GlobalInfoTable;
import com.sas.beans.ExtendedBeanInfo;
import com.sas.util.Strings;

import java.io.PrintWriter;
import java.text.Format;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.Collections;

/**
 * @author Scott Leslie
 * @see com.sas.util.SimpleTemplate
 * @version	1.0
 */

/**
 * The SimpleTemplate2 class serves as a generic string substitution and formatting
 * class.  A raw text string, referred to as the template, can contain variables,
 * references, text, and formatting directives.
 * <p>
 * The SimpleTemplate2 class also uses a dictionary that contains variables in the template
 * along with their associated values. Formatting the template replaces keys from the
 * dictionary with their corresponding values.	This dictionary is a collection object
 * that implements the java.util.Map interface. The values from this dictionary are
 * evaluated when looked up during the Template pattern evaluation by only calling
 * toString on the value dictionary objects once and caching the result
 * for use in all occurences of that dictionary object in the pattern.
 * <p>
 * The caching of value dictionary objects' toString results is enabled by default.
 * If toStringCaching is disabled then the value objects used in the dictionary will
 * be evaluated with a toString call once for each occurence in the pattern.
 * The toString cache is cleared whenever a format(), setValueDictionary() or
 * the clearToStringCache() mehod is called.
 * <p>
 * If the toString cache should NOT be reset upon the start of each format, then
 * setClearToStringCacheOnFormat(false).  When this is set to false, the cache will
 * only clear when a setValueDictionary() or clearToStringCache() is done.
 * <p>
 * The general form of a template is:
 * <pre>
 *   text %VAR(FLD1, FLD2, FLD3) more text
 * <pre>
 * where:
 * <p>
 *  VAR is the name of a variable that is defined in the dictionary
 *  supplied to the SimpleTemplate2. Variable names must begin with %. The next character
 *  must be a valid identifier start character as defined
 *  by {@link java.lang.Character#isJavaIdentifierStart(char ch)}.
 *  All subsequent characters must be valid identifier characters as defined by
 *  {@link java.lang.Character#isJavaIdentifierPart(char ch)}.
 *  Variable names are case sensitive.
 *	<p>
 *  NOTE:  The '%' is not part of the variable name (it acts as a delimiter). In order to specify a value
 *  for the variable, %PROC, one would use the following syntax where the dictionary
 *  key is specified as 'PROC' not '%PROC'. The '%' character may be changed by calling
 *  the {@link #setMarker(char)} method, or you can used the constructor which
 *  specifies the marker character.
 *
 * <ul>
 * <li> dictionary.set("PROC", "SORT");
 * </ul>
 * <p>
 * <p>
 *   FLD1 is an optional field that will be evaluated and displayed if VAR is
 *   defined in the dictionary and is not null. If VAR is not defined,
 *   FLD1 will be not be displayed. FLD1 can be designated as a combination of
 *   text and variables.
 * <p>
 *   FLD2 is an optional field that contains a default value that should be used
 *   if VAR is not defined in the dictionary or is null. FLD2 can be designated
 *   as a combination of text and variables.
 * <p>
 *   FLD3 is an optional field that designates a class that will perform special
 *   formatting. The class specified as FLD3 must extend java.text.Format.
 *   The class will be instantiated and the <code>format(Object)</code> method will be invoked with
 *   the value of FLD1 (if VAR is defined) or FLD2 (if VAR is not defined or is null).
 * <p>
 * <p>
 * NOTES:
 * <p>
 * Commas are used for field delineation. If you need to use a comma as part of
 * a field expression (FLD1,FLD2,FLD3), that field must be surrounded by quotes.
 * <p>
 * % is used for variable delineation. %% will be used to denote a % that is not
 * used for a variable name. If you change the marker character with   {@link #setMarker(char)},
 * you must double it to see it rendered. For example, if you call <code>setMarker('$')</code>
 * then you would use a template such as <code>"The price is $$$price"</code>, with
 * price = 10.75 in the dictionary, to yield <code>"$10.75"</code>. The check for double-delimiters
 * (e.g. %%) occurs in the template string only, so any substituted variable values will not be
 * modified as part of this replacement process.
 * <p>
 * The following are example of templates:
 * <p>
 * EXAMPLE 1:
 * <pre>
 *     "The woman is %status"
 * </pre>
 * If status = "very wealthy.", the expression will be processed to:
 * <pre>
 *     "The woman is very wealthy."
 * </pre>
 * If status is not defined, the expression will be processed to:
 * <pre>
 *     "The woman is"
 * </pre>
 * EXAMPLE 2:
 * <pre>
 *     "Today it is %DAY(%DAY, raining)."
 * </pre>
 * If the variable DAY is contained in the dictionary and set to "sunny", the
 * string will be processed to:
 * <pre>
 *     "Today it is sunny."
 * </pre>
 * If the variable DAY is not contained in the dictionary or is null, the string
 * will be processed to:
 * <pre>
 *     "Today it is raining."
 * </pre>
 * EXAMPLE 3:
 * <pre>
 *     "Hello. %person(%GREET %person?)"
 * </pre>
 * If person = "Mary" and GREET = "How are you", the expression will be
 * processed to:
 * <pre>
 *     "Hello. How are you Mary?"
 * </pre>
 * If person is not defined in the dictionary, the expression will be
 * processed to:
 * <pre>
 *     "Hello."
 * </pre>
 * If person = "Mary" and GREET is not defined, the expression will be
 * processed to:
 * <pre>
 *     "Hello. Mary?"
 * </pre>
 * <p>
 * EXAMPLE 4:
 * <pre>
 * %state(, is not defined)
 * </pre>
 * If state is defined in the dictionary and is not null, the expression
 * will be processed to:
 * <pre>
 *     ""  (string of zero length)
 * </pre>
 * If state is not defined in the dictionary the expression will be processed to:
 * <pre>
 *     "is not defined"
 * </pre>
 * Note that the comma is necessary for correct field delineation.
 * <p>
 * EXAMPLE 5:
 * <pre>
 *     "The stock market is %market(%direction, closed)."
 * </pre>
 * If market is defined and direction="up today" the expression will be processed to:
 * <pre>
 *     "The stock market is up today."
 * </pre>
 * As this example illustrates, the VAR field (market) does not have to appear in the FLD1
 * expression (direction). The VAR field acts first as a true/false operator that
 * dictates whether to evaluate FLD1 or FLD2 (FLD1 if VAR is in dictionary and not null,
 * FLD2 otherwise). If VAR appears in the expression to be evaluated, its value is
 * substituted in the expression.
 * <p>
 * EXAMPLE 6:
 * <pre>
 *     %test(%dvar,,com.sas.me.MyClass)
 * </pre>
 * An instance of com.sas.me.MyClass will be instantiated. If the class cannot be
 * instantiated, an exception will be thrown. If com.sas.me.MyClass extends
 * java.text.Format, the format(Object) method will be invoked with the value of dvar
 * (if test is not null) or a "" (if test is null). If com.sas.myClass does not
 * extend java.text.Format, the expression will not be evaluated
 * any further and will not appear in the formatted string. Note that the second
 * comma is necessary for correct field delineation.
 * <p>
 * The special characters '%', '(', and ')' can be replaced with other
 * characters on the SimpleTemplate2 constructor, or by setting the
 * {@link #setMarker(char) <var>marker</var>},
 * {@link #setOpen(char) <var>open</var>},
 * {@link #setClose(char) <var>close</var>}
 * properties. For example,
 * <pre>
 *  template.setMarker('$');
 *  template.setOpen('{');
 *  template.setClose('}');
 *  template.setTemplate("The stock market is $market{$direction, closed}.");
 * </pre>
 */

public class SimpleTemplate2 extends Object
{
//	public static final String RB_KEY = "SimpleTemplate."; //$NON-NLS-1$

	// Used to allocate extra space in the StringBuffer so that replaces
	// of the double delimiter case don't cause reallocations.
	protected static int DELIMITER_REPLACE_PAD_SIZE = 20;

	// Variable used for the submitString attribute
	protected String submit_string = "";

	// Variable used for the templateString attribute
	protected String template_string = "";

	// Variable used for tokenizing the template
	protected ArrayList text_and_token_string_array;
	protected ArrayList temp_string_array;

	protected boolean dirty = true;

	// Variable used for tokenizing the template
	protected String vdelimiter; // = "%";
	protected char openc; // = '(';
	protected char closec; // = ')';

	transient protected String openStr; // = new String( new char[] { openc } );
	transient protected String closeStr; // = new String( new char[] { closec } );

	// Variable used for the value to substitute (must implement Map interface)
	public Map dmapValues = null;

	// Variable used to enable class evaluation
	protected boolean allowClassLoading = false;

	// toString cache
	protected Map _toStringCache = new HashMap();

	// enabled means toStrings of objects will be cached
	protected boolean _toStringCachingEnabled = true;

	// enabled means format will clear toString cache before evaluation
	protected boolean _toStringCacheClearedOnFormat = true;

	// output stream to which tokenized values are sent.
	protected PrintWriter writer;

	// ------------------------------------------------------------------
	// CONSTRUCTORS
	// ------------------------------------------------------------------

	/**
	* Default Constructor
	*/
	public SimpleTemplate2() {
		init("",null,'%','(',')',false); /*I18nOK:EMS*/
	}

	/**
	* Construct a SimpleTemplate object with the specified template
	* @param pattern specified template
	*/

	public SimpleTemplate2(String pattern) {
		init(pattern,null,'%','(',')',false); /*I18nOK:EMS*/
	}

	/**
	* Construct a SimpleTemplate2 with the specified template and object that implements
	* the java.util.Map interface.
	* @param pattern specified template
	* @param dictionary collection object that implements java.util.Map interface
	*/

	public SimpleTemplate2(String pattern, Map dictionary)
	{
		init(pattern,dictionary,'%','(',')',false); /*I18nOK:EMS*/
	}

	/**
	* Construct a SimpleTemplate2 with the specified template and object that implements
	* the Map Interface.
	* @param pattern specified template
	* @param dictionary colleciton object that implements java.util.Map interface
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	*/

	public SimpleTemplate2(String pattern,
	              Map dictionary,
	              char marker,
	              char open,
	              char close)
	{
		init(pattern,dictionary,marker,open,close,false);
	}

	/**
	* Construct a SimpleTemplate2 with the specified template.
	* @param pattern specified template
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	*/

	public SimpleTemplate2(String pattern,
	              char marker,
	              char open,
	              char close)
	{
		init(pattern,null,marker,open,close,false);
	}

	/**
	* Construct a SimpleTemplate2 with the object that implements the Map interface.
	* @param dictionary collection object that implements java.util.Map interface
	*/

	public SimpleTemplate2(Map dictionary)
	{
		init("",dictionary,'%','(',')',false); /*I18nOK:EMS*/
	}

	/**
	* Construct a SimpleTemplate2 object with the specified template
	* @param allowClassLoading enable or disable the ability to load classes via the pattern
	*/

	public SimpleTemplate2(boolean allowClassLoading) {
		init("",null,'%','(',')',allowClassLoading); /*I18nOK:EMS*/
	}

	/**
	* Construct a SimpleTemplate2 object with the specified template
	* @param pattern specified template
	* @param allowClassLoading enable or disable the ability to load classes via the pattern
	*/

	public SimpleTemplate2(String pattern, boolean allowClassLoading ) {
		init(pattern,null,'%','(',')',allowClassLoading); /*I18nOK:EMS*/
	}

	/**
	* Construct a SimpleTemplate2 with the specified template and object that implements the Map Interface.
	* @param pattern specified template
	* @param dictionary object that implements java.util.Map interface
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	* @param allowClassLoading enable or disable the ability to load classes via the pattern
	*/

	public SimpleTemplate2(String pattern,
	              Map dictionary,
	              char marker,
	              char open,
	              char close,
		          boolean allowClassLoading)
	{
		init(pattern,dictionary,marker,open,close,allowClassLoading);
	}

	/**
	* Construct a SimpleTemplate2 with the specified template.
	* @param pattern specified template
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	* @param allowClassLoading enable or disable the ability to load classes via the pattern
	*/

	public SimpleTemplate2(String pattern,
	              char marker,
	              char open,
	              char close,
		          boolean allowClassLoading)
	{
		init(pattern,null,marker,open,close,allowClassLoading);
	}

	/**
	* Construct a SimpleTemplate2 with the object that implements the Map Interface.
	* @param dictionary object that implements java.util.Map interface
	* @param allowClassLoading enable or disable the ability to load classes via the pattern
	*/

	public SimpleTemplate2(Map dictionary, boolean allowClassLoading )
	{
		init("",dictionary,'%','(',')',allowClassLoading); /*I18nOK:EMS*/
	}

	// ------------------------------------------------------------------
	// MISC METHODS
	// ------------------------------------------------------------------

	protected void init( String pattern,
						Map dictionary,
						char marker,
						char open,
						char close,
						boolean enableClassLoading )
	{
		// Handle the original parameters and use null for the textAndTemplate
		init( pattern, dictionary, marker, open, close, enableClassLoading, null );
	}

	protected void init( String pattern,
						Map dictionary,
						char marker,
						char open,
						char close,
						boolean enableClassLoading,
						List textAndTemplate )
	{
		allowClassLoading = enableClassLoading;
		vdelimiter = new String(new char[] { marker });
		openc = open;
		openStr = new String( new char[] { open } );
		closec = close;
		closeStr = new String( new char[] { close } );
		dmapValues = dictionary;
		if (pattern == null)
			template_string = "";
		else
			template_string = pattern;

		// Allocate the list for the parsed text and tokens
		if ( textAndTemplate != null )
		{
			text_and_token_string_array = new ArrayList(textAndTemplate.size());
			// If the user specified a text and template list, act as if we've already parsed
			text_and_token_string_array.addAll( textAndTemplate );
			dirty = false;
		}
		else
		{
			text_and_token_string_array = new ArrayList();
		}

	}

	/**
	 * Handles when the template or something else (marker, open, close) has changes and the template
	 * needs to be reparsed.
	 */
	private void nowDirty()
	{
		dirty = true;
	}

	/**
	 * Forces the template to be re-evaluated for tokens and the changes
	 * written to the Names Dictionary
	 */

	// Serialization
	private void readObject(java.io.ObjectInputStream stream)
		throws java.io.IOException, ClassNotFoundException
	{
		stream.defaultReadObject();
	}

	/**
	* This sets a map of values, other than the one returned by getNameDictionary
	* to read values from when formatting the values into the template.  It will
	* clear the toString cache.
	*/
	public void setValueDictionary( Map dictionary )
	{
		dmapValues = dictionary;
		clearToStringCache();
	}

	/**
	* This returns a secondary map of values other than the NameMap.  If no secondary values
	* dictionary has been set, then null will be returned.	And values set on the NameMap will
	* be used for values formatted.
	* @return Map interface to the object holding the values, if no object set with setValueDictionary then null will be returned.
	*/
	public Map getValueDictionary()
	{
		return dmapValues;
	}

	/**
	* This returns true if the class loading for the pattern string evaluation is enabled.
	* The default is to disable it.
	* @parm allowClassLoading true if enabled to load classes via the pattern
	*/
	public boolean getClassLoading()
	{
		return allowClassLoading;
	}

	/**
	* This enables the class loading for the pattern string evaluation
	* @parm enable enable or disable the ability to load classes via the pattern
	*/
	public void setClassLoading( boolean enable )
	{
		allowClassLoading = enable;
		nowDirty();
	}

	/**
	* Returns the template
	* @return a string containing variables and formatting directives
	*/

	public synchronized String getTemplate()
	{
		return template_string;
	}

	/**
	* Sets the template
	* @param string_val string containing variables and formatting directives
	*/

	public synchronized void setTemplate(String string_val)
	{
	    if (string_val == null)
	      string_val = "";

	    if ( string_val.equals(template_string) == false ) {
	    	template_string =  string_val;
			nowDirty();
	    }//if changed
	}

	/**
	* Set the marker property.
	* @param marker the new value for the marker property.
	*/
	public synchronized void setMarker(char marker)
	{
		char oldMarker = getMarker();
		if (oldMarker == marker)
			return;

		vdelimiter = new String(new char[] { marker });

		nowDirty();
	}


	/**
	* Return the value of the marker property.
	* @return the value of the marker property.
	*/
	public char getMarker()
	{
	    return vdelimiter.charAt(0);
	}

	/**
	* Set the PrintWriter to which the formatted template should be written.
	* @param writer the new value for the writer property.
	*/
	public void setWriter(PrintWriter writer)
	{
		this.writer = writer;
	}


	/**
	* Return the PrintWriter to which the formatted template should be written.
	* @return the value of the writer property.
	*/
	public PrintWriter getWriter()
	{
	    return this.writer;
	}


	/**
	* Set the close property.
	* @param close the new value for the close property.
	*/
	public synchronized void setClose(char close)
	{
		char oldClose = getClose();
		if (oldClose == close)
			return;

		closec = close;
		closeStr = new String( new char[] { close } );

		nowDirty();
	}


	/**
	* Return the value of the close property.
	* @return the value of the close property.
	*/
	public char getClose()
	{
	    return closec;
	}


	/**
	* Set the open property.
	* @param open the new value for the open property.
	*/
	public synchronized void setOpen(char open)
	{
		char oldOpen = getOpen();
		if (oldOpen == open)
			return;

		openc = open;
		openStr = new String( new char[] { open } );

		nowDirty();
	}

	/**
	* Return the value of the open property.
	* @return the value of the open property.
	*/
	public char getOpen()
	{
	    return openc;
	}

	/**
	 * Returns the ArrayList holding the text and tokens parsed from the template.
	 * @return the ArrayList holding the text and tokens parsed from the template
	 */
	public List getTextAndTokens()
	{
		// Tokenize the template if necessary
//		tokenizeTemplate();

		return (List) ((ArrayList)text_and_token_string_array).clone();
	}

	/**
	 * Sets the ArrayList holding the text and tokens instead of having them parsed from the template.
	 * The List contents are copied into an internal List so that changes to the passed in list
	 * do not effect the SimpleTemplate2.
	 * @param textAndTokens The ArrayList holding the text and tokens
	 */
	public synchronized void setTextAndTokens( List textAndTokens )
	{
		text_and_token_string_array.clear();
		text_and_token_string_array.addAll( textAndTokens );

		// note that the template and the tokenized version in text_and_token_string_array
		// may no longer match up.  if a reparse occurs, text_and_token_string_array will
		// be updated off the template.

		dirty = false;
	}

	// ------------------------------------------------------------------
	// PARSING CODE
	// ------------------------------------------------------------------
	/**
	* Tests a string for the presence of field delimiter(comma).
	* If present, then call ParseFields to separate string into
	* individual fields
	* @param nstring string to be parsed
	* @return array containing fields
	*/

	protected String[] findVarDelimiter(String nstring)
	{
		//search for commas
		// System.out.println("In findVarDel");
		String [] subvar = {null, null, null};
//		int [] cpos = Strings.positions(nstring,0,-1,',');
		int [] cpos = Strings.positions(nstring,',');
		if (cpos.length > 0) {
		  for (int i=0; i < 2; i++) { //parse for max of 3 fields
			  // System.out.println("nstring =" + nstring);
			  String[] temp = parseFields(nstring);
			  switch (i){
				  case 0:
					  subvar[0] = temp[0];
					  break;
				  case 1:
					  subvar[1] = temp[0];
					  break;
				  case 3:
					  subvar[2] = temp[0];
					  break;
			  } //switch
			  if (temp[1] == null) //no more string to parse
				  break;
			  else
				  nstring = temp[1];
		  } //for
		} //if cpos.length
		else
		  subvar[0] = nstring;

		for (int j=0; j < 3; j++) {
		  if (subvar[j] != null) {
//			  int [] hpos = Strings.positions(subvar[j],0,-1,'"');
			  int [] hpos = Strings.positions(subvar[j],'"');
			  if (hpos.length >=2)
				  subvar[j] = subvar[j].substring(hpos[0]+1, hpos[hpos.length-1]);

		  } //if
		} // for

		return subvar;
	}

	/**
	* Parse string into a field using comma as a field delimiter
	* @param mstring string to be parsed
	* @return array containing the field and the remainder of the string
	*/

	protected String[] parseFields(String mstring)
	{
		// System.out.println("mstring =" + mstring);
		String [] holder = {null, null};
		int k=0;
//		int [] cpos = Strings.positions(mstring,0,-1,",", false);
		int [] cpos = Strings.positions(mstring,',');

		int cposLength = cpos.length;
		if (cposLength > 0) {
			int mstringLength = mstring.length();
			for (int j = 0; j < cposLength; j++) {	
				String cstring = mstring.substring(k,cpos[j]);
				// System.out.println("cstring = " + cstring);
//				int [] qpos = Strings.positions(cstring,0,-1,'"');
				int [] qpos = Strings.positions(cstring,'"');
				if (qpos.length !=1) { //no quotes or matched quotes
					int cstringLength = cstring.length();
					if (cstringLength > 0) {
//						int [] ppos5 = Strings.positions(cstring,0,-1,openStr, false);
						int [] ppos5 = positions(cstring,openStr);
//						int [] ppos6 = Strings.positions(cstring,0,-1,closeStr, false); // new String( new char[] { cValue } );
						int [] ppos6 = positions(cstring,closeStr); // new String( new char[] { cValue } );
						if (ppos5.length == ppos6.length){ //comma not located between ()
							holder[0] = mstring.substring(k,cpos[j]);
							if (cpos[j]+1 < mstringLength)
								holder[1]  = mstring.substring(cpos[j]+1);

							break;
						} //if ppos5.length
						else {  //comma is located between ()
							holder[0] = mstring;
						}
					} //if cstring.length
					else if (cpos[j]+1 < mstringLength) {
						holder[1]  = mstring.substring(cpos[j]+1);
						break;
					} //else if
				} //if qpos.length
			} //for
		} //cpos.length
		else  //no commas
			holder[0] = mstring;

		return holder;
	}

	/**
	* Separate variable fields from other text in string
	* @return value indicating whether string was successfully parsed
	*/

	protected boolean parseStringTokens(String fstring, List strname)
	{
		boolean kvalue = false;
		int fstringLength = fstring.length();
		if (fstringLength > 0) {
			//String testString = fstring.toUpperCase();
			boolean flag = true;
			//check for lone %
			if (fstring.equals(vdelimiter)) {
				strname.add(fstring);
				kvalue = true;
			} //if
			else {

				for (int i=1; i < fstringLength; i++) {
					char c = fstring.charAt(i);
					//if ((Strings.lookingAt(testString, i, validChars)))
					if ((i==1) && (Character.isJavaIdentifierStart(c)))
						flag = true;
					else if ((i > 1) && (Character.isJavaIdentifierPart(c)))
						flag = true;
					else { //non-JavaIdentifier found
						if ( c != openc ) { //not the open character [default: '(')
							strname.add(fstring.substring(0,i));
							strname.add(fstring.substring(i));
							kvalue = true;
							flag=false;
							break;
						} //if fstring

						else { //found a parenthesis
							flag=false;
//							int [] ppos1 = Strings.positions(fstring,0,-1,openStr, false);
							int [] ppos1 = positions(fstring,openStr);
//							int [] ppos2 = Strings.positions(fstring,0,-1,closeStr, false);
							int [] ppos2 = positions(fstring,closeStr);
							// if ((ppos1.length > 0 && ppos2.length > 0) && (ppos1.length == ppos2.length)) {
							int ppos1Length = ppos1.length;
							int ppos2Length = ppos2.length;
							if (ppos1Length > 0 && ppos2Length > 0) { //one or more ( and )
								if (ppos1Length ==1 && ppos2Length ==1) { //one set of parentheses
									strname.add(fstring.substring(0,ppos2[0]+1)); /*I18nOK:LINE*/
									if (ppos2[0] + 1 <= fstringLength-1)
										strname.add(fstring.substring(ppos2[0]+1));

									kvalue = true;
								} //if ppos1.length
								else if (ppos1Length > 1) {
									if (ppos1[1] > ppos2[0]){ // ( )(
										strname.add(fstring.substring(0,ppos2[0]+1)); /*I18nOK:LINE*/
										if (ppos2[0] + 1 <= fstringLength-1)
											strname.add(fstring.substring(ppos2[0]+1));

										kvalue = true;
									} //if ppos1[1]
									else if (ppos1Length == ppos2Length) {
										//more than one set of matched parentheses
										strname.add(fstring.substring(0,ppos2[ppos2Length-1]+1)); /*I18nOK:LINE*/
										if (ppos2[ppos2Length-1] + 1 <= fstringLength-1)
											strname.add(fstring.substring(ppos2[ppos2Length-1]+1));
										kvalue = true;
									} //else if
								} //else if
								else if (ppos1Length < ppos2Length) {
									strname.add(fstring.substring(0,ppos2[0]+1)); /*I18nOK:LINE*/
									if (ppos2[0] + 1 <= fstringLength-1)
										strname.add(fstring.substring(ppos2[0]+1));
									kvalue = true;
								}
							} //if ppos1.length
						} //else found a parenthesis
						break;
					} //else
				} //for
				if (flag) { //all alphanumerics
					strname.add(fstring);
					kvalue = true;
				} //if flag
			} //else
		} //if fstring.length

		return kvalue;
	}

	/**
	* get the value for this name
	*/
	protected Object getValue(String vstring)
	{
		Object tempobj = null;

		Map values = getValueDictionary();
		if (  values != null ) {
			// System.out.println("Dictionary count = " + values.count());
			// System.out.println("Looking up '" + vstring +"' in Values Dictionary" );
			tempobj = values.get(vstring);
		} //if values dictionary defined

		return checkToStringCache( tempobj ); // check to see if this object has been cached, or should be cached.
	}


	/**
	* Process token into text
	* @param pstring string to be parsed
	* @return text
	*/

	protected String parseToken( String pstring )   {
		boolean var1Set = false;
		boolean var2Set = false;
		boolean var3Set = false;
		String resolvedString = null;

		//check if not simple string - no %delimiters and not just %
		if ( pstring.startsWith(vdelimiter) && pstring.length() > 1)
		{
			// not simple string
			String vstring; //get conditional name
			int pos = pstring.indexOf(openc);
			if (pos == -1)
			{
				//not complex(no parenthesis)
				vstring = pstring.substring(1);
			}
			else { //(pos >= 0)
				//it is complex (contains parentheses)
				vstring = pstring.substring(1,pos);
			}

//			String tstring = null;  //get variable value
			Object tempobj = getValue(vstring);
//			if (tempobj != null)
//				tstring =  tempobj.toString();

			if (pos == -1) {
//				resolvedString = formatWith( tstring );
				resolvedString = formatWith( tempobj );
			}
			else //tstring is conditional
			{
				//separate string into substitutionString,Default,and ClassFormatter
				String nstring = pstring.substring(pos+1); //new token
//				int [] mpos = Strings.positions(nstring,0,-1,closec); //search for )
				int [] mpos = Strings.positions(nstring,closec); //search for )
				nstring = nstring.substring(0, mpos[mpos.length -1]); //strip outer )
				nstring = nstring.trim();

				if (nstring.length() == 0)
				{
//					resolvedString = formatWith( tstring );
					resolvedString = formatWith( tempobj );
				}
				else { //nstring.length > 0
					String var[] = this.findVarDelimiter(nstring);

					
					List var_string_array = null;
					if ( var[0] != null ) {
						var1Set = true;
//						if ( var_string_array == null )
							var_string_array = new ArrayList();

						tokenizeField(var[0], var_string_array);

						if ( var_string_array.isEmpty() ) {
							var[0] = "";
						}
						else {
							StringBuffer hold = new StringBuffer(32);
							int varStringArraySize = var_string_array.size();
							for ( int i = 0; i < varStringArraySize; i++ ) {
								String out = parseToken( (String) var_string_array.get(i));
								if ( out != null )
									hold.append(out);
							}
							var[0] = hold.toString();
						}
					} //if var[0]

					if ( var[1] != null) {
						var2Set = true;
						if ( var_string_array == null )
							var_string_array = new ArrayList();
						else
							var_string_array.clear();

						tokenizeField(var[1], var_string_array);

						if ( var_string_array.isEmpty() ) {
							var[1] = "";
						}
						else {
							StringBuffer hold = new StringBuffer(32);
							int varStringArraySize = var_string_array.size();
							for ( int i = 0; i < varStringArraySize; i++ ) {
								String out = parseToken( (String) var_string_array.get(i));
								if ( out != null )
									hold.append(out);
							}
							var[1] = hold.toString();
						}
					} //if var[1]


					if ( var[2] != null) {
						var3Set = true;
						if ( var_string_array == null )
							var_string_array = new ArrayList();
						else
							var_string_array.clear();

						tokenizeField(var[2], var_string_array);
						if ( var_string_array.isEmpty() ) {
							var[2] = "";
						}
						else {
							StringBuffer hold = new StringBuffer(32);
							int varStringArraySize = var_string_array.size();
							for ( int i = 0; i < varStringArraySize; i++ ) {
								String out = parseToken( (String) var_string_array.get(i));
								if ( out != null )
									hold.append(out);
							}
							var[2] = hold.toString();
						}
					} //if var[2]

					// System.out.println("tstring = " + tstring);
					// System.out.println("var[0] = " + var[0]);
					// System.out.println("var[1] = " + var[1]);
					// System.out.println("var[2] = " + var[2]);

					if ( var3Set ) {
//						resolvedString = formatWith( tstring, var[0], var[1], var[2] );
						resolvedString = formatWith( tempobj, var[0], var[1], var[2] );
					}
					else if ( var2Set ) {
//						resolvedString = formatWith( tstring, var[0], var[1] );
						resolvedString = formatWith( tempobj, var[0], var[1] );
					}
					else if ( var1Set ) {
//						resolvedString = formatWith( tstring, var[0] );
						resolvedString = formatWith( tempobj, var[0] );
					}
					else
						return null;
				} //else nstring.length > 0
			} //else pos.length() > 0
		} //if pstring
		else
			resolvedString = formatWith( pstring );

		// System.out.println("resolvedString = " + resolvedString);
		return resolvedString;
	}

	/**
	 * get the string to insert for the object, defaults to calling toString
	 */
	protected String formatValueObject( Object tObject )
	{
		String tstring = null;

		if ( tObject != null )
		{
			tstring = tObject.toString();
//			System.out.println("Called ToString on value object and got [" + tstring + "]" );
		}

		return tstring;
	}

	protected String formatWith( Object tObject )
	{
		return formatWith( formatValueObject(tObject) );
	}

	protected String formatWith( Object tObject, String notNullValue )
	{
		return formatWith( formatValueObject(tObject), notNullValue );

	}

	protected String formatWith( Object tObject, String notNullValue, String isNullValue  )
	{
		return formatWith( formatValueObject(tObject), notNullValue, isNullValue );
	}

	protected String formatWith( Object tObject, String notNullValue, String isNullValue, String formatterClass  )
	{
		return formatWith( formatValueObject(tObject), notNullValue, isNullValue, formatterClass );
	}

	/**
	 * Allows the text to be replace a singleton placehold to be munged just before insertion.
	 * Default behaivor is to just insert the string.
	 * @return value string that is going to replace variable
	 */
	protected String formatWith( String origValue )	{ return origValue; }


 	/**
	 * Allows the text to be replace a placehold with 1 parameter to be munged just before insertion
	 * Default behavior is insert the second parameter if the first has a value.
	 * @return value string that is going to replace variable
	 */
	protected String formatWith( String origValue, String notNullValue )
	{
		if ( origValue == null )
			return null;
		else
			return notNullValue; //return substitutionString
	}


 	/**
	 * Allows the text to be replace a placehold with 2 parameters to be munged just before insertion
	 * Default behavior is insert the second parameter if the first has a value, otherwise insert the third if null
	 * @return value string that is going to replace variable
	 */
	protected String formatWith( String origValue, String notNullValue, String isNullValue )
	{
		if (origValue == null)  //conditional is null
			return isNullValue; //return default
		else //conditional not null
			return notNullValue; //return substitutionString
	}

 	/**
	 * Allows the text to be replace a placehold with 2 parameters to be munged just before insertion by using a format on a formatter class
	 * Default behavior is format the orignal value if parms 1 and 2 are empty
	 * If the original value is null, then format the second value
	 * If the original value is not null, then use it
	 * The first parameter is not used in order to maintain complience with the original Template spec
	 * @return value string that is going to replace variable
	 */
	protected String formatWith( String origValue, String notNullValue, String isNullValue, String formatterClass )
	{
		// origValue - value of the placeholder being replaced
		// notNullValue - if value is not null then use this
		// isNullValue - if value is null then use this
		// formatter - use this class to format
		if ( allowClassLoading )
		{
			try {
				Format fd = (Format) (GlobalInfoTable.loadClass(formatterClass).newInstance());
				String formatted;

				if (origValue != null)
					//if ( notNullValue != null )
					//	formatted = fd.format((Object)notNullValue);
					//else
						formatted = fd.format((Object)origValue);
				else
					formatted = fd.format((Object)isNullValue);

				return formatted;
			} //try
			catch (InstantiationException e) {
				return null;
			} //catch
			catch (IllegalAccessException e) {
				return null;
			} //catch
			catch (ClassNotFoundException e) {
				return null;
			} //catch
		}
		else
		{
			return formatWith( origValue, notNullValue, isNullValue );
		}
	}

	/**
	* Create a string from the template where any
	* variables have been replaced with their values
	*/

	protected void processText() // throws FormatException
	{
		//parse template string for variable substitution
		//Note this also performs double-delimiter substitution.
		tokenizeTemplate();

		// clear out the output array that will contain the text parts
		if ( temp_string_array != null )
		{
			temp_string_array.clear();
		}

		boolean empty = true;
		PrintWriter currentWriter = getWriter();
		int temp_string_array_totalstringlength = 0;
		for (int npos =0, nsize=text_and_token_string_array.size(); npos < nsize; npos++){
			String pstring = (String) text_and_token_string_array.get(npos);

//			System.out.println("  pstring = " + pstring);
			String tmp = parseToken(pstring);

//			System.out.println("    tmp = " + tmp);
	        if ( tmp != null ) {
				if ( currentWriter != null ) {
					empty = false;
					// SPL: Took out the Strings.replace call since the parsing prevents
					// double delimiters from being in the same string in the array
//					currentWriter.print(Strings.replace(tmp, 0, -1, doubleDelim, vdelimiter));
					currentWriter.print(tmp);
				}
				else {
					// Lazy allocate the array since sometimes we're just writing to a stream
					if ( temp_string_array == null )
					{
						temp_string_array = new ArrayList( nsize );
					}
					// Save what we are going to append together later
					temp_string_array.add(tmp);
					// Keep track of the total size since we are putting the output in a StringBuffer
					temp_string_array_totalstringlength += tmp.length();
				}
			}
		}  //for

		if ( currentWriter != null ) {
			if ( empty ) {
				currentWriter.print("");
			}
			return;
		}

		//Handle the case of no text
		if ( temp_string_array == null )
		{
			submit_string = "";
			return;
		}

		//recreate the submit string
		int tempStringArraySize = temp_string_array.size();
		StringBuffer temp_string_buffer = new StringBuffer(temp_string_array_totalstringlength + DELIMITER_REPLACE_PAD_SIZE);
		for (int k = 0; k < tempStringArraySize; k++) {
			temp_string_buffer.append(temp_string_array.get(k) ); /*I18nOK:LINE*/
		}

		// Convert to a string and replace the double delimiters with a single one
//		String temp_string = temp_string_buffer.toString()/*I18nOK:MTH*/;
//		if (temp_string_buffer.equals(submit_string) == false )
		{
//			submit_string = Strings.replace(temp_string, 0, -1, doubleDelim, vdelimiter);
			submit_string = temp_string_buffer.toString();
		} //if

		// Clear the list so we don't hold onto data too long (since it could be quite large)
		temp_string_array.clear();
	}

//	private StringBuffer replace( StringBuffer buffer, String search, String replace )
//	{
////System.out.println( "buffer = " + buffer );
//		int start = 0;
//		int searchLength = search.length();
//		while ( start < buffer.length() )
//		{
//			int index = buffer.indexOf( search, start );
////System.out.println( "   index = " + index );
//			if ( index == -1 )
//			{
//				return buffer;
//			}
//			else
//			{
////System.out.println( "   start = " + start);
////System.out.println( "   searchLength = " + searchLength );
//				buffer.replace( index, index + searchLength, replace );
//				start = index + replace.length();
//			}
//		}
//		return buffer;
//	}

	/**
	* Tokenize the field
	*/

	protected void tokenizeField(String strtemp, List collname) // throws FormatException
	{
//	    int [] tpos = Strings.positions(strtemp,0,-1,vdelimiter, false);
	    int [] tpos = positions(strtemp, vdelimiter);
		int tposLength = tpos.length;
		if (tposLength > 0) {
			int k=0;
			int j;
			for (j = 0; j < tposLength; j++) {
				String fstring = strtemp.substring(k,tpos[j]);
				// System.out.println("fstring = " + fstring);
				if (parseStringTokens(fstring, collname))
					k = tpos[j];

			} //for
			//parse remainder of string
			/*boolean cvar = */parseStringTokens(strtemp.substring(k,strtemp.length()), collname);
		}//if tpos.length

		else
			collname.add(strtemp);

		// for (int i=0; i< collname.count(); i++)
			// System.out.println("field array value=" + collname.getString(i));
	}

	/**
	 * Similar to Strings.positions except does not support overlapping and always
	 * does the whole source.
	 */
	private int[] positions( String source, String delimiter )
	{
		if ( delimiter.length() == 1 )
		{
//			//***************************
//			// Start Test code
//			//***************************
//			int[] myPos = positions2( source, delimiter.charAt(0) );
//			int[] strPos = Strings.positions( source, 0, -1, delimiter, false );
//
//			{
//				System.out.println( "Search: " + source + " : " + delimiter );
//				System.out.println( "  Lengths: " + myPos.length + " : " + strPos.length );
//				for (int i = 0; i < myPos.length; i++ )
//				{
//					System.out.println( "    " + myPos[i] + " : " + strPos[i] );
//				}
//			}
//			//***************************
//			// End Test code
//			//***************************

			return Strings.positions( source, delimiter.charAt(0) );
		}

		return Strings.positions( source, 0, -1, delimiter, false );
	}

	/**
	 * Similar to Strings.positions except always does the whole source.
	 */
//	private static int POSITION_CACHE_SIZE = 10;
//	public static int[] positions2( String source, char delim )
//	{
//		// buffer the first 'n' tokens
////		int[] posCache = null;
//		int[] posCache = new int[POSITION_CACHE_SIZE];
//
//		// Count the delimiters.
//		// Keep track of the last found index in case there is only one.
//		// Cache some number of indices so we don't have to go through String twice to
//		// count and then store the indices in the proper size array
//		int count = 0;
//		int last = -1;
//		for (int i=0, n=source.length(); i < n; i++ )
//		{
//			if ( source.charAt(i) == delim )
//			{
//				// Cache the index if we're within the cache size
//				if ( count < POSITION_CACHE_SIZE )
//				{
//					// If we've found the second index, we need to make the cache array now.
//					// We can wait until we've found the second since we're storing the first
//					// index in 'last'.
////					if ( count == 1 )
////					{
////						posCache = new int[POSITION_CACHE_SIZE];
////						posCache[0] = last;
////					}
//					// Store the index if we're caching.  This won't happen if we only have one
//					// index so far.
////					if ( posCache != null )
//					{
//						posCache[count] = i;
//					}
//				}
//				last = i;
//				count++;
//			}
//		}
//
//		// Handle some easy special cases
//		if ( count == 0 )
//		{
//			return new int[0];
//		}
//		else if ( count == 1 )
//		{
//			return new int[] { last };
//		}
//		else if ( count == POSITION_CACHE_SIZE )
//		{
//			// No need to copy data or allocate a new array if we filled the cache
//			return posCache;
//		}
//
//		// Handle the cached case more efficiently
//		if ( count <= POSITION_CACHE_SIZE )
//		{
//			int[] output = new int[ count ];
//			System.arraycopy( posCache, 0, output, 0, count );
//			return output;
//		}
//
//		// Make the return array
//		// This scans the source string again.
//		// This avoids making the ArrayList that Strings.positions() creates.
//		// Make use of our cache to avoid rescanning all of the String.
//		int[] output = new int[ count ];
//		System.arraycopy( posCache, 0, output, 0, POSITION_CACHE_SIZE );
//		count -= POSITION_CACHE_SIZE;
//		int index = POSITION_CACHE_SIZE;
//		for (int i=posCache[POSITION_CACHE_SIZE-1]+1, n=source.length(); i < n  &&  count > 0; i++ )
//		{
//			if ( source.charAt(i) == delim )
//			{
//				output[index++] = i;
//				count--;
//			}
//		}
//
//		return output;
//	}

//	private static int POSITION_CACHE_SIZE = 10;
//	public static int[] positions( String source, char delim )
//	{
//		// buffer the first 'n' tokens
////		int[] posCache = null;
//		int[] posCache = new int[POSITION_CACHE_SIZE];
//
//		// Count the delimiters.
//		// Keep track of the last found index in case there is only one.
//		// Cache some number of indices so we don't have to go through String twice to
//		// count and then store the indices in the proper size array
//		int count = 0;
//		int last = -1;
//		for (int i=0, n=source.length(); i < n; )
//		{
//			int index = source.indexOf( delim, i );
//			if ( index != -1 )
//			{
//				// Cache the index if we're within the cache size
//				if ( count < POSITION_CACHE_SIZE )
//				{
//					// If we've found the second index, we need to make the cache array now.
//					// We can wait until we've found the second since we're storing the first
//					// index in 'last'.
//					if ( count == 1 )
//					{
//						posCache = new int[POSITION_CACHE_SIZE];
//						posCache[0] = last;
//					}
//					// Store the index if we're caching.  This won't happen if we only have one
//					// index so far.
//					if ( posCache != null )
//					{
//						posCache[count] = index;
//					}
//				}
//				last = index;
//				i = index + 1;
//				count++;
//			}
//			else
//			{
//				break;
//			}
//		}
//
//		// Handle some easy special cases
//		if ( count == 0 )
//		{
//			return new int[0];
//		}
//		else if ( count == 1 )
//		{
//			return new int[] { last };
//		}
//		else if ( count == POSITION_CACHE_SIZE )
//		{
//			// No need to copy data or allocate a new array if we filled the cache
//			return posCache;
//		}
//
//		// Handle the cached case more efficiently
//		if ( count <= POSITION_CACHE_SIZE )
//		{
//			int[] output = new int[ count ];
//			System.arraycopy( posCache, 0, output, 0, count );
//			return output;
//		}
//
//		// Make the return array
//		// This scans the source string again.
//		// This avoids making the ArrayList that Strings.positions() creates.
//		// Make use of our cache to avoid rescanning all of the String.
//		int[] output = new int[ count ];
//		System.arraycopy( posCache, 0, output, 0, POSITION_CACHE_SIZE );
//		count -= POSITION_CACHE_SIZE;
//		int index = POSITION_CACHE_SIZE;
//		for (int i=posCache[POSITION_CACHE_SIZE-1]+1, n=source.length(); i < n  &&  count > 0; i++ )
//		{
//			if ( source.charAt(i) == delim )
//			{
//				output[index++] = i;
//				count--;
//			}
//		}
//
//		return output;
//	}

	/**
	* Tokenize the template.
	* This method first performs substitution of any double-delimiters 
	*/
	protected void tokenizeTemplate() // throws FormatException
	{
//if (1==1) throw new IllegalArgumentException( "just checking");
		if ( !dirty )
		{
			return;
		}
		
		// S0241653: Perform double-delimiter substitution on the template string
		// *before* any tokenization and variable substitutions are done, so that we don't 
		// alter the values of those variables (e.g. a variable with '%%%' as a value).
		String doubleDelim = vdelimiter+vdelimiter;
		String template_string2 = Strings.replace(template_string, 0, -1, doubleDelim, vdelimiter);


		// We're starting over and rebuilding this
		text_and_token_string_array.clear();

//		int [] tpos = Strings.positions(template_string2,0,-1,vdelimiter, false);
		int [] tpos = positions(template_string2,vdelimiter);
		String tempstring = template_string2;
//		System.out.println("template string = " + template_string2);

		int tposLength = tpos.length;
		if (tposLength > 0) {
			int k=0;
			for (int j = 0; j < tposLength; j++) {
				String fstring = tempstring.substring(k,tpos[j]);
//				System.out.println("    fstring = " + fstring);
				if (parseStringTokens(fstring, text_and_token_string_array))
					k = tpos[j];

			} //for
			//parse remainder of string
//			System.out.println("    remaining fstring = " + tempstring.substring(k,tempstring.length()));
			boolean cvar = parseStringTokens(tempstring.substring(k,tempstring.length()), text_and_token_string_array);

		}//if tpos.length
		else
		{
			// No delimiters, so just add the whole string to the list
			text_and_token_string_array.add(template_string2);
		}

		// The text and token array is now up to date
		dirty = false;
	}

	// ------------------------------------------------------------------
	// FORMAT METHODS
	// ------------------------------------------------------------------

	/**
	* Returns the text created from the template
	* @return string whose variables have been
	* replaced with their values
	*/

	public synchronized String format()
	{
		if ( _toStringCacheClearedOnFormat )
			clearToStringCache();

		processText();  // formats
		return submit_string;
	}

	/**
	* This will do all the work of formatting a new object with a new valuesMap for the current template.
	* @param collection object that implements the java.util.Map interface
	* @return the formatted string from current pattern with the dictionary provided
	*/
	public String format( Map dictionary )
	{
		setValueDictionary( dictionary );
		return format();
	}


	/////////////////////////////////////////////////////////////////////
	// Static Methods
	/////////////////////////////////////////////////////////////////////

	/**
	* Conveniance method.  This will do all the work of formatting a specified pattern string
	* and object that implements the Map interface.
	* @param pattern specified template
	* @param dictionary collection object that implements java.util.Map interface
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	* @return the formatted string from the specified pattern with the dictionary provided
	*/
	static public String format(  String pattern,
	                              Map dictionary,
	                              char marker,
	                              char open,
	                              char close)
	{
		SimpleTemplate2 tmp = new SimpleTemplate2( pattern, dictionary, marker, open, close );
		return tmp.format();
	}

	/**
	* Conveniance method.  This will do all the work of formatting a specified pattern string
	* and object that implements the Map interface.
	* @param pattern specified template
	* @param dictionary collection object that implements java.util.Map interface
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	* @param out The PrintWriter to which the formatted template should be written. 
	* @return the formatted string from the specified pattern with the dictionary provided
	*/
	static public void format(  String pattern,
	                              Map dictionary,
	                              char marker,
	                              char open,
	                              char close,
								  PrintWriter out)
	{
		SimpleTemplate2 tmp = new SimpleTemplate2( pattern, dictionary, marker, open, close );
		tmp.setWriter(out);
		tmp.format();
	}

	/**
	* Conveniance method.  This will do all the work of formatting a specified pattern string
	* and object that implements the Map interface.
	* @param pattern specified template
	* @param dictionary colleciton object that implements the java.util.Map interface
	* @return the formatted string from the specified pattern with the dictionary provided
	*/
	static public String format(  String pattern,
	                              Map dictionary)
	{
		return SimpleTemplate2.format( pattern, dictionary, '%', '(', ')' ); /*I18nOK:EMS*/
	}

	/**
	* Conveniance method.  This will do all the work of formatting a specified pattern string
	* and object that implements the Map interface.
	* @param pattern specified template
	* @param dictionary colleciton object that implements the java.util.Map interface
	* @param out The PrintWriter to which the formatted template should be written.
	* @return the formatted string from the specified pattern with the dictionary provided
	*/
	static public void format(  String pattern,
	                              Map dictionary,
								  PrintWriter out)
	{
		SimpleTemplate2.format( pattern, dictionary, '%', '(', ')', out ); /*I18nOK:EMS*/
	}

//	/**
//	* Returns TRUE if there are names in the template that are not in the values dictionary
//	* @return boolean TRUE if there are Names in the template that are not in the Values provided
//	*/
//
//	public boolean notFound()
//	{
//		return ( ! getNotFoundNames().isEmpty() );
//	}
//
//	/**
//	* Returns a SET of Names from the template that are not defined for the current
//	* values dictionary
//	* @return a java.util.Set of the names not found
//	*/
//
//	public Set getNotFoundNames()
//	{
//		Map mapNames = getNameDictionary();
//		HashSet setNameKeys = new HashSet(mapNames.keySet());
//
//		Map mapValues = getValueDictionary();
//		if ( mapValues != null )
//		{
//			Set setValueKeys = mapValues.keySet();
//			setNameKeys.removeAll( setValueKeys );
//		}
//		return setNameKeys;
//	}

	// ------------------------------------------------------------------
	// CACHING CODE
	// ------------------------------------------------------------------

	/**
	* Clear the cached toStrings if any
	*/
	public void clearToStringCache() { _toStringCache.clear(); }


	/**
	* If caching is enabled, find the cached toString and if not found then it is generated and saved.
	* @param object object to be looked up in the toString cache
	* @return object either the original object if cachine not enabled or the string that is cached
	*/
	protected Object checkToStringCache( Object object )
	{
		if ( object == null ) return null;
		Object result = object;
		if ( isToStringCachingEnabled() )
		{
			if ( ! ( object instanceof String ) )
			{
				if ( _toStringCache.containsKey( object ) )
				{
					result = _toStringCache.get( object );
				}
				else
				{
					String value = object.toString();
					_toStringCache.put( object, value);
					result = value;
				}
			}
		}
		return result;
	}

	/**
	* Test to see if toString cashing is enabled.
	* @return boolean true if toStrings will be cached
	*/
	public boolean isToStringCachingEnabled() { return _toStringCachingEnabled; }

	/**
	* enable or disable the toString cashing
	* @param boolean true will enable the toString Cache
	*/
	public void setToStringCachingEnabled( boolean enable )
	{
		_toStringCachingEnabled = enable;
	}

	/**
	* Test to see if toString cashing will be cleared on each call to format
	* @return boolean true if cached toStrings will be cleared when format begins filing in of the template
	*/
	public boolean isToStringCacheClearedOnFormat() { return _toStringCacheClearedOnFormat; }

	/**
	* enable or disable the clearing of the toString cashe on each call to format
	* @param boolean true will enable the toString Cache
	*/
	public void setToStringCacheClearedOnFormat( boolean enable )
	{
		_toStringCacheClearedOnFormat = enable;
	}

}  //class SimpleTemplate2

