/* $Id: SimpleTemplateToPrintWriter.java,v 1.1.4.1 2004/12/21 17:41:46 sasnxb Exp $ */
// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import com.sas.GlobalInfoTable;
import com.sas.beans.ExtendedBeanInfo;
import com.sas.util.Strings;

import java.io.IOException;
import java.io.PrintWriter;
import java.text.Format;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.Collections;

/**
 * @author Neil Bell
 * @version	1.0
 */

/**
 * The SimpleTemplateToPrintWriter class serves as a generic string substitution
 * class that sends the formatted template to a PrintWriter object.  A raw text string,
 * referred to as the template, can contain variables, references, and text.
 * 
 * <p>
 * The SimpleTemplateToPrintWriter class also uses a dictionary that contains variables
 * in the template along with their associated values. Writing the template replaces keys
 * from the dictionary with their corresponding values.	This dictionary is a collection
 * object that implements the java.util.Map interface.
 * </p>
 * 
 * <p>
 * When a value from this dictionary implements the 
 * {@link com.sas.util.WriteToPrintWriterInterface}, it writes it substitution string
 * to the PrintWriter object itself when its
 * {@link com.sas.util.WriteToPrintWriterInterface#write(java.io.PrintWriter)} method
 * is called by this class. Otherwise, a value from this dictionary is resolved via a
 * single call to its toString method.  The result of the toString operation is cached
 * for use in all occurences of that dictionary object in the pattern.
 * </p>
 * 
 * <p>
 * The caching of toString results is enabled by default. If toStringCaching is disabled
 * then the value objects used in the dictionary will be evaluated with a toString call
 * once for each occurence in the pattern. The toString cache is cleared whenever a
 * write(), format(), setValueDictionary() or the clearToStringCache() mehod is called.
 * </p>
 * 
 * <p>
 * If the toString cache should NOT be reset upon the start of each write, then
 * setClearToStringCacheOnFormat(false).  When this is set to false, the cache will
 * only clear when a setValueDictionary() or clearToStringCache() is done.
 * </p>
 * 
 * <p>
 * The general form of a template is:
 * <pre>
 *   text %VAR(FLD1, FLD2) more text
 * <pre>
 * where:
 * <p>
 *  VAR is the name of a variable that is defined in the dictionary
 *  supplied to the SimpleTemplateToPrintWriter. Variable names must begin with %.
 *  The next character must be a valid identifier start character as defined
 *  by {@link java.lang.Character#isJavaIdentifierStart(char ch)}.
 *  All subsequent characters must be valid identifier characters as defined by
 *  {@link java.lang.Character#isJavaIdentifierPart(char ch)}.
 *  Variable names are case sensitive.
 *	<p>
 *  NOTE:  The '%' is not part of the variable name (it acts as a delimiter). In order to
 *  specify a value for the variable, %PROC, one would use the following syntax where the
 *  dictionary key is specified as 'PROC' not '%PROC'. The '%' character may be changed
 *  by calling the {@link #setMarker(char)} method, or you can used the constructor which
 *  specifies the marker character.
 *
 * <ul>
 * <li> dictionary.set("PROC", "SORT");
 * </ul>
 * <p>
 * <p>
 *   FLD1 is an optional field that will be evaluated and displayed if VAR is
 *   defined in the dictionary and is not null. If VAR is not defined,
 *   FLD1 will be not be displayed. FLD1 can be designated as a combination of
 *   text and variables.
 * <p>
 *   FLD2 is an optional field that contains a default value that should be used
 *   if VAR is not defined in the dictionary or is null. FLD2 can be designated
 *   as a combination of text and variables.
 * <p>
 * <p>
 * NOTES:
 * <p>
 * Commas are used for field delineation. If you need to use a comma as part of
 * a field expression (FLD1,FLD2), that field must be surrounded by quotes.
 * <p>
 * % is used for variable delineation. %% will be used to denote a % that is not
 * used for a variable name. If you change the marker character with   {@link #setMarker(char)},
 * you must double it to see it rendered. For example, if you call <code>setMarker('$')</code>
 * then you would use a template such as <code>"The price is $$$price"</code>, with
 * price = 10.75 in the dictionary, to yield <code>"$10.75"</code>
 * <p>
 * The following are example of templates:
 * <p>
 * EXAMPLE 1:
 * <pre>
 *     "The woman is %status"
 * </pre>
 * If status = "very wealthy.", the expression will be processed to:
 * <pre>
 *     "The woman is very wealthy."
 * </pre>
 * If status is not defined, the expression will be processed to:
 * <pre>
 *     "The woman is"
 * </pre>
 * EXAMPLE 2:
 * <pre>
 *     "Today it is %DAY(%DAY, raining)."
 * </pre>
 * If the variable DAY is contained in the dictionary and set to "sunny", the
 * string will be processed to:
 * <pre>
 *     "Today it is sunny."
 * </pre>
 * If the variable DAY is not contained in the dictionary or is null, the string
 * will be processed to:
 * <pre>
 *     "Today it is raining."
 * </pre>
 * EXAMPLE 3:
 * <pre>
 *     "Hello. %person(%GREET %person?)"
 * </pre>
 * If person = "Mary" and GREET = "How are you", the expression will be
 * processed to:
 * <pre>
 *     "Hello. How are you Mary?"
 * </pre>
 * If person is not defined in the dictionary, the expression will be
 * processed to:
 * <pre>
 *     "Hello."
 * </pre>
 * If person = "Mary" and GREET is not defined, the expression will be
 * processed to:
 * <pre>
 *     "Hello. Mary?"
 * </pre>
 * <p>
 * EXAMPLE 4:
 * <pre>
 * %state(, is not defined)
 * </pre>
 * If state is defined in the dictionary and is not null, the expression
 * will be processed to:
 * <pre>
 *     ""  (string of zero length)
 * </pre>
 * If state is not defined in the dictionary the expression will be processed to:
 * <pre>
 *     "is not defined"
 * </pre>
 * Note that the comma is necessary for correct field delineation.
 * <p>
 * EXAMPLE 5:
 * <pre>
 *     "The stock market is %market(%direction, closed)."
 * </pre>
 * If market is defined and direction="up today" the expression will be processed to:
 * <pre>
 *     "The stock market is up today."
 * </pre>
 * As this example illustrates, the VAR field (market) does not have to appear in the FLD1
 * expression (direction). The VAR field acts first as a true/false operator that
 * dictates whether to evaluate FLD1 or FLD2 (FLD1 if VAR is in dictionary and not null,
 * FLD2 otherwise). If VAR appears in the expression to be evaluated, its value is
 * substituted in the expression.
 * <p>
 * <p>
 * The special characters '%', '(', and ')' can be replaced with other
 * characters on the SimpleTemplateToPrintWriter constructor, or by setting the
 * {@link #setMarker(char) <var>marker</var>},
 * {@link #setOpen(char) <var>open</var>},
 * {@link #setClose(char) <var>close</var>}
 * properties. For example,
 * <pre>
 *  template.setMarker('$');
 *  template.setOpen('{');
 *  template.setClose('}');
 *  template.setTemplate("The stock market is $market{$direction, closed}.");
 * </pre>
 */

public class SimpleTemplateToPrintWriter
	extends SimpleTemplate2
{
	static final String RB_KEY = "SimpleTemplateToPrintWriter.";

	protected boolean empty;

	/**
	* Construct a SimpleTemplateToPrintWriter with the specified template and object that implements
	* the java.util.Map interface.
	* @param pattern specified template
	* @param dictionary collection object that implements java.util.Map interface
	* @param out The PrintWriter to which the template should be written.
	*/
	public SimpleTemplateToPrintWriter(String pattern, Map dictionary, PrintWriter out)
	{
		super(pattern, dictionary);
		setWriter(out);
	}

	/**
	* Construct a SimpleTemplateToPrintWriter with the specified template and object that implements
	* the Map Interface.
	* @param pattern specified template
	* @param dictionary colleciton object that implements java.util.Map interface
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	* @param out The PrintWriter to which the template should be written.
	*/
	public SimpleTemplateToPrintWriter(String pattern,
	              Map dictionary,
	              char marker,
	              char open,
	              char close,
				  PrintWriter out)
	{
		super(pattern, dictionary, marker, open, close);
		setWriter(out);
	}

	/**
	 * Writes the token to the output PrintWriter.
	 * @param pstring a token from the template pattern.
	 */
	protected void writeToken( String pstring )   {

		if ( !(pstring.startsWith(vdelimiter)) || (pstring.length() <= 1) ) {
			getWriter().print( pstring );
			return;
		}

		int pos = pstring.indexOf(openc);
		if (pos == -1) {
			//not complex(no parenthesis)
			writeValue( getValueFromDictionary(pstring.substring(1)) );
			return;
		}

//		String nstring = preFormatToken(pstring, pos);
		String nstring = pstring.substring( pos+1, pstring.lastIndexOf( closec ) ).trim();
		if (nstring.length() == 0)
		{
			writeValue( getValueFromDictionary(pstring.substring(1,pos)) );
		}
		else { //nstring.length > 0
			Object value = getValueFromDictionary(pstring.substring(1,pos));

			//separate string into substitutionString,Default,and ClassFormatter
			String var[] = this.findVarDelimiter(nstring);
			if ( (value != null) && (var[0] != null) ) {
				writeVar( var[0] );
			}
			else if ( (value == null) && (var[1] != null) ) {
				writeVar( var[1] );
			}
		} //else nstring.length > 0

		return;
	}

	private void writeValue( Object value )
	{
		if ( value == null ) {
			return;
		}

		empty = false;
		if ( value instanceof WriteToPrintWriterInterface ) {
			try {
				((WriteToPrintWriterInterface)value).write(getWriter());
			}
			catch( IOException ioe ) {
				throw new RuntimeException(ioe);
			}
		}
		else {
			getWriter().print( checkToStringCache( value ) );
		}
	}

	private void writeVar( String var )
	{
		ArrayList var_string_array = new ArrayList();
		tokenizeField(var, var_string_array);

		if ( var_string_array.isEmpty() ) {
			empty = false;
			getWriter().print( "" );
		}
		else {
			int varStringArraySize = var_string_array.size();
			for ( int i = 0; i < varStringArraySize; i++ ) {
				empty = false;
				writeToken( (String)var_string_array.get(i) );
			}
		}
	}

	/**
	* Write a string from the template where any
	* variables have been replaced with their values
	*/
	protected void writeText() // throws FormatException
	{
		//parse template string for variable substitution
		tokenizeTemplate();

		empty = true;
		for (int npos =0, nsize=text_and_token_string_array.size(); npos < nsize; npos++){
			String pstring = (String) text_and_token_string_array.get(npos);
			writeToken(pstring);
		}

		if ( empty ) {
			getWriter().print("");
		}
	}

	/**
	* Writes the text created from the template
	*/
	public synchronized void write()
	{
		if ( getWriter() == null ) {
			throw new IllegalStateException(RB.getStringResource(RB_KEY, "noWriter.txt"));
		}

		if ( _toStringCacheClearedOnFormat )
			clearToStringCache();

		writeText();
		return;
	}

	/**
	* Writes the text created from the template
	*/
	static public void write(String pattern, Map dictionary, PrintWriter out)
	{
		SimpleTemplateToPrintWriter.write( pattern, dictionary, '%', '(', ')', out ); /*I18nOK:EMS*/
	}

	/**
	* Conveniance method.  This will do all the work of writing a specified pattern string
	* and object that implements the Map interface.
	* @param pattern specified template
	* @param dictionary collection object that implements java.util.Map interface
	* @param marker the character to used to mark template expansion strings
	* in the template. The default is '%'
	* @param open the character to use instead of the default '(' to open
	* expansion options after the marker/name
	* @param close the character to use instead of the default ')' to close
	* expansion options after the marker/name
	* @param out The PrintWriter to which the formatted template should be written. 
	* @return the formatted string from the specified pattern with the dictionary provided
	*/
	static public void write(  String pattern,
	                              Map dictionary,
	                              char marker,
	                              char open,
	                              char close,
								  PrintWriter out)
	{
		SimpleTemplateToPrintWriter tmp = new SimpleTemplateToPrintWriter( pattern, dictionary, marker, open, close, out );
		tmp.write();
	}

	private Object getValueFromDictionary(String key)
	{
		Map values = getValueDictionary();
		if (  values != null ) {
			return values.get(key);
		}
		return null;
	}

//	private String preFormatToken(String token, int startPos)
//	{
//		String nstring = token.substring(startPos+1); //new token
//		int [] mpos = Strings.positions(nstring,closec); //search for )
//		nstring = nstring.substring(0, mpos[mpos.length -1]); //strip outer )
//		return nstring.trim();
//	}

}  //class SimpleTemplateToPrintWriter

