/* $Id$ */
// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Comparator;
//import java.util.Vector;
import java.util.Hashtable;

/**
 * Various string manipulation methods not found in the <code>{@link java.lang.String}</code> package.
 * @author David Biesack
 * @version 1.0
 * @date 15Oct96
 */

public class Strings
{
  public static final String RB_KEY = "Strings.";

  /**
   * A typical pad character for {@link #split(String, int, Character)}
   * that consists of the space '&nbsp;' character.
   */
  public static final Character SPACE = new Character(' ');

  // private constructor prevents instantiation;
  // this class has only static methods and there should be no instances.
  private Strings()
  {
  }

  /**
   * Test if a string matches a pattern. This is not a fully featured
   * regular expression matching (i.e. it does not do character sets,
   * character syntax, grouping, etc,) but does simple wild card matching.
   * <p>
   * For example, <pre>
   * Strings.match("this is an absurd example", "*", true),
   * Strings.match("A", "a*", true);
   * Strings.match("Voila!", "Voila?", true);
   * Strings.match("A very simple match", "a*", true);
   * Strings.match("NCG2224", "NC?22*", true)</pre>
   * all return <code>true</code>, whereas <pre>
   * Strings.match("aab", "a*a", true);
   * Strings.match("match asterisk exactly", "match \\* exactly", true),
   * Strings.match("eh?", "a?", true),
   * </pre> all return <code>false</code>.
   * <p>
   * Note that if you want to match a <code>'?'</code> or <code>'*'</code> character exactly (i.e. not
   * treat them as special wildcards), you should escape them by
   * preceeding them with a <code>'\'</code> character in the pattern string
   * (which means you probably need <code>\\?</code> or <code>\\*</code> in string constants
   * used for the pattern string, since <code>'\'</code> is a Java
   * string literal escape character). For example, to test if a string
   * contains a '?' after a '*' character, use
   * <code>Strings.match(string, "*\\?*\\**")</code>.
   * @param string a string which is matched against the pattern.
   * @param pattern the pattern string. Within the pattern string,
   * a '?' character matches exactly one single character in <code>string</code>,
   * and a '*' will match zero or more substrings in <code>string</code>.
   * @param ignoreCase if <code>true</code>, differences in case are ignored.
   * @return <code>true</code> if the <var>string</var> matches <var>pattern</var>; <code>false</code> otherwise.
   */
  public static boolean match(String string, String pattern, boolean ignoreCase)
    {
      return match(string.toCharArray(), 0, string.length(),
                   pattern.toCharArray(), 0, pattern.length(), ignoreCase, '\\');
    }

  /**
   * Match a string against a pattern.
   * Like <code>{@link #match(String string, String pattern, boolean ignoreCase)}</code>,
   * but allows you to specify an escape character
   * other than <code>'\\'</code>. This is useful especially when trying to match strings containing
   * filenames on certain operating systems where the directory separator is <code>'\\'</code>.
   * @param string a string which is matched against the pattern.
   * @param pattern the pattern string.
   * @param ignoreCase if true, differences in case are ignored.
   * @param escapeChar an escape character to use instead of <var>'\\'</var>.
   * If <var>escapeChar</var> is -1, then no escapes are allowed.
   * @return <code>true</code> if the <var>string</var> matches <var>pattern</var>; <code>false</code> otherwise.
   */
  public static boolean match(String string, String pattern, boolean ignoreCase, int escapeChar)
    {
      return match(string.toCharArray(), 0, string.length(),
                   pattern.toCharArray(), 0, pattern.length(), ignoreCase, escapeChar);
    }

    /**
    * Test if character arrays match. This does the work for
    * <code>{@link #match(String string, String pattern, boolean ignoreCase)}</code>,
    * but is useful if you want to do matches of substrings of the input
    * string, without calling <code>string.substring(...)</code>
    * @param string an array holding a string value
    * @param sStart where to begin comparing characters within string
    * (normally, <var>sStart</var> == 0)
    * @param sEnd one past the last character to compare within string
    * (normally <code><var>sEnd</var> == <var>string</var>.length</code>).
    * @param pattern an array holding the pattern characters
    * @param pStart where in the pattern to start comparing characters
    * (normally, <code><var>pStart</var> == 0</code>)
    * @param pEnd one past the last pattern character in the pattern array.
    * (normally <code><var>pEnd</var> == <var>pattern</var>.length()</code>)
    * @param ignoreCase if true, differences in case are ignored.
    * @return match(string, sStart, sEnd, pattern, pStart, pEnd, ignoreCase, '\\')
    * @see com.sas.util.Strings#match(String,String,boolean)
    */
  public static boolean match(char string[], int sStart, int sEnd, char pattern[], int pStart, int pEnd, boolean ignoreCase)
    {
      return match(string, sStart, sEnd, pattern, pStart, pEnd, ignoreCase, '\\');
    }

    /**
    * Test if character arrays match. This does the work for
    * <code>{@link #match(String string, String pattern, boolean ignoreCase)}</code>,
    * but is useful if you want to do matches of substrings of the input
    * string, without calling <code>string.substring(...)</code>
    * @param string an array holding a string value
    * @param sStart where to begin comparing characters within string
    * (normally, <code><var>sStart</var> == 0</code>)
    * @param sEnd one past the last character to compare within string
    * (normally <code><var>sEnd</var> == <var>string</var>.length()</code>).
    * @param pattern an array holding the pattern characters
    * @param pStart where in the pattern to start comparing characters
    * (normally, <code><var>pStart</var> == 0</code>)
    * @param pEnd one past the last pattern character in the pattern array.
    * (normally <code><var>pEnd</var> == <var>pattern</var>.length</code>)
    * @param escapeChar an escape character to use instead of <code>'\\'</code>.
    * if <var>escapeChar</var> is -1, then no escapes are allowed.
    * @param ignoreCase if <code>true</code>, differences in case are ignored.
    * @return <code>true</code> if the <var>string</var> matches <var>pattern</var>; <code>false</code> otherwise.
    * @see #match(String,String,boolean)
    */
  public static boolean match(char string[], int sStart, int sEnd, char pattern[], int pStart, int pEnd, boolean ignoreCase, int escapeChar)
    {
      int sIndex = sStart;
      for (int pIndex = pStart; pIndex < pEnd; pIndex++, sIndex++)
        {
          char pChar = pattern[pIndex];
          boolean special = true;
          if (   pChar == escapeChar
              && ((pIndex + 1)) < pEnd
              && ((pattern[pIndex+1] == '*') || (pattern [pIndex+1] == '?')))
            {
              special = false;
              pIndex++;
              pChar = pattern[pIndex];
            }
          if (pChar == '?' && special)
            {
              if (sIndex == sEnd)
                return false;
              else
                return match(string, sIndex+1, sEnd, pattern, pIndex+1, pEnd, ignoreCase, escapeChar);
            }
          else if (pChar == '*' && special)
            {
              while (sIndex < sEnd)
                {
                  if (match(string, sIndex, sEnd, pattern, pIndex+1, pEnd, ignoreCase, escapeChar))
                    return true;
                  sIndex++;
                }
              return match(string, sIndex, sEnd, pattern, pIndex+1, pEnd, ignoreCase, escapeChar);
            }
          else if (sIndex < sEnd)
            {
              char sChar = string[sIndex];

              if (ignoreCase)
                {
                  sChar = Character.toUpperCase(sChar);
                  pChar = Character.toUpperCase(pChar);
                }
              if (pChar != sChar)
                return false;
            }
        }
      return sIndex == sEnd;
    }

  /**
   * Count the number of times a character occurs in a substring
   * of an input string.
   * @param inputString the input string in which we search for the character
   * @param startPos where within the input string to start searching
   * @param endPos the last position in the input string; normally this
   *        would be <code><var>inputString</var>.length()-1</code>.  If <var>endPos</var> is -1, then
   *        the default <code><var>inputString</var>.length()-1</code> is used.
   * @param c the character to search for.
   * @return the number of times the character <var>c</var> exists
   */
  public static int count(String inputString, int startPos, int endPos, char c)
  {
    int count = 0;
    if (endPos == -1)
      endPos = inputString.length()-1;
    if (startPos < 0)
      startPos = 0;
    while (startPos <= endPos)
      if (c == inputString.charAt(startPos++))
        count ++;
    return count;
  }

  /**
   * Count the number of times the search string occurs in a substring
   * of an input string.
   * @param inputString the input string in which we search for substrings
   * @param startPos where within the input string to start searching
   * @param endPos the last position in the input string; normally this
   *        would be <code><var>inputString</var>.length()-1</code>.  If <var>endPos</var> is -1, then
   *        the default <code><var>inputString</var>.length()-1</code> is used.
   * @param searchString the substring to search for
   * @param allowOverlap if <code>true</code>, allow the search string to overlap with itself
   *               in the input string. For example, searching <code>"cababababcababab"</code>
   *               for <code>"abab"</code> would find five overlapping substrings but
   *               only three non-overlapping substrings.
   * @return the number of times the <var>searchString</var> exists
   */
  public static int count(String inputString, int startPos, int endPos,
                          String searchString,
                          boolean allowOverlap)
  {
    int count = 0;
    int searchStringLength = searchString.length();
    if (searchStringLength == 0)
      return 0;
    if (searchStringLength == 1)
      return count(inputString, startPos, endPos, searchString.charAt(0));
    if (endPos == -1)
      endPos = inputString.length()-1;
    int last = endPos+1 - searchStringLength;
    while (startPos <= last)
      {
        int index = inputString.indexOf(searchString, startPos);
        if (index == -1)
          break;
        if (allowOverlap)
          startPos = index + 1;
        else
          startPos = index + searchStringLength;
        count ++;
      }
    return count;
  }

  /**
   * Find all occurrences of a character in an input string.
   * @param inputString the input string in which we search for the character
   * @param startPos where within the input string to start searching
   * @param endPos the last position in the input string; normally this
   *        would be <code><var>inputString</var>.length()-1</code>.  If <var>endPos</var> is -1, then
   *        the default <code><var>inputString</var>.length()-1</code> is used.
   * @param c the character to search for
   * @return an array of indices where <var>c</var> is found. This is an
   * empty array if no occurrences were found.
   */
  public static int[] positions(String inputString, int startPos, int endPos, char c)
    {
      int count = count(inputString, startPos, endPos, c);
      if (count == 0)
        return new int[0];
      int positions[] = new int[count];
      if (endPos == -1)
        endPos = inputString.length() - 1;
      for (int index = 0; index < count; startPos++)
        if (c == inputString.charAt(startPos))
          positions[index++] = startPos;
      return positions;
    }

  /**
   * Find all occurrences of a search string in an input string.
   * @param inputString the input string in which we search for substrings
   * @param startPos where within the input string to start searching
   * @param endPos the last position in the input string; normally this
   *        would be <code><var>inputString</var>.length()-1</code>.  If <var>endPos</var> is -1, then
   *        the default inputString.length()-1 is used.
   * @param searchString the substring to search for
   * @param allowOverlap if <code>true</code>, allow the search string to overlap with itself
   *        in the input string. For example, searching <code>"cababababcababab"</code>
   *        for <code>"abab"</code> would find five overlapping substrings but
   *        only three non-overlapping substrings.
   * @return an array of ints. This is an empty array if no occurrences
   * were found or if <var>searchString</var> is empty.
   */
  public static int[] positions(String inputString, int startPos, int endPos,
                                     String searchString,
                                     boolean allowOverlap)
  {
    /* It's not clear if the better implementation would be to call count(...)
     * first, and preallocate the vector first, with 2N calls to indexOf, or just
     * use a Vector, and only have N calls to indexOf, but also N calls to add
     * and a second loop of N copies. Since indexOf is probably O(N+M) and perhaps
     * O(N*M), I'm guessing it will outweigh the other O(N) operations, so I use
     * a Vector and only one pass through the inputString instead of two.
     * Another option is to use an array if size (inputString.length/searchString.length)
     * then truncate that array.
     */
    int searchStringLength = searchString.length();
    if (searchStringLength == 0)
      return new int[0];
    if (searchStringLength == 1)
      return positions(inputString, startPos, endPos, searchString.charAt(0));
    if (endPos == -1)
      endPos = inputString.length()-1;
    int last = endPos+1 - searchStringLength;
//    Util.tracing = true;
//    Util.println("Strings::positions Start finding positions of {0} in inputString {1}[{2},{3}] = {4}",
//                 searchString, inputString,
//                 new Integer(startPos), new Integer(endPos+1),
//                 inputString.substring(startPos, endPos+1));
    ArrayList positions = null;
    while (startPos <= last)
      {
        int index = inputString.indexOf(searchString, startPos);
        if (index == -1 || index >= endPos)
          break;
        // Util.println("Strings::positions found search string at position {0}", new Integer(index));
        if (allowOverlap)
          startPos = index + 1;
        else
          startPos = index + searchStringLength;
        if (positions == null)
          positions = new ArrayList((last - startPos + 1) / searchString.length());
        positions.add(new Integer(index));
      }
    if (positions == null)
      return new int[0];
    int v[] = new int[positions.size()];
    for (int i = positions.size()-1; i>= 0; i--)
      v[i] = ((Integer) (positions.get(i))).intValue();
    return v;
  }

	private static int POSITION_CACHE_SIZE = 10;
 	/**
	 * 
	 */
	public static int[] positions( String source, char delim )
	{
		// Buffer for the first 'n' tokens.
		// Don't allocate it until we have at least two indexes though (see below)
		int[] posCache = null;

		// Count the delimiters.
		// Keep track of the last found index in case there is only one.
		// Cache some number of indices so we don't have to go through String twice to
		// count and then store the indices in the proper size array
		int count = 0;
		int last = -1;
		{
			int i = 0;
			while ( true )
			{
				i = source.indexOf( delim, i );
				if ( i != -1 )
				{
					// Cache the index if we're within the cache size
					if ( count < POSITION_CACHE_SIZE )
					{
						// If we've found the second index, we need to make the cache array now.
						// We can wait until we've found the second since we're storing the first
						// index in 'last'.
						if ( count == 1 )
						{
							posCache = new int[POSITION_CACHE_SIZE];
							posCache[0] = last;
						}
						// Store the index if we're caching.  This won't happen if we only have one
						// index so far.
						if ( posCache != null )
						{
							posCache[count] = i;
						}
					}
					// remember the last index in case there is only one
					last = i;
					count++;
					// Skip past the search character we just found
					i++;
				}
				else
				{
					break;
				}
			}
		}

		// Handle some easy special cases
		if ( count == 0 )
		{
			return new int[0];
		}
		else if ( count == 1 )
		{
			return new int[] { last };
		}
		else if ( count == POSITION_CACHE_SIZE )
		{
			// No need to copy data or allocate a new array if we filled the cache exactly
			return posCache;
		}

		// Handle the "everything has been cached" case more efficiently
		if ( count < POSITION_CACHE_SIZE )
		{
			int[] output = new int[ count ];
			System.arraycopy( posCache, 0, output, 0, count );
			return output;
		}

		// Make the return array
		// This scans part of the source string again.
		// Make use of our cache to avoid rescanning all of the input String twice.
		int[] output = new int[ count ];
		System.arraycopy( posCache, 0, output, 0, POSITION_CACHE_SIZE );
		count -= POSITION_CACHE_SIZE;
		int index = POSITION_CACHE_SIZE;
		for (int i=posCache[POSITION_CACHE_SIZE-1]+1; count > 0; )
		{
			// scan for the search character
			i = source.indexOf( delim, i );
			// store the index and decrement the count we're looking for
			output[index++] = i;
			count--;
			// skip past the search character just found
			i++;
		}

		return output;
	}


  /**
   * Replace all occurances of search string with a replacement string in the StringBuffer.
   * @param buffer The StringBuffer to perform the replacements in
   * @param search The String to search for
   * @param replace The String to replace the occurances of search string with
   * @return The StringBuffer passed in
   */
	// This routine does not create an new objects to do it's work unless the StringBuffer passed in
	// does not have enough space in it's internal array to perform the replacements.
	public static StringBuffer replace( StringBuffer buffer, String search, String replace )
	{
//System.out.println( "buffer = " + buffer );
		int start = 0;
		int searchLength = search.length();
		while ( start < buffer.length() )
		{
			int index = buffer.indexOf( search, start );
//System.out.println( "   index = " + index );
			if ( index == -1 )
			{
				return buffer;
			}
			else
			{
//System.out.println( "   start = " + start);
//System.out.println( "   searchLength = " + searchLength );
				buffer.replace( index, index + searchLength, replace );
				start = index + replace.length();
			}
		}
		return buffer;
	}


	/**
   * Replace all occurrences of the search string with a replacement string.
   * @param inputStrings An array of strings on which the search/replace is performed.
   * @param startPos where within the input string to start searching/replacing
   * @param endPos the last position in the input string; normally this
   *        would be <code><var>inputStrings[i]</var>.length()-1</code>.  If <var>endPos</var> is -1, then
   *        the default inputStrings[i].length()-1 is used.
   * @param firstString index of first element of inputStrings to replace
   * @param lastString index of last element on inputStrings (or -1 means last) to replace
   * @param searchString the string to search for and replace with replaceString
   * @param replaceString the string which replaces the replaceString
   * @return another array of strings, with replacements performed.
   */
  public static String[] replace(String[] inputStrings, int startPos, int endPos,
                               int firstString, int lastString,
                               String searchString,
                               String replaceString)
  {
    String result[] = new String[inputStrings.length];
    if (lastString == -1) lastString = inputStrings.length - 1 ;
    for (int i = firstString; i <= lastString; i++)
      if (i < firstString || i > lastString )
        result[i] = inputStrings[i];
      else
        result[i] = replace(inputStrings[i], startPos, endPos, searchString, replaceString);
    return result;
  }


  /**
   * Replace all occurrences of the search string with a replacement string.
   * @param inputString the string on which the search/replace is performed.
   * @param startPos where within the input string to start searching/replacing
   * @param endPos the last position in the input string; normally this
   *        would be <code><var>inputString</var>.length()-1</code>.  If <var>endPos</var> is -1, then
   *        the default inputString.length()-1 is used.
   * @param searchString the string to search for and replace with replaceString.
   *        All non-overlapping occurrences of the search string in the input string
   *        are replaced. The replaced text is not a candidate for additional
   *        replacement (i.e. if the search string occurs within the replacement string).
   *        If null or "", return <var>inputString</var> without changing it.
   * @param replaceString the string which replaces the replaceString
   *        if the <var>searchString</var> equals the <var>replaceString</var>,
   *        return <var>inputString</var> without changing it.
   * @return a string with the string replacements performed.
   */
  public static String replace(String inputString, int startPos, int endPos,
                               String searchString,
                               String replaceString)
  {
    if (searchString == null)
      return inputString;
    if (replaceString == null)
      replaceString = "";
    if (searchString.length() == 0 || searchString.equals(replaceString))
      return inputString;
    if (endPos == -1)
      endPos = inputString.length()-1;
    // Util.println("Strings::replace replace {0} in inputString {1}", searchString, inputString);
    int positions[] = positions(inputString, startPos, endPos,
                                             searchString, false);
    if (positions.length == 0)
      return inputString;

    // Util.println("Strings::found {0} occurrences of inputString {1}." +  new Integer(positions.length), inputString);

    int searchStringLength  = searchString.length();
    int replaceStringLength = replaceString.length();
    int numReplacements     = positions.length;
    int targetPos           = 0;
    int inputPos            = 0;
    int resultLength        = (inputString.length()
                                - numReplacements * searchStringLength
                                + numReplacements * replaceStringLength);
    char buffer[]           = new char[resultLength];

    for (int subStringCount = 0;
         subStringCount < numReplacements;
         subStringCount++)
      {
        int subStringStartPos = positions[subStringCount];
        if (subStringStartPos > inputPos)
          {
            inputString.getChars(inputPos, subStringStartPos, buffer, targetPos);
            int copiedStringLength = (subStringStartPos - inputPos);
            targetPos += copiedStringLength;
            inputPos  += copiedStringLength;
          }
        if (replaceStringLength > 0)
          {
            replaceString.getChars(0, replaceStringLength, buffer, targetPos);
            targetPos += replaceStringLength;
          }
        inputPos += searchStringLength;
      }
    if (targetPos < resultLength)
      inputString.getChars(inputPos, inputString.length(), buffer, targetPos);
    return new String(buffer);
  }

  /**
   * Test if the specified character in a string
   * is one of the characters in the character set
   * @param s a string from which a character is tested
   * @param pos test the character from s at this position
   * @param charSet a string representing a set of characters
   * @return true if <code>charSet.indexOf(<var>s</var>.charAt(<var>pos</var>)) >= 0</code>;
   */
  public static boolean lookingAt(String s, int pos, String charSet)
  {
    return charSet.indexOf(s.charAt(pos)) >= 0;
  }

  /**
   * Test two String arrays for equality.
   * @param a an array of strings. May be <code>null</code> or contain <code>null</code> items.
   * @param b an array of strings. May be <code>null</code> or contain <code>null</code> items.
   * @return <code>true</code> if <code><var>a</var> == <var>b</var></code> (even if both are <code>null</code>)
   *         or <code>true</code> if and only if <code><var>a</var>.length==<var>b</var>.length</code> and
   *         <code><code>a</code>[i].equals(<var>b</var>[i])</code> for all <code>i</code> in <code>0 .. <code>a</code>.length</code>
   * @deprecated use equal(String a[], String b[])
   */
  public final static boolean equals(String a[], String b[])
    {
      return equal(a, b);
    }

  /**
   * Test two String arrays for equality.
   * @param a an array of strings. May be <code>null</code> or contain <code>null</code> items.
   * @param b an array of strings. May be <code>null</code> or contain <code>null</code> items.
   * @return <code>true</code> if <code><var>a</var> == <var>b</var></code> (even if both are <code>null</code>)
   *         or <code>true</code> if and only if <code><var>a</var>.length==<var>b</var>.length</code> and
   *         <code><var>a</var>[i].equals(<var>b</var>[i])</code> for all <code>i</code> in <code>0 .. <var>a</var>.length</code>
   */
  public final static boolean equal(String a[], String b[])
  {
    if (a == b)
      return true;
    if (a == null || b == null)
      return false;
    synchronized(a)
      {
        synchronized(b)
          {
            int length = a.length;
            if (length != b.length)
              return false;
            for (int i = 0; i < length; i++)
              {
                String stringA = a[i];
                String stringB = b[i];
                if (stringA == stringB)
                  continue;
                if (stringA == null || stringB == null)
                  return false;
                if ( ! stringA.equals(stringB) )
                  return false;
              }
          }
      }
    // no reason to think otherwise, so...
    return true;
  }

  /**
    * Quote a string such that it is a valid SAS string literal.
    * The <var>quoteChar</var> is
    * used to quote the string, and any occurences of the <var>quoteChar</var>
    * within the string will be replaced with two quote characters.
    * @param string an input string; may be null
    * @param quoteChar the quote character, either <code>'\''</code> or <code>'"'</code>
    * @param embed if true will embed the resuling string in outer quotes
    *              consisting of <var>quoteChar</var>
    * @return a SAS string literal useful for SAS code generation
    *         or for use in WHERE clauses, etc.
    * For example,
    * <code>quoteForSAS("O'Conner", '"', true)</code>
    * will return the ten characters <code>"O'Conner"</code>,
    * <code>quoteForSAS("O'Conner", '\'', true)</code>
    * will return the eleven character string <code>'O''Conner'</code>, and
    * <code>quoteForSAS("Tom O'Conner said, \"All is well at Mary's job.\".", '"', true)</code>
    * will return the 51 character string <code>"Tom O'Conner said, ""All is well at Mary's job""."</code>
    * @exception java.io.IllegalArgumentException if quoteChar is
    * not one of <code>'\''</code>  or <code>'"'</code>
    */
  public static String quoteForSAS(String string, char quoteChar, boolean embed)
    {
      int len = string.length();
      if (quoteChar != '\'' && quoteChar != '"')
        throw new IllegalArgumentException(MessageFormat.format(RB.getStringResource(RB_KEY, "notQuoteChar.ex.txt"),
                                                          new Object[] {new Character(quoteChar)}));
      int count = count(string, 0, -1, quoteChar);
      if (embed == false && count == 0)
        return string;
      StringBuffer buffer = new StringBuffer(len + count + 2);
      if (embed)
        buffer.append(quoteChar);
      if (count > 0)
        {
          for (int i = 0; i < len; i++)
            {
              char c = string.charAt(i);
              if (c == quoteChar)
                buffer.append(quoteChar);
              buffer.append(c);
            }
        }
      else
        buffer.append(string);
      if (embed)
        buffer.append(quoteChar);
      return buffer.toString()/*I18nOK:MTH*/;
    }

  private static char xmlSpecial[]        = { '&',     '"',      '<',    '>' };
  private static String xmlReplacements[] = { "&amp;", "&quot;", "&lt;", "&gt;" };

  /**
   * Quote a string for embedding withing XML.
   * This performs the following replacements:
   *    <table border="0">
   *    <tr><th>Replace:</th><th>With:</th></tr>
   *    <tr><td>&amp;</td><td>&amp;amp;</td></tr>
   *    <tr><td>&quot;</td><td>&amp;quot;</td></tr>
   *    <tr><td>&lt;</td><td>&amp;lt;</td></tr>
   *    <tr><td>&gt;</td><td>&amp;gt;</td></tr>
   *    </table>
   * @param string a string
   * @return the string after making the above replacements.
   */
  public static String quoteForXML(String string)
    {
      return Strings.replace(string, xmlSpecial, xmlReplacements);
    }

  /**
   * Replace special characters in a string with corresponding replacements.
   * Each character in <var>string</var> is compared agains each
   * of the <var>specialCharacters</var>. If the character is special,
   * it is replaced with the corresponding string from <var>replacementStrings</var>.
   * Thus, the <var>specialCharacters</var> and <var>replacementStrings</var> arrays
   * must be the same size.
   * @param specialCharacters an array of characters which must be replaced
   * @param replacementStrings an array of replacements values, one for each character
   * in the  <var>specialCharacters</var> array. This array cannot contain null values.
   * @return a string in which all occurrences of all the <var>specialCharacters</var>
   * have been replaced with corresponding replacements.  However, the replacements
   * themselves are <em>not</em> subject to further replacement; if they contain
   * more special characters, these values are ignored.  If none of the characters
   * in the <var>string</var> are special charactes, no replacements are performed
   * and the original <var>string</var> is returned.
   */
  public static String replace(String string, char specialCharacters[], String replacementStrings[])
    {
		if ( containsCharacters(string, specialCharacters) == false )
			return string;

		int len = string.length();

      // we could count the matches and compute the full required length
      // of the new string, but that complicates things and requires a double scan,
      // so just assume the minimum plus a 10% fudge.
      StringBuffer buffer = new StringBuffer(len * 11 / 10 );
	  int numChars = specialCharacters.length;
      int numReplacements = 0;
      for (int i = 0; i < len; i++)
        {
          char c = string.charAt(i);
          boolean replaced = false;
          for (int j = 0; j < numChars; j++)
            {
              char special = specialCharacters[j];
              if (c == special)
                {
                  replaced = true;
                  buffer.append(replacementStrings[j]);
                  numReplacements++;
                  break;
                }
            }
          if ( ! replaced )
            buffer.append(c);
        }
      if (numReplacements == 0)
        return string;
      else
        return buffer.toString()/*I18nOK:MTH*/;
    }

	/**
	 * Returns true if any of the chars in specialCharacters are present in
	 * the specified String. Otherwise, return false.
	 * 
	 * @param string The string to scan for special characters.
	 * @param specialCharacters The array of special characters.
	 * @return A boolean indicating whether any of the chars in
	 * specialCharacters are present in the specified String
	 */
	private static boolean containsCharacters(String string, char specialCharacters[])
	{
		int numChars = specialCharacters.length;
		for (int i = 0; i < numChars; i++)
		{
			if ( string.indexOf( specialCharacters[i] ) != -1 )
			{
				return true;
			}
		}

		return false;
	}

  /**
    * Quote a string with double quotes such that it is a valid SAS string literal.
    * @param string an input string; may be null
    * @return a SAS string literal (embedded in double quotes) useful for SAS code generation
    *         or for use in WHERE clauses, etc. This method simply calls
    *         <code>quoteForSAS(string, '"', true)</code>
    * @see #quoteForSAS(String, char, boolean)
    */
  public static String quoteForSAS(String string)
    {
      return quoteForSAS(string, '"', true);
    }

  /**
    * Quote a string such that its printed representation
    * is a valid Java literal.
    * @param string an input string; may be <code>null</code>
    * @return a Java string literal useful for code generation.
    */
  public static String quote(String string)
    {
      if (string == null)
        return "null";
      else
        {
          int length = string.length();
          StringBuffer buffer = new StringBuffer(length + 4);
          buffer.append('"'); // open quote
          for (int i = 0; i < length; i++)
            {
              char c = string.charAt(i);
              if (c == '"' || c == '\\')
                {
                  buffer.append('\\');
                  buffer.append(c);
                }
              else if (c == '\b') buffer.append("\\b");
              else if (c == '\f') buffer.append("\\f");
              else if (c == '\n') buffer.append("\\n");
              else if (c == '\r') buffer.append("\\r");
              else if (c == '\t') buffer.append("\\t");
              else if ( Character.isISOControl(c) || (c > 255) )
                {
                  String hex = Integer.toHexString((int) c);
                  // Util.println("ISO control c: {0}, hex {1} at {2} in string \"{3}\"", new Character(c), hex, new Integer(i), string);
                  buffer.append("\\u");
                  buffer.append("000".substring(hex.length() - 1));
                  buffer.append(hex);
                }
              else
                buffer.append(c);
            } // end for
          buffer.append('\"'); // close quote
          return buffer.toString()/*I18nOK:MTH*/;
        }
    }

  /**
    * Test is a string should be quoted when printed.
    * @return true if the string contains leading or trailing blanks
    * or any special characters.
    */
  public static boolean needsQuoting(String string)
    {
      if (string == null)
        return false;
      int length = string.length();

      if (   (length > 0)
          && (   Character.isWhitespace(string.charAt(0))
              || Character.isWhitespace(string.charAt(length-1))))
        return true;

      for (int i = 0; i < length; i++)
        {
          char c = string.charAt(i);
          if (   (c == '\"')
              || (c == '\\') || (c == '\b') || (c == '\f')
              || (c == '\n') || (c == '\r') || (c == '\t')
              || (Character.isISOControl(c)
              || ( c > 255)))
            return true;
        }
      return false;
    }


  /**
   * Remove all white space, including interior white space, from a string.
   * @param string a String.
   * @param a new string, derived from <var>string</var> with the white space removed.
   * @see java.lang.String#trim
   */
  public static String removeWhitespace(String string)
    {
      int len = string.length();
      StringBuffer newString = new StringBuffer(len);
      for ( int i = 0; i < len; i++ )
        {
          char c = string.charAt(i);
          if ( ! Character.isWhitespace( c ) )
            newString.append(c);
        }
      return newString.toString()/*I18nOK:MTH*/;
    }

  /**
   * Return the length of a trimmed string. This method
   * is roughly equivalent to calling <code><var>string</var>.{@link
   * java.lang.String#trim() trim()}.{@link java.lang.String#length() length()}</code>
   * but is more efficient; it does not have to
   * allocate a new String object (which is what trim()</code> may do).
   * The behavior is not exactly the same; this method trims whitespace
   * according to {@link java.lang.Character#isWhitespace(char)}, whereas
   * <code>String.trim()</code> trims whitespace and control characters.
   * @param string a string whose trimmed length you want to compute.
   *        If <var>string</var> is null, return 0.
   * @return The length of <var>string</var> with leading and trailing
   *         whitespace removed.
   * @see java.lang.String#trim()
   */
  public static int trimmedLength(String string)
    {
      if (string == null)
        return 0;
      int left = 0, right = string.length()-1;
      while (left < right && Character.isWhitespace(string.charAt(left)))
        left++;
      while (right >= left && Character.isWhitespace(string.charAt(right)))
        right--;
      return right - left + 1;
    }

  /**
   * Compare two String values, optionally trimming the strings and ignoring case.
   * @param a a String to compare to <var>b</var>
   * @param b a String to compare to <var>a</var>
   * @param trimStrings if <code>true</code>, compare the strings as if leading and trailing
   * white space in both strings <var>a</var> and <var>b</var>
   * were first trimmed.
   * @param ignoreCase if <code>true</code>, compare the relevant characters
   * accoring to the rules of {@link java.lang.String#equalsIgnoreCase}
   * @return true if <code>(<var>a</var> == <var>b</var>)</code> or if
   * the two strings <var>a</var> and <var>b</var>
   * are equal under the conditions specified by <var>trimStrings</var>
   * and <var>ignoreCase</var>
   * @see java.lang.String#equalsIgnoreCase(java.lang.String)
   * @see java.lang.String#trim()
   */
  public static boolean equal(String a, String b,
                              boolean trimStrings,
                              boolean ignoreCase)
    {
      if (a == null || b == null)
        return a == b;
      int lenA = a.length();
      int lenB = b.length();
      int startA = 0, startB = 0;
      int len;
      if (trimStrings)
        {
          int trimmedLenA = trimmedLength(a);
          int trimmedLenB = trimmedLength(b);
          if (trimmedLenA != trimmedLenB)
            return false;
          len = trimmedLenA;
          while (startA < lenA && Character.isWhitespace(a.charAt(startA)))
            startA++;
          while (startB < lenB && Character.isWhitespace(b.charAt(startB)))
            startB++;
        }
      else
        {
          if (lenA != lenB)
            return false;
          len = lenA;
        }
      return a.regionMatches(ignoreCase, startA, b, startB, len);
    }

  private static Hashtable ht = new Hashtable();
  /**
   * Intern a string, similar but unrelated to {@link java.lang.String#intern()}.
   * This method does not have the internal limits of <code>String.intern</code>
   * found in some Java Virtual Machines.
   * @param string a non-null string to intern
   * If the <var>string</var> has already been interned
   * (or another <code>String</code> that is equals() to this one),
   * return that previously interned string.
   * If the string has not been interned already, it is
   * interned and returned.
   */
  public static String intern(String string)
    {
      String interned = (String) ht.get(string);
      if (interned != null)
        return interned;
      ht.put(string, string);
      return string;
    }

  /**
   * Test if a string has been interned.
   * @param string a non-null string
   * @return true if <var>string</var> has been interned,
   *         false if not.
   */
    public boolean interned(String string)
    {
      return ht.get(string) != null;
    }

  /**
   * Concatenate the strings in a string array
   * into one string, with a delimiter after each
   * string.
   * @param strings an array of strings.
   *        null items are allowed.
   * @param prefix a prefix string which is
   *        concatenated first, if non-null.
   *        The delimiter does not appear after the prefix.
   * @param delimiter string to be placed after
   *        each string in <var>array</var>
   * @param suffix a prefix string
   * @param prefix a suffix string which is
   *        concatenated last, if non-null.
   *        The delimiter does not appear before the prefix.
   * @see #split(String string, String prefix, String delimiter, String suffix)
   */
  public static String concatenate(String[] strings,
                                   String prefix, String delimiter, String suffix )
  {
    if (strings == null || strings.length == 0)
      return "";
    // Quickly count the string lengths for efficient
    // StringBuffer allocation, to avoid internal reallocation.
    int numStrings = strings.length;
    if (delimiter == null) delimiter = "";
    if (prefix == null)    prefix = "";
    if (suffix == null)    suffix = "";
    int len = prefix.length() + (numStrings * delimiter.length()) + suffix.length();
    for (int i = 0; i < numStrings; i++)
      {
        String s = strings[i];
        len += (s == null ? 0 : s.length());
      }

    StringBuffer b = new StringBuffer(len);
    b.append(prefix);
    for (int i = 0, numStrings_1 = numStrings - 1 ; i < numStrings; i++)
      {
        String s = strings[i];
        b.append(s == null ? "" : s);
        if (i < numStrings_1)
          b.append(delimiter);
      }
    b.append(suffix);
    return b.toString()/*I18nOK:MTH*/;
  }

  /**
   * Split a string into an array of strings.
   * This method is the inverse of the {@link #concatenate(String[],String,String,String)}
   * method.
   * Ignoring the <var>prefix</var>
   * and <var>suffix</var> strings, the remainder of <var>string</var>
   * is split at the <var>delimiter</var>.
   * For example,
   * <code>Strings.split("[a==b==c==d]", "[", "==", "]")</code>
   * yields the array <code>{"a", "b", "c", "d"}</code>
   * @param string a string to be split. This string must be non-null
   * @param prefix a prefix of <var>string</var> to ignore.
   *               The prefix may be null or "". Note that
   *               with <code>Strings.split("/usr/local/bin", "/", "/", null)</code>
   *               yields the array <code>{"usr", "local", "bin"}</code> because the first
   *               "/" substring prefix is ignored, whereas
   *               <code>Strings.split("/usr/local/bin", null, "/", null)</code>
   *               yields the array <code>{"", "usr", "local", "bin"}</code> because the first
   *               "/" substring is not ignored and the string before it is returned.
   * @param delimiter a non-null string that separates the substrings to extract.
   *                  Note that unlike {@link java.util.StringTokenizer}, this is not
   *                  a string of single character delimiters. In the first example, above,
   *                  the delimiter was the two character string, "==".
   * @param suffix A suffix substring of <var>string</var> to ignore.
   *               The prefix may be null or "".
   * @exception NullPointerException if <var>string</var> or <var>delimiter</var> is null.
   * @see #concatenate(String[],String,String,String)
   */
  public static String[] split(String string, String prefix, String delimiter, String suffix)
  {
    // System.out.println("split("+string+", "+prefix +", " +delimiter+ ", "+suffix+")");
    int start = 0;
    int end = string.length() - 1;
    int delimLen = delimiter.length();
    if (delimLen == 0)
      throw new IllegalArgumentException("\"" + delimiter + "\".length() == 0");
    if (prefix == null || prefix.length() == 0)
      start = 0;
    else if (string.startsWith(prefix))
      start = prefix.length();
    if (   suffix != null
        && suffix.length() != 0
        && string.endsWith(suffix))
      end -= suffix.length();
    if (end < start)
      return new String[0];
    int positions[] = positions(string, start, end, delimiter, false);
    int n = positions.length;
    for (int i = 0; i < n; i++)
      {
        int pos = positions[i];
        // System.out.println("positions[" + i + "] = " + pos);
      }

    String result[] = new String[n+1];
    if (n == 0)
      {
        result[0] = string.substring(start, end+1);
        // System.out.println("result[0] = " + result[0]);
      }
    else
      {
        int indices[] = new int[n+2];
        indices[0] = start;
        System.arraycopy(positions, 0, indices, 1, n);
        indices[n+1] = end+1;
        for (int i = 0, offset = 0; i <= n; i++)
          {
            // System.out.println("\"" + string + "\".substring(" + indices[i] + "+" + offset + ", " + indices[i] + ")");
            result[i] = string.substring(indices[i] + offset, indices[i+1]);
            // System.out.println("result[" + i + "] = " + result[i]);
            offset = delimLen;
          }
      }
    return result;
  }

  /**
   * Split the string into segments of <var>segmentLength</var> characters each.
   * The last segment may be less than <var>segmentLength</var> characters if
   * <var>padChar</var> is null, otherwise it is padded on the right with
   * the <var>padChar</var>. Note that this method is <em>not</em>
   * locale aware and
   * may break words and may even break "double" characters such as
   * a "ch" in a Spanish locale. Instead, this
   * method treats the string as a sequence of individual characters.
   * Use a {@link java.text.BreakIterator} for splitting text using
   * Locale aware boundaries.
   * For example, <code>Strings.split("JanFebMarAprMayJunJulAugSepOctNovDec", 3, null)</code>
   * will return <code>{"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
   * "Aug", "Sep", "Oct", "Nov", "Dec", }</code>. (But you should not use this to get
   * a list of month abbreviations, since it is not Locale/Calendar aware!)
   * @param string the string to split
   * @param segmentLength the size of the substrings to extract from the <var>string</var>
   * @param padChar If not null, use {@link Character#charValue() <var>padChar</var>.charValue()}
   * as the pad character to pad the final string to a full <var>segmentLength</var>,
   * if it would otherwise be less than <var>segmentLength</var> characters.
   * For example, <code>Strings.split("This is a sentence split into 8 character segments.", 8, Strings.SPACE)</code>
   * returns  the 7 segments
   * <code>{"This is ", "a senten", "ce split", " into 8 ", "characte", "r segmen", "ts.     "  }</code>
   * while <code>Strings.split("This is a sentence split into 8 character segments.", 8, null)</code>
   * returns the 7 segments
   * <code>{"This is ", "a senten", "ce split", " into 8 ", "characte", "r segmen", "ts."  }</code>
   */
  public static String[] split(String string, int segmentLength, Character padChar)
  {
    if (segmentLength < 1)
      throw new IllegalArgumentException("segmentLength " + segmentLength + " < 1");
    int len = string.length();
    int numSegments = (len / segmentLength) + (((len % segmentLength) == 0) ? 0 : 1);
    String result[] = new String[numSegments];
    for (int i = 0, start = 0, end = segmentLength;
         i < numSegments;
         i++, start+=segmentLength, end+=segmentLength)
      {
        result[i] = string.substring(start, Math.min(end, len));
        // System.out.println("result[" + i + "] = " + result[i]);
      }
    if (padChar != null)
      {
        StringBuffer buffer = new StringBuffer(result[numSegments - 1]);
        for (int i = buffer.length(); i < segmentLength; i++)
          buffer.append(padChar.charValue());
        result[numSegments - 1] = buffer.toString()/*I18nOK:MTH*/;
        // System.out.println("result[" + (numSegments - 1) + "] = " + result[(numSegments - 1)]);
      }
    return result;
  }

   /**
    * Find the position of a string in an array of strings.
    * @param string a string to search for. This String may be null.
    * @param strings an array of strings in which to search for <var>string</var>.
    *        Items in the array may be null, but the array itself cannot be null.
    * @return the position where <var>string</var> was found, or -1 if not found
    */
   public static int indexOf(String string, String[] strings)
   {
     return indexOf(string, strings, 0, strings.length, null);
   }

   /**
    * Find the position of a string in an array of strings.
    * @param string a string to search for. This String may be null.
    * @param strings an array of strings in which to search for <var>string</var>.
    *        Items in the array may be null, but the array itself cannot be null.
    * @param startPos the starting position within <var>strings</var>
    *        to begin the search. Use 0 to start at the beginning.
    * @param endPos One past the last position to compare. Use <var>strings.length</var>
    *        to search all items.
    * @param Comparator comparator a comparator to use for searching.
    *        If null, {@link Util#equal(Object,Object)} is used.
    * @return the position where <var>string</var> was found, or -1 if not found
    */
   public static int indexOf(String string, String[] strings, int startPos, int endPos, Comparator comparator)
   {
     if (comparator == null)
       for (int i = startPos; i < endPos; i++)
         {
           if (Util.equal(string, strings[i]))
             return i;
         }
     else
       for (int i = startPos; i < endPos; i++)
         {
           if (0 == comparator.compare(string, strings[i]))
             return i;
         }
     return -1;
   }


}
