package com.sas.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.UndeclaredThrowableException;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.security.PrivilegedActionException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.ByteArrayOutputStream;
import com.sas.util.Classes;

/**
 * UncheckedException is a base exception for exceptions which
 * &quot;wrap&quot; or chain other other exceptions.
 * UncheckedException, which extends from {@link java.lang.RuntimeException},
 * is an unchecked exception.
 * For example, if you wish to implement some interface which
 * has methods with no checked exceptions, but the implementation
 * invokes methods which throw checked exceptions (such as {@link java.io.IOException}),
 * you can wrap those exceptions in an <code>UncheckedException</code> and throw
 * the UncheckedException. Note that since callers of the method do
 * not excpect any exceptions, they may not have a catch clause to
 * catch <code>UncheckedException</code>.
 *
 * The exception wrapped by an <code>UncheckedException</code> exception may be a checked or an unchecked exception.
 * <p>
 * You may use {@link #getRootException()} to get the
 * underlying or root cause from which this <code>UncheckedException</code>
 * was constructed.
 * <p>
 * UncheckedException contains some static methods for printing
 * the entire stack chain. You may print the exception chain
 * to either a {@link PrintStream}, {@link PrintWriter}, or a {@link StringBuffer}.
 * Be careful to not reimplement a UncheckedException subclass by changing
 * one of it's {@link Throwable#printStackTrace()} methods to call
 * one of these static methods, since that will result in an infinite
 * loop (because those methods call back to the base <code>printStackTrace()</code>
 * methods).
 * <p>
 * @see ChainedException
 * @see java.lang.reflect.InvocationTargetException
 * @see java.lang.reflect.UndeclaredThrowableException
 * @see java.sql.SQLException
 * @see java.sql.SQLWarning
 * @author sasdjb <A HREF="http://sww.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */
public class UncheckedException
       extends RuntimeException
       implements ChainedExceptionInterface
{


  //////////////////////////////// Instance Variables ////////////////////////////////

  private Throwable uncheckedException;
  private boolean printing;


  //////////////////////////////// Constructors ////////////////////////////////

  /**
   * Create a UncheckedException that contains no other exception
   * or exception message.
   */
  public UncheckedException()
  {
  }

  /**
   * Create a UncheckedException that is not based on another exception.
   * @param message the text describing the exception.
   * @see Throwable#getMessage()
   */
  public UncheckedException(String message)
  {
    super(message);
  }


  //////////////////////////////// Instance Methods ////////////////////////////////
  /**
   * Create a UncheckedException based on another exception.
   * This is the only way to set the target exception that
   * is returned by {@link #getRootException()}.
   * @param rootException An exception which this
   * UncheckedException will chain to via {@link #getRootException()}.
   * It may itself be
   * another UncheckedException.
   * @param message the text describing the exception.
   * If null, then the <var>rootException</var>'s
   * message is used.
   * @see Throwable#getMessage()
   * @see #getTargetException()
   */
  public UncheckedException(Throwable rootException, String message)
  {
    super(  (message == null && rootException == null)
          ? null
          : (  (message == null)
             ? rootException.getMessage()
             : message));
    this.uncheckedException = rootException;
  }

  /**
   * Return the exception with which this ChainedExceptionInterface was constructed
   * via {@link #UncheckedException(Throwable, String)}.
   */
  public Throwable getTargetException()
  {
    return uncheckedException;
  }


  /**
   * Return the root cause of this UncheckedException.
   * @param the underlying exception; i.e. the exception that
   * has no other chained/invocation/next exception.
   */
  public Throwable getRootException()
  {
    return ChainedException.getRootException(this);
  }

  /**
   * Return this <code>UncheckedException</code> as a <code>String</code>
   * @return the value of <code>toStringBuffer(this, new StringBuffer()).toString()</code>
   */
  public String toString()
  {
    return ChainedException.toStringBuffer(this, new StringBuffer()).toString()/*I18nOK:MTH*/;
  }


}
