/* $Id$ */
// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util;

import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.ByteArrayOutputStream;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
import java.util.MissingResourceException;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.Vector;
import java.util.ResourceBundle;
import java.util.Locale;

/**
 * General purpose utility functions not found elsewhere.
 * @author <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 */

public class Util
{
  public static final String RB_KEY = "Util.";
  private static long commandCount;

  // private constructor prevents insantiation
  private Util()
  {
  }

  /**
   * Compare two objects for equality, with tests for null as well.
   * @param a one object
   * @param b another object
   * @return <code>true</code> if <code><var>a</var> ==
   *         <var>b</var></code> or
   *         <code><var>a</var>.equals(<var>b</var>)</code>,
   *         <code>false</code> if either (but not both)
   *         is <code>null</code>.
   */
  public static final boolean equal(Object a, Object b)
  {
    return (   (a == b)
            || (   (a != null)
                && (b != null)
                   && ( (a.getClass().isArray() && b.getClass().isArray())
                       ? arrayEqual(a, b)
                       : a.equals(b))));
  }

  /**
   * Compare two objects for inequality, with tests for null as well.
   * @param a one object
   * @param b another object
   * @return <code>false</code> if <code><var>a</var> == <var>b</var></code>
   *         or <code><var>a</var>.equals(<var>b</var>)</code>, <code>true</code> if either (but not both)
   *         is <code>null</code>.
   */
  public static final boolean unequal(Object a, Object b)
  {
    return !equal(a,b);
  }

  /**
   * Compare two Vectors for equality, with tests for <code>null</code> as well.
   * This performs an element-wise comparison.
   * @param v1 a Vector
   * @param v2 another Vector
   * @return <code>true</code> if <var>v1</var><code>==</code><var>v2</var>
   *         or if the vectors each have the same size and
   *         corresponding elements of each vector are equals
   *         <code>false</code> otherwise.
   */
  public static final boolean equal(Vector v1, Vector v2)
  {
    if (v1 == v2)
      return true;
    if (v1 == null || v2 == null || (v1.size() != v2.size()))
      return false;
    synchronized(v1) {
      synchronized(v2) {
        for (int i = 0, v1Len = v1.size(); i < v1Len; i++)
          {
            Object o1 = (Object) v1.elementAt(i);
            Object o2 = (Object) v2.elementAt(i);
            if (unequal(o1, o2))
              return false;
          }
      }
    }
    return true;
  }

  /**
   * Compare two Enumerations for equality, with tests for null as well.
   * @param e1 an Enumeration
   * @param e2 an Enumeration
   * @return <code>true</code> if the <var>e1</var><code>==</code><var>e2</var>, or if items in each enumeration are equal
   *         and both enumerations return the same number of items;
   *         <code>false</code> otherwise.
   */
  public static final boolean equal(Enumeration e1, Enumeration e2)
  {
    if (e1 == e2)
      return true;
    if (e1 == null || e2 == null)
      return false;
    while (e1.hasMoreElements() && e2.hasMoreElements())
      {
        Object a = e1.nextElement();
        Object b = e2.nextElement();
        if (unequal(a, b) )
          return false;
      }
    return ( ! e1.hasMoreElements() && ! e2.hasMoreElements() );
  }

  /**
   * Controls whether tracing is enabled or not.
   * This is public; there are no accessor methods;
   * just go ahead and set it: <code>Util.tracing = true;</code>
   * if you want to enable tracing. By default, <code>tracing</code>
   * is disabled <code>(tracing==false)</code>
   */
  public static boolean tracing = false;

  /*
   * ... By default, tracing is disabled (tracing==false) unless the application
   * is run with the DEBUGTRACE property variable
   * set to "true". i.e. invoke with
   * <pre>
   *    java -DDEBUGTRACE=true
   * <pre>
   *
   * but the DEBUGTRACE check was removed (see below) due to a MS VM
   * problem. sasdjb 31Mar98
   */

  /**
   * Stream where <code>Util.print</code> and <code>Util.println</code> output goes.
   * @deprecated use printWriter instead
   */
  public static final PrintStream printStream = System.err;

  /**
   * Stream where println output goes.
   * This is public; there are no accessor methods;
   * just go ahead and set it: <code>Util.printStream = System.out;</code>
   * for example. The <code>println</code> methods flush the <code>printWriter</code>
   * after writing. <code>printWriter</code> defaults to a <code>printWriter</code>
   * based on <a href="../../../java/lang/System.html#err">System.err</a>
   */
  public static PrintWriter printWriter = new PrintWriter(new OutputStreamWriter(System.err));

  /**
   * Flush the printWriter if tracing is enabled
   * @see #tracing
   * @see #printWriter
   */
  public static void flush()
    {
      if (!tracing) return;
      checkPW();
      printWriter.flush();
    }

    private static void checkPW()
    {
      if (printWriter == null && printStream != null)
        printWriter = new PrintWriter(printStream);
    }

  /**
   * Print a string to printWriter if tracing is enabled.
   * @param string a message to print
   * @see #tracing
   * @see #printWriter
   */
  public static final void print(String message)
  {
    if (!tracing) return;
    checkPW();
    printWriter.print(message);
    flush();
  }

  /**
   * Print a string and newline to printWriter if tracing is enabled.
   * @param string a message to print
   * @see #tracing
   * @see #printWriter
   */
  public static final void println(String message)
  {
    if (!tracing) return;
    checkPW();
    printWriter.println(message);
    flush();
  }

  /**
   * Print a newline to printWriter if tracing is enabled.
   * @see #tracing
   * @see #printWriter
   */
  public static final void println()
  {
    if (!tracing) return;
    checkPW();
    printWriter.println();
    flush();
  }

  /**
   * If tracing is enabled,
   * format and print a message by applying the format directives
   * in the unformatted text to the format args.
   * The existing format args in the object, if any,
   * are replaced with the new args.
   * @param pattern the new raw format string
   * @param arg0 First value to be formatted with the message.
   * @see #tracing
   * @see #printWriter
   */
  public static final void println(String pattern, Object arg0)
  {
    if (!tracing) return;
    printWriter.print( MessageFormat.format( pattern, new Object[] {arg0} ));
    flush();
  }

  /**
   * If tracing is enabled,
   * format and print a message by applying the format directives
   * in the unformatted text to the format args.
   * @param pattern the new raw format string
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @see #tracing
   * @see #printWriter
   */
  public static final void println(String pattern, Object arg0, Object arg1)
  {
    if (!tracing) return;
    printWriter.print( MessageFormat.format( pattern, new Object[] {arg0, arg1} ));
    flush();
  }

  /**
   * If tracing is enabled,
   * format and print a message by applying the format directives
   * in the unformatted text to the format args.
   * @param pattern the new raw format string
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @see #tracing
   * @see #printWriter
   */
  public static final void println(String pattern, Object arg0, Object arg1, Object arg2)
  {
    if (!tracing) return;
    printWriter.print( MessageFormat.format( pattern, new Object[] {arg0, arg1, arg2} ));
    flush();
  }

  /**
   * If tracing is enabled,
   * format and print a message by applying the format directives
   * in the unformatted text to the format args.
   * @param pattern the new raw format string
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @see #tracing
   * @see #printWriter
   */
  public static final void println(String pattern,
                       Object arg0, Object arg1, Object arg2, Object arg3)
  {
    if (!tracing) return;
    printWriter.print( MessageFormat.format( pattern, new Object[] {arg0, arg1, arg2, arg3} ));
    flush();
  }

  /**
   * If tracing is enabled,
   * format and print a message by applying the format directives
   * in the unformatted text to the format args.
   * @param pattern the new raw format string
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @see #tracing
   * @see #printWriter
   */
  public static final void println(String pattern,
                       Object arg0, Object arg1, Object arg2, Object arg3, Object arg4)
  {
    if (!tracing) return;
    printWriter.print( MessageFormat.format( pattern, new Object[] {arg0, arg1, arg2, arg3, arg4} ));
    flush();
  }

  /**
   * If tracing is enabled,
   * format and print a message by applying the format directives
   * in the unformatted text to the format args.
   * @param pattern the new raw format string
   * @param arg0 First value to be formatted with the message.
   * @param arg1 Next value to be formatted with the message.
   * @param arg2 Next value to be formatted with the message.
   * @param arg3 Next value to be formatted with the message.
   * @param arg4 Next value to be formatted with the message.
   * @param arg5 Next value to be formatted with the message.
   * @see #tracing
   * @see #printWriter
   */
  public static final void println(String pattern,
                       Object arg0, Object arg1, Object arg2, Object arg3, Object arg4, Object arg5)
  {
    if (!tracing) return;
    printWriter.print( MessageFormat.format( pattern, new Object[] {arg0, arg1, arg2, arg3, arg4, arg5} ));
    flush();
  }

// Doing below causes security problems in MS VM in a browser;
// it prints an exception traceback and causes some other
// error that the catch block does not catch.
// disabled, sasdjb 31Mar98
// As a workaround, set com.sas.util.tracing = true;
// Or, try java com.sas.util.Debug myclass args
// (Debug.java is in $/SAS/Test/com/sas/util/Debug.java)

//   static
//   {
//     // See if we can get system properties and test for a
//     // DEBUG environment variable. This will throw a
//     // SecurityException when run in an applet, so
//     // we allow for that.
//     try
//       {
//         String debug = System.getProperty("DEBUGTRACE").toUpperCase();
//         tracing =  (   (debug != null)
//                     && (   debug.equals("TRUE")
//                         || debug.equals("T")
//                         || debug.equals("Y")
//                         || debug.equals("YES")));
//       }
//     catch (Exception e)
//       { }
//   }

  /**
    * Get the millisecond value of a calendar. This is the
    * number of milliseconds since the
    * <a href="com.sas.util.TimerInterface#Epoch">Epoch</a>
    * @param calendar a Calendar representing a date/time
    * @return the number of milliseconds between the Epoch
    * and the <var>calendar</var>. This is negative if the
    * <var>calendar</var> is before the Epoch.
    */
  public static long toMillis(Calendar calendar)
    {
      return calendar.getTime().getTime();
    }

  /**
    * Calculate the number of milliseconds until a <code>java.util.Calendar</code>
    * date/time.
    * @param calendar a java.util.Calendar representing a date/time
    * @return the number of milliseconds between the current time and the calendar.
    */
  public static long millisUntil(Calendar calendar)
    {
      long currentTime = System.currentTimeMillis();
      return calendar.getTime().getTime() - currentTime;
    }


    /////////////////////////////////////////////////////////////////////
    // Resource Bundles and Internationalization
    /////////////////////////////////////////////////////////////////////

  /**
   * Return a string resource from the common resource bundle.
   * (This is the String resource bundle shared across the com.sas
   * component hierarchy).
   * @param key a resource key
   * @return String a string resource obtained from the key.
   */
   static public String getCommonStringResource(String key)
   {
      return getCommonResources().getString(key);
   }

   static private ResourceBundle _commonResourceBundle;
   /**
    * Get the common resource bundle.
    * (This is the String resource bundle shared across the com.sas
    * component hierarchy).
    * @return the default resource bundle.
    */
   static public ResourceBundle getCommonResources()
   {
     if (_commonResourceBundle == null)
        _commonResourceBundle = ResourceBundle.getBundle("com.sas.util.CommonResources");
     return _commonResourceBundle;
   }

  /**
   * Copy an array. This creates a new array that has the same elements
   * as the input <var>array</var> items
   * @param array an array to copy.
   * @return a new array of the same type as the input <var>array</var>,
   * populated with all the items from the input <var>array</var>
   * The array is not copied recursively.
   * @exception java.lang.IllegalArgumentException if <var>array</var> is not really an array
   */
  public static Object copyArray(Object array)
    {
      return copyArray(array, 0, Array.getLength(array));
    }

  /**
   * Copy an array. This creates a new array that has the same elements
   * as the input <var>array</var> items
   * <code><var>array</var>[<var>start</var>]</code> through
   * <code><var>array</var>[<var>end</var> - 1]</code>
   * @param array an array to copy.
   * @param start the index of the first item from <var>array</var> to copy.
   * @param end one greater than the index of the last item from <var>array</var> to copy.
   * @return a new array of the same type as the input <var>array</var>,
   * populated with items <code><var>array</var>[<var>start</var>]</code> through <code><var>array</var>[<var>end</var> - 1]</code>
   * (if <var>start</var> == <var>end</var>, the new array has length == 0)
   * @exception java.lang.IllegalArgumentException if <var>array</var> is not really an array,
   * or if <var>start</var> or <var>end</var> are not in the range <code>[0, <var>array</var>.length-1]</code>, or if
   * <var>end</var> < 0
   */
  public static Object copyArray(Object array, int start, int end)
    {
      Object newArray = null;
      Class componentType = null;
      try
        {
          componentType = array.getClass().getComponentType();
          newArray = Array.newInstance(componentType, end - start);
          System.arraycopy(array, start, newArray, 0, end - start);
        }
      catch (NullPointerException e)
        {
          throw e;
        }
      catch (RuntimeException e)
        {
          if (componentType == null)
            throw new IllegalArgumentException(RB.getStringResource(RB_KEY, "notArray.ex.txt"));
          int length = Array.getLength(array);
          if (start < 0 || start >= length)
            throw new IllegalArgumentException(MessageFormat.format(RB.getStringResource(RB_KEY, "index.ex.txt"),
                                                                new Object[] {new Integer(start), new Integer(0), new Integer(length)} ));
          if (end < start || end >= length)
            throw new IllegalArgumentException(MessageFormat.format(RB.getStringResource(RB_KEY, "index.ex.txt"),
                                                                new Object[] {new Integer(end), new Integer(start), new Integer(length)} ));
        }

      return newArray;
    }

  /**
   * Resize an object array to a new size.
   * @param array an object array
   * @param newSize the new array size. If the new size
   * is larger than <code><var>array</var>.length</code>, the new items will be <code>null</code>.
   */
  public static Object[] resizeArray(Object array[], int newSize)
    {
      Object result[] = new Object[newSize];
      System.arraycopy(array, 0, result, 0, Math.min(newSize, array.length));
      return result;
    }


 /**
   * Determines if a given array contains a given object.
   * @param array The array to search.
   * @param obj The object to search for; may be <code>null</code>.
   * @return <code>true</code> if <var>array</var> contains <var>obj</var>,
   *    and <code>false</code> otherwise.
  */
 public static boolean arrayContains (Object[] array, Object obj)
    {
    for (int i = 0; i < array.length; i++)
       if (equal (array[i], obj))
          return true;

    return false;
    }


  /**
   * Test two arrays for equality.
   * It uses {@link #unequal(java.lang.Object, java.lang.Object)
   * <code>Util.unequal(<var>arrayA</var>[<var>i</var>], <var>arrayB</var>[<var>i</var>])</code>}
   * to compare each pair of objects in the
   * array (if the arrays are object arrays), else the primitive values are compared directly.
   * If <code><var>arrayA</var>.getClass() != <var>arrayB</var>.getClass()</code>,
   * returns <code>false</code>
   * In case of primitve float and double values, objects are compared using
   * == method, so if the arrays conatin any <code>NAN</code>
   * values the result will be false.
   * eg.
   * <pre>
   * double[] a = {Double.NaN};
   * double[] b = {Double.NaN};
   * </pre>
   * The result of <code>arrayEqual(a, b)</code> is <code>false</code>.
   * @param arrayA an array of objects. Can be <code>null</code> or contain <code>null</code>s.
   * @param arrayB an array of objects. Can be <code>null</code> or contain <code>null</code>s.
   * @return <code>true</code> if the two arrays are equal, <code>false</code> otherwise.
   * @Exception java.lang.IllegalArgumentException if either object passed in
   *            is not an array or null.
   **/
  public static boolean arrayEqual(Object arrayA, Object arrayB)
    {
      if (arrayA == arrayB )
        return true;
      if (arrayA == null || arrayB == null)
        return false;

      synchronized (arrayA)
        {
          synchronized (arrayB)
            {
              Class classA = arrayA.getClass();
              Class classB = arrayB.getClass();
              if (!(classA.isArray()) && !(classB.isArray()))
                throw new java.lang.IllegalArgumentException(RB.getStringResource(RB_KEY, "notArray.ex.txt"));
              if (classA != classB)
                return false;

              Class classComponentA = classA.getComponentType();
              Class classComponentB = classB.getComponentType();
              if (classComponentA != classComponentB)
                return false;

              if (classComponentA.isPrimitive())
                {
                  if (classComponentA == Integer.TYPE)
                    return arrayEqual((int[])arrayA, (int[])arrayB);
                  else if (classComponentA == Boolean.TYPE)
                    return arrayEqual((boolean[])arrayA, (boolean[])arrayB);
                  else if (classComponentA == Character.TYPE)
                    return arrayEqual((char[])arrayA, (char[])arrayB);
                  else if (classComponentA == Byte.TYPE)
                    return arrayEqual((byte[])arrayA, (byte[])arrayB);
                  else if (classComponentA == Short.TYPE)
                    return arrayEqual((short[])arrayA, (short[])arrayB);
                  else if (classComponentA == Long.TYPE)
                    return arrayEqual((long[])arrayA, (long[])arrayB);
                  else if (classComponentA == Float.TYPE)
                    return arrayEqual((float[])arrayA, (float[])arrayB);
                  else if (classComponentA == Double.TYPE)
                    return arrayEqual((double[])arrayA, (double[])arrayB);
                }

              else if ( classComponentA.isArray() && classComponentB.isArray() )
                {
                  int lenA = Array.getLength(arrayA);
                  int lenB = Array.getLength(arrayB);
                  if (lenA != lenB)
                    return false;
                  for (int i = 0; i < lenA; i++)
                    {
                      Object subA = Array.get(arrayA, i);
                      Object subB = Array.get(arrayB, i);
                      if (! Util.arrayEqual(subA, subB))
                        return false;
                    }
                  return true;
                }

              // fall back: cast arrays to object arrays
              Object[] objecta = (Object [])arrayA;
              Object[] objectb = (Object [])arrayB;

              int length = objecta.length;
              if (length != objectb.length)
                return false;
              else
                {
                  for (int i = 0; i < length; i++)
                    {
                      if (Util.unequal(objecta[i], objectb[i]))
                        return false;
                    }
                }
            }
        }
      return true;
    }


  /**
   * Test equality for two integer arrays
   **/
  private static boolean arrayEqual(int[] intArrayA, int[] intArrayB)
    {
      if (intArrayA.length != intArrayB.length)
        return false;
      else
        {
          for (int i = intArrayA.length-1; i >= 0; i--)
            {
              if (intArrayA[i] != intArrayB[i])
                return false;
            }
        }
      return true;
    }

  /**
   * Test equality for two boolean arrays
   **/
  private static boolean arrayEqual(boolean[] booleanArrayA, boolean[] booleanArrayB)
    {
      if (booleanArrayA.length != booleanArrayB.length)
        return false;
      else
        {
          for (int i = booleanArrayA.length-1; i >= 0; i--)
            {
              if (booleanArrayA[i] != booleanArrayB[i])
                return false;
            }
        }
      return true;
    }

  /**
   * Test equality for two byte arrays
   **/
  private static boolean arrayEqual(byte[] byteArrayA, byte[] byteArrayB)
    {
      if (byteArrayA.length != byteArrayB.length)
        return false;
      else
        {
          for (int i = byteArrayA.length-1; i >= 0; i--)
            {
              if (byteArrayA[i] != byteArrayB[i])
                return false;
            }
        }
      return true;
    }

  /**
   * Test equality for two character arrays
   **/
  private static boolean arrayEqual(char[] charArrayA, char[] charArrayB)
    {
      if (charArrayA.length != charArrayB.length)
        return false;
      else
        {
          for (int i = charArrayA.length-1; i >= 0; i--)
            {
              if (charArrayA[i] != charArrayB[i])
                return false;
            }
        }
      return true;
    }


  /**
   * Test equality for two short arrays
   **/
  private static boolean arrayEqual(short[] shortArrayA, short[] shortArrayB)
    {
      if (shortArrayA.length != shortArrayB.length)
        return false;
      else
        {
          for (int i = shortArrayA.length-1; i >= 0; i--)
            {
              if (shortArrayA[i] != shortArrayB[i])
                return false;
            }
        }
      return true;
    }

  /**
   * Test equality for two long arrays
   **/
  private static boolean arrayEqual(long[] longArrayA, long[] longArrayB)
    {
      if (longArrayA.length != longArrayB.length)
        return false;
      else
        {
          for (int i = longArrayA.length-1; i >= 0; i--)
            {
              if (longArrayA[i] != longArrayB[i])
                return false;
            }
        }
      return true;
    }

  /**
   * Test equality for two float arrays
   **/
  private static boolean arrayEqual(float[] floatArrayA, float[] floatArrayB)
    {
      if (floatArrayA.length != floatArrayB.length)
        return false;
      else
        {
          for (int i = floatArrayA.length-1; i >= 0; i--)
            {
              if (floatArrayA[i] != floatArrayB[i])
                return false;
            }
        }
      return true;
    }

  /**
   * Test equality for two double arrays
   **/
  private static boolean arrayEqual(double[] doubleArrayA, double[] doubleArrayB)
    {
      if (doubleArrayA.length != doubleArrayB.length)
        return false;
      else
        {
          for (int i = doubleArrayA.length-1; i >= 0; i--)
            {
              if (doubleArrayA[i] != doubleArrayB[i])
                return false;
            }
        }
      return true;
    }

  /**
   * If an InvocationTargetException is passed in, it is "unwrapped" until a
   * non-InvocationTargetException is found.
   * @param exception The exception to check for InvocationTargetException.
   * @return A throwable.  Either the target exception of the InvocationTargetException passed in
   * or the exception passed in.
   */
  public static Throwable getTargetException( Throwable exception )
    {
      while ( exception instanceof InvocationTargetException )
        {
          exception = ((InvocationTargetException)exception).getTargetException();
        } // while

      return exception;
    }

  // see canAccessAwtEventQueue(); if void, have not called canAccessAwtEventQueue() yet
  private static Boolean qAccess;

  /**
   * Test if the application/applet can access the AWT event queue.
   * This method remembers this information and is thus more efficient
   * than trying to access the event queue and catching security exceptions
   * each time you wish to access the event queue.
   * @return true if the application/applet can access the AWT event queue.
   */
  public static boolean canAccessAwtEventQueue()
    {
      if (qAccess != null)
        return qAccess.booleanValue();

      // Assume we can acces the event queue
      qAccess = Boolean.TRUE;

      // Determine if we have access to the AWT event queue.
      SecurityManager sm = System.getSecurityManager();
      if (sm != null)
        try
          { sm.checkAwtEventQueueAccess(); }
	    catch (SecurityException t)
		  { qAccess = Boolean.FALSE; }
      return qAccess.booleanValue();
    }

  /**
   * Constant Java Virtual Machine vendor identifier for Sun Microsystem's JVM.
   * Use as a parameter for the JVMVendorIs method.
   */
  public static String JVM_VENDOR_SUN_MICROSYSTEMS = "SUN MICROSYSTEMS";

  /**
   * Constant Java Virtual Machine vendor identifier for Netscape Communications JVM,
   * within Netscape Navigator.
   * Use as a parameter for the JVMVendorIs method.
   */
  public static String JVM_VENDOR_NETSCAPE_COMMUNICATIONS = "NETSCAPE";

  /**
   * Constant Java Virtual Machine vendor identifier for Microsoft's JVM.
   * Use as a parameter for the JVMVendorIs method.
   */
  public static String JVM_VENDOR_MICROSOFT = "MICROSOFT";

  /**
   * Constant Java Virtual Machine vendor identifier for Java VM's
   * other than Microsoft or JavaSoft.
   * Use as a parameter for the JVMVendorIs method.
   */
  public static String JVM_VENDOR_OTHER = "<<unknown>>";

  // store the computed JVM vendor id in _vendor.
  // The default value null is a marker meaning JVMVendorIs has not been called
  private static String _vendor;

  /**
   * Check the Java Virtual Machine vendor.
   * @param vendor one of {@link #JVM_VENDOR_SUN_MICROSYSTEMS}
   * or {@link #JVM_VENDOR_MICROSOFT} or
   * or {@link #JVM_VENDOR_NETSCAPE_COMMUNICATIONS} or {@link #JVM_VENDOR_OTHER}
   * @return true if the Java VM vendor matches the requested vendor.
   */
  public static boolean JVMVendorIs(String vendor)
    {
      if (_vendor == null)
        {
          String vendorProp = System.getProperty("java.vendor").toUpperCase()/*I18nOK:MTH*/;
          if (vendorProp.indexOf(JVM_VENDOR_SUN_MICROSYSTEMS) >= 0)
            _vendor = JVM_VENDOR_SUN_MICROSYSTEMS;
          else if (vendorProp.indexOf(JVM_VENDOR_NETSCAPE_COMMUNICATIONS) >= 0)
            _vendor = JVM_VENDOR_NETSCAPE_COMMUNICATIONS;
          else if (vendorProp.indexOf(JVM_VENDOR_MICROSOFT) >= 0)
            _vendor = JVM_VENDOR_MICROSOFT;
          else
            _vendor = JVM_VENDOR_OTHER;
        }
      vendor = vendor.toUpperCase()/*I18nOK:MTH*/;
      return _vendor == vendor || _vendor.indexOf(vendor) >= 0;
    }

  /**
   * Test if the Java Virtual Machine vendor is Microsoft
   * @return true if the JVM vendor is Microsoft
   */
  public static boolean JVMVendorIsMicrosoft()
    {
      return JVMVendorIs( JVM_VENDOR_MICROSOFT );
    }

  /**
   * Test if the Java Virtual Machine vendor is Netscape
   * @return true if the JVM vendor is Netscape
   */
  public static boolean JVMVendorIsNetscape()
    {
      return JVMVendorIs( JVM_VENDOR_NETSCAPE_COMMUNICATIONS );
    }

  /**
   * Test if the Java Virtual Machine vendor is Sun Microsystems.
   * @return true if the JVM vendor is Sun Microsystems.
   */
  public static boolean JVMVendorIsSunMicrosystems()
    {
      return JVMVendorIs( JVM_VENDOR_SUN_MICROSYSTEMS );
    }



  /**
   * Constant operating system identifier for Sun Microsystem's Solaris.
   * Use as a parameter for the osTypeIs method.
   */
  public static String OS_SUN_MICROSYSTEMS_SOLARIS = "SOLARIS";

  /**
   * Constant operating system identifier for Microsoft's Windows
   * Use as a parameter for the osTypeIs method.
   */
  public static String OS_MICROSOFT_WINDOWS = "WINDOWS";

  /**
   * Constant operating system identifier for os
   * other than Microsoft Windows or Sun Solaris.
   * Use as a parameter for the osTypeIs method.
   */
  public static String OS_OTHER = "<<unknown>>";

  // store the computed os id in _os.
  // The default value null is a marker meaning osTypeIs has not been called
  private static String _opSys;

  /**
   * Check the operating system.
   * @param opSys one of {@link #OS_SUN_MICROSYSTEMS_SOLARIS}
   * or {@link #OS_MICROSOFT_WINDOWS}
   * or {@link #OS_OTHER}
   * @return true if the current OS matches the requested operating system.
   *         Note: this method return false if the operating system
   *         cannot be determined.
   */
  public static boolean osTypeIs(String opSys)
    {
      if (_opSys == null)
        {
          try {
            String osProp = System.getProperty("os.name").toUpperCase()/*I18nOK:MTH*/;
            if (osProp.indexOf(OS_SUN_MICROSYSTEMS_SOLARIS) >= 0)
              _opSys = OS_SUN_MICROSYSTEMS_SOLARIS;
            else if (osProp.indexOf(OS_MICROSOFT_WINDOWS) >= 0)
              _opSys = OS_MICROSOFT_WINDOWS;
            else
              _opSys = OS_OTHER;
          }
          catch (SecurityException se)
            { _opSys = ""; }
        }
      opSys = opSys.toUpperCase()/*I18nOK:MTH*/;
      return _opSys == opSys || _opSys.indexOf(opSys) >= 0;
    }

  /**
   * Test if the operating system is Microsoft Windows.
   * @return true if the JVM vendor is Microsoft Windows.
   */
  public static boolean osTypeIsWindows()
    {
      return osTypeIs( OS_MICROSOFT_WINDOWS );
    }

  /**
   * Test if the operating system is Sun Microsystems Solaris.
   * @return true if the JVM vendor is Sun Microsystems Solaris.
   */
  public static boolean osTypeIsSolaris()
    {
      return osTypeIs( OS_SUN_MICROSYSTEMS_SOLARIS );
    }

  private static String localeListDelimiter;
  private static Locale defaultLocale;
  /**
   * Return the string which is used as a delimiter or separator in lists.
   * This is a Localizable string in the properties file, com/sas/util/CommonResources.properties.
   * The returned string should be a single character if possible.
   * @return the string which is used as a delimiter or separator in lists.
   *		 return "," if the resource is missing.
   */
  public static String getListDelimiter()
	{
	  Locale currentLocale = Locale.getDefault();
	  if (localeListDelimiter == null || currentLocale != defaultLocale)
		try
		  {
			// System.out.println("get Delimiter.txt for locale " + currentLocale);
			defaultLocale = currentLocale;
			_commonResourceBundle = null;
			localeListDelimiter = getCommonResources().getString("Delimiter.txt");
		  }
 		catch (MissingResourceException e) { }
	  if (localeListDelimiter == null)
		localeListDelimiter = ",";
	  return localeListDelimiter;
	}

  /**
   * Return the character which is used as a delimiter or separator in lists.
   * This is the first character from getListDelimiter()
   * @return the character which is used as a delimiter or separator in lists.
   */
  public static char getListDelimiterChar()
	{
	  return getListDelimiter().charAt(0);
	}

  /**
   * Convert a Throwable's stack trace to a String.
   * @param a Throwable, including Exceptions and Errors.
   * If null, a new Throwable is created, thus returning the
   * current stack trace (including this method and the
   * Throwable's constructor).
   * @return the traceback, as a single String.
   */
  public static String getStackTraceAsString(Throwable t, boolean includeMessage)
  {
    ByteArrayOutputStream baoStream = new ByteArrayOutputStream(1000);
    PrintStream stream = new PrintStream(baoStream);
    if (t == null)
      t = new Throwable();
    if (includeMessage)
      stream.println(t.getMessage());
    t.printStackTrace(stream);
    stream.close();
    return new String(baoStream.toByteArray());
  }

  /**
   * Returns the name of the class passed in, if the class is an array,
   * it reformats the name to be easily interpretted.
   * @return the name of the class passed in, formatted in a user-friendly way
   **/
	public static String getClassDeclaration( Class classObj )
	{
		if ( classObj.isArray() == false )
			return classObj.getName();

		String arrayDimension = "";
		String className = classObj.getName();
		while ( className.startsWith("[") )
		{
			arrayDimension += "[]";

			className = className.substring( 1 );
		}

		if ( className.length() == 1 )
		{
			if ( className.equals( "I" ) )
			{
				className = "int";
			}
			else if( className.equals( "B" ) )
			{
				className = "byte";
			}
			else if( className.equals( "C" ) )
			{
				className = "char";
			}
			else if( className.equals( "D" ) )
			{
				className = "double";
			}
			else if( className.equals( "F" ) )
			{
				className = "float";
			}
			else if( className.equals( "J" ) )
			{
				className = "long";
			}
			else if( className.equals( "S" ) )
			{
				className = "short";
			}
			else if( className.equals( "Z" ) )
			{
				className = "boolean";
			}
		}
		else
		{
			className = className.substring( 1, className.length()-1 );
		}
		return className + arrayDimension/*I18nOK:COS*/;
	}

	/**
	 * Returns the next unique sequential number. The number can be
	 * used to uniquely identify an object with in the java vm.
	 * @return unique sequential number.
	 **/

	public static synchronized long getNumber()
	{
		return commandCount++;
	}

	/**
	 * Returns a unique name with in the java vm.
	 * The string passed in is concatinated with the getNumber()
	 * method, returning a unique name.
	 * @return unique string name.
	 **/

	public static synchronized String getName(String name)
	{
		return name + getNumber();
	}

	/**
	 * Removes a character which is an invalid as a Java/JavaScript identifier and returns a valid identifier.
	 * eg. '3test' will return 'test'; 'test name' will return 'testname'
	 * If name is null, null will be returned
	 * @param name A Java or JavaScript identifier name to be checked
	 * @return valid Java/JavaScript identifier
	 */
	/**
	 * Removes a character which is an invalid as a Java/JavaScript identifier and returns a valid identifier.
	 * If the string starts with a number, "_" will be appended in fornt of it: '34234' will return '_34234'
	 * If the string contains invalid char such such as %, $, #, etc.., the character will be replaced with "_" plus the hex value of the char: 'test1&test2' returns 'test1_26test2'
	 * A string containing "_" will be replaced with "__"
	 * If name is null, null will be returned
	 * @param name A Java or JavaScript identifier name to be checked
	 * @return valid Java/JavaScript identifier
	 */
	public static String createJavaIdentifier(String name)
	{
		if (name == null)
			return name;

		char chars[]=
		{
			'_'
		}
		;
		String strings[]=
		{
			"__"
		}
		;
		String newName=Strings.replace(name, chars, strings);

		StringBuffer tempBffr=new StringBuffer(newName);
		char start=tempBffr.charAt(0);
		if (Character.isDigit(start))
		{
			tempBffr.insert(0, "_");
		}
		else if (!Character.isJavaIdentifierStart(start))
		{
			tempBffr.replace(0, 1, "_" + Integer.toHexString((int)start));
		}


		for (int j=tempBffr.length() - 1 ; j >= 1 ; j--)
		{
			char c=tempBffr.charAt(j);
			if (!Character.isJavaIdentifierPart(c))
			{
				tempBffr.replace(j, j + 1, "_" + Integer.toHexString((int)c));
				//tempBffr.deleteCharAt(j);
			}
		}
		return tempBffr.toString()/*I18nOK:MTH*/;
	}

}


