/* Copyright (c) 2002 by SAS Institute Inc., Cary, NC 27513 */

package com.sas.util.login;

import java.util.Vector;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;

/**
 * The Login bean is a model encapsulating the properties and behaviors of a user's
 * attempts to login to a system. The specifics of the system are intentionally
 * omitted to allow for a general purpose bean. Other classes exist elsewhere
 * to provide useful implementations in Swing and JSP scenarios.
 * <P>
 * The following example shows a possible configuration of the bean:
 * <PRE>
 * LoginBean bean = new LoginBean();
 * bean.setTitle("Enter your username and password:");
 * 
 * // Provide a default value for the username.
 * bean.setUsername("James");
 * 
 * // Allow 2 retry attempts before failing (default is 3).
 * bean.setAttempts(2);
 * </PRE>
 * <P>
 * Since the bean is a model, it needs to be associated with some viewer to be useful.
 * See the LoginBeanSwingUtil class in this same package for a more complete example that
 * uses the LoginBean in a Swing dialog.
 */
public class LoginBean
{
	public static String RB_KEY = "LoginBean.";

	/** A descriptive label at the top of the prompt elements. */
	protected String title;

	/** Holds the user-supplied username value. */
	protected String username;
	/** Holds the user-supplied password value. */
	protected String password;

	/** The value shown as the label for the username. */
	protected String usernameLabel;
	/** The value shown as the label for the password. */
	protected String passwordLabel;
	/** The value shown as the label for the OK button. */
	protected String okButtonLabel;
	/** The value shown as the label for the Cancel button. */
	protected String cancelButtonLabel;
	/** The value shown as the label for the Close button. */
	protected String closeButtonLabel;

	/** The current number of login attempts that a user has made. */
	protected int attemptsCurrent;
	/** The total number of login attempts allowed. */
	protected int attempts;
	/** The message to be displayed instructing the user to try again after a failed login attempt. */
	protected String retryMessage;

	/** The action handler. */
	protected LoginBeanActionInterface actionHandler;
	/** The validation handler. */
	protected LoginBeanValidateInterface validateHandler;

	/** The enumeration value representing the OK button. */
	public static final int BUTTON_OK = 0;
	/** The enumeration value representing the Cancel button. */
	public static final int BUTTON_CANCEL = 1;
	/** The enumeration value representing the Close button. */
	public static final int BUTTON_CLOSE = 2;

	/** The type of failure encountered. */
	protected int failureType;
	/** The string message to be displayed for an unknown failure. */
	protected String failureMessageUnknown;
	/** The string message to be displayed for an invalid host failure. */
	protected String failureMessageInvalidHost;
	/** The string message to be displayed for an invalid login failure. */
	protected String failureMessageInvalidLogin;
	/** The string message to be displayed for a refused connection failure. */
	protected String failureMessageConnectionRefused;
	/** The string message to be displayed for an unavailable service failure. */
	protected String failureMessageNotAvailable;
	/** The enumeration value for an undefined failure type. */
	public static final int FAILURE_TYPE_NONE = 0;
	/** The enumeration value for an unknown failure type. */
	public static final int FAILURE_TYPE_UNKNOWN = 1;
	/** The enumeration value for an invalid login failure type. */
	public static final int FAILURE_TYPE_INVALID_LOGIN = 2;
	/** The enumeration value for an invalid login failure type. */
	public static final int FAILURE_TYPE_INVALID_HOST = 3;
	/** The enumeration value for an invalid host failure type. */
	public static final int FAILURE_TYPE_CONNECTION_REFUSED = 4;
	/** The enumeration value for an refused connection failure type. */
	public static final int FAILURE_TYPE_NOT_AVAILABLE = 5;


	/**
	 * Default constructor.
	 */
	public LoginBean()
	{
		super();

		username = null;
		password = null;
		usernameLabel = RB.getStringResource(RB_KEY, "usernameLabel.txt");
		passwordLabel = RB.getStringResource(RB_KEY, "passwordLabel.txt");

		attempts = 3;
		attemptsCurrent = 0;
		retryMessage = RB.getStringResource(RB_KEY, "tryAgainMessage.txt");
		failureMessageUnknown = RB.getStringResource(RB_KEY, "failureUnknownMessage.txt");
		failureMessageInvalidHost = RB.getStringResource(RB_KEY, "failureInvalidHostMessage.txt");
		failureMessageInvalidLogin = RB.getStringResource(RB_KEY, "failureInvalidLoginMessage.txt");
		failureMessageConnectionRefused = RB.getStringResource(RB_KEY, "failureConnectRefusedMessage.txt");
		failureMessageNotAvailable = RB.getStringResource(RB_KEY, "failureNotAvailableMessage.txt");

		failureType = FAILURE_TYPE_NONE;

		okButtonLabel = RB.getStringResource(RB_KEY, "okButtonLabel.txt");
		cancelButtonLabel = RB.getStringResource(RB_KEY, "cancelButtonLabel.txt");
		closeButtonLabel = RB.getStringResource(RB_KEY, "closeButtonLabel.txt");
	}

	/**
	 * Sets the value for the title of the bean.
	 * @param value The title to be displayed describing the prompt elements.
	 * @see #getTitle()
	 */
	public void setTitle(String value)
	{
	   title = value;
	}

	/**
	 * Gets the title specified for the bean.
	 * @return The title describing the prompt elements.
	 * @see #setTitle(String)
	 */
	public String getTitle()
	{
	   return title;
	}

	/**
	 * Sets the default value for the username.
	 * @param value The username value.
	 * @see #getUsername()
	 */
	public void setUsername(String value)
	{
		username = value;
	}

	/**
	 * Returns the value for the username.
	 * @return The username value.
	 * @see #setUsername(String)
	 */
	public String getUsername()
	{
		return username;
	}

	/**
	 * Sets the default value for the password.
	 * @param value The password value.
	 * @see #getPassword()
	 */
	public void setPassword(String value)
	{
		password = value;
	}

	/**
	 * Returns the value for the password.
	 * @return The password value.
	 * @see #setPassword(String)
	 */
	public String getPassword()
	{
		return password;
	}

	/**
	 * Sets the label that describes the username
	 * @param label The username label value.
	 * @see #getUsernameLabel()
	 */
	public void setUsernameLabel(String label)
	{
		usernameLabel = label;
	}

	/**
	 * Returns the label that describes the username.
	 * @return The username label value.
	 * @see #setUsernameLabel(String)
	 */
	public String getUsernameLabel()
	{
		return usernameLabel;
	}

	/**
	 * Sets the label that describes the password.
	 * @param label The password label value.
	 * @see #getPasswordLabel()
	 */
	public void setPasswordLabel(String label)
	{
		passwordLabel = label;
	}

	/**
	 * Returns the label that describes the password.
	 * @return The password label value.
	 * @see #setPasswordLabel(String)
	 */
	public String getPasswordLabel()
	{
		return passwordLabel;
	}

	/**
	 * Sets the label for the OK button.
	 * @param label The label value.
	 * @see #getOKButtonLabel()
	 */
	public void setOKButtonLabel(String label)
	{
		okButtonLabel = label;
	}

	/**
	 * Returns the label for the OK button.
	 * @return The label value.
	 * @see #setOKButtonLabel(String)
	 */
	public String getOKButtonLabel()
	{
		return okButtonLabel;
	}

	/**
	 * Sets the label for the Cancel button.
	 * @param label The label value.
	 * @see #getCancelButtonLabel()
	 */
	public void setCancelButtonLabel(String label)
	{
		cancelButtonLabel = label;
	}

	/**
	 * Returns the label for the Cancel button.
	 * @return The label value.
	 * @see #setCancelButtonLabel(String)
	 */
	public String getCancelButtonLabel()
	{
		return cancelButtonLabel;
	}

	/**
	 * Sets the label for the Close button.
	 * @param label The label value.
	 * @see #getCloseButtonLabel()
	 */
	public void setCloseButtonLabel(String label)
	{
		closeButtonLabel = label;
	}

	/**
	 * Returns the label for the Close button.
	 * @return The label value.
	 * @see #setCloseButtonLabel(String)
	 */
	public String getCloseButtonLabel()
	{
		return closeButtonLabel;
	}

	/**
	 * Sets the message that gets displayed when a login retry
	 * attempt needs to be performed.
	 * @param value The message to display.
	 * @see #getRetryMessage()
	 */
	public void setRetryMessage(String value)
	{
	   retryMessage = value;
	}

	/**
	 * Returns the message that is displayed when a login retry
	 * attempt is performed.
	 * @return The message that is displayed.
	 * @see #setRetryMessage(String)
	 */
	public String getRetryMessage()
	{
		return retryMessage;
	}

	/**
	 * Sets the handler object for notification of Action events.
	 * Action events consist of notifications that can occur during the lifetime
	 * of the login bean, such as OK, Cancel, Success, Failure and Retry attempts.
	 * 
	 * You can specify a handler so that you'll receive notification whenever one
	 * of these actions occurs.
	 * 
	 * @param handler An object that implements LoginBeanActionInterface.
	 * @see #getActionHandler()
	 */
	public void setActionHandler(LoginBeanActionInterface handler)
	{
		actionHandler = handler;
	}

	/**
	 * Returns the handler object that receives notification of Action events.
	 * @return The object that implements LoginBeanActionInterface.
	 * @see #setActionHandler(LoginBeanActionInterface)
	 */
	public LoginBeanActionInterface getActionHandler()
	{
		return actionHandler;
	}

	/**
	 * Sets the handler object which will perform login validation.
	 * 
	 * You should specify a handler if you would like the bean to validate
	 * the credentials once they have been entered in by the user.
	 * 
	 * @param handler An object that implements LoginBeanValidateInterface.
	 * @see #getValidateHandler()
	 */
	public void setValidateHandler(LoginBeanValidateInterface handler)
	{
		validateHandler = handler;
	}

	/**
	 * Returns the handler object which performs login validation.
	 * @return The object that implements LoginBeanValidateInterface.
	 * @see #setValidateHandler(LoginBeanValidateInterface)
	 */
	public LoginBeanValidateInterface getValidateHandler()
	{
		return validateHandler;
	}

	/**
	 * Sets the number of attempts the user is allowed during the login
	 * process. This property is only relevent when you are using a Validation handler.
	 * @param value The number of login attempts.
	 * @see #getAttempts()
	 */
	public void setAttempts(int value)
	{
		attempts = value;
	}

	/**
	 * Returns the number of attempts the user is allowed during the login
	 * process.
	 * @return The number of login attempts.
	 * @see #setAttempts(int)
	 */
	public int getAttempts()
	{
		return attempts;
	}

	/**
	 * Sets the type of login failure currently associated with this login attempt.
	 * @param value The type of failure, which can be one of the following values:
	 *   FAILURE_TYPE_NONE
	 *   FAILURE_TYPE_UNKNOWN
	 *   FAILURE_TYPE_INVALID_LOGIN
	 *   FAILURE_TYPE_INVALID_HOST
	 *   FAILURE_TYPE_CONNECTION_REFUSED
	 *   FAILURE_TYPE_NOT_AVAILABLE
	 * @see #getFailureType()
	 */
	public void setFailureType(int value)
	{
		failureType = value;
	}

	/**
	 * Returns the type of login failure currently associated with this login attempt.
	 * @return The type of failure.
	 * @see #setFailureType(int)
	 */
	public int getFailureType()
	{
		return failureType;
	}

	/**
	 * Sets the failure message based that should be used when the
	 * specified type of failure occurs.
	 * @param type The type of failure.
	 * @param msg The associated message for the failure.
	 * @see #setFailureType(int)
	 */
	public void setFailureMessage(int type, String msg)
	{
		switch (type)
		{
			case FAILURE_TYPE_UNKNOWN:
				failureMessageUnknown = msg;
				break;
			case FAILURE_TYPE_INVALID_HOST:
				failureMessageInvalidHost = msg;
				break;
			case FAILURE_TYPE_INVALID_LOGIN:
				failureMessageInvalidLogin = msg;
				break;
			case FAILURE_TYPE_CONNECTION_REFUSED:
				failureMessageConnectionRefused = msg;
				break;
			case FAILURE_TYPE_NOT_AVAILABLE:
				failureMessageNotAvailable = msg;
				break;
		}
	}

	/**
	 * Returns the message that is associated with the specified
	 * type of failure. 
	 * @param type The type of failure.
	 * @return The associated message for the failure.
	 * @see #setFailureType(int)
	 */
	public String getFailureMessage(int type)
	{
		switch (type)
		{
			case FAILURE_TYPE_UNKNOWN:
				return failureMessageUnknown;
			case FAILURE_TYPE_INVALID_HOST:
				return failureMessageInvalidHost;
			case FAILURE_TYPE_INVALID_LOGIN:
				return failureMessageInvalidLogin;
			case FAILURE_TYPE_CONNECTION_REFUSED:
				return failureMessageConnectionRefused;
			case FAILURE_TYPE_NOT_AVAILABLE:
				return failureMessageNotAvailable;
		}

		return null;
	}

	/**
	 * Returns the message that is associated with the current
	 * type of failure for this login.
	 * @return The associated message for the current failure.
	 */
	public String getCurrentFailureMessage()
	{
		return getFailureMessage(getFailureType());
	}

	/**
	 * Performs validation of the login credentials on the specified Validation handler.
	 * If the validation succeeds, the bean calls the successAction method on 
	 * the specified Action handler.
	 * <BR>
	 * If the validation fails and there are additional login attempts allowed,
	 * the bean calls the retryAction method on the specified Action handler.
	 * <BR>
	 * If the validation fails and there are no more login attempts allowed,
	 * the bean calls the failureAction method on the specified Action handler.
	 * @return Returns true if the validation was successful, false otherwise.
	 * @see #setValidateHandler(LoginBeanValidateInterface)
	 */
	public boolean validate()
	{
		if (validateHandler == null)
			return false;

		boolean valid = validateHandler.validate(getUsername(), getPassword());
		attemptsCurrent++;

		if (actionHandler != null)
		{
			if (valid)
			{
				actionHandler.successAction();
				attemptsCurrent = 0;
			}
			else
			{
				// Have we run out of attempts?
				if (attemptsCurrent == attempts)
				{
					actionHandler.failureAction();
					attemptsCurrent = 0;
				}
				else
					actionHandler.retryAction();
			}
		}

		return valid;
	}
}
