// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513


package com.sas.util.transforms;

import com.sas.ComponentException;
import com.sas.text.Message;

import java.beans.PropertyChangeEvent;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;


/**
 * The Transforms class manages the list of registered transforms used by the property linking
 * mechanism of the com.sas.LinkPropertiesInterface and com.sas.PropertyLinkAdaptor.
 */
public class BaseTransforms
{
	public static final String RB_KEY = "BaseTransforms.";

	/**
	 * Returns the default instance of the BaseTransforms class.  This instance will register the set of default transforms.
	 * In general, this instance should be used when the set of default transforms are needed.  
	 */
	public static BaseTransforms defaultInstance = new BaseTransforms();

	/**
	 * This flag is used by all public methods of this class to determine if the default
	 * transforms have already be registered.  This allow lazy registration of the
	 * default transforms so that applications which do not use property linking will
	 * not have to pay the price for registering the transforms.
	 */
	protected boolean defaultTransformsRegistered;

	protected Class registryConfigRelativeClass;
	protected String registryConfigName;

	public BaseTransforms()
	{
		this( com.sas.util.transforms.BaseTransforms.class, "BaseTransformRegistry.config" );
	}

	public BaseTransforms( Class registryConfigRelativeClass, String registryConfigName )
	{
		this.registryConfigRelativeClass = registryConfigRelativeClass;
		this.registryConfigName = registryConfigName;
	}

	/**
	 * registerTransforms registers all the transforms listed in the ResourceBundle
	 * passed in.
	 */
	public void registerTransforms( Properties properties )
	{
		String number = properties.getProperty( "Transforms.count" );
		int transformCount;
		try
		{
			transformCount = Integer.parseInt( number );
		}
		catch (NumberFormatException nfe)
		{
			 throw new IllegalStateException( nfe.getMessage() );
		}

		try
		{
			for ( int i = 1; i <= transformCount; i++ )
			{
				String registryInfo = properties.getProperty( "transform" + i );
//				System.out.println( registryInfo );
				int sourceCommaIndex = registryInfo.indexOf(",");
				if ( sourceCommaIndex != -1 )
				{
					String sourceClassName = registryInfo.substring( 0, sourceCommaIndex ).trim();
					registryInfo = registryInfo.substring( sourceCommaIndex+1 );

					int targetCommaIndex = registryInfo.indexOf(",");
					if ( targetCommaIndex != -1 )
					{
						String targetClassName = registryInfo.substring( 0, targetCommaIndex ).trim();
						String transformClassName = registryInfo.substring( targetCommaIndex+1 ).trim();

//						System.out.println( "Source=" + sourceClassName );
//						System.out.println( "    Target=" + targetClassName );
//						System.out.println( "    Transform=" + transformClassName );

						try
						{
							Class sourceClass = forName( sourceClassName );
							Class targetClass = forName( targetClassName );
							Class transformClass = forName( transformClassName );

							TransformInterface transform = (TransformInterface) transformClass.newInstance();
							registerTransform( sourceClass, targetClass, transform );
						}
						catch (IllegalArgumentException iarge)
						{
							throw new ComponentException(Message.format(RB.getStringResource(RB_KEY, "errorLoadingClass.fmt.txt"), iarge.getMessage()));
						} // catch
						catch (ClassNotFoundException cnfe)
						{
							throw new ComponentException(Message.format(RB.getStringResource(RB_KEY, "errorLoadingClass.fmt.txt"), cnfe.getMessage()));
						} // catch
						catch (IllegalAccessException iae)
						{
							throw new ComponentException(Message.format(RB.getStringResource(RB_KEY, "errorInstantiatingClass.fmt.txt"), iae.getMessage()));
						} // catch
						catch (InstantiationException ie)
						{
							throw new ComponentException(Message.format(RB.getStringResource(RB_KEY, "errorInstantiatingClass.fmt.txt"), ie.getMessage()));
						} // catch
					} // if
					else
					{
						throw new IllegalStateException(Message.format(RB.getStringResource(RB_KEY, "invalidTransformFormat.fmt.txt"), new Integer(i)));
					}
				} // if
				else
				{
						throw new IllegalStateException(Message.format(RB.getStringResource(RB_KEY, "invalidTransformFormat.fmt.txt"), new Integer(i)));
				}
			} // for
		}
		catch (StringIndexOutOfBoundsException sioobe)
		{
			throw new IllegalStateException(Message.format(RB.getStringResource(RB_KEY, "invalidTransformFormat.fmt.txt"), sioobe.getMessage()));
		}
	}

	/**
	 * registerDefaultTransforms registers a set of default transforms so that there is a
	 * base set of transforms to use in linking properties.
	 */
	protected void registerDefaultTransforms()
	{
		if ( defaultTransformsRegistered )
			return;

		defaultTransformsRegistered = true;

		Properties properties = new Properties();
		try
		{
			properties.load( registryConfigRelativeClass.getResourceAsStream( registryConfigName ) );
		}
		catch ( IOException ioe )
		{
			throw new IllegalArgumentException( ioe.getMessage() );
		}
		registerTransforms( properties );
	}

	protected Object[][] primitiveClassNamesMap = new Object[][]
	{
		{ "byte", byte.class },
		{ "char", char.class },
		{ "int", int.class },
		{ "short", short.class },
		{ "long", long.class },
		{ "float", float.class },
		{ "double", double.class },
		{ "boolean", boolean.class },
		{ "java.lang.String[]", java.lang.String[].class },
	};

	/**
	 * forName is a helper method that returns a Class object based on the name passed in.
	 */
	protected Class forName( String name ) throws ClassNotFoundException
	{

		for ( int i = 0; i < primitiveClassNamesMap.length; i++ )
		{
			if ( name.equals( primitiveClassNamesMap[i][0] ) )
			{
				return (Class) primitiveClassNamesMap[i][1];
			}
		}

		return Class.forName( name );
	}

	/**
	 * inputClasses is a Hashtable which maintains a hash of Class objects to
	 * HashTables.  These inner Hashtables contain a hash of Class objects to
	 * objects which implement the TransformInterface.
	 * @see #registerTransform
	 * @see #unregisterTransform
	 * @see #lookupTransform
	 */
	protected Hashtable inputClasses = new Hashtable();

	/**
	 * Registers a transformation (converterClass) from inputClass to outputClass.
	 * @see #unregisterTransform
	 * @see #lookupTransform
	 */
	public void registerTransform( Class inputClass, Class outputClass, TransformInterface converterClass )
	{
		// Make sure the default transforms are registered

		registerDefaultTransforms();

		if ( inputClasses.containsKey( inputClass ) )
		{
			Hashtable outputClasses = (Hashtable) inputClasses.get( inputClass );

			// XXX: might ought to check for dups here...

			outputClasses.put( outputClass, converterClass );
		} // if
		else
		{
			Hashtable outputClasses = new Hashtable();
			outputClasses.put( outputClass, converterClass );
			inputClasses.put( inputClass, outputClasses );
		} // else
	}

	/**
	 * Unregisters a transformation (converterClass) from inputClass to outputClass.
	 * @see #registerTransform
	 * @see #lookupTransform
	 */
	public void unregisterTransform( Class inputClass, Class outputClass, TransformInterface converterClass )
	{
		// Make sure the default transforms are registered

		registerDefaultTransforms();

		if ( inputClasses.containsKey( inputClass ) )
		{
			Hashtable outputClasses = (Hashtable) inputClasses.get( inputClass );

			if ( outputClasses.containsKey( outputClass ) )
				outputClasses.remove( outputClass );
		} // if
	}


	/**
	 * isTrivialTransform returns true if no transformation is necessary to convert
	 * the inputClass to the outputClass.  It returns true if the input class and the
	 * output class are equal, if the outputClass is java.lang.Object, or if the outputClass
	 * is assignable from the inputClass.
	 */
	private boolean isTrivialTransform( Class inputClass, Class outputClass )
	{
		if ( inputClass == outputClass  ||
			 outputClass == java.lang.Object.class  ||
			 outputClass.isAssignableFrom( inputClass ) )
		{
//			System.out.println( "isTrivialTransform: " + inputClass + ", " + outputClass + " : true" );
			return true;
		}

//		System.out.println( "    isTrivialTransform: " + inputClass + ", " + outputClass + " : false" );
		return false;
	}

	/**
	 * Check for a transformation from inputClass to outputClass from the pool of registered transformations.
	 * @see #lookupTransform
	 * @see #registerTransform
	 * @see #unregisterTransform
	 */
    public boolean isTransformable( Class inputClass, Class outputClass )
    {
		// TO-DO:(SPL): Make this more efficient...

		if ( isTrivialTransform(inputClass, outputClass) )
		{
			return true;
		}

		return ( lookupTransform( inputClass, outputClass ) != null );
	}

	/**
	 * Finds a transformation from inputClass to outputClass from the pool of registered transformations.
	 * @see #registerTransform
	 * @see #unregisterTransform
	 */
	public TransformInterface lookupTransform( Class inputClass, Class outputClass )
	{
		// Handle common/easy cases (where no transform is necessary) quickly

		if ( isTrivialTransform(inputClass, outputClass) )
		{
			// return null;
			return IdentityTransform.identityTransform;
		}

		// Make sure the default transforms are registered

		registerDefaultTransforms();

		while( inputClass != null )
		{
			if ( inputClasses.containsKey( inputClass ) )
			{
				Hashtable outputClasses = (Hashtable) inputClasses.get( inputClass );

				if ( outputClasses != null  &&
				outputClasses.containsKey( outputClass ) )
				{
					Object outputClassNameOrObject = outputClasses.get( outputClass );
					if ( outputClassNameOrObject instanceof TransformInterface )
					{
						return (TransformInterface) outputClassNameOrObject;
					}
					else if ( outputClassNameOrObject instanceof String )
					{
						try
						{
							Class transformClass = Class.forName( (String) outputClassNameOrObject );
							TransformInterface transform = (TransformInterface) transformClass.newInstance();
							// Don't need to check and see if the newInstance returns
							// a valid TransformInterface since we'll get an exception
							// if it doesn't.
							return transform;
						}
						catch (Exception e)
						{
							// Ignore exception and fall through to return null below.
						}
					}
				} // if
			} // if

			inputClass = inputClass.getSuperclass();
		}
		return null;
	}


	/**
	 * Returns an array of unique registered transform objects.
	 * @return an array of unique registered transform objects.
	 */
	public TransformInterface[] getRegisteredTransforms()
	{
		// Make sure the default transforms are registered

		registerDefaultTransforms();

		Vector transformNames = new Vector();

		Enumeration inputEnum = inputClasses.keys();
		while (inputEnum.hasMoreElements())
		{
			Class inputClass = (Class) inputEnum.nextElement();
			Hashtable outputClasses = (Hashtable) inputClasses.get( inputClass );

			Enumeration outputEnum = outputClasses.keys();
			while (outputEnum.hasMoreElements())
			{
				Class outputClass = (Class) outputEnum.nextElement();
				TransformInterface transform = (TransformInterface) outputClasses.get( outputClass );
				if ( transformNames.contains( transform ) == false )
				{
					transformNames.addElement( transform );
				} // if
			} // while
		} // while

		TransformInterface[] transformArray = new TransformInterface[ transformNames.size() ];
		transformNames.copyInto( transformArray );
		return transformArray;
	}


    /**
     * Returns whether or not the specified source property can be transformed to the type of the
     * specified target property.
     * <P>
     * This is used by the IDE.
     * @see #registerTransform
     * @see #unregisterTransform
     * @see #lookupTransform
     */
    public boolean isTransformable( Object source, String sourceProperty, Object target, String targetProperty )
    {
//System.out.println("source = " + source + ", srcProp = " + sourceProperty );
//System.out.println("target = " + target + ", targetProp = " + targetProperty );
//Thread.currentThread().dumpStack();

//		// Make sure the default transforms are registered
//
//		registerDefaultTransforms();

		Class sourceType = null;
		Class targetType = null;
		java.beans.PropertyDescriptor targetPropertyDescriptor = null;

		try
		{
			// Lookup the PropertyDescriptor for the sourceProperty

			java.beans.BeanInfo sourceBeanInfo = com.sas.beans.Introspector.getBeanInfo( source.getClass() );
			java.beans.PropertyDescriptor sourcePropertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( sourceProperty, sourceBeanInfo );
			if ( sourcePropertyDescriptor == null )
			{
				throw new ComponentException(Message.format(RB.getStringResource(RB_KEY, "propertyDescSource.fmt.txt"), "Transforms::isTransformable()" /*I18nOK:EMS*/, sourceProperty));
			}
			sourceType = sourcePropertyDescriptor.getPropertyType();

			// Lookup the PropertyDescriptor for the targetProperty

			java.beans.BeanInfo targetBeanInfo = com.sas.beans.Introspector.getBeanInfo( target.getClass() );
			targetPropertyDescriptor = com.sas.beans.Introspector.getPropertyDescriptor( targetProperty, targetBeanInfo );
			if ( targetPropertyDescriptor == null )
			{
				String error = Message.format(RB.getStringResource(RB_KEY, "propertyDescTarget.fmt.txt"), "Transforms::isTransformable()" /*I18nOK:EMS*/, targetProperty);
				System.out.println(error);
				throw new ComponentException(error);
			}
			targetType = targetPropertyDescriptor.getPropertyType();

			if ( sourcePropertyDescriptor.getReadMethod() == null  ||
				 targetPropertyDescriptor.getWriteMethod() == null )
			{
				return false;
			} // if
		} // try
		catch ( Exception e )
		{
			// XXX
			System.out.println( Message.format(RB.getStringResource(RB_KEY, "propertyInfo.fmt.txt"), "Transforms::isTransformable()" /*I18nOK:EMS*/, e.toString()));
			e.printStackTrace();
		} // catch

		if ( !isTrivialTransform( sourceType, targetType ) )
		{
			TransformInterface converterInstance = lookupTransform( sourceType, targetType );
			if ( converterInstance == null )
			{
				converterInstance = lookupTransform( java.lang.Object.class, targetType );
				if ( converterInstance != null )
				{
					return true;
				} // if
			} // if
			else
			{
				return true;
			}
		} // if
		else
		{
			return true;
		} // else

		return false;
    } // isTransformable

	public Object transform( Object value, Class targetType ) throws TransformException
	{
		TransformInterface transform = lookupTransform( value.getClass(), targetType );
		if ( transform != null )
		{
			return transform.transform( value );
		}
		throw new TransformException( Message.format(RB.getStringResource(RB_KEY, "unableToTransform.fmt.txt"), targetType.getClass().getName()));
	}
}
