// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

/**
 * A composite transform, used to compose two transformation functions to act as one.
 * The first transformation function is applied, then the second applied to the result
 * of the first. To compose more transforms, use multiple CompositeTransform
 * objects:
 * <pre>
 * TransformInterface fgh  = new CompositeTransform(f, new CompositeTransform(g, h));
 * TransformInterface fghi = new CompositeTransform(new CompositeTransform(f, g),
 *                                                  new CompositeTransform(h, i));
 * </pre>
 * @author <A HREF="http://www.unx.sas.com/~sasdjb/Home.html">David J. Biesack</A>
 * @version 1.0
 */

/*
 * @support sasdjb
 * @history
 * <dl>
 * <dt>sasdjb 27Feb97<dd>Created
 * </dl>
 */

package com.sas.util.transforms;
import com.sas.util.transforms.TransformInterface;

public class CompositeTransform implements TransformInterface
{
	static final long serialVersionUID = 7593450837558504583L;

  private TransformInterface first;
  private TransformInterface second;

  /**
   * Create a default CompositeTransform.
   * Both the first and the second CompositeTransform will
   * be identity transforms.
   * @see com.sas.util.transforms.IdentityTransform
   */
  public CompositeTransform()
  {
    first = IdentityTransform.identityTransform;
    second = IdentityTransform.identityTransform;
  }

  /**
   * Get the first transform used by this composite transform.
   * @return the first transform object.
   */
  public TransformInterface getFirst()
  {
    return first;
  }

  /**
   * Replace the first transform used by this composite transform.
   * Take care to avoid creating cycles; this will result in infinite loops.
   * @param first the first transform. If null, it
   *        is set to IdentityTransform.identityTransform
   */
  public void setFirst(TransformInterface first)
  {
    this.first = first == null ? IdentityTransform.identityTransform : first;
  }

  /**
   * Get the second transform used by this composite transform.
   * @return the second transform object.
   */
  public TransformInterface getSecond()
  {
    return second;
  }

  /**
   * Replace the second transform used by this composite transform.
   * Take care to avoid creating cycles; this will result in infinite loops.
   * @param second the second transform. If null, it
   *        is set to IdentityTransform.identityTransform
   */
  public void setSecond(TransformInterface second)
  {
    this.second = second == null ? IdentityTransform.identityTransform : second;
  }

  /**
   * Create a CompositeTransform from two other transforms.
   * @param first the first transform. If null, it
   *        is set to IdentityTransform.identityTransform
   * @param second the second transform. If null, it
   *        is set to IdentityTransform.identityTransform
   * @see com.sas.util.transforms.IdentityTransform#identityTransform
   */
  public CompositeTransform(TransformInterface firstTransform, TransformInterface secondTransform )
  {
	this.first = firstTransform == null ? IdentityTransform.identityTransform : firstTransform;
	this.second = secondTransform == null ? IdentityTransform.identityTransform : secondTransform;
  }

  /**
   * The transform is performed by applying the first transform, then
   * applying the second transform to the result of the first.
   * Note that either of these transforms may produce null results.
   * @param obj any object
   * @return the result of applying the first and second subtransforms
   *         to obj.
   * @exception com.sas.util.transforms.TransformException if either
   *            transform function fails.
   */
  public Object transform(Object obj) throws TransformException
  {
    return second.transform(first.transform(obj));
  }

}
