// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util.transforms;

import com.sas.util.NumberUtil;

import java.util.Locale;

/**
 * The StringToDoubleTransform converts a String to a Double object.  If the value of the 
 * input to the transform is null or a blank/empty string, zero will be returned.
 * Otherwise, Double.valueOf() will be called to convert the string.
 * <P>
 * @see java.lang.Double#valueOf
 */
public class StringToDoubleTransform implements TransformInputOutputInterface
{
	static final long serialVersionUID = -7757250351074975763L;

	public static final StringToDoubleTransform defaultInstance = new StringToDoubleTransform();

	private Locale locale;
	
	/**
	 * Create a default instance of the transform
	 */
	public StringToDoubleTransform()
	{
		this(null);
	}

	/**
	 * Creates an instance of the transform and will use the given locale 
	 * when converting the String to a Double
	 * @param locale The locale used for converting the String to a Double, null to use the default locale
	 */
	public StringToDoubleTransform(Locale locale)
	{
		this.locale = locale;
	}

	public Object transform( Object input )
	{
		// null strings return zero
		if ( input == null )
			return new Double(0);

		// trim off the whitespace on both ends
		String trimmedInput = ((String)input).trim();

		// empty strings return zero
		if ( trimmedInput.length() == 0 )
			return new Double(0);
		
		if (locale == null)
			return new Double( NumberUtil.doubleValueOf( trimmedInput ) );
		else
			return new Double( NumberUtil.doubleValueOf( trimmedInput, locale ) );
	}

	public Class getInputClass()
	{
		return String.class;
	}

	public Class getOutputClass()
	{
		return Double.class;
	}
}
