// Copyright (c) 1998 by SAS Institute Inc., Cary, NC 27513

package com.sas.util.transforms;

import com.sas.util.NumberUtil;

import java.util.Locale;

/**
 * The StringToFloatTransform converts a String to a Float object.  If the value of the 
 * input to the transform is null or a blank/empty string, zero will be returned.
 * Otherwise, Float.valueOf() will be called to convert the string.
 * <P>
 * @see java.lang.Float#valueOf(String)
 */
public class StringToFloatTransform implements TransformInputOutputInterface
{
	static final long serialVersionUID = 4539523488635769533L;

	public static final StringToFloatTransform defaultInstance = new StringToFloatTransform();

	private Locale locale;
	
	/**
	 * Create a default instance of the transform
	 */
	public StringToFloatTransform()
	{
		this(null);
	}

	/**
	 * Creates an instance of the transform and will use the given locale 
	 * when converting the String to a float
	 * @param locale The locale used for converting the string to a float, null to use the default locale
	 */
	public StringToFloatTransform(Locale locale)
	{
		this.locale = locale;
	}

	public Object transform( Object input )
	{
		// null strings return zero
		if ( input == null )
			return new Float(0);

		String trimmedInput = ((String)input).trim();

		// empty strings return zero
		if ( trimmedInput.length() == 0 )
		{
			return new Float(0);
		} // if

		if (locale == null)
			return new Float( NumberUtil.floatValueOf( trimmedInput ) );
		else
			return new Float( NumberUtil.floatValueOf( trimmedInput, locale ) );
	}

	public Class getInputClass()
	{
		return String.class;
	}

	public Class getOutputClass()
	{
		return Float.class;
	}
}
