/* $Id: Resources.java,v 1.1.22.1 2005/08/04 17:40:39 sasspl Exp $ */
/* Copyright (c) 2002 by SAS Institute Inc., Cary, NC 27513 */
package com.sas.visuals;

import java.awt.Toolkit;
import java.awt.Image;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.PropertyResourceBundle;
import java.io.InputStream;
import java.io.IOException;
import java.util.Hashtable;
import javax.swing.ImageIcon;

import com.sas.text.Message;


/**
 * Convenience methods for accessing images and other resources.
 * This class provides a resource bundle for the <code>com.sas.visuals</code>
 * package, as well as a static
 * <code>{@link #getImage(java.lang.String)}</code>
 * method for loading
 * special images which are managed in a
 * more ad-hoc fashion, not as true <code>ResourceBundle</code> resources.
 * <p>
 * Images may be loaded by their base name, without the .gif extension,
 * such as <code>"SASSystem"</code>
 * or by a relative (to <code>com.sas.visuals.Resources</code>)
 * name with extension, such as <code>"res/SASSystem.gif"</code>
 * (See <code>{@link #PREFIX}</code> and <code>{@link #SUFFIX}</code>)
 * Case is significant.  Images are remembered locally, as well as
 * the names that have failed. Invoke the <code>{@link #reset()}</code>
 * method if the cache maintained by <code>com.sas.visuals.Resources</code>
 * becomes obsolete (i.e. the set of files on the server has changed.)
 * <p>
 * This resource bundle loads properties from <code>"Resources.properties"</code>
 * in this directory. Resource bundles for other locales will
 * not extend from this class but will be created automatically
 * by <code>java.util.ResourceBundle.getBundle("Resources");</code> for example,
 * if the default locale is <code>"en_UK"</code>, then <code>getBundle</code> will return
 * a <code>PropertyResourceBundle</code> based on <code>Resources_en_UK.properties</code> but it's
 * parent will be this bundle (or one based on <code>Resources_en.properties</code>,
 * if it exists.)
 * @author David Biesack
 * @version 1.2
 */

public class Resources
	extends PropertyResourceBundle
	/*
	 * support: sasdjb
	 * history:
	 * sasdjb 03Feb97
	 * Created
	 * sasdjb 11Aug97
	 * Rewrote to use JDK 1.1 Class.getResource() and Toolkit.getImage
	 * sasdjb 26Feb98
	 * Made into subclass of PropertyResourceBundle and added
	 * initialization from Resources.res.
	 * sasdjb 31Aug98
	 * Added caching and public reset() method;
	 * version change from 1.1 to 1.2
	 */
{
	public static final String RB_KEY = "Resources.";

	static final Class resClass = com.sas.GlobalInfoTable._loadClass("com.sas.visuals.Resources");
	/**
	 * Default constructor.
	 */
	public Resources() throws IOException
	{
		super(com.sas.security.BaseSecuritySupport.securitySupport.getResourceAsStream(resClass, "Resources.properties"));
	}

	/**
	 * Public constructor.
	 * @param stream a stream where the properties are found.
	 */
	public Resources(InputStream stream) throws IOException
	{
		super(stream);
	}

	/** Assumed prefix if an image is not found via a base name. */
	public static final String PREFIX = "res/";
	/** Assumed suffix if an image is not found via a base name. */
	public static final String SUFFIX = ".gif";

	/** images we've loaded successfully, to reduce server round trips. */
	private static Hashtable images;

	/** image icons we've loaded successfully, to reduce server round trips. */
	private static Hashtable imageIcons;

	private final static int BLOCK_SIZE = 1024;

	/** use this as the size param when requestin an image that is of 16 x 16 pixels, 
	 * standard size for small toolbar and menu images*/
	public final static int ICON_16 = 16;

	/** use this as the size param when requestin an image that is of 24 x 24 pixels*/
	public final static int ICON_24 = 24;

	/** use this as the size param when requestin an image that is of 32 x 32 pixels*/
	public final static int ICON_32 = 32;

	/** use this as the size param when requestin an image that is of 48 x 48 pixels*/
	public final static int ICON_48 = 48;

	/**
	 * Get the java.awt.Image from an InputStream.
	 * This method reads the InputStream to a byte array and then creates an Image
	 * object using the byte array.  Reading the InputStream is not done in another
	 * Thread so this will block until the image is read.
	 **/
	public static synchronized Image getImageFromStream( InputStream iStrm ) throws IOException
	{
		byte accumulator[] = {0};	// avoid NullPtrExc @arraycopy(oldAcc, accum)
		byte tmpBuffer[] = new byte[BLOCK_SIZE];
		int count = 0;
		int numRead = 0;

		while ((numRead = iStrm.read(tmpBuffer, 0, BLOCK_SIZE)) != -1)
		{
			byte oldAcc[] = accumulator;	// store the old location
			accumulator = new byte [numRead + count];

			// copy in the old values to the new accumulator
			System.arraycopy(oldAcc, 0, accumulator, 0, count);
			// append tmpBuffer to accumulator
			System.arraycopy(tmpBuffer, 0, accumulator, count, numRead);

			count += numRead;
		} // while()

		return (Toolkit.getDefaultToolkit().createImage(accumulator));
	}

	private static Image loadImage(Class loadClass, String imageName)
	{
		try
		{
			InputStream is = com.sas.security.BaseSecuritySupport.securitySupport.getResourceAsStream(loadClass, imageName);
			if (is != null)
			{
				Image image = getImageFromStream( is );
				images.put( loadClass.getName() + imageName, image ); /*I18nOK:LINE*/
				return image;
			}
		}
		catch (IOException e)
		{
			String msg = Message.format(RB.getStringResource(RB_KEY, "imageLoadError.ex.txt"), imageName, loadClass.getName());
			throw new IllegalArgumentException(msg);
		}
		return null;
	}

	private static Image loadImage(Class loadClass, String imageName, int size)
	{
		try
		{
			InputStream is = com.sas.security.BaseSecuritySupport.securitySupport.getResourceAsStream(loadClass, imageName);
			if (is != null)
			{
				Image image = getImageFromStream( is );
				images.put( loadClass.getName() + imageName + "#" + size, image ); /*I18nOK:LINE*/
				return image;
			}
		}
		catch (IOException e)
		{
			String msg = Message.format(RB.getStringResource(RB_KEY, "imageLoadError.ex.txt"), imageName, loadClass.getName());
			throw new IllegalArgumentException(msg);
		}
		return null;
	}



	/**
	 * Return an image.
	 * For historical purposes, a <code>"res/"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from this (<code>com.sas.visuals.Resources</code>) class.
	 * I.e. the image <code>"res/Folder.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/Folder.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * Image folder = Resources.getImage("Folder");
	 * or
	 * Image folder = Resources.getImage(Resources.PREFIX + "Folder" + Resources.SUFFIX);
	 * </pre>
	 * Note that this routine will only load images from the sas.icons.jar when used with the
	 * Eclipse classloader.
	 * @param imageName the name of an image file; may not be null
	 * @return an Image object.
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized Image getImage(String imageName)
	{
		return getImage( SASIcons.class, imageName );
	}

	/**
	 * Return an image.
	 * For historical purposes, a <code>"res/"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from the class passed in.
	 * I.e. the image <code>"res/Folder.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/Folder.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * Image folder = Resources.getImage(com.sas.visuals.SASIcons.class, "Folder");
	 * or
	 * Image folder = Resources.getImage(com.sas.visuals.SASIcons.class, Resources.PREFIX + "Folder" + Resources.SUFFIX);
	 * </pre>
	 * @param imageName the name of an image file; may not be null
	 * @param loadClass the class that the imageName is relative to; may not be null
	 * @return an Image object.
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized Image getImage(Class loadClass, String imageName)
	{
		if (images == null) 
			images = new Hashtable();

		Image image = null;
		String[] possibleImageNames = getValidImageNameKeys(imageName);
		for (int i = 0; i < possibleImageNames.length; i++)
		{
			image = (Image)images.get(loadClass.getName() + possibleImageNames[i]); /*I18nOK:LINE*/
			if (image != null)
			{
				return image;
			}
		}

		for (int i = 0; i < possibleImageNames.length; i++)
		{
			image = loadImage(loadClass, possibleImageNames[i]);
			if (image != null)
			{
				return image;
			}
		}

		String msg = Message.format(RB.getStringResource(RB_KEY, "noSuchImage.ex.txt"), imageName, loadClass.getName());
		throw new IllegalArgumentException(msg);
	}

	/**
	 * Return an image.
	 * A <code>"res/GIF_(size)"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from this (<code>com.sas.visuals.Resources</code>) class.
	 * I.e. the image <code>"res/GIF_16/StoredProcess.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/GIF_16/StoredProcess.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * Image folder = Resources.getImage("StoredProcess", 16);
	 * or
	 * Image folder = Resources.getImage(Resources.PREFIX + "GIF_16/StoredProcess" + Resources.SUFFIX, 16);
	 * </pre>
	 * Note that this routine will only load images from the sas.icons.jar when used with the
	 * Eclipse classloader.
	 * @param imageName the name of an image file; may not be null
	 * @return an Image object.
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized Image getImage(String imageName, int size)
	{
		return getImage( SASIcons.class, imageName, size);
	}

	/**
	 * Return an image.
	 * A <code>"res/GIF_(size)"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from the class passed in.
	 * I.e. the image <code>"res/GIF_16/StoredProcess.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/GIF_16/StoredProcess.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * Image folder = Resources.getImage(com.sas.visuals.SASIcons.class, "StoredProcess", 16);
	 * or
	 * Image folder = Resources.getImage(com.sas.visuals.SASIcons.class, Resources.PREFIX + "GIF_16/StoredProcess" + Resources.SUFFIX, 16);
	 * </pre>
	 * @param imageName the name of an image file; may not be null
	 * @param loadClass the class that the imageName is relative to; may not be null
	 * @return an Image object.
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized Image getImage(Class loadClass, String imageName, int size)
	{
		if (images == null) 
			images = new Hashtable();

		Image image = null;
		String[] possibleImageNames = getValidImageNameKeys(imageName, size);
		for (int i = 0; i < possibleImageNames.length; i++)
		{
			image = (Image)images.get(loadClass.getName() + possibleImageNames[i] + "#" + size); /*I18nOK:LINE*/
			if (image != null)
			{
				return image;
			}
		}

		for (int i = 0; i < possibleImageNames.length; i++)
		{
			image = loadImage(loadClass, possibleImageNames[i], size);
			if (image != null)
			{
				return image;
			}
		}

		String msg = Message.format(RB.getStringResource(RB_KEY, "noSuchImage.ex.txt"), imageName, loadClass.getName());
		throw new IllegalArgumentException(msg);
	}

	/**
	 * Reset the internal cache of images that have been loaded,
	 * as well as resetting the list of image names that have
	 * failed. Use this if the files on the server have changed.
	 */
	public static synchronized void reset()
	{
		images = new Hashtable();
		imageIcons = new Hashtable();
	}

	/**
	 * Return a javax.swing.ImageIcon for the path to the image passed in.
	 * For historical purposes, a <code>"res/"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from this (<code>com.sas.visuals.Resources</code>) class.
	 * I.e. the image <code>"res/Folder.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/Folder.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * ImageIcon folder = Resources.getImageIcon("Folder");
	 * or
	 * ImageIcon folder = Resources.getImageIcon(Resources.PREFIX + "Folder" + Resources.SUFFIX);
	 * </pre>
	 * Note that this routine will only load images from the sas.icons.jar when used with the
	 * Eclipse classloader.
	 * @param imageName the name of an image file; may not be null
	 * @return an ImageIcon object if the image exists, otherwise null
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized ImageIcon getImageIcon(String imageName)
	{
		return getImageIcon(SASIcons.class, imageName);
	}

	/**
	 * Return a javax.swing.ImageIcon for path of the image relative to the class passed in.
	 * For historical purposes, a <code>"res/"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from this (<code>com.sas.visuals.Resources</code>) class.
	 * I.e. the image <code>"res/Folder.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/Folder.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * ImageIcon folder = Resources.getImageIcon(com.sas.visuals.SASIcons.class, "Folder");
	 * or
	 * ImageIcon folder = Resources.getImageIcon(com.sas.visuals.SASIcons.class, Resources.PREFIX + "Folder" + Resources.SUFFIX);
	 * </pre>
	 * @param imageName the name of an image file; may not be null
	 * @return an ImageIcon object if the image exists, otherwise null
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized ImageIcon getImageIcon(Class loadClass, String imageName)
	{
		if (imageIcons == null) 
			imageIcons = new Hashtable();

		ImageIcon imageIcon = null;
		String[] possibleImageNames = getValidImageNameKeys(imageName);
		for (int i = 0; i < possibleImageNames.length; i++)
		{
			imageIcon = (ImageIcon)imageIcons.get(loadClass.getName() + possibleImageNames[i]); /*I18nOK:LINE*/
			if (imageIcon != null)
			{
				return imageIcon;
			}
		}

		Image image = getImage(loadClass, imageName);

		if (image != null)
		{
			imageIcon = new ImageIcon(image);
			imageIcons.put(loadClass.getName() + imageName, imageIcon); /*I18nOK:LINE*/
		}

		return imageIcon;
	}

	/**
	 * Return a javax.swing.ImageIcon for the path to the image passed in.
	 * For historical purposes, a <code>"res/"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from the class passed in.
	 * I.e. the image <code>"res/Folder.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/Folder.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * Image folder = Resources.getImageIcon(com.sas.visuals.SASIcons.class, "Folder");
	 * or
	 * Image folder = Resources.getImageIcon(com.sas.visuals.SASIcons.class, Resources.PREFIX + "Folder" + Resources.SUFFIX);
	 * </pre>
	 * Note that this routine will only load images from the sas.icons.jar when used with the
	 * Eclipse classloader.
	 * @param imageName the name of an image file; may not be null
	 * @return an ImageIcon object if the image exists, otherwise null
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized ImageIcon getImageIcon(String imageName, int size)
	{
		return getImageIcon(SASIcons.class, imageName, size);
	}

	/**
	 * Return a javax.swing.ImageIcon for path of the image relative to the class passed in.
	 * A <code>"res/GIF_(size)"</code> prefix and a <code>".gif"</code> suffix is
	 * assumed if the resource does not exist as <code>"<var>imageName</var>"</code>.
	 * The image should be specified as a relative pathname
	 * from the class passed in.
	 * I.e. the image <code>"res/GIF_16/StoredProcess.gif"</code> is
	 * located in the archive at <code>com/sas/visuals/res/GIF_16/StoredProcess.gif</code>
	 * You can fetch this image with either
	 * <pre>
	 * Image folder = Resources.getImageIcon(com.sas.visuals.SASIcons.class, "StoredProcess", 16);
	 * or
	 * Image folder = Resources.getImageIcon(com.sas.visuals.SASIcons.class, Resources.PREFIX + "GIF_16/StoredProcess" + Resources.SUFFIX, 16);
	 * </pre>
	 * @param imageName the name of an image file; may not be null
	 * @return an ImageIcon object if the image exists, otherwise null
	 * @exception java.lang.IllegalArgumentException if the
	 * image could not be loaded.
	 * @see #PREFIX
	 * @see #SUFFIX
	 */
	public static synchronized ImageIcon getImageIcon(Class loadClass, String imageName, int size)
	{
		if (imageIcons == null) 
			imageIcons = new Hashtable();

		ImageIcon imageIcon = null;
		String[] possibleImageNames = getValidImageNameKeys(imageName, size);
		for (int i = 0; i < possibleImageNames.length; i++)
		{
			imageIcon = (ImageIcon)imageIcons.get(loadClass.getName() + possibleImageNames[i] + "#" + size); /*I18nOK:LINE*/
			if (imageIcon != null)
			{
				return imageIcon;
			}
		}

		Image image = getImage(loadClass, imageName, size);

		if (image != null)
		{
			imageIcon = new ImageIcon(image);
			imageIcons.put(loadClass.getName() + imageName + "#" + size, imageIcon); /*I18nOK:LINE*/
		}

		return imageIcon;	
	}

	/**
	 * Get the java.awt.Image from an InputStream and return a javax.swing.ImageIcon with it as the image.
	 * This method calls getImageFromSteam to get the image, then it creates a 
	 * javax.swing.ImageIcon with the image.
	 * @param istrm the InputStream in which the image is loaded from
	 * @return an ImageIcon from the InputStream passed in, null if the iamge can not be loaded
	 * @see #getImageFromStream(InputStream)
	 **/
	public static synchronized ImageIcon getImageIconFromStream(InputStream iStrm) throws IOException
	{
		Image image = getImageFromStream(iStrm);
		ImageIcon imageIcon = null;

		if (image != null)
			imageIcon = new ImageIcon(image);

		return imageIcon;
	}

	private static String[] getValidImageNameKeys(String imageName)
	{
		return getValidImageNameKeys(imageName, -1);
	}

	private static String[] getValidImageNameKeys(String imageName, int size)
	{
		if (imageName != null)
		{
			List names = new ArrayList();
			names.add(imageName);

			// Name passed in failed. Try alt name if imageName
			// does not already begin/end with the PREFIX/SUFFIX
			boolean suffixAdded = false;
			if (!imageName.endsWith(SUFFIX))
			{
				if (imageName.indexOf(".") < 0)
				{
					String altName = imageName + SUFFIX; /*I18nOK:LINE*/
					names.add(altName);
					suffixAdded = true;
				}
			}

			if (size == -1)
			{
				if (!imageName.startsWith(PREFIX))
				{
					String altName = PREFIX + imageName; /*I18nOK:LINE*/
					names.add(altName);
					if (suffixAdded)
					{
						names.add(altName + SUFFIX); /*I18nOK:LINE*/
					}
				}
			}
			else
			{
				if (!imageName.startsWith(PREFIX + "GIF_" + size + "/")) /*I18nOK:LINE*/
				{
					String altName = PREFIX + "GIF_" + size + "/" + imageName /*I18nOK:LINE*/;
					names.add(altName);
					if (suffixAdded)
					{
						names.add(altName + SUFFIX); /*I18nOK:LINE*/
					}
				}

			}

			return (String[])names.toArray(new String[0]);
		}
		return new String[0];
	}
}